package AntTasks;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.Vector;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;

/**
 * Takes a set of web pages and saves them locally.
 * Main purpose is to take jsp pages and store them as html.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DownloadUrl extends Task {

    @SuppressWarnings("unused")
	private static final String ENCODING_ISO_8859_1 = "ISO-8859-1";
    private static final String ENCODING_UTF8 = "UTF-8";
	private String fromurl = "file://.";

	/**
	 * @param fromurl
	 *            the root directory of where to read the source files from.
	 *            Default value is the current directory.
	 */
	public void setFromurl(final String fromurl) {
		this.fromurl = fromurl;
	}

	private String todir = ".";

	/**
	 * @param toDir
	 *            the root directory of where to write the destination files to.
	 *            Default value is the current directory.
	 */
	public void setTodir(final String toDir) {
		this.todir = toDir;
	}

	private boolean clean = false;

	/**
	 * @param clean
	 *            if <code>true</code>, instead of creating the destination
	 *            files, all destination files are removed. Default is
	 *            <code>false</code>.
	 */
	public void setClean(final boolean clean) {
		this.clean = clean;
	}

	/** 
	 * Defines a web page to download. The URL should be given relative to fromurl. 
	 * 
	 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class Page {
		public Page() {
			// Nothing to do
		}

		public Page(final String url) {
			this.url = url;
		}

		String url;

		public void setUrl(final String url) {
			this.url = url;
		}

		public String getUrl() {
			return url;
		}
	}

	Vector<Page> pages = new Vector<Page>();

	public Page createPage() {
		Page page = new Page();
		pages.add(page);
		return page;
	}

	public Page createPage(final String url) {
		final Page page = new Page(url);
		pages.add(page);
		return page;
	}

	public void download(final String srcUrl, String destfile) throws IOException {
		try {
			final URL url = new URL(srcUrl);
			log("Opening connection to " + srcUrl + "...", Project.MSG_INFO);
//            URLConnection urlC = url.openConnection();
			// Copy resource to local file, use remote file
			// if no local file name specified
			final InputStream is = new BufferedInputStream(url.openStream());
			final InputStreamReader isw = new InputStreamReader(is, ENCODING_UTF8);
			// Print info about resource
//            System.out.print("Copying resource (type: " + urlC.getContentType());
//            Date date = new Date(urlC.getLastModified());
//            System.out.println(", modified on: " + date.toLocaleString() + ")...");
//            System.out.flush();
			BufferedOutputStream os = null;
			if (null == destfile) {
				// Get only file name
				final StringTokenizer st = new StringTokenizer(url.getFile(), "/");
				while (st.hasMoreTokens())
					destfile = st.nextToken();
			}
			os = new BufferedOutputStream(new FileOutputStream(destfile));
			OutputStreamWriter osw = new OutputStreamWriter(os, ENCODING_UTF8);
//			log("Encoding read is " + isw.getEncoding() + ", for writing is: " + osw.getEncoding());
			int oneChar;
			@SuppressWarnings("unused")
			int count = 0;
			while ((oneChar = isw.read()) != -1) {
				osw.write(oneChar);
				count++;
			}
			isw.close();
			osw.close();
//            System.out.println(count + " byte(s) copied");
		} catch (final MalformedURLException e) {
			log(e, Project.MSG_ERR);
			throw e;
		} catch (final IOException e) {
			log(e, Project.MSG_ERR);
			throw e;
		}
	}

	public void execute(final String jspName) throws IOException {
		final String srcUrl = fromurl + "/" + jspName;
		final String destPath = todir + "/" + FileAction.changeFilename(jspName, null, null, null, "html");
		log("Current directory is " + new File(".").getAbsolutePath(), Project.MSG_DEBUG);
		if (clean) {
			FileAction.remove(destPath);
			log("removing " + destPath, Project.MSG_INFO);
		} else {
			download(srcUrl, destPath);
		}
	}

	/**
	 * Executes the Ant task. In this case, iterates through the fileset
	 * and executes the task for each file, depending on the file compare with the destination file.
	 * 
	 * @see #executeIfDifferent(File, String[], int[])
	 */
	@Override
	public void execute() {
		log("todir = " + todir, Project.MSG_DEBUG);
		try {
			for (final Iterator<Page> itp = pages.iterator(); itp.hasNext();) {
				final Page page = itp.next();
				execute(page.getUrl());
			}
		} catch (final IOException e) {
			throw new BuildException(e);
		}
	}

	/**
	 * @return the usage for command line calling of this class for human readers
	 */
	public String usage() {
		final StringBuffer s = new StringBuffer();
		s.append("usage: java " + getClass().getName() + " [-d todir] [-s fromurl] [-clean|+clean] page(s)\n");
		s.append(" -d todir : The destination root path. Default is " + todir + "\n");
		s.append(" -s fromurl : The source root URL. URLs are read starting from there. Default is " + fromurl + "\n");
		s.append(" -clean : Turns cleaning mode off (default).\n");
		s.append(" +clean : Turns cleaning mode on. No files are created, instead all destination files are deleted.\n");
		return s.toString();
	}

	public static void main(final String[] args) {
		final DownloadUrl task = new DownloadUrl();
		if (args.length < 1) {
			System.err.println(task.usage());
			System.exit(1);
		}
		String mode = "";
		for (int i = 0; i < args.length; i++) {
			final String arg = args[i];
			if ("-d".equals(mode)) {
				task.setTodir(arg);
				mode = "";
			} else if ("-s".equals(mode)) {
				task.setFromurl(arg);
				mode = "";
			} else if ("-d".equals(arg)) {
				mode = arg;
			} else if ("-s".equals(arg)) {
				mode = arg;
			} else if ("-clean".equals(arg)) {
				task.setClean(false);
			} else if ("+clean".equals(arg)) {
				task.setClean(true);
			} else {
				task.createPage(arg);
			}
		}
		task.execute();
	}
}
