package AntTasks;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.Vector;

import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.types.FileSet;

/**
 * Abstract class to define an Ant task executing something on a fileset.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class FileAction extends Task {

	private String todir = ".";

	/**
	 * @param toDir
	 *            the root directory of where to write the destination files to.
	 *            Default value is the current directory.
	 */
	public void setTodir(final String toDir) {
		this.todir = toDir;
	}

	private String comparismMode = "filedate";

	/**
	 * @param comparismMode
	 *            the comparism mode to define if the source file is different
	 *            or newer than the destination file. Legal values are all
	 *            values in @link{FileCompare#COMPARISM_MODE_NAMES}, or a comma
	 *            separated list thereof. Default value is filedate.
	 * @see FileCompare#comparismModes(String)
	 */
	public void setComparismmode(final String comparismMode) {
		this.comparismMode = comparismMode;
	}

	private Vector<FileSet> filesets = new Vector<FileSet>();

	/**
	 * Adds a fileset of source files to look at.
	 * 
	 * @param fileset
	 *            the fileset to add
	 */
	public void addFileset(final FileSet fileset) {
		filesets.add(fileset);
	}

	private boolean preservelastmodified = true;

	/**
	 * @param preservelastmodified
	 *            if <code>true</code>, the created destination files get the
	 *            same last modified time as the original source files. Default
	 *            is <code>true</code>.
	 */
	public void setPreservelastmodified(final boolean preservelastmodified) {
		this.preservelastmodified = preservelastmodified;
	}

	private boolean clean = false;

	/**
	 * @param clean
	 *            if <code>true</code>, instead of creating the destination
	 *            files, all destination files are removed. Default is
	 *            <code>false</code>.
	 */
	public void setClean(final boolean clean) {
		this.clean = clean;
	}

	/**
	 * A helper method for converting file names.
	 * 
	 * @param filename
	 *            the original file name to start with. Should have an extension
	 * @param prefix
	 *            a prefix to add to the original filename
	 * @param suffix
	 *            a suffix to add to the original filename
	 * @param infixToRemove
	 *            an infix in the original filename that will be removed if present
	 * @param extension
	 *            a new extension for the new filename. If <code>null</code>,
	 *            the extension from filename is kept.
	 * @return the converted file name
	 */
	public static String changeFilename(
			final String filename, 
			final String prefix, 
			final String suffix, 
			final String infixToRemove, 
			final String extension) {
		final int idx = filename.lastIndexOf('.');
		String filenameWithoutExt = filename.substring(0, idx);
		if (null != infixToRemove) {
			filenameWithoutExt = filenameWithoutExt.replaceFirst(infixToRemove, "");
		}
		final String changedFilename = (null == prefix ? "" : prefix) 
			+ filenameWithoutExt 
			+ (null == suffix ? "" : suffix) + "." 
			+ (null == extension ? filename.substring(idx + 1) : extension);
		return changedFilename;
	}

	/**
	 * Calculates the destination filename based on the source filename
	 * (relative to the file base). The default implementation returns just the
	 * source filename. Override if you need something else.
	 * 
	 * @param srcFilename
	 *            the source filename, relative to the root of the fileset
	 * @return the same as srcFilename (as default implementation)
	 */
	public String getDestFilename(final String srcFilename) {
		return srcFilename;
	}

	/**
	 * Does a binary file copy.
	 * 
	 * @param srcFilename
	 *            the complete source file name and path
	 * @param destFilename
	 *            the complete destination file name and path
	 */
	public void copy(final String srcFilename, final String destFilename) {
		try {
			final BufferedInputStream is = new BufferedInputStream(new FileInputStream(srcFilename));
			final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(destFilename));
			int oneChar;
			while ((oneChar = is.read()) != -1) {
				os.write(oneChar);
			}
			is.close();
			os.close();
		} catch (final IOException e) {
			log(e.toString(), Project.MSG_ERR);
		}
	}

	/**
	 * Removes the file.
	 * 
	 * @param destFilename
	 *            the file to remove
	 */
	public static void remove(final String destFilename) {
		final File f = new File(destFilename);
		f.delete();
	}

	/**
	 * @return a human readable description of the action
	 */
	public abstract String getName();

	/**
	 * Executes the file action to the destination file, and uses the source
	 * file as source.
	 * 
	 * @param srcFilename
	 *            the complete source file name and path
	 * @param destFilename
	 *            the complete destination file name and path
	 */
	public abstract void execute(final String srcFilename, final String destFilename);

	/**
	 * Iterates through the files, compares each of them to the destination
	 * file, and if different, executes the file action.
	 * 
	 * @param srcBase
	 *            the root of all source file names
	 * @param srcPathsFromBase
	 *            the source file names, relative to srcBase
	 * @param comparismModes
	 *            the comparism mode that defines if there is something to do
	 */
	public void executeIfDifferent(final File srcBase, final String[] srcPathsFromBase, final int[] comparismModes) {
		for (int i = 0; i < srcPathsFromBase.length; i++) {
			final File srcPath = new File(srcBase, srcPathsFromBase[i]);
			final String destPath = todir + File.separator + getDestFilename(srcPathsFromBase[i]);
			if (clean) {
				remove(destPath);
				log("removing " + destPath, Project.MSG_INFO);
			} else {
				final boolean isDifferent = FileCompare.isDifferent(srcPath.getPath(), destPath, comparismModes);
				if (isDifferent) {
					log(getName() + " " + srcPath.getPath() + " to " + destPath, Project.MSG_INFO);
					execute(srcPath.getPath(), destPath);
					if (preservelastmodified) {
						File destfile = new File(destPath);
						destfile.setLastModified(srcPath.lastModified());
					}
				}
			}
		}
	}

	/**
	 * Executes the Ant task. In this case, iterates through the fileset and
	 * executes the task for each file, depending on the file compare with the
	 * destination file.
	 * 
	 * @see #executeIfDifferent(File, String[], int[])
	 */
	@Override
	public void execute() {
		final String message = getProject().getProperty("ant.project.name");
		log("Here is project '" + message + "'.", Project.MSG_DEBUG);
		log("I am used in: " + getLocation(), Project.MSG_DEBUG);
		log("todir = " + todir, Project.MSG_DEBUG);
		log("comparismmode = " + comparismMode, Project.MSG_DEBUG);
		int[] comparismModes = FileCompare.comparismModes(comparismMode);
		for (int cm : comparismModes) {
			log("Comparism mode " + cm, Project.MSG_DEBUG);
		}
		for (Iterator<FileSet> itFSets = filesets.iterator(); itFSets.hasNext();) { // 2
			final FileSet fs = (FileSet) itFSets.next();
			final DirectoryScanner ds = fs.getDirectoryScanner(getProject()); // 3
			final File base = ds.getBasedir(); // 5
			final String[] includedFiles = ds.getIncludedFiles();
			executeIfDifferent(base, includedFiles, comparismModes);
		}
	}

	/**
	 * If no files to handle are given on the command line, use the following
	 * list as default.
	 * 
	 * @return the default implementation returns an empty list, but subclasses
	 *         may overload this behavior
	 */
	public String[] defaultFiles() {
		return new String[] {};
	}

	/**
	 * Executes a fileset given by the parameters. For usage from non-ant
	 * environments.
	 * 
	 * @param basedir
	 *            the base directory of all source files
	 * @param files
	 *            the source file names relative to the base directory
	 */
	public void execute(final String basedir, String[] files) {
		if ((null == files) || (0 == files.length)) {
			files = defaultFiles();
		}
		log("todir = " + todir, Project.MSG_DEBUG);
		log("comparismmode = " + comparismMode, Project.MSG_DEBUG);
		final int[] comparismModes = FileCompare.comparismModes(comparismMode);
		for (final int cm : comparismModes) {
			log("Comparism mode " + cm, Project.MSG_DEBUG);
		}
		executeIfDifferent(new File(basedir), files, comparismModes);
	}

	/**
	 * Executes a fileset given by the parameters. For usage from non-ant
	 * environments.
	 * 
	 * @param basedir
	 *            the base directory of all source files
	 * @param files
	 *            the source file names relative to the base directory
	 */
	public void execute(final String basedir, final Vector<String> files) {
		String[] filesArr = new String[files.size()];
		filesArr = files.toArray(filesArr);
		execute(basedir, filesArr);
	}

	/**
	 * Outputs the usage for command line calling of this class for human
	 * readers.
	 * 
	 * @param paramsList
	 *            additional parameters, or <code>null</code>
	 * @param paramsExplanation
	 *            explanation for those additional parameters, or
	 *            <code>null</code>
	 * @return a human-readable string explaining the usage of this command
	 */
	public String usage(final String paramsList, final String paramsExplanation) {
		final StringBuffer s = new StringBuffer();
		s.append("usage: java " + getClass().getName()
				+ "[-d todir] [-s fromdir] [-c comparismmode] [-p|+p] [+clean|-clean] "
				+ ((null == paramsList) ? "" : paramsList.trim() + " ") + "file(s)\n");
		s.append(" -d todir : The destination root path.\n");
		s.append(" -s fromdir : The source root path. Files are read starting from there.\n");
		s.append(" -c comparismmode : The comparism mode. A comma separated list of:\n");
		s.append("        ");
		for (final String fcs : FileCompare.COMPARISM_MODE_NAMES) {
			s.append(fcs);
			s.append(" ");
		}
		s.append(" -p : Turns preservefilename off.\n");
		s.append(" +p : Turns preservefilename on (default).\n");
		s.append(" -clean : Turns cleaning mode off (default).\n");
		s.append(" +clean : Turns cleaning mode on. No files are created, instead all destination files are deleted.\n");
		if (null != paramsExplanation) {
			s.append(paramsExplanation);
		}
		return s.toString();
	}

	/**
	 * Parses the args as if coming from the command line, and executes this
	 * task accordingly.
	 * 
	 * @param args
	 *            the args as coming from command line
	 */
	public void executeMain(final String[] args) {
		if (args.length < 1) {
			System.err.println(usage(null, null));
			System.exit(1);
		}
		String fromdir = ".";
		String mode = "";
		final Vector<String> files = new Vector<String>();
		for (int i = 0; i < args.length; i++) {
			final String arg = args[i];
			if ("-d".equals(mode)) {
				setTodir(arg);
				mode = "";
			} else if ("-s".equals(mode)) {
				fromdir = arg;
				mode = "";
			} else if ("-c".equals(mode)) {
				setComparismmode(arg);
				mode = "";
			} else if ("-d".equals(arg)) {
				mode = arg;
			} else if ("-s".equals(arg)) {
				mode = arg;
			} else if ("-c".equals(arg)) {
				mode = arg;
			} else if ("-p".equals(arg)) {
				setPreservelastmodified(false);
			} else if ("+p".equals(arg)) {
				setPreservelastmodified(true);
			} else if ("-clean".equals(arg)) {
				setClean(false);
			} else if ("+clean".equals(arg)) {
				setClean(true);
			} else {
				files.add(arg);
			}
		}
		execute(fromdir, files);
	}
}
