package AntTasks;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;

/**
 * Compares two files, depending on several comparism modes.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FileCompare {

	/**
	 * Compares two files, which has the later modification.
	 * 
	 * @param srcFile
	 *            source file
	 * @param destFile
	 *            destination file
	 * @return <code>true</code> if the source is newer than the destination
	 *         file
	 */
	public static boolean isDifferentFileDate(final File srcFile, final File destFile) {
		final long srcTimestamp = srcFile.lastModified();
		final long destTimestamp = destFile.lastModified();
//        System.out.println(srcFile.getName() + " Timestamp src=" + srcTimestamp + " dest=" + destTimestamp);
		return (srcTimestamp > destTimestamp);
	}

	/**
	 * Compares two files, which has the later modification.
	 * 
	 * @param srcFilename
	 *            path and name of the source file
	 * @param destFilename
	 *            path and name of the destination file
	 * @return <code>true</code> if the source is newer than the destination
	 *         file
	 */
	public static boolean isDifferentFileDate(final String srcFilename, final String destFilename) {
		final File srcFile = new File(srcFilename);
		final File destFile = new File(destFilename);
		return isDifferentFileDate(srcFile, destFile);
	}

	/**
	 * Compares two files by binary comparism.
	 * 
	 * @param srcFilename
	 *            path and name of the source file
	 * @param destFilename
	 *            path and name of the destination file
	 * @return <code>true</code> if both files differ
	 */
	public static boolean isDifferentBinary(final String srcFilename, final String destFilename) {
		boolean ret = false;
		try {
			final BufferedInputStream is1 = new BufferedInputStream(new FileInputStream(srcFilename));
			final BufferedInputStream is2 = new BufferedInputStream(new FileInputStream(destFilename));
			int is1char, is2char;
			do {
				is1char = is1.read();
				is2char = is2.read();
				if (is1char != is2char) {
					ret = true;
				}
			} while (!ret && (-1 != is1char) && (-1 != is2char));
			is1.close();
			is2.close();
		} catch (final IOException e) {
			ret = true;
		}
		return ret;
	}

	public static final int COMPARISM_MODE_LAST_MODIFIED = 0;
	public static final int COMPARISM_MODE_BINARY = 1;
	public static final int COMPARISM_MODE_TEXT = 2;
	public static final int COMPARISM_MODE_TEXT_IGNORE_WHITECHARS = 3;
	public static final int COMPARISM_MODE_ALWAYS_EQUAL = 4;
	public static final int COMPARISM_MODE_ALWAYS_UNEQUAL = 5;
	public static final int COMPARISM_MODE_IGNORE = -1;

	public static final String[] COMPARISM_MODE_NAMES = new String[] { 
			"filedate", "binary", "text", "textIgnoreWhitechars", "false", "true" };

	/**
	 * Compares two files, depending on the comparism mode.
	 * 
	 * @param srcFilename
	 *            the source file
	 * @param destFilename
	 *            the destination file
	 * @param comparismMode
	 *            defines the way how to compare those two files. One of
	 *            {@link #COMPARISM_MODE_LAST_MODIFIED},
	 *            {@link #COMPARISM_MODE_BINARY}, {@link #COMPARISM_MODE_TEXT},
	 *            {@link #COMPARISM_MODE_TEXT_IGNORE_WHITECHARS},
	 *            {@link #COMPARISM_MODE_ALWAYS_EQUAL},
	 *            {@link #COMPARISM_MODE_ALWAYS_UNEQUAL}.
	 * @return <code>true</code> if the files are different
	 */
	public static boolean isDifferent(final String srcFilename, final String destFilename, final int comparismMode) {
		boolean ret = false;
		switch (comparismMode) {
		case COMPARISM_MODE_LAST_MODIFIED: {
			ret = isDifferentFileDate(srcFilename, destFilename);
		}
			break;
		case COMPARISM_MODE_BINARY: {
			ret = isDifferentBinary(srcFilename, destFilename);
		}
			break;
		case COMPARISM_MODE_TEXT:
		case COMPARISM_MODE_TEXT_IGNORE_WHITECHARS: {
			try {
				final BufferedReader is1 = new BufferedReader(new FileReader(srcFilename));
				final BufferedReader is2 = new BufferedReader(new FileReader(destFilename));
				int is1char, is2char;
				do {
					do {
						is1char = is1.read();
					} while ((COMPARISM_MODE_TEXT_IGNORE_WHITECHARS == comparismMode) && Character.isWhitespace(is1char));
					do {
						is2char = is2.read();
					} while ((COMPARISM_MODE_TEXT_IGNORE_WHITECHARS == comparismMode) && Character.isWhitespace(is2char));
					if (is1char != is2char) {
						ret = true;
					}
				} while (!ret && (-1 != is1char) && (-1 != is2char));
				is1.close();
				is2.close();
			} catch (final IOException e) {
				ret = true;
			}
		}
		break;
		case COMPARISM_MODE_ALWAYS_EQUAL: 
			ret = false;
			break;
		case COMPARISM_MODE_ALWAYS_UNEQUAL: 
			ret = true;
			break;
		}
//        System.out.println("Comparism mode " + comparismMode + " resulted in " + ret);
		return ret;
	}

	/**
	 * Compares two files, depending on the comparism mode.
	 * 
	 * @param srcFilename
	 *            the source file
	 * @param destFilename
	 *            the destination file
	 * @param comparismModes
	 *            defines the way how to compare those two files. To consider
	 *            two files as equal, all given comparism modes need to prove equality.
	 *            Each element of the array could be one of
	 *            {@link #COMPARISM_MODE_LAST_MODIFIED},
	 *            {@link #COMPARISM_MODE_BINARY}, {@link #COMPARISM_MODE_TEXT},
	 *            {@link #COMPARISM_MODE_TEXT_IGNORE_WHITECHARS},
	 *            {@link #COMPARISM_MODE_ALWAYS_EQUAL},
	 *            {@link #COMPARISM_MODE_ALWAYS_UNEQUAL}.
	 * @return <code>true</code> if the files are different
	 */
	public static boolean isDifferent(final String srcFilename, final String destFilename, final int[] comparismModes) {
		boolean ret = true;
		for (final int mode : comparismModes) {
			if (COMPARISM_MODE_IGNORE != mode) {
				final boolean diff = isDifferent(srcFilename, destFilename, mode);
				if (!diff) {
					return false;
				}
			}
		}
		return ret;
	}

	/**
	 * Parses a string containing the comparism modes.
	 * 
	 * @param comparismMode
	 *            the string containing a comma separated list of comparism
	 *            modes
	 * @return an array of integer constants containing the comparism modes to
	 *         apply
	 */
	public static int[] comparismModes(final String comparismMode) {
		final String[] tokens = comparismMode.split(",");
		final int[] ret = new int[tokens.length];
		for (int i = 0; i < tokens.length; i++) {
			final String token = tokens[i].trim();
			boolean isModeFound = false;
			for (int mode = 0; !isModeFound && (mode < COMPARISM_MODE_NAMES.length); mode++) {
				if (COMPARISM_MODE_NAMES[mode].equalsIgnoreCase(token)) {
					ret[i] = mode;
					isModeFound = true;
				}
			}
			if (!isModeFound) {
				ret[i] = COMPARISM_MODE_IGNORE;
			}
		}
		return ret;
	}
}
