package AntTasks;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

import org.apache.tools.ant.Project;

/**
 * Removes the "required" annotation from java source files.
 * This is necessary for an incompatibility between XJC in JDK 1.8 (which creates that annotation) 
 * and java compilers of older JDKs (which don't understand that annotation).
 * 
 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FileRemoveJavaRequiredAnnotation extends FileAction {

	@Override
	public String getName() {
		return "adapting";
	}

	/**
	 * @param srcFilename
	 *            the file that will be changed, so this file is source and
	 *            destination.
	 * @param destFilename
	 *            ignored.
	 */
	@Override
	public void execute(String srcFilename, String destFilename) {
        try {
        	final BufferedReader is = new BufferedReader(new FileReader(srcFilename));
        	final ArrayList<String> lines = new ArrayList<String>();
            String oneLine;
//            is.readLine();
            while ((oneLine = is.readLine()) != null) {
        		lines.add(oneLine);
            }
            is.close();
            adapt1(lines);
            adapt2(lines);
            adapt3(lines);
            adapt4(lines, "Ruft den Wert der ", "-Eigenschaft ab.", "@return", "     * @return ", "");
            adapt4(lines, "Legt den Wert der ", "-Eigenschaft fest.", "@param", "     * @param value ", "");
            adapt5(lines);
            adapt6(lines);
            adapt7(lines);
        	final BufferedWriter os = new BufferedWriter(new FileWriter(srcFilename));
        	for (final String line : lines) {
                os.write(line);
                os.newLine();
            }
            os.close();
        } catch (final IOException e) {
            log(e.toString(), Project.MSG_ERR);
        }
	}
	
	/**
	 * Removes &quot;@required&quot; attribute where JAXB 1.8 creates them, to be backwards compatible for JDK 1.6.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt1(final ArrayList<String> lines) {
		for (int i = 0; i < lines.size(); i++) {
			final String line = lines.get(i);
	    	if (line.contains(PATTERN)) {
				lines.set(i, line.replace(PATTERN, ""));
	    		mylog(line);
	    	}
		}
	}

	private final String PATTERN = ", required = false";

	/**
	 * Adds javadoc &quot;@return&quot; where JAXB does not create them, so javadoc 1.8 does not barf.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt2(final ArrayList<String> lines) {
		for (int i = 0; i < lines.size() - 1; i++) {
			final String line = lines.get(i);
	    	if (line.contains(PATTERN21) && lines.get(i + 1).contains(PATTERN22)) {
				lines.set(i, line.replace(PATTERN21, "* @return"));
	    		mylog(line);
	    	}
		}
	}
	
	private final String PATTERN22 = "Objects of the following type(s) are allowed in the list";
	private final String PATTERN21 = "* <p>";

	/**
	 * Replaces bad closing bracket comments, to fix a bug in JAXB 1.8.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt3(final ArrayList<String> lines) {
		for (int i = 0; i < lines.size() - 1; i++) {
			String line = lines.get(i);
			boolean wasChanged = false;
			if (line.contains("&lt;")) {
				line = line.replaceAll(">", "&gt;");
				wasChanged = true;
			}
			if (wasChanged) {
				lines.set(i,  line);
				mylog(line);
			}
		}
	}

	/**
	 * Adds javadoc &quot;@return&quot; where JAXB does not create them, so javadoc 1.8 does not barf.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt4(final ArrayList<String> lines, String openerPrefix, String openerPostfix, String canceler, String replacementPrefix, String replacementPostfix) {
		String name = "";
		for (int i = 0; i < lines.size() - 1; i++) {
			String line = lines.get(i);
			final int idx1 = line.indexOf(openerPrefix);
			final int idx2 = line.indexOf(openerPostfix);
			if ((0 <= idx1) && (idx1 < idx2)) {
				name = line.substring(idx1 + openerPrefix.length(), idx2);
			}
			if (line.contains(canceler)) {
				name = "";
			}
			if (line.contains("*/")) {
				if (!name.isEmpty()) {
					final String lline = replacementPrefix + name + replacementPostfix;
					lines.set(i - 1,  lline);
					mylog(lline);
				}
				name = "";
			}
		}
	}
	
	/**
	 * Removes empty HTML paragraphs created by JAXB, so javadoc 1.8 does not barf.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt5(final ArrayList<String> lines) {
		for (int i = 0; i < lines.size() - 1; i++) {
			String line = lines.get(i);
			if (line.equals(" * <p>") && lines.get(i + 1).equals(" * <pre>")) {
				final String lline = " * ";
				lines.set(i,  lline);
				mylog(lline);
			}
		}
	}
		
	/**
	 * Adds javadoc &quot;@return&quot; where JAXB does not create them, so javadoc 1.8 does not barf.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt6(final ArrayList<String> lines) {
		for (int i = 0; i < lines.size() - 1; i++) {
			String line = lines.get(i);
			if (line.startsWith("     * Create an instance of ")) {
				final String lline = line.replaceFirst("Create", "@return");
				lines.set(i,  lline);
				mylog(lline);
			}
		}
	}
		
	/**
	 * Adds javadoc &quot;@param&quot; where JAXB does not create them, so javadoc 1.8 does not barf.
	 * 
	 * @param lines
	 *            the lines of the java source file to modify
	 */
	private void adapt7(final ArrayList<String> lines) {
		for (int i = 0; i < lines.size() - 1; i++) {
			String line = lines.get(i);
			if (line.startsWith("    public JAXBElement<")
					&& line.contains(" create") && line.contains("value)")) {
				for (int j = 3; j <= 4; j++) {
					if (lines.get(i - j).equals("     * ")) {
						final String lline = "     * @param value egal";
						lines.set(i - j, lline);
						mylog(lline);
						break;
					}
				}
			}
		}
	}
	
	private void mylog(final String line) {
//		log(line);
	}
		
    public static void main(String[] args) {
    	FileRemoveJavaRequiredAnnotation task = new FileRemoveJavaRequiredAnnotation();
        task.executeMain(args);
    }
}
