package AntTasks;

import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.imageio.ImageIO;

/**
 * Brightens a jpg to fade out to white to the right.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImgBrightenToRight extends FileAction {

	/**
	 * Brightens a single color value.
	 * 
	 * @param argb
	 *            the alpha, red, green or blue value. Should be between 0x00
	 *            and 0xff.
	 * @param howmuch
	 *            0 leaves as it is. 100 sets to white, -100 to black.
	 * @return the brightened red, green or blue value
	 */
	public static int brightenARGB(final int argb, final int howmuch) {
		int ret = argb;
		if (howmuch > 0) {
			ret += (0xff - ret) * howmuch / 100;
		} else {
			ret -= ret * howmuch / 100;
		}
		return ret & 0xff;
	}

	/**
	 * Brightens a color (all rgb components).
	 * 
	 * @param argb
	 *            the RGB color value
	 * @param howmuch
	 *            0 leaves as it is. 100 sets to white, -100 to black.
	 * @return the brightened RGB color value
	 */
	public static int brighten(final int argb, final int howmuch) {
		int r = ((argb >> 16) & 0xff);
		int g = ((argb >>  8) & 0xff);
		int b = ((argb >>  0) & 0xff);
		r = brightenARGB(r, howmuch);
		g = brightenARGB(g, howmuch);
		b = brightenARGB(b, howmuch);
		int ret = (argb & 0xff000000) | (r << 16) | (g << 8) | b;
		return ret;
	}

	/**
	 * Brightens a buffered image
	 * 
	 * @param bi
	 *            the image to brighten
	 * @param start
	 *            the brighten factor on the left side
	 * @param end
	 *            the brighten factor on the right side
	 */
	public static void brighten(final BufferedImage bi, final int start, final int end) {
		final int h = bi.getHeight();
		final int w = bi.getWidth();
		for (int x = 0; x < w; x++) {
			final int howmuch = start + (end - start) * x / w;
			for (int y = 0; y < h; y++) {
				int rgb = bi.getRGB(x, y);
				bi.setRGB(x, y, brighten(rgb, howmuch));
			}
		}
	}

	@Override
	public String getName() {
		return "brightening";
	}

	@Override
	public void execute(final String fullSrcFilename, final String fullDestFilename) {
		try {
			final BufferedInputStream is = new BufferedInputStream(new FileInputStream(fullSrcFilename));
			final BufferedImage bi = ImageIO.read(is);
			brighten(bi, 25, 100);
			final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(fullDestFilename));
			ImageIO.write(bi, "JPG", os);
			os.close();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param srcFilename
	 *            the source file name
	 * @return the corresponding destination filename. Current implementation
	 *         appends a "v" to the filename.
	 */
	@Override
	public String getDestFilename(final String srcFilename) {
		return changeFilename(srcFilename, "", "v", null, null);
	}

	@Override
	public String[] defaultFiles() {
		return new String[] { "dev/P9270090-c1.jpg" };
	}

	public static void main(final String[] args) {
		final ImgBrightenToRight task = new ImgBrightenToRight();
		task.executeMain(args);
	}
}
