package AntTasks;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.imageio.ImageIO;

/**
 * Converts a GIF to PNG and sets the red color as transparent color.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImgConvertGifToPngAndSetRedToTransparent extends FileAction {

	private static final String MAGENTA_TO_TRANSPARENT = ".MagentaToTransparent";
	protected static final int COLOR_SET_TO_TRANSPARENT = Color.red.getRGB();
	protected static final int COLOR_SET_TO_TRANSPARENT2 = Color.magenta.getRGB();

	/**
	 * Sees that a pixel in the image with the given transparentColor gets 100%
	 * transparent. The code is taken from a <a href=
	 * "http://forum.java.sun.com/thread.jspa?threadID=5227456&tstart=150">thread
	 * in the java forum</a>.
	 * 
	 * @param x
	 *            unused
	 * @param y
	 *            unused
	 * @param argb
	 *            the ARGB color of the pixel to make transparent, if equal to
	 *            transparent color
	 * @param transparentColor
	 *            the ARGB color to compare with
	 * @return the transformed ARGB color value
	 */
	public static final int filterRGB(final int x, final int y, final int argb, final int transparentColor) {
		if ((argb | 0xFF000000) == (transparentColor | 0xFF000000)) {
			// Mark the alpha bits as zero - transparent
			return 0x00FFFFFF & argb;
		} else {
			// Mark the alpha bits as 100% opaque
			return 0xFF000000 | argb;
		}
	}

	public static boolean convertGIFtoPNGandSetTransparentColor(final String inputFilename, final String outputFilename,
			final int transparentColorRGB) {
		boolean ret = false;
		try {
			final BufferedInputStream is = new BufferedInputStream(new FileInputStream(inputFilename));
			final BufferedImage bi = ImageIO.read(is);
			final int h = bi.getHeight();
			final int w = bi.getWidth();
			final BufferedImage bid = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
			for (int x = 0; x < w; x++) {
				for (int y = 0; y < h; y++) {
					final int rgb = bi.getRGB(x, y);
					final int newRGB = filterRGB(x, y, rgb, transparentColorRGB);
//					System.out.print(((newRGB & 0xff000000) != 0x00000000) ? '*' : '-');
					bid.setRGB(x, y, newRGB);
				}
//				System.out.println();
			}
			final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(outputFilename));
			ImageIO.write(bid, "PNG", os);
			os.close();
			ret = true;
		} catch (final IOException e) {
			e.printStackTrace();
		}
		return ret;
	}

	@Override
	public String getName() {
		return "Iconizing";
	}

	/**
	 * @param srcFilename
	 *            the source file name
	 * @return the corresponding destination filename. Current implementation
	 *         changes the filename extension to png.
	 */
	@Override
	public String getDestFilename(final String srcFilename) {
		return changeFilename(srcFilename, null, null, MAGENTA_TO_TRANSPARENT, "png");
	}

	@Override
	public void execute(final String fullSrcFilename, final String fullDestFilename) {
		if (fullSrcFilename.contains(MAGENTA_TO_TRANSPARENT)) {
//			System.out.println("Magenta is 0x" + Integer.toHexString(COLOR_SET_TO_TRANSPARENT2));
			convertGIFtoPNGandSetTransparentColor(fullSrcFilename, fullDestFilename, COLOR_SET_TO_TRANSPARENT2);
		} else {
			convertGIFtoPNGandSetTransparentColor(fullSrcFilename, fullDestFilename, COLOR_SET_TO_TRANSPARENT);
		}
	}

	@Override
	public String[] defaultFiles() {
		return new String[]{
			"de/ped/tools/dev/Action.Help.MagentaToTransparent.png",
		};
	}

	public static void main(final String[] args) {
		final ImgConvertGifToPngAndSetRedToTransparent task = new ImgConvertGifToPngAndSetRedToTransparent();
		task.executeMain(args);
	}
}
