package AntTasks;

import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;

import javax.imageio.ImageIO;

/**
 * Thumbnails an image.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImgThumbnail extends FileAction {

	protected static final double[] DEFAULT_SNIPPET = new double[] { 0.0, 0.0, 1.00, 1.00, }; 

	protected static final HashMap<String, double[]> snippets = new HashMap<String, double[]>();
	static {
		// %offset-x, %offset-y, %width, %height
		snippets.put("1995-11-Naa.jpg", 				new double[] { 0.00, 0.10, 1.00, 0.72, } );
		snippets.put("1996-Karikatur_von_Jantiff.jpg",	new double[] { 0.65, 0.20, 0.35, 0.60, } );
		snippets.put("1998-Karikatur.gif", 				new double[] { 0.00, 0.10, 1.00, 0.72, } );
		snippets.put("1998-Vilamoura.jpg", 				new double[] { 0.10, 0.00, 0.90, 1.00, } );
		snippets.put("2000-Santos.jpg", 				new double[] { 0.00, 0.00, 1.00, 0.40, } );
		snippets.put("2003-11-17-Mit_Katzen.jpg", 		new double[] { 0.10, 0.15, 0.80, 0.60, } );
		snippets.put("2004-09-Egeskov-pastoral.jpg", 	new double[] { 0.25, 0.50, 0.50, 0.45, } ); // 0.89
		snippets.put("2004-10-Passfoto.jpg", 			new double[] { 0.00, 0.10, 1.00, 0.76, } );
		snippets.put("2006-08-07_20-07-45.jpg", 		new double[] { 0.20, 0.00, 0.65, 1.00, } );
		snippets.put("2009-07-30-Karikatur.gif", 		new double[] { 0.00, 0.00, 1.00, 0.90, } );
		snippets.put("2018-05-28-Beim_Segeln.jpg", 		new double[] { 0.00, 0.00, 0.80, 1.00, } );
		snippets.put("2018-12-31-Beim_Kochen.jpg", 		new double[] { 0.00, 0.00, 0.90, 0.80, } );	
	}
	
    protected static final int THUMBNAIL_MAX_WIDTH = 40;
    protected static final int THUMBNAIL_MAX_HEIGHT = THUMBNAIL_MAX_WIDTH; 
    
    protected static final boolean IS_ALIASING_THUMBNAIL = true;

    /** Switches between jpg and png thumbnail creation. */
    protected static final boolean IS_THUMBNAIL_PNG = false;

    public static final int weightedRGB(final double weightedColor, final double weightedSum) {
        return ((int)(weightedColor / weightedSum + 0.5)) & 0xFF;
    }

    @Override
    public String getName() {
        return "Thumbnailing";
    }

	/**
	 * @param srcFilename
	 *            the source file name
	 * @return the corresponding destination filename. Current implementation
	 *         appends a "_t" to the name and changes the extension.
	 */
	@Override
    public String getDestFilename(final String srcFilename) {
        return changeFilename(srcFilename, null, "_t", null, (IS_THUMBNAIL_PNG ? "png" : "jpg"));
    }
    
    @Override
    public void execute(final String fullSrcFilename, final String fullDestFilename) {
        try {
        	final BufferedInputStream is = new BufferedInputStream(new FileInputStream(fullSrcFilename));
        	final BufferedImage bi = ImageIO.read(is);
        	final String filename = new File(fullSrcFilename).getName();
    		double[] snippetPercent = snippets.get(filename);
    		if (null == snippetPercent) {
    	    	snippetPercent = DEFAULT_SNIPPET;
        	} else {
        		log("Taking snippet only of " + filename);
        	}
        	double[] snippetAbs = { 
        			(double)bi.getWidth() * snippetPercent[0], 
        			(double)bi.getHeight() * snippetPercent[1], 
        			(double)bi.getWidth() * snippetPercent[2], 
        			(double)bi.getHeight() * snippetPercent[3], 
        	};
            // Get the right factor how much to squeeze the image.
            final double factorX = snippetAbs[2] / (double)THUMBNAIL_MAX_WIDTH;
            final double factorY = snippetAbs[3] / (double)THUMBNAIL_MAX_HEIGHT;
            final double factor = Math.max(1.0, Math.max(factorX, factorY));
            // Calculate the radius for aliasing.
            final double radius = factor / 1.44;
            final int upperRadius = (int)radius + 1;
            // Calculate the size of the destination thumbnail.
            final int dw = (int)(snippetAbs[2] / factor + 0.5);
            final int dh = (int)(snippetAbs[3] / factor + 0.5);
            final double sx_left = snippetAbs[0] + (snippetAbs[2] - (factor * dw)) / 2.0;
            final double sy_top  = snippetAbs[1] + (snippetAbs[3] - (factor * dh)) / 2.0;
            final BufferedImage bid = new BufferedImage(dw, dh, (IS_THUMBNAIL_PNG ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB));
            for (int tx = 0; tx < dw; tx++) {
            	final double sx = sx_left + factor * tx;
                for (int ty = 0; ty < dh; ty++) {
                	final double sy = sy_top + factor * ty;
                	final int isx = (int)(sx + 0.5);
                	final int isy = (int)(sy + 0.5);
                    boolean isColorDefined = false;
                    double sumColorR = 0.0;
                    double sumColorG = 0.0;
                    double sumColorB = 0.0;
                    double sumWeight = 0.0;
                    try {
                        if (IS_ALIASING_THUMBNAIL) {
                            for (int rx = -upperRadius; rx <= upperRadius; rx++) { 
                                for (int ry = -upperRadius; ry <= upperRadius; ry++) {
                                	final double dist = Math.sqrt(rx * rx + ry * ry);
                                	final double weight = Math.max(0.0, radius - dist);
                                    try {
                                    	final int argb = bi.getRGB(isx + rx, isy + ry);
                                        isColorDefined = true;
                                        sumWeight += weight;
                                        sumColorR += (double)((argb >> 16) & 0xFF) * weight;
                                        sumColorG += (double)((argb >>  8) & 0xFF) * weight;
                                        sumColorB += (double)((argb >>  0) & 0xFF) * weight;
                                    } catch (final ArrayIndexOutOfBoundsException e) {
                                        // Nothing to do.
                                    }
                                }
                            }
                        } else {
                        	final int argb = bi.getRGB(isx, isy);
                            isColorDefined = true;
                            sumWeight += 1.0;
                            sumColorR += (double)((argb >> 16) & 0xFF);
                            sumColorG += (double)((argb >>  8) & 0xFF);
                            sumColorB += (double)((argb >>  0) & 0xFF);
                        }
                    } catch (final ArrayIndexOutOfBoundsException e) {
                        // Nothing to do.
                    }
                    int argb = 0x00FFFFFF; // 100% transparent white
                    if (isColorDefined) {
                        argb = weightedRGB(sumColorR, sumWeight) << 16 |
                               weightedRGB(sumColorG, sumWeight) <<  8 |
                               weightedRGB(sumColorB, sumWeight) <<  0;
                        if (IS_THUMBNAIL_PNG) {
                            argb |= 0xFF000000;
                        }
                    }
                    bid.setRGB(tx, ty, argb);
                }
            }
            // Write it all to file.
            final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(fullDestFilename));
            ImageIO.write(bid, (IS_THUMBNAIL_PNG ? "PNG" : "JPG"), os);
            os.close();
        } catch (final IOException e) {
            e.printStackTrace();
        }
    }

    @Override
    public String[] defaultFiles() {
        return new String[]{ 
            "pictures/1969-Baby.jpg",
            "pictures/1994-selbstversunken.gif",
            "pictures/1995-11-Geburtstagsfeier222.gif",
            "pictures/1995-11-Naa.jpg",
            "pictures/1995-Selbstportrait.gif",
            "pictures/1996-Schnellportrait_mit_Raballion.jpg",
            "pictures/1996-Karikatur_von_Jantiff.jpg",
            "pictures/1998-Vilamoura.jpg",
            "pictures/1998-Karikatur.gif",
            "pictures/2000-Santos.jpg",
            "pictures/2000-SantosFace.jpg",
            "pictures/2004-09-Egeskov-pastoral.jpg",
            "pictures/2004-10-Passfoto.jpg",
            "pictures/2006-08-07_20-07-45.jpg",
            "pictures/Baer_und_Tiger_Janosch.gif",
            "pictures/2003-11-17-Mit_Katzen.jpg",
            "pictures/2003-11-17-Mit_Katzen_Face.jpg",
        };
    }
    
    public static void main(final String[] args) {
    	final ImgThumbnail task = new ImgThumbnail();
        task.executeMain(args);
    }
}
