package de.ped.advent.logic;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Random;

import de.ped.tools.HtmlHelper;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;

/**
 * The business logic behind the advent JSP.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach
 */
public class Advent {

	/**
	 * The encapsulation of a database access
	 * @author Copyright (c) 2022 Peter Diefenbach
	 */
	public static class DBA {
		private static final String DRIVER = "com.mysql.jdbc.Driver"; 
		private final String url; 
		private final String dbName;
		private final String userName; 
		private final String password;
		public DBA(final String url, final String dbName, final String userName, final String password) {
			this.url = url;
			this.dbName = dbName;
			this.userName = userName;
			this.password = password;
		}

		/**
		 * Gets a connection to this database.
		 * @return the connection
		 * @throws SQLException if a database access error occurs
		 * @throws InstantiationException if this Class named in {@link #DRIVER} 
		 * 		represents an abstract class, 
		 * 		an interface, an array class, a primitive type, or void;
		 * 		or if the class has no nullary constructor;
		 * 		or if the instantiation fails for some other reason. 
		 * @throws IllegalAccessException if the class named in {@link #DRIVER} or its nullary
		 * 		constructor is not accessible. 
		 * @throws ClassNotFoundException if the class named in {@link #DRIVER} cannot be located
		 */
		public Connection getConnection() throws SQLException, InstantiationException, IllegalAccessException, ClassNotFoundException {
			Class.forName(DRIVER).newInstance();
			return DriverManager.getConnection(url + dbName, userName, password);
		}
	}

	private DBA[] dbas = {
		new DBA("jdbc:mysql://s001.profi1.de:3306/", "c171_db", "c171_mcblack", "IvHzBTE7"),
		new DBA("jdbc:mysql://localhost:3306/", "usr_web468_1", "web468", "W57xNXx5"),
	};

	public static final int DEFAULT_YEAR = 2023;
	public static final String DEFAULT_CONTENT = "<p>Was machst Du hier?!?</p>";
	public static final String PARAM_KEY_DATE  = "item";
	public static final String PARAM_KEY_FOR   = "for";
	public static final String PARAM_KEY_DEMO  = "demo";
	public static final String PARAM_KEY_RESET = "reset";
	public static final String PARAM_KEY_ALL   = "all";
	public static final String PARAM_KEY_LOCAL = "local";

	/** Parameter for the christmas story handling. */
	public static final String PARAM_KEY_STORY = "story";
	public static final String PARAM_KEY_PART  = "part";

	public static final String PARAM_VAL_FOR_LEONIE  = "leonie";
	public static final String PARAM_VAL_FOR_FAMILY  = "family";
	public static final String PARAM_VAL_FOR_MARTINA = "martina";
	public static final String PARAM_VAL_FOR_EDITH   = "edith";
	public static final String PARAM_VAL_FOR_URSULA  = "ursula";
	public static final String PARAM_VAL_FOR_PAUL    = "paul";

	public static final String PARAM_VAL_TRUE = "true";

	public static final String DEFAULT_DATE_FORMAT_STRING = "yyyy-MM-dd";
	public static final SimpleDateFormat DEFAULT_DATE_FORMAT = new SimpleDateFormat(DEFAULT_DATE_FORMAT_STRING);

	private static final String ESC_KITTEN = "@Kitten ";
	private static final String ESC_TEXTAREA = "@Textarea ";

	public static String createKey(final int year, final int month, final int day) {
		return String.format("%04d-%02d-%02d", year, month, day);
	}

	public static String href(final String valueItem, final String valueFor, final String valueDemo) {
		String ret = "index.jsp";
		if (null != valueItem) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_DATE, valueItem);
		}
		if (null != valueFor) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_FOR, valueFor);
		}
		if (null != valueDemo) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_DEMO, valueDemo);
		}
		ret = HtmlHelper.sgmlify(ret);
		return ret;
	}

	/**
	 * Parses a parameter and returns a normalized form of the value.
	 * 
	 * @param key
	 *            the parameter key to parse
	 * @param valueRead
	 *            the parameter value to parse
	 * @return the normalized parameter value
	 */
	public static String parseParam(final String key, final String valueRead) {
		String ret = null;
		if (PARAM_KEY_DATE.equalsIgnoreCase(key)) {
			ret = parseDate(valueRead);
		} else if (PARAM_KEY_DEMO.equalsIgnoreCase(key)) {
			ret = parseDate(valueRead);
		} else if (PARAM_KEY_FOR.equalsIgnoreCase(key)) {
			if ((null == valueRead) || (valueRead.isEmpty()) || (PARAM_VAL_FOR_LEONIE.equalsIgnoreCase(valueRead))) {
				ret = PARAM_VAL_FOR_LEONIE;
			} else if (PARAM_VAL_FOR_MARTINA.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_FOR_MARTINA;
			} else if (PARAM_VAL_FOR_EDITH.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_FOR_EDITH;
			} else if (PARAM_VAL_FOR_URSULA.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_FOR_URSULA;
			} else if (PARAM_VAL_FOR_PAUL.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_FOR_PAUL;
			} else {
				ret = PARAM_VAL_FOR_FAMILY;
			}
		} else if (PARAM_KEY_ALL.equalsIgnoreCase(key)) {
			if (PARAM_VAL_TRUE.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_TRUE;
			}
		} else if (PARAM_KEY_LOCAL.equalsIgnoreCase(key)) {
			if (PARAM_VAL_TRUE.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_TRUE;
			}
		} else if (PARAM_KEY_RESET.equalsIgnoreCase(key)) {
			if (null != valueRead) {
				ret = PARAM_VAL_TRUE;
			}
		} else if (PARAM_KEY_STORY.equalsIgnoreCase(key)) {
			ret = valueRead;
		} else if (PARAM_KEY_PART.equalsIgnoreCase(key)) {
			ret = valueRead;
		} else {
			throw new IllegalArgumentException("Unknown parameter key: " + key);
		}
		return ret;
	}

	private static String parseDate(final String valueRead) {
		String ret = null;
		Date parsedDate = null;
		if (null != valueRead) {
			if (null == parsedDate) {
				parsedDate = parseDate(valueRead, DEFAULT_DATE_FORMAT_STRING);
			}
			if (null == parsedDate) {
				parsedDate = setDecember(parseDate(valueRead, "yyyy-dd"));
			}
			if (null == parsedDate) {
				parsedDate = parseDate(valueRead, "yyyyMMdd");
			}
			if (null == parsedDate) {
				parsedDate = setDecember(parseDate(valueRead, "yyyydd"));
			}
			if (null == parsedDate) {
				try {
					final int number = Integer.parseInt(valueRead);
					if ((1 <= number) && (number <= 24)) {
						ret = String.format(Integer.toString(DEFAULT_YEAR) + "-12-%02d", number);
					} else if (1000 <= number) {
						ret = String.format("%04d", number);
					}
				} catch (final NumberFormatException ex) {
					// Nothing to do
				}
			}
			if (null != parsedDate) {
				ret = DEFAULT_DATE_FORMAT.format(parsedDate);
			}
		}
		return ret;
	}

	private static Date setDecember(Date parsedDate) {
		if (null != parsedDate) {
			final GregorianCalendar cal = new GregorianCalendar();
			cal.setTime(parsedDate);
			cal.set(Calendar.MONTH, Calendar.DECEMBER);
			parsedDate = cal.getTime();
		}
		return parsedDate;
	}

	private static Date parseDate(final String value, final String format) {
		Date ret = null;
		final SimpleDateFormat df = new SimpleDateFormat(format);
		try {
			ret = df.parse(value);
		} catch (ParseException e) {
			// Just ignore
		}
		return ret;
	}

	private Messages messages = null;

	public String getTitle(final String vFor) {
		return messages().getString(vFor, null);
	}

	private String getDateAsHeading(final String vItem) {
		return "<h2>" + getDateFormatted(vItem) + "</h2>";
	}

	public String getContent(final String vItem, final String vFor, final String vDemo, final String vAll) {
		String ret = null;
		if (null == vAll) {
			ret = getDateAsHeading(vItem) + getContent(vItem, vFor, vDemo);
		} else {
			final StringBuffer buf = new StringBuffer();
			for (int i = 1; i <= 24; i++) {
				buf.append(getContent(vItem.substring(0, vItem.length() - 2) + String.format("%02d", i), vFor, vDemo, null));
			}
			ret = buf.toString();
		}
		return ret;
	}

	public String getContent(final String vItem, final String vFor, final String vDemo) {
		final String key = vFor + "_" + vItem;
		String ret = parseContent(vItem, vFor, vDemo, messages().getString(key, null));
		if (null == ret) {
			ret = DEFAULT_CONTENT;
		}
		return ret;
	}

	/**
	 * Returns the given date in the format as it will be presented to the user
	 * on the page.
	 * 
	 * @param key
	 *            the normalized date key
	 * @return The date string
	 */
	public String getDateFormatted(final String key) {
		final SimpleDateFormat df = new SimpleDateFormat("dd. MMMM yyyy");
		String ret;
		try {
			ret = df.format(DEFAULT_DATE_FORMAT.parse(key));
		} catch (ParseException e) {
			ret = e.getLocalizedMessage();
		}
		return ret;
	}

	private Messages messages() {
		if (null == messages) {
			messages = new Messages("de.ped.advent.logic.messages",
					"Advent",
					"1.2", 
					"Copyright " + Integer.toString(DEFAULT_YEAR) + " Peter Diefenbach (peter@pdiefenbach.de)",
					"Adventskalender",
					"Adventskalender",
					Messages.DE_LOCALES,
					null,
					0);
		}
		return messages;
	}

	private String parseContent(final String vItem, final String vFor, final String vDemo, String ret) {
		if (null != ret) {
			if (ret.startsWith(ESC_KITTEN)) {
				final String kitten = messages().getString("Kitten", null);
				if (null != kitten) {
					final String level = ret.split(" ")[1];
					ret = kitten.replaceFirst("value=\"1.0\"", "value=\"" + level + "\"");
				}
			}
			if (ret.startsWith(ESC_TEXTAREA)) {
				final String form = messages().getString("Textarea", null);
				if (null != form) {
					final int parts = calcNumOfStories(vItem, vFor);
					final String title = ret.split(" ")[1];
					final int part = 1;
					String content = loadStory(vItem, vFor, part);
					StringBuffer buf = new StringBuffer();
					buf.append("<input type=\"hidden\" id=\"item\" name=\"item\" value=\"@IIIII\" />");
					buf.append("\r\n");
					buf.append("<input type=\"hidden\" id=\"for\"  name=\"for\"  value=\"@FFFFF\" />");
					buf.append("\r\n");
					buf.append("<input type=\"hidden\" id=\"demo\" name=\"demo\" value=\"@DDDDD\" />");
					buf.append("\r\n");
					buf.append("<input type=\"hidden\" id=\"part\" name=\"part\" value=\"1\" />");
					buf.append("\r\n");
					buf.append("<textarea id=\"story\" name=\"story\" rows=\"25\" cols=\"80\" wrap=\"soft\" placeholder=\"Los, tipp was...\" maxlength=\"64000\" readonly=\"readonly\">@XXXXX</textarea>");
					buf.append("\r\n");
					buf.append("<br>");
					buf.append("\r\n");
					buf.append("<div>");
					for (int i = 1; i <= parts; i++) {
						if (i == 1) {
							buf.append("  <button type=\"submit\" id=\"save\" name=\"save\">Speichern</button>");
						} else if (i == parts + 1) {
							buf.append("  <button type=\"submit\" id=\"new\"  name=\"new\" >Neu</button>");
						} else {
							buf.append("  <button type=\"submit\" id=\"part" + i + "\" name=\"part" + i + "\">Abschnitt " + i + "</button>");
						}
					}
					buf.append("</div>");
					buf.append("\r\n");
					ret = form;
					ret = ret.replaceAll("<textarea>", denull(buf.toString()));
					ret = ret.replaceAll("readonly=\"readonly\"", denull(""));
					ret = ret.replaceAll("@NNNNN", denull(title));
					ret = ret.replaceAll("@IIIII", denull(vItem));
					ret = ret.replaceAll("@FFFFF", denull(vFor));
					ret = ret.replaceAll("@DDDDD", denull(vDemo));
					ret = ret.replaceAll("@XXXXX", denull(content));
				}
			}
		}
		return ret;
	}

	private String denull(String string) {
		return (string == null ? "" : string);
	}

	public int[] shuffle(final int year) {
		final int[] ret = new int[24];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = i + 1;
		}
		Random random = new Random(year);
		MathUtil.shuffle(ret, random);
		return ret;
	}

	public boolean isDate(final String key) {
		return ((null != key) && (4 < key.length()));
	}

	public int getYear(final String key) {
		int ret = DEFAULT_YEAR;
		if (null != key) {
			ret = Integer.parseInt(key.substring(0, 4));
		}
		return ret;
	}

	public int getDay(final String key) {
		int ret = 0;
		if (isDate(key)) {
			ret = Integer.parseInt(key.substring(key.length() - 2));
		}
		return ret;
	}

	/**
	 * Decides if this door is already open. Rules to follow:
	 * <ul>
	 *   <li>If the date to show is only a year without full date info, then the result is always <em>no</em>.</li>
	 *   <li>If <em>demoDateStr</em> is <code>null</code>, then the door is open if <em>dateToShowStr</em> is in the past or today.</li>
	 *   <li>If <em>demoDateStr</em> is not <code>null</code>, then the door is open if <em>dateToShowStr</em> is in the past or equal to <em>demoDateStr</em>.</li>
	 * </ul>
	 * 
	 * @param dateToShowStr
	 *            the date representation of the door to look for
	 * @param demoDateStr
	 *            the date representation of the demo mode (to test future
	 *            doors), or <code>null</code> for today
	 * @return if this door is already open
	 */
	public boolean isOpen(final String dateToShowStr, final String demoDateStr) {
		boolean ret = false;
		if ((null != dateToShowStr) && (isDate(dateToShowStr))) {
			Date demoDate;
			if (null == demoDateStr) {
				demoDate = new Date();
			} else {
				try {
					demoDate = DEFAULT_DATE_FORMAT.parse(demoDateStr);
				} catch (ParseException e) {
					demoDate = new Date();
				}
			}
			if (dateToShowStr.compareToIgnoreCase(DEFAULT_DATE_FORMAT.format(demoDate)) <= 0) {
				ret = true;
			}
		}
		return ret;
	}

	private int calcNumOfStories(final String vItem, final String vFor) {
		return 1;
	}

//	private DBA dbaForLog = null;
	private Connection getConnection() throws Exception{
		Connection conn = null;
		Exception exc = null;
		final DBA dba;
		if (isLocalDB) {
			dba = dbas[1];
		} else {
			dba = dbas[0];
		}
		try {
			conn = dba.getConnection();
//				dbaForLog = dba;
		} catch (SQLException e) {
			exc = e;
		} catch (InstantiationException e) {
			exc = e;
		} catch (IllegalAccessException e) {
			exc = e;
		} catch (ClassNotFoundException e) {
			exc = e;
		}
		if (null == conn) {
			throw exc;
		}
		return conn;
	}

	public String loadStory(final String vItem, final String vFor, final int vPart) {
		final StringBuffer sb = new StringBuffer();
		try { 
			final Connection conn = getConnection();
			final int day = getDay(vItem);
			final Statement st = conn.createStatement(); 
//			final ResultSet res = st.executeQuery("SELECT storyId, partId, subpartId, content FROM stories ORDER BY storyId, partId, subpartId ASC"); 
			final ResultSet res = st.executeQuery("SELECT content FROM stories WHERE storyId=0 AND partID=" + day + " AND subpartId=" + vPart); 
			while (res.next()) { 
				final String content = res.getString("content"); 
				sb.append(content);
//				sb.append("-->DBA=" + dbaForLog.url + "<--");
			} 
			conn.close(); 
		} catch (final Exception e) { 
			sb.append(e.toString()); 
		} 
		return sb.toString();
	}

	public String saveStory(final String vItem, final String vFor, final String vPart, final String content) {
		String ret = null;
		if ((null != content) && (!content.isEmpty())) {
			try { 
				final Connection conn = getConnection();
				final int day = getDay(vItem);
				final PreparedStatement prepped = conn.prepareStatement("INSERT into stories(storyId, partId, subpartId, content) VALUES (?, ?, ?, ?) ON DUPLICATE KEY UPDATE content=?;");
				prepped.setInt(1, 0);
				prepped.setInt(2, day);
				prepped.setInt(3, Integer.parseInt(vPart));
				prepped.setString(4, content);
				prepped.setString(5, content);
				prepped.executeUpdate();

//				final Statement st = conn.createStatement(); 
//				final int val = st.executeUpdate("INSERT into stories(storyId, partId, subpartId, content) VALUES (" 
//						+ 0 + "," + day + "," + vPart + "," + "'" + content + "'" + ") ON DUPLICATE KEY UPDATE content=" + "'" + content + "'" + ";");
//				if ((val != 1) && (val != 2)){
//					ret = "Updated " + val + " rows?!?";
//				}
				conn.close(); 
			} catch (final Exception e) { 
				ret = e.toString(); 
			} 
		}
		return ret;
	}

	@SuppressWarnings("unused")
	private String getFilename(final String vItem, final String vFor, final String vPart) {
		return vFor + vItem + "_" + vPart + ".txt";
	}

	public void setLocalDB(final String vLocal) {
		setLocalDB(PARAM_VAL_TRUE.equalsIgnoreCase(vLocal));
	}

	public void setLocalDB(final boolean isLocalDB) {
		this.isLocalDB = isLocalDB;
	}

	private boolean isLocalDB = false;
}
