package de.ped.atari8bit.logic;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Arrays;



import de.ped.tools.Assert;
import de.ped.tools.TextFormatter;

/**
 * Defines an Atari8bit disk image in ATR format.
 * 
 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ATRDiskImage extends DiskImage {
	/**
	 * This struct is from Nick Kennedy's readme.txt for SIO2PC.
	 * 
	 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class ATRHeader {
		public static final int ATR_SIGNATURE = 0x0296;

		public ATRHeader() {
			// Nothing to do
		}

		public final int getOffset() {
			return 0;
		}

		public final int getSize() {
			return 16;
		}

		@SuppressWarnings("unused")
		private byte getByte(final int idx) {
			return ByteArrayUtil.get8bit(rawdata, getOffset() + idx);
		}

		@SuppressWarnings("unused")
		private void setByte(final int idx, final byte value) {
			ByteArrayUtil.set8bit(rawdata, getOffset() + idx, value);
		}

		private int getWord(final int idx) {
			return ByteArrayUtil.get16bitLSF(rawdata, getOffset() + idx);
		}

		private void setWord(final int idx, final int value) {
			ByteArrayUtil.set16bitLSF(rawdata, getOffset() + idx, value);
		}

		@SuppressWarnings("unused")
		private int getDWord(final int idx) {
			return ByteArrayUtil.get32bitLSF(rawdata, getOffset() + idx);
		}

		@SuppressWarnings("unused")
		private void setDWord(final int idx, final int value) {
			ByteArrayUtil.set32bitLSF(rawdata, getOffset() + idx, value);
		}

		/**
		 * The content of the Atari indicator.
		 * 
		 * @return should always be {@link #ATR_SIGNATURE}
		 */
		public int getAtrNickAtari() {
			// WORD = special code indicating this is an Atari disk file.
			return getWord(0);
		}

		private void setAtrNickAtari(final int b) {
			setWord(0, b);
		}

		/** 
		 * @return the size of this disk image, in paragraphs (size/16)
		 */
		public int getParagraphCount() {
			// WORD = size of this disk image, in paragraphs (size/16),
			// WORD = high part of size, in paragraphs (added by REV 3.00)
			return (ByteArrayUtil.get16bitLSF(rawdata, getOffset() + 2)
					| (ByteArrayUtil.get8bit(rawdata, getOffset() + 7) << 16));
		}

		private void setParagraphCount(final int val) {
			// System.out.println("Setting ParagraphCount to " + val);
			ByteArrayUtil.set16bitLSF(rawdata, getOffset() + 2, (val & 0xffff));
			ByteArrayUtil.set8bit(rawdata, getOffset() + 7, (byte) ((val >>> 16) & 0xff));
		}

		/**
		 * @return the sector size in bytes (either 128 or 256)
		 */
		public int getSectorSize() {
			// WORD = sector size. (128 or 256) bytes/sector
			return ByteArrayUtil.get16bitLSF(rawdata, getOffset() + 4);
		}

		public void setSectorSize(final int val) {
			ByteArrayUtil.set16bitLSF(rawdata, getOffset() + 4, val);
		}

		/**
		 * @return the 1st (or typical) bad sector, or 0 if there is no bad sector on this disk
		 */
		public int getCrc32() {
			// WORD=1st (or typical) bad sector; see copy protection chapter.
			return ByteArrayUtil.get32bitLSF(rawdata, getOffset() + 8);
		}

		/**
		 * @param val
		 *            the 1st (or typical) bad sector, or 0 if there is no bad
		 *            sector on this disk (which is the normal case)
		 */
		public void setCrc32(final int val) {
			ByteArrayUtil.set32bitLSF(rawdata, getOffset() + 8, val);
		}

		/**
		 * @return a bitfield of disk flags such as copy protection and write
		 *         protect (containing the flags {@link DiskImage#D_ENHDENS},
		 *         {@link DiskImage#D_RESERVED}, {@link DiskImage#D_DBLDENS},
		 *         {@link DiskImage#D_MOTORON}, {@link DiskImage#D_WRTPROT}, ,
		 *         {@link DiskImage#D_WRTERR}, {@link DiskImage#D_BADDATA},
		 *         {@link DiskImage#D_BADCOMM})
		 */
		public int getStatus() {
			// BYTE = disk flags such as copy protection and write protect; see
			// copy protection chapter
			return ByteArrayUtil.get8bit(rawdata, getOffset() + 15);
		}

		/**
		 * @param val
		 *            a bitfield of disk flags such as copy protection and write
		 *            protect (containing the flags {@link DiskImage#D_ENHDENS},
		 *            {@link DiskImage#D_RESERVED}, {@link DiskImage#D_DBLDENS},
		 *            {@link DiskImage#D_MOTORON}, {@link DiskImage#D_WRTPROT},
		 *            , {@link DiskImage#D_WRTERR}, {@link DiskImage#D_BADDATA},
		 *            {@link DiskImage#D_BADCOMM})
		 */
		public void setStatus(final int val) {
			ByteArrayUtil.set8bit(rawdata, getOffset() + 15, (byte) (val & 0xff));
		}

		public int getSectorCount() {
			// see Nick Kennedy's readme.txt for SIO2PC
			final int ss = getSectorSize();
			Assert.assertTrue(ss >= 0);
			if (0 == ss) {
				return 3;
			} else {
				return (16 * getParagraphCount() - 3 * SIOConstants.SIO_SD_SECTOR_SIZE) / ss + 3;
			}
		}

		public void format(final int sectors, final int bytesPerSector) {
			setAtrNickAtari(ATR_SIGNATURE);
			setParagraphCount((rawdata.length - atrh.getSize()) / 16);
			setSectorSize(bytesPerSector);
			setCrc32(0);
			setStatus(1);
			Assert.assertTrue(isLegal());
			Assert.assertEquals(getSectorCount(), sectors);
			Assert.assertEquals(getSectorSize(), bytesPerSector);
			Assert.assertEquals(getSectorCount(), sectors);
		}

		public boolean isLegal() {
			if (getAtrNickAtari() != ATR_SIGNATURE) {
				System.out.println("ATR Signature should be: " + ATR_SIGNATURE + " but it is " + getAtrNickAtari());
				return false;
			}
			return true;
		}

		public String dump() {
			final StringBuffer buf = new StringBuffer();
			buf.append("Dump=");
			TextFormatter.hexdump(buf, rawdata, getOffset(), getSize());
			buf.append("\natrNickAtari=");
			buf.append(getAtrNickAtari());
			buf.append("\nParagraphCount=");
			buf.append(getParagraphCount());
			buf.append("\nSectorSize=");
			buf.append(getSectorSize());
			buf.append("\nCrc32=");
			buf.append(getCrc32());
			buf.append("\nStatus=");
			buf.append(getStatus());
			buf.append("\n");
			return buf.toString();
		}

		/*
		WORD atrNickAtari;     // 0x0296
		WORD paragraphCountLo;
		WORD sectorSize;
		BYTE paragraphCountHi; // v3.00
		DWORD crc32;           // APE extension
		BYTE unused[3];
		BYTE status;           // APE extension : bit 0 = write protect, bit 1 = sealed/authenticated
		 */
	}

	ATRHeader atrh = new ATRHeader();

	public ATRHeader getATRHeader() {
		return atrh;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(rawdata);
		return result;
	}

	/**
	 * @param obj
	 *            the object to compare with
	 * @return <code>true</code> if both objects are of same class and the raw data (disk size and content) is bitwise identical
	 */
	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final ATRDiskImage other = (ATRDiskImage) obj;
		if (!Arrays.equals(rawdata, other.rawdata))
			return false;
		return true;
	}

	@Override
	public void load(final InputStream is) {
		final ByteArrayOutputStream os = new ByteArrayOutputStream();
		int c;
		try {
			while ((c = is.read()) != -1) {
				os.write(c);
			}
			rawdata = os.toByteArray();
			os.close();
		} catch (final IOException e) {
			System.out.println(e);
			Assert.assertTrue(false);
		}
		Assert.assertTrue(atrh.isLegal());
	}

	@Override
	public void save(final OutputStream os) {
		try {
			for (int i = 0; i < rawdata.length; i++) {
				os.write(rawdata[i]);
			}
		} catch (final IOException e) {
			System.out.println(e);
			Assert.assertTrue(false);
		}
	}

	@Override
	public int getNumOfSectors() {
		return atrh.getSectorCount();
	}

	@Override
	public int getSectorSize() {
		return atrh.getSectorSize();
	}

	/**
	 * @param sector the sector number (a positive number)
	 * @return the position of the first byte of sector <em>sector</em> within this disk image
	 */
	protected int getSectorOffset(final int sector) {
		Assert.assertTrue(sector > 0);
		int off = atrh.getSize();
		if (sector < 4) {
			off += (sector - 1) * SIOConstants.SIO_SD_SECTOR_SIZE;
		} else {
			// berspringe die ersten Sektoren konstanter Gre
			off += 3 * SIOConstants.SIO_SD_SECTOR_SIZE; 
			// Und nun alle anderen Sektoren
			off += (sector - 4) * atrh.getSectorSize(); 
		}
		return off;
	}

	@Override
	public byte[] getSector(final int sector) {
		final int off = getSectorOffset(sector);
		final int size = getSectorSize(sector);
		final byte[] ret = new byte[size];
		ByteArrayUtil.copyPart(ret, 0, rawdata, off, ret.length);
		return ret;
	}

	@Override
	public void format(final int sectors, final int bytesPerSector) {
		int sizeInBytes = atrh.getSize();
		for (int i = 1; i <= sectors; i++) {
			sizeInBytes += getSectorSize(i, bytesPerSector);
		}
		rawdata = new byte[sizeInBytes];
		atrh.format(sectors, bytesPerSector);
	}

	@Override
	public void setSector(final int sector, final byte[] arr) {
		Assert.assertEquals(arr.length, getSectorSize(sector));
		Assert.assertTrue(sector > 0);
		Assert.assertTrue(sector <= atrh.getSectorCount());
		final int offset = getSectorOffset(sector);
		final int max = getSectorSize(sector);
		ByteArrayUtil.copyPart(rawdata, offset, arr, 0, max);
	}

	private boolean isSectorEmpty(final int sectnum) {
		final byte[] arr = getSector(sectnum);
		for (int i = 0; i < arr.length; i++) {
			if (arr[i] != 0) {
				return false;
			}
		}
		return true;
	}

	public void shorten() {
		int sectnum = getNumOfSectors();
		while (isSectorEmpty(sectnum)) {
			sectnum--;
		}
		// Now sectnum contains the last non-empty sector.
		if (sectnum == getNumOfSectors()) {
			return; // Nothing to shorten
		}
		final ATRDiskImage copy = new ATRDiskImage();
		copy.format(sectnum, getSectorSize());
		ByteArrayUtil.copyPart(copy.rawdata, 0, rawdata, 0, copy.rawdata.length);
		copy.atrh.format(sectnum, getSectorSize());
		this.rawdata = copy.rawdata;
	}

	public byte[] getRawdata() {
		return rawdata;
	}

	private byte[] rawdata;
}
