package de.ped.atari8bit.logic;

/**
 * Helps to deal with byte alignment when writing 16-bit and 32-bit values into a byte array.
 * 
 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ByteArrayUtil {

	/**
	 * Reads an 8-bit value.
	 * @param arr the array to read from
	 * @param idx the index within the array to start reading
	 * @return the decoded value
	 */
	public static byte get8bit(final byte[] arr, final int idx) {
		return arr[idx];
	}

	public static void set8bit(final byte[] arr, final int idx, final byte value) {
		arr[idx] = value;
	}

	public static void set8bit(final byte[] arr, final int idx, final int value) {
		arr[idx] = (byte) (value & 0xff);
	}

	/**
	 * Reads a 16-bit value, least signifant byte first.
	 * @param arr the array to read from
	 * @param idx the index within the array to start reading
	 * @return the decoded value
	 */
	public static int get16bitLSF(final byte[] arr, final int idx) {
		return (((arr[idx + 0] & 0xff) << 0) | ((arr[idx + 1] & 0xff) << 8));
	}

	/**
	 * Reads a 16-bit value, most signifant byte first.
	 * @param arr the array to read from
	 * @param idx the index within the array to start reading
	 * @return the decoded value
	 */
	public static int get16bitMSF(final byte[] arr, final int idx) {
		return (((arr[idx + 0] & 0xff) << 8) | ((arr[idx + 1] & 0xff) << 0));
	}

	public static void set16bitLSF(final byte[] arr, final int idx, final int value) {
		arr[idx] = (byte) ((value >>> 0) & 0xff);
		arr[idx + 1] = (byte) ((value >>> 8) & 0xff);
	}

	public static void set16bitMSF(final byte[] arr, final int idx, final int value) {
		arr[idx] = (byte) ((value >>> 8) & 0xff);
		arr[idx + 1] = (byte) ((value >>> 0) & 0xff);
	}

	/**
	 * Reads a 32-bit value, least signifant byte first.
	 * @param arr the array to read from
	 * @param idx the index within the array to start reading
	 * @return the decoded value
	 */
	public static int get32bitLSF(final byte[] arr, final int idx) {
		return ((get16bitLSF(arr, idx + 0)) | (get16bitLSF(arr, idx + 2) << 16));
	}

	/**
	 * Reads a 32-bit value, most signifant byte first.
	 * @param arr the array to read from
	 * @param idx the index within the array to start reading
	 * @return the decoded value
	 */
	public static int get32bitMSF(final byte[] arr, final int idx) {
		return ((get16bitMSF(arr, idx + 2)) | (get16bitMSF(arr, idx + 0) << 16));
	}

	public static void set32bitLSF(final byte[] arr, final int idx, final int value) {
		arr[idx + 0] = (byte) ((value >>> 0) & 0xff);
		arr[idx + 1] = (byte) ((value >>> 8) & 0xff);
		arr[idx + 2] = (byte) ((value >>> 16) & 0xff);
		arr[idx + 3] = (byte) ((value >>> 24) & 0xff);
	}

	public static void set32bitMSF(final byte[] arr, final int idx, final int value) {
		arr[idx + 0] = (byte) ((value >>> 24) & 0xff);
		arr[idx + 1] = (byte) ((value >>> 16) & 0xff);
		arr[idx + 2] = (byte) ((value >>> 8) & 0xff);
		arr[idx + 3] = (byte) ((value >>> 0) & 0xff);
	}

	public static void copyPart(final byte[] dest, final int destStartIndex, final byte[] source,
			final int sourceStartIndex, final int size) {
		for (int i = 0; i < size; i++) {
			dest[destStartIndex + i] = source[sourceStartIndex + i];
		}
	}
}
