package de.ped.atari8bit.logic;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.LinkedList;

import de.ped.tools.Assert;
import de.ped.tools.log.Logger;

/**
 * Defines an Atari8bit disk image.
 * 
 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class DiskImage {

	protected transient Logger logger = Logger.getLogger(getClass());

	/**
	 * What kinds of data are there a sector can hold.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum SectorType {
		/** Boot sector */
		BOOT,
		/** Sector containing the file allocation table */
		FAT,
		/** Sector containing the root directory */
		DIR,
		/** Some buffer sector */
		BUFFER,
		/** Normal data sector */
		DATA;
	}

	/**
	 * What kinds of ATARI DOSes are supported, defining how the disk content is organized.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum DOS {
		/** DOS 2.0 and DOS 2.5 by Atari */
		DOS2x,
		/** DOS 3.0 by Atari */
		DOS3x,
		/** Percom DOS and others */
		PERCOM;
	}

	/** Device status bit for Enhanced density */
	public static final int D_ENHDENS = 0x80;

	/** Device status bit for Reserved bit */
	public static final int D_RESERVED = 0x40;

	/** Device status bit for Double density */
	public static final int D_DBLDENS  = 0x20;

	/** Device status bit for Motor on */
	public static final int D_MOTORON  = 0x10;

	/** Device status bit for Write protection */
	public static final int D_WRTPROT  = 0x08;

	/** Device status bit for Write error */
	public static final int D_WRTERR   = 0x04;

	/** Device status bit for Bad data frame */
	public static final int D_BADDATA  = 0x02;

	/** Device status bit for Bad command frame */
	public static final int D_BADCOMM  = 0x01;

	/**
	 * Internal data structure for DOS2.x to know which parts of FAT sectors keep sector availability information.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class FATSectorData {
		public final int sectnum;
		public final int offset;
		public final int firstSector;
		public final int lastSector;
		public final int posOfNumOfFreeSectsPrecalculated;
		public final int firstSectorForPrecalculated;
		public final int lastSectorForPrecalculated;
		public FATSectorData(
				int sectnum, 
				int offset, 
				int firstSector, 
				int lastSector, 
				int posOfNumOfFreeSectsPrecalculated,
				int firstSectorForPrecalculated,
				int lastSectorForPrecalculated) {
			this.sectnum = sectnum;
			this.offset = offset;
			this.firstSector = firstSector;
			this.lastSector = lastSector;
			this.posOfNumOfFreeSectsPrecalculated = posOfNumOfFreeSectsPrecalculated;
			this.firstSectorForPrecalculated = firstSectorForPrecalculated;
			this.lastSectorForPrecalculated = lastSectorForPrecalculated;
		}
	}

	public static final FATSectorData SECTORDATA_FAT_LOW = new FATSectorData(
			0x168, 
			0x0a, 0x00, SIOConstants.SIO_SSSD_SECTOR_NUM, 
			0x03, 0x00, SIOConstants.SIO_SSSD_SECTOR_NUM);
	public static final FATSectorData SECTORDATA_FAT_HIGH = new FATSectorData(
			0x400, 
			0x00, 0x30, 0x400, 
			0x7a, SIOConstants.SIO_SSSD_SECTOR_NUM, 0x400);

	/**
	 * @param sector
	 *            the sector number (a positive number)
	 * @return the sector type for this sector, assuming that DOS2x or PERCOM is
	 *         the DOS is use
	 */
	public SectorType getSectorType(final int sector) {
		SectorType ret = SectorType.DATA;
		if ((0x000 < sector) && (sector < 0x004)) {	
			ret = SectorType.BOOT;
		} else if ((sector == SECTORDATA_FAT_LOW.sectnum) || (sector == SECTORDATA_FAT_HIGH.sectnum)) {
			ret = SectorType.FAT;
		} else if ((sector > 0x168) && (sector <= 0x170)) {
			ret = SectorType.DIR;
		} else if (sector == 0x000) {
			ret = SectorType.BUFFER;
		}
		return ret;
	}

	public boolean isWriteProtected() {
		return false;
	}

	public static Density density(final int sectors, final int bytesPerSector) {
		final Density density;
		if (bytesPerSector == 0x100) {
			density = Density.DOUBLEDENS;
		} else if (sectors >= SIOConstants.SIO_SSED_SECTOR_NUM) {
			density = Density.ENHANCDENS;
		} else { // single density
			density = Density.SINGLEDENS;
		}
		return density;
	}

	public Density getDensity() {
		return density(getNumOfSectors(), getSectorSize());
	}

	public void setWriteProtected(final boolean isWriteProtected) {
		// Default implementation does nothing
	}

	private boolean isFreeFATEntry(final int sector, final FATSectorData fat) {
		boolean ret = false;
		if ((fat.firstSector <= sector) && (sector < fat.lastSector)) {
			final byte[] data = getSector(fat.sectnum);
			final int offset = fat.offset + ((sector - fat.firstSector) >> 3);
			final int mask= (0x80 >> (sector & 0x07));
			if (0 != (data[offset] & mask)) {
				ret = true;
			}
			logger.trace("Searching FAT 0x" + Integer.toHexString(fat.sectnum) + " for sector " + sector + ": offset=" + offset + ", byte=" + Integer.toHexString(data[offset] & 0xff) + ", mask=" + Integer.toHexString(mask) + " was " + (ret ? "used" : "free"));
		} else {
			throw new IllegalArgumentException("FAT entry out of range: FAT " + Integer.toHexString(fat.sectnum) + ", sector=" + sector);
		}
		return ret;
	}

	private void setFreeFATEntry(final int sector, final FATSectorData fat, final boolean isFree) {
		final int mask= (0x80 >> (sector & 0x07));
		if ((fat.firstSector <= sector) && (sector < fat.firstSector + fat.lastSector)) {
			final byte[] data = getSector(fat.sectnum);
			final int offset = fat.offset + ((sector - fat.firstSector) >> 3);
			if (isFree) {
				data[offset] |= mask;
			} else {
				data[offset] &= ~mask;
			}
			setSector(fat.sectnum, data);
		}
	}

	public boolean isFreeFATEntry(final int sector) {
		boolean ret = false;
		boolean retLow = false;
		boolean isLowFAT = false;
		boolean retHigh = false;
		boolean isHighFAT = false;
		try {
			retLow = isFreeFATEntry(sector, SECTORDATA_FAT_LOW);
			isLowFAT = true;
		} catch (final IllegalArgumentException ex) {
			// Nothing to do
		}
		try {
			retHigh = isFreeFATEntry(sector, SECTORDATA_FAT_HIGH);
			isHighFAT = true;
		} catch (final IllegalArgumentException ex) {
			// Nothing to do
		} catch (final ArrayIndexOutOfBoundsException ex) {
			// Nothing to do - this disk is not enhanced density
		}
		if (isHighFAT && isLowFAT && (retLow != retHigh)) {
			throw new IllegalStateException("FAT twist!");
		}
		if (isLowFAT) {
			ret = retLow;
		} else {
			ret = retHigh;
		}
		return ret;
	}

	public void setFreeFATEntry(final int sector, final boolean isFree) {
		setFreeFATEntry(sector, SECTORDATA_FAT_LOW, isFree);
		setFreeFATEntry(sector, SECTORDATA_FAT_HIGH, isFree);
	}

	/**
	 * Searches the FAT for the next free sector, marks it as used, and returns
	 * the number of that sector.
	 * 
	 * @param currentSector
	 *            the sector number (a positive number) where to start searching
	 * @return the sector number of the next free sector
	 */
	private int reserveFreeFATSector(final int currentSector) {
		int ret = -1;
		int sectnum = currentSector;
		final int numOfSectors = getNumOfSectors();
		for (int i = 0; (ret < 0) && (i < numOfSectors); i++) {
			boolean isFreeFatEntry = false;
			isFreeFatEntry = isFreeFATEntry(sectnum);
			if (isFreeFatEntry) {
				setFreeFATEntry(sectnum, false);
				ret = sectnum;
    			logger.trace("Reserved FAT entry " + ret);
			} else {
				sectnum++;
				if (numOfSectors < sectnum) {
					sectnum = 0;
				}
			}
		}
		return ret;
	}

	public int getNumOfFreeSectorsFromFATPrecalculated() {
		final int retLow = ByteArrayUtil.get16bitLSF(getSector(SECTORDATA_FAT_LOW.sectnum), SECTORDATA_FAT_LOW.posOfNumOfFreeSectsPrecalculated);
		int retHigh = 0;
		try {
			retHigh = ByteArrayUtil.get16bitLSF(getSector(SECTORDATA_FAT_HIGH.sectnum), SECTORDATA_FAT_HIGH.posOfNumOfFreeSectsPrecalculated);
		} catch (final ArrayIndexOutOfBoundsException ex) {
			// Nothing to do - this disk is not enhanced density
		}
		return retHigh + retLow;
	}

	private void refreshNumOfFreeSectorsFromFATPrecalculated(final FATSectorData fat) {
		int freeSectsCounter = 0;
		final byte[] data = getSector(fat.sectnum);
		for (int i = fat.firstSectorForPrecalculated; i < fat.lastSectorForPrecalculated; i++) {
			if (isFreeFATEntry(i, fat)) {
				freeSectsCounter++;
			}
		}
		ByteArrayUtil.set16bitLSF(data, fat.posOfNumOfFreeSectsPrecalculated, freeSectsCounter);
		setSector(fat.sectnum, data);
	}

	private void refreshNumOfFreeSectorsFromFATPrecalculated() {
		refreshNumOfFreeSectorsFromFATPrecalculated(SECTORDATA_FAT_LOW);
		try {
			refreshNumOfFreeSectorsFromFATPrecalculated(SECTORDATA_FAT_HIGH);
		} catch (final ArrayIndexOutOfBoundsException ex) {
			// Nothing to do - this disk is not enhanced density
		}
	}

	public int getNumOfFreeSectorsFromFAT() {
		int ret = 0;
		final int numOfSectors = getNumOfSectors();
		for (int i = 0; (i < numOfSectors); i++) {
			try {
				if (isFreeFATEntry(i)) {
					ret++;
				}
			} catch (final IllegalArgumentException ex) {
				// Nothing to do
			}
		}
		return ret;
	}

	public int getNumOfFreeSectorsForEmptyDisk() {
		final byte[] fat = getSector(SECTORDATA_FAT_LOW.sectnum);
		return ByteArrayUtil.get16bitLSF(fat, 0x01);
	}

	/**
	 * A directory entry on that disk. Works only for DOS2x and PERCOM.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class DirectoryEntry {
		public static final int FIRST_DIRECTORY_SECTOR = 0x169;
		public static final int ENTRY_SIZE = 0x10;
		public static final int SECTOR_SIZE = 0x80;
		public static final int ENTRIES_PER_SECTOR = SECTOR_SIZE / ENTRY_SIZE;
		public static final int ENTRIES = 0x40;

		public static final int STATUS_BITFIELD_DELETED = 0x80;
		public static final int STATUS_BITFIELD_WRITEPROTECT = 0x20;
		public static final int STATUS_NORMALFILE = 0x42;

		private final int sectorIndex;
		private final int sectorOffset;
		public final int index;
		private final byte[] rawData = new byte[ENTRY_SIZE];

		public String nameFromRaw(final String rawName) {
			final StringBuffer s = new StringBuffer();
			s.append(rawName.substring(0, 8).trim());
			s.append('.');
			s.append(rawName.substring(8).trim());
			return s.toString();
		}

		public String nameToRaw(final String name) {
			final StringBuffer s = new StringBuffer(11);
			int dotpos = name.lastIndexOf(".");
			if (dotpos < 0) {
				dotpos = name.length();
			}
			for (int i = 0; i < 8; i++) {
				char c = ' ';
				if (i < dotpos) {
					c = name.charAt(i);
				}
				s.append(c);
			}
			for (int i = 0; i < 3; i++) {
				char c = ' ';
				final int pos = dotpos + 1 + i;
				if (pos < name.length()) {
					c = name.charAt(pos);
				}
				s.append(c);
			}
			return s.toString();
		}

		public DirectoryEntry(final int index) {
			this.index = index;
			if ((index < 0) || (ENTRIES <= index)) {
				throw new IllegalArgumentException("Index " + index + " out of range [0:" + (ENTRIES - 1) + "]");
			}
			sectorIndex = FIRST_DIRECTORY_SECTOR + (index / ENTRIES_PER_SECTOR);
			sectorOffset = ENTRY_SIZE * (index % ENTRIES_PER_SECTOR);
			final byte[] sector = getSector(sectorIndex);
			ByteArrayUtil.copyPart(rawData, 0, sector, sectorOffset, ENTRY_SIZE);
		}

		public byte[] getRawData() {
			return rawData;
		}

		public boolean isRealEntry() {
			boolean ret = false;
			for (int i = 0; i < rawData.length; i++) {
				if (rawData[i] != 0) {
					ret = true;
				}
			}
			return ret;
		}

		protected void writeRawData() {
			final byte[] sector = getSector(sectorIndex);
			ByteArrayUtil.copyPart(sector, sectorOffset, rawData, 0, ENTRY_SIZE);
			setSector(sectorIndex, sector);
		}

		public String getRawName() {
			final StringBuffer s = new StringBuffer(11);
			for (int i = 0; i < 11; i++) {
				final byte b = rawData[5 + i];
				s.append((char) (b));
			}
			return s.toString();
		}

		public void setRawName(final String rawName) {
			for (int i = 0; i < 11; i++) {
				rawData[5 + i] = (byte) (rawName.charAt(i));
			}
		}

		public String getName() {
			return nameFromRaw(getRawName());
		}

		public void setName(final String name) {
			setRawName(nameToRaw(name));
		}

		public byte getStatus() {
			return ByteArrayUtil.get8bit(rawData, 0);
		}

		public void setStatus(final byte status) {
			ByteArrayUtil.set8bit(rawData, 0, status);
			writeRawData();
		}

		public int getLengthInSectors() {
			return ByteArrayUtil.get16bitLSF(rawData, 1);
		}

		public void setLengthInSectors(final int length) {
			ByteArrayUtil.set16bitLSF(rawData, 1, length);
			writeRawData();
		}

		public int getFirstSectorIndex() {
			return ByteArrayUtil.get16bitLSF(rawData, 3);
		}

		public void setFirstSectorIndex(final int firstSector) {
			ByteArrayUtil.set16bitLSF(rawData, 3, firstSector);
			writeRawData();
		}

		public String dump() {
			return "DirEntry " + index + " " + getName() + " Status=" + Integer.toHexString(getStatus()) + " First="
					+ getFirstSectorIndex() + " Len=" + getLengthInSectors();
		}
	}

	/**
	 * A data sector on this disk.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class DataSector {
		public final int sectorIndex;
		private byte[] rawData = null;

		public int getMaxDataLength() {
			return getSectorSize(sectorIndex) - 3;
		}

		public DataSector(final int sectorIndex) {
			this.sectorIndex = sectorIndex;
			final SectorType sectorType = getSectorType(sectorIndex);
			if (SectorType.DATA != sectorType) {
				throw new IllegalArgumentException("Wrong sector type " + sectorType.name() + " for sector " + sectorIndex + " - expected type " + SectorType.DATA.name());
			}
			this.rawData = getSector(sectorIndex);
		}

		protected void writeRawData() {
			setSector(sectorIndex, rawData);
		}

		protected int getDataLength() {
			final int ret = ByteArrayUtil.get8bit(rawData, rawData.length - 1) & 0xff;
			logger.trace("Sector " + sectorIndex + " data length = " + ret);
			return ret;
		}

		protected void setDataLength(final int length) {
			ByteArrayUtil.set8bit(rawData, rawData.length - 1, Math.min(length, getMaxDataLength()));
			writeRawData();
		}

		public byte[] getData() {
			byte[] ret = new byte[getDataLength()];
			ByteArrayUtil.copyPart(ret, 0, rawData, 0, ret.length);
			return ret;
		}

		public void setData(final byte[] data) {
			ByteArrayUtil.copyPart(rawData, 0, data, 0, Math.min(data.length, getMaxDataLength()));
			writeRawData();
		}

		public int getNextSectorIndex() {
			return ByteArrayUtil.get16bitMSF(rawData, rawData.length - 3) & 0x03ff;
		}

		public void setNextSectorIndex(final int nextSectorIndex) {
			ByteArrayUtil.set16bitMSF(rawData, rawData.length - 3, nextSectorIndex);
			writeRawData();
		}

		public int getDirEntryIndex() {
			return ByteArrayUtil.get8bit(rawData, rawData.length - 3) >>> 2;
		}

		public void setDirEntryIndex(final int index) {
			int b = ByteArrayUtil.get8bit(rawData, rawData.length - 3) & 0xff;
			b = (b & 0x03) | ((b & 0x3f) << 2);
			ByteArrayUtil.set8bit(rawData, rawData.length - 3, b);
			writeRawData();
		}

		public String dump() {
			return "DataSector " + sectorIndex + ": DataLen=" + getDataLength() + ", Entry=" + getDirEntryIndex()
					+ ", Next=" + getNextSectorIndex();
		}
	}

	/**
	 * @param index
	 *            the directory entry index
	 * @return the directory entry with the given index
	 */
	public DirectoryEntry getDirEntry(final int index) {
		return new DirectoryEntry(index);
	}

	/**
	 * @param filename
	 *            the filename (not raw file name) to find within the directory
	 * @return the directory entry with the given file name
	 */
	public DirectoryEntry getDirEntry(final String filename) {
		DirectoryEntry ret = null;
		for (int i = 0; (null == ret) && (i < DirectoryEntry.ENTRIES); i++) {
			DirectoryEntry entry = getDirEntry(i);
			String name = entry.getName();
			if (name.equals(filename)) {
				ret = entry;
			}
		}
		return ret;
	}

	/**
	 * @return the first empty directory entry
	 */
	private DirectoryEntry getEmptyDirEntry() {
		DirectoryEntry ret = null;
		for (int i = 0; (null == ret) && (i < DirectoryEntry.ENTRIES); i++) {
			final DirectoryEntry entry = getDirEntry(i);
			final byte stat = entry.getStatus();
			if ((0 == stat) || (false)) {
				ret = entry;
			}
		}
		return ret;
	}

	private static final DirectoryEntry[] template = new DirectoryEntry[0];

	public DirectoryEntry[] getDirEntries() {
		final LinkedList<DirectoryEntry> entries = new LinkedList<DirectoryEntry>();
		for (int i = 0; i < DirectoryEntry.ENTRIES; i++) {
			final DirectoryEntry entry = getDirEntry(i);
			if (entry.isRealEntry()) {
				entries.add(entry);
			} else {
				break;
			}
		}
		return entries.toArray(template);
	}

	/**
	 * Loads a disk from an input stream.
	 * 
	 * @param is
	 *            the input stream from where to load
	 */
	public abstract void load(final InputStream is);

	/**
	 * Loads a disk from a file.
	 * 
	 * @param filename
	 *            the (path and) name of the file from where to load
	 */
	public void load(final String filename) {
		try {
			final FileInputStream is = new FileInputStream(filename);
			load(is);
			is.close();
		} catch (final FileNotFoundException e) {
			System.out.println(e);
			Assert.assertTrue(false);
		} catch (final IOException e) {
			System.out.println(e);
			Assert.assertTrue(false);
		}
	}

	/**
	 * Saves a disk to an output stream.
	 * 
	 * @param os
	 *            the output stream to write this object into
	 */
	public abstract void save(OutputStream os);

	/**
	 * Saves a disk to a file.
	 * 
	 * @param filename
	 *            the (path and) name of the file where to write this object
	 *            into
	 */
	public void save(String filename) {
		try {
			final FileOutputStream os = new FileOutputStream(filename);
			try {
				save(os);
				os.close();
			} catch (final IOException e) {
				System.out.println(e);
				Assert.assertTrue(false);
			}
		} catch (final FileNotFoundException e) {
			System.out.println(e);
			Assert.assertTrue(false);
		}
	}

	/** 
	 * @return the size (in bytes) of normal sectors for this disk (all sectors but 1-3)
	 */
	public abstract int getSectorSize();

	/**
	 * Returns the size (in bytes) of sector <em>sectnum</em>, if the normal
	 * sector size is <em>defaultSectorSize</em>.
	 * 
	 * @param sector
	 *            the sector number (a positive number)
	 * @param defaultSectorSize
	 *            the default sector size
	 * @return the size (in bytes) of the given sector
	 */
	public static int getSectorSize(final int sector, final int defaultSectorSize) {
		Assert.assertTrue(sector > 0);
		return ((sector <= 3) ? SIOConstants.SIO_SD_SECTOR_SIZE : defaultSectorSize);
	}

	/**
	 * @param sector
	 *            the sector number (a positive number)
	 * @return the size (in bytes) of the given sector
	 */
	public int getSectorSize(final int sector) {
		return getSectorSize(sector, getSectorSize());
	}

	/**
	 * @param sector
	 *            the sector number (a positive number)
	 * @return the content of the sector
	 */
	public abstract byte[] getSector(final int sector);

	/** 
	 * @return the number of sectors
	 */
	public abstract int getNumOfSectors();

	/**
	 * Not all physically accessable sectors are used in DOS 2.x enhanced mode,
	 * and FAT knows this.
	 * 
	 * @return the number of sectors FAT knows about
	 */
	public int getNumOfSectorsForFat() {
		return Math.min(getNumOfSectors(), 0x400);
	}

	/**
	 * Does a hard formatting of the disk. All old content is lost. The disk is
	 * formatted according to the parameters.
	 * 
	 * @param sectors
	 *            the number of sectors the newly formatted disk will contain
	 * @param bytesPerSector
	 *            the bytes per sector the newly formatted disk will contain
	 */
	public abstract void format(final int sectors, final int bytesPerSector);

	public void format(final Density density) {
		format(density.sectors, density.bytesPerSector);
	}

	/**
	 * Writes the sector <em>sectnum</em>.
	 * 
	 * @param sector
	 *            the sector number (a positive number)
	 * @param arr
	 *            the new content of the sector
	 */
	public abstract void setSector(final int sector, byte[] arr);

	/**
	 * Initializes an empty formatted disk with all the data structures necesary
	 * for the given DOS.
	 * 
	 * @param dos
	 *            the DOS to initialize for
	 */
	public void initDiskForDos(final DOS dos) {
		if (getSectorSize() > 0x80) {
			throw new IllegalArgumentException("Double density disks currently not supported");
		}
		for (PredefinedSector sect : PredefinedSectors.instance().sectors(dos, getDensity())) {
			setSector(sect.sector, sect.content());
		}
	}

	/**
	 * Compares two disks. Two disks are identical if they have the same sector
	 * size, the same number of sectors, and all sectors are bit identical.
	 * 
	 * @param other
	 *            the object to compare with
	 * @return <code>true</code> if the disks are identical
	 */
	public boolean equalsTo(final Object other) {
		if (other instanceof DiskImage) {
			DiskImage odi = (DiskImage) other;
			if (getSectorSize() != odi.getSectorSize()) {
				System.out.println("Different SectorSizes! " + getSectorSize() + " != " + odi.getSectorSize());
				return false;
			}
			if (getNumOfSectors() != odi.getNumOfSectors()) {
				System.out.println("Different NumOfSectors! " + getNumOfSectors() + " != " + odi.getNumOfSectors());
				return false;
			}
			for (int s = 1; s <= getNumOfSectors(); s++) {
				final byte[] sarr = getSector(s);
				final byte[] darr = odi.getSector(s);
				Assert.assertEquals(sarr.length, darr.length);
				for (int i = 0; i < sarr.length; i++) {
					if (sarr[i] != darr[i]) {
						System.out.println("Different Content of sector " + s + " at pos " + i + ": " + sarr[i] + " != "
								+ darr[i]);
						return false;
					}
				}
			}
			return true;
		}
		return false;
	}

	/**
	 * Copies the content of one disk to another. There may be a loss of
	 * information as it does a format change if necessary.
	 * 
	 * @param source
	 *            the disk to copy from
	 */
	public void copyFrom(final DiskImage source) {
		format(source.getNumOfSectors(), source.getSectorSize());
		for (int i = 1; i <= source.getNumOfSectors(); i++) {
			setSector(i, source.getSector(i));
		}
	}

	public void importFile(final String xlfilename, final InputStream is) throws IOException {
		if (isWriteProtected()) {
			throw new IllegalStateException("Disk is write protected");
		}
		if (null != getDirEntry(xlfilename)) {
			throw new IllegalStateException("File already exists");
		}
		final DirectoryEntry entry = getEmptyDirEntry();
		if (null == entry) {
			throw new IllegalStateException("Disk full (maximum number of files reached)");
		}
		entry.setName(xlfilename);
		entry.setStatus((byte) DirectoryEntry.STATUS_NORMALFILE);
		DataSector lastSector = null;
		int sectorsWrittenCounter = 0;
		int readBytes = 0;
		boolean isDone = false;
		do {
			final byte[] arr = new byte[125]; // TODO: This does not work for double density
			readBytes = is.read(arr);
			if (readBytes <= 0) {
				isDone = true;
			} else {
				final int currentSectornum = reserveFreeFATSector(0);
				if (currentSectornum <= 0) {
					isDone = true;
				} else {
					final DataSector currentSector = new DataSector(currentSectornum);
					if (null == lastSector) {
						entry.setFirstSectorIndex(currentSectornum);
					} else {
						lastSector.setNextSectorIndex(currentSectornum);
					}
					currentSector.setData(arr);
					currentSector.setDataLength(readBytes);
					currentSector.setDirEntryIndex(entry.index);
					lastSector = currentSector;
					sectorsWrittenCounter++;
				}
			}
		} while (!isDone);
		entry.setLengthInSectors(sectorsWrittenCounter);
		refreshNumOfFreeSectorsFromFATPrecalculated();
		is.close();
	}

	public void exportFile(final int directoryIndex, final OutputStream os) throws IOException {
		final DirectoryEntry entry = getDirEntry(directoryIndex);
		int sectnum = entry.getFirstSectorIndex();
		while (0 < sectnum) {
			final DataSector sect = new DataSector(sectnum);
			logger.trace(sect.dump());
			os.write(sect.getData());
			sectnum = sect.getNextSectorIndex();
		}
		os.close();
	}

	public void exportFiles(final String path) throws IOException {
		final ATRDiskImage.DirectoryEntry[] dir = getDirEntries();
		for (final ATRDiskImage.DirectoryEntry e : dir) {
			final String filename = path + File.separator + e.getName();
			final FileOutputStream fos = new FileOutputStream(filename);
			exportFile(e.index, fos);
		}
	}

	public boolean containsSameData(final DiskImage other) {
		if (other.getNumOfSectors() != getNumOfSectors()) {
			return false;
		}
		for (int i = 1; i < getNumOfSectors(); i++) {
			if (!Arrays.equals(getSector(i), other.getSector(i))) {
				return false;
			}
		}
		return true;
	}
}
