package de.ped.atari8bit.logic;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.util.Arrays;

import de.ped.tools.Assert;
import de.ped.tools.AssertionException;
import de.ped.tools.TextFormatter;

/**
 * Defines an Atari8bit disk image in XLST format.
 * 
 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class XLSTDiskImage extends DiskImage {
	/**
	 * The block defining basic disk properties.
	 * 
	 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class PercomBlock {
		/** Hardware status bit for Door open */
		public static final int F_DOOROPEN = 0x80;
		/** Hardware status bit for INV: Write protection */
		public static final int FI_WRTPROT = 0x40;
		/** Hardware status bit for INV: Record Type */
		public static final int FI_RCDTYPE = 0x20;
		/** Hardware status bit for INV: RNF Error */
		public static final int FI_RNFERROR = 0x10;
		/** Hardware status bit for INV: CRC Error */
		public static final int FI_CRCERROR = 0x08;
		/** Hardware status bit for INV: Data Lost */
		public static final int FI_DATALOST = 0x04;
		/** Hardware status bit for INV: Data Request */
		public static final int FI_DATARQST = 0x02;
		/** Hardware status bit for INV: FDC busy */
		public static final int FI_FDCBUSY = 0x01;

		public final int getOffset() {
			return 10;
		}

		public final int getSize() {
			return 10 * 2 + 2 * 1;
		}
		// 10 uint, 2 ubyte ==> 10*2 + 2 = 24 byte (aligned)

		public int getSectorLen() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 0);
		}

		public void setSectorLen(final int sectorLen) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 0, sectorLen);
		}

		public int getNumOfTracks() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 2);
		}

		public void setNumOfTracks(final int numOfTracs) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 2, numOfTracs);
		}

		public int getSectsPerTrack() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 4);
		}

		public void setSectsPerTrack(final int sectsPerTrack) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 4, sectsPerTrack);
		}

		public int getDiskSides() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 6);
		}

		public void setDiskSides(final int diskSides) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 6, diskSides);
		}

		public int getDensity() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 8);
		}

		public void setDensity(final int density) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 8, density);
		}

		public int getOnline() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 10);
		}

		public void setOnline(final int online) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 10, online);
		}

		public int getStepRate() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 12);
		}

		public void setStepRate(final int stepRate) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 12, stepRate);
		}

		public int getBaudRate() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 14);
		}

		public void setBaudRate(final int baudRate) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 14, baudRate);
		}

		public int getReserv0() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 16);
		}

		public void setReserv0(final int reserv0) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 16, reserv0);
		}

		public int getReserv1() {
			return ByteArrayUtil.get16bitMSF(rawdata, getOffset() + 18);
		}

		public void setReserv1(final int reserv1) {
			ByteArrayUtil.set16bitMSF(rawdata, getOffset() + 18, reserv1);
		}

		public int getDeviceStatus() {
			return ByteArrayUtil.get8bit(rawdata, getOffset() + 20);
		}

		public void setDeviceStatus(final int status) {
			ByteArrayUtil.set8bit(rawdata, getOffset() + 20, status);
		}

		/**
		 * @return a bitfield containing {@link #F_DOOROPEN},
		 *         {@link #FI_WRTPROT}, {@link #FI_RCDTYPE},
		 *         {@link #FI_RNFERROR}, {@link #FI_CRCERROR},
		 *         {@link #FI_DATALOST}, {@link #FI_DATARQST},
		 *         {@link #FI_FDCBUSY}
		 */
		public int getFStat() {
			return ByteArrayUtil.get8bit(rawdata, getOffset() + 21);
		}

		/**
		 * @param fstat
		 *            a bitfield containing {@link #F_DOOROPEN},
		 *            {@link #FI_WRTPROT}, {@link #FI_RCDTYPE},
		 *            {@link #FI_RNFERROR}, {@link #FI_CRCERROR},
		 *            {@link #FI_DATALOST}, {@link #FI_DATARQST},
		 *            {@link #FI_FDCBUSY}
		 */
		public void setFStat(final int fstat) {
			ByteArrayUtil.set8bit(rawdata, getOffset() + 21, fstat);
		}

		public void init(final int sectors, final int bytesPerSector) {
			Assert.assertTrue(bytesPerSector == 0x80 || bytesPerSector == 0x100);
			final Density density = density(sectors, bytesPerSector);
			setSectorLen(bytesPerSector);
			setSectsPerTrack(density.sectorsPerTrack);
			setDensity(density.density);
			setDeviceStatus(density.deviceStatus);
			setNumOfTracks(0x28);
			setDiskSides(0x00);
			// Ab jetzt eigentlich egal, da extrem hardwarenah
			setOnline(0x01);
			setStepRate(0x01);
			setBaudRate(0x01);
			setReserv0(0x01);
			setReserv1(0x01);
			setFStat(FI_FDCBUSY);
			setDeviceStatus(getDeviceStatus() | D_MOTORON);
		}

		public String dump() {
			final StringBuffer buf = new StringBuffer();
			buf.append("\nSectorLen=");
			buf.append(getSectorLen());
			buf.append("\nNumOfTracks=");
			buf.append(getNumOfTracks());
			buf.append("\nSectsPerTrack=");
			buf.append(getSectsPerTrack());
			buf.append("\nDiskSides=");
			buf.append(getDiskSides());
			buf.append("\nDensity=");
			buf.append(getDensity());
			buf.append("\n");
			return buf.toString();
		}
		/*
	    uint		sectorlen;			// Bytes per sector
		uint		numtracks;			// Number of tracks
		uint		sectspertrack;		// Sectors per track
		uint		disksides;			// Disk sides-1
		uint		density;			// Density (0=FM, 4=MFM)
		uint		online;				// Station ON LINE
		uint		steprate;			// Steprate des Kopfes
		uint		baudrate;			// bertragungsrate
		uint		reserv0;			// Reserved byte #0
		uint		reserv1;			// Reserved byte #1
										//---Statusbytes:-----------
		ubyte		dstat;				// Device Status
		ubyte		fstat;				// Hardware Status
		 */
	}

	/**
	 * The specs defining basic disk properties.
	 * 
	 * @author Copyright (c) 2002 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class DiskSpec {
		private PercomBlock percomBlock = new PercomBlock();

		public final int getOffset() {
			return 0;
		}

		public final int getSize() {
			return percomBlock.getOffset() + percomBlock.getSize() + 20;
		}

		/** 
		 * @return the total number of sectors
		 */
		public int getNumSects() {
			return ByteArrayUtil.get16bitMSF(rawdata, 4);
		}

		public void setNumSects(final int numSects) {
			ByteArrayUtil.set16bitMSF(rawdata, 4, numSects);
		}

		/**
		 * @return the total number of bytes
		 */
		public int getNumBytes() {
			return ByteArrayUtil.get32bitMSF(rawdata, 6);
		}

		public void setNumBytes(final int numBytes) {
			ByteArrayUtil.set32bitMSF(rawdata, 6, numBytes);
		}

		public PercomBlock getPercomBlock() {
			return percomBlock;
		}

		/** 
		 * @return the disk name to remind you of the content
		 */
		public String getDiskName() {
			final StringBuffer buf = new StringBuffer();
			final int offset = percomBlock.getOffset() + percomBlock.getSize();
			for (int i = 0; i < 20; i++) {
				final char c = (char) (ByteArrayUtil.get8bit(rawdata, offset + i));
				if (c >= ' ') {
					buf.append(c);
				}
			}
			return buf.toString();
		}

		/**
		 * Sets the disk name.
		 * 
		 * @param name
		 *            the new disk name (up to 20 characters)
		 */
		public void setDiskName(final String name) {
			final int offset = percomBlock.getOffset() + percomBlock.getSize();
			for (int i = 0; i < 20; i++) {
				ByteArrayUtil.set8bit(rawdata, offset + i, ((i < name.length()) ? name.charAt(i) : ' '));
			}
		}

		public void init(final int sectors, final int bytesPerSector) {
			percomBlock.init(sectors, bytesPerSector);
			setNumSects(
					percomBlock.getNumOfTracks() * percomBlock.getSectsPerTrack() * (1 + percomBlock.getDiskSides()));
			// Sektor 0 ist reserviert, muss dazu! Daher Sektoren 0..numsects,
			// daher "+1"
			setNumBytes(percomBlock.getSectorLen() * (1 + getNumSects()));
			setDiskName("");
		}

		public String dump() {
			final StringBuffer buf = new StringBuffer();
			buf.append("XLST Disk Image DiskSpec:\n");
			TextFormatter.hexdump(buf, rawdata, getOffset(), getSize());
			buf.append("\nDiskName=");
			buf.append(getDiskName());
			buf.append("\nNumSects=");
			buf.append(getNumSects());
			buf.append("\nNumBytes=");
			buf.append(getNumBytes());
			buf.append(percomBlock.dump());
			buf.append("\n");
			return buf.toString();
		}
		/* alter C Code:
	    ubyte		*diskmem;			// Zeiger auf Disk-Daten
		uint		numsects;			// Total number of sectors
		size_t		numbytes;			// Total number of bytes
		PERCOMBLOCK	percom;				// Einmal Percom-Block bitte
		char		diskname[20];		// DiskName zur Erinnerung
		 */
	}

	private DiskSpec diskspec = new DiskSpec();

	public String dumpDiskSpec() {
		return diskspec.dump();
	}

	@Override
	public boolean isWriteProtected() {
		return (0 != (diskspec.percomBlock.getFStat() & PercomBlock.FI_WRTPROT));
	}

	@Override
	public void setWriteProtected(final boolean isWriteProtected) {
		int fstat = diskspec.percomBlock.getFStat();
		if (isWriteProtected) {
			fstat |= PercomBlock.FI_WRTPROT;
		} else {
			fstat &= ~PercomBlock.FI_WRTPROT;
		}
		diskspec.percomBlock.setFStat(fstat);
	}

	@Override
	public void load(final InputStream is) {
		final ByteArrayOutputStream os = new ByteArrayOutputStream();
		int c;
		try {
			while ((c = is.read()) != -1) {
				os.write(c);
			}
			rawdata = os.toByteArray();
			os.close();
		} catch (final IOException e) {
			logger.error("Unable to load disk", e);
			Assert.assertTrue(false);
		}
		/* Old C Code:
		int			handle, err;
		uint		sector;
		size_t		sectlen;
		long		len;
		ubyte		*adr;
		DISKSPEC	*dptr;
		err =0;
		dptr=&disklist[disk];
		if ((handle=open(filename,O_RDONLY))<0)
		{	errorcode=ErrFILENAME;
			strncpy (errorhelp,filename,80);
			return -1;
		}
		len=read (handle, dptr, sizeof (DISKSPEC));
		if (len<sizeof(DISKSPEC))
		{	errorcode=ErrREAD;
			strncpy (errorhelp,filename,80);
			err=-1;
		}
		sectlen=dptr->percom.sectorlen;
		if (dptr->diskmem!=NULL)
		{	dptr->diskmem=NULL;
			if (initdisk(disk,-1,FALSE)) err=-1;
			while (read(handle,&sector,sizeof(sector))==sizeof(sector) &&
					(adr=getsectadr(disk,sector))!=NULL)
			{	len=read (handle,adr,sectlen);
				if (len!=sectlen)
				{	errorcode=ErrFULLDST;
					strncpy (errorhelp,filename,80);
					err=-1;
				}
			}
		}
		close (handle);
		return err;
		 */
	}

	@Override
	public void save(final OutputStream os) {
		throw new AssertionException("Not implemented yet");
	}

	@Override
	public int getNumOfSectors() {
		return diskspec.getNumSects();
	}

	@Override
	public int getSectorSize() {
		return diskspec.getPercomBlock().getSectorLen();
	}

	/**
	 * @param sector
	 *            the sector number (a positive number)
	 * @return the position of the first byte of the given sector
	 *         within this disk image, or -1 if not found
	 */
	protected int getSectorOffset(final int sector) {
		Assert.assertTrue(sector > 0);
		int off = diskspec.getSize();
		while (off < rawdata.length) {
			int sectnum = ByteArrayUtil.get16bitMSF(rawdata, off);
			off += 2;
			logger.trace("Checking Sector #" + sectnum + " at offset=" + off);
			if (sectnum == sector) {
				logger.trace("Found Sector #" + sectnum);
				return off;
			}
			off += diskspec.getPercomBlock().getSectorLen();
		}
		return -1;
		/*
		int off = atrh.getRawBytes().length;
			while (read(handle,&sector,sizeof(sector))==sizeof(sector) &&
					(adr=getsectadr(disk,sector))!=NULL)
			{	len=read (handle,adr,sectlen);
				if (len!=sectlen)
				{	errorcode=ErrFULLDST;
					strncpy (errorhelp,filename,80);
					err=-1;
				}
			}
		 */
	}

	@Override
	public byte[] getSector(final int sector) {
		final byte[] retval = new byte[getSectorSize(sector)];
		final int off = getSectorOffset(sector);
		if (off > 0) {
			for (int i = 0; i < retval.length; i++) {
				retval[i] = rawdata[off + i];
			}
		}
		return retval;
	}

	@Override
	public void setSector(final int sector, final byte[] arr) {
		Assert.assertEquals(arr.length, getSectorSize(sector));
		Assert.assertTrue(sector > 0);
		int off = getSectorOffset(sector);
		if (off < 0) {
			final byte[] newrawdata = new byte[rawdata.length + 2 + diskspec.getPercomBlock().getSectorLen()];
			ByteArrayUtil.copyPart(newrawdata, 0, rawdata, 0, rawdata.length);
			ByteArrayUtil.set16bitMSF(newrawdata, rawdata.length, sector);
			off = rawdata.length + 2;
			rawdata = newrawdata;
		}
		ByteArrayUtil.copyPart(rawdata, off, arr, 0, arr.length);
	}

	@Override
	public void format(final int sectors, final int bytesPerSector) {
		final int sizeInBytes = diskspec.getSize() + sectors * (2 + bytesPerSector);
		rawdata = new byte[sizeInBytes];
		diskspec.init(sectors, bytesPerSector);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + Arrays.hashCode(rawdata);
		return result;
	}

	/**
	 * @param obj
	 *            the object to compare with
	 * @return <code>true</code> if both objects are of same class and the raw data (disk size and content) is bitwise identical
	 */
	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (!super.equals(obj))
			return false;
		if (getClass() != obj.getClass())
			return false;
		final XLSTDiskImage other = (XLSTDiskImage) obj;
		if (!Arrays.equals(rawdata, other.rawdata))
			return false;
		return true;
	}

	private byte[] rawdata;
}
