package de.ped.atari8bit.test;

import static org.junit.Assert.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Collection;

import org.junit.*;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import de.ped.atari8bit.logic.ATRDiskImage;
import de.ped.atari8bit.logic.Density;
import de.ped.atari8bit.logic.DiskImage;
import de.ped.atari8bit.logic.DiskImage.DOS;
import de.ped.tools.MathUtil;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;


/**
 * Tests the abstract {@link DiskImage} and its subclasses.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
@RunWith(value=Parameterized.class)
public class DiskImageTest {
	private transient Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
	}

	private DiskImage testling;

	@Parameters
	public static Collection<DiskImage[]> data() {
		return Arrays.asList(new DiskImage[][] { { new ATRDiskImage(), },
//        		{ new XLSTDiskImage(), },
		});
	}

	public DiskImageTest(final DiskImage testling) {
		this.testling = testling;
	}

	public void checkDiskFormatting(final Density density, final DOS dos) throws IOException {
		final String testcase = toString() + "," + density + "," + dos;
		testling.format(density);
		final int freeSectsOfEmptyDisk = (density == Density.ENHANCDENS ? 1010 : 707);
		assertEquals(testcase, density.sectors, testling.getNumOfSectors());
		assertEquals(testcase, density.bytesPerSector, testling.getSectorSize());
		assertEquals(testcase, 0, testling.getDirEntries().length);
		testling.initDiskForDos(dos);
		if (dos != DOS.DOS3x) {
			assertEquals(testcase, 0, testling.getDirEntries().length);
			assertEquals(testcase, freeSectsOfEmptyDisk, testling.getNumOfFreeSectorsForEmptyDisk());
			assertEquals(testcase, freeSectsOfEmptyDisk, testling.getNumOfFreeSectorsFromFAT());
			assertEquals(testcase, freeSectsOfEmptyDisk, testling.getNumOfFreeSectorsFromFATPrecalculated());

			final String xlfilename = "HALLO.TXT";
			final byte[] data = new byte[1983];
			for (int i = 0; i < data.length; i++) {
				data[i] = (byte) (MathUtil.random(0x100) & 0xff);
			}

			{ // Import file a first time, should work fine.
				final InputStream is = new ByteArrayInputStream(data);
				testling.importFile(xlfilename, is);
				final DiskImage.DirectoryEntry[] entries = testling.getDirEntries();
				assertEquals(testcase, 1, entries.length);
				final DiskImage.DirectoryEntry entry = entries[0];
				final String rawname = "HALLO   TXT";
				assertEquals(testcase, xlfilename, entry.nameFromRaw(entry.nameToRaw(xlfilename)));
				assertEquals(testcase, rawname, entry.getRawName());
				assertEquals(testcase, xlfilename, entry.getName());
				assertEquals(testcase, true, entry.isRealEntry());
				assertEquals(testcase, 16, entry.getLengthInSectors());
				assertEquals(testcase, 4, entry.getFirstSectorIndex());
				assertEquals(testcase, freeSectsOfEmptyDisk - 16, testling.getNumOfFreeSectorsFromFAT());
				assertEquals(testcase, freeSectsOfEmptyDisk - 16, testling.getNumOfFreeSectorsFromFATPrecalculated());
			}
			{ // Import same file a second time, should fail.
				final InputStream is = new ByteArrayInputStream(data);
				try {
					testling.importFile(xlfilename, is);
					fail(testcase + " could import the same filename twice");
				} catch (final IllegalStateException ex) {
					// That's what we expect
				}
			}
			{ // Export file, should stay the same
				final ByteArrayOutputStream os = new ByteArrayOutputStream();
				testling.exportFile(0, os);
				assertArrayEquals(data, os.toByteArray());
			}
			logger.debug(testcase + " hat geklappt");
		}
	}

	@Test
	public void testDiskImage() throws IOException {
		checkDiskFormatting(Density.SINGLEDENS, DOS.DOS2x);
		checkDiskFormatting(Density.ENHANCDENS, DOS.DOS2x);
		checkDiskFormatting(Density.SINGLEDENS, DOS.PERCOM);
		checkDiskFormatting(Density.SINGLEDENS, DOS.DOS3x);
		checkDiskFormatting(Density.ENHANCDENS, DOS.DOS3x);
		// Double density not supported by initDiskForDos
//		checkDiskFormatting(Density.DOUBLEDENS, DOS.PERCOM);
	}

	@Override
	public String toString() {
		return testling.getClass().getSimpleName();
	}
}
