package de.ped.conquest.gui;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Polygon;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.ArrayList;
import java.util.Vector;

import javax.swing.JPanel;

import de.ped.conquest.logic.Board;
import de.ped.conquest.logic.Game;
import de.ped.conquest.logic.Playfield;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.log.Logger;

/**
 * The area where the playfield is painted.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ConquestMainCanvas extends JPanel implements MouseListener {
	private static final long serialVersionUID = 1L;

	private static Logger logger = Logger.getRootLogger();

	public static final Dimension PREFERRED_SIZE = new Dimension(350, 400);
	public static final Dimension MINIMUM_SIZE = PREFERRED_SIZE;

	public static final Color COLOR_FIELD_WALL = Color.black;
	public static final Color COLOR_FIELD_FREE = Color.gray;
	public static final Color COLOR_FIELD_PLAYER0 = Color.red;
	public static final Color COLOR_FIELD_PLAYER1 = Color.green;
	public static final Color COLOR_DECORATION_START = Color.white;
	public static final Color COLOR_DECORATION_GROW = Color.magenta;
	public static final Color COLOR_DECORATION_JUMP = Color.blue;

	/** Accessor to the logic layer that the GUI presents. */
	private Game model;

	private final ConquestMainWindow parent;
	
	/**
	 * A new game has started. This may have changed the topology.
	 * 
	 * @param game
	 *            the new model
	 */
	public void setModel(final Game game) {
		model = game;
		playfieldPainter = createPlayfieldPainter(model);
	}

	protected Game getModel() {
		return model;
	}

	/** The playfield painter for the topology of the game. */
	private PlayfieldPainter playfieldPainter;

	public ConquestMainCanvas(final ConquestMainWindow parent) {
		this.parent = parent;
		setBackground(Color.black);
		addMouseListener(this);
	}

	@Override
	public Dimension getPreferredSize() {
		return PREFERRED_SIZE;
	}

	@Override
	public Dimension getMinimumSize() {
		return MINIMUM_SIZE;
	}

	/**
	 * Chooses the color to paint for a playfield, based on the content of that
	 * field.
	 * 
	 * @param playfieldContent
	 *            the content of the playfield
	 * @return the color to paint for the playfield
	 */
	protected Color chooseColor(final int playfieldContent) {
		switch (playfieldContent) {
		case Playfield.WALL:
			return COLOR_FIELD_WALL;
		case Playfield.FREE:
			return COLOR_FIELD_FREE;
		default:
			if (playfieldContent == Playfield.owned(0)) {
				return COLOR_FIELD_PLAYER0;
			} else if (playfieldContent == Playfield.owned(1)) {
				return COLOR_FIELD_PLAYER1;
			} else {
				throw new ArrayIndexOutOfBoundsException();
			}
		}
	}

	@Override
	public void paintComponent(final Graphics g) {
		super.paintComponent(g);
		final Game model = getModel();
		if ((null == model) || (null == playfieldPainter)) {
			return;
		}
		final Vector<PlayfieldPosition> possibleStartPositions = model
				.getPossibleStartPositions();
		final Vector<PlayfieldPosition> possibleDestinationsGrow = model
				.getPossibleDestinationsInDistance(Game.DISTANCE_GROW);
		final Vector<PlayfieldPosition> possibleDestinationsJump = model
				.getPossibleDestinationsInDistance(Game.DISTANCE_JUMP);
		if (null != possibleDestinationsGrow) {
			logger.trace("possibleDestinationsGrow.size = " + possibleDestinationsGrow.size());
		}
		if (null != possibleDestinationsJump) {
			logger.trace("possibleDestinationsJump.size = " + possibleDestinationsJump.size());
		}
		final PlayfieldPosition pos = new PlayfieldPosition();
		final Board board = model.getBoard();
		playfieldPainter.recalcCenter(board, getSize());
		for (int x = 0; x < board.getSize().getWidth(); x++) {
			pos.setX(x);
			for (int y = 0; y < board.getSize().getHeight(); y++) {
				pos.setY(y);
				int val = Playfield.WALL;
				try {
					val = board.getPlayfieldValueAt(pos);
				} catch (final ArrayIndexOutOfBoundsException exc) {
					// Nothing to do
				}
				if (val != Playfield.WALL) {
					final Color mainColor = chooseColor(board.getPlayfieldValueAt(pos));
					final Color secondColor;
					boolean isFilled = true;
					if (pos.equals(model.getCurrentMoveStartPos())) {
						secondColor = COLOR_DECORATION_START;
					} else if ((null != possibleDestinationsGrow) && (possibleDestinationsGrow.contains(pos))) {
						secondColor = COLOR_DECORATION_GROW;
					} else if ((null != possibleDestinationsJump) && (possibleDestinationsJump.contains(pos))) {
						secondColor = COLOR_DECORATION_JUMP;
					} else if ((null != possibleStartPositions) && (possibleStartPositions.contains(pos))) {
						secondColor = COLOR_DECORATION_START;
						isFilled = false;
					} else {
						secondColor = null;
					}
					playfieldPainter.paintPlayfield(g, board, pos, mainColor, secondColor, isFilled);
				}
			}
		}
		final ArrayList<Polygon> hack = PlayfieldTiling.getPolys();
		if (null != hack) {
			g.setColor(Color.blue);
			for (Polygon poly : hack) {
				g.fillPolygon(poly);
			}
		}
		// Paint the last move of the opponent
		final PlayfieldPosition lzsp = model.getLastMoveStartPos();
		final PlayfieldPosition lzep = model.getLastMoveEndPos();
		if ((null != lzsp) && (null != lzep)) {
			playfieldPainter.paintLastMove(g, lzsp, lzep);
		}
	}

	/**
	 * Reacts on mouse clicks into this canvas.
	 * 
	 * @param e
	 *            the mouse event to react to
	 */
	@Override
	public void mouseClicked(final MouseEvent e) {
		logger.trace("mouseClicked: " + e.paramString());
		if ((null != model) && (null != playfieldPainter)) {
			final PlayfieldPosition pos = playfieldPainter.pixel2position(e.getPoint());
			if ((null != pos) && (model.getBoard().isValid(pos))) {
				try {
					setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
					model.onClick(pos);
				} finally {
					setCursor(Cursor.getDefaultCursor());
					parent.updateStatus();
					parent.repaint();
				}
			}
		}
	}

	@Override
	public void mousePressed(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseReleased(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseEntered(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseExited(final MouseEvent e) {
		// Nothing to do
	}

	/**
	 * Internal factory for playfield painters.
	 * 
	 * @param model
	 *            the game to create the playfield painter for
	 * @return the playfield painter
	 */
	private PlayfieldPainter createPlayfieldPainter(final Game model) {
		PlayfieldPainter retval = null;
		if (null != model) {
			retval = new PlayfieldPainter(model.getBoard().getTopologyMode());
		}
		return retval;
	}
}
