package de.ped.conquest.gui;

import java.awt.BorderLayout;
import java.util.ArrayList;
import java.util.Locale;

import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JScrollPane;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.ResourceFinder.Folder;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.TextFormatter.ColumnInfo;
import de.ped.tools.TextFormatter.TableCellInfo;
import de.ped.tools.TextFormatter.TableInfo;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PropertyElement;
import de.ped.tools.log.Logger;
import de.ped.conquest.logic.BoardFactory;
import de.ped.conquest.logic.Game;
import de.ped.conquest.logic.Player;

/**
 * The main window, including menu.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ConquestMainWindow extends ApplicationMainWindow {
	private static final long serialVersionUID = 1L;
	private static Logger logger = Logger.getRootLogger();

	public ConquestMainWindow(final ApplicationEnvironment appEnv) {
		super(appEnv);
		getContentPane().setLayout(new BorderLayout());
		mainCanvas = new ConquestMainCanvas(this);
		setModel();
		getContentPane().add(new JScrollPane(mainCanvas), BorderLayout.CENTER);
		statusLabel = new JLabel();
		getContentPane().add(statusLabel, BorderLayout.SOUTH);
		setJMenuBar(createMenuBar());
		pack(); // Resize the window to the preferred size of its components.
		switchLocale(getMessages().getCurrentLocale());
	}

	public void setModel() {
		model = new Game(boardFactory.createBoard(chosenBoard), chosenStrategy0, chosenStrategy1);
		mainCanvas.setModel(model);
	}

	public void updateStatus() {
		if (null != model) {
			final String statekey = model.getStateOfMove().i18nkey;
			final Messages messages = getMessages();
			final String statestr = messages.getText(statekey);
			final String currentPlayerKey = (0 == model.getGameState().getCurrentPlayerIndex() ? "Game.Player0" : "Game.Player1");
			final String currentPlayerStr = messages.getText(currentPlayerKey);
			final String statusLabelText = currentPlayerStr + " ("
					+ messages.getText(model.getCurrentPlayer().getPlayerTypeName()) + "), " + statestr;
			statusLabel.setText(statusLabelText);
			statusLabel.setHorizontalAlignment(JLabel.CENTER);
			statusLabel.repaint();
		}
	}

	@Override
	protected void startNewGame() {
		final boolean wasRunning = isRunning();
		pauseOrResume(false);
		final NewGameSelectionDialog dialog = getOrCreateNewGameSelectionDialog();
		dialog.enterDialog(chosenBoard, chosenStrategy0, chosenStrategy1);
		logger.debug("New game selection dialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (dialog.wasAcceptedDialogChanges()) {
			chosenBoard = dialog.getChosenBoard();
			chosenStrategy0 = dialog.getChosenStrategy(0);
			chosenStrategy1 = dialog.getChosenStrategy(1);
			setModel();
			updateStatus();
			repaint();
			pauseOrResume(true);
		} else {
			pauseOrResume(wasRunning);
		}
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		doc.setTitle(getMessages().getString("HelpWindow.Title"));
		doc.setHeading("Help");
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");
		doc.addContentPart(2, "Help.Operation");

		doc.addContentPart(2, "Help.Colors");
		final String[] colorRowKeys = new String[] { 
				"Help.Colors.Fields",
				"Game.Field.WALL", 
				"Game.Field.EMPTY", 
				"Game.Field.PLAYER0", 
				"Game.Field.PLAYER1", 
				"Help.Colors.Decorations",
				"Game.Decoration.Potential.Start",
				"Game.Decoration.Selected.Start", 
				"Game.Decoration.Potential.GrowTo",
				"Game.Decoration.Potential.JumpTo", 
				"Game.Decoration.OtherPlayerMove",
		};
		doc.beginTable(null, new TableInfo(doc, false, false, "", new ColumnInfo[] { 
								new ColumnInfo(4, TextFormatter.CENTER, doc.messages(), "Help.Colors.Fields.Color"),
								new ColumnInfo(80, TextFormatter.LEFT, doc.messages(), "Help.Colors.Fields.Desc"), }));
		for (final String rowKey: colorRowKeys) {
			addColorTableRow(doc, rowKey);
		}
		doc.endTable();
		
		final String[] topologyRowKeys = new String[] { 
				"GameProperties.MapTopologyMode.HEXAGON", 
				"GameProperties.MapTopologyMode.TRIANGLE", 
				"GameProperties.MapTopologyMode.SQUARE4", 
				"GameProperties.MapTopologyMode.SQUARE8", 
		};
		final PropertyElement[] pes = new PropertyElement[topologyRowKeys.length];
		for (int i = 0; i < topologyRowKeys.length; i++) {
			pes[i] = new PropertyElement(topologyRowKeys[i], null, null, topologyRowKeys[i], Folder.COMMON);
		}
		buildPropertyDescriptions(doc, 2, "Help.Topologies", pes, false);
		buildMenuDescriptions(doc, getJMenuBar());
		buildInputDescriptions(doc, null, false); // Should do nothing
		buildPropertyDescriptions(doc, getOrCreateNewGameSelectionDialog());
		buildPropertyDescriptions(doc, 3, "GameProperties.Players", Player.getStrategies(), false);
		doc.endDocument();
	}

	private void addColorTableRow(final TextDocumentBuilder doc, final String rowKey) {
		ArrayList<TableCellInfo> cellInfos = new ArrayList<TableCellInfo>();
		if (rowKey.startsWith("Help.")) {
			cellInfos.add(new TableCellInfo(
					doc.messages().getText(new I18NStringWithFillIns(rowKey, null, null)), null,
					TextFormatter.CENTER, 2, 1));
		} else {
			cellInfos.add(new TableCellInfo(doc.htmlTextWithImg("",
					TextDocumentBuilder.getCodedImageFilename(rowKey + ".png", ResourceFinder.Folder.APP))));
			cellInfos.add(new TableCellInfo(doc.messages().getText(new I18NStringWithFillIns(rowKey, null, null))));
		}
		doc.addTableRow(cellInfos);
	}

	@Override
	public void switchLocale(final Locale locale) {
		super.switchLocale(locale);
		newGameSelectionDialog = null;
		updateStatus();
		repaint();
	}

	private JMenuBar createMenuBar() {
		final JMenuBar menuBar = createJMenuBar();
		final JMenu menuFile = createJMenu("Menu.File");
		createActionItem(menuFile, startNewGameAction);
		menuFile.addSeparator();
		menuFile.add(createLocalesMenu());
		menuFile.addSeparator();
		createActionItem(menuFile, quitAction);
		menuBar.add(menuFile);
		final JMenu menuHelp = createJMenu("Menu.Help");
		createActionItem(menuHelp, helpAction);
		createActionItem(menuHelp, aboutAction);
		menuBar.add(menuHelp);
		try {
			menuBar.setHelpMenu(menuHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}
		menuBar.add(menuHelp);
		return menuBar;
	}

	private NewGameSelectionDialog getOrCreateNewGameSelectionDialog() {
		if (null == newGameSelectionDialog) {
			newGameSelectionDialog = new NewGameSelectionDialog(this);
		}
		return newGameSelectionDialog;
	}

	BoardFactory getBoardFactory() {
		return boardFactory;
	}

	/** Access to the model to display. */ 
	private Game model;

	private final BoardFactory boardFactory = new BoardFactory();

	private int chosenBoard = 3;
	private int chosenStrategy0 = 0;
	private int chosenStrategy1 = 2;

	private NewGameSelectionDialog newGameSelectionDialog;

	/** The canvas painting the game board. */
	private ConquestMainCanvas mainCanvas;

	/** The label to show status messages. */
	private JLabel statusLabel;
}
