package de.ped.conquest.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Polygon;

import de.ped.conquest.logic.Board;
import de.ped.conquest.logic.Playfield;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Logger;

/**
 * Maps pixel to position on playfield and paints the playfield.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
class PlayfieldPainter {
	PlayfieldPainter(TopologyMode topologyMode) {
		final int stretch = 24;
		switch(topologyMode) {
		case SQUARE8:
		case SQUARE4:
			ptd = new PlayfieldTiling(topologyMode, topologyMode, stretch, stretch, 1, 1, 1);
			break;
		case HEXAGON:
		case HEXAGON_H:
			ptd = new PlayfieldTiling(topologyMode, topologyMode, stretch, stretch, 1, 1, 4);
			break;
		case TRIANGLE:
			ptd = new PlayfieldTiling(topologyMode, topologyMode, stretch / 2, stretch, 2, 1, 2);
			break;
		default:
			ptd = null;
		}
	}

	/**
	 * Calculates (after change of model or window size) the translation to keep
	 * the board centered within the canvas. Affects how all the following calls are executed.
	 * 
	 * @param board
	 *            the board to use for calibration
	 * @param windowSize
	 *            the window dimension (in pixels)
	 */
	public void recalcCenter(final Board board, final Dimension windowSize) {
		final Logger logger = Logger.getLogger(getClass());
		final double[] rect = getSurroundingRectangle(board);
		logger.trace("recalcCenter: surroundingRect=[" + rect[0] + "," + rect[1] + "," + rect[2] + "," + rect[3] + "]");
		final double bsx = rect[2] - rect[0];
		final double bsy = rect[3] - rect[1];
		final double cx = rect[0] + bsx / 2.0;
		final double cy = rect[1] + bsy / 2.0;
		final int stretchStep = ptd.stretchStep;
		final int xstretchNew = Math.max(STRETCH_MIN, (int) (((double) windowSize.width / (bsx + 1.0)) / (double) stretchStep) * stretchStep);
		final int ystretchNew = Math.max(STRETCH_MIN, (int) (((double) windowSize.height / (bsy + 1.0)) / (double) stretchStep) * stretchStep);
		final int stretchNew = Math.min(xstretchNew, ystretchNew);
		final int xstr = (TopologyMode.TRIANGLE == ptd.getPositionTopology() ? stretchNew / 2 : stretchNew); // FIXME: this code is ugly
		ptd.setSpace(xstr, stretchNew);
		final int xtransNew = windowSize.width / 2 - (int)(cx * stretchNew);
		final int ytransNew = windowSize.height / 2 - (int)(cy * stretchNew);
		logger.trace("recalcCenter: middle=[" + cx * stretchNew + "," + cy * stretchNew + "], window size=[" + windowSize.width + "," + windowSize.height + "]");
		logger.trace("recalcCenter: stretch=" + stretchNew + ", trans=[" + xtransNew + "," + ytransNew + "]");
		trans.setLocation(xtransNew, ytransNew);
	}

	/**
	 * Paints one playfield.
	 * 
	 * @param g
	 *            the graphics object to paint into
	 * @param board
	 *            the board to paint the playfield from
	 * @param pos
	 *            the playfield position to paint
	 * @param playfieldColor
	 *            the main color to paint the playfield with
	 * @param decorationColor
	 *            the color to mark move options with. Set to <code>null</code> if no move option to paint.
	 * @param isDecorationToFill
	 *            if <code>true</code>, the decoration will be drawn filled, otherwise only outlined.
	 */
	@SuppressWarnings("unused")
	public void paintPlayfield(
			final Graphics g, 
			final Board board,
			final PlayfieldPosition pos, 
			final Color playfieldColor, 
			final Color decorationColor,
			final boolean isDecorationToFill) {
		final Point sc = new Point();
		position2pixelCenter(pos, sc);
		g.setColor(playfieldColor);
		final Point sul = new Point();
		ptd.map2canvas(sul, pos, trans, PlayfieldTiling.TOP_LEFT);
		final Polygon poly = ptd.polygon(pos, sul, 0);
		g.fillPolygon(poly);
		if ((null != decorationColor) && !playfieldColor.equals(decorationColor)) {
			g.setColor(decorationColor);
//			final int diag = Math.min (ptd.getSurroundingWidth(), ptd.getSurroundingHeight()) / 4;
//			g.fillRect(sc.x - diag / 2, sc.y - diag / 2, diag, diag);
			final Polygon innerPoly = ptd.polygon(pos, sul, -75);
			if (isDecorationToFill) {
				g.fillPolygon(innerPoly);
			} else {
				g.drawPolygon(innerPoly);
			}
		}
		g.setColor(Color.yellow);
		g.drawPolygon(poly);
		if (false) { // Debugging: paints the Position of the playfield into the playfield
			final String str = pos.toString();
			g.setColor(Color.white);
			g.drawString(str, sc.x - ptd.getSpaceWidth() / 2, sc.y);
		}
		if (false) { // Debugging: Paints distance to position(0,0) into the playfield
			final int dist = board.calcDistance(new PlayfieldPosition(0, 0), pos);
			g.setColor(Color.white);
			g.drawString(Integer.toString(dist), sc.x, sc.y);
		}
	}

	/**
	 * Paints the last move of the opponent.
	 * 
	 * @param g
	 *            the graphics object to paint into
	 * @param posStart
	 *            the playfield position where the move started
	 * @param posEnd
	 *            the playfield position where the move ended
	 */
	public void paintLastMove(final Graphics g, final PlayfieldPosition posStart, final PlayfieldPosition posEnd) {
		final Point s = new Point();
		final Point e = new Point();
		position2pixelCenter(posStart, s);
		position2pixelCenter(posEnd, e);
		g.setColor(Color.darkGray);
		g.drawLine(s.x, s.y, e.x, e.y);
		final int diag = Math.min (ptd.getSurroundingWidth(), ptd.getSurroundingHeight()) / 2;
		g.drawOval(e.x - diag / 2, e.y - diag / 2, diag, diag);
	}

	/**
	 * @param board the board to calculate the surrounding for
	 * @return an array of size 4, returning the upper, left, right and lower boundary of the whole board, in board coordinates.
	 */
	private double[] getSurroundingRectangle(final Board board) {
		final double[] ret = new double[] { Double.MAX_VALUE, Double.MAX_VALUE, Double.MIN_VALUE, Double.MIN_VALUE };
		final double[] fieldsurrect = new double[] {
				(double) ptd.getSurroundingWidth() / ptd.getSpaceWidth(),
				(double) ptd.getSurroundingHeight() / ptd.getSpaceHeight(),
		};
		final double[] curfieldsurrect = new double[4];
		final double[] boardCoords = new double[2];
		final Logger logger = Logger.getLogger(getClass());
		final PlayfieldDimension size = board.getSize();
		final PlayfieldPosition pos = new PlayfieldPosition(); // re-used for performance reasons
		final Point origin = new Point();
		for (int y = 0; y < size.getHeight(); y++) {
			for (int x = 0; x < size.getWidth(); x++) {
				pos.setLocation(x, y);
				final int pfv = board.getPlayfieldValueAt(pos);
				if (pfv != Playfield.WALL) {
					final Point s = new Point();
					ptd.map2canvas(s, pos, origin, PlayfieldTiling.CENTER);
					boardCoords[0] = (double) s.x / ptd.getSpaceWidth()  / ptd.getRatioWidth();
					boardCoords[1] = (double) s.y / ptd.getSpaceHeight() / ptd.getRatioHeight();

					curfieldsurrect[0] = boardCoords[0] - fieldsurrect[0] / 2.0;
					curfieldsurrect[1] = boardCoords[1] - fieldsurrect[1] / 2.0;
					curfieldsurrect[2] = boardCoords[0] + fieldsurrect[0] / 2.0;
					curfieldsurrect[3] = boardCoords[1] + fieldsurrect[1] / 2.0;
					if (ret[0] > curfieldsurrect[0]) {
						ret[0] = curfieldsurrect[0];
						logger.trace("getSourroundingRectangle: LEFT  pos=" + pos.toString() + ", boardCoords=[" + boardCoords[0] + "," + boardCoords[1] + ")");
					}
					if (ret[1] > curfieldsurrect[1]) {
						ret[1] = curfieldsurrect[1];
						logger.trace("getSourroundingRectangle: UP    pos=" + pos.toString() + ", boardCoords=[" + boardCoords[0] + "," + boardCoords[1] + ")");
					}
					if (ret[2] < curfieldsurrect[2]) {
						ret[2] = curfieldsurrect[2];
						logger.trace("getSourroundingRectangle: RIGHT pos=" + pos.toString() + ", boardCoords=[" + boardCoords[0] + "," + boardCoords[1] + ")");
					}
					if (ret[3] < curfieldsurrect[3]) {
						ret[3] = curfieldsurrect[3];
						logger.trace("getSourroundingRectangle: DOWN  pos=" + pos.toString() + ", boardCoords=[" + boardCoords[0] + "," + boardCoords[1] + ")");
					}
				}
			}
		}
		// Just for safety - if no board coords were found...
		for (int i = 0; i < ret.length; i++) {
			final double val = ret[i];
			if ((Double.MIN_VALUE < val) && (val < Double.MAX_VALUE)) {
				// all is good
			} else {
				ret[i] = (i < 2 ? 0.0 : ret[i - 2] + 1.0);
			}
		}
		return ret;
	}

	/**
	 * @param point the pixel position
	 * @return the playfield for the pixel. If the pixel is not part of any
	 *         playfield, the result is <code>null</code>.
	 */
	public PlayfieldPosition pixel2position(final Point point) {
		return ptd.canvas2map(point.x, point.y, trans);
	}

	/**
	 * @param pos the playfield position
	 * @param ret output parameter: the point on the screen of the center of this playfield
	 */
	protected void position2pixelCenter(final PlayfieldPosition pos, final Point ret) {
		ptd.map2canvas(ret, pos, trans, PlayfieldTiling.CENTER);
	}

	/** Smaller than that (in pixel per field) the board will not be drawn. */
	protected final int STRETCH_MIN = 12;

	/** offset from canvas origin to board origin */
	protected Point trans = new Point(5, 5);
	protected final PlayfieldTiling ptd;
}
