package de.ped.conquest.logic;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;

/**
 * A container knowing all the playfields, their positions, and the topology,
 * what in this case means the distance between the playfields.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Board implements Cloneable {
	/**
	 * A helper class to store the key for the topography lookup. This is
	 * necessary for speed optimization, as otherwise the system would spend
	 * most of its time re-calculating the neighboring relationships on the
	 * board.
	 * 
	 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class LookupKey {
		public final PlayfieldPosition pos;
		public final int distance;

		public LookupKey(final PlayfieldPosition pos, final int depth) {
			this.pos = pos;
			this.distance = depth;
		}

		@Override
		public String toString() {
			return pos.toString() + distance;
		}

		@Override
		public boolean equals(final Object o) {
			if (o instanceof LookupKey) {
				LookupKey other = (LookupKey) o;
				return ((distance == other.distance) && pos.equals(other.pos));
			}
			return false;
		}

		@Override
		public int hashCode() {
			return pos.hashCode();
		}
	}

	/**
	 * @return how many playfields are in the board - including walls
	 */
	public int getNumOfIndices() {
		return playfields.length;
	}

	/**
	 * @return the maximum size of the board, measured in playfields
	 */
	public PlayfieldDimension getSize() {
		return size;
	}

	/**
	 * @param pos
	 *            the (two-dimensional) playfield position
	 * @return <code>true</code> if this position is a legal position on the board
	 *         (means: we find an index for it)
	 */
	public boolean isValid(final PlayfieldPosition pos) {
		return pos.isWithin(getSize());
	}

	/**
	 * calculates the one-dimensional index of the given position. The inverse
	 * operation is @link{#index2position}.
	 * 
	 * @param pos
	 *            the (two-dimensional) playfield position
	 * @return the (one-dimensional) index of the playfield
	 */
	public int position2index(final PlayfieldPosition pos) {
		if (!isValid(pos)) {
			throw new ArrayIndexOutOfBoundsException(pos.toString());
		}
		return pos.getY() * getSize().getWidth() + pos.getX();
	}

	/**
	 * Calculates the position based on the one-dimensional index. The inverse
	 * operation is @link{#position2index}.
	 * 
	 * @param idx
	 *            the (one-dimensional) index of the playfield
	 * @return the (two-dimensional) playfield position
	 */
	public PlayfieldPosition index2position(final int idx) {
		if ((idx < 0) || (idx >= getNumOfIndices())) {
			throw new ArrayIndexOutOfBoundsException();
		}
		final int width = getSize().getWidth();
		return new PlayfieldPosition(idx % width, idx / width);
	}

	public TopologyMode getTopologyMode() {
		return topologyMode;
	}

	/**
	 * @return the maximum number of possible neighbors
	 */
	public int getMaxNumOfNeighbors() {
		return getTopologyMode().neighbors;
	}

	public int getPlayfieldValueAt(final int idx) {
		return playfields[idx];
	}

	public void setPlayfieldValueAt(final int idx, final int val) {
		if (!Playfield.isValid(val)) {
			throw new ArrayIndexOutOfBoundsException();
		}
		playfields[idx] = val;
	}

	public int getPlayfieldValueAt(final PlayfieldPosition pos) {
		return getPlayfieldValueAt(position2index(pos));
	}

	public void setPlayfieldValueAt(final PlayfieldPosition pos, final int val) {
		setPlayfieldValueAt(position2index(pos), val);
	}

	/**
	 * @param startpos
	 *            the first position
	 * @param endpos
	 *            the second position
	 * @return the distance between the two positions. For this game it's okay
	 *         to only detect distance 1 and 2 correctly.
	 */
	public int calcDistance(final PlayfieldPosition startpos, final PlayfieldPosition endpos) {
		return PlayfieldTiling.calcAirDistanceScalar(getTopologyMode(), startpos, PlayfieldPosition.distance(startpos, endpos));
	}

	/**
	 * @param startpos
	 *            the start position of the move
	 * @param distance
	 *            the distance to reach from the startpos.
	 * @return all positions of playfields in the given distance to the given
	 *         start position.
	 */
	private Vector<PlayfieldPosition> calcAllPlayfieldsInDistanceTo(final PlayfieldPosition startpos,
			final int distance) {
		if (distance < 0) {
			throw new ArrayIndexOutOfBoundsException();
		}
		if (null == startpos) {
			return null;
		}
		final Vector<PlayfieldPosition> v = new Vector<PlayfieldPosition>();
		if (0 == distance) { // Only the same position has distance 0.
			v.add(startpos);
		} else {
			// Brute force generate-and-test
			for (int i = 0; i < playfields.length; i++) {
				final PlayfieldPosition endpos = index2position(i);
				if (calcDistance(startpos, endpos) == distance) {
					v.add(endpos);
				}
			}
		}
		return v;
	}

	/**
	 * @param startpos
	 *            the start position of the move
	 * @param distance
	 *            the distance to reach from the startpos.
	 * @return all positions of playfields in the given distance to the given
	 *         start position.
	 *         <p>
	 *         The reader is not allowed to change the returned data. If you
	 *         must, clone the returned data first.
	 */
	public Vector<PlayfieldPosition> getAllPlayfieldsInDistanceTo(final PlayfieldPosition startpos, final int distance) {
		if (distance < 0) {
			throw new ArrayIndexOutOfBoundsException();
		}
		if (null == startpos) {
			return null;
		}
		final LookupKey key = new LookupKey(startpos, distance);
		Vector<PlayfieldPosition> value = neighborsLookup.get(key);
		if (null == value) {
			value = calcAllPlayfieldsInDistanceTo(startpos, distance);
			if (null != value) {
				neighborsLookup.put(key, value);
			}
		}
		return value;
	}

	/**
	 * If someone wants to change the topography during the game, this method should be called. 
	 * However, call it sparsely, or the performance of the computer players goes to its knees.
	 */
	protected void topographyChanged() {
		neighborsLookup.clear();
	}

	/**
	 * @param value the playfield content to count
	 * @return the number of playfields with a given value (Only for
	 *         statistics)
	 */
	public int countPlayfieldsOfValue(final int value) {
		int counter = 0;
		for (int playfield : playfields) {
			if (playfield == value) {
				counter++;
			}
		}
		return counter;
	}

	/**
	 * Creates a new board with the starting situation for the game.
	 * @param topologyMode the topology of the board
	 * @param x the width of the board
	 * @param y the height of the board
	 */
	public Board(final TopologyMode topologyMode, final int x, final int y) {
		this.topologyMode = topologyMode;
		this.size = new PlayfieldDimension(x, y);
		playfields = new int[x * y];
		// All fields are empty by default. This is done by setting Playfield.EMPTY to zero.
	}

	/**
	 * Creates a copy of the board. The board playfields themselves are cloned,
	 * so they can change independently, but the topography (neighboring
	 * relationships between the playfields) stay shared.
	 * 
	 * @param other
	 *            the board to clone from
	 */
	protected Board(final Board other) {
		this.topologyMode = other.topologyMode;
		size = other.size.clone();
		playfields = other.playfields.clone();
		neighborsLookup = other.neighborsLookup; // Shallow copy only, we assume the topography won't change
	}

	/**
	 * @return a human-readable text describing the current topology lookup
	 *         (neighboring) data structure
	 */
	public String topologyLookupToString() {
		final StringBuffer s = new StringBuffer();
		final Iterator<Map.Entry<LookupKey, Vector<PlayfieldPosition>>> topologyLookupIterator = neighborsLookup.entrySet().iterator();
		while (topologyLookupIterator.hasNext()) {
			final Map.Entry<LookupKey, Vector<PlayfieldPosition>> entry = topologyLookupIterator.next();
			s.append("\n\t");
			s.append(entry.getKey().toString());
			s.append(": ");
			s.append(entry.getValue().size());
			final Iterator<PlayfieldPosition> neighborsIterator = entry.getValue().iterator();
			while (neighborsIterator.hasNext()) {
				PlayfieldPosition pp = neighborsIterator.next();
				s.append(pp.toString());
			}
		}
		return s.toString();
	}

	@Override
	public String toString() {
		final StringBuffer s = new StringBuffer();
		s.append(getSize().getWidth());
		s.append(',');
		s.append(getSize().getHeight());
		s.append('[');
		for (int playfield : playfields) {
			s.append(Playfield.toString(playfield));
		}
		s.append(']');
		return s.toString();
	}

	/**
	 * @return a clone of this board.
	 */
	@Override
	public Object clone() throws CloneNotSupportedException {
		return new Board(this);
	}

	/**
	 * Two boards are equal if the topology, the {@link #getSize()} and all the playfields are identical.
	 * @param o the board compare this with
	 * @return <code>true</code> if the boards are equal
	 */
	@Override
	public boolean equals(final Object o) {
		if (null == o) {
			return false;
		}
		if (!(o instanceof Board)) {
			return false;
		}
		final Board other = (Board) o;
		if (!this.getTopologyMode().equals(other.getTopologyMode())) {
			return false;
		}
		if (!size.equals(other.size)) {
			return false;
		}
		if (playfields.length == other.playfields.length) {
			for (int i = 0; i < playfields.length; i++) {
				if (playfields[i] != other.playfields[i]) {
					return false;
				}
			}
			return true;
		}
		return false;
	}

	/**
	 * @return hash value is the EOR of all playfields owned by players (using 2
	 *         bits for each playfield, shifting through the int)
	 */
	@Override
	public int hashCode() {
		int hashCode = 0x00;
		for (int i = 0; i < playfields.length; i++) {
			final int shift = 2 * (i % 16);
			final int playfield = playfields[i];
			int value = 0x00;
			if (playfield == Playfield.owned(0)) {
				value = 0x01;
			} else if (playfield == Playfield.owned(1)) {
				value = 0x02;
			}
			hashCode ^= (value << shift);
		}
		return hashCode;
	}

	/**
	 * A board contains many playfields.
	 */
	private int[] playfields;

	/**
	 * The board contains neighboring relationships between the playfields.
	 * These are calculated only once and retrieved again from here.
	 */
	private HashMap<LookupKey, Vector<PlayfieldPosition>> neighborsLookup = new HashMap<LookupKey, Vector<PlayfieldPosition>>();

	private final TopologyMode topologyMode;
	private final PlayfieldDimension size;
}
