package de.ped.conquest.logic;

import java.util.ArrayList;

import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;

/**
 * Creates the boards and knows all createable boards.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class BoardFactory {

	/**
	 * Holds all data to define a board.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class BoardData {
		public BoardData(final String key, final TopologyMode topologyMode, final String[] boardData) {
			this.key = key;
			this.topologyMode = topologyMode;
			this.boardData = boardData;
		}
		public final String key;
		public final TopologyMode topologyMode;
		public final String[] boardData;
	}

	private ArrayList<BoardData> data = new ArrayList<BoardData>();
	{
		data.add(new BoardData("Game.Board.Board01", TopologyMode.SQUARE4, new String[] {
			"oooooo",
			"1ooooo",
			"oooooo",
			"oooooo",
			"ooooo2",
			"oooooo",
		}));
		data.add(new BoardData("Game.Board.Board02", TopologyMode.SQUARE8, new String[] {
			"1ooooooooo",
			"oooooooooo",
			"oooooooooo",
			"oooooooooo",
			"oooo##oooo",
			"oooo##oooo",
			"oooooooooo",
			"oooooooooo",
			"oooooooooo",
			"ooooooooo2",
		}));
		data.add(new BoardData("Game.Board.Board03", TopologyMode.HEXAGON, new String[] {
			"####o####",
			"##ooooo##",
			"ooooooooo",
			"o1ooooooo",
			"oooo#oooo",
			"ooooooooo",
			"ooooooo2o",
			"#ooooooo#",
			"###ooo###",
		}));
		data.add(new BoardData("Game.Board.Board04", TopologyMode.HEXAGON, new String[] {
			"o#o#o#o#o#o#o#o",
			"ooooooooooooooo",
			"ooooooooooooooo",
			"ooo#ooooooooooo",
			"oo###ooooooo#oo",
			"ooo#1oooooo#1#o",
			"ooo#ooo#ooo#o#o",
			"oooooo###oooooo",
			"oooooo###oooooo",
			"ooooo##o##ooooo",
			"oooooo###oooooo",
			"ooooooo#ooooooo",
			"ooooooo#ooooooo",
			"ooo#2oooooo#2#o",
			"oo###oooooo###o",
			"ooo#ooooooooooo",
			"ooo#ooooooooooo",
			"ooooooooooooooo",
			"ooooooooooooooo",
		}));
		data.add(new BoardData("Game.Board.Board05", TopologyMode.TRIANGLE, new String[] {
			"ooooooo",
			"1oooooo",
			"ooooooo",
			"ooooooo",
			"oooooo2",
			"ooooooo",
		}));
		/**
		 * <table>
		 *   <caption>The board</caption>
		 *   <tbody>
		 *     <tr><td> PLR0 </td><td> WALL </td></tr>
		 *     <tr><td> FREE </td><td> PLR1 </td></tr>
		 *   </tbody>
		 * </table>
		 * a board with only one move possible for the next player, and that move wins.
		 */
		data.add(new BoardData("Game.Board.Board06", TopologyMode.SQUARE4, new String[] {
			"1o",
			"#2",
		}));
		/**
		 * <table>
		 *   <caption>The board</caption>
		 *   <tbody>
		 *     <tr><td> WALL </td><td> PLR1 </td><td> WALL </td></tr>
		 *     <tr><td> FREE </td><td> FREE </td><td> PLR0 </td></tr>
		 *   </tbody>
		 * </table>
		 * a board with two moves possible for player 0, 
		 * and one of those moves wins, and the other one lets player 1 win.
		 */
		data.add(new BoardData("Game.Board.Board07", TopologyMode.SQUARE4, new String[] {
			"#2#",
			"oo1",
		}));
		/**
		 * <table>
		 *   <caption>The board</caption>
		 *   <tbody>
		 *     <tr><td> PLR1 </td><td> FREE </td><td> FREE </td></tr>
		 *     <tr><td> FREE </td><td> FREE </td><td> PLR0 </td></tr>
		 *   </tbody>
		 * </table>
		 * a board small enough to calculate all possibilities, but interesting enough to have no immediate winner.
		 */
		data.add(new BoardData("Game.Board.Board08", TopologyMode.SQUARE4, new String[] {
			"22o",
			"oo1",
		}));
		data.add(new BoardData("Game.Board.Board09", TopologyMode.HEXAGON_H, new String[] {
				"##ooooo##",
				"##o1oooo#",
				"#ooooooo#",
				"#oooooooo",
				"oooo#oooo",
				"#oooooooo",
				"#ooooooo#",
				"##oooo2o#",
				"##ooooo##",
			}));
	}

	/**
	 * @return the keys to all board names
	 */
	public String[] getNameKeysOfBoardTemplates() {
		final String[] ret = new String[data.size()];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = data.get(i).key;
		}
		return ret;
	}

	/**
	 * Factory method for boards. Delivers the board with the given index.
	 * 
	 * @param which
	 *            the index of the board to create. In the range from 0 to
	 *            {@link #getNameKeysOfBoardTemplates()} - 1.
	 * @return the board
	 */
	public Board createBoard(final int which) {
		final BoardData d = data.get(which);
		final String[] board = d.boardData;
		final PlayfieldDimension size = new PlayfieldDimension(board[0].length(), board.length);
		final Board b = new Board(d.topologyMode, size.getWidth(), size.getHeight());
		for (int y = 0; y < size.getHeight(); y++) {
			final String s = board[y];
			for (int x = 0; x < size.getWidth(); x++) {
				final char c = s.charAt(x);
				switch(c) {
				case '1':
					b.setPlayfieldValueAt(new PlayfieldPosition(x, y), Playfield.owned(0));
					break;
				case '2':
					b.setPlayfieldValueAt(new PlayfieldPosition(x, y), Playfield.owned(1));
					break;
				case '#':
					b.setPlayfieldValueAt(new PlayfieldPosition(x, y), Playfield.WALL);
					break;
				default:
					// it's an empty field per default
				}
			}
		}
		return b;
	}
}
