package de.ped.conquest.logic;

import java.util.Vector;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;

/**
 * Implements a computer strategy that just looks the half-move ahead and picks the move that gains the most land.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ComputerPlayerGreedy extends Player {
	static String STRATEGY_NAME = Player.STRATEGY_NAME_GREEDY;

	protected PlayfieldPosition bestStartPos;
	protected PlayfieldPosition bestEndPos;

	/**
	 * Rates a move. The higher the better.
	 * 
	 * @param posStart
	 *            the start position of the move
	 * @param posEnd
	 *            the end position of the move
	 * @return the rating for the move
	 */
	protected double calcMoveRating(final PlayfieldPosition posStart, final PlayfieldPosition posEnd) {
		// Winning more land is good.
		final Vector<PlayfieldPosition> vn = getGame().getBoard().getAllPlayfieldsInDistanceTo(posEnd, 1);
		int count = 0;
		for (PlayfieldPosition pn : vn) {
			if (getGame().getBoard().getPlayfieldValueAt(pn) == Playfield.owned(getGame().getGameState().getCurrentOpponentIndex())) {
				count++;
			}
		}
		double rating = count;
		// Grow is better than move.
		if (getGame().getBoard().calcDistance(posStart, posEnd) == 1) {
			rating += 2.0;
		}
		return rating;
	}

	/**
	 * Rates all possible moves. The best possible move is stored in bestStartPos and bestEndPos.
	 * (If there are more than one with highest rating, one is picked randomly.)
	 */
	protected void findBestMove() {
		double bestRating = -1000.0;
		final Vector<PlayfieldPosition> bestStartPoses = new Vector<PlayfieldPosition>();
		final Vector<PlayfieldPosition> bestEndPoses   = new Vector<PlayfieldPosition>();
		final Vector<PlayfieldPosition> vstart = getGame().getPossibleStartPositions();
		for (int istart = 0; istart < vstart.size(); istart++) {
			final PlayfieldPosition posStart = (PlayfieldPosition)vstart.elementAt(istart);
			final Vector<PlayfieldPosition> vend = getGame().getGameState().findAllPossibleDestinationsFrom(posStart);
			for (int iend = 0; iend < vend.size(); iend++) {
				final  PlayfieldPosition posEnd = (PlayfieldPosition)vend.elementAt(iend);
				final double rating = calcMoveRating(posStart, posEnd);
				if (rating > bestRating) {
					// Forget all the others, we found a new best one.
					bestStartPoses.clear();
					bestEndPoses.clear();
					bestStartPoses.add(posStart);
					bestEndPoses.add(posEnd);
					bestRating   = rating;
				} else if (rating == bestRating) {
					// Another one with the same highest rating - keep it as well.
					bestStartPoses.add(posStart);
					bestEndPoses.add(posEnd);
				}
			}
		}
		// Pick one from the best
		final int idx = MathUtil.random(bestStartPoses.size());
		bestEndPos = (PlayfieldPosition) bestEndPoses.elementAt(idx);
		bestStartPos = (PlayfieldPosition) bestStartPoses.elementAt(idx);
	}

	@Override
	protected PlayfieldPosition suggestStartPos() {
		findBestMove();
		return bestStartPos;
	}

	@Override
	protected PlayfieldPosition suggestEndPos() {
		return bestEndPos;
	}

	@Override
	public String getPlayerTypeName() {
		return STRATEGY_NAME;
	}
}
