package de.ped.conquest.logic;

import java.util.ArrayList;
import java.util.HashMap;

import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Level;

/**
 * Implements a computer strategy using the minimax algorithm.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ComputerPlayerMinimax extends Player {

	static String STRATEGY_NAME = Player.STRATEGY_NAME_MINIMAX;

	public static final double RATING_WON = +1000000;
	public static final double RATING_LOST = -RATING_WON;
	public static final double RATING_UNDEFINED = Double.NEGATIVE_INFINITY;

	protected PlayfieldPosition bestStartPos;
	protected PlayfieldPosition bestEndPos;

	private int maxDepth = 2;

	/**
	 * @return the maximum depth of half-moves look into the game
	 */
	public int getMaxDepth() {
		return maxDepth;
	}

	/**
	 * @param maxDepth the maximum depth of half-moves look into the game
	 */
	public void setMaxDepth(final int maxDepth) {
		this.maxDepth = maxDepth;
	}

	/**
	 * A game move with a rating for minimax search.
	 * 
	 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class RatedMove extends GameMove {
		public double rating;

		RatedMove(
				final GameState fromState, 
				final PlayfieldPosition start, 
				final PlayfieldPosition end, 
				final int distance) {
			super(fromState, start, end, distance);
			this.rating = RATING_UNDEFINED;
		}

		@Override
		public String toString() {
			return super.toString() + ": " + rating;
		}
	}

	/**
	 * A factory for game moves with a rating for minimax search.
	 * 
	 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class RatedMoveFactory extends GameMoveFactory {
		@Override
		public GameMove createGameMove(
				final GameState fromState, 
				final PlayfieldPosition start, 
				final PlayfieldPosition end, int distance) {
			return new RatedMove(fromState, start, end, distance);
		}
	}

	/**
	 * @param gameState
	 *            the current game state
	 * @return all moves possible from this game state for the current player
	 */
	public ArrayList<RatedMove> calcPossibleMoves(final GameState gameState) {
		final RatedMoveFactory factory = new RatedMoveFactory();
		final ArrayList<RatedMove> moves = new ArrayList<RatedMove>();
		final ArrayList<GameMove> source = gameState.findAllPossibleMoves(factory);
		for (final GameMove move : source) {
			moves.add((RatedMove) move);
		}
		return moves;
	}

	private HashMap<GameState, Double> gameStateLookup = new HashMap<GameState, Double>();

	/**
	 * @param gameState
	 *            the current game state
	 * @return the rating for the player that has to move right now
	 */
	public double calcStateRating(final GameState gameState) {
		Double rating = gameStateLookup.get(gameState);
		if (null == rating) {
			final int numOfPossibleMoves = gameState.calcNumOfPossibleMoves();
			if (0 == numOfPossibleMoves) {
				rating = new Double(RATING_LOST); // No move possible --> Lost
			} else {
				rating = new Double(numOfPossibleMoves);
			}
			if (gameStateLookup.size() < 30000) {
				gameStateLookup.put(gameState, rating);
			}
		}
		return rating.doubleValue();
	}

	/**
	 * Returns the rating (non-recursively) for the player that has to move
	 * right now.
	 * 
	 * @param move
	 *            The move to try
	 * @return the rating
	 */
	public double calcMoveRating(final RatedMove move) {
		final GameState newState = move.fromState.doMove(move.start, move.end);
		if (null == newState) {
			return RATING_LOST; // Move was impossible --> Lost
		}
		newState.nextPlayer();
		return -calcStateRating(newState);
	}

	private int ratedMovesCounter;

	public RatedMove findBestNextMove(final GameState gameState, final int depth) {
		ratedMovesCounter++;
		final int logLevel = Level.TRACE;
		if (logger.isEnabledFor(logLevel)) {
			logger.log(logLevel, TextFormatter.indent("- ", depth, gameState.toString()));
		}
		final ArrayList<RatedMove> moves = calcPossibleMoves(gameState);
		RatedMove maxMove = null;
		// First see if the next move wins.
		for (final RatedMove move : moves) {
			move.rating = calcMoveRating(move);
			if (logger.isEnabledFor(logLevel))
				logger.log(logLevel, TextFormatter.indent("- ", depth, "* " + move.toString()));
			if ((null == maxMove) || (move.rating > maxMove.rating)) {
				maxMove = move;
			}
		}
		// If not, we need to do minimax deep search
		if ((null != maxMove) && (RATING_LOST < maxMove.rating) && (maxMove.rating < RATING_WON) && (depth < maxDepth)) {
			for (RatedMove move : moves) {
				if ((RATING_LOST < move.rating) && (move.rating < RATING_WON)) {
					final GameState newState = move.fromState.doMove(move.start, move.end);
					if (null != newState) {
						newState.nextPlayer();
						final RatedMove oppMove = findBestNextMove(newState, depth + 1);
						// move.rating *= MathUtil.sgn(oppMove.rating);
						move.rating = -oppMove.rating;
						// Future wins or losses degrade over time
						move.rating -= MathUtil.sgn(move.rating);
						// Penalty for leaving the opponent options
						move.rating -= MathUtil.sgn(move.rating) * calcStateRating(newState);
						if (logger.isEnabledFor(logLevel))
							logger.log(logLevel, TextFormatter.indent("- ", depth, "+ " + move.toString()));
					}
				}
			}
			maxMove = null;
			for (final RatedMove move : moves) {
				if ((null == maxMove) || (move.rating > maxMove.rating)) {
					maxMove = move;
				}
			}
		}
		if (logger.isEnabledFor(logLevel)) {
			logger.log(logLevel, TextFormatter.indent("- ", depth, maxMove.toString()));
		}
		return maxMove;
	}

	public void findBestMove() {
		gameStateLookup = new HashMap<GameState, Double>();
		ratedMovesCounter = 0;
		final RatedMove move = findBestNextMove(getGame().getGameState(), 0);
		logger.debug("Rated " + ratedMovesCounter + " moves.");
		bestStartPos = move.start;
		bestEndPos = move.end;
		gameStateLookup = new HashMap<GameState, Double>();
	}

	@Override
	protected PlayfieldPosition suggestStartPos() {
		findBestMove();
		return bestStartPos;
	}

	@Override
	protected PlayfieldPosition suggestEndPos() {
		return bestEndPos;
	}

	@Override
	public String getPlayerTypeName() {
		return STRATEGY_NAME;
	}
}
