package de.ped.conquest.logic;

import java.util.Vector;

import de.ped.tools.PlayfieldPosition;

/**
 * Represents all the rules of the game, knows the state of the game,
 * and contains all other logical elements of the game, like board or players.
 * Is central starting point for any call into the logical layer from higher layers.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Game {

	/**
	 * The state of the move.
	 * 
	 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum StateOfMove {
		/**
		 * Intermediate state of sanitizing and evaluating the input of the
		 * player. From there the system transitions independently into one of
		 * the other states.
		 */
		CHECKING("Game.State.Checking"),
		/** Expecting the start position of the move from the player. */
		STARTPOS("Game.State.Startpos"),
		/** Expecting the end position of the move from the player. */
		ENDPOS  ("Game.State.Endpos"),
		/**
		 * Indicating that there is no possible move for the player, so the game
		 * is over.
		 */
		GAMEOVER("Game.State.GameOver");
		
		public final String i18nkey;
		
		private StateOfMove(final String key) {
			this.i18nkey = key;
		}
	};

	public static final int DISTANCE_GROW = 1;
	public static final int DISTANCE_JUMP = 2;

	public static final int NUMBER_OF_PLAYERS = 2;

	private StateOfMove stateOfMove = StateOfMove.GAMEOVER;

	private GameState gameState;

	public StateOfMove getStateOfMove() {
		return stateOfMove;
	}

	public GameState getGameState() {
		return gameState;
	}

	/** A game has two players. */
	private Player[] players;

	public Player getPlayer(final int idx) {
		return players[idx];
	}

	public Board getBoard() {
		return gameState.getBoard();
	}

	public Game(final Board board, final int player1type, final int player2type) {
		gameState = new GameState(board);
		players = new Player[2];
		players[0] = Player.createPlayer(this, player1type);
		players[1] = Player.createPlayer(this, player2type);
		setState(StateOfMove.CHECKING);
	}

	public Player getCurrentPlayer() {
		return players[gameState.getCurrentPlayerIndex()];
	}

	public Player getCurrentOpponent() {
		return players[gameState.getCurrentOpponentIndex()];
	}

	// ------------------------------------------------------------------------
	// Methods depending on the current state of the half-move.

	/**
	 * @return all allowed start positions, if in state STATE_STARTPOS. Otherwise, <code>null</code>.
	 */
	public Vector<PlayfieldPosition> getPossibleStartPositions() {
		if (stateOfMove != StateOfMove.STARTPOS) {
			return null;
		}
		return gameState.findAllPossibleStartPositions();
	}

	/**
	 * @param distance the distance in playfields to consider
	 * @return all legal destinations in the given distance to the current start
	 *         position of the current player, or <code>null</code> if no start
	 *         position was defined
	 */
	public Vector<PlayfieldPosition> getPossibleDestinationsInDistance(final int distance) {
		if (stateOfMove == StateOfMove.ENDPOS) {
			return gameState.findAllPossibleDestinationsInDistance(getCurrentMoveStartPos(), distance);
		} else {
			return null;
		}
	}

	/**
	 * Simulates a mouse click somewhere in an empty area on the playfield,
	 * if the current player does not need a positioned click (e.g., computer players).
	 * @see Player#needsPositionOnClick()
	 */
	private void clickAutomaticallyIfPossible() {
		if (getCurrentPlayer().needsPositionOnClick()) {
			// Nothing to do: This player needs clicks.
		} else {
			if (getCurrentOpponent().needsPositionOnClick()) {
				getCurrentPlayer().onClick(null);
			} else {
				// Nothing to do: If both players need no clicks, as a viewer I still want to stop after each half-move to see the progress of the game.
			}
		}
	}

	/**
	 * Does a transition from the current state to the newState. This may start
	 * a cascade of followup status changes, if the new state needs no user
	 * interaction. So the caller should not rely on that after the call of this
	 * method the state really is newState.
	 * 
	 * @param newState the new state the game should transition into
	 */
	private void setState(final StateOfMove newState) {
		if (stateOfMove == newState) {
			return;
		}
		switch (newState) {
		case CHECKING:
			if (stateOfMove == StateOfMove.ENDPOS) {
				gameState.nextPlayer();
			}
			// This state is only transitional, we'll leave in another state.
			stateOfMove = newState;
			lastMoveStartPos = currentMoveStartPos;
			lastMoveEndPos = currentMoveEndPos;
			currentMoveStartPos = null;
			currentMoveEndPos = null;
			if (gameState.canMove()) {
				setState(StateOfMove.STARTPOS);
			} else {
				setState(StateOfMove.GAMEOVER);
			}
			break;
		case ENDPOS:
			if ((null != currentMoveStartPos) && gameState.isLegalStartPos(currentMoveStartPos)) {
				stateOfMove = newState;
			}
			clickAutomaticallyIfPossible();
			break;
		case STARTPOS:
			// Nothing special to do
			stateOfMove = newState;
			clickAutomaticallyIfPossible();
			break;
		case GAMEOVER:
			// Nothing special to do
			stateOfMove = newState;
			break;
		default:
			throw new ArrayIndexOutOfBoundsException();
		}
	}

	private PlayfieldPosition lastMoveStartPos;
	private PlayfieldPosition lastMoveEndPos;
	private PlayfieldPosition currentMoveStartPos;
	private PlayfieldPosition currentMoveEndPos;

	public PlayfieldPosition getLastMoveStartPos() {
		return lastMoveStartPos;
	}

	public PlayfieldPosition getLastMoveEndPos() {
		return lastMoveEndPos;
	}

	public PlayfieldPosition getCurrentMoveStartPos() {
		return currentMoveStartPos;
	}

	public void setCurrentMoveStartPos(final PlayfieldPosition pos) {
		if (stateOfMove == StateOfMove.STARTPOS) {
			if (gameState.isLegalStartPos(pos)) {
				currentMoveStartPos = pos;
				setState(StateOfMove.ENDPOS);
			}
		}
	}

	public void setCurrentMoveEndPos(final PlayfieldPosition pos) {
		if (stateOfMove == StateOfMove.ENDPOS) {
			final GameState newState = gameState.doMove(currentMoveStartPos, pos);
			if (null != newState) {
				gameState = newState;
				currentMoveEndPos = pos;
				setState(StateOfMove.CHECKING);
			} else {
				currentMoveStartPos = null;
				setState(StateOfMove.STARTPOS);
			}
		}
	}

	/**
	 * The game needs to know how to react on a click onto a playfield.
	 * Typically it delegates this to the current player.
	 * 
	 * @param pos
	 *            the playfield position that has been clicked
	 */
	public void onClick(final PlayfieldPosition pos) {
		getCurrentPlayer().onClick(pos);
	}
}
