package de.ped.conquest.logic;

import java.util.ArrayList;
import java.util.Vector;

import de.ped.tools.CollectionUtil;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;

/**
 * Represents the state of the game, that changes from move to move.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameState implements Cloneable {

	/** A game state has a board. */
	private Board board;

	private int currentPlayerIndex = 0;

	public GameState(final Board board) {
		try {
			this.board = (Board) board.clone();
		} catch (final CloneNotSupportedException e) {
			this.board = null;
		}
	}

	protected GameState(final GameState other) {
		try {
			this.board = (Board) other.board.clone();
		} catch (final CloneNotSupportedException e) {
			this.board = null;
		}
		this.currentPlayerIndex = other.currentPlayerIndex;
	}

	@Override
	public GameState clone() {
		return new GameState(this);
	}

	public Board getBoard() {
		return board;
	}

	/** 
	 * @return the index (0 or 1) of the current player 
	 */
	public int getCurrentPlayerIndex() {
		return currentPlayerIndex;
	}

	/**
	 * @return the index (0 or 1) of the opponent of the current player
	 */
	public int getCurrentOpponentIndex() {
		return (1 - currentPlayerIndex);
	}

	/** Sets the next player as the current player. */
	public void nextPlayer() {
		currentPlayerIndex = 1 - currentPlayerIndex;
	}

	// ------------------------------------------------------------------------
	// Calculate all legal moves of the current player,
	// no matter if the player has already halfway decided his move.

	/**
	 * Checks if the current player could move from startpos, but only by
	 * checking for the the state of start playfield.
	 * 
	 * @param startpos
	 *            the playfield position to try to start a move from
	 * @return <code>true</code> if it could be a legal move
	 */
	private boolean isPossibleStartposOnPlayfieldState(final PlayfieldPosition startpos) {
		if (!(board.getPlayfieldValueAt(startpos) == Playfield.owned(currentPlayerIndex))) {
			return false;
		}
		return true;
	}

	/**
	 * Checks if the current player could move to endpos, but only by checking
	 * for the the state of end playfield.
	 * 
	 * @param endpos
	 *            the playfield position to try to end the move
	 * @return <code>true</code> if it could be a legal move
	 */
	private boolean isPossibleDestinationOnPlayfieldState(final PlayfieldPosition endpos) {
		if (board.getPlayfieldValueAt(endpos) != Playfield.FREE) {
			return false;
		}
		return true;
	}

	/**
	 * Checks if the current player could move from startpos to endpos. In
	 * detail, checks for the the state of start and end playfield, and for the
	 * distance between them.
	 * 
	 * @param startpos
	 *            the playfield position to try to start the move from
	 * @param endpos
	 *            the playfield position to try to end the move
	 * @return <code>true</code> if this is a legal move
	 */
	private boolean isValidMove(final PlayfieldPosition startpos, final PlayfieldPosition endpos) {
		if (!isPossibleDestinationOnPlayfieldState(endpos)) {
			return false;
		}
		final int distance = board.calcDistance(startpos, endpos);
		switch (distance) {
		case Game.DISTANCE_GROW:
		case Game.DISTANCE_JUMP:
			return true;
		default:
			return false;
		}
	}

	/**
	 * @param startpos
	 *            the playfield position to start the move from
	 * @param v
	 *            some possible playfield positions to end the move
	 * @return only those elements in the given vector that can be reached
	 *         directly from startpos
	 */
	private Vector<PlayfieldPosition> filterPossibleDestinations(final PlayfieldPosition startpos, final Vector<PlayfieldPosition> v) {
		Vector<PlayfieldPosition> ret = null;
		if (null != v) {
			ret = new Vector<PlayfieldPosition>();
			for (int i = 0; i < v.size(); i++) {
				final PlayfieldPosition endpos = v.get(i);
				if (isPossibleDestinationOnPlayfieldState(endpos)) {
					ret.add(endpos);
				}
			}
		}
		return ret;
	}

	/**
	 * @param startpos
	 *            the playfield position to start the move from
	 * @param distance
	 *            the distance to the start position ({@link Game#DISTANCE_GROW}
	 *            or {@link Game#DISTANCE_JUMP})
	 * @return all legal destinations in the given distance to the given start
	 *         position.
	 */
	public Vector<PlayfieldPosition> findAllPossibleDestinationsInDistance(
			final PlayfieldPosition startpos, 
			final int distance) {
		if (((Game.DISTANCE_GROW == distance) || (Game.DISTANCE_JUMP == distance)) && (isPossibleStartposOnPlayfieldState(startpos))) {
			final Vector<PlayfieldPosition> v = board.getAllPlayfieldsInDistanceTo(startpos, distance);
			return filterPossibleDestinations(startpos, v);
		} else {
			return null;
		}
	}

	/**
	 * @param startpos
	 *            the playfield position to start the move from
	 * @return all legal destinations for the starting position
	 */
	public Vector<PlayfieldPosition> findAllPossibleDestinationsFrom(final PlayfieldPosition startpos) {
		final Vector<PlayfieldPosition> v1 = findAllPossibleDestinationsInDistance(startpos, Game.DISTANCE_GROW);
		final Vector<PlayfieldPosition> v2 = findAllPossibleDestinationsInDistance(startpos, Game.DISTANCE_JUMP);
		if (null != v1) {
			if (null != v2) {
				v1.addAll(v2);
			}
			return v1;
		} else {
			return v2;
		}
	}

	/**
	 * @param startpos
	 *            the playfield position to start the move from
	 * @return <code>true</code> if <code>startpos</code> is a legal starting
	 *         position for a half-move. In detail, checks if the starting
	 *         position is owned by the current player, and if there is at least
	 *         one destination field in reach.
	 */
	public boolean isLegalStartPos(final PlayfieldPosition startpos) {
		final Vector<PlayfieldPosition> vEnd = findAllPossibleDestinationsFrom(startpos);
		return !CollectionUtil.isNullOrEmpty(vEnd);
	}

	public Vector<PlayfieldPosition> findAllPossibleStartPositions() {
		final Vector<PlayfieldPosition> vStart = new Vector<PlayfieldPosition>();
		for (int i = 0; i < board.getNumOfIndices(); i++) {
			final PlayfieldPosition startpos = board.index2position(i);
			if (isLegalStartPos(startpos)) {
				vStart.add(startpos);
			}
		}
		return vStart;
	}

	/**
	 * @param gameMoveFactory
	 *            the game move factory
	 * @return all moves possible from this game state for the current player
	 */
	public ArrayList<GameMove> findAllPossibleMoves(final GameMoveCreateable gameMoveFactory) {
		final ArrayList<GameMove> vMoves = new ArrayList<GameMove>();
		for (int startposIndex = 0; startposIndex < board.getNumOfIndices(); startposIndex++) {
			final PlayfieldPosition startpos = board.index2position(startposIndex);
			final Vector<PlayfieldPosition> v1 = findAllPossibleDestinationsInDistance(startpos, Game.DISTANCE_GROW);
			final Vector<PlayfieldPosition> v2 = findAllPossibleDestinationsInDistance(startpos, Game.DISTANCE_JUMP);
			if (null != v1) {
				for (int iDests = 0; iDests < v1.size(); iDests++) {
					vMoves.add(gameMoveFactory.createGameMove(this, startpos, v1.get(iDests), Game.DISTANCE_GROW));
				}
			}
			if (null != v2) {
				for (int iDests = 0; iDests < v2.size(); iDests++) {
					vMoves.add(gameMoveFactory.createGameMove(this, startpos, v2.get(iDests), Game.DISTANCE_JUMP));
				}
			}
		}
		return vMoves;
	}

	/**
	 * @return the number of all moves possible from this game state for the
	 *         current player
	 */
	public int calcNumOfPossibleMoves() {
		int ret = 0;
		for (int i = 0; i < board.getNumOfIndices(); i++) {
			final PlayfieldPosition startpos = board.index2position(i);
			final Vector<PlayfieldPosition> v1 = findAllPossibleDestinationsInDistance(startpos, Game.DISTANCE_GROW);
			final Vector<PlayfieldPosition> v2 = findAllPossibleDestinationsInDistance(startpos, Game.DISTANCE_JUMP);
			if (null != v1) {
				ret += v1.size();
			}
			if (null != v2) {
				ret += v2.size();
			}
		}
		return ret;
	}

	/**
	 * Checks if the current player is able to do any move at all.
	 * 
	 * @return <code>true</code> if the player is able to move
	 */
	public boolean canMove() {
		final PlayfieldDimension size = board.getSize();
		final PlayfieldPosition startpos = new PlayfieldPosition(); // re-used for performance reasons
		for (int y = 0; y < size.getHeight(); y++) {
			for (int x = 0; x < size.getWidth(); x++) {
				startpos.setLocation(x, y);
				if (isLegalStartPos(startpos)) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Executes the given move, if possible. Be aware: If the move was not
	 * possible, the move may nevertheless be half-done and this may represent a
	 * state that should not be possible when following the game rules.
	 * 
	 * @param startpos
	 *            the playfield position to start the move from
	 * @param endpos
	 *            the playfield position where to end the move
	 * @return <code>true</code> if the move was done successfully
	 */
	private boolean isMoveDone(final PlayfieldPosition startpos, final PlayfieldPosition endpos) {
		if (!isValidMove(startpos, endpos)) {
			return false;
		}
		final int distance = board.calcDistance(startpos, endpos);
		switch (distance) {
		case 1: // Grow
			// start field stays the same
			break;
		case 2: // Move
			board.setPlayfieldValueAt(startpos, Playfield.FREE);
			break;
		default:
			return false;
		}
		// Set destination field
		board.setPlayfieldValueAt(endpos, Playfield.owned(currentPlayerIndex));
		// Set the neighbors
		final Vector<PlayfieldPosition> v = board.getAllPlayfieldsInDistanceTo(endpos, 1);
		for (final PlayfieldPosition p : v) {
			final int val = board.getPlayfieldValueAt(p);
			if ((val != Playfield.WALL) && (val != Playfield.FREE)) {
				board.setPlayfieldValueAt(p, Playfield.owned(currentPlayerIndex));
			}
		}
		return true;
	}

	/**
	 * Executes (if allowed) the given move for the current player.
	 * 
	 * @param startpos
	 *            the playfield position to start the move from
	 * @param endpos
	 *            the playfield position where to end the move
	 * @return A new game state, if the move was allowed. <code>null</code>
	 *         otherwise
	 */
	public GameState doMove(final PlayfieldPosition startpos, final PlayfieldPosition endpos) {
		GameState newState = new GameState(this);
		final boolean isDone = newState.isMoveDone(startpos, endpos);
		if (!isDone) {
			newState = null;
		}
		return newState;
	}

	@Override
	public String toString() {
		final StringBuffer s = new StringBuffer();
		s.append(currentPlayerIndex);
		s.append(',');
		s.append(board.toString());
		return s.toString();
	}

	/**
	 * @param o
	 *            the object to compare with
	 * @return <code>true</code> if the boards are identical, and the index of the
	 *         current player is the same
	 */
	@Override
	public boolean equals(final Object o) {
		if (o instanceof GameState) {
			final GameState other = (GameState) o;
			return ((currentPlayerIndex == other.currentPlayerIndex) && (board.equals(other.board)));
		}
		return false;
	}

	@Override
	public int hashCode() {
		final int mask = (currentPlayerIndex == 0 ? 0x00 : 0xff);
		return (board.hashCode() ^ mask);
	}
}
