package de.ped.conquest.logic;

import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Abstract class for all strategies.
 * Main goal is to deliver a legal half-move.
 * Each subclass can decide if to wait for user input or find a move on its own.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Player {
	protected final Logger logger = Logger.getLogger(getClass()); {
		logger.setLogLevel(Level.WARN);
	}

	/** The game the player is in. */
	private Game game;

	protected Game getGame() {
		return game;
	}

	private void setGame(final Game game) {
		this.game = game;
	}

	/**
	 * @return if the player takes hints (clicks on positions) from the human
	 *         interface. Default implementation is <code>false</code> . Any
	 *         strategy with human interaction should overload this method and
	 *         return <code>true</code>.
	 */
	public boolean needsPositionOnClick() {
		return false;
	}

	/**
	 * The user has clicked a playfield on the board. What happens now? This
	 * depends from the state of the game (is start or end position of a move
	 * expected?)
	 * 
	 * @param pos
	 *            the playfield position that was clicked
	 */
	public void onClick(final PlayfieldPosition pos) {
		switch (game.getStateOfMove()) {
		case STARTPOS:
			game.setCurrentMoveStartPos(suggestStartPos());
			break;
		case ENDPOS:
			game.setCurrentMoveEndPos(suggestEndPos());
			break;
		default:
			logger.debug("Player.onClick: game.getState()==" + game.getStateOfMove());
			// Nothing to do with clicking...
		}
	}

	/**
	 * @return the start position of this player's move
	 */
	protected abstract PlayfieldPosition suggestStartPos();

	/**
	 * @return the end position of this player's move
	 */
	protected abstract PlayfieldPosition suggestEndPos();

	public abstract String getPlayerTypeName();

	/**
	 * Type parameter for the parameterized factory method.
	 * @see #createPlayer
	 */
	public static final int PLAYER_TYPE_HUMAN = 0;

	/**
	 * Type parameter for the parameterized factory method.
	 * @see #createPlayer
	 */
	public static final int PLAYER_TYPE_COMPUTER_RANDOM = 1;

	/**
	 * Type parameter for the parameterized factory method.
	 * @see #createPlayer
	 */
	public static final int PLAYER_TYPE_COMPUTER_SIMPLE = 2;

	/**
	 * Type parameter for the parameterized factory method.
	 * @see #createPlayer
	 */
	public static final int PLAYER_TYPE_COMPUTER_MINIMAX = 3;

	static String STRATEGY_NAME_HUMAN = "Game.Player.Human";
	static String STRATEGY_NAME_RANDOM = "Game.Player.ComputerRandom";
	static String STRATEGY_NAME_GREEDY = "Game.Player.ComputerSimple";
	static String STRATEGY_NAME_MINIMAX = "Game.Player.ComputerMinimax";

	/**
	 * Human readable list of all possible strategies in the order in which to present it to the user to pick one.
	 * The strings returned are message keys for internationalization.
	 * @see #createPlayer
	 */
	private static final String[] namesOfStrategies = {
		STRATEGY_NAME_HUMAN,
		STRATEGY_NAME_RANDOM,
		STRATEGY_NAME_GREEDY,
		STRATEGY_NAME_MINIMAX,
	};

	/**
	 * @return a human readable list of all possible strategies in the order in
	 *         which to present it to the user to pick one. The strings returned
	 *         are message keys for internationalization.
	 * @see #createPlayer
	 */
	public static final String[] getStrategies() {
		return namesOfStrategies;
	}

	/**
	 * Parameterized factory method to create a player.
	 * 
	 * @param game
	 *            the game to create the player in
	 * @param strategy
	 *            the strategy the player will follow. One of
	 *            {@link #PLAYER_TYPE_COMPUTER_MINIMAX},
	 *            {@link #PLAYER_TYPE_COMPUTER_SIMPLE},
	 *            {@link #PLAYER_TYPE_COMPUTER_RANDOM} or
	 *            {@link #PLAYER_TYPE_HUMAN}.
	 * @return the new player
	 */
	public static Player createPlayer(final Game game, final int strategy) {
		Player player;
		switch (strategy) {
		case PLAYER_TYPE_COMPUTER_MINIMAX:
			player = new ComputerPlayerMinimax();
			break;
		case PLAYER_TYPE_COMPUTER_SIMPLE:
			player = new ComputerPlayerGreedy();
			break;
		case PLAYER_TYPE_COMPUTER_RANDOM:
			player = new ComputerPlayerRandom();
			break;
		case PLAYER_TYPE_HUMAN:
			player = new HumanPlayer();
			break;
		default:
			throw new IllegalArgumentException("strategy out of range!");
		}
		player.setGame(game);
		return player;
	}
}
