package de.ped.conquest.test;

import static org.junit.Assert.*;

import java.util.Vector;

import org.junit.Test;

import de.ped.conquest.logic.Board;
import de.ped.conquest.logic.BoardFactory;
import de.ped.conquest.logic.Playfield;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;
import de.ped.tools.test.GenericTest;

/**
 * Automated unit tests for the {@link Board} class and its subclasses.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class BoardTest {

	public static void validateBoard(final Board board) {
		final int idxs = board.getNumOfIndices();
		// Test transformation from index to position.
		for (int idx = 0; idx < idxs; idx++) {
			PlayfieldPosition pos = board.index2position(idx);
			assertEquals(idx, board.position2index(pos));
			assertEquals(board.getPlayfieldValueAt(idx), board.getPlayfieldValueAt(pos));
		}
		board.toString(); // Tests the toString() method, no need to store the
							// result
		final int maxNumOfNeighbors = board.getMaxNumOfNeighbors();
		for (int idx = 0; idx < idxs; idx++) {
			final PlayfieldPosition pos = board.index2position(idx);
			final Vector<PlayfieldPosition> v1 = board.getAllPlayfieldsInDistanceTo(pos, 1);
			assertTrue(v1.size() <= maxNumOfNeighbors);
			final Vector<PlayfieldPosition> v2 = board.getAllPlayfieldsInDistanceTo(pos, 2);
			assertTrue(v2.size() <= 2 * maxNumOfNeighbors);
		}
	}

	@Test
	public void testNullSquaresBoard() {
		Board nullBoard = new Board(TopologyMode.SQUARE4, 0, 0);
		assertEquals(0, nullBoard.getSize().getWidth());
		assertEquals(0, nullBoard.getSize().getHeight());
		validateBoard(nullBoard);
		try {
			nullBoard.setPlayfieldValueAt(0, 0);
			fail();
		} catch (final ArrayIndexOutOfBoundsException ex) {
			// That's what we wanted
		}
	}

	@Test
	public void testSquaresBoard4() {
		Board board = new Board(TopologyMode.SQUARE4, 3, 4);
		assertEquals(3, board.getSize().getWidth());
		assertEquals(4, board.getSize().getHeight());
		validateBoard(board);
		board.setPlayfieldValueAt(new PlayfieldPosition(1, 1), Playfield.owned(0));
		assertEquals(1, board.countPlayfieldsOfValue(Playfield.owned(0)));
		assertEquals(3 * 4 - 1, board.countPlayfieldsOfValue(Playfield.FREE));
		assertEquals(1, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 0).size());
		assertEquals(4, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 1).size());
		assertEquals(5, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 2).size());
		assertEquals(2, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 3).size());
		assertEquals(2, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 1).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 2).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 1), 1).size());
	}

	@Test
	public void testSquaresBoard8() {
		Board board = new Board(TopologyMode.SQUARE8, 3, 4);
		assertEquals(3, board.getSize().getWidth());
		assertEquals(4, board.getSize().getHeight());
		validateBoard(board);
		board.setPlayfieldValueAt(new PlayfieldPosition(1, 1), Playfield.owned(0));
		assertEquals(1, board.countPlayfieldsOfValue(Playfield.owned(0)));
		assertEquals(3 * 4 - 1, board.countPlayfieldsOfValue(Playfield.FREE));
		assertEquals(1, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 0).size());
		assertEquals(8, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 1).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 2).size());
		assertEquals(0, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 3).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 1).size());
		assertEquals(5, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 2).size());
		assertEquals(5, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 1), 1).size());
	}

	@Test
	public void testHexagonsBoard() {
		Board board = new Board(TopologyMode.HEXAGON, 3, 4);
		assertEquals(3, board.getSize().getWidth());
		assertEquals(4, board.getSize().getHeight());
		validateBoard(board);
		validateBoard(board); // This is a kind of test that validating has no side effects.
		board.setPlayfieldValueAt(new PlayfieldPosition(1, 1), Playfield.owned(0));
		assertEquals(1, board.countPlayfieldsOfValue(Playfield.owned(0)));
		assertEquals(3 * 4 - 1, board.countPlayfieldsOfValue(Playfield.FREE));
		assertEquals(1, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 0).size());
		assertEquals(6, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 1).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 2).size());
		assertEquals(2, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 3).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 1).size());
		assertEquals(4, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 2).size());
		assertEquals(4, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 1), 1).size());
	}

	@Test
	public void testTrianglesBoard() {
		Board board = new Board(TopologyMode.TRIANGLE, 3, 4);
		assertEquals(3, board.getSize().getWidth());
		assertEquals(4, board.getSize().getHeight());
		validateBoard(board);
		board.setPlayfieldValueAt(new PlayfieldPosition(1, 1), Playfield.owned(0));
		assertEquals(1, board.countPlayfieldsOfValue(Playfield.owned(0)));
		assertEquals(3 * 4 - 1, board.countPlayfieldsOfValue(Playfield.FREE));
		assertEquals(1, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 0).size());
		assertEquals(3, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 1).size());
		assertEquals(4, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 2).size());
		assertEquals(1, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(1, 1), 3).size());
		assertEquals(1, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 1).size());
		assertEquals(2, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 0), 2).size());
		assertEquals(2, board.getAllPlayfieldsInDistanceTo(new PlayfieldPosition(0, 1), 1).size());
	}

	@Test
	public void testEqualsInPlayfields() {
		final BoardFactory boardFactory = new BoardFactory();
		Board board1 = boardFactory.createBoard(6);
		Board board2;
		try {
			board2 = (Board) board1.clone();
		} catch (final CloneNotSupportedException e) {
			board2 = null;
		}
		assertEquals(board1, board2);
		assertEquals(board1.hashCode(), board2.hashCode());
		assertEquals(Playfield.WALL, board2.getPlayfieldValueAt(0));
		board2.setPlayfieldValueAt(0, Playfield.FREE);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsSquare() {
		Board board1 = new Board(TopologyMode.SQUARE8, 3, 4);
		Board board2 = new Board(TopologyMode.SQUARE8, 3, 4);
		assertEquals(board1, board2);
		assertEquals(board1.hashCode(), board2.hashCode());
	}

	@Test
	public void testEqualsSquareMaxX() {
		Board board1 = new Board(TopologyMode.SQUARE8, 3, 4);
		Board board2 = new Board(TopologyMode.SQUARE8, 4, 4);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsSquareMaxY() {
		Board board1 = new Board(TopologyMode.SQUARE8, 3, 4);
		Board board2 = new Board(TopologyMode.SQUARE8, 3, 3);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsSquareIsEightNeigbors() {
		Board board1 = new Board(TopologyMode.SQUARE8, 3, 4);
		Board board2 = new Board(TopologyMode.SQUARE4, 3, 4);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsTriangles() {
		Board board1 = new Board(TopologyMode.TRIANGLE, 3, 4);
		Board board2 = new Board(TopologyMode.TRIANGLE, 3, 4);
		assertEquals(board1, board2);
		assertEquals(board1.hashCode(), board2.hashCode());
	}

	@Test
	public void testEqualsTrianglesMaxX() {
		Board board1 = new Board(TopologyMode.TRIANGLE, 3, 4);
		Board board2 = new Board(TopologyMode.TRIANGLE, 4, 4);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsTrianglesMaxY() {
		Board board1 = new Board(TopologyMode.TRIANGLE, 3, 4);
		Board board2 = new Board(TopologyMode.TRIANGLE, 3, 3);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsHexagons() {
		Board board1 = new Board(TopologyMode.HEXAGON, 3, 4);
		Board board2 = new Board(TopologyMode.HEXAGON, 3, 4);
		assertEquals(board1, board2);
		assertEquals(board1.hashCode(), board2.hashCode());
	}

	@Test
	public void testEqualsHexagonsMaxX() {
		Board board1 = new Board(TopologyMode.HEXAGON, 3, 4);
		Board board2 = new Board(TopologyMode.HEXAGON, 4, 4);
		assertNotSame(board1, board2);
	}

	@Test
	public void testEqualsHexagonsMaxY() {
		Board board1 = new Board(TopologyMode.HEXAGON, 3, 4);
		Board board2 = new Board(TopologyMode.HEXAGON, 3, 3);
		assertNotSame(board1, board2);
	}

	/**
	 * @see Board#equals(Object)
	 * @see Board#clone()
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(true, new Board(TopologyMode.HEXAGON, 3, 4), new Board(TopologyMode.HEXAGON, 3, 3));
		GenericTest.checkObjectBasics(true, new Board(TopologyMode.SQUARE8, 3, 4), new Board(TopologyMode.SQUARE4, 3, 4));
		GenericTest.checkObjectBasics(true, new Board(TopologyMode.TRIANGLE, 3, 4), new Board(TopologyMode.TRIANGLE, 3, 3));
	}
}
