package de.ped.conquest.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotSame;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Arrays;

import org.junit.BeforeClass;
import org.junit.Test;

import de.ped.conquest.logic.Board;
import de.ped.conquest.logic.BoardFactory;
import de.ped.conquest.logic.ComputerPlayerMinimax;
import de.ped.conquest.logic.Game;
import de.ped.conquest.logic.GameState;
import de.ped.conquest.logic.Player;
import de.ped.conquest.logic.Playfield;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Tests {@link ComputerPlayerMinimax}.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ComputerPlayerMinimaxTest {
	private static final double EPSILON = 1e-5;

	private final Logger logger = Logger.getLogger(this.getClass());
	{
		logger.setLogLevel(Level.INFO);
	}

	private void startTest(final String testName) {
		logger.debug("--- Starting " + testName + " ---");
		this.testName = testName;
	}

	@BeforeClass
	public static void setUp() {
		boardFactory = new BoardFactory();
	}
	private static BoardFactory boardFactory;

	@Test
	public void testOnlyOneMovePossible() {
		startTest("testOnlyOneMovePossible");
		final Board b = boardFactory.createBoard(5);
		BoardTest.validateBoard(b);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		final Player player = game.getCurrentPlayer();
		final ComputerPlayerMinimax p = (ComputerPlayerMinimax) player;
		// Do we find the only possible move?
		final int numOfMoves = game.getGameState().calcNumOfPossibleMoves();
		assertEquals(1, numOfMoves);
		final ArrayList<ComputerPlayerMinimax.RatedMove> moves = p.calcPossibleMoves(game.getGameState());
		assertEquals(1, moves.size());
		final PlayfieldPosition start = moves.get(0).start;
		final PlayfieldPosition end = moves.get(0).end;
		assertEquals(0, start.getX());
		assertEquals(0, start.getY());
		assertEquals(1, end.getX());
		assertEquals(0, end.getY());
		{ // Rate that move - step by step
			final GameState won = game.getGameState().doMove(start, end);
			won.nextPlayer();
			assertEquals(1, won.getCurrentPlayerIndex());
			final double rating = p.calcStateRating(won);
			assertEquals(ComputerPlayerMinimax.RATING_LOST, rating, EPSILON);
		}
		{ // Rate that move - now in one go
			final double rating = p.calcMoveRating(moves.get(0));
			assertEquals(ComputerPlayerMinimax.RATING_WON, rating, EPSILON);
		}
		{ // Find the best move
			final ComputerPlayerMinimax.RatedMove move = p.findBestNextMove(game.getGameState(), 0);
			assertEquals(0, move.start.getX());
			assertEquals(0, move.start.getY());
			assertEquals(1, move.end.getX());
			assertEquals(0, move.end.getY());
		}
		logger.debug(game.getGameState().getBoard().topologyLookupToString());
	}

	@Test
	public void testOneMoveToWin() {
		startTest("testOneMoveToWin");
		final Board b = boardFactory.createBoard(6);
		BoardTest.validateBoard(b);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		final Player player = game.getCurrentPlayer();
		final ComputerPlayerMinimax p = (ComputerPlayerMinimax) player;
		final int numOfMoves = game.getGameState().calcNumOfPossibleMoves();
		assertEquals(2, numOfMoves);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		assertEquals(b, game.getGameState().getBoard());
		{ // Rate the moves
			final ArrayList<ComputerPlayerMinimax.RatedMove> moves = p.calcPossibleMoves(game.getGameState());
			assertEquals(2, moves.size());
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			final double rating1 = p.calcMoveRating(moves.get(0));
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			final double rating2 = p.calcMoveRating(moves.get(1));
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			if (rating1 == ComputerPlayerMinimax.RATING_WON) {
				assertEquals(-2, rating2, EPSILON);
			} else if (rating2 == ComputerPlayerMinimax.RATING_WON) {
				assertEquals(-2, rating1, EPSILON);
			} else {
				assertTrue(false);
			}
		}
		// Do we find the winning move?
		final ArrayList<ComputerPlayerMinimax.RatedMove> moves = p.calcPossibleMoves(game.getGameState());
		assertEquals(2, moves.size());
		ComputerPlayerMinimax.RatedMove move = p.findBestNextMove(game.getGameState(), 0);
		assertEquals(2, move.start.getX());
		assertEquals(1, move.start.getY());
		assertEquals(1, move.end.getX());
		assertEquals(1, move.end.getY());
		assertEquals(ComputerPlayerMinimax.RATING_WON, p.calcMoveRating(move), EPSILON);
		assertEquals(ComputerPlayerMinimax.RATING_WON, move.rating, EPSILON);
		logger.debug(game.getGameState().getBoard().topologyLookupToString());
	}

	@Test
	public void testBoardSquares4N2x3FullGameWithMinimaxDepth0() {
		startTest("testBoardSquares4N2x3FullGameWithMinimaxDepth0");
		final Board b = boardFactory.createBoard(7);
		BoardTest.validateBoard(b);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		final Player player = game.getCurrentPlayer();
		final ComputerPlayerMinimax p = (ComputerPlayerMinimax) player;
		final int numOfMoves = game.getGameState().calcNumOfPossibleMoves();
		assertEquals(3, numOfMoves);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		assertEquals(b, game.getGameState().getBoard());
		{ // Rate the moves
			final ArrayList<ComputerPlayerMinimax.RatedMove> moves = p.calcPossibleMoves(game.getGameState());
			assertEquals(3, moves.size());
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			final double[] ratings = new double[moves.size()];
			for (int i = 0; i < moves.size(); i++) {
				ratings[i] = p.calcMoveRating(moves.get(i));
			}
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			Arrays.sort(ratings);
			assertEquals(-3, ratings[0], EPSILON);
			assertEquals(-2, ratings[1], EPSILON);
			assertEquals(-2, ratings[2], EPSILON);
		}
		logger.debug(game.getGameState().getBoard().topologyLookupToString());
	}

	@Test
	public void testBoardSquares4N2x3FullGameWithMinimaxDepth2() {
		startTest("testBoardSquares4N2x3FullGameWithMinimaxDepth2");
		final Board b = boardFactory.createBoard(7);
		BoardTest.validateBoard(b);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		final Player player = game.getCurrentPlayer();
		ComputerPlayerMinimax p = (ComputerPlayerMinimax) player;
		p.setMaxDepth(2);
		final int numOfMoves = game.getGameState().calcNumOfPossibleMoves();
		assertEquals(3, numOfMoves);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		assertEquals(b, game.getGameState().getBoard());
		{ // Rate the moves
			final ArrayList<ComputerPlayerMinimax.RatedMove> moves = p.calcPossibleMoves(game.getGameState());
			assertEquals(3, moves.size());
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			final double[] ratings = new double[moves.size()];
			for (int i = 0; i < moves.size(); i++) {
				ratings[i] = p.calcMoveRating(moves.get(i));
			}
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			Arrays.sort(ratings);
			assertEquals(-3, ratings[0], EPSILON);
			assertEquals(-2, ratings[1], EPSILON);
			assertEquals(-2, ratings[2], EPSILON);
		}
		final ComputerPlayerMinimax.RatedMove move = p.findBestNextMove(game.getGameState(), 0);
		assertTrue(ComputerPlayerMinimax.RATING_LOST < move.rating);
	}

	@Test
	public void testBoardSquares4N2x3FullGameBug1() {
		startTest("testBoardSquares4N2x3FullGameBug1");
		final Board b = new Board(TopologyMode.SQUARE4, 3, 2);
		b.setPlayfieldValueAt(new PlayfieldPosition(0, 0), Playfield.owned(1));
		b.setPlayfieldValueAt(new PlayfieldPosition(0, 1), Playfield.owned(1));
		b.setPlayfieldValueAt(new PlayfieldPosition(1, 1), Playfield.owned(1));
		b.setPlayfieldValueAt(new PlayfieldPosition(1, 0), Playfield.owned(0));
		BoardTest.validateBoard(b);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		final Player player = game.getCurrentPlayer();
		final ComputerPlayerMinimax p = (ComputerPlayerMinimax) player;
		p.setMaxDepth(2);
		final int numOfMoves = game.getGameState().calcNumOfPossibleMoves();
		assertEquals(2, numOfMoves);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		assertEquals(b, game.getGameState().getBoard());
		{ // Rate the moves, without looking ahead. In this case, the wrong move would be selected.
			final ArrayList<ComputerPlayerMinimax.RatedMove> moves = p.calcPossibleMoves(game.getGameState());
			assertEquals(2, moves.size());
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			final double[] ratings = new double[moves.size()];
			for (int i = 0; i < moves.size(); i++) {
				ratings[i] = p.calcMoveRating(moves.get(i));
			}
			assertEquals(0, game.getGameState().getCurrentPlayerIndex());
			assertEquals(b, game.getGameState().getBoard());
			if ((ratings[0] == -3) && (ratings[1] == -2)) {
				assertEquals(0, moves.get(1).end.getY());
			} else if ((ratings[0] == -2) && (ratings[1] == -3)) {
				assertEquals(0, moves.get(0).end.getY());
			} else {
				Arrays.sort(ratings);
				assertEquals(-3, ratings[0], EPSILON);
				assertEquals(-2, ratings[1], EPSILON);
			}
		}
		final ComputerPlayerMinimax.RatedMove move = p.findBestNextMove(game.getGameState(), 0);
		assertEquals(1, move.start.getX());
		assertEquals(0, move.start.getY());
		assertEquals(2, move.end.getX());
		assertEquals(1, move.end.getY());
		assertTrue(ComputerPlayerMinimax.RATING_LOST < move.rating);
	}

	@Test
	public void testBoardSquares4N2x3FullGameBug2() {
		startTest("testBoardSquares4N2x3FullGameBug2");
		final Board b = new Board(TopologyMode.SQUARE4, 3, 2);
		b.setPlayfieldValueAt(new PlayfieldPosition(2, 0), Playfield.owned(1));
		b.setPlayfieldValueAt(new PlayfieldPosition(0, 1), Playfield.owned(1));
		b.setPlayfieldValueAt(new PlayfieldPosition(2, 1), Playfield.owned(1));
		b.setPlayfieldValueAt(new PlayfieldPosition(1, 1), Playfield.owned(0));
		BoardTest.validateBoard(b);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		final Player player = game.getCurrentPlayer();
		ComputerPlayerMinimax p = (ComputerPlayerMinimax) player;
		final int numOfMoves = game.getGameState().calcNumOfPossibleMoves();
		assertEquals(2, numOfMoves);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		assertEquals(b, game.getGameState().getBoard());
		final ComputerPlayerMinimax.RatedMove move = p.findBestNextMove(game.getGameState(), 0);
		assertEquals(1, move.start.getX());
		assertEquals(1, move.start.getY());
		assertEquals(0, move.end.getX());
		assertEquals(0, move.end.getY());
		assertTrue(ComputerPlayerMinimax.RATING_LOST < move.rating);
	}

	@Test
	public void testBoardSquares4N2x3FullGameEndless() {
		startTest("testBoardSquares4N2x3FullGameEndless");
		final Board b = boardFactory.createBoard(7);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		GameState state = game.getGameState();
		final ComputerPlayerMinimax p = (ComputerPlayerMinimax) game.getCurrentPlayer();
		final ComputerPlayerMinimax p2 = (ComputerPlayerMinimax) game.getCurrentOpponent();
		for (int depth = 1; depth < 5; depth++) {
			p.setMaxDepth(depth);
			p2.setMaxDepth(depth);
			for (int i = 0; i < 50; i++) {
				assertEquals(Game.StateOfMove.STARTPOS, game.getStateOfMove());
				game.getCurrentPlayer().onClick(null);
				assertEquals(Game.StateOfMove.ENDPOS, game.getStateOfMove());
				game.getCurrentPlayer().onClick(null);
				GameState nextState = game.getGameState();
				assertNotSame(state, nextState);
				state = nextState;
			}
			assertEquals(Game.StateOfMove.STARTPOS, game.getStateOfMove());
		}
	}

	@Test
	public void testBoardSquares8N10x10CenterHoleFullGameEndless() {
		startTest("testBoardSquares8N10x10CenterHoleFullGameEndless");
		final long startTime = System.currentTimeMillis();
		final Board b = boardFactory.createBoard(1);
		final Game game = new Game(b, Player.PLAYER_TYPE_COMPUTER_MINIMAX, Player.PLAYER_TYPE_COMPUTER_MINIMAX);
		assertEquals(0, game.getGameState().getCurrentPlayerIndex());
		GameState state = game.getGameState();
		final ComputerPlayerMinimax p = (ComputerPlayerMinimax) game.getCurrentPlayer();
		final ComputerPlayerMinimax p2 = (ComputerPlayerMinimax) game.getCurrentOpponent();
		final int depth = 4;
		p.setMaxDepth(depth);
		p2.setMaxDepth(depth);
		assertEquals(Game.StateOfMove.STARTPOS, game.getStateOfMove());
		game.getCurrentPlayer().onClick(null);
		assertEquals(Game.StateOfMove.ENDPOS, game.getStateOfMove());
		game.getCurrentPlayer().onClick(null);
		final GameState nextState = game.getGameState();
		assertNotSame(state, nextState);
		state = nextState;
		final long endTime = System.currentTimeMillis();
		logger.info("Game " + testName + " took " + (endTime - startTime) + " ms.");
	}
	
	private String testName;
}
