package de.ped.deinbac.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.Locale;

import javax.swing.ButtonGroup;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.FinishableThread;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.ResourceFinder.Folder;
import de.ped.tools.TextFormatter;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.CheckBoxMenuItemAction;
import de.ped.tools.gui.GUIAlgorithms;
import de.ped.tools.gui.GameThread;
import de.ped.tools.gui.HtmlPaneDialog;
import de.ped.tools.gui.JHtmlPane;
import de.ped.tools.gui.JScrollAndZoomPane;
import de.ped.tools.gui.PedAction;
import de.ped.tools.gui.PedJMenu;
import de.ped.deinbac.logic.Bacterion;
import de.ped.deinbac.logic.DeinbacTrendChartEntry;
import de.ped.deinbac.logic.GameModel;
import de.ped.deinbac.logic.GameProperties;
import de.ped.deinbac.logic.SimulationType;

/**
 * The main window of the application.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DeinbacMainWindow extends ApplicationMainWindow {
	private static final long serialVersionUID = 1L;

	/**
	 * The action to set the simulation speed.
	 * 
	 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class SetSpeedAction extends PedAction {
		private static final long serialVersionUID = 1L;
		public final int index;
		public final int sleepDelay;
		public final boolean isRunning;

		public SetSpeedAction(final int index, final boolean isRunning, final int sleepDelay, final String key, final Folder folder) {
			super(key, DeinbacMainWindow.this, folder);
			this.index = index;
			this.sleepDelay = sleepDelay;
			this.isRunning = isRunning;
		}
	}

	/** 
	 * The action to show the genealogy.
	 *
	 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class SingleStepAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public SingleStepAction() {
			super("Action.SetSpeedSingle", DeinbacMainWindow.this, ResourceFinder.Folder.APP);
		}

		@Override
		public boolean isToBeEnabled() {
			return (!setSpeedActions[speedIndex].isRunning);
		}
	}

	/** 
	 * The action to change the zoom level of the screen.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewZoomLevelAction extends PedAction {
		private static final long serialVersionUID = 1L;
		public final int index;

		public ViewZoomLevelAction(final int i) {
			super("Action.ViewZoomLevel.Level" + i, DeinbacMainWindow.this, ResourceFinder.Folder.COMMON, false);
			index = i;
		}
	}

	/** 
	 * The action to show the trend chart.
	 *
	 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ShowTrendChartAction extends CheckBoxMenuItemAction {
		private static final long serialVersionUID = 1L;

		public ShowTrendChartAction() {
			super(false, "Action.ShowTrendChart", DeinbacMainWindow.this, ResourceFinder.Folder.APP);
		}

		@Override
		public boolean isToBeEnabled() {
			return ((null != model) && (model.getProps().getSimulation() == SimulationType.DEINBAC) 
					&& (null != model.deinbac) 
					&& (0 < model.deinbac.getProps().getTrendChartFrequency()));
		}
	}

	/** 
	 * The action to show the genealogy.
	 *
	 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ShowGenealogyAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ShowGenealogyAction() {
			super("Action.ShowGenealogy", DeinbacMainWindow.this, ResourceFinder.Folder.APP);
		}

		@Override
		public boolean isToBeEnabled() {
			return ((null != model) && (model.getProps().getSimulation() == SimulationType.DEINBAC) 
					&& (null != model.deinbac));
		}
	}

	public DeinbacMainWindow(final ApplicationEnvironment appEnv) {
		super(appEnv);
		for (int i = 0; i < viewZoomLevelActions.length; i++) {
			viewZoomLevelActions[i] = new ViewZoomLevelAction(i);
		}
		// Update menu status
		model = new GameModel(props);
		getContentPane().setLayout(new BorderLayout());
		worldCanvas = new DeinbacMainCanvas(this);
		statusText = new JTextField();
		statusText.setEditable(false);
		worldScrollPane = new JScrollAndZoomPane(this, worldCanvas);
		getContentPane().add(worldScrollPane, BorderLayout.CENTER);
		getContentPane().add(statusText, BorderLayout.SOUTH);
		showWorldAction.setOn(worldCanvas.isDisplayed());
		showTrendChartAction.setOn(false);
		setJMenuBar(createMenuBar());
		updateZoomLevelSubmenuSelected();
		updateStatusText();
		pack(); // Resize the window to the preferred size of its components.
		switchLocale(getMessages().getCurrentLocale());
		pauseOrResume(true);
		updateAllActionsEnabled();
	}

	private void updateStatusText() {
		final String key = (isAlive ? "Status.Alive" : "Status.Dead");
		final String isAliveStr = getMessages().getText(key);
		statusText.setText(getMessages().getText("Status", "%0", new String[] { Long.toString(model.getTime()), isAliveStr }));
	}

	private JMenuBar createMenuBar() {
		final ButtonGroup speedButtonGroup = new ButtonGroup();
		final JMenuBar menuBar = createJMenuBar();
		final JMenu menuFile = createJMenu("Menu.File");
		createActionItem(menuFile, startNewGameAction);
		menuFile.addSeparator();
		menuFile.add(createLocalesMenu());
		menuFile.addSeparator();
		createActionItem(menuFile, quitAction);
		menuBar.add(menuFile);
		final JMenu menuRun = createJMenu("Menu.Run");
		for (int i = 0; i < setSpeedActions.length; i++) {
			final JRadioButtonMenuItem button = createRadioButtonActionItem(menuRun, setSpeedActions[i], speedButtonGroup);
			if (i == 0) {
				button.setSelected(true);
				createActionItem(menuRun, singleStepAction);
			}
			speedButtons[i] = button;
		}
		menuRun.addSeparator();
		createActionItem(menuRun, showWorldAction);
		final PedJMenu zoomLevelsSubmenu = createJMenu("Menu.ViewZoomLevel", ResourceFinder.Folder.COMMON);
		final ButtonGroup zoomButtonGroup = new ButtonGroup();
		for (final ViewZoomLevelAction action : viewZoomLevelActions) {
			zoomLevelButtons[action.index] = createRadioButtonActionItem(zoomLevelsSubmenu, action, zoomButtonGroup);
		}
		menuRun.add(zoomLevelsSubmenu);
		createActionItem(menuRun, showTrendChartAction);
		menuRun.addSeparator();
		createActionItem(menuRun, showGenealogyAction);
		menuBar.add(menuRun);
		final JMenu menuHelp = createJMenu("Menu.Help");
		createActionItem(menuHelp, helpAction);
		createActionItem(menuHelp, aboutAction);
		menuBar.add(menuHelp);
		try {
			menuBar.setHelpMenu(menuHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}
		return menuBar;
	}

	private void updateSpeedButtons() {
		for (int i = 0; i < speedButtons.length; i++) {
			speedButtons[i].setSelected(i == speedIndex);
		}
	}

	private void updateAllActionsEnabled() {
		final boolean isToBeEnabled = showTrendChartAction.isToBeEnabled();
		if (!isToBeEnabled) {
			showTrendChart(false);
			showTrendChartAction.setOn(false);
		}
		showTrendChartAction.updateEnabled();
		showGenealogyAction.updateEnabled();
		singleStepAction.updateEnabled();
	}

	@Override
	public void pedAbstractActionPerformed(final PedAction action, final ActionEvent event) {
		super.pedAbstractActionPerformed(action, event);
		if (action instanceof SetSpeedAction) {
			setSpeedIndex(((SetSpeedAction) action).index);
		} else if (action == singleStepAction) {
			performSimulationStep();
		} else if (action == showWorldAction) {
			showWorld(((CheckBoxMenuItemAction) action).isOn());
		} else if (action instanceof ViewZoomLevelAction) {
			setZoomLevel(((ViewZoomLevelAction) action).index);
		} else if (action == showTrendChartAction) {
			showTrendChart(((CheckBoxMenuItemAction) action).isOn());
		} else if (action == showGenealogyAction) {
			showGenealogy();
		}
	}

	private SetSpeedAction[] setSpeedActions = new SetSpeedAction[] {
			new SetSpeedAction(0, false, 500, "Action.Pause", ResourceFinder.Folder.COMMON),
			new SetSpeedAction(1, true, 500, "Action.SetSpeedSlow", ResourceFinder.Folder.APP),
			new SetSpeedAction(2, true,  50, "Action.SetSpeedMedium", ResourceFinder.Folder.APP),
			new SetSpeedAction(3, true,   5, "Action.SetSpeedFast", ResourceFinder.Folder.APP),
	};
	private JRadioButtonMenuItem[] speedButtons = new JRadioButtonMenuItem[setSpeedActions.length];
	private SingleStepAction singleStepAction = new SingleStepAction();
	private CheckBoxMenuItemAction showWorldAction = new CheckBoxMenuItemAction(true, "Action.ShowWorld", this, null);
	private ViewZoomLevelAction[] viewZoomLevelActions = new ViewZoomLevelAction[DeinbacMainCanvas.MAGNIFICATIONS.length];
	private JRadioButtonMenuItem[] zoomLevelButtons = new JRadioButtonMenuItem[viewZoomLevelActions.length];

	private ShowTrendChartAction showTrendChartAction = new ShowTrendChartAction();
	private ShowGenealogyAction showGenealogyAction = new ShowGenealogyAction();

	private void updateZoomLevelSubmenuSelected() {
		for (int i = 0; i < viewZoomLevelActions.length; i++) {
			zoomLevelButtons[i].setSelected(worldCanvas.getMagnification() == DeinbacMainCanvas.MAGNIFICATIONS[i]);
		}
	}

	/**
	 * Sets the model.
	 * 
	 * @param game
	 *            the new model
	 */
	public void setModel(final GameModel game) {
		model = game;
	}

	protected GameModel getModel() {
		return model;
	}

	@Override
	public void switchLocale(final Locale locale) {
		super.switchLocale(locale);
		updateStatusText();
		repaint();
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		final Messages messages = getMessages();
		doc.setTitle(messages.getString("HelpWindow.Title"));
		doc.setHeading("Help");
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");
		doc.addContentPart(3, "Help.Introduction.Deinbac");
		doc.addContentPart(3, "Help.Introduction.Life");
		buildMenuDescriptions(doc, getJMenuBar());
		final NewGameSelectionWizard d = new NewGameSelectionWizard(this,
				messages.getText("GamePropertiesWindow.Title"));
		buildPropertyDescriptions(doc, 2, "Help.Properties", null, true);
		d.buildHelpDocumentSections(doc, 3, false);
		buildInputDescriptions(doc, null, false);
		final String[] columns = new String[] {
				"Bacterion.ID",
				"Position.Position",
				"Bacterion.Energy",
				"Bacterion.DateOfBirth",
				"Bacterion.DateOfLastEvent",
				"Bacterion.MotherID",
				"Bacterion.FatherID",
				"Bacterion.Genes",
				"Bacterion.ChildIDs",
		};
		buildPropertyDescriptions(doc, 2, "Help.Genealogy", columns, false);
		doc.endDocument();
	}

	private void resizeWorldCanvas() {
		worldCanvas.setSize(worldCanvas.getPreferredSize());
		worldScrollPane.setSize(worldCanvas.getPreferredSize());
		pack();
		repaint();
	}

	@Override
	protected void startNewGame() {
		logger.trace("Action: startNewGame");
		final Messages messages = getMessages();
		setSpeedIndex(0);
		updateSpeedButtons();
		final NewGameSelectionWizard d = new NewGameSelectionWizard(this,
				messages.getText("GamePropertiesWindow.Title"));
		d.enterDialog();
		logger.trace("New game selection dialog " + (d.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (d.wasAcceptedDialogChanges()) {
			d.readUIFieldsIntoProperties();
			props = d.getModel();
			model.init(props);
			isAlive = true;
			updateAllActionsEnabled();
			updateStatusText();
			resizeWorldCanvas();
		}
	}

	@Override
	public long getSleepDelay() {
		return setSpeedActions[speedIndex].sleepDelay;
	}

	private void setSpeedIndex(final int index) {
		this.speedIndex = index;
		updateAllActionsEnabled();
	}

	@Override
	public void step() {
		if (setSpeedActions[speedIndex].isRunning) {
			performSimulationStep();
		}
	}

	private void performSimulationStep() {
		isAlive = model.getSimulatable().step();
		worldCanvas.repaint();
		updateStatusText();
		if ((null != model) && (model.getProps().getSimulation() == SimulationType.DEINBAC)
				&& (null != trendChartWindow)) {
			trendChartWindow.stepSimul();
		}
	}

	private void showWorld(final boolean isToShow) {
		worldCanvas.setDisplayed(isToShow);
		worldCanvas.repaint();
	}

	private void setZoomLevel(final int zoomLevel) {
		worldCanvas.setZoomLevel(zoomLevel);
		resizeWorldCanvas();
	}

	@Override
	public void changeZoomLevel(final int amount) {
		final int old = worldCanvas.getZoomLevel();
		final int level = MathUtil.bound(old + amount, 0, DeinbacMainCanvas.MAGNIFICATIONS.length - 1);
		worldCanvas.setZoomLevel(level);
		resizeWorldCanvas();
		updateZoomLevelSubmenuSelected();
	}

	public void showGenealogy() {
		final TextDocumentBuilder doc = new TextDocumentBuilder(getMessages());
		doc.setInventoryMaxDepth(0);
		doc.setTitle(getMessages().getString(new I18NStringWithFillIns("Genealogy.Title", null,
				new Object[] { getMessages().getApplicationTitle(), Long.toString(model.deinbac.getTime()), })));
		doc.beginDocument();
		final List<Bacterion> bacs = model.deinbac.getGenealogySnapshot();
		if (!bacs.isEmpty()) {
			buildBacterionsTable(doc, bacs, true);
		}
		doc.endDocument();

		final JFrame window = new JFrame(doc.getTitle());
		final JHtmlPane htmlPane = new JHtmlPane(doc.result());
		final JScrollPane scrollPane = new JScrollPane(htmlPane);
		window.getContentPane().add(scrollPane);
		window.pack();
		GUIAlgorithms.center(window);
		window.setVisible(true);
		// HtmlPaneDialog d = new HtmlPaneDialog(this, title,
		// content.toString());
		// d.setVisible(true);
	}

	public void showTrendChart(final boolean isToShow) {
		logger.trace("Action: showTrendChart");
		if (null == trendChartWindow) {
			trendChartWindow = new DeinbacTrendChartWindow(this, "This should never show up!");
		}
		trendChartWindow.updateTitle();
		trendChartWindow.setVisible(isToShow);
	}

	public String genesToString(final Bacterion b) {
		final StringBuffer content = new StringBuffer();
		final int[] genes = b.getGenes();
		for (int i = 0; i < genes.length; i++) {
			final int g = genes[i];
			content.append((char) ((g < 10) ? '0' + g : 'A' + g - 10));
		}
		return content.toString();
	}

	private String bacterionHref(final TextFormatter tf, final int id) {
		final StringBuffer ret = new StringBuffer();
		if (id < 0) {
			ret.append("---");
		} else {
			ret.append(tf.href("#bac" + id, Integer.toString(id)));
		}
		return ret.toString();
	}

	private static final TextFormatter.ColumnInfo[] COLUMN_INFO_TEMPLATE = new TextFormatter.ColumnInfo[] {};

	/**
	 * Builds a document table containing the given bacterions.
	 * 
	 * @param doc
	 *            the document to write into
	 * @param bacs
	 *            the bacterions to list
	 * @param isToShowChildren
	 *            if <code>true</code>, an extra column is created where for
	 *            each bacterion the direct childrens are listed
	 */
	private void buildBacterionsTable(
			final TextDocumentBuilder doc, 
			final List<Bacterion> bacs, 
			final boolean isToShowChildren) {
		final ArrayList<TextFormatter.ColumnInfo> cis = new ArrayList<TextFormatter.ColumnInfo>(); 
		final Messages messages = getMessages();
		cis.add(new TextFormatter.ColumnInfo( 5, TextFormatter.RIGHT,  messages, "Bacterion.ID"));
		cis.add(new TextFormatter.ColumnInfo(10, TextFormatter.CENTER, messages, "Position.Position"));
		cis.add(new TextFormatter.ColumnInfo( 6, TextFormatter.RIGHT,  messages, "Bacterion.Energy"));
		cis.add(new TextFormatter.ColumnInfo(10, TextFormatter.RIGHT,  messages, "Bacterion.DateOfBirth"));
		cis.add(new TextFormatter.ColumnInfo(10, TextFormatter.RIGHT,  messages, "Bacterion.DateOfLastEvent"));
		cis.add(new TextFormatter.ColumnInfo( 5, TextFormatter.RIGHT,  messages, "Bacterion.MotherID"));
		cis.add(new TextFormatter.ColumnInfo( 5, TextFormatter.RIGHT,  messages, "Bacterion.FatherID"));
		cis.add(new TextFormatter.ColumnInfo( 6, TextFormatter.LEFT,   messages, "Bacterion.Genes"));
		if (isToShowChildren) {
			cis.add(new TextFormatter.ColumnInfo(30, TextFormatter.LEFT, messages, "Bacterion.ChildIDs"));
		}
		final TextFormatter tf = doc.tf();
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true,
				TextFormatter.htmlAttribute(TextFormatter.CLASS, "bacterions"), cis.toArray(COLUMN_INFO_TEMPLATE));
		doc.beginTable(null, ti);
		final String isDeadMsg = messages.getText("Bacterion.Dead");
		final ListIterator<Bacterion> it = bacs.listIterator();
		while (it.hasNext()) {
			final Bacterion b = it.next();
			final int id = b.getID();
			final int energy = b.getEnergy();
			final ArrayList<TextFormatter.TableCellInfo> cells = new ArrayList<TextFormatter.TableCellInfo>(cis.size());
			cells.add(new TextFormatter.TableCellInfo(tf.anchor("bac" + id, Integer.toString(id))));
			cells.add(new TextFormatter.TableCellInfo("(" + b.getX() + ",&nbsp;" + b.getY() + ")"));
			cells.add(new TextFormatter.TableCellInfo((energy < 0) ? isDeadMsg : Integer.toString(energy)));
			cells.add(new TextFormatter.TableCellInfo(Long.toString(b.getDateOfBirth())));
			cells.add(new TextFormatter.TableCellInfo(Long.toString(b.getDateOfLastEvent())));
			cells.add(new TextFormatter.TableCellInfo(bacterionHref(tf, b.getMotherID())));
			cells.add(new TextFormatter.TableCellInfo(bacterionHref(tf, b.getFatherID())));
			cells.add(new TextFormatter.TableCellInfo(genesToString(b)));
			if (isToShowChildren) {
				final StringBuffer buf = new StringBuffer();
				final ListIterator<Bacterion> itChilds = bacs.listIterator(it.nextIndex());
				while (itChilds.hasNext()) {
					final Bacterion bc = itChilds.next();
					if ((bc.getMotherID() == id) || (bc.getFatherID() == id)) {
						buf.append(bacterionHref(tf, bc.getID()) + " ");
					}
				}
				buf.append("&nbsp;");
				cells.add(new TextFormatter.TableCellInfo(buf.toString()));
			}
			doc.addTableRow(cells);
		}
		doc.endTable();
	}

	public void showPositionInfo(final int x, final int y) {
		final TextDocumentBuilder doc = new TextDocumentBuilder(getMessages());
		doc.setInventoryMaxDepth(0);
		doc.setTitle(getMessages().getString(
			new I18NStringWithFillIns("Position.Title", null, new Object[] {
				getMessages().getApplicationTitle(),
				Long.toString(model.deinbac.getTime()),
				Integer.toString(x),
				Integer.toString(y),
			} )));
		doc.beginDocument();
		doc.beginParagraph(null);
		doc.add(getMessages().getString((0 != model.deinbac.getPlayfieldAt(x, y) ? "Position.HasMana" : "Position.HasNoMana")));
		doc.endParagraph();
		final List<Bacterion> bacs = model.deinbac.getBacterionsOverPosition(x, y);
		if (!bacs.isEmpty()) {
			final String key = "Position.Bacteria";
			doc.addContentPart(2, key);
			buildBacterionsTable(doc, bacs, false);
		}
		doc.endDocument();
		final HtmlPaneDialog d = new HtmlPaneDialog(this, doc.getTitle(), doc.result());
		d.enterDialog();
	}

	public ArrayList<DeinbacTrendChartEntry> getTrend() {
		return model.deinbac.getTrendSnapshot();
	}

	public List<Bacterion> getGenealogy() {
		return model.deinbac.getGenealogySnapshot();
	}

	@Override
	public boolean isRunning() {
		return isRunningRun();
	}

	@Override
	public FinishableThread createRunThread() {
		return new GameThread(this);
	}

	@Override
	public void pauseOrResume(final boolean isToResume) {
		pauseOrResumeRun(isToResume);
	}

	private int speedIndex = 0;
	private boolean isAlive = true;

	private GameProperties props = new GameProperties();

	/** Access to the logic layer. */
	private GameModel model;

	/** The canvas painting the mana world. */
	private DeinbacMainCanvas worldCanvas;
	private JScrollAndZoomPane worldScrollPane;
	private JTextField statusText;

	private DeinbacTrendChartWindow trendChartWindow;
}
