package de.ped.deinbac.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.ArrayList;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import de.ped.deinbac.logic.DeinbacGame;
import de.ped.deinbac.logic.DeinbacTrendChartEntry;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.gui.GUIAlgorithms;
import de.ped.tools.log.Logger;

/**
 * The trend chart window.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DeinbacTrendChartWindow extends JFrame implements WindowListener {
	private static final long serialVersionUID = 1L;

	DeinbacMainWindow mainWindow;
	DeinbacTrendChartCanvas canvas;
	JScrollPane scrollPane;

	DeinbacTrendChartWindow(final DeinbacMainWindow mainWindow, final String title) {
		super(title);
		this.mainWindow = mainWindow;
		canvas = new DeinbacTrendChartCanvas(mainWindow);
		scrollPane = new JScrollPane(canvas);
		getContentPane().add(scrollPane);
		pack();
		GUIAlgorithms.center(this);
		addWindowListener(this);
	}

	public void stepSimul() {
		if (canvas.step()) {
			final Dimension oldScrollPaneSize = scrollPane.getSize();
			canvas.setSize(canvas.getPreferredSize());
			scrollPane.setPreferredSize(oldScrollPaneSize);
			scrollPane.getHorizontalScrollBar().setValue(scrollPane.getHorizontalScrollBar().getMaximum());
			updateTitle();
			pack();
			repaint();
		}
	}

	public void updateTitle() {
		final Messages messages = mainWindow.getMessages();
		final String title = messages.getString(new I18NStringWithFillIns("TrendChart.Title", null,
				new Object[] { messages.getApplicationTitle(), Long.toString(mainWindow.getModel().getTime()), }));
		setTitle(title);
	}

	@Override
	public void windowActivated(final WindowEvent e) {
		logger.trace("Trend chart window activated.");
	}

	@Override
	public void windowClosed(final WindowEvent e) {
		logger.trace("Trend chart window closed.");
	}

	@Override
	public void windowClosing(final WindowEvent e)	{
		logger.trace("Trend chart window closing.");
		mainWindow.showTrendChart(false);
	}

	@Override
	public void windowDeactivated(final WindowEvent e)	{
		logger.trace("Trend chart window deactivated.");
	}

	@Override
	public void windowDeiconified(final WindowEvent e) {
		logger.trace("Trend chart window deiconified.");
	}

	@Override
	public void windowIconified(final WindowEvent e) {
		logger.trace("Trend chart window iconified.");
	}

	@Override
	public void windowOpened(final WindowEvent e) {
		logger.trace("Trend chart window opened.");
	}

	private final Logger logger = Logger.getLogger(this.getClass());
}

/**
 * The trend chart canvas within the trend chart window.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
class DeinbacTrendChartCanvas extends JPanel {
	private static final long serialVersionUID = 1L;

	private static final Color[] COLORS = new Color[] {
			Color.white,
			Color.magenta,
			Color.yellow,
			Color.gray,
	};

	private DeinbacMainWindow mainWindow;

	private ArrayList<DeinbacTrendChartEntry> trend;
	private long[] maxes = new long[DeinbacTrendChartEntry.SIZE];

	DeinbacTrendChartCanvas(final DeinbacMainWindow mainWindow) {
		this.mainWindow = mainWindow;
		setDoubleBuffered(true);
		clearScaling();
	}

	@Override
	public void paintComponent(final Graphics g) {
		super.paintComponent(g);
		g.setColor(Color.black);
		final int height = getHeight();
		final int width = getWidth();
		g.fillRect(0, 0, width, height);
		final int xaxis = height;
		g.setColor(COLORS[0]);
		g.drawLine(0, xaxis, width, xaxis);
		final int trendChartFrequency = mainWindow.getModel().deinbac.getProps().getTrendChartFrequency();
		if ((null != trend) && (0 < trendChartFrequency)) {
			for (int i = 0; i < trend.size(); i++) {
				final DeinbacTrendChartEntry e = trend.get(i);
				final long t = e.getTime();
				final int x = (int) (t / trendChartFrequency);
				if (99 == (i % 100)) {
					g.setColor(COLORS[0]);
					g.drawLine(x, xaxis, x, xaxis - 2);
					g.drawString(Long.toString(t), x, xaxis - 6);
				}
				for (int j = DeinbacTrendChartEntry.SIZE - 1; j > 0; j--) {
					final long val = e.get(j);
					maxes[j] = Math.max(maxes[j], val);
					final int y = xaxis - (int) (val * height / maxes[j]);
					g.setColor(COLORS[j]);
					g.drawLine(x, y, x, y);
				}
			}
		}
	}

	private int getLength() {
		return (null == trend ? 0 : trend.size());
	}

	/**
	 * @return the preferred size of this canvas
	 */
	@Override
	public Dimension getPreferredSize() {
		final int width = Math.max(getLength(), 200);
		return new Dimension(width, 100);
	}

	@Override
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	public boolean step() {
		boolean ret = false;
		final DeinbacGame model = mainWindow.getModel().deinbac;
		final int trendChartFrequency = model.getProps().getTrendChartFrequency();
		if ((0 < trendChartFrequency) && (0 == (model.getTime() % trendChartFrequency))) {
			final int oldSize = getLength();
			trend = mainWindow.getTrend();
			final int newSize = getLength();
			if (oldSize != newSize) {
				ret = true;
				if (oldSize > newSize) {
					clearScaling();
				}
			}
		}
		return ret;
	}

	private void clearScaling() {
		for (int i = 0; i < maxes.length; i++) {
			maxes[i] = 1L;
		}
	}
}