package de.ped.deinbac.logic;

import java.awt.Dimension;

import de.ped.tools.Assert;
import de.ped.tools.MathUtil;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A bacterion in our simulation. It can crawl, eat, split and have sex with other bacterions.
 * It has genes that influence the behavior of the bacterion - especially how often to change directions. 
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Bacterion {
	private static Logger logger = Logger.getRootLogger();
	static {
		logger.setLogLevel(Level.INFO);
	}

	/** The x size of one bacterion. Should be divisible by 2, so that all mana can be eaten. */
	public static final int XSIZE = 4;
	/** The y size of one bacterion. Should be divisible by 2, so that all mana can be eaten. */
	public static final int YSIZE = 4;
	/** Movement vectors for all allowed directions. */
	public static final Dimension[] DIRECTIONS = {
			new Dimension( 0,-2),    
			new Dimension( 2,-1),    
			new Dimension( 2, 1),    
			new Dimension( 0, 2),    
			new Dimension(-2, 1),    
			new Dimension(-2,-1),    
	};
	/** The number of directions a bacterion can head. */
	public static final int NUMBER_OF_DIRECTIONS = DIRECTIONS.length;
	/** The number of genes each bacterion has. */
	public static final int NUMBER_OF_GENES = NUMBER_OF_DIRECTIONS;
	/** Each gene contains a number between 0 and this value. */
	private static final int GENE_VALUE_MAX = 19;

	/** A virtual ID value meaning none. */
	public static final int NULL_ID = -1;

	private static int nextUnusedID = 0;

	private static int getNextUnusedID() {
		return nextUnusedID++;
	}

	public static void resetNextUnusedID() {
		nextUnusedID = 0;
	}

	/** A unique ID of this bacterion. */
	private int id;
	private final int motherID;
	private final int fatherID;
	/** The genes of this bacterion. */
	private int[] genes = new int[NUMBER_OF_GENES];
	/** Helper for performance reasons - contains the integral of probability chunks up to this gene. */
	private long[] probIntegral = new long[NUMBER_OF_GENES];
	/** Where am i? */
	private int x;
	/** Where am i? */
	private int y;
	/** Where am i heading? */
	private int direction;
	/** How much energy do I have? */
	private int energy;
	/** How many generations am I away from the initial population? */
	private int generation;
	/** The date of the last meaningful event (like birth, split, sex, death) */
	private long dateOfLastEvent;
	/** I was born at simulation heartbeat */
	private long dateOfBirth;

	private final DeinbacProperties props;

	/**
	 * Creates a new bacterion at random position.
	 * 
	 * @param props
	 *            these properties are used to create the bacterion
	 */
	public Bacterion(final DeinbacProperties props) {
		this.props = props;
		do {
			// follow position matrix constraints
			x = 2 * (MathUtil.random((props.getXsize() / 2) - 2));
			y = 2 * (MathUtil.random((props.getYsize() / 2) - 2)) + ((x % 4) / 2);
		} while (!Bacterion.isInsideWorldBorders(x, y, props));
		for (int j = 0; j < Bacterion.NUMBER_OF_GENES; j++) {
			genes[j] = MathUtil.random(8);
		}
		energy = Math.min(props.getInitialEnergy(), props.getMaxEnergy());
		generation = 0;
		dateOfBirth = 0; // Adams & Evas
		dateOfLastEvent = dateOfBirth;
		direction = MathUtil.random(NUMBER_OF_DIRECTIONS);
		motherID = NULL_ID;
		fatherID = NULL_ID;
		afterInit();
		createGenealogyEntry(this, null, null, 0);
	}

	/**
	 * this will become a clone of other.
	 * 
	 * @param other
	 *            the source to clone this from
	 */
	protected void copyFrom(final Bacterion other) {
		for (int i = 0; i < NUMBER_OF_GENES; i++) {
			genes[i] = other.genes[i];
			probIntegral[i] = other.probIntegral[i];
		}
		x = other.x;
		y = other.y;
		direction = other.direction;
		energy = other.energy;
		generation = other.generation;
		dateOfLastEvent = other.dateOfLastEvent;
		dateOfBirth = other.dateOfBirth;
	}

	public int getID() {
		return id;
	}

	public static boolean isInsideWorldBorders(final int x, final int y, final DeinbacProperties props) {
		return ((x >= 0) && (x <= props.getXsize() - XSIZE) && (y >= 0) && (y <= props.getYsize() - YSIZE));
	}

	public boolean isOverPosition(final int xpos, final int ypos, final DeinbacProperties props) {
		return ((MathUtil.modulo(xpos - x, props.getXsize()) < XSIZE)
				&& (MathUtil.modulo(ypos - y, props.getYsize()) < YSIZE));
	}

	public Bacterion(final Bacterion mother, final long timestamp) {
		props = mother.props;
		mother.energy /= 2;
		copyFrom(mother);
		dateOfBirth = timestamp;
		dateOfLastEvent = dateOfBirth;
		mother.dateOfLastEvent = dateOfBirth;
		generation++;
		direction = MathUtil.random(Bacterion.NUMBER_OF_DIRECTIONS);
		motherID = mother.id;
		fatherID = NULL_ID;
		int j = MathUtil.random(Bacterion.NUMBER_OF_GENES);
		genes[j] += -1 + MathUtil.random(2) * 2;
		afterInit();
		createGenealogyEntry(this, mother, null, timestamp);
	}

	public Bacterion(final Bacterion mother, final Bacterion father, final long timestamp) {
		props = mother.props;
		int energ = (mother.energy + father.energy) / 3;
		mother.energy = (int) (mother.energy * 2.0 / 3.0);
		father.energy = (int) (father.energy * 2.0 / 3.0);
		copyFrom(mother);
		energy = energ;
		dateOfBirth = timestamp;
		dateOfLastEvent = dateOfBirth;
		mother.dateOfLastEvent = dateOfBirth;
		father.dateOfLastEvent = dateOfBirth;
		generation++;
		direction = MathUtil.random(Bacterion.NUMBER_OF_DIRECTIONS);
		motherID = mother.id;
		fatherID = father.id;
		int j = NUMBER_OF_GENES / 2;
		for (int i = 0; i < Bacterion.NUMBER_OF_GENES; i++) {
			if ((j >= 0) && ((i + j >= Bacterion.GENE_VALUE_MAX) || (0 == MathUtil.random(2)))) {
				genes[i] = father.genes[i];
				j--;
			}
		}
		afterInit();
		createGenealogyEntry(this, mother, father, timestamp);
	}

	/**
	 * Lets this bacterion die.
	 * 
	 * @param timestamp
	 *            the current timestamp (in world steps)
	 */
	public void die(final long timestamp) {
		energy = -1;
		dateOfLastEvent = timestamp;
		createGenealogyEntry(this, null, null, timestamp);
	}

	/** 
	 * Norms the genes and does some pre-calculations on them to speed up simulation.
	 * Should be called after each creation of a bacterion. 
	 */
	protected void afterInit() {
		id = getNextUnusedID();
		int min = genes[0];
		for (int i = 0; i < NUMBER_OF_GENES; i++) {
			if (genes[i] < min) {
				min = genes[i];
			}
		}
		int cnt = 0;
		for (int i = 0; i < NUMBER_OF_GENES; i++) {
			genes[i] = Math.min(GENE_VALUE_MAX, genes[i] - min);
			cnt = cnt + (1 << genes[i]);
			probIntegral[i] = cnt;
		}
	}

	public void createGenealogyEntry(
			final Bacterion child, 
			final Bacterion mother, 
			final Bacterion father, 
			final long timestamp) {
		String s;
		if (child.getEnergy() < 0) {
			s = "Dying " + child.id;
		} else if (null == mother) {
			s = "Creating " + child.id;
		} else if (null == father) {
			s = "Birth (split) " + child.id + " of " + mother.id;
		} else {
			s = "Birth (sex) " + child.id + " of " + mother.id + " and " + father.id;
		}
		logger.debug(timestamp + ": " + s);
	}

	public int getX() {
		return x;
	}

	public int getY() {
		return y;
	}

	public void setXY(final int x, final int y) {
		this.x = x;
		this.y = y;
	}

	public int getEnergy() {
		return energy;
	}

	public int addEnergy(final int addition) {
		energy = MathUtil.bound(energy + addition, 0, props.getMaxEnergy());
		return energy;
	}

	public int getMotherID() {
		return motherID;
	}

	public int getFatherID() {
		return fatherID;
	}

	public long getDateOfBirth() {
		return dateOfBirth;
	}

	public long getDateOfLastEvent() {
		return dateOfLastEvent;
	}

	public int[] getGenes() {
		return genes;
	}

	/**
	 * @param timestamp
	 *            the current timestamp (in world steps)
	 * @return the duration (in world steps) since the last meaningful event (like birth, split,
	 *         sex)
	 */
	public long getTimeSinceLastEvent(long timestamp) {
		return (timestamp - dateOfLastEvent);
	}

	/**
	 * @param timestamp
	 *            the current timestamp (in world steps)
	 * @return if this bacterion is healthy and old enough to split
	 */
	public boolean isSplittable(final long timestamp) {
		return ((energy >= props.getEnergyForSplit()) && (getTimeSinceLastEvent(timestamp) >= props.getAgeForSplit()));
	}

	/**
	 * @param timestamp
	 *            the current timestamp (in world steps)
	 * @return if this bacterion is healthy and old enough to be a sex partner
	 */
	public boolean isParentable(final long timestamp) {
		return ((energy >= props.getEnergyForSex()) && (getTimeSinceLastEvent(timestamp) >= props.getAgeForSex()));
	}

	public boolean isPartnerInRangeForSex(final Bacterion partner) {
		return ((Math.abs(x - partner.x) + Math.abs(y - partner.y)) <= 4);
	}

	public void decideDirection() {
		int zf = MathUtil.random((int) probIntegral[Bacterion.NUMBER_OF_GENES - 1]);
		int idx = 0;
		for (int i = Bacterion.NUMBER_OF_GENES - 1; i >= 0; i--) {
			if (zf < probIntegral[i]) {
				idx = i;
			}
		}
		direction = (direction + idx) % Bacterion.NUMBER_OF_DIRECTIONS;
	}

	public void move() {
		decideDirection();
		int xnew = x + Bacterion.DIRECTIONS[direction].width;
		int ynew = y + Bacterion.DIRECTIONS[direction].height;
		if (props.isWorldTorus()) {
			xnew = MathUtil.modulo(xnew, props.getXsize());
			ynew = MathUtil.modulo(ynew, props.getYsize());
		} else {
			if (!isInsideWorldBorders(xnew, ynew, props)) {
				xnew = x;
				ynew = y;
			}
		}
		x = xnew;
		y = ynew;
		Assert.assertTrue(x >= 0);
		Assert.assertTrue(x < props.getXsize());
		Assert.assertTrue(y >= 0);
		Assert.assertTrue(y < props.getYsize());
	}
}
