package de.ped.deinbac.logic;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;

/**
 * The simulation logic, and access point for all other logic items.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DeinbacGame extends Simulatable {

	private final DeinbacProperties props;

	private DeinbacStatus status = new DeinbacStatus();

	/** 
	 * The list of living bacterions for the next simulation step.
	 * So, bacterions dying in the current step are removed, 
	 * and bacterions being born are added.
	 */
	private List<Bacterion> nextBacterions = new ArrayList<Bacterion>();
	/** The genealogy, keeping track of all bacterions that ever lived. */
	private List<Bacterion> genealogy = new ArrayList<Bacterion>();
	/** The trend chart data. */
	private List<DeinbacTrendChartEntry> trend = new ArrayList<DeinbacTrendChartEntry>();

	/**
	 * A bacterion at the given position eats the mana.
	 * 
	 * @param x
	 *            the x position within the world
	 * @param y
	 *            the y position within the world
	 * @return the amount of mana eaten
	 */
	public int eatAt(final int x, final int y) {
		final int xsize = props.getXsize();
		final int ysize = props.getYsize();
		int bites = 0;
		for (int xi = 0; xi < Bacterion.XSIZE; xi++) {
			for (int yi = 0; yi < Bacterion.YSIZE; yi++) {
				int xpos = MathUtil.modulo(x + xi, xsize);
				int ypos = MathUtil.modulo(y + yi, ysize);
				if (status.board.getAndSetAt(xpos, ypos, false)) {
					bites++;
				}
			}
		}
		return bites;
	}

	/**
	 * @return the logical size of the playfield
	 */
	@Override
	public PlayfieldDimension getPlayfieldSize() {
		return new PlayfieldDimension(props.getXsize(), props.getYsize());
	}

	/**
	 * @param x
	 *            the x position within the world
	 * @param y
	 *            the y position within the world
	 * @return if at position (x, y) is Mana
	 */
	@Override
	public int getPlayfieldAt(final int x, final int y) {
		return (status.board.getAt(x, y) ? 1 : 0);
	}

	/**
	 * @return a snapshot of the currently living population of Bacterions
	 */
	public synchronized List<Bacterion> getBacterions() {
		final ArrayList<Bacterion> ret = new ArrayList<Bacterion>(status.currentBacterions);
		return ret;
	}

	/**
	 * @param x
	 *            the x position within the world
	 * @param y
	 *            the y position within the world
	 * @return all bacterions being over position (x, y)
	 */
	public synchronized List<Bacterion> getBacterionsOverPosition(final int x, final int y) {
		final ArrayList<Bacterion> ret = new ArrayList<Bacterion>();
		final Iterator<Bacterion> it = status.currentBacterions.iterator();
		while (it.hasNext()) {
			final Bacterion b = it.next();
			if (b.isOverPosition(x, y, props)) {
				ret.add(b);
			}
		}
		return ret;
	}

	public DeinbacProperties getProps() {
		return props;
	}

	@Override
	public long getTime() {
		return status.time;
	}

	public long getAge(final Bacterion bacterion) {
		return bacterion.getTimeSinceLastEvent(status.time);
	}

	public synchronized List<Bacterion> getGenealogySnapshot() {
		return new ArrayList<Bacterion>(genealogy);
	}

	public synchronized ArrayList<DeinbacTrendChartEntry> getTrendSnapshot() {
		return new ArrayList<DeinbacTrendChartEntry>(trend);
	}

	/** Let's rain mana onto the world. */
	protected void rainMana() {
		status.manaToGrow += props.getManaToGrowEachRound();
		while (1.0 <= status.manaToGrow) {
			status.board.growMana();
			status.manaToGrow--;
		}
		status.manaToGrowInCockaigne += props.getManaToGrowEachRoundInCockaigne();
		while (1.0 <= status.manaToGrowInCockaigne) {
			status.board.growManaInCockaigne();
			status.manaToGrowInCockaigne--;
		}
	}

	/**
	 * Moves a bacterion, including eating, dying, splitting and sex
	 * 
	 * @param bacterion
	 *            the bacterion to move
	 */
	protected void moveBacterion(final Bacterion bacterion) {
		if (bacterion.getEnergy() <= 0) {
			bacterion.die(status.time);
			nextBacterions.remove(bacterion);
		} else if (bacterion.isSplittable(status.time)) {
			final Bacterion child = new Bacterion(bacterion, status.time);
			nextBacterions.add(child);
			genealogy.add(child);
		} else if (bacterion.getDateOfBirth() != status.time) {
			final int bite = eatAt(bacterion.getX(), bacterion.getY());
			bacterion.addEnergy(props.getManaEnergy() * bite - 1);
			bacterion.move();
			if (bacterion.isParentable(status.time)) {
				// Mating partner available?
				final Iterator<Bacterion> it = status.currentBacterions.iterator();
				while (it.hasNext()) {
					final Bacterion partner = it.next();
					if ((partner.getID() != bacterion.getID())
							&& (partner.isParentable(status.time))
							&& bacterion.isPartnerInRangeForSex(partner)) {
						final Bacterion child = new Bacterion(bacterion, partner, status.time);
						nextBacterions.add(child);
						genealogy.add(child);
						break;
					}
				}
			}
		}
	}

	protected void createTrendRecord() {
		int cnt = 0;
		long sum = 0;
		final Iterator<Bacterion> it = status.currentBacterions.iterator();
		while (it.hasNext()) {
			final Bacterion b = it.next();
			sum += b.getEnergy();
			cnt++;
		}
		final int energ = (cnt > 0 ? Math.round((float) sum / cnt) : 0);
		final DeinbacTrendChartEntry e = new DeinbacTrendChartEntry(status.time, cnt, energ, status.board.getAmountOfMana());
		trend.add(e);
	}

	/**
	 * Does a simulation step.
	 * 
	 * @return <code>true</code> iff there are still bacterions in the world
	 */
	@Override
	public synchronized boolean step() {
		boolean ret = !status.currentBacterions.isEmpty();
		if (ret) {
			status.time++;
			nextBacterions.clear();
			nextBacterions.addAll(status.currentBacterions);
			final Iterator<Bacterion> it = status.currentBacterions.iterator();
			while (it.hasNext()) {
				final Bacterion b = it.next();
				moveBacterion(b);
			}
			status.currentBacterions.clear();
			status.currentBacterions.addAll(nextBacterions);
			rainMana();
			final int trendChartFrequency = props.getTrendChartFrequency();
			if ((0 < trendChartFrequency) && (0 == status.time % trendChartFrequency)) {
				createTrendRecord();
			}
		}
		return ret;
	}

	public DeinbacGame(final DeinbacProperties props) {
		this.props = props;
		Bacterion.resetNextUnusedID();
		status.time = 0;
		status.manaToGrow = 0.0;
		status.manaToGrowInCockaigne = 0.0;
		status.board = new ManaWorld(props.getXsize(), props.getYsize());
		double m = props.getInitialManaToGrowPerField();
		if (0.05 < m) { // Too many fields to set to randomly pick positions
			for (int x = 0; x < props.getXsize(); x++) {
				for (int y = 0; y < props.getYsize(); y++) {
					if (!status.board.getAt(x,  y) && Math.random() < m) {
						status.board.setAt(x,  y,  true);
					}
				}
			}
		} else { // Only sparse mana, so it's okay to randomly pick positions
			for (int i = 0; i < props.getInitialManaToGrow(); i++) {
				status.board.growMana();
			}
		}
		for (int i = 0; i < props.getInitialNumberOfBacterions(); i++) {
			final Bacterion child = new Bacterion(props);
			status.currentBacterions.add(child);
			genealogy.add(child);
		}
	}
}

/**
 * Contains the world including mana.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
class ManaWorld {
	/** The x-size of the world. */
	private int xsize;
	/** The y-size of the world. */
	private int ysize;
	/** The world itself. */
	private boolean[][] pixel;

	/** The current amount of mana in this world. */
	private long amountOfMana = 0;

	/**
	 * Creates an empty world.
	 * 
	 * @param xsize
	 *            the x-size of the world
	 * @param ysize
	 *            the y-size of the world
	 */
	public ManaWorld(final int xsize, final int ysize) {
		this.xsize = xsize;
		this.ysize = ysize;
		pixel = new boolean[xsize][ysize];
	}

	public boolean getAt(final int x, final int y) {
		return pixel[x][y];
	}

	public void setAt(final int x, final int y, final boolean value) {
		getAndSetAt(x, y, value);
	}

	/**
	 * Sets the new value at the given position, and returns the old value.
	 * 
	 * @param x
	 *            the x position in the world of the pixel to change
	 * @param y
	 *            the y position in the world of the pixel to change
	 * @param value
	 *            the new value to set. <code>true</code> represents mana,
	 *            <code>false</code> no mana.
	 * @return if there was mana before this operation. <code>true</code> represents
	 *         mana, <code>false</code> no mana.
	 */
	public boolean getAndSetAt(final int x, final int y, final boolean value) {
		final boolean oldValue = pixel[x][y];
		pixel[x][y] = value;
		if (oldValue) {
			amountOfMana--;
		}
		if (value) {
			amountOfMana++;
		}
		return oldValue;
	}

	/** 
	 * @return the current amount of mana in this world
	 */
	public long getAmountOfMana() {
		return amountOfMana;
	}

	/** Lets mana grow at a random spot. */
	public void growMana() {
		final int x = MathUtil.random(xsize);
		final int y = MathUtil.random(ysize);
		setAt(x, y, true);
	}

	/** Lets mana grow within cockaigne - a small area of the world. */
	public void growManaInCockaigne() {
		final int cockaigneXSize = xsize / DeinbacProperties.COCKAIGNE_RATIO;
		final int cockaigneYSize = ysize / DeinbacProperties.COCKAIGNE_RATIO;
		final int x = MathUtil.random(cockaigneXSize) + ((xsize - cockaigneXSize) / 2);
		final int y = MathUtil.random(cockaigneYSize) + ((ysize - cockaigneYSize) / 2);
		setAt(x, y, true);
	}
}

/**
 * Contains the basic status of the simulation.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
class DeinbacStatus {
	public List<Bacterion> currentBacterions = new ArrayList<Bacterion>();
	/** The number of heartbeats the simulation has been running. */
	public long time;
	/** How much mana is outstanding to grow. If it reaches 1.0, mana will grow somewhere. */
	public double manaToGrow;
	/** How much mana is outstanding to grow in cockaigne. If it reaches 1.0, mana will grow in cockaigne. */
	public double manaToGrowInCockaigne;
	public ManaWorld board;
}
