package de.ped.deinbac.logic;

import de.ped.tools.Assert;

/** 
 * Contains all properties that remain constant for a run. 
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DeinbacProperties implements Cloneable {
	/** The 1/<em>n</em> part of x-size and y-size will be used for cockaigne. */
	public static final int COCKAIGNE_RATIO = 8;
	/** The x-size of the world. Must be divisible by 4. */
	private int xsize = 320;
	/** The y-size of the world. Must be divisible by 2. */
	private int ysize = 200;
	/** Is the world a isWorldTorus - compared to having border walls? */
	private boolean isWorldTorus = true;
	/** Initial ratio of mana in the world per field. */
	private double initialManaToGrowPerField = 0.0078;
	/** The ratio of mana to grow each round per field. */
	private double manaToGrowEachRoundPerField = 0.0000047;
	/** The ratio of mana to grow each round per field. */
	private double manaToGrowEachRoundInCockaignePerField = 0.000320;
	/** The amount of energy each mana gives to a bacterion. */
	private int manaEnergy = 40;
	/** The initial amount of energy the first generation of bacterions starts with. */
	private int initialEnergy = 100;
	/** The maximum amount of energy a bacterion can get. After that, eating does not increase the energy level. */
	private int maxEnergy = 1500;
	/** The duration a bacterion must wait after a major event (birth, split, sex) before it can have sex again. */
	private int ageForSex = 300;
	/** The amount of energy a bacterion must have to have sex. */
	private int energyForSex = 750;
	/** The duration a bacterion must wait after a major event (birth, split, sex) before it can split again. */
	private int ageForSplit = 800;
	/** The amount of energy a bacterion must have to split. */
	private int energyForSplit = 1000;
	/** The number of bacterions to start with. */
	private int initialNumberOfBacterions = 50;
	/** To create a genealogy graph? */
	private boolean isGenealogyGraph = false;
	/** How often to create a new trend chart record? 0 means never. */
	private int trendChartFrequency = manaEnergy;

	public int getAgeForSex() {
		return ageForSex;
	}
	public void setAgeForSex(final int ageForSex) {
		this.ageForSex = ageForSex;
	}
	public int getAgeForSplit() {
		return ageForSplit;
	}
	public void setAgeForSplit(final int ageForSplit) {
		this.ageForSplit = ageForSplit;
	}
	public int getEnergyForSex() {
		return energyForSex;
	}
	public void setEnergyForSex(final int energyForSex) {
		this.energyForSex = energyForSex;
	}
	public int getEnergyForSplit() {
		return energyForSplit;
	}
	public void setEnergyForSplit(final int energyForSplit) {
		this.energyForSplit = energyForSplit;
	}
	public int getTrendChartFrequency() {
		return trendChartFrequency;
	}
	public void setTrendChartFrequency(final int trendChartFrequency) {
		this.trendChartFrequency = trendChartFrequency;
	}
	public int getInitialEnergy() {
		return initialEnergy;
	}
	public void setInitialEnergy(final int initialEnergy) {
		this.initialEnergy = initialEnergy;
	}
	public int getInitialManaToGrow() {
		return (int) Math.round(fromPerField(initialManaToGrowPerField, 1));
	}
//	public void setInitialManaToGrow(final int initialManaToGrow) {
//		this.initialManaToGrowPerField = toPerField(initialManaToGrow, 1);
//	}
	public double getInitialManaToGrowPerField() {
		return initialManaToGrowPerField;
	}
	public void setInitialManaToGrowPerField(final double initialManaToGrowPerField) {
		this.initialManaToGrowPerField = initialManaToGrowPerField;
	}
	public double getManaToGrowEachRound() {
		return fromPerField(manaToGrowEachRoundPerField, 1);
	}
//	public void setManaToGrowEachRound(final double manaToGrowEachRound) {
//		this.manaToGrowEachRoundPerField = toPerField(manaToGrowEachRound, 1);
//	}
	public double getManaToGrowEachRoundPerField() {
		return manaToGrowEachRoundPerField;
	}
	public void setManaToGrowEachRoundPerField(final double manaToGrowEachRoundPerField) {
		this.manaToGrowEachRoundPerField = manaToGrowEachRoundPerField;
	}
	public double getManaToGrowEachRoundInCockaigne() {
		return fromPerField(manaToGrowEachRoundInCockaignePerField, COCKAIGNE_RATIO);
	}
//	public void setManaToGrowEachRoundInCockaigne(final double manaToGrowEachRound) {
//		this.manaToGrowEachRoundInCockaignePerField = toPerField(manaToGrowEachRound, COCKAIGNE_RATIO);
//	}
	public double getManaToGrowEachRoundInCockaignePerField() {
		return manaToGrowEachRoundInCockaignePerField;
	}
	public void setManaToGrowEachRoundInCockaignePerField(final double manaToGrowEachRoundPerField) {
		this.manaToGrowEachRoundInCockaignePerField = manaToGrowEachRoundPerField;
	}
//	private double toPerField(final double absoluteValue, final int ratio) {
//		return absoluteValue / xsize / ysize * ratio * ratio;
//	}
	private double fromPerField(final double perFieldValue, final int ratio) {
		return perFieldValue * xsize * ysize / ratio / ratio;
	}
	public int getInitialNumberOfBacterions() {
		return initialNumberOfBacterions;
	}
	public void setInitialNumberOfBacterions(final int initialNumberOfBacterions) {
		this.initialNumberOfBacterions = initialNumberOfBacterions;
	}
	public boolean isGenealogyGraph() {
		return isGenealogyGraph;
	}
	public void setGenealogyGraph(final boolean isGenealogyGraph) {
		this.isGenealogyGraph = isGenealogyGraph;
	}
	public boolean isWorldTorus() {
		return isWorldTorus;
	}
	public void setWorldTorus(final boolean isWorldTorus) {
		this.isWorldTorus = isWorldTorus;
	}
	public int getManaEnergy() {
		return manaEnergy;
	}
	public void setManaEnergy(final int manaEnergy) {
		this.manaEnergy = manaEnergy;
	}
	public int getMaxEnergy() {
		return maxEnergy;
	}
	public void setMaxEnergy(final int maxEnergy) {
		this.maxEnergy = maxEnergy;
	}
	public int getXsize() {
		return xsize;
	}
	public void setXsize(final int xsize) {
		this.xsize = xsize;
	}
	public int getYsize() {
		return ysize;
	}
	public void setYsize(final int ysize) {
		this.ysize = ysize;
	}

	@Override
	public DeinbacProperties clone() {
		DeinbacProperties ret = null;
		try {
			ret = (DeinbacProperties) super.clone();
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}
}
