package de.ped.deinbac.logic;

import java.util.Arrays;

import de.ped.deinbac.logic.LifeProperties.Pattern;
import de.ped.tools.Bitfield32;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;

/**
 * The simulation logic, and access point for all other logic items.
 * 
 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class LifeGame extends Simulatable {
	public enum ImplementationMode {
		BOOL_ARRAY		 (false, false),
		BIT_FIELD		 (true,  false),
		BIT_FIELD_PRECALC(true,  true);

		private ImplementationMode(final boolean isBitFieldImpl, final boolean isPrecalc) {
			this.isBitFieldImpl = isBitFieldImpl;
			this.isPrecalc = isPrecalc;
		}
		final boolean isBitFieldImpl;
		final boolean isPrecalc;
	}

	public static ImplementationMode implMode = ImplementationMode.BIT_FIELD_PRECALC;
	public static int[] preCalculated;

	public static final int PRECALC_BIT_OFFSET = 1;
	public static final int PRECALC_CHUNKSIZE = 5;

	public static void init() {
		if ((null == preCalculated) && (ImplementationMode.BIT_FIELD_PRECALC == implMode)) {
			long startTime = System.currentTimeMillis();
			final LifeProperties pr = new LifeProperties();
			final int srcChunksize = PRECALC_CHUNKSIZE + 2;
			pr.setXsize(srcChunksize);
			pr.setYsize(3);
			pr.setWorldTorus(false);
			pr.setPattern(null);
			final LifeGame game = new LifeGame(pr);
			final int dimsize = (1 << (3 * srcChunksize));
			final int[] ret = new int[dimsize];
			game.status.initCopyPlane();
			for (int hash = 0; hash < ret.length; hash++) {
				final LifeWorldBitArray world = (LifeWorldBitArray) game.status.getWorld(false);
				final LifeWorldBitArray newworld = (LifeWorldBitArray) game.status.getWorld(true);
				game.status.initCopyPlane();
				final int bitmask = Bitfield32.bitmask(PRECALC_BIT_OFFSET - 1, srcChunksize);
				int upper  = (hash >> (0 * srcChunksize)) & bitmask;
				int middle = (hash >> (1 * srcChunksize)) & bitmask;
				int lower  = (hash >> (2 * srcChunksize)) & bitmask;
				world.pixels[0][0] = upper;
				world.pixels[1][0] = middle;
				world.pixels[2][0] = lower;
				newworld.pixels[1][0] = middle;
				for (int x = srcChunksize - 2; x >= 1; x--) {
					game.calcCellStep(x, 1);
				}
				int value = newworld.pixels[1][0] & Bitfield32.bitmask(PRECALC_BIT_OFFSET, PRECALC_CHUNKSIZE);
				ret[hash] = value;
				game.status.switchPlane();
			}
			preCalculated = ret;
			long endTime = System.currentTimeMillis();
			System.out.println("Initialization took " + (endTime - startTime) + "ms for " + ret.length + " hash entries.");
		}
	}

	private final LifeProperties props;
	private final LifeStatus status;

	public LifeGame(final LifeProperties props) {
		this.props = props;
		status = new LifeStatus(props.getXsize(), props.getYsize());
		addPattern(props.getPattern());
	}

	private void addPattern(final Pattern pattern) {
		if (null != pattern) {
			final String pat;
			if (Pattern.FREE_TEXT == pattern) {
				pat = props.getFreePattern();
			} else {
				pat = pattern.spec;
			}
			final String[] spec = pat.replaceFirst("" + LifeProperties.PATTERN_DELIM, "").replaceAll("\n", "")
					.split("\\" + LifeProperties.LINE_DELIM);
			final int xul = (props.getXsize() - spec[0].length()) / 2;
			final int yul = (props.getYsize() - spec.length) / 2;
			for (int y = 0; y < spec.length; y++) {
				int dx = 0;
				final String specRow = spec[y];
				for (int sx = 0; sx < specRow.length(); sx++) {
					int amount = 1;
					char c = specRow.charAt(sx);
					if (Character.isDigit(c)) {
						amount = 0; 
						while (Character.isDigit(c)) {
							amount = amount * 10 + Character.digit(c, 10); 
							sx++;
							if (sx < specRow.length()) {
								c = specRow.charAt(sx);
							} else { // A number followed by nothing is ignored.
								c = ' ';
								amount = 0;
							}
						}
					}
					while (0 < amount) {
						setPlayfieldAt(xul + dx, yul + y, (c == LifeProperties.ALIVE_CELL));
						dx++;
						amount--;
					}
				}
			}
		}
	}

	/**
	 * @return the logical size of the playfield
	 */
	@Override
	public PlayfieldDimension getPlayfieldSize() {
		return new PlayfieldDimension(props.getXsize(), props.getYsize());
	}

	public String getBoardString(boolean isHiddenPlane) {
		return status.getWorld(isHiddenPlane).toString();
	}

	/**
	 * @param x
	 *            the x position within the world
	 * @param y
	 *            the y position within the world
	 * @return if at position (x, y) is Mana
	 */
	@Override
	public int getPlayfieldAt(final int x, final int y) {
		return (status.getWorld().getAt(x, y) ? 1 : (status.getHull().getAt(x, y) ? -1 : 0));
	}
	public boolean getPlayfieldAt(final int x, final int y, final boolean isHiddenPlane) {
		return status.getWorld(isHiddenPlane).getAt(x, y);
	}

	public void setPlayfieldAt(int x, int y, boolean b) {
		status.getWorld().setAt(x, y, b);
	}

	public void togglePlayfieldAt(int x, int y) {
		status.getWorld().setAt(x, y, !status.getWorld().getAt(x, y));
	}

	public LifeProperties getProps() {
		return props;
	}

	@Override
	public long getTime() {
		return status.time;
	}

	/**
	 * Does a simulation step.
	 * 
	 * @return <code>true</code> iff the step resulted in any change to the world
	 */
	@Override
	public synchronized boolean step() {
		boolean ret = false;
		if (0 == status.time) {
			status.initCopyPlane();
		}
		status.time++;
		final int ylast = props.getYsize() - 1;
		final LifeWorldBitArray world;
		final LifeWorldBitArray newworld;
		if (implMode.isBitFieldImpl) {
			world = (LifeWorldBitArray) status.getWorld();
			newworld = (LifeWorldBitArray) status.getWorld(true);
		} else {
			world = null;
			newworld = null;
		}
		final int xsize = props.getXsize();
		final int srcChunksize = PRECALC_CHUNKSIZE + 2;
		final int smask = Bitfield32.bitmask(0, srcChunksize);
		final int dmask = Bitfield32.bitmask(PRECALC_BIT_OFFSET, PRECALC_CHUNKSIZE);
		for (int y = ylast; y >= 0; y--) {
			if ((null == preCalculated) || (null == world) || (0 == y) || (ylast == y)) {
				for (int x = xsize - 1; x >= 0; x--) {
					ret = calcCellStep(x, y) || ret;
				}
			} else {
				final int lastPrecalcX = xsize & ~(Integer.SIZE - 1);
				for (int x = 0; x < xsize; x++) {
					final int xm = x & (Integer.SIZE - 1); // The bit in the word
					if ((PRECALC_BIT_OFFSET <= xm) && (xm < PRECALC_BIT_OFFSET + PRECALC_CHUNKSIZE) 
							&& (x < lastPrecalcX)) { // Right border is classical, for torus worlds
						if (xm % PRECALC_CHUNKSIZE == PRECALC_BIT_OFFSET) { 
							final int wordIdx = x >>> 5; // (ln2(Integer.SIZE))
							final int shifts = (xm / PRECALC_CHUNKSIZE) * PRECALC_CHUNKSIZE;
							final int upperWord  = world.pixels[y - 1][wordIdx];
							final int middleWord = world.pixels[y    ][wordIdx];
							final int lowerWord  = world.pixels[y + 1][wordIdx];
							if ((0 == shifts) // first chunk of the word
									&& (0 == upperWord) && (0 == middleWord) && (0 == lowerWord) // empty part of world
									&& (0 == (props.getRuleset() & 0x1))) { // ruleset creates nothing out of empty part of world
								// If this part of the world is totally empty, then we don't need to look at single bits
								// So skip most other bits of the word and stop at the final border bit.
								x += Integer.SIZE - 3; // One off for increment in for loop, 2 off for border bits
								continue; 
							}
							final int upper  = (upperWord  >>> shifts) & smask;
							final int middle = (middleWord >>> shifts) & smask;
							final int lower  = (lowerWord  >>> shifts) & smask;
							if (upper != 0) {
								int i = 1; // TODO: Only there for breakpoint
							}
							final int hash = (upper << (0 * srcChunksize)) 
									| (middle << (1 * srcChunksize))
									| (lower  << (2 * srcChunksize));
							final int val = (preCalculated[hash] << shifts);
							final int dsmask = dmask << shifts;
							final int oldval = newworld.pixels[y][wordIdx];
							final int newval = (oldval & ~dsmask) | val;
							newworld.pixels[y][wordIdx] = newval;
							ret = (newval != oldval) || ret;
							// Skip the other bits in chunk, they are done already. One off for increment in for loop
							x += PRECALC_CHUNKSIZE - 1; 
						}
					} else {
						ret = calcCellStep(x, y) || ret;
					}
				}
			}
		}
		status.switchPlane();
		return ret;
	}

	private boolean calcCellStep(int x, int y) {
		boolean ret = false;
		boolean isLiving = status.getWorld().getAt(x, y);
		int n = 0;
		n += checkNeighbor(x - 1, y - 1);
		n += checkNeighbor(x - 1, y    );
		n += checkNeighbor(x - 1, y + 1);
		n += checkNeighbor(x    , y - 1);
		n += checkNeighbor(x    , y + 1);
		n += checkNeighbor(x + 1, y - 1);
		n += checkNeighbor(x + 1, y    );
		n += checkNeighbor(x + 1, y + 1);
		final int rulesetBits = props.getRuleset() & LifeProperties.getRulesetBit(isLiving, n);
		if (isLiving) { // Survival mode
			if (0 == rulesetBits) {
				status.getWorld(true).setAt(x, y, false);
				ret = true;
			}
		} else { // Birth mode
			if (0 != rulesetBits) {
				status.getWorld(true).setAt(x, y, true);
				ret = true;
			}
		}
		return ret;
	}

	private int checkNeighbor(final int x, final int y) {
		int ret = 0;
		if (props.isWorldTorus()) {
			int xnew = MathUtil.modulo(x, props.getXsize());
			int ynew = MathUtil.modulo(y, props.getYsize());
			if (status.getWorld().getAt(xnew, ynew)) {
				ret = 1;
			}
		} else {
			try {
				if (status.getWorld().getAt(x, y)) {
					ret = 1;
				}
			} catch (ArrayIndexOutOfBoundsException ex) {
				// Nothing to do
			}
		}
		return ret;
	}
}

/**
 * Contains the basic status of the simulation.
 * 
 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
 */
class LifeStatus {
	public long time;
	public LifeWorld board[];
	public LifeWorld hullBoard;
	private int bufIdx = 0;

	LifeStatus(int x, int y) {
		time = 0L;
		final boolean isBitField = LifeGame.implMode.isBitFieldImpl;
		board = new LifeWorld[]{ LifeWorld.factory(x, y, isBitField), LifeWorld.factory(x, y, isBitField) } ;
		hullBoard = LifeWorld.factory(x, y, isBitField);
	}

	public LifeWorld getWorld() {
		return getWorld(false);
	}

	public LifeWorld getWorld(final boolean isOtherBoard) {
		return board[isOtherBoard ? 1 - bufIdx : bufIdx];
	}
	public LifeWorld getHull() {
		return hullBoard;
	}

	public boolean getAt(final int x, final int y, final boolean isOtherBoard) {
		return getWorld(isOtherBoard).getAt(x, y);
	}

	public void setAt(final int x, final int y, final boolean value, final boolean isOtherBoard) {
		getWorld(isOtherBoard).setAt(x, y, value);
	}

	public void initCopyPlane() {
		getWorld(false).copyTo(getWorld(true));
		getWorld(false).copyTo(hullBoard);
	}

	public void switchPlane() {
		bufIdx = 1 - bufIdx;
		getWorld(false).copyTo(getWorld(true));
		hullBoard.uniteWith(getWorld(false));
	}
}

abstract class LifeWorld {
	public static LifeWorld factory(final int xsize, final int ysize, final boolean isBitImpl) {
		if (isBitImpl) {
			return new LifeWorldBitArray(xsize, ysize);
		} else {
			return new LifeWorldBooleanArray(xsize, ysize);
		}
	}
	
	public abstract int getXSize();
	public abstract int getYSize();
	public abstract boolean getAt(final int x, final int y);
	public abstract void setAt(final int x, final int y, final boolean value);
	public abstract boolean getAndSetAt(final int x, final int y, final boolean value);
	public abstract void copyTo(LifeWorld destination);
	public abstract void uniteWith(LifeWorld source);

	@Override
	public String toString() {
		StringBuffer eb = new StringBuffer();
		for (int y = 0; y < getYSize(); y++) {
			for (int x = 0; x < getXSize(); x++) {
				char c = (getAt(x, y) ? LifeProperties.ALIVE_CELL : LifeProperties.DEAD_CELL);
				eb.append(c);
			}
			if (y < getYSize() - 1) {
				eb.append(LifeProperties.LINE_DELIM);
			} else {
				eb.append(LifeProperties.PATTERN_DELIM);
			}
		}
		return eb.toString();
	}
}

class LifeWorldBooleanArray extends LifeWorld implements  Cloneable {
	/** The world itself, double-buffered */
	boolean[][] pixel;

	/**
	 * Creates an empty world.
	 * 
	 * @param xsize
	 *            the x-size of the world
	 * @param ysize
	 *            the y-size of the world
	 */
	public LifeWorldBooleanArray(final int xsize, final int ysize) {
		pixel = new boolean[xsize][ysize];
	}

	@Override
	public int getXSize() {
		return pixel.length;
	}

	@Override
	public int getYSize() {
		return pixel[0].length;
	}

	@Override
	public boolean getAt(final int x, final int y) {
		return pixel[x][y];
	}

	@Override
	public void setAt(final int x, final int y, final boolean value) {
		getAndSetAt(x, y, value);
	}

	/**
	 * Sets the new value at the given position, and returns the old value.
	 * 
	 * @param x
	 *            the x position in the world of the pixel to change
	 * @param y
	 *            the y position in the world of the pixel to change
	 * @param value
	 *            the new value to set. <code>true</code> represents mana,
	 *            <code>false</code> no mana.
	 * @return if there was mana before this operation. <code>true</code> represents
	 *         mana, <code>false</code> no mana.
	 */
	@Override
	public boolean getAndSetAt(final int x, final int y, final boolean value) {
		final boolean oldValue = pixel[x][y];
		pixel[x][y] = value;
		return oldValue;
	}

	@Override
	public void copyTo(final LifeWorld destination) {
		LifeWorldBooleanArray dest = (LifeWorldBooleanArray) destination;
		for (int x = pixel.length - 1; x >= 0; x--) {
        	System.arraycopy(pixel[x], 0, dest.pixel[x], 0, pixel[x].length);
		}
	}

	@Override
	public void uniteWith(final LifeWorld source) {
		LifeWorldBooleanArray s = (LifeWorldBooleanArray) source;
		for (int x = pixel.length - 1; x >= 0; x--) {
			for (int y = pixel[x].length - 1; y >= 0; y--) {
				pixel[x][y] = pixel[x][y] || s.pixel[x][y];
			}
		}
	}

	@Override
	public String toString() {
		StringBuffer eb = new StringBuffer();
		for (int y = 0; y < pixel.length; y++) {
			for (int x = 0; x < pixel[y].length; x++) {
				char c = (getAt(x, y) ? LifeProperties.ALIVE_CELL : LifeProperties.DEAD_CELL);
				eb.append(c);
			}
			if (y < pixel.length - 1) {
				eb.append(LifeProperties.LINE_DELIM);
			} else {
				eb.append(LifeProperties.PATTERN_DELIM);
			}
		}
		return eb.toString();
	}

	/** 
	 * Does a deep copy of a 2D array. 
	 * See <a href="https://stackoverflow.com/questions/1564832/how-do-i-do-a-deep-copy-of-a-2d-array-in-java#1564856">How do I do a deep copy of a 2D array in Java</a> for reference.
	 * @param original the array to clone
	 * @return the deep copy clone
	 */
	public static boolean[][] deepCopy(boolean[][] original) {
	    if (original == null) {
	        return null;
	    }
	    final boolean[][] result = new boolean[original.length][];
	    for (int i = 0; i < original.length; i++) {
	        result[i] = Arrays.copyOf(original[i], original[i].length);
	    }
	    return result;
	}	
}

class LifeWorldBitArray extends LifeWorld implements  Cloneable {
	/** The world itself, double-buffered */
	int[][] pixels;
	int xsize;

	/**
	 * Creates an empty world.
	 * 
	 * @param xsize
	 *            the x-size of the world
	 * @param ysize
	 *            the y-size of the world
	 */
	public LifeWorldBitArray(final int xsize, final int ysize) {
		this.xsize = xsize;
		final int words = (xsize + 31) >> 0x5;
		pixels = new int[ysize][words];
	}

	@Override
	public int getXSize() {
		return xsize;
	}

	@Override
	public int getYSize() {
		return pixels.length;
	}

	@Override
	public boolean getAt(final int x, final int y) {
		final int xword = x >>> 0x5;
		final int bitno = (x & 0x1f);
		return Bitfield32.isBit(pixels[y][xword], bitno);
	}

	@Override
	public void setAt(final int x, final int y, final boolean value) {
		getAndSetAt(x, y, value);
	}

	/**
	 * Sets the new value at the given position, and returns the old value.
	 * 
	 * @param x
	 *            the x position in the world of the pixel to change
	 * @param y
	 *            the y position in the world of the pixel to change
	 * @param value
	 *            the new value to set. <code>true</code> represents mana,
	 *            <code>false</code> no mana.
	 * @return if there was mana before this operation. <code>true</code> represents
	 *         mana, <code>false</code> no mana.
	 */
	@Override
	public boolean getAndSetAt(final int x, final int y, final boolean value) {
		final int xword = x >>> 0x5;
		final int bitno = (x & 0x1f);
		final int word = pixels[y][xword];
		final boolean oldValue = Bitfield32.isBit(word, bitno);
		pixels[y][xword] = Bitfield32.setBit(word, bitno, value);
		return oldValue;
	}

	@Override
	public void copyTo(final LifeWorld destination) {
		LifeWorldBitArray dest = (LifeWorldBitArray) destination;
		for (int y = pixels.length - 1; y >= 0; y--) {
        	System.arraycopy(pixels[y], 0, dest.pixels[y], 0, pixels[y].length);
		}
	}

	@Override
	public void uniteWith(final LifeWorld source) {
		LifeWorldBitArray s = (LifeWorldBitArray) source;
		for (int y = pixels.length - 1; y >= 0; y--) {
			for (int x = pixels[y].length - 1; x >= 0; x--) {
				pixels[y][x] |= s.pixels[y][x];
			}
		}
	}
}
