package de.ped.dsatool.dev;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

import de.ped.dsatool.dsa.generated.*;
import de.ped.dsatool.gui.DSATool;
import de.ped.tools.log.Logger;

/**
 * Creates a whole bunch of constant Java data structures from the data.xml file.
 * The idea was to be able to run without the overhead of an XML parser.
 * Only worked out for JABX 1, but is useless for JAXB 2.
 * @author Copyright (c) 2004 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class StandaloneImplInitializationSourcecodeGenerator {
	private static Logger logger = Logger.getRootLogger();

	/** How to indent the source code. */
	private static final String INDENT = "\t";

	/**
	 * @param n How often to concatenate indent.
	 * @return <em>n</em> times the {@link #INDENT}.
	 */
	public static String indent(final int n) {
		final StringBuffer s = new StringBuffer();
		for (int i = 0; i < n; i++) {
			s.append(INDENT);
		}
		return s.toString();
	}

	/** 
	 * Converts a string in a way that it can appear in source code.
	 * So it includes the whole string in double quotes, and inside it tranforms: 
	 * <table>
	 *   <caption>Transformations</caption>
	 *   <tbody>
	 *     <tr><th>From</th><th>To</th></tr>
	 *     <tr><td><code>\</code></td><td><code>\\</code></td>
	 *     <tr><td><code>&quot;</code></td><td><code>\&quot;</code></td>
	 *     <tr><td><code>Line feed</code></td><td><code>\n</code></td>
	 *     <tr><td><code>Tab</code></td><td><code>\t</code></td>
	 *   </tbody>
	 * </table> 
	 * 
	 * @param str the string to convert
	 * @return the converted string
	 */
	public static String sourcify(final String str) {
		return "\"" + str.replaceAll("\\\\", "\\\\\\\\").replaceAll("\"", "\\\\\"").replaceAll("\n", "\\\\n").replaceAll("\t", "\\\\t") + "\"";
	}

	/**
	 * Adds <em>indent</em> indentation steps beforehand, and adds a line break at the end.
	 * @param indent The number of indent steps to prefix.
	 * @param str The string to modify.
	 * @return The concatenation of Indentation, string and line break.
	 */
	public static String linify(final int indent, final String str) {
		return indent(indent) + str + "\n";
	}

	protected static final Class<?> FLOAT_CLASS = new Float(0).getClass();
	protected static final Class<?> OBJECT_CLASS = new Object().getClass();

	/** Used to create unique variable names. */
	private int varNameCounter = 0;

	/** Keeps all the generated variable names for objects to generate. */
	protected HashMap<Object, String> varNames = new HashMap<Object, String>();
	
	/**
	 * Return type for findOrCreateVarName and sourcifyValue.
	 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de),
	 */
	protected class VarNameCreationResult {
		/** The found or created variable name. If <code>null</code>, this is no variable on its own, but a list. */
		public String varName;
		
		/** <code>true</code> iff the variable name didn't exist before and had to be newly created. */
		public boolean isNewlyCreated;
	};

	/**
	 * Checks if for a given object a variable name is defined.
	 * If not, it gets created.
	 * @param o The object that needs a local variable name.
	 * @return The variable name for the object, 
	 *   plus the information if the variable name had to be newly created or existed before.
	 */
	protected VarNameCreationResult findOrCreateVarName(final Object o) {
		String varName = (String)varNames.get(o);
		boolean isnc = false;
		if (null == varName) {
			varName = "o";
			try {
				final BusinessObjectType bo = (BusinessObjectType)o;
				varName += bo.getID();
			} catch (final Exception e) {
				varName += varNameCounter;
				varNameCounter++;
			}
			varNames.put(o, varName);
			isnc = true;
		}
		final VarNameCreationResult ret = new VarNameCreationResult(); 
		ret.varName = varName;
		ret.isNewlyCreated = isnc;
		return ret;
	}

	/** The writer where we collect all the attributes (variables) for the new objects. */
	protected Writer swDeclarations = new StringWriter();
	
	/** We need several writers for several initialization methods. Here we store them. */
	protected HashMap<Class<?>, StringWriter> writers = new HashMap<Class<?>, StringWriter>();

	/**
	 * Finds or creates a writer where to write the source code for this object to.
	 * This is necessary because Java has a maximum limit of how long a method is allowed to be,
	 * so we need to creatively split the definition code.
	 * One writer represents later one method containing initialization code.
	 * This implementation creates a writer (and thus an initialization method) for each class, 
	 * so objects of the same class are initialized in the same method.
	 * Change this implementation of you create huge amounts of objects of the same class.
	 * @param o The object that will be defined in source code.
	 * @return The writer where to write the source code for this object to.
	 */
	protected Writer findOrCreateDefinitionWriter(final Object o) {
		final Class<?> key = o.getClass();
		StringWriter ret = writers.get(key);
		if (null == ret) {
			ret = new StringWriter();
			writers.put(key, ret);
		}
		return ret;
	}
	
	/**
	 * Tries to express the object o as source code for a constant value.
	 * So for strings, it puts hyphens around, for objects it adds the name of the variable, etc.
	 * @param codeLine Where to append the source code for the constant value.
	 * @param o The object to represent as source code.
	 * @param isToObjectifyBuiltinTypes If <code>true</code>, represents the object always as is, e.g., as from type Integer.
	 *   If <code>false</code>, does builtin code representation as int, if possible.
	 *   The difference is necessary, as Collections expect objects, 
	 *   but the setter methods expect builtin types.
	 * @return Indicator what else is there to do. 
	 *   If varName is <code>null</code>, then the object was a list, and all the work has to happen elsewhere.
	 *   If isNewlyCreted, then <em>o</em> is a new object, and that should be declared and defined elsewhere.
	 */
	protected VarNameCreationResult sourcifyValue(
			final StringBuffer codeLine, 
			final Object o, 
			final boolean isToObjectifyBuiltinTypes) {
		VarNameCreationResult ret = null;
		if (null == o) {
			codeLine.append("null");
		} else if (o instanceof String) {
			codeLine.append(sourcify(o.toString()));
//		} else if (o instanceof XMLSerializable) { // Is it a child XML object?
//			ret = findOrCreateVarName(o);
//			codeLine.append(ret.varName);
		} else if (o instanceof List) {
			ret = new VarNameCreationResult();
			ret.isNewlyCreated = false;
			ret.varName = null;
			// codeLine gets nothing appended, this must be handled elsewhere.
		} else { // Here I assume Integer, Float or other Object representants of basic types.
			if (isToObjectifyBuiltinTypes) {
				codeLine.append("new ");
				codeLine.append(o.getClass().getName());
				codeLine.append("(");
			}
			codeLine.append(o);
			if (o.getClass().equals(FLOAT_CLASS)) {
				codeLine.append("F");
			}
			if (isToObjectifyBuiltinTypes) {
				codeLine.append(")");
			}
		}
		return ret;
	}
    
    private final static String METHOD_PREFIX_GET = "get";
    private final static String METHOD_PREFIX_IS = "is";
	
	/**
	 * Creates java source code that builds up the data contained in the XML file.
	 * @param parentName The name of the variable where to add this object. 
	 *   If <code>null</code>, this object is the root object.
	 * @param objectName The object to create becomes this name as local variable.
	 * @param o The object to create in source code.
	 * @throws IOException Whatever {@link java.io.StringWriter#write} may throw.
	 */
	public void build(final String parentName, final String objectName, final Object o) throws IOException {
		// Constructor
		final Class<?> clazz = o.getClass();
		final String fullClassName = clazz.getName();
		final String[] classNameParts = fullClassName.split("\\.");
		final String className = classNameParts[classNameParts.length - 1];
		final Writer swDefinitions = (Writer)findOrCreateDefinitionWriter(o);
	    swDeclarations.write(linify(1, className + " " + objectName + " = new " + className + "();"));
	    // Attributes
        final Method[] methods = clazz.getMethods();
        for (int i = 0; i < methods.length; i++) {
        	Method method = methods[i];
        	String methodName = method.getName();
        	if ((null != methodName) 
        			&& ((methodName.startsWith(METHOD_PREFIX_GET) 
        					&& !methodName.equals("getPrimaryInterface") 
        					&& !methodName.equals("getClass")) 
        			|| (methodName.startsWith(METHOD_PREFIX_IS)))) {
        	    final String returnTypeName = method.getReturnType().getName();
        		logger.debug(methodName + " -> " + returnTypeName);
        	    Object returnValue = null;
				try {
					returnValue = method.invoke(o, (Object[]) null );
				} catch (final IllegalArgumentException e) {
					e.printStackTrace();
				} catch (final IllegalAccessException e) {
					e.printStackTrace();
				} catch (final InvocationTargetException e) {
					e.printStackTrace();
				}
				logger.debug((null == returnValue) ? "null" : returnValue.toString());
				// Now build the statement
				boolean isToCreateStatement = true;
				final StringBuffer codeLine = new StringBuffer(objectName);
				codeLine.append(".set");
                int indexAfterMethodNamePrefix = 3;
                if (methodName.startsWith(METHOD_PREFIX_IS)) {
                    indexAfterMethodNamePrefix = 2;
                }
				codeLine.append(methodName.substring(indexAfterMethodNamePrefix));
				codeLine.append("(");
				VarNameCreationResult creationResult = sourcifyValue(codeLine, returnValue, false);
				if (null != creationResult) {
					if (null == creationResult.varName) {
						// It's a list
	            		isToCreateStatement = false;
	            		final List<?> l = (List<?>)returnValue;
	            		final Iterator<?> it = l.iterator();
	            		while (it.hasNext()) {
	            			final Object listElement = it.next();
	            			final StringBuffer listElementCodeLine = new StringBuffer(objectName);
	            			listElementCodeLine.append(".");
	            			listElementCodeLine.append(methodName);
	            			listElementCodeLine.append("().add(");
	            			
	            			final VarNameCreationResult listElementCreationResult = sourcifyValue(listElementCodeLine, listElement, true);
	        				if (null != listElementCreationResult) {
	        					if (null == listElementCreationResult.varName) {
	        						// It's a list within a list - this is not implemented.
	        						throw new ArrayIndexOutOfBoundsException("A list within a list is beyond my abilities.");
	        					} else if (listElementCreationResult.isNewlyCreated) {
		    						// It's a new single child object
		                            build(objectName, listElementCreationResult.varName, listElement);
		    					}
	        				}
	            			listElementCodeLine.append(");");
	                        swDefinitions.write(linify(2, listElementCodeLine.toString()));
	            		}
					} else if (creationResult.isNewlyCreated) {
						// It's a new single child object
                        build(objectName, creationResult.varName, returnValue);
					}
				}
				codeLine.append(");");
				if (isToCreateStatement) {
				    swDefinitions.write(linify(2, codeLine.toString()));
				}
        	}
        }
        swDefinitions.write(linify(2, ""));
	}
	
	/**
	 * Generates the complete source code for a factory class that builds the whole DSA structure.
	 * It takes an XML file as input. The resulting code is independent from XML 
	 * and can thus be used in a context where XML reading capability would be to 
	 * complex to integrate.
	 */
	public void generateDsaFactory() {
        try  {
        	final JAXBContext jc = JAXBContext.newInstance( "de.ped.dsatool.dsa.generated" ); //$NON-NLS-1$
        	final Unmarshaller u = jc.createUnmarshaller();
        	InputStream s = null;
            if ((null == DSATool.getInstance()) || (DSATool.isApplication())) {
        		s = new FileInputStream( "src\\resources\\data.xml" ); //$NON-NLS-1$
            } else {
            	final URL codebase = DSATool.getInstance().getCodeBase();
            	final URL filename = new URL(codebase.toString() + "data.xml"); //$NON-NLS-1$
            	logger.debug(filename.toString());
            	s = new DataInputStream( filename.openStream() );
            }
            final DsaType delegate = (DsaType)u.unmarshal(s);
            
            build(null, "ret", delegate);

            final FileWriter ow = new FileWriter("src\\de\\ped\\dsatool\\dsa\\generated\\standaloneimpl\\StandaloneImplInitializer.java");
            ow.write(linify(0, "package de.ped.dsatool.dsa.generated.standaloneimpl;"));
            ow.write(linify(0, ""));
            ow.write(linify(0, "import de.ped.dsatool.dsa.generated.*;"));
            ow.write(linify(0, ""));
            ow.write(linify(0, "public class StandaloneImplInitializer"));
            ow.write(linify(0, "{"));
            ow.write(swDeclarations.toString());
            final Iterator<StringWriter> it = writers.values().iterator();
            ow.write(linify(0, ""));
            int initMethods = 0;
            while (it.hasNext()) {
            	final StringWriter writer = (StringWriter)it.next();
                ow.write(linify(1, "private void init" + initMethods + "()"));
                ow.write(linify(1, "{"));
                ow.write(writer.toString());
                ow.write(linify(1, "}"));
                ow.write(linify(0, ""));
            	initMethods++;
            }
            ow.write(linify(1, "public DsaType createDsa()"));
            ow.write(linify(1, "{"));
            for (int i = 0; i < initMethods; i++) {
                ow.write(linify(2, "init" + i + "();"));
            }
            ow.write(linify(2, "return ret;"));
            ow.write(linify(1, "}"));
            ow.write(linify(0, "}"));
            ow.close();
        } catch (final JAXBException je ) {
            je.printStackTrace();
        } catch (final IOException ioe ) {
            ioe.printStackTrace();
        }
	}
	
	/**
	 * Creates the source code for a factory class by calling generateDsaFactory().
	 * @param args not interpreted.
	 * @see #generateDsaFactory()
	 */
	public static void main(final String[] args) {
		final StandaloneImplInitializationSourcecodeGenerator instance = new StandaloneImplInitializationSourcecodeGenerator();
		instance.generateDsaFactory();
	}
}
