package de.ped.dsatool.gui;

import java.awt.Canvas;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.JFrame;

import de.ped.tools.MathUtil;
import de.ped.tools.log.Logger;

/**
 * Draws a map and reacts on user input.
 * 
 * @author Copyright (c) 2007 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AventuriaMapCanvas extends Canvas implements KeyListener, MouseMotionListener, MouseListener {
	private static final long serialVersionUID = 1L;

	private static Logger logger = DSATool.getLogger();

	/** The maximum width the aventuria map image should get. */
	static final int WINDOW_MAX_WIDTH =  500; //1024;

	/** The maximum height the aventuria map image should get. */
	static final int WINDOW_MAX_HEIGHT = 500; //800;

	/** Where I'm embedded */
	private JFrame parent;

	/** What to display */
	private AventuriaMapUIModel map;

	public AventuriaMapCanvas(final JFrame parent, final AventuriaMapUIModel map) {
		this.parent = parent;
		this.map = map;
		setSize(WINDOW_MAX_WIDTH, WINDOW_MAX_HEIGHT);
		addKeyListener(this);
		addMouseMotionListener(this);
		addMouseListener(this);
	}

	@Override
	public void paint(final Graphics graphics) {
		graphics.drawImage(map.getDisplay(), 0, 0, null);
	}

	/** Should be called (and is called internally) if any of the data to be displayed could have changed. */
	public void recalculate() {
		if (null != map) {
			// (If map is smaller, make the window fitting.)
			final Dimension windowSize = getSize();
			windowSize.width = MathUtil.bound((int) windowSize.getWidth(), 0, WINDOW_MAX_WIDTH);
			windowSize.height = MathUtil.bound((int) windowSize.getHeight(), 0, WINDOW_MAX_HEIGHT);
			map.setSize(windowSize);

			map.recalculate();
		}
		if (null != parent) {
			parent.repaint();
		}
	}

	@Override
	public Dimension getPreferredSize() {
		Dimension ret;
		if (null == map) {
			ret = getMaximumSize();
		} else {
			ret = map.getSize();
		}
		return ret;
	}

	@Override
	public Dimension getMaximumSize() {
		return new Dimension(WINDOW_MAX_WIDTH, WINDOW_MAX_HEIGHT);
	}

	@Override
	public Dimension getMinimumSize() {
		return new Dimension(50, 50);
	}

	@Override
	public void keyPressed(final KeyEvent e) {
		logger.debug("KeyTyped: keyChar=" + e.getKeyChar() + ", keyCode=" + e.getKeyCode());
		boolean isChanged = false;
		if (null != map) {
			switch (e.getKeyCode()) {
			case KeyEvent.VK_LEFT:
				if (e.isShiftDown()) {
					isChanged = map.modifySelectedPatch(0, -AventuriaMapUIModel.MODIFY_STEP);
				} else if (e.isControlDown()) {
					isChanged = map.modifySelectedPatch(2, -AventuriaMapUIModel.MODIFY_STEP);
				} else {
					isChanged = map.moveHotspot(-1, 0);
				}
				break;
			case KeyEvent.VK_RIGHT:
				if (e.isShiftDown()) {
					isChanged = map.modifySelectedPatch(0, AventuriaMapUIModel.MODIFY_STEP);
				} else if (e.isControlDown()) {
					isChanged = map.modifySelectedPatch(2, AventuriaMapUIModel.MODIFY_STEP);
				} else {
					isChanged = map.moveHotspot(1, 0);
				}
				break;
			case KeyEvent.VK_UP:
				if (e.isShiftDown()) {
					isChanged = map.modifySelectedPatch(1, -AventuriaMapUIModel.MODIFY_STEP);
				} else if (e.isControlDown()) {
					isChanged = map.modifySelectedPatch(3, -AventuriaMapUIModel.MODIFY_STEP);
				} else {
					isChanged = map.moveHotspot(0, -1);
				}
				break;
			case KeyEvent.VK_DOWN:
				if (e.isShiftDown()) {
					isChanged = map.modifySelectedPatch(1, AventuriaMapUIModel.MODIFY_STEP);
				} else if (e.isControlDown()) {
					isChanged = map.modifySelectedPatch(3, AventuriaMapUIModel.MODIFY_STEP);
				} else {
					isChanged = map.moveHotspot(0, 1);
				}
				break;
			case KeyEvent.VK_MINUS:
				map.decrementZoomStepIndex();
				isChanged = true;
				break;
			case KeyEvent.VK_PLUS:
				map.incrementZoomStepIndex();
				isChanged = true;
				break;
			case KeyEvent.VK_G:
				map.switchGridState();
				isChanged = true;
				break;
			case KeyEvent.VK_S:
				map.togglePatchSelectionDisplayed();
				isChanged = true;
				break;
			case KeyEvent.VK_P:
				map.selectNextPatch();
				isChanged = true;
				break;
			}
			if (isChanged) {
				recalculate();
			}
		}
	}

	@Override
	public void keyReleased(final KeyEvent e) {
		// Nothing to do
	}

	@Override
	public void keyTyped(final KeyEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseDragged(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseMoved(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseClicked(final MouseEvent e) {
		if (null != map) {
			map.setMousePoint(e.getPoint());
			recalculate();
		}
	}

	@Override
	public void mouseEntered(final MouseEvent e) {
		if (null != map) {
			map.setMousePoint(e.getPoint());
			recalculate();
		}
	}

	@Override
	public void mouseExited(MouseEvent e) {
		if (null != map) {
			map.setMousePoint(null);
			recalculate();
		}
	}

	@Override
	public void mousePressed(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseReleased(final MouseEvent e) {
		// Nothing to do
	}
}
