package de.ped.dsatool.gui;

import java.util.Enumeration;
import java.util.List;

import javax.swing.tree.DefaultMutableTreeNode;

import de.ped.dsatool.dsa.generated.BasicAreaType;
import de.ped.dsatool.dsa.generated.MapLayerAreaType;
import de.ped.dsatool.dsa.generated.MapLayerPatchType;
import de.ped.dsatool.dsa.generated.MapLayerType;
import de.ped.dsatool.logic.DSAEnv;
import de.ped.tools.CollectionUtil;
import de.ped.tools.Messages;
import de.ped.tools.gui.CheckTreeNode;
import de.ped.tools.gui.TristateCheckBox;

/**
 * Represents the data behind a tree node for the layer selection in the AventuriaMap.
 * 
 * @author Copyright (c) 2007 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AventuriaMapLayerTreeNode extends DefaultMutableTreeNode implements CheckTreeNode {
	private static final long serialVersionUID = 1L;

	public static final int ROOT_LEVEL = 0;
	public static final int LAYER_LEVEL = 1;
	public static final int PATCH_LEVEL = 2;

	/**
	 * The mode the node can be in.
	 * 
	 * @author Copyright (c) 2007 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Mode {
		PATCH_MODE, AREA_MODE, GRID_MODE, CROSSHAIR_MODE,
	}

	private int level;
	private Mode mode;
	private MapLayerType mapLayer;
	private MapLayerPatchType mapLayerPatch;
	private MapLayerAreaType mapLayerArea;
	private String name;

	/** Only used for leafs. */
	private boolean isSelected;

	private boolean isSelectedDefault;

	private AventuriaMapUIModel changeListener;

	/**
	 * Constructor.
	 * 
	 * @param listener
	 *            the listener who gets informed about a change of the selection
	 *            state
	 * @param level
	 *            the level the node is in
	 * @param mode
	 *            the mode the node is in
	 * @param ml
	 *            the map layer for the node. Only evaluated if the node is in
	 *            {@link Mode#GRID_MODE} and at least {@link #LAYER_LEVEL}.
	 * @param mlp
	 *            the map layer patch for the node. Only evaluated if the node
	 *            is in {@link Mode#GRID_MODE} and on {@link #PATCH_LEVEL}.
	 * @param mla
	 *            the map layer area for the node. Only evaluated if the node is
	 *            in {@link Mode#AREA_MODE} and on {@link #PATCH_LEVEL}.
	 * @param name
	 *            the message key of the node name
	 * @param isSelected
	 *            if <code>true</code>, the node is initially selected. Only
	 *            evaluated if the node is a leaf.
	 */
	public AventuriaMapLayerTreeNode(
			final AventuriaMapUIModel listener, 
			final int level, 
			final Mode mode, 
			final MapLayerType ml, 
			final MapLayerPatchType mlp, 
			final MapLayerAreaType mla, 
			final String name, 
			final boolean isSelected) {
		this.changeListener = listener;
		this.level = level;
		this.mode = mode;
		this.mapLayer = ml;
		this.mapLayerPatch = mlp;
		this.mapLayerArea = mla;
		this.isSelectedDefault = isSelected;
		this.isSelected = isSelected;
		this.name = calculateName(name);
		// TODO: This is a hack to show undefined areas.
		if (null != mla) {
			List<BasicAreaType> bats = mla.getArea();
			if (CollectionUtil.isNullOrEmpty(bats)) {
				this.name = "[!]" + this.name;
			}
		}
	}

	/**
	 * Returns the selection state of this node. 
	 * For leafs, this is straight-forward. For non-leafs it is calculated recursively, depending on the states of the children.
	 * TODO: Don't calculate this every time - only if the model changes.
	 */
	@Override
	public TristateCheckBox.State getSelectionState() {
		TristateCheckBox.State ret = TristateCheckBox.DONT_CARE;
		if (isLeaf()) {
			ret = (isSelected ? TristateCheckBox.SELECTED : TristateCheckBox.NOT_SELECTED);
		} else {
			final Enumeration<?> e = this.children();
			boolean areAllSelected = true;
			boolean areAllDeselected = true;
			while (e.hasMoreElements() && (areAllSelected || areAllDeselected)) {
				AventuriaMapLayerTreeNode child = (AventuriaMapLayerTreeNode) e.nextElement();
				TristateCheckBox.State childState = child.getSelectionState();
				if (TristateCheckBox.NOT_SELECTED == childState) {
					areAllSelected = false;
				} else if (TristateCheckBox.SELECTED == childState) {
					areAllDeselected = false;
				} else {
					areAllSelected = false;
					areAllDeselected = false;
				}
			}
			if (areAllSelected) {
				ret = TristateCheckBox.SELECTED;
			} else if (areAllDeselected) {
				ret = TristateCheckBox.NOT_SELECTED;
			}
		}
		return ret;
	}

	/** 
	 * @return if the selection state is fully selected for this node 
	 */
	public boolean isSelected() {
		return (TristateCheckBox.SELECTED == getSelectionState());
	}

	/**
	 * Sets the selection state of this node. Does not inform the change
	 * listener.
	 * 
	 * @param value
	 *            <code>true</code> for selected, otherwise unselected.
	 * @param deepen
	 *            if <code>true</code>, and this node is no leaf, this sets the
	 *            selection state of all the children recursively
	 * @return <code>true</code> if some selection state of a node really
	 *         changed
	 */
	protected boolean setSelected(final boolean value, final boolean deepen) {
		boolean ret = false;
		if (isLeaf()) {
			ret = (isSelected != value);
			isSelected = value;
		} else {
			if (deepen) {
				final Enumeration<?> e = this.children();
				while (e.hasMoreElements()) {
					final AventuriaMapLayerTreeNode child = (AventuriaMapLayerTreeNode) e.nextElement();
					final boolean cret = child.setSelected(value, true);
					ret = ret || cret;
				}
			}
		}
		return ret;
	}

	/**
	 * Sets the selection state of this node. If this node is no leaf, this sets
	 * the selection state of all the children recursively. The listener is only
	 * informed about the selection state change for this node, not for all the
	 * children.
	 * 
	 * @param value
	 *            <code>true</code> for selected, otherwise unselected
	 */
	@Override
	public void setSelected(final boolean value) {
		final boolean changed = setSelected(value, true);
		if (changed) {
			changeListener.nodeChanged(this);
		}
	}

	protected boolean setSelectedToDefault(final boolean deepen) {
		boolean ret = false;
		if (isLeaf()) {
			ret = (isSelected != isSelectedDefault);
			isSelected = isSelectedDefault;
		} else {
			if (deepen) {
				final Enumeration<?> e = this.children();
				while (e.hasMoreElements()) {
					final AventuriaMapLayerTreeNode child = (AventuriaMapLayerTreeNode)e.nextElement();
					final boolean cret = child.setSelectedToDefault(true);
					ret = ret || cret;
				}
			}
		}
		return ret;
	}

	/**
	 * Sets the selection state of this node (and recursively all nodes below) back to default values.
	 * The listener is only informed about the selection state change for this node,
	 * not for all the children.
	 */
	public void setSelectedToDefault() {
		final boolean changed = setSelectedToDefault(true);
		if (changed) {
			changeListener.nodeChanged(this);
		}
	}

	/**
	 * This text will be presented in the tree as node name.
	 * 
	 * @param name
	 *            the key postfix
	 * @return the text to show
	 */
	private String calculateName(final String name) {
		final Messages messages = DSAEnv.instance().messages();
		final String ret = messages.getString("MapLayers." + name, name);
		return ret;
	}

	/**
	 * @return the level - is one of {@link #ROOT_LEVEL}, {@link #LAYER_LEVEL}
	 *         or {@link #PATCH_LEVEL}
	 */
	@Override
	public int getLevel() {
		return level;
	}

	/**
	 * @return the MapLayer.
	 *   Only used on {@link #LAYER_LEVEL} and {@link #PATCH_LEVEL}, 
	 *   and only used in {@link Mode#PATCH_MODE}.
	 */
	public MapLayerType getMapLayer() {
		return mapLayer;
	}

	/**
	 * @return the MapLayerPatch. Only used on {@link #PATCH_LEVEL}, and only
	 *         used in {@link Mode#PATCH_MODE}.
	 */
	public MapLayerPatchType getMapLayerPatch() {
		return mapLayerPatch;
	}

	/**
	 * @return the MapLayerArea. Only used on {@link #PATCH_LEVEL}, and only
	 *         used in {@link Mode#AREA_MODE}.
	 */
	public MapLayerAreaType getMapLayerArea() {
		return mapLayerArea;
	}

	/**
	 * @return the mode - is one of {@link Mode#PATCH_MODE},
	 *         {@link Mode#CROSSHAIR_MODE} or {@link Mode#GRID_MODE}.
	 */
	public Mode getMode() {
		return mode;
	}

	/**
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * @return the text that will show up in the tree as node name
	 */
	@Override
	public String toString() {
		return name;
	}
}
