package de.ped.dsatool.gui;

// TODO: Catch resize event and redraw.

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Composite;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.imageio.ImageIO;

import de.ped.tools.Assert;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.TextDocumentType;
import de.ped.tools.MathUtil;
import de.ped.tools.MathVector;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.MapLayerAreaType;
import de.ped.dsatool.dsa.generated.MapType;
import de.ped.dsatool.logic.DSA;
import de.ped.dsatool.logic.DSAEnv;
import de.ped.dsatool.logic.Map;
import de.ped.dsatool.logic.Model;
import de.ped.dsatool.dsa.generated.MapLayerType;
import de.ped.dsatool.dsa.generated.MapLayerPatchType;

/**
 * Draws a scrollable and zoomable map.
 * 
 * @author Copyright (c) 2005 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AventuriaMapUIModel {

	private static Logger logger = DSATool.getLogger();

	/************************************************************/
	/* Zoom model                                               */
	/**
	 * @return the current zoom step index
	 */
	public int getZoomStepIndex() {
		return dataModel.getAventuriaMapZoomStepIndex();
	}
	
	public void setZoomStepIndex(int value) {
		dataModel.getOrCreateAventuriaMapModel().setZoomStepIndex(value);
	}

	public void incrementZoomStepIndex() {
		dataModel.getOrCreateAventuriaMapModel().incrementZoomStepIndex();
	}

	public void decrementZoomStepIndex() {
		dataModel.getOrCreateAventuriaMapModel().decrementZoomStepIndex();
	}

	/************************************************************/
	/* Position model                                           */
	
	/** Amount of the whole image a step in a direction will go. */
	private static final double MOVE_STEP = 0.01;

	/************************************************************/
	/* Crosshair model                                          */

	private static final int CROSSHAIR_MIN = 8;
	private static final int CROSSHAIR_STEPSIZE = 5;
	private static final int CROSSHAIR_STEPS = 6;
	private static final int CROSSHAIR_THICKNESS = 3;

	/**
	 * Draws a crosshair.
	 * 
	 * @param graphics
	 *            the graphics object to draw to
	 * @param hotSpotX
	 *            the x position where the crosshair points to
	 * @param hotSpotY
	 *            the x position where the crosshair points to
	 */
	@SuppressWarnings("unused")
	private void drawCrosshair(final Graphics graphics, final int hotSpotX, final int hotSpotY) {
		int pos = CROSSHAIR_MIN;
		final int thicknessOffset = CROSSHAIR_THICKNESS/2;
		final int thickness = CROSSHAIR_THICKNESS - 1;
		for (int i = 0; i < CROSSHAIR_STEPS; i++) {
			graphics.setColor((i % 2 == 0) ? Color.WHITE : Color.BLACK);
			if (1 == CROSSHAIR_THICKNESS) {
				graphics.drawLine(
						hotSpotX, hotSpotY - pos, 
						hotSpotX, hotSpotY - (pos + CROSSHAIR_STEPSIZE - 1));
				graphics.drawLine(
						hotSpotX, hotSpotY + pos, 
						hotSpotX, hotSpotY + (pos + CROSSHAIR_STEPSIZE - 1));
				graphics.drawLine(
						hotSpotX - pos, hotSpotY, 
						hotSpotX - (pos + CROSSHAIR_STEPSIZE - 1), hotSpotY);
				graphics.drawLine(
						hotSpotX + pos, hotSpotY, 
						hotSpotX + (pos + CROSSHAIR_STEPSIZE - 1), hotSpotY);
			} else {
				graphics.drawRect(hotSpotX - thicknessOffset, hotSpotY - pos, 
						thickness, (CROSSHAIR_STEPSIZE - 1));
				graphics.drawRect(hotSpotX - thicknessOffset, hotSpotY + pos - CROSSHAIR_STEPSIZE, 
						thickness, (CROSSHAIR_STEPSIZE - 1));
				graphics.drawRect(hotSpotX - pos, hotSpotY - thicknessOffset, 
						(CROSSHAIR_STEPSIZE - 1), thickness);
				graphics.drawRect(hotSpotX + pos - CROSSHAIR_STEPSIZE, hotSpotY - thicknessOffset, 
						(CROSSHAIR_STEPSIZE - 1), thickness);
			}
			pos += CROSSHAIR_STEPSIZE;
		}
		graphics.setColor(Color.WHITE);
		graphics.drawOval(hotSpotX - pos, hotSpotY - pos, 2 * pos - 1, 2 * pos - 1);
	}

	/************************************************************/
	/* Color mapping                                            */

	private static HashMap<String, Color> colorMap;

	private static Color findColor(final String colorName) {
		if (null == colorMap) {
			// First initialize completely, then set static variable, to avoid multithreading problems.
			final HashMap<String, Color> cm = new HashMap<String, Color>();
			cm.put("black", Color.black);
			cm.put("blue", Color.blue);
			cm.put("cyan", Color.cyan);
			cm.put("white", Color.white);
			cm.put("darkGrey", Color.darkGray);
			cm.put("grey", Color.gray);
			cm.put("green", Color.green);
			cm.put("lightGrey", Color.lightGray);
			cm.put("magenta", Color.magenta);
			cm.put("orange", Color.orange);
			cm.put("pink", Color.pink);
			cm.put("red", Color.red);
			cm.put("white", Color.white);
			cm.put("yellow", Color.yellow);

			cm.put("darkBlue", Color.blue.darker());
			cm.put("darkCyan", Color.cyan.darker());
			cm.put("darkWhite", Color.white.darker());
			cm.put("darkGreen", Color.green.darker());
			cm.put("darkMagenta", Color.magenta.darker());
			cm.put("darkOrange", Color.orange.darker());
			cm.put("darkPink", Color.pink.darker());
			cm.put("darkRed", Color.red.darker());
			cm.put("darkYellow", Color.yellow.darker());

			cm.put("lightBlue", Color.blue.brighter().brighter());
			cm.put("lightCyan", Color.cyan.brighter());
			cm.put("lightWhite", Color.white.brighter());
			cm.put("lightGreen", Color.green.brighter());
			cm.put("lightMagenta", Color.magenta.brighter());
			cm.put("lightOrange", Color.orange.brighter());
			cm.put("lightPink", Color.pink.brighter());
			cm.put("lightRed", Color.red.brighter());
			cm.put("lightYellow", Color.yellow.brighter());
			colorMap = cm;
		}
		Color c = (Color) colorMap.get(colorName);
		if (null == c) {
			c = Color.gray;
		}
		return c;
	}

	/************************************************************/
	/* Patch images                                             */

	private static HashMap<MapLayerPatchType, Image> patchToImg = new HashMap<MapLayerPatchType, Image>();

	/**
	 * Loads all the patch image files of all layers of the map. Must be called
	 * after or during creation of the object, before using it.
	 * 
	 * @param map
	 *            the map to load
	 */
	private static void load(final MapType map) {
		// For performance reasons, start all file loads at once.
		final Iterator<MapLayerType> mit = map.getMapLayer().iterator();
		while (mit.hasNext()) {
			final MapLayerType ml = (MapLayerType) mit.next();
			final Iterator<MapLayerPatchType> lit = ml.getMapLayerPatch().iterator();
			while (lit.hasNext()) {
				final MapLayerPatchType mlp = (MapLayerPatchType) lit.next();
				final String filename = ml.getPath() + "/" + mlp.getFilename();
				final URL filenameUrl = DSAEnv.instance().resources().getURLToResource(filename);
				logger.info("Loading img file from " + filenameUrl.toString());
				try {
					final BufferedImage bi = ImageIO.read(filenameUrl);
					patchToImg.put(mlp, bi);
				} catch (final IOException e) {
					logger.error("Unable to read image file " + filenameUrl, e);
				}
			}
		}
	}

	/**
	 * Norms a patch. Must be called after creation of the object, before using
	 * it.
	 * 
	 * @param mlp
	 *            the map layer patch to look at
	 * @param ml
	 *            the map layer the patch is part of
	 */
	private static void norm(final MapLayerPatchType mlp, final MapLayerType ml) {
		if (mlp.getOffset().isIsAbsolute()) {
			Assert.assertTrue(false);
//            normedOffsetX = (float)absOffsetX / (float)layerData.absoluteSizeX; 
//            normedOffsetY = (float)absOffsetY / (float)layerData.absoluteSizeY; 
//            isInAbsoluteMode = false;
		}
	}

	/**
	 * Returns the position of the upper left corner and width and size in normed coordinates.
	 * 
	 * @param mlp
	 *            the map layer patch to look at
	 * @param ml
	 *            the map layer the patch is part of
	 * @return an array of doubles containing the following information:
	 *   <table>
	 *     <caption>Array content</caption>
	 *     <tbody>
	 *       <tr><th>Index</th><th>Content</th></tr>
	 *       <tr><td>0</td><td>The x ordinate of the upper left corner of the patch.</td></tr>
	 *       <tr><td>1</td><td>The y ordinate of the upper left corner of the patch.</td></tr>
	 *       <tr><td>2</td><td>The width of the patch.</td></tr>
	 *       <tr><td>3</td><td>The height of the patch.</td></tr>
	 *     </tbody>
	 *   </table>
	 */
	private static double[] normedPatchCoordinates(final MapLayerPatchType mlp, final MapLayerType ml) {
		final double[] ret = { 0.0, 0.0, 0.0, 0.0 };
		final List<?> coord = mlp.getOffset().getCoordinate();
		final int length = coord.size();
		if (2 == length) {
			final Image img = (Image) patchToImg.get(mlp);
			ret[2] = (double) img.getWidth(null) / (double) ml.getAbsoluteSizeX();
			ret[3] = (double) img.getHeight(null) / (double) ml.getAbsoluteSizeY();
		} else if (4 == length) {
			ret[2] = DSA.toDoubleValue(coord.get(2));
			ret[3] = DSA.toDoubleValue(coord.get(3));
		} else {
			Assert.assertTrue(false);
		}
		ret[0] = DSA.toDoubleValue(coord.get(0));
		ret[1] = DSA.toDoubleValue(coord.get(1));
		return ret;
	}

	/**
	 * Norms all patches. Must be called after creation of the object, before
	 * using it.
	 * 
	 * @param ml
	 *            the map layer the patch is part of
	 */
	private static void norm(final MapLayerType ml) {
		final Iterator<MapLayerPatchType> it = ml.getMapLayerPatch().iterator();
		while (it.hasNext()) {
			final MapLayerPatchType mlp = (MapLayerPatchType) it.next();
			norm(mlp, ml);
		}
	}

	protected MapLayerPatchType selectedPatch = null;

	protected MapLayerType selectedLayer = null;

	/** By this step width (in normed coordinates) the position of a patch is modified. */
	public static final float MODIFY_STEP = 0.0001f;

	private boolean isPatchSelectionDisplayed = true;

	/** -1 indicates that no patch is selected. */
	private int patchSelectedIndex = 0;

	private int patchSelectedMaxIndex = 0;

	/** Turns marking of the selected patch on or off. */
	public void togglePatchSelectionDisplayed() {
		isPatchSelectionDisplayed = !isPatchSelectionDisplayed;
	}

	/** Selects the next patch. */
	public void selectNextPatch() {
		patchSelectedIndex++;
		if (patchSelectedMaxIndex <= patchSelectedIndex) {
			patchSelectedIndex = 0;
		}
	}

	/**
	 * Modifies the selected patch.
	 * 
	 * @param idx
	 *            the index of the coordinate to change
	 * @param change
	 *            the translation of the coordinate
	 * @return <code>true</code> if something was changed
	 */
	public boolean modifySelectedPatch(final int idx, final float change) {
		boolean ret = false;
		if (null != selectedPatch) {
			final List<String> coord = selectedPatch.getOffset().getCoordinate();
			float v = Float.parseFloat(coord.get(idx));
			v += change;
			coord.set(idx, Float.toString(v));
			ret = true;
		}
		return ret;
	}

	/************************************************************/
	/* Others                                                   */

	/** The status text to show if there is no status to show. */
	static final String EMPTY_STATUS_TEXT = "---"; 

	/** The basic stroke to return to when all my drawing is done, to prevent future drawers to draw with our stroke. */
	private static final BasicStroke bsDefault = new BasicStroke();

	/** The data model containing this aventuria map model. */
	private Model dataModel;

	/**
	 * @return the data model containing this aventuria map model
	 */
	public Model getModel() {
		return dataModel;
	}

	/** What's visible on the screen. */
	private BufferedImage display;

	public Image getDisplay() {
		return display;
	}

	private AventuriaMapLayerTreeNode topNode;
	private AventuriaMapLayerTreeNode gridNode5Percent;
	private AventuriaMapLayerTreeNode gridNode1Percent;
	private AventuriaMapLayerTreeNode crosshairNode;

	protected MapType map = null;

	private String statusText = EMPTY_STATUS_TEXT;

	public String getStatusText() {
		return statusText;
	}

	/** 
	 * This layer defines the ratio between x and y size of the whole map.
	 * For that we assume that a pixel of this layer is a cube - all sides have same length. 
	 */
	protected MapLayerType layer0;

	private void calcZoomXY(final double[] arr) {
		double zoom = dataModel.getOrCreateAventuriaMapModel().getZoom();
		arr[0] = zoom * (double) layer0.getAbsoluteSizeX();
		arr[1] = zoom * (double) layer0.getAbsoluteSizeY();
	}

	private Dimension size;

	public Dimension getSize() {
		return size;
	}

	public void setSize(final Dimension size) {
		if ((null == display) || (display.getWidth() != size.width) || (display.getHeight() != size.height)) {
			this.size = size;
			logger.debug("I'm still living before creation of BufferedImage.");
			display = new BufferedImage(size.width, size.height, BufferedImage.TYPE_INT_RGB);
			logger.debug("I'm still living after creation of BufferedImage.");
			Graphics2D graphics2D = display.createGraphics();
			logger.debug("I'm still living after createGraphics.");
			graphics2D.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
			logger.debug("I'm still living after setRenderingHint.");
			recalculate();
		}
	}

	// TODO: Switch to other image loading mechanism, like http://java.sun.com/j2se/1.4.2/docs/guide/imageio/spec/intro.fm5.html

	public AventuriaMapUIModel(final Model dataModel, final Dimension size) {
		this.dataModel = dataModel;
		this.size = size;

		final DSA dsa = DSA.instance();
		map = dsa.getMap();
		Assert.assertNotNull(map);
		final Iterator<MapLayerType> it = map.getMapLayer().iterator();
		while (it.hasNext()) {
			MapLayerType ml = (MapLayerType) it.next();
			norm(ml);
		}
		load(map);
		layer0 = (MapLayerType) map.getMapLayer().get(0);

		topNode = new AventuriaMapLayerTreeNode(
				this,
				AventuriaMapLayerTreeNode.ROOT_LEVEL, 
				AventuriaMapLayerTreeNode.Mode.PATCH_MODE, 
				null, null, null, "Layers", true);
		AventuriaMapLayerTreeNode layerNode = null;
		AventuriaMapLayerTreeNode patchNode = null;
		final Iterator<MapLayerType> mit = dsa.getMap().getMapLayer().iterator();
		boolean isFirstLayer = true;
		while (mit.hasNext()) {
			final MapLayerType ml = (MapLayerType) mit.next();
			final boolean isLayerDisplayed = isFirstLayer || ml.isDefaultIsDisplayed();
			layerNode = new AventuriaMapLayerTreeNode(
					this,
					AventuriaMapLayerTreeNode.LAYER_LEVEL, 
					AventuriaMapLayerTreeNode.Mode.PATCH_MODE, 
					ml, null, null, Map.calculateName(ml), isLayerDisplayed);
			topNode.add(layerNode);
			final Iterator<MapLayerPatchType> mlpIt = ml.getMapLayerPatch().iterator();
			while (mlpIt.hasNext()) {
				final MapLayerPatchType mlp = (MapLayerPatchType)mlpIt.next();
				final boolean isDisplayed = ((null == mlp.isDefaultIsDisplayed()) ? isLayerDisplayed : mlp.isDefaultIsDisplayed().booleanValue());
				patchNode = new AventuriaMapLayerTreeNode(
						this,
						AventuriaMapLayerTreeNode.PATCH_LEVEL, 
						AventuriaMapLayerTreeNode.Mode.PATCH_MODE, 
						ml, mlp, null, Map.calculateName(mlp), isDisplayed);
				layerNode.add(patchNode);
			}
			final Iterator<MapLayerAreaType> mlaIt = ml.getMapLayerArea().iterator();
			while (mlaIt.hasNext()) {
				final MapLayerAreaType mla = (MapLayerAreaType)mlaIt.next();
				final boolean isDisplayed = ((null == mla.isDefaultIsDisplayed()) ? isLayerDisplayed : mla.isDefaultIsDisplayed().booleanValue());
				patchNode = new AventuriaMapLayerTreeNode(
						this,
						AventuriaMapLayerTreeNode.PATCH_LEVEL, 
						AventuriaMapLayerTreeNode.Mode.AREA_MODE, 
						ml, null, mla, Map.calculateName(mla), isDisplayed);
				layerNode.add(patchNode);
			}
			isFirstLayer = false;
		}
		layerNode = new AventuriaMapLayerTreeNode(
				this,
				AventuriaMapLayerTreeNode.LAYER_LEVEL, 
				AventuriaMapLayerTreeNode.Mode.GRID_MODE, 
				null, null, null, "Grid", false);
		topNode.add(layerNode);
		gridNode5Percent = new AventuriaMapLayerTreeNode(
				this,
				AventuriaMapLayerTreeNode.PATCH_LEVEL, 
				AventuriaMapLayerTreeNode.Mode.GRID_MODE, 
				null, null, null, "5%", false);
		layerNode.add(gridNode5Percent);
		gridNode1Percent = new AventuriaMapLayerTreeNode(
				this,
				AventuriaMapLayerTreeNode.PATCH_LEVEL, 
				AventuriaMapLayerTreeNode.Mode.GRID_MODE, 
				null, null, null, "1%", false);
		layerNode.add(gridNode1Percent);

		crosshairNode = new AventuriaMapLayerTreeNode(
				this,
				AventuriaMapLayerTreeNode.LAYER_LEVEL, 
				AventuriaMapLayerTreeNode.Mode.CROSSHAIR_MODE, 
				null, null, null, "Crosshair", true);
		topNode.add(crosshairNode);

		setSize(size);
		logger.debug("I'm still living after constructor.");
	}

	/**
	 * Draws the map layer tree nodes to the graphics2D.
	 * 
	 * @param graphics2D
	 *            where to draw into
	 * @param x
	 *            the absolute coordinate of the upper left corner
	 * @param y
	 *            the absolute coordinate of the upper left corner
	 * @param width
	 *            the absolute width of the area to draw
	 * @param height
	 *            the absolute height of the area to draw
	 * @param zoom
	 *            an array of two doubles, containing the x- and y-zooming
	 *            factor between normed and absolute coordinates.
	 */
	public void draw(final Graphics2D graphics2D, final int x, final int y, final int width, final int height, final double[] zoom) {
		final double zoomX = zoom[0];
		final double zoomY = zoom[1];
		final Composite acDefault = graphics2D.getComposite();
		graphics2D.clearRect(0, 0, width, height);
		final Enumeration<?> en = topNode.depthFirstEnumeration();
		int patchIndex = 0;
		while (en.hasMoreElements()) {
			final AventuriaMapLayerTreeNode node = (AventuriaMapLayerTreeNode) en.nextElement();
			if (	(node.getMode() == AventuriaMapLayerTreeNode.Mode.PATCH_MODE)
				&& (node.getLevel() == AventuriaMapLayerTreeNode.PATCH_LEVEL))
			{
				final MapLayerType layer = node.getMapLayer();
				final MapLayerPatchType patch = node.getMapLayerPatch();
				final Image img = (Image) patchToImg.get(patch);
				Assert.assertNotNull(img);
				final double[] normedPC = normedPatchCoordinates(patch, layer);
				final int wulx = (int) Math.round(zoomX * normedPC[0]) - x;
				final int wuly = (int) Math.round(zoomY * normedPC[1]) - y;
				final int wsx = (int) Math.round(zoomX * normedPC[2]);
				final int wsy = (int) Math.round(zoomY * normedPC[3]);
				if (node.isSelected()) {
					graphics2D.drawImage(img, wulx, wuly, wsx, wsy, null);
				}
				if (layer.getID().equals("MAPLAYER_DETAILS_GERMAN")) {
					if (isPatchSelectionDisplayed && (patchIndex == patchSelectedIndex)) {
						graphics2D.setColor(Color.ORANGE);
						graphics2D.drawRect(wulx, wuly, wsx, wsy);
						graphics2D.drawRect(wulx - 2, wuly - 2, wsx + 4, wsy + 4);
						selectedPatch = patch;
						selectedLayer = layer;
					}
					patchIndex++;
				}
			} else if ((node.getMode() == AventuriaMapLayerTreeNode.Mode.AREA_MODE)
					&& (node.getLevel() == AventuriaMapLayerTreeNode.PATCH_LEVEL)
					&& (node.isSelected())) {
				final MapLayerAreaType patch = node.getMapLayerArea();
				final String colorName = DSA.toStringValue(patch.getColor());
                logger.trace(colorName);
				final Color c = findColor(colorName);
				final AlphaComposite acp = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f);
				//final float[] dash = new float[]{ 5.0f, 5.0f };
				//final BasicStroke bsp = new BasicStroke(5, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL, bsDefault.getMiterLimit(), dash, 0.0f);
				final BasicStroke bsp = new BasicStroke(5, BasicStroke.CAP_ROUND, 
						BasicStroke.JOIN_BEVEL, bsDefault.getMiterLimit());
				graphics2D.setColor(c);
				graphics2D.setStroke(bsp);
				graphics2D.setComposite(acp);
				Map.draw(patch, zoomX, zoomY, -x, -y, graphics2D);
				graphics2D.setStroke(bsDefault);
				graphics2D.setComposite(acDefault);
			} else if ((node.getMode() == AventuriaMapLayerTreeNode.Mode.GRID_MODE)
					&& (node.isSelected())) {
				final double[] normedDistance = { 0.050, 0.010 };
				final float[][] dash = new float[][]{null, { 1.0f, 1.0f }};
				final Color[] gridColors = {Color.BLUE, Color.WHITE};
				final int distanceIndex = (node == gridNode1Percent ? 1 : 0);
				final BasicStroke bsp = new BasicStroke(1, BasicStroke.CAP_ROUND, 
						BasicStroke.JOIN_BEVEL, bsDefault.getMiterLimit(), 
						dash[distanceIndex], 0.0f);
				graphics2D.setColor(gridColors[distanceIndex]);
				graphics2D.setStroke(bsp);
				for (double pos = normedDistance[distanceIndex]; pos < 1.0; pos += normedDistance[distanceIndex]) {
//                    boolean isToDraw = true;
//                    if (0 < distanceIndex)
//                    {
//                        isToDraw = (0.001 < MathUtil.modulo(pos, normedDistance[0], 0.0));
//                    }
//                    if (isToDraw)
//                    {
					final int gridX = (int) Math.round(zoomX * pos) - x;
					graphics2D.drawLine(gridX, 0, gridX, height);
					final int gridY = (int) Math.round(zoomY * pos) - y;
					graphics2D.drawLine(0, gridY, width, gridY);
//                    }
				}
				graphics2D.setStroke(bsDefault);
			} else if ((node.getMode() == AventuriaMapLayerTreeNode.Mode.CROSSHAIR_MODE)
					&& (node.isSelected())) {
				final MathVector normedPos = dataModel.getCurrentPositionOnAventuriaMap();
				final int hotspotX = (int)Math.round(zoomX * normedPos.get(MathVector.X)) - x;
				final int hotspotY = (int)Math.round(zoomY * normedPos.get(MathVector.Y)) - y;
				drawCrosshair(graphics2D, hotspotX, hotspotY);
			}
		}
		patchSelectedMaxIndex = patchIndex;
	}

	/**
	 * Recalculates the whole image.
	 * Is called internally whenever the things to display or parameters of display (e.g., zoom) changed. 
	 */
	public void recalculate() {
		logger.trace("Beginning recalculate");
		// Calculate absolute position of upper left corner of the visible
		// window within the layer.
		final double[] zoom = new double[2];
		calcZoomXY(zoom);
		final double zoomX = zoom[0];
		final double zoomY = zoom[1];
		// TODO: layerUL is representing the hotspot, not the upper left corner!
		MathVector normedPos = dataModel.getCurrentPositionOnAventuriaMap();
		final int hotspotX = (int) Math.round(zoomX * normedPos.get(MathVector.X));
		final int hotspotY = (int) Math.round(zoomY * normedPos.get(MathVector.Y));
		// Define the "hot spot" within the visible window.
		int windowX = hotspotX - size.width / 2;
		int windowY = hotspotY - size.height / 2;
		windowX = MathUtil.bound(windowX, 0, (int) Math.round(zoomX) - size.width);
		windowY = MathUtil.bound(windowY, 0, (int) Math.round(zoomY) - size.height);
		// Clear the display.
		final Graphics2D graphics2D = (Graphics2D) display.getGraphics();
		draw(graphics2D, windowX, windowY, size.width, size.height, zoom);
		logger.trace("Survived traversing the tree");

		final Messages messages = DSAEnv.instance().messages();
		final TextDocumentBuilder doc = new TextDocumentBuilder(messages);
		doc.setTextDocumentType(TextDocumentType.HTML401_TRANSITIONAL);
		doc.setEmbedded(false);
		doc.setInventoryMaxDepth(0);
		doc.beginDocument();
		doc.beginParagraph(null);
		doc.add(messages.getText(new I18NStringWithFillIns("Map.Position", "%0",
				new String[] { Model.formatPosition(dataModel.getCurrentPositionOnAventuriaMap()), })));
		doc.add(doc.tf().newLine());
		doc.add(dataModel.allMapLayerAreasAtCurrentPositionOnAventuriaMap());
		doc.endParagraph();
		if (null != selectedPatch) {
			doc.beginParagraph(null);
			final double[] normedPC = normedPatchCoordinates(selectedPatch, selectedLayer);
			final double roundFactor = 10000.0;
			final long x = Math.round(normedPC[0] * roundFactor);
			final long y = Math.round(normedPC[1] * roundFactor);
			final long w = Math.round(normedPC[2] * roundFactor);
			final long h = Math.round(normedPC[3] * roundFactor);
			doc.add(selectedPatch.getFilename() + " (" + x + ", " + y + ", " + w + ", " + h + ")");
			if (null != mousePoint) {
				final double dmx = (mousePoint.getX() + hotspotX - windowX) / zoomX;
				final double dmy = (mousePoint.getY() + hotspotY - windowY) / zoomY;
				final long mx = Math.round(dmx * roundFactor);
				final long my = Math.round(dmy * roundFactor);
				doc.add(" - Mouse(" + mx + ", " + my + ")");
			}
//            final int textbarWidth = 400;
//            final int textbarULX = WINDOW_MAX_WIDTH - textbarWidth;
//            graphics2D.clearRect(textbarULX, 0, textbarWidth, 55);
//            graphics2D.setColor(Color.ORANGE);
//            graphics2D.drawString(statusText, textbarULX + 5, 50);
			doc.endParagraph();
		}
		doc.endDocument();
		statusText = doc.result();
		//repaint();
		logger.trace("Survived recalculate");
	}

	public AventuriaMapLayerTreeNode getSelectionTreeRootNode() {
		return topNode;
	}

	public boolean moveHotspot(final int x, final int y) {
		boolean ret = false;
		final double moveStep = MOVE_STEP / dataModel.getOrCreateAventuriaMapModel().getZoom();
		final MathVector normedPos = dataModel.getCurrentPositionOnAventuriaMap();
		double newX = normedPos.get(MathVector.X);
		double newY = normedPos.get(MathVector.Y);
		newX += moveStep * x;
		newY += moveStep * y;
		newX = Math.max(newX, 0.0);
		newX = Math.min(newX, 1.0);
		newY = Math.max(newY, 0.0);
		newY = Math.min(newY, 1.0);
		if ((newX != normedPos.get(MathVector.X)) || (newY != normedPos.get(MathVector.Y))) {
			normedPos.setAllFrom(newX, newY);
			ret = true;
		}
		return ret;
	}

	public void nodeChanged(final AventuriaMapLayerTreeNode node) {
		recalculate();
	}

	/**
	 * Each call switches the state of the grid to the next of three states:
	 * No grid displayed, wide grid displayed, dense grid displayed.
	 */
	public void switchGridState() {
		final int stateBitmap = (gridNode5Percent.isSelected() ? 2 : 0) 
								+ (gridNode1Percent.isSelected() ? 1 : 0);
		final int[] nextState = new int[] { 2, 2, 3, 0 };
		final int nextBitmap = nextState[stateBitmap];
		gridNode5Percent.setSelected((nextBitmap & 2) != 0);
		gridNode1Percent.setSelected((nextBitmap & 1) != 0);
	}

	private Point mousePoint = null;

	public void setMousePoint(final Point point) {
		mousePoint = point;
	}
}
