package de.ped.dsatool.gui;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.Hashtable;

import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JTree;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.ped.dsatool.logic.AventuriaMapModel;
import de.ped.dsatool.logic.DSAEnv;
import de.ped.dsatool.logic.Model;
import de.ped.tools.Assert;
import de.ped.tools.Messages;
import de.ped.tools.TextFormatter;
import de.ped.tools.gui.CheckTreeManager;
import de.ped.tools.log.Logger;

/**
 * Draws a scrollable and zoomable map.
 * 
 * @author Copyright (c) 2007 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AventuriaMapWindow extends JFrame implements WindowListener, ActionListener, KeyListener, ChangeListener, ComponentListener {
	private static final long serialVersionUID = 1L;

	private static Logger logger = DSATool.getLogger();

	private AventuriaMapUIModel map;
	private AventuriaMapCanvas mapCanvas;

	/** Text field for status messages. */
	private JEditorPane statusLabel;

	/** The arrow buttons to navigate on the map. */
	private class ArrowButtonInfo {
		/** Direction the arrow points to, as defined in @link{javax.swing.SwingConstants} */
		int direction;
		/** The position in a border layout where to place the button */
		public String position;
		/** How to move the x component of the cursor when the button is pressed */
		public int moveX;
		/** How to move the y component of the cursor when the button is pressed */
		public int moveY;

		public ArrowButtonInfo(
				final int direction, 
				final String text, 
				final String filename, 
				final String position, 
				final int moveX, 
				final int moveY) {
			this.direction = direction;
			this.position = position;
			this.moveX = moveX;
			this.moveY = moveY;
		}
	}

	private ArrowButtonInfo[] arrowButtonInfo = {
			new ArrowButtonInfo(SwingConstants.WEST,  "<-", "v-l.gif", BorderLayout.WEST, -1, 0),
			new ArrowButtonInfo(SwingConstants.EAST,  "->", "v-r.gif", BorderLayout.EAST, +1, 0),
			new ArrowButtonInfo(SwingConstants.NORTH, "^",  "v-t.gif", BorderLayout.NORTH, 0, -1),
			new ArrowButtonInfo(SwingConstants.SOUTH, "v",  "v-b.gif", BorderLayout.SOUTH, 0, +1),
	};

	private JButton[] arrowButton = new JButton[arrowButtonInfo.length];
	private JSlider zoomSlider;
	private JLabel zoomLevelLabel = new JLabel("600.00%");
	private JLabel zoomLabel = new JLabel("Zoom:");
	private Dimension zoomLevelLabelSize;
	private JButton zoomInButton = new JButton("+");
	private JButton zoomOutButton = new JButton("-");
	private JButton setSelectionToDefaultButton = new JButton("Reset");

	private JTree layerTree;
	CheckTreeManager checkTreeManager;

	public static final Dimension AVENTURIA_MAP_MODEL_SIZE = new Dimension(100, 100);

	public AventuriaMapWindow(final Model dataModel) {
		final Messages messages = DSAEnv.instance().messages();
		setTitle(messages.getApplicationTitle() + " " + messages.getText("Action.ViewMap"));
		map = new AventuriaMapUIModel(dataModel, AVENTURIA_MAP_MODEL_SIZE);
		mapCanvas = new AventuriaMapCanvas(this, map);
		Assert.assertNotNull(map);

		// Create the map and surrounding scroll arrows.
		// TODO: Why not use just scollbars instead?
		final Container cont = new Container();
		cont.setLayout(new BorderLayout());
//        final String path = DSA.instance().getPathToResources();
		final Dimension arrowButtonWESize = new Dimension(ArrowIcon.W + 4, 200);
		final Dimension arrowButtonNSSize = new Dimension(arrowButtonWESize.height, arrowButtonWESize.width);
		for (int i = 0; i < arrowButtonInfo.length; i++) {
//            String filename = path + "/icons/" + arrowButtonInfo[i].filename;
//            logger.debug(filename);
//            ImageIcon icon = new ImageIcon(filename, arrowButtonInfo[i].text);
			final ArrowIcon aicon = new ArrowIcon(arrowButtonInfo[i].direction);

			arrowButton[i] = new JButton(aicon); // , arrowButtonInfo[i].text);
			arrowButton[i].setPreferredSize((i < 2) ? arrowButtonWESize : arrowButtonNSSize);
			final JPanel buttonPanel = new JPanel(new GridBagLayout());
			buttonPanel.add(arrowButton[i], new GridBagConstraints());
			cont.add(buttonPanel, arrowButtonInfo[i].position);

			arrowButton[i].addActionListener(this);
			arrowButton[i].addKeyListener(mapCanvas);
		}
		cont.add(mapCanvas, BorderLayout.CENTER);

		// Create the other elements to control the look of the map.
		final JPanel control = new JPanel(new FlowLayout());

		// Zoom section
		zoomSlider = new JSlider(0, AventuriaMapModel.ZOOM_STEPS.length - 1, map.getZoomStepIndex());
		zoomSlider.setMinorTickSpacing(1);
		zoomSlider.setPaintTicks(true);
		zoomSlider.setSnapToTicks(true);
		zoomSlider.addChangeListener(this);
		final Hashtable<Integer, JLabel> labelTable = new Hashtable<Integer, JLabel>();
		final int[] indexes = AventuriaMapModel.ZOOM_STEPS_MAJOR_INDEXES;
		for (int i = 0; i < indexes.length; i++) {
			labelTable.put(new Integer(indexes[i]), new JLabel(AventuriaMapModel.calcZoomStepIndexString(indexes[i])));
		}
		setSelectionToDefaultButton.addActionListener(this);
		zoomSlider.setLabelTable(labelTable);
		zoomSlider.setPaintLabels(true);
		zoomOutButton.addActionListener(this);
		zoomOutButton.addKeyListener(mapCanvas);
		zoomInButton.addActionListener(this);
		zoomInButton.addKeyListener(mapCanvas);
		control.add(setSelectionToDefaultButton);
		control.add(zoomLabel);
		control.add(zoomOutButton);
		control.add(zoomSlider);
		control.add(zoomInButton);
		control.add(zoomLevelLabel);

		statusLabel = new JEditorPane(TextFormatter.CONTENT_TYPE_HTML, AventuriaMapUIModel.EMPTY_STATUS_TEXT);
		statusLabel.setSelectionStart(0);
		statusLabel.setSelectionEnd(0);
		statusLabel.setEditable(false);
		final JScrollPane statusScrollPane = new JScrollPane(statusLabel);
		statusScrollPane.setPreferredSize(new Dimension(800, 100));

		// Put all the parts together.
		setLayout(new BorderLayout());
		add(cont, BorderLayout.CENTER);
		add(control, BorderLayout.NORTH);
		add(statusScrollPane, BorderLayout.SOUTH);

		layerTree = new JTree(map.getSelectionTreeRootNode());
		checkTreeManager = new CheckTreeManager(layerTree, false); 
//        TreeModel treeModel = layerTree.getModel();
//        Object treeRoot = treeModel.getRoot();
//        logger.debug("Root = " + treeRoot);
//        int treeSiblings = treeModel.getChildCount(treeRoot);
//        for (int i = 0; i < treeSiblings; i++)
//        {
//            Object treeS1 = treeModel.getChild(treeRoot, i);
//            logger.debug("Sibling " + i + " = " + treeS1);
//        }

		final JScrollPane treeView = new JScrollPane(layerTree);
		add(treeView, BorderLayout.WEST);

		pack(); //Resize the window to the preferred size of its components.
//        logger.debug("Label size=" + zoomLevelLabel.getSize().width + "," + zoomLevelLabel.getSize().height);
//        logger.debug("Label pref size=" + zoomLevelLabel.getPreferredSize().width + "," + zoomLevelLabel.getPreferredSize().height);
		zoomLevelLabelSize = zoomLevelLabel.getSize();

		// setSize(WINDOW_MAX_WIDTH, WINDOW_MAX_HEIGHT);
		addWindowListener(this);
		cont.addKeyListener(mapCanvas);
		repaint();
		logger.debug("I'm still living after constructor of " + this.getClass().getName());
	}

	@Override
	public void repaint() {
		if (null != map) {
			if (null != statusLabel) {
				statusLabel.setText(map.getStatusText());
				statusLabel.setSelectionStart(0);
				statusLabel.setSelectionEnd(0);
				statusLabel.repaint();
			}
			zoomLevelLabel.setText(AventuriaMapModel.calcZoomStepIndexString(map.getZoomStepIndex()));
			zoomLevelLabel.setPreferredSize(zoomLevelLabelSize);
			zoomLevelLabel.repaint();
			final int idx = map.getZoomStepIndex();
			Assert.assertNotNull(zoomSlider);
			zoomSlider.setValue(idx);
			zoomSlider.repaint();
			zoomOutButton.setEnabled(0 < idx);
			zoomOutButton.repaint();
			zoomInButton.setEnabled(idx < AventuriaMapModel.ZOOM_STEPS.length - 1);
			zoomInButton.repaint();
			mapCanvas.repaint();
			final Dimension mapSize = mapCanvas.getSize();
			final Dimension arrowButtonWESize = new Dimension(ArrowIcon.W + 4, mapSize.height / 2);
			final Dimension arrowButtonNSSize = new Dimension(mapSize.width / 2, arrowButtonWESize.width);
			for (int i = 0; i < arrowButtonInfo.length; i++) {
				arrowButton[i].setPreferredSize((i < 2) ? arrowButtonWESize : arrowButtonNSSize);
			}
		}
	}

	@Override
	public void stateChanged(ChangeEvent e) {
		if (e.getSource() == zoomSlider) {
			if (!zoomSlider.getValueIsAdjusting()) {
				final int tick = (int) zoomSlider.getValue();
				map.setZoomStepIndex(tick);
				mapCanvas.recalculate();
			}
		}
	}

	@Override
	public void actionPerformed(final ActionEvent e) {
		logger.debug(e.paramString());
		boolean isChanged = false;
		final Object source = e.getSource();
		for (int i = 0; i < arrowButtonInfo.length; i++) {
			if (source == arrowButton[i]) {
				isChanged = map.moveHotspot(arrowButtonInfo[i].moveX, arrowButtonInfo[i].moveY);
				break;
			}
		}
		if (source == zoomOutButton) {
			map.decrementZoomStepIndex();
			isChanged = true;
		} else if (source == zoomInButton) {
			map.incrementZoomStepIndex();
			isChanged = true;
		} else if (source == setSelectionToDefaultButton) {
			map.getSelectionTreeRootNode().setSelectedToDefault();
			isChanged = true;
		}
		if (isChanged) {
			mapCanvas.recalculate();
		}
	}

	@Override
	public void keyPressed(final KeyEvent e) {
		// Nothing to do
	}

	@Override
	public void keyReleased(final KeyEvent e) {
		// Nothing to do
	}

	@Override
	public void keyTyped(final KeyEvent e) {
		// Nothing to do
	}

	@Override
	public void windowActivated(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowClosed(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowClosing(final WindowEvent e) {
		// Nothing to do
//        System.exit(0);
		// to get the paths that were checked
//        TreePath checkedPaths[] = checkTreeManager.getSelectionModel().getSelectionPaths(); 
//        for (int i = 0; i < checkedPaths.length; i++)
//        {
//            logger.debug("Selected: " + checkedPaths[i].toString());
//        }
	}

	@Override
	public void windowDeactivated(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowDeiconified(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowIconified(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowOpened(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void componentHidden(final ComponentEvent e) {
		// Nothing to do
	}

	@Override
	public void componentMoved(final ComponentEvent e) {
		// Nothing to do
	}

	@Override
	public void componentResized(final ComponentEvent e) {
		logger.debug("componentResized()");
		repaint();
	}

	@Override
	public void componentShown(final ComponentEvent e) {
		// Nothing to do
	}
}
