package de.ped.dsatool.gui;

import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFormattedTextField;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.FormattedTextFieldInputIntegerVerifier;
import de.ped.tools.gui.JInfoDialog;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.CalendarType;
import de.ped.dsatool.logic.DSA;
import de.ped.dsatool.logic.DSACalendar;
import de.ped.dsatool.logic.DSADateCalendarTwelve;
import de.ped.dsatool.logic.DSADateComplete;
import de.ped.dsatool.logic.DSAEnv;
import de.ped.dsatool.logic.DSAException;
import de.ped.dsatool.logic.Model;
import de.ped.dsatool.logic.Weather;

/**
 * A dialog asking the user for the start date (and weather)
 * from where to create a calender for the next few Aventurian days.
 * Allows to calculate the representation of the year in several Aventurian calendar schemes.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class CalendarDialog extends PedJDialog implements ActionListener, FocusListener, ItemListener {
	private static final long serialVersionUID = 1L;
	private static Logger logger = DSATool.getLogger();

	private Weather         weather;
	private DSADateComplete returnDate;

	private JButton buttonRandom;
	private JComboBox<String> choiceMonth;
	private JFormattedTextField textDay = new JFormattedTextField();
	private JTextField textYear  = new JTextField(5);
	private JComboBox<String> choiceCalendar;
	private ChoiceWeather choiceWeather;
	private JComboBox<String> choiceWind;
	private JComboBox<String> choiceTemperature;

	/**
	 * Constructor.
	 * 
	 * @param parent
	 *            from this frame the dialog was called
	 * @param date
	 *            this date will be edited
	 */
	public CalendarDialog(final ApplicationMainWindow parent, final DSADateComplete date) {
		super(parent, parent);
		returnDate = date;
		final Messages messages = getMessages();
		final String titleKey = "Action.CreateDate";
		setTitle(messages.getText(titleKey));
		weather = (Weather)date.getWeather().clone();

		choiceMonth = new JComboBox<String>(Model.getAllMonthStrings());
		choiceCalendar = new JComboBox<String>(DSACalendar.getAllStrings());
		choiceWeather = new ChoiceWeather();
		final DSA dsa = DSA.instance();
		final List<String> windList = dsa.getWeather().getWindStrengthDescription();
		choiceWind = new JComboBox<String>(DSA.toStringArrayWithCounter(windList));
		final List<String> tempList = dsa.getWeather().getTemperatureDescription();
		choiceTemperature = new JComboBox<String>(DSA.toStringArrayWithCounter(tempList));

		textDay.setInputVerifier(new FormattedTextFieldInputIntegerVerifier(this, 1, 30, 1));

		setPropertyElements(new PropertyElement[] {
				new PropertyElement("CalendarDialog.Date", null, new PropertyElement[] {
						new PropertyElement("CalendarDialog.Date.Day", textDay),
						new PropertyElement("CalendarDialog.Date.Month", choiceMonth),
						new PropertyElement("CalendarDialog.Date.Year", textYear),
						new PropertyElement("CalendarDialog.Date.Calendar", choiceCalendar),
				}),
				new PropertyElement("CalendarDialog.Weather", null, new PropertyElement[] {
						new PropertyElement("CalendarDialog.Weather.Weather", choiceWeather),
						new PropertyElement("CalendarDialog.Weather.Wind", choiceWind),
						new PropertyElement("CalendarDialog.Weather.Temperature", choiceTemperature),
				}),
		});

		setTitlePanel(messages.getText("CalendarDialog.CalendarDialog.Date.Heading"));
		// Eingabefeld:
		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);
		addAllPropertyElementsTo(mainPanel, c);

		getContentPane().add(new JScrollPane(mainPanel), BorderLayout.CENTER);
		// Button-Leiste
		addHelpButton(titleKey);
		buttonRandom = getMainWindow().createButton("CalendarDialog.Weather.Random");
		addButton(buttonRandom);
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		finishLayout();
		choiceWeather.saveSize();
		choiceWind.setPreferredSize(choiceWeather.getPreferredSize());
		choiceTemperature.setPreferredSize(choiceWeather.getPreferredSize());

		writeDateFields();
		choiceCalendar.addItemListener(this);
		choiceMonth.addItemListener(this);
		buttonRandom.addActionListener(this);
		textYear.addFocusListener(this);
		textDay.addActionListener(this);
		textYear.addActionListener(this);
		enterDialog();
	}

	/**
	 * Invoked when an item state change event occurs.
	 * @param e the event to handle
	 */
	@Override
	public void itemStateChanged(final ItemEvent e) {
		final Object source = e.getSource();
		if (source == choiceCalendar) {
			// We must handle this separately: If the user changes the calendar, the year must be recalculated.
			// But if the user changes the year, a logic check must occur (if the actual calendar has a year 0)
			final int idx = choiceCalendar.getSelectedIndex();
			final CalendarType cal = (CalendarType)DSA.findBOByIDnum(DSA.instance().getCalendar(), idx);
			weather.setCalendar(cal);
			writeDateFields();
			repaint();
		} else if (source == choiceMonth) {
			// Change of month may mean new season and therefore new weather
			if (readDateFields()) {
				writeDateFields();
				repaint();
			}
		}
	}

	@Override
	public void focusGained(FocusEvent e) {
		// Nothing to do
	}

	@Override
	public void focusLost(final FocusEvent e) {
		if (e.getSource() == textYear) {
			// If the user changes the year, we must act immediatly
			// to check if the year is valid (some calendars don't have a year
			// 0)
			if (readDateFields()) {
				writeDateFields();
				repaint();
			}
		}
	}

	/**
	 * Invoked when an action occurs.
	 * @param e the event to handle
	 */
	@Override
	public void actionPerformed(final ActionEvent e) {
		logger.debug(e.paramString());
		final Object source = e.getSource();
		if (source == buttonRandom) {
			if (readDateFields()) {
				weather.setWeatherRandom();
				writeDateFields();
				repaint();
			}
		} else if ((source == getButtonOk()) || (source == textDay) || (source == textYear)) {
			// Pressing return in a text field should be handled as pressing the OK button
			if (readDateFields()) {
				returnDate.setWeather(weather);
				onOk();
			}
		} else if (source == getButtonCancel()) {
			onCancel();
		}
	}

	/**
	 * Sets the internal data structure according to the actual settings of the
	 * dialog fields. If the data in the dialog fields makes no sense (e.g., a
	 * year 0 in a calender without a zero year), a message box is shown to the
	 * user.
	 * 
	 * @return if reading the date fields was possible without error
	 */
	private boolean readDateFields() {
		final int month = choiceMonth.getSelectedIndex() + 1;
		logger.debug("Month: " + Integer.toString(month)); 
		int day = ((Integer)(textDay.getValue())).intValue();
		day = MathUtil.bound(day, 1, (month == 13 ? 5 : 30));
		logger.debug("Day: " + textDay.getText() + " -> " + Integer.toString(day)); 
		final int year = Integer.parseInt(textYear.getText());
		logger.debug("Year: " + textYear.getText() + " -> " + Integer.toString(year)); 
		try {
			final DSADateCalendarTwelve cal = DSACalendar.getCalendars().get(weather.getCalendar());
			logger.debug("Current calendar is " + cal.getCalendar().getID());
			cal.setDSADate(weather.getDate()); // We want to share the same date object.
			cal.setDayMonthYear(day, month, year);
			logger.debug("Year (Field) " + year + " -> (Date) " + cal.getYear()); 
			weather.setWeather(choiceWeather.getSelectedIndex() + 1);
			weather.setWind(choiceWind.getSelectedIndex());
			weather.setTemperature(choiceTemperature.getSelectedIndex());
			writeDateFields();
		} catch (final DSAException e) {
			logger.info(e.toString());
			final Messages messages = DSAEnv.instance().messages();
			final PedJDialog t = new JInfoDialog(
					(Frame)getParent(), 
					messages.getText("CalendarDialog.Exception.Title"), 
					messages.getText("CalendarDialog.Exception.Year0"),
					JOptionPane.ERROR_MESSAGE);
			t.setVisible(true);
			return false;
		}
		return true;
	}

	/**
	 * Sets the actual settings of the dialog fields according to the internal data structure.
	 */
	private void writeDateFields() {
		final DSADateCalendarTwelve cal = DSACalendar.getCalendars().get(weather.getCalendar());
		cal.setDSADate(weather.getDate());
		setValue(textDay, new Integer(cal.getDay()), 2);
		int monthIndex = cal.getMonthAsNumber() - 1;
		if (monthIndex != choiceMonth.getSelectedIndex()) {
			choiceMonth.setSelectedIndex(monthIndex);
		}
		textYear.setText(Integer.toString(cal.getYear()));
		logger.debug("Jahr (Datum) " + cal.getYear() + " -> (Feld) " + textYear.getText());
		final int calendarIndex = weather.getCalendar().getIDnum();
		if (calendarIndex != choiceCalendar.getSelectedIndex()) {
			choiceCalendar.setSelectedIndex(calendarIndex);
		}
		// For the weather the field list changes according to the season
		choiceWeather.adapt(weather);
		choiceWeather.setSelectedIndex(weather.getWeather() - 1);
		choiceWind.setSelectedIndex(weather.getWind());
		choiceTemperature.setSelectedIndex(weather.getTemperature());
	}
}
