package de.ped.dsatool.gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;

import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JEditorPane;
import javax.swing.JFileChooser;

import java.io.File;
import java.io.FileWriter;
import java.io.BufferedWriter;
import java.io.IOException;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.Assert;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextDocumentType;
import de.ped.tools.TextFormatter;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.ConfigurableFileFilter;
import de.ped.tools.gui.GUIAlgorithms;
import de.ped.tools.gui.HtmlPaneDialog;
import de.ped.tools.gui.PedAction;
import de.ped.tools.log.Logger;
import de.ped.dsatool.logic.DSAEnv;
import de.ped.dsatool.logic.MasterDataModel;
import de.ped.dsatool.logic.Model;
import de.ped.dsatool.logic.NamesModel;
import de.ped.dsatool.logic.PlantsModel;
import de.ped.dsatool.logic.MasterDataModel.MasterDataType;
import de.ped.dsatool.logic.Model.ModelMode;

/**
 * The main window of the DSA tool
 * 
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSAToolMainWindow extends ApplicationMainWindow {
	private static final long serialVersionUID = 1L;

	private static Logger logger = DSATool.getLogger();

	private static final int WIDTH = 800;
	private static final int HEIGHT = 600;

	/**
	 * Shows one of the model parts like quest, plants, town, names.
	 * 
	 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private class ViewModelPartAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewModelPartAction(final ModelMode mode, final String messageKey) {
			super(messageKey, DSAToolMainWindow.this, null, false);
			this.mode = mode;
		}

		final ModelMode mode;
	}

	/**
	 * Shows one of the master data tables.
	 * 
	 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private class ViewMasterDataAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewMasterDataAction(final MasterDataType mode) {
			super(null, DSAToolMainWindow.this, null, false);
			this.masterDataMode = mode;
		}

		final MasterDataType masterDataMode;
	}

	private Action exportTextAction     = new PedAction("Action.ExportText",     this, null, true); 
	private Action createDateAction     = new PedAction("Action.CreateDate",     this, null, true); 
	private Action createDateUndoAction = new PedAction("Action.CreateDateUndo", this, null, false); 
	private Action createDateNextAction = new PedAction("Action.CreateDateNext", this, null, false); 
	private Action createTownAction     = new PedAction("Action.CreateTown",     this, null, true); 
	private Action createMapAction      = new PedAction("Action.CreateMap",      this, null, false); 
	private Action createNamesAction    = new PedAction("Action.CreateNames",    this, null, true); 
	private Action createPlantsAction   = new PedAction("Action.CreatePlants",   this, null, true); 
	private Action createQuestAction    = new PedAction("Action.CreateQuest",    this, null, false); 
	private Action createDiceAction     = new PedAction("Action.CreateDice",     this, null, true);
	private Action viewMapAction        = new PedAction("Action.ViewMap",        this, null, true); 
	private Action viewDateAction   = new ViewModelPartAction(Model.ModelMode.DATE,  "Action.ViewDate"); 
	private Action viewTownAction   = new ViewModelPartAction(Model.ModelMode.TOWN,  "Action.ViewTown");
	private Action viewNamesAction  = new ViewModelPartAction(Model.ModelMode.NAMES, "Action.ViewNames");
	private Action viewPlantsAction = new ViewModelPartAction(Model.ModelMode.PLANT, "Action.ViewPlants");
	private Action viewQuestAction  = new ViewModelPartAction(Model.ModelMode.QUEST, "Action.ViewQuest");
	private Action viewDiceAction   = new ViewModelPartAction(Model.ModelMode.DICE,  "Action.ViewDice");
	private Action viewMasterDataAction[] = new ViewMasterDataAction[MasterDataModel.MasterDataType.values().length];
	private JRadioButtonMenuItem miViewDate;
	private JRadioButtonMenuItem miViewTown;
	private JRadioButtonMenuItem miViewNames;
	private JRadioButtonMenuItem miViewPlants;
	private JRadioButtonMenuItem miViewQuest;
	private JRadioButtonMenuItem miViewDice;
	private JRadioButtonMenuItem miMasterDatas[] = new JRadioButtonMenuItem[MasterDataModel.MasterDataType.values().length];

	@Override
	public void pedAbstractActionPerformed(final PedAction action, final ActionEvent event) {
		super.pedAbstractActionPerformed(action, event);
		if (action == exportTextAction) {
			exportText();
		} else if (action == createDateAction) {
			createDate();
		} else if (action == createDateUndoAction) {
			createDateUndo();
		} else if (action == createDateNextAction) {
			createDateNext();
		} else if (action == createTownAction) {
			createTown();
		} else if (action == createMapAction) {
			createMap();
		} else if (action == createNamesAction) {
			createNames();
		} else if (action == createPlantsAction) {
			createPlants();
		} else if (action == createQuestAction) {
			createQuest();
		} else if (action == createDiceAction) {
			createDice();
		} else if (action == viewMapAction) {
			viewMap();
		} else if (action instanceof ViewModelPartAction) {
			dataModel.setMode(((ViewModelPartAction) action).mode);
			actualizeView();
		} else if (action instanceof ViewMasterDataAction) {
			dataModel.setMode(Model.ModelMode.MASTER_DATA);
			dataModel.getMasterDataModel().setMode(((ViewMasterDataAction) action).masterDataMode);
			actualizeView();
		}
	}

	/** The text may be too large to display at once, so we need a scroll pane around it. */
	public JScrollPane textScrollPane;

	/** The big text area displaying the logical model. */
	private JEditorPane textBuf;

	private Frame mapFrame;

	public Frame getOrCreateMapFrame() {
		if (null == mapFrame) {
			mapFrame = new AventuriaMapWindow(dataModel);
		}
		return mapFrame;
	}

	/**
	 * Constructor.
	 * 
	 * @param appEnv
	 *            the application environment
	 */
	public DSAToolMainWindow(final ApplicationEnvironment appEnv) {
		super(appEnv);

		// Data structures
		dataModel = new Model(true);

		final ButtonGroup modeButtonGroup = new ButtonGroup();

		try {
			setBounds(0, 0, WIDTH, HEIGHT + 80);
		} catch (final NoSuchMethodError e) {
			logger.info(e + "Ah, using an old browser, eh? Ok, I'm ignoring the bounds of the root frame...");
		}

		for (int i = 0; i < MasterDataType.values().length; i++) {
			viewMasterDataAction[i] = new ViewMasterDataAction(MasterDataType.values()[i]);
			putValues(viewMasterDataAction[i], MasterDataModel.getMessageKey(i));
		}

		// Menu bar
		final JMenuBar menuBar = createJMenuBar();
		setJMenuBar(menuBar);

		final JMenu menuEntryFile = createJMenu("Menu.File");
		menuBar.add(menuEntryFile);
		createActionItem(menuEntryFile, exportTextAction);
		exportTextAction.setEnabled(appEnv.isExitable());
		menuEntryFile.addSeparator();
		createActionItem(menuEntryFile, viewMapAction);
		menuEntryFile.addSeparator();
		menuEntryFile.add(createLAFMenu());
		menuEntryFile.addSeparator();
		createActionItem(menuEntryFile, quitAction);

		final JMenu menuEntryCalendar = createJMenu("Menu.Calendar");
		menuBar.add(menuEntryCalendar);
		final JMenu menuEntryTown = createJMenu("Menu.Town");
		menuBar.add(menuEntryTown);
		final JMenu menuEntryNames = createJMenu("Menu.Names");
		menuBar.add(menuEntryNames);
		final JMenu menuEntryPlants = createJMenu("Menu.Plants");
		menuBar.add(menuEntryPlants);
		final JMenu menuEntryQuest = createJMenu("Menu.Quest");
		menuBar.add(menuEntryQuest);
		final JMenu menuEntryDice = createJMenu("Menu.Dice");
		menuBar.add(menuEntryDice);
		final JMenu menuEntryMasterData = createJMenu("Menu.MasterData");
		menuBar.add(menuEntryMasterData);

		miViewDate = createRadioButtonActionItem(menuEntryCalendar, viewDateAction, modeButtonGroup);
		createActionItem(menuEntryCalendar, createDateAction);
		createActionItem(menuEntryCalendar, createDateUndoAction);
		createActionItem(menuEntryCalendar, createDateNextAction);
		
		miViewTown = createRadioButtonActionItem(menuEntryTown, viewTownAction, modeButtonGroup);
		createActionItem(menuEntryTown, createTownAction);
		createActionItem(menuEntryTown, createMapAction);

		miViewNames = createRadioButtonActionItem(menuEntryNames, viewNamesAction, modeButtonGroup);
		createActionItem(menuEntryNames, createNamesAction);

		miViewPlants = createRadioButtonActionItem(menuEntryPlants, viewPlantsAction, modeButtonGroup);
		createActionItem(menuEntryPlants, createPlantsAction);

		miViewQuest = createRadioButtonActionItem(menuEntryQuest, viewQuestAction, modeButtonGroup);
		createActionItem(menuEntryQuest, createQuestAction);

		miViewDice = createRadioButtonActionItem(menuEntryDice, viewDiceAction, modeButtonGroup);
		createActionItem(menuEntryDice, createDiceAction);

		for (int i = 0; i < MasterDataModel.MasterDataType.values().length; i++) {
			miMasterDatas[i] = createRadioButtonMenuItem(modeButtonGroup, viewMasterDataAction[i]);
			menuEntryMasterData.add(miMasterDatas[i]);
		}

		final JMenu menuEntryHelp = createJMenu("Menu.Help");
		createActionItem(menuEntryHelp, helpAction);
		createActionItem(menuEntryHelp, aboutAction);
		menuBar.add(menuEntryHelp);
		try {
			menuBar.setHelpMenu(menuEntryHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}

		// Text buffer
		textBuf = new JEditorPane();
		textScrollPane = new JScrollPane(textBuf);
		final Dimension d = new Dimension(WIDTH, HEIGHT);
		textScrollPane.setPreferredSize(d);
		getContentPane().add(textScrollPane);
		setFormat(dataModel.getFormatStyle());

		switchLocale(getMessages().getCurrentLocale());

		GUIAlgorithms.center(this, getParent());
		textBuf.addKeyListener(this);

		actualizeRadioButtons();
		actualizeView();
	}
	
	/**
	 * Initializes the correct selection of the radio buttons.
	 * Called only once during initialization.
	 */
	private void actualizeRadioButtons() {
		logger.debug("actualizeRadioButtons called."); 
		switch(dataModel.getMode()) {
		case DATE:
			miViewDate.setSelected(true);
			break;
		case TOWN:
			miViewTown.setSelected(true);
			break;
		case NAMES:
			miViewNames.setSelected(true);
			break;
		case PLANT:
			miViewPlants.setSelected(true);
			break;
		case QUEST:
			miViewQuest.setSelected(true);
			break;
		case DICE:
			miViewDice.setSelected(true);
		case MASTER_DATA:
//                miViewMasterData.setSelected(true);
			int masterDataMode = dataModel.getMasterDataModel().getMode().ordinal();
			miMasterDatas[masterDataMode].setSelected(true);
			break;
		default:
			Assert.assertTrue(false);
		}
	}

	private void setFormat(final TextDocumentType aFormat)	{
		switch(aFormat)	{
		case FIXEDFONT:
			dataModel.setFormatStyle(aFormat);
			textBuf.setContentType(TextFormatter.CONTENT_TYPE_PLAINTEXT);
			textBuf.setFont(new Font(Font.MONOSPACED, Font.PLAIN, 12));
			break;
		case HTML401_TRANSITIONAL:
		case XHTML10_STRICT:
		case HTML5:
			dataModel.setFormatStyle(aFormat);
			textBuf.setContentType(TextFormatter.CONTENT_TYPE_HTML); 
			break;
		default:
			Assert.assertTrue(false);
		}
	}

	private void actualizeView() {
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			final String text = dataModel.write(DSAEnv.instance().messages().getApplicationTitleLong());
			textBuf.setText(text);
			// Set viewport of the ScrollPane to the beginning of the text.
			textBuf.setSelectionStart(0);
			textBuf.setSelectionEnd(0);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
	}

	public void exportText() {
		final JFileChooser chooser = new JFileChooser();
		final ConfigurableFileFilter filter = new ConfigurableFileFilter();
		final TextDocumentType formatStyle = dataModel.getFormatStyle();
		filter.setAllowedExtensions(formatStyle.getTypicalFileExtensions());
		filter.setDescription(formatStyle.getTypicalFileFilterDescription());
		chooser.setFileFilter(filter);
		int returnVal = chooser.showSaveDialog(this);
		if (JFileChooser.APPROVE_OPTION == returnVal) {
			final File file = filter.correctExtensionOfSelectedFile(chooser.getSelectedFile());
			try {
				logger.info("You chose to write to this file: " + file.getName());
				final FileWriter fw = new FileWriter(file);
				final BufferedWriter bw = new BufferedWriter(fw);
				bw.write(TextDocumentType.HTML401_TRANSITIONAL.doctypeTag);
				bw.write(TextFormatter.NL);
				bw.write(textBuf.getText());
				bw.close();
				fw.close();
			} catch (final IOException e) {
				logger.error("Error exporting file " + file.getName() + ": " + e);
			}
		}
	}

	public void createDate() {
		CalendarDialog cd = new CalendarDialog(this, dataModel.getFirstDay());
		if (cd.wasAcceptedDialogChanges()) {
			dataModel.setFirstCalendarDay(dataModel.getFirstDay());
		}
		actualizeRadioButtons();
		actualizeView();
	}

	public void createDateUndo() {
		dataModel.recalcCalendarTimeframe();
		actualizeRadioButtons();
		actualizeView();
	}

	public void createDateNext() {
		dataModel.calcNextCalendarTimeframe();
		actualizeRadioButtons();
		actualizeView();
	}

	public void createTown() {
		// The dialog gets a reference to a member of the model and changes it
		// eventually
		final TownDialog td = new TownDialog(this, dataModel.getTown());
		if (td.wasAcceptedDialogChanges()) {
			// Tell the model that it has changed
			dataModel.setTown(td.getTown());
		}
		actualizeRadioButtons();
		actualizeView();
	}

	public void createMap() {
		dataModel.changeTownMap();
		actualizeRadioButtons();
		actualizeView();
	}

	public void createNames() {
		final NamesModel namesModel = dataModel.getNamesModel();
		final NamesDialog nd = new NamesDialog(this, namesModel);
		nd.enterDialog();
		if (nd.wasAcceptedDialogChanges()) {
			dataModel.setNamesModel(namesModel);
		}
		actualizeRadioButtons();
		actualizeView();
	}

	public void createPlants() {
		final PlantsModel plantsModel = dataModel.getPlantsModel();
		final PlantsDialog nd = new PlantsDialog(this, plantsModel);
		nd.enterDialog();
		if (nd.wasAcceptedDialogChanges()) {
			dataModel.setPlantsModel(plantsModel);
		}
		actualizeRadioButtons();
		actualizeView();
	}

	public void createQuest() {
		dataModel.changeQuest();
		actualizeRadioButtons();
		actualizeView();
	}

	public void createDice() {
		final DiceDialog td = new DiceDialog(this, dataModel.getDice().getDiceRollType());
		td.enterDialog();
		if (td.wasAcceptedDialogChanges()) {
			// Tell the model that it has changed
			dataModel.rollDice(td.getDiceRollType());
		}
		actualizeRadioButtons();
		actualizeView();
	}

	public void viewMap() {
		mapFrame = getOrCreateMapFrame();
		mapFrame.setVisible(true);
		actualizeRadioButtons();
		actualizeView();
	}

	@Override
	public void showAboutBox() {
		final Messages messages = DSAEnv.instance().messages();
		final TextDocumentBuilder doc = new TextDocumentBuilder(messages);
		doc.setTextDocumentType(TextDocumentType.HTML401_TRANSITIONAL);
		doc.setEmbedded(false);
		doc.setInventoryMaxDepth(0);
		doc.setTitle(messages.getApplicationTitle());
		doc.beginDocument();
		doc.add(messages.getLongDescription("Copyright"));
        doc.addContentPart(2, "Help.VersionHistory");
        doc.addContentPart(2, "Help.KnownProblems");
		doc.endDocument();
		HtmlPaneDialog d = new HtmlPaneDialog(this, doc.getTitle(), doc.result());
		d.enterDialog();
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		final DSAToolUIHelpTextBuilder builder = new DSAToolUIHelpTextBuilder(this);
		builder.buildHelpText(doc);
	}

	@Override
	public void setLookAndFeel(final String className) {
		super.setLookAndFeel(className);
		final Dimension d = textScrollPane.getSize();
		textScrollPane.setSize(d);
		actualizeView();
	}

	/** 
	 * Handle the key typed event from the text field.
	 * @param e the event to handle
	 */
	@Override
	public void keyTyped(final KeyEvent e) {
		logger.debug("KeyTyped: keyChar=" + e.getKeyChar());
		if (e.getKeyChar() == 1) { //Ctrl-A
			textBuf.selectAll();
			textBuf.repaint();
		}
	}

	/** The logical model of the document to display in the text area. */
	private Model dataModel;
}
