package de.ped.dsatool.gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Iterator;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JPanel;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.TownType;
import de.ped.dsatool.logic.DSA;
import de.ped.dsatool.logic.Model;
import de.ped.dsatool.logic.NamesModel;
import de.ped.dsatool.logic.Town;

/**
 * The dialog that asks the user what Aventurian names he/she is interested in.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class NamesDialog extends PedJDialog implements ActionListener, ItemListener {
	private static final long serialVersionUID = 1L;
	
	private static final String LABEL_PREFIX = "        ";

	/**
	 * Constructor
	 * @param parent the parent window
	 * @param namesModel the model to use for the dialog
	 */
	public NamesDialog(final ApplicationMainWindow parent, final NamesModel namesModel) {
		super(parent, parent);
		model = namesModel;
		final Messages messages = getMessages();
		setTitle(messages.getText("Action.CreateNames"));
		setTitlePanel(messages.getText("NamesDialog.Title")); 
		// Eingabefeld:
		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);
		final ButtonGroup modeButtonGroup = new ButtonGroup();
		choiceLanguage = new ChoiceLanguage();
		choiceRegion = new ChoiceRegion();
		choiceTownPredefined = new JComboBox<String>();
		final Iterator<TownType> it = DSA.instance().getTown().iterator();
		while (it.hasNext()) {
			final TownType town = (TownType)it.next();
			choiceTownPredefined.addItem(town.getName());
		}
		buttonMapPosition = getMainWindow().createButton("NamesDialog.MapPosition");
		buttonMapPosition.addActionListener(this);
		
		modeElems = new ModeElement[NamesModel.Mode.values().length];
		for (int i = 0; i < modeElems.length; i++) {
			modeElems[i] = new ModeElement();
		}
		setOptionBlock(mainPanel, gridbag, c, modeButtonGroup, NamesModel.Mode.LANGUAGE, choiceLanguage, "NamesDialog.For.Language", "NamesDialog.Language", null); 
		setOptionBlock(mainPanel, gridbag, c, modeButtonGroup, NamesModel.Mode.REGION, choiceRegion, "NamesDialog.For.Region", "NamesDialog.Region", null); 
		setOptionBlock(mainPanel, gridbag, c, modeButtonGroup, NamesModel.Mode.TOWN, choiceTownPredefined, "NamesDialog.For.Town", "NamesDialog.Town", null); 
		setOptionBlock(mainPanel, gridbag, c, modeButtonGroup, NamesModel.Mode.MAP_POSITION, buttonMapPosition, "NamesDialog.For.MapPosition", null, Model.formatPosition(0.0, 0.0)); 
		getContentPane().add(new JScrollPane(mainPanel), BorderLayout.CENTER);
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		addHelpButton("Action.CreateNames");
		finishLayout();
		updateUIFieldsBasedOnProperties();
		updateView();
	}

	private void setOptionBlock(
			final JPanel mainPanel, 
			final GridBagLayout gridbag, 
			final GridBagConstraints c,
			final ButtonGroup modeButtonGroup, 
			final NamesModel.Mode mode,
			final Component component, 
			final String selectionKey, 
			final String labelKey,
			final String labelAlternative) {
		final ModeElement el = modeElems[mode.ordinal()];
		el.radioButton = getMainWindow().createRadioButton(new I18NStringWithFillIns(selectionKey, null, null));
		el.radioButton.addItemListener(this);
		modeButtonGroup.add(el.radioButton);
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		gridbag.setConstraints(el.radioButton, c);
		mainPanel.add(el.radioButton);

		final JLabel spacer = new JLabel(LABEL_PREFIX);
		spacer.setSize(20, 10);
		c.gridwidth = 1;
		gridbag.setConstraints(spacer, c);
		mainPanel.add(spacer);
		if (null != labelKey) {
			el.label = getMainWindow().createLabel(labelKey);
		} else {
			el.label = new JLabel(labelAlternative);
		}
		c.gridwidth = 1;
		gridbag.setConstraints(el.label, c);
		mainPanel.add(el.label);

		el.component = component;
		final JPanel p = new JPanel();
		p.setLayout(new FlowLayout(FlowLayout.LEFT));
		p.add(el.component);
		c.gridwidth = GridBagConstraints.REMAINDER; //end row
		gridbag.setConstraints(p, c);
		mainPanel.add(p);
		el.label.setLabelFor(el.component);
	}

	/** Updates the view, especially disables or enables fields after a radio button was pressed. */
	private void updateView() {
		final NamesModel.Mode mode = readModeFromRadioButtons();
		for (int i = 0; i < modeElems.length; i++) {
			final boolean isActive = NamesModel.Mode.values()[i] == mode;
			modeElems[i].label.setEnabled(isActive);
			modeElems[i].component.setEnabled(isActive);
		}
	}

	/** 
	 * Invoked when an action occurs.
	 * @param e the event to handle
	 */
	@Override
	public void actionPerformed(final ActionEvent e) {
		logger.debug(e.paramString());
		final Object source = e.getSource();
		if (source == getButtonOk()) {
			// Pressing return in a text field should be handled as pressing the OK button
			readUIFieldsIntoProperties();
			onOk();
		} else if (source == getButtonCancel()) {
			onCancel();
		} else if (source == buttonMapPosition) {
			logger.debug("Übernehme Daten von Map");
			model.setAventuriaMapPositionFromDataModel();
			modeElems[NamesModel.Mode.MAP_POSITION.ordinal()].label
					.setText(Model.formatPosition(model.getAventuriaMapPosition()));
		}
	}

	/** 
	 * Handles the item state changes of the checkable menu items.
	 * @param e the event to handle
	 */
	@Override
	public void itemStateChanged(final ItemEvent e) {
		for (int i = 0; i < modeElems.length; i++) {
			if (e.getSource() == modeElems[i].radioButton) {
				updateView();
				break;
			}
		}
	}

	@Override
	public void readUIFieldsIntoProperties() {
		model.setRegion(choiceRegion.getSelectedRegion());
		model.setLanguage(choiceLanguage.getSelectedLanguage());
		final TownType predefinedTown = Town.findPredefinedTown((String)choiceTownPredefined.getSelectedItem());
		final TownType town = model.getTown();
		if ((null != predefinedTown) && (!predefinedTown.equals(town))) {
			model.setTown(predefinedTown);
		}
		model.setMode(readModeFromRadioButtons());
		updateUIFieldsBasedOnProperties();
	}

	protected NamesModel.Mode readModeFromRadioButtons() {
		NamesModel.Mode mode = NamesModel.Mode.LANGUAGE;
		for (int i = 0; i < modeElems.length; i++) {
			if (modeElems[i].radioButton.isSelected()) {
				mode = NamesModel.Mode.values()[i];
				break;
			}
		}
		return mode;
	}

	@Override
	public void updateUIFieldsBasedOnProperties() {
		modeElems[model.getMode().ordinal()].radioButton.setSelected(true);
		choiceLanguage.setSelectedLanguage(model.getLanguage());
		choiceRegion.setSelectedRegion(model.getRegion());
		final TownType town = model.getTown();
		if (null != town) {
			choiceTownPredefined.setSelectedItem(model.getTown().getName());
		}
		modeElems[NamesModel.Mode.MAP_POSITION.ordinal()].label.setText(model.getAventuriaMapPositionString(false));
	}

	public NamesModel getModel() {
		return model;
	}

	private NamesModel model;

	private class ModeElement {
		public JRadioButton radioButton;
		public Component component;
		public JLabel label;
	}

	private ModeElement[] modeElems = new ModeElement[NamesModel.Mode.values().length];

	private ChoiceRegion choiceRegion;
	private JButton buttonMapPosition;
	private JComboBox<String> choiceTownPredefined;
	private ChoiceLanguage  choiceLanguage;

	private final Logger logger = DSATool.getLogger();
}
