package de.ped.dsatool.logic;

/**
 * Knows zooming levels for the Aventurian map.
 * 
 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AventuriaMapModel {

	/************************************************************/
	/* Zooming model                                            */
	
	/** The zooming levels the user can select. */
	public static final double [] ZOOM_STEPS =  {
			0.200, 0.220, 0.250, 0.300, 0.400, 0.500, 0.667, 0.750, 0.875, 1.000, 
			1.125, 1.250, 1.375, 1.500, 1.750, 2.000, 
			2.250, 2.500, 2.750, 3.000, 
			3.250, 3.500, 3.750, 4.000, 
			4.250, 4.500, 4.750, 5.000,
			5.500, 6.000,
	};
	/** The zooming levels highlighted when the list of zooming levels are presented to the user. */
	public static final int [] ZOOM_STEPS_MAJOR_INDEXES = {
			0, 9, 19, 29,
	};

	/**
	 * Returns the textual representation for a zoom step index.
	 * 
	 * @param zoomStepIndex
	 *            the index of {@link ZOOM_STEPS} to show
	 * @return the textual representation for a zoom step index
	 */
	public static String calcZoomStepIndexString(final int zoomStepIndex) {
		final double zoomFactor = ZOOM_STEPS[zoomStepIndex];
		final long zoom = Math.round(zoomFactor * 100);
		// return TextFormatter.format(zoom, 6, 1).toString() + "%";
		return Long.toString(zoom) + "%";
	}

	/**
	 * @return a textual representation of all zoom step indexes, e.g., for a selection box
	 */
	public static String[] getZoomStepIndexStrings() {
		if (null == zoomStepIndexStrings) {
			// First initialize completely, then set static variable, to avoid multithreading problems.
			final String[] zsis = new String[ZOOM_STEPS.length];
			for (int i = 0; i < ZOOM_STEPS.length; i++) {
				zsis[i] = calcZoomStepIndexString(i);
			}
			zoomStepIndexStrings = zsis;
		}
		return zoomStepIndexStrings;
	}

	private static String[] zoomStepIndexStrings;

	/**
	 * @return the current zoom step index
	 */
	public int getZoomStepIndex() {
		return zoomStepIndex;
	}

	/**
	 * Sets the current zoom step index. Clips illegal values to the border of
	 * the allowed interval.
	 * 
	 * @param value
	 *            the new zoom step index
	 */
	public void setZoomStepIndex(int value) {
		value = Math.max(value, 0);
		value = Math.min(value, AventuriaMapModel.ZOOM_STEPS.length - 1);
		zoomStepIndex = value;
	}

	public void incrementZoomStepIndex() {
		setZoomStepIndex(zoomStepIndex + 1);
	}

	public void decrementZoomStepIndex() {
		setZoomStepIndex(zoomStepIndex - 1);
	}

	/** 
	 * @return the current zoom level
	 */
	public double getZoom() {
		return AventuriaMapModel.ZOOM_STEPS[zoomStepIndex];
	}

	/** 
	 * Zoom step index defining how detailed the display shows the map.
	 * @see AventuriaMapModel#ZOOM_STEPS
	 */
	private int zoomStepIndex = 9;
}
