package de.ped.dsatool.logic;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.List;
import java.util.Iterator;

import javax.xml.bind.JAXBElement;
import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlEnum;

import de.ped.dsatool.dsa.generated.*;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.HtmlHelper;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Logger;

/**
 * Root class for the whole DSA data.
 * Implements the Singleton pattern from the Gang of Four.
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSA extends DsaType {
	/**
	 * Implements the Singleton pattern from the Gang of Four.
	 * 
	 * @return the singleton instance
	 */
	public static DSA instance() {
		return DSAEnv.instance().dsa();
	}

	public static final String DATA_XML_FILENAME = "data.xml";

	public String getContentText(final BusinessObjectType o, final TextFormatter tf) {
		return o.getName();
	}

	public String getHeadingText(final BuildingType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Building");
	}

	public String getHeadingText(final DayOfWeekType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.DayOfWeek");
	}

	public String getHeadingText(final DeityType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Deity");
	}

	public String getHeadingText(final DiceRollType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.DiceRoll");
	}

	public String getContentText(final DiceRollType o, final TextFormatter tf) {
		final StringBuffer offsetText = new StringBuffer();
		final int offset = o.getOffset();
		if (0 != offset) {
			if (offset >= 0) {
				offsetText.append('+');
			}
			offsetText.append(o.getOffset());
		}
		return DSAEnv.instance().messages().getText("DSA.DiceRoll.Content", "%0W%1%2",
				new Object[] { Integer.valueOf(o.getNumber()), Integer.valueOf(o.getEyes()), offsetText, });
	}

	public String getMoonPhaseString(final DSADate date, final TextFormatter tf) {
		final int moonday = date.getMoonday();
		final int lunarEclipseDay = date.getDaysSinceLastLunarEclipse();
		final List<String> moonPhase = getMoonPhase();
		final String phaseString = toStringValue(moonPhase.get(moonday));
		final I18NStringWithFillIns eclipse;
		if (0 == lunarEclipseDay) {
			eclipse = new I18NStringWithFillIns("DSA.MoonDay.Content.LunarEclipse", ", %0", new I18NStringWithFillIns[] { 
					new I18NStringWithFillIns("DSA.MoonDay.LunarEclipse", null, null),
			});
		} else {
			eclipse = I18NStringWithFillIns.EMPTY;
		}
		return DSAEnv.instance().messages().getText("DSA.MoonDay.Content", "%0 (%1%2)", new Object[] { 
				TextFormatter.format(moonday + 1, 2), 
				tf.formatIfPlainText(phaseString, 12), 
				eclipse,
		});
	}

	public String getHeadingText(final DSADateCalendar o, final TextDocumentBuilder doc) {
		return (o.getHeadingText(doc));
	}

	public String getContentText(final DSADateCalendar o, final TextDocumentBuilder doc) {
		return (o.getContentText(doc));
	}

	public String getHeadingText(final DSADateComplete o, final TextDocumentBuilder doc) {
		return getHeadingText(o.getWeather(), doc);
	}

	public String getContentText(final DSADateComplete o, final TextDocumentBuilder doc) {
		final StringBuffer ret = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final Iterator<HolidayType> itHolidays = Holiday.findHolidaysAt(o.getDate()).iterator();
		ret.append(tf.beginPreformatted());
		final DSADateCalendarTwelve cal = new DSADateCalendarTwelve(o.getDate());
		ret.append(getContentText(cal, doc));
		ret.append("\n\t");
		ret.append(getContentText(o.getWeather(), tf));
		ret.append("\n\t");
		while (itHolidays.hasNext()) {
			final HolidayType holiday = (HolidayType) itHolidays.next();
			ret.append(getContentText(holiday, tf));
			ret.append("\n\t");
		}
		ret.append(getContentText(o.getRandomDayEvent(), tf));
		ret.append(tf.endPreformatted());
		ret.append("\n\n\n");
		return (ret.toString());
	}

	public String getContentText(final HolidayType o, final TextFormatter tf) {
		final StringBuffer ret = new StringBuffer();
		ret.append(o.getName());
		final String rem = o.getRemark();
		if (!CollectionUtil.isNullOrEmpty(rem)) {
			ret.append(" (");
			ret.append(rem);
			ret.append(")");
		}
		return (ret.toString());
	}

	public String getHeadingText(final LanguageType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Language"); 
	}

	public String getContentText(final LanguageType o, final TextFormatter tf) {
		return o.getAbbr() + " (" + o.getName() + ")";
	}

	public String getHeadingText(final LanguageDistributionType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.LanguageDistribution"); 
	}

	public String getContentText(final TextDocumentBuilder doc, final LanguageDistributionType o) {
		final StringBuffer ret = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final List<LanguageProbabilityType> probs = o.getLanguageProbability();
		final Object[] arr = probs.toArray();
		final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[arr.length];
		for (int c = 0; c < arr.length; c++) {
			final LanguageProbabilityType langprob = (LanguageProbabilityType) arr[c];
			final LanguageType lang = (LanguageType) langprob.getLanguage();
			ci[c] = new TextFormatter.ColumnInfo(6, TextFormatter.RIGHT,
					tf.href(MasterDataModel.getHRefFor(lang), lang.getAbbr()));
		}
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
		ret.append(tf.beginTable(null, ti));
		ret.append(tf.formatTableHeading(ci));
		ret.append(tf.beginTableBody());
		ret.append(tf.beginTableRow());
		for (int c = 0; c < arr.length; c++) {
			final LanguageProbabilityType langprob = (LanguageProbabilityType) arr[c];
			float probval = langprob.getProbability();
			ret.append(tf.formatTableCell(ci[c], TextFormatter.format(probval, 5, 1) + "%"));
		}
		ret.append(tf.endTableRow());
		ret.append(tf.endTableBody());
		ret.append(tf.endTable());
		return ret.toString();
	}

	public String getHeadingText(final MonthType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Month"); 
	}

	public String getHeadingText(final PersonType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Person"); 
	}

	public String getContentText(final PersonType o, final TextFormatter tf) {
		Assert.assertNotNull(o);
		return o.getName();
	}

	public String getHeadingText(final PlantType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Plant"); 
	}

	public String getHeadingText(final RandomDayEvent o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.RandomDayEvent"); 
	}

	public String getContentText(final RandomDayEvent o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DateModel.RandomDayEvent", "%0: %1", new Object[] { 
				TextFormatter.format(o.getDiceEyes(), 3),
				tf.formatIfPlainText(o.message(), 60),
		} );
	}

	public String getHeadingText(final RegionType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Region"); 
	}

	public String getContentText(final RegionType o, final TextFormatter tf) {
		return o.getName() + " (" + o.getDescription() + ")"; 
	}

	public String getHeadingText(final SeasonType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Season"); 
	}

	public String getHeadingText(final ShopType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Shop"); 
	}

	public String getHeadingText(final TempleType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Temple"); 
	}

	public String getHeadingText(final TerrainType o, final TextFormatter tf) {
		return DSAEnv.instance().messages().getText("DSA.Terrain"); 
	}

	public String getHeadingText(final TownType o, final TextFormatter tf) {
		return o.getName();
	}

	/**
	 * @param doc
	 *            The document to add the text to
	 * @param town
	 *            The town
	 */
	public void addContentText(final TextDocumentBuilder doc, final TownType town) {
		final Messages messages = doc.messages();
		final RegionType region = (RegionType) town.getRegion();
		final TownSizeCategoryType cat = Town.calcSizeCategory(town, false);
		final TextFormatter tf = doc.tf();
		// Header
		doc.beginParagraph(null);
		doc.add(messages.getText(new I18NStringWithFillIns("Town.Inhabitants", null, new Object[] { 
				TextFormatter.format(town.getNumberOfInhabitants(), 5).toString(),
				tf.href(MasterDataModel.getHRefFor(cat), Integer.toString(cat.getIDnum())),
		})));
		doc.add(tf.newLine());
		doc.add(messages.getText("Town.Region", "%0",
				new Object[] { tf.href(MasterDataModel.getHRefFor(region), DSA.instance().getContentText(region, tf)), }));
		doc.endParagraph();
		// language probability distribution
		final LanguageDistributionType langDist = town.getLanguageDistribution();
		final int depth = 2 + 1;
		if (null != langDist) {
			doc.addSectionHeading(depth, "DSA.LanguageDistribution");
			doc.add(getContentText(doc, langDist));
		}

		doc.addSectionHeading(depth, "Town.Infrastructure");
		// Open table if HTML style
		final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[] {
				new TextFormatter.ColumnInfo(25, TextFormatter.LEFT, messages, "Shop.Specialty"),
				new TextFormatter.ColumnInfo(30, TextFormatter.LEFT, messages, "Town.Details"),
				new TextFormatter.ColumnInfo(30, TextFormatter.LEFT, messages, "Building.Chief"),
				new TextFormatter.ColumnInfo(25, TextFormatter.LEFT, messages, "Town.PriceFactor"),
				new TextFormatter.ColumnInfo(35, TextFormatter.LEFT, messages, "Shop.Quality"), };
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
		doc.beginTable(null, ti);

		// Temples
		{
			final TownSizeCategoryType borderTownSizeCategory = (TownSizeCategoryType) getTownSizeCategory().get(5);
			if (borderTownSizeCategory.getMinimumNumberOfInhabitants() <= town.getNumberOfInhabitants()) {
				doc.add(tf.formatTableFullTextRow(messages.getText("DSA.Temple"),
						messages.getText("Town.Reference.Box.Land")));
			} else {
				if (town.getTemple().isEmpty()) {
					doc.add(tf.formatTableFullTextRow(messages.getText("DSA.Temple"), messages.getText("Town.None")));
				}
			}
			final Iterator<TempleType> templeIt = town.getTemple().iterator();
			while (templeIt.hasNext()) {
				final TempleType t = (TempleType) templeIt.next();
				final PersonType chief = (PersonType) t.getChief();
				String chiefString = "";
				if (null != chief) {
					chiefString = getContentText(chief, tf);
				}
				doc.add(tf.beginTableRow());
				doc.add(tf.formatTableCell(ci[0], messages.getText("DSA.Temple")));
				doc.add(tf.formatTableCell(ci[1], Temple.getDeityListString(t, tf)));
				doc.add(tf.formatTableCell(ci[2], chiefString));
				doc.add(tf.formatTableCell(ci[3], (String) null));
				doc.add(tf.formatTableCell(ci[4], (String) null));
				doc.add(tf.endTableRow());
			}
		}

		// Hotels, Bars, Handcrafts, Merchands, Services, Healer, Enterprises
		boolean isDisplayedEnterpriseRemark = false;
		boolean isDisplayedTravellersRemark = false;
		final Iterator<ShopType> it = town.getShop().iterator();
		boolean isToFinishLoop = false;
		while (!isToFinishLoop) {
			boolean isToDisplayEnterpriseRemark = false;
			boolean isToDisplayTravellersRemark = false;
			ShopType shop = null;
			if (!it.hasNext()) {
				isToDisplayEnterpriseRemark = true;
				isToDisplayTravellersRemark = true;
				isToFinishLoop = true;
			} else {
				shop = (ShopType) it.next();
				if ((Shop.getCategory(shop).getIDnum() >= Shop.CATEGORY_ENTERPRISE) && (!isDisplayedEnterpriseRemark)) {
					isToDisplayEnterpriseRemark = true;
				}
				if ((Shop.getCategory(shop).getIDnum() >= Shop.CATEGORY_TRAVELLER) && (!isDisplayedTravellersRemark)) {
					isToDisplayTravellersRemark = true;
				}
			}
			if (isToDisplayEnterpriseRemark && (!isDisplayedEnterpriseRemark)) {
				final String enterprisesRemark = Town.getEnterprisesRemark(town);
				if (null != enterprisesRemark) {
					doc.add(tf.formatTableFullTextRow(Town.getEnterprisesRemark(town), null));
				}
				isDisplayedEnterpriseRemark = true;
			}
			if (isToDisplayTravellersRemark && (!isDisplayedTravellersRemark)) {
				final String travellersRemark = Town.getTravellersRemark(town);
				if (null != travellersRemark) {
					doc.add(tf.formatTableFullTextRow(messages.getText("Town.Travellers"), travellersRemark));
				}
				isDisplayedTravellersRemark = true;
			}
			if (!isToFinishLoop) {
				doc.add(tf.beginTableRow());
				String shopCatString = "";
				ShopCategoryType shopCategory = Shop.getCategory(shop);
				if (shopCategory.getIDnum() == Shop.CATEGORY_TRAVELLER) {
					shopCatString = shopCategory.getName();
				} else {
					final String specialtyName = Shop.getSpecialtyName(shop);
					shopCatString = ((specialtyName.equals("")) ? shopCategory.getName() : specialtyName);
				}
				doc.add(tf.formatTableCell(ci[0], shopCatString));
				shopCatString = "";
				if (Shop.getCategory(shop).getIDnum() == Shop.CATEGORY_TRAVELLER) {
					shopCatString = Shop.getSpecialtyName(shop);
				} else {
					final String shopName = shop.getName();
					if (!CollectionUtil.isNullOrEmpty(shopName)) {
						shopCatString = shopName;
					} else {
						// Nothing to show
					}
				}
				doc.add(tf.formatTableCell(ci[1], shopCatString));
				doc.add(tf.formatTableCell(ci[2], getContentText((PersonType) shop.getChief(), tf)));
				doc.add(tf.formatTableCell(ci[3], Shop.getPricingName(shop)));
				doc.add(tf.formatTableCell(ci[4], Shop.getQualityName(shop)));
				doc.add(tf.endTableRow());
			}
		}
		doc.add(tf.formatTableFullTextRow(messages.getText("Town.MarketdaysPerMonth"), Town.getMarketDaysRemark(town)));
		doc.endTable();
	}

	public String getHeadingText(final Weather o, final TextDocumentBuilder doc) {
		final DSADateCalendarTwelve cal = new DSADateCalendarTwelve(o.getDate());
		return getHeadingText(cal, doc);
	}

	public String getContentText(final Weather o, final TextFormatter tf) {
		final Messages messages = DSAEnv.instance().messages();
		final StringBuffer ret = new StringBuffer();
		final WeatherTableEntryType ent = o.getTableEntry();
		if (null != ent) {
			ret.append(messages.getText(new I18NStringWithFillIns("Weather.Weather", null, new String[] {
					TextFormatter.format(o.getWeather(), 2).toString(),
					TextFormatter.format(o.getWind(), 1).toString(),
					TextFormatter.format(o.getTemperature(), 1).toString(),
					ent.getName(),
					toStringValue(getWeather().getWindStrengthDescription().get(o.getWind())),
					toStringValue(getWeather().getTemperatureDescription().get(o.getTemperature())),
			} ))); 
		}
		return (ret.toString());
	}

	/**
	 * Converts the object to a double value.
	 * 
	 * @param o
	 *            The object to convert. Supported are JAXBElement, String,
	 *            Double, Float and Integer.
	 * @return The double value, or 0.0 if o has a type where no conversion is
	 *         implemented.
	 */
	public static double toDoubleValue(final Object o) {
		double ret = 0.0;
		if (o instanceof JAXBElement) {
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement) o;
			ret = toDoubleValue(e.getValue());
		} else if (o instanceof String) {
			ret = Double.parseDouble((String) o);
		} else if (o instanceof Double) {
			final Double d = (Double) o;
			ret = d.doubleValue();
		} else if (o instanceof Float) {
			final Float d = (Float) o;
			ret = d.doubleValue();
		} else if (o instanceof Integer) {
			final Integer d = (Integer) o;
			ret = d.doubleValue();
		}
		return ret;
	}

	/**
	 * Returns the same value.
	 * 
	 * @param i
	 *            The object to return.
	 * @return i.
	 */
	public static int toIntValue(final int i) {
		return i;
	}

	/**
	 * Converts the object to an int value.
	 * 
	 * @param o
	 *            The object to convert. Supported are JAXBElement, String and
	 *            Integer.
	 * @return The value, or 0 if o has a type where no conversion is
	 *         implemented.
	 */
	public static int toIntValue(final Object o) {
		int ret = 0;
		if (o instanceof JAXBElement) {
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement) o;
			Integer i = (Integer) e.getValue();
			ret = i.intValue();
		} else if (o instanceof String) {
			ret = Integer.parseInt((String) o);
		} else if (o instanceof Integer) {
			final Integer i = (Integer) o;
			ret = i.intValue();
		}
		return ret;
	}

	/**
	 * Converts the object to a String value.
	 * 
	 * @param o
	 *            The object to convert. Supported are JAXBElement and String.
	 * @return The value, or <code>null</code> if o has a type where no
	 *         conversion is implemented.
	 */
	public static String toStringValue(final Object o) {
		String ret = null;
		if (o instanceof JAXBElement) { // SAXB 2.x
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement) o;
			ret = (String) e.getValue();
		} else if (o instanceof String) {
			ret = (String) o;
		} else if (o.getClass().isAnnotationPresent(XmlEnum.class)) { // SAXB
																		// 2.x
			// How to do this generally?
			if (o instanceof Color) {
				Color c = (Color) o;
				ret = c.value();
			}
		}
		return ret;
	}

	/**
	 * Converts the object to a List value.
	 * 
	 * @param o
	 *            The object to convert. Supported are JAXBElement and List.
	 * @return The value, or <code>null</code> if o has a type where no
	 *         conversion is implemented.
	 */
	public static List<?> toListValue(final Object o) {
		List<?> ret = null;
		if (o instanceof JAXBElement) {
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement) o;
			ret = (List<?>) e.getValue();
		} else if (o instanceof List) {
			ret = (List<?>) o;
		}
		return ret;
	}

	/**
	 * Converts the object to a PersonNamePartType value.
	 * 
	 * @param o
	 *            The object to convert. Supported are JAXBElement and
	 *            PersonNamePartType.
	 * @return The value, or <code>null</code> if o has a type where no
	 *         conversion is implemented.
	 */
	public static PersonNamePartType toPersonNamePartValue(final Object o) {
		PersonNamePartType ret = null;
		if (o instanceof JAXBElement) {
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement) o;
			ret = (PersonNamePartType) e.getValue();
		} else if (o instanceof PersonNamePartType) {
			ret = (PersonNamePartType) o;
		}
		return ret;
	}

	/**
	 * Converts the object to a QuestTextPartType value.
	 * 
	 * @param o
	 *            The object to convert. Supported are JAXBElement and
	 *            QuestTextPartType.
	 * @return The value, or <code>null</code> if o has a type where no
	 *         conversion is implemented.
	 */
	public static QuestTextPartType toQuestTextPartValue(final Object o) {
		QuestTextPartType ret = null;
		if (o instanceof JAXBElement) {
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement) o;
			ret = (QuestTextPartType) e.getValue();
		} else if (o instanceof QuestTextPartType) {
			ret = (QuestTextPartType) o;
		}
		return ret;
	}

	/**
	 * Finder for business objects. Searches the <em>list</em> for an object with the given <em>idnum</em>.
	 * @param list The list of business objects. Should only contain one type of business objecs, otherwise the idnum may not be unique.
	 * @param idnum The IDnum to search for.
	 * @return The object with that IDnum.
	 */
	public static BusinessObjectType findBOByIDnum(final List<? extends BusinessObjectType> list, final int idnum) {
		final Iterator<? extends BusinessObjectType> it = list.iterator();
		while (it.hasNext()) {
			final BusinessObjectType bo = (BusinessObjectType)it.next();
			if (bo.getIDnum() == idnum) {
				return bo;
			}
		}
		return null;
	}

	/**
	 * Finder for business objects. Searches the <em>list</em> for an object with the given <em>id</em>.
	 * @param list The list of business objects.
	 * @param id The ID to search for.
	 * @return The object with that ID.
	 */
	public static BusinessObjectType findBOByID(final List<? extends BusinessObjectType> list, final String id) {
		final Iterator<? extends BusinessObjectType> it = list.iterator();
		while (it.hasNext()) {
			final BusinessObjectType bo = (BusinessObjectType)it.next();
			final String s = bo.getID();
			if ((null != s) && s.equals(id)) {
				return bo;
			}
		}
		return null;
	}

	/**
	 * Finder for business objects. Searches the <em>list</em> for an object with the given <em>name</em>.
	 * @param list The list of business objects. Should only contain one type of business objecs, otherwise the name may not be unique.
	 * @param name The name to search for.
	 * @return The object with that name.
	 */
	public static BusinessObjectType findBOByName(final List<? extends BusinessObjectType> list, final String name) {
		final Iterator<? extends BusinessObjectType> it = list.iterator();
		while (it.hasNext()) {
			final BusinessObjectType bo = (BusinessObjectType)it.next();
			final String s = bo.getName();
			if ((null != s) && s.equals(name)) {
				return bo;
			}
		}
		return null;
	}

	/**
	 * Given a list of business objects, returns an array containing all the IDnums of the objects in that list.
	 * @param list The list of business objects. Should only contain one type of business objecs.
	 * @return an array containing all the IDnums of the objects in <em>list</em>.
	 */
	public static int[] toIDnumArray(final List<BusinessObjectType> list) {
		int[] arr;
		if (null == list) {
			arr = new int[0];
		} else {
			final int size = list.size();
			arr = new int[size];
			final Iterator<BusinessObjectType> it = list.iterator();
			int i = 0;
			while((it.hasNext()) && (i < size)) {
				BusinessObjectType o = (BusinessObjectType)it.next();
				arr[i] = o.getIDnum();
				i++;
			}
		}
		return arr;
	}

	/**
	 * Given a business object, returns an array containing the IDnum of the object.
	 * @param o The single business object.
	 * @return an array containing the IDnum of the object.
	 */
	public static int[] toIDnumArray(final BusinessObjectType o) {
		int[] arr;
		if (null == o) {
			arr = new int[0];
		} else {
			arr = new int[1];
			arr[0] = o.getIDnum();
		}
		return arr;
	}

	public static String[] toStringArrayWithCounter(final List<?> list) {
		final Iterator<?> it = list.iterator();
		final int max = list.size();
		final String[] ret = new String[max];
		int i = 0;
		while (it.hasNext()) {
			final String s = toStringValue(it.next());
			ret[i] = TextFormatter.format(i, 2) + ": " + s; 
			i++;
		}
		return ret;
	}

	/**
	 * Creates a HTML snippet containing a selection, filled with the given
	 * values.
	 * 
	 * @param selectionId
	 *            Under this name the select object can be referenced within
	 *            HTML. The parameter is used as value of the <code>name</code>
	 *            tag and <code>id</code> tag.
	 * @param list
	 *            Those items will be filled into the selection. Objects
	 *            instance of {@link BusinessObjectType} or {@link TownType} are
	 *            treated special, otherwise method
	 *            {@link java.lang.Object#toString()} is used as textual
	 *            representation.
	 * @param selectedIdx
	 *            The index of the item to preselect.
	 * @return HTML snippet.
	 */
	public static String toHtmlSelect(final String selectionId, List<?> list, final int selectedIdx) {
		final StringBuffer s = new StringBuffer();
		s.append("<select" 
				+ TextFormatter.htmlAttribute(TextFormatter.ID, selectionId) 
				+ TextFormatter.htmlAttribute(TextFormatter.NAME, selectionId) + ">\n");
		if (null != list) {
			final Iterator<?> it = list.iterator();
			int cnt = 0;
			while (it.hasNext()) {
				final Object o = it.next();
				int value = -1;
				String text = null;
				boolean isSelected = false;
				if (o instanceof BusinessObjectType) {
					BusinessObjectType bo = (BusinessObjectType) o;
					value = bo.getIDnum();
					text = bo.getName();
				} else if (o instanceof TownType) {
					final TownType t = (TownType) o;
					value = cnt;
					text = t.getName();
				} else {
					value = cnt;
					text = o.toString();
				}
				if (value == selectedIdx) {
					isSelected = true;
				}
				s.append("<option" + TextFormatter.htmlAttribute(TextFormatter.VALUE, Integer.toString(value)));
				s.append(HtmlHelper.selectedIf(isSelected));
				s.append(">");
				s.append(text);
				s.append("</option>\n");
				cnt++;
			}
		}
		s.append("</select>\n");
		return s.toString();
	}

	/**
	 * Creates a HTML snippet containing a selection, filled with the given
	 * values.
	 * 
	 * @param list
	 *            Those items will be filled into the selection. Objects must be
	 *            instance of {@link BusinessObjectType}.
	 * @param bo
	 *            The object to preselect.
	 * @param selectionId
	 *            Under this name the select object can be referenced within
	 *            HTML.
	 * @return HTML snippet.
	 */
	public static String toHtmlSelect(final String selectionId, final List<?> list, final BusinessObjectType bo) {
		final int idx = ((null == bo) ? -1 : bo.getIDnum());
		return toHtmlSelect(selectionId, list, idx);
	}

	/**
	 * Creates a HTML snippet containing a selection, filled with the given
	 * values.
	 * 
	 * @param list
	 *            Those items will be filled into the selection.
	 * @param selectedIdx
	 *            The index of the item to preselect.
	 * @param selectionId
	 *            Under this name the select object can be referenced within
	 *            HTML.
	 * @return HTML snippet.
	 */
	public static String toHtmlSelect(final String selectionId, final String[] list, final int selectedIdx) {
		final StringBuffer s = new StringBuffer();
		s.append("<select" 
				+ TextFormatter.htmlAttribute(TextFormatter.ID, selectionId) 
				+ TextFormatter.htmlAttribute(TextFormatter.NAME, selectionId) + ">\n");
		if (null != list) {
			for (int i = 0; i < list.length; i++) {
				s.append("<option" + TextFormatter.htmlAttribute(TextFormatter.VALUE, Integer.toString(i)));
				s.append(HtmlHelper.selectedIf(i == selectedIdx));
				s.append(">");
				s.append(list[i]);
				s.append("</option>\n");
			}
		}
		s.append("</select>\n");
		return s.toString();
	}

	/**
	 * Creates a HTML snippet containing a multi-selection (which may be
	 * represented as lots of checkboxes), filled with the given values.
	 * 
	 * @param id
	 *            The id (or prefix) of the HTML elements.
	 * @param list
	 *            Those items will be filled into the multi-selection.
	 * @param checkedIDs
	 *            Those IDs are pre-selected.
	 * @return HTML snippet.
	 * @see HtmlHelper#toHtmlCheckList(String, String[], int[])
	 */
	public static String toHtmlCheckList(final String id, List<? extends BusinessObjectType> list,
			final int[] checkedIDs) {
		final StringBuffer s = new StringBuffer();
		if (null != list) {
			final int max = list.size();
			final String[] texts = new String[max];
			final Iterator<? extends BusinessObjectType> it = list.iterator();
			int i = 0;
			while ((i < max) && it.hasNext()) {
				final BusinessObjectType bo = (BusinessObjectType) it.next();
				texts[i] = bo.getName();
				i++;
			}
			s.append(HtmlHelper.toHtmlCheckList(id, texts, checkedIDs));
		}
		return s.toString();
	}

	/**
	 * Checks one complete list of business objects of one type for constraints
	 * that could not be checked within XML. Current checks performed are:
	 * <ul>
	 * <li>Is IDnum for each object unique?
	 * <li>Are the IDnums starting by 0 and counting up?
	 * </ul>
	 * 
	 * @param list
	 *            The complete list of business objects of that type. Should
	 *            contain only one type of business objecs, otherwise the name
	 *            may not be unique. Even more, it should contain all business
	 *            objects of that type.
	 */
	protected void checkBOs(final List<? extends BusinessObjectType> list) {
		final Iterator<? extends BusinessObjectType> it = list.iterator();
		int i = 0;
		while (it.hasNext()) {
			final BusinessObjectType bo = (BusinessObjectType) it.next();
			Assert.assertEquals(bo.getIDnum().intValue(), i);
			i++;
		}
	}

	/**
	 * This class implements the Singleton pattern from the Gang of Four.
	 */
	protected DSA() {
		// Nothing to do, all interesting things are done later
	}

	public String outputObject(final Object o) {
		final StringBuffer buf = new StringBuffer();
		final Logger logger = DSAEnv.instance().logger();
		try {
			buf.append("Object=" + o.toString() + "\n");
			buf.append(jaxbWrapper.marshalObject(o));
			buf.append("\nI've been here");
		} catch (final NoClassDefFoundError ncdfe) {
			logger.info(ncdfe.toString());
			ncdfe.printStackTrace();
			buf.append(ncdfe.toString());
		} catch (final IllegalArgumentException iae) {
			logger.info(iae.toString());
			iae.printStackTrace();
			buf.append(iae.toString());
		}
		return buf.toString();
	}

	static void init(DSA instance) {
		instance.init();
	}

	private void init() {
		if (null == delegate) {
			final Logger logger = DSAEnv.instance().logger();
			try {
				InputStream dataXmlStream = null;
				URL filename = DSAEnv.instance().resources().getURLToResource(DATA_XML_FILENAME);
				logger.debug("Loading " + DATA_XML_FILENAME + " file from " + filename);
				dataXmlStream = new DataInputStream(filename.openStream());
				jaxbWrapper = new JAXBWrapper();
				delegate = jaxbWrapper.loadDsaType(dataXmlStream);
			} catch (final NoClassDefFoundError ncdfe) {
				logger.fatal("NoClassDefFoundError: " + ncdfe.getMessage(), ncdfe);
			} catch (final IOException ioe) {
				logger.fatal("IOException: " + ioe.getMessage(), ioe);
			} catch (final IllegalArgumentException iae) {
				logger.fatal("IllegalArgumentException: " + iae.getMessage(), iae);
			}
			if (null == delegate) {
				logger.fatal("Couldn't load XML data file.");
			} else {
				check();
				logger.trace("Using XML data file.");
			}
		}
	}

	/**
	 * Checks if <em>o</em> is a JAXB object (or is built otherwise for
	 * stand-alone).
	 * 
	 * @param o
	 *            The object to check.
	 * @return <code>true</code> if annotated to be an {@link XmlAccessorType} (JAXB
	 *         2.x)
	 */
	public boolean isImpl(final Object o) {
		boolean ret = false;
		try {
			if (o.getClass().isAnnotationPresent(XmlAccessorType.class)) { 
				// SAXB 2.x
				ret = true;
			}
		} catch (final NoClassDefFoundError e) {
			// Nothing to do
		}
		return ret;
	}

	private void logState() {
		final Logger logger = DSAEnv.instance().logger();
		logger.trace("JAXBWrapper"
				+ ((null == jaxbWrapper) ? "=null" : ((jaxbWrapper.isActive()) ? " is active" : " is inactive")));
	}

	public LanguageDistributionType createLanguageDistributionType() {
		logState();
		if ((null != jaxbWrapper) && (jaxbWrapper.isActive())) {
			return jaxbWrapper.createLanguageDistributionType();
		} else {
			return null;
		}
	}

	public LanguageProbabilityType createLanguageProbabilityType() {
		logState();
		if ((null != jaxbWrapper) && (jaxbWrapper.isActive())) {
			return jaxbWrapper.createLanguageProbabilityType();
		} else {
			return null;
		}
	}

	public PersonType createPersonType() {
		logState();
		if ((null != jaxbWrapper) && (jaxbWrapper.isActive())) {
			return jaxbWrapper.createPersonType();
		} else {
			return null;
		}
	}

	public ShopType createShopType() {
		logState();
		if ((null != jaxbWrapper) && (jaxbWrapper.isActive())) {
			return jaxbWrapper.createShopType();
		} else {
			return null;
		}
	}

	public TempleType createTempleType() {
		logState();
		if ((null != jaxbWrapper) && (jaxbWrapper.isActive())) {
			return jaxbWrapper.createTempleType();
		} else {
			return null;
		}
	}

	public TownType createTownType() {
		logState();
		if ((null != jaxbWrapper) && (jaxbWrapper.isActive())) {
			return jaxbWrapper.createTownType();
		} else {
			return null;
		}
	}

	@Override
	public List<DeityType> getDeity() {
		return delegate.getDeity();
	}

	@Override
	public List<CalendarType> getCalendar() {
		return delegate.getCalendar();
	}

	@Override
	public List<PlantCategoryType> getPlantCategory() {
		return delegate.getPlantCategory();
	}

	@Override
	public List<TerrainType> getTerrain() {
		return delegate.getTerrain();
	}

	@Override
	public List<String> getMoonPhase() {
		return delegate.getMoonPhase();
	}

	@Override
	public List<TownSizeCategoryType> getTownSizeCategory() {
		return delegate.getTownSizeCategory();
	}

	@Override
	public List<QualityCategoryType> getQualityCategory() {
		return delegate.getQualityCategory();
	}

	@Override
	public List<String> getTempleDistribution() {
		return delegate.getTempleDistribution();
	}

	@Override
	public List<SexType> getSex() {
		return delegate.getSex();
	}

	@Override
	public List<LanguageType> getLanguage() {
		return delegate.getLanguage();
	}

	@Override
	public List<ShopCategoryType> getShopCategory() {
		return delegate.getShopCategory();
	}

	@Override
	public List<DayOfWeekType> getDayOfWeek() {
		return delegate.getDayOfWeek();
	}

	@Override
	public List<ShopSpecialtyType> getShopSpecialty() {
		return delegate.getShopSpecialty();
	}

	@Override
	public List<PlantType> getPlant() {
		return delegate.getPlant();
	}

	@Override
	public List<HolidayType> getHoliday() {
		return delegate.getHoliday();
	}

	@Override
	public List<RegionType> getRegion() {
		return delegate.getRegion();
	}

	@Override
	public List<SeasonType> getSeason() {
		return delegate.getSeason();
	}

	@Override
	public List<QualityNamesTableType> getQualityNamesTable() {
		return delegate.getQualityNamesTable();
	}

	@Override
	public List<PricingCategoryType> getPricingCategory() {
		return delegate.getPricingCategory();
	}

	@Override
	public List<MonthType> getMonth() {
		return delegate.getMonth();
	}

	@Override
	public List<TownType> getTown() {
		return delegate.getTown();
	}

	@Override
	public List<RandomDayEventTableEntryType> getRandomDayEventTableEntry() {
		return delegate.getRandomDayEventTableEntry();
	}

	@Override
	public WeatherType getWeather() {
		return delegate.getWeather();
	}

	@Override
	public void setWeather(WeatherType o) {
		delegate.setWeather(o);
	}

	@Override
	public PubNameType getPubName() {
		return delegate.getPubName();
	}

	@Override
	public void setPubName(PubNameType o) {
		delegate.setPubName(o);
	}

	@Override
	public PersonNameType getPersonName() {
		return delegate.getPersonName();
	}

	@Override
	public void setPersonName(PersonNameType o) {
		delegate.setPersonName(o);
	}

	@Override
	public MapType getMap() {
		return delegate.getMap();
	}

	@Override
	public void setMap(MapType o) {
		delegate.setMap(o);
	}

	@Override
	public QuestTextType getQuestText() {
		return delegate.getQuestText();
	}

	@Override
	public void setQuestText(final QuestTextType o) {
		delegate.setQuestText(o);
	}

	/** Additional check for the moon phases object, if the list has the right size. */
	protected void checkMoonPhases() {
		Assert.assertEquals(28, getMoonPhase().size()); 
	}

	/** Additional checks on language probabilities of regions. */
	protected void checkRegions() {
		final Iterator<RegionType> it = getRegion().iterator();
		while (it.hasNext()) {
			final RegionType region = it.next();
			LanguageDistribution.checkLanguageProbability(this, region.getLanguageDistribution());
		}
	}

	/** Additional checks on language probabilities of towns. */
	protected void checkTowns() {
		final Iterator<TownType> it = getTown().iterator();
		while (it.hasNext()) {
			final TownType town = it.next();
			LanguageDistribution.checkLanguageProbability(this, town.getLanguageDistribution());
		}
	}

	/**
	 * Checks for all constraints in the XML file that could not be checked within XML.
	 */
	public void check() {
		checkBOs(getDeity());
		checkBOs(getSeason());
		checkBOs(getMonth());
		checkBOs(getDayOfWeek());
		checkBOs(getCalendar());
		checkMoonPhases();
		checkBOs(getSex());
		checkBOs(getLanguage());
		checkBOs(getRegion());
		//checkBOs(getPerson());
		checkRegions();
		checkBOs(getTerrain());
		checkBOs(getPlantCategory());
		checkBOs(getPlant());
		checkBOs(getShopCategory());
		checkBOs(getShopSpecialty());
		checkRegions();
		checkTowns();
		if (null != getPubName()) {
			checkBOs(getPubName().getGender());
		}
	}

	/** We delegate the data-centric methods to a generated object. */
	protected DsaType delegate;

	private JAXBWrapper jaxbWrapper = null;
}
