package de.ped.dsatool.logic;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Observable;

import de.ped.dsatool.dsa.generated.SeasonType;
import de.ped.tools.Assert;
import de.ped.tools.MathUtil;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A class representing a single unique day in Aventuria.
 * This representation is independent of any calendar or other time counting practices 
 * and should be used to translate between the different calendars in use in Aventuria.
 * It uses an integer to count the days since 
 * Praios 1st of the year of Bosparans fall (0 BF).
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Der Aventurische Kalender" (p. 60..63)</p>
 *
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSADate extends Observable implements Cloneable {
	private Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
		logger.removeAppender(Logger.DEFAULT_STRING_WRITER_APPENDER);
	}

	private int daysSinceBF = 0;

	/** Creates a date representing Praios 1st, 0 BF. */
	public DSADate() {
		daysSinceBF = 0;
	}

	/**
	 * Creates a date with the given number of days distance to Praios 1st, 0 BF.
	 * 
	 * @param daysSinceBF
	 *            the number of days distance to Praios 1st, 0 BF.
	 */
	public DSADate(final int daysSinceBF) {
		this.daysSinceBF = daysSinceBF;
	}

	/**
	 * Creates an Aventurian date based on the date on earth.
	 * 
	 * @param earthDate
	 *            The date on earth to convert to Aventurian time
	 * @see #setFromEarthDate(Date)
	 */
	public DSADate(final Date earthDate) {
		setFromEarthDate(earthDate);
	}

	/**
	 * @return the days since Praios 1st, 0 BF
	 */
	public int getDaysSinceBF() {
		return daysSinceBF;
	}

	/**
	 * @param daysSinceBF
	 *            the days since Praios 1st, 0 BF
	 */
	public void setDaysSinceBF(final int daysSinceBF) {
		this.daysSinceBF = daysSinceBF;
		setChanged();
	}

	/** The duration between two new moons in days. */
	public static final int MOON_MONTH_DAYS = 28;

	/** The duration between two new lunar eclipses in moon cycles. */
	public static final int MOON_MONTHS_BETWEEN_LUNAR_ECLIPSES = 240;

	/** The duration between two new lunar eclipses in days. */
	public static final int DAYS_BETWEEN_LUNAR_ECLIPSES = MOON_MONTH_DAYS * MOON_MONTHS_BETWEEN_LUNAR_ECLIPSES;

	/**
	 * @return the day of the moon phase. 0 represents new moon.
	 */
	public int getMoonday() {
		final int moonDayAtPraios1st0BF = 16;
		return MathUtil.modulo(daysSinceBF + moonDayAtPraios1st0BF, MOON_MONTH_DAYS);
	}

	/**
	 * @return the number of days since the last lunar eclipse. 0 represents
	 *         lunar eclipse. Lunar eclipse always happens two days after new
	 *         moon, but only rarely.
	 */
	public int getDaysSinceLastLunarEclipse() {
		final int daysSinceLastLunarEclipseAtPraios1st0BF = (15 - 1) + (24 - 1) * MOON_MONTH_DAYS;
		return MathUtil.modulo(daysSinceBF + daysSinceLastLunarEclipseAtPraios1st0BF, DAYS_BETWEEN_LUNAR_ECLIPSES);
	}

	/** The duration in days of a standard earth year. */
	public static final int DAYS_PER_EARTH_YEAR = 365;

	/** The duration in days of half a standard earth year (rounded up). */
	private static final int DAYS_PER_EARTH_HALF_YEAR = (DAYS_PER_EARTH_YEAR + 1) / 2;

	/** The julian earth day of February 29 (if it's a leap year). */
	private static final int JDAY_FEB29 = 60;

	/**
	 * Supports in dealing with the gregorian change when converting between DSA and Earth time.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class GregorianCalendarHelper {
		public static final int DAYS_SKIPPED = 10;

		private GregorianCalendar calendar;
		private int gregorianChangeYear;
		private int gregorianChangeDayOfYear;

		public GregorianCalendarHelper(final GregorianCalendar calendar) {
			setCalendar(calendar);
			final Date changeDate = calendar.getGregorianChange();
			calendar.setTime(changeDate);
			gregorianChangeYear = calendar.get(Calendar.YEAR);
			gregorianChangeDayOfYear = calendar.get(Calendar.DAY_OF_YEAR);
			// System.out.println("Gregorian change at " + gregorianChangeYear + "/" + gregorianChangeDayOfYear);
		}

		public void setCalendar(final GregorianCalendar calendar) {
			this.calendar = calendar;

		}

		public GregorianCalendar getCalendar() {
			return calendar;
		}

		/** @return the year in which the gregorian change took place. */
		public int getGregorianChangeYear() {
			return gregorianChangeYear;
		}

		/** @return the first day after the skipped days. */
		public int getGregorianChangeDayOfYear() {
			return gregorianChangeDayOfYear;
		}

		/**
		 * In the gregorian change year, all days after the skipped days need to
		 * be corrected for the skipped days.
		 * 
		 * @param year
		 *            The earth year following Gregorian calendar
		 * @param dayOfYear
		 *            The day of year following Gregorian calendar
		 * @return <code>true</code> if to correct
		 */
		public boolean isToCorrect(final int year, final int dayOfYear) {
			return ((getGregorianChangeYear() == year) && (getGregorianChangeDayOfYear() <= dayOfYear));
		}

		/**
		 * Behavior of {@link GregorianCalendar} changed between JDK 1.6 and
		 * 1.7. To take care of this, we need this helper function.
		 * 
		 * @return <code>true</code> for versions of Java older than 1.7.
		 */
		public boolean isJavaAsymetrical() {
			final String javaVersionInfo = System.getProperty("java.version");
			final boolean ret = (0 < "1.7".compareTo(javaVersionInfo));
			return ret;
		}
	}

	private static final GregorianCalendarHelper GREGORIAN = new GregorianCalendarHelper(new GregorianCalendar());

	/**
	 * @return the date on earth from the DSA date. Calculation is based on the
	 *         year change method, that means, January 1st is always equal to
	 *         Praios 1st. There is no leap year in Aventuria, so 1st of March
	 *         is the same as Feb 29. Furthermore, before 1996 (or 1019 BF), one
	 *         earth year was two Aventurian years, and since 1996, one earth
	 *         year is one Aventurian year.
	 * @see <a href="http://www.dasschwarzeauge.de/index.php?id=49">Aventurische
	 *      Zeitrechnung auf dasschwarzeauge.de</a>
	 * @see <a href=
	 *      "http://www.wiki-aventurica.de/index.php?title=Kalender">Zeitrechnung
	 *      auf wiki-aventurica.de</a>
	 */
	public Date toEarthDate() {
		final int daysSinceBF = getDaysSinceBF();
		final int avYear = MathUtil.divisio(daysSinceBF, DAYS_PER_EARTH_YEAR);
		final int avJday = MathUtil.modulo(daysSinceBF, DAYS_PER_EARTH_YEAR);
		int earthYear;
		int earthJday;
		if (1019 <= avYear) { 
			// Since 1996 (or 1019 BF), one earth year is one Aventurian year
			earthYear = 1996 - 1019 + avYear;
			earthJday = avJday;
		} else {
			earthYear = 1996 - (1019 + 1 - avYear) / 2; // +1 to round correctly
			if (0 == MathUtil.modulo(avYear, 2)) {
				// odd avYears come first, so even avYears represent the second half of the earth year
				earthJday = (avJday + DAYS_PER_EARTH_YEAR) / 2;
			} else { 
				// Now the easy part, the first half of the earth year
				earthJday = avJday / 2;
			}
			if (GREGORIAN.isToCorrect(earthYear, earthJday) && (!GREGORIAN.isJavaAsymetrical())) {
				earthJday -= GregorianCalendarHelper.DAYS_SKIPPED;
			}
		}
		final GregorianCalendar earthCal = (GregorianCalendar) GREGORIAN.getCalendar().clone();
		final boolean isLeapYear = earthCal.isLeapYear(earthYear);
		if (isLeapYear && earthJday >= JDAY_FEB29) { 
			// Get rid of February 29th.
			earthJday++;
		}
		earthCal.set(GregorianCalendar.YEAR, earthYear);
		earthCal.set(GregorianCalendar.DAY_OF_YEAR, earthJday + 1);
		// We choose noon to avoid problems at night when switching summer and winter time
		earthCal.set(GregorianCalendar.HOUR_OF_DAY, 12);
		final Date earthDate = earthCal.getTime();
		logger.debug("Aventurian date " + daysSinceBF + " (Year " + avYear + " JDay " + avJday + ") --> " + earthDate);
		return earthDate;
	}

	/**
	 * Calculates the DSA date from the date on earth. Calculation is based on
	 * the year change method, that means, January 1st is always equal to Praios
	 * 1st. There is no leap year in Aventuria, so 1st of March is the same as
	 * Feb 29. Furthermore, before 1996 (or 1019 BF), one earth year was two
	 * aventurian years, and since 1996, one earth year is one aventurian year.
	 * 
	 * @param earthDate
	 *            The date on earth to create the DSA date from
	 * @see <a href="http://www.dasschwarzeauge.de/index.php?id=49">Aventurische
	 *      Zeitrechnung auf dasschwarzeauge.de</a>
	 * @see <a href=
	 *      "http://www.wiki-aventurica.de/index.php?title=Kalender">Zeitrechnung
	 *      auf wiki-aventurica.de</a>
	 */
	public void setFromEarthDate(final Date earthDate) {
		final GregorianCalendar earthCal = (GregorianCalendar) GREGORIAN.getCalendar().clone();
		earthCal.setTime(earthDate);
		final int earthYear = earthCal.get(GregorianCalendar.YEAR);
		final int earthDayOfYear = earthCal.get(GregorianCalendar.DAY_OF_YEAR);
		final boolean isLeapYear = earthCal.isLeapYear(earthYear);
		int jday = earthDayOfYear - 1; // jday is 0 for January 1st.
		if (isLeapYear && jday >= JDAY_FEB29) {
			// Get rid of February 29th.
			jday--;
		}
		int avYear;
		int avJday;
		if (1996 <= earthYear) {
			// Since 1996 (or 1019 BF), one earth year is one Aventurian year
			avYear = 1019 - 1996 + earthYear;
			avJday = jday;
		} else {
			// Before 1996 (or 1019 BF), one earth year was two Aventurian years
			avYear = 1019 - 2 * (1996 - earthYear);
			if (DAYS_PER_EARTH_HALF_YEAR < jday) {
				avYear++;
				avJday = 2 * (jday - DAYS_PER_EARTH_HALF_YEAR) + 1;
			} else {
				avJday = 2 * jday;
			}
			if (GREGORIAN.isToCorrect(earthYear, earthDayOfYear)) {
				// Correct the 10 days inserted by Pope Gregor
				avJday += 2 * GregorianCalendarHelper.DAYS_SKIPPED;
			}
		}
		final int daysSinceBF = avYear * DAYS_PER_EARTH_YEAR + avJday;
		logger.debug("Earth date " + earthDate.toString() + " --> " + daysSinceBF + " (Year " + avYear + " JDay " + avJday + ")");
		setDaysSinceBF(daysSinceBF);
	}

	/**
	 * @return The actual season, as defined for the calendars based on the twelve gods.
	 */
	public SeasonType season() {
		final int jday = MathUtil.modulo(daysSinceBF, DAYS_PER_EARTH_YEAR);
		int seasonID = 1;
		if (jday < 60) {
			seasonID = 1; // Summer
		} else if (jday < 60 + 90) {
			seasonID = 2; // Autumn
		} else if (jday < 60 + 90 + 90) {
			seasonID = 3; // Winter
		} else if (jday < 60 + 90 + 90 + 90) {
			seasonID = 0; // Sprint
		} else {
			seasonID = 1; // Summer
		}
		return (SeasonType) DSA.findBOByIDnum(DSA.instance().getSeason(), seasonID);
	}

	/**
	 * @param other
	 *            the date with which is compared
	 * @return how many days is this date after the other in the calendar. (The
	 *         result may get negative if this date is before the other.)
	 */
	public int daysAfter(final DSADate other) {
		return (daysSinceBF - other.daysSinceBF);
	}

	/**
	 * Adds days to the actual date.
	 * 
	 * @param days
	 *            so many days will go by. Is allowed to be a negative number as
	 *            well.
	 */
	public void addDays(final int days) {
		daysSinceBF += days;
		setChanged();
	}

	/**
	 * @param other
	 *            The date with which is compared.
	 * @return if this date is after the other in the calendar
	 */
	public boolean isAfter(final DSADate other) {
		return (daysSinceBF > other.daysSinceBF);
	}

	/**
	 * @param other
	 *            The date with which is compared.
	 * @return if this date is before the other in the calendar
	 */
	public boolean isBefore(final DSADate other) {
		return (daysSinceBF < other.daysSinceBF);
	}

	/**
	 * @param other
	 *            The date with which is compared.
	 * @return if this date is the same as the other
	 */
	@Override
	public boolean equals(final Object other) {
		if (other instanceof DSADate) {
			final DSADate o = (DSADate) other;
			logger.trace("Comparing days since BF " + daysSinceBF + " with " + o.daysSinceBF);
			return (daysSinceBF == o.daysSinceBF);
		} else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return daysSinceBF;
	}

	@Override
	public DSADate clone() {
		DSADate ret = null;
		try {
			ret = (DSADate) super.clone();
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}

	@Override
	public String toString() {
		return "DSADate(" + daysSinceBF + ")";
	}
}
