package de.ped.dsatool.logic;

import de.ped.dsatool.dsa.generated.SeasonType;
import de.ped.tools.MathUtil;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.Textable;
import de.ped.tools.log.Logger;

/**
 * A base class for all calendars or other days counting practices in Aventuria.
 * We assume that all days counting practices have the concept of a year -
 * that is, the counting repeats after roughly the same timeframe each time.
 * This basic implementation sets the duration of a year to 365 days 
 * and starts counting at 1st of Praios 0 BF - 
 * so it really is a calendar for Bosparans fall, but without the knowledge of months,
 * as we can't assume this concept to be valid for all subclasses.
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Der Aventurische Kalender" (p. 60..63)</p>
 *
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSADateCalendar implements Textable {
	protected Logger getLogger() {
		return DSAEnv.instance().logger();
	}

	/** The number of days the year has for the calendar Bosparans fall. */
	public static final int DAYS_PER_YEAR_BF = 365;

	/** The calendar Bosparans fall has a year zero. */
	public static final boolean HAS_YEAR_ZERO_BF = true;

	/** The date this calendar knows as current. */
	private DSADate date;

	private int daysFromYear0ToBF = 0;

	private int daysPerYear = DAYS_PER_YEAR_BF;

	private boolean hasYear0 = HAS_YEAR_ZERO_BF;

	private String name = "Basiskalender";

	/** 
	 * Creates a date representing Praios 1st, 0 BF.
	 * @see DSADate#DSADate() 
	 */
	public DSADateCalendar() {
		date = new DSADate();
	}

	/**
	 * Creates a view following this calendar to the given date.
	 * 
	 * @param date
	 *            the DSA date to be represented using this calendar
	 */
	public DSADateCalendar(final DSADate date) {
		this.date = date;
	}

	/**
	 * @return the {@link DSADate} that is represented using this calendar.
	 */
	public DSADate getDSADate() {
		return date;
	}

	/**
	 * Sets another {@link DSADate} to be represented using this calendar.
	 * 
	 * @param date
	 *            the date to be represented using this calendar
	 */
	public void setDSADate(final DSADate date) {
		this.date = date;
	}

	/** 
	 * @return the name of this calendar. 
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets the name of this calendar.
	 * 
	 * @param name
	 *            the name of this calendar
	 */
	public void setName(final String name) {
		this.name = name;
	}

	/** 
	 * @return the number of days a year in this calendar has. 
	 */
	public int getDaysPerYear() {
		return daysPerYear;
	}

	/**
	 * Sets the number of days a year in this calendar has.
	 * 
	 * @param daysPerYear
	 *            the number of days a year in this calendar has.
	 */
	public void setDaysPerYear(final int daysPerYear) {
		this.daysPerYear = daysPerYear;
	}

	/** 
	 * @return <code>true</code> if this calender knows a year 0 (or <code>false</code> if year -1 follows year +1). 
	 */
	public boolean hasYear0() {
		return hasYear0;
	}

	/**
	 * Sets if this calender knows a year 0 (or if otherwise to year -1 follows
	 * year +1).
	 * 
	 * @param hasYear0
	 *            defines if this calender knows a year 0
	 */
	public void setHasYear0(final boolean hasYear0) {
		this.hasYear0 = hasYear0;
	}

	/** 
	 * @return the offset in days between BF calendar and this calendar. 
	 */
	public int getDaysFromYear0ToBF() {
		return daysFromYear0ToBF;
	}

	/**
	 * Sets the offset in days between BF calendar and this calendar.
	 * 
	 * @param daysFromYear0ToBF
	 *            the offset in days between BF calendar and this calendar.
	 */
	public void setDaysFromYear0ToBF(final int daysFromYear0ToBF) {
		this.daysFromYear0ToBF = daysFromYear0ToBF;
	}

	/**
	 * @return the days since Praios 1st, 0 BF.
	 * @see DSADate#getDaysSinceBF()
	 */
	public int getDaysSinceBF() {
		return date.getDaysSinceBF();
	}

	/**
	 * Sets the days since Praios 1st, 0 BF.
	 * 
	 * @param daysSinceBF
	 *            the days since Praios 1st, 0 BF.
	 * @see DSADate#getDaysSinceBF()
	 */
	public void setDaysSinceBF(final int daysSinceBF) {
		date.setDaysSinceBF(daysSinceBF);
	}

	/**
	 * @return the day of the year (julian day) for Bosparans fall calendar, ranges from 1 to {@link #DAYS_PER_YEAR_BF}.
	 */
	public int getJDayBF() {
		return MathUtil.modulo(getDaysSinceBF(), DAYS_PER_YEAR_BF) + 1;
	}

	/**
	 * @return the year of the Bosparans fall calendar.
	 */
	public int getYearBF() {
		return MathUtil.divisio(getDaysSinceBF(), DAYS_PER_YEAR_BF);
	}

	/**
	 * Sets the date according to Bosparans fall.
	 * 
	 * @param jday
	 *            The day of the year (julian day), ranges from 1 to
	 *            {@link #getDaysPerYear()}.
	 * @param year
	 *            The year (according to BF).
	 * @exception DSAException
	 *                may be thrown, depending on implementation of subclass.
	 *                The default implementation does not throw anything.
	 */
	public void setJDayYearBF(final int jday, final int year) throws DSAException {
//        if ((jday < 1) || (jday > daysPerYear())) {
//             throw new DSAException("jday out of range");
//        }
		setDaysSinceBF((year * DAYS_PER_YEAR_BF) + (jday - 1));
	}

	/**
	 * @return the days since the beginning of year 0 of this calendar.
	 */
	public int getDaysSinceYear0() {
		return getDaysSinceBF() - daysFromYear0ToBF;
	}

	/**
	 * Sets the days since the beginning of year 0 of this calendar.
	 * 
	 * @param daysSinceYear0
	 *            the days since the beginning of year 0 of this calendar.
	 */
	public void setDaysSinceYear0(final int daysSinceYear0) {
		setDaysSinceBF(daysSinceYear0 + daysFromYear0ToBF);
	}

	/**
	 * @return The day of the year (julian day) for this calendar, ranges from 1 to {@link #getDaysPerYear()}.
	 */
	public int getJDay() {
		return MathUtil.modulo(getDaysSinceYear0(), getDaysPerYear()) + 1;
	}

	/**
	 * @return The year of this calendar.
	 */
	public int getYear() {
		int year = MathUtil.divisio(getDaysSinceYear0(), getDaysPerYear());
		if (!hasYear0()) {
			if (0 >= year) { 
				// If there is no year zero, we start counting at year 1.
				year--;
			}
		}
		return year;
	}

	/**
	 * Sets the date.
	 * 
	 * @param jday
	 *            The day of the year (julian day), ranges from 1 to
	 *            {@link #getDaysPerYear()}.
	 * @param year
	 *            The year (according to BF).
	 * @exception DSAException
	 *                may be thrown, depending on implementation of subclass.
	 *                The default implementation does not throw anything.
	 */
	public void setJDayYear(final int jday, int year) throws DSAException {
//        if ((jday < 1) || (jday > daysPerYear())) {
//             throw new DSAException("jday out of range");
//        }
		if (!hasYear0()) {
			if (0 == year) {
				throw new DSADateYearZeroException(getName());
			} else if (0 > year) { // If there is no year zero, we start
									// counting at year 1.
				year++;
			}
		}
		final int daysSinceYear0 = (year * getDaysPerYear()) + (jday - 1);
		setDaysSinceYear0(daysSinceYear0);
	}

	public String getYearString(final int year, final String positiveUnitName, final String negativeUnitName) {
		final StringBuffer s = new StringBuffer();
		if (year < 0) {
			s.append(TextFormatter.format(-year, 5));
			s.append(' ');
			s.append(negativeUnitName);
		} else {
			s.append(TextFormatter.format(year, 5));
			s.append(' ');
			s.append(positiveUnitName);
		}
		return s.toString();
	}

	/**
	 * @return the day of the moon.
	 * @see DSADate#getMoonday()
	 */
	public int getMoonday() {
		return date.getMoonday();
	}

	/**
	 * @return the current season.
	 * @see DSADate#season()
	 */
	public SeasonType season() {
		return date.season();
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		return "Year " + getYear() + " JDay " + getJDay() + " BF";
	}

	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		final StringBuffer s = new StringBuffer();
		final TextFormatter tf = doc.tf();
		s.append(getHeadingText(doc));
		s.append(": ");
		s.append(doc.messages().getText("DSA.MoonDay"));
		s.append(": ");
		s.append(DSA.instance().getMoonPhaseString(getDSADate(), tf));
		return (s.toString());
	}

	/**
	 * Some basic information about a part of the date representation in this calendar.
	 *
	 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class FieldInfo {
		public String name;
		public int minimumValue;
		public int maximumValue;
	}

	@Override
	public String toString() {
		return "DSADateCalendar " + getDaysSinceBF() + " (Year " + getYearBF() + " JDay " + getJDayBF() + " BF)";
	}
}
