package de.ped.dsatool.logic;

import de.ped.tools.MathUtil;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;

/**
 * A class representing the calendar or days counting practice in Aventuria
 *  as used by Saurians ("Echsen" in German).
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Der Aventurische Kalender" (p. 60..63)</p>
 *
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSADateCalendarSaurians extends DSADateCalendar {
	/** 
	 * Creates a calendar showing a date representing Praios 1st, 0 BF.
	 * @see DSADateCalendar#DSADateCalendar() 
	 */
	public DSADateCalendarSaurians() {
		super();
		init();
	}

	/** 
	 * Creates a view following this calendar to the given date.
	 * 
	 * @param date
	 *            the DSA date to be represented using this calendar
	 * @see DSADateCalendar#DSADateCalendar(DSADate) 
	 */
	public DSADateCalendarSaurians(final DSADate date) {
		super(date);
		init();
	}

	protected void init() {
		setName("Echsisch");
		// Der 1. Praios 1000 BF entspricht dem 
		// 18. Tag des 219. Monats im Abschnitt des Drachen (also 5. Abschnitt), 
		// im 4. Ehhn des 4. Tsiina und ist ein Gzht'G. 
		setDaysFromYear0ToBF(1000 * DSADateCalendar.DAYS_PER_YEAR_BF - 
				((18 - 1) + DAYS_PER_MONTH * 
						((219 - 1) + MONTHS_PER_SECTION *
								((5 - 1) + SECTIONS_PER_EHHN * 
										((4 - 1) + EHHNS_PER_TSIINA *
												((4 - 1)))))));
		setHasYear0(true);
		setDaysPerYear(DAYS_PER_MONTH * MONTHS_PER_SECTION);
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		final int section = getSection();
		final int weekday = getWeekday();
		final StringBuffer s = new StringBuffer();
		final TextFormatter tf = doc.tf();
		// Der 1. Praios 1000 BF entspricht dem 
		// 18. Tag des 219. Monats im Abschnitt des Drachen (also 5. Abschnitt), 
		// im 4. Ehhn des 4. Tsiina und ist ein Gzht'G (3.) 
		s.append(tf.formatIfPlainText(weekdayNames[weekday], 6));
		s.append(" (");
		s.append(TextFormatter.format(weekday + 1, 1));
		s.append(".)");
		s.append(", ");
		s.append(TextFormatter.format(getDay() + 1, 2));
		s.append(". ");
		s.append("Tag des ");
		s.append(TextFormatter.format(getMonth() + 1, 3));
		s.append(". ");
		s.append("Monats im Abschnitt ");
		s.append(' ');
		s.append(tf.formatIfPlainText(sectionNames[section], 20));
		s.append(" (");
		s.append(TextFormatter.format(section + 1, 2));
		s.append(".) ");
		s.append("im");
		s.append(TextFormatter.format(getEhhn() + 1, 2));
		s.append(". ");
		s.append("Ehhn des ");
		s.append(TextFormatter.format(getTsiina() + 1, 5));
		s.append(". ");
		s.append("Tsiina");
		return (s.toString());
	}

	public String getYearString() {
		return getYearString(getYear(), "gS", "v.gS");
	}

	public static final int DAYS_PER_MONTH = 33;

	public static final int MONTHS_PER_SECTION = 553;

	public static final int SECTIONS_PER_EHHN = 10;

	public static final int EHHNS_PER_TSIINA = 33;

	public int getWeekday() {
		return MathUtil.modulo(getDaysSinceBF() + 2, DAYS_PER_WEEK);
	}

	public int getDay() {
		return getPart(1, DAYS_PER_MONTH);
	}

	public int getMonth() {
		return getPart(DAYS_PER_MONTH, MONTHS_PER_SECTION);
	}

	public int getSection() {
		return getPart(DAYS_PER_MONTH * MONTHS_PER_SECTION, SECTIONS_PER_EHHN);
	}

	public int getEhhn() {
		return getPart(DAYS_PER_MONTH * MONTHS_PER_SECTION * SECTIONS_PER_EHHN, EHHNS_PER_TSIINA);
	}

	public int getTsiina() {
		return getPart(DAYS_PER_MONTH * MONTHS_PER_SECTION * SECTIONS_PER_EHHN * EHHNS_PER_TSIINA, 0);
	}

	public int getPart(final int divisor, final int modulo) {
		if (0 == modulo) {
			return MathUtil.divisio(getDaysSinceYear0(), divisor);
		} else {
			return MathUtil.divisio(MathUtil.modulo(getDaysSinceYear0(), divisor * modulo), divisor);
		}
	}

	public static final String[] weekdayNames = new String[] {
			"Sz'G", 
			"Drs'G", 
			"Gzht'G", 
			"Lhn'G", 
			"Rsz'G",         
	};

	public static final int DAYS_PER_WEEK = weekdayNames.length;

	public static final String[] sectionNames = DSAEnv.CALENDAR_SAURION_SECTION_NAMES;
}
