package de.ped.dsatool.logic;

import de.ped.dsatool.dsa.generated.CalendarType;
import de.ped.dsatool.dsa.generated.DayOfWeekType;
import de.ped.dsatool.dsa.generated.MonthType;
import de.ped.tools.MathUtil;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;

/**
 * A class representing all calendars or days counting practices in Aventuria 
 *  which are based on the twelve gods.
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Der Aventurische Kalender" (p. 60..63)</p>
 *
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSADateCalendarTwelve extends DSADateCalendar {
	/** 
	 * Creates a calendar based on Bosparans fall, showing a date representing Praios 1st, 0 BF.
	 * @see DSADateCalendar#DSADateCalendar() 
	 */
	public DSADateCalendarTwelve() {
		super();
		init(DSACalendar.findDefault());
	}

	/** 
	 * Creates a view following Bosparans fall calendar to the given date.
	 * 
	 * @param date
	 *            the DSA date to be represented using this calendar
	 * @see DSADateCalendar#DSADateCalendar(DSADate) 
	 */
	public DSADateCalendarTwelve(final DSADate date) {
		super(date);
		init(DSACalendar.findDefault());
	}

	/**
	 * Creates a calendar based on the given calendarID, showing a date
	 * representing Praios 1st, 0 BF.
	 * 
	 * @param calendarID
	 *            the calendarID
	 * @see DSADateCalendar#DSADateCalendar()
	 */
	public DSADateCalendarTwelve(final String calendarID) {
		super();
		init(DSACalendar.findByID(calendarID));
	}

	/** 
	 * Creates a view following the given calendar to the given date.
	 * 
	 * @param date
	 *            the DSA date to be represented using this calendar
	 * @param calendarID
	 *            the calendarID
	 * @see DSADateCalendar#DSADateCalendar(DSADate) 
	 */
	public DSADateCalendarTwelve(final DSADate date, final String calendarID) {
		super(date);
		init(DSACalendar.findByID(calendarID));
	}

	protected void init(final CalendarType calendar) {
		this.calendar = calendar;
		setName(calendar.getName());
		setDaysFromYear0ToBF(-calendar.getYearZeroBFis() * DSADateCalendar.DAYS_PER_YEAR_BF);
		setHasYear0(calendar.isHasYearZero());
		setDaysPerYear(DSADateCalendar.DAYS_PER_YEAR_BF);
	}

	/**
	 * @return the calendar for the output of the date.
	 */
	public CalendarType getCalendar() {
		return calendar;
	}

	/**
	 * @param calendar
	 *            the calendar for the output of the date.
	 */
	public void setCalendar(CalendarType calendar) {
		init(calendar);
	}

	/**
	 * @return the calendar for the output of the date.
	 */
	public int getCalendarIDnum() {
		return calendar.getIDnum();
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		final StringBuffer s = new StringBuffer();
		final TextFormatter tf = doc.tf();
		s.append(TextFormatter.format(getDay(), 2));
		s.append('.');
		s.append(TextFormatter.format(getMonthAsNumber(), 2));
		s.append('.');
		s.append(" (");
		final MonthType month = getMonth();
		s.append(tf.formatIfPlainText(DSA.instance().getContentText(month, tf), 10));
		s.append('/');
		s.append(tf.formatIfPlainText(month.getNameInRealLife(), 10));
		s.append(") ");
		s.append(getYearString());
		return (s.toString());
	}

	/**
	 * Calculates the julian day from day and month (in DSA months)
	 * 
	 * @param day
	 *            The day of the month, ranges from 1 to 30.
	 * @param month
	 *            The month of the year, ranges from 1 to 13.
	 * @return The day of the year (julian day), ranges from 1 to 365.
	 */
	public int jday(final int day, final int month) {
//        if ((month < 1) || (month > 13) || (day < 1) || (day > 30)) {
//            throw new DSAException("day or month out of range");
//        }
		return ((month - 1) * 30 + day);
	}

	/**
	 * Sets the date.
	 * 
	 * @param day
	 *            The day of the month, ranges from 1 to 30.
	 * @param month
	 *            The month of the year, ranges from 1 to 13.
	 * @param year
	 *            The year (according to BF).
	 * @exception DSADateYearZeroException
	 *                year==0, but calendar does not support a year 0.
	 */
	   public void setDayMonthYear(final int day, final int month, final int year) throws DSAException {
		setJDayYear(jday(day, month), year);
	}

	/**
	 * @return The day of the month, ranges from 1 to 30.
	 */
	public int getDay() {
		return (MathUtil.modulo(getJDay() - 1, 30) + 1);
	}

	/**
	 * @return The month of the year.
	 */
	public MonthType getMonth() {
		return (MonthType) DSA.instance().getMonth().get(getMonthAsNumber() - 1);
	}

	/**
	 * @return The month of the year, ranges from 1 to 13.
	 */
	public int getMonthAsNumber() {
		return (((getJDay() - 1) / 30) + 1);
	}

	public String getYearString() {
		return getYearString(getYear(), calendar.getPositiveUnitName(), calendar.getNegativeUnitName());
	}

	/** The calendar the date is displayed in. */
	private CalendarType calendar;

	/**
	 * @return The day of the week.
	 */
	public DayOfWeekType getDayOfWeek() {
		// The 1st of Praios 0 BF was a Praiostag.
		return DayOfWeek.findByIDnum(MathUtil.modulo((getDaysSinceBF() + 6), 7));
	}
}
