package de.ped.dsatool.logic;

import java.util.Observable;

import de.ped.tools.Assert;

/**
 * Bundles all information related to calendar dates.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSADateComplete extends Observable implements Cloneable {
	/**
	 * Constructor.
	 * 
	 * @param weather
	 *            The weather (including the date).
	 * @param randomDayEvent
	 *            The random day event of that date.
	 */
	public DSADateComplete(final Weather weather, final RandomDayEvent randomDayEvent) {
		this.weather  = weather;
		this.dayEvent = randomDayEvent;
	}

	/**
	 * Constructor. The random day event will be set randomly.
	 * 
	 * @param weather
	 *            The weather (including the date).
	 */
	public DSADateComplete(final Weather weather) {
		this.weather  = weather;
		this.dayEvent = new RandomDayEvent();
	}

	/**
	 * Constructor. The weather for that date will be set randomly.
	 * 
	 * @param date
	 *            the DSA date to be represented using this calendar
	 * @param randomDayEvent
	 *            The random day event of that date.
	 */
	public DSADateComplete(final DSADate date, final RandomDayEvent randomDayEvent) {
		this.weather  = new Weather(date);
		this.dayEvent = randomDayEvent;
	}

	/**
	 * Constructor. The weather and the random day event for that date will be
	 * set randomly.
	 * 
	 * @param date
	 *            the DSA date to be represented using this calendar
	 */
	public DSADateComplete(final DSADate date) {
		this.weather  = new Weather(date);
		this.dayEvent = new RandomDayEvent();
	}

	/**
	 * Default constructor.
	 * The date will be the one of the default constructor of DSADate.
	 * The weather and the random day event for that date will be set randomly.
	 * @see DSADate#DSADate
	 */
	public DSADateComplete() {
		this.weather  = new Weather(new DSADate());
		this.dayEvent = new RandomDayEvent();
	}

	/**
	 * @return the weather for this day
	 */
	public Weather getWeather() {
		return weather;
	}

	/**
	 * @return the random day event for this day
	 */
	public RandomDayEvent getRandomDayEvent() {
		return dayEvent;
	}

	/**
	 * @return the DSA date
	 */
	public DSADate getDate() {
		return weather.getDate();
	}

	/**
	 * Sets the weather
	 * @param weather the weather
	 */
	public void setWeather(final Weather weather) {
		this.weather = weather;
		super.setChanged();
	}

	/**
	 * @return The following day of this day.
	 */
	public DSADateComplete nextDay() {
		return new DSADateComplete(weather.createNextDay());
	}
	
	private Weather			weather;
	private RandomDayEvent	dayEvent;

	@Override
	public DSADateComplete clone() {
		DSADateComplete ret = null;
		try {
			ret = (DSADateComplete)super.clone();
			ret.weather  = (Weather)weather.clone();
			ret.dayEvent = (RandomDayEvent)dayEvent.clone();
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}
}
