package de.ped.dsatool.logic;

import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;

/**
 * Creates the HTML help text for the DSA tool.
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSAToolHelpTextBuilder {
	private static final int FUNCTION = 3;
	private static final int ACTION = 4;

	/**
	 * Builds the full help text into the text document builder
	 * @param doc the text document builder to build the help text into
	 */
	public void buildHelpText(final TextDocumentBuilder doc) {
		final Messages messages = doc.messages();
		final String title = messages.getApplicationName(Messages.I18NedOrNot.I18N) + " " + messages.getText("Help");
		doc.setTitle(title);
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");
		buildHelpTextContent(doc);
		doc.addContentPart(2, "Help.Finish");
		doc.endDocument();
	}

	/**
	 * Builds the specific help text into the text document builder.
	 * Default implementation builds for the web interface; override for more specific implementations.
	 * @param doc the text document builder to build the help text into
	 */
	protected void buildHelpTextContent(final TextDocumentBuilder doc) {
		doc.addContentPart(2, "Help.Menu");
		addSection(doc, FUNCTION, "Action.ViewDate");
		addSection(doc, ACTION, "Action.CreateDate");
		addUIElements(doc, new String[] { "CalendarDialog.Weather.Random", "CalendarDialog.Weather.Own" });
		addSection(doc, ACTION, "Action.CreateDateUndo");
		addSection(doc, ACTION, "Action.CreateDateNext");
		addSection(doc, FUNCTION, "Action.ViewTown");
		addSection(doc, ACTION, "Action.CreateTown");
		addSection(doc, ACTION, "Action.CreateMap");
		addSection(doc, FUNCTION, "Action.ViewNames");
		addSection(doc, ACTION, "Action.CreateNames");
		addSection(doc, FUNCTION, "Action.ViewPlants");
		addSection(doc, ACTION, "Action.CreatePlants");
		addUIElements(doc, new String[] { "PlantsDialog.IsToFilterFor.TerrainProb", "PlantsDialog.IsToFilterFor.Date",
				"PlantsDialog.IsToFilterFor.Botany" });
		addSection(doc, FUNCTION, "Action.ViewQuest");
		addSection(doc, ACTION, "Action.CreateQuest");
		addSection(doc, FUNCTION, "Action.ViewDice");
		addSection(doc, ACTION, "Action.CreateDice");
		addSection(doc, FUNCTION, "Action.ViewMap");
		addSection(doc, FUNCTION, "Menu.MasterData");
		final int max = MasterDataModel.MasterDataType.values().length - 1;
		for (int i = 0; i <= max; i++) {
			addSection(doc, ACTION, MasterDataModel.getMessageKey(i));
		}
		closeActionsDd(doc);
		closeFunctionsDl(doc);
	}

	/** 
	 * Formats a DSA tool function heading or action heading.
	 * @param doc the document to work with
	 * @param type either {@link #FUNCTION} or {@link #ACTION}.
	 * @param key the message key of the action or function
	 */
	protected void addSection(final TextDocumentBuilder doc, final int type, final String key) {
		final Messages messages = doc.messages();
		switch(type) {
		case FUNCTION: {
			final TextFormatter tf = doc.tf();
			closeActionsDd(doc);
			closeFunctionsDl(doc);
			doc.addSectionHeading(type, key, messages.getText(key));
			doc.add(tf.ppBeginLine());
			doc.add(doc.convertToTextDocumentType(messages.getLongDescription(key)));
			doc.add(tf.ppEndLine());
			break;
		}
		case ACTION: {
			closeActionsDd(doc);
			if (!isFunctionsDlOpen) {
				isFunctionsDlOpen = true;
				doc.beginDL(null);
			}
			doc.addDT(messages.getText(key));
			doc.beginDD();
			isActionsDdOpen = true;
			doc.add(messages.getShortDescription(key));
			break;
		}
		default:
			// Nothing to do
		}
	}

	protected void closeFunctionsDl(final TextDocumentBuilder doc) {
		if (isFunctionsDlOpen) {
			isFunctionsDlOpen = false;
			doc.endDL();
		}
	}

	protected void closeActionsDd(final TextDocumentBuilder doc) {
		if (isActionsDdOpen) {
			isActionsDdOpen = false;
			doc.endDD();
		}
	}

	protected void addUIElements(final TextDocumentBuilder doc, final String[] keys) {
		if ((null != keys) && (0 < keys.length)) {
			final Messages messages = doc.messages();
			doc.beginDL(null);
			for (String key : keys) {
				doc.addDT(messages.getText(key));
				doc.beginDD();
				doc.add(messages.getShortDescription(key));
				doc.endDD();
			}
			doc.endDL();
		}
	}

	private boolean isFunctionsDlOpen = false;
	private boolean isActionsDdOpen = false;
}
