package de.ped.dsatool.logic;

import java.util.Date;
import java.util.Iterator;
import java.util.List;

import de.ped.tools.Assert;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.Textable;
import de.ped.dsatool.dsa.generated.DayOfWeekType;
import de.ped.dsatool.dsa.generated.HolidayType;
import de.ped.dsatool.dsa.generated.MonthType;

/**
 * A model to keep the actual calendar time frame.
 * It always calculates one more day than NUM_OF_DAYS,
 * so that this extra day can be the start point for the following time frame.
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DateModel implements Cloneable, Textable {
	/** How many days has a time frame? */
	public static final int NUM_OF_DAYS = 14;

	public DateModel() {
		final Date today = new Date();
		final DSADate date = new DSADate(today);
		days[0] = new DSADateComplete(date);
		recalculate();
	}

	/**
	 * Recalculates the same time frame. The start day is unaffected,
	 * but for all following days the weather and all other random effects are re-diced.
	 */
	public void recalculate() {
		for (int i = 0; i < NUM_OF_DAYS; i++) {
			days[i + 1] = days[i].nextDay();
		}
	}

	/**
	 * Takes the last day of the actual timeframe as start point,
	 * and calculates the days of the following timeframe.
	 */
	public void calcNextTimeframe() {
		days[0] = days[NUM_OF_DAYS];
		recalculate();
	}

	/**
	 * Sets the first day of the time frame. The rest of the timeframe is
	 * re-diced.
	 * 
	 * @param date
	 *            the first day
	 */
	public void setFirstDay(final DSADateComplete date) {
		days[0] = date;
		recalculate();
	}

	/** 
	 * @return the first day of the time frame
	 */
	public DSADateComplete getFirstDay() {
		return days[0];
	}

	/** 
	 * @return the last day of the time frame - the seed for a following time frame 
	 */
	public DSADateComplete getLastDay() {
		return days[NUM_OF_DAYS];
	}

	/**
	 * @param i the index
	 * @return the day of index <em>i</em> of the actual time frame. 
	 */
	public DSADateComplete getDay(final int i) {
		return days[i];
	}

	private DSADateCalendarTwelve cal12;

	public DSADateCalendarTwelve getDialogCalendar() {
		if (null == cal12) {
			cal12 = new DSADateCalendarTwelve();
		}
		cal12.setCalendar(days[0].getWeather().getCalendar());
		cal12.setDSADate(days[0].getWeather().getDate());
		return cal12;
	}

	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		final TextFormatter tf = doc.tf();
		final StringBuffer ret = new StringBuffer();
		final Messages messages = doc.messages();
		final DSA dsa = DSA.instance();
		switch (tf.getLocalDocType()) {
		case FIXEDFONT:
			for (int i = 0; i < NUM_OF_DAYS + 1; i++) {
				ret.append(dsa.getContentText(days[i], doc));
			}
			break;
		case XHTML10_STRICT:
		case HTML5:
		default:
			final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[] {
					new TextFormatter.ColumnInfo(3, TextFormatter.RIGHT, messages, "DSA.DayOfMonth"),
					new TextFormatter.ColumnInfo(20, TextFormatter.LEFT, messages, "DSA.Month"),
					new TextFormatter.ColumnInfo(20, TextFormatter.LEFT, messages, "DSA.Year"),
					new TextFormatter.ColumnInfo(20, TextFormatter.LEFT, messages, "DSA.DayOfWeek"), };
			final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
			ret.append(tf.beginTable(null, ti));
			ret.append(tf.formatTableHeading(ci));
			ret.append(tf.beginTableBody());
			for (int i = 0; i < NUM_OF_DAYS + 1; i++) {
				final DSADateComplete day = days[i];
				final DSADateCalendarTwelve sday = new DSADateCalendarTwelve(day.getDate());
				final MonthType month = sday.getMonth();
				final DayOfWeekType dayOfWeek = sday.getDayOfWeek();
				final Weather weather = day.getWeather();
//				ret.append(dsa.getContentText(days[i], tf));
				final I18NStringWithFillIns i18nMonth = new I18NStringWithFillIns("DateModel.Month", "%0.", new String[] { 
						TextFormatter.format(sday.getMonthAsNumber(), 2).toString(),
						tf.href(MasterDataModel.getHRefFor(month), dsa.getContentText(month, tf)),
						month.getNameInRealLife(),
				});

				ret.append(tf.beginTableRow());
				ret.append(tf.formatTableCell(ci[0], tf.beginBold() + Integer.toString(sday.getDay()) + "." + tf.endBold()));
				ret.append(tf.formatTableCell(ci[1], tf.beginBold() + messages.getText(i18nMonth) + tf.endBold()));
				ret.append(tf.formatTableCell(ci[2], tf.beginBold() + sday.getYearString() + tf.endBold()));
				ret.append(tf.formatTableCell(ci[3], tf.beginBold() + tf.href(MasterDataModel.getHRefFor(dayOfWeek), dsa.getContentText(dayOfWeek, tf)) + tf.endBold()));
				ret.append(tf.endTableRow());
				// Show weather, all holidays and random events for the current date
				StringBuffer s = new StringBuffer();
				s.append(messages.getText(new I18NStringWithFillIns("DSA.MoonDay.All", "%0: %1", new Object[] { 
						new I18NStringWithFillIns("DSA.MoonDay", null, null), 
						dsa.getMoonPhaseString(sday.getDSADate(), tf), }
				)));
				s.append(tf.newLine());
				s.append(dsa.getContentText(weather, tf));
				s.append(tf.newLine());
				final Iterator<HolidayType> itHolidays = Holiday.findHolidaysAt(sday.getDSADate()).iterator();
				while (itHolidays.hasNext()) {
					final HolidayType holiday = (HolidayType) itHolidays.next();
					s.append(dsa.getContentText(holiday, tf));
					s.append(tf.newLine());
				}
				s.append(dsa.getContentText(day.getRandomDayEvent(), tf));
				s.append(tf.newLine());
				s.append("&nbsp;");
				ret.append(tf.formatTableFullTextRow(s.toString(), null));
			}
			ret.append(tf.endTableBody());
			ret.append(tf.endTable());
		}
		final DSADate d = (DSADate) days[0].getDate().clone();
		final int depth = 2 + 1;
		ret.append(tf.addSectionHeading(depth, messages.getText("DateModel.Calendars.Heading")));
		final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[] {
				new TextFormatter.ColumnInfo(38, TextFormatter.LEFT, messages, "DSA.Calendar"),
				new TextFormatter.ColumnInfo(107, TextFormatter.LEFT, messages, "DateModel.DateInCalendar.Heading"), };
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
		ret.append(tf.beginTable("table0", ti));
		ret.append(tf.formatTableHeading(ci));
		ret.append(tf.beginTableBody());
		final List<DSADateCalendar> cals = DSACalendar.createAllCalendarsFor(d);
		for (final DSADateCalendar cal : cals) {
			ret.append(tf.beginTableRow());
			String calName = cal.getName();
			if (cal instanceof DSADateCalendarTwelve) {
				DSADateCalendarTwelve calt = (DSADateCalendarTwelve) cal;
				calName = tf.href(MasterDataModel.getHRefFor(calt.getCalendar()), calName);
			}
			ret.append(tf.formatTableCell(ci[0], calName));
			ret.append(tf.formatTableCell(ci[1], cal.getHeadingText(doc)));
			ret.append(tf.endTableRow());
		}
		ret.append(tf.endTableBody());
		ret.append(tf.endTable());
		return ret.toString();
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		final DSADateCalendarTwelve d = new DSADateCalendarTwelve(days[0].getDate());
		return doc.messages().getText(new I18NStringWithFillIns("DateModel.Intro.Heading", null,
				new String[] { DSA.instance().getHeadingText(d, doc), }));
	}

	@Override
	public DateModel clone() {
		DateModel ret = null;
		try {
			ret = (DateModel)super.clone();
			for (int i = 0; i < NUM_OF_DAYS + 1; i++) {
				ret.days[i] = (DSADateComplete)days[i].clone();
			}
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}

	public String[] getWeatherTable() {
		return getFirstDay().getWeather().getWeatherNames();
	}

	/** The days created. */
	private DSADateComplete	days[] = new DSADateComplete[NUM_OF_DAYS + 1];
}
