package de.ped.dsatool.logic;

import de.ped.dsatool.dsa.generated.DiceRollType;
import de.ped.tools.MathUtil;

/**
 * Contains regulatories concerning the outcome of rolling dices.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Dice {
	/** Description for probability on W20 - in German DSA documentation "sehr häufig". */
	public static final int VERY_OFTEN = 16;
	/** Description for probability on W20 - in German DSA documentation "häufig". */
	public static final int OFTEN = 8;
	/** Description for probability on W20 - in German DSA documentation "gelegentlich". */
	public static final int OCCASIONLY = 4;
	/** Description for probability on W20 - in German DSA documentation "selten". */
	public static final int RARE = 2;
	/** Description for probability on W20 - in German DSA documentation "sehr selten". */
	public static final int VERY_RARE = 1;

	/**
	 * Simulates the behavior of n dices, with m sides each. So in DSA notation,
	 * roll(n, m, offset, factor) results in factor * (nWm + offset).
	 * 
	 * @param dr
	 *            the structure containing the number of dices, the eyes per
	 *            dice, and the offset.
	 * @return the sum of eyes thrown + offset, multiplied with factor.
	 */
	public static int roll(final DiceRollType dr) {
		return roll(dr.getNumber(), dr.getEyes(), dr.getOffset(), dr.getFactor());
	}

	/**
	 * Simulates the behavior of n dices, with m sides each. So in DSA notation,
	 * roll(n, m, offset, factor) results in factor * (nWm + offset).
	 * 
	 * @param n
	 *            so many dices to throw
	 * @param m
	 *            so many sides has each dice
	 * @param offset
	 *            this offset is added to the result.
	 * @param factor
	 *            this factor the result is multiplied with.
	 * @return the sum of eyes thrown + offset, multiplied with factor.
	 */
	public static int roll(final int n, final int m, final int offset, final int factor) {
		return factor * roll(n, m, offset);
	}

	/**
	 * Simulates the behavior of n dices, with m sides each. So in DSA notation,
	 * roll(n, m, offset, factor) results in factor * (nWm + offset).
	 * 
	 * @param eyes
	 *            the eyes already thrown for the dices
	 * @param sgn
	 *            the sign to apply to the sum of eyes (e.g., because the number of dices was negative)
	 * @param offset
	 *            this offset is added to the result.
	 * @param factor
	 *            this factor the result is multiplied with.
	 * @return the sum of eyes thrown + offset, multiplied with factor.
	 */
	public static int roll(final int[] eyes, final int sgn, final int offset, final int factor) {
		int sum = 0;
		for (int i = 0; i < eyes.length; i++) {
			sum += eyes[i];
		}
		return factor * (sum * sgn + offset);
	}

	/**
	 * Simulates the behavior of n dices, with m sides each. So in DSA notation,
	 * roll(n, m, offset) rolls nWm+offset.
	 * 
	 * @param n
	 *            so many dices to throw
	 * @param m
	 *            so many sides has each dice
	 * @param offset
	 *            this offset is added to the result.
	 * @return the sum of eyes thrown + offset
	 */
	public static int roll(final int n, final int m, final int offset) {
		return roll(n, m) + offset;
	}

	/**
	 * Simulates the behavior of n dices, with m sides each. So in DSA notation,
	 * roll(n, m) rolls nWm.
	 * 
	 * @param n
	 *            so many dices to throw
	 * @param m
	 *            so many sides has each dice
	 * @return the sum of eyes thrown
	 */
	public static int roll(final int n, final int m) {
		if ((0 == n) || (0 == m)) {
			return 0;
		}
		if (n < 0) {
			return -roll(-n,m);
		}
		if (m < 0) {
			return -roll(n,-m);
		}
		int sum = 0;
		for (int i = 0; i < n; ++i) {
			sum += (MathUtil.random(m) + 1);
		}
		return sum;
	}

	/**
	 * Simulates the behavior of n dices, with 6 sides each. So in DSA notation,
	 * roll(n) rolls nW6.
	 * 
	 * @param n
	 *            so many dices to throw
	 * @return the sum of eyes thrown
	 */
	public static int roll(final int n) {
		return roll(n, 6);
	}

	/**
	 * Simulates the behavior of 1 dice with 6 sides. So in DSA notation, roll()
	 * rolls 1W6.
	 * 
	 * @return the eyes thrown
	 */
	public static int roll() {
		return roll(1);
	}

	/**
	 * Checks the thrown W20 dices for very lucky or very unlucky results.
	 * 
	 * @param dices
	 *            The W20 dices eyes thrown.
	 * @exception DSATestVeryLuckyException
	 *                if it was a very lucky test. 
	 *                <ul>
	 *                <li>In a 1W20 test this happens if the dice shows a 1.</li>
	 *                <li>In a 3W20 test this happens if two or more dices show a 1.</li>
	 *                <li>In a 3W20 test, if all three dices show a 1, a DSATestExtremlyLuckyException (a
	 *                subclass) is thrown.</li>
	 *                </ul>
	 * @exception DSATestVeryUnluckyException
	 *                if it was a very unlucky test. 
	 *                <ul>
	 *                <li>In a 1W20 test this happens if the dice shows a 20.</li>
	 *                <li>In a 3W20 test this happens if two or more dices show a 20.</li>
	 *                <li>In a 3W20 test, if all three dices show a 20, a DSATestExtremlyLuckyException
	 *                (a subclass) is thrown.</li>
	 *                </ul>
	 */
	public static void checkForExceptions(final int[] dices) throws DSATestVeryLuckyException, DSATestVeryUnluckyException {
		int count01 = 0;
		int count20 = 0;
		for (int i = 0; i < dices.length; i++) {
			if (1 == dices[i]) {
				count01++;
			}
			if (20 == dices[i]) {
				count20++;
			}
		}
		// Check for very lucky or unlucky tests:
		switch(dices.length) {
		case 1:
			if (1 == count01) {
				throw(new DSATestVeryLuckyException());
			}
			if (1 == count20) {
				throw(new DSATestVeryUnluckyException());
			}
			break;
		case 3:
			switch(count01) {
			case 3:
				throw(new DSATestExtremlyLuckyException());
			case 2:
				throw(new DSATestVeryLuckyException());
			default:
				// Nothing to do
			}
			switch(count20) {
			case 3:
				throw(new DSATestExtremlyUnluckyException());
			case 2:
				throw(new DSATestVeryUnluckyException());
			default:
				// Nothing to do
			}
		default:
			// No rules defined, so ignore very lucky or unlucky results
		}
	}

	/**
	 * Checks a roll of dices against hero properties, like all talent or spell
	 * tests (3W20) or tests against attack, parade, positive or negative hero
	 * properties (1W20).
	 * 
	 * @param dices
	 *            The W20 dices eyes thrown.
	 * @param properties
	 *            An array of hero properties. Contains all the properties that
	 *            are involved in the test. 
	 *            <ul>
	 *            <li>For a 3W20 test this will consist of 3 positive properties.</li>
	 *            <li>For a 1W20 test this will consist of the single property involved (e.g., parade value).</li>
	 *            </ul>
	 *            But theoretically this method is not limited to 3 or 1 dices but
	 *            any positive number of dices.
	 * @param talent
	 *            The talent value of the hero for this test
	 * @param modifier
	 *            Any modifications to the test to make it more easy or
	 *            difficult. Positive numbers will make it more difficult for
	 *            the hero to succeed.
	 * @return if &gt;= 0, the test succeeded. The value says how well it
	 *         succeeded. If the value is below 0, the test failed. The value
	 *         says how bad it failed.
	 */
	public static int testForProperties(final int[] dices, final int[] properties, final int talent, final int modifier) {
		// Do I have to check for good or bad dice eyes?
		final int talentAndModifier = talent - modifier;
		boolean allDicesSmaller = true;
		for (int i = 0; i < dices.length; i++) {
			if (dices[i] > properties[i]) {
				allDicesSmaller = false;
			}
		}
		int result = talentAndModifier;
		if (allDicesSmaller) {
			for (int i = 0; i < dices.length; i++) {
				result += (properties[i] - dices[i]);
			}
		} else {
			for (int i = 0; i < dices.length; i++) {
				if (dices[i] > properties[i]) {
					result += (properties[i] - dices[i]);
				}
			}
		}
		return result;
	}

	/**
	 * The DSA rules often need a roll of dices against hero properties, like
	 * all talent or spell tests (3W20) or tests against attack, parade,
	 * positive or negative hero properties (1W20). This method throws the same
	 * amount of dices as <em>properties</em> has elements, and analyzes the
	 * roll.
	 * 
	 * @param properties
	 *            An array of hero properties. Contains all the properties that
	 *            are involved in the test. 
	 *            <ul>
	 *            <li>For a 3W20 test this will consist of 3 positive properties.</li>
	 *            <li>For a 1W20 test this will consist of the single property involved (e.g., parade value).</li>
	 *            </ul>
	 *            But theoretically this method is not limited to 3 or 1 dices but
	 *            any positive number of dices.
	 * @param talent
	 *            The talent value of the hero for this test
	 * @param modifier
	 *            Any modifications to the test to make it more easy or
	 *            difficult. Positive numbers will make it more difficult for
	 *            the hero to succeed.
	 * @return if &gt;= 0, the test succeeded. The value says how well it
	 *         succeeded. If the value is below 0, the test failed. The value
	 *         says how bad it failed.
	 * @exception DSATestVeryLuckyException
	 *                if it was a very lucky test. 
	 *                <ul>
	 *                <li>In a 1W20 test this happens if the dice shows a 1.</li>
	 *                <li>In a 3W20 test this happens if two or more dices show a 1.</li>
	 *                <li>In a 3W20 test, if all three dices show a 1, a DSATestExtremlyLuckyException (a
	 *                subclass) is thrown.</li>
	 *                </ul>
	 * @exception DSATestVeryUnluckyException
	 *                if it was a very unlucky test. 
	 *                <ul>
	 *                <li>In a 1W20 test this happens if the dice shows a 20.</li>
	 *                <li>In a 3W20 test this happens if two or more dices show a 20.</li>
	 *                <li>In a 3W20 test, if all three dices show a 20, a DSATestExtremlyLuckyException
	 *                (a subclass) is thrown.</li>
	 *                </ul>
	 * @see #checkForExceptions(int[])
	 * @see #testForProperties(int[], int[], int, int)
	 */
	public static int rollForProperties(final int[] properties, final int talent, final int modifier) throws DSATestVeryLuckyException, DSATestVeryUnluckyException {
		// First throw 3W20 and memorize them
		final int[] dices = new int[properties.length];
		for (int i = 0; i < dices.length; i++) {
			dices[i] = roll(1, 20);
		}
		checkForExceptions(dices);
		return testForProperties(dices, properties, talent, modifier);
	}
}
