package de.ped.dsatool.logic;

import de.ped.dsatool.dsa.generated.DiceRollType;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.Textable;

/**
 * Contains the model of the outcome of rolling dices.
 *
 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DiceModel implements Textable {
	/**
	 * Constructor for a dice roll of 1W6 with lots of additional information about the dice roll.
	 */
	public DiceModel() {
		final DiceRollType input = new DiceRollType();
		setInput(input);
	}

	/**
	 * Constructor for a dice roll with lots of additional information about the dice roll.
	 * @param input the dice roll to execute
	 */
	public DiceModel(final DiceRollType input) {
		setInput(input);
	}

	@Override
	public String getContentText(TextDocumentBuilder doc) {
		final StringBuffer s = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final Messages messages = doc.messages();
		final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[] {
				new TextFormatter.ColumnInfo(3, TextFormatter.RIGHT, messages, "Dice.Min.Heading"),
				new TextFormatter.ColumnInfo(5, TextFormatter.RIGHT, messages, "Dice.Avg.Heading"), 
				new TextFormatter.ColumnInfo(3, TextFormatter.RIGHT, messages, "Dice.Max.Heading"),
				new TextFormatter.ColumnInfo(3, TextFormatter.RIGHT, messages, "Dice.Result.Heading"),
				new TextFormatter.ColumnInfo(30, TextFormatter.LEFT, messages, "Dice.Details.Heading"), };
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
		s.append(tf.beginTable(null, ti));
		s.append(tf.formatTableHeading(ci));
		s.append(tf.beginTableBody());
		s.append(tf.beginTableRow());
		s.append(tf.formatTableCell(ci[0], Integer.toString(getMin())));
		s.append(tf.formatTableCell(ci[1], Float.toString(getAvg())));
		s.append(tf.formatTableCell(ci[2], Integer.toString(getMax())));
		s.append(tf.formatTableCell(ci[3], tf.beginBold() + Integer.toString(result) + tf.endBold()));
		s.append(tf.formatTableCell(ci[4], getDiceResultsString()));
		s.append(tf.endTableRow());
		s.append(tf.endTableBody());
		s.append(tf.endTable());
		return s.toString();
	}

	@Override
	public String getHeadingText(TextDocumentBuilder doc) {
		return toString();
	}

	/** @return the dice roll that was thrown. */
	public DiceRollType getDiceRollType() {
		return input;
	}

	/** @return the minimum value for that dice roll. */
	public int getMin() {
		return getMinOrMax(true);
	}

	/** @return the maximum value for that dice roll. */
	public int getMax() {
		return getMinOrMax(false);
	}

	public int getMinOrMax(final boolean isMin) {
		final int off = MathUtil.sgn(input.getEyes());
		final int v1 = input.getFactor() * (input.getNumber() * off + input.getOffset());
		final int v2 = input.getFactor() * (input.getNumber() * input.getEyes() + input.getOffset());
		final int ret;
		if (isMin) {
			ret = Math.min(v1, v2);
		} else {
			ret = Math.max(v1, v2);
		}
		return ret;
	}

	/** @return average or expectation value for that dice roll. */
	public float getAvg() {
		final int off = MathUtil.sgn(input.getEyes());
		return input.getFactor() * (input.getNumber() * (float)(off + input.getEyes()) / 2.0f + input.getOffset());
	}

	/** @return string representing the formula for this specific dice roll. */
	@Override
	public String toString() {
		String str = Integer.toString(input.getNumber()) + "W" + Integer.toString(input.getEyes());
		if (0 != input.getOffset()) {
			str += (input.getOffset() < 0 ? "" : "+") + Integer.toString(input.getOffset());
		}
		if (1 != input.getFactor()) {
			str = Integer.toString(input.getFactor()) 
					+ " x (" 
					+ str
					+ ")";
		}
		return str;
	}

	/** @return a string representation for the dice results */
	public String getDiceResultsString() {
		final StringBuffer buf = new StringBuffer();
		buf.append("(");
		for (int i = 0; i < diceResults.length; i++) {
			if (0 < i) {
				buf.append(",");
			}
			buf.append(diceResults[i]);
		}
		buf.append(")");
		if (20 == input.getEyes()) {
			try {
				Dice.checkForExceptions(diceResults);
			} catch(final DSATestVeryLuckyException ex) {
				buf.append(" " + ex.getMessage());
			} catch(final DSATestVeryUnluckyException ex) {
				buf.append(" " + ex.getMessage());
			}
		}
		return buf.toString();
	}

	/** @return dice roll result that was thrown. */
	public int getResult() {
		return result;
	}

	public void setInput(final DiceRollType input) {
		this.input = input;
		reroll();
	}

	public void reroll() {
		final int normedNumber = Math.abs(input.getNumber());
		if ((null == diceResults) || (diceResults.length != normedNumber)) {
			diceResults = new int[normedNumber];
		}
		for (int i = 0; i < diceResults.length; i++) {
			diceResults[i] = Dice.roll(1, input.getEyes());
		}
		this.result = Dice.roll(diceResults, MathUtil.sgn(input.getNumber()), input.getOffset(), input.getFactor());
	}

	/** The dice roll that was thrown. */
	private DiceRollType input;
	private int[] diceResults;
	private int result;
}
