package de.ped.dsatool.logic;

import java.io.InputStream;
import java.io.StringWriter;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import de.ped.dsatool.dsa.generated.DsaType;
import de.ped.dsatool.dsa.generated.LanguageDistributionType;
import de.ped.dsatool.dsa.generated.LanguageProbabilityType;
import de.ped.dsatool.dsa.generated.ObjectFactory;
import de.ped.dsatool.dsa.generated.PersonType;
import de.ped.dsatool.dsa.generated.ShopType;
import de.ped.dsatool.dsa.generated.TempleType;
import de.ped.dsatool.dsa.generated.TownType;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A helper class necessary to run without all the JAXB classes.
 * Contains all references to JAXB classes and abstracts from them,
 * where no other interface or abstract class mechanism is available. 
 * Any method using a JAXB class or throwing a JAXB exception is wrapped herein. 
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
class JAXBWrapper {
	private Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
	}

	/** The JAXB generated object factory. If <code>null</code>, not using JAXB but stand-alone implementation. */
	protected ObjectFactory objectFactory;

	public static final String JAXB_PATHNAME = "de.ped.dsatool.dsa.generated";

	public String marshalObject(final Object obj) {
		final StringWriter w = new StringWriter();
		try {
			final JAXBContext jc = JAXBContext.newInstance(JAXB_PATHNAME);
			final Marshaller m = jc.createMarshaller();
			m.marshal(obj, w);
		} catch (final JAXBException je) {
			logger.info(je.toString());
			w.write(je.toString());
			je.printStackTrace();
		} catch (final NoClassDefFoundError ncdfe) {
			logger.info(ncdfe.getClass().getName() + ": " + ncdfe.getMessage());
			w.write(ncdfe.toString());
			ncdfe.printStackTrace();
		}
		return w.toString();
	}

	/**
	 * Loads the whole XML data file. This should be the first method called in
	 * the wrapper. If loading the data was fine, the state of the wrapper is
	 * set to active.
	 * 
	 * @param dataXmlStream
	 *            the stream to read the XML data file from
	 * @return the data structure, or <code>null</code> if that failed.
	 * @see #isActive()
	 */
	public DsaType loadDsaType(final InputStream dataXmlStream) {
		DsaType ret = null;
		try {
			final JAXBContext jc = JAXBContext.newInstance(JAXB_PATHNAME);
			final Unmarshaller u = jc.createUnmarshaller();
			final Object o = u.unmarshal(dataXmlStream);
			logger.trace("Unmarshalled " + o.getClass() + ": " + o);
			@SuppressWarnings("rawtypes")
			final JAXBElement e = (JAXBElement)o;
			logger.debug("Unmarshalled name " + e.getName().getClass() + ": " + e.getName());
			final Object v = e.getValue();
			logger.debug("Unmarshalled value " + v.getClass() + ": " + v);
			logger.trace("Unmarshalled class " + e.getClass().getClass() + ":" + e.getClass());
			ret = (DsaType) v;
			objectFactory = new ObjectFactory();
			// Test code: Marshall back to somewhere. Only works for XML root elements.
			// StringWriter w = new StringWriter();
			// Marshaller m = jc.createMarshaller();
			// m.marshal(ret, w);
			// logger.debug("Marshalled to " + w.toString());
		} catch (final JAXBException je) {
			logger.info(je.getClass().getName() + ": " + je.getMessage());
			je.printStackTrace();
		} catch (final NoClassDefFoundError ncdfe) {
			logger.info(ncdfe.getClass().getName() + ": " + ncdfe.getMessage());
			ncdfe.printStackTrace();
		}
		return ret;
	}

	/**
	 * @return if loading the XML data was successful.
	 */
	public boolean isActive() {
		return (null != objectFactory);
	}

	/**
	 * @return a newly created JAXB object, or <code>null</code> if that failed.
	 */
	public LanguageDistributionType createLanguageDistributionType() {
		try {
			final LanguageDistributionType ld = objectFactory.createLanguageDistributionType();
			logger.trace("Creating languageDistribution " + ld);
			return ld;
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public LanguageProbabilityType createLanguageProbabilityType() {
		try {
			final LanguageProbabilityType lp = objectFactory.createLanguageProbabilityType();
			logger.trace("Creating languageProbability " + lp);
			return lp;
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public PersonType createPersonType() {
		try {
			final PersonType p = objectFactory.createPersonType();
			logger.trace("Creating person " + p);
			return p;
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public ShopType createShopType() {
		try {
			final ShopType s = objectFactory.createShopType();
			logger.trace("Creating shop " + s);
			return s;
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public TempleType createTempleType() {
		try {
			final TempleType t = objectFactory.createTempleType();
			logger.trace("Creating temple " + t);
			return t;
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public TownType createTownType() {
		try {
			final TownType t = objectFactory.createTownType();
			logger.trace("Creating town " + t);
			return t;
		} catch (final Exception e) {
			e.printStackTrace();
			return null;
		}
	}
}
