package de.ped.dsatool.logic;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.ped.dsatool.dsa.generated.LanguageDistributionType;
import de.ped.dsatool.dsa.generated.LanguageProbabilityType;
import de.ped.dsatool.dsa.generated.LanguageType;
import de.ped.tools.Assert;
import de.ped.tools.MathUtil;

/**
 * A class representing the distribution of all known living languages among the population.
 * This class affects the generation of persons' names for a town or a region.
 * <p>This class is designed that the language distribution may be specified incomplete.
 * The rest of the languages will be filled up equally between the rest of the population.
 *
 * <p><b>See:</b> "???", Section "???" (p. ??..??)</p>
 *
 * @author Copyright (c) 1999 Peter Diefenbach (peter@pdiefenbach.de),
 *		based on algorithms by Urs Beeli
 */
public abstract class LanguageDistribution {
	protected static void checkLanguageProbability(final DSA dsa, final List<LanguageProbabilityType> probs) {
		// Check for uniqueness of each language.
		Iterator<LanguageProbabilityType> it = probs.iterator();
		final int langs = dsa.getLanguage().size();
		boolean[] langsUsed = new boolean[langs];
		while (it.hasNext()) {
			final LanguageProbabilityType langprob = it.next();
			final LanguageType lang = (LanguageType) langprob.getLanguage();
			final int langID = lang.getIDnum();
			Assert.assertFalse(langsUsed[langID]);
			langsUsed[langID] = true;
		}
		// Check for percentages out of range - single values and sum.
		it = probs.iterator();
		double sum = 0.0;
		while (it.hasNext()) {
			final LanguageProbabilityType langprob = it.next();
			final double prob = langprob.getProbability();
			Assert.assertTrue(0.0 <= prob);
			Assert.assertTrue(100.0 >= prob);
			sum += prob;
		}
		Assert.assertTrue(0.0 <= sum);
		Assert.assertTrue(100.0 >= sum);
	}

	/**
	 * Checks if the language probability contains no bad data errors, e.g., if
	 * no language is given twice, no language probability is out of range, or
	 * the sum of all language probabilities is out of range.
	 * 
	 * @param dsa
	 *            The dsa object holding together all the DSA logic layer data
	 * @param languageDistribution
	 *            The language distribution. An empty distribution
	 *            (<code>null</code>) or an incomplete distribution (sum is less
	 *            than 100 percent) is considered okay, as it can be filled up.
	 */
	public static void checkLanguageProbability(final DSA dsa, final LanguageDistributionType languageDistribution) {
		if (null != languageDistribution) {
			checkLanguageProbability(dsa, languageDistribution.getLanguageProbability());
		}
	}

	public static float getLanguagePercentage(final LanguageDistributionType languageDistribution, final String languageID) {
		final List<LanguageProbabilityType> probs = languageDistribution.getLanguageProbability();
		final Iterator<LanguageProbabilityType> it = probs.iterator();
		while (it.hasNext()) {
			final LanguageProbabilityType langprob = it.next();
			final LanguageType lang = (LanguageType)langprob.getLanguage(); 
			if (lang.getID().equals(languageID)) {
				return langprob.getProbability();
			}
		}
		return 0.0f;
	}

	public static void setLanguagePercentage(
			final LanguageDistributionType languageDistribution,
			final String languageID, 
			final float probability) {
		Assert.assertTrue(probability >= 0.0f);
		Assert.assertTrue(probability <= 100.0f);
		final List<LanguageProbabilityType> probs = languageDistribution.getLanguageProbability();
		final Iterator<LanguageProbabilityType> it = probs.iterator();
		LanguageProbabilityType foundLangProb = null;
		while (it.hasNext()) {
			final LanguageProbabilityType langprob = (LanguageProbabilityType)it.next();
			final LanguageType lang = (LanguageType)langprob.getLanguage(); 
			if (lang.getID().equals(languageID)) {
				foundLangProb = langprob;
				break;
			}
		}
		final DSA dsa = DSA.instance();
		if (null == foundLangProb) {
			foundLangProb = dsa.createLanguageProbabilityType();
			final LanguageType lang = Language.findByID(languageID);
			foundLangProb.setLanguage(lang);
		}
		foundLangProb.setProbability(probability);
		checkLanguageProbability(dsa, languageDistribution.getLanguageProbability());
	}

	public static void setLanguagePercentage(
			final LanguageDistributionType languageDistribution, 
			final int langIDnum, 
			final float probability) {
		final List<LanguageType> languages = DSA.instance().getLanguage();
		final Iterator<LanguageType> it = languages.iterator();
		LanguageType lang = null;
		while (it.hasNext() && (null == lang)) {
			final LanguageType l = (LanguageType)it.next();
			if (l.getIDnum() == langIDnum) {
				lang = l;
			}
		}
		setLanguagePercentage(languageDistribution, lang.getID(), probability);
	}

	public static LanguageDistributionType clone(final LanguageDistributionType languageDistribution) {
		final DSA dsa = DSA.instance();
		final LanguageDistributionType ret = dsa.createLanguageDistributionType();
		final List<LanguageProbabilityType> destlp = ret.getLanguageProbability();
		final List<LanguageProbabilityType> srclp = languageDistribution.getLanguageProbability();
		final Iterator<LanguageProbabilityType> it = srclp.iterator();
		while (it.hasNext()) {
			final LanguageProbabilityType srcprob = (LanguageProbabilityType)it.next();
			final LanguageProbabilityType destprob = dsa.createLanguageProbabilityType();
			destprob.setLanguage(srcprob.getLanguage());
			destprob.setProbability(srcprob.getProbability());
			destlp.add(destprob);
		}
		return ret;
	}

	/**
	 * @param languageDistribution
	 *            from that distribution a language is chosen
	 * @return The language key of a language. The language is chosen randomly
	 *         according to the language distribution.
	 */
	public static int chooseLanguageID(final LanguageDistributionType languageDistribution) {
		final List<LanguageProbabilityType> langprobs = languageDistribution.getLanguageProbability(); 
		final int size = langprobs.size();
		final int[] langArr = new int[size];
		double[] probArr = new double[size];
		final Iterator<LanguageProbabilityType> it = langprobs.iterator();
		int idx = 0;
		while (it.hasNext()) {
			final LanguageProbabilityType langprob = (LanguageProbabilityType)it.next();
			final LanguageType lang = (LanguageType)langprob.getLanguage();
			langArr[idx] = lang.getIDnum();
			probArr[idx] = langprob.getProbability();
			idx++;
		}
		idx = MathUtil.randomDistributed(probArr);
		return langArr[idx];
	}

	/**
	 * @param languageDistribution
	 *            from that distribution a language is chosen
	 * @return A language chosen randomly according to the language
	 *         distribution.
	 */
	public static LanguageType chooseLanguage(final LanguageDistributionType languageDistribution) {
		final int langID = LanguageDistribution.chooseLanguageID(languageDistribution);
		return Language.findByIDnum(langID);
	}

	/** 
	 * @return the default language distribution. Currently this is equal probability for all languages. 
	 */
	public static LanguageDistributionType createDefaultLanguageDistribution() {
		final DSA dsa = DSA.instance();
		final LanguageDistributionType ret = dsa.createLanguageDistributionType();
		final List<LanguageType> languages = dsa.getLanguage();
		final float probval = 100f / languages.size();
		final Iterator<LanguageType> it = languages.iterator();
		while (it.hasNext()) {
			final LanguageType lang = (LanguageType)it.next();
			final LanguageProbabilityType prob = dsa.createLanguageProbabilityType();
			prob.setLanguage(lang);
			prob.setProbability(probval);
			ret.getLanguageProbability().add(prob);
		} 
		return ret;
	}

	/**
	 * @param languageID
	 *            the language to use
	 * @return a language distribution containing only one language with a peak
	 *         of 100%, and all other languages 0%.
	 */
	public static LanguageDistributionType createSingleTongueLanguageDistribution(final int languageID) {
		final DSA dsa = DSA.instance();
		final LanguageDistributionType ret = dsa.createLanguageDistributionType();
		final List<LanguageType> languages = dsa.getLanguage();
		final Iterator<LanguageType> it = languages.iterator();
		while (it.hasNext()) {
			final LanguageType lang = (LanguageType)it.next();
			final LanguageProbabilityType prob = dsa.createLanguageProbabilityType();
			prob.setLanguage(lang);
			prob.setProbability((languageID == lang.getIDnum()) ? 100.0f : 0.0f);
			ret.getLanguageProbability().add(prob);
		}
		return ret;
	}

	/**
	 * If the language distribution still contains undefined values, fill them
	 * up. The still undefined values will be distributed equally among the rest
	 * of the population.
	 * 
	 * @param me
	 *            contains the well-defined language distribution of the
	 *            population
	 */
	public static void fillUp(final LanguageDistributionType me) {
		fillUpWith(me, createDefaultLanguageDistribution());
	}

	/**
	 * If the language distribution still contains undefined values, fill them
	 * up. The still undefined values will be distributed among the rest of the
	 * population as defined in the given other language distribution.
	 * 
	 * @param me
	 *            contains the well-defined language distribution of the
	 *            population
	 * @param other
	 *            controls the distribution to fill up with the so far undefined
	 *            rest of the population.
	 */
	public static void fillUpWith(final LanguageDistributionType me, final LanguageDistributionType other) {
		double divisor  =   0.0;
		double dividend = 100.0;
		// find all languages which are part of other, but not of me. 
		final LinkedList<LanguageProbabilityType> newProbs = new LinkedList<LanguageProbabilityType>();
		final Iterator<LanguageProbabilityType> otherIterator = other.getLanguageProbability().iterator();
		while (otherIterator.hasNext()) {
			final LanguageProbabilityType otherLangProb = (LanguageProbabilityType)otherIterator.next();
			final LanguageType otherLanguage = (LanguageType)otherLangProb.getLanguage();
			
			final Iterator<LanguageProbabilityType> myIterator = me.getLanguageProbability().iterator();
			LanguageProbabilityType myFoundLanguageProp = null;
			while (myIterator.hasNext() && (null == myFoundLanguageProp)) {
				final LanguageProbabilityType myLangProb = (LanguageProbabilityType)myIterator.next();
				final LanguageType myLanguage = (LanguageType)myLangProb.getLanguage();
				if (myLanguage.getID().equals(otherLanguage.getID())) {
					myFoundLanguageProp = myLangProb;
				}
			}
			if (null == myFoundLanguageProp) {
				// This language is to add
				divisor += otherLangProb.getProbability();
				newProbs.add(otherLangProb);
			} else {
				// We already counted that language
				dividend -= myFoundLanguageProp.getProbability();
			}
		}
		final int howMany = newProbs.size();
		if (howMany > 0) {
			// catch division by zero - if there are no other folks around,
			// assume there really are other folks around, and equally distributed.
			double factor = dividend / howMany;
			if (0.0 != divisor) {
				factor = dividend / divisor;
			}
			// Now add all the other languages, to fill up the population.
			final Iterator<LanguageProbabilityType> it = newProbs.iterator();
			while (it.hasNext()) {
				final LanguageProbabilityType newProb = (LanguageProbabilityType)it.next();
				final LanguageProbabilityType prob = DSA.instance().createLanguageProbabilityType();
				prob.setLanguage(newProb.getLanguage());
				prob.setProbability((float)(newProb.getProbability() * factor));
				me.getLanguageProbability().add(prob);
			}
		}
	}
}
