package de.ped.dsatool.logic;

import java.awt.Graphics2D;
import java.awt.Polygon;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.ped.dsatool.dsa.generated.BasicAreaType;
import de.ped.dsatool.dsa.generated.BusinessObjectType;
import de.ped.dsatool.dsa.generated.MapLayerAreaType;
import de.ped.dsatool.dsa.generated.MapLayerPatchType;
import de.ped.dsatool.dsa.generated.MapLayerType;
import de.ped.dsatool.dsa.generated.MapType;
import de.ped.dsatool.dsa.generated.Shape;
import de.ped.dsatool.dsa.generated.TownType;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.MathVector;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * The map of Aventuria.
 * 
 * @author Copyright (c) 2007 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Map {
	private static Logger logger = Logger.getLogger("Map");
	static {
		logger.setLogLevel(Level.INFO);
	}

	/** The default zooming in x dimension. Defines precision of identity of positions. */
	public static final int ZOOM_X = 10000;

	/** The default zooming in y dimension. Defines precision of identity of positions. */
	public static final int ZOOM_Y = 10000;

	/**
	 * Creates a polygon from the map layer area list of points. The polygon is
	 * zoomed and moved on the plane.
	 * 
	 * @param points
	 *            The list of points to create the polygon from.
	 * @param zoomX
	 *            The zoom in X dimension.
	 * @param zoomY
	 *            The zoom in Y dimension.
	 * @param transX
	 *            The translation (moving) in X dimension.
	 * @param transY
	 *            The translation (moving) in Y dimension.
	 * @return The polygon created, or <code>null</code> if not possible.
	 */
	private static Polygon createPolygon(final List<?> points, final double zoomX, final double zoomY, final int transX, final int transY) {
		Polygon ret = null;
		if (!CollectionUtil.isNullOrEmpty(points)) {
			final double[] coords = toDoubleArray(points);
			final int edges = coords.length / 2;
			final int[] xpoints = new int[edges];
			final int[] ypoints = new int[edges];
			for (int i = 0; i < edges; i++) {
				xpoints[i] = (int) Math.round(zoomX * coords[2 * i + 0]) + transX;
				ypoints[i] = (int) Math.round(zoomY * coords[2 * i + 1]) + transY;
			}
			ret = new Polygon(xpoints, ypoints, edges);
		}
		return ret;
	}

	/**
	 * Translates an array of coordinates from normed double to screen
	 * coordinates. To achieve that, all coordinates are zoomed and moved on the
	 * plane.
	 * 
	 * @param coords
	 *            The normed coordinates to transform.
	 * @param zoomX
	 *            The zoom in X dimension.
	 * @param zoomY
	 *            The zoom in Y dimension.
	 * @param transX
	 *            The translation (moving) in X dimension.
	 * @param transY
	 *            The translation (moving) in Y dimension.
	 * @return An array of same size as coords, with translated and moved
	 *         values.
	 */
	private static int[] toIntArray(final double[] coords, final double zoomX, final double zoomY, final int transX,
			final int transY) {
		int[] ret = null;
		if (null != coords) {
			ret = new int[coords.length];
			// First translate all the x/y pairs
			final int pairsIdx = coords.length & 0xfffe;
			for (int i = 0; i < pairsIdx; i++) {
				if ((i & 1) == 0) {
					ret[i] = (int) Math.round(zoomX * coords[i]) + transX;
				} else {
					ret[i] = (int) Math.round(zoomY * coords[i]) + transY;
				}
			}
			// Now the radius of the circle may be left over.
			if (pairsIdx < coords.length) {
				ret[pairsIdx] = (int) Math.round(zoomX * coords[pairsIdx]);
			}
		}
		return ret;
	}

	/**
	 * Returns the points as a simple array of doubles. Used as helping method
	 * to convert a map layer area to doubles.
	 * 
	 * @param points
	 *            The list of points to convert.
	 * @return The array of doubles read from the list.
	 */
	private static double[] toDoubleArray(final List<?> points) {
		double[] ret = null;
		final Iterator<?> pit = points.iterator();
		if (!CollectionUtil.isNullOrEmpty(points)) {
			final int edges = points.size();
			final double[] coords = new double[edges * 2];
			int i = 0;
			while (pit.hasNext() && (i < edges * 2)) {
				final List<?> al = DSA.toListValue(pit.next());
				Assert.assertTrue(2 == al.size());
				coords[i + 0] = DSA.toDoubleValue(al.get(0));
				coords[i + 1] = DSA.toDoubleValue(al.get(1));
				i += 2;
			}
			ret = coords;
		}
		return ret;
	}

	/**
	 * Returns the points as a simple array of doubles.
	 * The coding is as follows:
	 * <dl>
	 *   <dt>Circles</dt>
	 *   <dd>represented by 3 doubles: center x, center y and radius.</dd>
	 *   <dt>Rectangles</dt>
	 *   <dd>represented by 4 doubles: upper, left, lower, right. </dd>
	 *   <dt>Polygons and Polylines</dt>
	 *   <dd>always of even size. First x, then y value of the first point, then x and y value of the second point, and so on.</dd>
	 * </dl>
	 * 
	 * @param ba The basic area type to represent.
	 * @return A simple double coding of the points, or <code>null</code> if nothing fits.
	 */
	private static double[] toDoubleArray(final BasicAreaType ba) {
		double[] ret = null;
		final Shape shape = ba.getShape();
		final List<?> points = ba.getPoint();
		final Iterator<?> pit = points.iterator();
		if (shape.equals(Shape.CIRCLE)) {
			Assert.assertEquals(1, points.size());
			final List<?> al = DSA.toListValue(pit.next());
			Assert.assertEquals(2, al.size());
			ret = new double[3];
			ret[0] = DSA.toDoubleValue(al.get(0));
			ret[1] = DSA.toDoubleValue(al.get(1));
			ret[2] = ba.getRadius();
		} else if (shape.equals(Shape.RECTANGLE)) {
			// Read the coordinates of upper left and lower right corner
			final double[] coords = toDoubleArray(points);
			Assert.assertTrue(4 == coords.length);
			// Check if we really have upper left and lower right corner
			for (int i = 0; i < 2; i++) {
				if (coords[0 + i] > coords[2 + i]) {
					double buf = coords[0 + i];
					coords[0 + i] = coords[2 + i];
					coords[2 + i] = buf;
				}
			}
			ret = coords;
		} else if (shape.equals(Shape.POLYGON) || shape.equals(Shape.POLYLINE)) {
			ret = toDoubleArray(points);
		}
		return ret;
	}

	/**
	 * Returns whether the given area contains the given position.
	 * 
	 * @param ba
	 *            The area.
	 * @param pos
	 *            The position.
	 * @return <code>true</code> iff the area contains the position.
	 */
	private static boolean contains(final BasicAreaType ba, final MathVector pos) {
		boolean ret = false;
		final Shape shape = ba.getShape();
		if (shape.equals(Shape.CIRCLE)) {
			final double[] coords = toDoubleArray(ba);
			// Use good old Pythagoras to decide: x^2 + y^2 = r^2
			final double diffX = coords[0] - pos.get(0);
			final double diffY = coords[1] - pos.get(1);
			final double r = coords[2];
			if (diffX * diffX + diffY * diffY < r * r) {
				ret = true;
			}
		} else if (shape.equals(Shape.POLYGON)) {
			final Polygon pol = createPolygon(ba.getPoint(), ZOOM_X, ZOOM_Y, 0, 0);
			if (pol.contains(ZOOM_X * pos.get(0), ZOOM_Y * pos.get(1))) {
				ret = true;
			}
		} else if (shape.equals(Shape.RECTANGLE)) {
			final double[] coords = toDoubleArray(ba);
			if (	(coords[0] < pos.get(0)) && (pos.get(0) < coords[2]) 
				&&  (coords[1] < pos.get(1)) && (pos.get(1) < coords[3]))
			{
				ret = true;
			}
		}
		return ret;
	}

	/**
	 * Returns whether the given area contains the given position.
	 * 
	 * @param mla
	 *            The area.
	 * @param pos
	 *            The position.
	 * @return <code>true</code> iff the area contains the position.
	 */
	public static boolean contains(final MapLayerAreaType mla, final MathVector pos) {
		boolean ret = false;
		final Iterator<BasicAreaType> batsi = mla.getArea().iterator();
		while (!ret && batsi.hasNext()) {
			final BasicAreaType bat = batsi.next();
			if (contains(bat, pos)) {
				ret = true;
			}
		}
		return ret;
	}

	/**
	 * Finds the first object of type clazz that fits to the given position. The
	 * order criterion of the search is the depth-first traversal of the
	 * Aventuria map.
	 * 
	 * @param pos
	 *            The position to search for.
	 * @param clazz
	 *            The class to search for.
	 * @return The map area linking to the object, or <code>null</code> if none found.
	 */
	public static MapLayerAreaType firstAt(MathVector pos, Class<?> clazz) {
		MapLayerAreaType ret = null;
		MapType map = DSA.instance().getMap();
		Iterator<MapLayerType> mlit = map.getMapLayer().iterator();
		while (mlit.hasNext() && (null == ret)) {
			final MapLayerType ml = mlit.next();
			final Iterator<MapLayerAreaType> mlait = ml.getMapLayerArea().iterator();
			while (mlait.hasNext() && (null == ret)) {
				final MapLayerAreaType mla = mlait.next();
				final Object ref = mla.getRef();
				if ((null != ref) && ref.getClass().equals(clazz)) {
					logger.debug("Looking at area " + calculateName(mla));
					if (contains(mla, pos)) {
						logger.debug("Area " + calculateName(mla) + " matches position.");
						ret = mla;
					}
				}
			}
		}
		return ret;
	}

	/**
	 * Finds the first object of one of the clazzes that fits to the given
	 * position. First order criterion of the search is the order of classes,
	 * second the depth-first traversal of the Aventuria map.
	 * 
	 * @param pos
	 *            The position to search for.
	 * @param clazzes
	 *            The classes to search for.
	 * @return The map area linking to the object, or <code>null</code> if none found.
	 */
	public static MapLayerAreaType firstAt(final MathVector pos, final Class<?>[] clazzes) {
		MapLayerAreaType ret = null;
		for (int i = 0; (i < clazzes.length) && (null == ret); i++) {
			ret = firstAt(pos, clazzes[i]);
		}
		return ret;
	}

	/**
	 * Finds all map layer areas containing the given position.
	 * 
	 * @param pos
	 *            The position to search for.
	 * @return A list of map layer areas.
	 */
	public static List<MapLayerAreaType> allMapLayerAreasAt(final MathVector pos) {
		final LinkedList<MapLayerAreaType> ret = new LinkedList<MapLayerAreaType>();
		final MapType map = DSA.instance().getMap();
		final Iterator<MapLayerType> mlit = map.getMapLayer().iterator();
		while (mlit.hasNext()) {
			final MapLayerType ml = mlit.next();
			final Iterator<MapLayerAreaType> mlait = ml.getMapLayerArea().iterator();
			while (mlait.hasNext()) {
				final MapLayerAreaType mla = mlait.next();
				if (contains(mla, pos)) {
					ret.add(mla);
				}
			}
		}
		return ret;
	}

	/**
	 * Finds all map layer areas containing the given position.
	 * 
	 * @param pos
	 *            The position to search for.
	 * @param tf
	 *            The text formatter to format the output with.
	 * @return A string describing the map layer areas found.
	 */
	public static String allMapLayerAreasAt(final MathVector pos, final TextFormatter tf) {
		final StringBuffer ret = new StringBuffer();
		ret.append(tf.newLine());
		final MapType map = DSA.instance().getMap();
		final Iterator<MapLayerType> mlit = map.getMapLayer().iterator();
		while (mlit.hasNext()) {
			final MapLayerType ml = mlit.next();
			boolean isLayerNew = true;
			final Iterator<MapLayerAreaType> mlait = ml.getMapLayerArea().iterator();
			while (mlait.hasNext()) {
				final MapLayerAreaType mla = mlait.next();
				if (contains(mla, pos)) {
					if (isLayerNew) {
						ret.append(tf.beginBold());
						ret.append(calculateName(ml));
						ret.append(tf.endBold());
						ret.append(tf.newLine());
						isLayerNew = false;
					}
					ret.append("- " + calculateName(mla));
					ret.append(tf.newLine());
				}
			}
		}
		return ret.toString();
	}

	/**
	 * Draws a basic area type.
	 * 
	 * @param bat
	 *            The basic area type to draw.
	 * @param zoomX
	 *            The zoom in X dimension.
	 * @param zoomY
	 *            The zoom in Y dimension.
	 * @param transX
	 *            The translation (moving) in X dimension.
	 * @param transY
	 *            The translation (moving) in Y dimension.
	 * @param graphics2D
	 *            The graphics object to draw to.
	 */
	private static void draw(final BasicAreaType bat, final double zoomX, final double zoomY, final int transX,
			final int transY, final Graphics2D graphics2D) {
		if (bat.getShape().equals(Shape.CIRCLE)) {
			final int[] coords = toIntArray(toDoubleArray(bat), zoomX, zoomY, transX, transY);
			final int left = coords[0] - coords[2];
			final int upper = coords[1] - coords[2];
			final int width = 2 * coords[2];
			final int height = 2 * coords[2];
			graphics2D.fillOval(left, upper, width, height);
			graphics2D.drawOval(left, upper, width, height);
		} else if (bat.getShape().equals(Shape.POLYGON)) {
			final Polygon poly = createPolygon(bat.getPoint(), zoomX, zoomY, transX, transY);
			if (null != poly) {
				graphics2D.fill(poly);
				graphics2D.draw(poly);
			}
		} else if (bat.getShape().equals(Shape.POLYLINE)) {
			logger.debug("I want to draw a Polyline.");
		} else if (bat.getShape().equals(Shape.RECTANGLE)) {
			final int[] coords = toIntArray(toDoubleArray(bat), zoomX, zoomY, transX, transY);
			final int width = coords[2] - coords[0];
			final int height = coords[3] - coords[1];
			graphics2D.fillRect(coords[0], coords[1], width, height);
			graphics2D.drawRect(coords[0], coords[1], width, height);
		}
	}

	/**
	 * Draws a patch. The color and stroke are taken from graphics2D and are not
	 * explicitly set.
	 * 
	 * @param patch
	 *            The map layer area type to draw.
	 * @param zoomX
	 *            The zoom in X dimension.
	 * @param zoomY
	 *            The zoom in Y dimension.
	 * @param transX
	 *            The translation (moving) in X dimension.
	 * @param transY
	 *            The translation (moving) in Y dimension.
	 * @param graphics2D
	 *            The graphics object to draw to.
	 */
	public static void draw(final MapLayerAreaType patch, final double zoomX, final double zoomY, final int transX, final int transY, final Graphics2D graphics2D) {
		final Iterator<BasicAreaType> bats = patch.getArea().iterator();
		while (bats.hasNext()) {
			final BasicAreaType bat = bats.next();
			draw(bat, zoomX, zoomY, transX, transY, graphics2D);
		}
	}

	/**
	 * Returns the name of the area. If a name attribute is given, that is
	 * taken. Otherwise the ref attribute is taken and the name taken from
	 * there.
	 * 
	 * @param patch
	 *            The area to find a name for.
	 * @return The name of the area.
	 */
	public static String calculateName(final MapLayerAreaType patch) {
		String ret = patch.getName();
		if (CollectionUtil.isNullOrEmpty(ret)) {
			final Object o = patch.getRef();
			if (o instanceof BusinessObjectType) {
				final BusinessObjectType bo = (BusinessObjectType) o;
				ret = bo.getName();
			}
			if (o instanceof TownType) {
				final TownType bo = (TownType) o;
				ret = bo.getName();
			}
		}
		return ret;
	}

	/**
	 * Returns the name of the patch. If a name attribute is given, that is
	 * taken. Otherwise the filename is returned as name.
	 * 
	 * @param patch
	 *            The patch to find a name for.
	 * @return The name of the patch.
	 */
	public static String calculateName(final MapLayerPatchType patch) {
		String ret = patch.getName();
		if (CollectionUtil.isNullOrEmpty(ret)) {
			ret = patch.getFilename();
		}
		return ret;
	}

	/**
	 * Returns the name of the patch. If a name attribute is given, that is
	 * taken. Otherwise the filename is returned as name.
	 * 
	 * @param patch
	 *            The patch to find a name for.
	 * @return The name of the patch.
	 */
	public static String calculateName(final MapLayerType patch) {
		String ret = patch.getName();
		if (CollectionUtil.isNullOrEmpty(ret)) {
			ret = patch.getID();
		}
		return ret;
	}
}
