package de.ped.dsatool.logic;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.List;

import de.ped.dsatool.dsa.generated.BusinessObjectType;
import de.ped.tools.Assert;
import de.ped.tools.HtmlHelper;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.Textable;
import de.ped.tools.log.Logger;

/**
 * A model to keep the actual calendar year.
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MasterDataModel implements Textable {
	@SuppressWarnings("unused")
	private static Logger logger = DSAEnv.instance().logger();

	/**
	 * Defines what type of master data to show.
	 * 
	 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum MasterDataType {
		CALENDAR          ("CalendarType"),
		DAY_OF_WEEK       ("DayOfWeekType"),
		DEITY             ("DeityType"),
		HOLIDAY           ("HolidayType"),
		LANGUAGE          ("LanguageType"),
		MONTH             ("MonthType"),
		PLANT_CATEGORY    ("PlantCategoryType"),
		PLANT             ("PlantType"),
		REGION            ("RegionType"),
		SEASON            ("SeasonType"),
		SEX               ("SexType"),
		SHOP_CATEGORY     ("ShopCategoryType"),
		SHOP_SPECIALTY    ("ShopSpecialtyType"),
		TERRAIN           ("TerrainType"),
		TOWN_SIZE_CATEGORY("TownSizeCategoryType");

		private MasterDataType(final String className) {
			this.className = className;
		}

		private final String className;
	}

	public static final String MESSAGE_PREFIX = "MasterData";

	public static String getMessageKey(final String className) {
		return MESSAGE_PREFIX + "." + className;
	}

	public static String getMessageKey(final int mode) {
		return getMessageKey(MasterDataType.values()[mode]);
	}

	public static String getMessageKey(final MasterDataType mode) {
		return MESSAGE_PREFIX + "." + mode.className;
	}

	public String getMessageKey() {
		return MESSAGE_PREFIX + "." + mode.className;
	}

	private MasterDataType mode = MasterDataType.CALENDAR;

	public MasterDataType getMode() {
		return mode;
	}

	public void setMode(final MasterDataType value) {
		mode = value;
	}

	/**
	 * Defines all the key information about a column to show for the master
	 * data model.
	 * 
	 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ColumnInfo extends TextFormatter.ColumnInfo {
		/** The mode this column info belongs to. */
		public MasterDataType mode = null;

		/** The getter method to fetch the data. */
		public String getter = null;

		/** 
		 * Default constructor - useless in itself, at least set mode and getter afterwards. 
		 */
		public ColumnInfo() {
			// Nothing to do
		}

		/** 
		 * Constructor defining the minimum information for this object to be useful.
		 * @param mode the mode this column info belongs to
		 * @param getter the getter method to fetch the data
		 */
		public ColumnInfo(final MasterDataType mode, final String getter) {
			this.mode = mode;
			this.getter = getter;
		}

		/** 
		 * Constructor defining the minimum information for this object to be useful. 
		 * @param mode the mode this column info belongs to
		 * @param getter the getter method to fetch the data
		 * @param defaultLength the default length of the column for fixed font formatting
		 */
		public ColumnInfo(final MasterDataType mode, final String getter, final int defaultLength) {
			this.mode = mode;
			this.getter = getter;
			this.columnDefaultLength = defaultLength;
		}

		/**
		 * Returns the unformatted heading name.
		 * It looks for several places for the heading:
		 * <ol>
		 *   <li>
		 *     <em>columnHeadingName</em>, if not <code>null</code>.
		 *   <li>
		 *     Message property named like the attribute 
		 *     within a section named like the class name 
		 *     within the section named {@link MasterDataModel#MESSAGE_PREFIX},
		 *     if defined.
		 *   <li>
		 *     Message property named like the attribute 
		 *     within the section named {@link MasterDataModel#MESSAGE_PREFIX},
		 *     if defined.
		 *   <li>
		 *     The name of the attribute itself, derived from the name of the getter method.
		 * </ol>
		 * 
		 * @return the unformatted heading name
		 */
		@Override
		public String getHeadingName() {
			initNameAndToolTip();
			return super.getHeadingName();
		}

		@Override
		public String getHeadingToolTip() {
			initNameAndToolTip();
			return super.getHeadingToolTip();
		}

		protected void initNameAndToolTip() {
			if (null == super.getHeadingName()) {
				final Messages messages = DSAEnv.instance().messages();
				String text = getHeadingKeyBase();
				final String key = getHeadingKey();
				if (null != key) {
					text = messages.getText(key, null);
					setHeadingToolTip(messages.getShortDescription(key));
				}
				setHeadingName(text);
			}
		}

		protected String getHeadingKey() {
			final Messages messages = DSAEnv.instance().messages();
			final String keyBase = getHeadingKeyBase();
			String ret = keyBase;
			String value = messages.getText(ret, null);
			if (null == value) {
				ret = MasterDataModel.getMessageKey(mode) + "." + keyBase;
				value = messages.getText(ret, null);
			}
			if (null == value) {
				ret = MESSAGE_PREFIX + "." + keyBase;
				value = messages.getText(ret, null);
			}
			if (null == value) {
				ret = null;
			}
			return ret;
		}

		protected String getHeadingKeyBase() {
			String keyBase = null;
			if (null != headingName) {
				keyBase = headingName;
			} else if (getter.startsWith("get")) {
				keyBase = getter.substring(3);
			} else if (getter.startsWith("is")) {
				keyBase = getter.substring(2);
			} else {
				keyBase = getter;
			}
			return keyBase;
		}
	}

	private static final int COLUMN_WIDTH_IDNUM = 3;
//    private static final int COLUMN_WIDTH_ID = 35;
	private static final int COLUMN_WIDTH_NAME = 32;
	private static final int COLUMN_WIDTH_ABBR = 5;
	private static final int COLUMN_WIDTH_REFERENCE = 40;

	private static final ColumnInfo[][] COLUMN_INFOS = new ColumnInfo[][] {
		{ //MODE_CALENDAR
			new ColumnInfo(MasterDataType.CALENDAR, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.CALENDAR, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.CALENDAR, "getName", 38),
			new ColumnInfo(MasterDataType.CALENDAR, "getPositiveUnitName", 25),
			new ColumnInfo(MasterDataType.CALENDAR, "getNegativeUnitName", 25),
			new ColumnInfo(MasterDataType.CALENDAR, "getYearZeroBFis", 9),
			new ColumnInfo(MasterDataType.CALENDAR, "isHasYearZero", 6),
//			new ColumnInfo(MasterDataType.CALENDAR, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_DAY_OF_WEEK
			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getAbbr", COLUMN_WIDTH_ABBR),
			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getNameInRealLife", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getAbbrInRealLife", COLUMN_WIDTH_ABBR),
//			new ColumnInfo(MasterDataType.DAY_OF_WEEK, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_DEITY
			new ColumnInfo(MasterDataType.DEITY, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.DEITY, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.DEITY, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.DEITY, "getAbbr", COLUMN_WIDTH_ABBR),
//			new ColumnInfo(MasterDataType.DEITY, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_HOLIDAY
			new ColumnInfo(MasterDataType.HOLIDAY, "getMonth", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.HOLIDAY, "getFirstDay", 2),
			new ColumnInfo(MasterDataType.HOLIDAY, "getDuration", 2),
			new ColumnInfo(MasterDataType.HOLIDAY, "getName", 80),
			new ColumnInfo(MasterDataType.HOLIDAY, "getRemark", 110),
		},
		{ //MODE_LANGUAGE
			new ColumnInfo(MasterDataType.LANGUAGE, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.LANGUAGE, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.LANGUAGE, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.LANGUAGE, "getAbbr", COLUMN_WIDTH_ABBR),
			new ColumnInfo(MasterDataType.LANGUAGE, "getFemalePercentage", 5),
//			new ColumnInfo(MasterDataType.LANGUAGE, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_MONTH
			new ColumnInfo(MasterDataType.MONTH, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.MONTH, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.MONTH, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.MONTH, "getDeity", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.MONTH, "getNameInRealLife", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.MONTH, "getAbbrInRealLife", COLUMN_WIDTH_ABBR),
			new ColumnInfo(MasterDataType.MONTH, "getSeason", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.MONTH, "getDuration", 2),
//			new ColumnInfo(MasterDataType.MONTH, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_PLANT_CATEGORY
			new ColumnInfo(MasterDataType.PLANT_CATEGORY, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.PLANT_CATEGORY, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.PLANT_CATEGORY, "getName", COLUMN_WIDTH_NAME),
//			new ColumnInfo(MasterDataType.PLANT_CATEGORY, "getAbbr", COLUMN_WIDTH_ABBR),
//			new ColumnInfo(MasterDataType.PLANT_CATEGORY, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_PLANT
			new ColumnInfo(MasterDataType.PLANT, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.PLANT, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.PLANT, "getName", COLUMN_WIDTH_NAME),
//			new ColumnInfo(MasterDataType.PLANT, "getAbbr", COLUMN_WIDTH_ABBR),
			new ColumnInfo(MasterDataType.PLANT, "getCategory", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.PLANT, "getRegionsDescription", 105),
			new ColumnInfo(MasterDataType.PLANT, "getTerrainsDescription", 70),
			new ColumnInfo(MasterDataType.PLANT, "getHowHardToFind", 3),
			new ColumnInfo(MasterDataType.PLANT, "getHarvestTimeDescription", 65),
			new ColumnInfo(MasterDataType.PLANT, "getComment", 85),
			new ColumnInfo(MasterDataType.PLANT, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_REGION
			new ColumnInfo(MasterDataType.REGION, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.REGION, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.REGION, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.REGION, "getAbbr", COLUMN_WIDTH_ABBR),
			new ColumnInfo(MasterDataType.REGION, "getDescription", 60),
			new ColumnInfo(MasterDataType.REGION, "getMainLanguage", COLUMN_WIDTH_NAME),
//			new ColumnInfo(MasterDataType.REGION, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_SEASON
			new ColumnInfo(MasterDataType.SEASON, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.SEASON, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.SEASON, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.SEASON, "getAbbr", COLUMN_WIDTH_ABBR),
//			new ColumnInfo(MasterDataType.SEASON, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_SEX
			new ColumnInfo(MasterDataType.SEX, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.SEX, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.SEX, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.SEX, "getAbbr", COLUMN_WIDTH_ABBR),
//			new ColumnInfo(MasterDataType.SEX, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_SHOP_CATEGORY
			new ColumnInfo(MasterDataType.SHOP_CATEGORY, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.SHOP_CATEGORY, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.SHOP_CATEGORY, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.SHOP_CATEGORY, "getChiefTitle", COLUMN_WIDTH_NAME),
//			new ColumnInfo(MasterDataType.SHOP_CATEGORY, "getQualityNamesTable", 20),
//			new ColumnInfo(MasterDataType.SHOP_CATEGORY, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_SHOP_SPECIALTY
			new ColumnInfo(MasterDataType.SHOP_SPECIALTY, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.SHOP_SPECIALTY, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.SHOP_SPECIALTY, "getName", COLUMN_WIDTH_NAME),
			new ColumnInfo(MasterDataType.SHOP_SPECIALTY, "getShopCategory", COLUMN_WIDTH_NAME),
//			new ColumnInfo(MasterDataType.SHOP_SPECIALTY, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_TERRAIN
			new ColumnInfo(MasterDataType.TERRAIN, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.TERRAIN, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.TERRAIN, "getName", COLUMN_WIDTH_NAME),
//			new ColumnInfo(MasterDataType.TERRAIN, "getAbbr", COLUMN_WIDTH_ABBR),
//			new ColumnInfo(MasterDataType.TERRAIN, "getReference", COLUMN_WIDTH_REFERENCE),
		},
		{ //MODE_TOWN_SIZE_CATEGORY
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getIDnum", COLUMN_WIDTH_IDNUM),
//			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getID", COLUMN_WIDTH_ID),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getMinimumNumberOfInhabitants"),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getEnterprisesRemark", 85),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getTravellersRemark", 35),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getMarketDaysRemark", 40),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getQuality", 2),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getPricesMerchands", 2),
			new ColumnInfo(MasterDataType.TOWN_SIZE_CATEGORY, "getPricesHandcrafts", 2),
		},
	};

	public List<?> getDataList() {
		List<?> ret = null;
		final DSA dsa = DSA.instance();
		switch (mode) {
		case CALENDAR:
			ret = dsa.getCalendar();
			break;
		case DAY_OF_WEEK:
			ret = dsa.getDayOfWeek();
			break;
		case DEITY:
			ret = dsa.getDeity();
			break;
		case HOLIDAY:
			ret = dsa.getHoliday();
			break;
		case LANGUAGE:
			ret = dsa.getLanguage();
			break;
		case MONTH:
			ret = dsa.getMonth();
			break;
		case PLANT_CATEGORY:
			ret = dsa.getPlantCategory();
			break;
		case PLANT:
			ret = dsa.getPlant();
			break;
		case REGION:
			ret = dsa.getRegion();
			break;
		case SEASON:
			ret = dsa.getSeason();
			break;
		case SEX:
			ret = dsa.getSex();
			break;
		case SHOP_CATEGORY:
			ret = dsa.getShopCategory();
			break;
		case SHOP_SPECIALTY:
			ret = dsa.getShopSpecialty();
			break;
		case TERRAIN:
			ret = dsa.getTerrain();
			break;
		case TOWN_SIZE_CATEGORY:
			ret = dsa.getTownSizeCategory();
			break;
		default:
			Assert.assertTrue(false);
		}
		return ret;
	}

	protected static final Class<?> BOOLEAN_CLASS = new Boolean(false).getClass();

	/**
	 * @param clazz
	 *            the class to get the name from
	 * @return the class name of the given class - stripped by any package
	 *         information and a possible trailing &quot;Impl&quot; in the name
	 */
	protected static String getClassName(final Class<?> clazz) {
		final String fullClassName = clazz.getName();
		final String[] classNameParts = fullClassName.split("\\.");
		String partClassName = classNameParts[classNameParts.length - 1];
		if (partClassName.endsWith("Impl")) {
			partClassName = partClassName.substring(0, partClassName.length() - 4);
		}
		return partClassName;
	}

	/**
	 * For the given class, finds the mode in which this class is displayed.
	 * 
	 * @param clazz
	 *            the class
	 * @return the mode, or <code>null</code> if not found
	 */
	public static MasterDataType findMode(final Class<?> clazz) {
		MasterDataType ret = null;
		final String className = getClassName(clazz);
		for (int i = 0; (i < MasterDataType.values().length) && (null == ret); i++) {
			if (className.startsWith(MasterDataType.values()[i].className)) {
				ret = MasterDataType.values()[i];
			}
		}
		return ret;
	}

	public static String getHRefFor(final BusinessObjectType bot) {
		String href = HtmlHelper.addParameter(Model.URL, Model.ACTION, Model.ACTION_MASTER_DATA_SHOW);
		href = HtmlHelper.addParameter(href, Model.MASTER_DATA_MODE,
				Integer.toString(findMode(bot.getClass()).ordinal()));
		href += "#" + bot.getID();
		return href;
	}

	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		final StringBuffer s = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final ColumnInfo[] arrCI = COLUMN_INFOS[mode.ordinal()];
		final List<?> list = getDataList();
		final Class<?> clazz = list.get(0).getClass();
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), null);
		s.append(tf.beginTable("table0", ti));
		// Table heading row
		s.append(tf.formatTableHeading(arrCI));
		s.append(tf.beginTableBody());
		// Table content rows
		final Iterator<?> it = list.iterator();
		while (it.hasNext()) {
			s.append(tf.beginTableRow());
			final Object o = it.next();
			for (int c = 0; c < arrCI.length; c++) {
				String value = "null";
				int alignment = TextFormatter.LEFT;
				try {
					final String methodName = arrCI[c].getter;
					Method method = clazz.getMethod(methodName, (Class[])null);
//                    final String returnTypeName = method.getReturnType().getName();
//                    logger.debug(methodName + " -> " + returnTypeName);
					final Object returnValue = method.invoke(o, (Object[]) null );
//                    logger.debug((null == returnValue) ? "null" : returnValue.toString());
					if (null == returnValue) {
						value = "";
					} else if (returnValue instanceof String) {
						final String str = returnValue.toString();
						if (methodName.equals("getID")) {
							value = tf.anchor(str, str);
						} else {
							value = tf.sgmlify(str);
						}
					} else if (DSA.instance().isImpl(returnValue)) { 
						// It is a child XML object
						if (returnValue instanceof BusinessObjectType) {
							final BusinessObjectType bot = (BusinessObjectType) returnValue;
							value = tf.href(getHRefFor(bot), bot.getName());
						} else {
							value = "!" + returnValue.getClass().getName();
						}
					} else if (returnValue instanceof List) {
						value = "List not implemented";
					} else { // Here I assume Integer, Float or other Object
								// representants of basic types.
						if (returnValue.getClass().equals(BOOLEAN_CLASS)) {
							final Boolean b = (Boolean) returnValue;
							final Messages messages = DSAEnv.instance().messages();
							value = messages.getText(b.booleanValue() ? "MasterData.True" : "MasterData.False");
							alignment = TextFormatter.CENTER;
						} else {
							value = returnValue.toString();
							alignment = TextFormatter.RIGHT;
						}
					}
				} catch (final SecurityException e1) {
					e1.printStackTrace();
				} catch (final NoSuchMethodException e1) {
					e1.printStackTrace();
				} catch (final IllegalArgumentException e) {
					e.printStackTrace();
				} catch (final IllegalAccessException e) {
					e.printStackTrace();
				} catch (final InvocationTargetException e) {
					e.printStackTrace();
				}
				s.append(tf.formatTableCell(arrCI[c], value, alignment));
			}
			s.append(tf.endTableRow());
		}
		s.append(tf.endTableBody());
		s.append(tf.endTable());
		return s.toString();
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		return doc.messages().getText(getMessageKey());
	}
}
