package de.ped.dsatool.logic;

import java.util.Vector;

import de.ped.tools.Assert;
import de.ped.tools.MathVector;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextDocumentType;
import de.ped.tools.TextFormatter;
import de.ped.tools.Textable;
import de.ped.dsatool.dsa.generated.CalendarType;
import de.ped.dsatool.dsa.generated.DiceRollType;
import de.ped.dsatool.dsa.generated.LanguageType;
import de.ped.dsatool.dsa.generated.MonthType;
import de.ped.dsatool.dsa.generated.RegionType;
import de.ped.dsatool.dsa.generated.TownType;

/**
 * Holds the concept of what is actually in the output window.
 * As such it is also a facade for lots of user interface actions affecting any part of the model.
 * 
 * @author Copyright (c) 1999 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Model implements Textable {

	public static final String URL = "index.jsp";
	public static final String KEY_MODEL = "model";
	public static final String ACTION = "action"; 
	public static final String ACTION_DATE_SHOW     = "date"; 
	public static final String ACTION_TOWN_SHOW     = "town"; 
	public static final String ACTION_NAMES_SHOW    = "names"; 
	public static final String ACTION_PLANTS_SHOW   = "plants";
	public static final String ACTION_QUEST_SHOW    = "quest"; 
	public static final String ACTION_DICE_SHOW     = "dice"; 
	public static final String ACTION_MASTER_DATA_SHOW = "masterdata"; 
	public static final String ACTION_DATE_NEW      = "date_new"; 
	public static final String ACTION_DATE_UNDO     = "date_undo"; 
	public static final String ACTION_DATE_NEXT     = "date_next"; 
	public static final String ACTION_TOWN_NEW      = "town_new"; 
	public static final String ACTION_TOWN_NEWMAP   = "town_newmap"; 
	public static final String ACTION_NAMES_NEW     = "names_new"; 
	public static final String ACTION_PLANT_NEW     = "plant_new"; 
	public static final String ACTION_QUEST_NEW     = "quest_new";
	public static final String ACTION_DICE_NEW      = "dice_new";
//	public static final String ACTION_FORMAT_TEXT   = "format_text"; 
	public static final String ACTION_FORMAT_HTML   = "format_html"; 
	public static final String ACTION_FORMAT_FIXEDFONT = "format_fixedfont"; 
//	public static final String ACTION_FORMAT_XML    = "format_xml"; 
	public static final String ACTION_MAP_SHOW      = "map";
	public static final String CALENDAR_DAY         = "day"; 
	public static final String CALENDAR_MONTH       = "month"; 
	public static final String CALENDAR_YEAR        = "year"; 
	public static final String CALENDAR_CALENDAR    = "calendar"; 
	public static final String CALENDAR_WEATHER     = "weather"; 
	public static final String CALENDAR_WIND        = "wind"; 
	public static final String CALENDAR_TEMPERATURE = "temperature"; 
//	public static final String CALENDAR_RANDOM_WEATHER = "random_weather"; 
	public static final String CALENDAR_IS_WEATHER_RANDOM         = "random_weather";
	public static final String CALENDAR_IS_WEATHER_RANDOM_RANDOM  = "random_weather_yes";
	public static final String CALENDAR_IS_WEATHER_RANDOM_BY_HAND = "random_weather_no";
	public static final String TOWN_IS_PREDEF       = "is_predefined"; 
	public static final String TOWN_IS_PREDEF_PREDEF= "predefined"; 
	public static final String TOWN_IS_PREDEF_OWN   = "own"; 
	public static final String TOWN_TOWN            = "town"; 
	public static final String TOWN_NAME            = "town_name"; 
	public static final String TOWN_INHABITANTS     = "town_inhabitants"; 
	public static final String TOWN_REGION          = "region"; 
	public static final String NAMES_FOR            = "names_for"; 
	public static final String NAMES_FOR_LANGUAGE   = "names_for_language"; 
	public static final String NAMES_FOR_REGION     = "names_for_region"; 
	public static final String NAMES_FOR_TOWN       = "names_for_town"; 
	public static final String NAMES_FOR_POSITION   = "names_for_position"; 
	public static final String NAMES_LANGUAGE       = "language"; 
	public static final String NAMES_REGION         = "region"; 
	public static final String NAMES_TOWN           = "town"; 
	public static final String NAMES_POSITION       = "position"; 
	public static final String PLANT_CHECK_IS_TO_TEST_FOR_TERRAIN_PROB = "checkIsToTestForTerrainProb"; 
	public static final String PLANT_CHECK_IS_TO_FILTER_FOR_DATE       = "checkIsToFilterForDate"; 
	public static final String PLANT_CHECK_IS_TO_TEST_FOR_BOTANY       = "checkIsToTestForBotany"; 
	public static final String PLANT_MONTH                             = "month"; 
	public static final String PLANT_REGION                            = "region"; 
	public static final String PLANT_TERRAIN                           = "terrain"; 
	public static final String PLANT_TEST_FOR_BOTANY                   = "test_for_botany"; 
	public static final String DICE_FACTOR          = "dice_factor";
	public static final String DICE_OFFSET          = "dice_offset";
	public static final String DICE_NUMBER          = "dice_number";
	public static final String DICE_EYES            = "dice_eyes";
	public static final String MASTER_DATA_MODE     = "master_data_mode"; 

	public static final String MAP_POS_X = "x";
	public static final String MAP_POS_Y = "y";
	public static final String MAP_ZOOM  = "map_zoom";

	/**
	 * The part of the model that is currently displayed.
	 * 
	 * @author Copyright (c) 1999 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum ModelMode {
		DATE, TOWN, NAMES, PLANT, QUEST, DICE, MASTER_DATA;
	}

	public Model(boolean isLocalApplication) {
		// Force loading of the persistent data
		final DSA dsa = DSA.instance();
		// If using the old Java html renderer, stick to old HTML.
		final TextDocumentType doctype;
		if (isLocalApplication) {
			doctype = TextDocumentType.HTML401_TRANSITIONAL;
		} else {
			doctype = TextDocumentType.HTML5;
		}
		textDocumentType = doctype;
		// Now we can initialize the models
		calendarModel = new DateModel();
		namesModel = new NamesModel(this);
		plantsModel = new PlantsModel();
		masterDataModel = new MasterDataModel();
		final String townName = DSAEnv.DEFAULT_TOWN_NAME;
		final int townInh = 36;
		final RegionType region = Region.findByIDorDefault("REGION_BO");
		townModel = new TownModel();
		townModel.setTown(Town.create(townInh, townName, region));
		namesModel.setRegion(region);
		plantsModel.setRegion(region);
		final DSADateCalendarTwelve cal = new DSADateCalendarTwelve(calendarModel.getFirstDay().getDate());
		final int monthIDnum = cal.getMonthAsNumber() - 1;
		final MonthType month = (MonthType) DSA.findBOByIDnum(dsa.getMonth(), monthIDnum);
		plantsModel.setMonth(month);
		plantsModel.recalculate();
		quest = new Quest();
		diceModel = new DiceModel();
		mode = ModelMode.DATE;
	}

	public DateModel getDateModel() {
		return calendarModel;
	}

	public static Vector<String> getAllMonthStrings() {
		return Month.getAllStrings();
	}

	public String[] getWeatherTable() {
		return calendarModel.getWeatherTable();
	}

	public DSADateComplete getFirstDay() {
		return calendarModel.getFirstDay();
	}

	/**
	 * Recalculates the calendar, starting with the given day.
	 * @param day the first day of the calendar period
	 */
	public void setFirstCalendarDay(final DSADateComplete day) {
		calendarModel.setFirstDay(day);
		mode = ModelMode.DATE;
	}

	/**
	 * Sets the first calendar day according to the given data.
	 * @param day the day of the month. In the range of 1 to 30 (or 5, depending of the month).
	 * @param monthID the IDnum of the month. In the range of 0 to 12.
	 * @param year the year, in the given calendar
	 * @param calendarID the IDnum of the calendar
	 * @param weatherID the index of the weather. In the range of 0 to 19.
	 * @param windID the index of the wind. In the range of 0 to 8.
	 * @param temperatureID the index of the temperature. In the range of 0 to 8.
	 * @param isWeatherRandom if <code>true</code>, parameters weatherID, windID and temperatureID are ignored
	 *      and those values chosen randomly
	 * @throws DSADateYearZeroException if the year is zero, and the given calendar supports no year zero
	 */
	public void setFirstCalendarDay(
			final int day, 
			final int monthID, 
			final int year, 
			final int calendarID, 
			final int weatherID, 
			final int windID, 
			final int temperatureID, 
			final boolean isWeatherRandom) throws DSAException {
		final CalendarType calendar = (CalendarType) DSA.findBOByIDnum(DSA.instance().getCalendar(), calendarID);
		final Weather weather = getFirstDay().getWeather();
		weather.setCalendar(calendar);
		final DSADateCalendarTwelve c12 = new DSADateCalendarTwelve(weather.getDate(), calendar.getID());
		c12.setDayMonthYear(day, monthID + 1, year);
		if (isWeatherRandom) {
			weather.setWeatherRandom();
		} else {
			weather.setWeather(weatherID + 1, windID, temperatureID);
		}
		recalcCalendarTimeframe();
	}

	/**
	 * Recalculates the calendar, starting with the same day as before.
	 * But all the following days may change.
	 */
	public void recalcCalendarTimeframe() {
		calendarModel.recalculate();
		mode = ModelMode.DATE;
	}

	/**
	 * Recalculates the calendar, starting with the last day of the actual time frame.
	 */
	public void calcNextCalendarTimeframe() {
		calendarModel.calcNextTimeframe();
		mode = ModelMode.DATE;
	}

	public TownModel getTownModel() {
		return townModel;
	}

	/**
	 * @return the town in the model
	 */
	public TownType getTown() {
		return this.townModel.getTown();
	}

	/**
	 * @param theTown the town to store in the model
	 */
	public void setTown(final TownType theTown) {
		townModel.setTown(theTown);
		mode = ModelMode.TOWN;
	}

	public void setTown(final int predefTownID) {
		setTown(Town.create((TownType) DSA.instance().getTown().get(predefTownID)));
	}

	public void setTown(final String name, final int inhabitants, final int regionID) {
		setTown(Town.create(inhabitants, name, (RegionType) DSA.findBOByIDnum(DSA.instance().getRegion(), regionID)));
	}

	/**
	 * Creates a new random map for the actual town.
	 */
	public void changeTownMap() {
		townModel.changeMap();
		mode = ModelMode.TOWN;
	}

	/**
	 * @return the names model
	 */
	public NamesModel getNamesModel() {
		return namesModel;
	}

	/**
	 * Sets the names model.
	 * @param actualNames the names model
	 */
	public void setNamesModel(final NamesModel actualNames) {
		namesModel = actualNames;
		mode = ModelMode.NAMES;
	}

	public void setNamesModelForLanguageID(final int languageIDnum) {
		namesModel.setLanguage((LanguageType) DSA.findBOByIDnum(DSA.instance().getLanguage(), languageIDnum));
		namesModel.setMode(NamesModel.Mode.LANGUAGE);
		mode = ModelMode.NAMES;
	}

	public void setNamesModelForRegionID(final int regionIDnum) {
		namesModel.setRegion((RegionType) DSA.findBOByIDnum(DSA.instance().getRegion(), regionIDnum));
		namesModel.setMode(NamesModel.Mode.REGION);
		mode = ModelMode.NAMES;
	}

	public void setNamesModelForTownID(final int predefinedTownIdx) {
		namesModel.setTown(predefinedTownIdx);
		namesModel.setMode(NamesModel.Mode.TOWN);
		mode = ModelMode.NAMES;
	}

	public void setNamesModelForPosition() {
		namesModel.setMode(NamesModel.Mode.MAP_POSITION);
		mode = ModelMode.NAMES;
	}

	public String allMapLayerAreasAtCurrentPositionOnAventuriaMap() {
		final Messages messages = DSAEnv.instance().messages();
		final TextDocumentBuilder doc = new TextDocumentBuilder(messages);
		doc.setTextDocumentType(getFormatStyle());
		doc.setEmbedded(true);
		doc.add(Map.allMapLayerAreasAt(currentPositionOnAventuriaMap, doc.tf()));
		return doc.result();
	}

	/**
	 * @return the plants model
	 */
	public PlantsModel getPlantsModel() {
		return plantsModel;
	}

	/**
	 * Sets the plants model.
	 * @param plantsModel the plants model to set
	 */
	public void setPlantsModel(PlantsModel plantsModel) {
		this.plantsModel = plantsModel;
		mode = ModelMode.PLANT;
	}

	public void setPlantsModel(
			final int regionID, 
			final int[] terrainIDs, 
			final boolean isToTestForTerrainProbability, 
			final boolean isToTestForBotany, 
			final int testForBotany, 
			final boolean isToFilterForDate, 
			final int monthID) {
		final DSA dsa = DSA.instance();
		final RegionType region = (RegionType) DSA.findBOByIDnum(dsa.getRegion(), regionID);
		plantsModel.setRegion(region);
		plantsModel.setTerrainIDs(terrainIDs);
		plantsModel.setToTestForTerrainProbability(isToTestForTerrainProbability);
		plantsModel.setToTestForBotany(isToTestForBotany);
		plantsModel.setTestForBotanySuccess(testForBotany);
		plantsModel.setToFilterForDate(isToFilterForDate);
		final MonthType month = (MonthType) DSA.findBOByIDnum(dsa.getMonth(), monthID);
		plantsModel.setMonth(month);
		plantsModel.recalculate();
		mode = ModelMode.PLANT;
	}

	/**
	 * @return the master data model
	 */
	public MasterDataModel getMasterDataModel() {
		return masterDataModel;
	}

	/**
	 * Sets the master data model, and sets the mode to master data mode.
	 */
	public void setMasterDataModel() {
		mode = ModelMode.MASTER_DATA;
	}

	/**
	 * @return the quest in the model
	 */
	public Quest getQuest() {
		return quest;
	}

	/**
	 * Creates a new random quest.
	 */
	public void changeQuest() {
		quest.setAllRandom();
		mode = ModelMode.QUEST;
	}

	/** @return the dice in the model */
	public DiceModel getDice() {
		return diceModel;
	}

	public void rollDice(final DiceRollType dr) {
		diceModel.setInput(dr);
		mode = ModelMode.DICE;
	}

	/**
	 * @return the model mode
	 */
	public ModelMode getMode() {
		return mode;
	}

	/**
	 * @param theMode the new model mode
	 */
	public void setMode(final ModelMode theMode) {
		mode = theMode;
	}

	/**
	 * @return the actual model object, depending on the model model. May be a
	 *         calendar model, a town, a names model, a plants search model, a
	 *         quest, or a master data model.
	 */
	public Textable getActualModelObject() {
		switch (mode) {
		case DATE:
			return getDateModel();
		case TOWN:
			return townModel;
		case NAMES:
			return namesModel;
		case QUEST:
			return getQuest();
		case PLANT:
			return getPlantsModel();
		case DICE:
			return getDice();
		case MASTER_DATA:
			return getMasterDataModel();
		default:
			Assert.assertTrue(false);
			return null;
		}
	}

	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		return getActualModelObject().getContentText(doc);
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		return getActualModelObject().getHeadingText(doc);
	}

	public boolean isMap() {
		return isMap;
	}

	public void setMap(final boolean value) {
		isMap = value;
	}

	private boolean isMap = false;

	private AventuriaMapModel aventuriaMapModel;

	public AventuriaMapModel getOrCreateAventuriaMapModel() {
		if (null == aventuriaMapModel) {
			aventuriaMapModel = new AventuriaMapModel();
		}
		return aventuriaMapModel;
	}

	/** Hotspot starts in the middle of the image */
	private MathVector currentPositionOnAventuriaMap = MathVector.create(0.5, 0.5);

	public MathVector getCurrentPositionOnAventuriaMap() {
		return currentPositionOnAventuriaMap;
	}

	public void setCurrentPositionOnAventuriaMap(final MathVector currentPositionOnAventuriaMap) {
		this.currentPositionOnAventuriaMap = currentPositionOnAventuriaMap;
	}

	public static String[] getAventuriaMapZoomStepIndexStrings() {
		return AventuriaMapModel.getZoomStepIndexStrings();
	}

	public int getAventuriaMapZoomStepIndex() {
		return getOrCreateAventuriaMapModel().getZoomStepIndex();
	}

	public String write(final String title) {
		final Messages messages = DSAEnv.instance().messages();
		final TextDocumentBuilder doc = new TextDocumentBuilder(messages);
		doc.setTextDocumentType(getFormatStyle());
		doc.setEmbedded(false);
		doc.setInventoryMaxDepth(0);
		doc.setTitle(getHeadingText(doc));
		doc.setHeading(null, null);
		doc.beginDocument();
		writeAsHtmlSnippet(doc, this, 1, true);
		doc.endDocument();
		String content = doc.result(null);
		return content;
	}

	public String writeAsHtmlSnippet(final int headingDepth) {
		final Messages messages = DSAEnv.instance().messages();
		final TextDocumentBuilder doc = new TextDocumentBuilder(messages);
		doc.setTextDocumentType(getFormatStyle());
		doc.setEmbedded(true);
		doc.setInventoryMaxDepth(0);
		doc.setMinIndent(2);
		writeAsHtmlSnippet(doc, this, headingDepth, true);
		String content = doc.result(null);
		return content;
	}

	/**
	 * Writes an HTML code snippet containing the text description of the
	 * textable into <em>doc</em>.
	 * 
	 * @param doc
	 *            the document to write into
	 * @param textable
	 *            the textable to write
	 * @param headingDepth
	 *            the heading will be created with this heading depth
	 * @param isEmbedded
	 *            if it is embedded in a HTML page - if yes, pure text will be
	 *            surrounded by HTML preformatted tag
	 */
	protected void writeAsHtmlSnippet(final TextDocumentBuilder doc, final Textable textable, final int headingDepth,
			final boolean isEmbedded) {
		final TextFormatter tf = doc.tf();
		doc.add(tf.addSectionHeading(headingDepth, getHeadingText(doc)));
		doc.add(getContentText(doc));
	}

	public static String formatPosition(final MathVector pos) {
		return formatPosition(pos.get(MathVector.X), pos.get(MathVector.Y));
	}

	public static String formatPosition(final double x, final double y) {
		final StringBuffer buf = new StringBuffer("(");
		buf.append(TextFormatter.format((int) (1000 * x), 4));
		buf.append(", ");
		buf.append(TextFormatter.format((int) (1000 * y), 4));
		buf.append(")");
		return buf.toString();
	}

	public TextDocumentType getFormatStyle() {
		return textDocumentType;
	}

	public void setFormatStyle(TextDocumentType textDocumentType) {
		this.textDocumentType = textDocumentType;
	}

	private TextDocumentType textDocumentType;

	/** The part of the model keeping a calendar. */
	private DateModel calendarModel;
	/** The part of the model keeping a town. */
	private TownModel townModel;
	/** The part of the model keeping a quest. */
	private Quest quest;
	/** The part of the model keeping dices to roll. */
	private DiceModel diceModel;
	/** The part of the model keeping names. */
	private NamesModel namesModel;
	/** The part of the model keeping plants. */
	private PlantsModel plantsModel;
	/** The part of the model keeping plants. */
	private MasterDataModel masterDataModel;
	/** Which part of the model is actually the active part and displayed? */
	private ModelMode mode;
}
