package de.ped.dsatool.logic;

import de.ped.tools.Assert;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.MathVector;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.Textable;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.LanguageDistributionType;
import de.ped.dsatool.dsa.generated.LanguageType;
import de.ped.dsatool.dsa.generated.MapLayerAreaType;
import de.ped.dsatool.dsa.generated.PersonType;
import de.ped.dsatool.dsa.generated.RegionType;
import de.ped.dsatool.dsa.generated.SexType;
import de.ped.dsatool.dsa.generated.TownType;

/**
 * A model to create names of Aventurian characters.
 * The model can create names of a given language,
 * or of a given region (based on the language distribution of that region).
 * The gender distribution is taken into account, so e.g. for dwarves you'll get more male names.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class NamesModel implements Cloneable, Textable {
	@SuppressWarnings("unused")
	private static Logger logger = DSAEnv.instance().logger();

	/**
	 * The kind of object the names are created for.
	 *
	 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Mode {
		/** The names are created only in the given language. */
		LANGUAGE,
		/** The names are created following the language distribution of that region. */
		REGION,
		/** The names are created following the language distribution of that town. */
		TOWN,
		/** The names are created following the language distribution of that position on the Aventuria map. */
		MAP_POSITION,
	};

	private Class<?>[] TOWN_OR_REGION = new Class<?>[] { TownType.class, RegionType.class };

	/**
	 * A model to create names of Aventurian characters.
	 * The model can create names of a given language,
	 * or of a given region (based on the language distribution of that region).
	 * The gender distribution is taken into account, so e.g. for dwarves you'll get more male names.
	 *
	 * @author Copyright (c) 2007 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private class AventuriaMapPositionData {
		MathVector aventuriaMapPosition;
		Mode mode = Mode.LANGUAGE;
		RegionType region = null;
		TownType town = null;
		LanguageDistributionType langDist = null;

		AventuriaMapPositionData() {
			setToPosition(MathVector.create(0.5, 0.5));
		}

		void setToPosition(final MathVector pos) {
			aventuriaMapPosition = pos;
			MapLayerAreaType area = Map.firstAt(pos, TOWN_OR_REGION);
			if (null != area) {
				final Object ref = area.getRef();
				if (ref instanceof RegionType) {
					region = (RegionType) ref;
					langDist = region.getLanguageDistribution();
					mode = Mode.REGION;
				} else if (ref instanceof TownType) {
					town = (TownType) ref;
					langDist = Town.getOrCreateLanguageDistribution(town);
					mode = Mode.TOWN;
				}
			} else {
				mode = Mode.LANGUAGE;
				langDist = LanguageDistribution.createDefaultLanguageDistribution();
			}
		}

		String getModeContent() {
			final String ret;
			final Messages messages = DSAEnv.instance().messages();
			switch (mode) {
			case REGION:
				ret = messages.getText(new I18NStringWithFillIns("NamesModel.Position.Description.Region", "%0 %1",
						new String[] { messages.getText("NamesDialog.Region"), region.getName(), }));
				break;
			case TOWN:
				ret = messages.getText(new I18NStringWithFillIns("NamesModel.Position.Description.Town", "%0 %1",
						new String[] { messages.getText("NamesDialog.Town"), town.getName(), }));
				break;
			default:
				ret = messages.getText("NamesModel.Position.Description.Undefined");
			}
			return ret;
		}
	}

	/** How many names to keep? */
	private static final int NUM_OF_NAMES = 20;

	/** 
	 * Creates a names model for the region Bornland.
	 * @param parent the parent model 
	 */
	public NamesModel(final Model parent) {
		this.parent = parent;
		recalculateNames();
	}

	/**
	 * Returns what kind of names are created.
	 * @return one of @link{#MODE_LANGUAGE}, @link{#MODE_REGION}, @link{#MODE_TOWN} or @link{#MODE_MAP_POSITION}
	 */
	public Mode getMode() {
		return mode;
	}

	/**
	 * Tells what kind of names will be created, and creates names of that kind.
	 * @param mode should be one of @link{#MODE_LANGUAGE}, @link{#MODE_REGION}, @link{#MODE_TOWN} or @link{#MODE_MAP_POSITION}
	 */
	public void setMode(final Mode mode) {
		switch(mode) {
		case LANGUAGE:
		case REGION:
		case TOWN:
		case MAP_POSITION:
			this.mode = mode;
			recalculateNames();
			break;
		default:
			throw new ArrayIndexOutOfBoundsException();
		}
	}

	public LanguageType getLanguage() {
		return language;
	}

	/** 
	 * Sets the language, and creates names of that language.
	 * @param language the language to set 
	 */
	public void setLanguage(final LanguageType language) {
		this.language = language;
		recalculateNames();
	}

	public RegionType getRegion() {
		return region;
	}

	/** 
	 * Sets the region, and creates names of that region. 
	 * @param region the region to set
	 */
	public void setRegion(final RegionType region) {
		this.region = region;
		recalculateNames();
	}

	public TownType getTown() {
		return town;
	}

	/** 
	 * Sets the town (by cloning and filling up), and creates names of that town.
	 * @param town the town to set
	 */
	public void setTown(final TownType town) {
		this.town = Town.create(town);
		recalculateNames();
	}

	/** 
	 * Sets the predefined town (by cloning and filling up), and creates names of that town. 
	 * @param predefinedTownIdx the index of the predefined town
	 */
	public void setTown(final int predefinedTownIdx) {
		TownType town = DSA.instance().getTown().get(predefinedTownIdx);
		setTown(town);
	}

	public int getPredefinedTownIdx() {
		int ret = -1;
		if (null != town) {
			ret = Town.findPredefinedTownIdx(town.getName());
		}
		return ret;
	}

	public MathVector getAventuriaMapPosition() {
		return aventuriaMapPositionData.aventuriaMapPosition;
	}

	public void setAventuriaMapPosition(final MathVector aventuriaMapPosition) {
		aventuriaMapPositionData.setToPosition(aventuriaMapPosition);
		recalculateNames();
	}

	public void setAventuriaMapPositionFromDataModel() {
		setAventuriaMapPosition(parent.getCurrentPositionOnAventuriaMap());
	}

	/**
	 * Calculates new names, based on mode and the other information.
	 */
	public void recalculateNames() {
		LanguageDistributionType langDist = null;
		switch(mode) {
		case LANGUAGE:
			langDist = LanguageDistribution.createSingleTongueLanguageDistribution(language.getIDnum());
			break;
		case REGION:
			langDist = region.getLanguageDistribution();
			break;
		case TOWN:
			langDist = town.getLanguageDistribution();
			break;
		case MAP_POSITION:
			langDist = aventuriaMapPositionData.langDist;
			break;
		default:
			Assert.assertTrue(false);
		}
		Assert.assertTrue(null != langDist);
		for (int i = 0; i < NUM_OF_NAMES; i++) {
			final LanguageType lang = LanguageDistribution.chooseLanguage(langDist);
			names[i] = Person.createPerson(lang);
		}
	}

	/** 
	 * @param isToEnrichByRegion if <code>true</code>, information about the best fitting region or town is given
	 * @return a string explaining where on the map we are. If a region or town fits, this is listed as well. 
	 */
	public String getAventuriaMapPositionString(final boolean isToEnrichByRegion) {
		final String ret;
		final String pos = Model.formatPosition(aventuriaMapPositionData.aventuriaMapPosition);
		if (isToEnrichByRegion) {
			final Messages messages = DSAEnv.instance().messages();
			ret = messages.getText(new I18NStringWithFillIns("NamesModel.Position.Description", "%0 - %1",
					new String[] { pos, aventuriaMapPositionData.getModeContent(), }));
		} else {
			ret = pos;
		}
		return ret;
	}

	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		final StringBuffer ret = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final Messages messages = doc.messages();
		final String namesHeadingKey;
		final int depth = 2 + 1;
		final DSA dsa = DSA.instance();
		switch(mode) {
		case LANGUAGE:
			// Nothing to do.
			namesHeadingKey = "NamesDialog.For.Language";
			break;
		case REGION:
			ret.append(tf.addSectionHeading(depth, messages.getText("NamesModel.LanguageDistribution.Region"))); 
			ret.append(dsa.getContentText(doc, (LanguageDistributionType)region.getLanguageDistribution()));
			namesHeadingKey = "NamesDialog.For.Region";
			break;
		case TOWN:
			ret.append(tf.addSectionHeading(depth, messages.getText("NamesModel.LanguageDistribution.Town")));
			ret.append(dsa.getContentText(doc, (LanguageDistributionType) town.getLanguageDistribution()));
			namesHeadingKey = "NamesDialog.For.Town";
			break;
		case MAP_POSITION:
			ret.append(tf.beginParagraph());
			ret.append(getAventuriaMapPositionString(true));
			ret.append(tf.endParagraph());
			ret.append(tf.newLineIfPlainText());
			ret.append(tf.addSectionHeading(depth, messages.getText("DSA.LanguageDistribution")));
			ret.append(dsa.getContentText(doc, aventuriaMapPositionData.langDist));
			namesHeadingKey = "NamesDialog.For.MapPosition";
			break;
		default:
			namesHeadingKey = null;
			Assert.assertTrue(false);
		}
		ret.append(tf.newLineIfPlainText());
		ret.append(tf.addSectionHeading(depth, messages.getText(namesHeadingKey)));
		final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[] {
				new TextFormatter.ColumnInfo(8, TextFormatter.LEFT, messages, "NamesModel.Properties.Heading"),
				new TextFormatter.ColumnInfo(50, TextFormatter.LEFT, messages, "NamesModel.Name.Heading"), };
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
		ret.append(tf.beginTable(null, ti));
		ret.append(tf.formatTableHeading(ci));
		ret.append(tf.beginTableBody());
		for (int i = 0; i < NUM_OF_NAMES; i++) {
			final PersonType name = names[i];
			final LanguageType language = (LanguageType) name.getLanguage();
			final SexType sex = (SexType) name.getSex();
			final String langAndSex = messages.getText(new I18NStringWithFillIns("NamesModel.Properties", "(%0, %1)",
					new String[] { 
							tf.href(MasterDataModel.getHRefFor(sex), sex.getAbbr()),
							tf.href(MasterDataModel.getHRefFor(language), language.getAbbr()), }));
			ret.append(tf.beginTableRow());
			ret.append(tf.formatTableCell(ci[0], langAndSex));
			ret.append(tf.formatTableCell(ci[1], name.getName()));
			ret.append(tf.endTableRow());
		}
		ret.append(tf.endTableBody());
		ret.append(tf.endTable());
		return ret.toString();
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		final StringBuffer ret = new StringBuffer();
		final DSA dsa = DSA.instance();
		final Messages messages = doc.messages();
		final TextFormatter tf = doc.tf();
		switch(mode) {
		case LANGUAGE:
			ret.append(messages.getText(new I18NStringWithFillIns("NamesModel.For.Language", null,
					new String[] { dsa.getContentText(language, tf), })));
			break;
		case REGION:
			ret.append(messages.getText(new I18NStringWithFillIns("NamesModel.For.Region", null,
					new String[] { dsa.getContentText(region, tf), })));
			break;
		case TOWN:
			ret.append(messages.getText(new I18NStringWithFillIns("NamesModel.For.Town", null,
					new String[] { dsa.getHeadingText(town, tf), })));
			break;
		case MAP_POSITION:
			ret.append(messages.getText(new I18NStringWithFillIns("NamesModel.For.MapPosition", null,
					new String[] { aventuriaMapPositionData.aventuriaMapPosition.toString(), })));
			break;
		default:
			Assert.assertTrue(false);
		}
		return ret.toString();
	}

	@Override
	public NamesModel clone() {
		NamesModel ret = null;
		try {
			ret = (NamesModel)super.clone();
			for (int i = 0; i < NUM_OF_NAMES; i++) {
				ret.names[i] = Person.createPerson(names[i]);
			}
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}

	/** The mode telling what names to create. */
	private Mode mode = Mode.REGION;

	/** The language the names base upon. Only set in MODE_LANGUAGE. */
	private LanguageType language = Language.findByID("LANGUAGE_BORNLANDISH"); 

	/** The region the names base upon. Only set in MODE_REGION. */
	private RegionType region = Region.findByIDorDefault("REGION_BO"); 

	/** The town the names base upon. Only set in MODE_TOWN. */
	private TownType town = null;
	
	/** The position on the Aventuria map the names base upon. */
//    private Coord2D aventuriaMapPosition = new Coord2D(0.5, 0.5);

	/** The area found at the Aventuria map position. */
	private AventuriaMapPositionData aventuriaMapPositionData = new AventuriaMapPositionData(); 

	/** The names created. */
	private PersonType names[] = new PersonType[NUM_OF_NAMES];

	private Model parent;
}
