package de.ped.dsatool.logic;

import java.util.Iterator;
import java.util.List;

import de.ped.dsatool.dsa.generated.LanguageType;
import de.ped.dsatool.dsa.generated.PersonNamePartType;
import de.ped.dsatool.dsa.generated.PersonNamePatternType;
import de.ped.dsatool.dsa.generated.PersonNameType;
import de.ped.dsatool.dsa.generated.PersonType;
import de.ped.dsatool.dsa.generated.SexType;
import de.ped.tools.CollectionUtil;
import de.ped.tools.MathUtil;

/**
 * This class represents a name of a person (and the characteristics that lead to that name).
 * Names in DSA depend on the sex and the language of the person.
 *
 * <p><b>See:</b> "???", Section "???" (p. ??..??)</p>
 *
 * @author Copyright (c) 1999 Peter Diefenbach (peter@pdiefenbach.de),
 *		based on algorithms by Urs Beeli
 */
public abstract class Person {
	/**
	 * @return a Garethian person of random sex.
	 */
	public static PersonType createPerson() {
		final PersonType ret = new PersonType();
		final LanguageType lang = Language.findDefault();
		ret.setLanguage(lang);
		ret.setSex(randomSex(lang));
		recalculateName(ret);
		return ret;
	}

	/**
	 * @param language
	 *            The language the name is in.
	 * @return a person of random sex and given language.
	 */
	public static PersonType createPerson(final LanguageType language) {
		return createPerson(randomSex(language), language);
	}

	/**
	 * @param sex
	 *            The sex of the person.
	 * @param language
	 *            The language the name is in.
	 * @return a person of given sex and language.
	 */
	public static PersonType createPerson(final SexType sex, final LanguageType language) {
		final PersonType p = DSA.instance().createPersonType();
		p.setLanguage(language);
		p.setSex(sex);
		recalculateName(p);
		return p;
	}

	/**
	 * Clones <em>p</em> in content, but not in ID.
	 * 
	 * TODO: Handle ID in a proper way.
	 * 
	 * @param person
	 *            The person to clone.
	 * @return The copy of <em>p</em>.
	 */
	public static PersonType createPerson(final PersonType person) {
		final PersonType ret = createPerson();
		ret.setSex(person.getSex());
		ret.setLanguage(person.getLanguage());
		ret.setName(person.getName());
		return ret;
	}

	/**
	 * @param person
	 *            the person to define the name for
	 * @return The name of this person.
	 */
	public static String getName(final PersonType person) {
		String name = person.getName();
		if (CollectionUtil.isNullOrEmpty(name)) {
			person.setName(calculateName((SexType)person.getSex(), (LanguageType)person.getLanguage()));
			name = person.getName();
		}
		return name;
	}

	/**
	 * Set the sex of a person. If this changes the object, the name will be
	 * recalculated.
	 * 
	 * @param person
	 *            the person to define the language for
	 * @param aSex
	 *            the new sex of the person
	 */
	public static void setSex(final PersonType person, final SexType aSex) {
		if (!person.getSex().equals(aSex)) {
			person.setSex(aSex);
			person.setName(null);
		}
	}

	/**
	 * Sets the language the name of the person is in. If this changes the
	 * object, the name will be recalculated.
	 * 
	 * @param person
	 *            the person to define the language for
	 * @param aLanguage
	 *            The language the name of the person is in.
	 */
	public static void setLanguage(final PersonType person, final LanguageType aLanguage) {
		if (!person.getLanguage().equals(aLanguage)) {
			person.setLanguage(aLanguage);
			person.setName(null);
		}
	}

	/**
	 * Chooses the sex randomly. The language is given to determine the race or
	 * nation, as some races don't have the sexes equally distributed.
	 * 
	 * @param language
	 *            The language the name of the person is in.
	 * @return the chosen sex
	 */
	private static SexType randomSex(final LanguageType language) {
		// Dwarfs have much lesser women
		return randomSex(language.getFemalePercentage());
	}

	/**
	 * Chooses the sex randomly.
	 * 
	 * @param femalePercent
	 *            The percentage of females in the population (from 0.0 to 100.0)
	 * @return the chosen sex
	 */
	public static SexType randomSex(final float femalePercent) {
		final int sexIndex = (MathUtil.random(100.0F) < femalePercent ? 1 : 0);
		return (SexType)DSA.instance().getSex().get(sexIndex);
	}

	/**
	 * Recalculates the name of the person, based on sex and language.
	 * 
	 * @param person
	 *            The person that gets a new name.
	 */
	public static void recalculateName(final PersonType person) {
		String name = calculateName((SexType)person.getSex(), (LanguageType)person.getLanguage());
		person.setName(name);
	}

	/**
	 * Calculates and generate the full name of a person.
	 * 
	 * @param sex
	 *            The sex of the person.
	 * @param language
	 *            The language the name of the person is in.
	 * @return the generated name
	 */
	private static String calculateName(final SexType sex, final LanguageType language) {
		final StringBuffer s = new StringBuffer();
		// Find the right build pattern for that kind of person
		PersonNamePatternType buildPattern = null;
		final PersonNameType personName = DSA.instance().getPersonName();
		if ((null != personName) && (null != personName.getName())) {
			final Iterator<PersonNamePatternType> it = personName.getName().iterator();
			while (it.hasNext() && (null == buildPattern)) {
				final PersonNamePatternType pattern = (PersonNamePatternType)it.next();
				final LanguageType patternLanguage = (LanguageType)pattern.getLanguage();
				final SexType patternSex = (SexType)pattern.getSex();
				if ((patternLanguage.getIDnum().equals(language.getIDnum()))
						&& (patternSex.getIDnum().equals(sex.getIDnum()))) {
					buildPattern = pattern;
				}
			}
		}
		if (null != buildPattern) {
			// Build the name following the build pattern
			final Iterator<?> partIt = buildPattern.getNamePart().iterator();
			while (partIt.hasNext()) {
				final PersonNamePartType personNamePart = DSA.toPersonNamePartValue(partIt.next());
				final List<String> parts = personNamePart.getPart();
				if (!parts.isEmpty()) {
					final int index = MathUtil.random(parts.size());
					s.append(parts.get(index));
				}
			}
		}
		return (s.toString());
	}
}
