package de.ped.dsatool.logic;

import java.util.Iterator;
import java.util.List;

import de.ped.dsatool.dsa.generated.MonthIntervalType;
import de.ped.dsatool.dsa.generated.MonthType;
import de.ped.dsatool.dsa.generated.PlantType;
import de.ped.dsatool.dsa.generated.RegionListType;
import de.ped.dsatool.dsa.generated.RegionType;
import de.ped.dsatool.dsa.generated.TerrainProbabilityType;
import de.ped.dsatool.dsa.generated.TerrainType;
import de.ped.tools.CollectionUtil;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.log.Logger;

/**
 * The different plants interesting to look for, e.g., healing or poisonous plants.
 * The plant data is following the Herbarium Aventuricum from 1997.
 *
 * @author Copyright (c) 2004 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Plant {
	private static Logger logger = Logger.getRootLogger();

	public static List<PlantType> findAll() {
		return DSA.instance().getPlant();
	}

	public static PlantType findByIDnum(final int id) {
		return (PlantType)(DSA.findBOByIDnum(DSA.instance().getPlant(), id));
	}

	public static PlantType findByName(final String name) {
		return (PlantType)(DSA.findBOByName(DSA.instance().getPlant(), name));
	}

	public static PlantType findByID(final String id) {
		return (PlantType)(DSA.findBOByID(DSA.instance().getPlant(), id));
	}

	/**
	 * @param plant
	 *            the plant to look at
	 * @return a human-readable description of the harvest time, and sometimes
	 *         what parts of the plant are usually harvested. It is built from
	 *         the month intervals and (if given) from the harvest time
	 *         description field.
	 */
	public static String getHarvestTimeDescription(final PlantType plant) {
		final Messages messages = DSAEnv.instance().messages();
		final StringBuffer buf = new StringBuffer();
		final Iterator<MonthIntervalType> it = plant.getHarvestTime().iterator();
		boolean isFirst = true;
		while (it.hasNext()) {
			final MonthIntervalType interval = (MonthIntervalType)it.next();
			if (isFirst) {
				isFirst = false;
			} else {
				buf.append(", ");
			}
			final MonthType startMonth = (MonthType)interval.getStartMonth();
			final MonthType endMonth = (MonthType)interval.getEndMonth();
			if ((null != endMonth) && (endMonth.getIDnum() - startMonth.getIDnum() == 12)) {
				buf.append(messages.getText("Plant.AllYear")); 
			} else {
				if ((null != endMonth) && (!startMonth.getID().equals(endMonth.getID()))) {
					buf.append(messages.getText(new I18NStringWithFillIns("Plant.MonthsSpan", null,
							new String[] { startMonth.getName(), endMonth.getName(), })));
				} else { // Single month
					buf.append(startMonth.getName());
				}
			}
			final String remark = interval.getRemark();
			if (!CollectionUtil.isNullOrEmpty(remark)) {
				buf.append(" ");
				buf.append(remark);
			}
		}
		final String desc = plant.getHarvestTimeDescription();
		if (!CollectionUtil.isNullOrEmpty(desc)) {
			buf.append(" - ");
			buf.append(desc);
		}
		return buf.toString();
	}

	/**
	 * @param plant
	 *            the plant to look at
	 * @param month
	 *            The month of the year.
	 * @return If the plant can be harvested in the given month.
	 */
	public static boolean canBeHarvestedIn(final PlantType plant, final MonthType month) {
		final Iterator<MonthIntervalType> it = plant.getHarvestTime().iterator();
		while (it.hasNext()) {
			final MonthIntervalType interval = (MonthIntervalType)it.next();
			final MonthType startMonth = (MonthType)interval.getStartMonth();
			MonthType endMonth = (MonthType)interval.getEndMonth();
			if (null == endMonth) {
				endMonth = startMonth;
			}
			if (startMonth.getIDnum() <= endMonth.getIDnum()) {
				logger.debug("Checking " + plant.getName() + ": " + startMonth + "<= " + month + "<=" + endMonth); 
				// The interval does not cross the year boundary.
				if ((startMonth.getIDnum() <= month.getIDnum()) && (month.getIDnum() <= endMonth.getIDnum())) {
					logger.debug(plant.getName() + ": " + startMonth + "<= " + month + "<=" + endMonth); 
					return true;
				}
			} else {
				// The interval does cross the year boundary.
				if ((startMonth.getIDnum() <= month.getIDnum()) || (month.getIDnum() <= endMonth.getIDnum())) {
					logger.debug(plant.getName() + ": " + startMonth + "<- " + month + "<-" + endMonth); 
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Tests if this plant is there where the parameters describe. Is is there
	 * if
	 * <ol>
	 * <li>The plant grows in the given region,</li>
	 * <li>The plant grows in the given terrains (with any likelyhood &gt; 0),</li>
	 * <li>A test thrown against the terrain likelyhood is successfull.</li>
	 * </ol>
	 * 
	 * @param plant
	 *            the plant to look at
	 * @param region
	 *            The region where the hero is.
	 * @param terrainIDs
	 *            The terrains where the hero looks for plants.
	 * @param isToTestForTerrainProb
	 *            If <code>true</code>, throws dices against the terrain probability for the
	 *            plant. If <code>false</code>, passes the test.
	 * @return An array containing the terrain ID at index 0 and the terrain
	 *         probability at index 1. If nothing found, <code>null</code>.
	 */
	public static int[] isThere(final PlantType plant, final RegionType region, final int[] terrainIDs, final boolean isToTestForTerrainProb) {
		// Are we in the right region?
		boolean isInRightRegion = false;
		final RegionListType regionList = plant.getRegionList();
		final List<Object> regions = regionList.getRegions();
		if (null != regions) {
			if (regionList.isIsInverse()) {
				isInRightRegion = !regions.contains(region);
			} else {
				isInRightRegion = regions.contains(region);
			}
		}
		if (isInRightRegion) {
			// Are we in the right terrain?
			if ((null != terrainIDs) && (null != plant.getTerrainProbability())) {
				/*
				Set terrainSet = new HashSet(terrainIDs.length);
				for (int i = 0; i < terrainIDs.length; i++) {
					terrainSet.add(Terrain.findByID(terrainIDs[i]));
				}
				Set depositSet = new HashSet(deposit.length);
				for (int i = 0; i < terrainIDs.length; i++) {
					depositSet.add(Terrain.findByID(deposit[i][0]));
				}
				Set intersection = new HashSet(terrainSet);
				intersection.retainAll(depositSet);
				*/
				final Iterator<TerrainProbabilityType> it = plant.getTerrainProbability().iterator();
				while (it.hasNext()) {
					final TerrainProbabilityType deposit = (TerrainProbabilityType)it.next();
					final TerrainType depositTerrain = (TerrainType)deposit.getTerrain();
					final int depositTerrainIdx = depositTerrain.getIDnum();
					for (int terrainIdx = 0; terrainIdx < terrainIDs.length; terrainIdx++) {
						if (depositTerrainIdx == terrainIDs[terrainIdx]) {
							// Okay, we have an intersection.
							if (isToTestForTerrainProb) {
								// Is the terrain probability good enough?
								final int diceEyes = Dice.roll(1, 20);
								if (diceEyes <= deposit.getProbability()) {
									// It is!
									int[] ret = new int[2];
									ret[0] = depositTerrainIdx;
									ret[1] = deposit.getProbability();
									return ret;
								}
							} else {
								int[] ret = new int[2];
								ret[0] = depositTerrainIdx;
								ret[1] = deposit.getProbability();
								return ret;
							}
						}
					}
				}
			}
		}
		// Nothing found so far
		return null;
	}

	/**
	 * Tests if this plant is there where the parameters describe. Is is there
	 * if
	 * <ol>
	 * <li>The plant grows in the given region,</li>
	 * <li>The plant grows in the given terrains (with any likelyhood &gt;
	 * 0),</li>
	 * <li>A test thrown against the terrain likelyhood is successfull,</li>
	 * <li>The plant can be harvested at the given time of year.</li>
	 * </ol>
	 * 
	 * @param plant
	 *            the plant to look at
	 * @param region
	 *            The region where the hero is.
	 * @param terrainIDs
	 *            The terrains where the hero looks for plants.
	 * @param isToTestForTerrainProb
	 *            If <code>true</code>, throws dices against the terrain probability for the
	 *            plant. If <code>false</code>, passes the test.
	 * @param month
	 *            The month the heros look for plants, or <code>null</code> if
	 *            unspecified.
	 * @return An array containing the terrain ID at index 0 and the terrain
	 *         probability at index 1. If nothing found, <code>null</code>.
	 */
	public static int[] isThere(final PlantType plant, final RegionType region, final int[] terrainIDs, final boolean isToTestForTerrainProb, final MonthType month) {
		int[] terrainInfo = isThere(plant, region, terrainIDs, isToTestForTerrainProb);
		if (null != terrainInfo) {
			if ((null != month) && (!Plant.canBeHarvestedIn(plant, month))) {
				// Test failed
				terrainInfo = null;
			} else {
				// Test passed
			}
		}
		return terrainInfo;
	}

	/**
	 * Tests if this plant is there where the parameters describe. Is is there
	 * if
	 * <ol>
	 * <li>The plant grows in the given region,</li>
	 * <li>The plant grows in the given terrains (with any likelyhood &gt; 0),</li>
	 * <li>A test thrown against the terrain likelyhood is successfull,</li>
	 * <li>The test result of the hero is good enough for the plant,</li>
	 * <li>The plant can be harvested at the given time of year.</li>
	 * </ol>
	 * 
	 * @param plant
	 *            the plant to look at
	 * @param region
	 *            The region where the hero is.
	 * @param terrainIDs
	 *            The terrains where the hero looks for plants.
	 * @param isToTestForTerrainProb
	 *            If <code>true</code>, throws dices against the terrain
	 *            probability for the plant. If <code>false</code>, passes the
	 *            test.
	 * @param month
	 *            The month the heros look for plants, or <code>null</code> if
	 *            unspecified.
	 * @param testResult
	 *            The result of the test of the hero for botany.
	 * @return An array containing the terrain ID at index 0 and the terrain
	 *         probability at index 1. If nothing found, <code>null</code>.
	 */
	public static int[] isThere(final PlantType plant, final RegionType region, final int[] terrainIDs, final boolean isToTestForTerrainProb, final MonthType month, final int testResult) {
		int[] terrainInfo = isThere(plant, region, terrainIDs, isToTestForTerrainProb, month);
		if (null != terrainInfo) {
			if (plant.getHowHardToFind() <= testResult) {
				// Test passed
			} else {
				// Test failed
				terrainInfo = null;
			}
		}
		return terrainInfo;
	}

	/**
	 * Delivers a list of plants to find if some hero tests for botany somewhere
	 * in Aventuria.
	 * 
	 * @param region
	 *            The region where the hero is.
	 * @param terrainIDs
	 *            The terrains where the hero looks for plants.
	 * @param testResult
	 *            The result of the test of the hero for botany.
	 * @param isToTestForTerrainProb
	 *            If <code>true</code>, throws dices against the terrain probability for the
	 *            plant. If <code>false</code>, passes the test.
	 * @param isToTestForBotany
	 *            If <code>true</code>, throws dices against <em>testResult</em>. If <code>false</code>,
	 *            passes the test.
	 * @param month
	 *            The month the heros look for plants, or <code>null</code> if
	 *            unspecified.
	 * @return An two-dimensional array. The first dimension defines the plant -
	 *         the index referes to the plant with this ID. The second dimension
	 *         describes at index 0 the terrain ID and the terrain probability
	 *         at index 1. If nothing found for a plant, <code>null</code>.
	 */
	public static int[][] whatPlantsAreThere(final RegionType region, final int[] terrainIDs, final int testResult, final boolean isToTestForTerrainProb, final boolean isToTestForBotany, final MonthType month) {
		final List<PlantType> plants = DSA.instance().getPlant();
		final int size = plants.size();
		final int[][] ret = new int[size][2];
		final Iterator<PlantType> it = plants.iterator();
		for (int i = 0; i < size; i++) {
			final PlantType plant = (PlantType)it.next();
			final int[] terrainInfo;
			if (isToTestForBotany) {
				terrainInfo = isThere(plant, region, terrainIDs, isToTestForTerrainProb, month, testResult);
			} else {
				terrainInfo = isThere(plant, region, terrainIDs, isToTestForTerrainProb, month);
			}
			ret[i] = terrainInfo;
		}
		return ret;
	}
}
