package de.ped.dsatool.logic;

import de.ped.dsatool.dsa.generated.MonthType;
import de.ped.dsatool.dsa.generated.PlantCategoryType;
import de.ped.dsatool.dsa.generated.PlantType;
import de.ped.dsatool.dsa.generated.RegionType;
import de.ped.dsatool.dsa.generated.TerrainType;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.Textable;

/**
 * Model for the Plants/Botany dialog.
 *
 * @author Copyright (c) 2004 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PlantsModel implements Cloneable, Textable {
	/** The region the party is searching for plants. */
	private RegionType region = Region.findByIDorDefault("REGION_BO"); 

	/** The IDs of the terrains where the group looks for plants. */
	private int[] terrainIDs;

	/** Is to test for the terrain probability of the plant? */
	private boolean isToTestForTerrainProbability = false;

	/**
	 * Is to test for the party's ability for botany?
	 * @see #testForBotanySuccess
	 */
	private boolean isToTestForBotany = false;

	/** How good was the party's test for botany? */
	private int testForBotanySuccess = 15;

	/**
	 * Is the search to be filtered for the date (month) when the search happens?
	 * @see #month
	 */
	private boolean isToFilterForDate = false;

	/**
	 * Month when the search happens.
	 * Must be in the range from Deity.PRAIOS to Deity.NAMENLOSER.
	 * When <em>isToFilterForDate</em> is <code>true</code>,
	 * the search is filtered for plants
	 * that can be harvested at that time of the year.
	 * @see #isToFilterForDate
	 */
	private MonthType month = (MonthType) DSA.instance().getMonth().get(0);

	/** The result set of the search for plants. */
	private int[][] plants;

	/** 
	 * Creates an empty model. 
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 */
	public PlantsModel() {
		final TerrainType t = Terrain.findByID("TERRAIN_FOREST"); 
		terrainIDs = DSA.toIDnumArray(t);
	}

	public RegionType getRegion() {
		return region;
	}

	/** 
	 * Sets the region. 
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param aRegion the region to set
	 */
	public void setRegion(final RegionType aRegion) {
		this.region = aRegion;
	}

	/**
	 * @return Returns the terrainIDs.
	 */
	public int[] getTerrainIDs() {
		return terrainIDs;
	}

	/**
	 * Sets the terrain IDs to consider. 	 
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param terrainIDs the terrainIDs to consider
	 */
	public void setTerrainIDs(final int[] terrainIDs) {
		this.terrainIDs = terrainIDs;
	}

	/**
	 * @return if to test for the terrain probability of the plant
	 */
	public boolean isToTestForTerrainProbability() {
		return isToTestForTerrainProbability;
	}

	/**
	 * Sets if to test for the terrain probability of the plant.
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param isTestForTerrainProbability if to test for the terrain probability of the plant
	 */
	public void setToTestForTerrainProbability(final boolean isTestForTerrainProbability) {
		this.isToTestForTerrainProbability = isTestForTerrainProbability;
	}

	/**
	 * @return if to test for the party's ability for botany
	 */
	public boolean isToTestForBotany() {
		return isToTestForBotany;
	}

	/**
	 * Sets if to test for the party's ability for botany.
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param isTestForBotany if to test for the party's ability for botany
	 */
	public void setToTestForBotany(final boolean isTestForBotany) {
		this.isToTestForBotany = isTestForBotany;
	}

	/**
	 * @return how good the party's test for botany was
	 */
	public int getTestForBotanySuccess() {
		return testForBotanySuccess;
	}

	/**
	 * Sets how good the party's test for botany was.
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param testForBotanySuccess how good the party's test for botany was
	 */
	public void setTestForBotanySuccess(final int testForBotanySuccess) {
		this.testForBotanySuccess = testForBotanySuccess;
	}

	/**
	 * @return if this search is filtered by the month the search takes place
	 */
	public boolean isToFilterForDate() {
		return isToFilterForDate;
	}

	/**
	 * Sets if this search is filtered by the month the search takes place.
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param isToFilterForDate if this search is filtered by the month the search takes place
	 */
	public void setToFilterForDate(final boolean isToFilterForDate) {
		this.isToFilterForDate = isToFilterForDate;
	}

	/**
	 * @return the month when the search happens
	 * @see PlantsModel#setMonth(MonthType)
	 * @see PlantsModel#isToFilterForDate()
	 */
	public MonthType getMonth() {
		return month;
	}

	/**
	 * Sets the month when the search happens.
	 * Call {@link #recalculate()} to finish construction, when done with all the setters to change the object.
	 * @param month the month to set
	 * @see PlantsModel#setToFilterForDate(boolean)
	 */
	public void setMonth(final MonthType month) {
		this.month = month;
	}

	/**
	 * @return the plants found doing the query
	 */
	public int[][] getPlants() {
		return plants;
	}

	/**
	 * Recalculates the search for plants.
	 * Is called internally after a change of the query parameters,
	 * but may be called externally to force a new throwing of dice.
	 */
	public void recalculate() {
		plants = Plant.whatPlantsAreThere(region, terrainIDs, testForBotanySuccess,
				isToTestForTerrainProbability, isToTestForBotany,
				(isToFilterForDate ? month : null));
	}

	public String getPlantsText(final TextDocumentBuilder doc) {
		final StringBuffer ret = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final Messages messages = doc.messages();
		PlantType plant = null;
		final String enrichedRemarkHeading = messages.getText(new I18NStringWithFillIns("PlantsModel.Remark",
				"%0; %1; %2",
				new I18NStringWithFillIns[] { 
						new I18NStringWithFillIns("MasterData.RegionsDescription", null, null),
						new I18NStringWithFillIns("MasterData.HarvestTimeDescription", null, null),
						new I18NStringWithFillIns("MasterData.Remark", null, null), }));
		final TextFormatter.ColumnInfo[] ci = new TextFormatter.ColumnInfo[] {
				new TextFormatter.ColumnInfo( 2, TextFormatter.RIGHT, messages, "PlantsModel.Botany"),
				new TextFormatter.ColumnInfo(12, TextFormatter.LEFT, messages, "PlantsModel.Terrain"),
				new TextFormatter.ColumnInfo(30, TextFormatter.LEFT, messages, "DSA.Plant"),
				new TextFormatter.ColumnInfo(10, TextFormatter.LEFT, messages, "PlantsModel.PlantCategory"),
//            new TextFormatter.ColumnInfo(90, TextFormatter.LEFT, messages, "MasterData.RegionsDescription"),
//            new TextFormatter.ColumnInfo(100, TextFormatter.LEFT, messages, "MasterData.HarvestTimeDescription"),
//            new TextFormatter.ColumnInfo(80, TextFormatter.LEFT, messages, "MasterData.Remark"),
				new TextFormatter.ColumnInfo(135, TextFormatter.LEFT, enrichedRemarkHeading),
				new TextFormatter.ColumnInfo(35, TextFormatter.LEFT, messages, "MasterData.Reference"),
		};
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true, TextFormatter.htmlAttribute(TextFormatter.CLASS, "dsatool"), ci);
		ret.append(tf.beginTable("table0", ti));
		ret.append(tf.formatTableHeading(ci));
		ret.append(tf.beginTableBody());
		for (int i = 0; i < plants.length; i++) {
			if (null != plants[i]) {
				plant = Plant.findByIDnum(i);
				final PlantCategoryType pct = (PlantCategoryType)plant.getCategory();
				final TerrainType ter = Terrain.findByIDnum(plants[i][0]);
				final DSA dsa = DSA.instance();
				final String terrainInfo = messages.getText(
						new I18NStringWithFillIns("PlantsModel.TerrainInfo", "%0(%1)",
								new String[] {
										tf.href(MasterDataModel.getHRefFor(ter),
												TextFormatter.format(dsa.getContentText(ter, tf), 4).toString()),
										Integer.toString(plants[i][1]), }));
				String remark = plant.getRegionsDescription();
				String addition = Plant.getHarvestTimeDescription(plant);
				if (!CollectionUtil.isNullOrEmpty(remark) && !CollectionUtil.isNullOrEmpty(addition)) {
					remark += "; ";
				}
				remark += addition;
				addition = plant.getComment();
				if (!CollectionUtil.isNullOrEmpty(remark) && !CollectionUtil.isNullOrEmpty(addition)) {
					remark += "; ";
				}
				remark += addition;
				ret.append(tf.beginTableRow());
				ret.append(tf.formatTableCell(ci[0], Integer.toString(plant.getHowHardToFind())));
				ret.append(tf.formatTableCell(ci[1], terrainInfo));
				ret.append(tf.formatTableCell(ci[2], tf.href(MasterDataModel.getHRefFor(plant), dsa.getContentText(plant, tf))));
				ret.append(tf.formatTableCell(ci[3], tf.href(MasterDataModel.getHRefFor(pct), pct.getName())));
				ret.append(tf.formatTableCell(ci[4], remark));
				ret.append(tf.formatTableCell(ci[5], plant.getReference()));
				ret.append(tf.endTableRow());
			}
		}
		ret.append(tf.endTableBody());
		ret.append(tf.endTable());
		return ret.toString();
	}

	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		final StringBuffer ret = new StringBuffer();
		final TextFormatter tf = doc.tf();
		final Messages messages = doc.messages();
		ret.append(tf.beginParagraph());
		ret.append(messages.getText("PlantsModel.SearchCriteria")); 
		ret.append(tf.newLine());
		ret.append(messages.getText("PlantsModel.Terrains")); 
		for (int i = 0; i < terrainIDs.length; i++) {
			if (0 < i) {
				ret.append(", "); 
			}
			ret.append(DSA.instance().getContentText(Terrain.findByIDnum(terrainIDs[i]), tf));
		}
		ret.append("."); 
		ret.append(tf.newLine());
		if (isToTestForTerrainProbability) {
			ret.append(messages.getText("PlantsModel.TestForAvailability.With")); 
		} else {
			ret.append(messages.getText("PlantsModel.TestForAvailability.Without")); 
		}
		ret.append(tf.newLine());
		if (isToTestForBotany) {
			String testStr = "";
			if (0 < testForBotanySuccess) {
				testStr = "+";
			}
			testStr += testForBotanySuccess;
			ret.append(messages.getText(
					new I18NStringWithFillIns("PlantsModel.TestForBotany.With", null, new String[] { testStr, })));
		} else {
			ret.append(messages.getText("PlantsModel.TestForBotany.Without")); 
		}
		ret.append(tf.newLine());
		if ((isToFilterForDate) && (null != month)) {
			ret.append(messages.getText(new I18NStringWithFillIns("PlantsModel.Month.With", null,
					new Object[] { Month.toString(month), })));
		} else {
			ret.append(messages.getText("PlantsModel.Month.Without")); 
		}
		ret.append("\n");
		ret.append(tf.endParagraph());
		boolean isEmpty = true;
		for (int i = 0; i < plants.length; i++) {
			if (null != plants[i]) {
				isEmpty = false;
				break;
			}
		}
		if (isEmpty) {
			ret.append(tf.beginParagraph());
			ret.append(messages.getText("PlantsModel.NothingFound")); 
			ret.append(tf.endParagraph());
		} else {
			ret.append(getPlantsText(doc));
		}
		return ret.toString();
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		final Messages messages = doc.messages();
		final String ret = messages.getText("PlantsModel.SearchInRegion", null, 
				new String[] { region.getName(), });
		return ret;
	}

	@Override
	public PlantsModel clone() {
		PlantsModel ret = null;
		try {
			ret = (PlantsModel)super.clone();
			ret.terrainIDs = (int[])terrainIDs.clone();
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}
}
