package de.ped.dsatool.logic;

import java.util.List;
import java.util.Iterator;

import de.ped.tools.Assert;
import de.ped.tools.TextDocumentType;
import de.ped.tools.Textable;

/**
 * A class representing all kinds of predefined business objects -
 * that means classes which have a constant set of constant objects.
 * So it's impossible to create new objects of that class,
 * or to change existing objects of that class.
 *
 * Each of these predefined business objects has at least two attributes:
 * <ol>
 * <li>The object ID (as primary key) of all objects counts from 0 to the
 *     number of objects of that subclass minus one.
 * </li>
 * <li>The name of all objects are non-null, non-empty, and unique.
 * </li>
 * </ol>
 *
 * To retrieve the number of objects of a subclass, you typically ask for
 * <em>&lt;subclass&gt;</em><code>.findAll().length</code>.
 *
 * @author Copyright (c) 1999 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class PredefinedBusinessObject implements Textable {
	/**
	 * Sets the object to the given object index. No legality checking is done.
	 * 
	 * @param anID
	 *            The ID of the new object
	 * @param aName
	 *            The name of the new object
	 */
	protected PredefinedBusinessObject(final int anID, final String aName) {
		id = anID;
		name = aName;
	}

	/**
	 * @param names
	 *            the list to search
	 * @param name
	 *            the name of the item to find
	 * @return the index of <em>name</em> within <em>names</em>.
	 * @exception ArrayIndexOutOfBoundsException
	 *                No reference object fits the object name.
	 */
	public static int findIdByName(final String[] names, final String name) {
		for (int i = 0; i < names.length; i++) {
			if (names[i].equalsIgnoreCase(name)) {
				return i;
			}
		}
		throw new ArrayIndexOutOfBoundsException();
	}

	/**
	 * @param objs
	 *            list to search
	 * @param name
	 *            the name of the item to find
	 * @return the object found, or <code>null</code> if none found.
	 */
	public static PredefinedBusinessObject findByName(final PredefinedBusinessObject[] objs, final String name) {
		for (int i = 0; i < objs.length; i++) {
			if (objs[i].getName().equalsIgnoreCase(name)) {
				return objs[i];
			}
		}
		return null;
	}

	/**
	 * @param objs
	 *            list to search
	 * @param name
	 *            the name of the item to find
	 * @return the object found, or <code>null</code> if none found.
	 */
	public static PredefinedBusinessObject findByName(final List<PredefinedBusinessObject> objs, final String name) {
		final Iterator<PredefinedBusinessObject> it = objs.iterator();
		while (it.hasNext()) {
			final PredefinedBusinessObject o = (PredefinedBusinessObject) it.next();
			if (o.getName().equalsIgnoreCase(name)) {
				return o;
			}
		}
		return null;
	}

	/**
	 * The list of names of all defined reference objects of this class.
	 * 
	 * @param objs
	 *            list of objects to compile the names from
	 * @return list of names
	 */
	public static String[] listOfObjectNames(final PredefinedBusinessObject[] objs) {
		final String[] listOfObjectNames = new String[objs.length];
		for (int i = 0; i < objs.length; i++) {
			listOfObjectNames[i] = objs[i].getName();
		}
		return listOfObjectNames;
	}

	/**
	 * Inverts a list of IDs, based on the fact that the IDs of predefined
	 * business objects are always counting from 0 to the number of objects
	 * minus one.
	 * 
	 * @param idArr
	 *            The positive array of IDs. <code>null</code> indicates an empty
	 *            array.
	 * @param numOfIDs
	 *            The number of objects of that class, so object IDs are in the
	 *            range of 0 .. numOfIDs - 1. May NOT contain <code>null</code>
	 *            values, duplicates, or out-of-range IDs.
	 * @return The negative array of IDs.
	 */
	public static int[] invert(final int[] idArr, final int numOfIDs) {
		final int[] negArr = new int[numOfIDs - idArr.length];
		int negIdx = 0;
		for (int allIdx = 0; allIdx < numOfIDs; allIdx++) {
			boolean isFound = false;
			for (int arrIdx = 0; arrIdx < idArr.length; arrIdx++) {
				if (allIdx == idArr[arrIdx]) {
					isFound = true;
					break;
				}
			}
			if (!isFound) {
				negArr[negIdx] = allIdx;
				negIdx++;
			}
		}
		Assert.assertEquals(negArr.length, negIdx);
		return negArr;
	}

	/**
	 * @return The index of this object in the list of names of creatable objects.
	 */
	public int getID() {
		return id;
	}

	/**
	 * @return The name of this object.
	 */
	public String getName() {
		return name;
	}

	public String getContentText(final TextDocumentType formatStyle) {
		return getName();
	}

	@Override
	public String toString() {
		return getContentText(TextDocumentType.FIXEDFONT);
	}

	/** ID of the object. */
	protected int id;

	/** Person of the object. */
	protected String name;
}
