package de.ped.dsatool.logic;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.ped.dsatool.dsa.generated.QuestTextPartType;
import de.ped.dsatool.dsa.generated.QuestTextPatternType;
import de.ped.dsatool.dsa.generated.QuestTextType;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.MathUtil;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.Textable;
import de.ped.tools.log.Logger;
import de.ped.tools.TextFormatter;

/**
 * Randomly generates a plot for a quest. Don't take the resulting quests too serious :-)
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de),
 *  algorithms based on the author of the DSA-o-mat, whoever that is :-)
 */
public class Quest implements Cloneable, Textable {
	private static Logger logger = Logger.getRootLogger();

	/** This ID of a QuestTextPartType object is treated special. */
	private static final String QUESTTEXT_PART_NEWLINE      = "QUESTTEXT_PART_NEWLINE";

	/** This ID of a QuestTextPartType object is treated special. */
	private static final String QUESTTEXT_PART_NEWPARAGRAPH = "QUESTTEXT_PART_NEWPARAGRAPH";

	/**
	 * Standard constructor.
	 * Builds a random quest.
	 */
	public Quest() {
		setAllRandom();
	}

	@Override
	public Quest clone() {
		Quest ret = null;
		try {
			ret = (Quest)super.clone();
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}

	/**
	 * Builds a random quest.
	 */
	public void setAllRandom() {
		questParts = null;
		final QuestTextType qtt = DSA.instance().getQuestText();
		if (null != qtt) {
			final QuestTextPatternType qtpattern = qtt.getQuestTextPattern();
			if (null != qtpattern) {
				final List<?> questTextParts = qtpattern.getQuestTextPart();
				if (!CollectionUtil.isNullOrEmpty(questTextParts)) {
					questParts = new LinkedList<Object>();
					final Iterator<?> it = questTextParts.iterator();
					while (it.hasNext()) {
						final QuestTextPartType qtp = DSA.toQuestTextPartValue(it.next());
						final String qtpid = qtp.getID();
						logger.trace(qtpid);
						if (qtpid.equals(QUESTTEXT_PART_NEWLINE)) {
							questParts.add(qtp);
						} else if (qtpid.equals(QUESTTEXT_PART_NEWPARAGRAPH)) {
							questParts.add(qtp);
						} else {
							final List<String> parts = qtp.getPart();
							if (!CollectionUtil.isNullOrEmpty(parts)) {
								final int idx = MathUtil.random(parts.size());
								final String part = parts.get(idx);
								logger.trace("" + idx + ": " + part);
								questParts.add(part);
							}
						}
					}
				}
			}
		}
	}

	/**
	 * @param doc
	 *            the document to add to
	 * @return the quest text.
	 */
	@Override
	public String getContentText(final TextDocumentBuilder doc) {
		final StringBuffer s = new StringBuffer();
		final TextFormatter tf = doc.tf();
		s.append(tf.beginParagraph());
		if (null != questParts) {
			final Iterator<Object> it = questParts.iterator();
			while (it.hasNext()) {
				final Object o = it.next();
				if (o instanceof QuestTextPartType) {
					final QuestTextPartType qtp = (QuestTextPartType)o;
					final String qtpid = qtp.getID();
					if (qtpid.equals(QUESTTEXT_PART_NEWLINE)) {
						s.append(tf.newLineIfPlainText());
					} else if (qtpid.equals(QUESTTEXT_PART_NEWPARAGRAPH)) {
						s.append(tf.endParagraph());
						s.append(tf.beginParagraph());
					} else {
						final List<String> parts = qtp.getPart();
						if (!CollectionUtil.isNullOrEmpty(parts)) {
							final String part = (String)parts.get(MathUtil.random(parts.size()));
							s.append(part);
						}
					}
				} else if (o instanceof String) {
					s.append((String)o);
				}
			}
		}
		s.append(tf.endParagraph());
		return (s.toString());
	}

	@Override
	public String getHeadingText(final TextDocumentBuilder doc) {
		return doc.messages().getText("DSA.Quest");
	}

	/** 
	 * The text parts this particular quest text is created from.
	 * May contain objects of type String or QuestTextPartType.
	 */
	private LinkedList<Object> questParts;
}
