package de.ped.dsatool.logic;

import de.ped.tools.CollectionUtil;
import de.ped.tools.TextFormatter;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.dsatool.dsa.generated.PersonType;
import de.ped.dsatool.dsa.generated.PricingCategoryType;
import de.ped.dsatool.dsa.generated.QualityCategoryType;
import de.ped.dsatool.dsa.generated.QualityNamesTableType;
import de.ped.dsatool.dsa.generated.ShopCategoryType;
import de.ped.dsatool.dsa.generated.ShopSpecialtyType;
import de.ped.dsatool.dsa.generated.ShopType;

/**
 * The concept of single shops, hotels, bars etc. for towns.
 *
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Stadt- und Dorfentwicklung" (p.20..23)</p>
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Shop {
	public static final int CATEGORY_HOTEL      =  0;
	public static final int CATEGORY_BAR        =  1;
	public static final int CATEGORY_HANDCRAFT1 =  2;
	public static final int CATEGORY_HANDCRAFT2 =  3;
	public static final int CATEGORY_HANDCRAFT3 =  4;
	public static final int CATEGORY_MERCHAND1  =  5;
	public static final int CATEGORY_MERCHAND2  =  6;
	public static final int CATEGORY_MERCHAND3  =  7;
	public static final int CATEGORY_SERVICE1   =  8;
	public static final int CATEGORY_SERVICE2   =  9;
	public static final int CATEGORY_HEALER     = 10;
	public static final int CATEGORY_ENTERPRISE = 11;
	public static final int CATEGORY_TRAVELLER  = 12;

	/**
	 * @param specialty
	 *            The profession of the shop.
	 * @param theQualityOffset
	 *            This offset will be added to the quality encoding of the shop.
	 *            This value is allowed to be negative.
	 * @param thePricesOffset
	 *            This offset will be added to the prices encoding of the shop.
	 *            This value is allowed to be negative.
	 * @return a shop of the given category and specialty. The quality and
	 *         pricing will be chosen randomly, affected by the given offsets.
	 *         For details of the quality and prices encoding see the DSA book
	 *         referred to in the general class comment.
	 */
	public static ShopType create(final ShopSpecialtyType specialty, final int theQualityOffset, final int thePricesOffset) {
		final DSA dsa = DSA.instance();
		final ShopType shop = dsa.createShopType();
		Shop.setSpecialty(shop, specialty);
		final int quality = Dice.roll(1, 20) + theQualityOffset;
		shop.setQuality(MathUtil.bound(quality, 1, 20));
		final int qualityCategoryIndex = shop.getQuality() - 1;
		final QualityCategoryType qc = (QualityCategoryType)dsa.getQualityCategory().get(qualityCategoryIndex); 
		shop.setPricing(MathUtil.bound(Dice.roll(1, 20) + thePricesOffset + qc.getQualityToPriceModifier(), 1, 20));
		return shop;
	}

	public static ShopType createShop(final ShopType sourceShop) {
		final ShopType newShop = DSA.instance().createShopType();
		newShop.setChief(sourceShop.getChief());
		newShop.setComment(sourceShop.getComment());
		newShop.setName(sourceShop.getName());
		newShop.setPricing(MathUtil.bound(sourceShop.getPricing(), 1, 20));
		newShop.setQuality(MathUtil.bound(sourceShop.getQuality(), 1, 20));
		setSpecialty(newShop, sourceShop.getSpecialty());
		newShop.getInhabitant().addAll(sourceShop.getInhabitant());
		return newShop;
	}

	/**
	 * @param shop
	 *            the shop
	 * @return The shop category.
	 */
	public static ShopCategoryType getCategory(final ShopType shop) {
		final ShopSpecialtyType specialty = (ShopSpecialtyType)shop.getSpecialty();
		final ShopCategoryType category = (ShopCategoryType)specialty.getShopCategory();
		return category;
	}

	/**
	 * @param shop
	 *            the shop
	 * @return The correct title for the person being the chief of this
	 *         building, e.g., "bartender"
	 */
	public static String getChiefTitle(final ShopType shop) {
		return getCategory(shop).getChiefTitle();
	}

	/**
	 * @param shop
	 *            the shop
	 * @param specialty
	 *            The profession of the shop.
	 * @see #getSpecialtyName(ShopType)
	 */
	public static void setSpecialty(final ShopType shop, final Object specialty) {
		final ShopSpecialtyType spec = (ShopSpecialtyType)specialty; 
		final ShopCategoryType category = (ShopCategoryType)spec.getShopCategory(); 
		if (category.getIDnum() == CATEGORY_HOTEL || category.getIDnum() == CATEGORY_BAR) {
			shop.setName(PubNameGenerator.generatePubName());
		} else {
			shop.setName(null);
		}
		shop.setSpecialty(specialty);
	}

	/**
	 * @param shop
	 *            the shop
	 * @return The shop specialty in human-readable form.
	 * @see ShopType#getSpecialty
	 */
	public static String getSpecialtyName(final ShopType shop) {
		final ShopSpecialtyType s = (ShopSpecialtyType)shop.getSpecialty();
		return s.getName();
	}

	/**
	 * @param shop
	 *            the shop
	 * @return The shop quality in human-understandable form.
	 * @see ShopType#getQuality
	 */
	public static String getQualityName(final ShopType shop) {
		final QualityNamesTableType table = (QualityNamesTableType)getCategory(shop).getQualityNamesTable();
		final int qualityCategoryIndex = MathUtil.bound(shop.getQuality(), 1, 20) - 1;
		final QualityCategoryType qcat = (QualityCategoryType)DSA.instance().getQualityCategory().get(qualityCategoryIndex); 
		final int qualityIndex = qcat.getQualityIndex();
		return DSA.toStringValue(table.getQualityName().get(qualityIndex));
	}

	/**
	 * @param shop
	 *            the shop
	 * @return The shop prices encoding in human-understandable form.
	 * @see ShopType#getPricing
	 * @see #getPricingFactor
	 */
	public static String getPricingName(final ShopType shop) {
		final int pricingIndex = MathUtil.bound(shop.getPricing(), 1, 20) - 1;
		final PricingCategoryType t = (PricingCategoryType)DSA.instance().getPricingCategory().get(pricingIndex);
		return t.getName() + " (" + t.getFactor() + "%)"; 
	}

	/**
	 * @param shop
	 *            the shop
	 * @return The pricing factor for that shop that should be multiplied with the standard prices for the common goods.
	 * @see ShopType#getPricing
	 * @see #getPricingName
	 */
	public static double getPricingFactor(final ShopType shop) {
		final int pricingIndex = MathUtil.bound(shop.getPricing(), 1, 20) - 1;
		final PricingCategoryType t = (PricingCategoryType)DSA.instance().getPricingCategory().get(pricingIndex);
		return (double)t.getFactor() / 100.0;
	}

	/**
	 * @param shop
	 *            the shop
	 * @param tf
	 *            the text formatter to use
	 * @return A string describing all important aspects of this shop.
	 */
	public static String getContentText(final ShopType shop, final TextFormatter tf) {
		final Messages messages = DSAEnv.instance().messages();
		final StringBuffer ret = new StringBuffer();
		ShopCategoryType category = getCategory(shop);
		final int indent = 9;
		ret.append("\t"); 
		ret.append(messages.getText("Shop.Category")); 
		ret.append(": "); 
		final String specialtyName = getSpecialtyName(shop);
		ret.append((specialtyName.equals("")) ? category.getName() : specialtyName); 
		final String shopName = shop.getName();
		if(!CollectionUtil.isNullOrEmpty(shopName)) {
			ret.append(TextFormatter.NL);
			ret.append("\t"); 
			ret.append(TextFormatter.format(messages.getText("Shop.ShopName"), indent, TextFormatter.LEFT, '.')); 
			ret.append(": "); 
			ret.append(shopName);
		}
		if(null != shop.getChief()) {
			ret.append(TextFormatter.NL);
			ret.append("\t"); 
			ret.append(TextFormatter.format(getChiefTitle(shop), indent, TextFormatter.LEFT, '.'));
			ret.append(": "); 
			ret.append(DSA.instance().getContentText((PersonType)shop.getChief(), tf));
		}
		ret.append(TextFormatter.NL);
		ret.append("\t"); 
		ret.append(TextFormatter.format(messages.getText("Shop.Quality"), indent, TextFormatter.LEFT, '.')); 
		ret.append(":"); 
		ret.append(TextFormatter.format(shop.getQuality(), 3));
		ret.append(" ("); 
		ret.append(getQualityName(shop));
		ret.append(")"); 
		ret.append(TextFormatter.NL);
		ret.append("\t"); 
		ret.append(TextFormatter.format(messages.getText("Shop.Prices"), indent, TextFormatter.LEFT, '.')); 
		ret.append(":"); 
		ret.append(TextFormatter.format(shop.getPricing(), 3));
		ret.append(" ("); 
		ret.append(getPricingName(shop));
		ret.append(")"); 
		ret.append(TextFormatter.NL);
		return ret.toString();
	}
}
