package de.ped.dsatool.logic;

import java.util.Iterator;
import java.util.List;

import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.MathUtil;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.BuildingType;
import de.ped.dsatool.dsa.generated.DeityType;
import de.ped.dsatool.dsa.generated.DiceRollType;
import de.ped.dsatool.dsa.generated.LanguageDistributionType;
import de.ped.dsatool.dsa.generated.LanguageType;
import de.ped.dsatool.dsa.generated.NumberOfShopsDiceRollType;
import de.ped.dsatool.dsa.generated.PersonType;
import de.ped.dsatool.dsa.generated.RegionType;
import de.ped.dsatool.dsa.generated.ShopCategoryType;
import de.ped.dsatool.dsa.generated.ShopSpecialtyType;
import de.ped.dsatool.dsa.generated.ShopType;
import de.ped.dsatool.dsa.generated.TempleType;
import de.ped.dsatool.dsa.generated.TownSizeCategoryType;
import de.ped.dsatool.dsa.generated.TownType;

/**
 * A system to create small towns.
 * This is an implementation of the suggestion given by the authors
 * of DSA in the listed source.  It's intention was to be easy
 * for people throwing dices, but it isn't quite appropiate for
 * the more sophisticated calculation possibilities of computers.
 *
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Stadt- und Dorfentwicklung" (p.20..23)</p>
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Town {
	private static Logger logger = Logger.getLogger("Town");
	static {
		logger.setLogLevel(Level.INFO);
	}

	/**
	 * @param numberOfInhabitants
	 *            The number of inhabitants
	 * @param name
	 *            The town name
	 * @param region
	 *            The Aventurian region the town is in
	 * @return a town with given name, size and region.
	 */
	public static TownType create(final int numberOfInhabitants, final String name, final RegionType region) {
		TownType t = DSA.instance().createTownType();
		t.setName(name);
		setRegion(t, region);
		t.setNumberOfInhabitants(numberOfInhabitants);
		fillUp(t);
		return t;
	}

	/**
	 * Creates a town based on a predefined town. It does so by cloning the
	 * predefined town and filling up the missing values.
	 * 
	 * @param sourceTown
	 *            The predefined town to start with.
	 * @return A fully created town.
	 */
	public static TownType create(final TownType sourceTown) {
		Assert.assertNotNull(sourceTown);
		final TownType myTown = createTown(sourceTown);
		fillUp(myTown);
		return myTown;
	}

	/**
	 * Reads the language distribution for the town. If it is not defined so
	 * far, it creates a plausible language distribution. The language
	 * distribution of the town itself is not changed by this method.
	 * 
	 * @param town
	 *            The town the language distribution is for.
	 * @return The language distribution for the town.
	 */
	public static LanguageDistributionType getOrCreateLanguageDistribution(final TownType town) {
		LanguageDistributionType ret = town.getLanguageDistribution();
		// How's the language distribution of the town?
		if (null == ret) {
			ret = DSA.instance().createLanguageDistributionType();
		}
		final RegionType region = (RegionType) town.getRegion();
		LanguageDistribution.fillUpWith(ret, region.getLanguageDistribution());
		return ret;
	}

	public static void fillUp(final TownType town) {
		// How's the language distribution of the town?
		if (null == town.getLanguageDistribution()) {
			town.setLanguageDistribution(getOrCreateLanguageDistribution(town));
		}
		final RegionType region = (RegionType) town.getRegion();
		LanguageDistribution.fillUpWith(town.getLanguageDistribution(), region.getLanguageDistribution());
		// Fill up the predefined buildings with missing information
		final Iterator<TempleType> templeIt = town.getTemple().iterator();
		while (templeIt.hasNext()) {
			final TempleType temple = (TempleType) templeIt.next();
			if (null == temple.getChief()) {
				temple.setChief(generateRandomPerson(town));
			}
		}
		final Iterator<ShopType> shopIt = town.getShop().iterator();
		while (shopIt.hasNext()) {
			final ShopType shop = (ShopType) shopIt.next();
			if (null == shop.getChief()) {
				shop.setChief(generateRandomPerson(town));
			}
		}
		// Fill up buildings
		addTemples(town);
		addShops(town);
	}

	/**
	 * Searches the predefined towns for the given town name. The return value
	 * must be treated as read-only. So, to create a fully blown-up town out of
	 * it, clone it first.
	 * 
	 * @param townName
	 *            The town name to search for.
	 * @return The predefined town, or <code>null</code> if not found.
	 */
	public static TownType findPredefinedTown(final String townName) {
		final Iterator<TownType> it = DSA.instance().getTown().iterator();
		TownType sourceTown = null;
		while ((null == sourceTown) && it.hasNext()) {
			final TownType town = (TownType)it.next();
			final String tName = town.getName();
			if ((null != tName) && (tName.equalsIgnoreCase(townName))) {
				sourceTown = town;
			}
		}
		return sourceTown;
	}

	/**
	 * Searches the predefined towns for the given town name. The return value
	 * must be treated as read-only. So, to create a fully blown-up town out of
	 * it, clone it first.
	 * 
	 * @param townName
	 *            The town name to search for.
	 * @return The predefined town, or <code>null</code> if not found.
	 */
	public static int findPredefinedTownIdx(final String townName) {
		final Iterator<TownType> it = DSA.instance().getTown().iterator();
		int ret = -1;
		int idx = 0;
		while ((ret < 0) && it.hasNext()) {
			final TownType town = (TownType) it.next();
			final String tName = town.getName();
			if ((null != tName) && (tName.equalsIgnoreCase(townName))) {
				ret = idx;
			}
			idx++;
		}
		return ret;
	}

	public static TownType createTown(final TownType sourceTown) {
		final TownType t = DSA.instance().createTownType();
		setRegion(t, sourceTown.getRegion());
		t.setLanguageDistribution(sourceTown.getLanguageDistribution());
		t.setName(sourceTown.getName());
		t.setNumberOfInhabitants(sourceTown.getNumberOfInhabitants());
		t.setReference(sourceTown.getReference());
		t.setMap(sourceTown.getMap());
		final Iterator<TempleType> templeIt = sourceTown.getTemple().iterator();
		while (templeIt.hasNext()) {
			final TempleType sourceTemple = (TempleType)templeIt.next();
			t.getTemple().add(Temple.createTemple(sourceTemple));
		}
		final Iterator<ShopType> shopIt = sourceTown.getShop().iterator();
		while (shopIt.hasNext()) {
			final ShopType sourceShop = (ShopType)shopIt.next();
			t.getShop().add(Shop.createShop(sourceShop));
		}
		return t;
	}

//	/**
//	 * set method for property
//	 */
//	public void setNumberOfInhabitants(int numberOfInhabitants)
//	{
//		final boolean resized = (getNumberOfInhabitants() != numberOfInhabitants);
//		super.setNumberOfInhabitants(numberOfInhabitants);
//		if (resized)
//		{
//			rebuild(true, true);
//		}
//	}

	/**
	 * @param town
	 *            The town 
	 * @param region
	 *            The Aventurian region the town is in
	 */
	public static void setRegion(final TownType town, final Object region) {
		final RegionType oldRegion = (RegionType) town.getRegion();
		final RegionType newRegion = (RegionType) region;
		town.setRegion(region);
		if ((null == town.getLanguageDistribution()) || (null == oldRegion) || (!newRegion.getID().equals(oldRegion.getID()))) {
			// Changing the region also sets the language distribution to the same as the new region.
			town.setLanguageDistribution(LanguageDistribution.clone(newRegion.getLanguageDistribution()));
		}
	}

	public static String getEnterprisesRemark(final TownType town) {
		return calcSizeCategory(town, false).getEnterprisesRemark();
	}

	public static String getTravellersRemark(final TownType town) {
		return calcSizeCategory(town, isFuzzyMode).getTravellersRemark();
	}

	public static String getMarketDaysRemark(final TownType town) {
		return calcSizeCategory(town, isFuzzyMode).getMarketDaysRemark();
	}

	public static void addContentText(final TextDocumentBuilder doc, final TownType town) {
		DSA.instance().addContentText(doc, town);
	}

	/**
	 * @param inhabitants
	 *            The size of the town in inhabitants
	 * @param doItFuzzy
	 *            If <code>false</code>, behavior is exactly as in the DSA books; if
	 *            <code>true</code>, it scales more smoothly between town sizes
	 * @return the size category of the town that's appropiate to the size
	 */
	public static TownSizeCategoryType calcSizeCategory(final int inhabitants, final boolean doItFuzzy) {
		final Iterator<TownSizeCategoryType> it = DSA.instance().getTownSizeCategory().iterator();
		TownSizeCategoryType next = null;
		TownSizeCategoryType here = null;
		TownSizeCategoryType last = null;
		boolean found = false;
		while(it.hasNext() && !found) {
			last = here;
			here = next;
			next = (TownSizeCategoryType)it.next();
			if(inhabitants < next.getMinimumNumberOfInhabitants()) {
				found = true;
			}
		}
		if (null == here) {
			here = next;
		}
		if (doItFuzzy) {
			if ((null != last)
					&& (null != here)
					&& (null != next)
					&& (MathUtil.random(next.getMinimumNumberOfInhabitants() - here.getMinimumNumberOfInhabitants()) 
							> (inhabitants - here.getMinimumNumberOfInhabitants()))) {
				here = last;
			}
		}
		// Just for safety
		if (null == here) {
			here = last;
		}
		logger.debug("TownSizeCategory=" + here.getIDnum()); 
		return here;
	}

	/**
	 * @param inhabitants
	 *            The size of the town in inhabitants.
	 * @return the size category, based on the given number of inhabitants. Uses
	 *         fuzzy mode, depending on the settings.
	 * @see #getFuzzyMode
	 */
	public static TownSizeCategoryType calcSizeCategory(final int inhabitants) {
		return calcSizeCategory(inhabitants, isFuzzyMode);
	}

	/**
	 * @param town
	 *            The town
	 * @return the size category, based on the number of inhabitants. Uses fuzzy
	 *         mode, depending on the settings.
	 * @see #getFuzzyMode
	 */
	public static TownSizeCategoryType calcSizeCategory(final TownType town) {
		return calcSizeCategory(town.getNumberOfInhabitants(), isFuzzyMode);
	}

	/**
	 * @param town
	 *            The town
	 * @param doItFuzzy
	 *            If <code>false</code>, behavior is exactly as in the DSA books; if
	 *            <code>true</code>, it scales more smoothly between town sizes
	 * @return the size category, based on the number of inhabitants.
	 */
	public static TownSizeCategoryType calcSizeCategory(final TownType town, final boolean doItFuzzy) {
		return calcSizeCategory(town.getNumberOfInhabitants(), doItFuzzy);
	}

	/**
	 * @param town
	 *            The town
	 * @return a person generated according to the language distribution
	 */
	public static PersonType generateRandomPerson(final TownType town) {
		LanguageDistributionType langDist = town.getLanguageDistribution();
		if (null == langDist) {
			RegionType region = (RegionType)town.getRegion();
			langDist = region.getLanguageDistribution();
		}
		final int languageID = LanguageDistribution.chooseLanguageID(langDist);
		final LanguageType language = Language.findByIDnum(languageID);
		return Person.createPerson(language);
	}

	/**
	 * Adds a single building to the town
	 * 
	 * @param town
	 *            The town to change
	 * @param building
	 *            The building to add
	 */
	public static void addBuilding(final TownType town, final BuildingType building) {
		if (building instanceof TempleType) {
			town.getTemple().add((TempleType)building);
		} else {
			town.getShop().add((ShopType)building);
		}
	}

	/**
	 * Adds a single temple to the town.
	 * 
	 * @param town
	 *            The town to change
	 * @param deity
	 *            The deity the temple worships to add
	 */
	public static void addTemple(final TownType town, final DeityType deity) {
		Assert.assertNotNull(deity);
		final TempleType temple = DSA.instance().createTempleType();
		temple.getDeities().add(deity);
		temple.setChief(generateRandomPerson(town));
		town.getTemple().add(temple);
	}

	/**
	 * Add some temples randomly, based on the DSA temples distribution
	 * 
	 * @param town
	 *            The town to change
	 * @param howMany
	 *            So many temples are added.
	 */
	public static void addTemples(final TownType town, final int howMany) {
		final RegionType region = (RegionType)town.getRegion();
		List<String> templeProbs = DSA.instance().getTempleDistribution();
		final List<String> regionTempleProbs = region.getTempleDistribution();
		if (!CollectionUtil.isNullOrEmpty(regionTempleProbs)) {
			templeProbs = regionTempleProbs;
		}
		final int[] probability = new int[templeProbs.size()];
		String logmessage = "Temple probabilities: "; 
		for (int i = 0; i < probability.length; i++) {
			probability[i] = DSA.toIntValue(templeProbs.get(i));
			logmessage += probability[i] + " "; 
		}
		logger.debug(logmessage);
		for (int t = 0; t < howMany; t++) {
			try {
				final int index = MathUtil.randomDistributed(probability);
				probability[index] = 0; //We don't allow two temples for the same deity
				addTemple(town, Deity.findByIDnum(index));
			} catch (final ArrayIndexOutOfBoundsException e) {
				logger.debug("All deities have their temples - so no more to add"); 
				break;
			}
		}
	}

	/**
	 * Add some temples randomly, based on the DSA temples distribution. How
	 * many temples will be added, depends on the size of the town. If the town
	 * already contains temples, these are kept, and the number is filled up.
	 * 
	 * @param town
	 *            The town to change
	 */
	public static void addTemples(final TownType town) {
		// If we want the reference remark to the archive of big cities, don't do anything
		final DSA dsa = DSA.instance();
		final TownSizeCategoryType borderTownSizeCategory = (TownSizeCategoryType) dsa.getTownSizeCategory().get(5);
		if (borderTownSizeCategory.getMinimumNumberOfInhabitants() <= town.getNumberOfInhabitants()) {
			return;
		}
		// How many many temples should be there?
		final TownSizeCategoryType townSizeCategory = calcSizeCategory(town, isFuzzyMode);
		final DiceRollType templesDiceRoll = townSizeCategory.getNumberOfTemplesDiceRoll();
		final int howOften = Dice.roll(templesDiceRoll);
		if (0 < howOften) {
			logger.debug("Temple: " + dsa.getContentText(templesDiceRoll, null) + " ==> " + howOften); 
			// How many temples are already there? Add the difference.
			addTemples(town, howOften - town.getTemple().size());
		}
	}

	/**
	 * Adds a single shop to the town.
	 * 
	 * @param town
	 *            The town to change
	 * @param specialty
	 *            The specialty of the shop to add
	 */
	protected static void addShop(final TownType town, final ShopSpecialtyType specialty) {
		int priceOffset;
		final ShopCategoryType category = (ShopCategoryType)specialty.getShopCategory();
		final TownSizeCategoryType townSizeCategory = calcSizeCategory(town, false);
		switch(category.getIDnum()) {
		case Shop.CATEGORY_MERCHAND1:
		case Shop.CATEGORY_MERCHAND2:
		case Shop.CATEGORY_MERCHAND3:
			priceOffset = townSizeCategory.getPricesMerchands();
			break;
		default:
			priceOffset = townSizeCategory.getPricesHandcrafts();
		}
		final ShopType shop = Shop.create(specialty, townSizeCategory.getQuality(), priceOffset);
		shop.setChief(generateRandomPerson(town));
		town.getShop().add(shop);
	}

	/**
	 * Counts how many shops of that specialty already exist.
	 * 
	 * @param town
	 *            The town to change
	 * @param specialty
	 *            The shop specialty to count.
	 * @return The number of shops of that specialty in this town.
	 */
	protected static int countShops(final TownType town, final ShopSpecialtyType specialty) {
		int foundShops = 0;
		final Iterator<ShopType> it = town.getShop().iterator();
		while (it.hasNext()) {
			final ShopType shop = (ShopType)it.next();
			final ShopSpecialtyType spec = (ShopSpecialtyType)shop.getSpecialty(); 
			if (spec.getID().equals(specialty.getID())) {
				foundShops++;
			}
		}
		return foundShops;
	}

	/**
	 * Adds a number of shops to the town. All shops are of the same category/specialty,
	 * but may differ in quality and pricing.
	 * 
	 * The number of shops generated is random (based on suggestions for towns of that size).
	 * If the town already contains shops of that kind, these are kept, and the number is filled up.
	 *
	 * @param town
	 *            The town to change
	 * @param specialty
	 *            The shop specialty to count.
	 */
	protected static void addShops(final TownType town, final ShopSpecialtyType specialty) {
		// How many shops of that type should be there?
		final ShopCategoryType shopCategory = (ShopCategoryType)specialty.getShopCategory();
		final NumberOfShopsDiceRollType numberOfShopsDiceRoll = (NumberOfShopsDiceRollType)calcSizeCategory(town, isFuzzyMode).getNumberOfShopsDiceRoll().get(shopCategory.getIDnum());
		int howOften = Dice.roll(numberOfShopsDiceRoll);
		if (0 < howOften) {
			final ShopCategoryType t = (ShopCategoryType)numberOfShopsDiceRoll.getShopCategory();
			logger.debug(t.getName() + ": " + DSA.instance().getContentText(numberOfShopsDiceRoll, null) + " ==> " + howOften); 
			// How many shops are already there? Add the difference.
			howOften -= countShops(town, specialty);
			for (int i = 0; i < howOften; ++i) {
				addShop(town, specialty);
			}
		}
	}

	/**
	 * Adds the whole palette of shops to the town. The number of shops
	 * generated is random (based on suggestions for towns of that size). If the
	 * town already contains shops, these are kept, and the number is filled up.
	 * 
	 * @param town
	 *            The town to change
	 */
	public static void addShops(final TownType town) {
		final Iterator<ShopSpecialtyType> it = DSA.instance().getShopSpecialty().iterator();
		while (it.hasNext()) {
			final ShopSpecialtyType specialty = (ShopSpecialtyType)it.next();
			addShops(town, specialty);
		}
	}

	/**
	 * @return if to use fuzzy mode on town size based calculations
	 */
	public static boolean getFuzzyMode() {
		return isFuzzyMode;
	}

	/**
	 * @param fuzzyMode
	 *            if to use fuzzy mode on town size based calculations
	 */
	public static void setFuzzyMode(final boolean fuzzyMode) {
		isFuzzyMode = fuzzyMode;
	}

	/**
	 * Is the size category of the town calculated as in the DSA books, with hard borders between categories
	 * (then this variable is <code>false</code>), or is it smoothened by fuzzy logic?
	 */
	private static boolean isFuzzyMode = true;
}
