package de.ped.dsatool.logic;

import de.ped.dsatool.dsa.generated.TownType;
import de.ped.tools.TextDocumentType;
import de.ped.tools.MathUtil;
import de.ped.tools.MathVector;
import de.ped.tools.Messages;
import de.ped.tools.TextFormatter;

/**
 * A class for calculating a map of a small town or village.
 * 
 * @author Martin Hocke (hocke@aea.ruhr-uni-bochum.de) wrote the algorithms in C++ for Linux,
 * @author Peter Diefenbach (peter@pdiefenbach.de) ported it 1998 to Java and changed the behavior of some routines
 */
public class TownMap {
	/** Part of the infrastructure bit field, defining an empty infrastructure bit field. */
	public static final int INFRA_EMPTY     = 0x0;
	/** Part of the infrastructure bit field, defining bits for a market place. */
	public static final int INFRA_MARKET    = 0x1;
	/** Part of the infrastructure bit field, defining bits for a town wall. */
	public static final int INFRA_TOWN_WALL = 0x2;
	/** Part of the infrastructure bit field, defining bits for a town hall. */
	public static final int INFRA_TOWN_HALL = 0x4;
	/** Part of the infrastructure bit field, defining bits for temples. */
	public static final int INFRA_TEMPLES   = 0x8;

	/** Output character for a street. */
	public static final char REP_STREET	= ' ';
	/** Output character for a house. */
	public static final char REP_HOUSE = 174;
	/** Output character for a market place. */
	public static final char REP_MARKET	= 'M';
	/** Output character for a town wall. */
	public static final char REP_WALL = '#';
	/** Output character for a town gate. */
	public static final char REP_GATE = '^';
	/** Output character for a town hall. */
	public static final char REP_TOWN_HALL = 'R';
	/** Output character for a temple. */
	public static final char REP_TEMPLE	= 'T';
	/** Output character for something undefined. */
	public static final char REP_UNDEF = ' ';

	public static final String TAB = "  "; 

	/**
	 * Creates a map of given size, infrastructure bit field, and number of temples and houses.
	 * 
	 * @param sizex
	 *            the x size of the town map in house grid coordinates
	 * @param sizey
	 *            the y size of the town map in house grid coordinates
	 * @param infra
	 *            {@link #INFRA_EMPTY} for none, or the OR-ed values of the
	 *            to-be-built infrastructure {@link #INFRA_MARKET},
	 *            {@link #INFRA_TOWN_WALL}, {@link #INFRA_TOWN_HALL},
	 *            {@link #INFRA_TEMPLES}
	 * @param temples
	 *            the number of temples to build (only is considered if
	 *            {@link #INFRA_TEMPLES} is set)
	 * @param houses
	 *            the number of houses to build
	 * @see #rebuild
	 */
	public TownMap(final int sizex, final int sizey, final int infra, final int temples, final int houses) {
		inhabitants = -1;
		rebuild(sizex, sizey, infra, temples, houses);
	}

	/**
	 * @param town
	 *            the town to create the map for
	 */
	public TownMap(final TownType town) {
		inhabitants = town.getNumberOfInhabitants().intValue();
		int sizex = 2 + (int)(Math.round(Math.sqrt(inhabitants)));
		int sizey = 2 + (int)(Math.round(Math.sqrt(inhabitants)));
		int infra = INFRA_EMPTY;
		if (Town.calcSizeCategory(town, false).getIDnum() > 2) {
			infra |= INFRA_MARKET;
		}
		if (MathUtil.random(6) < Town.calcSizeCategory(town).getIDnum()) {
			infra |= INFRA_TOWN_WALL;
			sizex += 2;
			sizey += 2;
		}
		if (MathUtil.random(6) < Town.calcSizeCategory(town).getIDnum()) {
			infra |= INFRA_TOWN_HALL;
		}
		infra |= INFRA_TEMPLES;
		rebuild(sizex, sizey, infra, town.getTemple().size(), (inhabitants + 4) / 5);
	}

	/**
	 * Does all the work for building or rebuilding a town.
	 * 
	 * @param sizex
	 *            the x size of the town map in house grid coordinates
	 * @param sizey
	 *            the y size of the town map in house grid coordinates
	 * @param infra
	 *            {@link #INFRA_EMPTY} for none, or the OR-ed values of the
	 *            to-be-built infrastructure {@link #INFRA_MARKET},
	 *            {@link #INFRA_TOWN_WALL}, {@link #INFRA_TOWN_HALL},
	 *            {@link #INFRA_TEMPLES}
	 * @param temples
	 *            the number of temples to build (only is considered if
	 *            {@link #INFRA_TEMPLES} is set)
	 * @param houses
	 *            the number of houses to build
	 */
	protected void rebuild(final int sizex, final int sizey, final int infra, int temples, int houses) {
		temples = ((temples < 0) ? (MathUtil.random(sx*sy/200) + 1) : temples);
		sx   = sizex;
		sy   = sizey;
		mp   = new char[sizey * sizex];
		clearMap();
		boolean mapHasWall = true;
		if (0 != (infra & INFRA_TOWN_WALL)) {
			x = MathUtil.random(sx / 10);
			y = MathUtil.random(sy / 10);
			placeWall();
		} else {
			mapHasWall = false;
		}
		if (0 != (infra & INFRA_MARKET)) {
			int marktgroesse = getMarketMinSize() + MathUtil.random((sx * sy) / 1000);
			x = randx();
			y = randy();
			placeMarket(marktgroesse);
		}
		if (houses < 0) {
			houses = sx * sy / getHouseMinSize() / (6 - (mapHasWall ? 0 : 1) - MathUtil.random(2));
		}
		if (inhabitants < 0) {
			inhabitants = houses * 5;
		}
		x = MathUtil.random(sx / 10);
		y = MathUtil.random(sy / 10);
		placeHouse(houses, mapHasWall);
		placePublicBuilding(infra & INFRA_TOWN_HALL, mapHasWall);
		for (int i = 0; i < temples; i++) {
			placePublicBuilding(infra & INFRA_TEMPLES, mapHasWall);
		}
	}

	/**
	 * @param tf
	 *            the text formatter to use
	 * @return a text representation of the town map
	 */
	public String getContentText(final TextFormatter tf) {
		final StringBuffer ret = new StringBuffer();
		boolean isPreformatted = false;
		if (TextDocumentType.FIXEDFONT != tf.getLocalDocType()) {
			ret.append(tf.beginPreformatted());
			isPreformatted = true;
		}
		ret.append(tf.beginParagraph());
		for (int i = 0; i < sy; i++) {
			for (int j = 0; j < sx; j++) {
				ret.append(mp[posxy(j, i)]);
			}
			ret.append(tf.newLine());
		}
		ret.append(tf.endParagraph());
		ret.append(tf.beginParagraph());
		ret.append(getLegend());
		ret.append(tf.endParagraph());
		if (isPreformatted) {
			ret.append(tf.endPreformatted());
		}
		return ret.toString();
	}

	/**
	 * @return a legend to explain all graphical elements
	 */
	public static String getLegend() {
		final Messages messages = DSAEnv.instance().messages();
		final StringBuffer ret = new StringBuffer();
		ret.append(messages.getText("TownMap.Legend"));
		ret.append(TextFormatter.NL);
		ret.append(createLegendEntry(REP_STREET, "TownMap.Legend.Street"));
		ret.append(createLegendEntry(REP_HOUSE,  "TownMap.Legend.Building"));
		ret.append(createLegendEntry(REP_MARKET, "TownMap.Legend.Marketplace"));
		ret.append(createLegendEntry(REP_TOWN_HALL, "TownMap.Legend.Townhall"));
		ret.append(createLegendEntry(REP_TEMPLE, "TownMap.Legend.Temple"));
		ret.append(createLegendEntry(REP_WALL,   "TownMap.Legend.Wall"));
		ret.append(createLegendEntry(REP_GATE,   "TownMap.Legend.Gate"));
		ret.append(createLegendEntry(REP_UNDEF,  "TownMap.Legend.Others"));
		return ret.toString();
	}

	private static String createLegendEntry(char rep, String key) {
		final StringBuffer ret = new StringBuffer();
		ret.append(rep);
		ret.append(TAB);
		ret.append(DSAEnv.instance().messages().getText(key));
		ret.append(TextFormatter.NL);
		return ret.toString();
	}

	/**
	 * @param xp
	 *            the x-position in house grid coordinates
	 * @param yp
	 *            the y-position in house grid coordinates
	 * @return the item that's at position (xp, yp)
	 */
	public char getItem (final int xp, final int yp) {
		return (mp[posxy(xp,yp)]);
	}

	public int getMarketMinSize() {
		if (inhabitants < 0) {
			return 36;
		}
		int cat = Town.calcSizeCategory(inhabitants, false).getIDnum();
		return (cat * cat);
	}

	public int getTownHallSize() {
		if (inhabitants < 0) {
			return 4;
		}
		return ((1 + Town.calcSizeCategory(inhabitants, false).getIDnum()) / 2);
	}

	public int getTempleSize() {
		if (inhabitants < 0) {
			return 4;
		}
		return ((1 + Town.calcSizeCategory(inhabitants, false).getIDnum()) / 2);
	}

	public int getHouseMinSize() {
		return 1;
	}

	protected int posxy(int xp, int yp) {
		xp = MathUtil.bound(xp, 0, sx-1);
		yp = MathUtil.bound(yp, 0, sy-1);
		return (yp * sx + xp);
	}

	protected int randx() {
		return (MathUtil.random(sx - sx/10) + sx/10);
	}

	protected int randy() {
		return (MathUtil.random(sy - sy/10) + sy/10);
	}

	/**
	 * Up or down? Left or right?
	 * 
	 * @return -1 or +1 (equally distributed)
	 */
	protected int randMoveAlways() {
		return ((MathUtil.random(2) == 0) ? -1 : 1);
	}

	/**
	 * Up, stay or down? Left, stay or right?
	 * 
	 * @return -1, 0 or +1 (equally distributed)
	 */
	protected int randMoveOften() {
		return (MathUtil.random(3) - 1);
	}

	/**
	 * Up, stay or down? Left, stay or right?
	 * 
	 * @return -1, 0 or +1 (with 0 in 7 of 9 cases)
	 */
	protected int randMoveSeldom() {
		switch (MathUtil.random(9)) {
		case 0:
			return -1;
		case 1:
			return  1;
		default:
			return  0;
		}
	}

	/**
	 * Clears the whole town map.
	 */
	protected void clearMap() {
		for (int i = 0; i < sx; i++) {
			for (int j = 0; j < sy; j++) {
				mp[posxy (i,j)] = REP_UNDEF;
			}
		}
	}

	/**
	 * If there are town walls, people prefer to live inside the walls. So this
	 * routine decides for a spot, if it's a preferred area. It's not always
	 * correct, to make living more interesting.
	 * 
	 * @param xp
	 *            the x-position in house grid coordinates
	 * @param yp
	 *            the y-position in house grid coordinates
	 * @param mapHasWall
	 *            if the town has a surrounding wall
	 * @return <code>true</code>, if the actual position is inside the town
	 */
	protected boolean isPreferredSpot(int xp, int yp, final boolean mapHasWall) {
		if (!mapHasWall) {
			// The question does not make sense, as there is no wall and no inside. So each spot is equally preferred.
			return true;
		}
		MathVector direction = MathVector.create(0.0, -1.0);
		int innenRichtung = 4;
		while(true) {
			if (mp[posxy(xp,yp)] != REP_UNDEF) {
				// Turn around
				direction.rotateSelfQuarter(MathVector.X, MathVector.Y, 1);
				innenRichtung--;
			}
			if ((xp <= 1) || (yp <= 1) || (xp > sx-1) || (yp > sy-1)) {
				return false;
			}
			if (innenRichtung == 0) {
			 	return true;
			}
			// Move to the next position
			xp += Math.round(direction.get(0));
			yp += Math.round(direction.get(1));
		}
	}

	/**
	 * Puts the marketplace of given size in place. The size is not guaranteed.
	 * <p>
	 * Changed by Peter Diefenbach: The marketplace now doesn't overwrite town walls.
	 * 
	 * @param size the size the marketplace should have 
	 */
	protected void placeMarket(int size) {
		boolean directionX = true;
		int     nx, ny;
		while(0 < size) {
			// First, find a new position in the neighborhood of the old position
			nx = x;
			ny = y;
			if (directionX) {
				nx = x + randMoveAlways();
				if (nx > sx) {
					nx -= 2;
				}
				if (nx < 0) {
					nx += 2;
				}
			} else {
				ny = y + randMoveAlways();
				if (ny > sy) {
					ny -= 2;
				}
				if (ny < 0) {
					ny += 2;
				}
			}
			directionX = (!directionX);
			// Now, set the dot as marketplace, if possible
			size--;
			char c = mp[posxy(nx, ny)];
			if (c == REP_WALL || c == REP_GATE) {
				// do nothing
			} else {
				x = nx;
				y = ny;
				mp[posxy(x,y)] = REP_MARKET;
			}
		}
	}

	/**
	 * Draws a closed polygon, defined by the strongpoints. Each line of the
	 * polygon is a bit randomized, but every strongpoint is reached exactly.
	 * 
	 * @param strongpoints
	 *            the corners of the polygon
	 */
	protected void intrapolateWall(final MathVector[] strongpoints) {
		// Now we draw the lines between them.
		final MathVector actualPos = new MathVector(strongpoints[0]);
		final MathVector diff      = new MathVector(2);
		final double[] probXY = new double[2];
		int gatesCount = (sx + sy) / (1 + MathUtil.random(3));
		char wallChar;
		for (int thisIndex = 0; thisIndex < strongpoints.length; thisIndex++) {
			final int nextIndex = (thisIndex + 1) % strongpoints.length;
			do {
				// Is it a gate or a normal wall?
				wallChar = REP_WALL;
				gatesCount--;
				if (gatesCount <= 3) {
					if (gatesCount == 0) {
						gatesCount = (sx + sy) / (1 + MathUtil.random(3));
					} else {
						wallChar = REP_GATE;
					}
				}
				// Set wall or gate to the actual position
				mp[posxy(MathUtil.bound((int)(actualPos.get(0)), 0, sx-1),
						 MathUtil.bound((int)(actualPos.get(1)), 0, sy-1))] = wallChar;
				// Where to go next?
				diff.copyFrom(strongpoints[nextIndex]);
				diff.subtractSelf(actualPos);
				if (diff.isNullVector(0.2)) {
					break;
				}
				for (int i = 0; i < 2; i++) {
					probXY[i] = Math.abs(diff.get(i));
				}
				final int directionXY = MathUtil.randomDistributed(probXY);
				actualPos.set(directionXY, actualPos.get(directionXY) +
				MathUtil.sgn(diff.get(directionXY)));
			} while(true);
		}
	}

	/**
	 * Defines and draws the town wall.
	 */
	protected void placeWall() {
		// These points are element of the wall. Between those we will draw randomized lines.
		final MathVector[] strongpoints;
		if (0 == MathUtil.random(4)) {
			// We do the wall squared
			final int steps = 5;
			final int xpoints = 1 + (sx - 2) / steps;
			final int ypoints = 1 + (sy - 2) / steps;
			strongpoints = new MathVector[2 * (xpoints + ypoints)];
			for (int i = 0; i < xpoints; i++) {
				strongpoints[i                        ] = MathVector.create(     1 + steps * i,      1);
				strongpoints[i +     xpoints + ypoints] = MathVector.create(sx - 2 - steps * i, sy - 2);
			}
			for (int i = 0; i < ypoints; i++) {
				strongpoints[i +     xpoints          ] = MathVector.create(sx - 2,      1 + steps * i);
				strongpoints[i + 2 * xpoints + ypoints] = MathVector.create(     1, sy - 2 - steps * i);
			}
		} else {
			// We draw a circle or ellipse as town wall
			strongpoints = new MathVector[(sx + sy) / 3];
			for (int i = 0; i < strongpoints.length; i++) {
				final double angle = (2.0 * Math.PI * i) / strongpoints.length;
				strongpoints[i] = MathVector.create(
						Math.floor(0.5 + ((double)(sx) / 2.0) * (1 + 0.9 * Math.sin(angle))),
						Math.floor(0.5 + ((double)(sy) / 2.0) * (1 + 0.9 * Math.cos(angle))));
			}
		}
		// Now, don't do it too exactly, shake the points a bit
		for (int i = 0; i < strongpoints.length; i++) {
			for (int j = 0; j < 2; j++) {
				final int index = j;
				strongpoints[i].set(index, strongpoints[i].get(j) + randMoveOften());
			}
		}
		intrapolateWall(strongpoints);
	}

	protected void placeHouse(int housesCount, final boolean mapHasWall) {
		int x1 = randx(), y1 = randy();
		int dx = randMoveOften(), dy = randMoveOften();
		int watchdog = 40000;
		for(watchdog = 40000; (watchdog > 0) && (housesCount > 0); watchdog--) {
			int str = 0, hstr = 0, dstr = 0;
			while ((str <= 8) && ((dx == 0 && dy == 0) || (MathUtil.random(8) == 0)  || (mp[posxy(x + dx, y + dy)] == REP_HOUSE))) {
				str++;
				dx = randMoveOften();
				dy = randMoveOften();
			}
			// Calculate new position, handle overflows
			x1 =x1 + dx;
			if (x1 >= sx-1) {
				x1 = 1;
				dx = randMoveOften();
				dy = randMoveOften();
			}
			if (x1 < 1) {
				x1 = sx-2;
				dx = randMoveOften();
				dy = randMoveOften();
			}
			y1 = y1 + dy;
			if (y1 >= sy-1) {
				y1 = 1;
				dx = randMoveOften();
				dy = randMoveOften();
			}
			if (y1 < 1) {
				y1 = sy-2;
				dx = randMoveOften();
				dy = randMoveOften();
			}
			// May I set a house?
			if (isPreferredSpot(x1, y1, mapHasWall) && mp[posxy(x1,y1)]==REP_UNDEF) {
				if (mp[posxy(x1-1,y1  )]==REP_UNDEF) str++;
				if (mp[posxy(x1  ,y1-1)]==REP_UNDEF) str++;
				if (mp[posxy(x1+1,y1  )]==REP_UNDEF) str++;
				if (mp[posxy(x1  ,y1+1)]==REP_UNDEF) str++;
				if (mp[posxy(x1-1,y1  )]==REP_HOUSE) hstr++;
				if (mp[posxy(x1  ,y1-1)]==REP_HOUSE) hstr++;
				if (mp[posxy(x1+1,y1  )]==REP_HOUSE) hstr++;
				if (mp[posxy(x1  ,y1+1)]==REP_HOUSE) hstr++;
				if (mp[posxy(x1-1,y1-1)]==REP_HOUSE) dstr++;
				if (mp[posxy(x1+1,y1-1)]==REP_HOUSE) dstr++;
				if (mp[posxy(x1+1,y1+1)]==REP_HOUSE) dstr++;
				if (mp[posxy(x1-1,y1+1)]==REP_HOUSE) dstr++;
				// Is a gate in the direct neighborhood?
				if (mp[posxy(x1-1,y1  )]!=REP_GATE &&
					mp[posxy(x1  ,y1-1)]!=REP_GATE &&
					mp[posxy(x1+1,y1  )]!=REP_GATE &&
					mp[posxy(x1  ,y1+1)]!=REP_GATE &&
					mp[posxy(x1-1,y1-1)]!=REP_GATE &&
					mp[posxy(x1+1,y1-1)]!=REP_GATE &&
					mp[posxy(x1+1,y1+1)]!=REP_GATE &&
					mp[posxy(x1-1,y1+1)]!=REP_GATE)
				{
					if ((dstr==1 && hstr<=1 && str>=1 && str<=3 && (dx!=0) && (dy!=0)) ||
						(str<=2 && dstr<=2 && hstr==3) ||
						(str<=3 && str>1 && (0 == dstr)) ||
						((0 == hstr) && (0 == dstr)))
					{
						mp[posxy(x1,y1)] = REP_HOUSE;
						housesCount--;
					}
				}
			}
		}
	}

	/**
	 * Puts a public building (temple or town hall) in place.
	 * <p>
	 * Change by Peter Diefenbach: The building now doesn't overwrite town walls
	 * 
	 * @param inf
	 *            one of {@link #INFRA_TOWN_HALL} or {@link #INFRA_TEMPLES}
	 * @param mapHasWall
	 *            if the map has town walls
	 */
	protected void placePublicBuilding(final int inf, final boolean mapHasWall) {
		char buildingType = REP_UNDEF;
		int  size;
		int  sizex = 0, sizey = 0;
		char c;
		if (inf == INFRA_TEMPLES || inf == INFRA_TOWN_HALL) {
			if (inf == INFRA_TEMPLES) {
				buildingType = REP_TEMPLE;
				size   = getTempleSize();
			} else {
				buildingType = REP_TOWN_HALL;
				size   = getTownHallSize();
			}
			// Try to find a spot for the building.
			boolean okay = false;
			for (int tries = 5; ((tries > 0) && !okay); tries++) {
				okay = true;
				sizex = MathUtil.bound((size + MathUtil.random(3) - 1), 1, 5);
				sizey = MathUtil.bound((size + MathUtil.random(3) - 1), 1, 5);
				do {
					x = MathUtil.random(sx - sizex);
					y = MathUtil.random(sy - sizey);
				}
				while (!isPreferredSpot (x, y, mapHasWall));
				for (int i = 0; ((i < sizex) && okay); i++) {
					for (int j = 0; ((j < sizey) && okay); j++) {
						c = mp[posxy(x + i, y + j)];
						if (c == REP_WALL || c == REP_GATE) {
							okay = false;
						}
					}
				}
			}
			// Set the building
			for (int i = 0; i < sizex; i++) {
				for (int j = 0; j < sizey; j++) {
					mp[posxy(x + i, y + j)] = buildingType;
				}
			}
		}
	}

	/** The internal representation of the map. */
	private char[] mp;
	private int x;
	private int y;
	/** The size of the map in characters in x direction. */
	private int sx;
	/** The size of the map in characters in y direction. */
	private int sy;
	/** The size of the town in inhabitants. */
	private int inhabitants;
}
