package de.ped.dsatool.logic;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import de.ped.tools.Assert;
import de.ped.tools.MathUtil;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.CalendarType;
import de.ped.dsatool.dsa.generated.SeasonType;
import de.ped.dsatool.dsa.generated.SeasonalWeatherType;
import de.ped.dsatool.dsa.generated.WeatherTableEntryType;
import de.ped.dsatool.dsa.generated.WeatherType;

/**
 * A generic weather system used in Aventurien (DSA):
 * An appropriate weather class
 * This is an implementation of the suggestion given by the Authors
 * of DSA in the listed source.  It's intention was to be easy
 * for people throwing dices, but it isn't quite appropiate for
 * the more sophisticated calculation possibilities of computers.
 *
 * <p><b>See:</b> "Das Handbuch für den Reisenden", Section "Wie wird das Wetter?" (p. 11..12)</p>
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Weather implements Cloneable {
	private static Logger logger = Logger.getRootLogger();

	/**
	 * Creates the weather randomly for the first Praios in the year 0 Hal.
	 */
	public Weather() {
		this.date = new DSADate();
		setWeatherRandom();
	}

	/**
	 * Creates the weather randomly for the given date.
	 * 
	 * @param date
	 *            The DSA data
	 */
	public Weather(final DSADate date) {
		this.date = date;
		setWeatherRandom();
	}

	/**
	 * Creates the weather as given.
	 * 
	 * @param date
	 *            The DSA date for which the weather is defined
	 * @param theWeather
	 *            The weather dice eyes (in the range from 1 to 20)
	 * @param theWind
	 *            The wind category (in the range from 0 to 8)
	 * @param theTemperature
	 *            The temperature category (in the range from 0 to 8)
	 */
	public Weather(final DSADate date, final int theWeather, final int theWind, final int theTemperature) {
		this.date = date;
		try {
			setWeather(theWeather, theWind, theTemperature);
		} catch (final IndexOutOfBoundsException e) {
			setWeatherRandom();
		}
	}

	/**
	 * @return the day for which the weather is defined.
	 */
	public DSADate getDate() {
		return date;
	}

	/**
	 * @return the weather dice eyes (in the range from 1 to 20)
	 */
	public int getWeather() {
		return weather;
	}

	/**
	 * @return The wind category (in the range from 0 to 8)
	 */
	public int getWind() {
		return wind;
	}

	/**
	 * @return The temperature category (in the range from 0 to 8)
	 */
	public int getTemperature() {
		return temperature;
	}

	/**
	 * @param date
	 *            The DSA date for which the weather is defined
	 */
	public void setDate(final DSADate date) {
		this.date = date;
	}

	/**
	 * @param weather
	 *            The weather dice eyes (in the range from 1 to 20)
	 */
	public void setWeather(final int weather) throws IndexOutOfBoundsException {
		if ((weather < 1) || (weather > 20)) {
			throw new IndexOutOfBoundsException("Weather out of bounds"); 
		}
		this.weather = weather;
	}

	/**
	 * @param wind
	 *            The wind category (in the range from 0 to 8)
	 */
	public void setWind(final int wind) throws IndexOutOfBoundsException {
		if ((wind < 0) || (wind > 8)) {
			throw new IndexOutOfBoundsException("Wind out of bounds"); 
		}
		this.wind = wind;
	}

	/**
	 * @param temperature
	 *            The temperature category (in the range from 0 to 8)
	 */
	public void setTemperature(final int temperature) throws IndexOutOfBoundsException {
		if ((temperature < 0) || (temperature > 8)) {
			throw new IndexOutOfBoundsException("Temperature out of bounds"); 
		}
		this.temperature = temperature;
	}

	/**
	 * Sets the weather, wind and temperature condition randomly.
	 */
	public void setWeatherRandom() {
		weather		= Dice.roll(1,20);
		wind		= MathUtil.random(5) + 2;
		temperature	= MathUtil.random(5) + 2;
	}

	/**
	 * Sets the weather, wind and temperature condition as given.
	 * @param weather
	 *            The weather dice eyes (in the range from 1 to 20)
	 * @param wind
	 *            The wind category (in the range from 0 to 8)
	 * @param temperature
	 *            The temperature category (in the range from 0 to 8)
	 */
	public void setWeather(final int weather, final int wind, final int temperature) throws IndexOutOfBoundsException {
		setWeather(weather);
		setWind(wind);
		setTemperature(temperature);
	}

	/**
	 * @return the next day, based on the weather of this day.
	 */
	public Weather createNextDay() {
		// Initialize the new date, incremented by one
		final Weather newWeather = new Weather((DSADate)getDate().clone());
		newWeather.date.addDays(1);
		// Calculate the new general weather scheme
		try {
			newWeather.weather = MathUtil.modulo(getWeather(), 20, 1);
			final WeatherTableEntryType ent = newWeather.getTableEntry();
			if (null != ent) {
				newWeather.setWeather(MathUtil.modulo(newWeather.weather + Dice.roll(1,6) + ent.getChange(), 20, 1));
				// Calculate the new wind
				newWeather.setWind(MathUtil.bound(doubleDice(getWind()) + ent.getWind(), 0, 8));
				// Calculate the new temperature
				newWeather.setTemperature(MathUtil.bound(doubleDice(getTemperature()) + ent.getTemperature(), 0, 8));
			}
		} catch (final IndexOutOfBoundsException e) {
			logger.fatal("This should never happen: ", e); 
		}
		return (newWeather);
	}

	/**
	 * @return the right WeatherTableEntry. This decision is based on the values
	 *         of getDate() and getWeather() of the object.
	 */
	protected WeatherTableEntryType getTableEntry() {
		// Get the right weather tab, dependent on the season
		WeatherTableEntryType ent = null;
		final List<WeatherTableEntryType> weathers = Weather.getWeatherTableFor(getDate().season());
		if (null != weathers) {
			final Iterator<WeatherTableEntryType> it = weathers.iterator();
			// Find the right entry
			while (it.hasNext()) {
				ent = (WeatherTableEntryType)(it.next());
				if (ent.getTillDiceEyes() >= getWeather()) {
					break;
				}
			} 
		}
		return ent;
	}

	/**
	 * throws two dices and chooses the result nearer to the given previous
	 * value. This procedure is needed for the algorithm for finding the weather
	 * of the next day.
	 * 
	 * @param oldValue
	 *            the previous value
	 * @return an intermediate result for the current value
	 */
	protected int doubleDice(int oldValue) {
		final int w1 = Dice.roll(1, 6);
		final int w2 = Dice.roll(1, 6);
		int diff = Math.abs(w1 - oldValue) - Math.abs(w2 - oldValue);
		if (0 == diff) {
			// If both dices have the same distance, flip a coin to choose the dice.
			diff = ((Dice.roll(1, 2) == 0) ? w1 : w2);
		}
		if (0 < diff) {
			return (w1);
		}
		return (w2);
	}

	/**
	 * @return the names of all possible weather states.
	 * 
	 *         This method can't be static as the list values change with the
	 *         season
	 */
	public String[] getWeatherNames() {
		return getWeatherNames(this.date);
	}

	/**
	 * @param date
	 *            The DSA date for which the weather is defined
	 * @return the names of all possible weather states for the given date
	 */
	public static String[] getWeatherNames(final DSADate date) {
		final String[] ret = new String[20];
		final Weather w = new Weather(date);
		for (int i = 0; i < 20; i++) {
			w.weather = i + 1;
			ret[i] = TextFormatter.format(i + 1, 2) + ": " + w.getTableEntry().getName();
		}
		return ret;
	}

	public static Vector<String> getAllWeatherNames() {
		final Vector<String> v = new Vector<String>();
		final Iterator<SeasonalWeatherType> itSeasons = DSA.instance().getWeather().getSeasonalWeather().iterator();
		while (itSeasons.hasNext()) {
			final SeasonalWeatherType weather = (SeasonalWeatherType)itSeasons.next();
			final Iterator<WeatherTableEntryType> itEntry = weather.getWeatherTableEntry().iterator();
			while (itEntry.hasNext()) {
				final WeatherTableEntryType wte = (WeatherTableEntryType)itEntry.next();
				v.add(TextFormatter.format(0, 2) + ": " + wte.getName());
			}
		}
		return v;
	}

	@Override
	public Weather clone() {
		Weather ret = null;
		try {
			ret = (Weather)super.clone();
			ret.date = (DSADate)date.clone();
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}

	/**
	 * @param season
	 *            For this season the weather table is searched.
	 * @return the weather table for the given season.
	 */
	public static List<WeatherTableEntryType> getWeatherTableFor(final SeasonType season) {
		final WeatherType wt = DSA.instance().getWeather();
		if (null == wt) {
			return null;
		} else {
			final Iterator<SeasonalWeatherType> it = wt.getSeasonalWeather().iterator();
			SeasonalWeatherType forOtherSeasons = null;
			SeasonalWeatherType forThisSeason = null;
			while (it.hasNext() && (null == forThisSeason)) {
				final SeasonalWeatherType weather = (SeasonalWeatherType)it.next();
				final SeasonType s = (SeasonType)weather.getSeason();
				if (null == s) {
					forOtherSeasons = weather;
				} else if (s.getIDnum().equals(season.getIDnum())) {
					forThisSeason = weather;
				}
			}
			if ((null == forThisSeason) && (null != forOtherSeasons)) {
				forThisSeason = forOtherSeasons;
			}
			if (null == forThisSeason) {
				return null;
			} else {
				return forThisSeason.getWeatherTableEntry();
			}
		}
	}

	public CalendarType getCalendar() {
		return calendar;
	}

	public void setCalendar(final CalendarType calendar) {
		this.calendar = calendar;
	}

	private DSADate date;
	private CalendarType calendar = DSACalendar.findDefault();
	private int weather;
	private int wind;
	private int temperature;
}
