package de.ped.dsatool.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.junit.Test;

import de.ped.dsatool.dsa.generated.CalendarType;
import de.ped.dsatool.logic.DSA;
import de.ped.dsatool.logic.DSACalendar;
import de.ped.dsatool.logic.DSADate;
import de.ped.dsatool.logic.DSADateCalendarTwelve;
import de.ped.dsatool.logic.DSAException;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.test.GenericTest;

/**
 * Tests {@link DSADate}.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DSADateTest {
	private Logger logger = Logger.getLogger(getClass());

	public DSADateTest() {
		logger.setLogLevel(Level.INFO);
	}

	@SuppressWarnings("unused")
	private boolean isEven(int n) {
		return (0 == (n & 0x1));
	}

	public static final int avDateOfGregorianChange = 70289;

	public void checkToAndFromEarthDate(DSADate dsaDate) {
		final Date earthDate = dsaDate.toEarthDate();
		logger.trace(dsaDate + " --> " + earthDate);
		final DSADate dsaDate2 = new DSADate(earthDate);
		final GregorianCalendar earthCal = new GregorianCalendar();
		earthCal.setTime(earthDate);
		final int earthYear = earthCal.get(GregorianCalendar.YEAR);
		if (earthYear < 1996) {
			final int daysSinceBF = dsaDate.getDaysSinceBF();
			if ((avDateOfGregorianChange - 2 * DSADate.GregorianCalendarHelper.DAYS_SKIPPED <= daysSinceBF) 
					&& (daysSinceBF < avDateOfGregorianChange)) {
				// The gregorian calendar change phase on earth - 
				// these 10 days are chaotic, so we don't demand bidirectional uniqueness.
			} else {
				final int diff = dsaDate2.getDaysSinceBF() - daysSinceBF;
				if ((0 == diff) || (diff == - 1)) {
					// Nothing to do
				} else {
					assertEquals(daysSinceBF, dsaDate2.getDaysSinceBF());
				}
			}
		} else {
			assertEquals(dsaDate.getDaysSinceBF(), dsaDate2.getDaysSinceBF());
		}
	}

	public void checkFromEarthDate(GregorianCalendar cal, int year, int month, int day) {
		final SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
		final String expected = String.format("%04d-%02d-%02d", year, month, day);
		cal.set(year, month - 1, day);
		assertEquals(expected, dateFormat.format(cal.getTime()));
		final Date earthDate = cal.getTime();
		final DSADate dsaDate = new DSADate(earthDate);
		final Date earthDate2 = dsaDate.toEarthDate();
		cal.setTime(earthDate2);
		assertEquals(expected, dateFormat.format(cal.getTime()));
	}

	@Test
	public void testFromEarthDate() {
		final GregorianCalendar cal = new GregorianCalendar();
		// From the past
		checkFromEarthDate(cal, 1581, 12, 31);
		checkFromEarthDate(cal, 1582, 05, 01);
		checkFromEarthDate(cal, 1582, 10, 03);
		checkFromEarthDate(cal, 1582, 10, 04);
//    	checkFromEarthDate(cal, 1582, 10, 05);

		// From the future
		checkFromEarthDate(cal, 1983, 05, 01);
		checkFromEarthDate(cal, 1583, 01, 01);
		checkFromEarthDate(cal, 1582, 12, 31);
		checkFromEarthDate(cal, 1582, 12, 01);
		checkFromEarthDate(cal, 1582, 11, 01);
	}

	@Test
	public void testToAndFromEarthDate() {
		for (int i = 993; i <= 1031; i++) {
			checkToAndFromEarthDate(new DSADate(i * DSADate.DAYS_PER_EARTH_YEAR));
		}
		checkToAndFromEarthDate(new DSADate());
		for (int i = -200 * DSADate.DAYS_PER_EARTH_YEAR; i <= 2000 * DSADate.DAYS_PER_EARTH_YEAR; i++) {
			checkToAndFromEarthDate(new DSADate(i));
		}
	}

	@Test
	public void testSetCalendar() throws DSAException {
		final CalendarType calendar = DSACalendar.findDefault();
		final HashMap<CalendarType, DSADateCalendarTwelve> cals = DSACalendar.getCalendars();
		final DSADateCalendarTwelve cal = cals.get(calendar);
		cal.setDayMonthYear(12, 03, 20);
		final List<CalendarType> calendars = DSA.instance().getCalendar();
		final Iterator<CalendarType> it = calendars.iterator();
		while (it.hasNext()) {
			final CalendarType calender2 = (CalendarType) it.next();
			final DSADateCalendarTwelve cal2 = DSACalendar.getCalendars().get(calender2);
			cal2.setDSADate(cal.getDSADate());
			assertEquals(12, cal2.getDay());
			assertEquals(03, cal2.getMonthAsNumber());
		}
	}

	@Test
	public void testFindDefaultCalendar() {
		final CalendarType calendar = DSACalendar.findDefault();
		assertNotNull(calendar);
		assertEquals(DSACalendar.DEFAULT_CALENDAR_ID, calendar.getID());
		assertTrue(calendar.isHasYearZero());
	}

	@Test
	public void testCalendarGetAllStrings() {
		final List<String> strs = DSACalendar.getAllStrings();
		assertNotNull(strs);
		assertTrue(0 < strs.size());
		assertEquals(DSA.instance().getCalendar().size(), strs.size());
	}

	/**
	 * @see DSADate#equals(Object)
	 * @see DSADate#clone()
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(true, new DSADate(0), new DSADate(1));
	}
}
