package de.ped.dsatool.web;

import java.util.Enumeration;

import javax.servlet.http.HttpServletRequest;

import de.ped.dsatool.logic.DSA;
import de.ped.dsatool.logic.DSAEnv;
import de.ped.dsatool.logic.DSAException;
import de.ped.dsatool.logic.MasterDataModel;
import de.ped.dsatool.logic.Model;
import de.ped.dsatool.logic.Model.ModelMode;
import de.ped.dsatool.logic.MasterDataModel.MasterDataType;
import de.ped.tools.TextDocumentType;
import de.ped.tools.HtmlHelper;
import de.ped.tools.MathVector;
import de.ped.tools.Messages;
import de.ped.tools.TextFormatter;

/**
 * For the web interface, all parameter keys and values as constants,
 * and how they are read and parsed. 
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Parameters {


	public static final String JSP_PRE = Model.URL + "?";

	/**
	 * Builds up HTML code representing the main menu for DSATool.
	 * @param model The model containing the current master data model.
	 * @return HTML code.
	 */
	public static String buildMainMenu(final Model model) {
		final StringBuffer ret = new StringBuffer();
		ret.append("<span");
		ret.append(TextFormatter.htmlAttribute(TextFormatter.CLASS, "tabbar"));
		ret.append(">");
		ret.append(
				mainTab(model, "Action.ViewDate", Model.ACTION_DATE_SHOW, Model.ModelMode.DATE) + " " +
				mainTab(model, "Action.ViewTown", Model.ACTION_TOWN_SHOW, Model.ModelMode.TOWN) + " " +
				mainTab(model, "Action.ViewNames", Model.ACTION_NAMES_SHOW, Model.ModelMode.NAMES) + " " +
				mainTab(model, "Action.ViewPlants", Model.ACTION_PLANTS_SHOW, Model.ModelMode.PLANT) + " " +
				mainTab(model, "Action.ViewQuest", Model.ACTION_QUEST_SHOW, Model.ModelMode.QUEST) + " " +
				mainTab(model, "Action.ViewDice", Model.ACTION_DICE_SHOW, Model.ModelMode.DICE) + " " +
				mainTab(model, "Action.ViewMap", Model.ACTION_MAP_SHOW, null) + " " +
				mainTab(model, "Menu.MasterData", Model.ACTION_MASTER_DATA_SHOW, Model.ModelMode.MASTER_DATA));
		ret.append("</span>");
		return ret.toString();
	}

	public static String mainTab(final Model model, final String textKey, final String action, ModelMode modelMode) {
		final boolean isLink;
		if (null == modelMode) {
			isLink = !model.isMap();
		} else {
			isLink = model.isMap() || (modelMode != model.getMode());
		}
		return tab("mainMenuEntry", textKey, action, null, isLink);
	}

	/**
	 * Builds up HTML code representing the submenu for master data modes.
	 * @param model The model containing the current master data model.
	 * @return HTML code.
	 */
	public static String buildMasterDataMenu(final Model model) {
		final int mode = model.getMasterDataModel().getMode().ordinal();
		final StringBuffer ret = new StringBuffer();
		final int max = MasterDataModel.MasterDataType.values().length - 1;
		ret.append("<span");
		ret.append(TextFormatter.htmlAttribute(TextFormatter.CLASS, "tabbar"));
		ret.append(">");
		for (int i = 0; i <= max; i++) {
			if (0 < i) {
				ret.append(" ");
			}
			ret.append(tab("subMenuEntry", MasterDataModel.getMessageKey(i), Model.ACTION_MASTER_DATA_SHOW,
					Integer.toString(i), (mode != i)));
		}
		ret.append("</span>");
		return ret.toString();
	}

	public static String tab(final String cssClass, final String textKey, final String action,
			final String masterDataMode, final boolean isLink) {
		final StringBuffer ret = new StringBuffer();
		final Messages messages = DSAEnv.instance().messages();
		String href = HtmlHelper.addParameter(Model.URL, Model.ACTION, action);
		if (null != masterDataMode) {
			href = HtmlHelper.addParameter(href, Model.MASTER_DATA_MODE, masterDataMode);
		}
		ret.append("<span");
		ret.append(TextFormatter.htmlAttribute(TextFormatter.CLASS,
				(null == cssClass ? "" : cssClass + " ") + "tab " + (isLink ? "deselected" : "selected")));
		ret.append(TextFormatter.htmlAttribute(TextFormatter.TITLE, messages.getShortDescription(textKey)));
		ret.append(">");
		ret.append(HtmlHelper.linkOrShow(
				TextFormatter.sgmlify(TextDocumentType.XHTML10_STRICT, messages.getText(textKey)),
				href, isLink));
		ret.append("</span>");
		return ret.toString();
	}

	/**
	 * Reads all parameters for index.jsp.
	 * @param request The request containing the parameters.
	 * @param model The model to update.
	 * @return A string describing in human words if something went wrong.
	 */
	public static String readParameters(final HttpServletRequest request, final Model model) {
		final StringBuffer errorMsgs = new StringBuffer();
		// Browsers send a click to a map (clickable image) in a strange format -
		// a hardcoded <path>?<x>,<y> as URL. So I have to catch this very early.
		MathVector pos = null;
		final Enumeration<?> params = request.getParameterNames();
		while (params.hasMoreElements() && (null == pos)) {
			final String name = (String)params.nextElement();
			final String[] nums = name.split(",");
			if (nums.length >= 2) {
				try {
					int xpos = Integer.parseInt(nums[0]);
					int ypos = Integer.parseInt(nums[1]);
					pos = MathVector.create(xpos, ypos);
				} catch (final NumberFormatException ex) {
					// Parameters are optional, it's okay to not get them.
				} catch (final NullPointerException ex) {
					// Parameters are optional, it's okay to not get them.
				}
			}
		}
		if (null == pos) {
			// Try to read parameters called "x" and "y" instead
			try {
				final int xpos = Integer.parseInt(request.getParameter(Model.MAP_POS_X));
				final int ypos = Integer.parseInt(request.getParameter(Model.MAP_POS_Y));
				pos = MathVector.create(xpos, ypos);
			} catch (final NumberFormatException ex) {
				// Parameters are optional, it's okay to not get them.
			} catch (final NullPointerException ex) {
				// Parameters are optional, it's okay to not get them.
			}
		}
		if (null != pos) {
			pos.setAllFrom((pos.get(0) - 1) / 1000.0, (pos.get(1) - 1) / 1500.0);
			model.setCurrentPositionOnAventuriaMap(pos);
			model.getNamesModel().setAventuriaMapPosition(pos);
		}
		// Here we look for our own well-defined parameters.
		final String action = request.getParameter(Model.ACTION);
		if (null != action) {
			if (action.equalsIgnoreCase(Model.ACTION_DATE_SHOW)) {
				model.setMap(false);
				model.setMode(Model.ModelMode.DATE);
			} else if (action.equalsIgnoreCase(Model.ACTION_TOWN_SHOW)) {
				model.setMap(false);
				model.setMode(Model.ModelMode.TOWN);
			} else if (action.equalsIgnoreCase(Model.ACTION_NAMES_SHOW)) {
				model.setMap(false);
				model.setMode(Model.ModelMode.NAMES);
			} else if (action.equalsIgnoreCase(Model.ACTION_PLANTS_SHOW)) {
				model.setMap(false);
				model.setMode(Model.ModelMode.PLANT);
			} else if (action.equalsIgnoreCase(Model.ACTION_QUEST_SHOW)) {
				model.setMap(false);
				model.setMode(Model.ModelMode.QUEST);
			} else if (action.equalsIgnoreCase(Model.ACTION_DICE_SHOW)) {
				model.setMap(false);
				model.setMode(Model.ModelMode.DICE);
			} else if (action.equalsIgnoreCase(Model.ACTION_MASTER_DATA_SHOW)) {
				try {
					final int masterDataMode = Integer.parseInt(request.getParameter(Model.MASTER_DATA_MODE));
					model.getMasterDataModel().setMode(MasterDataType.values()[masterDataMode]);
				} catch (final NumberFormatException ex) {
					// This is normal behavior - 
					// it is allowed to give no master data mode when switching to 
					// master data, as the model already has a value.
				} catch (final NullPointerException ex) {
					// This is normal behavior - 
					// it is allowed to give no master data mode when switching to 
					// master data, as the model already has a value.
				}
				model.setMap(false);
				model.setMode(Model.ModelMode.MASTER_DATA);
			} else if (action.equalsIgnoreCase(Model.ACTION_DATE_NEW)) {
				//String paramRandomWeather = request.getParameter(Parameters.CALENDAR_RANDOM_WEATHER);
				try {
					final int day        = Integer.parseInt(request.getParameter(Model.CALENDAR_DAY));
					final int monthID    = Integer.parseInt(request.getParameter(Model.CALENDAR_MONTH));
					final int year       = Integer.parseInt(request.getParameter(Model.CALENDAR_YEAR));
					final int calendarID = Integer.parseInt(request.getParameter(Model.CALENDAR_CALENDAR));
					final int weatherID  = Integer.parseInt(request.getParameter(Model.CALENDAR_WEATHER));
					final int windID     = Integer.parseInt(request.getParameter(Model.CALENDAR_WIND));
					final int tempID     = Integer.parseInt(request.getParameter(Model.CALENDAR_TEMPERATURE));
					final String[] expectedValues = {Model.CALENDAR_IS_WEATHER_RANDOM_RANDOM, Model.CALENDAR_IS_WEATHER_RANDOM_BY_HAND};
					final int idxWeatherRandom = HtmlHelper.readRadioParameter(request, Model.CALENDAR_IS_WEATHER_RANDOM, expectedValues, errorMsgs);
					if (0 <= idxWeatherRandom) {
						model.setMap(false);
						model.setFirstCalendarDay(day, monthID, year, calendarID, weatherID, windID, tempID, (0 == idxWeatherRandom));
					}
				} catch (final DSAException ex) {
					errorMsgs.append(ex.getLocalizedMessage());
					errorMsgs.append("\n");
				} catch (final NumberFormatException ex) {
					if (null != errorMsgs) {
						errorMsgs.append("Expected a number ");
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				} catch (final NullPointerException ex) {
					if (null != errorMsgs) {
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				}
			} else if (action.equalsIgnoreCase(Model.ACTION_DATE_UNDO)) {
				model.setMap(false);
				model.recalcCalendarTimeframe();
			} else if (action.equalsIgnoreCase(Model.ACTION_DATE_NEXT)) {
				model.setMap(false);
				model.calcNextCalendarTimeframe();
			} else if (action.equalsIgnoreCase(Model.ACTION_TOWN_NEW)) {
				try {
					final String[] expectedValues = {Model.TOWN_IS_PREDEF_PREDEF, Model.TOWN_IS_PREDEF_OWN};
					final int idxIsPredef = HtmlHelper.readRadioParameter(request, Model.TOWN_IS_PREDEF, expectedValues, errorMsgs);
					if (0 == idxIsPredef) {
						final int townID = Integer.parseInt(request.getParameter(Model.TOWN_TOWN));
						model.setMap(false);
						model.setTown(townID);
					} else if (1 == idxIsPredef) {
						final String name = request.getParameter(Model.TOWN_NAME);
						final int inhabitants = Integer.parseInt(request.getParameter(Model.TOWN_INHABITANTS));
						final int regionID = Integer.parseInt(request.getParameter(Model.TOWN_REGION));
						model.setMap(false);
						model.setTown(name, inhabitants, regionID);
					}
				} catch (final NumberFormatException ex) {
					if (null != errorMsgs) {
						errorMsgs.append("Expected a number ");
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				} catch (final NullPointerException ex) {
					if (null != errorMsgs) {
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				}
			} else if (action.equalsIgnoreCase(Model.ACTION_TOWN_NEWMAP)) {
				model.setMap(false);
				model.changeTownMap();
			} else if (action.equalsIgnoreCase(Model.ACTION_NAMES_NEW)) {
				try {
					final String[] expectedValues = {Model.NAMES_FOR_LANGUAGE, Model.NAMES_FOR_REGION, Model.NAMES_FOR_TOWN, Model.NAMES_FOR_POSITION};
					final int idxIsPredef = HtmlHelper.readRadioParameter(request, Model.NAMES_FOR, expectedValues, errorMsgs);
					if (0 == idxIsPredef) {
						final int languageID = Integer.parseInt(request.getParameter(Model.NAMES_LANGUAGE));
						model.setMap(false);
						model.setNamesModelForLanguageID(languageID);
					} else if (1 == idxIsPredef) {
						final int regionID = Integer.parseInt(request.getParameter(Model.NAMES_REGION));
						model.setMap(false);
						model.setNamesModelForRegionID(regionID);
					} else if (2 == idxIsPredef) {
						final int townID = Integer.parseInt(request.getParameter(Model.NAMES_TOWN));
						model.setMap(false);
						model.setNamesModelForTownID(townID);
					} else if (3 == idxIsPredef) {
						model.setMap(false);
						model.setNamesModelForPosition();
					}
				} catch (final NumberFormatException ex) {
					if (null != errorMsgs) {
						errorMsgs.append("Expected a number ");
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				} catch (final NullPointerException ex) {
					if (null != errorMsgs) {
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				}
			} else if (action.equalsIgnoreCase(Model.ACTION_PLANT_NEW)) {
				try {
					final int regionID = Integer.parseInt(request.getParameter(Model.NAMES_REGION));
					final int[] terrainIDs = HtmlHelper.readCheckListParams(request, Model.PLANT_TERRAIN, DSA.instance().getTerrain().size());
					final boolean isToTestForTerrainProb = (null != request.getParameter(Model.PLANT_CHECK_IS_TO_TEST_FOR_TERRAIN_PROB));
					final boolean isToFilterForDate      = (null != request.getParameter(Model.PLANT_CHECK_IS_TO_FILTER_FOR_DATE));
					final boolean isToTestForBotany      = (null != request.getParameter(Model.PLANT_CHECK_IS_TO_TEST_FOR_BOTANY));
					final int monthID = Integer.parseInt(request.getParameter(Model.PLANT_MONTH));
					final int testForBotany = Integer.parseInt(request.getParameter(Model.PLANT_TEST_FOR_BOTANY));
					model.setMap(false);
					model.setPlantsModel(regionID, terrainIDs, isToTestForTerrainProb, isToTestForBotany, testForBotany, isToFilterForDate, monthID);
				} catch (final NumberFormatException ex) {
					if (null != errorMsgs) {
						errorMsgs.append("Expected a number ");
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				} catch (final NullPointerException ex) {
					if (null != errorMsgs) {
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				}
			} else if (action.equalsIgnoreCase(Model.ACTION_QUEST_NEW)) {
				model.setMap(false);
				model.changeQuest();
			} else if (action.equalsIgnoreCase(Model.ACTION_DICE_NEW)) {
				try {
					final int factor     = Integer.parseInt(request.getParameter(Model.DICE_FACTOR));
					final int offset     = Integer.parseInt(request.getParameter(Model.DICE_OFFSET));
					final int number     = Integer.parseInt(request.getParameter(Model.DICE_NUMBER));
					final int eyes       = Integer.parseInt(request.getParameter(Model.DICE_EYES));
					model.getDice().getDiceRollType().setFactor(factor);
					model.getDice().getDiceRollType().setOffset(offset);
					model.getDice().getDiceRollType().setNumber(number);
					model.getDice().getDiceRollType().setEyes(eyes);
				} catch (final NumberFormatException ex) {
					if (null != errorMsgs) {
						errorMsgs.append("Expected a number ");
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				} catch (final NullPointerException ex) {
					if (null != errorMsgs) {
						errorMsgs.append(ex.getLocalizedMessage());
						errorMsgs.append("\n");
					}
				}
				model.setMap(false);
				model.rollDice(model.getDice().getDiceRollType());
			} else if (action.equalsIgnoreCase(Model.ACTION_FORMAT_HTML)) {
				model.setFormatStyle(TextDocumentType.HTML5);
			} else if (action.equalsIgnoreCase(Model.ACTION_FORMAT_FIXEDFONT)) {
				model.setFormatStyle(TextDocumentType.FIXEDFONT);
			} else if (action.equalsIgnoreCase(Model.ACTION_MAP_SHOW)) {
				model.setMap(true);
			}
		}
		return errorMsgs.toString();
	}
}
