package de.ped.empire.dev;

import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;

import de.ped.empire.logic.ZoomLevel;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.TileTemplate;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;

/**
 * Helper for building the Empire project. Reads in image templates files and
 * image files for all zoom sizes for squares, and outputs them for hexagon
 * maps.
 * 
 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImgHexagonTemplates extends ImgTemplates {
	public static void main(final String[] args) {
		ImgTemplates task = new ImgHexagonTemplates();
		task.executeMain(args);
	}

	protected ImgHexagonTemplates() {
		props.setTopologyMode(TopologyMode.HEXAGON);
	}

	/** Adds a prefix to the file name and changes the filename extension to png. */
	@Override
	public String getDestFilename(final String srcFilename) {
		return changeFilename(srcFilename, "map", "Hex", "png");
	}

	@Override
	public String getName() {
		return "Updating Empire hexagon icon templates";
	}

	@Override
	protected int getBits() {
		return 6;
	}

	@Override
	protected void fillUpUndefinedTileParts(final BufferedImage destImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant, final int id, final Color color) {
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		if (((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()))) {
			final Point iconPos = getDestIconPos(imageKey, variant, id);
			final Point d = calcDestPosUpperLeftInPixels(zoomLevel, iconPos.x, iconPos.y, true, ORIGIN);
			for (int x = 0; x < playfieldTiling.getSurroundingWidth(); x++) {
				final int yoff = getYOffset(zoomLevel, x);
				for (int y = yoff; y < playfieldTiling.getSurroundingHeight() - yoff; y++) {
					destImg.setRGB(d.x + x,  d.y + y, calcColorCode(destImg, zoomLevel, imageKey, variant, id, x,  y, destImg.getRGB(d.x + x,  d.y + y), color));
				}
			}
			logger.trace("fillUpUndefinedTileParts(" + zoomLevel + "," + imageKey + ") - " + d + ", color=" + (null == color ? null : String.format("%x", color.getRGB())));
		}
	}

	private int calcColorCode(final BufferedImage destImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant, final int id, final int x,
			final int y, final int oldColor, final Color newColor) {
		int ret = oldColor;
		final boolean isOldTransparent = (ImageKey.getBits(oldColor, TileTemplate.BA_ALPHA) == 0);
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		final int normalGray = 0xff808080;
		final int lightGray = 0xffc0c0c0;
		final int darkRed = 0xff800000;
		final int template = imageKey.getBits(ImageKey.BA_TEMPLATE);
		final int type = imageKey.getBits(ImageKey.BA_TTYPE);
		final int surroundingWidth = playfieldTiling.getSurroundingWidth();
		if ((type == ImageKey.TType.Unit.ordinal()) // All moveable units have a dark red tile look
				&& (template != ImageKey.CMDL_ID_AIRBASE)
				&& (template != ImageKey.CMDL_ID_CITY)
				|| ((type == ImageKey.TType.Field.ordinal()) // Non-empty cities also have a dark red tile look
						&& (template == ImageKey.FIELD_ID_CITY)
						&& (id != 0))
				|| ((type == ImageKey.TType.Unit.ordinal()) // Non-empty airports also have a dark red tile look
						&& (template == ImageKey.CMDL_ID_AIRBASE)
						&& (id != 0))) {
			final int sgn;
			final int borderColor;
			if (x < surroundingWidth / 2) { // left half
				sgn = -1;
				if ((type == ImageKey.TType.Field.ordinal()) || (template == ImageKey.CMDL_ID_AIRBASE)) { // City
					borderColor = normalGray;
				} else {
					borderColor = lightGray;
				}
			} else {
				sgn = 1;
				borderColor = normalGray;
			}
			if (isOldTransparent) {
				// Left edge gets filled up with light gray, right edge with normal gray
				if (x < getOffset(zoomLevel).x) {
					ret = borderColor;
				} else if (x >= surroundingWidth - getOffset(zoomLevel).x){
					ret = borderColor;
				}
			} 
			// Icrease the dark red background to the left and right to make that also look more hexagonal
			final int thickness = (zoomLevel.resolution <= 16 ? 1 : 2);
			final int xoffset = (zoomLevel.resolution <= 16 ? 1 : 2);
			final int yoff = Math.max(thickness, getYOffset(zoomLevel, x + sgn * xoffset));
			if ((yoff <= y) && (y < playfieldTiling.getSurroundingHeight() - yoff)) {
				if (ret == borderColor) {
					ret = darkRed;
				} 
			} else {
				if (ret == darkRed) {
					ret = borderColor;
				} 
			}
		} else if (isOldTransparent) { // Only fill the left and right edges
			if (null != newColor) {
				ret = newColor.getRGB();
			} else { // border, outdated overlay
				// Continue the 4x2 px pattern - no matter the resolution
				final Point iconPos = getDestIconPos(imageKey, variant, id);
				final Point d = calcDestPosUpperLeftInPixels(zoomLevel, iconPos.x, iconPos.y, true, ORIGIN);
				final int sgn = (x < surroundingWidth / 2 ? 1 : -1);
				final int xoff = Math.max(8, getOffset(zoomLevel).x);
				ret = destImg.getRGB(d.x + x + sgn * xoff, d.y + y);
			}
		}
		return ret;
	}

	@Override
	protected boolean createFieldBlock(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel, 
			final ImageKey imageKey, final int variant) {
		final int fieldsPerBlock = (1 << getBits());
//		final TileTemplateData ttd = getTTD(imageKey);
		final int[][] copyParamsUpDownSides = calcCopyParamsUpDownSides(zoomLevel);
		final int zoom = zoomLevel.resolution;
		final Point d = getDestIconPos(imageKey, variant);
		boolean ret = false;
		if (ImageKey.UNEXPLORED_NEIGHBOR.equals(imageKey)) {
			for (int id = 0; id < fieldsPerBlock; id++) { // whiten fields
				copyTile(destImg, srcImg, zoomLevel, imageKey, 0, id, destTilesPerBlock*0, destTilesPerBlock*0, 0,0,NONE); // fog
			}
			copyTiles(destImg, srcImg, zoomLevel, imageKey, 0, copyParamsUpDownSides);
			ret = true;
		} else if (ImageKey.SEA.equals(imageKey)) {
			for (int id = 0; id < fieldsPerBlock; id++) { // whiten fields
				copyTile(destImg, srcImg, zoomLevel, imageKey, variant, id, destTilesPerBlock*0, destTilesPerBlock*0, 0,0,NONE); // sea
			}
			copyTile(destImg, srcImg, zoomLevel, imageKey, variant, 0, 0, 0, 0,0,ALL); // sea without coast
			copyTiles(destImg, srcImg, zoomLevel, imageKey, variant, copyParamsUpDownSides);
			ret = true;
		} else if (ImageKey.RIVER.equals(imageKey)) {
			logger.trace("createRivers(" + zoomLevel + "," + variant + ")");
			if (isNeighborMod(zoomLevel)) {
				for (int id = 0; id < fieldsPerBlock; id++) {
					final Point iconPos = getDestIconPos(imageKey, variant, id);
					// Use LAND as background
					copyTile(destImg, srcImg, zoomLevel, iconPos.x, iconPos.y, 18, variant + 8, 0, 0, ALL);
					// Find the middle of the tile as origin / bending / meeting point
					final Point c = calcDestPosUpperLeftInPixels(zoomLevel, 
							iconPos.x, iconPos.y,
							true, new Point(zoom / 2, zoom / 2));
					switch(props.getTopologyMode()) {
					case HEXAGON:
						if (PlayfieldTiling.IS_HEX_COMPACT) {
							// Nothing to do
						} else {
							c.translate(0, -zoom / 2);
						}
						final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
						if (((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()))) {
							c.translate(zoom / 8, 0);
						}
						break;
					default:
						// Nothing to do
					}
					int numOfRivers = 0;
					for (int bit = 1; bit < fieldsPerBlock; bit = bit << 1) {
						if (0 != (id & bit)) {
							numOfRivers++;
						}
					}
					final double[] randomWalkProbsPerNumOfRivers = { 0.00, 0.25, 0.00, 0.10, 0.20, 0.30, 0.40, };  
					// index 0 = sinus wave, index 1 = random walk
					final double[] probs = { 1.0, randomWalkProbsPerNumOfRivers[numOfRivers] };
					for (int bit = 1; bit < fieldsPerBlock; bit = bit << 1) {
						if (0 != (id & bit)) { 
							int bend = 0;
							final int maxwidth = zoom / 8;
							final int algo = MathUtil.randomDistributed(probs);
							final int bends = MathUtil.random(4);
							for (int r = zoom / 2 - 1; r >=0; r--) { // Draw inwards, so borders fit, even if inner is chaotic
								if ((HEX_N == id) || (HEX_S == id) || ((HEX_N | HEX_S) == id)) {
									// Nothing to do, we copy those from source
								} else {
									if (0 == algo) { // sinus wave
										bend = (int) Math.round(maxwidth * 
												Math.sin(bends * Math.PI * r / (zoom / 2.0)) // Bending frequency
												* Math.sin(Math.PI * r / (zoom / 2.0))); // Hull
									} else { // random walk
//					    				int[] bendDist = new int[]{ 1, 4, 6, 9, 6, 3, 1};
										bend += MathUtil.random(3) - 1;
//					    				switch(zoom) {
//					    				case 8:
//						    				bendDist = new int[]{ 1, 6, 1};
//					    					break;
//					    				case 16:
//						    				bendDist = new int[]{ 1, 4, 1};
//					    					break;
//					    				case 24:
//						    				bendDist = new int[]{ 1, 4, 6, 3, 1};
//					    					break;
//					    				default:
//					    					// Nothing to do
//					    				}
//					    				int bendNew = MathUtil.randomDistributed(bendDist) - bendDist.length / 2;
									}
									for (int width = 0; width < maxwidth; width++) {
										if (bit == HEX_N) { // North
											try {
												destImg.setRGB(c.x + bend - width, c.y - r, calcRiverRGB());
											} catch (ArrayIndexOutOfBoundsException ex) {
												logger.error("destImg:" + destImg.getWidth() + "x" + destImg.getHeight());
												logger.fatal("zoom=" + zoom + ", id=" + id + ", c=" + c + ", r=" + r + ", bend=" + bend + ", width=" + width, ex);
											}
										}
										if (bit == HEX_NW) { // North-west
											destImg.setRGB(c.x - r, c.y + bend - r / 2 - width, calcRiverRGB());
										}
										if (bit == HEX_SW) { // South-west
											destImg.setRGB(c.x - r, c.y + bend + r / 2 - width, calcRiverRGB());
										}
										if (bit == HEX_S) { // South
											destImg.setRGB(c.x + bend - width, c.y + r, calcRiverRGB());
										}
										if (bit == HEX_SE) { // South-east
											destImg.setRGB(c.x + r, c.y + bend + r / 2 - width, calcRiverRGB());
										}
										if (bit == HEX_NE) { // North-east
											destImg.setRGB(c.x + r, c.y + bend - r / 2 - width, calcRiverRGB());
										}
									}
								}
							}
						}
					}
				}
			} else { // Just copy the river block
				copyBlock(destImg, srcImg, zoomLevel, imageKey, variant);
			}
			copyTiles(destImg, srcImg, zoomLevel, imageKey, variant, copyParamsUpDown);
			// River without neighbor is lake
			final Rectangle s = getSrcIconRect(imageKey, variant);
			copyTile(destImg, srcImg, zoomLevel, d.x, d.y, s.x, s.y, 0, 0, ALL);
			ret = true;
		} else if (ImageKey.FOREST.equals(imageKey)) {
			final Color[] colors = new Color[] { 
					new Color(0xff008000), // Dark green
					Color.green, new Color(0xff808000), // Olive
					new Color(0xff808080), // Dark gray
					new Color(0xffc0c0c0), // Light gray
					Color.red, Color.black, };
			treatNeighborLandMods(destImg, srcImg, zoomLevel, imageKey, variant, fieldsPerBlock, colors);
			ret = true;
		} else if (ImageKey.ROUGH.equals(imageKey)) {
			final Color[] colors = new Color[] { 
					new Color(0xff008000), // Dark green
					Color.green, new Color(0xff808000), // Olive
					new Color(0xff808080), // Dark gray
					new Color(0xffc0c0c0), // Light gray
					Color.red, Color.black, };
			treatNeighborLandMods(destImg, srcImg, zoomLevel, imageKey, variant, fieldsPerBlock, colors);
			ret = true;
		} else if (ImageKey.MOUNTAIN.equals(imageKey)) {
			final Color[] colors = new Color[grayColors.length + 2];
			for (int i = 0; i < grayColors.length; i++) {
				colors[i] = grayColors[i];
			}
			colors[grayColors.length + 0] = new Color(0xff008000); // Dark green
			colors[grayColors.length + 1] = new Color(0xff808000); // Olive;
			treatNeighborLandMods(destImg, srcImg, zoomLevel, imageKey, variant, fieldsPerBlock, colors);
			ret = true;
		}
		return ret;
	}

	protected void treatNeighborLandMods(final BufferedImage destImg, final BufferedImage srcImg,
			final ZoomLevel zoomLevel, final ImageKey imageKey, final int variant, final int fieldsPerBlock,
			final Color[] colors) {
		copyBlock(destImg, srcImg, zoomLevel, imageKey, variant);
		if (isNeighborMod(zoomLevel)) {
			for (int id = 0; id < fieldsPerBlock; id++) {
				if ((HEX_N == id) || (HEX_S == id) || ((HEX_N | HEX_S) == id)) {
					// Nothing to do, we copy those from source
				} else {
					final int[] borders = new int[] { HEX_N, HEX_NW, HEX_SW, HEX_S, HEX_SE, HEX_NE, };
					for (int border : borders) {
						if (0 != (id & border)) {
							recolorSingleBorder(destImg, zoomLevel, imageKey, variant, id, border, colors);
						}
					}
				}
			}
		}
	}

	@Override
	protected void copyBlock(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant) {
		final Point d = getDestIconPos(imageKey, variant);
		final Rectangle s = getSrcIconRect(imageKey, variant);
		// Fill all with empty fields
		for (int x = 0; x < destTilesPerBlock; x++) {
			for (int y = 0; y < destTilesPerBlock; y++) {
				copyTile(destImg, srcImg, zoomLevel, d.x + x, d.y + y,
						s.x, s.y, 0, 0, ALL);
			}
		}
		// Fill the initial block
		copyTiles(destImg, srcImg, zoomLevel, imageKey, variant, copyParamsUpDown);
	}

	@Override
	protected int[][] calcCopyParamsUpDownSides(final ZoomLevel zoomLevel) {
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		if (((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()))) {
			return copyParamsUpDown;
		} else {
			return appendCopyParams(copyParamsUpDown, copyParamsSides);
		}
	}

	private final int[][] copyParamsUpDown = new int[][] { // dest pos, source pos
		{ 1, 0, 1, 0, ALL, },
		{ 0, 1, 2, 0, ALL, },
		{ 1, 1, 3, 0, ALL, },
	};

	final int[][] copyParamsSides = new int[][] {
		{ 2, 0, 0, 1, UPPER, },
		{ 4, 0, 0, 1, LOWER, },
		{ 0, 4, 0, 2, UPPER, },
		{ 0, 2, 0, 2, LOWER, },
	};
}
