package de.ped.empire.dev;

import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;

import javax.imageio.ImageIO;

import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.ZoomLevel;
import de.ped.tools.TopologyMode;

/**
 * Helper for building the Empire project. Reads in an image templates file for
 * a given zoom size, the according screenshot of a template map from the
 * original Empire, and stores a copy of the image templates file where the
 * tiles from the empire map was copied into.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImgSquareTemplates extends ImgTemplates {

	public static void main(final String[] args) {
		final ImgTemplates task = new ImgSquareTemplates();
		task.executeMain(args);
	}

	protected ImgSquareTemplates() {
		props.setTopologyMode(TopologyMode.SQUARE8);
	}

	/** Adds a prefix to the file name and changes the filename extension to png. */
	@Override
	public String getDestFilename(final String srcFilename) {
		return changeFilename(srcFilename, "map", null, "png");
	}

	@Override
	public String getName() {
		return "Updating Empire square icon templates";
	}

	@Override
	public void execute(final String srcFilename, final String destFilename) {
		try {
			final int zoom = findZoom(srcFilename).zoom;
			final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[ImgTemplates.findZoomLevelIndexForResolution(zoom)];
			final BufferedInputStream is = new BufferedInputStream(new FileInputStream(srcFilename));
			final BufferedImage srcImg = ImageIO.read(is);
			is.close();
			final Dimension sizeInPixel = props.getMapSizeInPixels(zoomLevel, true);
			final BufferedImage destImg = new BufferedImage(sizeInPixel.width, sizeInPixel.height, srcImg.getType());
			// Copy the whole source image
			destImg.setData(srcImg.getData());
//        	copyTilesFromTemplate(srcFilename.replace("icons", "Template"), destImg);
			increaseToHexagonSize(destImg, zoomLevel);
			
			mergeNeighborModifications(destImg, zoomLevel, ImageKey.UNEXPLORED_NEIGHBOR, 0);
			for (int variant = 0; variant < 2; variant++) {
				mergeNeighborModifications(destImg, zoomLevel, ImageKey.SEA, variant);
			}

			final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(destFilename));
			ImageIO.write(destImg, "PNG", os);
			os.close();
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	protected void increaseToHexagonSize(final BufferedImage destImg, final ZoomLevel zoomLevel) {
		convertBlock(destImg, destImg, zoomLevel, 8*6+1, 8*0,  4, 8, 12, 4); // cities and armies
		for (int i = 0; i < 4; i++) {
			convertBlock(destImg, destImg, zoomLevel, 8*6+i, 5, 16+0, 8+i, 1, 1); // fight overlay
		}
		convertBlock(destImg, destImg, zoomLevel, 8*6+0, 8*0, 16+2, 8+0, 1, 1); // land1
		convertBlock(destImg, destImg, zoomLevel, 8*6+0, 8*1, 16+2, 8+1, 1, 1); // land2
		convertBlock(destImg, destImg, zoomLevel, 8*6+0,   1, 16+1, 8+0, 1, 4); // border, unexplored, crosshair, outdated overlay
		convertBlock(destImg, destImg, zoomLevel, 8*5  , 8*1, 16, 4,  4, 4); // mountain2 
		convertBlock(destImg, destImg, zoomLevel, 8*5  , 8*0, 16, 0,  4, 4); // mountain1 
		convertBlock(destImg, destImg, zoomLevel, 8*4  , 8*1, 12, 4,  4, 4); // rough2 
		convertBlock(destImg, destImg, zoomLevel, 8*4  , 8*0, 12, 0,  4, 4); // rough1 
		convertBlock(destImg, destImg, zoomLevel, 8*3  , 8*1,  8, 4,  4, 4); // forest 
		convertBlock(destImg, destImg, zoomLevel, 8*3  , 8*0,  8, 0,  4, 4); // forest 
		convertBlock(destImg, destImg, zoomLevel, 8*2  , 8*1,  4, 4,  4, 4); // river
		convertBlock(destImg, destImg, zoomLevel, 8*2  , 8*0,  4, 0,  4, 4); // river
		convertBlock(destImg, destImg, zoomLevel, 8*1  , 8*1,  0, 4,  4, 4); // sea
		convertBlock(destImg, destImg, zoomLevel, 8*1  , 8*0,  0, 0,  4, 4); // sea
		convertBlock(destImg, destImg, zoomLevel, 8*0  , 8*0,  0, 8,  4, 4); // fog of war 
		convertBlock(destImg, destImg, zoomLevel, 8*6+1, 8*1, 8*6+ 1, 2, 1, 2); // cities variant1
		convertBlock(destImg, destImg, zoomLevel, 8*6+2,   2, 8*6+15, 2, 1, 2); // just make a white patch transparent for beauty
	}

	@SuppressWarnings("unused")
	private void copyTilesFromTemplate(final String templateFilename, final BufferedImage destImg) 
			throws FileNotFoundException, IOException {
		final BufferedInputStream tis = new BufferedInputStream(new FileInputStream(templateFilename));
		final BufferedImage ti = ImageIO.read(tis);
		final ZoomDataValue zoomData = findZoom(templateFilename);
		final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[ImgTemplates.findZoomLevelIndexForResolution(zoomData.zoom)];
		// Copy tiles
		final ArrayList<int[]> list = new ArrayList<int[]>();
		list.add(new int[] { 0, 0 + 0,  1 + 0, 1, }); // Sea 0
		list.add(new int[] { 0, 0 + 4,  1 + 4, 1, }); // Sea 1
		list.add(new int[] { 3, 8 + 0,  2, 4, }); // Land 0
		list.add(new int[] { 3, 8 + 1,  2, 5, }); // Land 1
		addRivers(list, new int[][] {
				{ 4 + 0, 0, 25,  4, }, // River 0 0000
				{ 4 + 0, 1, 23,  4, }, // River 0 0001
				{ 4 + 0, 2, 27,  4, }, // River 0 0010
				{ 4 + 1, 0, 25,  2, }, // River 0 0100
				{ 4 + 2, 0, 25,  6, }, // River 0 1000
				{ 4 + 0, 3, 24, 13, }, // River 0 0011
				{ 4 + 1, 1, 26, 11, }, // River 0 0101
				{ 4 + 1, 2, 27, 14, }, // River 0 0110
				{ 4 + 1, 3, 25, 15, }, // River 0 0111
				{ 4 + 2, 1, 23, 12, }, // River 0 1001
				{ 4 + 2, 2, 24, 15, }, // River 0 1010
				{ 4 + 2, 3, 25, 11, }, // River 0 1011
				{ 4 + 3, 0, 25, 12, }, // River 0 1100
				{ 4 + 3, 1, 27, 13, }, // River 0 1101
				{ 4 + 3, 2, 23, 13, }, // River 0 1110
				{ 4 + 3, 3, 25, 13, }, // River 0 1111
		});
		addTerrains(list, new int[][] {
				{  8, 0, 10,  2, }, // Forest
				{ 12, 0,  6,  2, }, // Rough
				{ 16, 0, 14,  2, }, // Mountain
		});
		for (final int[] tile : list) {
		    copyTile(destImg, ti, zoomLevel, tile[0], tile[1], tile[2], tile[3], zoomData.sOffsetX, zoomData.sOffsetY, ALL);
		}
	}

	private void addRivers(final ArrayList<int[]> list, final int[][] parts) {
		for (final int[] part : parts) {
			list.add(new int [] {part[0], part[1] + 0, part[2] + 0, part[3], });
			list.add(new int [] {part[0], part[1] + 4, part[2] + 8, part[3], });
		}
	}
	
	private void addTerrains(final ArrayList<int[]> list, final int[][] terrainOffsets) {
		final int[][] parts = new int[][] {
				// 1x1 isles
				{ 0, 0 + 0, 0, 16, }, // 0000 0
				{ 0, 0 + 4, 2, 16, }, // 0000 1
				// Horizontal isles
				{ 0, 1 + 0, 2,  0, }, // 0001 0
				{ 0, 1 + 4, 2,  6, }, // 0001 1
				{ 0, 2 + 0, 0,  0, }, // 0010 0
				{ 0, 2 + 4, 0,  6, }, // 0010 1
				{ 0, 3 + 0, 1,  0, }, // 0011 0
				{ 0, 3 + 4, 1,  6, }, // 0011 1
				// Vertical isles
				{ 1, 0 + 0, 0,  4, }, // 0100 0
				{ 1, 0 + 4, 2,  4, }, // 0100 1
				{ 2, 0 + 0, 0,  2, }, // 1000 0
				{ 2, 0 + 4, 2,  2, }, // 1000 1
				{ 3, 0 + 0, 0,  3, }, // 1100 0
				{ 3, 0 + 4, 2,  3, }, // 1100 1
				// Huge isles
				{ 1, 1 + 0, 2,  8 + 2, }, // 0101 0
				{ 1, 1 + 4, 2, 12 + 2, }, // 0101 1
				{ 1, 2 + 0, 0,  8 + 2, }, // 0110 0
				{ 1, 2 + 4, 0, 12 + 2, }, // 0110 1
				{ 1, 3 + 0, 1,  8 + 2, }, // 0111 0
				{ 1, 3 + 4, 1, 12 + 2, }, // 0111 1
				{ 2, 1 + 0, 2,  8 + 0, }, // 1001 0
				{ 2, 1 + 4, 2, 12 + 0, }, // 1001 1
				{ 2, 2 + 0, 0,  8 + 0, }, // 1010 0
				{ 2, 2 + 4, 0, 12 + 0, }, // 1010 1
				{ 2, 3 + 0, 1,  8 + 0, }, // 1011 0
				{ 2, 3 + 4, 1, 12 + 0, }, // 1011 1
				{ 3, 1 + 0, 2,  8 + 1, }, // 1101 0
				{ 3, 1 + 4, 2, 12 + 1, }, // 1101 1
				{ 3, 2 + 0, 0,  8 + 1, }, // 1110 0
				{ 3, 2 + 4, 0, 12 + 1, }, // 1110 1
				{ 3, 3 + 0, 1,  8 + 1, }, // 1111 0
				{ 3, 3 + 4, 1, 12 + 1, }, // 1111 1
		};
		for (final int[] terrain : terrainOffsets) {
			for (final int[] part : parts) {
				list.add(new int [] {part[0] + terrain[0], part[1] + terrain[1], part[2] + terrain[2], part[3] + terrain[3], });
			}
		}
	}

	@Override
	protected int getBits() {
		return 4;
	}

	@Override
	protected void copyBlock(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant) {
		// Nothing to do
	}

	@Override
	protected int[][] calcCopyParamsUpDownSides(final ZoomLevel zoomLevel) {
		return null;
	}
}
