package de.ped.empire.dev;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import javax.imageio.ImageIO;

import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.ImageSet;
import de.ped.empire.logic.TileTemplate;
import de.ped.empire.logic.TileTemplate.TileTemplateData;
import de.ped.empire.logic.ZoomLevel;
import de.ped.tools.BorderMode;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Logger;

/**
 * Helper for building the Empire project. Parent class for creating map icons files.
 * 
 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class ImgTemplates {

	protected static final int NONE  = 0x0;
	protected static final int UPPER = 0x1;
	protected static final int LOWER = 0x2;
	protected static final int ALL   = 0x3;

	/**
	 * Compares two RGB values and decides which of those values to choose.
	 * 
	 * @author Copyright (c) 2016 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	protected interface ColorComparator {
		public boolean isToCopyDest(int sourceRGB, int destRGB);
	}

	/**
	 * Compares in a way that the brighter pixel wins.
	 * 
	 * @author Copyright (c) 2016 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	protected class BrighterComparator implements ColorComparator {
		@Override
		public boolean isToCopyDest(int sourceRGB, int destRGB) {
			return ((0xffffff != (sourceRGB & 0xffffff)) // Source is not white
					&& ((0xffffff == (destRGB & 0xffffff)) // AND Destination is white
							|| (getBrightness(sourceRGB) < getBrightness(destRGB))));
		}
	}

	/**
	 * Compares in a way that the greener pixel wins.
	 * 
	 * @author Copyright (c) 2016 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	protected class GreenerComparator implements ColorComparator {
		@Override
		public boolean isToCopyDest(int sourceRGB, int destRGB) {
			return ((0xffffff != (sourceRGB & 0xffffff)) // Source is not white
					&& ((0xffffff == (destRGB & 0xffffff)) // AND Destination is white
							|| (((sourceRGB >>  8) & 0xff) > ((destRGB >>  8) & 0xff)))); // OR Source is greener than the other side 
		}
	}

	/**
	 * Compares in a way that the destination pixel is taken iff the source pixel is white.
	 * 
	 * @author Copyright (c) 2016 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	protected class WhiteScreenComparator implements ColorComparator {
		@Override
		public boolean isToCopyDest(int sourceRGB, int destRGB) {
			return (0xffffff == (destRGB & 0xffffff));
		}
	}

	/** 
	 * @param resolution the resolution (field size) to search for
	 * @return the index of the zoom level containing that resolution, or -1 if none found
	 */
	protected static int findZoomLevelIndexForResolution(final int resolution) {
		int ret = -1;
		for (int i = 0; i < ZoomLevel.INSTANCES.length; i++) {
			if (resolution == ZoomLevel.INSTANCES[i].resolution) {
				ret = i;
				break;
			}
		}
		return ret;
	}

	private static final HashMap<String, ZoomDataValue> ZOOM_DATA = new HashMap<String, ZoomDataValue>();
	static {
		ZOOM_DATA.put("32", new ZoomDataValue(32,  -1,  -1)); // No old game map exists for that zoom value
		ZOOM_DATA.put("24", new ZoomDataValue(24,  12,  90));
		ZOOM_DATA.put("16", new ZoomDataValue(16,   8,  86));
		ZOOM_DATA.put("08", new ZoomDataValue( 8, 176, 230));
		ZOOM_DATA.put("04", new ZoomDataValue( 4,  -1,  -1)); // No old game map exists for that zoom value
		ZOOM_DATA.put("02", new ZoomDataValue( 2,  -1,  -1)); // No old game map exists for that zoom value
	}

	protected static ZoomDataValue findZoom(final String filename) {
		for (final String key : ZOOM_DATA.keySet()) {
			if (filename.contains(key)) {
				return ZOOM_DATA.get(key);
			}
		}
		throw new IllegalArgumentException("Unknown zoom format!");
	}

	protected ImgTemplates() {
		props.setBorderMode(BorderMode.NONE);
		props.setSize(PlayfieldPosition.X, 8 * 8);
		props.setSize(PlayfieldPosition.Y, 8 * 2);
	}

	/**
	 * Parses the args as if coming from the command line, and executes this
	 * task accordingly.
	 * 
	 * Is based on the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @param args
	 *            the args as coming from command line
	 */
	public void executeMain(final String[] args) {
		String fromdir = "src/de/ped/empire/dev";
		setTodir(fromdir);
		String mode = "";
		final ArrayList<String> files = new ArrayList<String>();
		for (int i = 0; i < args.length; i++) {
			final String arg = args[i];
			if ("-d".equals(mode)) {
				setTodir(arg);
				mode = "";
			} else if ("-s".equals(mode)) {
				fromdir = arg;
				mode = "";
			} else {
				files.add(arg);
			}
		}
		execute(fromdir, files);
	}

	/**
	 * Executes a fileset given by the parameters. For usage from non-ant
	 * environments.
	 * 
	 * Is currently identical to the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @param basedir
	 *            the base directory of all source files
	 * @param files
	 *            the source file names relative to the base directory
	 */
	public void execute(final String basedir, final ArrayList<String> files) {
		String[] filesArr = new String[files.size()];
		filesArr = files.toArray(filesArr);
		execute(basedir, filesArr);
	}

	/**
	 * Executes a fileset given by the parameters. For usage from non-ant
	 * environments.
	 * 
	 * Is based on the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @param basedir
	 *            the base directory of all source files
	 * @param files
	 *            the source file names relative to the base directory
	 */
	public void execute(final String basedir, String[] files) {
		if ((null == files) || (0 == files.length)) {
			files = defaultFiles();
		}
		logger.debug("todir = " + todir);
		executeIfDifferent(new File(basedir), files, null);
	}

	/**
	 * Iterates through the files, compares each of them to the destination
	 * file, and if different, executes the file action.
	 * 
	 * Is based on the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @param srcBase
	 *            the root of all source file names
	 * @param srcPathsFromBase
	 *            the source file names, relative to srcBase
	 * @param comparismModes
	 *            the comparism mode that defines if there is something to do
	 */
	public void executeIfDifferent(final File srcBase, final String[] srcPathsFromBase, final int[] comparismModes) {
		for (int i = 0; i < srcPathsFromBase.length; i++) {
			final File srcPath = new File(srcBase, srcPathsFromBase[i]);
			final String destPath = todir + File.separator + getDestFilename(srcPathsFromBase[i]);
			logger.info(getName() + " " + srcPath.getPath() + " to " + destPath);
			execute(srcPath.getPath(), destPath);
		}
	}

	/**
	 * If no files to handle are given on the command line, use the following
	 * list as default.
	 * 
	 * Is currently identical to the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @return the default implementation returns an empty list, but subclasses
	 *         may overload this behavior
	 */
	public String[] defaultFiles() {
		return new String[]{
			"icons32.png",
			"icons24.png",
			"icons16.png",
			"icons08.png",
			"icons04.png",
			"icons02.png",
		};
	}

	/**
	 * Calculates the destination filename based on the source filename
	 * (relative to the file base). The default implementation returns just the
	 * source filename. Override if you need something else.
	 * 
	 * Is currently identical to the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @param srcFilename
	 *            the source filename, relative to the root of the fileset
	 * @return the same as srcFilename (as default implementation)
	 */
	public String getDestFilename(final String srcFilename) {
		return srcFilename;
	}

	/**
	 * A helper method for converting file names.
	 * 
	 * Is currently identical to the implentation in <code>AntTasks.FileAction</code>.
	 * 
	 * @param filename
	 *            the original file name to start with. Should have an extension
	 * @param prefix
	 *            a prefix to add to the original filename
	 * @param suffix
	 *            a suffix to add to the original filename
	 * @param extension
	 *            a new extension for the new filename. If <code>null</code>,
	 *            the extension from filename is kept.
	 * @return the converted file name
	 */
	public static String changeFilename(
			final String filename, 
			final String prefix, 
			final String suffix, 
			final String extension) {
		final int idx = filename.lastIndexOf('.');
		final String filenameWithoutExt = filename.substring(0, idx);
		final String changedFilename = (null == prefix ? "" : prefix) + 
			filenameWithoutExt + 
			(null == suffix ? "" : suffix) + "." + 
			(null == extension ? filename.substring(idx + 1) : extension);
		return changedFilename;
	}

	/** 
	 * @return a human readable description of the action
	 */
	public abstract String getName();

	/**
	 * @param toDir
	 *            the root directory of where to write the destination files to.
	 *            Default value is the current directory.
	 */
	public void setTodir(final String toDir) {
		this.todir = toDir;
	}

	/**
	 * Data model for tile types.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ImageKeyData {
		public final ImageKey imageKey;
		public final Rectangle srcTileUL; 
		public final Point srcLookVector;
		public final Color colorForId0;
		public final Color colorForIdNon0;
		public final Color[] colorsForNeighborMods;
		private ImageKeyData(final ImageKey imageKey, final Rectangle srcTileUL, final Point srcLookVector,
				final Color colorForId0, final Color colorForIdNon0, final Color[] colorsForNeighborMods) {
			this.srcTileUL = srcTileUL;
			this.srcLookVector = srcLookVector;
			this.imageKey = imageKey;
			this.colorForId0 = colorForId0;
			this.colorForIdNon0 = colorForIdNon0;
			this.colorsForNeighborMods = colorsForNeighborMods;
		}
		private ImgTemplates getOuterType() {
			return ImgTemplates.this;
		}
		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + getOuterType().hashCode();
			result = prime * result + ((imageKey == null) ? 0 : imageKey.hashCode());
			return result;
		}
		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			ImageKeyData other = (ImageKeyData) obj;
			if (!getOuterType().equals(other.getOuterType())) {
				return false;
			}
			if (imageKey == null) {
				if (other.imageKey != null) {
					return false;
				}
			} else if (!imageKey.equals(other.imageKey)) {
				return false;
			}
			return true;
		}
		@Override
		public String toString() {
			return "ImageKeyData [imageKey=" + imageKey + "]";
		}
	}

	protected final Color landColor = Color.green.darker().darker();

	/** Shades of gray */
	protected final Color[] grayColors = new Color[] { 
			Color.black, new Color(0xff404040), new Color(0xff808080), new Color(0xffa0a0a0), Color.white, 
	}; 

	protected final HashMap<ImageKey, ImageKeyData> imageKeyDataColl = new HashMap<ImageKey, ImageKeyData>();
	{
		imageKeyDataColl.put(ImageKey.UNEXPLORED_NEIGHBOR, 
				new ImageKeyData(ImageKey.UNEXPLORED_NEIGHBOR, 
						getSrcBlockRectangle(0, 2), SRC_TRANS_VECTOR_LOOK, 
						Color.white, Color.white,
						grayColors // From black to white (=transparent) some gray tones
				));
		imageKeyDataColl.put(ImageKey.SEA,
				new ImageKeyData(ImageKey.SEA, 
						getSrcBlockRectangle(0, 0), SRC_TRANS_VECTOR_LOOK, 
						new Color(0xff000080), Color.white,
						new Color[] { new Color(0xff008000), new Color(0xff003060), Color.blue, new Color(0xff000080), Color.white, } // Green, Olive, blue, dark blue, white (for transparent)
				));
		imageKeyDataColl.put(ImageKey.RIVER, 
				new ImageKeyData(ImageKey.RIVER, 
						getSrcBlockRectangle(1, 0), SRC_TRANS_VECTOR_LOOK, 
						landColor, landColor, null
				));
		imageKeyDataColl.put(ImageKey.FOREST, 
				new ImageKeyData(ImageKey.FOREST, 
						getSrcBlockRectangle(2, 0), SRC_TRANS_VECTOR_LOOK, 
						landColor, landColor, null
				));
		imageKeyDataColl.put(ImageKey.ROUGH, 
				new ImageKeyData(ImageKey.ROUGH, 
						getSrcBlockRectangle(3, 0), SRC_TRANS_VECTOR_LOOK, 
						landColor, landColor, null
				));
		imageKeyDataColl.put(ImageKey.MOUNTAIN, 
				new ImageKeyData(ImageKey.MOUNTAIN, 
						getSrcBlockRectangle(4, 0), SRC_TRANS_VECTOR_LOOK, 
						landColor, landColor, null
				));
//    	imageKeyDataColl.put(ImageKey.LAND, 
//    			new ImageKeyData(ImageKey.LAND, 
//    					new Rectangle(18, 8, 1, 1), new Point(0, 1), 
//    					landColor, landColor, null
//    			));
	}

	protected Rectangle getSrcBlockRectangle(final int x, final int y) {
		return new Rectangle(x * srcTilesPerBlock, y * srcTilesPerBlock, srcTilesPerBlock, srcTilesPerBlock);
	}

	/**
	 * Executes the file action to the destination file, and uses the source
	 * file as source.
	 * 
	 * @param srcFilename
	 *            the complete source file name and path
	 * @param destFilename
	 *            the complete destination file name and path
	 */
	public void execute(String srcFilename, String destFilename) {
		try {
			final int zoom = findZoom(srcFilename).zoom;
			final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[ImgTemplates.findZoomLevelIndexForResolution(zoom)];
			// Init
			final BufferedInputStream is = new BufferedInputStream(new FileInputStream(srcFilename));
			final BufferedImage srcImg = ImageIO.read(is);
			is.close();
			final Dimension sizeInPixel = props.getMapSizeInPixels(zoomLevel, true);
			logger.trace("SizeInPixel=" + sizeInPixel);
			final BufferedImage destImg = new BufferedImage(sizeInPixel.width, sizeInPixel.height, srcImg.getType());

			final ImageKey[] landTypes = { ImageKey.UNEXPLORED_NEIGHBOR, ImageKey.SEA, ImageKey.RIVER, ImageKey.FOREST, ImageKey.ROUGH, ImageKey.MOUNTAIN, };
			for (int block = 0; block < landTypes.length; block++) {
				final ImageKey imageKey = landTypes[block];
				final ImageKeyData imageKeyData = imageKeyDataColl.get(imageKey);
				final TileTemplateData tileTemplateData = getTTD(imageKey);
				final int variants = (tileTemplateData.iconModifier.equals(TileTemplate.TRANS_VECTOR_NONE) ? 1 : 2);
				for (int variant = 0; variant < variants; variant++) {
					final boolean isDone = createFieldBlock(destImg, srcImg, zoomLevel, imageKey, variant);
					if (!isDone) {
						copyBlock(destImg, srcImg, zoomLevel, imageKey, variant);
					}
					for (int id = 0; id < (1 << getBits()); id++) {
						final Color color = (id == 0 ? imageKeyData.colorForId0 : imageKeyData.colorForIdNon0);
						if (null != color) {
							fillUpUndefinedTileParts(destImg, zoomLevel, imageKey, variant, id, color);
						}
					}
					final Color[] colList = imageKeyData.colorsForNeighborMods;
					if (null != colList) {
						createSingleNeighborModifications(destImg, srcImg, zoomLevel, imageKey, variant, colList);
						mergeNeighborModifications(destImg, zoomLevel, imageKey, variant);
					}
				}
			}

			// Transform all the rest
			for (int variant = 0; variant < 2; variant++) {
				final Point d = getDestIconPos(ImageKey.LAND, variant);
				// land without terrain type
				copyTile(destImg, srcImg, zoomLevel, d.x, d.y, 18, 8 + variant, 0, 0, ALL);
				fillUpUndefinedTileParts(destImg, zoomLevel, ImageKey.LAND, variant, 0, landColor);
			}
			final ImageKey[] data = { ImageKey.BORDER, ImageKey.CROSSHAIR, ImageKey.UNEXPLORED, ImageKey.OUTDATED };
			for (int i = 0; i < data.length; i++) {
				final Point d = getDestIconPos(data[i], 0);
				copyTile(destImg, srcImg, zoomLevel, d.x, d.y, 17, 8 + i, 0, 0, ALL);
			}
			fillUpUndefinedTileParts(destImg, zoomLevel, ImageKey.BORDER, 0, 0, null);
			fillUpUndefinedTileParts(destImg, zoomLevel, ImageKey.UNEXPLORED, 0, 0, Color.black);
			fillUpUndefinedTileParts(destImg, zoomLevel, ImageKey.OUTDATED, 0, 0, null);

			final ImageKey[] units = new ImageKey[] { ImageKey.CITY, ImageKey.ARMY, ImageKey.TANK, ImageKey.FIGHTER,
					ImageKey.BOMBER, ImageKey.TRANSPORTER, ImageKey.DESTROYER, ImageKey.SUBMARINE, ImageKey.CRUISER,
					ImageKey.BATTLESHIP, ImageKey.CARRIER, ImageKey.AIRBASE, };
			final Color unitColor = Color.gray;
			for (int x = 0; x < units.length; x++) { // military units
				final ImageKey imageKey = units[x];
				final TileTemplateData tileTemplateData = getTTD(imageKey);
				final int variants = (tileTemplateData.iconModifier.equals(TileTemplate.TRANS_VECTOR_NONE) ? 1 : 2);
				for (int variant = 0; variant < variants; variant++) {
					final Point p = getDestIconPos(imageKey, variant);
					final int bitNonEmpty = ImageKey.getBits(tileTemplateData.normalizationMask, ImageKey.BA_NONEMPTY);
					// empty and loaded
					for (int i = 0; i < 1 + bitNonEmpty; i++) { 
						copyTile(destImg, srcImg, zoomLevel, p.x, p.y + i, 4 + x, 8 + (2 * variant) + i, 0, 0, ALL);
						fillUpUndefinedTileParts(destImg, zoomLevel, imageKey, variant, destTilesPerBlock * i,
								unitColor);
					}
				}
			}
			final TileTemplateData fight = getTTD(ImageKey.FIGHT);
			for (int i = 0; i < 4; i++) { // fight overlay
				copyTile(destImg, srcImg, zoomLevel, fight.iconX + i, fight.iconY, 4 + 12, 8 + i, 0, 0, ALL);
			}
			// Finish
			final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(destFilename));
			ImageIO.write(destImg, "PNG", os);
			os.close();
			System.out.println("done.");
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	protected static final int HEX_N = 0x01;
	protected static final int HEX_NW = 0x02;
	protected static final int HEX_SW = 0x04;
	protected static final int HEX_S = 0x08;
	protected static final int HEX_SE = 0x10;
	protected static final int HEX_NE = 0x20;

	protected void createSingleNeighborModifications(final BufferedImage destImg, final BufferedImage srcImg,
			final ZoomLevel zoomLevel, final ImageKey imageKey, final int variant, final Color[] colors) {
		for (int id : new int[] { HEX_NW, HEX_SW, HEX_SE, HEX_NE, }) { // All directions added per hex support
			recolorSingleBorder(destImg, zoomLevel, imageKey, variant, id, id, colors);
		}
	}

	protected void recolorSingleBorder(final BufferedImage destImg, final ZoomLevel zoomLevel, final ImageKey imageKey,
			final int variant, final int id, final int borderId, final Color[] colors) {
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		if (((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()))) {
			final Point iconPos = getDestIconPos(imageKey, variant, id);
			final Point c = calcDestPosUpperLeftInPixels(zoomLevel,  iconPos.x, iconPos.y, true, ORIGIN);
			final int tileWidth = playfieldTiling.getSurroundingWidth();
			final int[][] distribs = getColorDistribution(zoomLevel, imageKey);
			int ysgn = 1;
			int ytrans = 0;
			int xsgn = 1;
			int xtrans = 0;
			if ((borderId == HEX_NW) || (borderId == HEX_NE) || (borderId == HEX_N)) {
				ysgn = 1;
				ytrans = 0;
			} else {
				ysgn = -1;
				ytrans = playfieldTiling.getSurroundingHeight() - 1;
			}
			if ((borderId == HEX_NW) || (borderId == HEX_SW) || (borderId == HEX_N) || (borderId == HEX_S)) {
				xsgn = 1;
				xtrans = 0;
			} else {
				xsgn = -1;
				xtrans = playfieldTiling.getSurroundingWidth() - 1;
			}
			if ((borderId == HEX_N) || (borderId == HEX_S)) { 
				for (int width = 0; width < distribs.length; width++) {
					for (int x = tileWidth / 4; x < tileWidth * 3 / 4; x++) {
						final Color color;
						if (ImageKey.FOREST.equals(imageKey)) {
							color = new Color(destImg.getRGB(c.x + xtrans + xsgn * x, c.y + ytrans + ysgn * width + ysgn * zoomLevel.resolution / 3));
						} else {
							final int colorIndex = MathUtil.randomDistributed(distribs[width]);
							color = colors[colorIndex];
						}
						destImg.setRGB(c.x + xtrans + xsgn * x, c.y + ytrans + ysgn * width, color.getRGB());
					}
				}
			} else { // All directions added per hex support
				int x = -distribs.length;
				final int yoff0 = getYOffset(zoomLevel, x);
				int yoff1 = getYOffset(zoomLevel, x + 1);
				for (int y = yoff0; y >= 0; y--) {
					if ((1 != y) && (y <= yoff1 + 1)) {
						x++;
						yoff1 = getYOffset(zoomLevel, x + 1);
					}
					for (int width = 0; width < distribs.length; width++) {
						if ((0 <= x) || ((playfieldTiling.getSurroundingHeight() - 1 - y) >= getYOffset(zoomLevel, x + width))) {
							final Color color;
							if (ImageKey.FOREST.equals(imageKey)) {
								color = new Color(destImg.getRGB(
										c.x + xtrans + xsgn * (x + width) + xsgn * zoomLevel.resolution / 6, 
										c.y + ytrans + ysgn * y + ysgn * zoomLevel.resolution / 3));
							} else {
								final int colorIndex = MathUtil.randomDistributed(distribs[width]);
								color = colors[colorIndex];
							}
							destImg.setRGB(c.x + xtrans + xsgn * (x + width), c.y + ytrans + ysgn * y, color.getRGB());
						}
					}
				}
			}
		}
	}

	protected int[][] getColorDistribution(final ZoomLevel zoomLevel, final ImageKey imageKey) {
		final int[][] distribs;
		if (ImageKey.SEA.equals(imageKey)) {
			if (zoomLevel.resolution <= 16) {
				distribs = new int[][] {
					{ 9, 3, 1, 0, 1},
					{ 3, 1, 3, 0, 3},
				};
			} else if (zoomLevel.resolution <= 24) {
				distribs = new int[][] {
					{ 9, 3, 1, 0, 1},
					{ 6, 2, 3, 0, 2},
					{ 3, 1, 1, 0, 3},
				};
			} else {
				distribs = new int[][] {
					{ 9, 3, 1, 0, 1},
					{ 6, 2, 3, 0, 2},
					{ 6, 2, 3, 0, 2},
					{ 3, 1, 1, 0, 3},
				};
			}
		} else if (ImageKey.UNEXPLORED_NEIGHBOR.equals(imageKey)) {
			if (zoomLevel.resolution <= 16) {
				distribs = new int[][] {
					{ 9, 1, 1, 0, 0},
					{ 3, 3, 2, 1, 1},
					{ 1, 2, 3, 2, 5},
				};
			} else if (zoomLevel.resolution <= 24) {
				distribs = new int[][] {
					{ 9, 1, 1, 0, 0},
					{ 5, 4, 3, 2, 1},
					{ 3, 3, 3, 1, 3},
					{ 1, 2, 3, 2, 5},
				};
			} else {
				distribs = new int[][] {
					{ 9, 1, 1, 0, 0},
					{ 9, 7, 5, 1, 1},
					{ 3, 3, 2, 1, 1},
					{ 1, 2, 2, 1, 3},
					{ 1, 2, 3, 1, 5},
				};
			}
		} else if (ImageKey.FOREST.equals(imageKey)) {
			final int[] distrib = new int[] { 0, 1, 0, 0, 0, 0, 0, }; 
			distribs = new int[zoomLevel.resolution / 8][];
			for (int i = 0; i < distribs.length; i++) {
				distribs[i] = distrib;
			}
		} else if (ImageKey.ROUGH.equals(imageKey)) {
			final int[] distrib = new int[] { 4, 1, 8, 16, 8, 1, 0, };
			distribs = new int[zoomLevel.resolution / 8][];
			for (int i = 0; i < distribs.length; i++) {
				distribs[i] = distrib;
			}
		} else if (ImageKey.MOUNTAIN.equals(imageKey)) {
			final int[] distrib = new int[] { 1, 3, 5, 20, 24, 8, 8, };
			distribs = new int[zoomLevel.resolution / 6][];
			for (int i = 0; i < distribs.length; i++) {
				distribs[i] = distrib;
			}
		} else {
			distribs = null;
		}
		return distribs;
	}

	protected TileTemplateData getTTD(final ImageKey imageKey) {
		return TileTemplate.getTTD(imageKey.getAll());
	}

	protected void fillUpUndefinedTileParts(final BufferedImage destImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant, final int id, final Color color) {
		// Default implementation does nothing
	}

	protected ArrayList<int[]> createLeftRightTopBottomMergeList(final int bits, final int columns) {
		final ArrayList<int[]> list = new ArrayList<int[]>();
		for (int bit = 0; bit < bits; bit++) {
			final int srcId = (1 << bit);
			for (int id = 1; id < (1 << bits); id++) {
				if ((0 != (id & srcId)) && (id != srcId)) {
					final int[] item = new int[] { srcId % columns, srcId / columns, id % columns, id / columns, };
					list.add(item);
					logger.trace("Merging (" + item[0] + "," + item[1] + ") into (" + item[2] + "," + item[3] + ")");
				}
			}
		}
		return list;
	}

	protected ArrayList<int[]> createBackgroundMergeList(final int bits, final int columns) {
		final int amount = (1 << bits);
		final ArrayList<int[]> list = new ArrayList<int[]>(amount);
		for (int id = 1; id < amount; id++) {
			list.add(new int[] { 0, 0, id % columns, id / columns, });
		}
		return list;
	}

	protected void mergeNeighborModifications(final BufferedImage destImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, int variant) {
		if (ImageKey.UNEXPLORED_NEIGHBOR.equals(imageKey)) {
			logger.trace("mergeNeighborModifications(" + zoomLevel.resolution + "," + imageKey + "," + variant + ")");
			final ArrayList<int[]> list = createLeftRightTopBottomMergeList(getBits(), destTilesPerBlock);
			final Point d = getDestIconPos(imageKey, variant);
			for (final int[] tilecopy : list) {
				mergeTile(destImg, zoomLevel, d, tilecopy, new BrighterComparator());
			}
		} else if (ImageKey.SEA.equals(imageKey)) {
			logger.trace("mergeNeighborModifications(" + zoomLevel.resolution + "," + imageKey + "," + variant + ")");
			final ArrayList<int[]> rltbList = createLeftRightTopBottomMergeList(getBits(), destTilesPerBlock);
			final ArrayList<int[]> backList = createBackgroundMergeList(getBits(), destTilesPerBlock);
			final Point d = getDestIconPos(imageKey, variant);
			for (final int[] tilecopy : rltbList) {
				mergeTile(destImg, zoomLevel, d, tilecopy, new GreenerComparator());
			}
			for (final int[] tilecopy : backList) {
				mergeTile(destImg, zoomLevel, d, tilecopy, new WhiteScreenComparator());
			}
		} else {
			// Nothing to do
		}
	}

	protected Point getDestIconPos(final ImageKey imageKey, int variant) {
		return getDestIconPos(imageKey, variant, 0);
	}

	protected Point getDestIconPos(final ImageKey imageKey, int variant, int id) {
		final TileTemplateData ttd = getTTD(imageKey);
		return new Point(
				ttd.iconX + variant * ttd.iconModifier.x + (id % destTilesPerBlock), 
				ttd.iconY + variant * ttd.iconModifier.y + (id / destTilesPerBlock));
	}

	protected Rectangle getSrcIconRect(final ImageKey imageKey, int variant) {
		return getSrcIconRect(imageKey, variant, 0);
	}

	protected Rectangle getSrcIconRect(final ImageKey imageKey, int variant, int id) {
		final ImageKeyData ikd = imageKeyDataColl.get(imageKey);
		final Rectangle rect = ikd.srcTileUL;
		return new Rectangle (
				rect.x + variant * ikd.srcLookVector.x + (id % rect.width), 
				rect.y + variant * ikd.srcLookVector.y + (id / rect.width),
				rect.width,
				rect.height);
	}

	/**
	 * Moves a block of old square topology format (4x4 blocks) to the adequate
	 * positions in the new hexagon-adapted square topology format (8x8 blocks).
	 * 
	 * @param destImg
	 *            the image to modify in-place
	 * @param srcImg the source image
	 * @param zoomLevel
	 *            the size of a tile in pixels
	 * @param dLeft
	 *            the x (in tile coordinates) of the most-left destination tile
	 * @param dTop
	 *            the y (in tile coordinates) of the most-top destination tile
	 * @param sLeft
	 *            the x (in tile coordinates) of the most-left source tile
	 * @param sTop
	 *            the y (in tile coordinates) of the most-top source tile
	 * @param sWidth
	 *            the width (in tiles) of the block
	 * @param sHeight
	 *            the height (in tiles) of the block 
	 */
	protected void convertBlock(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel, 
			final int dLeft, final int dTop, 
			final int sLeft, final int sTop, 
			final int sWidth, final int sHeight) {
		final int dwidth = (sWidth == 4 ? destTilesPerBlock : sWidth);
		for (int y = 0; y < sHeight; y++) {
			for (int x = 0; x < sWidth; x++) {
				final int id;
				if ((4 == sWidth) && (4 == sHeight)) { 
					id = x * sWidth + y; // re-shuffle landscape tiles to fit to hexagon logic
				} else {
					id = y * sWidth + x;
				}
				final int dx = dLeft + (id % dwidth);
				final int dy = dTop  + (id / dwidth);
				final int sx = sLeft + x;
				final int sy = sTop + y;
				if ((dx == sx) && (dy == sy)) {
					// Nothing to do, everything is in place
				} else {
					copyTile(destImg, srcImg, zoomLevel, dx, dy, sx, sy, 0, 0, ALL);
					if (destImg == srcImg) {
						clearTile(srcImg, zoomLevel, sLeft + x, sTop + y);
					}
				}
			}
		}
	}

	/**
	 * Copies a tile (or only half a tile) from the source image to the destination image.
	 * @param destImg the destimation image
	 * @param srcImg the source image
	 * @param zoomLevel the size of a tile in pixels
	 * @param imageKey the icon type this tile represents
	 * @param variant the look of the image key, if there is more than one look (e.g., fields)
	 * @param id the id of the modification of the icon type, e.g., for sea units with nearby land
	 * @param sTileX the x coordinate of the source tile
	 * @param sTileY the y coordinate of the source tile
	 * @param sOffsetX the x offset of the (0, 0) tile in the source image
	 * @param sOffsetY the y offset of the (0, 0) tile in the source image
	 * @param halfs {@link #UPPER} for upper half, {@link #LOWER} for lower half, or {@link #ALL} for both halfs - the rest is filled with white color.
	 */
	protected void copyTile(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel, 
			final ImageKey imageKey, final int variant, final int id, 
			final int sTileX, final int sTileY, final int sOffsetX, final int sOffsetY,
			final int halfs) {
		final Point p = getDestIconPos(imageKey, variant, id);
		copyTile(destImg, srcImg, zoomLevel, p.x, p.y, sTileX, sTileY, sOffsetX, sOffsetY, halfs);
	}

	/**
	 * Copies a tile (or only half a tile) from the source image to the destination image.
	 * @param destImg the destimation image
	 * @param srcImg the source image
	 * @param zoomLevel the size of a tile in pixels
	 * @param dTileX the x coordinate of the destination tile
	 * @param dTileY the y coordinate of the destination tile
	 * @param sTileX the x coordinate of the source tile
	 * @param sTileY the y coordinate of the source tile
	 * @param sOffsetX the x offset of the (0, 0) tile in the source image
	 * @param sOffsetY the y offset of the (0, 0) tile in the source image
	 * @param halfs {@link #UPPER} for upper half, {@link #LOWER} for lower half, or {@link #ALL} for both halfs - the rest is filled with white color.
	 */
	protected void copyTile(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel, final int dTileX,
			final int dTileY, final int sTileX, final int sTileY, final int sOffsetX, final int sOffsetY,
			final int halfs) {
		final int zoom = zoomLevel.resolution;
		final Point d = calcDestPosUpperLeftInPixels(zoomLevel, dTileX, dTileY, true, ORIGIN);
		final Point s = calcDestPosUpperLeftInPixels(zoomLevel, sTileX, sTileY, (destImg == srcImg), new Point(sOffsetX, sOffsetY));
		logger.trace("copyTile: d=" + d + ", s=" + s);
		final Point o = getOffset(zoomLevel);
		for (int x = 0; x < zoom; x++) {
			final int yoff = getYOffset(zoomLevel, x + o.x);
			for (int y = yoff; y < zoom - yoff; y++) {
				int val = Color.white.getRGB();
				if ((y < zoom / 2) && (0 != (halfs & UPPER)) || (y >= zoom / 2) && (0 != (halfs & LOWER))) {
					val = srcImg.getRGB(s.x + x, s.y + y);
				}
				try {
					destImg.setRGB(d.x + x + o.x, d.y + y + o.y, val);
				} catch (ArrayIndexOutOfBoundsException ex) {
					logger.fatal("zoom=" + zoom + ", dTile=(" + dTileX + "," + dTileY + "), d=" + d + ", o=" + o + ", x=" + x + ", y=" + y, ex);
				}
			}
		}
	}

	protected void clearTile(final BufferedImage destImg, final ZoomLevel zoomLevel, final int dTileX, final int dTileY) {
		final int zoom = zoomLevel.resolution;
		final Point d = calcDestPosUpperLeftInPixels(zoomLevel, dTileX, dTileY, true, ORIGIN);
		final Point o = getOffset(zoomLevel);
		for (int x = 0; x < zoom; x++) {
			final int yoff = getYOffset(zoomLevel, x + o.x);
			for (int y = yoff; y < zoom - yoff; y++) {
				destImg.setRGB(d.x + x + o.x, d.y + y + o.y, 0x00000000);
			}
		}
	}

	/**
	 * Merges two tiles.
	 * 
	 * @param destImg
	 *            image to modify. Source and destination tile are both part of
	 *            that image.
	 * @param zoomLevel
	 *            the size of a tile in pixels
	 * @param tileOffset
	 *            The offset (in tile coordinates) of the top-left tile to modify
	 * @param is
	 *            an array of length 4. The first two entries contain x and y
	 *            tile position of the source tile, the other two entries
	 *            contain x and y tile position of the destination tile.
	 * @param comparator
	 *            the comparator to decide if source or destination pixel wins
	 */
	protected void mergeTile(final BufferedImage destImg, final ZoomLevel zoomLevel, final Point tileOffset, final int[] is, final ColorComparator comparator) {
		final int zoom = zoomLevel.resolution;
		final int sxpos = (tileOffset.x + is[0]);
		final int sypos = (tileOffset.y + is[1]);
		final int dxpos = (tileOffset.x + is[2]);
		final int dypos = (tileOffset.y + is[3]);
		final Point d = calcDestPosUpperLeftInPixels(zoomLevel, dxpos, dypos, true, ORIGIN);
		final Point s = calcDestPosUpperLeftInPixels(zoomLevel, sxpos, sypos, true, ORIGIN);
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		logger.trace("Merging tile (" + sxpos + "," + sypos + ") into (" + dxpos + "," + dypos + ")");
		for (int x = 0; x < playfieldTiling.getSurroundingWidth(); x++) {
			final int yoff = getYOffset(zoomLevel, x);
			for (int y = yoff; y < zoom - yoff; y++) {
				final int sval = destImg.getRGB(s.x + x, s.y + y);
				final int dval = destImg.getRGB(d.x + x, d.y + y);
				if (comparator.isToCopyDest(sval, dval)) {
					destImg.setRGB(d.x + x, d.y + y, sval);
				}
			}
		}
	}

	protected Point calcDestPosUpperLeftInPixels(final ZoomLevel zoomLevel, final int dTileX, final int dTileY, final boolean isInplaceDest, final Point offset) {
		final Point d = new Point();
		if (!isInplaceDest) {
			d.x = dTileX * zoomLevel.resolution;
			d.y = dTileY * zoomLevel.resolution;
			d.translate(offset.x, offset.y);
		} else {
			props.map2canvas(d, new PlayfieldPosition(dTileX, dTileY), offset, zoomLevel, PlayfieldTiling.TOP_LEFT);
			ImageSet.translateToImageSet(d, props.getTopologyMode(), zoomLevel);
		}
		return d;
	}

	/**
	 * @param zoomLevel the size of a tile in pixels
	 * @param x the pixel column of this tile, in surrounding box coordinates
	 * @return the first row for the given column that contains data and is not transparent
	 */
	public int getYOffset(ZoomLevel zoomLevel, int x) {
		int ret = 0;
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		if (x < playfieldTiling.getSurroundingWidth() / 2) {
			if (((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()))) {
				ret = Math.max(0, playfieldTiling.getSurroundingHeight() / 2 - 2 * x - 1);
			}
		} else {
			ret = getYOffset(zoomLevel, playfieldTiling.getSurroundingWidth() - 1 - x);
		}
		return ret;
	}

	protected Point getOffset(final ZoomLevel zoomLevel) {
		int ret = 0;
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(props.getTopologyMode());
		if (((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()))) {
			ret = zoomLevel.resolution / 8;
		}
		return new Point(ret, 0);
	}

	protected boolean isNeighborMod(final ZoomLevel zoomLevel) {
		return (8 <= zoomLevel.resolution);
	}

	protected int[][] appendCopyParams(final int[][] arr1, final int[][] arr2) {
		final int[][] ret = new int[arr1.length + arr2.length][arr1[0].length];
		for (int i = 0; i < arr1.length; i++) {
			ret[i] = arr1[i];
		}
		for (int i = 0; i < arr2.length; i++) {
			ret[arr1.length + i] = arr2[i];
		}
		return ret;
	}

	protected int calcRiverRGB() {
		final Color[] c = new Color[] { 
				Color.white, 
				Color.blue.brighter().brighter(), 
				Color.blue.brighter(), 
				Color.blue, 
				Color.blue.darker(), 
				Color.blue.darker().darker(), 
				Color.blue.darker().darker().darker(), 
				Color.black, };
		final int[] probls = new int[] { 0, 1, 2, 4, 8, 12, 6, 2}; 
		return c[MathUtil.randomDistributed(probls)].getRGB();
	}

	protected abstract int getBits();

	/**
	 * @param destImg the destimation image
	 * @param srcImg the source image
	 * @param zoomLevel the size of a tile in pixels
	 * @param imageKey the image key representing the landscape block
	 * @param variant The look of the tile, if there are several looks (e.g., for fields)
	 * @return if the landscape block has been defined by this method. Default implementation returns <code>false</code>.
	 */
	protected boolean createFieldBlock(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel, 
			final ImageKey imageKey, final int variant) {
		return false;
	}

	protected abstract void copyBlock(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant);

	protected abstract int[][] calcCopyParamsUpDownSides(final ZoomLevel zoomLevel);

	public void copyTiles(final BufferedImage destImg, final BufferedImage srcImg, final ZoomLevel zoomLevel,
			final ImageKey imageKey, final int variant, final int[][] copyParams) {
		final Point d = getDestIconPos(imageKey, variant);
		final Rectangle s = getSrcIconRect(imageKey, variant);
		for (int[] copyParam : copyParams) {
			copyTile(destImg, srcImg, zoomLevel,
					d.x + copyParam[0],
					d.y + copyParam[1],
					s.x + copyParam[2], 
					s.y + copyParam[3], 0, 0,
					copyParam[4]);
		}
	}

	protected float getBrightness(final int rgb) {
		final int r = (rgb >> 16) & 0xff;
		final int g = (rgb >>  8) & 0xff;
		final int b = (rgb >>  0) & 0xff;
		float[] ret = Color.RGBtoHSB(r, g, b, null);
		return ret[2];
	}

	protected final GameProperties props = new GameProperties();

	protected static final int srcTilesPerBlock = 4;
	protected static final int destTilesPerBlock = 8;

	protected static final Point ORIGIN = new Point(0, 0);
	protected static final Point SRC_TRANS_VECTOR_LOOK = new Point(0, srcTilesPerBlock);

	protected final Logger logger = Logger.getLogger(getClass().getName());
	private String todir = ".";
}
