package de.ped.empire.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Font;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.SwingConstants;
import javax.swing.border.Border;

import de.ped.empire.logic.TileTemplate;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.StatusReportPlayerData;
import de.ped.empire.logic.CommandableTemplate;
import de.ped.empire.logic.Templates;
import de.ped.empire.logic.ZoomLevel;
import de.ped.tools.Assert;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.TopologyMode;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A dialog to show and change the current city production.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class CityProductionDialog extends PedJDialog {

	private static final long serialVersionUID = 1L;

	/**
	 * Contains all data necessary for a table row of this table to build it data driven.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ProductionUnitModel {
		StatusReportPlayerData.StatusReportUnitData data;
		JRadioButton radioButton;
		JLabel iconLabel;
		JLabel turnsLabel;
		JLabel inProductionLabel;
		JLabel nextReadyInLabel;
		private final EmpireMainWindow empWindow;

		public ProductionUnitModel(final EmpireMainWindow empireWindow, final ButtonGroup modeButtonGroup) {
			this.empWindow = empireWindow;
			radioButton = new JRadioButton();
			radioButton.setText("----------");
			iconLabel = new JLabel("", new ImageIcon(ZoomLevel.DIALOG_ZOOM_LEVEL.getImageSet(TopologyMode.SQUARE8).getTransparentImage()), JLabel.LEFT);
			turnsLabel = new JLabel("000");
			inProductionLabel = new JLabel("000");
			nextReadyInLabel = new JLabel("000");
			modeButtonGroup.add(radioButton);
			iconLabel.setLabelFor(radioButton);
			turnsLabel.setLabelFor(radioButton);
			inProductionLabel.setLabelFor(radioButton);
			nextReadyInLabel.setLabelFor(radioButton);
		}

		public void setData(final StatusReportPlayerData.StatusReportUnitData data) {
			this.data = data;
			Assert.assertNotNull(data);
			Assert.assertNotNull(data.gameView);
			empWindow.putValues(radioButton, 
					new I18NStringWithFillIns(data.template.messageKeyPrefix, data.template.messageKeyPrefix, null));
			setTurns(Integer.MAX_VALUE);
			if (!data.template.isUndefinedOrNothing()) {
				final ImageKey imageKey = new ImageKey(ImageKey.TType.Unit, data.template.id);
				imageKey.setBits(ImageKey.BA_PLAYER_COLOR, data.gameView.getPlayer().getPlayerColorId());
				iconLabel.setIcon(new ImageIcon(TileTemplate.createImage(imageKey.getAll(), ZoomLevel.DIALOG_ZOOM_LEVEL, TopologyMode.SQUARE8)));
			}
			inProductionLabel.setText(Integer.toString(data.inProduction));
			final String turnsText = (data.nextReadyIn == Integer.MAX_VALUE ? "--" : Integer.toString(data.nextReadyIn));
			nextReadyInLabel.setText(turnsText);
		}
		
		public void setTurns(final int turnsToProduce) {
			final String turnsText = (turnsToProduce == Integer.MAX_VALUE ? "--" : Integer.toString(turnsToProduce));
			turnsLabel.setText(turnsText);
		}
	}

	private final EmpireMainWindow empWindow;
	private StatusReportPlayerData data;
	private LinkedList<ProductionUnitModel> productionUnitModels = new LinkedList<ProductionUnitModel>();
	private final FieldCity city;

	public CityProductionDialog(
			final ApplicationMainWindow mainWindow, 
			final Frame parent, 
			final FieldCity city) {
		super(mainWindow, parent);
		this.city = city;
		final Messages messages = getMessages();
		setTitle(messages.getString("Game.City.Title"));
		Assert.assertNotNull(messages);
		this.empWindow = (EmpireMainWindow) mainWindow;
		setResizable(false);
		
		// Title panel: City name and productivity
		final GridLayout titleLayout = new GridLayout(2, 2, 10, 5);
		final JPanel titlePanel = new JPanel();
		titlePanel.setLayout(titleLayout);
		Assert.assertNotNull(city);
		final JLabel titleLabel = new JLabel(messages.getString(city.getName()));
		final Font plainFont = titleLabel.getFont().deriveFont(Font.PLAIN);
		final Font boldFont = plainFont.deriveFont(Font.BOLD);
		{
			final JLabel townNameLabel = new JLabel(messages.getString("Game.Field.City.Text"));
			townNameLabel.setHorizontalAlignment(SwingConstants.RIGHT);
			townNameLabel.setFont(plainFont);
			titlePanel.add(townNameLabel);
		}
		{
			titleLabel.setFont(boldFont);
			titlePanel.add(titleLabel);
		}
		{
			final JLabel productivityNameLabel = new JLabel(messages.getString("Game.Field.City.Productivity.Text"));
			productivityNameLabel.setToolTipText(messages.getShortDescription("Game.Field.City.Productivity"));
			productivityNameLabel.setHorizontalAlignment(SwingConstants.RIGHT);
			productivityNameLabel.setFont(plainFont);
			titlePanel.add(productivityNameLabel);
		}
		{
			final JLabel productivityLabel = new JLabel(Integer.toString(city.getProductivity()) + "%");
			productivityLabel.setFont(plainFont);
			titlePanel.add(productivityLabel);
		}
		final Border lineBorder = BorderFactory.createLineBorder(Color.gray);
		titlePanel.setBorder(lineBorder);
		getContentPane().add(titlePanel, BorderLayout.NORTH);

		// Main panel: Units to produce
		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);
		
		final List<CommandableTemplate> unitTemplates = city.getUnitsToProduce();
		init(mainPanel, gridbag, city, unitTemplates);
		addAllPropertyElementsTo(mainPanel, c);

		add(mainPanel, BorderLayout.CENTER);
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		setDefaultButton(getButtonOk());
		finishLayout();
	}

	protected void init(
			final JPanel panel, 
			final GridBagLayout gridbag, 
			final FieldCity city, 
			final List<CommandableTemplate> unitTemplates) {
		final ButtonGroup modeButtonGroup = new ButtonGroup();
		final GridBagConstraints c = new GridBagConstraints();
		c.ipadx = 10;
		// Heading
		final GridBagTableColumn columns[] = new GridBagTableColumn[] {
				new GridBagTableColumn(2, GridBagConstraints.WEST, "Game.City.UnitToProduce.Heading"),
				new GridBagTableColumn(1, GridBagConstraints.CENTER, "Game.City.TurnsToFinish.Heading"),
				new GridBagTableColumn(1, GridBagConstraints.CENTER, "Game.Status.Units.InProduction.Heading"),
				new GridBagTableColumn(GridBagConstraints.REMAINDER, 
										GridBagConstraints.CENTER, "Game.Status.Units.NextReadyIn.Heading"),
			};
		for (final GridBagTableColumn column : columns) {
			c.gridwidth = column.gridwidth;
			c.anchor = column.anchor; 
			final JLabel label = empWindow.createLabel(column.key);
			gridbag.setConstraints(label, c);
			panel.add(label);
		}
		Logger logger = Logger.getLogger(getClass());
		logger.setToShowStack(true);
		logger.setLogLevel(Level.INFO);
		logger.debug("City productivity=" + city.getProductivity() + ", preferred unit=" + city.getPreferredUnit().messageKeyPrefix);
		// Content
		data = empWindow.getGameFacade().getStatusReportPlayerData(city.getPlayerId(), true);
		for (final CommandableTemplate template : unitTemplates) {
			final ProductionUnitModel pum = new ProductionUnitModel(empWindow, modeButtonGroup);
			StatusReportPlayerData.StatusReportUnitData unitData = null;
			for (final StatusReportPlayerData.StatusReportUnitData ud : data.units) {
				if (ud.template.equals(template)) {
					unitData = ud;
					break;
				}
			}
			pum.setData(unitData);
			pum.radioButton.setSelected(pum.data.template == city.getUnitInProduction());
			pum.setTurns(city.getTurnsToProduce(pum.data.template));
			productionUnitModels.add(pum);
			{
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
				gridbag.setConstraints(pum.radioButton, c);
				if ((!pum.data.template.isUndefinedOrNothing()) && (pum.data.template.equals(city.getPreferredUnit()))) {
					logger.debug("Found preferred unit " + pum.data.template.messageKeyPrefix);
					pum.radioButton.setForeground(Color.blue);
					pum.radioButton.setText(pum.radioButton.getText() + " *");
				}
				panel.add(pum.radioButton);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
				gridbag.setConstraints(pum.iconLabel, c);
				panel.add(pum.iconLabel);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.turnsLabel, c);
				panel.add(pum.turnsLabel);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.inProductionLabel, c);
				panel.add(pum.inProductionLabel);
			} {
				c.gridwidth = GridBagConstraints.REMAINDER; //end row
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.nextReadyInLabel, c);
				panel.add(pum.nextReadyInLabel);
			}
			pum.radioButton.addActionListener(this);
		}
		getButtonOk().setEnabled(Templates.instance().UNDEFINED != city.getUnitInProduction());
	}

	/**
	 * When the dialog was finished, returns the production chosen by the user.
	 * 
	 * @return the production chosen by the user
	 */
	public CommandableTemplate getProduction() {
		CommandableTemplate ret = Templates.instance().UNDEFINED;
		for (final ProductionUnitModel model : productionUnitModels) {
			if (model.radioButton.isSelected()) {
				ret = model.data.template;
				break;
			}
		}
		return ret;
	}

	@Override
	public void actionPerformed(final ActionEvent e) {
		super.actionPerformed(e);
		if (e.getSource() instanceof JRadioButton) {
			getButtonOk().setEnabled(true);
		}
	}

	public FieldCity getCity() {
		return city;
	}
}
