package de.ped.empire.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;

import javax.sound.sampled.Clip;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollBar;
import javax.swing.SwingUtilities;

import de.ped.empire.logic.AnimationPart;
import de.ped.empire.logic.CircularReferenceException;
import de.ped.empire.logic.Commandable;
import de.ped.empire.logic.CommandableKey;
import de.ped.empire.logic.FieldTemplate;
import de.ped.empire.logic.GameFacade;
import de.ped.empire.logic.GameNotification;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.GameObserver;
import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.GameState;
import de.ped.empire.logic.GameStateBasics;
import de.ped.empire.logic.GameViewProperties;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.MapViewMode;
import de.ped.empire.logic.MilitaryBranch;
import de.ped.empire.logic.MoveResult;
import de.ped.empire.logic.SoundTypes;
import de.ped.empire.logic.TileTemplate;
import de.ped.empire.logic.MoveResult.Role;
import de.ped.empire.logic.Order;
import de.ped.empire.logic.OrderDefineMovePaths;
import de.ped.empire.logic.OrderTemplate;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.ProcessorSpeedType;
import de.ped.empire.logic.RulesComplexity;
import de.ped.empire.logic.Templates;
import de.ped.empire.logic.Tileable;
import de.ped.empire.logic.Unit;
import de.ped.empire.logic.CommandableTemplate;
import de.ped.empire.logic.ZoomLevel;
import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextDocumentType;
import de.ped.tools.FinishableThread;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.PlayerColor;
import de.ped.tools.Messages.I18NedOrNot;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.ResourceFinder;
import de.ped.tools.ResourceFinder.Folder;
import de.ped.tools.TextFormatter;
import de.ped.tools.TextFormatter.TableCellInfo;
import de.ped.tools.TopologyMode;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.CheckBoxMenuItemAction;
import de.ped.tools.gui.ConfigurableFileFilter;
import de.ped.tools.gui.GameThread;
import de.ped.tools.gui.HtmlPaneDialog;
import de.ped.tools.gui.JInfoDialog;
import de.ped.tools.gui.JScrollAndZoomPane;
import de.ped.tools.gui.PedAction;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PedJMenu;
import de.ped.tools.gui.UserGameActionInput;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.sound.SoundClip;

/**
 * The main window of the application.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class EmpireMainWindow extends ApplicationMainWindow implements GameObserver {
	private static final long serialVersionUID = 1L;

	static final TopologyMode[] TOPOLOGY_MODES = new TopologyMode[] { TopologyMode.SQUARE8, TopologyMode.SQUARE4, TopologyMode.HEXAGON, TopologyMode.HEXAGON_H, };

	/** 
	 * The action to start a new game.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	static class StartNewGameAction extends ApplicationMainWindow.StartNewGameAction {
		private static final long serialVersionUID = 1L;

		public StartNewGameAction(final ApplicationMainWindow parent) {
			super(parent);
		}

		@Override
		public boolean isToBeEnabled() {
			return (!((EmpireMainWindow) getParent()).isGameRunning());
		}
	}

	/** 
	 * The action to exit the current game.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	static class ExitGameAction extends ApplicationMainWindow.ExitGameAction {
		private static final long serialVersionUID = 1L;

		public ExitGameAction(final ApplicationMainWindow parent) {
			super(parent);
		}

		@Override
		public boolean isToBeEnabled() {
			return ((EmpireMainWindow) getParent()).isGameRunning();
		}
	}

	/** 
	 * The action to resign the current player.
	 *
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ResignPlayerAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ResignPlayerAction() {
			super("Action.ResignPlayer", EmpireMainWindow.this, ResourceFinder.Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			resignCurrentPlayer();
		}

		@Override
		public boolean isToBeEnabled() {
			return (Player.ID_NONE != getCurrentPlayerId());
		}
	}

	/** 
	 * The action to load a game.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class LoadGameAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public LoadGameAction() {
			super("Action.LoadGame", EmpireMainWindow.this, ResourceFinder.Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			loadGame();
		}

		@Override
		public boolean isToBeEnabled() {
			return ((!isGameRunning()) && (getApplicationEnvironment().isExitable()));
		}
	}

	/** 
	 * The action to load a game.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class SaveGameAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public SaveGameAction() {
			super("Action.SaveGame", EmpireMainWindow.this, ResourceFinder.Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			saveGame(this);
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			return (getApplicationEnvironment().isExitable() && (Player.ID_NONE != playerId)
					&& (gameFacadeProxy.getCurrentPlayerId() == playerId));
		}
	}

	/** 
	 * The action to center screen on the current commandable in focus.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewCenterAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewCenterAction() {
			super("Action.ViewCenter", EmpireMainWindow.this, ResourceFinder.Folder.APP, false);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			centerScreen(true);
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			return ((Player.ID_NONE != playerId) && (!gameFacadeProxy.isFinishTurnEnabled(playerId)));
		}
	}

	/** 
	 * The action to show the context menu on the current commandable in focus.
	 *
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewContextMenuAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewContextMenuAction() {
			super("Action.ViewContextMenu", EmpireMainWindow.this, Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			final int playerId = getCurrentPlayerId();
			if (Player.ID_NONE != playerId) {
				final PlayfieldPosition pos = gameFacadeProxy.getPositionInFocus(playerId);
				actionAtPosition(pos);
			}
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			boolean ret = false;
			if (!isBusy() && (Player.ID_NONE != playerId) && gameFacadeProxy.isDefaultPerspective(playerId)) {
				final GameFacade.UserMapTileContextOptions options = gameFacadeProxy.collectMapTileContextOptions(playerId, null);
				final List<Order> orderOptions = options.orders;
				final List<Commandable> commandableOptions = options.commandables;
				if (CollectionUtil.isNullOrEmpty(orderOptions) && CollectionUtil.isNullOrEmpty(commandableOptions)) {
					// Nothing to do
				} else {
					ret = true;
				}
			}
			return ret;
		}
	}

	/** 
	 * The action to change the zoom level of the screen.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewZoomLevelAction extends PedAction {
		private static final long serialVersionUID = 1L;

		final int index;

		public ViewZoomLevelAction(final int i) {
			super("Action.ViewZoomLevel.Level" + i, EmpireMainWindow.this, ResourceFinder.Folder.COMMON, false);
			index = i;
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			setZoomLevel(index);
		}

		@Override
		public boolean isToBeEnabled() {
			return ((Player.ID_NONE != getCurrentPlayerId()) && isGameRunning());
		}
	}

	/** 
	 * The action to change the perspective of the map.
	 *
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewPerspectiveAction extends PedAction {
		private static final long serialVersionUID = 1L;

		final int index;
		private final JRadioButtonMenuItem button;

		/**
		 * Constructor.
		 * @param i 0 for world perspective, 1 to 6 for player perspectives
		 * @param buttonGroup the button group that perspective button will be part of
		 */
		public ViewPerspectiveAction(final int i, final ButtonGroup buttonGroup) {
			super("Action.ViewPerspective" + (0 == i ? ".World" : ".Player"), EmpireMainWindow.this, (0 == i ? Folder.APP : null), false);
			index = i;
			if (0 != i) {
				final ImageKey imageKey = new ImageKey(ImageKey.TType.Unit, Templates.instance().CMDL_CITY.id);
				imageKey.setBits(ImageKey.BA_PLAYER_COLOR, getGameProperties().getPlayers()[i - 1].getPlayerColorId());
				putValue(Action.SMALL_ICON, new ImageIcon(TileTemplate.createImage(imageKey.getAll(), ZoomLevel.DIALOG_ZOOM_LEVEL, TopologyMode.SQUARE8)));
			}
			button = createRadioButtonMenuItem(buttonGroup, this);
		}

		@Override
		protected Object[] getI18NParams() {
			final String[] ret;
			if (0 == index) {
				ret = null;
			} else {
				ret = new String[] { gameFacadeProxy.getGameProperties().getPlayers()[index - 1].getName() };
			}
			return ret;
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			gameFacadeProxy.setPerspective(getCurrentPlayerId(), index);
			notifyMapChanged();
			onOrderOptionsChanged();
		}

		@Override
		public boolean isToBeEnabled() {
			boolean ret = false;
			if (!isBusy() && gameFacadeProxy.isPerspectiveEnabled(getCurrentPlayerId(), index)) {
				ret = true;
			}
			return ret;
		}

		@Override
		public boolean isToBeDefined() {
			final boolean ret;
			if (0 == index) {
				ret = true;
			} else {
				final GameProperties props = gameFacadeProxy.getGameProperties();
				ret = ((null != props) && (props.getPlayers()[index - 1].isPlayer()));
			}
			return ret;
		}

		public void setSelected(final boolean b) {
			button.setSelected(b);
		}
	}

	/** 
	 * The action to switch between the map views defined in {@link MapViewMode}.
	 *
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewMapViewModeAction extends PedAction {
		private static final long serialVersionUID = 1L;

		private final MapViewMode mode;
		private final JRadioButtonMenuItem button;

		public ViewMapViewModeAction(final MapViewMode mode, final ButtonGroup buttonGroup) {
			super(mode.messageKeyPrefix, EmpireMainWindow.this, Folder.APP, false);
			this.mode = mode;
			button = createRadioButtonActionItem(null, this, buttonGroup);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			final int playerId = getCurrentPlayerId();
			if (Player.ID_NONE != playerId) {
				gameFacadeProxy.setMapViewMode(playerId, mode);
			}
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			boolean ret = false;
			if (!isBusy() && (Player.ID_NONE != playerId)) {
				if (gameFacadeProxy.isMapViewModeEnabled(playerId, mode)) {
					ret = true;
				}
			}
			return ret;
		}

		public void setSelected(final boolean b) {
			button.setSelected(b);
		}
	}

	/** 
	 * The action to show status report.
	 *
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewStatusReportAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewStatusReportAction() {
			super("Action.ViewStatusReport", EmpireMainWindow.this, Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			final int playerId = getCurrentPlayerId();
			if (Player.ID_NONE != playerId) {
				showStatusReportDialog(playerId);
			}
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			return ((Player.ID_NONE != playerId) && (gameFacadeProxy.getCurrentPlayerId() == playerId));
		}
	}

	/** 
	 * The action to show log book report.
	 *
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewLogReportAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewLogReportAction() {
			super("Action.ViewLogReport", EmpireMainWindow.this, Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			final int playerId = getCurrentPlayerId();
			if (Player.ID_NONE != playerId) {
				showLogReportDialog(playerId);
			}
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			return ((Player.ID_NONE != playerId) && (gameFacadeProxy.getCurrentPlayerId() == playerId));
		}
	}

	/** 
	 * The action to show the game properties dialog.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewGamePropertiesAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewGamePropertiesAction() {
			super("Action.ViewGameProperties", EmpireMainWindow.this, Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			showGamePropertiesDialog();
		}

		@Override
		public boolean isToBeEnabled() {
			return (isGameRunning());
		}
	}

	/** 
	 * The action to show the player view properties dialog.
	 *
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class ViewPlayerPropertiesAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ViewPlayerPropertiesAction() {
			super("Action.ViewPlayerProperties", EmpireMainWindow.this, Folder.APP, true);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			showGameViewPropertiesDialog();
		}

		@Override
		public boolean isToBeEnabled() {
			return (isGameRunning());
		}
	}

	/** 
	 * The action when the user wants to end his/her turn.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class FinishTurnAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public FinishTurnAction() {
			super("Action.FinishTurn", EmpireMainWindow.this, ResourceFinder.Folder.APP, false);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			onActionFinishTurn(true);
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			boolean result = false;
			final boolean isBusy = isBusy();
			if (!isBusy && (Player.ID_NONE != playerId) && (gameFacadeProxy.isFinishTurnEnabled(playerId))) {
				result = true;
			}
			logger.trace("finishTurnAction.updateEnabled for player " + playerId + " resulted in " + result
					+ " with busy=" + isBusy);
			return result;
		}
	}

	/** 
	 * The action when the button for the next commandable was pressed.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class NextCommandableAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public NextCommandableAction() {
			super("Action.NextCommandable", EmpireMainWindow.this, ResourceFinder.Folder.APP, false);
		}

		@Override
		public void securedActionPerformed(final ActionEvent event) {
			onActionNext();
		}

		@Override
		public boolean isToBeEnabled() {
			final int playerId = getCurrentPlayerId();
			boolean result = false;
			if (!isBusy() && (Player.ID_NONE != playerId)
					&& (gameFacadeProxy.isDefaultPerspective(playerId))
					&& (gameFacadeProxy.getMapViewMode(playerId) == MapViewMode.NORMAL)) {
				if (null != gameFacadeProxy.getVisibleCommandableInFocus(playerId)) {
					if (!gameFacadeProxy.isInMoveOver()) {
						result = true;
					}
				}
			}
			return result;
		}
	}

	/** 
	 * The action to select a commandable, e.g., from a popup menu when there were several commandables to choose from.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class SelectCommandableAction extends PedAction {
		private static final long serialVersionUID = 1L;
		private final Commandable commandable;

		SelectCommandableAction(final Commandable commandable) {
			super(getMessages().getString(commandable.getName()));
			this.commandable = commandable;
			setLogger(logger);
		}

		@Override
		public void securedActionPerformed(final ActionEvent ev) {
			onActionSelect(commandable);
		}
	}

	/** 
	 * The action to select an order for the current unit.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class SelectOrderAction extends PedAction {
		private static final long serialVersionUID = 1L;
		public final Order order;

		private SelectOrderAction(final Order order) {
			super(getMessages().getString(order.getToolTip((Player.ID_NONE == getCurrentPlayerId() ? null
					: gameFacadeProxy.getVisibleCommandableInFocus(getCurrentPlayerId())))));
			this.order = order;
			setMessageKeyPrefix(order.template.messageKeyPrefix, EmpireMainWindow.this, ResourceFinder.Folder.APP);
			setStartingDialog(order.template.isStartingDialog);
			setLogger(logger);
		}

		SelectOrderAction(final byte order) {
			this(Order.createGenericOrder(order));
		}

		@Override
		public void securedActionPerformed(final ActionEvent ev) {
			onActionSelect(getCurrentPlayerId(), order);
		}

		@Override
		public boolean isToBeEnabled() {
			boolean result = false;
			final int playerId = getCurrentPlayerId();
			if (Player.ID_NONE != playerId) {
				if (isGameRunning()) {
					if (!isBusy()) {
						result = gameFacadeProxy.isOrderActionEnabled(playerId, order);
					}
				}
			}
			return result;
		}
	}

	/** 
	 * The action to start a mode to select the target for a targeted action for the current unit.
	 *
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class SelectTargetModeAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public final OrderTemplate template;

		SelectTargetModeAction(final OrderTemplate orderTemplate) {
			super(orderTemplate.messageKeyPrefix, EmpireMainWindow.this, ResourceFinder.Folder.APP, false);
			this.template = orderTemplate;
		}

		@Override
		public void securedActionPerformed(final ActionEvent ev) {
			onActionSelectTargetMode(template);
		}

		@Override
		public boolean isToBeEnabled() {
			boolean result = false;
			final int playerId = getCurrentPlayerId();
			if ((Player.ID_NONE != playerId) && gameFacadeProxy.isOrderActionTargetModeEnabled(playerId, template)) {
				result = true;
			}
			return result;
		}
	}

	private final Logger logger = Logger.getLogger(getClass());
	{
		// This is for debugging purposes only; remove when no longer needed.
//		logger.addAppender(new PrintWriterAppender(new File("EmpireMainWindowLogger.log")));
	}

	private GameFacadeProxy gameFacadeProxy = new GameFacadeProxy();
	{
		gameFacadeProxy.setEnvironment(getApplicationEnvironment());
	}

	private GameProperties props = new GameProperties();

	private UserGameActionInput[] userGameActionInputs = null;

	private EmpireMainCanvas worldCanvas;
	private JScrollAndZoomPane worldScrollPane;
	private final JLabel statusBar;
	private final JButton finishTurnButton;
	private final JButton nextButton;
	private JMenu perspectivesSubmenu;

	private ResignPlayerAction resignPlayerAction = new ResignPlayerAction();
	private LoadGameAction loadGameAction = new LoadGameAction();
	private SaveGameAction saveGameAction = new SaveGameAction();
	private CheckBoxMenuItemAction soundOnOffAction = new CheckBoxMenuItemAction(true, "Action.SoundOnOff", this, ResourceFinder.Folder.COMMON);

	private ViewCenterAction viewCenterAction = new ViewCenterAction();
	private ViewZoomLevelAction[] viewZoomLevelActions = new ViewZoomLevelAction[ZoomLevel.INSTANCES.length];
	private JRadioButtonMenuItem[] zoomLevelButtons = new JRadioButtonMenuItem[viewZoomLevelActions.length];
	private ViewPerspectiveAction[] viewPerspectiveActions = new ViewPerspectiveAction[1];
	private ArrayList<ViewMapViewModeAction> viewMapViewModeActions = new ArrayList<ViewMapViewModeAction>(MapViewMode.values().length);
	private ViewStatusReportAction viewStatusReportAction = new ViewStatusReportAction();
	private ViewContextMenuAction viewContextMenuAction = new ViewContextMenuAction();
	private ViewLogReportAction viewLogReportAction = new ViewLogReportAction();
	private ViewGamePropertiesAction viewGamePropertiesAction = new ViewGamePropertiesAction();
	private ViewPlayerPropertiesAction viewPlayerPropertiesAction = new ViewPlayerPropertiesAction();

	private FinishTurnAction finishTurnAction = new FinishTurnAction();
	private NextCommandableAction nextCommandableAction = new NextCommandableAction();
	private SelectOrderAction orderChangeProductionAction;
	private SelectOrderAction orderSkipMoveAction;
	private SelectOrderAction orderClearOrdersAction;
	private SelectOrderAction orderClearAlertsAction;
	private SelectOrderAction orderDisbandAction;
	private SelectOrderAction orderSentryAction;
	private SelectOrderAction orderLoadAction;
	private SelectOrderAction orderUnloadAction;
	private SelectOrderAction orderBuildAirbaseAction;
	private SelectOrderAction orderRenameAction;
	private SelectOrderAction orderHomeAction;
	private SelectOrderAction orderExploreAction;
	private SelectOrderAction orderDefineMovePathAction;
	private SelectTargetModeAction orderGoToTargetAction = new SelectTargetModeAction(OrderTemplate.GOTO);
	private SelectTargetModeAction orderEscortTargetAction = new SelectTargetModeAction(OrderTemplate.ESCORT);

//	private SelectTargetModeAction orderDefineMovePathTargetAction = new SelectTargetModeAction(OrderTemplate.CHANGE_MOVE_PATHS);

	private int zoomLevelIdx;
	/** Stores the index of the locale to use when no player is actively displayed. */
	private int globalLocaleIndex = Messages.DEFAULT_LOCALE_INDEX;

	/**
	 * Is toggling controlled by timer to enable blinking of the commandable in focus (or the next turn button).
	 * Non-persistent. 
	 */
	private BlinkingState blinkingState = new BlinkingState();

	private OrderTemplate targetSelectionOrder = null;

	private boolean wasInitialized = false;

	private int step = 0;

	public EmpireMainWindow(final ApplicationEnvironment appEnv) throws IOException {
		super(appEnv);
		logger.setLogLevel(Level.INFO);
		logger.setToShowThread(true);
		initColors();
		worldCanvas = new EmpireMainCanvas(this);
		worldScrollPane = new JScrollAndZoomPane(this, worldCanvas);
		worldScrollPane.setMinimumSize(worldCanvas.getMinimumSize());
		getContentPane().add(worldScrollPane, BorderLayout.CENTER);

		addComponentListener(new ComponentAdapter() {
			@Override
			public void componentResized(final ComponentEvent e) {
				centerScreen(false);
			}
		});

		startNewGameAction = new StartNewGameAction(this);
		exitGameAction = new ExitGameAction(this);
		statusBar = new JLabel("--");
		finishTurnButton = new JButton(finishTurnAction);
		nextButton = new JButton(nextCommandableAction);
		// Remove focus of buttons, Otherwise other keyboard events won't come through
		nextButton.setFocusable(false);
		finishTurnButton.setFocusable(false);

		final Container southButtons = new Container();
		southButtons.setLayout(new GridLayout());
		southButtons.add(nextButton);
		southButtons.add(finishTurnButton);

		Dimension endTurnBarSize = new Dimension(50, 20);
		finishTurnButton.setMinimumSize(endTurnBarSize);
		finishTurnButton.setPreferredSize(endTurnBarSize);
		final Container southBar = new Container();
		southBar.setLayout(new BorderLayout());
		southBar.add(southButtons, BorderLayout.EAST);
		southBar.add(statusBar, BorderLayout.CENTER);
		getContentPane().add(southBar, BorderLayout.SOUTH);

		for (int i = 0; i < viewZoomLevelActions.length; i++) {
			viewZoomLevelActions[i] = new ViewZoomLevelAction(i);
		}
		final ButtonGroup viewPerspectiveButtonGroup = new ButtonGroup();
		for (int i = 0; i < viewPerspectiveActions.length; i++) {
			viewPerspectiveActions[i] = new ViewPerspectiveAction(i, viewPerspectiveButtonGroup);
		}
		final ButtonGroup mapViewModeButtonGroup = new ButtonGroup();
		for (MapViewMode mode : MapViewMode.values()) {
			if (mode.isEnabled) {
				viewMapViewModeActions.add(new ViewMapViewModeAction(mode, mapViewModeButtonGroup));
			}
		}

		if (getApplicationEnvironment().isSilent()) {
			soundOnOffAction.setOn(false);
		}

		orderChangeProductionAction = new SelectOrderAction(OrderTemplate.TYPE_CHANGE_PRODUCTION);
		orderDefineMovePathAction = new SelectOrderAction(OrderTemplate.TYPE_CHANGE_MOVE_PATHS);
		orderHomeAction = new SelectOrderAction(OrderTemplate.TYPE_HOME);
		orderExploreAction = new SelectOrderAction(OrderTemplate.TYPE_EXPLORE);
		orderSkipMoveAction = new SelectOrderAction(OrderTemplate.TYPE_SKIP_MOVE);
		orderClearOrdersAction = new SelectOrderAction(OrderTemplate.TYPE_CLEAR_ORDERS);
		orderClearAlertsAction = new SelectOrderAction(OrderTemplate.TYPE_CLEAR_ALERTS);
		orderSentryAction = new SelectOrderAction(OrderTemplate.TYPE_SENTRY);
		orderLoadAction = new SelectOrderAction(OrderTemplate.TYPE_LOAD);
		orderUnloadAction = new SelectOrderAction(OrderTemplate.TYPE_UNLOAD);
		orderBuildAirbaseAction = new SelectOrderAction(OrderTemplate.TYPE_BUILD_AIRBASE);
		orderDisbandAction = new SelectOrderAction(OrderTemplate.TYPE_DISBAND);
		orderRenameAction = new SelectOrderAction(OrderTemplate.TYPE_RENAME);

		setJMenuBar(createMenuBar());
		pack(); // Resize the window to the preferred size of its components.
		final Messages messages = getMessages();
		switchLocale(messages.getCurrentLocale());

		setZoomLevel(0);
		final int playerId = getCurrentPlayerId();
		// This is a hack - because creation of model already updates status, but model is still null...
		notifyStatusMessageChanged(playerId); 
		notifyMapSizeChanged();
		notifyMapChanged();
		onOrderOptionsChanged();
		notifyMapViewModeChanged(playerId);
		redrawTurnButton();
		addKeyListener(this);
		pauseOrResume(true);
		logger.info(messages.getApplicationTitle() + " " + messages.getApplicationCopyrightMessage());
	}

	private void initColors() {
		PlayerColor.init(
				new PlayerColor[] {
						new PlayerColor(0, "GameProperties.Players.Color.0", "FF0000", "800000"), // red
						new PlayerColor(1, "GameProperties.Players.Color.1", "00FF00", "008000"), // green 
						new PlayerColor(2, "GameProperties.Players.Color.2", "00F0F0", "008080"), // cyan
						new PlayerColor(3, "GameProperties.Players.Color.3", "FF00FF", "600060"), // magenta
						new PlayerColor(4, "GameProperties.Players.Color.4", "8080FF", "000040"), // blue
						new PlayerColor(5, "GameProperties.Players.Color.5", "F0F000", "404000"), // yellow
						new PlayerColor(Player.ID_NONE, "GameProperties.Players.Color.None", "FFFFFF", "808080"), // white
				}
		);
	}

	/**
	 * Manual saving of a game state.
	 * @param action the save game action this was initiated from
	 */
	public void saveGame(final PedAction action) {
		boolean isToStart = false;
		final boolean wasRunning = pauseBegin();
		final ConfigurableFileFilter filter = createFileFilter();
		JFileChooser chooser = createFileChooser(filter);
		int returnVal = chooser.showSaveDialog(this);
		if (JFileChooser.APPROVE_OPTION == returnVal) {
			final File file = filter.correctExtensionOfSelectedFile(chooser.getSelectedFile());
			if (isOutputFileValid(action, file)) {
				gameFacadeProxy.saveGame(this, file);
			}
		}
		pauseEnd(wasRunning, isToStart);
	}

	/**
	 * Presents a yes/no dialog to the user, asking her/him if a file save action really should be executed.
	 * 
	 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	protected class FileOverwriteSafetyQuestionDialog extends JInfoDialog {

		private static final long serialVersionUID = 1L;

		public FileOverwriteSafetyQuestionDialog(final ApplicationMainWindow parent, final PedAction action, final File file) {
			super(parent);
			setTitle(getMessages().getText("Action.SaveGame.Overwrite" + Messages.TEXT_POSTFIX));
			setOptionPaneParams(
					parent.getMessages().getString(
							"Action.SaveGame.Overwrite.SafetyQuestion", null,
							new String[] { file.getName() }), 
					JOptionPane.QUESTION_MESSAGE);
			setOptionType(JOptionPane.OK_CANCEL_OPTION);
			finishLayout();
		}
	}
	
	protected boolean isOutputFileValid(final PedAction action, final File file) {
		boolean isValid = false;
		if (file.exists()) {
			final JInfoDialog dialog = new FileOverwriteSafetyQuestionDialog(this, action, file);
			dialog.enterDialog();
			isValid = dialog.wasAcceptedDialogChanges();
		} else {
			isValid = true;
		}
		return isValid;
	}

	public void loadGame() {
		boolean isToStart = false;
		final boolean wasRunning = pauseBegin();
		final ConfigurableFileFilter filter = createFileFilter();
		JFileChooser chooser = createFileChooser(filter);
		int returnVal = chooser.showOpenDialog(this);
		if (JFileChooser.APPROVE_OPTION == returnVal) {
			final File file = chooser.getSelectedFile();
			logger.info("You chose to load this file: " + file.getName());
			gameFacadeProxy.loadGame(this, file);
			setCurrentPlayerId(gameFacadeProxy.getCurrentPlayerId());
			onGameChanged();
		}
		pauseEnd(wasRunning, isToStart);
	}

	private JFileChooser createFileChooser(final ConfigurableFileFilter filter) {
		JFileChooser chooser = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			chooser = new JFileChooser();
			chooser.setFileFilter(filter);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		return chooser;
	}

	private ConfigurableFileFilter createFileFilter() {
		final ConfigurableFileFilter filter = new ConfigurableFileFilter();
		filter.setAllowedExtensions(new String[] { GameFacade.SAVED_GAME_FILE_EXTENSION });
		filter.setDescription("Empire saved games (*." + GameFacade.SAVED_GAME_FILE_EXTENSION + ")");
		return filter;
	}

	/** 
	 * Centers the screen on the commandable in focus, if there is any.
	 * @param isAlwaysCentering if <code>true</code>, then the screen is centered to the commandable, even if the commandable was visible before. 
	 *   If <code>false</code>, the viewport is changed with the least effort to get the commandable on the screen.
	 */
	public void centerScreen(boolean isAlwaysCentering) {
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			final PlayfieldPosition pos = gameFacadeProxy.getPositionInFocus(playerId);
			final PlayfieldDimension size = gameFacadeProxy.getGameProperties().getSize();
			centerScrollbar(pos, size, PlayfieldPosition.X, worldScrollPane.getHorizontalScrollBar(), isAlwaysCentering);
			centerScrollbar(pos, size, PlayfieldPosition.Y, worldScrollPane.getVerticalScrollBar(), isAlwaysCentering);
		}
	}

	private void centerScrollbar(final PlayfieldPosition pos, final PlayfieldDimension size, final int dimension, final JScrollBar bar, final boolean isAlwaysCentering) {
		final int barSize = bar.getMaximum() - bar.getMinimum();
		float val = ((float) (pos.get(dimension)) / (float) (size.get(dimension)));
		final TopologyMode topologyMode = props.getTopologyMode();
		switch(topologyMode) {
		case HEXAGON:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				// Nothing to do
			} else {
				if (PlayfieldPosition.Y == dimension) {
					final float x = ((float) (pos.getX()) / (float) (size.getWidth()));
					val -= x / 2.0f;
				}
			}
			break;
		case HEXAGON_H:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				// Nothing to do
			} else {
				if (PlayfieldPosition.X == dimension) {
					final float y = ((float) (pos.getY()) / (float) (size.getHeight()));
					val -= y / 2.0f;
				}
			}
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			// Nothing to do
		}

		final int ival = (int)(val * barSize) - bar.getVisibleAmount() / 2;
		final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[zoomLevelIdx];
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(topologyMode);
		int value = Integer.MIN_VALUE;
		final int spaceDimension = playfieldTiling.getSpaceDimension(dimension);
		if (isAlwaysCentering) {
			final int ixb = MathUtil.bound(ival, 0, barSize);
			value = bar.getMinimum() + ixb;
		} else {
			final int barValue = bar.getValue() - bar.getMinimum();
			final int ldiff = ival - (barValue - (bar.getVisibleAmount() / 2 - 2 * spaceDimension));
			logger.trace("centerScrollbar " + (dimension == 0 ? "left" : "up") + ": " + ival + " - " + barValue + "-" + bar.getVisibleAmount() / 2 + " --> " + ldiff + (ldiff < 0 ? "outbound" : "inbound"));
			if (ldiff < 0) {
				value = bar.getMinimum() + MathUtil.bound(barValue + ldiff, 0, barSize);
			} else {
				final int rdiff = ival - (barValue + (bar.getVisibleAmount() / 2 - 2 * spaceDimension - playfieldTiling.getSurroundingDimension(dimension)));
				logger.trace("centerScrollbar " + (dimension == 0 ? "right" : "down") + ": " + ival + " - " + barValue + "+" + bar.getVisibleAmount() / 2 + " --> " + rdiff + (rdiff > 0 ? "outbound" : "inbound"));
				if (0 < rdiff) {
					value = bar.getMinimum() + MathUtil.bound(barValue + rdiff, 0, barSize);
				} else {
					// Nothing to do, we are already inbound
				}
			}
		}
		logger.trace("centerScrollbar(" + pos.toString() + "," + size.toString() + "," + dimension + ","
				+ isAlwaysCentering + ") --> " + value + "(" + value / (float) spaceDimension + ")");
		if (Integer.MIN_VALUE != value) {
			bar.setValue(value);
		}
	}

	private void onGameChanged() {
		final SelectOrderAction[] orderActions = new SelectOrderAction[] {
				orderChangeProductionAction,
				orderDefineMovePathAction,
				orderSkipMoveAction,
				orderClearOrdersAction,
				orderClearAlertsAction,
				orderDisbandAction,
				orderSentryAction,
				orderLoadAction,
				orderUnloadAction,
				orderBuildAirbaseAction,
				orderRenameAction,
				orderHomeAction,
				orderExploreAction,
		};
		final GameState gameState = gameFacadeProxy.getGameState();
		for (final SelectOrderAction orderAction : orderActions) {
			orderAction.order.setGameState(gameState);
		}
		updateAllActionsEnabled();
		updatePerspectivesSubmenuContent();
	    updateZoomLevelSubmenuSelected();
		final int playerId = getCurrentPlayerId();
		updateLocaleIndex(playerId);
		updateZoomLevel(playerId);
		notifyMapSizeChanged();
		notifyMapChanged();
		notifyOrderOptionsChanged();
		redrawTurnButton();
		notifyMapViewModeChanged(playerId);
		notifyStatusMessageChanged(playerId);
		centerScreen(true);
		repaint();
		handleTitleStateTransition(TitleStateTransition.STOP);
	}

	private void updateAllActionsEnabled() {
		for (final Action action : allActions) {
			if (action instanceof PedAction) {
				final PedAction a = (PedAction) action;
				a.updateEnabled();
			}
		}
	}

	private void updatePerspectivesSubmenuContent() {
		for (final ViewPerspectiveAction action : viewPerspectiveActions) {
			allActions.remove(action);
		}
		perspectivesSubmenu.removeAll();
		List<ViewPerspectiveAction> views = new LinkedList<ViewPerspectiveAction>();
		final ButtonGroup viewPerspectiveButtonGroup = new ButtonGroup();
		for (int i = 0; i < 7; i++) {
			ViewPerspectiveAction action = new ViewPerspectiveAction(i, viewPerspectiveButtonGroup);
			if (action.isToBeDefined()) {
				action.updateEnabled();
				views.add(action);
			}
		}
		final ViewPerspectiveAction[] template = new ViewPerspectiveAction[views.size()];
		viewPerspectiveActions = views.toArray(template);
		for (final ViewPerspectiveAction action : viewPerspectiveActions) {
			perspectivesSubmenu.add(action.button);
			putValues(action);
			allActions.add(action);
		}
		updatePerspectiveSubmenuSelected();
	}

	@Override
	public void startNewGame() {
		boolean isToStart = false;
		final boolean wasRunning = pauseBegin();
		final Messages messages = getMessages();
		final NewGameSelectionWizard wizard = new NewGameSelectionWizard(this,
				messages.getText("GamePropertiesWindow.Title"), true);
		wizard.enterDialog();

		logger.debug("New game selection dialog " + (wizard.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (wizard.wasAcceptedDialogChanges()) {
			boolean wasOk = false;
			final int tries;
			try {
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				if (wizard.isTutorial()) {
					tries = gameFacadeProxy.createGame(this, globalLocaleIndex, null, GameFacade.TUTORIAL_LEVEL_DEFAULT);
				} else {
					wizard.readUIFieldsIntoProperties();
					props = wizard.getModel();
					tries = gameFacadeProxy.createGame(this, globalLocaleIndex, props, GameFacade.TUTORIAL_LEVEL_NONE);
				}
				wasOk = (GameFacade.FAILED != tries);
			} finally {
				setCursor(Cursor.getDefaultCursor());
			}
			if (wasOk) {
				targetSelectionOrder = null;
				// Finding the game views might be more complicated...
				setCurrentPlayerId(gameFacadeProxy.getCurrentPlayerId());
				onGameChanged();
				isToStart = true;
			} else {
				final PedJDialog t = new JInfoDialog(this,
						messages.getString("GamePropertiesWindow.UnableToCreateWorld.Title"),
						messages.getString("GamePropertiesWindow.UnableToCreateWorld.Text"), JOptionPane.ERROR_MESSAGE);
				t.setVisible(true);
			}
		}
		pauseEnd(wasRunning, isToStart);
	}

	public int getZoomLevel() {
		return zoomLevelIdx;
	}

	private void updateZoomLevel(final int playerId) {
		if (Player.ID_NONE != playerId) {
			zoomLevelIdx = gameFacadeProxy.getGameViewProperties(playerId).getZoomLevel();
			notifyMapSizeChanged();
			notifyMapChanged();
			hasToCenterScreen = true;
		}
	}

	/** 
	 * Sets the zoom level of the current player to the given index. 
	 * Can be safely called by a controller to change the value for the current player.
	 * @param index the zoom level index
	 */
	private void setZoomLevel(final int index) {
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			gameFacadeProxy.setZoomLevel(playerId, index);
			updateZoomLevel(playerId);
		}
	}

	/** 
	 * Zooms in or out for the current player.
	 * Can be safely called by a controller to change the value for the current player.
	 * @param amount the amount by which to change the zoom level - typically -1 or +1.
	 */
	@Override
	public void changeZoomLevel(final int amount) {
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			final int oldIndex = gameFacadeProxy.getGameViewProperties(playerId).getZoomLevel();
			final int index = MathUtil.bound(oldIndex + amount, 0, ZoomLevel.INSTANCES.length - 1);
			gameFacadeProxy.setZoomLevel(playerId, index);
			updateZoomLevel(playerId);
			updateZoomLevelSubmenuSelected();
		}
	}

	private void updateLocaleIndex(final int playerId) {
		if (Player.ID_NONE != playerId) {
			final int localeIndex = gameFacadeProxy.getGameViewProperties(playerId).getLocaleIndex();
			switchLocaleReally(getMessages().getAvailableLocales()[localeIndex]);
			logger.trace("updateLocaleIndex(" + playerId + ") --> localeIndex " + localeIndex);
			onOrderOptionsChanged();
			notifyStatusMessageChanged(playerId);
		} else {
			final int localeIndex = globalLocaleIndex;
			switchLocaleReally(getMessages().getAvailableLocales()[localeIndex]);
			logger.trace("updateLocaleIndex(" + playerId + ") --> localeIndex " + localeIndex);
		}
	}

	private JMenuBar createMenuBar() {
		final JMenuBar menuBar = createJMenuBar();
		final JMenu menuFile = createJMenu("Menu.File");
		createActionItem(menuFile, startNewGameAction);
		createActionItem(menuFile, loadGameAction);
		createActionItem(menuFile, saveGameAction);
		createActionItem(menuFile, resignPlayerAction);
		createActionItem(menuFile, exitGameAction);
		menuFile.addSeparator();
		menuFile.add(createLocalesMenu());
		menuFile.add(createLAFMenu());

		menuFile.addSeparator();
		createActionItem(menuFile, soundOnOffAction);
		menuFile.addSeparator();
		createActionItem(menuFile, quitAction);
		menuBar.add(menuFile);

		final JMenu menuView = createJMenu("Menu.View");
		final PedJMenu zoomLevelsSubmenu = createJMenu("Menu.ViewZoomLevel", ResourceFinder.Folder.COMMON);
		final ButtonGroup zoomButtonGroup = new ButtonGroup();
		for (final ViewZoomLevelAction action : viewZoomLevelActions) {
			zoomLevelButtons[action.index] = createRadioButtonActionItem(zoomLevelsSubmenu, action, zoomButtonGroup);
		}
		perspectivesSubmenu = createJMenu("Menu.ViewPerspective", Folder.APP);
		updatePerspectivesSubmenuContent();
		createActionItem(menuView, viewCenterAction);
		menuView.add(zoomLevelsSubmenu);
		menuView.add(perspectivesSubmenu);
		menuView.addSeparator();
		for (final ViewMapViewModeAction action : viewMapViewModeActions) {
			menuView.add(action.button);
		}
		menuView.addSeparator();
		createActionItem(menuView, viewContextMenuAction);
		createActionItem(menuView, viewStatusReportAction);
		createActionItem(menuView, viewLogReportAction);
		createActionItem(menuView, viewGamePropertiesAction);
		createActionItem(menuView, viewPlayerPropertiesAction);
		menuBar.add(menuView);

		final JMenu menuRun = createJMenu("Menu.Run");
		createActionItem(menuRun, orderChangeProductionAction);
		createActionItem(menuRun, orderDefineMovePathAction);
//		createActionItem(menuRun, orderDefineMovePathTargetAction);
		menuRun.addSeparator();
		createActionItem(menuRun, orderGoToTargetAction);
		createActionItem(menuRun, orderEscortTargetAction);
		createActionItem(menuRun, orderHomeAction);
		createActionItem(menuRun, orderExploreAction);
		createActionItem(menuRun, orderSentryAction);
		createActionItem(menuRun, orderLoadAction);
		createActionItem(menuRun, orderUnloadAction);
		createActionItem(menuRun, orderBuildAirbaseAction);
		createActionItem(menuRun, orderClearOrdersAction);
		createActionItem(menuRun, orderClearAlertsAction);
		createActionItem(menuRun, orderDisbandAction);
		createActionItem(menuRun, orderRenameAction);
		menuRun.addSeparator();
		createActionItem(menuRun, nextCommandableAction);
		createActionItem(menuRun, orderSkipMoveAction);
		menuRun.addSeparator();
		createActionItem(menuRun, finishTurnAction);
		menuBar.add(menuRun);
		final JMenu menuHelp = createJMenu("Menu.Help");
		createActionItem(menuHelp, helpAction);
		createActionItem(menuHelp, aboutAction);
		menuBar.add(menuHelp);
		try {
			menuBar.setHelpMenu(menuHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}
		updateZoomLevelSubmenuSelected();
		return menuBar;
	}

	private void updateZoomLevelSubmenuSelected() {
		final GameViewProperties props = gameFacadeProxy.getGameViewProperties(getCurrentPlayerId());
		int selectedIndex = -1;
		if (null != props) {
			selectedIndex = props.getZoomLevel();
		}
		for (int i = 0; i < zoomLevelButtons.length; i++) {
			zoomLevelButtons[i].setSelected(i == selectedIndex);
		}
	}

	public GameFacade getGameFacade() {
		return gameFacadeProxy.getGameFacade();
	}

	public GameFacadeProxy getGameFacadeProxy() {
		return gameFacadeProxy;
	}

	/** 
	 * @return if the game is still running.
	 * TODO: Don't ask the game facade every time. Collect it once or via notification. 
	 */
	boolean isGameRunning() {
		return gameFacadeProxy.isGameRunning();
	}

	/**
	 * @return the current game properties (or the default settings, if there was no previous game)
	 */
	GameProperties getGameProperties() {
		final GameProperties serverProps = gameFacadeProxy.getGameProperties();
		if (null != serverProps) {
			props = (GameProperties) serverProps.clone();
		}
		return props;
	}

	public static final int ACTION_MASK_FINISH_TURN = 0x10;

	@Override
	public void switchLocale(final Locale locale) {
		super.switchLocale(locale);
		final Messages messages = getMessages();
		final Locale currentLocale = messages.getCurrentLocale();
		final Locale[] availableLocales = messages.getAvailableLocales();
		final int idx = CollectionUtil.indexOf(availableLocales, currentLocale);
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			gameFacadeProxy.setLocaleIndex(playerId, idx);
		} else {
			globalLocaleIndex = idx;
		}
		updateLocaleIndex(playerId);
		repaint();
	}

	public void switchLocaleReally(final Locale locale) {
		super.switchLocale(locale);
		final Messages messages = getMessages();
		// Add all input possibilities not handled by menu actions
		userGameActionInputs = new UserGameActionInput[] {
			new UserGameActionInput(messages, "Game.Player0L"  + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD4, 0, DigitalJoystickDirection.MASK_LEFT),
			new UserGameActionInput(messages, "Game.Player0R"  + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD6, 0, DigitalJoystickDirection.MASK_RIGHT),
			new UserGameActionInput(messages, "Game.Player0U"  + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD8, 0, DigitalJoystickDirection.MASK_UP),
			new UserGameActionInput(messages, "Game.Player0D"  + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD2, 0, DigitalJoystickDirection.MASK_DOWN),
			new UserGameActionInput(messages, "Game.Player0UL" + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD7, 0, DigitalJoystickDirection.MASK_UP_LEFT),
			new UserGameActionInput(messages, "Game.Player0UR" + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD9, 0, DigitalJoystickDirection.MASK_UP_RIGHT),
			new UserGameActionInput(messages, "Game.Player0DL" + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD1, 0, DigitalJoystickDirection.MASK_DOWN_LEFT),
			new UserGameActionInput(messages, "Game.Player0DR" + Messages.KEYSTROKE_POSTFIX, KeyEvent.VK_NUMPAD3, 0, DigitalJoystickDirection.MASK_DOWN_RIGHT),
		};
		repaint();
	}

	private void suggestUserGameAction(final int playerAction) {
		logger.trace("User game action " + playerAction);
		final int playerId = getCurrentPlayerId();
		if (gameFacadeProxy.isAcceptingUserInput(playerId)) {
			if (EmpireMainWindow.ACTION_MASK_FINISH_TURN == playerAction) {
				if (gameFacadeProxy.isFinishTurnEnabled(playerId)) {
					onActionFinishTurn(true);
				}
			} 
			if (gameFacadeProxy.getMapViewMode(playerId) == MapViewMode.NORMAL) {
				if (0 != (playerAction & DigitalJoystickDirection.MASK_ALL_DIRECTIONS)) {
					final Commandable commandableInFocus = gameFacadeProxy.getVisibleCommandableInFocus(playerId);
					if (null != commandableInFocus) {
						if (commandableInFocus instanceof Unit) {
							final Unit unit = (Unit) commandableInFocus;
							final PlayfieldPosition pos = unit.getPosition();
							final GameProperties props = gameFacadeProxy.getGameProperties();
							final DigitalJoystickDirection kdir = DigitalJoystickDirection.getDirection(playerAction);
							final DigitalJoystickDirection dir = props.translateUserInputDirection(props.getTopologyMode(), pos, kdir);
							if (null != dir) {
								final PlayfieldPosition dest = props.wrap(pos, dir);
								final Order order = gameFacadeProxy.getMoveOrAttackOrder(unit, dest);
								onActionSelect(playerId, order);
							}
						}
					}
				}
			} else { // mapViewMode != MapViewMode.NORMAL
				if (0 != (playerAction & DigitalJoystickDirection.MASK_ALL_DIRECTIONS)) {
					final DigitalJoystickDirection dir = DigitalJoystickDirection.getDirection(playerAction);
					if (null != dir) {
						gameFacadeProxy.setPositionInFocus(playerId, dir);
					}
				}
			}
		} else { // null != moveResultToShow
			logger.debug("Ignored user action during animation.");
		}
	}

	@Override
	public void keyPressed(final KeyEvent keyEvent) {
		try {
			logger.trace("Key pressed: " + KeyEvent.getKeyText(keyEvent.getKeyCode()));
			for (final UserGameActionInput tsv : userGameActionInputs) {
				if (keyEvent.getKeyCode() == tsv.keyCode) {
					suggestUserGameAction(tsv.playerAction);
					break;
				}
			}
		} catch (final Throwable ex) {
			logger.errorWithStackTrace("This should never happen!", ex);
		}
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		final Messages messages = getMessages();
		doc.setInventoryMaxDepth(4);
		doc.setTitle(getMessages().getString("HelpWindow.Title"));
		doc.setHeading("Help");
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");

		final NewGameSelectionDialog gameSelectionDialog = new NewGameSelectionDialog(this, false);
		buildPropertyDescriptions(doc, gameSelectionDialog);
		buildPropertyDescriptions(doc, 3, "GameProperties.Players", gameSelectionDialog.getTableProperties(), true);

		buildMenuDescriptions(doc, getJMenuBar());
		buildInputDescriptions(doc, userGameActionInputs, false);
		doc.addContentPart(2, "Help.Fields");
		for (final FieldTemplate template : Templates.instance().allGameMapFieldTemplates) {
			doc.addContentPart(3, template.messageKeyPrefix, true, messages.getText(template.messageKeyPrefix),
					template.imageResourceName);
		}
		final String[] rowKeys = new String[] {
				"Game.Unit.Type",
				"Game.Unit.TurnsToProduce",
				"Game.Unit.SightingRange",
				"Game.Unit.Cargo",
				"Game.Unit.MovesPerTurn",
				"Game.Unit.MovingRange",
				"Game.Unit.CanMoveOver",
				"Game.Unit.MaximumDamage",
				"Game.Unit.CanCoastalAttack",
				"Game.Unit.AttackStrength",
				"Game.Unit.DefenseStrength",
				"Game.Unit.ShotDamage",
		};
		final TextFormatter.ColumnInfo[] ciTemplate = new TextFormatter.ColumnInfo[0];
		final List<TextFormatter.ColumnInfo> rowInfosAL = new ArrayList<TextFormatter.ColumnInfo>(rowKeys.length);
		for (final String rowKey : rowKeys) {
			rowInfosAL.add(new TextFormatter.ColumnInfo(10, TextFormatter.LEFT, messages, rowKey, true));
		}

		final List<TextFormatter.ColumnInfo> columnInfosAL = new ArrayList<TextFormatter.ColumnInfo>(1 + RulesComplexity.values().length);
		columnInfosAL.add(new TextFormatter.ColumnInfo(40, TextFormatter.LEFT, ""));
		for (RulesComplexity rc : RulesComplexity.values()) {
			columnInfosAL.add(new TextFormatter.ColumnInfo(25, TextFormatter.CENTER, messages, RulesComplexity.KEY + "." + rc.name(), true));
		}
		final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true,
				TextFormatter.htmlAttribute(TextFormatter.CLASS, "unit"), 
				columnInfosAL.toArray(ciTemplate));
		buildPropertyDescriptions(doc, 2, "Help.Units", rowKeys, true);
		final List<CommandableTemplate> unitTypes = CommandableTemplate.allUnitsToShowStatsFor(RulesComplexity.ADVANCED);
		final String na = messages.getText("Game.NotApplicable");
		final String[] nas = new String[] {
				na,
				na,
				na,
				na,
				na,
				na,
				na,
				na,
				na,
				na,
				na,
				na,
		};
		for (final CommandableTemplate template : unitTypes) {
			doc.addContentPart(3, template.messageKeyPrefix, true, messages.getText(template.messageKeyPrefix),
					template.imageResourceName);
			final Object[][] tableContent = new Object[RulesComplexity.values().length][];
			for (RulesComplexity gameRulesComplexity : RulesComplexity.values()) {
				final Object[] values;
				final int hitsTillDeath = template.getHitsTillDeath(gameRulesComplexity);
				if (0 == hitsTillDeath) {
					values = nas;
				} else {
					final int stepsTillDeath = template.getStepsTillDeath(gameRulesComplexity);
					final int sightingRange = template.getSightingRange(gameRulesComplexity, MilitaryBranch.LAND_UNIT);
					final int sightingRangeAir = template.getSightingRange(gameRulesComplexity, MilitaryBranch.AIR_UNIT);
					final int stepsPerTurn = template.getMaximumStepsPerTurn(gameRulesComplexity);
					values = new Object[] {
							// "Game.Unit.Type",
							doc.htmlTextWithImg(messages.getText(template.militaryBranch.messageKey), template.militaryBranch.messageKey + ".png"),
							// "Game.Unit.TurnsToProduce",
							Integer.toString(template.turnsToProduce),
							// "Game.Unit.SightingRange",
							((0 < sightingRange) ?
									((sightingRange == sightingRangeAir) ?
											new I18NStringWithFillIns("Game.Unit.SightingRange.Standard.Text", "%0", 
													new String[] { Integer.toString(sightingRange) } ) :
											new I18NStringWithFillIns("Game.Unit.SightingRange.Complex.Text", "%0 (%1)", 
													new String[] { Integer.toString(sightingRange), Integer.toString(sightingRangeAir) } )) :
									messages.getText("Game.NotApplicable")),
							// "Game.Unit.Cargo",
							messages.getText(template.getCargoString(doc, gameRulesComplexity)),
							// "Game.Unit.MovesPerTurn",
							Integer.toString(stepsPerTurn),
							// "Game.Unit.MovingRange",
							((stepsTillDeath < 1000) ?  
									Integer.toString(stepsTillDeath) : 
									messages.getText("Game.NotApplicable")),
							// "Game.Unit.CanMoveOver",
							messages.getText(template.canMoveOver ? "Game.Yes" : "Game.No"),
							// "Game.Unit.MaximumDamage",
							Integer.toString(hitsTillDeath),
							// "Game.Unit.CanCoastalAttack",
							messages.getText(
								(template.militaryBranch == MilitaryBranch.SEA_UNIT ? 
									(template.canCoastalAttack(gameRulesComplexity) ? "Game.Yes" : "Game.No") : "Game.NotApplicable")),
							// "Game.Unit.AttackStrength",
							((0 < stepsPerTurn) ?
									Integer.toString(template.attack) :
									messages.getText("Game.NotApplicable")),
							// "Game.Unit.DefenseStrength",
							Integer.toString(template.defense),
							// "Game.Unit.ShotDamage",
							Integer.toString(template.shotDamage),
					};
				}
				tableContent[gameRulesComplexity.ordinal()] = values;
			}
			doc.beginTable(null, ti);
			for (int rowCnt = 0; rowCnt < tableContent[0].length; rowCnt++) {
				final ArrayList<TableCellInfo> cells = new ArrayList<TableCellInfo>(tableContent.length + 1);
				// Left heading
				cells.add(new TableCellInfo(new I18NStringWithFillIns(rowKeys[rowCnt], null, null), TextFormatter.ROW,
						TextFormatter.UNDEFINED, TextFormatter.SPAN_DEFAULT, TextFormatter.SPAN_DEFAULT));
				// Iterate rules complexities
				for (int i = 0; i < tableContent.length; i++) {
					final int rowspan;
					int colspan = TextFormatter.SPAN_DEFAULT;
					if ((tableContent[i] == nas)) {
						// merge all cells in a column containing a unit not applicable for this rules complexity
						if (0 < rowCnt) {
							rowspan = 0; // Indicates an already merged cell, so nothing to output
						} else {
							rowspan = nas.length;
						}
					} else {
						rowspan = TextFormatter.SPAN_DEFAULT;
						// Look for content equal for several rules complexities to merge cells in a row
						for (int j = i + 1; j < tableContent.length; j++) {
							if (tableContent[i][rowCnt].equals(tableContent[j][rowCnt])) {
								colspan++;
							}
						}
					}
					if (0 < rowspan) {
						final TableCellInfo tci = new TableCellInfo(
								messages.getString(null, "%0", new Object[] { tableContent[i][rowCnt] }), null,
								TextFormatter.UNDEFINED, colspan, rowspan);
						cells.add(tci);
					}
					if (1 < colspan) { // Skip merged cells
						i += (colspan - 1);
					}
				}
				doc.addTableRow(cells);
			}
			doc.endTable();
		}
		// Build help for player settings
		PedJDialog dialog = new GameViewPropertiesDialog(this, Player.ID_NONE);
		buildPropertyDescriptions(doc, 2, "GameViewProperties", dialog.getAllPropertyElements(), true);
		doc.endDocument();
	}

	@Override
	protected void buildKeyboardActionsDescription(final TextDocumentBuilder doc, final UserGameActionInput[] userGameActionInputs, final boolean isDifferentPerPlayer) {
		if (null != userGameActionInputs) {
			final Messages messages = getMessages();
			doc.addContentPart(3, "Help.KeyboardActions");
			final ArrayList<TextFormatter.ColumnInfo> cis = new ArrayList<TextFormatter.ColumnInfo>();
			if (isDifferentPerPlayer) {
				cis.add(new TextFormatter.ColumnInfo(10, TextFormatter.RIGHT, messages, "Game.Player"));
			}
			cis.add(new TextFormatter.ColumnInfo(40, TextFormatter.LEFT, messages, "Game.PlayerAction"));
			cis.add(new TextFormatter.ColumnInfo(20, TextFormatter.LEFT, messages, "Game.KeyStroke"));
			for (final TopologyMode tm : TOPOLOGY_MODES) {
				cis.add(new TextFormatter.ColumnInfo(10, TextFormatter.CENTER, messages, TopologyMode.KEY + "." + tm.name()));
			}
			final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true,
					TextFormatter.htmlAttribute(TextFormatter.CLASS, "keyboardActions inv"), 
					cis.toArray(CI_TEMPLATE));
			doc.beginTable("keys", ti);
			final GameProperties p = new GameProperties();
			for (final UserGameActionInput ugai : userGameActionInputs) {
				final ArrayList<TextFormatter.TableCellInfo> cs = new ArrayList<TextFormatter.TableCellInfo>(3);
				if (isDifferentPerPlayer) {
					cs.add(new TextFormatter.TableCellInfo(Integer.toString(ugai.playerIdx + 1)));
				}
				cs.add(new TextFormatter.TableCellInfo(getKeyboardActionDescriptionText(ugai)));
				cs.add(new TextFormatter.TableCellInfo(ugai.keyCodeString));
				for (final TopologyMode tm : TOPOLOGY_MODES) {
					p.setTopologyMode(tm);
					final String txt = (null == p.translateUserInputDirection(tm, new PlayfieldPosition(), 
							DigitalJoystickDirection.getDirection(ugai.playerAction)) ? "&nbsp;" : "x");
							cs.add(new TextFormatter.TableCellInfo(txt));
				}
				doc.addTableRow(cs);
			}
			doc.endTable();
		}
	}

	@Override
	protected String getKeyboardActionDescriptionText(final UserGameActionInput ugai) {
		return getMessages().getString(DigitalJoystickDirection.getDirection(ugai.playerAction).key);
	}

	@Override
	public void showAboutBox() {
		final Messages messages = getMessages();
		final TextDocumentBuilder doc = new TextDocumentBuilder(messages);
		doc.setTextDocumentType(TextDocumentType.HTML401_TRANSITIONAL);
		doc.setEmbedded(false);
		doc.setInventoryMaxDepth(0);
		doc.setTitle(messages.getApplicationName(I18NedOrNot.I18N));
		doc.beginDocument();
		doc.add(messages.getApplicationCopyrightMessage());
		doc.addContentPart(2, "Help.VersionHistory");
		doc.addContentPart(2, "Help.KnownProblems");
		doc.endDocument();
		HtmlPaneDialog d = new HtmlPaneDialog(this, doc.getTitle(), doc.result());
		d.enterDialog();
	}

	@Override
	public FinishableThread createRunThread() {
		final GameThread thread = new GameThread(this);
		final Logger threadLogger = Logger.getLogger(thread.getClass());
		threadLogger.setToShowThread(true);
		thread.setLogger(threadLogger);
		return thread;
	}

	@Override
	protected void pauseOrResume(final boolean isToResume) {
		pauseOrResumeRun(isToResume);
	}

	private boolean isCommandableInFocusPossiblyToPlayPromptSound = false;

	/** 
	 * Called by the GameThread.
	 * <p>
	 * In its current standalone implementation, this is the thread that is 
	 * able to do things not started by user interactions.
	 * So it summarizes things that in the full version will be run in two threads:
	 * The server thread and the client thread.
	 * Goal is to separate them in a way, that first the server does its job,
	 * and then the client can do followup actions.
	 * The client is allowed to call server methods any time, 
	 * but the server shall only deliver information by the GameNotification object.
	 */
	@Override
	public synchronized void step() {
		final int playerId = getCurrentPlayerId();
		if (isGameRunning()) {
			if (blinkingState.step(PST)) {
				onFocusBlinking();
			}
			// Hack, until client and server get clearly separated: 
			// Let the computer player a chance, as it will never be a visible game view for a client.
			gameFacadeProxy.stepIfNot(playerId, PST);
			logger.trace("after stepIfNot: playerId=" + playerId + ", state is " + isGameRunning());
			if ((Player.ID_NONE != playerId) && (!isGetReady()) && !isUnitLocked()) {
				final boolean isToChangeView = gameFacadeProxy.step(playerId, PST);
				logger.trace("after step: playerId=" + playerId + ", state is " + isGameRunning());
				if (isToChangeView) {
					currentPlayerInfo.setPlayerAndTurn(gameFacadeProxy.getCurrentPlayerId(), gameFacadeProxy.getTurn());
				}
			}
			redrawTurnButton();
			if (hasToRepaintMap) {
				hasToRepaintMap = false;
				logger.trace("Repaint!");
				repaint();
			}
			if (hasToCenterScreen) {
				hasToCenterScreen = false;
				centerScreen(true);
			}
			if (hasToChangeOrderOptions) {
				hasToChangeOrderOptions = false;
				onOrderOptionsChanged();
			}
			if (!wasInitialized) {
				logger.trace("step(): calling notifyCommandableInFocusChanged");
				notifyCommandableInFocusChanged(getCurrentPlayerId());
				wasInitialized = true;
			}
			if (Player.ID_NONE != getCurrentPlayerId()) {
				if (isCommandableInFocusPossiblyToPlayPromptSound) {
					// TODO: Should have been triggered by a notifyUnitProptsForInput()
					GameFacade.SoundToPlayData data = gameFacadeProxy.getSoundToPlay(getCurrentPlayerId());
					if (null != data) {
						final String filename = data.filename;
						if (null != filename) {
							playSound(filename);
						}
						isCommandableInFocusPossiblyToPlayPromptSound = false;
					}
				}
			}
		} else {
			playTitleMusic(step);
		}
		step++;
	}

	public static final ProcessorSpeedType PST = ProcessorSpeedType.FAST;

	/** 
	 * Called by the GameThread to determine how long to sleep.
	 * @return the number of milliseconds to sleep to the next call to step
	 */
	@Override
	public long getSleepDelay() {
		if (isGameRunning()) {
			return PST.msPerStep;
		} else {
			return 100L;
		}
	}

	private void onActionFinishTurn(final boolean isWithPotentialSaveGame) {
		try {
			final int playerId = getCurrentPlayerId();
			if (gameFacadeProxy.isPlayerAlive(playerId)) {
				if (gameFacadeProxy.isPlayerWinner(playerId)) {
					notifyGameEndedForPlayer(playerId, true);
				}
			} else {
				notifyGameEndedForPlayer(playerId, false);
			}
			gameFacadeProxy.autoSaveGame();
			final boolean isGameAlive = gameFacadeProxy.finishTurn(playerId);
			if (!isGameAlive) {
				exitGameWithoutAsking();
			}
		} catch (final IllegalStateException ex) {
			// Nothing to do
		}
	}

	public void onActionNext() {
		final int playerId = getCurrentPlayerId();
		gameFacadeProxy.nextCommandable(playerId);
		notifyMapChanged();
		notifyOrderOptionsChanged();
	}

	public void onActionSelect(final Commandable commandable) {
		logger.trace("onActionSelect: " + getMessages().getString(commandable.getName()));
		final int playerId = getCurrentPlayerId();
		gameFacadeProxy.nextCommandable(playerId, commandable);
	}

	/**
	 * Called when an order was selected for the unit in focus. 
	 * If needed, asks a safety question to the user.
	 * @param playerId the player who selected the action
	 * @param order the order to set for the unit in focus
	 */
	private void onActionSelect(final int playerId, final Order order) {
		if (gameFacadeProxy.onActionSelectOrderMakesSense(playerId, order)) {
			final String key = order.getSafetyQuestion();
			boolean isToExecute = true;
			if (null != key) {
				final OrderSafetyQuestionDialog dialog = new OrderSafetyQuestionDialog(this, order);
				dialog.enterDialog();
				if (!dialog.wasAcceptedDialogChanges()) {
					isToExecute = false;
				}
			}
			if (isToExecute) {
				logger.debug("Command set");
				gameFacadeProxy.setOrder(playerId, order);
			}
		}
	}

	/**
	 * Called when the user initiated setting a targeted order and now needs to define the target.
	 * @param orderModel the kind of order that the user wants to define
	 */
	public void onActionSelectTargetMode(final OrderTemplate orderModel) {
		logger.trace("onActionSelectTargetMode " + orderModel.messageKeyPrefix);
		setTargetSelectionOrder(orderModel);
	}

	/**
	 * Called whenever the user clicks onto a field. 
	 * @param pos the position of the field
	 */
	public void actionAtPosition(final PlayfieldPosition pos) {
		final int playerId = getCurrentPlayerId();
		if (!isBusy()) {
			final GameProperties props = gameFacadeProxy.getGameProperties();
			if ((null != props) && (props.isWithin(pos))) {
				final OrderTemplate targetSelection = getTargetSelectionOrder();
				if (null != targetSelection) {
					actionTargetSelectionAtPosition(playerId, targetSelection, pos);
				} else {
					actionAtPosition(playerId, pos);
				}
			}
		}
	}

	private void actionAtPosition(final int playerId, final PlayfieldPosition position) {
		logger.trace("actionAtPosition " + playerId + position.toString());
		final Messages messages = getMessages();
		gameFacadeProxy.setPositionInFocus(playerId, position);
		final GameFacade.UserMapTileContextOptions options = gameFacadeProxy.collectMapTileContextOptions(playerId, position);
		final List<Order> orderOptions = options.orders;
		final List<Commandable> commandableOptions = options.commandables;
		if (orderOptions.isEmpty() && commandableOptions.isEmpty()) {
			gameFacadeProxy.setStatusMessageToField(playerId, position);
		} else {
			boolean isStatusMessageSet = false;
			final JPopupMenu popup = new JPopupMenu();
			int cnt = 0;
			if (!orderOptions.isEmpty()) {
				final Tileable to = gameFacadeProxy.getOrderOptionsObject(playerId, position);
				isStatusMessageSet = true;
				gameFacadeProxy.setStatusMessageToToolTip(playerId, to);
				final String ts = getMessages().getString(to.getName()) + ":";
				final JMenuItem title = new JMenuItem(ts);
				title.setEnabled(false);
				title.setIcon(createImageIcon(playerId, to));
				popup.add(title);
				for (final Order order : orderOptions) {
					popup.add(new JMenuItem(new SelectOrderAction(order)));
					cnt++;
				}
			}
			if (!commandableOptions.isEmpty()) {
				if (0 < cnt) {
					popup.addSeparator();
				}
				// Add all possible actions starting on the clicked location
				final JMenuItem title = new JMenuItem(messages.getText("Action.SetFocusOn"));
				title.setEnabled(false);
				popup.add(title);
				for (final Commandable commandable : commandableOptions) {
					JMenuItem item = new JMenuItem(new SelectCommandableAction(commandable));
					item.setMnemonic(KeyEvent.VK_1 + cnt);
					item.setIcon(createImageIcon(playerId, commandable));
					cnt++;
					popup.add(item);
				}
			}
			if (!isStatusMessageSet) {
				gameFacadeProxy.setStatusMessageToField(playerId, position);
			}
			logger.debug("Showing popup");
			final Point point = worldCanvas.map2canvas(position, PlayfieldTiling.CENTER);
			popup.show(worldCanvas, point.x, point.y);
		}
	}

	private ImageIcon createImageIcon(final int playerId, final Tileable tileable) {
		final ImageKey imageKey;
		if (tileable instanceof Commandable) {
			imageKey = new ImageKey(ImageKey.TType.Unit, ((Commandable) tileable).getKey().getTemplateId());
		} else {
			imageKey = new ImageKey(tileable.getImageKey());
		}
		final GameProperties props = gameFacadeProxy.getGameProperties();
		final int colorId = props.getPlayers()[playerId].getPlayerColorId();
		imageKey.setBits(ImageKey.BA_PLAYER_COLOR, colorId);
		final ImageIcon icon = new ImageIcon(TileTemplate.createImage(imageKey.getAll(), ZoomLevel.MENU_ZOOM_LEVEL, TopologyMode.SQUARE8));
		return icon;
	}

	private void actionTargetSelectionAtPosition(final int playerId, final OrderTemplate targetSelection, final PlayfieldPosition pos) {
		Order order = null;
		try {
			order = gameFacadeProxy.actionTargetSelectionAtPosition(playerId, targetSelection, pos);
		} catch (final CircularReferenceException ex) {
			final Messages messages = getMessages();
			final String title = messages.getText("Action.GoTo.Message.CircularReference.Title");
			final String text = messages.getText("Action.GoTo.Message.CircularReference.Text");
			final JInfoDialog dialog = new JInfoDialog(this, title, text, JOptionPane.ERROR_MESSAGE);
			dialog.enterDialog();
		}
		if (null != order) {
			if (order.makesSense()) {
				logger.debug("actionAt: order made sense");
				onActionSelect(playerId, order);
			}
		}
		setTargetSelectionOrder(null);
	}

	private void notifyStatusMessageChanged(final int playerId) {
		if (null != statusBar) {
			final String text;
			if ((Player.ID_NONE != playerId) && (!isGetReady())) {
				text = gameFacadeProxy.getStatusMessage(playerId);
			} else {
				final GameProperties props = gameFacadeProxy.getGameProperties();
				if (isGameRunning() && (null != props)){
					final I18NStringWithFillIns i18n = new I18NStringWithFillIns("Game.Status.NotCurrentPlayer.Text", "...", 
							new String[] { props.getPlayers()[gameFacadeProxy.getCurrentPlayerId()].getName(), } );
					text = getMessages().getString(i18n);
				} else {
					text = "";
				}
			}
			statusBar.setText(text);
			statusBar.repaint();
			logger.trace("Setting status bar Player" + playerId + isGetReady() + " to: " + text);
		}
	}

	private void notifyMapViewModeChanged(final int playerId) {
		final MapViewMode mapViewMode = gameFacadeProxy.getMapViewMode(playerId);
		for (ViewMapViewModeAction action : viewMapViewModeActions) {
			action.updateEnabled();
			action.setSelected(mapViewMode == action.mode);
		}
		notifyMapChanged();
		notifyOrderOptionsChanged();
	}

	private void onOrderOptionsChanged() {
		updateAllActionsEnabled();
	}

	private void redrawTurnButton() {
		Color color = Color.white;
		String text = "--";
		final int currPlayerId = getCurrentPlayerId();
		if (Player.ID_NONE != currPlayerId) {
			final int turn = gameFacadeProxy.getTurn();
			int playerId;
			if (null != animationMoveResult) {
				playerId = animationMoveResult.getCommandable(Role.Active).getPlayerId();
			} else {
				playerId = gameFacadeProxy.getCurrentPlayerId();
			}
			if (playerId != Player.ID_NONE) {
				final GameProperties props = gameFacadeProxy.getGameProperties();
				final Player player = props.getPlayers()[playerId];
				color = player.getForeground();
				if (isCommandableInFocusBlinking() && (gameFacadeProxy.isFinishTurnEnabled(currPlayerId))) {
					color = player.getBackground();
				}
			}
			text = Integer.toString(turn);
		}
		finishTurnButton.setBackground(color);
		finishTurnButton.setText(text);
		finishTurnButton.setAlignmentX(CENTER_ALIGNMENT);
		finishTurnButton.repaint();
	}

	private boolean hasToRepaintMap = false;
	private boolean hasToCenterScreen = false;
	private boolean hasToChangeOrderOptions = false;

	private void notifyMapChanged() {
		hasToRepaintMap = true;
	}

	private void notifyOrderOptionsChanged() {
		hasToChangeOrderOptions = true;
		logger.trace("notifyOrderOptionsChanged" /* , new Exception("OrderOptions changed") */);
	}

	private void notifyCommandablePromptsForInput(final int playerId, final Commandable commandable) {
		if ((null != commandable) && (commandable.getKey().getTemplateId() != ImageKey.CMDL_ID_CITY)) {
			final GameViewProperties props = gameFacadeProxy.getGameViewProperties(playerId);
			if ((null != props) && (props.isPlayingSounds(SoundTypes.PROMPT.ordinal()))) {
				playSound(((Unit) commandable).getUnitTemplate().promptSoundFileName);
			}
		}
	}

	private boolean isBusy() {
		final int playerId = getCurrentPlayerId();
		return ((Player.ID_NONE == playerId) || !gameFacadeProxy.isAcceptingUserInput(playerId));
	}

	private void notifyUnitMoveStarted(final int playerId, final MoveResult moveResult) {
		animationMoveResult = moveResult;
		redrawTurnButton();
		notifyMapChanged();
		notifyOrderOptionsChanged();
		logger.trace("notifyUnitMoveStarted: Cursor set to wait");
		worldCanvas.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	}

	/** 
	 * Executes one unit move animation step. 
	 * @param playerId the player to execute the animation for
	 * @param moveResult the move result to show
	 * @param seq the animation part within the move to show
	 * @param step the step wihtin the animation part
	 * @return <code>true</code> if this was the final step of the whole move result
	 */
	private boolean notifyUnitMoveStep(final int playerId, final MoveResult moveResult, final AnimationPart seq, final int step) {
		boolean ret = false;
		notifyMapChanged();
		final Unit unit = moveResult.getUnit();
		final Commandable enemy = moveResult.getCommandable(Role.Affected);
		logger.trace("Unit move step: unit=" + unit + ", enemy=" + enemy + ", seq=" + seq + ", step=" + step);		
		if (0 == step) {
			CommandableTemplate enemyTemplate = null;
			if (null != enemy) {
				enemyTemplate = CommandableTemplate.findTemplateById(enemy.getKey().getTemplateId());
			}
			String soundFileName = null;
			switch(seq) {
			case FIGHT_INTRO:
				soundFileName = unit.getUnitTemplate().fightIntroSoundFileName;
				break;
			case FIGHT_HIT_OWN:
				soundFileName = unit.getUnitTemplate().fightHitSoundFileName;
				break;
			case FIGHT_HIT_ENEMY:
				soundFileName = enemyTemplate.fightHitSoundFileName;
				break;
			case DEATH_OWN:
				soundFileName = unit.getUnitTemplate().deathSoundFileName;
				break;
			case DEATH_ENEMY:
				soundFileName = enemyTemplate.deathSoundFileName;
				break;
			case MOVE:
				soundFileName = unit.getUnitTemplate().moveSoundFileName;
				break;
			case DOCK:
			case LOAD:
			case GARRISON:
			case SLOWED_DOWN:
			case FUEL_EMPTY:
				// Nothing to do
				break;
			case ALERT_IMPEDIMENT:
				soundFileName = unit.getUnitTemplate().alertImpedimentSoundFileName;
				break;
			case ALERT_SIGHTING:
				soundFileName = unit.getUnitTemplate().alertSightingSoundFileName;
				break;
			case END_ANIMATION:
				// We assume hard-coded that this animation step only needs one step.
				ret = true;
				break;
			}
			if ((null != soundFileName) && (null != seq.soundType)) {
				final GameViewProperties props = gameFacadeProxy.getGameViewProperties(playerId);
				if ((null != props) && props.isPlayingSounds(seq.soundType.ordinal())) {
					playSound(soundFileName);
				}
			}
		}
		notifyStatusMessageChanged(playerId);
		return ret;
	}

	private void notifyUnitMoveFinished() {
		notifyMapChanged();
		notifyOrderOptionsChanged();
		animationMoveResult = null;
		redrawTurnButton();
		worldCanvas.setCursor(Cursor.getDefaultCursor());
		logger.trace("notifyUnitMoveFinished: Cursor set to default");
		notifyOrderOptionsChanged();
	}

	private void notifyMapSizeChanged() {
		final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[zoomLevelIdx];
		final TopologyMode topologyMode = props.getTopologyMode();
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(topologyMode);
		worldScrollPane.getHorizontalScrollBar().setUnitIncrement(playfieldTiling.getSpaceWidth());
		worldScrollPane.getVerticalScrollBar().setUnitIncrement(playfieldTiling.getSpaceHeight());
		worldCanvas.notifyMapSizeChanged();
		worldScrollPane.repaint();
	}

	@Override
	public void notifyGameNotification(final int playerId, final GameNotification notif) {
		logger.trace("notifyGameNotification(" + playerId + "," + notif + ") started.");
		onGameNotification(playerId, notif);
		logger.trace("notifyGameNotification(" + playerId + "," + notif + ") finished.");
	}

	private ArrayList<GameNotification> notifications = new ArrayList<GameNotification>();

	private void onGameNotification(final int playerId, final GameNotification notif) {
		if (null != notif) {
			synchronized (notifications) {
				notifications.add(notif);
			}
			logger.trace("onGameNotification added for player " + playerId + " " + notif + ", now containing " + notifications.size() + " elements.");
		}
		if(SwingUtilities.isEventDispatchThread()){
			checkGameNotifications(playerId);
		} else {
			SwingUtilities.invokeLater(new Runnable() {
				@Override
				public void run() {
					checkGameNotifications(playerId);
				}
			});
		}
	}

	private synchronized void checkGameNotifications(final int playerId) {
		logger.trace("checkGameNotifications(" + playerId + ")");
		while (!notifications.isEmpty()) {
			GameNotification notification = null;
			synchronized (notifications) {
				if (!notifications.isEmpty()) {
					notification = notifications.get(0);
				}
			}
			if (null != notification) {
				handleGameNotification(playerId, notification);
			}
			synchronized (notifications) {
				if (!notifications.isEmpty()) {
					final GameNotification head = notifications.get(0);
					if (head.isEvent(GameNotification.MASK_PLAYER_IN_FOCUS)) {
						logger.trace("checkGameNotifications checks: " + playerId + " " + head);
					}
					if (head.isEmpty()) {
						notifications.remove(0);
					} else {
						// Notification could not be finished, e.g., because of pending dialog
						break;
					}
				}
			}
		}
		logger.trace("checkGameNotifications(" + playerId + ") finished");
	}

	private void handleGameNotification(final int playerId, final GameNotification notif) {
		logger.debug("handleGameNotification(playerId=" + playerId + ", " + notif + ")");
		if (notif.isEvent(GameNotification.MASK_PLAYER_IN_FOCUS) || notif.isEvent(GameNotification.MASK_NEW_GAME_STARTED)) {
			final boolean isNewGameStarted = notif.isEvent(GameNotification.MASK_NEW_GAME_STARTED);
			notif.clearEvent(GameNotification.MASK_NEW_GAME_STARTED);
			notif.clearEvent(GameNotification.MASK_PLAYER_IN_FOCUS);
			logger.trace("handleGameNotification with player change");
			notifyPlayerChanged(playerId, notif.getPlayerId(), isNewGameStarted);
			if (isWaitingForModal()) {
				return;
			}
		}
		if (notif.isEvent(GameNotification.MASK_POSITION_IN_FOCUS)) {
			notifyPositionInFocusChanged(playerId);
			notif.clearEvent(GameNotification.MASK_POSITION_IN_FOCUS);
			if (isWaitingForModal()) {
				return;
			}
		}
		if (notif.isEvent(GameNotification.MASK_MAP_MODE)) {
			notifyMapViewModeChanged(playerId);
			notif.clearEvent(GameNotification.MASK_MAP_MODE);
			if (isWaitingForModal()) {
				return;
			}
		}
		if (notif.isEvent(GameNotification.MASK_COMMANDABLE_IN_FOCUS) 
				|| notif.isEvent(GameNotification.MASK_UNIT_PRODUCED)) {
			boolean isCmdlFound = notifyCommandableInFocusChanged(playerId);
			if (isCmdlFound) {
				logger.trace("handleGameNotification(): clearing focus");
				notif.clearEvent(GameNotification.MASK_UNIT_PRODUCED);
				notif.clearEvent(GameNotification.MASK_COMMANDABLE_IN_FOCUS);
			} else if (notif.isEvent(GameNotification.MASK_COMMANDABLE_IN_FOCUS)) {
				logger.trace("handleGameNotification(): clearing focus for unknown unit");
				notif.clearEvent(GameNotification.MASK_COMMANDABLE_IN_FOCUS);
			}
			if (isWaitingForModal()) {
				return;
			}
		}
		final MoveResult moveResult = notif.getAnimationMoveResult();
		if ((null != moveResult) && (!notif.isEmpty())) {
			if (null == animationMoveResult) {
				notifyUnitMoveStarted(playerId, moveResult);
			}
			if (null != notif.getAnimationPart()) {
				notifyUnitMoveStep(playerId, moveResult, notif.getAnimationPart(), notif.getAnimationPartStep());
			}
		} else {
			if (null != animationMoveResult) {
				logger.trace("handleGameNotification(playerId=" + playerId + ", " + notif + "): moveResult was null");
				notifyUnitMoveFinished();
			}
		}
		final CommandableKey cmdlKeyPrompting = notif.getCommandablePrompting();
		if (Player.ID_NONE != playerId) {
			if (null != cmdlKeyPrompting) {
				final Commandable cmdl = gameFacadeProxy.findCommandable(playerId, cmdlKeyPrompting);
				notifyCommandablePromptsForInput(playerId, cmdl);
			}
		}
		if (notif.isEvent(GameNotification.MASK_RENAME_REQ)) {
			notif.clearEvent(GameNotification.MASK_RENAME_REQ);
			if (null != cmdlKeyPrompting) {
				final Commandable cmdl = gameFacadeProxy.findCommandable(playerId, cmdlKeyPrompting);
				showNamingDialog(playerId, cmdl, null);
				if (isWaitingForModal()) {
					return;
				}
			}
		}
		if (notif.isEvent(GameNotification.MASK_CHANGE_PROD_REQ) && (!notif.isAnimationRunning())) {
			notif.clearEvent(GameNotification.MASK_CHANGE_PROD_REQ);
			if (null != cmdlKeyPrompting) {
				final Commandable cmdl = gameFacadeProxy.findCommandable(playerId, cmdlKeyPrompting);
				Assert.assertNotNull(cmdl);
				showCityProductionDialog(playerId, (FieldCity) cmdl);
				if (isWaitingForModal()) {
					return;
				}
			}
		}
		// Animation and change production are coupled, so that the change production dialog is shown after the animation of the fight that conquered that city.
		notif.clearEvent(GameNotification.MASK_CHANGE_PROD_REQ);
		notif.cleanAnimationRunning();
		
		if (notif.isEvent(GameNotification.MASK_EDIT_MOVE_PATH_REQ)) {
			notif.clearEvent(GameNotification.MASK_EDIT_MOVE_PATH_REQ);
			if (null != cmdlKeyPrompting) {
				final Commandable cmdl = gameFacadeProxy.findCommandable(playerId, cmdlKeyPrompting);
				showMovePathsDialog(playerId, cmdl);
				if (isWaitingForModal()) {
					return;
				}
			}
		}
		if (notif.isEvent(GameNotification.MASK_ORDERS)) {
			notifyOrderOptionsChanged();
			notif.clearEvent(GameNotification.MASK_ORDERS);
		}
		if (notif.isEvent(GameNotification.MASK_STATUS_MESSAGE)) {
			notifyStatusMessageChanged(playerId);
			notif.clearEvent(GameNotification.MASK_STATUS_MESSAGE);
		}
		if (notif.isEvent(GameNotification.MASK_GAME_LOST)) {
			notifyGameEndedForPlayer(playerId, false);
			notif.clearEvent(GameNotification.MASK_GAME_LOST);
		}
		if (notif.isEvent(GameNotification.MASK_GAME_WON)) {
			notifyGameEndedForPlayer(playerId, true);
			notif.clearEvent(GameNotification.MASK_GAME_WON);
		}
		logger.trace("handleGameNotification(playerId=" + playerId + ", " + notif + ") finished.");
	}

	private void notifyGameEndedForPlayer(final int playerId, final boolean isWon) {
		logger.trace("notifyGameEndedForPlayer(" + playerId + "," + isWon + ")");
		showGameEndedForPlayerDialog(playerId, isWon);
	}

	private static class GameStateLocal implements Comparable<GameStateLocal>{
		public final int turn;
		public final int playerId;
		public final boolean hasShownPlayerGetReadyDialog;
		public GameStateLocal(final int turn, final int playerId, final boolean hasShownPlayerGetReadyDialog) {
			this.turn = turn;
			this.playerId = playerId;
			this.hasShownPlayerGetReadyDialog = hasShownPlayerGetReadyDialog;
		}
		@Override
		public int compareTo(final GameStateLocal o) {
			if (this.turn != o.turn) {
				return this.turn - o.turn;
			}
			if (this.playerId != o.playerId) {
				return this.playerId - o.playerId;
			}
			if (this.hasShownPlayerGetReadyDialog != o.hasShownPlayerGetReadyDialog) {
				return (this.hasShownPlayerGetReadyDialog ? 1 : 0) - (o.hasShownPlayerGetReadyDialog ? 1 : 0);
			}
			return 0;
		}
		@Override
		public String toString() {
			return "(" + turn + "," + playerId + "," + hasShownPlayerGetReadyDialog + ")";
		}
	}

	private GameStateLocal gameStateLocal = new GameStateLocal(-1, 0, false);
	
	private void notifyPlayerChanged(final int playerId, final int notifPlayerId, final boolean hasNewGameStarted) {
		final boolean isGameRunning = isGameRunning();
		int playerIdToSet = Player.ID_NONE;
		if ((isGameRunning) && (notifPlayerId != Player.ID_NONE)) {
			playerIdToSet = notifPlayerId;
		}
		final boolean hasViewChanged = setCurrentPlayerId(playerIdToSet);
		if (playerIdToSet != getCurrentPlayerId()) {
			worldCanvas.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		} else {
			worldCanvas.setCursor(Cursor.getDefaultCursor());
		}
		if (hasViewChanged) {
			updateLocaleIndex(getCurrentPlayerId());
		}
		finishTurnAction.updateEnabled();
		redrawTurnButton();
		final int currentPlayerId = getCurrentPlayerId();
		if (isGameRunning && (playerId == playerIdToSet)) {
			final GameStateLocal gsl = new GameStateLocal(gameFacadeProxy.getTurn(), playerIdToSet, false);
			logger.debug("notifyPlayerChanged(playerId=" + playerId 
					+ ", playerIdToSet=" + playerIdToSet 
					+ ", notifPlayerId=" + notifPlayerId 
					+ ", hasNewGameStarted=" + hasNewGameStarted 
					+ ", hasViewChanged=" + hasViewChanged 
					+ ", currentPlayerId=" + currentPlayerId
					+ ", gameStateLocal=" + gameStateLocal
					+ ", gsl=" + gsl
					+ ", compare=" + gameStateLocal.compareTo(gsl));
			if (hasNewGameStarted || ((playerIdToSet == currentPlayerId) && (gameStateLocal.compareTo(gsl) < 0))) {
				showPlayerGetReadyDialog();
			}
			gameStateLocal = gsl;
		}
		if (hasViewChanged) {
			updatePerspectiveSubmenuSelected();
		    updateZoomLevelSubmenuSelected();
			updateZoomLevel(playerId);
			centerScreen(true);
			notifyOrderOptionsChanged();
		}
	}

	private void updatePerspectiveSubmenuSelected() {
		int ret = 0;
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			ret = gameFacadeProxy.getPerspective(playerId);
		}
		for (final ViewPerspectiveAction action : viewPerspectiveActions) {
			action.button.setSelected(ret == action.index);
		}
	}

	private MoveResult animationMoveResult = null;

	private void notifyPositionInFocusChanged(final int playerId) {
		if ((Player.ID_NONE != playerId) && (playerId == getCurrentPlayerId())) {
			final PlayfieldPosition pos = gameFacadeProxy.getPositionInFocus(playerId);
			logger.trace("notifyPositionInFocusChanged: " + (null == pos ? pos : pos.toString() + " for game view Player" + playerId));
			blinkingState.reset();
			centerScreen(false);
			notifyMapChanged();
			notifyOrderOptionsChanged();
		}
	}

	private boolean notifyCommandableInFocusChanged(final int playerId) {
		logger.trace("notifyCommandableInFocusChanged(playerId=" + currentPlayerInfo.getPlayerId() + ") with current player being " + getCurrentPlayerId());
		boolean ret = false;
		if ((Player.ID_NONE != playerId) && (playerId == getCurrentPlayerId())) {
			isCommandableInFocusPossiblyToPlayPromptSound = true;
			final Commandable commandable = gameFacadeProxy.getVisibleCommandableInFocus(playerId);
			if (null != commandable) {
				if (commandable.getKey().getTemplateId() == ImageKey.CMDL_ID_CITY) {
					ret = true;
					final Commandable unitProduced = gameFacadeProxy.getUnitProduced(playerId, commandable);
					logger.trace("notifyCommandableInFocusChanged(playerId=" + currentPlayerInfo.getPlayerId() + ") --> cmdl=" + commandable.getNickName() + ", unit=" + unitProduced);
					playSound(null);
					if (null != unitProduced) {
						showUnitProducedDialog(playerId, (FieldCity) commandable, (Unit) unitProduced);
					}
				} else {
					logger.trace("notifyCommandableInFocusChanged(playerId=" + currentPlayerInfo.getPlayerId() + ") --> cmdl=" + commandable.getKey());
				}
			} else {
				logger.trace("notifyCommandableInFocusChanged(playerId=" + currentPlayerInfo.getPlayerId() + ") --> cmdl=null");
				// If animation is running, visible commandable in focus == null, but that result should be kept for afterwards.
				if (gameFacadeProxy.isFinishTurnEnabled(playerId)) {
					ret = true;
				}
				playSound(null);
			}
			notifyOrderOptionsChanged();
			setTargetSelectionOrder(null);
		}
		return ret;
	}

	@Override
	public void exitGame() {
		if (isGameRunning()) {
//	    	boolean wasRunning = isRunning();
//	    	pauseOrResume(false);
			boolean isToExecute = true;
			PedJDialog dialog = null;
			try {
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				dialog = new SafetyQuestionDialog(this, exitGameAction);
			} finally {
				setCursor(Cursor.getDefaultCursor());
			}
			dialog.enterDialog();
			if (!dialog.wasAcceptedDialogChanges()) {
				isToExecute = false;
			}
			if (isToExecute) {
				exitGameWithoutAsking();
			}
			// pauseOrResume(wasRunning);
		}
	}

	private void exitGameWithoutAsking() {
		step = 0;
		setCurrentPlayerId(Player.ID_NONE);
		gameFacadeProxy.exitGame(this);
		targetSelectionOrder = null;
		onGameChanged();
	}

	public void resignCurrentPlayer() {
//    	boolean wasRunning = isRunning();
//    	pauseOrResume(false);
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			boolean isToExecute = true;
			PedJDialog dialog = null;
			try {
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				dialog = new SafetyQuestionDialog(this, resignPlayerAction);
			} finally {
				setCursor(Cursor.getDefaultCursor());
			}
			dialog.enterDialog();
			if (!dialog.wasAcceptedDialogChanges()) {
				isToExecute = false;
			}
			if (isToExecute) {
				gameFacadeProxy.resign(playerId);
				onActionFinishTurn(false);
			}
//	        pauseOrResume(wasRunning);
		}
	}

	@Override
	public synchronized void notifyUnitProduced(final int playerId, final FieldCity city, final Unit unit) {
		logger.info("NotifyUnitProduced: playerId=" + playerId + ", city=" + city + ", unit=" + unit);
	}

	private void showUnitProducedDialog(final int playerId, final FieldCity city, final Unit unit) {
		logger.info("showUnitProducedDialog(" + playerId + "," + city.getNickName() + "," + unit.getKey() + ")");
		if ((Player.ID_NONE == playerId) || (playerId == getCurrentPlayerId())) {
			final GameViewProperties props = gameFacadeProxy.getGameViewProperties(playerId);
			if ((null != props) && (props.isShowingProductionReports(unit.getUnitTemplate().militaryBranch.ordinal()))) {
				worldCanvas.repaint();
				showNamingDialog(playerId, unit, city);
			}
			gameFacadeProxy.clearUnitProduced(playerId, city);
			gameFacadeProxy.nextCommandable(playerId);
		}
	}

	private void showNamingDialog(final int playerId, final Commandable commandable, final FieldCity producingCity) {
		if (isGameRunning()) {
//	    	boolean wasRunning = isRunning();
//	    	pauseOrResume(false);
			if (null == commandable) {
				throw new IllegalArgumentException("null commandable");
			}
			NamingDialog dialog = null;
			try {
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				dialog = new NamingDialog(this, commandable, producingCity,
						orderChangeProductionAction.getMessageKeyPrefix());
			} finally {
				setCursor(Cursor.getDefaultCursor());
			}
			dialog.enterDialog();
//	        pauseOrResume(wasRunning);
		}
	}

	private void showCityProductionDialog(final int playerId, final FieldCity city) {
//    	boolean wasRunning = isRunning();
//    	pauseOrResume(false);
		notifyPositionInFocusChanged(playerId);
		redrawTurnButton();
		worldCanvas.repaint();
		CityProductionDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = new CityProductionDialog(this, this, city);
			logger.trace("showCityProductionDialog(" + playerId + ", " + city.toString() + ")");
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.enterDialog();
	}

	private void unlockUnit(final NamingDialog dialog) {
		logger.info("UnitProducedDialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		final Commandable commandable = dialog.getCommandable();
		if (dialog.wasAcceptedDialogChanges()) {
			commandable.setNickName(dialog.getName());
		}
		gameFacadeProxy.setStatusMessageToToolTip(commandable.getPlayerId(), commandable);
		if (dialog.wasProductionButtonPressed()) {
			final int playerId = getCurrentPlayerId();
			showCityProductionDialog(playerId, dialog.getProducingCity());
		}
	}

	private void unlockUnit(final CityProductionDialog dialog) {
		logger.trace("CityDialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		gameFacadeProxy.setCityProduction(getCurrentPlayerId(), dialog.getCity().getKey(), dialog.getProduction(), dialog.wasAcceptedDialogChanges());
		repaint();
	}

	private void showPlayerGetReadyDialog() {
		final int playerId = getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
//	    	boolean wasRunning = isRunning();
//	    	pauseOrResume(false);
			PlayerGetReadyDialog dialog = null;
			try {
				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				dialog = new PlayerGetReadyDialog(this);
			} finally {
				setCursor(Cursor.getDefaultCursor());
			}
			try {
				notifyStatusMessageChanged(playerId);
				worldCanvas.repaint();
				logger.trace("showPlayerGetReadyDialog(" + playerId + ")");
				dialog.enterDialog();
			} finally {
				// Nothing left to do, remove sometimes
			}
			worldCanvas.repaint();
//	        pauseOrResume(wasRunning);
		}
	}

	private void showMovePathsDialog(final int playerId, final Commandable cmd) {
//    	boolean wasRunning = isRunning();
//    	pauseOrResume(false);
		notifyPositionInFocusChanged(playerId);
		redrawTurnButton();
		worldCanvas.repaint();
		MovePathsDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = new MovePathsDialog(this, cmd);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.enterDialog();
		// pauseOrResume(wasRunning);
	}

	private void showMovePathCircularReferenceMessageBox() {
		final Messages messages = getMessages();
		final String title = messages.getText("Action.ChangeMovePaths.Message.CircularReference.Title");
		final String message = messages.getText("Action.ChangeMovePaths.Message.CircularReference.Text");
		final JInfoDialog infoDialog = new JInfoDialog(this, title, message, JOptionPane.ERROR_MESSAGE);
		infoDialog.enterDialog();
	}

	public void showLogReportDialog(final int playerId) {
		// Stop the game
		final boolean wasRunning = pauseBegin();
		// Show the box
		LogReportDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = new LogReportDialog(this, this, playerId);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.enterDialog(playerId);
		logger.debug("LogReportDialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (dialog.wasAcceptedDialogChanges()) {
//        	wasRunning = true;
		} else {
			// Nothing to do
		}
		// Resume the game
		pauseEnd(wasRunning, false);
	}

	public void showGamePropertiesDialog() {
		// Stop the game
		final boolean wasRunning = pauseBegin();
		// Show the box
		NewGameSelectionDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = new NewGameSelectionDialog(this, false);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.enterDialog();
		if (dialog.wasAcceptedDialogChanges()) {
			// wasRunning = true;
		} else {
			// Nothing to do
		}
		// Resume the game
		pauseEnd(wasRunning, false);
	}

	private void showGameViewPropertiesDialog() {
		// Stop the game
		final boolean wasRunning = pauseBegin();
		// Show the box
		final int playerId = getCurrentPlayerId();
		GameViewPropertiesDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = new GameViewPropertiesDialog(this, playerId);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.enterDialog();
		if (dialog.wasAcceptedDialogChanges()) {
			dialog.readUIFieldsIntoProperties();
			gameFacadeProxy.setGameViewProperties(playerId, dialog.getGamesViewProperties());
		} else {
			// Nothing to do
		}
		// Resume the game
		pauseEnd(wasRunning, false);
	}

	/** @return if in target selection mode, the order a target shall be selected for; or <code>null</code> if not in target selection mode */
	public OrderTemplate getTargetSelectionOrder() {
		return targetSelectionOrder;
	}

	public void showStatusReportDialog(final int playerId) {
		// Stop the game
		final boolean wasRunning = pauseBegin();
		// Show the box
		StatusReportDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = new StatusReportDialog(this, this, playerId);
			logger.debug("StatusReportDialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.enterDialog();
		if (dialog.wasAcceptedDialogChanges()) {
			// wasRunning = true;
		} else {
			// Nothing to do
		}
		// Resume the game
		pauseEnd(wasRunning, false);
	}

	private void showGameEndedForPlayerDialog(final int playerId, final boolean isWon) {
		// boolean wasRunning = isRunning();
		// pauseOrResume(false);
		notifyPositionInFocusChanged(playerId);
		redrawTurnButton();
		worldCanvas.repaint();
		JDialog dialog = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			dialog = GameEndedForPlayerDialog.create(this, playerId, isWon);
			logger.trace("showGameEndedForPlayerDialog(" + playerId + ", " + isWon + ")");
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		dialog.setVisible(true);
	}

	@Override
	public void setWaitingForModal(final PedJDialog dialog, final boolean wasOpened) {
		logger.trace("setWaitingForModal(" + wasOpened + ", " + dialog.toString() + ")");
		if (!wasOpened) {
			if (isGetReady()) {
				final int playerId = getCurrentPlayerId();
				gameFacadeProxy.clearGetReady(playerId);
				notifyStatusMessageChanged(playerId);
				notifyOrderOptionsChanged();
			}
			if (isUnitLocked()) {
				if (dialog instanceof NamingDialog) {
					unlockUnit((NamingDialog) dialog);
				} else {
					unlockUnit((CityProductionDialog) dialog);
				}
			}
			if (dialog instanceof MovePathsDialog) {
				final MovePathsDialog movePathsDialog = (MovePathsDialog) dialog;
				final Commandable commandable = movePathsDialog.getCommandable();
				logger.debug("MovePathsDialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
				if (movePathsDialog.wasAcceptedDialogChanges()) {
					final boolean wasCircular = gameFacadeProxy.setMovePaths(commandable, movePathsDialog.getResult(), true);
					if (wasCircular) {
						showMovePathCircularReferenceMessageBox();
					}
				} else if (movePathsDialog.wasButtonCreateNewPathPressed()) {
					final boolean wasCircular = gameFacadeProxy.setMovePaths(commandable, movePathsDialog.getResult(), false);
					if (wasCircular) {
						showMovePathCircularReferenceMessageBox();
					}
					onActionSelectTargetMode(OrderTemplate.CHANGE_MOVE_PATHS);
				}
				final Order order = commandable.getOrder();
				if ((null != order) && (order instanceof OrderDefineMovePaths)) {
					gameFacadeProxy.setOrder(commandable, null);
				}
				gameFacadeProxy.nextCommandable(commandable.getPlayerId());
				repaint();
			}
		}
		super.setWaitingForModal(dialog, wasOpened);
	}

	protected void setTargetSelectionOrder(final OrderTemplate targetSelectionOrder) {
		final OrderTemplate oldValue = this.targetSelectionOrder;
		this.targetSelectionOrder = targetSelectionOrder;
		final int playerId = getCurrentPlayerId();
		if (null == targetSelectionOrder) {
			if (null != oldValue) {
				gameFacadeProxy.popStatusMessage(playerId);
			}
		} else {
			final I18NStringWithFillIns str = new I18NStringWithFillIns(targetSelectionOrder.messageKeyPrefix + ".Mode" + Messages.TEXT_POSTFIX, null, null);
			gameFacadeProxy.pushStatusMessage(playerId, str);
		}
	}

	public SelectOrderAction createSelectOrderAction(final Order order) {
		return new SelectOrderAction(order);
	}

	public void onFocusBlinking() {
		notifyMapChanged();
	}

	boolean isCommandableInFocusBlinking() {
		return blinkingState.isVisible();
	}

	/** 
	 * Returns the player that currently has the focus in this client. 
	 * May be different to the game view of the current player, e.g., in
	 * network games or with computer opponents.
	 * @return the player that currently has the focus in this client
	 */
	int getCurrentPlayerId() {
		int ret = currentPlayerInfo.getPlayerId();
		return ret;
	}

	private boolean setCurrentPlayerId(final int playerId) {
		boolean ret = false;
		final GameProperties props = gameFacadeProxy.getGameProperties();
		if ((Player.ID_NONE == playerId) || (props.getPlayers()[playerId].isHumanPlayer())) {
			final int turn = gameFacadeProxy.getTurn();
			if ((currentPlayerInfo.getPlayerId() != playerId) 
					|| (currentPlayerInfo.getTurn() != turn)) {
				currentPlayerInfo.setPlayerAndTurn(playerId, turn);
				notifyMapSizeChanged();
				logger.trace("setCurrentPlayerId(" + playerId + ") sets to "+ currentPlayerInfo.getPlayerId());
				ret = true;
			}
		}
		return ret;
	}

	private GameStateBasics currentPlayerInfo = new GameStateBasics();

	@Override
	public synchronized void step(final int playerId) {
		// Is called regularly from the server. May be used when the separation continues.
	}

	public boolean isSoundOn() {
		return soundOnOffAction.isOn();
	}

	long getStep() {
		return step;
	}

	public boolean isGetReady() {
		return (getDialogWaitingFor() instanceof PlayerGetReadyDialog);
	}

	public boolean isUnitLocked() {
		final PedJDialog dialog = getDialogWaitingFor();
		return (dialog instanceof CityProductionDialog) 
				|| (dialog instanceof NamingDialog);
	}

	public void playSound(final String filename) {
		if (isSoundOn()) {
			final SoundClip sound = sounds.findSound(filename);
			logger.trace("playing sound " + filename);
			if (null != sound) {
				try {
					SoundClip.stop(currentClips[0]);
					currentClips[0] = sound.play();
				} catch (final NullPointerException ex) {
					logger.warn("Unexpected exception playing sound for event " + filename, ex);
				} catch (final OutOfMemoryError ex) {
					logger.warn("Unexpected exception playing sound for event " + filename, ex);
				}
			}
		}
	}

	private EmpireSounds sounds = new EmpireSounds(null);

	@Override
	protected void handleTitleStateTransition(final TitleStateTransition transition) {
		logger.trace("handleTitleStateTransition(" + transition.toString() + ") started");
		switch (transition) {
		case START:
			break;
		case STOP:
			for (int i = 0; i < currentClips.length; i++) {
				SoundClip.stop(currentClips[i]);
				currentClips[i] = null;
			}
			break;
		case PAUSE:
			for (int i = 0; i < currentClips.length; i++) {
				SoundClip.pauseOrResume(currentClips[i], false);
			}
			break;
		case RESUME:
			for (int i = 0; i < currentClips.length; i++) {
				SoundClip.pauseOrResume(currentClips[i], true);
			}
			break;
		}
	}

	private void playTitleMusic(final int beat) {
		if (isSoundOn()) {
			// First, create all new clips. This may be a costly operation.
			final Clip[] nextClips = new Clip[currentClips.length];
			for (int i = 0; i < nextClips.length; i++) {
				nextClips[i] = sounds.createTitleSoundClip(i, beat);
			}
			// All at once, we start the clips that should be started at that beat.
			for (int i = 0; i < nextClips.length; i++) {
				if (null != nextClips[i]) {
					SoundClip.stop(currentClips[i]);
					currentClips[i] = nextClips[i];
					SoundClip.play(currentClips[i]);
				}
			}
		}
	}

	private Clip[] currentClips = new Clip[4];
}
