package de.ped.empire.gui;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.Clip;

import de.ped.empire.logic.Resources;
import de.ped.tools.MathUtil;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationSounds;
import de.ped.tools.sound.AbstractWave;
import de.ped.tools.sound.ConstantWave;
import de.ped.tools.sound.DynamicSoundMonoWaveCreator;
import de.ped.tools.sound.EnvelopeWave;
import de.ped.tools.sound.MonoWave;
import de.ped.tools.sound.MultipliedWave;
import de.ped.tools.sound.Note;
import de.ped.tools.sound.PinkNoiseWave;
import de.ped.tools.sound.RepeatingWave;
import de.ped.tools.sound.SawDownWave;
import de.ped.tools.sound.SoundClip;
import de.ped.tools.sound.SquareWave;
import de.ped.tools.sound.TriangleWave;
import de.ped.tools.sound.Tuning;

/**
 * Bundles all the sounds for Empire.
 * 
 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class EmpireSounds extends ApplicationSounds {
	public EmpireSounds(final ResourceFinder resources) {
		super(resources);
	}

	/**
	 * Returns the sound clip for the given key.
	 * @param key the key (e.g., event) to find the sound for. Accepts {@link String} objects.
	 * @return the sound clip for the given key, or <code>null</code> if there is no such sound.
	 */
	@Override
	public SoundClip findSound(final Object key) {
		final String filename = (String) key;
		SoundClip sound = eventSounds.get(filename);
		if ((null == sound) && (null != filename)) {
			logger.trace("Loading sound " + filename);
			sound = new SoundClip(Resources.getURLToResource(filename));
			eventSounds.put(filename, sound);
		}
		return sound;
	}

	@Override
	public void init() {
		// Not needed, as all event sounds are sound files, and sound files are loaded lazily.
	}

	@Override
	public AudioInputStream createSound(final Object key) {
		// Not needed, as all event sounds are sound files, and sound files are loaded lazily.
		return null;
	}
	
	Clip createTitleSoundClip(final int instrument, final int beat) {
		// Lazy initialization
		while (titleClips.size() < 4) {
			titleClips.add(new HashMap<String, Clip>());
		}
		Clip ret = null;
		final AbstractWave.Params params = new AbstractWave.Params(new DynamicSoundMonoWaveCreator(), 0.0, 0.1);
		final int melodyBeat = beat / 2;
		final int melodyIndex = melodyBeat % HARMONY.length;
		final int round = (melodyBeat / HARMONY.length) % ROUNDS.length;
		String noteDescription = null;
		DynamicSoundMonoWaveCreator.Sound sound = null;
		switch(instrument) {
		case INSTRUMENT_ID_MELODY:
		case INSTRUMENT_ID_BASS:
		case INSTRUMENT_ID_XYLO:
			if (0 == beat % 2) {
				noteDescription = HARMONY[melodyIndex][instrument];
				if (ROUNDS[round][instrument] && (null != noteDescription)) {
					ret = titleClips.get(instrument).get(noteDescription);
					if (null == ret) {
						String nextNoteDescription = HARMONY[(melodyIndex + 1) % HARMONY.length][instrument];
						sound = createTitleMelodySound(instrument, params, noteDescription, nextNoteDescription);
					}
				}
			}
			break;
		case INSTRUMENT_ID_SNARE:
			noteDescription = "x";
			if ('x' == SNARE_BEAT.charAt(MathUtil.modulo(beat, SNARE_BEAT.length()))) {
				ret = titleClips.get(instrument).get(noteDescription);
				if (null == ret) {
					sound = createTitleMelodySound(instrument, params, noteDescription, null);
				}
			}
			break;
		}
		if ((null == ret) && (null != sound)) {
			ret = SoundClip.soundClip(params.smwc, sound).createClip();
			titleClips.get(instrument).put(noteDescription, ret);
		}
		return ret;
	}

	private DynamicSoundMonoWaveCreator.Sound createTitleMelodySound(
			final int instrument,
			final AbstractWave.Params params,
			final String noteDescription,
			final String nextNoteDescription) {
		DynamicSoundMonoWaveCreator.Sound sound = null;
		final List<MonoWave> mults = new LinkedList<MonoWave>();
		final double noteDuration;
		switch (instrument) {
		case INSTRUMENT_ID_MELODY:
			noteDuration = 0.2;
			mults.add(new TriangleWave(new RepeatingWave.Params(params, 
					Tuning.defaultTuning().getFrequency(Note.get(noteDescription))))); // Tone
			mults.add(new EnvelopeWave(new EnvelopeWave.Parameters(params, 
					noteDuration * 0.1, noteDuration * 0.1, noteDuration * 0.6, noteDuration * 0.2, 
					0.0, 0.3, 0.2))); // Hull
			sound = new DynamicSoundMonoWaveCreator.Sound(new MultipliedWave(params, mults), noteDuration);
			break;
		case INSTRUMENT_ID_BASS:
			if (null == nextNoteDescription) {
				noteDuration = 0.6;
			} else {
				noteDuration = 0.2;
			}
			final AbstractWave.Params params2 = new AbstractWave.Params(params.smwc, 0.0, params.duration * 2.0);
			mults.add(new TriangleWave(new RepeatingWave.Params(params2, 
					Tuning.defaultTuning().getFrequency(Note.get(noteDescription))))); // Tone
			mults.add(new EnvelopeWave(new EnvelopeWave.Parameters(params2, 
					noteDuration * 0.1, noteDuration * 0.1, noteDuration * 0.6, noteDuration * 0.2, 
					0.0, 0.5, 0.4))); // Hull
			sound = new DynamicSoundMonoWaveCreator.Sound(new MultipliedWave(params2, mults), noteDuration);
			break;
		case INSTRUMENT_ID_XYLO:
			noteDuration = 0.1;
			final double freq = Tuning.defaultTuning().getFrequency(Note.get(noteDescription));
			mults.add(new SquareWave(new RepeatingWave.Params(params, freq))); // Tone
			mults.add(new EnvelopeWave(new EnvelopeWave.Parameters(params, 
					noteDuration * 0.0, noteDuration * 1.0, noteDuration * 0.0, noteDuration * 0.0, 
					0.0, 0.03, 0.0))); // 0.03
			sound = new DynamicSoundMonoWaveCreator.Sound(new MultipliedWave(params, mults), noteDuration);
			break;
		case INSTRUMENT_ID_SNARE:
			noteDuration = 0.06;
			mults.add(new PinkNoiseWave(params.smwc, new ConstantWave(100.0))); // Tone
			mults.add(new SawDownWave(new RepeatingWave.Params(params, 30.0))); // "Snare" modulation
			mults.add(new EnvelopeWave(new EnvelopeWave.Parameters(params, 
					noteDuration * 0.0, noteDuration * 0.1, noteDuration * 0.7, noteDuration * 0.2, 
					0.0, 0.4, 0.3))); // Hull
			sound = new DynamicSoundMonoWaveCreator.Sound(new MultipliedWave(params, mults), noteDuration);
			break;
		}
		return sound;
	}

	private static final int INSTRUMENT_ID_MELODY = 0;
	private static final int INSTRUMENT_ID_BASS = 1;
	private static final int INSTRUMENT_ID_XYLO = 2;
	private static final int INSTRUMENT_ID_SNARE = 3;

	private static String[][] HARMONY = {
		{"C2", "C0",  "G3", }, // Block 1 C-dur
		{"E2", null,  "C4", },
		{"D2", null,  "B3", },
		{"G1", null,  "A3", },
		{"C2", "C0",  "G3", },
		{null, null,  "E3", },
		{"G2", null,  "D3", },
		{null, null,  "E3", },
		{"C2", "C0",  "G3", }, // Block 2 C-dur
		{"E2", null,  "C4", },
		{"D2", null,  "B3", },
		{"G1", "C0",  "A3", },
		{"C2", "C0",  "G3", },
		{null, null,  "E3", },
		{"C2", null,  "D3", },
		{null, null,  "C3", },
		{"C2", "C0",  "G3", }, // Block 3 C-dur
		{"E2", null,  "C4", },
		{"D2", null,  "B3", },
		{"G1", null,  "A3", },
		{"C2", "C0",  "G3", },
		{null, null,  "E3", },
		{"G2", null,  "D3", },
		{null, null,  "E3", },
		{"C2", "C0",  "G3", }, // Block 4 C-dur
		{"E2", null,  "A3", },
		{"D2", null,  "B3", },
		{"G1", "B-1", "D4", },
		{"C2", "C0",  "G3", },
		{null, null,  null, },
		{null, null,  "C4", },
		{null, null,  "G3", },
		{"C2", "A-1", "E3", }, // Block 5 a-moll
		{"E2", null,  "F3", },
		{"D2", null,  "G3", },
		{"A1", null,  "A3", },
		{"C2", "E0",  "B3", },
		{null, null,  "A3", },
		{"G2", null,  "C4", },
		{null, "D0",  "A3", },
		{"C2", "A-1", "E3", }, // Block 6 a-moll
		{"E2", null,  "F3", },
		{"D2", null,  "G3", },
		{"A1", null,  "A3", },
		{"C2", "E0",  "B3", },
		{null, null,  "A3", },
		{"C2", null,  "E3", },
		{null, null,  "A3", },
		{"C2", "F0",  null, }, // Block 7 F-dur G-dur
		{"D2", null,  "A3", },
		{"E2", null,  null, },
		{"A1", "F0",  "C4", },
		{"C2", "G0",  null, },
		{null, null,  "B3", },
		{"G2", null,  null, },
		{null, null,  "D4", },
		{"C2", "G-1", null, }, // Block 8 G-dur C-dur
		{"D2", null,  "B3", },
		{"E2", null,  null, },
		{"G1", "A-1", "D4", },
		{"C2", "C0",  null, },
		{null, null,  "C4", },
		{null, null,  "C4", },
		{null, "G-1", "C4", },
	};

	private static boolean[][] ROUNDS = new boolean[][] {
			{ true, true, false, },
			{ true, false, true, },
			{ false, true, true, },
			{ true, true, true, },
	};

	private static String SNARE_BEAT = "x...x...x...x.xxx.x...x.x.x...x.x.x...x.x.x...x.x.";

	private ArrayList<HashMap<String, Clip>> titleClips = new ArrayList<HashMap<String, Clip>>();

	private HashMap<String, SoundClip> eventSounds = new HashMap<String, SoundClip>();
}
