package de.ped.empire.gui;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

import javax.swing.JCheckBox;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import de.ped.empire.logic.GameFacade;
import de.ped.empire.logic.GameViewProperties;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.SoundTypes;
import de.ped.empire.logic.GameViewProperties.AnimationDelayType;
import de.ped.empire.logic.GameViewProperties.AnimationDelayValue;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.JIconComboBox;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;

/**
 * A dialog to edit player view settings.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameViewPropertiesDialog extends PedJDialog {
	private static final long serialVersionUID = 1L;

	/**
	 * Constructor.
	 * 
	 * @param mainWindow
	 *            The main window of the application.
	 * @param playerId
	 *            The ID of the player to call this for. 
	 *            If {@link Player#ID_NONE}, this dialog will only be created as
	 *            stub to create the help text.
	 */
	public GameViewPropertiesDialog(final ApplicationMainWindow mainWindow, final int playerId) {
		super(mainWindow, mainWindow);
		final EmpireMainWindow emp = (EmpireMainWindow) mainWindow;
		final GameFacade gameFacade = emp.getGameFacade();
		final Messages messages = getMessages();
		if (Player.ID_NONE != playerId) {
			this.properties = (GameViewProperties) gameFacade.getGameViewProperties(playerId).clone();
			setTitle(messages.getString("Action.ViewPlayerProperties.Title.Text", null,
					new String[] { gameFacade.getGameProperties().getPlayers()[playerId].getName(), }));
		} else {
			this.properties = null;
			setTitle(messages.getString("Action.ViewPlayerProperties.Text"));
		}
		animationDelayComboBoxes = new Vector<JIconComboBox<String>>(peAnimationDelays.length);
		for (int i = 0; i < animationDelayComboBoxes.capacity(); i++) {
			JIconComboBox<String> box = new JIconComboBox<String>();
			for (AnimationDelayValue value : AnimationDelayValue.values()) {
				box.addItem(messages.getString(value.messageKey));
			}
			animationDelayComboBoxes.add(box);
			final AnimationDelayType animationDelayType = AnimationDelayType.values()[i];
			peAnimationDelays[i] = new PropertyElement(animationDelayType.messageKey, box, null, animationDelayType.imageKey, ResourceFinder.Folder.APP);
//			box.setEnabled(false);
		}

		// Main panel: Units to produce
		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);
		final List<PropertyElement> pes = new ArrayList<PropertyElement>();
		pes.add(new PropertyElement("GameViewProperties.ProductionReport", null, peProductionReports, null, ResourceFinder.Folder.APP));
		pes.add(new PropertyElement("GameViewProperties.Sounds", null, peSounds, "Action.SoundOnOff", ResourceFinder.Folder.COMMON));
		pes.add(new PropertyElement("GameViewProperties.AnimationDelay", null, peAnimationDelays, null, ResourceFinder.Folder.APP));
		pes.add(new PropertyElement("GameViewProperties.Graphics", null, new PropertyElement[] { peDistanceInfo }, "GameProperties.ContinentSize.Max", ResourceFinder.Folder.APP));

		add(mainPanel, BorderLayout.CENTER);
		setPropertyElements(pes.toArray(new PropertyElement[] {}));
		addAllPropertyElementsTo(mainPanel, c);
		getContentPane().add(new JScrollPane(mainPanel), BorderLayout.CENTER);

		if (Player.ID_NONE != playerId) {
			updateUIFieldsBasedOnProperties();
		}
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		addHelpButton("GameViewProperties");
		setDefaultButton(getButtonOk());
		finishLayout();
	}

	@Override
	public void updateUIFieldsBasedOnProperties() {
		for (int i = 0; i < checkboxProductionReports.length; i++) {
			checkboxProductionReports[i].setSelected(properties.isShowingProductionReports(i));
		}
		for (int i = 0; i < checkboxSounds.length; i++) {
			checkboxSounds[i].setSelected(properties.isPlayingSounds(i));
		}
		for (int i = 0; i < animationDelayComboBoxes.size(); i++) {
			animationDelayComboBoxes.get(i)
					.setSelectedIndex(properties.getAnimationDelay(AnimationDelayType.values()[i]).ordinal());
		}
		checkboxDistanceInfo.setSelected(properties.isShowingDistanceInfo());
	}

	@Override
	public void readUIFieldsIntoProperties() {
		for (int i = 0; i < checkboxProductionReports.length; i++) {
			properties.setShowingProductionReports(i, checkboxProductionReports[i].isSelected());
		}
		for (int i = 0; i < checkboxSounds.length; i++) {
			properties.setPlayingSounds(i, checkboxSounds[i].isSelected());
		}
		for (int i = 0; i < animationDelayComboBoxes.size(); i++) {
			properties.setAnimationDelay(AnimationDelayType.values()[i],
					AnimationDelayValue.values()[animationDelayComboBoxes.get(i).getSelectedIndex()]);
		}
		properties.setShowingDistanceInfo(checkboxDistanceInfo.isSelected());
	}

	public GameViewProperties getGamesViewProperties() {
		return properties;
	}

	private final JCheckBox[] checkboxProductionReports = new JCheckBox[] {
			new JCheckBox(),
			new JCheckBox(),
			new JCheckBox(),
	};
	private final PropertyElement[] peProductionReports = new PropertyElement[] {
			new PropertyElement("GameViewProperties.ProductionReport.Land", checkboxProductionReports[0], null, "Game.Unit.Type.Land", ResourceFinder.Folder.APP),
			new PropertyElement("GameViewProperties.ProductionReport.Air", checkboxProductionReports[1], null, "Game.Unit.Type.Air", ResourceFinder.Folder.APP),
			new PropertyElement("GameViewProperties.ProductionReport.Sea", checkboxProductionReports[2], null, "Game.Unit.Type.Sea", ResourceFinder.Folder.APP),
	};

	private final JCheckBox[] checkboxSounds = new JCheckBox[SoundTypes.values().length];
	private final PropertyElement[] peSounds = new PropertyElement[checkboxSounds.length];
	{
		for (int i = 0; i < checkboxSounds.length; i++) {
			final SoundTypes st = SoundTypes.values()[i];
			checkboxSounds[i] = new JCheckBox();
			peSounds[i] = new PropertyElement(st.key, checkboxSounds[i], null, st.imageKey, st.imageFolder);
		}
	};

	private final Vector<JIconComboBox<String>> animationDelayComboBoxes;
	private final PropertyElement[] peAnimationDelays = new PropertyElement[GameViewProperties.AnimationDelayType.values().length];

	private final JCheckBox checkboxDistanceInfo = new JCheckBox();
	private final PropertyElement peDistanceInfo = new PropertyElement("GameViewProperties.Graphics.DistanceInfo", checkboxDistanceInfo, null, null, ResourceFinder.Folder.APP);

	private final GameViewProperties properties;
}
