package de.ped.empire.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.AbstractCellEditor;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;

import de.ped.empire.logic.Commandable;
import de.ped.empire.logic.TileTemplate;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.MilitaryBranch;
import de.ped.empire.logic.Unit;
import de.ped.empire.logic.ZoomLevel;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.TopologyMode;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.TableColumnModel;
import de.ped.tools.gui.TableModel;
import de.ped.tools.gui.TableTools;
import de.ped.tools.log.Logger;

/**
 * The dialog to edit the move paths starting from a game field.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MovePathsDialog extends PedJDialog {
	private static final long serialVersionUID = 1L;

	/**
	 * Represents the table to display the move paths starting from a game field.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class MovePathsTableModel extends TableModel {
		private static final long serialVersionUID = 1L;

		private static final TableColumnModel[] columns = new TableColumnModel[]{
			new TableColumnModel("Game.ChangeMovePathsDialog.Direction.Heading",	String.class, 30), 
			new TableColumnModel("Game.ChangeMovePathsDialog.Destination.Heading",	String.class, 20), 
			new TableColumnModel(MilitaryBranch.LAND_UNIT.messageKey, Boolean.class, 5, MilitaryBranch.LAND_UNIT.messageKey, ResourceFinder.Folder.APP),
			new TableColumnModel(MilitaryBranch.AIR_UNIT.messageKey, Boolean.class, 5, MilitaryBranch.AIR_UNIT.messageKey, ResourceFinder.Folder.APP),
			new TableColumnModel(MilitaryBranch.SEA_UNIT.messageKey, Boolean.class, 5, MilitaryBranch.SEA_UNIT.messageKey, ResourceFinder.Folder.APP),
		};

		private static final int ROW_IDX_NONE = 0;

		private static final int COL_IDX_DIRECTION = 0;
		private static final int COL_IDX_DESTINATION = 1;
		private static final int COL_IDX_MILITARY_BRANCH = 2;

		private final Object[][] data;

		public MovePathsTableModel(final Object[][] data, final Messages msg, final Commandable commandable) {
			super(msg);
			this.data = data;
		}

		@Override
		public TableColumnModel[] getColumns() {
			return columns;
		}

		@Override
		public int getRowCount() {
			return 1 + data.length;
		}

		@Override
		public Object getValueAt(final int rowIndex, final int columnIndex) {
			final Object ret;
			final int dataRowIndex = rowIndex - 1;
			switch(columnIndex) {
			case COL_IDX_DIRECTION:
				if (ROW_IDX_NONE == rowIndex) {
					ret = msg.getString("Game.ChangeMovePathsDialog.Direction.None.Text");
				} else {
					ret = msg.getString((I18NStringWithFillIns) data[dataRowIndex][1]);
				}
				break;
			case COL_IDX_DESTINATION:
				if (ROW_IDX_NONE == rowIndex) {
					ret = msg.getString("Game.ChangeMovePathsDialog.Destination.None.Text");
				} else {
					ret = msg.getString((I18NStringWithFillIns) data[dataRowIndex][2]);
				}
				break;
			default:
				if (ROW_IDX_NONE == rowIndex) {
					boolean isFound = false;
					for (Object[] dataRow : data) {
						final Boolean b = (Boolean) dataRow[1 + columnIndex];
						if (b.booleanValue()) {
							isFound = true;
							break;
						}
					}
					ret = Boolean.valueOf(!isFound);
				} else {
					ret = data[dataRowIndex][1 + columnIndex];
					logger.trace("getValueAt(" + rowIndex + "," + columnIndex + ") --> " + ret);
				}
			}
			return ret;
		}

		@Override
		public void setValueAt(final Object value, final int rowIndex, final int columnIndex) {
			logger.debug("setValueAt(" + rowIndex + "," + columnIndex + ") to " + value);
			final int dataRowIndex = rowIndex - 1;
			switch(columnIndex) {
			case COL_IDX_DIRECTION:
			case COL_IDX_DESTINATION:
				// Nothing to do
				break;
			default:
				// Update the model
				Boolean val = (Boolean) value;
				if (val.booleanValue()) {
					for (Object[] dataRow : data) {
						dataRow[1 + columnIndex] = Boolean.FALSE;
					}
					if (ROW_IDX_NONE < rowIndex) {
						data[dataRowIndex][1 + columnIndex] = Boolean.TRUE;
					}
					// A radio buttonCreateNewPath also affects all other cells belonging to the same buttonCreateNewPath group
					for (int i = 0; i <= data.length; i++) {
						fireTableCellUpdated(i, columnIndex);
					}
				}
			}
		}

		@Override
		public boolean isCellEditable(final int rowIndex, final int columnIndex) {
			boolean ret = false;
			final int dataRowIndex = rowIndex - 1;
			if (COL_IDX_MILITARY_BRANCH <= columnIndex) {
				if (ROW_IDX_NONE < rowIndex) {
					final Boolean b = (Boolean) data[dataRowIndex][4 + columnIndex];
					if (b.booleanValue()) {
						ret = true;
					}
				} else {
					ret = true;
				}
			}
			return ret;
		}
	}

	/**
	 * Knows how to show a move path selection cell (as a radio button).
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class MovePathsRenderer extends JRadioButton implements TableCellRenderer {
		private static final long serialVersionUID = 1L;

		public MovePathsRenderer() {
			// Nothing to do
		}

		@Override
		public Component getTableCellRendererComponent(
				final JTable table, 
				final Object value,
				final boolean isSelected, 
				final boolean hasFocus,
				final int row, 
				final int column) {
			return getCellRendererComponent(value, table.getModel().isCellEditable(row, column));
		}

		private Component getCellRendererComponent(final Object value, final boolean isEditable) {
			final Boolean val = (Boolean) value;
			setSelected(val.booleanValue());
			setEnabled(isEditable);
			setHorizontalAlignment(JLabel.CENTER);
			return this;
		}
	}

	/**
	 * Knows how to edit a move path selection cell (as a radio button).
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class MovePathsEditor extends AbstractCellEditor implements TableCellEditor, ItemListener {
		private static final long serialVersionUID = 1L;

		private final Logger logger = Logger.getLogger(getClass());

		private MovePathsRenderer renderer = null;	

		public MovePathsEditor() {
			// Nothing to do
		}

		@Override
		public Object getCellEditorValue() {
			Object ret = null;
			if (null != renderer) {
				ret = Boolean.valueOf(renderer.isSelected());
			}
			return ret;
		}

		@Override
		public Component getTableCellEditorComponent(
				final JTable table, 
				final Object value,
				final boolean isSelected, 
				final int row, 
				final int column) {
			final Boolean val = (Boolean) value;
			renderer = new MovePathsRenderer();
			renderer.setSelected(Boolean.valueOf(val));
			renderer.addItemListener(this);
			renderer.setHorizontalAlignment(JLabel.CENTER);
			return renderer;
		}

		@Override
		public void itemStateChanged(final ItemEvent event) {
			logger.debug("item selected: " + event);
			stopCellEditing();
		}
	}

	/**
	 * Creates the move path dialog.
	 * @param parent the parent window
	 * @param commandable the commandable to name
	 */
	public MovePathsDialog(
			final ApplicationMainWindow parent, 
			final Commandable commandable) {
		super(parent, parent);
		empWindow = (EmpireMainWindow) parent;
		this.commandable = commandable;
		final Messages messages = getMessages();
		setTitle(messages.getString("Game.ChangeMovePathsDialog.Title"));

		// model for template label describing the commandable to look at
		final TileTemplate image;
		if (commandable instanceof Unit) {
			final Unit unit = (Unit) commandable;
			image = unit;
		} else {
			image = (FieldCity) commandable;
		}
		final I18NStringWithFillIns cmdlName = commandable.getName();
		final I18NStringWithFillIns title = new I18NStringWithFillIns("Game.ChangeMovePathsDialog.Text", null, new I18NStringWithFillIns[] { cmdlName } );

		setTitlePanel(messages.getString(title));

		getTitleLabel().setIcon(new ImageIcon(image.createImage(ZoomLevel.DIALOG_ZOOM_LEVEL, TopologyMode.SQUARE8)));

		final JPanel tablePanel = new JPanel();
		{
			final Object[][] pathsData = empWindow.getGameFacade().getMovePaths(commandable);
			tableModel = new MovePathsTableModel(pathsData, empWindow.getMessages(), commandable);
			table = new JTable(tableModel);
			table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			table.setDefaultRenderer(Boolean.class, new MovePathsRenderer());
			table.setDefaultEditor(Boolean.class, new MovePathsEditor());
			TableTools.adjustColumns(table, 5, empWindow.getApplicationEnvironment().resources());
			ApplicationMainWindow.adjustRowHeightsToRenderers(table);
			tablePanel.setLayout(new BorderLayout());
			tablePanel.add(table, BorderLayout.CENTER);
			tablePanel.add(table.getTableHeader(), BorderLayout.NORTH);
			buttonCreateNewPath = parent.createButton("Game.ChangeMovePathsDialog.Path.New.Text");
			buttonCreateNewPath.addActionListener(this);
			tablePanel.add(ApplicationMainWindow.createPanel(buttonCreateNewPath), BorderLayout.SOUTH);
		}
		final JPanel mainPanel = createMainPanel();
		{
			final BorderLayout mainPanelLayout = new BorderLayout();
			mainPanel.setLayout(mainPanelLayout);
			mainPanel.add(tablePanel, BorderLayout.CENTER);
		}
		add(new JScrollPane(mainPanel), BorderLayout.CENTER);
		updateUIFieldsBasedOnProperties();
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		finishLayout();
	}

	@Override
	public void actionPerformed(final ActionEvent e) {
		super.actionPerformed(e);
		if (e.getSource() == buttonCreateNewPath) {
			wasButtonCreateNewPathPressed = true;
			leaveDialog();
		}
	}

	@Override
	public void enterDialog() {
		super.enterDialog();
		wasButtonCreateNewPathPressed = false;
	}

	public Object[][] getResult() {
		return tableModel.data;
	}

	public Commandable getCommandable() {
		return commandable;
	}

	public JButton getButtonCreateNewPath() {
		return buttonCreateNewPath;
	}

	public boolean wasButtonCreateNewPathPressed() {
		return wasButtonCreateNewPathPressed;
	}

	protected final EmpireMainWindow empWindow;
	protected final Commandable commandable;

	private final MovePathsTableModel tableModel;
	private final JTable table;
	private final JButton buttonCreateNewPath;
	private boolean wasButtonCreateNewPathPressed = false;
}
