package de.ped.empire.gui;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;

import de.ped.empire.logic.Commandable;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.Unit;
import de.ped.tools.Messages;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;

/**
 * Presents a dialog to the user to name or rename a commandable.
 * Has three modes to distinguish:
 * <ol>
 * <li>A new unit was created. The game informs the user about the creation, and the user can name the unit.</li>
 * <li>The user wants to rename a unit.</li>
 * <li>The user wants to rename a city.</li>
 * </ol> 
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class NamingDialog extends PedJDialog {

	private static final long serialVersionUID = 1L;

	protected final EmpireMainWindow empWindow;
	private JTextField nameField = new JTextField("Insert name");
	private JLabel templateLabel;
	private JButton productionButton = null;
	private boolean wasProductionButtonPressed = false;

	private final FieldCity producingCity;
	private final Commandable commandable;

	/**
	 * Creates the unit naming dialog.
	 * 
	 * @param parent
	 *            the parent window
	 * @param commandable
	 *            the commandable to name
	 * @param producingCity
	 *            the city where the unit was created, or <code>null</code> if the
	 *            commandable is just to rename
	 * @param changeProductionOrderMessagePrefix
	 *            the message key prefix for the "change production" button
	 *            (only makes sense if <em>producingCity</em> is set)
	 */
	public NamingDialog(
			final ApplicationMainWindow parent, 
			final Commandable commandable, 
			final FieldCity producingCity,
			final String changeProductionOrderMessagePrefix) {
		super(parent, parent);
		this.empWindow = (EmpireMainWindow) parent;
		this.commandable = commandable;
		this.producingCity = producingCity;
		final Messages messages = getMessages();
		setTitle(messages.getString(
				(null == producingCity ? 
						(commandable instanceof Unit ? 
								"Game.Naming.Unit.Title" : 
								"Game.Naming.City.Title") :
						"Game.Naming.ProductionMessage.Title")));
		templateLabel = OrderSafetyQuestionDialog.createLabelForCommandable(parent, commandable);
		
		// Title pane
		final String title;
		if (null == producingCity) {
			final String key = 
				(commandable instanceof Unit ? 
					"Game.Naming.Unit" : 
					"Game.Naming.City");
			title = messages.getText(key, null, new String[] { templateLabel.getText(), });
		} else {
			title = messages.getText("Game.Naming.ProductionMessage", null, 
					new Object[] {
							templateLabel.getText(),
							producingCity.getName(),
					});
		}
		setTitlePanel(title);
		
		// model for name property
		final String name;
		if (commandable.needsNickName()) {
			name = commandable.getNickNameSuggestion();
		} else {
			name = commandable.getNickName();
		}
		nameField.setText(name);
		nameField.setColumns(15);
		setPropertyElements(new PropertyElement[] { new PropertyElement("Game.Naming.Name", nameField), });

		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.CENTER;
		mainPanel.setLayout(gridbag);
		{ // First line contains the unit type
			c.gridwidth = GridBagConstraints.REMAINDER; //end row
			c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
			gridbag.setConstraints(templateLabel, c);
			mainPanel.add(templateLabel);
		}
		addAllPropertyElementsTo(mainPanel, c);
		add(mainPanel, BorderLayout.CENTER);

		int optionType = JOptionPane.OK_CANCEL_OPTION;
		if (null != producingCity) {
			optionType = JOptionPane.DEFAULT_OPTION;
			productionButton = parent.createButton(changeProductionOrderMessagePrefix);
			addButton(productionButton);
			productionButton.addActionListener(this);
		}
		setOptionType(optionType);
		finishLayout();
		setResizable(false);
	}

	@Override
	public void actionPerformed(final ActionEvent e) {
		super.actionPerformed(e);
		if (e.getSource() == productionButton) {
			wasProductionButtonPressed = true;
			onOk();
		}
	}

	public boolean wasProductionButtonPressed() {
		return wasProductionButtonPressed;
	}

	@Override
	public String getName() {
		return nameField.getText();
	}

	public Commandable getCommandable() {
		return commandable;
	}

	public FieldCity getProducingCity() {
		return producingCity;
	}
}
