package de.ped.empire.gui;

import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.PlayerHandicap;
import de.ped.empire.logic.PlayerType;
import de.ped.tools.Messages;
import de.ped.tools.PlayerColor;
import de.ped.tools.gui.TableCellComboBoxRenderer;
import de.ped.tools.gui.TableColumnModel;
import de.ped.tools.gui.TableModel;

/**
 * Represents the table to display for the list of players of the new game.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PlayersSelectionTableModel extends TableModel {

	private static final long serialVersionUID = 1L;

	private static final TableColumnModel[] COLUMNS = new TableColumnModel[]{
		new TableColumnModel("GameProperties.Players.Number.Heading",             Integer.class,         5), 
		new TableColumnModel("GameProperties.Players.Type.Heading",               PlayerType.class,     25), 
		new TableColumnModel("GameProperties.Players.Color.Heading",              PlayerColor.class,    15),
		new TableColumnModel("GameProperties.Players.HandicapFight.Heading",      PlayerHandicap.class, 35), 
		new TableColumnModel("GameProperties.Players.HandicapProduction.Heading", PlayerHandicap.class, 35),
		new TableColumnModel("GameProperties.Players.Name.Heading",               String.class,         25), 
	};

	private static final int COL_IDX_NUMBER = 0;
	private static final int COL_IDX_TYPE = 1;
	private static final int COL_IDX_COLOR = 2;
	private static final int COL_IDX_HANDICAP_FIGHT = 3;
	private static final int COL_IDX_HANDICAP_PRODUCTION = 4;
	private static final int COL_IDX_NAME = 5;

	private final Player[] players;
	private final boolean isEditable;

	public PlayersSelectionTableModel(
			final GameProperties properties, 
			final Messages msg, 
			final boolean isEditable) {
		super(msg);
		this.players = properties.getPlayers(); // The containing class NewGameSelectionDialog does the cloning already, so we work just on the same model.
		this.isEditable = isEditable;
	}

	@Override
	public TableColumnModel[] getColumns() {
		return COLUMNS;
	}

	@Override
	public int getRowCount() {
		return Player.MAX_NUMBER_OF_PLAYERS;
	}

	@Override
	public Object getValueAt(final int rowIndex, final int columnIndex) {
		logger.trace("getValueAt(" + rowIndex + "," + columnIndex + ")");
		final Player entry = players[rowIndex];
		final Object ret;
		switch(columnIndex) {
		case COL_IDX_NUMBER:
			ret = new Integer(rowIndex + 1);
			break;
		case COL_IDX_NAME:
			ret = entry.getName();
			break;
		case COL_IDX_TYPE:
			ret = PlayerType.values()[entry.getPlayerTypeId()];
			break;
		case COL_IDX_COLOR:
			ret = PlayerColor.get(entry.getPlayerColorId());
			break;
		case COL_IDX_HANDICAP_FIGHT:
			ret = PlayerHandicap.values()[entry.getHandicapFightId()];
			break;
		case COL_IDX_HANDICAP_PRODUCTION:
			ret = PlayerHandicap.values()[entry.getHandicapProductionId()];
			break;
		default:
			ret = "--";
		}
		return ret;
	}

	@Override
	public void setValueAt(final Object value, final int rowIndex, final int columnIndex) {
		logger.trace("setValueAt(" + rowIndex + "," + columnIndex + ")");
		final Player entry = players[rowIndex];
		switch(columnIndex) {
		case COL_IDX_NUMBER:
			// Nothing to do
			break;
		case COL_IDX_NAME:
			entry.setName((String) value);
			break;
		case COL_IDX_TYPE:
			setPlayerTypeId(rowIndex, TableCellComboBoxRenderer.getIdxOf(value));
			break;
		case COL_IDX_COLOR: {
			int colorId = Player.ID_NONE;
			if (value instanceof PlayerColor) {
				colorId = ((PlayerColor) value).getId();
			} else {
				colorId = ((Integer) value).intValue();
			}
			setPlayerColorId(rowIndex, colorId);
			break;
		}
		case COL_IDX_HANDICAP_PRODUCTION:
		case COL_IDX_HANDICAP_FIGHT: {
			int id = TableCellComboBoxRenderer.getIdxOf(value);
			if (COL_IDX_HANDICAP_FIGHT == columnIndex) {
				entry.setHandicapFightId(id);
			} else {
				entry.setHandicapProductionId(id);
			}
			break;
		}
		default:
			// Nothing to do
		}
	}

	/**
	 * Sets the colorId for a player. 
	 * But as all colors are unique among players, this swaps color with the other player previously using that color.
	 * @param playerId the player to get the new colorId
	 * @param colorId the new colorId
	 */
	protected void setPlayerColorId(final int playerId, final int colorId) {
		logger.trace("setPlayerColorId(" + playerId + "," + colorId + ")");
		final Player entry = players[playerId];
		int otherPlayerIdx = -1;
		for (int i = 0; i < players.length; i++) {
			if (players[i].getPlayerColorId() == colorId) {
				otherPlayerIdx = i;
				break;
			}
		}
		players[otherPlayerIdx].setPlayerColorId(entry.getPlayerColorId());
		entry.setPlayerColorId(colorId);
		fireTableCellUpdated(otherPlayerIdx, COL_IDX_COLOR);
	}

	protected void setPlayerTypeId(final int playerId, final int typeId) {
		final Player entry = players[playerId];
		entry.setPlayerTypeId(typeId);
		fireTableCellUpdated(playerId, COL_IDX_TYPE);
	}

	@Override
	public boolean isCellEditable(final int rowIndex, final int colIndex) {
		return isEditable && (COL_IDX_NUMBER != colIndex) 
				&& ((COL_IDX_TYPE == colIndex) 
						|| (players[rowIndex].getPlayerTypeId() != PlayerType.NONE.ordinal()));
	}
}