package de.ped.empire.gui;

import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.util.LinkedList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.border.Border;

import de.ped.empire.logic.GameFacade;
import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.TileTemplate;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.StatusReportPlayerData;
import de.ped.empire.logic.ZoomLevel;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.TopologyMode;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;

/**
 * A dialog to show the the global status for a player.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class StatusReportDialog extends PedJDialog {

	private static final long serialVersionUID = 1L;

	/**
	 * Contains all data necessary for a table row of this table to build it data driven.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class StatusUnitModel {
		JLabel nameLabel;
		JLabel inProductionLabel;
		JLabel nextReadyInLabel;
		JLabel cntInFieldLabel;
		JLabel cntLossesLabel;
		JLabel cntKillsLabel;

		public StatusUnitModel(final Border border) {
			nameLabel = new JLabel("---", new ImageIcon(ZoomLevel.DIALOG_ZOOM_LEVEL.getImageSet(TopologyMode.SQUARE8).getTransparentImage()), JLabel.LEFT);
			inProductionLabel = new JLabel("000");
			nextReadyInLabel = new JLabel("000");
			cntInFieldLabel = new JLabel("000");
			cntLossesLabel = new JLabel("000");
			cntKillsLabel = new JLabel("000");
			nameLabel.setBorder(border);
			inProductionLabel.setBorder(border);
			nextReadyInLabel.setBorder(border);
			cntInFieldLabel.setBorder(border);
			cntLossesLabel.setBorder(border);
			cntKillsLabel.setBorder(border);
		}

		public void setData(final StatusReportPlayerData.StatusReportUnitData data) {
			final ImageKey imageKey = new ImageKey(ImageKey.TType.Unit, data.template.id);
			imageKey.setBits(ImageKey.BA_PLAYER_COLOR, data.gameView.getPlayer().getPlayerColorId());
			nameLabel.setIcon(new ImageIcon(TileTemplate.createImage(imageKey.getAll(), ZoomLevel.DIALOG_ZOOM_LEVEL, TopologyMode.SQUARE8)));
			nameLabel.setText(data.gameView.getMessages().getString(data.template.getName()));
			final String turnsText = (data.nextReadyIn == Integer.MAX_VALUE ? "--" : Integer.toString(data.nextReadyIn));
			nextReadyInLabel.setText(turnsText);
			inProductionLabel.setText(Integer.toString(data.inProduction));
			cntInFieldLabel.setText(Integer.toString(data.cntInField));
			cntLossesLabel.setText(Integer.toString(data.cntLosses));
			cntKillsLabel.setText(Integer.toString(data.cntKills));
		}
	}

	protected final EmpireMainWindow empWindow;

	public StatusReportDialog(final ApplicationMainWindow mainWindow, final Frame parent, final int playerId) {
		super(mainWindow, parent);
		this.empWindow = (EmpireMainWindow) mainWindow;
		final Messages messages = getMessages();
		final GameFacade gameFacade = empWindow.getGameFacade();
		final GameProperties props = gameFacade.getGameProperties();
		data = gameFacade.getStatusReportPlayerData(playerId, false);
		final boolean isAlive = gameFacade.isPlayerAlive(playerId);
		final boolean isWinner = gameFacade.isPlayerWinner(playerId);

		final Player player = props.getPlayers()[playerId];
		final I18NStringWithFillIns title = new I18NStringWithFillIns("Game.Status.Title", "%0 %2", 
						new Object[] { 
							new I18NStringWithFillIns("Action.ViewStatusReport.Text", "Status", null),
							playerId,
							player.getName(),
				} );
		super.setTitle(messages.getString(title));

		final JPanel panel = new JPanel();
		panel.setLayout(new GridBagLayout());
		final GridBagConstraints positioningConstraints = new GridBagConstraints();
		positioningConstraints.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		positioningConstraints.gridwidth = GridBagConstraints.REMAINDER; //end row
		positioningConstraints.fill = GridBagConstraints.HORIZONTAL;

		final String key;
		if (isWinner) {
			key = "Game.Status.PlayerState.Value.Won.Text";
		} else if (isAlive) {
			key = "Game.Status.PlayerState.Value.Alive.Text";
		} else {
			key = "Game.Status.PlayerState.Value.Lost.Text";
		}
		final JLabel valuePlayerState = ApplicationMainWindow.createLabel(messages.getText(key, null, null), null);
		final PropertyElement pePlayerState = new PropertyElement("Game.Status.PlayerState.Label", valuePlayerState);
		empWindow.addPropertyElementTo(panel, pePlayerState, positioningConstraints);
		
		final int mapPercentage = (int) Math.round(100.0 * data.mapKnown / data.mapTotal);
		final JLabel valueMap = ApplicationMainWindow.createLabel(messages.getText("Game.Status.Map.Value", null, 
				new String[] { 
					Integer.toString(mapPercentage), 
					Integer.toString(data.mapKnown),
					Integer.toString(data.mapTotal),
				}), null);
		final PropertyElement peMap = new PropertyElement("Game.Status.Map.Label", valueMap);
		empWindow.addPropertyElementTo(panel, peMap, positioningConstraints);

		final int citiesPercentage = (int) Math.round(100.0 * data.citiesPlayer / data.citiesTotal);
		final JLabel valueCities = ApplicationMainWindow.createLabel(messages.getText("Game.Status.Cities.Value", null, 
				new String[] { 
					Integer.toString(citiesPercentage), 
					Integer.toString(data.citiesPlayer),
					Integer.toString(data.citiesTotal),
				}), null);
		final PropertyElement peCities = new PropertyElement("Game.Status.Cities.Label", valueCities);
		empWindow.addPropertyElementTo(panel, peCities, positioningConstraints);
		
		final JLabel valueTroops = ApplicationMainWindow.createLabel(messages.getText("Game.Status.Cities.Efficiency.Value", null, 
				new String[] { 
					Integer.toString(Math.round(100.0f * data.troopsFactor)), 
					Integer.toString(data.troopsPlayer), 
				}), null);
		final PropertyElement peTroops = new PropertyElement("Game.Status.Cities.Efficiency.Label", valueTroops);
		empWindow.addPropertyElementTo(panel, peTroops, positioningConstraints);

		addTitlePanel(panel);

		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);

		init(mainPanel, gridbag, playerId);
		addAllPropertyElementsTo(mainPanel, c);

		add(mainPanel, BorderLayout.CENTER);
		setOptionType(JOptionPane.DEFAULT_OPTION);
		finishLayout();
		setResizable(false);
	}

	private StatusReportPlayerData data;
	private List<StatusUnitModel> unitModels = new LinkedList<StatusUnitModel>();

	protected void init(
			final JPanel panel, 
			final GridBagLayout gridbag, 
			final int playerId) {
		final GridBagConstraints c = new GridBagConstraints();
		c.ipadx = 10;
		// Heading
		final GridBagTableColumn columns[] = new GridBagTableColumn[] {
			new GridBagTableColumn(1, GridBagConstraints.WEST, "Game.Status.Units.UnitType.Heading"),
			new GridBagTableColumn(1, GridBagConstraints.CENTER, "Game.Status.Units.InProduction.Heading"),
			new GridBagTableColumn(1, GridBagConstraints.CENTER, "Game.Status.Units.NextReadyIn.Heading"),
			new GridBagTableColumn(1, GridBagConstraints.CENTER, "Game.Status.Units.CntInField.Heading"),
			new GridBagTableColumn(1, GridBagConstraints.CENTER, "Game.Status.Units.CntLosses.Heading"),
			new GridBagTableColumn(GridBagConstraints.REMAINDER, 
									GridBagConstraints.CENTER, "Game.Status.Units.CntKills.Heading"),
		};
		final Border emptyBorder = BorderFactory.createEmptyBorder(2, 0, 2, 0);
		for (final GridBagTableColumn column : columns) {
			c.gridwidth = column.gridwidth;
			c.anchor = column.anchor; 
			final JLabel label = empWindow.createLabel(column.key);
			gridbag.setConstraints(label, c);
			label.setBorder(emptyBorder);
			panel.add(label);
		}
		// Content
		for (final StatusReportPlayerData.StatusReportUnitData unitData : data.units) {
			final StatusUnitModel pum = new StatusUnitModel(emptyBorder);
			pum.setData(unitData);
			unitModels.add(pum);
			{
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
				gridbag.setConstraints(pum.nameLabel, c);
				panel.add(pum.nameLabel);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.inProductionLabel, c);
				panel.add(pum.inProductionLabel);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.nextReadyInLabel, c);
				panel.add(pum.nextReadyInLabel);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.cntInFieldLabel, c);
				panel.add(pum.cntInFieldLabel);
			} {
				c.gridwidth = 1;
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.cntLossesLabel, c);
				panel.add(pum.cntLossesLabel);
			} {
				c.gridwidth = GridBagConstraints.REMAINDER; //end row
				c.anchor = GridBagConstraints.CENTER;
				gridbag.setConstraints(pum.cntKillsLabel, c);
				panel.add(pum.cntKillsLabel);
			}
		}
	}
}
