package de.ped.empire.logic;

import java.util.LinkedList;
import java.util.List;

import de.ped.empire.logic.MoveResult.CommandableInfo;
import de.ped.empire.logic.MoveResult.Role;
import de.ped.tools.CollectionUtil;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * An animation to show for a move result.
 * An animation consists of parts which again consist of steps of equal length.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Animation implements Marshallable {
	/** Default value for {@link #getDyingProgress()}. Is less or equal than lowest legal value. */
	public static final float NOT_DYING = 0.0f;
	/** Default value for {@link #getDyingProgress()}. Is less or equal than lowest legal value. */
	public static final float NOT_MOVING = 0.0f;

	/**
	 * Creates a list of animation parts based on the move result to show in
	 * this sequence.
	 * 
	 * @param moveResult
	 *            the move result to show in this sequence
	 * @return a list of animation parts
	 */
	private static List<AnimationPart> createSequence(final MoveResult moveResult) {
		final List<AnimationPart> ret = new LinkedList<AnimationPart>();
		if (null != moveResult) {
			final int resultMask = moveResult.getResultMask();
			if (0 != (resultMask & MoveResult.MOVE_MASK_FIGHTING)) {
				ret.add(AnimationPart.FIGHT_INTRO);
			}
			for (int i = 0; i < moveResult.getHits().size(); i++) {
				if (MoveResult.Role.Affected == moveResult.getHits().get(i)) {
					ret.add(AnimationPart.FIGHT_HIT_ENEMY);
				} else {
					ret.add(AnimationPart.FIGHT_HIT_OWN);
				}
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_WINNING_FIGHT)) {
				ret.add(AnimationPart.DEATH_ENEMY);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_LOSING_FIGHT)) {
				ret.add(AnimationPart.DEATH_OWN);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_SLOWED_DOWN)) {
				ret.add(AnimationPart.SLOWED_DOWN);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_MOVING)) {
				ret.add(AnimationPart.MOVE);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_FUEL_EMPTY)) {
				ret.add(AnimationPart.FUEL_EMPTY);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_ALERT_IMPEDIMENT)) {
				ret.add(AnimationPart.ALERT_IMPEDIMENT);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_ALERT_SIGHTING)) {
				ret.add(AnimationPart.ALERT_SIGHTING);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_DOCKING)) {
				ret.add(AnimationPart.DOCK);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_LOADING)) {
				ret.add(AnimationPart.LOAD);
			}
			if (0 != (resultMask & MoveResult.MOVE_MASK_GARRISON)) {
				ret.add(AnimationPart.GARRISON);
			}
			ret.add(AnimationPart.END_ANIMATION);
		}
		return ret;
	}

	/** Not persistent. */
	private Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
	}

	private final GameView gameView;

	Animation(final GameView gameView, final MoveResult moveResult) {
		this.gameView = gameView;
		setMoveResult(moveResult);
	}

	Animation(final GameView gameView, final Marshaller marshaller) {
		this.gameView = gameView;
		demarshal(marshaller);
	}

	/**
	 * Called whenever the game thread wakes up. Brings the current sequence
	 * showing the current move result one step further. When the sequence is
	 * done, clears the move result to show.
	 * @param pst the processor speed type, to adapt number of steps of visualizations to the step duration
	 * @return <code>true</code> if the sequence was finished, so now it's tme
	 *         to switch to the next animation sequence
	 */
	public synchronized boolean step(final ProcessorSpeedType pst) {
		boolean ret = false;
		if (null != moveResultToShow) {
			if (!hasBegun) {
				hasBegun = true;
				if (!hasBeenShown) {
					final GameNotification gameNotification = gameView.getGameNotification();
					gameNotification.setAnimationMoveResult(moveResultToShow);
				}
				gameView.pushStatusMessage(I18NStringWithFillIns.EMPTY);
			}
			AnimationPart seq = null;
			try {
				seq = sequence.get(0);
			} catch (IndexOutOfBoundsException ex) {
				logger.error("Trying to read animation part, but there isn't: " + "step(" + moveResultToShow.toString() + ",step=" + showStep + ")");
			}
			if (null != seq) {
				final int step = showStep;
				logger.debug("step(" + moveResultToShow.toString() + ",step=" + showStep + "->(" + seq + "," + step + "))");
				final int duration = seq.duration[pst.ordinal()][gameView.getGameViewProperties()
						.getAnimationDelay(seq.animationDelayType).ordinal()];
				switch(seq) {
				case FIGHT_INTRO:
					if (duration <= FightOverlay.NUM_OF_INSTANCES) {
						calculateFightOverlay(0, step * 2, 0);
						calculateFightOverlay(1, step * 2, (duration * 2 - FightOverlay.NUM_OF_INSTANCES));
					} else {
						final int div = duration / FightOverlay.NUM_OF_INSTANCES;
						calculateFightOverlay(0, step / div, 0);
						calculateFightOverlay(1, step / div, (duration / div - FightOverlay.NUM_OF_INSTANCES));
					}
					if (1 <= step) {
						calcMoveVector();
						moveProgress = 0.25f;
					} else {
						moveVector = null;
						moveProgress = NOT_MOVING;
					}
					break;
				case DEATH_OWN:
					if (0 == step) {
						messageAndLogDeath();
					}
					dyingProgress = (step + 1.0f) / (duration + 1.0f);
					moveVector = null;
					moveProgress = NOT_MOVING;
					// No break
				case FIGHT_HIT_OWN:
					if (duration < FightOverlay.NUM_OF_INSTANCES) {
						calculateFightOverlay(0, step, 0);
					} else {
						final int div = duration / FightOverlay.NUM_OF_INSTANCES;
						calculateFightOverlay(0, step / div, 0);
					}
					fightOverlays[1] = null;
					break;
				case DEATH_ENEMY:
					if (0 == step) {
						messageAndLogDeath();
					}
					dyingProgress = (step + 1.0f) / (duration + 1.0f);
					moveVector = null;
					moveProgress = NOT_MOVING;
					// No break
				case FIGHT_HIT_ENEMY:
					if (duration < FightOverlay.NUM_OF_INSTANCES) {
						calculateFightOverlay(1, step, 0);
					} else {
						final int div = duration / FightOverlay.NUM_OF_INSTANCES;
						calculateFightOverlay(1, step / div, 0);
					}
					fightOverlays[0] = null;
					break;
				case MOVE:
					fightOverlays[0] = null;
					fightOverlays[1] = null;
					float part = (step + 1.0f) / (duration + 1.0f);
					calcMoveVector();
					moveProgress = part;
					break;
				case DOCK:
					if (0 == step) {
						final String key = moveResultToShow.getUnit().getUnitTemplate().militaryBranch.messageKey + ".EnteredCity";
						final I18NStringWithFillIns message = new I18NStringWithFillIns(
								key, 
								null, 
								new Object[] { 
										moveResultToShow.getUnitName(Role.Active), 
										moveResultToShow.getUnitName(Role.Affected), });
	
						gameView.setStatusMessage(message);
						logSinglePlayerMessage(message);
					}
					break;
				case LOAD:
					if (0 == step) {
						gameView.setStatusMessage(new I18NStringWithFillIns(
								"Game.Unit.EnteredUnit", 
								null, 
								new Object[] { 
										moveResultToShow.getUnitName(Role.Active), 
										moveResultToShow.getUnitName(Role.Affected), }));
					}
					break;
				case GARRISON:
					break;
				case SLOWED_DOWN:
					break;
				case ALERT_IMPEDIMENT:
					break;
				case ALERT_SIGHTING:
					break;
				case FUEL_EMPTY:
					if (0 == step) {
						final I18NStringWithFillIns message = new I18NStringWithFillIns(
								"Game.Unit.FuelEmpty", 
								null, 
								new Object[] { 
										moveResultToShow.getUnitName(Role.Active), });
						gameView.setStatusMessage(message);
						logSinglePlayerMessage(message);
					}
					calculateFightOverlay(0, step, 0);
					fightOverlays[1] = null;
					break;
				case END_ANIMATION:
					if (0 == step) {
						gameView.popStatusMessage();
					}
					fightOverlays[0] = null;
					fightOverlays[1] = null;
					ret = true;
					showStep = 0;
					break;
				default:
					throw new IllegalStateException("Unknown move subsequence " + seq);
				}
				if (AnimationPart.END_ANIMATION.equals(seq)) {
					gameView.getGameNotification().setAnimationMoveResult(null);
				} else {
					gameView.getGameNotification().setAnimationPart(moveResultToShow, seq, step);
				}
				if (null != moveResultToShow) {
					showStep++;
					if (duration <= showStep) {
						sequence.remove(0);
						showStep = 0;
					}
				}
			}
		}
		return ret;
	}

	protected void calcMoveVector() {
		moveVector = gameView.getGameState().getProperties().renormDistanceVector(
				moveResultToShow.direction.getDistanceVector(),
				moveResultToShow.getCommandableInfo(Role.Active).position);
	}

	private void messageAndLogDeath() {
		final int[] killedUnits = moveResultToShow.getKilledUnits();
		Role role = null;
		CommandableInfo ci = null;
		// Create status message for current view
		for (final Role r : Role.values()) {
			ci = moveResultToShow.getCommandableInfo(r);
			if ((null != ci) && (gameView.getPlayer().getId() == ci.getPlayerId())) {
				role = r;
				break;
			}
		}
		if (null != role) {
			final I18NStringWithFillIns logMsg = moveResultToShow.getLogMessage(role, killedUnits[0], killedUnits[1]);
			if (!logMsg.isEmpty()) {
				gameView.setStatusMessage(logMsg);
			}
		}
		// Create log messages for both players
		final GameState gameState = gameView.getGameState();
		if (!hasBeenShown) {
			for (final Role r : Role.values()) {
				final I18NStringWithFillIns logMsg = moveResultToShow.getLogMessage(r, killedUnits[0], killedUnits[1]);
				if (!logMsg.isEmpty()) {
					final LogbookEntry logbookEntry = new LogbookEntry(gameState.getTurn(), moveResultToShow, r, logMsg);
					final int logPlayerId = moveResultToShow.getCommandableInfo(r).getPlayerId();
					if (Player.ID_NONE != logPlayerId) {
						final GameView logView = gameState.getGameView(logPlayerId);
						logView.onLogbookEntryWorthyEvent(logbookEntry);
					}
				}
			}
		}
	}
	
	private void logSinglePlayerMessage(final I18NStringWithFillIns message) {
		if (!hasBeenShown) {
			gameView.getLogbook().add(new LogbookEntry(
					gameView.getGameState().getTurn(),
					moveResultToShow.getUnit().getPlayerId(),
					Player.ID_NONE,
					moveResultToShow.getUnit().getPosition(),
					moveResultToShow.getUnit(),
					message));
		}
	}

	private void calculateFightOverlay(final int index, final int step, final int offset) {
		if ((step < offset) || (offset + FightOverlay.NUM_OF_INSTANCES <= step)) {
			fightOverlays[index] = null;
		} else {
			fightOverlays[index] = FightOverlay.instance(step - offset);
		}
	}

	/**
	 * Marshal this object into a byte stream. So far only the move result is
	 * persistent.
	 * 
	 * @param marshaller
	 *            the stream to write into
	 */
	@Override
	public void marshal(final Marshaller marshaller) {
		if (null == moveResultToShow) {
			throw new IllegalStateException("An animation without move result?!");
		}
		marshaller.writeMarshallable(moveResultToShow);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		moveResultToShow = new MoveResult(gameView, marshaller);
		reset();
		return this;
	}

	/** 
	 * @return the fight overlays (for both combatants) to display for the current animation step
	 */
	public synchronized FightOverlay[] getFightOverlays() {
		return fightOverlays;
	}

	/** 
	 * @return the move vector for the active combatant, or <code>null</code> if there is no unit to be shown outside its tile
	 */
	public synchronized PlayfieldDistance getMoveVector() {
		return moveVector;
	}

	/**
	 * @return the progress of the move sequence for the active combatant 
	 * 		   ({@link #NOT_MOVING} if unit is currently not moving, 
	 * 	       or in a range between 0.0 for start of moving sequence 
	 *         and 1.0 for end of moving sequence)
	 */
	public synchronized float getMoveProgress() {
		return moveProgress;
	}

	/**
	 * @return the progress of the dying sequence for the active combatant 
	 * 		   ({@link #NOT_DYING} if unit is currently not dying, 
	 * 	       or in a range between 0.0 for start of dying sequence 
	 *         and 1.0 for end of dying sequence)
	 */
	public synchronized float getDyingProgress() {
		return dyingProgress;
	}

	/** 
	 * @return the move result to show, or <code>null</code> if there is no move to show
	 */
	public synchronized MoveResult getMoveResult() {
		return moveResultToShow;
	}

	/** 
	 * Makes a move result ready for replay. 
	 */
	synchronized void reset() {
		hasBegun = false;
		hasBeenShown = true;
		moveVector = null;
		moveProgress = NOT_MOVING;
		dyingProgress = NOT_DYING;
		setMoveResult(moveResultToShow);
	}
	
	/** 
	 * Sets the next move result to show. A call while another move result is still not through all its display steps is illegal.
	 * @param moveResult the next move result to show
	 */
	private void setMoveResult(final MoveResult moveResult) {
		if (null == moveResult) {
			throw new IllegalArgumentException("Won't accept null for move result");
		}
		boolean hasChanged = (moveResult != moveResultToShow);
		if ((null != moveResultToShow) && (hasChanged)) {
			logger.fatal("Setting move result: " + (null == moveResult ? "null" : moveResult.toString()) + " on top of " + (null == moveResultToShow ? "null" : moveResultToShow.toString()));
		} else {
			logger.trace("Setting move result: " + (null == moveResult ? "null" : moveResult.toString()));
		}
		moveResultToShow = moveResult;
		sequence = createSequence(moveResult);
		if (sequence.isEmpty()) {
			throw new IllegalArgumentException("Empty sequence from move result " + moveResult.toString());
		}
		showStep = 0;
	}
	
	@Override
	public String toString() {
		String ret;
		if (null != moveResultToShow) {
			ret = moveResultToShow.toString();
		} else {
			ret = "No move result";
		}
		if (!CollectionUtil.isNullOrEmpty(sequence)) {
			ret += " " + sequence.get(0);
		}
		ret += " " + showStep;
		return ret;
	}

	/** If not <code>null</code>, indicates the animation to display. Not persistent. */
	private MoveResult moveResultToShow;

	/** If not <code>null</code>, indicates the animation to display. Not persistent. */
	private List<AnimationPart> sequence;

	/** The animation step of the animation to display. Only makes sense when {@link #moveResultToShow} is not <code>null</code>. Not persistent. */
	private int showStep = 0;

	/** Stores the fight overlays (for both combatants) to display for the current animation step. Not persistent. */
	private FightOverlay[] fightOverlays = new FightOverlay[2];

	/** Knows if this animation has been shown already. Only when shown first time, logbook entries and other things have to be done. */
	private boolean hasBeenShown = false;

	/** Knows if this animation has been started. */
	private boolean hasBegun = false;

	private PlayfieldDistance moveVector = null;
	private float moveProgress = NOT_MOVING;
	private float dyingProgress = NOT_DYING;
}