package de.ped.empire.logic;

import java.awt.Graphics;
import java.awt.Point;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;

/**
 * Anything on the map that can be given a command by the owning player.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public interface Commandable extends PositionedTileable, Marshallable {

	/**
	 * Desribes how a unit might attack another commandable.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum AttackMode {
		/** Attack is impossible. */
		IMPOSSIBLE,
		/** When the attacker wins, it won't move on the attacked field. */
		WITHOUT_MOVE,
		/** When the attacker wins, it will move on the attacked field if possible. */
		INCLUDING_MOVE,
	};

	@Override
	public abstract int getPlayerId();

	@Override
	public abstract PlayfieldPosition getPosition();

	@Override
	public abstract I18NStringWithFillIns getName();

	@Override
	public abstract I18NStringWithFillIns getToolTip(final GameView gameView);

	@Override
	public abstract void paint(final Graphics graphics, final Point offset, final ZoomLevel zoomLevel, final TopologyMode topologyMode);

	/** 
	 * @return the unique key for this commandable 
	 */
	public abstract CommandableKey getKey();

	/** 
	 * @return the pure name of the commandable, without any modifications for I18N. May be <code>null</code>.
	 */
	public abstract String getNickName();

	/** 
	 * @param name the pure name of the commandable, without any modifications for I18N. May be <code>null</code>.
	 */
	public abstract void setNickName(final String name);

	/** 
	 * @return <code>true</code> if the unit can be named and has received no name so far
	 */
	public abstract boolean needsNickName();

	/** 
	 * @return a suggestion for the pure name of the commandable. May be <code>null</code>, if there is no suggestion. 
	 */
	public abstract String getNickNameSuggestion();

	/**
	 * @param observed
	 *            for what military branch to view the sighting range shall be
	 *            looked for
	 * @return how far this commandable can see other commandables (or
	 *         {@link MilitaryBranch#NO_UNIT} for fields)
	 */
	public abstract int getSightingRange(final MilitaryBranch observed);

	/** 
	 * @return if this commandable has seen enemies since last turn
	 */
	public abstract boolean hasSeenEnemiesLately();

	/**
	 * @param value
	 *            <code>true</code> sets for this commandable that it has seen
	 *            enemies since last turn; <code>false</code> clears this
	 */
	public abstract void setSeenEnemiesLately(final boolean value);

	/** 
	 * Is called by the game engine when the turn of the player owning this unit starts. 
	 */
	public abstract void startTurn();

	/** 
	 * Is called by the game engine when the turn of the player owning this unit ends. 
	 */
	public abstract void finishTurn();

	/**
	 * @return <code>true</code> if the commandable should be considered in the
	 *         list of units presented to the player this turn
	 */
	public abstract boolean isNextable();

	/** 
	 * @return <code>true</code> if the commandable was presented to the player in this round of nexting 
	 */
	public abstract boolean wasNexted();

	/** 
	 * Called when all commandables were presented to the player. Initiates a new round of nexting. 
	 */
	public abstract void resetNext();

	/** 
	 * Is called when the system presents this commandable to the player. 
	 */
	public abstract void nextFocusGained();

	/** 
	 * Is called when the system presented this commandable to the player and now switches to another commandable. 
	 */
	public abstract void nextFocusLost();

	/**
	 * @param opponent
	 *            the opponent to attack
	 * @return if and how this unit can attack the opponent
	 */
	public abstract AttackMode getAttackModeAgainst(final Commandable opponent);
	
	public abstract float getFightStrength(final Commandable opponent, final boolean isAttacker);
	
	public abstract int getAttack();
	public abstract int getDefense();
	public abstract int getShotDamage();

	/** 
	 * @return if the commandable is still alive
	 */
	public abstract boolean isAlive();

	/** 
	 * @return if the commandable has been hit and needs repair
	 */
	public abstract boolean isDamaged();

	/**
	 * @return the number of hits this commandable can take before it is
	 *         killed/conquered
	 */
	public abstract int getHitsTillDeath();

	/**
	 * Changes the number of hits this commandable can take before it is
	 * killed/conquered.
	 * 
	 * @param numberOfHits
	 *            if positive, gets this commandable nearer to death; if
	 *            negative, heals the commandable
	 */
	public abstract void receiveHits(final int numberOfHits);

	/**
	 * Only makes sense for units, as only those can move.
	 * 
	 * @return if this commandable has moved this turn
	 */
	abstract boolean hasMovedThisTurn();
	abstract void setSkippingThisTurn(boolean value);

	/**
	 * Only makes sense for units, there commands lasting longer than one turn
	 * can be given.
	 * 
	 * @return the order this commandable is following, or <code>null</code> if
	 *         there is none
	 */
	public abstract Order getOrder();

	/**
	 * @param order
	 *            the order this commandable is following, or <code>null</code>
	 *            to unset any order
	 */
	public abstract void setOrder(final Order order);

	/**
	 * Evaluates the order this commandable is following.
	 * 
	 * @param gameView
	 *            the game view
	 * @param isNow
	 *            if <code>false</code>, it is only checked if this commandable
	 *            could sometimes take this unit; if <code>true</code>, it is
	 *            checked if this commandable can take this unit now, so it
	 *            takes e.g. already loaded units into account.
	 * @return the move result for the order
	 */
	abstract MoveResult evaluateOrder(final GameView gameView, final boolean isNow);

	/**
	 * Defines if this commandable can take the given unit type.
	 * 
	 * @param unitToLoad
	 *            the unit to load
	 * @param isNow
	 *            if <code>false</code>, it is only checked if this commandable
	 *            could sometimes take this unit; if <code>true</code>, it is
	 *            checked if this commandable can take this unit now, so it
	 *            takes e.g. already loaded units into account.
	 * @return if this commandable can load the given unit type
	 */
	abstract boolean canLoad(final UnitAlike unitToLoad, final boolean isNow);
}
