package de.ped.empire.logic;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;

/**
 * A unique key to identify a commandable. Necessary for marshalling and demarshalling, as direct references can't be marshalled.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class CommandableKey implements Marshallable, Cloneable, Comparable<CommandableKey> {

	/**
	 * A factory class for demarshalling of {@link CommandableKey} objects.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Factory implements Marshallable {
		
		public static CommandableKey create(final Marshaller marshaller) {
			return new CommandableKey(marshaller);
		}

		@Override
		public void marshal(final Marshaller marshaller) {
			throw new IllegalArgumentException("Factory is never marshalled.");
		}

		@Override
		public Marshallable demarshal(final Marshaller marshaller) {
			return create(marshaller);
		}
	}

	public static final Factory factory = new Factory();

	public static CommandableKey NULL_KEY = new CommandableKey(ImageKey.CMDL_ID_UNDEFINED, Player.ID_NONE, 0);
	
//	Started mem optimization to make the key fit in a single int, but seems not worth the effort
//	/** Contains the player color ID, including the ID for no player. Only used for commandables. */
//	public static final BitArea BA_PLAYER_ID = new BitArea( 0, 4);
//	/** Contains the info which template ID (depending on the ttype) to apply. */
//	public static final BitArea BA_TEMPLATE = new BitArea(4, 4);
//	/** The id of a unit is unique per game, player and unit type, while the id of a city is unique per game. */
//	public static final BitArea BA_ID = new BitArea(8, 24);

	private final int templateAndPlayerId;
	/** The id of a unit is unique per game and player. */
	private final int id;

	public CommandableKey(final int templateId, final int playerId, final int id) {
		final ImageKey imageKey = new ImageKey(ImageKey.TType.Unit, templateId);
		imageKey.setBits(ImageKey.BA_PLAYER_COLOR, playerId);
		this.templateAndPlayerId = imageKey.getAll();
		this.id = id;
	}

	public CommandableKey(final Marshaller marshaller) {
		this.templateAndPlayerId = marshaller.readInt();
		this.id = marshaller.readInt();
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(templateAndPlayerId);
		marshaller.writeInt(id);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw new IllegalStateException("Constant objects can't be demarshalled after creation - use constructor instead.");
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + templateAndPlayerId;
		result = prime * result + id;
		return result;
	}

	public int getTemplateId() {
		return ImageKey.getBits(templateAndPlayerId, ImageKey.BA_TEMPLATE);
	}

	public int getPlayerId() {
		return ImageKey.getBits(templateAndPlayerId, ImageKey.BA_PLAYER_COLOR);
	}

	public int getId() {
		return id;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		return (0 == compareTo((CommandableKey) obj));
	}

	/**
	 * Compares by template ID, player ID (ignored for cities, as that may change over time) and ID, in this order. 
	 */
	@Override
	public int compareTo(CommandableKey other) {
		if (templateAndPlayerId != other.templateAndPlayerId) {
			final int template = ImageKey.getBits(templateAndPlayerId, ImageKey.BA_TEMPLATE);
			final int otherTemplate = ImageKey.getBits(other.templateAndPlayerId, ImageKey.BA_TEMPLATE);
			if ((template == otherTemplate) && (template == ImageKey.CMDL_ID_CITY)) {
				// If it's a city, don't care about the player id, as that may change over time.
				return (template - otherTemplate);
			}
			return (templateAndPlayerId - other.templateAndPlayerId);
		}
		if (id != other.id) {
			return (id - other.id);
		}
		return 0;
	}

	@Override
	public CommandableKey clone() {
		try {
			final CommandableKey ret = (CommandableKey) super.clone();
			return ret;
		} catch (final CloneNotSupportedException e) {
			throw new IllegalStateException("Clone didn't work", e);
		}
	}

	@Override
	public String toString() {
		final int templateId = getTemplateId();
		final CommandableTemplate template = CommandableTemplate.findTemplateById(templateId);
		return "(t=" + templateId + " " + template.getLogRepresentation() + ",p=" + getPlayerId() + ",id=" + id + ")";
	}
}
