package de.ped.empire.logic;

import java.util.LinkedList;
import java.util.List;

import de.ped.empire.logic.Commandable.AttackMode;
import de.ped.tools.Assert;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.Bitfield32;
import de.ped.tools.TextDocumentBuilder;

/**
 * A class describing all the details of a military unit class.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class CommandableTemplate extends TileTemplate {

	/**
	 * Defines all parameters which might depend on the rules complexity.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	static class RulesComplexityDependentValues {
		/**
		 * Constructor.
		 * 
		 * @param loadDefinitions
		 *            the kind of load this unit may be able to carry
		 * @param hitConsequences
		 *            the length of this array tells how many hits this unit can
		 *            take. In the second dimension, the length is always 2,
		 *            listing the values which depend on the number of hits
		 *            taken, with index 0 keeping moves per turn possible, and
		 *            index 1 keeping the cargo hold size.
		 * @param stepsTillDeath
		 *            how many steps the unit can move before it will die of
		 *            fuel exhaustion. Only used for air units so far. For all
		 *            other units, {@link Integer#MAX_VALUE} indicates that this
		 *            value is not applicable.
		 * @param sightingRange
		 *            the sighting range of the unit for land and sea units. 
		 *            Even in a beginner game without fog of war this value makes sense, 
		 *            as it defines the warning for enemy sightings.
		 * @param sightingRangeOnAirUnits
		 *            the sighting range of the unit for air units 
		 * @param bitmaskAttackWithoutMove
		 *            a bitmask about the units this unit can attack with attack mode WITHOUT_MOVE
		 * @param bitmaskAttackIncludingMove
		 *            a bitmask about the units this unit can attack with attack mode INCLUDING_MOVE
		 * @param factorRuleE
		 *            advanced game only - attack ratings are x 1/2 when
		 *            attacking air units. 
		 *            See manual page 135 rule e for details.
		 */
		RulesComplexityDependentValues(
				final LoadDefinition[] loadDefinitions, 
				final int[][] hitConsequences, 
				final int stepsTillDeath, 
				final int sightingRange,
				final int sightingRangeOnAirUnits,
				final int bitmaskAttackWithoutMove,
				final int bitmaskAttackIncludingMove,
				final float factorRuleE) {
			this.loadDefinitions = loadDefinitions;
			this.hitConsequences = hitConsequences;
			this.stepsTillDeath = stepsTillDeath;
			this.sightingRange = sightingRange;
			this.sightingRangeOnAirUnits = sightingRangeOnAirUnits;
			this.bitmaskAttackWithoutMove = bitmaskAttackWithoutMove;
			this.bitmaskAttackIncludingMove = bitmaskAttackIncludingMove;
			this.factorRuleE = factorRuleE;
		}

		boolean isEmpty() {
			return ((null == hitConsequences) || (0 == hitConsequences.length));
		}

		private final int[][] hitConsequences;
		final int stepsTillDeath;
		private final int sightingRange;
		private final int sightingRangeOnAirUnits;
		final int bitmaskAttackWithoutMove;
		final int bitmaskAttackIncludingMove;
		private final LoadDefinition[] loadDefinitions;
		public final float factorRuleE;	// See manual page 135
	}

	static int MAP_ICONS_Y_BASE = 8;

	/**
	 * @param rulesComplexity
	 *            the rule complexity to list the unit types for
	 * @return all real unit types, meaning units that could move around and
	 *         being produced by cities
	 */
	public static List<CommandableTemplate> allRealUnitTypes(final RulesComplexity rulesComplexity) {
		LinkedList<CommandableTemplate> ret = new LinkedList<CommandableTemplate>();
		for (CommandableTemplate template : Templates.instance().allCommandableTemplates) {
			if ((0 < template.getMaximumStepsPerTurn(rulesComplexity)) 
					&& (0 < template.turnsToProduce)) {
				ret.add(template);
			}
		}
		return ret;
	}

	public static List<CommandableTemplate> allUnitsCitiesCanProduce(final RulesComplexity rulesComplexity) {
		List<CommandableTemplate> list = allRealUnitTypes(rulesComplexity);
		list.add(Templates.instance().NOTHING);
		return list;
	}

	public static List<CommandableTemplate> allUnitsToShowStatsFor(final RulesComplexity rulesComplexity) {
		List<CommandableTemplate> list = allRealUnitTypes(rulesComplexity);
		final CommandableTemplate airbase = Templates.instance().AIRBASE;
		if (0 < airbase.getStepsTillDeath(rulesComplexity)) {
			list.add(airbase);
		}
		return list;
	}

	/** The unique ID for this unit template. */
	public final int id;
	/** The kind of unit. */
	public final MilitaryBranch militaryBranch;
	/** The i18n prefix for this kind of unit. */
	public final String messageKeyPrefix;
	/** The i18n string for this kind of unit. */
	private final I18NStringWithFillIns name;
	/** If this kind of unit might be able to get the focus. */
	public final boolean isFocusable;
	/** How many turns does it take a standard city to produce this kind of unit. */
	public final int turnsToProduce;
	/** What other units can see this unit? */
	public final int visibilityBitmask;

	private final RulesComplexityDependentValues[] rcdvs;

	public final String imageResourceName;
	public final String promptSoundFileName;
	public final String alertImpedimentSoundFileName;
	public final String alertSightingSoundFileName;
	public final String moveSoundFileName;
	public final String fightIntroSoundFileName;
	public final String fightHitSoundFileName;
	public final String deathSoundFileName;
	private final String nameSuggestionsFileName;
	private final String[] nameSuggestionsDefaults;
	private NameSuggestion nameSuggestions; 
	/** Is this unit able to attack other units on coastal fields? */
	public final int attack;
	public final int defense;
	public final int shotDamage;
	/** This value helps to define which target to attack if there are several on the same field. A lower value is preferred. */
	public final int attackPriority;
	public final boolean canMoveOver;

	CommandableTemplate(
			final int id, 
			final MilitaryBranch militaryBranch, 
			final boolean isFocusable, 
			final String messageKeyPrefix, 
			final int turnsToProduce, 
			final RulesComplexityDependentValues[] rcdvs,
			final int visibilityBitmask,
			final boolean canMoveOver,
			final int attack,
			final int defense,
			final int shotDamage,
			final int attackPriority,
			final boolean hasImageResource,
			final String promptSoundFileName,
			final String alertImpedimentSoundFileName,
			final String alertSightingSoundFileName,
			final String moveSoundFileName,
			final String fightIntroSoundFileName,
			final String fightHitSoundFileName,
			final String deathSoundFileName,
			final String nameSuggestionsFileName,
			final String[] nameSuggestionsDefaults) {
		initImageKey(ImageKey.TType.Unit, id);
		this.id = id;
		this.militaryBranch = militaryBranch;
		this.isFocusable = isFocusable;
		this.messageKeyPrefix = messageKeyPrefix;
		this.name = new I18NStringWithFillIns(getNameKey(), null, null);
		this.turnsToProduce = turnsToProduce;
		this.rcdvs = rcdvs;
		this.visibilityBitmask = visibilityBitmask;
		this.canMoveOver = canMoveOver;
		if (hasImageResource) {
			this.imageResourceName = messageKeyPrefix + ".png";
		} else {
			this.imageResourceName = null;
		}
		this.promptSoundFileName = promptSoundFileName;
		this.alertImpedimentSoundFileName = alertImpedimentSoundFileName;
		this.alertSightingSoundFileName = alertSightingSoundFileName;
		this.moveSoundFileName = moveSoundFileName;
		this.fightIntroSoundFileName = fightIntroSoundFileName;
		this.fightHitSoundFileName = fightHitSoundFileName;
		this.deathSoundFileName = deathSoundFileName;
		this.nameSuggestionsFileName = nameSuggestionsFileName;
		this.nameSuggestionsDefaults = nameSuggestionsDefaults;
		this.attack = attack;
		this.defense = defense;
		this.shotDamage = shotDamage;
		this.attackPriority = attackPriority;
	}

	public int getHitsTillDeath(final RulesComplexity rulesComplexity) {
		Assert.assertNotNull(rulesComplexity);
		final RulesComplexityDependentValues rulesComplexityDependingValues = getRulesComplexityDependingValues(rulesComplexity);
		int ret = 0;
		if (null != rulesComplexityDependingValues.hitConsequences) {
			ret = rulesComplexityDependingValues.hitConsequences.length;
		}
		return ret;
	}

	public int getStepsPerTurn(final RulesComplexity rulesComplexity, final int hitsTillDeath) {
		return getHitsConsequencesValue(rulesComplexity, hitsTillDeath, 0);
	}

	public int getCapacity(final RulesComplexity rulesComplexity, final int hitsTillDeath) {
		return getHitsConsequencesValue(rulesComplexity, hitsTillDeath, 1);
	}

	public int getMaximumStepsPerTurn(final RulesComplexity rulesComplexity) {
		int ret = 0;
		final int hits = getHitsTillDeath(rulesComplexity);
		if (0 < hits) {
			ret = getStepsPerTurn(rulesComplexity, hits);
		}
		return ret;
	}

	public int getMaximumCapacity(final RulesComplexity rulesComplexity) {
		return getCapacity(rulesComplexity, getHitsTillDeath(rulesComplexity));
	}

	private int getHitsConsequencesValue(final RulesComplexity rulesComplexity, final int hitsTillDeath, final int index) {
		int ret = 0;
		final int[][] hitConsequences = getRulesComplexityDependingValues(rulesComplexity).hitConsequences;
		if ((null != hitConsequences) && (0 < hitConsequences.length) && (0 < hitsTillDeath)) {
			ret = hitConsequences[hitsTillDeath - 1][index];
		}
		return ret;
	}

	/**
	 * @param rulesComplexity
	 *            the ruleset to apply
	 * @param militaryBranch
	 *            the military branch of objects to sight, or
	 *            {@link MilitaryBranch#NO_UNIT} for fields
	 * @return how far this commandable can see other commandables of the given
	 *         military branch
	 */
	public int getSightingRange(final RulesComplexity rulesComplexity, final MilitaryBranch militaryBranch) {
		final RulesComplexityDependentValues rcdv = getRulesComplexityDependingValues(rulesComplexity);
		final int ret;
		switch(militaryBranch) {
		case AIR_UNIT:
			ret = rcdv.sightingRangeOnAirUnits;
			break;
		case SEA_UNIT:
		case LAND_UNIT:
		case NO_UNIT:	// For fields
		default:
			ret = rcdv.sightingRange;
			break;
		}
		return ret;
	}

	public int getStepsTillDeath(final RulesComplexity rulesComplexity) {
		return getRulesComplexityDependingValues(rulesComplexity).stepsTillDeath;
	}

	public boolean isDefinedFor(final RulesComplexity rulesComplexity) {
		return !getRulesComplexityDependingValues(rulesComplexity).isEmpty();
	}

	AttackMode getAttackModeAgainst(final RulesComplexity rulesComplexity, final CommandableTemplate opponentTemplate) {
		final RulesComplexityDependentValues rcdv = getRulesComplexityDependingValues(rulesComplexity);
		final AttackMode ret;
		if (Bitfield32.isBit(rcdv.bitmaskAttackIncludingMove, opponentTemplate.id)) {
			ret = AttackMode.INCLUDING_MOVE;
		} else if (Bitfield32.isBit(rcdv.bitmaskAttackWithoutMove, opponentTemplate.id)) {
			ret = AttackMode.WITHOUT_MOVE;
		} else {
			ret = AttackMode.IMPOSSIBLE;
		}
		return ret;
	}

	/**
	 * Returns if this unit could attack units on coastal land.
	 * <p>
	 * See Empire Deluxe Manual page 135, rule f: This unit may conduct shore
	 * bombardment.
	 * 
	 * @param rulesComplexity
	 *            the ruleset to apply
	 * @return if this unit could attack units on coastal land
	 */
	public boolean canCoastalAttack(final RulesComplexity rulesComplexity) {
		final boolean ret;
		if (militaryBranch.equals(MilitaryBranch.SEA_UNIT)) {
			if (AttackMode.WITHOUT_MOVE == getAttackModeAgainst(rulesComplexity, Templates.instance().ARMY)) {
				ret = true;
			} else {
				ret = false;
			}
		} else {
			ret = false;
		}
		return ret;
	}

	/**
	 * @param rulesComplexity
	 *            the ruleset to apply
	 * @param cargo
	 *            the kind of unit to hold
	 * @return how many cargo holds one unit of <em>cargo</em> takes within this
	 *         unit, or {@link Integer#MAX_VALUE} to indicate that this unit
	 *         can't take any of <em>cargo</em>
	 */
	public int getCargoHoldsFor(final RulesComplexity rulesComplexity, final CommandableTemplate cargo) {
		int ret = Integer.MAX_VALUE;
		final LoadDefinition[] loadDef = getRulesComplexityDependingValues(rulesComplexity).loadDefinitions;
		if (null != loadDef) {
			for (final LoadDefinition def : loadDef) {
				if (def.getUnitType().equals(cargo)) {
					ret = def.holdsTaken;
					break;
				}
			}
		}
		return ret;
	}

	/**
	 * @param rulesComplexity
	 *            the ruleset to apply
	 * @param cargo
	 *            the kind of unit to hold
	 * @return if a unit of this type is able to hold a unit of type
	 *         <em>cargo</em>
	 */
	public boolean canHold(final RulesComplexity rulesComplexity, final CommandableTemplate cargo) {
		final boolean ret = (Integer.MAX_VALUE != getCargoHoldsFor(rulesComplexity, cargo));
		return ret;
	}

	/**
	 * @param gameView
	 *            the game view to apply
	 * @return if this kind of unit can offer the load command
	 */
	public boolean canLoad(final GameView gameView) {
		final LoadDefinition[] loadDef = rcdvs[gameView.getProperties().getRulesComplexity().ordinal()].loadDefinitions;
		return ((null != loadDef) && (0 < loadDef.length));
	}

	/** 
	 * @param gameView
	 *            the game view to apply
	 * @return if this kind of unit can dock
	 */
	public boolean canDock(final GameView gameView) {
		final RulesComplexity rulesComplexity = gameView.getProperties().getRulesComplexity();
		return ((1 < getHitsTillDeath(rulesComplexity)) 
				|| (Integer.MAX_VALUE != getStepsTillDeath(rulesComplexity)));
	}

	public String getNameKey() {
		return messageKeyPrefix + Messages.TEXT_POSTFIX;
	}

	public String getShortKey() {
		return messageKeyPrefix + ".Short";
	}

	public String getMnemonicKey() {
		return messageKeyPrefix + Messages.MNEMONIC_POSTFIX;
	}

	@Override
	public I18NStringWithFillIns getName() {
		return name;
	}

	/**
	 * @param doc
	 *            the document to create the text for; is left untouched
	 * @param rulesComplexity
	 *            the ruleset to apply
	 * @return an i18n'ized string explaining what kind and amount of cargo this
	 *         unit can hold
	 */
	public I18NStringWithFillIns getCargoString(final TextDocumentBuilder doc, final RulesComplexity rulesComplexity) {
		final I18NStringWithFillIns  ret;
		final int capacity = getMaximumCapacity(rulesComplexity);
		if (0 < capacity) {
			ret = new I18NStringWithFillIns("Game.Unit.Cargo.Hold", null, new Object[]{
					Integer.toString(capacity),
					getCargoListString(doc, rulesComplexity),
			});
		} else {
			ret = new I18NStringWithFillIns("Game.Unit.Cargo.None", null, null);
		}
		return ret;
	}

	private I18NStringWithFillIns getCargoListString(final TextDocumentBuilder doc, final RulesComplexity rulesComplexity) {
		final StringBuffer ret = new StringBuffer();
		final LoadDefinition[] loadDef = getRulesComplexityDependingValues(rulesComplexity).loadDefinitions;
		I18NStringWithFillIns[] parts = null; 
		if ((null != loadDef) && (0 < loadDef.length)) {
			Assert.assertTrue(loadDef.length <= 10);
			parts = new I18NStringWithFillIns[loadDef.length]; 
			for (int i = 0; i < loadDef.length; i++) {
				final LoadDefinition load = loadDef[i];
				if (0 < i) {
					ret.append(", ");
				}
				ret.append("%" + i);
				final CommandableTemplate unitType = load.getUnitType();
				parts[i] = new I18NStringWithFillIns(null, "%0&nbsp;(%1)", 
						new Object[] { 
						doc.tf().href("#" + unitType.messageKeyPrefix, doc.messages().getText(unitType.getName())), 
						Integer.toString(load.holdsTaken), });
			}
		}
		return new I18NStringWithFillIns(null, ret.toString(), parts);
	}

	RulesComplexityDependentValues getRulesComplexityDependingValues(final RulesComplexity rulesComplexity) {
		return rcdvs[rulesComplexity.ordinal()];
	}

	public String getNameSuggestion() {
		if (null == nameSuggestions) {
			this.nameSuggestions = new NameSuggestion(Resources.loader(), nameSuggestionsFileName, nameSuggestionsDefaults);
		}
		return nameSuggestions.getNameSuggestion();
	}

	public boolean isUndefinedOrNothing() {
		return (MilitaryBranch.NO_UNIT == militaryBranch);
	}

	@Override
	public String toString() {
		return messageKeyPrefix;
	}

	public static CommandableTemplate findTemplateById(int unitId) {
		if (null == cmdlTemplatesMap) {
			CommandableTemplate[] map = null;
			map = new CommandableTemplate[16];
			for (CommandableTemplate template : Templates.instance().allCommandableTemplates) {
				map[template.id] = template;
			}
			cmdlTemplatesMap = map;
		}
		CommandableTemplate ret = cmdlTemplatesMap[unitId];
		if (null == ret) {
			ret = Templates.instance().NOTHING;
		}
		return ret;
	}

	private static CommandableTemplate[] cmdlTemplatesMap = null;

	Commandable getPreferredAttackTarget(Commandable target1, Commandable target2) {
		final Commandable ret;
		if (null == target1) {
			ret = target2;
		} else if (null == target2) {
			ret = target1;
		} else if (target1 instanceof FieldCity) {
			ret = target1;
		} else if (target2 instanceof FieldCity) {
			ret = target2;
		} else {
			CommandableTemplate t1 = ((Unit) target1).getUnitTemplate();
			CommandableTemplate t2 = ((Unit) target2).getUnitTemplate();
			if (t1.attackPriority <= t2.attackPriority) {
				ret = target1;
			} else {
				ret = target2;
			}
		}
		return ret;
	}
}
