package de.ped.empire.logic;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;

import de.ped.tools.Assert;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A very easy computer strategy. It's main purpose is to test the strategy interface.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ComputerStrategySimple implements GameObserver {

	/**
	 * A relevant part of the commandable to prevent double-visits of the same commandable within a turn.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class UnitStep {
		public UnitStep(final Commandable commandable) {
			this.commandableKey = commandable.getKey();
			if (commandable instanceof Unit) {
				final Unit unit = (Unit) commandable;
				this.stepsLeftThisTurn = unit.getStepsLeftThisTurn();
			} else {
				this.stepsLeftThisTurn = 0;
			}
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime
					* result
					+ ((commandableKey == null) ? 0 : commandableKey.hashCode());
			result = prime * result + stepsLeftThisTurn;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			UnitStep other = (UnitStep) obj;
			if (commandableKey == null) {
				if (other.commandableKey != null)
					return false;
			} else if (!commandableKey.equals(other.commandableKey))
				return false;
			if (stepsLeftThisTurn != other.stepsLeftThisTurn)
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "(" + commandableKey.toString() + ",step=" + stepsLeftThisTurn + ")";
		}

		private final CommandableKey commandableKey;
		private final int stepsLeftThisTurn;
	}

	/**
	 * Defines the states in which commandables are revisited in getting a new command.
	 * 
	 * @author Copyright (c) 2016 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private enum RevisitState {
		/** This commandable has been revisited so often, the turn should end. */
		FINISH,
		/** This commandable is visited for the first time. */
		VISIT,
		/** This commandable is re-visited, so special rules should apply to avoid endless loops. */
		REVISIT,
	}

	/**
	 * Defines an interface for the {@link CommandableRule} class.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static interface Applyable {
		/**
		 * Defines a rule that can be applied to a commandable.
		 * @param strategy the computer strategy object in which to act
		 * @param gameView the game view in which to act
		 * @param commandable the commandable to check the rule for
		 * @return the order to give to <em>commandable</em>, or <code>null</code> if this rule should not impact <em>commandable</em>
		 */
		public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable);
	}

	/**
	 * Defines all the rules that can be applied to a unit by a computer
	 * strategy. The order in which the rules are defined implies a preference;
	 * the first rule that returns a command for a commandable wins.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private enum CommandableRule implements Applyable {
		/**
		 * Rule for all units that need healing.
		 */
		HEAL {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable.isDamaged()) {
					if (gameView.getAt(commandable.getPosition()).getPlayerId() == commandable.getPlayerId()) {
						// I'm already at home, stay there
						ret = new OrderSkipMove(gameView.getGameState(), commandable);
					} else {
						ret = new OrderHome(gameView.getGameState(), commandable);
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for all units low on fuel.
		 */
		REFUEL {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					if (unit.isLowOnFuel()) {
						ret = new OrderHome(gameView.getGameState(), commandable);
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for units that can conquer cities and know an enemy city nearby.
		 */
		CONQUER_CITY_BY_LAND {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					if (MilitaryBranch.LAND_UNIT == unit.getUnitTemplate().militaryBranch) {
						PlayfieldPosition pos = strategy.findNearestEnemyCity(gameView, unit);
						if (null != pos) {
							ret = gameView.getMoveOrAttackOrder(unit, pos);
						}
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for transporters:
		 * The more cargo, the more likely to directly move to next non-own reachable (that means coastal) city.
		 */
		TRANSPORT_CONQUER_CITY_BY_SEA {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					if (Templates.instance().TRANSPORTER == unit.getUnitTemplate()) {
						final int remaining = unit.calcRemainingCapacity();
						final int max = Templates.instance().TRANSPORTER.getMaximumCapacity(gameView.getGameState().getProperties().getRulesComplexity());
						if (MathUtil.random(max) <= remaining) {
							// If with load, move to nearest enemy city
							PlayfieldPosition pos = strategy.findNearestEnemyCity(gameView, unit);
							if (null != pos) {
								ret = gameView.getMoveOrAttackOrder(unit, pos);
							}
						}
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for transporters:
		 * <ul>
		 *   <li>TODO: otherwise find units to load and move there
		 * </ul>
		 */
		TRANSPORT_LOAD {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					if (Templates.instance().TRANSPORTER == unit.getUnitTemplate()) {
						final List<FieldAlike> landUnitFields = new LinkedList<FieldAlike>();
						// Finds the nearest land unit to load.
						// TODO: Optimization: If there is more than one, rate this direction higher.
						for (final UnitAlike target : gameView.getUnits()) {
							final CommandableKey targetKey = target.getKey();
							if ((targetKey.getPlayerId() == unit.getKey().getPlayerId()) 
									&& unit.canLoad(target, true)) {
								landUnitFields.add(gameView.getAt(target.getPosition()));
							}
						}
						final PlayfieldPosition pos = strategy.findNearestOf(unit, landUnitFields);
						if (null != pos) {
							// FIXME: Find a landing spot, more there, and move the units there as well.
							ret = gameView.getMoveOrAttackOrder(unit, pos);
						}
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for land units:
		 * If a non-full transporter is reachable, move there.
		 */
		GET_LOADED {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					if (MilitaryBranch.LAND_UNIT == unit.getUnitTemplate().militaryBranch) {
						// Find nearest transporter to load
						final List<FieldAlike> transporterFields = new LinkedList<FieldAlike>();
						for (final UnitAlike target : gameView.getUnits()) {
							final CommandableKey targetKey = target.getKey();
							if ((targetKey.getPlayerId() == unit.getKey().getPlayerId()) 
									&& target.canLoad(unit, true)) {
								transporterFields.add(gameView.getAt(target.getPosition()));
							}
						}
						final PlayfieldPosition pos = strategy.findNearestOf(unit, transporterFields);
						if (null != pos) {
							ret = gameView.getMoveOrAttackOrder(unit, pos);
						}
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for all units:
		 * <ul>
		 *   <li>Attack a near enemy unit, depending on the probability of the first hit.
		 * </ul>
		 */
		FIGHT {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					UnitAlike opponent = null;
					unit.calculateForwardRating(StepRatingPurpose.NORMAL);
					float probability = 0.0f;
					for (final UnitAlike oppCandidate : gameView.getUnits()) {
						if (oppCandidate.getPlayerId() != unit.getPlayerId()) {
							final int dist = gameView.getPlayerMap().getProperties().calcAirDistance(unit.getPosition(), oppCandidate.getPosition());
							if (dist <= MAX_DISTANCE_FOR_ATTACK) {
								StepRatingPosition srp = unit.getStepRating(oppCandidate.getPosition());
								strategy.logger.debug("ruleForFighting found air dist " + dist + " between " + unit.getPosition().toString() + " and " + oppCandidate.getPosition().toString() + "; steprating=" + srp);
								if ((null != srp) && (srp.distanceFromStart <= MAX_DISTANCE_FOR_ATTACK)) {
									final float own = unit.getFightStrength(oppCandidate, true);
									final float opp = oppCandidate.getFightStrength(unit, false);
									final float prob = own / (own + opp);
									if (prob > probability) {
										opponent = oppCandidate;
										probability = prob;
									}
								}
							}
						}
					}
					if (MathUtil.random(1.0f) <= probability) {
						final PlayfieldPosition oppPos = opponent.getPosition();
						ret = gameView.getMoveOrAttackOrder(unit, oppPos);
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for all units:
		 * <ul>
		 *   <li>Explore the world.
		 * </ul>
		 */
		EXPLORE {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if ((commandable instanceof Unit) 
						&& (gameView.getProperties().getRulesComplexity().isWithFogOfUncertainty)) {
					final Unit unit = (Unit) commandable;
					final Order order = new OrderExplore(gameView.getGameState(), unit);
					if (order.makesSense() && (null != order.evaluate(true))) {
						ret = order;
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for cities:
		 * If on the same continent there is an enemy city, build army.
		 */
		CITY_BUILD_ARMIES {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				if (commandable instanceof FieldCity) {
					final FieldAlike nearest = strategy.existsEnemyCityOnSameIsland(gameView, (FieldCity) commandable);
					if (null != nearest) {
						final OrderChangeProduction ocp = new OrderChangeProduction(gameView.getGameState(), commandable);
						ocp.setUnitToProduce(Templates.instance().ARMY);
						ret = ocp;
					}
				}
				return ret;
			}
		},

		/**
		 * Rule for cities:
		 * If the continent is not fully explored, build either land or air unit.
		 * <ul>
		 *   <li>If the nearest unexplored field is within fuel range, and there is no air unit within fuel range, build air unit.
		 *   <li>Otherwise, if there are unexplored fields on this continent, build army. (There might be a city waiting to be discovered.)
		 * </ul>
		 */
		CITY_BUILD_EXPLORING_UNITS {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				CommandableTemplate unitToProduce = Templates.instance().NOTHING;
				if (gameView.isUsingFogOfWar() && (commandable instanceof FieldCity)) {
					// Find the best air unit template to pick
					final FieldCity city = (FieldCity) commandable;
					final List<CommandableTemplate> unitTemplates = city.getUnitsToProduce();
					final RulesComplexity rulesComplexity = gameView.getProperties().getRulesComplexity();
					int fuelDist = 0;
					for (final CommandableTemplate template : unitTemplates) {
						if (template.militaryBranch.equals(MilitaryBranch.AIR_UNIT)) {
							final int stepsTillDeath = template.getStepsTillDeath(rulesComplexity);
							if (fuelDist < stepsTillDeath) {
								fuelDist = stepsTillDeath;
								unitToProduce = template;
							}
						}
					}
					// Is there another air unit in fuel range?
					boolean isOwnUnitInFuelRange = false;
					for (final Commandable other : gameView.getGameState().findCommandables(gameView.getPlayer().getId())) {
						if (other instanceof Unit) {
							final Unit unit = (Unit) other;
							if (unit.getUnitTemplate().militaryBranch.equals(MilitaryBranch.AIR_UNIT)) {
								if (fuelDist / 2 <= gameView.getProperties().calcAirDistance(commandable.getPosition(), unit.getPosition())) {
									isOwnUnitInFuelRange = true;
									break;
								}
							}
						}
					}
					if (!Templates.instance().NOTHING.equals(unitToProduce) && !isOwnUnitInFuelRange) {
						// Is there an unexplored field in air distance?
						final int dist = strategy.calcAirDistanceToNearestOf(gameView, commandable, Templates.instance().UNEXPLORED, fuelDist / 2);
						if (0 < dist) {
							// All is fine, the unit will be produced.
						} else {
							unitToProduce = Templates.instance().NOTHING;
						}
					} else {
						unitToProduce = Templates.instance().NOTHING;
					}
					// So far with air units. Should we produce armies instead?
					if (Templates.instance().NOTHING.equals(unitToProduce)) {
						final boolean exists = strategy.existsUnexploredOnSameIsland(gameView, (FieldCity) commandable);
						if (exists) {
							unitToProduce = Templates.instance().ARMY;
						}
					}
					if (!Templates.instance().NOTHING.equals(unitToProduce)) {
						final OrderChangeProduction ocp = new OrderChangeProduction(gameView.getGameState(), commandable);
						ocp.setUnitToProduce(unitToProduce);
						ret = ocp;
					}
				}
				if (null != ret) {
					strategy.logger.debug("Applied rule to " + this.name() + "(" + unitToProduce + ") for " + commandable.getKey());
				}
				return ret;
			}
		},

		/**
		 * Rule for all units:
		 * Of all possible orders for this unit that can be fulfilled within this turn, pick one randomly.
		 */
		RANDOM {
			@Override
			public Order apply(final ComputerStrategySimple strategy, final GameView gameView, final Commandable commandable) { 
				Order ret = null;
				final List<Order> orders = new LinkedList<Order>();
				// Pick possible distant and steady orders
				final int maxDistance;
				if (commandable instanceof Unit) {
					maxDistance = ((Unit) commandable).getStepsLeftThisTurn();
				} else {
					maxDistance = 0;
				}
				List<FieldAlike> fields = gameView.getPlayerMap().getAllFieldsInAirDistance(gameView.getMap().getAt(commandable.getPosition()), maxDistance, true, 0);
				for (final FieldAlike field : fields) {
					List<Order> os = gameView.collectOrderOptions(field.getPosition());
					if (PlayfieldPosition.equals(field.getPosition(), commandable.getPosition())) {
						// steady orders
						for (final Order order : os) {
							if (order instanceof OrderSentry) {
								// Nothing to do
							} else {
								orders.add(order);
							}
						}
					} else {
						// distant orders
						orders.addAll(os);
					}
				}
				if (!orders.isEmpty()) {
					ret = orders.get(MathUtil.random(orders.size()));
					strategy.logger.debug("Applied random rule out of " + orders.size() + " possible for " + commandable.getKey());
				}
				return ret;
			}
		},
	};

	ComputerStrategySimple(final GameState gameState, final Messages messages) {
		this.gameState = gameState;
		this.messages = messages;
		logger.setToShowThread(true);
		logger.setLogLevel(Level.DEBUG);
	}

	protected FieldAlike existsEnemyCityOnSameIsland(final GameView gameView, final FieldCity city) {
		FieldAlike ret = null;
		final Unit pivot = new Unit(gameView.getGameState(), Templates.instance().ARMY, gameView.getPlayer().getId(), city);
		pivot.calculateForwardRating(StepRatingPurpose.CONQUER);
		final GameMap map = gameView.getMap();
		final GameProperties props = map.getProperties();
		final int width = map.getSize().getWidth();
		final PlayfieldPosition pos = new PlayfieldPosition(); // Re-used for performance reasons
		for (int x = 0; (x < width) && (null == ret); x++) {
			for (int y = props.getYMin(x); (y < props.getYMax(x)) && (null == ret); y++) {
				pos.setLocation(x, y);
				final FieldAlike field = map.getAt(pos);
				if ((null != pivot.getStepRating(pos))
						&& field.getTemplate().equals(Templates.instance().FIELD_CITY)
						&& (field.getPlayerId() != gameView.getPlayer().getId())) {
					ret = field;
				}
			}
		}
		return ret;
	}

	protected boolean existsUnexploredOnSameIsland(final GameView gameView, final FieldCity city) {
		boolean ret = false;
		final Unit pivot = new Unit(gameView.getGameState(), Templates.instance().ARMY, gameView.getPlayer().getId(), city);
		final Order order = new OrderExplore(gameView.getGameState(), pivot);
		if (order.makesSense() && (null != order.evaluate(true))) {
			ret = true;
		}
		return ret;
	}

	protected int calcAirDistanceToNearestOf(final GameView gameView, final Commandable commandable, final FieldTemplate fieldTemplate, final int fuelDist) {
		final int maxDistance = fuelDist;
		boolean found = false;
		int d;
		for (d = 1; (d <= maxDistance) && (!found); d++) {
			final List<FieldAlike> targetFieldCandidates = 
					gameView.getMap().getAllFieldsInAirDistance((FieldCity) commandable, d, false, 0);
			for (final FieldAlike targetField : targetFieldCandidates) {
				if (targetField.isUnexplored()) {
					found = true;
					break;
				}
			}
		}
		if (!found) {
			d = 0;
		}
		return d;
	}

	@Override
	public Messages getMessages() {
		return messages;
	}

	@Override
	public synchronized void notifyGameNotification(final int playerId, final GameNotification notif) {
		if ((null != notif) && (!notif.isEmpty())) {
			final GameView gameView = gameState.getGameView(playerId);
			logger.debug("notifyGameNotification " + notif);
			if (null != notif.getAnimationMoveResult()) {
				logger.error("TODO: Animation going on, clear it first");
			}
			if (notif.isEvent(GameNotification.MASK_PLAYER_IN_FOCUS)) {
				logger.trace("MASK_PLAYER_IN_FOCUS: " + gameState.getCurrentPlayerId());
			}
			if (notif.isEvent(GameNotification.MASK_MAP_MODE)) {
				logger.error("MASK_MAP_MODE: This event should never happen for computers?!?");
			}
			if (notif.isEvent(GameNotification.MASK_POSITION_IN_FOCUS)) {
				logger.trace("MASK_POSITION_IN_FOCUS: " + gameView.getPositionInFocus().toString());
			}
			if (notif.isEvent(GameNotification.MASK_STATUS_MESSAGE)) {
				logger.trace("MASK_STATUS_MESSAGE: " + gameView.getStatusMessage());
			}
			if (notif.isEvent(GameNotification.MASK_COMMANDABLE_IN_FOCUS)) {
				logger.trace("MASK_COMMANDABLE_IN_FOCUS: " + gameView.getCommandableInFocus());
//				setOrderToCommandableInFocus(gameView);
			}
		}
	}

	@Override
	public synchronized void step(final int playerId) {
		logger.debug("step");
		final GameView gameView = gameState.getGameView(playerId);
		/* final boolean hasFinishedTurn = */ setOrderToCommandableInFocus(gameView);
	}

	/**
	 * Tries to check and set new orders to the current commandable in focus.
	 * @param gameView the game view to work on
	 * @return <code>true</code> if this turn has ended for this player
	 */
	private boolean setOrderToCommandableInFocus(final GameView gameView) {
		boolean hasFinishedTurn = false;
		logger.trace("setOrderToCommandableInFocus(" + gameView + ")");
		if (gameView.isGetReady()) {
			gameView.clearGetReady();
		}
		final Commandable commandable = gameView.getVisibleCommandableInFocus();
		if ((null != commandable) && (commandable.isNextable())) {
			if (commandable.hasSeenEnemiesLately()) {
				commandable.setSeenEnemiesLately(false);
				if (commandable instanceof Unit) {
					commandable.setOrder(null);
				}
				gameView.nextCommandable();
			} else {
				final UnitStep us = new UnitStep(commandable);
				final RevisitState outcome;
				if (unitSteps.contains(us)) {
					final Order order = commandable.getOrder();
					if ((null != order) && (order instanceof OrderExplore)) {
						outcome = RevisitState.REVISIT;
					} else {
						outcome = RevisitState.FINISH;
					}
				} else {
					unitSteps.add(us);
					outcome = RevisitState.VISIT;
				}
				logger.trace("setOrderToCommandableInFocus: " + commandable.getKey().toString() + " outcome " + outcome.name());
				switch(outcome) {
				case FINISH:
					hasFinishedTurn = true;
					break;
				case VISIT:
				case REVISIT:
					final Order order = chooseOrder(gameView, commandable, outcome);
					logger.debug("setOrderToCommandableInFocus: Order chosen for " + commandable.getKey().toString() + " was " + order);
					if (null != order) {
						if (order instanceof OrderChangeProduction) {
							applyOrderChangeProduction(gameView, (FieldCity) commandable, (OrderChangeProduction) order);
							gameView.nextCommandable();
						} else {
							gameView.setOrder(order);
						}
					}
					break;
				default:
					// This should never happen
				}
			}
		} else if (gameView.isFinishTurnAllowed()) {
			hasFinishedTurn = true;
		}
		if (hasFinishedTurn) {
			logger.debug("setOrderToCommandableInFocus: nextPlayer");
			logger.debug("Visited units are: ");
			for (final UnitStep us : unitSteps) {
				logger.debug("- " + us.toString());
			}
			unitSteps.clear();
			if (shouldResign(gameView)) {
				gameView.resign();
				logger.info("I'm resigning!");
			}
			gameView.finishTurn();
		} else {
			gameView.nextMove();
		}
		return hasFinishedTurn;
	}

	/**
	 * Checks wheather this computer player should resign.
	 * Current implementation resigns if no city is left, and no land unit is left to capture a new city.
	 * 
	 * @param gameView
	 *            the game view to act in
	 * @return <code>true</code> iff the computer strategy should resign
	 */
	private boolean shouldResign(final GameView gameView) {
		boolean ret = false;
		boolean found = false;
		for (Commandable cmdl : gameView.getGameState().findCommandables(gameView.getPlayer().getId())) {
			final int templateId = cmdl.getKey().getTemplateId();
			if (ImageKey.CMDL_ID_CITY == templateId) {
				found = true;
			} else {
				final CommandableTemplate cmdlt = ((Unit) cmdl).getUnitTemplate();
				if (cmdlt.militaryBranch == MilitaryBranch.LAND_UNIT) {
					found = true;
				}
			}
			if (found) {
				break;
			}
		}
		if (!found) {
			ret = true;
		}
		return ret;
	}

	@Override
	public synchronized void notifyUnitProduced(final int playerId, final FieldCity city, final Unit unit) {
		logger.debug("notifyUnitProduced");
		final GameView gameView = gameState.getGameView(playerId);
		final Order order = chooseOrder(gameView, city, RevisitState.VISIT);
		applyOrderChangeProduction(gameView, city, order);
	}

	/**
	 * Applies a change production order to a city.
	 * @param gameView the game view to apply the change in
	 * @param city the city to change production for
	 * @param order the order to apply; must be of type {@link OrderChangeProduction}.
	 * 		If {@link OrderChangeProduction#getUnitToProduce()} is undefined or nothing,
	 * 		a unit template will be chosen randomly,
	 * 		otherwise that specified unit will be produced.
	 */
	private void applyOrderChangeProduction(final GameView gameView, final FieldCity city, final Order order) {
		final List<CommandableTemplate> unitsToProduce = city.getUnitsToProduce();
		final OrderChangeProduction ocp = (OrderChangeProduction) order;
		CommandableTemplate unitToProduce = ocp.getUnitToProduce();
		do {
			unitToProduce = unitsToProduce.get(MathUtil.random(unitsToProduce.size()));
		} while (unitToProduce.isUndefinedOrNothing());
		gameView.setCityProduction(city, unitToProduce);
	}

	private Order chooseOrder(final GameView gameView, final Commandable commandable, RevisitState revisitState) {
		Order ret = null;
		if (commandable.isAlive()) {
			Assert.assertEquals(gameView.getPlayer().getId(), commandable.getPlayerId());
			for (CommandableRule rule : CommandableRule.values()) {
				boolean isToTry = true;
				if ((CommandableRule.EXPLORE == rule) && (RevisitState.REVISIT == revisitState)) {
					isToTry = false;
				}
				if (isToTry) {
					ret = rule.apply(this, gameView, commandable);
					if (null != ret) {
						reasons.put(commandable, rule);
						logger.debug("Applied rule to " + rule.name() + " for " + commandable.getKey());
						break;
					}
				}
			}
		}
		return ret;
	}

	private PlayfieldPosition findNearestEnemyCity(final GameView gameView, final Unit unit) {
		final List<FieldCity> cities = unit.getGameState().getMap().getAllCities();
		final List<FieldAlike> fields = new LinkedList<FieldAlike>();
		for (final FieldCity city : cities) {
			if (city.getPlayerId() != gameView.getPlayer().getId()) {
				fields.add(city);
			}
		}
		return findNearestOf(unit, fields);
	}

	/**
	 * Returns the position of the nearest field for the given unit, based on the step rating.
	 * @param unit the unit to search the nearest field for
	 * @param fields the list of fields to take into account
	 * @return a single nearest position, or <code>null</code> if there is none
	 */
	private PlayfieldPosition findNearestOf(final Unit unit, final Collection<FieldAlike> fields) {
		final List<FieldAlike> nearest = unit.findNearestOf(fields);
		final int numOfNearest = nearest.size();
		PlayfieldPosition pos = null;
		if (0 < numOfNearest) {
			final FieldAlike field = nearest.get(MathUtil.random(numOfNearest));
			pos = field.getPosition();
		}
		return pos;
	}

	private static final int MAX_DISTANCE_FOR_ATTACK = 6;

	private final GameState gameState;

	private final Logger logger = Logger.getLogger(getClass());

	private Messages messages;

	private HashSet<UnitStep> unitSteps = new HashSet<UnitStep>();

	/** 
	 * Stores the reason why a commandable is doing what it's doing.
	 * This knowledge helps in overriding existing orders.
	 * TODO: So far unused. Might be helpful to match transporters to land units to load.
	 */
	private HashMap<Commandable, CommandableRule> reasons = new HashMap<Commandable, CommandableRule>();
}
