package de.ped.empire.logic;

import java.util.Collection;
import java.util.TreeSet;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.Bitfield32;

/**
 * The super class for anything that can be a tiled part of the world without being a military unit.
 * For each map, there can be only one such object per position.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Field extends TileTemplate implements FieldAlike {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	public static Field create(final Marshaller marshaller, final int x, final int y) {
		final Field ret;
		final int imageKey = marshaller.readInt();
		Bitfield32 bits = new Bitfield32(imageKey);
		switch(bits.getBits(ImageKey.BA_TEMPLATE)) {
		case ImageKey.FIELD_ID_UNEXPLORED:
		case ImageKey.FIELD_ID_BORDER:
		case ImageKey.FIELD_ID_SEA:
		case ImageKey.FIELD_ID_LAND:
		case ImageKey.FIELD_ID_ROUGH:
		case ImageKey.FIELD_ID_FOREST:
		case ImageKey.FIELD_ID_RIVER:
		case ImageKey.FIELD_ID_MOUNTAIN:
			ret = new Field(marshaller, x, y, imageKey);
			break;
		case ImageKey.FIELD_ID_CITY:
			ret = new FieldCity(marshaller, x, y, imageKey);
			break;
		default:
			throw new IllegalArgumentException("Unknown map type " + Integer.toHexString(imageKey));
		}
		return ret;
	}

	/** The position on the map of this field. Non-persistent. */
	private final transient PlayfieldPosition position;

	public Field(final int x, final int y, byte templateId) {
		position = new PlayfieldPosition(x, y);
		initImageKey(ImageKey.TType.Field, templateId);
		if (ImageKey.FIELD_ID_RIVER == templateId) {
			Bitfield32 bf = new Bitfield32(getImageKey());
			bf.setBits(ImageKey.BA_NEIGHBOR_MOD, 0xf);
			setImageKey(bf.getAll());
		}
	}

	protected Field(final Marshaller marshaller, final int x, final int y, final int imageKey) {
		setImageKey(imageKey);
		position = new PlayfieldPosition(x,  y);
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(getImageKey());
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	@Override
	public FieldTemplate getTemplate() {
		final FieldTemplate[] templates = Templates.instance().allGameMapFieldTemplates;
		final byte templateId = getTemplateId();
		FieldTemplate ret = null;
		for (final FieldTemplate temp : templates) {
			if (temp.id == templateId) {
				ret = temp;
				break;
			}
		}
		return ret;
	}

	@Override
	public byte getTemplateId() {
		return (byte) Bitfield32.getBits(getImageKey(), ImageKey.BA_TEMPLATE);
	}

	@Override
	public PlayfieldPosition getPosition() {
		return position;
	}

	@Override
	public I18NStringWithFillIns getName() {
		return getTemplate().name;
	}

	@Override
	public I18NStringWithFillIns getToolTip(final GameView gameView) {
		return getName();
		// Alternative implementation for debugging purposes:
//		return new I18NStringWithFillIns(null, "%0 (%1)", new Object[] { 
//				getTemplate().name,
//				Integer.toHexString(getImageKey()),
//		});
	}

	@Override
	public boolean isLand() {
		return getTemplate().isLand;
	}

	@Override
	public boolean isUnexplored() {
		return false;
	}

	@Override
	public boolean isCity() {
		return (getTemplate().id == ImageKey.FIELD_ID_CITY);
	}

	@Override
	public boolean canLoad(final UnitAlike unitToLoad, final boolean isNow) {
		final int capacity = getCapacity(unitToLoad.getUnit());
		final int holdsUsed;
		// We know fields can only hold 0, 1 or infinite. So we only need to take care of capacity 1.
		if (isNow && (0 < capacity) && (capacity < Integer.MAX_VALUE) 
				&& (unitToLoad.getUnit().getGameState().getMap().hasUnitsAtPosition(position))) {
			holdsUsed = 1; 
		} else {
			holdsUsed = 0;
		}
		final int holdsRemaining = capacity - holdsUsed;
		final int holdsNeeded = 1;
		return (holdsNeeded <= holdsRemaining);
	}

	@Override
	public int getCapacity(final Unit unit) {
		return getTemplate().getCapacity(unit.getGameState(), unit.getUnitTemplate());
	}

	@Override
	public Field getField() {
		return this;
	}

	@Override
	public int hashCode() {
		return (null == position ? 0 : position.hashCode());
	}

	@Override
	public int compareTo(final FieldAlike other) {
		final PlayfieldPosition that = this.position;
		return that.compareTo(other.getPosition());
	}

	@Override
	public String toString() {
		return (position.toString() + getTemplate().messageKeyPrefix);
	}

	@Override
	public Collection<UnitAlike> getUnits() {
		return units;
	}

	/** All the units on this field. Non-persistent. */
	private TreeSet<UnitAlike> units = new TreeSet<UnitAlike>();
}
