package de.ped.empire.logic;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.Bitfield32;
import de.ped.tools.log.Level;

/**
 * The representation of a city field.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FieldCity extends Field implements Commandable {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	private static final CommandableTemplate TEMPLATE = Templates.instance().CMDL_CITY;

	/** Persistent. */
	private CommandableKey key;
	/** Persistent. */
	private String name;
	/** Persistent. */
	private int productivity = 100;
	/** Persistent. */
	private CommandableTemplate preferredUnit = Templates.instance().UNDEFINED;
	/** The amount of unit production points already reached. Persistent. */
	private float productionProgress = 0.0f;
	/** Persistent. */
	private CommandableTemplate unitInProduction = Templates.instance().NOTHING;
	/** Persistent. */
	private boolean hasSeenEnemiesLately;
	/** The unit just produced and needs to be shown as production report. Persistent. */
	private CommandableKey unitProduced;
	/** Persistent. */
	private boolean wasNexted;
	/** Not persistent. */
	protected Order order;
	/** Not persistent. */
	private I18NStringWithFillIns i18nName = null;
	/** Not persistent. */
	private GameState gameState;

	public FieldCity(final int id, final int x, final int y) {
		super(x, y, ImageKey.FIELD_ID_CITY);
		this.key = new CommandableKey(TEMPLATE.id, Player.ID_NONE, id);
		this.name = TEMPLATE.getNameSuggestion();
		this.hasSeenEnemiesLately = false;
		this.unitProduced = null;
		this.order = null;
	}

	public FieldCity(final Marshaller marshaller, final int x, final int y, final int imageKey) {
		super(marshaller, x, y, imageKey);
		final int playerId = marshaller.readInt();
		final int id = marshaller.readInt();
		this.key = new CommandableKey(TEMPLATE.id, playerId, id);
		this.name = marshaller.readString();
		this.productivity = marshaller.readInt();
		this.preferredUnit = CommandableTemplate.findTemplateById(marshaller.readInt());
		this.productionProgress = marshaller.readFloat();
		this.unitInProduction = CommandableTemplate.findTemplateById(marshaller.readInt());
		if (!marshaller.readNullIndicator()) {
			this.unitProduced = new CommandableKey(marshaller);
		} else {
			this.unitProduced = null;
		}
		this.hasSeenEnemiesLately = marshaller.readBoolean();
		this.wasNexted = marshaller.readBoolean();
		this.order = null; // City orders are not persistent
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		super.marshal(marshaller);
		marshaller.writeInt(key.getPlayerId());
		marshaller.writeInt(key.getId());
		marshaller.writeString(name);
		marshaller.writeInt(productivity);
		marshaller.writeInt(preferredUnit.id);
		marshaller.writeFloat(productionProgress);
		marshaller.writeInt(unitInProduction.id);
		if (!marshaller.writeNullIndicator(unitProduced)) {
			marshaller.writeMarshallable(unitProduced);
		}
		marshaller.writeBoolean(hasSeenEnemiesLately);
		marshaller.writeBoolean(wasNexted);
	}

	@Override
	public CommandableKey getKey() {
		return key;
	}

	/**
	 * Called once at the beginning of a new game.
	 * 
	 * @param gameState
	 *            the game state
	 */
	void setGameState(final GameState gameState) {
		this.gameState = gameState;
	}

	void setProductivity() {
		if (gameState.getProperties().getRulesComplexity().isProductivityRandom 
				&& (getPlayerId() == Player.ID_NONE)) {
			productivity = 60 + MathUtil.random(100);
			if (0 == MathUtil.random(5)) {
				final List<CommandableTemplate> units = getUnitsToProduce();
				preferredUnit = units.get(MathUtil.random(units.size()));
			}
		}
	}

	public int getProductivity() {
		return productivity;
	}

	public CommandableTemplate getPreferredUnit() {
		return preferredUnit;
	}

	@Override
	public int getImageKey() {
		int imgKey = super.getImageKey();
		if (null != gameState) {
			imgKey = Bitfield32.setBits(imgKey, ImageKey.BA_PLAYER_COLOR, 
					gameState.getPlayer(getPlayerId()).getPlayerColorId());
			imgKey = Bitfield32.setBits(imgKey, ImageKey.BA_NONEMPTY, 
					((gameState.hasUnitsAtPosition(getPosition())) ? 1 : 0));
		}
		imgKey = Bitfield32.setBits(imgKey, ImageKey.BA_LOOK, MathUtil.modulo(key.getId(), 2));
		setImageKey(imgKey);
		return imgKey;
	}

	@Override
	public I18NStringWithFillIns getName() {
		if (null == i18nName) {
			i18nName = new I18NStringWithFillIns(null, name, null);
		}
		return i18nName;
	}

	@Override
	public String getNickName() {
		return name;
	}

	@Override
	public void setNickName(final String name) {
		this.name = name;
		this.i18nName = null;
	}

	@Override
	public boolean needsNickName() {
		return (CollectionUtil.isNullOrEmpty(name));
	}

	@Override
	public String getNickNameSuggestion() {
		final String suggestion = TEMPLATE.getNameSuggestion();
		return suggestion;
	}

	@Override
	public I18NStringWithFillIns getToolTip(final GameView gameView) {
		return getToolTip(gameView, this, getPlayerId(), getProductivity(),
				getUnitInProduction(), getTurnsToProduce(),
				hasSeenEnemiesLately());
	}

	public static I18NStringWithFillIns getToolTip(
			final GameView gameView, 
			final FieldAlike city,
			final int playerId,
			final int productivity,
			final CommandableTemplate unitInProduction,
			final int turnsToProduce,
			final boolean hasSeenEnemiesLately) {
		I18NStringWithFillIns namePart = new I18NStringWithFillIns("Game.City.Name", null, new I18NStringWithFillIns[] { city.getName() });
		I18NStringWithFillIns prodPart = I18NStringWithFillIns.EMPTY;
		I18NStringWithFillIns effiPart = I18NStringWithFillIns.EMPTY;
		I18NStringWithFillIns unitsPart = I18NStringWithFillIns.EMPTY;
		I18NStringWithFillIns sightPart = I18NStringWithFillIns.EMPTY;
		I18NStringWithFillIns ownerPart = I18NStringWithFillIns.EMPTY;
		if (city.getPlayerId() == gameView.getPlayer().getId()) {
			// Productivity and unit in production of a city is only shown for own cities, not for other players' cities.
			if (gameView.getProperties().getRulesComplexity().isProductivityRandom) {
				effiPart = new I18NStringWithFillIns("Game.City.Productivity", null, new Object[] { 
						Integer.toString(productivity),
				});
			}
			if (unitInProduction.isUndefinedOrNothing()) {
				// Nothing to do
			} else {
				prodPart = new I18NStringWithFillIns("Game.City.Production", null, new Object[] { 
						new I18NStringWithFillIns(unitInProduction.getNameKey(), null, null),
						Integer.toString(turnsToProduce),
				});
			}
			final PlayfieldPosition pos = city.getPosition();
			Collection<UnitAlike> units = gameView.findUnitsAtPosition(pos);
			if (!CollectionUtil.isNullOrEmpty(units)) {
				unitsPart = new I18NStringWithFillIns("Game.City.Units", null, 
						new String[] { Integer.toString(units.size()) });
			}
			if (hasSeenEnemiesLately) {
				sightPart = new I18NStringWithFillIns("Game.Unit.SightedEnemy", null, null);
			}
		} else {
			ownerPart = new I18NStringWithFillIns("Game.City.Player", null,
					new String[] { gameView.getPlayer(playerId).getName() });
		}
		logger.log(Level.TRACE, "getToolTip: unitInProduction=" + unitInProduction.messageKeyPrefix, new IllegalArgumentException("Just to get a stack trace"), true);
		return new I18NStringWithFillIns("Game.City.ToolTip", "%0%1%2%3%4%5", 
				new Object[] { namePart, effiPart, ownerPart, prodPart, unitsPart, sightPart, } );
	}

	@Override
	public int getPlayerId() {
		return key.getPlayerId();
	}

	private String logPrefix() {
		return "FieldCity." + getNickName() + ".";
	}

	public void setPlayerId(final int playerId) {
		logger.debug(logPrefix() + "setPlayerId(" + playerId + ")");
		if (playerId != key.getPlayerId()) {
			this.key = new CommandableKey(key.getTemplateId(), playerId, key.getId());
			this.productionProgress = 0.0f;
			if (playerId == Player.ID_NONE) {
				this.unitInProduction = Templates.instance().NOTHING;
			} else {
				this.unitInProduction = Templates.instance().UNDEFINED;
			}
			this.hasSeenEnemiesLately = false;
			this.unitProduced = null;
			this.resetNext();
		}
	}

	void setProduction(final CommandableTemplate unit) {
		logger.debug(logPrefix() + "setProduction " + unit);
		if (unit == unitInProduction) {
			// Nothing to do, do not throw away the work already done
		} else {
			unitInProduction = unit;
			productionProgress = 0.0f;
		}
		hasSeenEnemiesLately = false;
		unitProduced = null;
		wasNexted = true;
	}

	private static final float EPSILON = 0.001f;

	private int roundToTurns(final double turns) {
		return (int)(turns + 1.0 - EPSILON);
	}

	public int getTurnsToProduce() {
		return roundToTurns((1.0f - productionProgress) * getTurnsToProduceNext(unitInProduction));
	}

	public CommandableTemplate getUnitInProduction() {
		return unitInProduction;
	}

	public int getTurnsToProduce(final CommandableTemplate template) {
		int ret;
		if (template == unitInProduction) {
			ret = getTurnsToProduce();
		} else {
			ret = getTurnsToProduceNext(template);
		}
		return ret;
	}

	private static final double PREFERRED_UNIT_FACTOR_NONE = 1.0;
	private static final double PREFERRED_UNIT_FACTOR_PREFERRED = 0.8;
	private static final double PREFERRED_UNIT_FACTOR_NON_PREFERRED = 1.1;

	private static final double TROOPS_SUPPORT_CUTOFF = 0.01;

	private int getTurnsToProduceNext(final CommandableTemplate template) {
		int ret;
		Assert.assertNotNull(gameState);
		if (template.isUndefinedOrNothing()) {
			ret = Integer.MAX_VALUE;
		} else {
			final double handicap = gameState.getPlayer(getPlayerId()).getProductionHandicapValue();
			double preferredUnitFactor = PREFERRED_UNIT_FACTOR_NONE;
			if (!preferredUnit.isUndefinedOrNothing()) {
				if (template.equals(preferredUnit)) {
					preferredUnitFactor = PREFERRED_UNIT_FACTOR_PREFERRED;
				} else {
					preferredUnitFactor = PREFERRED_UNIT_FACTOR_NON_PREFERRED;
				}
			}
			double troopsSupportUnitFactor = gameState.getTroopsSupportFactor(getPlayerId());
			logger.trace("Turns to produce for " + template.messageKeyPrefix + " is " + handicap);
			if (troopsSupportUnitFactor < TROOPS_SUPPORT_CUTOFF) {
				ret = Integer.MAX_VALUE;
			} else {
				ret = roundToTurns(template.turnsToProduce 
						* handicap 
						/ troopsSupportUnitFactor 
						* preferredUnitFactor 
						* 100 / productivity);
			}
		}
		return ret;
	}

	@Override
	public void startTurn() {
		final int turns = getTurnsToProduce();
		logger.debug(logPrefix() + "startTurn() with " + MathUtil.round(productionProgress, EPSILON) + " production progress  and " + turns + " turns to produce");
		resetNext();
		if (!unitInProduction.isUndefinedOrNothing() && (-EPSILON <= productionProgress)) {
			productionProgress += 1.0f / getTurnsToProduceNext(unitInProduction);
			if (1.0f <= productionProgress) {
				Unit unit = Unit.create(gameState, this);
				unitProduced = unit.getKey();
				logger.debug(logPrefix() + "startTurn() produced " + unit.getUnitTemplate().getNameKey());
				gameState.addUnit(this, unit);
				productionProgress = (float) gameState.getProperties().getRulesComplexity().productionRepetitionBonusFactor;
			}
		}
	}

	@Override
	public void finishTurn() {
		logger.debug(logPrefix() + "finishTurn()");
		hasSeenEnemiesLately = false;
		unitProduced = null;
		resetNext();
	}

	/** As a city is owned by a player, it can only stack units of the same player. */
	@Override
	public int getCapacity(final Unit unit) {
		int ret;
		if (getPlayerId() == unit.getPlayerId()) {
			ret = super.getCapacity(unit);
		} else {
			ret = 0;
		}
		return ret;
	}

	@Override
	public boolean canLoad(final UnitAlike unitToLoad, final boolean isNow) {
		return true;
	}

	@Override
	public boolean isNextable() {
		return (Templates.instance().UNDEFINED == getUnitInProduction())
				|| hasBeenUnitProducedLately()
				|| hasSeenEnemiesLately;
	}

	@Override
	public boolean wasNexted() {
		return (wasNexted);
	}

	@Override
	public void resetNext() {
		wasNexted = false;
	}

	@Override
	public void nextFocusGained() {
		wasNexted = true;
	}

	@Override
	public void nextFocusLost() {
		logger.debug(logPrefix() + "nextFocusLost()");
		hasSeenEnemiesLately = false;
//		unitProduced = null;
	}

	boolean hasBeenUnitProducedLately() {
		return (null != unitProduced);
	}

	CommandableKey getUnitProduced() {
		return unitProduced;
	}

	void clearUnitProduced() {
		logger.debug(logPrefix() + "clearUnitProduced()");
		unitProduced = null;
	}

	@Override
	public Order getOrder() {
		return order;
	}

	@Override
	public void setOrder(final Order order) {
		this.order = order;
	}

	@Override
	public MoveResult evaluateOrder(final GameView gameView, final boolean isNow) {
		MoveResult ret = null;
		if (null != order) {
			ret = order.evaluate(isNow);
		}
		return ret;
	}

	protected boolean isCoastal(final GameMap map) {
		boolean ret = false;
		final PlayfieldPosition pos = getPosition();
		for (final DigitalJoystickDirection dir : map.getProperties().getNeighborDirections(pos)) {
			final PlayfieldPosition npos = map.getProperties().wrap(pos, dir);
			if (map.getAt(npos).getTemplateId() == ImageKey.FIELD_ID_SEA) {
				ret = true;
				break;
			}
		}
		return ret;
	}

	/**
	 * Keeps what kind of units can be produced in this city. 
	 * Is not persistent, but calculated when needed.
	 */
	private List<CommandableTemplate> unitsToProduce = null;

	/**
	 * @return what kind of units can be produced in this city
	 */
	public List<CommandableTemplate> getUnitsToProduce() {
		if (null == unitsToProduce) {
			List<CommandableTemplate> ret = new LinkedList<CommandableTemplate>();
			final boolean isCoastal = isCoastal(gameState.getMap());
			for (final CommandableTemplate template : CommandableTemplate.allUnitsCitiesCanProduce(gameState.getProperties().getRulesComplexity())) {
				if ((template.militaryBranch != MilitaryBranch.SEA_UNIT) || isCoastal) {
					ret.add(template);
				}
			}
			unitsToProduce = ret;
		}
		return unitsToProduce;
	}

	@Override
	public int getSightingRange(final MilitaryBranch observed) {
		Assert.assertNotNull(TEMPLATE);
		Assert.assertNotNull(gameState);
		int ret = TEMPLATE.getSightingRange(gameState.getProperties().getRulesComplexity(), observed);
		return ret;
	}

	@Override
	public boolean hasSeenEnemiesLately() {
		return hasSeenEnemiesLately;
	}

	@Override
	public void setSeenEnemiesLately(final boolean value) {
		hasSeenEnemiesLately = value;
	}

	@Override
	public AttackMode getAttackModeAgainst(Commandable opponent) {
		return AttackMode.IMPOSSIBLE;
	}

	@Override
	public int getHitsTillDeath() {
		return TEMPLATE.getHitsTillDeath(gameState.getProperties().getRulesComplexity());
	}

	@Override
	public void receiveHits(int numberOfHits) {
//		hitsTillDeath = MathUtil.bound(hitsTillDeath - numberOfHits, 0, MAX_HITS_TILL_DEATH);
	}

	@Override
	public boolean hasMovedThisTurn() {
		return false;
	}

	@Override
	public void setSkippingThisTurn(boolean value) {
		// Nothing to do
	}

	@Override
	public int getAttack() {
		return TEMPLATE.attack;
	}

	@Override
	public int getDefense() {
		return TEMPLATE.defense;
	}

	@Override
	public int getShotDamage() {
		return TEMPLATE.shotDamage;
	}

	@Override
	public float getFightStrength(final Commandable opponent, final boolean isAttacker) {
		float ret = getDefense();
		if (unitInProduction.isUndefinedOrNothing()) {
			ret -= 0.2f;
		} else {
			ret += 0.4f * productionProgress;
		}
		return ret;
	}

	@Override
	public boolean isAlive() {
		return true;
	}

	@Override
	public boolean isDamaged() {
		return false;
	}

	public void decreaseProductivity() {
		productivity = Math.max(Math.round(productivity * 0.9f), 10);
		logger.debug(logPrefix() + "decreaseProductivity()");
	}

	public void destroyUnitInProduction() {
		productionProgress = 0.0f;
		logger.debug(logPrefix() + "destroyUnitInProduction");
	}

	@Override
	public FieldTemplate getTemplate() {
		return Templates.instance().FIELD_CITY;
	}

	@Override
	public String toString() {
		return key.toString() + "=" + getNickName();
	}

	int getTroopSupportFactor() {
		final int ret;
		if (Templates.instance().NOTHING.equals(unitInProduction)) {
			ret = 2;
		} else {
			ret = 1;
		}
		return ret;
	}
}
