package de.ped.empire.logic;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;

/**
 * Some configuration data to distinguish the game map field types.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FieldTemplate {

	/**
	 * Defines all parameters which might depend on the rules complexity.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class RulesComplexityDependentValues {
		public RulesComplexityDependentValues(
				final LoadDefinition[] loadDefinitions,
				final int bitmaskUnitsSlowedDown,
				final int bitmaskDefenseBonusAgainst) {
			this.loadDefinitions = loadDefinitions;
			this.unitsSlowedDown = bitmaskUnitsSlowedDown;
			this.defenseBonusAgainst = bitmaskDefenseBonusAgainst;
		}
		
		final LoadDefinition[] loadDefinitions;
		final int unitsSlowedDown;
		final int defenseBonusAgainst;
	}

	public final byte id;
	public final String messageKeyPrefix;
	public final String imageResourceName;
	public final boolean isLand;
	public final boolean isRiverNeighbor;
	public final int maxCapacity;
	public final RulesComplexityDependentValues[] rcdvs;
	protected final I18NStringWithFillIns name;

	/**
	 * Constructor.
	 * @param id the id of the template
	 * @param messageKeyPrefix the i18n message key prefix
	 * @param isLand if a land field
	 * @param isRiverNeighbor needed for drawing rivers correctly
	 * @param maxCapacity the maximum capacity of units fields of this type can hold. Either 0, 1 or {@link Integer#MAX_VALUE}.
	 * @param rcdvs an array of the same size as we have different rule sets (currently 3), defining rule set depending values.
	 */
	FieldTemplate(
			final byte id, 
			final String messageKeyPrefix, 
			final boolean isLand,
			final boolean isRiverNeighbor,
			final int maxCapacity,
			final RulesComplexityDependentValues[] rcdvs) {
		this.id = id;
		this.messageKeyPrefix = messageKeyPrefix;
		this.imageResourceName = messageKeyPrefix + ".png";
		this.isLand = isLand;
		this.isRiverNeighbor = isRiverNeighbor;
		this.maxCapacity = maxCapacity;
		this.rcdvs = rcdvs;
		this.name = new I18NStringWithFillIns(messageKeyPrefix + Messages.TEXT_POSTFIX, null, null);
	}

	/**
	 * Calculates how many units of a given commandable an empty field of this type can hold.
	 * @param gameState the game state, defining the ruleset to follow
	 * @param cmdlTemplate the commandable this type of field should hold
	 * @return how many units of a given commandable an empty field of this type can hold
	 */
	int getCapacity(final GameState gameState, final CommandableTemplate cmdlTemplate) {
		int ret = 0;
		final RulesComplexity ruleComplexity = gameState.getProperties().getRulesComplexity();
		final FieldTemplate.RulesComplexityDependentValues rcdv =
				rcdvs[ruleComplexity.ordinal()];
		final LoadDefinition[] loadDefinitions = rcdv.loadDefinitions;
		if (null != loadDefinitions) {
			for (LoadDefinition ld : loadDefinitions) {
				if ((null != ld) && (ld.getUnitType().equals(cmdlTemplate))) {
					if (0 < ld.holdsTaken) {
						ret = maxCapacity / ld.holdsTaken;
					} else {
						ret = Integer.MAX_VALUE;
					}
					break;
				}
			}
		}
		return ret;
	}
}