package de.ped.empire.logic;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import javax.swing.JOptionPane;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.Assert;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.ResourceFinder;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * The facade for clients looking at the game state.
 * 
 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameFacade {

	public static final String SAVED_GAME_FILE_EXTENSION = "svg";
	public static final TutorialLevel TUTORIAL_LEVEL_NONE = null;
	public static final TutorialLevel TUTORIAL_LEVEL_DEFAULT = TutorialLevel.VISIB_IF_LOADED;

	/**
	 * Data structure to report which sound to play (if any).
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class SoundToPlayData {
		public final String filename;
		public final boolean hasBeenChanged;

		public SoundToPlayData(final String filename, final boolean hasBeenChanged) {
			this.filename = filename;
			this.hasBeenChanged = hasBeenChanged;
		}
	}

	public static final int TRIES = 10;
	public static final int FAILED = 0;

	/**
	 * Creates a new game.
	 * @param observer the observer who initiated the action
	 * @param localeIndex the index of the locale to start with for all players
	 * @param props the game properties the game will be based upon
	 * @param tutorialLevel the tutorial level, or {@link #TUTORIAL_LEVEL_NONE} for a non-tutorial game
	 * @return the number of tries that was necessary to create a map  (could be any non-zero positive number), 
	 * 		or {@link #FAILED} if no game with those parameters could be created in a reasonable number of tries.
	 */
	public synchronized int createGame(final GameObserver observer, final int localeIndex, final GameProperties props, final TutorialLevel tutorialLevel) {
		logger.debug("createGame()");
		boolean ok = false;
		int tries;
		for (tries = 0; !ok && (tries < TRIES); tries++) {
			logger.debug("createGame(): Starting try " + tries);
			try {
				if (TUTORIAL_LEVEL_NONE == tutorialLevel) {
					final GameMap map = new GameMap(props);
					map.addAtRandom(props);
					map.finish();
					gameState = new GameState(observer, map);
					if (props.isQuickstart()) {
						gameState.initForQuickStart();
					}
				} else {
					final Tutorial tutorial = new Tutorial();
					final GameMap map = tutorial.createMap(tutorialLevel);
					map.finish();
					gameState = new GameState(observer, map);
					tutorial.createTutorial(gameState, tutorialLevel);
				}
				ok = true;
			} catch (final GameInitializationException ex) {
				logger.debug("createGame(): Unable to create game in try " + tries + ", retrying.");
			}
		};
		if (ok) {
			final Player[] players = gameState.getProperties().getPlayers();
			for (int i = 0; i < players.length; i++) {
				final GameView view = gameState.getGameView(i);
				if (null != view) {
					view.getGameViewProperties().setLocaleIndex(localeIndex);
				}
			}
			gameState.startGame();
			gameState.init();
		} else {
			logger.warn("createGame(): Unable to create game after " + tries + " tries.");
		}
		return (ok ? tries : FAILED);
	}

	/**
	 * Immediately ends the running game.
	 * @param observer the observer who initiated the action
	 */
	public synchronized void exitGame(final GameObserver observer) {
		logger.debug("exitGame()");
		gameState = null;
	}

	public synchronized GameStateBasics getGameStateBasics() {
		logger.debug("getGameStateBasics()");
		GameStateBasics ret = null;
		if (null != gameState) {
			ret = (GameStateBasics) gameState.getGameStateBasics().clone();
		}
		return ret;
	}

	/**
	 * Loads a game from file.
	 * @param observer the observer who initiated the action
	 * @param file the file to load
	 * @return <code>true</code> if the operation was successful
	 */
	public synchronized boolean loadGame(final GameObserver observer, final File file) {
		logger.debug("loadGame(" + file + ")");
		boolean ret = false;
		try {
			gameState = new GameState(observer, Marshaller.loadFromFile(file));
			ret = true;
		} catch (final IOException e) {
		    logger.error("Error loading file " + file.getName() + ": " + e); 
		}
		return ret;
	}

	/**
	 * Saves a game to file.
	 * @param observer the observer who initiated the action
	 * @param file the file to save to
	 * @return <code>true</code> if the operation was successful
	 */
	public synchronized boolean saveGame(final GameObserver observer, final File file) {
		logger.debug("saveGame(" + file + ")");
		boolean ret = false;
		if (null == gameState) {
			logger.info("Unable to save null game :-)");
		} else {
			try {
				logger.info("You chose to write to this file: " + file.getName());
				final byte[] data = Marshaller.writeToFile(gameState, file);
				ret = true;
				logger.info("Bytes written: " + data.length);
			} catch (final IOException e) {
				logger.error("Error exporting file " + file.getName() + ": " + e);
			}
		}
		return ret;
	}

	public synchronized Object[][] getLogbookData(final int playerId) {
		logger.debug("getLogbookData(" + playerId + ")");
		final Object[][] ret;
		if (null == gameState) {
			ret = null;
		} else {
			final List<LogbookEntry> data = gameState.getGameView(playerId).getLogbook();
			ret = new Object[data.size()][5];
			int i = 0;
			for (final LogbookEntry entry : data) {
				ret[i][0] = new Integer(entry.turn);
				ret[i][1] = gameState.getPlayer(entry.currentPlayerId).getName();
				final String pos;
				if (null == entry.pos) {
					pos = "--";
				} else {
					pos = entry.pos.getX() + ", " + entry.pos.getY();
				}
				ret[i][2] = pos;
				final I18NStringWithFillIns cmdlName;
				if (null == entry.commandable) {
					cmdlName = new I18NStringWithFillIns(null, "--", null);
				} else {
					cmdlName = entry.commandable.getName();
				}
				ret[i][3] = cmdlName;
				ret[i][4] = entry.getEntry();
				i++;
			}
		}
		return ret;
	}

	public synchronized Collection<MovePath> getMovePaths(final int playerId) {
		logger.debug("getMovePaths(" + playerId + ")");
		final Collection<MovePath> ret;
		if (null == gameState) {
			ret = null;
		} else {
			ret = gameState.getMovePaths().findMovePaths(playerId);
		}
		return ret;
	}

	/**
	 * Gets the move paths data to be shown in the MovePaths dialog.
	 * @param commandable all move paths for which this commandable is the starting point are returned
	 * @return the data to display. It's a two-dimensional array containing paths on x axis and on y the following info:
	 * 	<ol>
	 * 	<li>end position
	 * 	<li>vector describing distance from start to end position
	 * 	<li>description or name of the field at end position
	 * 	<li>for the three real military branches, if this path is selected
	 * 	<li>for the three real military branches, if this path is editable
	 *  </ol>
	 */
	public synchronized Object[][] getMovePaths(final Commandable commandable) {
		logger.debug("getMovePaths(" + commandable + ")");
		final Object[][] ret;
		if (null == gameState) {
			ret = null;
		} else {
			final PlayfieldPosition startPosition = commandable.getPosition();
			final MovePaths movePaths = (MovePaths) gameState.getMovePaths().clone();
			final List<MovePath> movePathsStartingHere = movePaths.findMovePaths(commandable.getPlayerId(), commandable.getPosition());
			final GameView gameView = gameState.getGameView(commandable.getPlayerId());
			ret = new Object[movePathsStartingHere.size()][9];
			int i = 0;
			for (MovePath path : movePathsStartingHere) {
				final PlayfieldPosition endPosition = path.getEndPos();
				ret[i][0] = endPosition;
				ret[i][1] = OrderGoTo.createRelativeMovePathDescription(gameView.getMap(), startPosition, endPosition);
				final FieldAlike field = gameView.getMap().getAt(endPosition);
				final Commandable found = gameView.findCityOrAirbase(field);
				if ((null == found) || (found == field)) {
					ret[i][2] = field.getToolTip(gameView);
				} else {
					ret[i][2] = found.getName();
				}
				for (MilitaryBranch branch : MilitaryBranch.REAL_BRANCHES) {
					ret[i][3 + branch.ordinal()] = Boolean.valueOf(path.isForMilitaryBranch(branch));
					ret[i][6 + branch.ordinal()] = Boolean.valueOf(gameView.makesSense(path, branch));
				}
				i++;
			}
		}
		return ret;
	}

	/**
	 * Sets the move paths, after they have been modified from
	 * {@link #getMovePaths(Commandable)}.
	 * 
	 * @param commandable
	 *            all move paths for which this commandable is the starting
	 *            point
	 * @param data
	 *            the modified data
	 * @param isToCleanup
	 *            <code>true</code> if the editing of the paths from this
	 *            location is finished (and empty paths can be removed), or
	 *            <code>false</code> if further editing might follow (e.g.
	 *            because user wanted to add another path to another end point)
	 * @return <code>false</code> if everything worked fine, <code>true</code>
	 *         to indicate that now at least one circular path exists
	 */
	public synchronized boolean setMovePaths(final Commandable commandable, final Object[][] data, final boolean isToCleanup) {
		logger.debug("setMovePaths(" + commandable + "," + isToCleanup + ")");
		boolean ret = false;
		if (null == gameState) {
			// Nothing to do
		} else {
			final int playerId = commandable.getPlayerId();
			final PlayfieldPosition position = commandable.getPosition();
			final MovePaths movePaths = gameState.getMovePaths();
			
			movePaths.removeAll(playerId, position, false);
			try {
				for (final Object[] dataRow : data) {
					final MovePath path = new MovePath(playerId, position, 0x00, new PlayfieldPosition((PlayfieldPosition)(dataRow[0])));
					for (final MilitaryBranch branch : MilitaryBranch.REAL_BRANCHES) {
						final Boolean b = (Boolean) dataRow[3 + branch.ordinal()];
						if (b.booleanValue()) {
							path.setForMilitaryBranch(branch, true);
						}
						movePaths.add(path);
					}
				}
			} catch (final CircularReferenceException ex) {
				movePaths.removeAll(playerId, position, false);
				logger.info(ex.getMessage());
				ret = true;
			}
			if (isToCleanup) {
				movePaths.clean(playerId, position);
			}
			for (MovePath path : gameState.getMovePaths().findMovePaths(playerId, position)) {
			    logger.debug("Move path " + path);
			}
		}
		return ret;
	}

	/**
	 * Returns the game properties of the running game.
	 * @return the game properties, or <code>null</code> if no game is running
	 */
	public synchronized GameProperties getGameProperties() {
		logger.debug("getGameProperties()");
		GameProperties ret = null;
		if (null != gameState) {
			ret = gameState.getProperties();
		}
		return ret;
	}

	public synchronized GameViewProperties getGameViewProperties(final int playerId) {
		logger.debug("getGameViewProperties(" + playerId + ")");
		GameViewProperties ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.getGameViewProperties();
		}
		return ret;
	}

	public synchronized void setGameViewProperties(final int playerId, final GameViewProperties gamesViewProperties) {
		logger.debug("setGameViewProperties(" + playerId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.setGameViewProperties(gamesViewProperties);
		}
	}

	public synchronized GameState getGameState() {
		logger.debug("getGameState()");
		return gameState;
	}

	public synchronized boolean isPlayerAlive(final int playerId) {
		logger.debug("isPlayerAlive(" + playerId + ")");
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			ret = gameState.isAlive(gameState.getPlayer(playerId));
		}
		return ret;
	}

	public synchronized boolean isPlayerWinner(final int playerId) {
		logger.debug("isPlayerWinner(" + playerId + ")");
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			ret = gameState.isWinner(gameState.getPlayer(playerId));
		}
		return ret;
	}

	public synchronized boolean isInMoveOver() {
		logger.debug("isInMoveOver()");
		boolean ret = false;
		if (null != gameState) {
			ret = gameState.isInMoveOver();
		}
		return ret;
	}

	public synchronized StatusReportPlayerData getStatusReportPlayerData(final int playerId, final boolean isForCityProduction) {
		logger.debug("getStatusReportPlayerData(" + playerId + "," + isForCityProduction + ")");
		StatusReportPlayerData data = null;
		if (null != gameState) {
			final RulesComplexity rulesComplexity = gameState.getProperties().getRulesComplexity();
			final List<CommandableTemplate> unitTemplates;
			if (isForCityProduction) {
				unitTemplates = CommandableTemplate.allUnitsCitiesCanProduce(rulesComplexity);
			} else {
				unitTemplates = CommandableTemplate.allUnitsToShowStatsFor(rulesComplexity);
			}
			final GameView gameView = gameState.getGameView(playerId);
			Assert.assertNotEquals(Player.ID_NONE, playerId);
			data = new StatusReportPlayerData(gameView, playerId, unitTemplates);
		}
		return data;
	}

	/**
	 * The different types of messages to present to the user in the player get ready dialog.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum PlayerGetReadyMessage {
		NoReports(   "Game.PlayerReadyDialog.NoReports",    JOptionPane.INFORMATION_MESSAGE, null),
		FirstTurn(   "Game.PlayerReadyDialog.FirstTurn",    JOptionPane.INFORMATION_MESSAGE, null),
		SightReports("Game.PlayerReadyDialog.SightReports", JOptionPane.WARNING_MESSAGE, ResourceFinder.Folder.APP),
		FightReports("Game.PlayerReadyDialog.FightReports", JOptionPane.WARNING_MESSAGE, ResourceFinder.Folder.APP),
		GameLost(    "Game.PlayerReadyDialog.GameLost",     JOptionPane.WARNING_MESSAGE, ResourceFinder.Folder.APP),
		GameWon(     "Game.PlayerReadyDialog.GameWon",      JOptionPane.INFORMATION_MESSAGE, ResourceFinder.Folder.APP),
		;
		public final String key;
		public final int messageType;
		public final ResourceFinder.Folder imageFolder;
		private PlayerGetReadyMessage(final String key, final int messageType, final ResourceFinder.Folder imageFolder) {
			this.key = key;
			this.messageType = messageType;
			this.imageFolder = imageFolder;
		}
	}

	public synchronized PlayerGetReadyMessage getPlayerReadyMessage(final int playerId) {
		logger.debug("getPlayerReadyMessage(" + playerId + ")");
		PlayerGetReadyMessage ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.getPlayerReadyMessage();
		}
		return ret;
	}

	/**
	 * Changes the unit type the city produces.
	 * 
	 * @param playerId
	 *            the player who initiates the action
	 * @param cityKey
	 *            the city that should change production
	 * @param unitToProduce
	 *            the unit type that shall be produced
	 * @param isChange
	 *            <code>true</code> if the city shall really change production,
	 *            or <code>false</code> if only the order shall be cleared
	 *            without any change
	 */
	public synchronized void setCityProduction(final int playerId, final CommandableKey cityKey, final CommandableTemplate unitToProduce, final boolean isChange) {
		logger.debug("getPlayerReadyMessage(" + playerId + "," + unitToProduce + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId) && (cityKey.getPlayerId() == playerId)) {
			final FieldCity city = (FieldCity) gameState.findCommandable(cityKey);
			if (cityKey.getPlayerId() == city.getPlayerId()) {
				final GameView gameView = gameState.getGameView(city.getPlayerId());
				if (isChange) {
					gameView.setCityProduction(city, unitToProduce);
					gameView.setStatusMessage(city.getToolTip(gameView));
				}
				final Order order = city.getOrder();
				if ((null != order) && (order.template.id == OrderTemplate.TYPE_CHANGE_PRODUCTION)) {
					city.setOrder(null);
				}
				gameView.nextCommandable();
			}
		}
	}

	public synchronized PlayfieldPosition getPositionInFocus(final int playerId) {
		logger.debug("getPositionInFocus(" + playerId + ")");
		PlayfieldPosition ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.getPositionInFocus();
		}
		return ret;
	}

	public synchronized void setPositionInFocus(final int playerId, final PlayfieldPosition position) {
		logger.debug("setPositionInFocus(" + playerId + "," + position + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.setPositionInFocus(position);
		}
	}

	public synchronized void setPositionInFocus(final int playerId, final DigitalJoystickDirection direction) {
		logger.debug("setPositionInFocus(" + playerId + "," + direction + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			final GameProperties props = gameState.getProperties();
			gameView.setPositionInFocus(props.wrap(gameView.getPositionInFocus(), direction));
		}
	}

	public synchronized Commandable getVisibleCommandableInFocus(final int playerId) {
		logger.debug("getVisibleCommandableInFocus(" + playerId + ")");
		Commandable  ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.getVisibleCommandableInFocus();
		}
		return ret;
	}

	/**
	 * @param playerId
	 *            the player asking
	 * @param commandable
	 *            the city that might have produced something
	 * @return the commandable that was just produced, or <code>null</code> if
	 *         <em>commandable</em> was no city or didn't produce a unit this
	 *         turn
	 */
	public synchronized Commandable getUnitProduced(final int playerId, final Commandable commandable) {
		logger.debug("getUnitProduced(" + playerId + "," + commandable + ")");
		Commandable ret = null;
		if (null != commandable) {
			if (commandable instanceof FieldCity) {
				final FieldCity city = (FieldCity) commandable;
				final CommandableKey unitProducedKey = city.getUnitProduced();
				if (null != unitProducedKey) {
					final Commandable cmdl = findCommandable(playerId, unitProducedKey);
					ret = cmdl;
				}
			}
		}
		return ret;
	}

	/**
	 * Clears the unit produced marker.
	 * @param playerId the player asking
	 * @param commandable the city that might have produced something
	 */
	public synchronized void clearUnitProduced(final int playerId, final Commandable commandable) {
		logger.debug("clearUnitProduced(" + playerId + "," + commandable + ")");
		if (null != commandable) {
			if (commandable instanceof FieldCity) {
				final FieldCity city = (FieldCity) commandable;
				city.clearUnitProduced();
			}
		}
	}

	public synchronized Commandable findCommandable(final int playerId, final CommandableKey key) {
		logger.debug("findCommandable(" + playerId + "," + key + ")");
		Commandable ret = null;
		Assert.assertNotNull(key);
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.findCommandable(key);
		}
		logger.trace("findCommandable(" + playerId + "," + key + ") -> " + ret);
		return ret;
	}

	public synchronized Tileable getOrderOptionsObject(final int playerId, final PlayfieldPosition pos) {
		logger.debug("getOrderOptionsObject(" + playerId + "," + pos + ")");
		Tileable ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.getOrderOptionsObject(pos);
		}
		return ret;
	}

	public synchronized void nextCommandable(final int playerId) {
		logger.debug("nextCommandable()");
		if ((null != gameState) && (gameState.getCurrentPlayerId() == playerId)) {
			gameState.getGameView(playerId).nextCommandable();
		}
	}

	public synchronized void nextCommandable(final int playerId, final Commandable commandable) {
		logger.debug("nextCommandable(" + playerId + "," + commandable + ")");
		if ((null != gameState) && (gameState.getCurrentPlayerId() == playerId)) {
			gameState.getGameView(playerId).nextCommandable(commandable);
		}
	}

	public synchronized SoundToPlayData getSoundToPlay(final int playerId) {
		logger.debug("getSoundToPlay(" + playerId + ")");
		SoundToPlayData ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			if (gameView.isAcceptingUserInput() && (MapViewMode.NORMAL == gameView.getMapViewMode())) {
				// TODO: Should have been triggered by a notifyUnitProptsForInput()
				final Commandable commandable = gameView.getVisibleCommandableInFocus();
				if (null != commandable) {
					final String name = CommandableTemplate.findTemplateById(commandable.getKey().getTemplateId()).promptSoundFileName;
					if (null != name) {
						String filename = null;
						if (gameView.getGameViewProperties().isPlayingSounds(SoundTypes.PROMPT.ordinal())) {
							filename = name;
						}
						ret = new SoundToPlayData(filename, true);
					}
				}
			}
		}
		return ret;
	}

	public synchronized String getStatusMessage(final int playerId) {
		logger.debug("getStatusMessage(" + playerId + ")");
		String ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			ret = gameState.getGameView(playerId).getStatusMessage();
		}
		return ret;
	}

	public synchronized void popStatusMessage(final int playerId) {
		logger.debug("popStatusMessage(" + playerId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			gameState.getGameView(playerId).popStatusMessage();
		}
	}

	public synchronized void pushStatusMessage(final int playerId, final I18NStringWithFillIns statusMessage) {
		logger.debug("pushStatusMessage(" + playerId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			gameState.getGameView(playerId).pushStatusMessage(statusMessage);
		}
	}
	
	public synchronized I18NStringWithFillIns getToolTip(final int playerId, final Tileable tileable) {
		logger.debug("getToolTip(" + playerId + ")");
		I18NStringWithFillIns ret = null;
		if (null != gameState) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = tileable.getToolTip(gameView);
		}
		return ret;
	}

	public synchronized void setStatusMessageToToolTip(final int playerId, final Tileable tileable) {
		logger.debug("setStatusMessageToToolTip(" + playerId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.setStatusMessage(tileable.getToolTip(gameView));
		}
	}

	public synchronized void setStatusMessageToField(final int playerId, final PlayfieldPosition pos) {
		logger.debug("setStatusMessageToField(" + playerId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.setStatusMessage(gameView.getStatusMessageForField(pos));
		}
	}

	public synchronized MapViewMode getMapViewMode(final int playerId) {
		logger.debug("getMapViewMode(" + playerId + ")");
		MapViewMode ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			ret = gameState.getGameView(playerId).getMapViewMode();
		}
		return ret;
	}

	public synchronized boolean isMapViewModeEnabled(final int playerId, final MapViewMode mode) {
		logger.debug("isMapViewModeEnabled(" + playerId + "," + mode + ")");
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.isMapViewModeEnabled(mode);
		}
		return ret;
	}

	public synchronized void setMapViewMode(final int playerId, final MapViewMode mode) {
		logger.debug("setMapViewMode(" + playerId + "," + mode + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			gameState.getGameView(playerId).setMapViewMode(mode);
		}
	}

	public synchronized int getPerspective(final int playerId) {
		logger.debug("getPerspective(" + playerId + ")");
		int ret = 0;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			ret = gameState.getGameView(playerId).getPerspective();
		}
		return ret;
	}

	public boolean isDefaultPerspective(int playerId) {
		logger.debug("isOwnPerspective(" + playerId + ")");
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			ret = gameState.getGameView(playerId).isDefaultPerspective();
		}
		return ret;
	}

	public synchronized boolean isPerspectiveEnabled(final int playerId, final int perspectiveId) {
		logger.debug("isPerspectiveEnabled(" + playerId + "," + perspectiveId + ")");
		boolean ret = false;
		if (null != gameState) {
			final GameView gameView = gameState.getGameView(playerId);
			if (null != gameView) {
				final Player player = gameState.getPlayer(playerId);
				if (gameState.isWinner(player)) {
					ret = true;
				} else if (!gameState.isAlive(player)){
					ret = true;
				} else {
					// When alive, only the player's view is enabled
					if (gameView.getPlayer().getId() == perspectiveId - 1) {
						ret = true;
					} else {
						// With debug parameter, you're able to view other views all the time.
						if (null == applEnv) {
							ret = false;
						} else {
							final String debug = applEnv.findParameter(ApplicationEnvironment.PARAM_KEY_DEBUG).getValue();
							if (null != debug) {
								ret = true; 
							} else {
								ret = false; 
							}
						}
					}
				}
			}
		}
		return ret;
	}

	public synchronized void setPerspective(final int playerId, final int perspectiveId) {
		logger.debug("setPerspective(" + playerId + "," + perspectiveId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.setPerspective(perspectiveId);
		}
	}

	public synchronized void setZoomLevel(final int playerId, final int zoomLevel) {
		logger.debug("setZoomLevel(" + playerId + "," + zoomLevel + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			gameState.getGameView(playerId).getGameViewProperties().setZoomLevel(zoomLevel);
		}
	}

	public synchronized void setLocaleIndex(final int playerId, final int localeIndex) {
		logger.debug("setLocaleIndex(" + playerId + "," + localeIndex + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			gameState.getGameView(playerId).getGameViewProperties().setLocaleIndex(localeIndex);
		}
	}

	public synchronized GameMapInfo createGameMapInfo(final int playerId) {
		logger.debug("createGameMapInfo(" + playerId + ")");
		GameMapInfo ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = new GameMapInfo(gameView);
		}
		return ret;
	}

	public synchronized boolean isFinishTurnEnabled(final int playerId) {
		logger.debug("isFinishTurnEnabled(" + playerId + ")");
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.isFinishTurnAllowed();
		}
		return ret;
	}

	public synchronized void clearGetReady(final int playerId) {
		logger.debug("clearGetReady(" + playerId + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.clearGetReady();
		}
	}

	public synchronized boolean finishTurn(final int playerId) {
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.finishTurn();
		}
		return ret;
	}

	public synchronized void autoSaveGame() {
		final long timestamp = gameState.getProperties().getTimestamp();
		final String gameName = gameState.getProperties().getName();
		final String name = new SimpleDateFormat("yyyy-MM-dd-HHmmss").format(new Date(timestamp)) + "-"
				+ (null == gameName ? "" : (gameName + "-"))
				+ String.format("%05d", gameState.getTurn()) + "-" + gameState.getCurrentPlayerId();
		final File file = new File(name + "." + SAVED_GAME_FILE_EXTENSION);
		saveGame(null, file);
		logger.info("Auto-saving " + file.getName() + " was successfull.");
	}

	public synchronized void resign(final int playerId) {
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.resign();
		}
	}

	/**
	 * Data structure to return user selection options for the context menu on
	 * the map. Consists of a list of orders and a list of commandables.
	 * 
	 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class UserMapTileContextOptions {
		public final List<Order> orders;
		public final List<Commandable> commandables;

		public UserMapTileContextOptions(final List<Order> orders, final List<Commandable> commandables) {
			this.orders = orders;
			this.commandables = commandables;
		}
	}

	/**
	 * Returns all orders and commandables that shall be presented in a context
	 * menu for the given position.
	 * 
	 * @param playerId
	 *            the player for whom the orders shall be collected
	 * @param position
	 *            the position to collect the orders for, or <code>null</code>
	 *            for the position in focus for that player
	 * @return a structure containing a list of orders and a list of commandables
	 */
	public synchronized UserMapTileContextOptions collectMapTileContextOptions(final int playerId, final PlayfieldPosition position) {
		List<Order> orders = null;
		List<Commandable> commandables = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			final PlayfieldPosition pos;
			if (null == position) {
				pos = gameView.getPositionInFocus();
			} else {
				pos = position;
			}
			orders = gameView.collectOrderOptions(pos);
			commandables = gameView.collectCommandableOptions(pos);
		}
		return new UserMapTileContextOptions(orders, commandables);
	}

	public synchronized Order actionTargetSelectionAtPosition(final int playerId, final OrderTemplate targetSelection, final PlayfieldPosition pos) throws CircularReferenceException {
		Order ret = null;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.actionTargetSelectionAtPosition(gameView, targetSelection, pos);
		}
		return ret;
	}

	public synchronized boolean onActionSelectOrderMakesSense(final int playerId, final Order order) {
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			if (gameView.getMapViewMode() == MapViewMode.PRODUCTION) {
				final FieldAlike field = gameView.getAt(gameView.getPositionInFocus());
				if (field.isCity()) {
					gameView.nextCommandable((Commandable) field.getField());
				}
			}
			final Commandable commandableInFocus = gameView.getVisibleCommandableInFocus();
			logger.trace("onActionSelectOrderMakesSense: " + order.getToolTip(commandableInFocus));
			if ((null != commandableInFocus) && (order.makesSense(commandableInFocus))) {
				ret = true;
			}
		}
		return ret;
	}

	public synchronized boolean isOrderActionEnabled(final int playerId, final Order order) {
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.isOrderActionEnabled(order);
		}
		return ret;
	}

	public synchronized Order getMoveOrAttackOrder(final Unit unit, final PlayfieldPosition destination) {
		Order ret = null;
		final int playerId = unit.getPlayerId();
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.getMoveOrAttackOrder(unit, destination);
		}
		return ret;
	}

	public synchronized void setOrder(final int playerId, final Order order) {
		logger.trace("setOrder(playerId=" + playerId + "," + order + ")");
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			gameView.setOrder(order);
		}
	}

	public synchronized void setOrder(final Commandable commandable, final Order order) {
		logger.trace("setOrder(commandable=" + commandable + "," + order + ")");
		if (null != gameState) {
			commandable.setOrder(order);
		}
	}

	public synchronized boolean isOrderActionTargetModeEnabled(final int playerId, final OrderTemplate template) {
		boolean ret = false;
		if ((null != gameState) && ((Player.ID_NONE != playerId))) {
			final GameView gameView = gameState.getGameView(playerId);
			if (gameView.isAcceptingUserInput()) {
				final MapViewMode mapViewMode = gameView.getMapViewMode();
				if (mapViewMode == MapViewMode.NORMAL) {
					final Commandable cmdl = gameView.getVisibleCommandableInFocus();
					if ((null != cmdl) && (cmdl instanceof Unit)) {
						final Unit unit = (Unit) cmdl;
						ret = template.makesSense(gameView, unit);
					}
				}
			}
		}
		return ret;
	}

	public synchronized boolean isAcceptingUserInput(final int playerId) {
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			ret = gameView.isAcceptingUserInput();
		}
		return ret;
	}

	public synchronized void stepIfNot(final int playerId, final ProcessorSpeedType pst) {
		if (null != gameState) {
			if (!gameState.hasGameEnded()) {
				final int currentPlayerId = gameState.getCurrentPlayerId();
				if ((currentPlayerId != playerId) /* && (currentPlayerId != Player.ID_NONE) */) {
					gameState.getCurrentGameView().step(pst);
				}
			}
		}
	}

	public synchronized boolean step(final int playerId, final ProcessorSpeedType pst) {
		boolean ret = false;
		if ((null != gameState) && (Player.ID_NONE != playerId)) {
			final GameView gameView = gameState.getGameView(playerId);
			final boolean isFinishedAnimation = gameView.step(pst);
			if (isFinishedAnimation) {
				gameView.nextMove();
				if (gameView.isMyTurn()) {
					ret = true;
				}
			}
		}
		return ret;
	}

	private Logger logger = Logger.getLogger(getClass());
	{
		logger.setToShowThread(true);
		logger.setLogLevel(Level.INFO);
	}

	public synchronized void setEnvironment(final ApplicationEnvironment applEnv) {
		this.applEnv = applEnv;
	}

	private GameState gameState;
	private ApplicationEnvironment applEnv;
}
