package de.ped.empire.logic;

import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import de.ped.tools.Assert;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.Bitfield32;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A map where all the military units can act upon. With land, sea and cities, and even the military units themselves.
 * A map exists for all players' {@link GameView}s and the all-knowing view ({@link GameState}).
 * The all-knowing view's map knows exactly the content of the real map.
 * The player views represent the best knowledge these players currently have of this real map.
 * <p>
 * A map knows nothing of the dynamic aspects of the game (e.g., who's turn it is) or 
 * of the player's view mode (e.g., if in production view mode, flight path mode or normal mode).
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameMap implements Marshallable {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	/**
	 * Factory for {@link GameMap} when demarshalling.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Factory implements Marshallable {

		public static GameMap create(final Marshaller marshaller, final GameProperties props, final GameView view) {
			return new GameMap(marshaller, props, view);
		}

		public Factory(final GameProperties props, final GameView view) {
			this.props = props;
			this.view = view;
		}

		@Override
		public void marshal(final Marshaller marshaller) {
			throw new IllegalArgumentException("Factory is never marshalled.");
		}

		@Override
		public Marshallable demarshal(final Marshaller marshaller) {
			return create(marshaller, props, view);
		}

		private final GameProperties props;
		private final GameView view;
	}

	public static final PlayfieldDimension DEFAULT_SIZE = new PlayfieldDimension(15, 10);

	private Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
	}

	/** Persistent. */
	private FieldAlike[][] map;

	/** Non-persistent, as stored in GameState. */
	private final GameProperties props;

	/** Non-persistent. */
	private final GameView view;

	/** Non-persistent. */
	private boolean wasFinished;

	/**
	 * Creates a new game map of the defined size with only borders (if
	 * applicable) and sea.
	 * 
	 * @param properties
	 *            the game properties describing the game map size and border
	 *            settings
	 */
	public GameMap(final GameProperties properties) {
		this.props = properties;
		this.view = null;
		final PlayfieldDimension size = props.getSize();
		createMap(size);
		if (properties.hasBorder(PlayfieldPosition.Y)) {
			for (int x = 0; x < size.getWidth(); x++) {
				addField(new Field(x, props.getYMin(x),     ImageKey.FIELD_ID_BORDER));
				addField(new Field(x, props.getYMax(x) - 1, ImageKey.FIELD_ID_BORDER));
			}
		}
		final int ydo = props.getYMin(size.getWidth() - 1);
		if (properties.hasBorder(PlayfieldPosition.X)) {
			for (int y = 0; y < size.getHeight(); y++) {
				addField(new Field(0, y, ImageKey.FIELD_ID_BORDER));
				addField(new Field(size.getWidth() - 1, y + ydo, ImageKey.FIELD_ID_BORDER));
			}
		}
		final Rectangle seaArea = properties.getUsableMapArea(1);
		for (int x = seaArea.x; x < seaArea.x + seaArea.width; x++) {
			final int ya = seaArea.y + props.getYMin(x);
			final int ye = ya + seaArea.height;
			for (int y = ya; y < ye; y++) {
				addField(new Field(x, y, ImageKey.FIELD_ID_SEA));
			}
		}
		wasFinished = false;
	}

	/**
	 * Builds a player's game view map based on a game state map.
	 * 
	 * @param stateMap
	 *            the game state map
	 * @param view
	 *            the game view the view map will belong to
	 */
	GameMap(final GameMap stateMap, final GameView view) {
		this.props = stateMap.props;
		this.view = view;
		final boolean isWithFogOfUncertainty = stateMap.getProperties().getRulesComplexity().isWithFogOfUncertainty;
		final PlayfieldDimension size = props.getSize();
		createMap(size);
		for (int x = 0; x < map.length; x++) {
			for (int y = 0; y < map[0].length; y++) {
				final FieldAlike fa = stateMap.getAt(x, y);
				if (null != fa) {
					final FieldAlike field = new ViewField((Field) fa, isWithFogOfUncertainty);
					addField(field);
				}
			}
		}
		wasFinished = false;
	}

	public GameMap(final Marshaller marshaller, final GameProperties props, final GameView view) {
		this.props = props;
		this.view = view;
		final PlayfieldDimension size = props.getSize();
		createMap(size);
		logger.trace("GameMap(" + view + "): stream index before reading = " + marshaller.getReadPos());
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				if (null == view) {
					addField(Field.create(marshaller, x, y));
				} else {
					addField(ViewField.create(marshaller, view, x, y));
				}
			}
		}
		logger.trace("GameMap(" + view + "): stream index after reading = " + marshaller.getReadPos());
		wasFinished = true;
	}

	private void createMap(final PlayfieldDimension size) {
		final PlayfieldDimension memSize = props.calcMapMemorySize(size);
		map = new FieldAlike[memSize.getWidth()][memSize.getHeight()];
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		final PlayfieldDimension size = props.getSize();
		logger.debug("GameMap(" + view + "): : stream index before reading = " + marshaller.getWritePos());
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				getAt(x, y).marshal(marshaller);
			}
		}
		logger.debug("GameMap(" + view + "): : stream index after reading = " + marshaller.getWritePos());
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		return new GameMap(marshaller, props, view);
	}

	void marshalLivingUnits(final Marshaller marshaller) {
		final Collection<UnitAlike> livingUnits = getUnits();
		for (final UnitAlike unit : livingUnits) {
			marshaller.writeMarshallable(unit.getKey());
			if ((unit instanceof ViewUnit) && (isOtherView(unit))) {
				unit.marshal(marshaller);
				logger.debug("GameMap[" + view.getPlayer().getId()  + "] marshals ViewUnit " + unit.getKey().toString());
			}
		}
		marshaller.writeMarshallable(CommandableKey.NULL_KEY);
	}

	/**
	 * Demarshals all the units visible in this map.
	 * @param marshaller the marshaller to read the unit keys from
	 * @param gameState the game state the units are in
	 */
	void demarshalLivingUnits(final Marshaller marshaller, final GameState gameState) {
		while(true) {
			final CommandableKey key = (CommandableKey) marshaller.readMarshallable(CommandableKey.factory);
			if (CommandableKey.NULL_KEY.equals(key)) {
				break;
			}
			final Unit base = GameMap.findUnit(key, gameState.getAllUnits());
			Assert.assertNotNull(base);
			final UnitAlike unit;
			if (null == view) {
				unit = base;
			} else if (isOtherView(base)){ 
				unit = new ViewUnit(marshaller, gameState, key);
				logger.debug("GameMap[" + view.getPlayer().getId() + "] demarshals ViewUnit " + key.toString());
			} else {
				unit = new ViewUnit(base, gameState.getTurn(), isOtherView(base));
			}
			addUnit(unit);
		}
	}
	
	boolean isOtherView(final Commandable cmdl) {
		return (null != view) && (view.isUsingFogOfWar()) && (view.getPlayer().getId() != cmdl.getPlayerId());
	}

	public void finish() {
		addGraphicalSugar();
		wasFinished = true;
	}

	public void addAtRandom(final GameProperties props) {
		final Rectangle mapSize = props.getUsableMapArea(2); // Border and sea, we will never see land hit the border.
		final int areaSize = mapSize.width * mapSize.height;
		final int cityNeighbors = 6; // 5 is too low and leads to infinite loops in rare cases
		final int numberOfCities = Math.max(Player.MAX_NUMBER_OF_PLAYERS,
				Player.MAX_NUMBER_OF_PLAYERS + (int) ((long) Math.max(0, areaSize - Player.MAX_NUMBER_OF_PLAYERS * cityNeighbors)
						* (long) (props.getCityDensity()) / (100L * cityNeighbors)));
		final LinkedHashSet<FieldAlike> seeds = new LinkedHashSet<FieldAlike>();
		logger.debug("Cities to place: " + numberOfCities);
		for (int id = 0; id < numberOfCities; id++) {
			addSeed(seeds, createRandomCity(id));
		}
		Assert.assertEquals(numberOfCities, seeds.size());
		Assert.assertEquals(numberOfCities, getAllCities().size());
		final int numberOfLand = (int) ((long) (areaSize - numberOfCities) * (long) props.getLandRatio() / 150L);
		final int numberOfSeeds = (int) (numberOfLand * (110 - props.getContinentSize()) / 110); // * (100 - props.getCityDensity()) / 100;
		logger.debug("Lands seed to place: " + numberOfSeeds);
		for (int land = 0; land < numberOfSeeds; land++) {
			addSeed(seeds, createRandomLandSeed());
		}
		Assert.assertEquals(numberOfCities + numberOfSeeds, seeds.size());
		Assert.assertEquals(numberOfCities, getAllCities().size());
		logger.debug("Lands to place: " + (numberOfLand - numberOfSeeds));
		LinkedList<FieldAlike> seeds2 = new LinkedList<FieldAlike>(seeds);
		for (int land = numberOfSeeds; land < numberOfLand; land++) {
			for (int tries = 0; tries < 10; tries++) {
				final int index = MathUtil.random(seeds2.size());
				final FieldAlike field = growRandomLand(seeds2.get(index));
				if (field != null) {
					addSeed(seeds2, field);
					break;
				} else {
					seeds2.remove(index);
				}
			}
		}
		Assert.assertEquals(numberOfCities, getAllCities().size());
		addRivers();
		Assert.assertEquals(numberOfCities, getAllCities().size());
		logMap();
	}

	private void addSeed(final Collection<FieldAlike> seeds, final FieldAlike field) {
		for (Iterator<FieldAlike> it = seeds.iterator(); it.hasNext(); ) {
			FieldAlike fa = it.next();
			if (PlayfieldPosition.equals(fa.getPosition(), field.getPosition())) {
				throw new IllegalStateException("Position " + fa.getPosition().toString() + " already in use by " + fa.toString() + " for " + field.toString());
			}
		}
		seeds.add(field);
	}

	private void addRivers() {
		final Collection<Field> shores = findAllShores();
		for (final Field field : shores) {
			final PlayfieldPosition pos = field.getPosition();
			final List<DigitalJoystickDirection> dirs = props.getShoreNeighborDirections(pos);
			int dirIdx = MathUtil.random(dirs.size());
			DigitalJoystickDirection dir = dirs.get(dirIdx);
			PlayfieldPosition npos = props.wrap(pos, dir);
			final PlayfieldPosition gpos = props.wrap(pos, dir.getOppositeDirection());
			if (getAt(npos).isLand() 				// Rivers must be at least 2 fields long
					&& (!getAt(gpos).isLand()) 		// Rivers must float to the sea without bend on the coastal field
					&& (0 == MathUtil.random(3))) 	// Still too many rivers
			{
				setRiver(pos);
				setRiver(npos);
				boolean isFlowing;
				do {
					isFlowing = false;
					if (0 == MathUtil.random(4)) { // Sometimes bend a bit
						if (0 == MathUtil.random(2)) {
							dirIdx++; // turn right on your float inwards
						} else {
							dirIdx--; // turn left on your float inwards
						}
						dirIdx = MathUtil.bound(dirIdx, 0, dirs.size() - 1);
						dir = dirs.get(dirIdx);
					}
					if (0 < MathUtil.random(4)) { // Most times try to continue river, sometimes just stop
						npos = props.wrap(npos, dir);
						PlayfieldPosition n2pos = props.wrap(npos, dir);
						final FieldAlike nposField = getAt(npos);
						final FieldAlike n2posField = getAt(n2pos);
						// New field is land, but no river, and even following field is land, but no river (rivers don't cross or float from sea to sea)
						if (nposField.isLand() && (nposField.getTemplateId() != ImageKey.FIELD_ID_RIVER) 
								&& n2posField.isLand() && (n2posField.getTemplateId() != ImageKey.FIELD_ID_RIVER)) {
							isFlowing = true;
							setRiver(npos);
						}
					}
				} while (isFlowing);
			}
		}
	}

	private void setRiver(final PlayfieldPosition npos) {
		if (!(getAt(npos).isCity())) {
			setAt(new Field(npos.getX(), npos.getY(), ImageKey.FIELD_ID_RIVER));
		}
	}

	private Collection<Field> findAllShores() {
		final Collection<Field> ret = new ArrayList<Field>(); 
		final PlayfieldDimension size = props.getSize();
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final Field field = (Field) getAt(x, y);
				if (field.isLand()) {
					final Collection<FieldAlike> neighbors = getNeighborsOf(field, 
							props.getAllVectorsToAirDistance(props.getTopologyModeForShores(), field.getPosition(), 1, false));
					for (final FieldAlike neighbor : neighbors) {
						final Field nf = (Field) neighbor;
						if (nf.getTemplateId() == ImageKey.FIELD_ID_SEA) {
							ret.add(field);
							break;
						}
					}
				}
			}
		}
		return ret;
	}

	private void addField(final FieldAlike field) {
		Assert.assertEquals((field instanceof Field), (null == view));
		try {
			setAt(field);
		} catch (final Throwable ex) {
			logger.error("addField: ", ex);
		}
	}

	/**
	 * Creates a land seed at a random sea field.
	 * 
	 * @return the new land field
	 */
	public Field createRandomLandSeed() {
		int x = 0;
		int y = 0;
		boolean isOkay;
		final Rectangle area = props.getUsableMapArea(2);
		do {
			x = area.x + MathUtil.random(area.width);
			y = area.y + MathUtil.random(area.height) + props.getYMin(x);
			isOkay = true;
			final Field field = (Field) getAt(x, y);
			if (field.getTemplateId() != ImageKey.FIELD_ID_SEA) {
				isOkay = false;
				logger.debug("Land at (" + x + "," + y + ") impossible, because of " + field.getTemplate().messageKeyPrefix);
			}
		} while (!isOkay);
		final Field field = createLand(x, y);
		addField(field);
		return field;
	}

	public FieldAlike growRandomLand(FieldAlike seed) {
		FieldAlike fieldToUse = null;
		for (int distance = 1; distance < 5; distance++) {
			final Collection<FieldAlike> potentials = getAllFieldsInAirDistance(seed, distance, false, 2);
			final List<FieldAlike> options = new LinkedList<FieldAlike>();
			for (final FieldAlike field : potentials) {
				if (field.getTemplateId() == ImageKey.FIELD_ID_SEA) {
					options.add(field);
				}
			}
			if (!options.isEmpty()) {
				fieldToUse = options.get(MathUtil.random(options.size()));
				break;
			}
		}
		if (null != fieldToUse) { // Fill the field with land
			final PlayfieldPosition position = fieldToUse.getPosition();
			final Field field = createLand(position.getX(), position.getY());
			addField(field);
			fieldToUse = field;
		}
		return fieldToUse;
	}

	private static final int[] LAND_PROBS = new int[] { 40, 10, 00, 10, 05, };
	private static final byte[] LAND_IDS = new byte[] {
		ImageKey.FIELD_ID_LAND,
		ImageKey.FIELD_ID_ROUGH,
		ImageKey.FIELD_ID_RIVER, // rivers are created elsewhere
		ImageKey.FIELD_ID_FOREST,
		ImageKey.FIELD_ID_MOUNTAIN,
	};

	private Field createLand(int x, int y) {
		final Field field = new Field(x, y, LAND_IDS[MathUtil.randomDistributed(LAND_PROBS)]);
		return field;
	}

	/**
	 * Creates a city at a random field, but not near another city.
	 * 
	 * @param id
	 *            the city id
	 * @return the created city
	 */
	public Field createRandomCity(int id) {
		int x = 0;
		int y = 0;
		boolean isOkay;
		final Rectangle area = props.getUsableMapArea(2);
		do {
			x = area.x + MathUtil.random(area.width);
			y = area.y + MathUtil.random(area.height) + props.getYMin(x);
			final FieldAlike here = getAt(x, y);
			if (here.isCity()) {
				isOkay = false;
			} else {
				isOkay = true;
				for (final FieldAlike neighbor : getNeighborsOf(here)) {
					if (null == neighbor) {
						logger.error("Neighbor of " + here.getPosition().toString() + " was null!");
					}
					if (neighbor.isCity()) {
						isOkay = false;
						logger.debug("City at (" + x + "," + y + ") impossible, because of other city at " + neighbor.getPosition().toString());
					}
				}
			}
		} while (!isOkay);
		final Field ret = new FieldCity(id, x, y);
		addField(ret);
		return ret;
	}

	/**
	 * Restores that graphical sugar that is not stored and needs therefore be added after loading.
	 * Adds coastal lines and glues rivers, roughs, forests and mountains together.
	 */
	private void addGraphicalSugar() {
		final PlayfieldDimension size = props.getSize();
		final Templates templates = Templates.instance();
		final ConditionChecker checkerForSea = new ConditionCheckerForSea();
		final ConditionChecker checkerForRiver = new ConditionCheckerForRiver();
		final ConditionChecker checkerForRough = new ConditionCheckerFor(templates.ROUGH);
		final ConditionChecker checkerForForest = new ConditionCheckerFor(templates.FOREST);
		final ConditionChecker checkerForMountain = new ConditionCheckerFor(templates.MOUNTAIN);
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final Field field = (Field) getAt(x, y);
				final FieldTemplate template = field.getTemplate();
				ConditionChecker checker = null;
				if (template.equals(templates.SEA)) {
					checker = checkerForSea;
				} else if (template.equals(templates.RIVER)) {
					checker = checkerForRiver;
				} else if (template.equals(templates.ROUGH)) {
					checker = checkerForRough;
				} else if (template.equals(templates.FOREST)) {
					checker = checkerForForest;
				} else if (template.equals(templates.MOUNTAIN)) {
					checker = checkerForMountain;
				}
				int key = field.getImageKey();
				if (null != checker) {
					final int neighborMod = addKeyBitsForField(key, checker, field.getPosition());
					key = Bitfield32.setBits(key, ImageKey.BA_NEIGHBOR_MOD, neighborMod);
				}
				
				if (0 != (TileTemplate.getTTD(key).normalizationMask & Bitfield32.bitmask(ImageKey.BA_LOOK))
//						&& (0 != MathUtil.modulo(x + y, 2))
						&& (0 != MathUtil.random(2))
						) {
					key = Bitfield32.setBits(key, ImageKey.BA_LOOK, 1);
				}
				if (0x00 != key) {
					field.setImageKey(key);
				}
			}
		}
	}

	int addKeyBitsForUnexploredBorder(final PlayfieldPosition pos) {
		final ConditionChecker checker = new ConditionCheckerFor(Templates.instance().UNEXPLORED);
		return addKeyBitsForField(0, checker, pos);
	}

	private int addKeyBitsForField(int key, final ConditionChecker checker, final PlayfieldPosition pos) {
		key = 0;
		// Are we allowed to wrap borders? We aren't for border fields, those should be drawn without unknown neighbor on the edge of the world side.
		if (!checker.isBorderRelevant() || (getAt(pos).getTemplateId() != ImageKey.FIELD_ID_BORDER)) {
			final ArrayList<DigitalJoystickDirection> ndirs = props.getShoreNeighborDirections(pos);
			final int[] nbits = getBitsForShoreNeighborDirections();
			for (int i = 0; i < ndirs.size(); i++) {
				key |= addKeyBitForField(checker, pos, ndirs.get(i), (1 << nbits[i]));
			}
		}
		return key;
	}

	int[] getBitsForShoreNeighborDirections() {
		final int[] bits;
		switch(props.getTopologyMode()) {
		case HEXAGON:
			bits = new int[] { 0, 1, 2, 3, 4, 5, };
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			bits = new int[] { 2, 0, 3, 1, };
		}
		return bits;
	}

	/**
	 * Returns either 0 or <em>bit</em>, depending on the checker's output for the given neighbor of the given field
	 * @param checker the checker to run to decide
	 * @param pos the position to look the neighbors for
	 * @param dir the direction of the neighbor
	 * @param bit the bit to set if true
	 * @return either 0 or <em>bit</em>
	 */
	private int addKeyBitForField(final ConditionChecker checker, final PlayfieldPosition pos, final DigitalJoystickDirection dir, final int bit) {
		int ret = 0;
		final PlayfieldPosition np = pos.add(props.renormDistanceVector(dir.getDistanceVector(), pos));
		final PlayfieldPosition npos = props.wrap(np);
		final FieldAlike field = getAt(npos);
		if (null == field) {
			throw new NullPointerException(" at map position " + npos.toString());
		} else {
			if (checker.isToSetBit(field)) {
				ret = bit;
			}
			return ret;
		}
	}

	/**
	 * Helps to define if an image key bit needs to be set. Used only internally.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static interface ConditionChecker {
		public boolean isToSetBit(FieldAlike field);
		public boolean isBorderRelevant();
	}

	/**
	 * Helps to define if an image key bit needs to be set. Used only internally.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ConditionCheckerForSea implements ConditionChecker {
		@Override
		public boolean isToSetBit(FieldAlike field) {
			return field.isLand() && (Templates.instance().UNEXPLORED.id != field.getTemplate().id);
		}

		@Override
		public boolean isBorderRelevant() {
			return false;
		}
	}

	/**
	 * Helps to define if an image key bit needs to be set. Used only internally.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ConditionCheckerForRiver implements ConditionChecker {
		@Override
		public boolean isToSetBit(FieldAlike field) {
			return field.getTemplate().isRiverNeighbor;
		}

		@Override
		public boolean isBorderRelevant() {
			return false;
		}
	}

	/**
	 * Helps to define if an image key bit needs to be set. Used only internally.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ConditionCheckerFor implements ConditionChecker {
		private ConditionCheckerFor(final FieldTemplate fieldTemplate) {
			this.fieldTemplate = fieldTemplate;
		}
		
		private final FieldTemplate fieldTemplate;
		
		@Override
		public boolean isBorderRelevant() {
			return (Templates.instance().UNEXPLORED == fieldTemplate);
		}
		
		@Override
		public boolean isToSetBit(FieldAlike field) {
			return field.getTemplate().equals(fieldTemplate);
		}
	}

	public GameProperties getProperties() {
		return props;
	}

	public PlayfieldDimension getSize() {
		return props.getSize();
	}

	public FieldAlike getAt(final int x, final int y) {
		return map[x][y];
	}

	public FieldAlike getAt(final PlayfieldPosition pos) {
		return getAt(pos.getX(), pos.getY());
	}

	private void setAt(final FieldAlike field) {
		final PlayfieldPosition pos = field.getPosition();
		map[pos.getX()][pos.getY()] = field;
	}

	/**
	 * Initializes the map to start a new game. Currently its main job is to
	 * find starting cities for all players.
	 * 
	 * @param gameState
	 *            the game state (should be initial) initialize the map for
	 * @exception GameInitializationException if initialization was not possible
	 */
	public void initFor(final GameState gameState) throws GameInitializationException {
		findStartingCitiesFor(gameState);
		// Initialize cities
		for (final FieldCity city : getAllCities()) {
			city.setProductivity();
		}
	}

	private void findStartingCitiesFor(final GameState gameState) throws GameInitializationException {
		// Collect a list of all active players for the game.
		final ArrayList<Player> players = new ArrayList<Player>(6);
		for (final Player player : gameState.getProperties().getPlayers()) {
			if (player.isPlayer()) {
				players.add(player);
			}
		}
		// Find all pickable cities. Prefer coastal cities.
		final List<FieldCity> cities = findCities(Player.ID_NONE);
		final List<FieldCity> coastalCities = new ArrayList<FieldCity>();
		final List<FieldCity> continentalCities = new ArrayList<FieldCity>();
		for (FieldCity city : cities) {
			if (city.isCoastal(this)) {
				coastalCities.add(city);
			} else {
				continentalCities.add(city);
			}
		}
		final List<FieldCity> pickableCities;
		if (coastalCities.size() >= players.size()) {
			pickableCities = coastalCities;
		} else if (continentalCities.size() >= players.size()) {
//			pickableCities = continentalCities;
			// TODO: Are there at least 6 cities on the same continent? If yes, pick from them.
			// throw new GameInitializationException(); // FIXME: Get this line back
			pickableCities = cities;
		} else {
			throw new GameInitializationException();
		}
		// Pick the cities randomly, but not too randomly.
		List<FieldCity> bestList = null;
		int bestRating = 0;
		if (pickableCities.size() == players.size()) {
			bestList = new ArrayList<FieldCity>(pickableCities);
			MathUtil.shuffle(bestList, null);
		} else {
			for (int tries = 0; (tries < 30) || (bestList == null); tries++) {
				final List<FieldCity> picked = new LinkedList<FieldCity>();
				final List<FieldCity> toPick = new LinkedList<FieldCity>(pickableCities);
				for (int i = 0; i < players.size(); i++) {
					final FieldCity city = toPick.get(MathUtil.random(toPick.size()));
					picked.add(city);
					toPick.remove(city);
				}
				int rating = Integer.MAX_VALUE;
				for (int i = 1; i < picked.size(); i++) {
					for (int j = 0; j < i; j++) {
						int dist = props.calcAirDistance(picked.get(i).getPosition(), picked.get(j).getPosition());
						rating = Math.min(rating, dist);
					}
				}
				if (bestRating < rating) {
					bestRating = rating;
					bestList = picked;
				}
			}
		}
		int listIdx = 0;
		for (int i = 0; i < players.size(); i++) {
			bestList.get(listIdx).setPlayerId(players.get(i).getId());
			listIdx++;
		}
		logger.debug("Final cities rating was " + bestRating);
	}

	/**
	 * Marks all cities for this game state (needed to be treated as
	 * commandable). A game map that is part of a game needs to know its game
	 * state, but as it's not persistent, it needs to be set at the beginning of
	 * a game or after loading a game, for each view's map.
	 * 
	 * @param gameState
	 *            the game state to set
	 */
	void setGameState(final GameState gameState) {
		if (null == view) {
			for (final FieldCity city : getAllCities()) {
				city.setGameState(gameState);
			}
		}
	}

	public List<FieldCity> getAllCities() {
		// TODO: Could be calculated up-front to speed this up
		final List<FieldCity> cities = new LinkedList<FieldCity>();
		for (int x = 0; x < map.length; x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final FieldAlike field = getAt(x, y);
				if (null == field) {
					logger.warn("(" + x + "," + y + ")");
				}
				if (field.isCity()) {
					cities.add((FieldCity) field.getField());
				}
			}
		}
		return cities;
	}

	public Collection<FieldAlike> getNeighborsOf(final FieldAlike middleField) {
		return getNeighborsOf(middleField, 
				props.getAllVectorsToAirDistance(props.getTopologyMode(), middleField.getPosition(), 1, false));
	}

	private List<FieldAlike> getNeighborsOf(final FieldAlike middleField, final List<PlayfieldDistance> neighborVectors) {
		final List<FieldAlike> ret = new ArrayList<FieldAlike>(neighborVectors.size());
		final PlayfieldPosition pos = new PlayfieldPosition(); // re-used for performance reasons
		for (PlayfieldDistance neighborVector : neighborVectors) {
			pos.addInplace(middleField.getPosition(), neighborVector);
			props.wrapInside(pos);
			final FieldAlike field = getAt(pos);
			if (null == field) {
				logger.error("Found no neighbor for " + middleField.getPosition().toString() + " at " + pos.toString());
			}
			ret.add(field);
		}
		return ret;
	}

	/**
	 * Returns all fields that are in air distance <em>distance</em> (or less,
	 * if <em>isWithFilling</em> is <code>true</code>) to <em>middleField</em>.
	 * 
	 * @param middleField
	 *            the field to start from
	 * @param distance
	 *            the distance to collect the other fields in
	 * @param isWithFilling
	 *            if <code>true</code>, the result includes the whole rectangle
	 *            of fields (including the middle fields), otherwise the result
	 *            contains only the border of the rectangle
	 * @param edgeThickness
	 *            Fields this near or nearer to the edge of the map are left
	 *            out. This is needed for land creation.
	 * @return the list of fields in the given distance (or less)
	 */
	public List<FieldAlike> getAllFieldsInAirDistance(final FieldAlike middleField, final int distance, final boolean isWithFilling, final int edgeThickness) {
		final List<FieldAlike> ret = new LinkedList<FieldAlike>();
		if (0 == distance) {
			ret.add(middleField);
		} else if (0 < distance) {
			final List<PlayfieldDistance> vectors = props.getAllVectorsToAirDistance(middleField.getPosition(), distance, isWithFilling);
			for (PlayfieldDistance vector : vectors) {
				addIfWithinRange(ret, middleField, vector, edgeThickness);
			}
		}
		return ret;
	}

	private void addIfWithinRange(final List<FieldAlike> ret, final FieldAlike middleField, final PlayfieldDistance dist, final int edgeThickness) {
		final PlayfieldPosition npos = props.wrap(middleField.getPosition().add(dist));
		if (props.containsUsableMapArea(npos, edgeThickness)) {
			final int nposX = npos.getX();
			final int nposY = npos.getY();
			if ((0 <= nposX) && (nposX < map.length) && (0 <= nposY) && (nposY < map[0].length)) {
				final FieldAlike field = getAt(npos);
				if (null != field) {
					ret.add(field);
				}
			}
		}
	}

	public FieldCity findCity(final int id) {
		// TODO: Could take use of a pre-calculated getCities()
		for (int x = 0; x < map.length; x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final FieldAlike field = getAt(x, y);
				if (field.isCity()) {
					final FieldCity city = (FieldCity) field.getField();
					if (city.getKey().getId() == id) {
						return city;
					}
				}
			}
		}
		return null;
	}

	public List<FieldCity> findCities(final int playerId) {
		// TODO: Could take use of a pre-calculated getCities()
		final List<FieldCity> cities = new LinkedList<FieldCity>();
		for (int x = 0; x < map.length; x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final FieldAlike field = getAt(x, y);
				if (null == field) {
					throw new NullPointerException(" at map position (" + x + "," + y + ")");
				}
				if (field.isCity()) {
					if (field.getPlayerId() == playerId) {
						final FieldCity city = (FieldCity) field.getField();
						cities.add(city);
					}
				}
			}
		}
		return cities;
	}

	public void setField(final Field field) {
		Assert.assertNull(view);
		setAt(field);
		if (wasFinished) {
			addGraphicalSugar();
		}
	}

	public void logMap() {
		for (int y = 0; y < map[0].length; y++) {
			final StringBuffer buf = new StringBuffer();
			for (int x = 0; x < map.length; x++) {
				final FieldAlike f = getAt(x, y);
				buf.append((null == f) ? " " : f.getLogRepresentation());
			}
			logger.info(buf.toString());
		}
	}

	boolean isLandNearby(final FieldAlike middleField) {
		boolean ret = false;
		Collection<FieldAlike> neighbors = getNeighborsOf(middleField);
		for (final FieldAlike neighbor : neighbors) {
			if (neighbor.isLand()) {
				ret = true;
				break;
			}
		}
		return ret;
	}

	public boolean isCityToConquerNearby(final FieldAlike middleField, final int playerId) {
		boolean ret = false;
		final Collection<FieldAlike> neighbors = getNeighborsOf(middleField);
		for (final FieldAlike neighbor : neighbors) {
			if ((neighbor.isCity()) && (playerId != neighbor.getPlayerId())) {
				ret = true;
				break;
			}
		}
		return ret;
	}

	GameMap createViewMap(final GameView view) {
		GameMap ret = new GameMap(this, view);
		return ret;
	}

	/** 
	 * Returns the number of non-border fields on the map. 
	 * Needed for calculation of how much of the world is explored. 
	 * @return the number of non-border fields on the map
	 */
	public int countAllFields() {
		final Rectangle usable = props.getUsableMapArea(1);
		return (usable.width * usable.height);
	}

	/**
	 * Returns the numer of non-unexplored non-border fields on the map.
	 * Needed for calculation of how much of the world is explored. 
	 * @return the numer of non-unexplored non-border fields on the map
	 */
	public int countKnownFields() {
		int ret = 0;
		final Rectangle usable = props.getUsableMapArea(1);
		for (int x = usable.x; x < usable.x + usable.width; x++) {
			for (int y = usable.y + props.getYMin(x); y < usable.y + props.getYMin(x) + usable.height; y++) {
				if (ImageKey.FIELD_ID_UNEXPLORED != getAt(x, y).getTemplateId()) {
					ret++;
				}
			}
		}
		return ret;
	}

	/** 
	 * @return all units known as alive on this map, including sightings
	 */
	public Collection<UnitAlike> getUnits() {
		return livingUnits.values();
	}

	/** 
	 * @return all units known as alive on this map, including sightings
	 */
	public Map<CommandableKey, UnitAlike> getUnitsMap() {
		return livingUnits;
	}

	/** 
	 * Adds to or updates a unit or sighting to the map.
	 * @param unit the unit to add or update
	 */
	void addUnit(final UnitAlike unit) {
		final UnitAlike oldUnit = livingUnits.get(unit.getKey());
		if (null != oldUnit) {
			getAt(oldUnit.getPosition()).getUnits().remove(oldUnit);
		}
		getAt(unit.getPosition()).getUnits().add(unit);
		livingUnits.put(unit.getKey(), unit);
	}

	/** 
	 * Removes a unit from the list of living units.
	 * @param unit the unit to remove
	 * @return the unit removed, or <code>null</code> if there was none
	 */
	UnitAlike removeUnit(final UnitAlike unit) {
		getAt(unit.getPosition()).getUnits().remove(unit);
		return livingUnits.remove(unit.getKey());
	}

	private static final boolean IS_USING_LOCAL_UNIT_DATA = false;

	/**
	 * @param pos
	 *            the position to look for units
	 * @return a collection of all units at the given position (as far as this
	 *         map knows), or <code>null</code> if there is no known unit.
	 *         Beware: You get access to the inner parts of the map, so if you
	 *         want to change the collection, first clone it and do your
	 *         operations on the clone.
	 */
	public Collection<UnitAlike> findUnitsAtPosition(final PlayfieldPosition pos) {
		if (IS_USING_LOCAL_UNIT_DATA) {
			return getAt(pos).getUnits();
		} else {
			Collection<UnitAlike> ret = null;
			for (final UnitAlike unit : getUnits()) {
				// FIXME: unit.getPosition() should return a playfield position
				if (pos.equals(new PlayfieldPosition(unit.getPosition()))) {
					if (null == ret) {
						ret = new LinkedList<UnitAlike>();
					}
					ret.add(unit);
				}
			}
			return ret;
		}
	}

	public boolean hasUnitsAtPosition(final PlayfieldPosition pos) {
		if (IS_USING_LOCAL_UNIT_DATA) {
			 return !getAt(pos).getUnits().isEmpty();
		} else {
			boolean ret = false;
			for (final UnitAlike unit : getUnits()) {
				if (pos.equals(unit.getPosition())) {
					ret = true;
					break;
				}
			}
			return ret;
		}
	}

	static Unit findUnit(final CommandableKey key, final Collection<Unit> list) {
		Unit ret = null;
		for (final Unit u : list) {
			if (u.getKey().equals(key)) {
				ret = u;
				break;
			}
		}
		return ret;
	}

	/** 
	 * @return all living units and all cities, no matter what owner
	 */
	List<Commandable> findAllCommandables() {
		final List<Commandable> ret = new LinkedList<Commandable>();
		ret.addAll(getAllCities());
		ret.addAll(getUnits());
		return ret;
	}

	/**
	 * Marks the whole map as outdated. 
	 * For the non-global map, normally called at the start of a turn for the player.
	 */
	void outdate() {
		Assert.assertNotNull(view); // Never outdate the global world map
		if (view.isUsingFogOfWar()) {
			final int width = props.getSize().getWidth();
			for (int x = 0; x < width; x++) {
				for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
	//					final GameMapField worldField = gameState.getMap().getAt(x, y);
	//					if (0 != (worldField.getImageKey(null) & Bitfield32.bitmask(ImageKey.BA_OUTDATED))) {
	//						logger.fatal("World view is outdated for (" + x + "," + y + "): Bitfield=" + worldField.getImageKey(null));
	//					}
					final FieldAlike field = getAt(x, y);
	//					if (worldField == field) {
	//						logger.fatal("World view is identical for (" + x + "," + y + ")");
	//					}
					field.setImageKey(field.getImageKey() | Bitfield32.bitmask(ImageKey.BA_OUTDATED));
	//					if (0 != (worldField.getImageKey(null) & Bitfield32.bitmask(ImageKey.BA_OUTDATED))) {
	//						logger.fatal("Setting outdated also changed world view for (" + x + "," + y + "): Bitfield=" + worldField.getImageKey(null));
	//					}
				}
			}
			for (final UnitAlike unit : getUnits()) {
				if (unit.getPlayerId() != view.getPlayer().getId()) {
					unit.setImageKey(unit.getImageKey() | Bitfield32.bitmask(ImageKey.BA_OUTDATED));
				}
			}
		}
	}

	/**
	 * Returns how many move steps one move onto the field at the given position
	 * a given unit would cost. Needed for forward calculation.
	 * 
	 * @param cmdlTemplateId
	 *            the unit type to move onto that field
	 * @param targetPosition
	 *            the position of the field
	 * @return how many move steps one move onto the field at the given position
	 *         a given unit would cost
	 */
	int calcStepsToConsume(final int cmdlTemplateId, final PlayfieldPosition targetPosition) {
		int ret = 1;
		final RulesComplexity rulesComplexity = getProperties().getRulesComplexity();
		final int unitsSlowedDown = getAt(targetPosition).getTemplate().rcdvs[rulesComplexity.ordinal()].unitsSlowedDown; 
		if (Bitfield32.isBit(unitsSlowedDown, cmdlTemplateId)) {
			// Empire Deluxe Manual page 135, rule j: Move is reduced to 1 in forest, river and rough terrain.
			ret = 2;
		}
		return ret;
	}

	/**
	 * In case of the game state map, a list of all units alive;
	 * in case of game view maps, a list of all the unit representatives known in this game view. 
	 * Persistent as reference to allUnits. 
	 */
	private Map<CommandableKey, UnitAlike> livingUnits = new TreeMap<CommandableKey, UnitAlike>();
}
