package de.ped.empire.logic;

import java.util.ConcurrentModificationException;
import java.util.Set;

import de.ped.empire.logic.MoveResult.CommandableInfo;
import de.ped.empire.logic.MoveResult.Role;
import de.ped.tools.Bitfield32;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;

/**
 * All the data to draw the map content efficiently, 
 * as a data structure that can be transferred between game server and view clients.
 * Will be built by the GameView and sent to the Canvas when refactoring is done.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameMapInfo {
	public static final int UNDEFINED = 0xffffffff;
	public GameMapInfo(final GameView gameView) {
		if (null != gameView) {
			size = gameView.getMapSize();
			final PlayfieldDimension memSize = gameView.getMap().getProperties().calcMapMemorySize(size);
			mapTileObjectKeys = new int[memSize.getWidth()][memSize.getHeight()];
			final MapViewMode mapViewMode = gameView.getMapViewMode();
			final PlayfieldPosition fieldpos = new PlayfieldPosition(); // re-used for performance reasons
			for (int x = 0; x < size.getWidth(); x++) {
				for (int y = gameView.getMap().getProperties().getYMin(x); y < gameView.getMap().getProperties().getYMax(x); y++) {
					fieldpos.setLocation(x, y);
					final FieldAlike field = gameView.getAt(fieldpos);
					int imgKey = field.getImageKey();
					if ((MapViewMode.PRODUCTION == mapViewMode) 
							&& (field.isCity()) 
							&& (gameView.getGameState().getCurrentPlayerId() == field.getPlayerId())) {
						final FieldCity city = (FieldCity) field.getField();
						final CommandableTemplate template = city.getUnitInProduction();
						if (0 <= template.id) {
							imgKey = gameView.getCityUnitInProductionImageKey(city);
						}
					}
					setMapAt(fieldpos, imgKey);
				}
			}
			if (0x00 != mapViewMode.bitmaskUnitsToShow) {
				try {
					final Set<CommandableKey> commandablesInvolved = gameView.calcCommandablesInvolved();
					visibleCommandableInFocus = gameView.getVisibleCommandableInFocus();
					for (final UnitAlike unit : gameView.getUnits()) {
						if (Bitfield32.isBit(mapViewMode.bitmaskUnitsToShow, unit.getKey().getTemplateId())) {
							final PlayfieldPosition pos = unit.getPosition();
							boolean isToDraw = false;
							if (commandablesInvolved.contains(unit.getKey())) {
				    			// Don't draw a unit that is part of the current animation sequence - will be drawn later
							} else {
								if ((gameView.getAt(pos).isCity()) || (null != unit.getParentKey())) {
									// Draw a unit inside a city or another unit only for blinking on current commandable
									isToDraw = false;
									if ((mapViewMode == MapViewMode.NORMAL) 
											&& (null != visibleCommandableInFocus)
											&& (unit.getKey().equals(visibleCommandableInFocus.getKey()))) {
										setBlinking(unit);
									}
								} else {
									// Draw a unit on a normal field always, but not for blinking on current commandable
									if ((mapViewMode == MapViewMode.NORMAL) 
											&& (null != visibleCommandableInFocus)
											&& (unit.getKey().equals(visibleCommandableInFocus.getKey()))) {
										setBlinking(unit);
										if (visibleCommandableInFocus instanceof Unit) {
											final Unit parentUnit = (Unit) visibleCommandableInFocus;
											isToDraw = (unit.getKey().equals(parentUnit.getParentKey()));
										}
									} else {
										isToDraw = true;
									}
								}
							}
							if (isToDraw) {
								setMapAt(gameView, pos, unit);
							}
						}
					}
					if ((mapViewMode.isShowingAnimation())) {
						if (null != visibleCommandableInFocus) {
							final PlayfieldPosition position = visibleCommandableInFocus.getPosition();
							if (visibleCommandableInFocus instanceof Unit) {
//						    		final Unit unit = (Unit) visibleCommandableInFocus;
//						    		final StepRatingMap stepRatings = unit.getStepRatings();
							} else {
								setBlinking(new TileObjectDelegate(GameView.getCrosshair(), position));
							}
						}
						// Show outstanding units in the old positions and states
						for (final CommandableKey cmdl : commandablesInvolved) {
							final CommandableInfo cmdli = gameView.findAnimationInfo(cmdl);
							if (gameView.isVisible(cmdli)) {
								final Commandable c = cmdli.commandable;
								final boolean isToShow;
								if (null == c) {
									isToShow = false;
								} else if (c instanceof FieldCity) {
									isToShow = true;
								} else {
									if (cmdli.isCargo) {
										isToShow = false;
									} else {
										// Current animation will be drawn later
										final Animation currentAnimation = gameView.getCurrentAnimation();
										if (null != currentAnimation) {
											final MoveResult result = currentAnimation.getMoveResult();
											if (null != result) {
												boolean isCurrent = false;
												for (Role role : Role.values()) {
													final Commandable acmdl = result.getCommandable(role);
													if ((null != acmdl) && (acmdl.getKey().equals(c.getKey()))) {
														isCurrent = true;
														break;
													}
												}
												if (isCurrent) {
													isToShow = false;
												} else {
													isToShow = true;
												}
											} else {
												isToShow = true;
											}
										} else {
											isToShow = true;
										}
									}
								}
								if (isToShow) {
									setMapAt(cmdli.position, cmdli.imageKey);
								}
							}
						}
						final Animation animation = gameView.getCurrentAnimation();
						final MoveResult moveResult = (null == animation ? null : animation.getMoveResult());
						if (null != moveResult) {
							animationMoveResult = animation.getMoveResult();
							animationMoveVector = animation.getMoveVector();
							animationMoveProgress = animation.getMoveProgress();
							animationDyingProgress = animation.getDyingProgress();
							// Show fight animation
							final FightOverlay[] overlays = animation.getFightOverlays();
							if (null != overlays) {
								for (final Role role : Role.values()) {
									final CommandableInfo cmdi = moveResult.getCommandableInfo(role);
									final FightOverlay overlay = overlays[role.ordinal()];
									if (null != overlay) {
										if (gameView.isVisible(cmdi)) {
											fightOverlays[role.ordinal()] = new TileObjectDelegate(overlay, cmdi.position);
										}
									}
								}
							}
						}
					}
				} catch (final ConcurrentModificationException ex) {
					// Ignore; will be redrawn later
				}
			}
			if (mapViewMode.isShowingCrosshair()) {
				final PlayfieldPosition position = gameView.getPositionInFocus();
				setBlinking(new TileObjectDelegate(GameView.getCrosshair(), position));
			}
		}
	}

	private void setMapAt(final PlayfieldPosition fieldpos, final int imgKey) {
		mapTileObjectKeys[fieldpos.getX()][fieldpos.getY()] = imgKey;
	}

	private void setMapAt(final GameView gameView, final PlayfieldPosition fieldpos, final Tileable obj) {
		setMapAt(fieldpos, obj.getImageKey());
	}

	/** 
	 * @return the size of the map to draw
	 */
	public PlayfieldDimension getSize() {
		return size;
	}

	/**
	 * @param pos
	 *            the position on the map
	 * @return the field bitmap at the position
	 */
	public int getMapAt(final PlayfieldPosition pos) {
		int ret;
		if (null != mapTileObjectKeys) {
			try {
				ret = mapTileObjectKeys[pos.getX()][pos.getY()];
			} catch(final ArrayIndexOutOfBoundsException ex) {
				ret = 0;
			}
		} else {
			ret = UNDEFINED;
		}
		return ret;
	}

	public PositionedTileable getBlinking() {
		return blinking;
	}

	public Crosshair getCrosshair() {
		return GameView.getCrosshair();
	}

	private void setBlinking(final PositionedTileable blinking) {
		this.blinking = blinking;
	}

	public Commandable getVisibleCommandableInFocus() {
		return visibleCommandableInFocus;
	}

	/**
	 * @return an array of size 2 of potential fight overlays. Each entry may be
	 *         <code>null</code>, and in fact most of the time is. Index 0
	 *         represents the fight overlay on the active commandable, index 1
	 *         on the affected commandable (if any).
	 */
	public PositionedTileable[] getFightOverlays() {
		return fightOverlays;
	}

	public MoveResult getAnimationMoveResult() {
		return animationMoveResult;
	}

	/**
	 * @param role
	 *            the role
	 * @return If there is an animation that needs to draw the commandable with
	 *         the given role not exactly aligned with the grid (e.g., within a
	 *         move), this method returns a vector to a neighboring field to show the direction of
	 *         non-aligment. If there is no such animation, returns
	 *         <code>null</code>.
	 */
	public PlayfieldDistance getAnimationMoveVector(final Role role) {
		final PlayfieldDistance ret;
		if (Role.Active == role) {
			ret = animationMoveVector;
		} else {
			ret = null;
		}
		return ret;
	}

	public float getAnimationMoveProgress(final Role role) {
		float ret = Animation.NOT_MOVING;
		if (null != getAnimationMoveVector(role)) {
			ret = animationMoveProgress;
		}
		return ret;
	}

	public float getAnimationDyingProgress(final Role role) {
		final float ret;
		final Role dyingRole;
		final MoveResult result = getAnimationMoveResult();
		if (null  == result) {
			dyingRole = null;
		} else {
			final int mask = result.getResultMask();
			if (0 != (mask & MoveResult.MOVE_MASK_LOSING_FIGHT)) {
				dyingRole = Role.Active;
			} else if (0 != (mask & MoveResult.MOVE_MASK_WINNING_FIGHT)) {
				dyingRole = Role.Affected;
			} else {
				dyingRole = null;
			}
		}
		final Commandable cmdl = result.getCommandable(role);
		if ((role == dyingRole) && (null != cmdl) && (ImageKey.CMDL_ID_CITY != cmdl.getKey().getTemplateId())) {
			ret = animationDyingProgress;
		} else {
			ret = Animation.NOT_DYING;
		}
		return ret;
	}

	private PlayfieldDimension size;
	private int[][] mapTileObjectKeys;
	private Commandable visibleCommandableInFocus;
	private PositionedTileable blinking;
	private PositionedTileable[] fightOverlays = new PositionedTileable[2];
	private PlayfieldDistance animationMoveVector;
	private float animationMoveProgress;
	private float animationDyingProgress;
	private MoveResult animationMoveResult;
}