package de.ped.empire.logic;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Contains all the game needs to tell an observer what happened.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameNotification implements Cloneable, Marshallable {
	/** A new status message was created. */
	public static final int MASK_STATUS_MESSAGE       = 0x0001;
	/** A non-moving order has been executed. Should result in status message and order menu changes. */
	public static final int MASK_ORDERS               = 0x0002;
	/** The map mode changed. */
	public static final int MASK_MAP_MODE             = 0x0004;

	/** The viewport to the world may need to be adjusted, e.g., because the commandable in focus moved. */
	public static final int MASK_POSITION_IN_FOCUS    = 0x0010;
	/** There is a new commandable in focus. */
	public static final int MASK_COMMANDABLE_IN_FOCUS = 0x0020;
	/** The next player is shown on the screen. */
	public static final int MASK_PLAYER_IN_FOCUS      = 0x0040;
	/** A city has produced a unit. */
	public static final int MASK_UNIT_PRODUCED        = 0x0080;

	/** A rename order has been executed. */
	public static final int MASK_RENAME_REQ           = 0x0100;
	/** A change production order has been executed. */
	public static final int MASK_CHANGE_PROD_REQ      = 0x0200;
	/** A change move paths order has been executed. */
	public static final int MASK_EDIT_MOVE_PATH_REQ   = 0x0400;

	/** A new game has been started. */
	public static final int MASK_NEW_GAME_STARTED     = 0x1000;
	/** The player has lost the game. */
	public static final int MASK_GAME_LOST            = 0x2000;
	/** The player has won the game. */
	public static final int MASK_GAME_WON             = 0x4000;

	public GameNotification() {
		reset();
	}

	public GameNotification(final int mask, final int playerId) {
		reset();
		addEvent(mask, playerId);
	}

	public GameNotification(final Marshaller marshaller, final GameView gameView) {
		eventMask = marshaller.readInt();
		playerId = marshaller.readInt();
		if (!marshaller.readNullIndicator()) {
			setCommandablePrompting(new CommandableKey(marshaller));
		}
		isAnimationRunning = marshaller.readBoolean();
		if (!marshaller.readNullIndicator()) {
			animationMoveResult = new MoveResult(gameView, marshaller);
		}
		int ap = marshaller.readInt();
		animationPart = (ap == -1 ? null : AnimationPart.values()[ap]);
		animationPartStep = marshaller.readInt();
	}

	@Override
	public void marshal(Marshaller marshaller) {
		marshaller.writeInt(eventMask);
		marshaller.writeInt(playerId);
		marshaller.writeNullOrMarshallable(commandablePrompting);
		marshaller.writeBoolean(isAnimationRunning);
		marshaller.writeNullOrMarshallable(animationMoveResult);
		marshaller.writeInt(animationPart == null ? -1 : animationPart.ordinal());
		marshaller.writeInt(animationPartStep);
	}

	@Override
	public Marshallable demarshal(Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	synchronized void reset() {
		eventMask = 0x00;
		commandablePrompting = null;
		playerId = Player.ID_NONE;
		animationMoveResult = null;
		animationPart = null;
		animationPartStep = 0;
	}

	public synchronized boolean isEvent(final int event) {
		return (0 != (eventMask & event));
	}

	synchronized void addEvent(final int event) {
		eventMask = eventMask | event;
		setPlayerId(playerId);
		addFollowingEvents(event, MASK_NEW_GAME_STARTED, MASK_PLAYER_IN_FOCUS | MASK_COMMANDABLE_IN_FOCUS | MASK_POSITION_IN_FOCUS);
		addFollowingEvents(event, MASK_PLAYER_IN_FOCUS,  MASK_COMMANDABLE_IN_FOCUS | MASK_POSITION_IN_FOCUS);
		addFollowingEvents(event, MASK_COMMANDABLE_IN_FOCUS, MASK_POSITION_IN_FOCUS);
		if (0 != (event & MASK_PLAYER_IN_FOCUS)) {
			logger.trace("addEvent " + Integer.toHexString(event) + ", resulting in " + Integer.toHexString(eventMask));
		}
	}

	synchronized void addEvent(final int event, final int playerId) {
		setPlayerId(playerId);
		addEvent(event);
	}

	public synchronized void clearEvent(final int event) {
		eventMask = eventMask & (~event);
	}

	private void addFollowingEvents(final int event, final int pre, final int followings) {
		if (0 != (event & pre)) {
			eventMask = eventMask | followings;
		}
	}

	public int getPlayerId() {
		return playerId;
	}

	void setPlayerId(int playerId) {
		this.playerId = playerId;
	}

	public CommandableKey getCommandablePrompting() {
		return commandablePrompting;
	}

	synchronized void setCommandablePrompting(final CommandableKey commandableKey) {
		commandablePrompting = commandableKey;
	}

	public MoveResult getAnimationMoveResult() {
		return animationMoveResult;
	}

	public AnimationPart getAnimationPart() {
		return animationPart;
	}

	public int getAnimationPartStep() {
		return animationPartStep;
	}

	public boolean isAnimationRunning() {
		return isAnimationRunning;
	}

	synchronized public void cleanAnimationRunning() {
		setAnimationMoveResult(null);
		isAnimationRunning = false;
	}

	synchronized void setAnimationMoveResult(final MoveResult moveResult) {
		animationMoveResult = moveResult;
		animationPart = null;
		animationPartStep = 0;
		if (null != moveResult) {
			addEvent(GameNotification.MASK_POSITION_IN_FOCUS | GameNotification.MASK_STATUS_MESSAGE);
			isAnimationRunning = true;
		}
	}

	synchronized void setAnimationPart(final MoveResult moveResult, final AnimationPart part, final int step) {
		animationMoveResult = moveResult;
		animationPart = part;
		animationPartStep = step;
	}

	@Override
	public synchronized GameNotification clone() {
		try {
			return (GameNotification) super.clone();
		} catch (final CloneNotSupportedException e) {
			throw new IllegalStateException("Unable to clone");
		}
	}

//	@Override
//	public boolean equals(final Object obj) {
//		if (this == obj)
//			return true;
//		if (obj == null)
//			return false;
//		if (getClass() != obj.getClass())
//			return false;
//		return (0 == compareTo((GameNotification) obj));
//	}
//
//	public int compareTo(final GameNotification other) {
//		if (this.eventMask != other.eventMask) {
//			return (this.eventMask - other.eventMask);
//		}
//		if (this.playerId != other.playerId) {
//			return (this.playerId - other.playerId);
//		}
//		if (null == this.commandablePrompting) {
//			if (null == other.commandablePrompting) {
//				// Both null, nothing to do
//			} else {
//				return -1;
//			}
//		} else {
//			if (null == other.commandablePrompting) {
//				return 1;
//			} else {
//				int ckdiff = this.commandablePrompting.compareTo(other.commandablePrompting);
//				if (0 != ckdiff) {
//					return (ckdiff);
//				}
//			}
//		}
//		if (this.isAnimationRunning != other.isAnimationRunning) {
//			return ((this.isAnimationRunning ? 1 : 0) - (other.isAnimationRunning ? 1 : 0));
//		}
//		// private MoveResult animationMoveResult;
//		if (this.animationPart != other.animationPart) {
//			return (this.animationPart.ordinal() - other.animationPart.ordinal());
//		}
//		if (this.animationPartStep != other.animationPartStep) {
//			return (this.animationPartStep - other.animationPartStep);
//		}
//		return 0;
//	}

	/** 
	 * @return if this notification contains no value - which means no events and no animation
	 */
	public synchronized boolean isEmpty() {
		return (0 == eventMask) && (!isAnimationRunning);
	}

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append(super.toString());
		ret.append("(");
		ret.append("mask=0x");
		ret.append(Integer.toHexString(eventMask));
		ret.append("; ");
		ret.append("player=0x");
		ret.append(Integer.toHexString(playerId));
		ret.append("; ");
		ret.append("cmdl=");
		ret.append(commandablePrompting);
		ret.append("; ");
		ret.append("anim=");
		if (isAnimationRunning) {
			if (null == animationPart) {
				ret.append("null");
			} else {
				ret.append(animationPart.toString());
			}
			ret.append(animationPartStep);
		} else {
			ret.append("None");
		}
		ret.append(" result=");
		if (null == animationMoveResult) {
			ret.append("null");
		} else {
			ret.append(animationMoveResult.toString());
		}
		ret.append(")");
		return ret.toString();
	}

	/** The mask containing all the flags what changed. */
	private int eventMask;
	/** The ID of the player the event is primarily affecting */
	private int playerId;
	/** The unit needs orders from the player. */
	private CommandableKey commandablePrompting;

	private boolean isAnimationRunning;
	private MoveResult animationMoveResult;
	private AnimationPart animationPart;
	private int animationPartStep;

	private Logger logger = Logger.getLogger(getClass());
	{
		logger.setToShowThread(true);
		logger.setLogLevel(Level.DEBUG);
	}
}
