package de.ped.empire.logic;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.List;

import de.ped.tools.BorderMode;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.SliderDef;
import de.ped.tools.TopologyMode;

/**
 * Defines all the properties to create a new game.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameProperties implements Marshallable, Cloneable {

	/** 
	 * Defines the slider properties for playfield size to minimum 10, maximum 200, step size 2. 
	 * (Step size 2 because HEXAGON wrap on x won't fit otherwise; 
	 * max size 150 because HEXAGON topology wastes memory space and max space per dimension is 255)
	 */
	public static final SliderDef SLIDERDEF_SIZE    = new SliderDef(10, (PlayfieldTiling.IS_HEX_COMPACT ? 200 : 150), 2);
	/** Defines the slider properties for playfield setup to minimum 0, maximum 100, step size 1. */
	public static final SliderDef SLIDERDEF_CONTENT = new SliderDef( 0, 100, 1);

	public GameProperties() {
		rulesComplexity = RulesComplexity.BEGINNER;
		size.setLocation(SLIDERDEF_SIZE.check(50), SLIDERDEF_SIZE.check(50));
		setCityDensity(10);
		setLandRatio(30);
		setContinentSize(70);
		topologyMode = TopologyMode.SQUARE8;
		borderMode = BorderMode.ALL;
		players = new Player[Player.MAX_NUMBER_OF_PLAYERS];
		for (int i = 0; i < players.length; i++) {
			players[i] = new Player(i);
		}
	}

	void initForGame(final GameState gameState, final boolean isNewGame) {
		for (int i = 0; i < players.length; i++) {
			players[i].initForGame(gameState, isNewGame);
		}
	}

	/** @return the size of the board */
	public PlayfieldDimension getSize() {
		return size;
	}

	/** 
	 * Returns one dimension of the size of the board.
	 * @param dimension either {@link PlayfieldPosition#X} or {@link PlayfieldPosition#Y}
	 * @return the length of the world in the given dimension, in the range as defined in {@link #SLIDERDEF_SIZE}
	 */
	public int getSize(final int dimension) {
		return size.get(dimension);
	}

	public Player[] getPlayers() {
		return players;
	}

	/** 
	 * Sets one dimension of the size of the board.
	 * @param dimension either {@link PlayfieldPosition#X} or {@link PlayfieldPosition#Y}
	 * @param value the length of the world in the given dimension, in the range as defined in {@link #SLIDERDEF_SIZE}
	 */
	public void setSize(final int dimension, final int value) {
		size.set(dimension, SLIDERDEF_SIZE.check(value));
	}

	/** @return the city density, in the range of 0 (nearly inhabited) to 100 (crowded world). */
	public int getCityDensity() {
		return cityDensity;
	}

	/** @param cityDensity the city density, in the range of 0 (nearly inhabited) to 100 (crowded world). */
	public void setCityDensity(final int cityDensity) {
		this.cityDensity = SLIDERDEF_CONTENT.check(cityDensity);
	}

	/** @return the land to sea ratio, in the range of 0 (no land at all) to 100 (lots of land fields). */
	public int getLandRatio() {
		return landRatio;
	}

	/** @param landRatio the land to sea ratio, in the range of 0 (no land at all) to 100 (lots of land fields). */
	public void setLandRatio(final int landRatio) {
		this.landRatio = SLIDERDEF_CONTENT.check(landRatio);
	}

	/** @return the continent size, in the range of 0 (mostly islands or atolls) to 100 (larger land masses). */
	public int getContinentSize() {
		return continentSize;
	}

	/** @param continentSize the continent size, in the range of 0 (mostly islands or atolls) to 100 (larger land masses). */
	public void setContinentSize(final int continentSize) {
		this.continentSize = SLIDERDEF_CONTENT.check(continentSize);
	}

	public TopologyMode getTopologyMode() {
		return topologyMode;
	}

	public void setTopologyMode(final TopologyMode topologyMode) {
		this.topologyMode = topologyMode;
	}

	public BorderMode getBorderMode() {
		return borderMode;
	}

	public void setBorderMode(final BorderMode borderMode) {
		this.borderMode = borderMode;
	}

	public boolean hasBorder(final int dimension) {
		boolean ret;
		switch(dimension) {
		case PlayfieldPosition.X:
			ret = (BorderMode.ALL == borderMode);
			break;
		case PlayfieldPosition.Y:
			ret = (BorderMode.NONE != borderMode);
			break;
		default:
			throw new IllegalArgumentException("Unknown dimension " + dimension);
		}
		return ret;
	}

	public Rectangle getUsableMapArea(final int borderThickness) {
		int xl = 0;
		int xr = size.getWidth();
		if (hasBorder(PlayfieldPosition.X)) {
			xl += borderThickness;
			xr -= borderThickness;
		}
		int yt = 0;
		int yd = size.getHeight();
		if (hasBorder(PlayfieldPosition.Y)) {
			yt += borderThickness;
			yd -= borderThickness;
		}
		Rectangle ret = new Rectangle(xl, yt, xr - xl, yd - yt);
		return ret;
	}

	public boolean containsUsableMapArea(final PlayfieldPosition pos, final int edgeThickness) {
		final Rectangle area = getUsableMapArea(edgeThickness);
		final int x = pos.getX();
		if (area.contains(x, pos.getY() - getYMin(x))) {
			return true;
		}
		return false;
	}

	public RulesComplexity getRulesComplexity() {
		return rulesComplexity;
	}

	public void setRulesComplexity(final RulesComplexity rulesComplexity) {
		this.rulesComplexity = rulesComplexity;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(rulesComplexity.ordinal());
		size.marshal(marshaller); 
		marshaller.writeInt(topologyMode.ordinal());
		marshaller.writeInt(borderMode.ordinal());
		marshaller.writeInt(cityDensity);
		marshaller.writeInt(landRatio);
		marshaller.writeInt(continentSize);
		marshaller.writeLong(timestamp);
		marshaller.writeBoolean(isQuickstart);
		marshaller.writeString(name);
		// Write players
		marshaller.writeInt(players.length);
		for (int i = 0; i < players.length; i++) {
			marshaller.writeMarshallable(players[i]);
		}
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		rulesComplexity = RulesComplexity.values()[marshaller.readInt()];
		size.demarshal(marshaller);
		topologyMode = TopologyMode.values()[marshaller.readInt()];
		borderMode = BorderMode.values()[marshaller.readInt()];
		cityDensity = marshaller.readInt();
		landRatio = marshaller.readInt();
		continentSize = marshaller.readInt();
		timestamp = marshaller.readLong();
		isQuickstart = marshaller.readBoolean();
		name = marshaller.readString();
		// Read players
		players = new Player[marshaller.readInt()];
		for (int i = 0; i < players.length; i++) {
			players[i] = new Player(marshaller);
		}
		return this;
	}

	@Override
	public GameProperties clone() {
		GameProperties ret = null;
		try {
			ret = (GameProperties) super.clone();
			ret.size = new PlayfieldDimension(this.size); // deep copy
			ret.players = new Player[players.length];
			for (int i = 0; i < players.length; i++) {
				ret.players[i] = (Player) players[i].clone();
			}
		} catch (final CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return ret;
	}

	/**
	 * Norms the result of the position with added vector from direction to be
	 * within the game boundaries.
	 * 
	 * @param position
	 *            the start position
	 * @param dir
	 *            the direction to move from the start position, or
	 *            <code>null</code> for no direction at all
	 * @return a new array containing the wrapped target position
	 */
	public PlayfieldPosition wrap(final PlayfieldPosition position, final DigitalJoystickDirection dir) {
		final PlayfieldPosition ret;
		if (null == dir) {
			ret = new PlayfieldPosition(position);
		} else {
			ret = position.add(renormDistanceVector(dir.getDistanceVector(), position));
			wrapInside(ret);
		}
		return ret;
	}

	/**
	 * Norms the position to be within the game boundaries.
	 * @param positionToWrap the position to wrap; is left unchanged
	 * @return a new array containing the wrapped position
	 */
	public PlayfieldPosition wrap(final PlayfieldPosition positionToWrap) {
		PlayfieldPosition ret = new PlayfieldPosition(positionToWrap);
		wrapInside(ret);
		return ret;
	}

	/**
	 * Norms the position to be within the game boundaries. 
	 * Currently only works if the position to wrap is not further away than one map size.
	 * @param positionToWrap the position to wrap; the wrapping is done in place
	 */
	public void wrapInside(final PlayfieldPosition positionToWrap) {
		switch(getTopologyMode()) {
		case HEXAGON:
			final int x = positionToWrap.getX();
			int y = positionToWrap.getY();
			final int xsize = size.getWidth(); 
			if (x < 0) { // Out to the west
				if (PlayfieldTiling.IS_HEX_COMPACT) {
					// Nothing to do
				} else {
					y += (xsize / 2);
				}
				if (0 == (x | 0x1)) { // Odd rows are a bit lower than even rows 
					y++;
				}
			} else if (xsize <= x) { // Out to the east
				if (PlayfieldTiling.IS_HEX_COMPACT) {
					// Nothing to do
				} else {
					y -= (xsize / 2); 
				}
				if (0 == (x | 0x1)) { // Odd rows are a bit lower than even rows 
					y++;
				}
			}
			// Out to the north or south
			final int nx = MathUtil.modulo(x, size.getWidth());
			positionToWrap.setLocation(nx, MathUtil.modulo(y, size.getHeight(), getYMin(nx)));
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			for (int d = 0; d < PlayfieldPosition.COORDS; d++) {
				positionToWrap.set(d, MathUtil.modulo(positionToWrap.get(d), size.get(d)));
			}
		}
	}

	/**
	 * Checks if the given position is still within the map (including borders).
	 * @param pos The position to check
	 * @return <code>true</code> if the position is within the map
	 */
	public boolean isWithin(final PlayfieldPosition pos) {
		final int x = pos.getX();
		final int y = pos.getY();
		return ((0 <= x) && (x < size.getWidth()) && (getYMin(x) <= y) && (y < getYMax(x)));
	}

	public int getYMin(final int x) {
		int ret = 0;
		switch(getTopologyMode()) {
		case HEXAGON:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				// Nothing to do
			} else {
				ret = x / 2;
			}
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			// Nothing to do
		}
		return ret;
	}

	// TOOO: Delegate this to the game facade, so it won't be public
	public int getYMax(final int x) {
		int ret = size.getHeight();
		switch(getTopologyMode()) {
		case HEXAGON:
			ret = ret + getYMin(x);
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			// Nothing to do
		}
		return ret;
	}

	/**
	 * @return the surrounding rectangle to the map, relative to the upper left corner of tile (0, 0), and in tile size.
	 */
	public Rectangle2D.Float getMapBoundary() {
		return PlayfieldTiling.getMapBoundary(topologyMode, size);
	}

	/**
	 * Returns the map size in pixel coordinates
	 * @param zoomLevel The zoom level to apply to calculate the size
	 * @param isMemorySize if <em>true</em>, the memory size to allocate is returned 
	 * 		(e.g., needed to create template tables), otherwise the map size as presented on the screen is returned
	 * @return The dimension object containing width and height of the map in pixels
	 */
	public Dimension getMapSizeInPixels(final ZoomLevel zoomLevel, final boolean isMemorySize) {
		final Rectangle2D.Float mapBoundary = getMapBoundary();
		if (isMemorySize) {
			PlayfieldDimension memsize = calcMapMemorySize(size);
			mapBoundary.setRect(0.0f, 0.0f, 
					mapBoundary.width + memsize.getWidth() - size.getWidth(),
					mapBoundary.height + memsize.getHeight() - size.getHeight());
		}
		return zoomLevel.getPlayfieldTiling(topologyMode).getMapSizeInPixels(topologyMode, mapBoundary);
	}

	/**
	 * Translates from map position to canvas position.
	 * @param point output parameter takes the resulting canvas position
	 * @param position the map position to translate
	 * @param mapOffset the offset on the canvas of upper left corner of playfield (0, 0)
	 * @param zoomLevel the zoom level shown on the screen
	 * @param isCenter if {@link PlayfieldTiling#TOP_LEFT}, the canvas position of the upper left corner of the map tile is calculated; 
	 * 		if {@link PlayfieldTiling#CENTER}, the center of the map tile is calculated
	 */
	public void map2canvas(final Point point, final PlayfieldPosition position, final Point mapOffset, final ZoomLevel zoomLevel, final boolean isCenter) {
		zoomLevel.getPlayfieldTiling(topologyMode).map2canvas(point, position, mapOffset, isCenter);
	}

	/**
	 * Calculates the size of the memory needed to hold a game map of the given size.
	 * @param mapSize the map size to calculate the memory size for
	 * @return the needed memory array size
	 */
	public PlayfieldDimension calcMapMemorySize(final PlayfieldDimension mapSize) {
		final PlayfieldDimension mapMemorySize = new PlayfieldDimension();
		switch(getTopologyMode()) {
		case HEXAGON:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				mapMemorySize.setLocation(mapSize);
			} else {
				final int x = mapSize.getWidth();
				mapMemorySize.setLocation(x, mapSize.getHeight() + (x / 2 - 1));
			}
			break;
		case HEXAGON_H:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				mapMemorySize.setLocation(mapSize);
			} else {
				final int y = mapSize.getHeight();
				mapMemorySize.setLocation(y, mapSize.getWidth() + (y / 2 - 1));
			}
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			mapMemorySize.setLocation(mapSize);
			break;
		}
		return mapMemorySize;
	}

	/**
	 * Translates from canvas (e.g., mouse) position to on what map position this canvas position lies.
	 * @param cx the canvas x position
	 * @param cy the canvas y position
	 * @param mapOffset the offset on the canvas of upper left corner of playfield (0, 0)
	 * @param zoomLevel the zoom level shown on the screen
	 * @return the resulting map position. Might be out of the map boundaries
	 */
	public PlayfieldPosition canvas2map(final int cx, final int cy, final Point mapOffset, final ZoomLevel zoomLevel) {
		return zoomLevel.getPlayfieldTiling(topologyMode).canvas2map(cx, cy, mapOffset);
	}

	private static class VectorsToAirDistanceKey {
		public VectorsToAirDistanceKey(
				final TopologyMode mode,
				final int middlePositionType,
				final int distance,
				final boolean isWithFilling) {
			this.mode = mode;
			this.middlePositionType = middlePositionType;
			this.distance = distance;
			this.isWithFilling = isWithFilling;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + distance;
			result = prime * result + (isWithFilling ? 1231 : 1237);
			result = prime * result + middlePositionType;
			result = prime * result + ((mode == null) ? 0 : mode.hashCode());
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			VectorsToAirDistanceKey other = (VectorsToAirDistanceKey) obj;
			if (distance != other.distance) {
				return false;
			}
			if (isWithFilling != other.isWithFilling) {
				return false;
			}
			if (middlePositionType != other.middlePositionType) {
				return false;
			}
			if (mode != other.mode) {
				return false;
			}
			return true;
		}

		public final TopologyMode mode;
		public final int middlePositionType;
		public final int distance; 
		public final boolean isWithFilling;
	}

	private static HashMap<VectorsToAirDistanceKey, ArrayList<PlayfieldDistance>> vectorsToAirDistanceHash = new HashMap<VectorsToAirDistanceKey, ArrayList<PlayfieldDistance>>();

	/**
	 * Return all distance vectors to all the fields in the given air distance
	 * (or less, if <em>isWithFilling</em> is <code>true</code>).
	 * 
	 * @param topologyMode
	 *            the topology the map should use.
	 * @param middlePosition
	 *            the position to start from. Necessary for complex topologies
	 *            where the neighborhood depends on position.
	 * @param distance
	 *            the distance to collect the other fields in
	 * @param isWithFilling
	 *            if <code>true</code>, the result includes also all the middle fields with lesser distance, including the null-vector.
	 *            otherwise the result contains only fields with the exact air distance.
	 * @return the list of distance vectors with the given distance (or less). Treat them as read-only.
	 */
	public List<PlayfieldDistance> getAllVectorsToAirDistance(final TopologyMode topologyMode, final PlayfieldPosition middlePosition, final int distance, final boolean isWithFilling) {
		if (distance < 0) {
			throw new IllegalArgumentException("Negative distance not allowed!");
		}
		final int middlePositionType = getPositionType(topologyMode, middlePosition);
		final VectorsToAirDistanceKey key = new VectorsToAirDistanceKey(topologyMode, middlePositionType, distance, isWithFilling);
		ArrayList<PlayfieldDistance> ret = vectorsToAirDistanceHash.get(key);
		if (null == ret) {
			ret = new ArrayList<PlayfieldDistance>();
			if (0 == distance) {
				ret.add(new PlayfieldDistance(0, 0));
			} else if (0 < distance) {
				switch(topologyMode) {
				case SQUARE8:
					if (isWithFilling) {
						for (int x = -distance; x <= distance; x++) {
							for (int y = -distance; y <= distance; y++) {
								ret.add(new PlayfieldDistance(x, y));
							}
						}
					} else {
						for (int i = -distance; i < distance; i++) {
							ret.add(new PlayfieldDistance( i, -distance)); // northern fields (-,-) (0,-)
							ret.add(new PlayfieldDistance(-distance, -i)); // western  fields (-,+) (-,0)
							ret.add(new PlayfieldDistance( distance,  i)); // eastern  fields (+,-) (+,0)
							ret.add(new PlayfieldDistance(-i,  distance)); // southern fields (+,+) (0,+)
						}
					}
					break;
				case SQUARE4:
					if (isWithFilling) {
						for (int x = -distance; x <= distance; x++) {
							final int ax = Math.abs(x);
							for (int y = -distance + ax; y <= distance - ax; y++) {
								ret.add(new PlayfieldDistance(x, y));
							}
						}
					} else {
						for (int i = 0; i < distance; i++) {
							ret.add(new PlayfieldDistance(-i + distance, -i)); // north-western fields
							ret.add(new PlayfieldDistance(-i,  i - distance)); // north-eastern fields
							ret.add(new PlayfieldDistance( i, -i + distance)); // south-western fields
							ret.add(new PlayfieldDistance( i - distance,  i)); // south-eastern fields
						}
					}
					break;
				case HEXAGON:
					if (isWithFilling) {
						for (int y = -distance; y <= distance; y++) {
							final int ay = Math.abs(y); 
							if (y <= 0) { // north-western half
								for (int x = -distance; x <= distance - ay; x++) {
									ret.add(renormHex(new PlayfieldDistance(x, y), middlePositionType));
								}
							} else { // south-eastern half
								for (int x = -distance + ay; x <= distance; x++) {
									ret.add(renormHex(new PlayfieldDistance(x, y), middlePositionType));
								}
							}
						}
					} else {
						for (int i = 0; i < distance; i++) {
							ret.add(renormHex(new PlayfieldDistance( i - distance, -distance), middlePositionType)); // north-western fields
							ret.add(renormHex(new PlayfieldDistance( i,  i - distance), middlePositionType)); // north-eastern fields
							ret.add(renormHex(new PlayfieldDistance(-i, -i + distance), middlePositionType)); // south-western fields
							ret.add(renormHex(new PlayfieldDistance(-i + distance,  distance), middlePositionType)); // south-eastern fields
							ret.add(renormHex(new PlayfieldDistance(    -distance,  -i), middlePositionType)); // western fields
							ret.add(renormHex(new PlayfieldDistance(     distance,  i), middlePositionType)); // eastern fields
						}
					}
					break;
				case HEXAGON_H:
					// TODO: Adapt to HEXAGON_H
					if (isWithFilling) {
						for (int y = -distance; y <= distance; y++) {
							final int ay = Math.abs(y); 
							if (y <= 0) { // north-western half
								for (int x = -distance; x <= distance - ay; x++) {
									ret.add(renormHex(new PlayfieldDistance(x, y), middlePositionType));
								}
							} else { // south-eastern half
								for (int x = -distance + ay; x <= distance; x++) {
									ret.add(renormHex(new PlayfieldDistance(x, y), middlePositionType));
								}
							}
						}
					} else {
						for (int i = 0; i < distance; i++) {
							ret.add(renormHex(new PlayfieldDistance( i - distance, -distance), middlePositionType)); // north-western fields
							ret.add(renormHex(new PlayfieldDistance( i,  i - distance), middlePositionType)); // north-eastern fields
							ret.add(renormHex(new PlayfieldDistance(-i, -i + distance), middlePositionType)); // south-western fields
							ret.add(renormHex(new PlayfieldDistance(-i + distance,  distance), middlePositionType)); // south-eastern fields
							ret.add(renormHex(new PlayfieldDistance(    -distance,  -i), middlePositionType)); // western fields
							ret.add(renormHex(new PlayfieldDistance(     distance,  i), middlePositionType)); // eastern fields
						}
					}
					break;
				default:
					throw new IllegalStateException("Not implemented topology " + topologyMode);
				}
			}
			Collections.sort(ret, POSIITON_COMPARATOR);
			vectorsToAirDistanceHash.put(key, ret);
		}
		return ret;
	}

	public List<PlayfieldDistance> getAllVectorsToAirDistance(final PlayfieldPosition middlePosition, final int distance, final boolean isWithFilling) {
		return getAllVectorsToAirDistance(topologyMode, middlePosition, distance, isWithFilling);
	}

	/**
	 * A comparator that allows to sort two positions so that the top left position is first and the bottom right position is last.
	 * Can be used in views where it's important to draw fields first that are more "behind".
	 */
	public static class PositionComparator implements Comparator<PlayfieldPosition> {
		@Override
		public int compare(PlayfieldPosition that, PlayfieldPosition other) {
			if (that == other) {
				return 0;
			}
			if (null == that) {
				return 1;
			}
			if (null == other) {
				return -1;
			}
			// y counting upwards
			int valDiff = that.getY() - other.getY();
			if (0 != valDiff) {
				return valDiff;
			}
			// x counting downwards (necessary for hexagons)
			valDiff = -(that.getX() - other.getX());
			if (0 != valDiff) {
				return valDiff;
			}
			return 0;
		}
	}

	private static final PositionComparator POSIITON_COMPARATOR = new PositionComparator();

	public TopologyMode getTopologyModeForShores() {
		final TopologyMode ret;
		switch(topologyMode) {
		case SQUARE8:
		case SQUARE4:
			ret = TopologyMode.SQUARE4;
			break;
		case HEXAGON:
		case HEXAGON_H:
			ret = TopologyMode.HEXAGON;
			break;
		default:
			ret = null;
		}
		return ret;
	}

	/** 
	 * Calculates the air distance between two positions.
	 * Air distance means that obstacles are not taken into account.
	 * Wrapping on non-border map boundaries is taken into account.
	 * The result is always non-negative and depends on the chosen topology mode.
	 * @param pos1 one of the positions
	 * @param pos2 one of the positions
	 * @return the air distance between the positions
	 */
	public int calcAirDistance(final PlayfieldPosition pos1, final PlayfieldPosition pos2) {
		final PlayfieldDistance distanceVector = calcAirDistanceVector(pos1, pos2);
		final int ret = calcAirDistanceScalar(pos1, distanceVector);
		return ret;
	}

	/** 
	 * Calculates the air distance vector between two positions.
	 * Air distance means that obstacles are not taken into account. 
	 * Wrapping on non-border map boundaries is taken into account.
	 * @param pos1 starting position of the vector
	 * @param pos2 ending position of the vector
	 * @return the resulting air distance vector between the positions
	 */
	public PlayfieldDistance calcAirDistanceVector(final PlayfieldPosition pos1, final PlayfieldPosition pos2) {
		final PlayfieldDistance result = new PlayfieldDistance();;
		switch(topologyMode) {
		case SQUARE8:
		case SQUARE4:
			result.setLocation(calcAirDistanceCoordinate(PlayfieldPosition.X, pos1, pos2),
					calcAirDistanceCoordinate(PlayfieldPosition.Y, pos1, pos2));
			break;
		case HEXAGON:
		case HEXAGON_H:
			final int xb = (hasBorder(PlayfieldPosition.X) ? 0 : 1);
			final int yb = (hasBorder(PlayfieldPosition.Y) ? 0 : 1);
			final PlayfieldDistance naiveVector = PlayfieldPosition.distance(pos1, pos2);
			final PlayfieldDistance bestVector = new PlayfieldDistance();
			int bestAbs = 100000;
			final PlayfieldDistance currentVector = new PlayfieldDistance();
			for (int xf = -xb; xf <= xb; xf++) {
				for (int yf = -yb; yf <= yb; yf++) {
					int yoffset = 0;
					int xoffset = 0;
					if (!PlayfieldTiling.IS_HEX_COMPACT) {
						switch(topologyMode) {
						case HEXAGON:
							yoffset = xf * size.getWidth() / 2;
							break;
						case HEXAGON_H:
							xoffset = yf * size.getHeight() / 2;
							break;
						default:
							// Nothing to do
						}
					}
					currentVector.setLocation(
							naiveVector.get(PlayfieldPosition.X) + xf * size.getWidth()  + xoffset,
							naiveVector.get(PlayfieldPosition.Y) + yf * size.getHeight() + yoffset);
					final int currentAbs = calcAirDistanceScalar(pos1, currentVector);
					if (currentAbs < bestAbs) {
						bestVector.setLocation(currentVector);
						bestAbs = currentAbs;
					}
				}
			}
			result.setLocation(bestVector);
			break;
		default:
			throw new IllegalStateException("Not implemented topology " + topologyMode);
		}
		return result;
	}

	/** 
	 * Calculates the air distance absolute value based on a distance vector.
	 * Air distance means that obstacles are not taken into account.
	 * The result is always non-negative and depends on the chosen topology mode.
	 * @param startPos starting position of the vector
	 * @param distanceVector the distance vector
	 * @return the air distance between the positions
	 */
	public int calcAirDistanceScalar(final PlayfieldPosition startPos, final PlayfieldPosition distanceVector) {
		return PlayfieldTiling.calcAirDistanceScalar(topologyMode, startPos, distanceVector);
	}

	/**
	 * Calculates from the destination and the direction taken to the destination from which source field this move step came from.
	 * @param source will be filled with the source field coordinates
	 * @param dir the direction taken in the move step to reach destination
	 * @param destination the destination field of the move step
	 */
	public void calcSourcePositionOfMoveStep(final PlayfieldPosition source, DigitalJoystickDirection dir, final PlayfieldPosition destination) {
		source.addInplace(destination, renormDistanceVector(dir.getOppositeDirection().getDistanceVector(), destination));
	}

	/**
	 * De-normalizes a distance vector from a non-linear to a linear topology.
	 * @param distanceVector the distance vector
	 * @param position the start position, as a non-linear topology might be position-sensitive to the distance vector
	 * @return the modified distance vector
	 */
	public PlayfieldPosition denormDistanceVectorInplace(final PlayfieldPosition distanceVector, final PlayfieldPosition position) {
		distanceVector.setY(renormHexY(distanceVector.getX(), distanceVector.getY(), getPositionType(getTopologyMode(), position), 1));
		return distanceVector;
	}

	/**
	 * Normalizes a distance vector from a linear to a non-linear topology.
	 * @param distanceVector the distance vector
	 * @param position the start position, as a non-linear topology might be position-sensitive to the distance vector
	 * @return the modified distance vector
	 */
	public PlayfieldDistance renormDistanceVector(final PlayfieldDistance distanceVector, final PlayfieldPosition position) {
		final PlayfieldDistance ret = new PlayfieldDistance(distanceVector); 
		return renormHex(ret, getPositionType(getTopologyMode(), position));
	}

	private PlayfieldDistance renormHex(final PlayfieldDistance distanceVector, final int positionType) {
		distanceVector.setY(renormHexY(distanceVector.getX(), distanceVector.getY(), positionType, -1));
		return distanceVector;
	}

	private int renormHexY(final int xdist, final int ydist, final int positionType, final int sgn) {
		return PlayfieldTiling.renormHexY(xdist, ydist, positionType, sgn);
	}

	/**
	 * Calculates the air distance in one dimension only between the
	 * <em>source</em> and the <em>dest</em> position. 
	 * Air distance means that obstacles are not taken into account. In torus and
	 * globe worlds, the world boundaries are taken into account correctly. The
	 * result may be negative to indicate that the shortest way leads north or
	 * west.
	 * 
	 * @param dimension
	 *            {@link PlayfieldPosition#X} for horizontal position, {@link PlayfieldPosition#Y}
	 *            for vertical position
	 * @param source
	 *            an array of size two representing the source position
	 * @param dest
	 *            an array of size two representing the destination position
	 * @return the air distance for the given dimension
	 */
	private int calcAirDistanceCoordinate(final int dimension, final PlayfieldPosition source, final PlayfieldPosition dest) {
		int ret = dest.get(dimension) - source.get(dimension);
		if (!hasBorder(dimension)) {
			final int size = getSize(dimension);
			ret = MathUtil.modulo(ret, size, -(size / 2));
		}
		return ret;
	}

	private int getPositionType(final TopologyMode topologyMode, final PlayfieldPosition position) {
		return PlayfieldTiling.getPositionType(topologyMode, position);
	}

	public boolean isWithHumanPlayer() {
		return (Player.ID_NONE != getLastHumanPlayerIndex());
	}

	public int getLastHumanPlayerIndex() {
		int ret = Player.ID_NONE;
		for (final Player p : players) {
			if (p.isHumanPlayer()) {
				ret = p.getId();
			}
		}
		return ret;
	}

	public int getLocaleIndex() {
		return localeIndex;
	}

	public void setLocaleIndex(final int localeIndex) {
		this.localeIndex = localeIndex;
	}

	/**
	 * Gets the (optional) name for the game, used for naming manually saved games.
	 * @return the name of the game
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets the (optional) name for the game, used for naming manually saved games.
	 * @param name the name of the game
	 */
	public void setName(final String name) {
		this.name = name;
	}

	/**
	 * Gets the timestamp the game was created, used for naming auto-save games.
	 * @return the timestamp the game was created
	 */
	public long getTimestamp() {
		return timestamp;
	}

	/** Clears the timestamp the game was created. */
	public void setTimestampEmpty() {
		this.timestamp = 0L;
	}

	/** Sets the timestamp the game was created. */
	public void setTimestampNow() {
		this.timestamp = (new Date()).getTime();
	}

	/**
	 * @return if all players should start with a set of units to speed up the starting phase
	 */
	public boolean isQuickstart() {
		return isQuickstart;
	}

	/**
	 * @param isQuickstart if all players should start with a set of units to speed up the starting phase
	 */
	public void setQuickstart(boolean isQuickstart) {
		this.isQuickstart = isQuickstart;
	}

	/**
	 * @param pos the starting position to look for neighbors. Some topologies might be sensitive to this.
	 * @return The directions allowed to move in the current topology. The returned list should be treated as read-only, so copy if you want to modify.
	 */
	public ArrayList<DigitalJoystickDirection> getNeighborDirections(final PlayfieldPosition pos) {
		return getNeighborDirections(pos, getTopologyMode());
	}

	/**
	 * @param pos the starting position to look for neighbors. Some topologies might be sensitive to this.
	 * @return The directions used for graphical sugar like shores in the current topology. The returned list should be treated as read-only, so copy if you want to modify.
	 */
	public ArrayList<DigitalJoystickDirection> getShoreNeighborDirections(final PlayfieldPosition pos) {
		return getNeighborDirections(pos, getTopologyModeForShores());
	}

	/**
	 * @param pos the starting position to look for neighbors. Some topologies might be sensitive to this.
	 * @param topologyMode the topology mode to return the neighbor directions for
	 * @return The directions allowed in the current topology. The returned list should be treated as read-only, so copy if you want to modify.
	 *         The directions are in counter-clockwise order.
	 */
	private ArrayList<DigitalJoystickDirection> getNeighborDirections(final PlayfieldPosition pos, final TopologyMode topologyMode) {
		switch(topologyMode) {
		case SQUARE8:
			return DigitalJoystickDirection.getDirections8();
		case SQUARE4:
			return DigitalJoystickDirection.getDirections4();
		case HEXAGON:
		case HEXAGON_H:
			return getHexagonNeighborDirections();
		default:
			throw new IllegalStateException("Not implemented topology " + topologyMode);
		}
	}

	/**
	 * Translates (and filters) a user game action input direction into the proper direction for the given topology.
	 * @param topologyMode the topology mode to translate and filter for
	 * @param pos the starting position to look for neighbors. Some topologies might be sensitive to this.
	 * @param kdir the user game action input direction, allowed to be <code>null</code>
	 * @return the appropriate direction fitting to the coding of the topology, or <code>null</code>
	 */
	public DigitalJoystickDirection translateUserInputDirection(final TopologyMode topologyMode, final PlayfieldPosition pos, final DigitalJoystickDirection kdir) {
		DigitalJoystickDirection ret = null;
		switch(topologyMode) {
		case HEXAGON:
			ret = getHexagonUserInputTranslation().get(kdir);
			break;
		case SQUARE8:
			ret = kdir;
			break;
		case SQUARE4:
			if ((null != kdir) && kdir.is4Direction()) {
				ret = kdir;
			}
			break;
		default:
			// Nothing to do
		}
		return ret;
	}

	private HashMap<DigitalJoystickDirection, DigitalJoystickDirection> getHexagonUserInputTranslation() {
		if (null == HEXAGON_INPUT_TRANSLATION) {
			final HashMap<DigitalJoystickDirection, DigitalJoystickDirection> huit = new HashMap<DigitalJoystickDirection, DigitalJoystickDirection>(6);
			huit.put(DigitalJoystickDirection.UP, DigitalJoystickDirection.UP);
			huit.put(DigitalJoystickDirection.DOWN, DigitalJoystickDirection.DOWN);
			huit.put(DigitalJoystickDirection.UP_LEFT, DigitalJoystickDirection.UP_LEFT);
			huit.put(DigitalJoystickDirection.DOWN_RIGHT, DigitalJoystickDirection.DOWN_RIGHT);
			huit.put(DigitalJoystickDirection.UP_RIGHT, DigitalJoystickDirection.RIGHT);
			huit.put(DigitalJoystickDirection.DOWN_LEFT, DigitalJoystickDirection.LEFT);
			HEXAGON_INPUT_TRANSLATION = huit;
		}
		return HEXAGON_INPUT_TRANSLATION;
	}
	private HashMap<DigitalJoystickDirection, DigitalJoystickDirection> HEXAGON_INPUT_TRANSLATION;

	/** @return the 6 directions in counter-clockwise order. */
	private ArrayList<DigitalJoystickDirection> getHexagonNeighborDirections() {
		if (null == HEXAGON_NEIGHBOR_DIRECTIONS) {
			final ArrayList<DigitalJoystickDirection> hnd = new ArrayList<DigitalJoystickDirection>(6);
			hnd.add(DigitalJoystickDirection.UP);
			hnd.add(DigitalJoystickDirection.UP_LEFT);
			hnd.add(DigitalJoystickDirection.LEFT);
//			hnd.add(DigitalJoystickDirection.DOWN_LEFT);
			hnd.add(DigitalJoystickDirection.DOWN);
			hnd.add(DigitalJoystickDirection.DOWN_RIGHT);
			hnd.add(DigitalJoystickDirection.RIGHT);
//			hnd.add(DigitalJoystickDirection.UP_RIGHT);
			HEXAGON_NEIGHBOR_DIRECTIONS = hnd;
		}
		return HEXAGON_NEIGHBOR_DIRECTIONS;
	}
	private ArrayList<DigitalJoystickDirection> HEXAGON_NEIGHBOR_DIRECTIONS;

	private PlayfieldDimension size = new PlayfieldDimension();
	private RulesComplexity rulesComplexity;
	private BorderMode borderMode;
	private TopologyMode topologyMode;
	private int cityDensity;
	private int landRatio;
	private int continentSize;
	private Player[] players;
	private int localeIndex;
	private String name = "";
	private long timestamp;
	private boolean isQuickstart = true;
}
