package de.ped.empire.logic;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import de.ped.empire.logic.Commandable.AttackMode;
import de.ped.empire.logic.MoveResult.Role;
import de.ped.tools.CollectionUtil;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * The business logic of the game.
 * Knows the current state of the map, the units, and the rules.
 * Only public methods should be accessed by the user interface or strategies directly,
 * otherwise they might get information that they should not know or see right now.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameState implements Marshallable {
	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	static void marshalCommandableKey(final Marshaller marshaller, final Commandable commandable, final boolean isWithNull) {
		final boolean isNull = (null == commandable);
		if (isWithNull) {
			marshaller.writeBoolean(isNull);
		} else {
			if (isNull) {
				throw new IllegalStateException("Commandable is null but shouldn't!");
			}
		}
		if (!isNull) {
			marshaller.writeMarshallable(commandable.getKey());
		}
	}

	Commandable demarshalCommandableKey(final Marshaller marshaller, boolean isWithNull) {
		Commandable ret = null;
		final boolean isNull = (isWithNull ? marshaller.readBoolean() : false);
		if (!isNull) {
			final CommandableKey key = (CommandableKey) marshaller.readMarshallable(CommandableKey.factory);
			ret = findCommandable(key);
		}
		return ret;
	}

	public GameState(final GameObserver observer, final GameMap map) throws GameInitializationException {
		this.props = map.getProperties();
		this.map = map;
		paths = new MovePaths();
		map.setGameState(this);
		props.initForGame(this, true);
		if (map.getAllCities().size() == map.findCities(Player.ID_NONE).size()) {
			map.initFor(this);
		}
		isInMoveOver = false;
		// Views
		final Player[] players = getProperties().getPlayers();
		gameViews = new GameView[players.length];
		for (int i = 0; i < gameViews.length; i++) {
			final Player player = players[i];
			gameViews[i] = new GameView(this, player, createObserver(player, observer));
		}
	}

	public GameState(final GameObserver observer, final Marshaller marshaller) {
		props = new GameProperties();
		marshaller.readMarshallable(props);
		marshaller.readMarshallable(gameStateBasics);
		isInMoveOver = marshaller.readBoolean();
		// Basic game map
		map = new GameMap(marshaller, props, null);
		// All units (including dead ones)
		final int allUnitsSize = marshaller.readInt();
		for (int i = 0; i < allUnitsSize; i++) {
			allUnits.add(Unit.create(this, marshaller));
		}
		// Things on the map, like living units
		paths = new MovePaths();
		marshaller.readMarshallable(paths);
		map.setGameState(this);
		props.initForGame(this, false);
		map.demarshalLivingUnits(marshaller, this);
		// Views
		final Player[] players = getProperties().getPlayers();
		gameViews = new GameView[players.length];
		for (int i = 0; i < gameViews.length; i++) {
			final Player player = players[i];
			gameViews[i] = new GameView(this, player, createObserver(player, observer), marshaller);
		}
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeMarshallable(props);
		marshaller.writeMarshallable(gameStateBasics);
		marshaller.writeBoolean(isInMoveOver);
		// Basic game map
		marshaller.writeMarshallable(map);
		// All units (including dead ones)
		final int allUnitsSize = allUnits.size();
		marshaller.writeInt(allUnitsSize);
		for (final Unit unit : allUnits) {
			marshaller.writeMarshallable(unit);
		}
		// Things on the map, like living units
		marshaller.writeMarshallable(paths);
		map.marshalLivingUnits(marshaller);
		// Views
		for (int i = 0; i < gameViews.length; i++) {
			marshaller.writeMarshallable(gameViews[i]);
		}
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	private GameObserver createObserver(final Player player, final GameObserver observer) {
		GameObserver playerObserver = observer;
		if (player.isComputerPlayer()) {
			playerObserver = new ComputerStrategySimple(this, observer.getMessages());
			logger.debug("Player " + player.getId() + " is a computer player.");
		}
		return playerObserver;
	}

	public void initForQuickStart() {
		for (final Player player : getProperties().getPlayers()) {
			if (player.isPlayer()) {
				final List<FieldCity> cities = getMap().findCities(player.getId());
				if (!CollectionUtil.isNullOrEmpty(cities)) {
					gameStateBasics.setPlayerAndTurn(player.getId(), 0);
					logger.trace("initForQuickStart() for player " + player.getId());
					final FieldCity city = cities.get(0);
					switch(getProperties().getRulesComplexity()) {
					case ADVANCED:
						addUnit(city, new Unit(this, Templates.instance().BOMBER, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().TANK, city.getPlayerId(), city));
						// No break
					case STANDARD:
						addUnit(city, new Unit(this, Templates.instance().FIGHTER, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().SUBMARINE, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().CARRIER, city.getPlayerId(), city));
						// No break
					case BEGINNER:
						addUnit(city, new Unit(this, Templates.instance().TRANSPORTER, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().DESTROYER, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
						addUnit(city, new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city));
					}
					// Add random move paths
//					final MovePaths paths = getMap().getMovePaths();
//					for (MilitaryBranch branch : MilitaryBranch.values()) {
//						if (0x00 != branch.mask) {
//							final int[] endPos = Position.create();
//							endPos[0] = MathUtil.bound(city.getPosition()[0] + MathUtil.random(21) - 10, 1, getMap().getSize().getWidth() - 2);
//							endPos[1] = MathUtil.bound(city.getPosition()[0] + MathUtil.random(21) - 10, 1, getMap().getSize().getHeight() - 2);
//							final MovePath path = new MovePath(player.getId(), city.getPosition(), branch.mask, endPos);
//							if (getGameView(player.getId()).makesSense(path, branch)) {
//								try {
//									paths.add(path);
//					        	} catch (final CircularReferenceException ex) {
//					        		// Nothing to do
//					        	}
//							}
//						}
//					}
				}
			}
		}
	}

	/**
	 * Returns the preferred enemy to attack, if there are several; or <code>null</code>if there is no enemy to attack.
	 * @param attacker the attacking unit
	 * @param pos the position to attack
	 * @param commandables the list of commandables to search (provided for performance reasons), 
	 * 		or <code>null</code> if it should be calculated by this method.
	 * @return the preferred enemy from the list
	 */
	Commandable findPreferredEnemy(final Unit attacker, final PlayfieldPosition pos, final Collection<Commandable> commandables) {
		final Collection<Commandable> commandablesToUse;
		if (null == commandables) {
			commandablesToUse = findCommandablesAtPosition(pos, true);
		} else {
			commandablesToUse = commandables;
		}
		Commandable preferred = null;
		if (null != commandablesToUse) {
			for (final Commandable commandable : commandablesToUse) {
				if (commandable.getPlayerId() != getCurrentPlayerId()) {
					if (AttackMode.IMPOSSIBLE != attacker.getAttackModeAgainst(commandable)) {
						preferred = attacker.getUnitTemplate().getPreferredAttackTarget(preferred, commandable);
					}
				}
			}
		}
		return preferred;
	}

	boolean canHold(final Unit unit, final FieldAlike field, final Collection<Commandable> commandablesAtField, final boolean isNow) {
		boolean ret = false;
		// Can the field itself take it?
		int capacity = field.getCapacity(unit);
		if (isNow && (0 < capacity) && (capacity < Integer.MAX_VALUE)) {
			int unitsOnField = 0;
			if (null != commandablesAtField) {
				for (final Commandable c : commandablesAtField) {
					if (c instanceof Unit) {
						unitsOnField++;
					}
				}
			}
			// Check if the field, if stepped onto, would result in a legal move-over situation.
			// Legal means: Unit has capability, it's not the last step of turn, and at least one neighboring field is reachable.
			boolean isMoveOver = false;
			if ((0 < unitsOnField) && (unit.getUnitTemplate().canMoveOver) && (unit.getStepsLeftThisTurn() > 1)) {
				for (FieldAlike neighbor : getMap().getNeighborsOf(field)) {
					if (0 < field.getCapacity(unit)) {
						final PlayfieldPosition neighborPos = neighbor.getPosition();
						if (!hasUnitsAtPosition(neighborPos)) {
							isMoveOver = true;
							break;
						}
					}
				}
			}
			if (!isMoveOver) {
				capacity -= unitsOnField;
			}
		}
		if (0 < capacity) {
			ret = true;
		} else {
			// The field can not. Is there a unit to take it?
			for (final Commandable commandable : commandablesAtField) {
				if ((commandable.getPlayerId() == gameStateBasics.getPlayerId()) && (commandable != field)) {
					if (commandable.canLoad(unit, isNow)) {
						ret = true;
						break;
					}
				}
			}
		}
		return ret;
	}

	public void addUnit(final FieldCity city, final Unit unit) {
		allUnits.add(unit);
		map.addUnit(unit);
		unit.startTurn();
		final int playerId;
		if (null != city) {
			playerId = city.getPlayerId();
		} else { // Airbase creation
			playerId = unit.getPlayerId();
		}
		getGameView(playerId).onUnitAdded(city, unit);
	}

	/**
	 * Checks for the commandable if there is another unit nearby that can load this commandable
	 * @param potentialCargo the commandable that might get loaded
	 */
	void setLoadingCommandIfLoadingUnitNearby(final Commandable potentialCargo) {
		final Unit loadingUnit = findUnitToLoadInto(potentialCargo);
		if (null != loadingUnit) {
			try {
				potentialCargo.setOrder(new OrderGoTo(this, potentialCargo, loadingUnit.getKey()));
			} catch (final CircularReferenceException ex) {
				logger.error("Loading lead to circular exception?!?", ex);
			}
		}
	}

	List<Unit> findChildren(final Unit unit) {
		final List<Unit> ret = new LinkedList<Unit>();
		for (final UnitAlike u : map.getUnits()) {
			if (unit.getKey().equals(u.getUnit().getParentKey())) {
				ret.add(u.getUnit());
			}
		}
		return ret;
	}

	/**
	 * Checks for the commandable if there is another unit nearby that can load this commandable
	 * @param potentialCargo the commandable that might get loaded
	 * @return the unit that can load this commandable, or <code>null</code> if there is none
	 */
	Unit findUnitToLoadInto(final Commandable potentialCargo) {
		Unit ret = null;
		if (potentialCargo instanceof Unit) {
			final RulesComplexity rulesComplexity = getProperties().getRulesComplexity();
			final Unit unitToLoad = (Unit) potentialCargo;
			final Order unitToLoadOrder = unitToLoad.getOrder();
			if (((null == unitToLoadOrder) || (unitToLoadOrder.template.canBeOverriddenToBeLoaded()))) {
				for (final UnitAlike loadingUnit : getUnits()) {
					final Order loadingUnitOrder = loadingUnit.getOrder();
					if ((loadingUnit.getPlayerId() == unitToLoad.getPlayerId())
							&& loadingUnit.getUnitTemplate().canHold(rulesComplexity, unitToLoad.getUnitTemplate())
							&& (null != loadingUnitOrder) && (OrderTemplate.LOAD == loadingUnitOrder.template)) {
						final PlayfieldPosition upos = loadingUnit.getPosition();
						final int distance = props.calcAirDistance(unitToLoad.getPosition(), upos);
						if (1 == distance) {
							if (loadingUnit.canLoad(unitToLoad, true)) {
								ret = loadingUnit.getUnit();
							}
						}
					}
				}
			}
		}
		return ret;
	}

	List<Commandable> findCommandablesAtPosition(final PlayfieldPosition pos, final boolean isWithOtherPlayers) {
		final List<Commandable> ret = new LinkedList<Commandable>();
		final GameMap map = getMap();
		if (map.getProperties().isWithin(pos)) {
			// Add city at position, if any
			final FieldAlike field = map.getAt(pos);
			if (field instanceof Commandable) {
				final Commandable c = (Commandable)field;
				if (isWithOtherPlayers || (c.getPlayerId() == getCurrentPlayerId())) {
					ret.add(c);
				}
			}
			// Then add units at position, if any
			final Collection<UnitAlike> units = findUnitsAtPosition(pos);
			if (null != units) {
				for (final UnitAlike c : units) {
					if (isWithOtherPlayers || (c.getPlayerId() == getCurrentPlayerId())) {
						ret.add(c);
					}
				}
			}
		}
		return ret;
	}

	/** 
	 * @param playerId the player ID
	 * @return all living commandables of the player 
	 */
	List<Commandable> findCommandables(int playerId) {
		final List<FieldCity> currentPlayerCities = getMap().findCities(playerId);
		final List<Commandable> ret = new LinkedList<Commandable>();
		ret.addAll(currentPlayerCities);
		for (final UnitAlike unit : getUnits()) {
			if (unit.getPlayerId() == playerId) {
				ret.add(unit);
			}
		}
		return ret;
	}

	/** 
	 * @return all living units and all cities, no matter what owner
	 */
	private List<Commandable> findAllCommandables() {
		return map.findAllCommandables();
	}

	/**
	 * Returns the commandable (city or unit) for the given key.
	 * @param key the commandable key to search for
	 * @return the commandable belonging to that key, or <code>null</code> if not found
	 */
	Commandable findCommandable(final CommandableKey key) {
		Commandable ret = null;
		if (null != key) {
			if (Templates.instance().CMDL_CITY.id == key.getTemplateId()) {
				ret = getMap().findCity(key.getId());
			} else {
				ret = findUnit(key);
			}
		}
		return ret;
	}

	Collection<UnitAlike> findUnitsAtPosition(final PlayfieldPosition pos) {
		// FIXME: Switch when implemented
//		return map.getAt(pos).getUnits();
		return map.findUnitsAtPosition(pos);
	}

	boolean hasUnitsAtPosition(final PlayfieldPosition pos) {
		return map.hasUnitsAtPosition(pos);
	}

	/**
	 * Returns the unit (dead or alive) that fits to the given key, or <code>null</code> if it doesn't exist.
	 * @param key the key to search for
	 * @return the unit
	 */
	Unit findUnit(final CommandableKey key) {
		return GameMap.findUnit(key, getAllUnits());
	}

	/** 
	 * @return all living units
	 */
	Collection<UnitAlike> getUnits() {
		return map.getUnits();
	}

	public List<Unit> getAllUnits() {
		return allUnits;
	}

	/** 
	 * A list of all units that have existed so far in the game.
	 * The dead ones are kept to be referenced to in the logbook.
	 * Persistent. 
	 */
	private List<Unit> allUnits = new LinkedList<Unit>();

	/**
	 * Switches to the next player that can move, if there is any.
	 * If the game should continue can be seen in {@link #hasGameEnded()}.
	 */
	synchronized void switchPlayer() {
		logger.info("switchPlayer()");
		final int nextPlayerId = findNextPlayer();
		if (Player.ID_NONE == nextPlayerId) {
			// The end result of the war is a dead planet
			gameStateBasics.setGameEnding(true, nextPlayerId);
			logger.info("switchPlayer(): End result is a dead planet");
		} else if (gameStateBasics.getPlayerId() == nextPlayerId) {
			// Only one player left
			gameStateBasics.setGameEnding(true, nextPlayerId);
			logger.info("switchPlayer(): Only one player left");
		} else {
			boolean foundHumanAlive = false;
			for (int i = 0; i < Player.MAX_NUMBER_OF_PLAYERS; i++) {
				final Player p = getPlayer(i);
				if (p.isHumanPlayer() && isAlive(p)) {
					foundHumanAlive = true;
					break;
				}
			}
			if (!foundHumanAlive) {
				// Only computer players left - that would result in a non-responsive computer, as the computer opponents fight for world dominance.
				gameStateBasics.setGameEnding(true, Player.ID_NONE);
				logger.info("switchPlayer(): Only computer players left");
			} else {
				// Game continues...
				int nextTurn = gameStateBasics.getTurn();
				if (nextPlayerId <= gameStateBasics.getPlayerId()) {
					nextTurn++;
				}
				gameStateBasics.setPlayerAndTurn(nextPlayerId, nextTurn);
				final Player player = getPlayer(gameStateBasics.getPlayerId());
				if (!isAlive(player)) {
					removePlayerFromGame(player, false);
				}
				startTurn();
			}
		}
	}

	boolean hasGameEnded() {
		return gameStateBasics.hasGameEnded();
	}

	void removePlayerFromGame(final Player player, final boolean hasWon) {
		player.removeFromGame();
	}

	private int findNextPlayer() {
		int ret = gameStateBasics.getPlayerId();
		final Player[] players = getProperties().getPlayers();
		for (int i = 0; i < players.length; i++) {
			ret++;
			if (ret >= players.length) {
				ret = 0;
			}
			if (getPlayer(ret).isPartOfTheGame()) {
				return ret;
			}
		}
		return Player.ID_NONE;
	}

	/**
	 * Checks if the given player is the winner of the game. 
	 * @param player the player
	 * @return <code>true</code> if the player is the only player left
	 */
	public boolean isWinner(final Player player) {
		boolean ret = false;
		if (Player.ID_NONE != player.getId()) {
			final int winningPlayerId = gameStateBasics.getWinningPlayerId();
			if (Player.ID_NONE == winningPlayerId) {
				if (isAlive(player)) {
					ret = true;
					for (final Player p : getProperties().getPlayers()) {
						if (p.getId() != player.getId()) {
							if (p.isPartOfTheGame() && isAlive(p)) {
								ret = false;
								break;
							}
						}
					}
				}
			} else if (player.getId() == winningPlayerId) {
				ret = true;
			}
		}
		return ret;
	}

	/**
	 * Checks if the given player is alive and part of the game. This means, she has some commandable left (airbases don't count).
	 * @param player the player
	 * @return <code>true</code> if the player has at least one non-airbase commandable left
	 */
	public boolean isAlive(Player player) {
		boolean ret = false;
		if (player.isPlayer()) {
			for (final Commandable commandable : findCommandables(player.getId())) {
				if (commandable.getKey().getTemplateId() != ImageKey.CMDL_ID_AIRBASE) {
					ret = true;
					break;
				}
			}
		}
		return ret;
	}

	public void startGame() {
		gameStateBasics.setPlayerAndTurn(0, 1);
		startTurn();
	}

	/**
	 * Starts turn for this player.
	 */
	private void startTurn() {
		// Make all the commandables ready for this turn
		for (final Commandable commandable : findCommandables(gameStateBasics.getPlayerId())) {
			commandable.startTurn();
		}
		// Tell all the views that the player changed.
		for (final GameView view : gameViews) {
			view.onCurrentPlayerChanged();
		}
		// Tell the view to start the turn, taking use of the commandables
		getCurrentGameView().onStartTurn();
	}

	public int getCurrentPlayerId() {
		return gameStateBasics.getPlayerId();
	}

	Player getPlayer(final int playerId) {
		Player player = Player.PLAYER_NONE;
		if (playerId != Player.ID_NONE) {
			player = getProperties().getPlayers()[playerId];
		}
		return player;
	}

	public int getTurn() {
		return gameStateBasics.getTurn();
	}

	/* Be careful - this returns no clone, but a reference, so better don't change it. */
	public GameStateBasics getGameStateBasics() {
		return gameStateBasics;
	}

	/** The current turn, current player the game is in. Persistent. */
	private GameStateBasics gameStateBasics = new GameStateBasics();

	public boolean isInMoveOver() {
		return isInMoveOver;
	}

	/** If is in move-over or not. Persistent. */
	private boolean isInMoveOver;

	public GameMap getMap() {
		return map;
	}

	/** Persistent. */
	private final GameProperties props;

	/** The map used in this game. Persistent. */
	private GameMap map;

	public GameView getGameView(final int playerId) {
		return gameViews[playerId];
	}

	public GameView getCurrentGameView() {
		return gameViews[getCurrentPlayerId()];
	}

	synchronized void finishTurn() {
		for (final Commandable commandable : findAllCommandables()) {
			if (commandable.getPlayerId() == getCurrentPlayerId()) {
				commandable.finishTurn();
			}
		}
		final Player player = getPlayer(getCurrentPlayerId());
		final String msgKey;
		if (!isAlive(player)) {
			removePlayerFromGame(player, false);
			msgKey = "Game.Logbook.Entry.GameLost.Text";
		} else if (isWinner(player)) {
			msgKey = "Game.Logbook.Entry.GameWon.Text";
		} else {
			msgKey = null;
		}
		if (null != msgKey) {
			final I18NStringWithFillIns msg = new I18NStringWithFillIns(msgKey, null, 
					new String[] { getProperties().getPlayers()[player.getId()].getName(), } );
			for (Player p : getProperties().getPlayers()) {
				getGameView(p.getId()).onLogbookEntryWorthyEvent(new LogbookEntry(getTurn(), player.getId(), msg));
			}
		}
	}

	/** 
	 * Executes the move - changes position, removes killed combatants, updates status messages etc.
	 * Called when the move animation is finished. 
	 * @param moveResult the move to execute
	 */
	void executeMove(final MoveResult moveResult) {
		logger.trace("executeMove(" + moveResult + ")");
		if (null != moveResult) {
			for (final GameView gameView : gameViews) {
				gameView.onMoveResult(moveResult);
			}
			executeMoveInternally(moveResult);
			moveResult.getUnit().cleanRating();
		}
	}

	private void executeMoveInternally(final MoveResult moveResult) {
		final int mask = moveResult.getResultMask();
		logger.debug("Executing unit move result: 0x" + Integer.toHexString(moveResult.getResultMask()));
		int ownUnitsKilled = 0;
		int enemyUnitsKilled = 0;
		final Unit unit = moveResult.getUnit();
		if (0 != (mask & MoveResult.MOVE_MASK_FIGHTING)) {
			// A fight ends the current order - was an attack command anyways. 
			// Necessary for attacks without move on the target position, like fighters attacking production, or tanks attacking units on mountains.
			unit.setOrder(null);
			
			ownUnitsKilled = unitReceiveHits(moveResult.getCommandable(Role.Active), moveResult.getHits(Role.Active), getPlayer(moveResult.getCommandable(Role.Affected).getPlayerId()));
			enemyUnitsKilled = unitReceiveHits(moveResult.getCommandable(Role.Affected), moveResult.getHits(Role.Affected), unit.getPlayer());
			if (0 != (mask & MoveResult.MOVE_MASK_WINNING_FIGHT)) {
				final Commandable opponent = moveResult.getCommandable(Role.Affected);
				if (null != opponent) {
					if (opponent instanceof Unit) {
						killUnit((Unit) opponent, unit.getPlayer());
					} else { // The opponent is a city. Attack mode is reflected in move mask.
						if (0 != (mask & MoveResult.MOVE_MASK_MOVING)) {
							conquerCity(moveResult, (FieldCity) opponent);
						} else {
							// Empire Deluxe Manual page 135, rule m: may bomb cities
							attackCityProduction(moveResult, (FieldCity) opponent);
						}
					}
				}
				if (0 != (mask & MoveResult.MOVE_MASK_MOVING)) {
					moveUnit(moveResult);
				}
			} else {
				killUnit(unit, getPlayer(moveResult.getCommandable(Role.Affected).getPlayerId()));
			}
			unit.consumeStep();
		} else {
			if (0 != (mask & MoveResult.MOVE_MASK_MOVING)) {
				moveUnit(moveResult);
				unit.consumeStep();
			}
			if (0 != (mask & MoveResult.MOVE_MASK_LOSING_FIGHT)) {
				killUnit(unit, null);
			}
			if (0 != (mask & MoveResult.MOVE_MASK_DOCKING)) {
//				final RulesComplexity rulesComplexity = map.getProperties().getRulesComplexity();
//				if (Integer.MAX_VALUE != unit.getUnitTemplate().getStepsTillDeath(rulesComplexity)) {
					unit.consumeAllStepsThisTurn();
//				}
			}
			if (0 != (mask & MoveResult.MOVE_MASK_LOADING)) {
//				final RulesComplexity rulesComplexity = map.getProperties().getRulesComplexity();
//				if (Integer.MAX_VALUE != unit.getUnitTemplate().getStepsTillDeath(rulesComplexity)) {
					unit.setSkippingThisTurn(true);
//				}
			}
//			if (0 != (mask & MoveResult.MOVE_MASK_SKIP)) {
//				moveResult.getUnit().consumeAllStepsThisTurn();
//			}
			if (0 != (mask & MoveResult.MOVE_MASK_GARRISON)) {
				unit.consumeAllStepsThisTurn();
				map.removeUnit(unit);
				// TODO: Check, if not also remove unit from map of current player?
			}
			if (0 != (mask & MoveResult.MOVE_MASK_ALERT_IMPEDIMENT)) {
				unit.setOrder(null);
			}
			if (0 != (mask & MoveResult.MOVE_MASK_ALERT_SIGHTING)) {
				unit.setSeenEnemiesLately(false);
			}
			if (0 != (mask & MoveResult.MOVE_MASK_MOVE_OVER)) {
				isInMoveOver = true;
				logger.debug("IsToMoveOver=true");
			} else {
				isInMoveOver = false;
				logger.debug("IsToMoveOver=false");
			}
			if (0 != (mask & MoveResult.MOVE_MASK_FUEL_EMPTY)) {
				killUnit(unit, unit.getPlayer());
			}
			if (0 != (mask & MoveResult.MOVE_MASK_SLOWED_DOWN)) {
				unit.setSkippingThisTurn(true);
			}
		}
		moveResult.setKilledUnits(ownUnitsKilled, enemyUnitsKilled);
		// Clear order if fulfilled
		final Order order = unit.getOrder();
		if ((null != order) && (order.isFulfilled())) {
			unit.setOrder(null);
		}
		// Check if player involved is losing
		final Role roleToCheckForLosing;
		final Role roleToCheckForWinning;
		if ((0 != (mask & MoveResult.MOVE_MASK_FIGHTING)) 
			&& (0 != (mask & MoveResult.MOVE_MASK_WINNING_FIGHT))) {
			roleToCheckForLosing = Role.Affected;
			roleToCheckForWinning = Role.Active;
		} else {
			roleToCheckForLosing = Role.Active;
			roleToCheckForWinning = Role.Affected;
		}
		int playerToCheckForLosing = moveResult.getCommandableInfo(roleToCheckForLosing).getPlayerId();
		if (Player.ID_NONE != playerToCheckForLosing) {
			final Player player = getPlayer(playerToCheckForLosing);
			if (!isAlive(player)) {
				int umask = moveResult.getResultMask() | MoveResult.MOVE_MASK_LOSING_GAME;
				int playerToCheckForWinning = moveResult.getCommandableInfo(roleToCheckForWinning).getPlayerId();
				if (isWinner(getPlayer(playerToCheckForWinning))) {
					umask |= MoveResult.MOVE_MASK_WINNING_GAME;
					logger.debug("Setting MOVE_MASK_WINNING_GAME");
				}
				moveResult.setResultMask(umask);
				logger.debug("Setting MOVE_MASK_LOSING_GAME");
			}
		}
	}

	boolean calcFightDoesAttackerWin(final Unit unit, final Commandable commandable) {
		final float myStrength = unit.getFightStrength(commandable, true);
		final float oppStrength = commandable.getFightStrength(unit, false);
		final boolean hasAttackerWon = calcFightDoesAttackerWin(myStrength, oppStrength);
		return hasAttackerWon;
	}

	private boolean calcFightDoesAttackerWin(final float myStrength, final float oppStrength) {
		final float totalStrength = myStrength + oppStrength;
		final double rand = Math.random();
		final boolean hasAttackerWon = (rand < (myStrength / totalStrength));
		return hasAttackerWon;
	}

	private void moveUnit(final MoveResult moveResult) {
		final RulesComplexity rulesComplexity = map.getProperties().getRulesComplexity();
		final Unit unit = moveResult.getUnit();
		// Is this move an unloading?
		if (null != unit.getParentKey()) {
			unit.setParent(null);
		}
		// Do the move
		final DigitalJoystickDirection dir = moveResult.direction;
		unit.move(dir);
		for (final Unit cargo : findChildren(unit)) { // Also move containing units
			cargo.setXY(unit.getPosition());
		}
		// Is this move a loading?
		if (0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_LOADING)) {
			final Unit c2 = (Unit) moveResult.getCommandable(Role.Affected);
			final Unit container = findUnit(c2.getKey());
			unit.setParent(container.getKey());
			unit.setOrder(new OrderSentry(this, unit));
			final Order o = container.getOrder();
			if ((null != o) && (o.isFulfilled())) {
				container.setOrder(null);
			}
		}
		// Is this move a landing?
		if (Integer.MAX_VALUE != unit.getUnitTemplate().getStepsTillDeath(rulesComplexity)) {
			final FieldAlike field = map.getAt(unit.getPosition());
			if (((field.isCity()) && (field.getPlayerId() == unit.getPlayerId()))
					|| (null != unit.getParentKey())) {
				unit.setOrder(new OrderSkipMove(this, unit));
			}
		}
		// Has this move been sighted by enemies?
		for (final Commandable commandable : findAllCommandables()) {
			if ((commandable.getPlayerId() != gameStateBasics.getPlayerId()) 
					&& (commandable.getPlayerId() != Player.ID_NONE) 
					&& !commandable.hasSeenEnemiesLately()) {
				if (unit.isVisibleFor(unit.getPosition(), commandable, moveResult)) {
					commandable.setSeenEnemiesLately(true);
				}
			}
		}
	}

	/**
	 * Lets a commandable receive a number of hits.
	 * 
	 * @param commandable
	 *            the commandable to receive hits
	 * @param hits
	 *            the hits to take
	 * @param opponent
	 *            the opponent that shot
	 * @return the number of units killed. May be greater than 0, if this was a
	 *         transporting unit and the cargo is collateral damage
	 */
	private int unitReceiveHits(final Commandable commandable, final int hits, final Player opponent) {
		int ret = 0;
		commandable.receiveHits(hits);
		if (commandable instanceof Unit) {
			final Unit unit = (Unit) commandable;
			while (unit.calcRemainingCapacity() < 0) {
				final List<Unit> loadedUnits = findChildren(unit);
				ret += killUnit(loadedUnits.get(MathUtil.random(loadedUnits.size())), opponent);
			}
		}
		return ret;
	}

	/**
	 * Kills a unit.
	 * 
	 * @param unit
	 *            the unit to kill
	 * @param killingPlayer
	 *            the opponent that shot
	 * @return the number of units killed. May be greater than 1, if this was a
	 *         transporting unit and the cargo is collateral damage.
	 */
	int killUnit(final Unit unit, final Player killingPlayer) {
		int ret = 1;
		unit.consumeAllStepsThisTurn();
		unit.setOrder(null);
		for (final Unit u : findChildren(unit)) {
			ret += killUnit(u, killingPlayer);
		}
		int templateId = unit.getUnitTemplate().id;
		if ((templateId == ImageKey.CMDL_ID_CITY) || (templateId == ImageKey.CMDL_ID_AIRBASE)) {
			getMovePaths().removeAll(unit.getPlayerId(), unit.getPosition(), true);
		}
		unit.disband();
		getUnits().remove(unit);
		final CommandableTemplate unitTemplate = unit.getUnitTemplate();
		unit.getPlayer().incCntUnitsLost(unitTemplate);
		if ((null != killingPlayer) && killingPlayer.isPlayer() && (!killingPlayer.equals(unit.getPlayer()))) {
			killingPlayer.incCntUnitsKilled(unitTemplate);
		}
		return ret;
	}

	/**
	 * When a city is conquered, the unit conquering the city is removed, as it
	 * is needed to control the city. Furthermore, all other units present in
	 * the city are destroyed.
	 * 
	 * @param moveResult
	 *            the move result to evaluate
	 * @param city
	 *            the city which was conquered
	 * @return the number of units killed
	 */
	private int conquerCity(final MoveResult moveResult, final FieldCity city) {
		int ret = 0;
		final Unit unit = moveResult.getUnit();
		// Clear the city
		getMovePaths().removeAll(city.getPlayerId(), city.getPosition(), true);
		final List<Commandable> unitsInCity = findCommandablesAtPosition(city.getPosition(), true);
		for (final Commandable commandable : unitsInCity) {
			if (commandable instanceof Unit) {
				ret += killUnit((Unit) commandable, unit.getPlayer());
			}
		}
		// conquer the city
		final int prevPlayerId = city.getPlayerId();
		city.setPlayerId(gameStateBasics.getPlayerId());
		// garrison
		unit.consumeAllStepsThisTurn();
		map.removeUnit(unit);
		if (unit.getKey().getTemplateId() == ImageKey.CMDL_ID_TANK) {
			final Unit nunit = new Unit(this, Templates.instance().ARMY, city.getPlayerId(), city);
			addUnit(null, nunit);
			nunit.consumeAllStepsThisTurn();
		}
		// Update loser's map
		if (Player.ID_NONE != prevPlayerId) {
			final GameView view = getGameView(prevPlayerId);
			if (null != view) {
				view.updateViewMap(city, moveResult);
			}
		}
		return ret;
	}

	/**
	 * When a unit that cannot conquer the city is attacking, it can bombard the
	 * city production. Randomly destroys current production or reduces
	 * productivity.
	 * 
	 * @param moveResult
	 *            the move result to evaluate
	 * @param city
	 *            the city which was conquered
	 */
	private void attackCityProduction(final MoveResult moveResult, final FieldCity city) {
		logger.debug("GameState.attackCityProduction called.");
		if (0 == MathUtil.random(2)) {
			city.decreaseProductivity();
		} else {
			city.destroyUnitInProduction();
		}
	}

	@Override
	public String toString() {
		return super.toString();
	}

	/** 
	 * The game views (one for each player) that operate on this game state. Persistent. 
	 */
	private final GameView[] gameViews;

	/** Not persistent. */
	Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
	}

	/**
	 * Lets the current player resign. This removes all his/her units and sets
	 * all his/her cities to neutral.
	 * 
	 * @param player
	 *            the player to resign
	 */
	public void resign(Player player) {
		if (gameStateBasics.getPlayerId() == player.getId()) {
			for (final Commandable commandable : findCommandables(player.getId())) {
				if (commandable instanceof FieldCity) {
					final FieldCity city = (FieldCity) commandable;
					city.setPlayerId(Player.ID_NONE);
				} else {
					final Unit unit = (Unit) commandable;
					killUnit(unit, player);
				}
			}
		}
	}

	/** Persistent. */
	private MovePaths paths;

	public MovePaths getMovePaths() {
		return paths;
	}

	GameMap createViewMap(final GameView view) {
		return map.createViewMap(view);
	}

	public GameProperties getProperties() {
		return props;
	}

	public void init() {
		for (GameView gameView : gameViews) {
			if (null != gameView) {
				gameView.updateViewMap();
			}
		}
	}

	public float getTroopsSupportFactor(int playerId) {
		float ret = 1.0f;
		if (props.getRulesComplexity().isWithTroopSupportFactor) {
			final List<FieldCity> cities = map.findCities(playerId);
			int playerCities = 0;
			for (FieldCity city : cities) {
				playerCities += city.getTroopSupportFactor();
			}
			final int playerTroops = findCommandables(playerId).size() - cities.size();
			final int troops = playerTroops - playerCities;
			if (0 < troops) {
				ret = Math.max(0.0f, 1.0f - 0.1f * troops / playerCities);
			}
		}
		return ret;
	}
}
