package de.ped.empire.logic;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import de.ped.empire.logic.Commandable.AttackMode;
import de.ped.empire.logic.GameFacade.PlayerGetReadyMessage;
import de.ped.empire.logic.MoveResult.CommandableInfo;
import de.ped.empire.logic.MoveResult.Role;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.Messages;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.Bitfield32;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Contains a view to the game that the GUI can show.
 * Knows all about the game state, but also about some UI selections, current player etc.
 * <p>
 * Elements handled here:
 * <ul>
 *   <li>View mode (nomal, production, path) (non-persistent)
 *   <li>Current position (crosshair) (non-persistent)
 *   <li>Language (persisted in player)
 *   <li>Playback / animation status (non-persistent)
 * </ul>
 * There's a naming convention on the methods defined herein:
 * <ul>
 *   <li>methods starting with <em>on</em> are invoked by the {@link GameState} object, but not from the observer.
 * </ul>
 *
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameView implements Marshallable {
	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	/**
	 * The state within a round this player is in. Currently unused.
	 *
	 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum GameViewState {
		NotCurrent,
		GetReady,
		Active,
	}

	/**
	 * The crosshair object to draw if a city is in focus or when in survey
	 * view. Implemented as singleton.
	 * 
	 * @return the crosshair object
	 */
	public static Crosshair getCrosshair() {
		if (null == crosshair) {
			crosshair = new Crosshair();
		}
		return crosshair;
	}

	/**
	 * The crosshair object to draw if a city is in focus or when in survey view. 
	 * Lazy evaluation, not persistent.
	 */
	private static Crosshair crosshair;

	/**
	 * Returns the image key of the unit icon to draw instead of the city when
	 * in {@link MapViewMode#PRODUCTION}.
	 * 
	 * @param city
	 *            the city to change the icon for
	 * @return the image key of the unit icon to draw
	 */
	public int getCityUnitInProductionImageKey(final FieldCity city) {
		final int templateId = city.getUnitInProduction().id;
		final int ret;
		if (templateId == ImageKey.CMDL_ID_UNDEFINED) {
			ret = city.getImageKey();
		} else {
			final ImageKey key = new ImageKey(ImageKey.TType.Unit, templateId);
			key.setBits(ImageKey.BA_PLAYER_COLOR, getPlayer().getPlayerColorId());
			ret = key.getAll();
		}
		logger.trace("getCityUnitInProductionImageKey(" + city + ") -> 0x" + Integer.toHexString(ret));
		return ret;
	}

	/** Not persistent. */
	final Logger logger;

	/** The observer that will be informed of status changes. Non-persistent. */
	final private GameObserver observer;

	/** The i18n messages instance. Non-persistent. */
	final private Messages messages;

	/** 
	 * The map view the current user has selected.
	 * Non-persistent. 
	 */
	private MapViewMode mapViewMode = MapViewMode.NORMAL;

	private boolean isGetReady;

	/**
	 * The current commandable that is in focus, or <code>null</code> if there is none. 
	 * Persistent as reference to allUnits. 
	 */
	private Commandable commandableInFocus;

	/**
	 * Knows the position to focus on when not in {@link MapViewMode#NORMAL}.
	 * Non-persistent. 
	 */
	private PlayfieldPosition positionInFocus = new PlayfieldPosition();

	/** The stack of status messages. Persistent. */
	private final Stack<I18NStringWithFillIns> statusMessages;

	/** The logbook of important messages. Persistent. */
	private ArrayList<LogbookEntry> logbook = new ArrayList<LogbookEntry>();

	private final Player player;
	
	final private GameState gameState;

	/** Persistent. */
	private GameViewProperties gameViewProperties;

	/** Persistent. */
	private int viewPerspective;

	public GameView(final GameState gameState, final Player player, final GameObserver observer) {
		this.gameState = gameState;
		this.player = player;
		this.observer = observer;
		logger = Logger.getLogger(toString());
		logger.setLogLevel(Level.INFO);
		logger.setToShowThread(true);
		messages = observer.getMessages();
		statusMessages = new Stack<I18NStringWithFillIns>();
		statusMessages.push(I18NStringWithFillIns.EMPTY);
		commandableInFocus = null;
		gameViewProperties = new GameViewProperties();
		map = gameState.createViewMap(this);
		updateViewMap();
		viewPerspective = getDefaultPerspective();
		gameNotification.addEvent(GameNotification.MASK_NEW_GAME_STARTED);
	}

	public GameView(final GameState gameState, final Player player, final GameObserver observer, final Marshaller marshaller) {
		this.gameState = gameState;
		this.player = player;
		this.observer = observer;
		logger = Logger.getLogger(toString());
		logger.setLogLevel(Level.DEBUG);
		messages = observer.getMessages();
		statusMessages = new Stack<I18NStringWithFillIns>();
		statusMessages.push(I18NStringWithFillIns.EMPTY);
		demarshal(marshaller);
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeNullOrMarshallable(map);
		if (null != map) {
			map.marshalLivingUnits(marshaller);
		}
		// Write commandableInFocus
		GameState.marshalCommandableKey(marshaller, commandableInFocus, true);
		logger.trace("marshal: CommandableInFocus=" + (null == commandableInFocus ? commandableInFocus : commandableInFocus.getKey()));
		marshaller.writeMarshallable(gameViewProperties);
		{	// Write stack of status messages
			final int size = statusMessages.size();
			marshaller.writeInt(size);
			for (int i = 0; i < size; i++) {
				marshaller.writeMarshallable(statusMessages.get(i));
			}
		} { // Write logbook
			final int size = logbook.size();
			marshaller.writeInt(size);
			for (int i = 0; i < size; i++) {
				marshaller.writeMarshallable(logbook.get(i));
			}
		} { // Write latest enemy moves
			for (final List<Animation> animationSequence : animations) {
				final int size = animationSequence.size();
				marshaller.writeInt(size);
				for (int i = 0; i < size; i++) {
					marshaller.writeMarshallable(animationSequence.get(i));
				}
			}
		}
		marshaller.writeInt(viewPerspective);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		if (marshaller.readNullIndicator()) {
			map = null;
		} else {
			map = new GameMap(marshaller, gameState.getProperties(), this);
			map.setGameState(gameState);
			map.demarshalLivingUnits(marshaller, gameState);
		}
		// Read commandableInFocus
		commandableInFocus = gameState.demarshalCommandableKey(marshaller, true);
		if (null != commandableInFocus) {
			setPositionInFocus(commandableInFocus.getPosition());
		}
		logger.trace("demarshal: CommandableInFocus=" + (null == commandableInFocus ? commandableInFocus : commandableInFocus.getKey()));
		gameViewProperties = (GameViewProperties) marshaller.readMarshallable(new GameViewProperties());
		{ // Read stack of status messages
			final int size = marshaller.readInt();
			statusMessages.clear();
			statusMessages.ensureCapacity(size);
			for (int i = 0; i < size; i++) {
				statusMessages.add(new I18NStringWithFillIns(marshaller));
			}
		} { // Read logbook
			final int size = marshaller.readInt();
			logbook.clear();
			logbook.ensureCapacity(size);
			for (int i = 0; i < size; i++) {
				logbook.add(new LogbookEntry(this, marshaller));
			}
		} { // Read latest enemy moves
			for (final List<Animation> animationSequence : animations) {
				final int size = marshaller.readInt();
				animationSequence.clear();
	//			animationSequence.ensureCapacity(size);
				for (int i = 0; i < size; i++) {
					animationSequence.add(new Animation(this, marshaller));
				}
			}
		}
		viewPerspective = marshaller.readInt();
		return this;
	}

	public GameState getGameState() {
		return gameState;
	}

	public PlayfieldDimension getMapSize() {
		return getProperties().getSize();
	}

	/**
	 * @param position
	 *            the position
	 * @return the field at the given position from the map as the player knows
	 *         it
	 */
	public FieldAlike getAt(final PlayfieldPosition position) {
		return getMap().getAt(position);
	}

	public boolean isUsingFogOfWar() {
		final RulesComplexity rulesComplexity = gameState.getProperties().getRulesComplexity();
		final boolean ret = rulesComplexity.isWithFogOfUncertainty;
		return ret;
	}

	public GameProperties getProperties() {
		return gameState.getProperties();
	}

	/** 
	 * @return the map, as this player currently views it
	 */
	public GameMap getMap() {
		final GameMap ret;
		if (0 == viewPerspective) {
			ret = gameState.getMap();
		} else {
			ret = gameState.getGameView(viewPerspective - 1).getPlayerMap();
		}
		return ret;
	}

	/** 
	 * @return the map, as this player knows it 
	 */
	public GameMap getPlayerMap() {
		final GameMap ret;
		if (null != map) {
			ret = map;
		} else {
			ret = gameState.getMap();
		}
		return ret;
	}

	public Player[] getPlayers() {
		return gameState.getProperties().getPlayers();
	}

	/** 
	 * @return all units known as alive on this map
	 */
	public Collection<UnitAlike> getUnits() {
		if (0 != getPerspective()) {
			return getMap().getUnits();
		} else {
			return gameState.getUnits();
		}
	}

	public Messages getMessages() {
		return messages;
	}

	public GameViewProperties getGameViewProperties() {
		return gameViewProperties;
	}

	public void setGameViewProperties(GameViewProperties gamesViewProperties) {
		this.gameViewProperties.copyFrom(gamesViewProperties);
	}

	public Commandable findCommandable(final CommandableKey key) {
		return gameState.findCommandable(key);
	}

	/**
	 * Returns the unit (dead or alive) that fits to the given key, or <code>null</code> if it doesn't exist.
	 * @param key the key to search for
	 * @return the unit
	 */
	public Unit findUnit(final CommandableKey key) {
		return gameState.findUnit(key);
	}

	public Collection<UnitAlike> findUnitsAtPosition(final PlayfieldPosition pos) {
		final GameMap map = getMap();
		if (null != map) {
			// FIXME: Switch when implemented
//			return map.getAt(pos).getUnits();
			return map.findUnitsAtPosition(pos);
		} else {
			return gameState.findUnitsAtPosition(pos);
		}
	}

	/** 
	 * @return <code>true</code> if the player get ready dialog should be shown
	 */
	public boolean isGetReady() {
		return isGetReady;
	}

	/** Call this method when the player get ready dialog was closed. */
	public synchronized void clearGetReady() {
		isGetReady = false;
		if (null != map) {
			map.outdate();
			updateViewMap();
			if (!getAnimation(AnimationSequenceType.ENEMIES).isEmpty()) {
				setMapViewMode(MapViewMode.REPLAY);
			}
//			addAnimations(getAnimation(AnimationSequenceType.ENEMIES));
		}
		addGameNotificationEventCommandableInFocus(commandableInFocus);
		nextCommandable();
	}

	/** 
	 * Ends the current player's turn, selects the next player, and starts his/her turn. 
	 * @return <code>true</code> if there IS a next player, <code>false</code> if everybody's dead
	 */
	public synchronized boolean finishTurn() {
		if (player.getId() != gameState.getCurrentPlayerId()) {
			throw new IllegalStateException("You can't finish turn. Current player is " + gameState.getCurrentPlayerId() + ", but you are player " + player.getId());
		}
		logger.info("finishTurn() started with turn " + gameState.getTurn());
		setCommandableInFocus(null);
		for (final List<Animation> animation : animations) {
			animation.clear();
		}
		gameState.finishTurn();
		gameState.switchPlayer();
		setStatusMessage(new I18NStringWithFillIns("Game.Status.NotCurrentPlayer.Text", "...", 
				new String[] { getProperties().getPlayers()[gameState.getCurrentPlayerId()].getName(), } ));
		return !gameState.hasGameEnded();
	}

	/** 
	 * Called by game state whenever this game view gets current game view.
	 * <p>
	 * Remark: Don't synchronize this lightheartedly, as it may block. 
	 */
	void onStartTurn() {
		logger.info("onStartTurn() started with turn " + gameState.getTurn());
		setCommandableInFocus(null);
		if (null != commandableInFocus) {
			throw new IllegalStateException("startTurn found commandable in focus not null, was " + commandableInFocus.getKey());
		}
		setStatusMessage(new I18NStringWithFillIns("Game.Status.GetReady.Text", "Get ready, %0!",
				new String[] { player.getName() }));
		isGetReady = true;
		logger.trace("onStartTurn() finished with turn " + gameState.getTurn());
	}

	private GameNotification gameNotification = new GameNotification();
	
	GameNotification getGameNotification() {
		return gameNotification;
	}

	synchronized void setGameNotificationEvent(final int event, final CommandableKey cmdlKey) {
		gameNotification.addEvent(event);
		gameNotification.setCommandablePrompting(cmdlKey);
	}

	/**
	 * Finds the next commandable, if current player and in {@link MapViewMode#NORMAL}.
	 */
	public synchronized void nextCommandable() {
		if (this == gameState.getCurrentGameView()) {
			if (mapViewMode == MapViewMode.NORMAL) {
				loseFocusOnCommandable();
				final Commandable commandable = findNext();
				executeMove(commandable, true, (null != commandable) && (!commandable.hasSeenEnemiesLately()));
			}
		}
	}

	public synchronized void nextCommandable(final Commandable commandable) {
		Assert.assertTrue(isMyTurn());
		loseFocusOnCommandable();
		executeMove(commandable, true, false);
	}

	private void loseFocusOnCommandable() {
		if (null != commandableInFocus) {
			commandableInFocus.nextFocusLost();
		}
	}

	private void executeMove(final Commandable commandable, final boolean isToGainFocus, final boolean isToExecuteOrders) {
		if (null != commandable) {
			commandable.setSkippingThisTurn(false);
		}
		if (isToGainFocus && (null != commandable)) {
			commandable.nextFocusGained();
		}
		setCommandableInFocus(commandable);
		gameState.setLoadingCommandIfLoadingUnitNearby(commandable);
		if (null != commandableInFocus) {
			// Is city and has undefined production? --> get order to change production
			if (commandableInFocus.getKey().getTemplateId() == ImageKey.CMDL_ID_CITY) {
				if ((null == commandableInFocus.getOrder())) {
					final FieldCity city = (FieldCity) commandableInFocus;
					if (city.getUnitInProduction() == Templates.instance().UNDEFINED) {
						city.setOrder(new OrderChangeProduction(gameState, commandableInFocus));
					}
				}
			}
			// Execute orders
			MoveResult result = null;
			logger.trace("Cmdl " + commandableInFocus.toString() + " hasSeenEnemiesLately=" + commandableInFocus.hasSeenEnemiesLately() + " isToExecuteOrders=" + isToExecuteOrders);
			if (isToExecuteOrders) {
				if (commandableInFocus.hasSeenEnemiesLately()) {
					if (commandableInFocus instanceof Unit) { // TODO: Enhance this also for cities
						result = new MoveResult(this, (Unit) commandableInFocus, null);
						result.setResultMask(MoveResult.MOVE_MASK_ALERT_SIGHTING);
						gameState.executeMove(result);
					}
				} else {
					result = commandableInFocus.evaluateOrder(this, true);
					gameState.executeMove(result);
				}
			}
			logger.trace("Updating map");
			updateViewMap(commandableInFocus, result);
			if (commandableInFocus.hasSeenEnemiesLately()) {
				setStatusMessage(commandableInFocus.getToolTip(this));
			}
		}
	}

	private Commandable findNext() {
		Commandable ret = null;
		for (int loops = 0; loops < 2; loops++) {
			List<Commandable> commandables = gameState.findCommandables(player.getId());
			logger.trace("In findNext, in loop " + loops + ", found " + commandables.size() + " commandables.");
			Iterator<Commandable> it = commandables.iterator();
			while (it.hasNext()) {
				final Commandable commandable = it.next();
				if (commandable.isNextable() && !commandable.wasNexted()) {
					ret = choose(ret, commandable, getPositionInFocus());
				}
			}
			if (null == ret) {
				Iterator<Commandable> itc = commandables.iterator();
				while (itc.hasNext()) {
					itc.next().resetNext();
				}
			} else {
				break;
			}
		}
		logger.trace("findNext() finished, found " + ret);
		return ret;
	}

	/**
	 * Decides which commandable to prefer when deciding which commandable should be next presented to the player.
	 * The decision is based on <code>null</code>ness, city vs. unit, kind of wakeup for cities, and distance to current position.
	 * @param c1 the first commandable to pick from; <code>null</code> is allowed
	 * @param c2 the other commandable to pick from; <code>null</code> is allowed
	 * @param pos the current position on the map where the user was already paying attention
	 * @return the preferred commandable of the two
	 */
	private Commandable choose(final Commandable c1, final Commandable c2, final PlayfieldPosition pos) {
		Commandable ret = c1;
		if (null == c1) {
			ret = c2;
		} else if (null == c2) {
			ret = c1;
		} else if (c1 instanceof FieldCity) {
			if (c2 instanceof FieldCity) {
				// A city with a unit produced lately always wins over a city with just enemies seen lately.
				// As secondary criterion, distance to current position decides.
				final FieldCity city1 = (FieldCity) c1;
				final FieldCity city2 = (FieldCity) c2;
				if (city1.hasBeenUnitProducedLately()) {
					if (city2.hasBeenUnitProducedLately()) {
						ret = chooseByDistance(c1, c2, pos);
					} else {
						ret = city1;
					}
				} else {
					if (city2.hasBeenUnitProducedLately()) {
						ret = city2;
					} else {
						ret = chooseByDistance(c1, c2, pos);
					}
				}
			} else {
				// City always wins over unit.
				ret = c1;
			}
		} else if (c2 instanceof FieldCity) {
			// City always wins over unit.
			ret = c2;
		} else {
			ret = chooseByDistance(c1, c2, pos);
		}
		return ret;
	}

	private Commandable chooseByDistance(final Commandable c1, final Commandable c2, PlayfieldPosition pos) {
		Commandable ret;
		if (null == pos) {
			pos = ORIGIN;
		} 
		if (getPlayerMap().getProperties().calcAirDistance(c1.getPosition(), pos) 
			<= getPlayerMap().getProperties().calcAirDistance(c2.getPosition(), pos)) {
			ret = c1;
		} else {
			ret = c2;
		}
		return ret;
	}

	private static final PlayfieldPosition ORIGIN = new PlayfieldPosition(0, 0);

	public Player getPlayer(final int playerId) {
		return gameState.getPlayer(playerId);
	}

	void onUnitAdded(final FieldCity city, final Unit unit) {
		if (null != city) {
			final I18NStringWithFillIns txt = new I18NStringWithFillIns("Game.City.ProductionStatus" + Messages.TEXT_POSTFIX, null, 
					new Object[] { city.getName(), unit.getName(), });
			setStatusMessage(txt);
			final LogbookEntry entry = new LogbookEntry(getGameState().getTurn(), player.getId(), Player.ID_NONE, 
					city.getPosition(), unit, txt);
			logbook.add(entry);
		} else { // Airbase creation
			final I18NStringWithFillIns txt = new I18NStringWithFillIns("Game.Unit.ProductionStatus" + Messages.TEXT_POSTFIX, null, 
					new Object[] { unit.getName(), });
			setStatusMessage(txt);
			final LogbookEntry entry = new LogbookEntry(getGameState().getTurn(), player.getId(), Player.ID_NONE, 
					unit.getPosition(), unit, txt);
			logbook.add(entry);
		}
	}

	/**
	 * Called from game state whenever the current player for the game state changed, to inform the views about it.
	 * <p>
	 * Remark: Should not be synchronized lightheartedly, as this sometimes blocks between game state and game view.
	 */
	void onCurrentPlayerChanged() {
		final int cpid = gameState.getCurrentPlayerId();
		gameNotification.addEvent(GameNotification.MASK_PLAYER_IN_FOCUS, cpid);
		logger.trace("onCurrentPlayerChanged() finished with GameNotification.MASK_PLAYER_IN_FOCUS event for current player=" + cpid);
	}

	public List<Commandable> findCommandablesAtPosition(final PlayfieldPosition coords, final boolean isWithOtherPlayers) {
		return gameState.findCommandablesAtPosition(coords, isWithOtherPlayers);
	}

	/**
	 * The different animation sequence lists to be handled within a view.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	enum AnimationSequenceType {
		/** The current animation to be displayed. */
		CURRENT,
		/** The list of enemy moves within the last turn. */
		ENEMIES,
		/** The last own move. */
		OWN,
	}

	/** The current animations to display, if any. Not persistent. */
	private List<List<Animation>> animations = new ArrayList<List<Animation>>();
	{
		for (int i = 0; i < AnimationSequenceType.values().length; i++) {
			animations.add(new LinkedList<Animation>());
		}
	}

	/** 
	 * Offered for testing purposes.
	 * @return the number of current animations
	 */
	public int getNumberOfCurrentAnimations() {
		return getAnimation(AnimationSequenceType.CURRENT).size();
	}

	private List<Animation> getAnimation(final AnimationSequenceType animationSequenceType) {
		return animations.get(animationSequenceType.ordinal());
	}

	public Animation getCurrentAnimation() {
		Animation ret = null;
		final List<Animation> animationSequence = getAnimation(AnimationSequenceType.CURRENT);
		if (!CollectionUtil.isNullOrEmpty(animationSequence)) {
			ret = animationSequence.get(0);
		}
		return ret;
	}

	/**
	 * Called whenever a unit (no matter what player) moves.
	 * Gives this view a chance to update its map and animations lists.
	 * @param moveResult the move to weave into the current view of the world
	 */
	void onMoveResult(final MoveResult moveResult) {
		logger.trace("onMoveResult(" + moveResult + ")");
		if ((null != moveResult) && (player.isHumanPlayer())) {
			final Animation animation = new Animation(this, moveResult);
			if (gameState.getCurrentPlayerId() == player.getId()) {
				getAnimation(AnimationSequenceType.CURRENT).add(animation);
				final List<Animation> ownAnimations = getAnimation(AnimationSequenceType.OWN);
				ownAnimations.clear();
				ownAnimations.add(animation);
			} else {
				if (isMoveCurrentlyVisible(moveResult)) {
					getAnimation(AnimationSequenceType.ENEMIES).add(animation);
					if (isUsingFogOfWar()) {
						if (0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_WINNING_FIGHT)) {
							final Commandable affected = moveResult.getCommandable(Role.Affected);
							if (affected instanceof Unit) {
								map.removeUnit((Unit) affected);
							}
							final PlayfieldPosition position = map.getProperties().wrap(moveResult.getUnit().getPosition(), moveResult.direction);
							final UnitAlike unit = new ViewUnit(moveResult.getUnit(), gameState.getTurn(), true, position);
							map.addUnit(unit);
						}
						if (0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_LOSING_FIGHT)) {
							final Commandable affected = moveResult.getCommandable(Role.Affected);
							if (affected instanceof Unit) {
								final UnitAlike unit = new ViewUnit((Unit) affected, gameState.getTurn(), true);
								map.addUnit(unit);
							}
							if (null != moveResult.direction) {
								final Unit unit = findUnit(moveResult.getUnit().getKey());
								unit.move(moveResult.direction);
							}
						}
					}
				}
			}
			logger.trace("Adding animation.");
		}
	}

	/**
	 * Checks if a move is currently visible by this player.
	 * @param moveResult the move to look at
	 * @return <code>true</code> if there is a commandable owned by this player nearby that could observe the move
	 */
	boolean isMoveCurrentlyVisible(final MoveResult moveResult) {
		boolean ret;
		if (isUsingFogOfWar()) {
			ret = false;
			final Collection<PlayfieldPosition> affectedPositions = moveResult.getPositions();
			if (0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_FIGHTING)) {
				final int playerId = getPlayer().getId();
				if ((playerId == moveResult.getCommandable(Role.Affected).getPlayerId())
						|| (playerId == moveResult.getCommandable(Role.Active).getPlayerId())) {
					// A fight is always visible for the units involved.
					ret = true;
				}
			} 
			if (!ret) {
				for (final PlayfieldPosition pos : affectedPositions) {
					if (isVisible(moveResult.getUnit(), pos, moveResult)) {
						ret = true;
						break;
					}
				}
			}
		} else {
			ret = true;
		}
		if (ret) {
			if (moveResult.getResultMask() == MoveResult.MOVE_MASK_DOCKING) { // Docking without moving, just staying inside city
				ret = false;
			}
		}
		return ret;
	}

	boolean isPositionCurrentlyVisible(final PlayfieldPosition position, final Collection<Field> visibleFields) {
		boolean ret = true;
		if (isUsingFogOfWar() && (0 != viewPerspective)) {
			ret = false;
			if (visibleFields.contains(gameState.getMap().getAt(position))) {
				ret = true;
			}
		}
		return ret;
	}

	private List<Animation> getCurrentAnimations() {
		return getAnimation(AnimationSequenceType.CURRENT);
	}

	public CommandableInfo findAnimationInfo(final CommandableKey commandableKey) {
		CommandableInfo ret = null;
		for (final Animation animation : getCurrentAnimations()) {
			final MoveResult result = animation.getMoveResult();
			for (final Role role : Role.values()) {
				final CommandableInfo info = result.getCommandableInfo(role);
				if ((null != info) 
						&& (null != info.commandable) 
						&& (info.commandable.getKey().equals(commandableKey))) {
					ret = info;
				}
			}
			if (null != ret) {
				break;
			}
		}
		return ret;
	}

	public Set<CommandableKey> calcCommandablesInvolved() {
		return calcCommandablesInvolved(getCurrentAnimations());
	}

	private Set<CommandableKey> calcCommandablesInvolved(final List<Animation> animations) {
		final Set<CommandableKey> ret = new HashSet<CommandableKey>();
		for (final Animation animation : animations) {
			for (final Role role : Role.values()) {
				final Commandable commandable = animation.getMoveResult().getCommandable(role);
				if (null != commandable) {
					ret.add(commandable.getKey());
				}
			}
		}
		return ret;
	}

	/** 
	 * Called whenever the game thread wakes up. 
	 * @param pst the processor speed type, to adapt number of steps of visualizations to the step duration
	 * @return <code>true</code> if all animations are finished
	 */
	public synchronized boolean step(final ProcessorSpeedType pst) {
		boolean ret = false;
		final List<Animation> currentAnimations = getCurrentAnimations();
		if (!CollectionUtil.isNullOrEmpty(currentAnimations)) {
			gameNotification.addEvent(GameNotification.MASK_POSITION_IN_FOCUS);
			final boolean result = currentAnimations.get(0).step(pst);
			if (result) {
				currentAnimations.remove(0);
				if (CollectionUtil.isNullOrEmpty(currentAnimations)) {
					ret = true;
					gameNotification.cleanAnimationRunning();
				}
			}
		} else {
			if (mapViewMode.isReplay()) {
				setMapViewMode(MapViewMode.NORMAL);
			}
		}
		final GameNotification notif = (GameNotification) gameNotification.clone();
		if ((null != notif) && (!notif.isEmpty())) {
			logger.trace("GameNotif1: " + gameNotification.toString());
			gameNotification.reset();
			logger.trace("GameNotif2: " + gameNotification.toString());
			if (notif.isEvent(GameNotification.MASK_PLAYER_IN_FOCUS)) {
				logger.trace("Notif: " + notif.toString());
			}
			observer.notifyGameNotification(this.getPlayer().getId(), notif);
		}
		observer.step(this.getPlayer().getId());
		return ret;
	}

	public boolean isAcceptingUserInput() {
		boolean ret = false;
		if (isMyTurn()) {
			final List<Animation> list = getCurrentAnimations();
			if (CollectionUtil.isNullOrEmpty(list)) {
				ret = true;
				logger.trace("isAcceptingUserInput: currentAnimation is empty");
			} else {
				logger.trace("isAcceptingUserInput: currentAnimation head is " + list.get(0));
			}
		}
		return ret;
	}

	/** 
	 * Executes the next move, if possible. If this means switching to the next commandable, it does so.
	 */
	public synchronized void nextMove() {
		Assert.assertTrue(isMyTurn());
		final Commandable commandable = getVisibleCommandableInFocus(); 
		if ((null != commandable) && (!commandable.hasSeenEnemiesLately()) /* && (null != cmd.getOrder()) */) {
			nextMove(commandable);
		}
	}

	/** 
	 * Executes the next move, if possible. If this means switching to the next commandable, it does so.
	 * @param commandable the commandable to continue with, if possible
	 */
	private void nextMove(final Commandable commandable) {
		if ((commandable instanceof Unit) && !commandable.isNextable()) {
			nextCommandable();
		} else if (null != commandable.getOrder()) {
//			commandable.setSeenEnemiesLately(false);
			executeMove(commandable, false, true);
		} else {
			setStatusMessage(commandable.getToolTip(this));
		}
	}

	/**
	 * @param pos
	 *            the position
	 * @return only <code>true</code> if the field at the position is a city and
	 *         that city belongs to another player (or is neutral)
	 */
	boolean isOwnedByEnemy(final PlayfieldPosition pos) {
		boolean ret = false;
		final FieldAlike field = getPlayerMap().getAt(pos);
		if (field.isCity()) {
			// Plan cities in only if already owned
			if (field.getPlayerId() != getGameState().getCurrentPlayerId()) {
				ret = true;
			}
		}
		return ret;
	}

	/**
	 * Evaluates if the field at position (<em>x</em>, <em>y</em>) is hypothetically visitable by the given unit.
	 * It checks for field capacity for this type of unit.
	 * If the distance of the position is small enough to be reached within this turn, 
	 * also other units (of the same player as well as enemy units) are considered as obstacles.
	 * @param unit the unit to move
	 * @param pos the position where the unit is assumed to stand
	 * @param isFightAnOption <code>true</code> if for this move a fight against an enemy is considered a possible move
	 * @param stepRatingPurpose <code>true</code> if for this move the destination field needs to be free
	 * @return the result of the evaluation if to move
	 */
	boolean isPossiblePlanAheadMove(
			final Unit unit, 
			final PlayfieldPosition pos, 
			final boolean isFightAnOption, 
			final StepRatingPurpose stepRatingPurpose) {
		boolean ret = false;
		final int distance = getPlayerMap().getProperties().calcAirDistance(unit.getPosition(), pos);
		String debugStr = null;
		if (0 < distance) {
			final FieldAlike field = getPlayerMap().getAt(pos);
			if (stepRatingPurpose.isTraversingUnexploredFields || !field.isUnexplored()) {
				final List<Commandable> commandables = findCommandablesAtPosition(pos, true);
				final Commandable enemy = gameState.findPreferredEnemy(unit, pos, commandables);
				final boolean canHold = gameState.canHold(unit, field, commandables, false);
				if (canHold) { // Can I move onto field?
					if (1 == distance) {
						// Don't move a land unit from ship to ship
						boolean isMoveOkay = true;
						if (MilitaryBranch.LAND_UNIT == unit.getUnitTemplate().militaryBranch) {
							final FieldAlike startField = getPlayerMap().getAt(unit.getPosition());
							isMoveOkay = (field.isLand() || startField.isLand());
						}
						if (isMoveOkay) {
							if (null != enemy) {
								if (isFightAnOption) {
									final AttackMode attackMode = unit.getAttackModeAgainst(enemy);
									ret = (AttackMode.IMPOSSIBLE != attackMode);
									if (!ret) {
										debugStr = "within canHold=true: (isFightAnOption) was true, but ret was false";
									}
								} else {
									debugStr = "within canHold=true:  (isFightAnOption) was false";
								}
							} else { // Peaceful move - is there room left now on this field or a unit?
								if (stepRatingPurpose.isDestinationFieldToBeFreeNow) {
									ret = gameState.canHold(unit, field, commandables, true);
									if (!ret) {
										debugStr = "within canHold=true:  (isDestinationFieldToBeFreeNow) was true, but ret was false";
									}
								} else {
									ret = true;
								}
							}
						} else {
							debugStr = "within canHold=true:  (isMoveOkay) was false";
						}
					} else {
						ret = true;
					}
				} else { // If I can't move onto the field, can I attack it?
					boolean isMoveOkay = true;
					if (1 == distance) {
						// Don't move a land unit from ship to ship
						if (MilitaryBranch.LAND_UNIT == unit.getUnitTemplate().militaryBranch) {
							final FieldAlike startField = getPlayerMap().getAt(unit.getPosition());
							isMoveOkay = (field.isLand() || startField.isLand());
						}
					}
					if (isMoveOkay) {
						if (null != enemy) {
							if (isFightAnOption) {
								final AttackMode attackMode = unit.getAttackModeAgainst(enemy);
								ret = (AttackMode.IMPOSSIBLE != attackMode);
								if (!ret) {
									debugStr = "within canHold=false: (AttackMode.IMPOSSIBLE != attackMode) was false";
								}
							} else {
								debugStr = "within canHold=false: (isFightAnOption) was false";
							}
						} else {
							debugStr = "within canHold=false: (null != enemy) was false";
						}
					} else {
						debugStr = "within canHold=false: (isMoveOkay) was false";
					}
				}
			}
		} else {
			debugStr = "within canHold=false: (0 < distance) was false";
		}
		if (null != debugStr) {
			logger.trace("GameView.isPossiblePlanAheadMove(" + unit + "," + pos.toString() + "): " + debugStr);
		}
		return ret;
	}

	/**
	 * Evaluates a move of the given unit one step into the given direction.
	 * Currently only moved by route planning orders.
	 * 
	 * @param unit
	 *            the unit to move
	 * @param dir
	 *            the direction to move the unit to
	 * @param isFightAnOption
	 *            <code>true</code> if the unit is allowed to start a fight
	 *            within this move. (Can be <code>false</code> when executing a
	 *            long-lasting {@link Order}.)
	 * @param stepRatingPurpose
	 *            {@link StepRatingPurpose#NORMAL} if units in the way should be taken into account
	 * @return the result of the evaluation if to move
	 */
	MoveResult evaluateMove(
			final Unit unit, 
			final DigitalJoystickDirection dir, 
			final boolean isFightAnOption,
			final StepRatingPurpose stepRatingPurpose) {
		final MoveResult ret = new MoveResult(this, unit, dir);
		int mask = MoveResult.MOVE_MASK_NOTHING;
		if (null != dir) {
			final PlayfieldPosition upos = unit.getPosition();
			final PlayfieldPosition newpos = getProperties().wrap(upos, dir);
			logger.trace("evaluateMove(" + unit + "," + dir + "," + isFightAnOption + "," + stepRatingPurpose + ") " + upos.toString() + "->" + newpos.toString());
			boolean isMoveToHomeBase = false;
			if (isPossiblePlanAheadMove(unit, newpos, isFightAnOption, stepRatingPurpose)) {
				final RulesComplexity rulesComplexity = getProperties().getRulesComplexity();
				final Commandable enemy = getGameState().findPreferredEnemy(unit, newpos, null);
				if (null == enemy) {
					final FieldAlike field = getPlayerMap().getAt(newpos);
					if (unit.getStepsLeftThisTurn() < gameState.getMap().calcStepsToConsume(unit.getKey().getTemplateId(), field.getPosition())) {
						// Don't move
						if (unit.hasMovedThisTurn()) { // Maybe next turn the unit has enough steps left
							mask = MoveResult.MOVE_MASK_SLOWED_DOWN;
						} else { // This move will always be impossible for this unit
							mask = MoveResult.MOVE_MASK_NOTHING;
						}
					} else {
						mask = MoveResult.MOVE_MASK_MOVING;
						if (field.isCity()) {
							// If target is a city, then add the DOCK flag.
							if (unit.getUnitTemplate().canDock(this)) {
								ret.setAffectedCommandable((FieldCity) field.getField());
								mask = mask | MoveResult.MOVE_MASK_DOCKING;
								isMoveToHomeBase = true;
							}
						} else {
							// If target can be a unit, add the LOAD flag.
							final Collection<UnitAlike> units = findUnitsAtPosition(newpos);
							if (!CollectionUtil.isNullOrEmpty(units)) {
								for (final UnitAlike container : units) {
									if (container.canLoad(unit, true)) {
										ret.setAffectedCommandable(container.getUnit());
										mask = mask | MoveResult.MOVE_MASK_LOADING;
										isMoveToHomeBase = true;
										break;
									}
								}
								// Otherwise it must be move-over
								if (0 == (mask & MoveResult.MOVE_MASK_LOADING)) {
									mask = mask | MoveResult.MOVE_MASK_MOVE_OVER;
									logger.trace("Move over!");
								}
							}
						}
					}
				} else {
					if (isFightAnOption) {
						mask = MoveResult.MOVE_MASK_FIGHTING;
						ret.setAffectedCommandable(enemy);
						while ((ret.getHits(Role.Active) < unit.getHitsTillDeath()) &&
								(ret.getHits(Role.Affected) < enemy.getHitsTillDeath())) {
							final boolean hasAttackerWon = gameState.calcFightDoesAttackerWin(unit, enemy);
							if (hasAttackerWon) {
								ret.hit(Role.Affected, unit.getShotDamage());
							} else {
								ret.hit(Role.Active, enemy.getShotDamage());
							}
						}
						if (ret.getHits(Role.Active) >= unit.getHitsTillDeath()) {
							mask |= MoveResult.MOVE_MASK_LOSING_FIGHT;
						}
						if (ret.getHits(Role.Affected) >= enemy.getHitsTillDeath()) {
							mask |= MoveResult.MOVE_MASK_WINNING_FIGHT;
							if (AttackMode.INCLUDING_MOVE == unit.getAttackModeAgainst(enemy)) {
								// Looking just at the unit fought against, we could move onto the target field. 
								// But there might be other reasons to stay where we are.
								final FieldAlike field = getPlayerMap().getAt(newpos);
								boolean canMove = true;
								if (field.isCity()) {
									// Attacked a unit in a city, not the city itself -> stay where you are
									if (AttackMode.INCLUDING_MOVE != 
											unit.getUnitTemplate().getAttackModeAgainst(rulesComplexity, Templates.instance().CMDL_CITY)) {
										canMove = false;
									}
								} else {
									if (0 == field.getCapacity(unit)) { 
										// Am I allowed to move onto the field (might be forbidden, e.g., for tanks on mountains)
										canMove = false;
									} else if (Bitfield32.isBit(field.getTemplate().rcdvs[rulesComplexity.ordinal()].unitsSlowedDown, unit.getKey().getTemplateId()) 
											&& unit.hasMovedThisTurn()) {
										// Don't move
										canMove = false;
									}
								}
								if (canMove) {
									mask |= MoveResult.MOVE_MASK_MOVING;
								}
							}
						}
					}
				}
			}
			if (!isMoveToHomeBase && (1 >= unit.getStepsLeftTillDeath())) {
				mask |= MoveResult.MOVE_MASK_FUEL_EMPTY;
				logger.debug("evaluate: FuelEmpty set");
			}
		}
		ret.setResultMask(mask);
		return ret;
	}

	public Tileable getOrderOptionsObject(final PlayfieldPosition pos) {
		Tileable ret;
		switch(mapViewMode) {
		case NORMAL:
			ret = (TileTemplate) getVisibleCommandableInFocus();
			break;
		case SURVEY:
		case PRODUCTION:
			ret = getPlayerMap().getAt(pos);
			break;
		case PATH:
			ret = (TileTemplate) findCityOrAirbase(getPlayerMap().getAt(pos));
			break;
		default:
			ret = null;
		}
		return ret;
	}

	/**
	 * Returns the possible orders for the commandable in focus that would work
	 * for the given position and should be presented in a context menu.
	 * 
	 * @param pos
	 *            the position
	 * @return all orders to present to the user to select from
	 * @see #getVisibleCommandableInFocus()
	 */
	public List<Order> collectOrderOptions(final PlayfieldPosition pos) {
		Assert.assertTrue(isMyTurn());
		logger.trace("collectOrderOptions " + pos.toString());
		final FieldAlike field = getPlayerMap().getAt(pos);
		final List<Order> orderOptions = new LinkedList<Order>();
		switch(mapViewMode) {
		case NORMAL:
			final Commandable commandableInFocus = getVisibleCommandableInFocus();
			if (null != commandableInFocus) {
				final int distance = getPlayerMap().getProperties().calcAirDistance(commandableInFocus.getPosition(), pos);
				if (commandableInFocus instanceof Unit) {
					// Add all possible actions for the commandable in focus
					final Unit unit = (Unit) commandableInFocus;
					if (0 == distance) { // Just clicked on same field as commandableInFocus
						addOrderOption(orderOptions, new OrderSkipMove(gameState, unit));
						addOrderOption(orderOptions, new OrderSentry(gameState, unit));
						addOrderOption(orderOptions, new OrderLoad(gameState, unit));
						addOrderOption(orderOptions, new OrderUnload(gameState, unit));
						addOrderOption(orderOptions, new OrderClearOrders(gameState, unit));
						addOrderOption(orderOptions, new OrderClearAlerts(gameState, unit));
//						addOrderOption(orderOptions, unit, new OrderDisband());
					}
					addOrderOption(orderOptions, getMoveOrAttackOrder(unit, pos));
					if (OrderTemplate.ESCORT.makesSense(this, commandableInFocus)) {
						final List<Commandable> commandables = findCommandablesAtPosition(pos, false);
						for (final Commandable target : commandables) {
							if (OrderEscort.makesSense(gameState, commandableInFocus, target.getKey())) {
								try {
									final Order order = new OrderEscort(gameState, commandableInFocus, target.getKey());
									addOrderOption(orderOptions, order);
								} catch (final CircularReferenceException e) {
									// Nothing to do
								}
							}
						}
					}
				} else if (commandableInFocus instanceof FieldCity) {
					if (0 == distance) {
						addOrderOption(orderOptions, new OrderChangeProduction(gameState, commandableInFocus));
						addOrderOption(orderOptions, new OrderClearAlerts(gameState, commandableInFocus));
					}
				}
			}
			break;
		case SURVEY:
			// Nothing to add
			break;
		case PRODUCTION:
			if (field.isCity()) {
				final OrderChangeProduction order = new OrderChangeProduction(gameState, ((FieldCity) field.getField()));
				if (order.makesSense()) {
					orderOptions.add(order);
				}
			}
			break;
		case GROUP:
			// Nothing to add
			break;
		case PATH:
			Commandable possiblePathStart = findCityOrAirbase(field);
			if (null != possiblePathStart) {
				final OrderDefineMovePaths order = new OrderDefineMovePaths(gameState, possiblePathStart);
				if (order.makesSense()) {
					orderOptions.add(order);
				}
			}
			break;
		case REPLAY:
		case REPLAY_OWN:
			// Nothing to add
			break;
		default:
			throw new IllegalStateException("Unknown map view mode " + mapViewMode.name());
		}
		return orderOptions;
	}

	/**
	 * Returns either a move or attack order, or <code>null</code> if e.g. the distance was 0.
	 * @param unit the unit to move
	 * @param pos the position of the target field
	 * @return the goto or attack order to move to that field
	 */
	public Order getMoveOrAttackOrder(final Unit unit, final PlayfieldPosition pos) {
		Order ret = null;
		// FIXME: remove unnecessary PlayfieldPosition constructor when unit was transformed
//		final int distance = getPlayerMap().getProperties().calcAirDistance(new PlayfieldPosition(unit.getPosition()), pos);
		final int distance = getPlayerMap().getProperties().calcAirDistance(unit.getPosition(), pos);
		if (0 == distance) { // Just clicked on same field as commandableInFocus
			// Nothing to do
		} else if (1 == distance) {
			ret = getMoveOrAttackOrderForSingleStep(unit, pos);
		} else {
			ret = getMoveOrderToLoadingTargetUnit(unit, pos);
			if (null == ret) {
				ret = new OrderGoTo(gameState, unit, pos);
			}
		}
		return ret;
	}

	/**
	 * Tries to find a loading target unit for this moving unit at the target position.
	 * @param unit the unit to move
	 * @param pos the position of the target field
	 * @return the goto order to the fitting target unit at the target field, or <code>null</code> if none found
	 */
	private Order getMoveOrderToLoadingTargetUnit(final Unit unit, final PlayfieldPosition pos) {
		Order ret = null;
		final List<Commandable> targets = getGameState().findCommandablesAtPosition(pos, false);
		Commandable bestTarget = null;
		if (null != targets) {
			for (Commandable target : targets) {
				if (target.canLoad(unit, false)) {
					bestTarget = target;
					if (target.getKey().getTemplateId() == ImageKey.CMDL_ID_CITY) {
						break;
					}
				}
			}
		}
		if (null != bestTarget) {
			try {
				ret = new OrderGoTo(gameState, unit, bestTarget.getKey());
			} catch (CircularReferenceException e) {
				// Nothing to do
			}
		}
		return ret;
	}

	private Order getMoveOrAttackOrderForSingleStep(final Unit unit, final PlayfieldPosition pos) {
		Order order = null;
		final Commandable enemy = getGameState().findPreferredEnemy(unit, pos, null);
		if (null != enemy) {
			try {
				order = new OrderGoTo(gameState, unit, enemy.getKey());
				logger.trace("!!!This is an attack!!!");
			} catch (CircularReferenceException e) {
				// Nothing to do
			}
		}
		if (null == order) {
			order = new OrderGoTo(gameState, unit, pos);
		}
		return order;
	}

	private void addOrderOption(final List<Order> orderOptions, final Order order) {
		if ((null != order) && (order.makesSense())) {
			orderOptions.add(order);
		}
	}

	/**
	 * Returns the possible commandables for position <em>x, y</em> that the user might want to be the new commandable in focus.
	 * @param pos the position
	 * @return all commandables to present to the user to select from
	 * @see #getVisibleCommandableInFocus()
	 */
	public List<Commandable> collectCommandableOptions(final PlayfieldPosition pos) {
		Assert.assertTrue(isMyTurn());
		final List<Commandable> commandables = findCommandablesAtPosition(pos, false);
		final List<Commandable> commandableOptions = new LinkedList<Commandable>();
		final Commandable commandableInFocus = getVisibleCommandableInFocus();
		for (final Commandable commandable : commandables) {
			switch(mapViewMode) {
			case NORMAL:
				if (commandable != commandableInFocus) {
					if (commandable instanceof FieldCity) {
						commandableOptions.add(commandable);
					} else {
						final Unit unit = (Unit) commandable;
						if (unit.getUnitTemplate().isFocusable) {
							commandableOptions.add(commandable);
						}
					}
				}
				break;
			case SURVEY:
				commandableOptions.add(commandable);
				break;
			case PRODUCTION:
				// Nothing to show here - per definition the city is selected, if available.
				break;
			case PATH:
				// Nothing to show here - per definition the city or airbase is selected, if available.
				break;
			default:
				// Don't do anything
			}
		}
		return commandableOptions;
	}

	public boolean isMapViewModeEnabled(MapViewMode mode) {
		boolean ret = false;
		if (mode.isEnabled && isDefaultPerspective()) {
			switch(mode) {
			case REPLAY:
				ret = !getAnimation(AnimationSequenceType.ENEMIES).isEmpty();
				break;
			case REPLAY_OWN:
				ret = !getAnimation(AnimationSequenceType.OWN).isEmpty();
				break;
			case SURVEY:
				ret = !gameState.isInMoveOver();
				break;
			default:
				ret = true;
			}
		}
		return ret;
	}

	public MapViewMode getMapViewMode() {
		return mapViewMode;
	}

	public synchronized void setMapViewMode(final MapViewMode mode) {
		Assert.assertTrue(isMyTurn());
		logger.trace("setMapViewMode(" + mode + ") started, current mode is " + mapViewMode);
		// Leaving normal mode keeps position
		if ((mapViewMode == MapViewMode.NORMAL) && (mode != MapViewMode.NORMAL)) {
			if (commandableInFocus != null) {
				final PlayfieldPosition destination = positionInFocus;
				destination.setLocation(commandableInFocus.getPosition());
			}
		}
		// Entering replay mode
		if ((mapViewMode != MapViewMode.REPLAY) && (mode == MapViewMode.REPLAY)) {
			addAnimations(getAnimation(AnimationSequenceType.ENEMIES));
		} else if ((mapViewMode != MapViewMode.REPLAY_OWN) && (mode == MapViewMode.REPLAY_OWN)) {
			addAnimations(getAnimation(AnimationSequenceType.OWN));
		}
		// Change mode
		final MapViewMode prevMode = mapViewMode;
		mapViewMode = mode;
		// Entering normal mode seeks commandable in focus
		if ((prevMode != MapViewMode.NORMAL) && (mapViewMode == MapViewMode.NORMAL)) {
			logger.trace("setMapViewMode(" + mode + ") has commandableInFocus=" + commandableInFocus);
			if (commandableInFocus == null) {
				nextCommandable();
			}
			logger.trace("setMapViewMode(" + mode + ") found commandableInFocus=" + commandableInFocus);
		}
		logger.debug("Map mode " + mode.name() + ", showing units " + Integer.toBinaryString(mode.bitmaskUnitsToShow));
		gameNotification.addEvent(GameNotification.MASK_MAP_MODE);
		final I18NStringWithFillIns toolTip;
		switch(mapViewMode) {
		case NORMAL:
			if (null != commandableInFocus) {
				toolTip = commandableInFocus.getToolTip(this);
			} else {
				toolTip = null; // We don't want to kill the end of turn message 
			}
			break;
		case SURVEY:
		case PATH:
		case PRODUCTION:
			toolTip = getStatusMessageForField(getPositionInFocus());
			break;
		case REPLAY:
		case REPLAY_OWN:
			toolTip = null;
			break;
		default:
			toolTip = getStatusMessageForField(getPositionInFocus());
		}
		if (null != toolTip) {
			setStatusMessage(toolTip);
		}
	}

	private void addAnimations(final List<Animation> replayAnimations) {
		if (!CollectionUtil.isNullOrEmpty(replayAnimations)) {
			for (final Animation animation : replayAnimations) {
				animation.reset();
			}
			getCurrentAnimations().addAll(0, replayAnimations);
		}
	}

	/** 
	 * @return if the controller is allowed to finish the turn
	 */
	public boolean isFinishTurnAllowed() {
		final boolean isMyTurn = isMyTurn();
		final boolean isGetReady = isGetReady();
		final Commandable cmdl = getVisibleCommandableInFocus();
		logger.trace("isFinishTurnAllowed(): isMyTurn=" + isMyTurn + ", isGetReady=" + isGetReady + ", mapViewMode=" + mapViewMode.toString() + ", cmdl=" + cmdl);
		return (isMyTurn && !isGetReady 
				&& (mapViewMode == MapViewMode.NORMAL) 
				&& (null == cmdl));
	}

	public boolean isMyTurn() {
		return (gameState.getCurrentGameView() == this);
	}

	/**
	 * Returns the commandable in focus, if there is any. A reason for no commandable is
	 * <ul>
	 *   <li>There is no commandable left that can move; 
	 *   	in this case {@link #isFinishTurnAllowed()} is <code>true</code>, or
	 *   <li>The map mode is not {@link MapViewMode#NORMAL}, so commandable in focus makes no sense, or 
	 *   <li>An animation is displayed, so no user interaction is allowed; in this case {@link #getCurrentAnimation()} is not <code>null</code>.
	 * </ul>
	 * 
	 * @return the commandable in focus, or <code>null</code> if there is none
	 */
	public Commandable getVisibleCommandableInFocus() {
		Commandable ret = null;
		if (!isGetReady() && getCurrentAnimations().isEmpty()) {
			ret = getCommandableInFocus();
			logger.trace("getVisibleCommandableInFocus(): ret=" + ret);
		} else {
			logger.trace("getVisibleCommandableInFocus(): not ready or in animation.");
		}
		return ret;
	}

	/**
	 * @return the commandable in focus that will receive the next user input
	 * @see #getVisibleCommandableInFocus()
	 * @see #nextCommandable(Commandable)
	 */
	public Commandable getCommandableInFocus() {
		return commandableInFocus;
	}

	private void setCommandableInFocus(final Commandable commandable) {
		Assert.assertTrue(isMyTurn());
		boolean hasChanged = false;
		if (commandable != commandableInFocus) {
			commandableInFocus = commandable;
			hasChanged = true;
		} else {
			logger.debug("commandableInFocus has not changed.");
		}
		logger.debug("setCommandableInFocus " 
				+ (null == commandable ? commandable : messages.getString(commandable.getName()) + " with order " + commandable.getOrder()));
		Assert.assertEquals(player.getId(), gameState.getCurrentPlayerId());
		I18NStringWithFillIns status;
		if (null != commandableInFocus) {
			if (mapViewMode == MapViewMode.SURVEY) {
				setMapViewMode(MapViewMode.NORMAL);
			}
			positionInFocus.setLocation(commandableInFocus.getPosition());
			status = commandable.getToolTip(this);
		} else {
			if (gameState.isAlive(player)) {
				if (gameState.isWinner(player)) {
					status = new I18NStringWithFillIns("Game.Status.GameWon" + Messages.TEXT_POSTFIX, null, null);
					gameState.removePlayerFromGame(player, true);
				} else {
					status = new I18NStringWithFillIns("Game.Status.NoCommandable" + Messages.TEXT_POSTFIX, null, null);
				}
			} else {
				status = new I18NStringWithFillIns("Game.Status.GameLost" + Messages.TEXT_POSTFIX, null, null);
				gameState.removePlayerFromGame(player, false);
			}
		}
		setStatusMessage(status);
		if (hasChanged) {
			addGameNotificationEventCommandableInFocus(commandable);
		}
	}

	public void addGameNotificationEventCommandableInFocus(
			final Commandable commandable) {
		gameNotification.addEvent(GameNotification.MASK_COMMANDABLE_IN_FOCUS);
		if ((null != commandable) && (commandable.getKey().getTemplateId() == ImageKey.CMDL_ID_CITY)) {
			FieldCity city = (FieldCity) commandable;
			if (city.hasBeenUnitProducedLately()) {
				gameNotification.addEvent(GameNotification.MASK_UNIT_PRODUCED);
			}
		}
	}

	/**
	 * @return the position in focus. Never change the content of the array you receive, as this is only a narrow copy!
	 */
	public PlayfieldPosition getPositionInFocus() {
		final List<Animation> currentAnimations = getCurrentAnimations();
		if (!CollectionUtil.isNullOrEmpty(currentAnimations)) {
			final Animation currentAnimation = currentAnimations.get(0);
			final MoveResult moveResult = currentAnimation.getMoveResult();
			if (null != moveResult) {
				positionInFocus.setLocation(moveResult.getCommandableInfo(Role.Active).position);
			}
		} else {
			switch(mapViewMode) {
			case NORMAL:
				if (null != commandableInFocus) {
					positionInFocus.setLocation(commandableInFocus.getPosition());
				}
				break;
			default:
				// Nothing to do
			}
		}
		return positionInFocus;
	}

	public synchronized void setPositionInFocus(final PlayfieldPosition pos) {
		positionInFocus.setLocation(pos);
		final FieldAlike field = getPlayerMap().getAt(pos);
		I18NStringWithFillIns toolTip = null;
		switch(mapViewMode) {
		case PRODUCTION:
			if (field.isCity()) {
				final Commandable city = (Commandable) field.getField();
				final OrderChangeProduction order = new OrderChangeProduction(gameState, city);
				if (order.makesSense()) {
					setCommandableInFocus(city);
				}
			}
			break;
		case PATH:
			final Commandable commandable = findCityOrAirbase(field);
			if (null != commandable) {
				final OrderDefineMovePaths order = new OrderDefineMovePaths(gameState, commandable);
				if (order.makesSense()) {
					setCommandableInFocus(commandable);
				}
			}
			break;
		default:
			// Nothing to do
		}
		if (null == toolTip) {
			final FieldAlike f = getMap().getAt(positionInFocus);
			if (null != f) {
				toolTip = f.getToolTip(this);
			}
		}
		if (null != toolTip) {
			setStatusMessage(toolTip);
			gameNotification.addEvent(GameNotification.MASK_POSITION_IN_FOCUS);
		}
	}

	/**
	 * Sets the order for the commandable in focus.
	 * @param generalOrder the order to set
	 * @return the move result from setting that order
	 * @see #getVisibleCommandableInFocus()
	 */
	public synchronized MoveResult setOrder(final Order generalOrder) {
		Assert.assertTrue(isMyTurn());
		final Order personalizedOrder = generalOrder.cloneIfNeededForCommandable(commandableInFocus);
		final MoveResult result;
		if (personalizedOrder.template.isOrder) {
			logger.trace("setOrder(" + generalOrder + "): Commandable in focus is " + commandableInFocus);
			commandableInFocus.setOrder(personalizedOrder);
			result = commandableInFocus.evaluateOrder(this, true);
		} else {
			logger.trace("No commandable in focus");
			result = personalizedOrder.evaluate(true);
		}
		if (null != result) {
			gameState.executeMove(result);
		}
		updateViewMap(commandableInFocus, result);
		nextMove(commandableInFocus);
		return result;
	}

	public String getStatusMessage() {
		return messages.getString(statusMessages.peek());
	}

	/**
	 * Gets the status message for the given field; 
	 * takes the map view mode into account to see what commandables are visible right now.
	 * @param pos the position to set the status message for
	 * @return the status message for the given field 
	 */
	public I18NStringWithFillIns getStatusMessageForField(final PlayfieldPosition pos) {
		final I18NStringWithFillIns ret;
		final List<Commandable> commandables = findCommandablesAtPosition(pos, true);
		final FieldAlike field = getAt(pos);
		Tileable obj = null;
		if (!CollectionUtil.isNullOrEmpty(commandables)) {
			for (Commandable commandable : commandables) {
				if (commandable instanceof Unit) {
					final Unit unit = (Unit) commandable;
					if (Bitfield32.isBit(mapViewMode.bitmaskUnitsToShow, unit.getUnitTemplate().id)
							&& (null == unit.getParentKey())) {
						obj = unit;
						break;
					}
				} else { // It's a city
					obj = field;
					break;
				}
			}
		}
		if ((null == obj) || (obj instanceof FieldAlike)) {
			I18NStringWithFillIns outdated  = I18NStringWithFillIns.EMPTY;
			I18NStringWithFillIns sightings = I18NStringWithFillIns.EMPTY;
			if (null == obj) {
				for (final UnitAlike unit : getUnits()) {
					if (PlayfieldPosition.equals(unit.getPosition(), pos)) {
						final int turnOfSighting = unit.getTurnOfSighting();
						sightings = new I18NStringWithFillIns("Game.Field.ToolTip.Sighting", null, 
								new Object[] { 
										unit.getToolTip(this), 
										Integer.valueOf(turnOfSighting), 
										Integer.valueOf(gameState.getTurn() - turnOfSighting),
								});
						break;
					}
				}
			}
			if ((ImageKey.getBits(field.getImageKey(), ImageKey.BA_TEMPLATE) != ImageKey.FIELD_ID_UNEXPLORED)
					&& (ImageKey.getBits(field.getImageKey(), ImageKey.BA_OUTDATED) != 0)) {
				outdated = new I18NStringWithFillIns("Game.Field.ToolTip.Outdated", null, null);
			}
			final I18NStringWithFillIns posStr = new I18NStringWithFillIns(null, field.getPosition().toString(), null);
			ret = new I18NStringWithFillIns("Game.Field.ToolTip", "%0%1%2", 
					new I18NStringWithFillIns[] { 
							field.getToolTip(this), outdated, sightings, posStr, } );
		} else {
			// For a commandable, simply show the tool tip
			ret = obj.getToolTip(this);
		}
		return ret;
	}

	public void setStatusMessage(final I18NStringWithFillIns statusMessage) {
		if (statusMessage.equals(statusMessages.firstElement())) {
			// Nothing to do
		} else {
			logger.trace("setStatusMessage(" + statusMessage + ")");
			statusMessages.pop();
			statusMessages.push(statusMessage);
			gameNotification.addEvent(GameNotification.MASK_STATUS_MESSAGE);
		}
	}

	public void setCityProduction(final FieldCity city, final CommandableTemplate unit) {
		// This seems overly complicated, but that way it doesn't matter which view map copy of the city we get as parameter.
		final GameMap gameMap = gameState.getMap();
		final FieldCity gameCity = (FieldCity) gameMap.getAt(city.getPosition());
		gameCity.setProduction(unit);
		final GameMap viewMap = getPlayerMap();
		if (gameMap != viewMap) {
			final ViewField field = (ViewField) viewMap.getAt(city.getPosition());
			if (ImageKey.FIELD_ID_CITY == field.getTemplateId()) {
				field.setUnexplored(false);
			}
		}
	}

	public synchronized void pushStatusMessage(final I18NStringWithFillIns statusMessage) {
		logger.debug("push status message: " + statusMessage);
		statusMessages.push(statusMessage);
		gameNotification.addEvent(GameNotification.MASK_STATUS_MESSAGE);
	}

	public synchronized void popStatusMessage() {
		logger.debug("pop status message");
		statusMessages.pop();
		gameNotification.addEvent(GameNotification.MASK_STATUS_MESSAGE);
	}

	public List<LogbookEntry> getLogbook() {
		return logbook;
	}

	void onLogbookEntryWorthyEvent(LogbookEntry logbookEntry) {
		logbook.add(logbookEntry);
	}

	/**
	 * @return the player this view is attached to
	 */
	public Player getPlayer() {
		return player;
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + player.getId();
	}

	/**
	 * Checks if a move path makes sense for the given military branch.
	 * @param path the path to check
	 * @param branch the military branch to check
	 * @return <code>true</code> if the move path makes sense for the given military branch
	 */
	public boolean makesSense(final MovePath path, final MilitaryBranch branch) {
		boolean ret = false;
		Unit pivot = null;
		if (getMap().getAt(path.getEndPos()).getTemplate().id != ImageKey.FIELD_ID_UNEXPLORED) {
			switch(branch) {
			case LAND_UNIT:
				// Forbid land paths from airports
				boolean isOkay = true;
				final Commandable cmdl = findCityOrAirbase(getPlayerMap().getAt(path.getStartPos()));
				if (cmdl instanceof Unit) {
					final Unit unit = (Unit) cmdl;
					if (ImageKey.CMDL_ID_AIRBASE == unit.getUnitTemplate().id) {
						isOkay = false;
					}
				}
				if (isOkay) {
					pivot = new Unit(getGameState(), Templates.instance().ARMY, getPlayer().getId(), getAt(path.getStartPos()));
				}
				break;
			case SEA_UNIT:
				pivot = new Unit(getGameState(), Templates.instance().TRANSPORTER, getPlayer().getId(), getAt(path.getStartPos()));
				break;
			case AIR_UNIT:
				// OK if (1) distance is within max distance of all air units, and (2) destination is city or airbase
				final RulesComplexity rulesComplexity = getProperties().getRulesComplexity();
				int maxStepsTillDeath = 0;
				for (final CommandableTemplate template : CommandableTemplate.allUnitsCitiesCanProduce(rulesComplexity)) {
					if (branch.equals(template.militaryBranch)) {
						final int stepsTillDeath = template.getRulesComplexityDependingValues(rulesComplexity).stepsTillDeath;
						maxStepsTillDeath = Math.max(maxStepsTillDeath, stepsTillDeath);
					}
				}
				if (0 < maxStepsTillDeath) {
					if (getPlayerMap().getProperties().calcAirDistance(path.getStartPos(), path.getEndPos()) <= maxStepsTillDeath) {
						final FieldAlike field = getPlayerMap().getAt(path.getEndPos());
						final Commandable unit = findCityOrAirbase(field);
						if ((null != unit) && (path.getPlayerId() == unit.getPlayerId())) {
							ret = true;
						}
					}
				}
				break;
			case NO_UNIT:
				break;
			default:
				break;
			}
			if (null != pivot) {
				final OrderGoTo order = new OrderGoTo(getGameState(), pivot, path.getEndPos());
				order.setStepRatingPurpose(StepRatingPurpose.PATH);
				if (order.makesSense(pivot)) {
					ret = true;
				}
			}
		}
		return ret;
	}

	public Commandable findCityOrAirbase(final FieldAlike field) {
		Commandable ret = null;
		if (field.isCity()) {
			ret = (Commandable) field.getField();
		} else {
			final Collection<UnitAlike> units = findUnitsAtPosition(field.getPosition());
			if (null != units) {
				for (UnitAlike unit : units) {
					if (Bitfield32.isBit(MapViewMode.PATH.bitmaskUnitsToShow, unit.getUnitTemplate().id)) {
						ret = unit;
						break;
					}
				}
			}
		}
		return ret;
	}

	/**
	 * Lets the player resign. This removes all his/her units and sets all his/her cities to neutral.
	 */
	public void resign() {
		gameState.resign(getPlayer());
		nextCommandable();
	}

	/**
	 * @return the message type of the report to show in the player ready dialog
	 */
	public PlayerGetReadyMessage getPlayerReadyMessage() {
		PlayerGetReadyMessage ret = null; 
		if (null == ret) {
			if (gameState.isWinner(getPlayer())) {
				ret = PlayerGetReadyMessage.GameWon;
			}
		}
		if (null == ret) {
			for (final Animation anim : getAnimation(AnimationSequenceType.ENEMIES)) {
				final MoveResult moveResult = anim.getMoveResult();
				if ((null != moveResult)
						&& (moveResult.getCommandableInfo(Role.Affected).getPlayerId() == getPlayer().getId())
						&& (0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_FIGHTING))) {
					ret = PlayerGetReadyMessage.FightReports;
					if ((0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_WINNING_FIGHT))
							&& (0 != (moveResult.getResultMask() & MoveResult.MOVE_MASK_LOSING_GAME))) {
						ret = PlayerGetReadyMessage.GameLost;
						break;
					}
				}
			}
		}
		if (null == ret) {
			for (final Commandable unit : gameState.findCommandables(getPlayer().getId())) {
				if (unit.hasSeenEnemiesLately()) {
					ret = PlayerGetReadyMessage.SightReports;
					break;
				}
			}
		}
		if (null == ret) {
			if (gameState.getTurn() <= 1) {
				ret = PlayerGetReadyMessage.FirstTurn;
			}
		}
		if (null == ret) {
			ret = PlayerGetReadyMessage.NoReports;
		}
		return ret;
	}

	/**
	 * Returns if this position should be treated as home base (and routed around) when planning the route for a move.
	 * @param unit the unit to plan the route for
	 * @param dpos the position to look at
	 * @return <code>true</code> if to route around
	 */
	public boolean isHomeBase(Unit unit, PlayfieldPosition dpos) {
		boolean ret = false;
		final RulesComplexity rulesComplexity = getProperties().getRulesComplexity();
		if (Integer.MAX_VALUE != unit.getUnitTemplate().getStepsTillDeath(rulesComplexity)) {
			final FieldAlike field = getPlayerMap().getAt(dpos);
			if (field.isCity()) {
				if (field.getPlayerId() == unit.getPlayerId()) {
					ret = true;
				}
			} else {
				final Collection<UnitAlike> unitsAtPosition = findUnitsAtPosition(dpos);
				if (null != unitsAtPosition) {
					for (final UnitAlike obstacle : unitsAtPosition) {
						if (obstacle.canLoad(unit, false)) {
							ret = true;
						}
					}
				}
			}
		}
		return ret;
	}

	/**
	 * Updates the view map based on the views of all commandables of the player.
	 */
	public void updateViewMap() {
		updateViewMap(calcVisibleFields(), null);
	}

	/**
	 * Returns all the fields (of the global map) that this player can see right now.
	 * Needed to define if an enemy move is visible for this player.
	 * @return a collection of all the fields (of the global map) that this player can see right now,
	 * 	or <code>null</code> to indicate that in a game without fog of war everything is visible.
	 */
	private Collection<FieldAlike> calcVisibleFields() {
		final Set<FieldAlike> ret;
		if (isUsingFogOfWar()) {
			ret = new HashSet<FieldAlike>();
			for (final Commandable commandable : getGameState().findCommandables(getPlayer().getId())) {
				ret.addAll(calcVisibleFields(commandable));
			}
		} else { // In non-fog-of-war games, all fields are visible
			ret = null;
		}
		return ret;
	}

	/**
	 * Updates the view map based on the view of the given commandable of the player.
	 * Normally called after a unit has moved.
	 * @param commandable the commandable to update the map for
	 * @param moveResult the move result of the move why the update happens, or <code>null</code> at the start of a turn
	 */
	void updateViewMap(final Commandable commandable, final MoveResult moveResult) {
		updateViewMap(calcVisibleFields(commandable), moveResult);
	}

	/**
	 * @param commandable the commandable to update the map for
	 * @return all the fields (of the global map) that the given unit can see right now
	 */
	private Collection<FieldAlike> calcVisibleFields(final Commandable commandable) {
		final GameMap gameMap = gameState.getMap();
		final int sightingRangeMax = Math.max(
				commandable.getSightingRange(MilitaryBranch.NO_UNIT), 
				commandable.getSightingRange(MilitaryBranch.AIR_UNIT));
		return gameMap.getAllFieldsInAirDistance(
				gameMap.getAt(commandable.getPosition()), sightingRangeMax, true, 0);
	}

	/**
	 * Updates the view map for the given fields.
	 * @param fields a collection of all the fields (of the global map) that this player can see right now,
	 * 	or <code>null</code> to indicate that in a game without fog of war everything is visible.
	 * @param moveResult the move result of the move why the update happens, or <code>null</code> at the start of a turn
	 */
	private void updateViewMap(final Collection<FieldAlike> fields, final MoveResult moveResult) {
		if (null == fields) {
			final GameProperties props = getProperties();
			final int width = map.getSize().getWidth();
			for (int x = 0; x < width; x++) {
				for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
					updateViewMap(map.getAt(x, y).getPosition(), moveResult);
				}
			}
		} else {
			for (final FieldAlike field : fields) {
				updateViewMap(field.getPosition(), moveResult);
			}
		}
	}

	/**
	 * Updates the view map for a given position.
	 * @param pos the position to update
	 * @param moveResult the move result of the move why the update happens, or <code>null</code> at the start of a turn
	 */
	private void updateViewMap(final PlayfieldPosition pos, final MoveResult moveResult) {
		// Update the simple map
		if (isVisible(null, pos, moveResult)) {
			final ViewField viewField = (ViewField) map.getAt(pos);
			if (viewField.isUnexplored()) {
				viewField.setUnexplored(false);
				updateUnexploredBorders(map.getAt(pos));
				// Have I just discovered a new city? Inform near land units about it.
				if (viewField.isCity()) {
					for (final UnitAlike observingUnit : getUnits()) {
						if ((observingUnit.getPlayerId() == getPlayer().getId())
								&& (MilitaryBranch.LAND_UNIT == observingUnit.getUnitTemplate().militaryBranch)
								&& (map.getProperties().calcAirDistance(observingUnit.getPosition(), pos) <= observingUnit.getSightingRange(MilitaryBranch.NO_UNIT))) {
							observingUnit.setSeenEnemiesLately(true);
						}
					}
				}
			} else if (viewField.isCity()) {
				// TODO: Make image key calculation for city event driven, not automatically calculated
				viewField.setUnexplored(false);
				updateUnexploredBorders(viewField);
			} else {
				// Nothing to do
			}
			int viewKey = viewField.getImageKey();
			viewKey = Bitfield32.setBits(viewKey, ImageKey.BA_OUTDATED, 0);
			if (viewField.isCity()) {
				final FieldCity city = (FieldCity) viewField.getField(); 
				int cityKey = city.getImageKey();
				viewKey = ImageKey.setBits(viewKey, ImageKey.BA_PLAYER_COLOR, ImageKey.getBits(cityKey, ImageKey.BA_PLAYER_COLOR));
				viewKey = ImageKey.setBits(viewKey, ImageKey.BA_SENTRIED, 0);
				final int isNonEmpty;
				if (map.isOtherView(city)) {
					isNonEmpty = 0;
				} else {
					isNonEmpty = ImageKey.getBits(cityKey, ImageKey.BA_NONEMPTY);
					logger.trace("updateViewMap(" + pos.toString() + "," + moveResult + "): city " + city + " isNonEmpty=" + isNonEmpty);
				}
				viewKey = ImageKey.setBits(viewKey, ImageKey.BA_NONEMPTY, isNonEmpty);
				logger.trace("updateViewMap(" + pos.toString() + "," + moveResult + "): city " + city + " viewKey=" + Integer.toHexString(viewKey));
			}
			viewField.setImageKey(viewKey);
		}
		// Update units on that field. 
		// Remark: We need to operate on the original data, on copies won't work.
		final Collection<UnitAlike> oldUnits = findUnitsAtPosition(pos);
		final Collection<UnitAlike> newUnits = gameState.findUnitsAtPosition(pos);
		if (null != oldUnits) {
			for (final UnitAlike oldUnit : oldUnits) {
				final UnitAlike newUnit = findUnitForKey(newUnits, oldUnit);
				if (null == newUnit) {
					// Remove units which left the field
					final boolean isToUpdate = isVisible(oldUnit, pos, moveResult);
					if (isToUpdate) {
						map.removeUnit(oldUnit);
					}
				} else {
					if (oldUnit.getPlayerId() == getPlayer().getId()) {
						// Update own units still on the field
						oldUnit.setImageKey(newUnit.getImageKey());
					} else {
						// Outdate other units still on the field
						final ImageKey key = new ImageKey(oldUnit.getImageKey());
						key.setBits(ImageKey.BA_OUTDATED, 0x00);
						oldUnit.setImageKey(key.getAll());
					}
				}
			}
		}
		if (null != newUnits) {
			for (final UnitAlike newUnit : newUnits) {
				if (null == findUnitForKey(oldUnits, newUnit)) {
					// Add units which came onto the field
					final boolean isToUpdate = isVisible(newUnit, pos, moveResult);
					if (isToUpdate) {
						map.addUnit(new ViewUnit((Unit) newUnit, gameState.getTurn(), map.isOtherView(newUnit)));
						// Have I just discovered a new unit? Inform near units about it.
						if (newUnit.getPlayerId() != getPlayer().getId()) {
							for (final UnitAlike observingUnit : getUnits()) {
								if ((observingUnit.getPlayerId() == getPlayer().getId())
										&& (map.getProperties().calcAirDistance(observingUnit.getPosition(), pos) 
												<= observingUnit.getSightingRange(MilitaryBranch.NO_UNIT))) {
									observingUnit.setSeenEnemiesLately(true);
								}
							}
						}
					}
				}
			}
		}
	}

	private UnitAlike findUnitForKey(Collection<UnitAlike> units, UnitAlike unitToFind) {
		UnitAlike ret = null;
		if (null != units) {
			for (UnitAlike unit : units) {
				if (unit.getUnit().getKey().equals(unitToFind.getUnit().getKey())) {
					ret = unit;
					break;
				}
			}
		}
		return ret;
	}

	public Order actionTargetSelectionAtPosition(
			final GameView gameView, 
			final OrderTemplate targetSelection, 
			final PlayfieldPosition pos) 
			throws CircularReferenceException {
    	final Messages messages = getMessages();
		final Commandable commandable = gameView.getVisibleCommandableInFocus();
		Order order = null;
		Unit unit = null;
		switch(targetSelection.id) {
		case OrderTemplate.TYPE_GOTO:
			if ((null != commandable) && (commandable instanceof Unit)) {
				unit = (Unit) commandable;
				order = gameView.getMoveOrAttackOrder(unit, pos);
				unit.calculateForwardRating(StepRatingPurpose.NORMAL);
				logger.debug("actionAt: MoveOrAttack");
			}
			break;
		case OrderTemplate.TYPE_ESCORT:
			if ((null != commandable) && (commandable instanceof Unit)) {
				unit = (Unit) commandable;
				final LinkedList<UnitAlike> targets = new LinkedList<UnitAlike>();
				{
					final Collection<UnitAlike> allTargets = gameView.findUnitsAtPosition(pos);
					// Only ships may be escorted
					if (null != allTargets) {
					for (final UnitAlike target : allTargets) {
						if (MilitaryBranch.SEA_UNIT == target.getUnitTemplate().militaryBranch) {
							targets.add(target);
							}
						}
					}
				}
				if (!targets.isEmpty()) {
					// TODO: Show list of possible targets
					try {
						order = new OrderEscort(gameState, unit, targets.getFirst().getKey());
						unit.calculateForwardRating(StepRatingPurpose.NORMAL);
						logger.debug("actionAt: Escort target " + messages.getString(targets.getFirst().getName()));
					} catch (final CircularReferenceException ex) {
						logger.debug("actionAt: Escort target " + messages.getString(targets.getFirst().getName()) + " lead to circular reference");
						throw ex;
					}
				}
			}
			break;
		case OrderTemplate.TYPE_CHANGE_MOVE_PATHS:
			logger.debug("actionAt: newMovePath");
			final int playerId = commandable.getPlayerId();
			final PlayfieldPosition startPos = commandable.getPosition();
			try {
				gameView.getGameState().getMovePaths().add(new MovePath(playerId, startPos, 0x0, pos));
			} catch (final CircularReferenceException ex) {
				logger.warn(ex.getMessage(), ex);
			}
			order = new OrderDefineMovePaths(gameState, commandable);
			break;
		default:
			order = null;
		}
		return order;
	}

	/**
	 * Checks if a {@link CommandableInfo} of a {@link MoveResult} should be shown or updated in fog of war.
	 * @param commandableInfo the commandable info
	 * @return <code>true</code> if the commandable is to be updated
	 */
	boolean isVisible(final CommandableInfo commandableInfo) {
		boolean ret = ((null != commandableInfo) 
				&& ((null != commandableInfo.getUnit())
						&& (getPlayer().getId() == commandableInfo.getPlayerId())
						|| ((0 != (commandableInfo.moveResult.getResultMask() & MoveResult.MOVE_MASK_FIGHTING))
								))
				|| isVisible(commandableInfo.getUnit(), commandableInfo.position, null));
		return ret;
	}

	/**
	 * Checks if a unit or field to be shown in fog of war should be updated.
	 * 
	 * @param unit
	 *            the unit to show. May be <code>null</code>, e.g., if it's a city.
	 * @param pos
	 *            the position of the unit (or field) to show. May be different
	 *            than current unit position for animations.
	 * @param moveResult
	 *            the move result of the move why the update happens, or
	 *            <code>null</code> at the start of a turn
	 * @return <code>true</code> if the unit or field is to be updated
	 */
	boolean isVisible(final UnitAlike unit, final PlayfieldPosition pos, final MoveResult moveResult) {
		boolean isToUpdate;
		if (isUsingFogOfWar()) {
			isToUpdate = false;
			final int playerId = player.getId();
			// Within a move, both commandables involved in the move are able to see each other.
			if ((null != moveResult) && ((moveResult.isWith(unit) || moveResult.isWith(pos)))) {
				for (final Role role : Role.values()) {
					final CommandableInfo cmdli = moveResult.getCommandableInfo(role);
					if ((null != cmdli) && (cmdli.getPlayerId() == playerId)) {
						isToUpdate = true;
					}
				}
			} 
			if (!isToUpdate) {
				// A unit is visible for any observing commandable in range.
				for (final Commandable observer : gameState.getMap().findAllCommandables()) {
					if ((observer.getPlayerId() == playerId)) {
						if (null == unit) {
							// It's a field, not a unit
							if (gameState.getMap().getProperties().calcAirDistance(pos, observer.getPosition()) 
										<= observer.getSightingRange(MilitaryBranch.NO_UNIT)) {
								isToUpdate = true;
							}
						} else if (unit.getPlayerId() == playerId) {
							// It's this player's unit
							isToUpdate = true;
						} else if (unit.getUnit().isVisibleFor(pos, observer, moveResult)) {
							// Special visibility rules for enemy submarines - can only be seen by cities, destroyers, cruisers and submarines.
							isToUpdate = true;
						}
						if (isToUpdate) {
							break;
						}
					}
				}
			}
		} else { // not isUsingFogOfWar
			isToUpdate = true;
		}
		return isToUpdate;
	}

	/**
	 * Must be called when an previously unexplored field gets explored.
	 * @param at the field that was hidden behind the unexplored field
	 */
	private void updateUnexploredBorders(FieldAlike at) {
		// Look up, down, left, right if unexplored. If yes, add here unexplored borders.
		final int ubBits = map.addKeyBitsForUnexploredBorder(at.getPosition());
		final ImageKey imageKey = new ImageKey(at.getImageKey());
		imageKey.setBits(ImageKey.BA_UNEXPLORED_MOD, ubBits);
		at.setImageKey(imageKey.getAll());
		// In up, down, left and right field, remove adjacenting unexplored border.
		final PlayfieldPosition pos = at.getPosition();
		final ArrayList<DigitalJoystickDirection> ndirs = getProperties().getShoreNeighborDirections(pos);
		final int[] nbits = map.getBitsForShoreNeighborDirections();
		final GameProperties props = getProperties();
		final PlayfieldPosition npos = new PlayfieldPosition(); // re-used for performance reasons
		for (int i = 0; i < ndirs.size(); i++) {
			final int nbit = (1 << nbits[i]);
			props.calcSourcePositionOfMoveStep(npos, ndirs.get(i), pos);
			props.wrapInside(npos);
			final FieldAlike field = map.getAt(npos);
			final ImageKey bits = new ImageKey(field.getImageKey());
			final int old = bits.getBits(ImageKey.BA_UNEXPLORED_MOD);
			final int now = old & ~nbit;
			bits.setBits(ImageKey.BA_UNEXPLORED_MOD, now);
			field.setImageKey(bits.getAll());
		}
	}

	/** The map visible for this player. Persistent. */
	private GameMap map;

	/**
	 * Returns the perspective to show for this player.
	 * @return 0 for world view, 1 to 6 for player view
	 */
	public int getPerspective() {
		return viewPerspective;
	}

	/**
	 * Sets the perspective to show for this player.
	 * @param perspective 0 for world view, 1 to 6 for player view
	 */
	public void setPerspective(int perspective) {
		viewPerspective = perspective;
	}

	/**
	 * Returns the default perspective to show for this player.
	 * @return 0 for world view, 1 to 6 for player view
	 */
	public int getDefaultPerspective() {
		int ret = getPlayer().getId() + 1;
		return ret;
	}

	public boolean isDefaultPerspective() {
		return (getDefaultPerspective() == viewPerspective);
	}

	public synchronized boolean isOrderActionEnabled(final Order order) {
		boolean result = false;
		switch(getMapViewMode()) {
		case NORMAL: {
				if (OrderTemplate.TYPE_CHANGE_MOVE_PATHS != order.template.id) {
					final Commandable commandable = getVisibleCommandableInFocus();
					if (null != commandable) {
						result = order.makesSense(commandable);
					}
				}
			}
			break;
		case PRODUCTION: {
				if (OrderTemplate.TYPE_CHANGE_MOVE_PATHS != order.template.id) {
					final FieldAlike field = getAt(getPositionInFocus());
					if (field.isCity()) {
						if (field.getPlayerId() == getPlayer().getId()) {
							result = order.makesSense((Commandable) field.getField());
						}
					}
				}
			}
			break;
		case PATH: {
				final FieldAlike field = getAt(getPositionInFocus());
				final Commandable commandable = findCityOrAirbase(field);
				if ((null != commandable) && (getPlayer().getId() == commandable.getPlayerId())) {
					result = order.makesSense(commandable);
				}
			}
			break;
		default:
			// Nothing to do
		}
		return result;
	}
}
