package de.ped.empire.logic;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;

/**
 * The properties of a human player, how and what he/she gets presented of the game state.
 * This consists things like language, zoom level, sounds and optional reports.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameViewProperties implements Marshallable, Cloneable {

	/**
	 * Defines the different animations that can be viewed in different speed levels.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum AnimationDelayType {
		MESSAGES("GameViewProperties.AnimationDelay.Messages", null),
		MOVEMENT("GameViewProperties.AnimationDelay.Movement", "Action.GoTo"),
		BATTLE  ("GameViewProperties.AnimationDelay.Battle", null);

		public final String messageKey;
		public final String imageKey;

		private AnimationDelayType(final String messageKey, final String imageKey) {
			this.messageKey = messageKey;
			this.imageKey = imageKey;
		}
	}

	/**
	 * Defines the different speed levels the animations that can be viewed in.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum AnimationDelayValue {
		SLOW  ("GameViewProperties.AnimationDelay.0.Text"),
		MEDIUM("GameViewProperties.AnimationDelay.1.Text"),
		FAST  ("GameViewProperties.AnimationDelay.2.Text");

		public final String messageKey;

		private AnimationDelayValue(final String messageKey) {
			this.messageKey = messageKey;
		}
	}

	public GameViewProperties() {
		this.zoomLevel = ZoomLevel.DEFAULT_ZOOM_LEVEL_INDEX;
		this.localeIndex = 0;
		isShowingProductionReports[MilitaryBranch.SEA_UNIT.ordinal()] = true;
		for (int i = 0; i < isPlayingSounds.length; i++) {
			isPlayingSounds[i] = true;
		}
		for (int i = 0; i < AnimationDelayType.values().length; i++) {
			animationDelays[i] = AnimationDelayValue.SLOW;
		}
		isShowingDistanceInfo = false;
	}

	public GameViewProperties(final GameViewProperties other) {
		copyFrom(other);
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(zoomLevel);
		marshaller.writeInt(localeIndex);
		marshaller.writeFixedLengthBooleanArray(isShowingProductionReports);
		marshaller.writeFixedLengthBooleanArray(isPlayingSounds);
		for (final AnimationDelayType type : AnimationDelayType.values()) {
			marshaller.writeByte((byte) getAnimationDelay(type).ordinal());
		}
		marshaller.writeBoolean(isShowingDistanceInfo);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		zoomLevel = marshaller.readInt();
		localeIndex = marshaller.readInt();
		isShowingProductionReports = marshaller.readFixedLengthBooleanArray(MilitaryBranch.REAL_BRANCHES.length);
		isPlayingSounds = marshaller.readFixedLengthBooleanArray(SoundTypes.values().length);
		for (final AnimationDelayType type : AnimationDelayType.values()) {
			setAnimationDelay(type, AnimationDelayValue.values()[marshaller.readByte()]);
		}
		isShowingDistanceInfo = marshaller.readBoolean();
		return this;
	}

	public void copyFrom(final GameViewProperties from) {
		this.localeIndex = from.localeIndex;
		this.zoomLevel = from.zoomLevel;
		this.isPlayingSounds = from.isPlayingSounds.clone();
		this.isShowingProductionReports = from.isShowingProductionReports.clone();
		this.animationDelays = from.animationDelays.clone();
		this.isShowingDistanceInfo = from.isShowingDistanceInfo;
	}

	@Override
	public GameViewProperties clone() {
		final GameViewProperties other = new GameViewProperties(this);
		return other;
	}

	public int getZoomLevel() {
		return zoomLevel;
	}

	public void setZoomLevel(final int level) {
		zoomLevel = level;
	}

	public int getLocaleIndex() {
		return localeIndex;
	}

	public void setLocaleIndex(final int index) {
		localeIndex = index;
	}

	public AnimationDelayValue getAnimationDelay(final AnimationDelayType what) {
		return animationDelays[what.ordinal()];
	}

	public void setAnimationDelay(final AnimationDelayType what, final AnimationDelayValue value) {
		animationDelays[what.ordinal()] = value;
	}

	public boolean isShowingProductionReports(final int i) {
		return isShowingProductionReports[i];
	}

	public void setShowingProductionReports(final int i, final boolean value) {
		isShowingProductionReports[i] = value;
	}

	public boolean isPlayingSounds(final int i) {
		return isPlayingSounds[i];
	}

	public void setPlayingSounds(final int i, final boolean value) {
		isPlayingSounds[i] = value;
	}

	public boolean isShowingDistanceInfo() {
		return isShowingDistanceInfo;
	}

	public void setShowingDistanceInfo(final boolean value) {
		isShowingDistanceInfo = value;
	}

	/** The preferred zoom level of the player. Persistent. */
	private int zoomLevel;
	/** The preferred locale of the player. Persistent. */
	private int localeIndex;
	/** If a dialog should pop up whenever a city produced a new unit. Persistent. */
	private boolean[] isShowingProductionReports = new boolean[MilitaryBranch.REAL_BRANCHES.length];
	/** If sounds should be played. Persistent. */
	private boolean[] isPlayingSounds = new boolean[SoundTypes.values().length];
	/** How fast different animations are displayed. Persistent. */
	private AnimationDelayValue [] animationDelays = new AnimationDelayValue[AnimationDelayType.values().length];
	/** If to display distance information on map. Persistent. */
	private boolean isShowingDistanceInfo;
}
