package de.ped.empire.logic;

import de.ped.tools.BitArea;
import de.ped.tools.Bitfield32;

/**
 * Identifies uniquely exactly which icon in which flavor to draw.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImageKey extends Bitfield32 {
	/** Contains the player color ID, including the ID for no player. Only used for commandables. */
	public static final BitArea BA_PLAYER_COLOR = new BitArea( 0, 4);
	/** Contains the modifications of the look based on the neighboring fields. Only used for fields. */
	public static final BitArea BA_NEIGHBOR_MOD = new BitArea( 4, 6);
	/** Contains the modifications of the look based on neighboring unexplored fields. Only used for fields. */
	public static final BitArea BA_UNEXPLORED_MOD = new BitArea(10, 6);
	/** Contains the flavor of the look (many fields exist in two flavors to make the map more interesting). Only used for fields and cities. */
	public static final BitArea BA_LOOK = new BitArea(16, 1);
	/** Contains the info if the unit is to draw non-empty. Only used for units and cities. */
	public static final BitArea BA_NONEMPTY = new BitArea(17, 1);
	/** Contains the info if the unit is to draw sentried. Only used for units. */
	public static final BitArea BA_SENTRIED = new BitArea(18, 1);
	/** Contains the info if the image is to draw outdated. */
	public static final BitArea BA_OUTDATED = new BitArea(19, 1);
	/** Contains the info which template ID (depending on the ttype) to apply. */
	public static final BitArea BA_TEMPLATE = new BitArea(20, 4);
	/** Contains the template type. */
	public static final BitArea BA_TTYPE = new BitArea(24, 2);

	/**
	 * Identifies what kind of icon to draw. The ordinals of this enum must fit into {@link ImageKey#BA_TTYPE}.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum TType {
		/** Neither field nor unit */
		Overlay,
		/** A field on the map that cannot be owned */
		Field,
		/** A unit that may move around */
		Unit,
		/** A city or airbase. Currently unused. */
		Infra,
	}

	public static final int OVERLAY_ID_CROSSHAIR = 0x0;
	public static final int OVERLAY_ID_FIGHT = 0x1;
	public static final int OVERLAY_ID_UNEXPLORED_NEIGHBOR = 0x2; // Not used in game but in field templates creation
	public static final int OVERLAY_ID_OUTDATED = 0x3; // Not used in game but in field templates creation
	public static final byte FIELD_ID_UNEXPLORED = (byte) 0x0;
	public static final byte FIELD_ID_BORDER = (byte) 0x1;
	public static final byte FIELD_ID_SEA = (byte) 0x2;
	public static final byte FIELD_ID_LAND = (byte) 0x3;
	public static final byte FIELD_ID_ROUGH = (byte) 0x4;
	public static final byte FIELD_ID_RIVER = (byte) 0x5;
	public static final byte FIELD_ID_FOREST = (byte) 0x6;
	public static final byte FIELD_ID_MOUNTAIN = (byte) 0x7;
	public static final byte FIELD_ID_CITY = (byte) 0x8;
	public static final int CMDL_ID_UNDEFINED = 0xf;
	public static final int CMDL_ID_CITY = 0x0;
	public static final int CMDL_ID_ARMY = 0x1;
	public static final int CMDL_ID_TANK = 0x2;
	public static final int CMDL_ID_FIGHTER = 0x3;
	public static final int CMDL_ID_BOMBER = 0x4;
	public static final int CMDL_ID_TRANSPORTER = 0x5;
	public static final int CMDL_ID_DESTROYER = 0x6;
	public static final int CMDL_ID_SUBMARINE = 0x7;
	public static final int CMDL_ID_CRUISER = 0x8;
	public static final int CMDL_ID_BATTLESHIP = 0x9;
	public static final int CMDL_ID_CARRIER = 0xa;
	public static final int CMDL_ID_AIRBASE = 0xb;

	public static final ImageKey CROSSHAIR = new ImageKey(TType.Overlay, OVERLAY_ID_CROSSHAIR);
	public static final ImageKey FIGHT = new ImageKey(TType.Overlay, OVERLAY_ID_FIGHT);
	public static final ImageKey UNEXPLORED_NEIGHBOR = new ImageKey(TType.Overlay, OVERLAY_ID_UNEXPLORED_NEIGHBOR);
	public static final ImageKey OUTDATED = new ImageKey(TType.Overlay, OVERLAY_ID_OUTDATED);
	public static final ImageKey UNEXPLORED = new ImageKey(TType.Field, FIELD_ID_UNEXPLORED);
	public static final ImageKey BORDER = new ImageKey(TType.Field, FIELD_ID_BORDER);
	public static final ImageKey SEA = new ImageKey(TType.Field, FIELD_ID_SEA);
	public static final ImageKey LAND = new ImageKey(TType.Field, FIELD_ID_LAND);
	public static final ImageKey ROUGH = new ImageKey(TType.Field, FIELD_ID_ROUGH);
	public static final ImageKey RIVER = new ImageKey(TType.Field, FIELD_ID_RIVER);
	public static final ImageKey FOREST = new ImageKey(TType.Field, FIELD_ID_FOREST);
	public static final ImageKey MOUNTAIN = new ImageKey(TType.Field, FIELD_ID_MOUNTAIN);
	public static final ImageKey CITY = new ImageKey(TType.Field, FIELD_ID_CITY);
	public static final ImageKey ARMY = new ImageKey(TType.Unit, CMDL_ID_ARMY);
	public static final ImageKey TANK = new ImageKey(TType.Unit, CMDL_ID_TANK);
	public static final ImageKey FIGHTER = new ImageKey(TType.Unit, CMDL_ID_FIGHTER);
	public static final ImageKey BOMBER = new ImageKey(TType.Unit, CMDL_ID_BOMBER);
	public static final ImageKey TRANSPORTER = new ImageKey(TType.Unit, CMDL_ID_TRANSPORTER);
	public static final ImageKey DESTROYER = new ImageKey(TType.Unit, CMDL_ID_DESTROYER);
	public static final ImageKey SUBMARINE = new ImageKey(TType.Unit, CMDL_ID_SUBMARINE);
	public static final ImageKey CRUISER = new ImageKey(TType.Unit, CMDL_ID_CRUISER);
	public static final ImageKey BATTLESHIP = new ImageKey(TType.Unit, CMDL_ID_BATTLESHIP);
	public static final ImageKey CARRIER = new ImageKey(TType.Unit, CMDL_ID_CARRIER);
	public static final ImageKey AIRBASE = new ImageKey(TType.Unit, CMDL_ID_AIRBASE);
	public static final ImageKey UNDEFINED = new ImageKey(TType.Unit, CMDL_ID_UNDEFINED);

	/**
	 * Initializes this image key with the full given bitset.
	 * @param key the bitset to use
	 */
	public ImageKey(final int key) {
		super(key);
	}

	/**
	 * Initializes the image key with the right ttype and template (and otherwise empty).
	 * @param ttype the ttype to use
	 * @param template the template to use
	 */
	public ImageKey(final TType ttype, final int template) {
		super(0);
		setBits(BA_TTYPE, ttype.ordinal());
		setBits(BA_TEMPLATE, template);
	}

	static Player getPlayer(final GameState gameState, final int imageKey) {
		// Read the player back from the image key
		final int playerColorId = ImageKey.getBits(imageKey, BA_PLAYER_COLOR);
		Player player = Player.PLAYER_NONE;
		for (final Player p : gameState.getProperties().getPlayers()) {
			if (p.getPlayerColorId() == playerColorId) {
				player = p;
				break;
			}
		}
		return player;
	}

	public static String toString(final int imageKey) {
		return "0x" + Integer.toString(imageKey, 16)
		+ "(ttype=" + Bitfield32.getBits(imageKey, BA_TTYPE)
		+ ",template=" + Bitfield32.getBits(imageKey, BA_TEMPLATE)
		+ ",look=" + Bitfield32.getBits(imageKey, BA_LOOK)
		+ ",outdated=" + Bitfield32.getBits(imageKey, BA_OUTDATED)
		+ ",nonempty=" + Bitfield32.getBits(imageKey, BA_NONEMPTY)
		+ ",sentried=" + Bitfield32.getBits(imageKey, BA_SENTRIED)
		+ ",neighbors=" + Bitfield32.getBits(imageKey, BA_NEIGHBOR_MOD)
		+ ",unexplored=" + Bitfield32.getBits(imageKey, BA_UNEXPLORED_MOD)
		+ ",playercolor=" + Bitfield32.getBits(imageKey, BA_PLAYER_COLOR)
		+ ")";
	}

	@Override
	public String toString() {
		return toString(getAll());
	}
}