package de.ped.empire.logic;

import java.awt.Color;
import java.awt.Point;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.HashMap;

import de.ped.tools.BitArea;
import de.ped.tools.Bitfield32;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Logger;

/**
 * A set of icons belonging together in size and shape to be shown on the same map.
 * 
 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ImageSet {
	/**
	 * A unique key for identifying an {@link ImageSet}.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Key {
		public static final BitArea zoomLevelArea = new BitArea(0, 3);
		public static final BitArea topologyModeArea = new BitArea(3, 1);

		public static final int VALUE_SQUARE = 0;
		public static final int VALUE_HEXAGON = 1;

		public static int calcKey(final int zoomLevelIdx, final TopologyMode topologyMode) {
			int ret = 0;
			ret |= Bitfield32.setBits(ret, zoomLevelArea, zoomLevelIdx);
			ret |= Bitfield32.setBits(ret, topologyModeArea, getTopologyIndex(topologyMode));
			return ret;
		}

		public static int getTopologyIndex(final TopologyMode topologyMode) {
			return (TopologyMode.HEXAGON == topologyMode ? VALUE_HEXAGON : VALUE_SQUARE);
		}

		public Key(final int zoomLevelIdx, final TopologyMode topologyMode) {
			this(calcKey(zoomLevelIdx, topologyMode));
		}

		public Key(final int key) {
			this.key = key;
		}

		public int getZoomLevelIndex() {
			return Bitfield32.getBits(key, zoomLevelArea);
		}

		/** @return 0 for square, 1 for hexagon */
		public int getTopologyIndex() {
			return Bitfield32.getBits(key, topologyModeArea);
		}

		@Override
		public int hashCode() {
			return key;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			Key other = (Key) obj;
			if (key != other.key) {
				return false;
			}
			return true;
		}

		public final int key;
	}

	public ImageSet(final Key key, final String resourceName) {
		this.key = key;
		this.resourceName = resourceName;
	}

	/** 
	 * @return an empty (transparent) image of the resolution of this instance
	 */
	public BufferedImage getTransparentImage() {
		if (null == emptyIcon) {
			final TopologyMode tm = ((Key.VALUE_SQUARE == key.getTopologyIndex()) ? TopologyMode.SQUARE8 : TopologyMode.HEXAGON);
			final PlayfieldTiling playfieldTiling = getZoomLevel().getPlayfieldTiling(tm);
			BufferedImage img = new BufferedImage(playfieldTiling.getSurroundingWidth(), playfieldTiling.getSurroundingHeight(), BufferedImage.TYPE_INT_ARGB);
			for (int x = 0; x < img.getWidth(); x++) {
				for (int y = 0; y < img.getHeight(); y++) {
					img.setRGB(x, y, 0x00000000);
				}
			}
			emptyIcon = img;
		}
		return emptyIcon;
	}

	/**
	 * Snips a tile (sub-image) from the templates image.
	 * 
	 * @param tileX
	 *            horizontal position of the tile
	 * @param tileY
	 *            vertical position of the tile
	 * @return the sub-image
	 */
	BufferedImage getTemplateImageTile(final int tileX, final int tileY) {
		final BufferedImage templates = loadTemplatesImage();
		BufferedImage ret = null;
		if (null != templates) {
			final ZoomLevel zoomLevel = getZoomLevel();
			final TopologyMode tm = ((Key.VALUE_SQUARE == key.getTopologyIndex()) ? TopologyMode.SQUARE8 : TopologyMode.HEXAGON);
			final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(tm);
			final Point d = new Point();
			playfieldTiling.map2canvas(d, new PlayfieldPosition(tileX, tileY), ORIGIN, PlayfieldTiling.TOP_LEFT);
			translateToImageSet(d, tm, zoomLevel);
			logger.trace("getTemplateImageTile(" + tileX + "," + tileY + ") on zoom " + zoomLevel.resolution + " --> " + d);
			ret = templates.getSubimage(
					d.x, 
					d.y,
					playfieldTiling.getSurroundingWidth(), 
					playfieldTiling.getSurroundingHeight());
			if (TopologyMode.HEXAGON == playfieldTiling.getShapeTopology()) {
				// Clone the sub-image, as the sub-image was read-only
				BufferedImage clipped = new BufferedImage(ret.getWidth(), ret.getHeight(), BufferedImage.TYPE_INT_ARGB);
				for (int x = 0; x < ret.getWidth(); x++) {
					for (int y = 0; y < ret.getHeight(); y++) {
						clipped.setRGB(x, y, ret.getRGB(x, y));
					}
				}
				// Make corners totally transparent
				final int c = new Color(0x80, 0x80, 0x80, 0x00).getRGB(); // "Transparent" color
				for (int x = 0; x < ret.getWidth() / 2; x++) {
					final int offset = getYOffset(x);
					if (0 == offset) {
						break;
					}
					for (int y = offset - 1; y >= 0; y--) {
						clipped.setRGB(x, y, c);
						clipped.setRGB(x, ret.getHeight() - 1 - y, c);
						clipped.setRGB(ret.getWidth() - 1 - x, y, c);
						clipped.setRGB(ret.getWidth() - 1 - x, ret.getHeight() - 1 - y, c);
					}
				}
				ret = clipped;
			}
		}
		return ret;
	}

	public static void translateToImageSet(final Point d, final TopologyMode topologyMode, final ZoomLevel zoomLevel) {
		switch(topologyMode) {
		case HEXAGON:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				d.translate(0, zoomLevel.resolution / 2);
			} else {
				d.translate(0, d.x);
			}
			break;
		default:
			// Nothing to do
		}
	}

	/**
	 * A hexagon is represented the following way (here for 16x16)
	 * <pre>
	 * 0 2 | 6 8 |
	 * ....************.... 0
	 * ...**************...
	 * ...**************... 2
	 * ..****************..
	 * ..****************.. 4
	 * .******************.
	 * .******************. 6
	 * ********************
	 * ******************** 8
	 * .******************.
	 * .******************.
	 * ..****************..
	 * ..****************..
	 * ...**************...
	 * ...**************...
	 * ....************....
	 * </pre>
	 * @param x the pixel column of this tile, in surrounding box coordinates
	 * @return the first row for the given column that contains data and is not transparent
	 */
	public int getYOffset(int x) {
		int ret = 0;
		final TopologyMode tm = ((Key.VALUE_SQUARE == key.getTopologyIndex()) ? TopologyMode.SQUARE8 : TopologyMode.HEXAGON);
		final PlayfieldTiling playfieldTiling = getZoomLevel().getPlayfieldTiling(tm);
		if (x < playfieldTiling.getSurroundingWidth() / 2) {
			if ((TopologyMode.HEXAGON == playfieldTiling.getShapeTopology())) {
				ret = Math.max(0, playfieldTiling.getSurroundingHeight() / 2 - 2 * x - 1);
			}
		} else {
			ret = getYOffset(playfieldTiling.getSurroundingWidth() - 1 - x);
		}
		return ret;
	}

	protected final Point ORIGIN = new Point(0, 0);

	/**
	 * Loads and buffers an image from the resources
	 * 
	 * @return the image
	 */
	BufferedImage loadTemplatesImage() {
		BufferedImage readImage = null;
		if ((null == templatesImage) && (null != resourceName)) {
			try {
				readImage = TileTemplate.loadImage(resourceName);
			} catch (final IOException e) {
				logger.error("Unable to read image " + resourceName, e);
			}
		} 
		if (null == readImage) {
			readImage = templatesImage;
		}
		templatesImage = readImage;
		return templatesImage;
	}

	/**
	 * Stores an image in the image buffer for later re-use
	 * 
	 * @param key
	 *            the image key
	 * @param img
	 *            the image
	 */
	void storeImage(final ImageKey key, final BufferedImage img) {
		images.put(key, img);
	}

	/**
	 * Loads an image from the image buffer, if available
	 * 
	 * @param key
	 *            the image key
	 * @return the image
	 */
	BufferedImage getImage(final ImageKey key) {
		return images.get(key);
	}

	public ZoomLevel getZoomLevel() {
		return ZoomLevel.INSTANCES[key.getZoomLevelIndex()];
	}

	public final Key key;
	/** The name of the resource file holding the icons of this zoom level for square maps. */
	public final String resourceName;
	/** The content of the resource file for this zoom level and square maps. */
	private BufferedImage templatesImage;
	/** An empty (transparent) icon of fitting size for this zoom level. Needed e.g. in the city production dialog for the no production row. */
	private BufferedImage emptyIcon;
	/** A store for the pre-calculated icons */
	private HashMap<ImageKey, BufferedImage> images = new HashMap<ImageKey, BufferedImage>();
	
	/** For debugging purposes. */
	private final Logger logger = Logger.getLogger(getClass());
}