package de.ped.empire.logic;

import de.ped.empire.logic.MoveResult.Role;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;

/**
 * A log book entry for a player.
 * If one move of one unit affects several other units (e.g., sightings, or fight), 
 * then for every affected unit or player a log book entry is created.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class LogbookEntry implements Marshallable {
	public final int turn;
	public final int currentPlayerId;
	public final int affectedPlayerId;
	public final PlayfieldPosition pos;
	public final Commandable commandable;
	public final MoveResult moveResult;
	protected final I18NStringWithFillIns entry;

	public LogbookEntry(
			final int turn, 
			final int currentPlayerId, 
			final I18NStringWithFillIns msgkey) {
		this.turn = turn;
		this.currentPlayerId = currentPlayerId;
		this.affectedPlayerId = Player.ID_NONE;
		this.pos = null;
		this.commandable = null;
		this.entry = msgkey;
		this.moveResult = null;
	}

	public LogbookEntry(
			final int turn, 
			final int currentPlayerId, 
			final int affectedPlayerId, 
			final PlayfieldPosition position, 
			final Commandable commandable, 
			final I18NStringWithFillIns msgkey) {
		this.turn = turn;
		this.currentPlayerId = currentPlayerId;
		this.affectedPlayerId = affectedPlayerId;
		this.pos = new PlayfieldPosition(position); // Deep copy
		this.commandable = commandable;
		this.entry = msgkey;
		this.moveResult = null;
	}

	public LogbookEntry(
			final int turn,
			final MoveResult moveResult,
			final Role role,
			final I18NStringWithFillIns entry) {
		this.turn = turn;
		Commandable opponent = null;
		switch(role) {
		case Active:
			this.commandable = moveResult.getCommandable(Role.Active);
			opponent = moveResult.getCommandable(Role.Affected);
			break;
		case Affected:
			this.commandable = moveResult.getCommandable(Role.Affected);
			opponent = moveResult.getCommandable(Role.Active);
			break;
		default:
			throw new IllegalStateException("Unkown role " + role.name());
		}
		this.currentPlayerId = getPlayerId(commandable);
		this.affectedPlayerId = getPlayerId(opponent);
		Commandable cmdlForPos = moveResult.getCommandable(role);
		if (null == cmdlForPos) {
			cmdlForPos = moveResult.getCommandable(Role.Active);
		}
		this.pos = new PlayfieldPosition(moveResult.getCommandable(role).getPosition()); // Deep copy
		this.moveResult = moveResult;
		this.entry = entry;
	}

	private int getPlayerId(final Commandable commandable) {
		int ret;
		if (null == commandable) {
			ret = Player.ID_NONE;
		} else {
			ret = commandable.getPlayerId();
		}
		return ret;
	}

	public LogbookEntry(final GameView gameView, final Marshaller marshaller) {
		this.turn = marshaller.readInt();
		this.currentPlayerId = marshaller.readInt();
		this.affectedPlayerId = marshaller.readInt();
		this.entry = new I18NStringWithFillIns(marshaller);
		this.pos = PlayfieldPosition.demarshalWithNull(marshaller);
		{
			// Read commandableInFocus
			final boolean commandableIsNull = Marshaller.toBoolean(marshaller.readByte());
			if (commandableIsNull) {
				this.commandable = null;
			} else {
				final CommandableKey key = (CommandableKey) marshaller.readMarshallable(CommandableKey.factory);
				this.commandable = gameView.getGameState().findCommandable(key);
			}
		} {
			// Read moveResult
			final boolean moveResultIsNull = Marshaller.toBoolean(marshaller.readByte());
			if (moveResultIsNull) {
				this.moveResult = null;
			} else {
				this.moveResult = new MoveResult(gameView, marshaller);
			}
		}
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(turn);
		marshaller.writeInt(currentPlayerId);
		marshaller.writeInt(affectedPlayerId);
		marshaller.writeMarshallable(entry);
		PlayfieldPosition.marshalWithNull(marshaller, pos);
		marshaller.writeNullOrMarshallable(null == commandable ? commandable : commandable.getKey());
		marshaller.writeNullOrMarshallable(moveResult);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw new IllegalStateException("Please call the marshaller constructor instead.");
	}

	public I18NStringWithFillIns getEntry() {
		return entry;
	}
}