package de.ped.empire.logic;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;

/**
 * Defines a path as shown in path mode.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MovePath implements Marshallable, Cloneable {
	MovePath() {
		this.playerId = Player.ID_NONE;
		this.militaryBranchesMask = 0x00;
		this.startPos = new PlayfieldPosition();
		this.endPos = new PlayfieldPosition();
	}

	public MovePath(final int playerId, final PlayfieldPosition startPos, final int militaryBranchesMask, final PlayfieldPosition endPos) {
		this.playerId = playerId; 
		this.militaryBranchesMask = militaryBranchesMask;
		this.startPos = startPos;
		this.endPos = endPos;
	}

	@Override
	public void marshal(Marshaller marshaller) {
		marshaller.writeInt(playerId);
		marshaller.writeInt(militaryBranchesMask);
		startPos.marshal(marshaller);
		endPos.marshal(marshaller);
	}

	@Override
	public Marshallable demarshal(Marshaller marshaller) {
		playerId = marshaller.readInt();
		militaryBranchesMask = marshaller.readInt();
		startPos = new PlayfieldPosition();
		startPos.demarshal(marshaller);
		endPos = new PlayfieldPosition();
		endPos.demarshal(marshaller);
		return this;
	}

//		int getMilitaryBranchesMask() {
//			return militaryBranchesMask;
//		}
//		
	public boolean isForMilitaryBranch(final MilitaryBranch branch) {
		return (0 != (militaryBranchesMask & branch.mask));
	}

	public void setForMilitaryBranch(final MilitaryBranch branch, boolean value) {
		if (value) {
			militaryBranchesMask |= branch.mask;
		} else {
			militaryBranchesMask &= (~branch.mask); 
		}
	}

	/** 
	 * @return <code>true</code> if this path object represents no military branch
	 */
	boolean isEmpty() {
		return (0x00 == militaryBranchesMask);
	}

	boolean isSharingMilitaryBranch(final MovePath other) {
		return (0 != (militaryBranchesMask & other.militaryBranchesMask));
	}

	/** 
	 * Removes all military branches indicated in the mask. After this operation, this path might be empty.
	 * @param other the path that defines the branches to remove from this
	 */
	void removeMilitaryBranches(final MovePath other) {
		militaryBranchesMask &= (~other.militaryBranchesMask);
	}

	void addMilitaryBranches(final MovePath other) {
		militaryBranchesMask |= other.militaryBranchesMask;
	}

	public int getPlayerId() {
		return playerId;
	}

	public PlayfieldPosition getStartPos() {
		return startPos;
	}

	public PlayfieldPosition getEndPos() {
		return endPos;
	}

	/**
	 * @return a deep copy - start and end position are cloned as well
	 */
	@Override
	public MovePath clone() {
		MovePath ret;
		try {
			ret = (MovePath) super.clone();
			ret.startPos = new PlayfieldPosition(startPos);
			ret.endPos   = new PlayfieldPosition(endPos);
		} catch (final CloneNotSupportedException e) {
			e.printStackTrace();
			ret = null;
		}
		return ret;
	}

	@Override
	public String toString() {
		String ret = "MovePath(";
		ret += "playerId=" + playerId;
		final PlayfieldPosition that = startPos;
		ret += ", startPos=" + that.toString();
		final PlayfieldPosition that1 = endPos;
		ret += ", endPos=" + that1.toString();
		ret += ", mask=" + Integer.toBinaryString(militaryBranchesMask);
		return ret;
	}

	private int playerId;
	private PlayfieldPosition startPos;
	private int militaryBranchesMask;
	private PlayfieldPosition endPos;
}