package de.ped.empire.logic;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;

/**
 * An encapsulation of the collection of all move paths existing in a game.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MovePaths implements Marshallable, Cloneable {
	private final ArrayList<MovePath> paths = new ArrayList<MovePath>();

	@Override
	public synchronized void marshal(Marshaller marshaller) {
		marshaller.writeInt(paths.size());
		for (MovePath path : paths) {
			marshaller.writeMarshallable(path);
		}
	}

	@Override
	public synchronized Marshallable demarshal(Marshaller marshaller) {
		paths.clear();
		final int size = marshaller.readInt();
		paths.ensureCapacity(size);
		for (int i = 0; i < size; i++) {
			MovePath path = new MovePath();
			marshaller.readMarshallable(path);
			paths.add(path);
		}
		return this;
	}

	public synchronized Collection<MovePath> findMovePaths(final int playerId) {
		final Collection<MovePath> ret = new ArrayList<MovePath>(); 
		for (final MovePath path : paths) {
			if (playerId == path.getPlayerId()) {
				ret.add(path);
			}
		}
		return ret;
	}

	public synchronized List<MovePath> findMovePaths(final int playerId, final PlayfieldPosition startPos) {
		final List<MovePath> ret = new ArrayList<MovePath>();
		for (final MovePath path : paths) {
			if ((playerId == path.getPlayerId())
					&& (PlayfieldPosition.equals(startPos, path.getStartPos()))) {
				ret.add(path);
			}
		}
		return ret;
	}

	/**
	 * Tries to find the fitting move path.
	 * @param playerId the player ID of the player the move path is for
	 * @param startPos the start position of the move path
	 * @param branch the military branch the move path is for
	 * @return the move path found, or <code>null</code> if there was none
	 */
	public synchronized MovePath findMovePath(final int playerId, final PlayfieldPosition startPos, final MilitaryBranch branch) {
		MovePath ret = null;
		for (final MovePath path : paths) {
			if ((playerId == path.getPlayerId())
					&& (PlayfieldPosition.equals(startPos, path.getStartPos()))
					&& (path.isForMilitaryBranch(branch))) {
				ret = path;
				break;
			}
		}
		return ret;
	}

	/**
	 * Adds a new path. If this conflicts with other paths, the new path wins. 
	 * So the following rules apply to paths with the same player and start postition:
	 * <ul>
	 *   <li>
	 *     If there is an path that shares military branches, those branches are removed from the old path.
	 *     The old path is left, even if empty, but can be removed using {@link #cleanup()}.
	 *   <li>
	 *     If there is a path with other military branches, but with same target position,
	 *     those paths are merged into one.
	 * </ul>
	 * 
	 * @param movePath
	 *            the new path to add
	 * @throws CircularReferenceException
	 *             if the new move path would lead to a circular reference for
	 *             at least one military branch
	 */
	public synchronized void add(final MovePath movePath) throws CircularReferenceException {
		wouldCreateCircularReference(movePath);
		boolean isToAdd = true;
		// Remove or limit all paths so no double-defines are possible.
		for (int i = paths.size() - 1; 0 <= i; i--) {
			final MovePath path = paths.get(i);
			if ((movePath.getPlayerId() == path.getPlayerId())
					&& (PlayfieldPosition.equals(movePath.getStartPos(), path.getStartPos()))) {
				if (PlayfieldPosition.equals(movePath.getEndPos(), path.getEndPos())) {
					// We can merge two path objects.
					path.addMilitaryBranches(movePath);
					isToAdd = false;
				} else if (movePath.isSharingMilitaryBranch(path)) {
					// remove all intersecting branches. If no branch left, remove old path.
					path.removeMilitaryBranches(movePath); 
				}
			}
		}
		// Add new path
		if (isToAdd) {
			paths.add(movePath);
		}
	}

	/**
	 * Checks if adding this move path would result in a circular reference.
	 * 
	 * @param movePath
	 *            the move path to add
	 * @throws CircularReferenceException
	 *             if the new move path would lead to a circular reference for
	 *             at least one military branch
	 */
	void wouldCreateCircularReference(final MovePath movePath) throws CircularReferenceException {
		for (MilitaryBranch branch : MilitaryBranch.values()) { 
			if (movePath.isForMilitaryBranch(branch)) {
				MovePath path = movePath;
				while (null != path) {
					final PlayfieldPosition pos = path.getEndPos();
					if (PlayfieldPosition.equals(pos, movePath.getStartPos())) {
						throw new CircularReferenceException("Move path would lead to circular path routing!");
					}
					path = findMovePath(movePath.getPlayerId(), pos, branch);
				}
			}
		}
	}

	public void addAll(final Collection<MovePath> coll) throws CircularReferenceException {
		for (MovePath path : coll) {
			add(path);
		}
	}

	/**
	 * Removes any empty paths.
	 * @see MovePath#isEmpty()
	 */
	public synchronized void cleanup() {
		for (int i = paths.size() - 1; 0 <= i; i--) {
			if (paths.get(i).isEmpty()) {
				paths.remove(i);
			}
		}
	}
	
	public synchronized void remove(final MovePath movePath) {
		paths.remove(movePath);
	}

	/**
	 * Removes all paths for the given player that start or end at the given position.
	 * @param playerId the player ID
	 * @param position the position
	 * @param isWithPathEnd if <code>true</code>, also paths having position as end point will be removed
	 */
	public synchronized void removeAll(final int playerId, final PlayfieldPosition position, final boolean isWithPathEnd) {
		for (int i = paths.size() - 1; 0 <= i; i--) {
			final MovePath path = paths.get(i);
			if ((playerId == path.getPlayerId())
					&& ((PlayfieldPosition.equals(position, path.getStartPos()))
							|| (isWithPathEnd && PlayfieldPosition.equals(position, path.getEndPos())))) {
				paths.remove(i);
			}
		}
	}

	/**
	 * Removes all paths for the given player that start at the given position and have no military units.
	 * @param playerId the player ID
	 * @param position the position
	 */
	public synchronized void clean(final int playerId, final PlayfieldPosition position) {
		for (int i = paths.size() - 1; 0 <= i; i--) {
			final MovePath path = paths.get(i);
			if ((playerId == path.getPlayerId())
					&& PlayfieldPosition.equals(position, path.getStartPos())
					&& path.isEmpty()) {
				paths.remove(i);
			}
		}
	}

	/**
	 * @return a deep deep copy - each move path inside this collection gets cloned
	 */
	@Override
	public synchronized MovePaths clone() {
		final MovePaths other = new MovePaths();
		for (MovePath path : paths) {
			other.paths.add((MovePath) path.clone());
		}
		return other;
	}
}