package de.ped.empire.logic;

import java.util.ArrayList;
import java.util.List;

import de.ped.tools.Assert;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;

/**
 * Returns all the information of a move to replay it.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MoveResult implements Marshallable {
	/**
	 * The role a unit plays in this move result. Can uniquely identify all units that play a part in this.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Role {
		/** The acting unit */
		Active,
		/** The commandable that may have been affected, e.g., as load target or attacked commandable */
		Affected,
	};

	/**
	 * Contains all information of a commandable that played a role in this move.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static final class CommandableInfo implements Marshallable {
		/** The containing structure. Non-persistent. */
		public final MoveResult moveResult;

		/**
		 * The commandable involved in the move; might be <code>null</code> for the
		 * CommandableInfo for {@link Role#Affected}. Persistent by key.
		 */
		public final Commandable commandable;

		/** The position involved in the move. Persistent. */
		public final PlayfieldPosition position;

		/** The image key of the commandable involved. Persistent. */
		public final int imageKey;

		/**
		 * The player ID of the commandable involved at the start of the move.
		 * (The player ID of the commandable itself may change within the move,
		 * e.g. when cities are conquered.) Persistent.
		 */
		private final int playerId;

		/** If the commandable is cargo. Persistent. */
		public final boolean isCargo;

		public CommandableInfo(final MoveResult moveResult, final GameView gameView, final Commandable commandable) {
			this.moveResult = moveResult;
			this.commandable = commandable;
			if (null == commandable) {
				position = null;
				imageKey = 0;
				playerId = Player.ID_NONE;
				isCargo = false;
			} else {
				position = new PlayfieldPosition(commandable.getPosition());
				final Commandable viewCmdl = gameView.findCommandable(commandable.getKey());
				if (null == viewCmdl) { // This is the case for pivotal units for e.g. flight path calculations
					imageKey = 0;
				} else {
					imageKey = ((TileTemplate) viewCmdl).getImageKey();
				}
				playerId = commandable.getPlayerId();
				if (commandable instanceof Unit) {
					final Unit u = (Unit) commandable;
					if ((null != u.getParentKey())
							|| gameView.getGameState().getMap().getAt(u.getPosition()).isCity()) {
						isCargo = true;
					} else {
						isCargo = false;
					}
				} else {
					isCargo = false;
				}
			}
		}

		public CommandableInfo(final MoveResult moveResult, final GameView gameView, final Marshaller marshaller) {
			this.moveResult = moveResult;
			commandable = gameView.getGameState().demarshalCommandableKey(marshaller, true);
			if (null == commandable) {
				position = null;
				imageKey = 0;
				playerId = Player.ID_NONE;
				isCargo = false;
			} else {
				position = PlayfieldPosition.demarshalWithNull(marshaller);
				imageKey = marshaller.readInt();
				playerId = marshaller.readInt();
				isCargo = marshaller.readBoolean();
			}
		}

		@Override
		public void marshal(final Marshaller marshaller) {
			GameState.marshalCommandableKey(marshaller, commandable, true);
			if (null == commandable) {
				// Nothing to do
			} else {
				PlayfieldPosition.marshalWithNull(marshaller, position);
				marshaller.writeInt(imageKey);
				marshaller.writeInt(playerId);
				marshaller.writeBoolean(isCargo);
			}
		}

		@Override
		public Marshallable demarshal(final Marshaller marshaller) {
			throw Marshaller.createConstantMarshallableDemarshallingException();
		}

		public int getPlayerId() {
			return playerId;
		}

		/**
		 * @return the unit, if this info contains a unit, or <code>null</code>
		 *         otherwise (e.g., contains no commandable or a city)
		 */
		public Unit getUnit() {
			Unit ret = null;
			if ((null != commandable) && (commandable instanceof Unit)) {
				ret = (Unit) commandable;
			}
			return ret;
		}
	}

	/** A mask value for: Nothing happened, because move was impossible. */
	public static final int MOVE_MASK_NOTHING  = 0x0000;

	/** Mask value for: Unit has moved, that means changed the position into the direction desired. */
	public static final int MOVE_MASK_MOVING   = 0x0001;

	/** Mask value for: Unit has fought. Fighting sequence needs to be shown. */
	public static final int MOVE_MASK_FIGHTING = 0x0002;

	/** Mask value for: Unit has died. Death sequence for this unit needs to be shown. */
	public static final int MOVE_MASK_LOSING_FIGHT   = 0x0004;

	/** Mask value for: Opponent died. Death sequence for opponent needs to be shown. */
	public static final int MOVE_MASK_WINNING_FIGHT  = 0x0008;

	/** Mask value for: Stopped moving for the rest of the turn. */
//	public static final int MOVE_MASK_SKIP = 0x0020;

	/** Mask value for: Docked in city. */
	public static final int MOVE_MASK_DOCKING  = 0x0040;

	/** Mask value for: Loaded into other unit. */
	public static final int MOVE_MASK_LOADING  = 0x0080;

	/** Mask value for: Unit needs to alert the user, e.g., because a command was unable to be executed. */
	public static final int MOVE_MASK_ALERT_IMPEDIMENT = 0x0100;

	/** Mask value for: Unit needs to alert the user, e.g., because there was an enemy in sight. */
	public static final int MOVE_MASK_ALERT_SIGHTING = 0x0200;

	/** Mask value for: Unit needs to alert the user, e.g., because there was an enemy in sight. */
	public static final int MOVE_MASK_FUEL_EMPTY = 0x0400;

	/** Mask value for: Unit is moving over another unit during turn, so its turn cannot end here. */
	public static final int MOVE_MASK_MOVE_OVER = 0x0800;

	/** Mask value for: Unit is building something and is therefore disappearing as garrison. */
	public static final int MOVE_MASK_GARRISON = 0x1000;

	/** Mask value for: Unit is slowed down by rough land. */
	public static final int MOVE_MASK_SLOWED_DOWN = 0x2000;

	/** Mask value for: Move resulted in losing the game for the player who lost the unit. */
	public static final int MOVE_MASK_LOSING_GAME = 0x4000;

	/** Mask value for: Move resulted in winning the game for the player who didn't lose his/her unit. */
	public static final int MOVE_MASK_WINNING_GAME = 0x8000;

	/** The infos about all units involved in this move. Persistent. */
	private CommandableInfo[] cmdInfos = new CommandableInfo[2];

	/** The direction the active unit went. Persistent. */
	final DigitalJoystickDirection direction;

	/** The mask of things that happened. Persistent. */
	private int resultMask;

	/** The sequence of hits to display in a fight. Persistent. */
	private List<Role> hits = new ArrayList<Role>();

	/** The number of units killed within the fighting commandables. Persistent. */
	private int[] killedUnits = new int[] { 0, 0, };

	public MoveResult(final GameView gameView, final Unit unit, final DigitalJoystickDirection direction) {
		this.gameView = gameView;
		cmdInfos[Role.Active.ordinal()] = new CommandableInfo(this, gameView, unit);
		cmdInfos[Role.Affected.ordinal()] = new CommandableInfo(this, gameView, (Commandable) null);
		this.direction = direction;
		this.resultMask = MOVE_MASK_NOTHING;
	}

	public MoveResult(final GameView gameView, final Marshaller marshaller) {
		this.gameView = gameView;
		for (int i = 0; i < cmdInfos.length; i++) {
			cmdInfos[i] = new CommandableInfo(this, gameView, marshaller);
		}
		if (!marshaller.readNullIndicator()) {
			direction = DigitalJoystickDirection.getDirection(marshaller.readInt());
		} else {
			direction = null;
		}
		resultMask = marshaller.readInt();
		for (int i = 0; i < killedUnits.length; i++) {
			killedUnits[i] = marshaller.readInt();
		}
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		for (int i = 0; i < cmdInfos.length; i++) {
			marshaller.writeMarshallable(cmdInfos[i]);
		}
		if (!marshaller.writeNullIndicator(direction)) {
			marshaller.writeInt(direction.getId());
		}
		marshaller.writeInt(resultMask);
		for (int i = 0; i < killedUnits.length; i++) {
			marshaller.writeInt(killedUnits[i]);
		}
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	public int getResultMask() {
		return resultMask;
	}

	void setResultMask(final int mask) {
		this.resultMask = mask;
	}

	public CommandableInfo getCommandableInfo(final Role role) {
		return cmdInfos[role.ordinal()];
	}

	public Commandable getCommandable(final Role role) {
		return cmdInfos[role.ordinal()].commandable;
	}

	public String getPlayerName(final Role role) {
		String ret = null;
		final CommandableInfo info = cmdInfos[role.ordinal()];
		if (null != info.commandable) {
			ret = gameView.getPlayer(info.getPlayerId()).getName();
		}
		return ret;
	}

	public I18NStringWithFillIns getUnitName(final Role role) {
		I18NStringWithFillIns ret = null;
		final CommandableInfo info = cmdInfos[role.ordinal()];
		if (null != info.commandable) {
			ret = info.commandable.getName();
		}
		return ret;
	}

	void setAffectedCommandable(final Commandable commandable) {
		Assert.assertTrue(commandable instanceof TileTemplate);
		cmdInfos[Role.Affected.ordinal()] = new CommandableInfo(this, gameView, commandable);
	}

	public int getHits(final Role role) {
		int ret = 0;
		for (final Role r : hits) {
			if (role == r) {
				ret++;
			}
		}
		return ret;
	}

	void hit(final Role role, final int numberOfHits) {
		for (int i = 0; i < numberOfHits; i++) {
			hits.add(role);
		}
	}

	void setKilledUnits(final int active, final int affected) {
		killedUnits[0] = active;
		killedUnits[1] = affected;
	}

	int[] getKilledUnits() {
		return killedUnits;
	}

	/**
	 * @return the list of who hits who during the fight. <code>true</code>
	 *         means the unit hits the opponent, <code>false</code> means the
	 *         opponent hits the unit.
	 */
	List<Role> getHits() {
		return hits;
	}

	@Override
	public String toString() {
		return "MoveResult(mask=0x" + Integer.toHexString(resultMask) + ", unit=" + getCommandable(Role.Active) + ", dir=" + (null == direction ? "null" : direction.key) + ")";
	}

	/**
	 * Checks whether the given commandable is involved in this move.
	 * 
	 * @param commandable
	 *            the commandable to check for
	 * @return <code>true</code> if <em>commandable</em> is not <code>null</code> and is
	 *         involved in this move either in active or affected role
	 */
	public boolean isWith(final Commandable commandable) {
		boolean ret = false;
		if (null != commandable) {
			for (final CommandableInfo cmdInfo : cmdInfos) {
				if (commandable == cmdInfo.commandable) {
					ret = true;
					break;
				}
			}
		}
		return ret;
	}

	public boolean isWith(final PlayfieldPosition position) {
		boolean ret = false;
		if (null != position) {
			for (final PlayfieldPosition pos : getPositions()) {
				if (PlayfieldPosition.equals(position, pos)) {
					ret = true;
					break;
				}
			}
		}
		return ret;
	}

	public DigitalJoystickDirection getDirection() {
		return direction;
	}

	public Unit getUnit() {
		return cmdInfos[Role.Active.ordinal()].getUnit();
	}

	List<PlayfieldPosition> getPositions() {
		final List<PlayfieldPosition> ret = new ArrayList<PlayfieldPosition>(2);
		final PlayfieldPosition startPos = getCommandableInfo(Role.Active).position; 
		ret.add(startPos);
		if (null != direction) {
			final PlayfieldPosition endPos = gameView.getGameState().getProperties().wrap(startPos, direction);
			ret.add(endPos);
		}
		return ret;
	}

	/**
	 * Calculates the final status message for this move result. Called when the
	 * move animation is nearing completion.
	 * 
	 * @param role
	 *            the role for which the log message needs to be created
	 * @param activePlayersUnitsKilled
	 *            the number of acting player's units killed by this move
	 * @param affectedPlayersUnitsKilled
	 *            the number of affected player's units killed by this move
	 * @return the log message for the given role, or <code>null</code> if there
	 *         is no message to log
	 */
	I18NStringWithFillIns getLogMessage(
			final Role role, 
			final int activePlayersUnitsKilled, 
			final int affectedPlayersUnitsKilled) {
		final int mask = getResultMask();
		final Commandable affected = getCommandable(Role.Affected);
		final Object[] fillIns = new Object[] {
				getPlayerName(Role.Active), 				// %0=attacking player
				getPlayerName(Role.Affected),				// %1=defending player
				getUnitName(Role.Active), 					// %2=attacking unit 
				getUnitName(Role.Affected),					// %3=defending unit
				Integer.toString(getHits(Role.Active)),		// %4=attacking unit hits received
				Integer.toString(getHits(Role.Affected)),	// %5=defending unit hits received
				Integer.toString(activePlayersUnitsKilled),	// %6=attacker's units killed
				Integer.toString(affectedPlayersUnitsKilled),// %7=defender's units killed
		};
		final I18NStringWithFillIns ret;
		switch(role) {
		case Active: {
				final String msgFor = ".AttackerMsg";
				if (0 != (mask & MOVE_MASK_FIGHTING)) {
					if (0 != (mask & MOVE_MASK_LOSING_FIGHT)) { // The active unit has lost, the affected unit won
						final String key = (affected instanceof FieldCity 
								? "Game.AttackerLostAgainst.City"
								: "Game.AttackerLostAgainst.Unit");
						final I18NStringWithFillIns i18nLostOrWon = new I18NStringWithFillIns(key + msgFor, null, fillIns);
						// The lost unit gets less information about the opponent
						if (0 < activePlayersUnitsKilled) {
							final I18NStringWithFillIns i18slostUnitsAttacker = new I18NStringWithFillIns("Game.Unit.Attacker.LosingUnits" + msgFor, null, fillIns);
							ret = new I18NStringWithFillIns(null, "%0%1", new Object[] { i18nLostOrWon, i18slostUnitsAttacker, });
						} else {
							ret = i18nLostOrWon;
						}
					} else { // The active unit has won, the affected unit lost
						final String key;
						if (affected instanceof FieldCity) {
							if (0 != (resultMask & MOVE_MASK_MOVING)) {
								key = "Game.AttackerWonAgainst.City";
							} else {
								key = "Game.AttackerWonAgainst.City.Production";
							}
						} else {
							key = "Game.AttackerWonAgainst.Unit";
						}
						final I18NStringWithFillIns i18nLostOrWon = new I18NStringWithFillIns(key + msgFor, null, fillIns);
						I18NStringWithFillIns i18slostUnitsDefender = I18NStringWithFillIns.EMPTY;
						I18NStringWithFillIns i18sHitsReceived = I18NStringWithFillIns.EMPTY;
						I18NStringWithFillIns i18slostUnitsAttacker = I18NStringWithFillIns.EMPTY;
						if (0 < affectedPlayersUnitsKilled) {
							i18slostUnitsDefender = new I18NStringWithFillIns("Game.Unit.Defender.LosingUnits" + msgFor, null, fillIns);
						}
						if (0 < activePlayersUnitsKilled) {
							i18slostUnitsAttacker = new I18NStringWithFillIns("Game.Unit.Attacker.LosingUnits" + msgFor, null, fillIns);
						}
						if (0 < getHits(role)) {
							i18sHitsReceived = new I18NStringWithFillIns("Game.Unit.Attacker.ReceivingHits", null, fillIns);
						}
						ret = new I18NStringWithFillIns(null, "%0%1%2%3", new Object[] { i18nLostOrWon, i18slostUnitsDefender, i18sHitsReceived, i18slostUnitsAttacker, });
					}
				} else {
					if (0 != (mask & MOVE_MASK_LOSING_FIGHT)) {
						ret = new I18NStringWithFillIns("Game.Unit.Disbanded", null, fillIns);
					} else {
						ret = I18NStringWithFillIns.EMPTY;
					}
				}
			}
			break;
		case Affected: {
				final String msgFor = ".DefenderMsg";
				if (0 != (mask & MOVE_MASK_FIGHTING)) {
					if (0 != (mask & MOVE_MASK_LOSING_FIGHT)) { // The active unit has lost, the affected unit won
						final String key = (affected instanceof FieldCity 
								? "Game.AttackerLostAgainst.City"
								: "Game.AttackerLostAgainst.Unit");
						final I18NStringWithFillIns i18nLostOrWon = new I18NStringWithFillIns(key + msgFor, null, fillIns);
						I18NStringWithFillIns i18slostUnitsDefender = I18NStringWithFillIns.EMPTY;
						I18NStringWithFillIns i18sHitsReceived = I18NStringWithFillIns.EMPTY;
						I18NStringWithFillIns i18slostUnitsAttacker = I18NStringWithFillIns.EMPTY;
						if (0 < affectedPlayersUnitsKilled) {
							i18slostUnitsDefender = new I18NStringWithFillIns("Game.Unit.Defender.LosingUnits" + msgFor, null, fillIns);
						}
						if (0 < activePlayersUnitsKilled) {
							i18slostUnitsAttacker = new I18NStringWithFillIns("Game.Unit.Attacker.LosingUnits" + msgFor, null, fillIns);
						}
						if (0 < getHits(role)) {
							i18sHitsReceived = new I18NStringWithFillIns("Game.Unit.Defender.ReceivingHits", null, fillIns);
						}
						ret = new I18NStringWithFillIns(null, "%0%2%1%3", new Object[] { i18nLostOrWon, i18slostUnitsDefender, i18sHitsReceived, i18slostUnitsAttacker, });
					} else { // The active unit has won, the affected unit lost
						final String key;
						if (affected instanceof FieldCity) {
							if (0 != (resultMask & MOVE_MASK_MOVING)) {
								key = "Game.AttackerWonAgainst.City";
							} else {
								key = "Game.AttackerWonAgainst.City.Production";
							}
						} else {
							key = "Game.AttackerWonAgainst.Unit";
						}
						final I18NStringWithFillIns i18nLostOrWon = new I18NStringWithFillIns(key + msgFor, null, fillIns);
						// The lost unit gets less information about the opponent
						if (0 < affectedPlayersUnitsKilled) {
							final I18NStringWithFillIns i18slostUnitsDefender = new I18NStringWithFillIns("Game.Unit.Defender.LosingUnits" + msgFor, null, fillIns);
							ret = new I18NStringWithFillIns(null, "%0%1", new Object[] { i18nLostOrWon, i18slostUnitsDefender, });
						} else {
							ret = i18nLostOrWon;
						}
					}
				} else {
					ret = I18NStringWithFillIns.EMPTY;
				}
			}
			break;
		default:
			ret = I18NStringWithFillIns.EMPTY;
		}
		return ret;
	}

	/** Non-persistent. */
	private final GameView gameView;
}