package de.ped.empire.logic;

import java.util.List;

import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.Messages;

/**
 * Abstract superclass for all orders that can be given to a military {@link Unit}.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Order implements Marshallable, Cloneable {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	/** The standard postfix for a message key defining a safety question to ask the user if an order should really be given to a unit. */
	public static final String POSTIFX_SAFETY_QUESTION = ".SafetyQuestion";

	protected static final String ERRORMSG_TRANSIENT = "This is a transient order that is executed immediatly.";

	/**
	 * Factory for {@link Order} when demarshalling.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Factory implements Marshallable {
		/**
		 * Factory method to read an order from {@link Marshaller} while
		 * deserializing.
		 * 
		 * @param gameState
		 *            the game state to create the order for
		 * @param commandable
		 *            the commandable to create the order for
		 * @param marshaller
		 *            the marshaller to read the rest of the order from
		 * @return the order, or <code>null</code> if order was transient
		 */
		public static Order create(final GameState gameState, final Commandable commandable, final Marshaller marshaller) {
			final byte b = marshaller.readByte();
			switch(b) {
			case OrderTemplate.TYPE_CLEAR_ORDERS:	return null; // transient
			case OrderTemplate.TYPE_DISBAND:		return null; // transient
			case OrderTemplate.TYPE_SKIP_MOVE:		return null; // transient
			case OrderTemplate.TYPE_SENTRY:			return new OrderSentry(gameState, commandable, marshaller);
			case OrderTemplate.TYPE_LOAD:			return new OrderLoad(gameState, commandable, marshaller);
			case OrderTemplate.TYPE_UNLOAD:			return null; // transient
			case OrderTemplate.TYPE_BUILD_AIRBASE:	return null; // transient
			case OrderTemplate.TYPE_GOTO:			return new OrderGoTo(gameState, commandable, marshaller);
//			case OrderTemplate.TYPE_ATTACK:			return new OrderAttack(gameState, commandable, marshaller);
			case OrderTemplate.TYPE_HOME:			return new OrderHome(gameState, commandable, marshaller);
			case OrderTemplate.TYPE_EXPLORE:		return new OrderExplore(gameState, commandable, marshaller);
			case OrderTemplate.TYPE_ESCORT:			return new OrderEscort(gameState, commandable, marshaller);
			default: throw new IllegalArgumentException("Unknown order type " + Integer.toHexString(b));
			}
		}

		public Factory(final GameState gameState, final Commandable commandable) {
			this.gameState = gameState;
			this.commandable = commandable;
		}

		private final GameState gameState;
		private final Commandable commandable;

		@Override
		public void marshal(final Marshaller marshaller) {
			throw new IllegalArgumentException("Factory is never marshalled.");
		}

		@Override
		public Marshallable demarshal(final Marshaller marshaller) {
			return create(gameState, commandable, marshaller);
		}
	}

	/**
	 * Creates a generic order without a binding to game or commandable, to be
	 * used e.g. in UI menus.
	 * 
	 * @param type
	 *            the order type of the order to create
	 * @return the generic order of that type, or <code>null</code> if that
	 *         makes no sense
	 */
	public static Order createGenericOrder(final byte type) {
		switch(type) {
		case OrderTemplate.TYPE_CHANGE_PRODUCTION: 
											return new OrderChangeProduction(null, null);
		case OrderTemplate.TYPE_CHANGE_MOVE_PATHS: 
											return new OrderDefineMovePaths(null, null);
		case OrderTemplate.TYPE_CLEAR_ORDERS:
											return new OrderClearOrders(null, null);
		case OrderTemplate.TYPE_CLEAR_ALERTS:
											return new OrderClearAlerts(null, null);
		case OrderTemplate.TYPE_DISBAND:    return new OrderDisband(null, null); 
		case OrderTemplate.TYPE_SKIP_MOVE:  return new OrderSkipMove(null, null);
		case OrderTemplate.TYPE_SENTRY:     return new OrderSentry(null, null);
		case OrderTemplate.TYPE_LOAD:       return new OrderLoad(null, null);
		case OrderTemplate.TYPE_UNLOAD:     return new OrderUnload(null, null); 
		case OrderTemplate.TYPE_BUILD_AIRBASE:
											return new OrderBuildAirbase(null, null); 
		case OrderTemplate.TYPE_GOTO:       return null; // Targeted orders are handled elsewhere
		case OrderTemplate.TYPE_HOME:       return new OrderHome(null, null);
		case OrderTemplate.TYPE_EXPLORE:    return new OrderExplore(null, null);
		case OrderTemplate.TYPE_ESCORT:     return null; // Targeted orders are handled elsewhere
		case OrderTemplate.TYPE_RENAME:     return new OrderRename(null, null);
		default: throw new IllegalArgumentException("Unknown order type " + Integer.toHexString(type));
		}
	}

	/** Non-persistent. */
	protected Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.DEBUG);
	}

	/** ID persistent, represents the type */
	public final OrderTemplate template;
	/** Non-persistent, set by factory. */
	private GameState gameState;
	/** Non-persistent, set by factory. */
	private Commandable commandable;

	protected Order(final GameState gameState, final Commandable commandable, final OrderTemplate orderTemplate) {
		this.gameState = gameState;
		this.commandable = commandable;
		this.template = orderTemplate;
//		Assert.assertTrue((null == commandable) || (commandable.getPlayerId() != Player.ID_NONE));
	}

	protected Order(final GameState gameState, final Commandable commandable, final OrderTemplate orderTemplate, final Marshaller marshaller) {
		this.gameState = gameState;
		this.commandable = commandable;
		this.template = orderTemplate;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeByte(getOrderType());
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	protected byte getOrderType() {
		return template.id;
	}

	protected GameState getGameState() {
		return gameState;
	}

	protected GameView getGameView() {
		GameView ret;
		if (null == commandable) {
			ret = gameState.getCurrentGameView();
		} else {
			ret = gameState.getGameView(commandable.getKey().getPlayerId());
		}
		return ret;
	}

	public void setGameState(GameState gameState) {
		this.gameState = gameState;
		if (null != commandable) {
			throw new IllegalStateException("Changing game makes no sense for order linked to unit.");
		}
	}

	public Commandable getCommandable() {
		return commandable;
	}

	public Order cloneIfNeededForCommandable(final Commandable commandable) {
		Order ret = (Order) this.clone();
		if (null == getCommandable()) {
			ret.commandable = commandable;
		}
		return ret;
	}

	public String getMessageKeyPrefix() {
		return template.messageKeyPrefix;
	}

	/**
	 * Returns the tool tip to show to the user for the given commandable. Only
	 * suitable for those GUI orders shown in menus.
	 * 
	 * @param commandable
	 *            the commandable
	 * @return the tool tip to show to the user for the given commandable
	 */
	public I18NStringWithFillIns getToolTip(final Commandable commandable) {
		return new I18NStringWithFillIns(getMessageKeyPrefix() + Messages.TEXT_POSTFIX, null, null);
	}

	/** 
	 * @return the tool tip to show to the user
	 */
	public I18NStringWithFillIns getToolTip() {
		return getToolTip(getCommandable());
	}

	/**
	 * Returns if the command makes sense to present to the user as option to
	 * select. Only suitable for those GUI orders shown in menus.
	 * <p>
	 * Default implementation just asks the template. Should be overloaded if
	 * behavior needs to be more specific on parameters.
	 * 
	 * @param commandable
	 *            the commandable to check the order for; could be different
	 *            than the commandable given at creation time of this object.
	 * @return if the order makes sense to present to the user
	 * @see OrderTemplate#makesSense(GameView, Commandable)
	 */
	public boolean makesSense(final Commandable commandable) {
		if (null == this.commandable) {
			return template.makesSense(getGameView(), commandable);
		} else if (commandable.equals(this.commandable)) {
			if (commandable.getPlayerId() != Player.ID_NONE) {
				return template.makesSense(getGameView(), commandable);
			} else {
				return false;
			}
		} else {
			throw new IllegalStateException("Why do you call this for another commandable?");
		}
	}

	/**
	 * @return if the command makes sense to present to the user as option to
	 *         select, using the commandable given at creation time of this
	 *         order
	 * @see #makesSense(Commandable)
	 */
	public boolean makesSense() {
		return makesSense(commandable);
	}

	/**
	 * Returns if the order is fulfilled by the unit and should be cleared.
	 * Default implementation just always returns <code>true</code>. Should be
	 * overloaded if behavior needs to be more specific.
	 * 
	 * @return if the order is fulfilled by the unit
	 */
	public boolean isFulfilled() {
		return true;
	}

	/** 
	 * Returns the message key for a safety question, or <code>null</code> if the user should not be bothered.
	 * Default implementation returns <code>null</code>. 
	 * 
	 * @return the message key for a safety question, if any
	 */
	public String getSafetyQuestion() {
		return null;
	}

	/** 
	 * Offers an alternative implementation for {@link #getSafetyQuestion()},
	 * looking for an entry in the messages properties, 
	 * using a key concatenating {@link #getMessageKeyPrefix()} and {@link #POSTIFX_SAFETY_QUESTION}.
	 * 
	 * @return the default message key for a safety question
	 */
	protected String getSafetyQuestionDefault() {
		return getMessageKeyPrefix() + POSTIFX_SAFETY_QUESTION;
	}

	/**
	 * Prepares the evaluation of possible next steps.
	 * 
	 * @return <code>true</code> if the order goal was reachable and further
	 *         rating makes sense
	 */
	public boolean prepareRating() {
		return false;
	}

	/**
	 * If the subclass wants to really implement the evaluation of possible next moves,
	 * this is the first thing to call in its own implementation of {@link #prepareRating()}.
	 * It initializes the step rating field of the unit, so that the order can set the target points as seeds.
	 * @param stepRatingPurpose the purpose to rate the map for
	 */
	protected void calculateForwardRating(final StepRatingPurpose stepRatingPurpose) {
		final Unit unit = (Unit) getCommandable();
		if (null == unit.getStepRatings()) {
			unit.calculateForwardRating(stepRatingPurpose);
		}
	}

	/**
	 * Tries to derive a next step for the unit based on this command.
	 * 
	 * @param isNow
	 *            <code>true</code> if the move could be executed immediatly,
	 *            <code>false</code> if the command is given now, but executed
	 *            later
	 * @return a next step for the unit based on this command
	 */
	public abstract MoveResult evaluate(final boolean isNow);

	@Override
	public Order clone() {
		try {
			final Order ret = (Order) super.clone();
			return ret;
		} catch (final CloneNotSupportedException e) {
			throw new IllegalStateException("Clone didn's work", e);
		}
	}

	@Override
	public String toString() {
		return template.toString();
	}

	protected MoveResult findNextMove(final Unit unit) {
		MoveResult ret = null;
		if (unit != getCommandable()) {
			throw new IllegalStateException("Wrong unit");
		}
		// Trace backwards
		if (prepareRating()) {
			logger.trace("findNextMove: Calculating backwards rating for " + unit.toString() + " with order " + toString());
			unit.calculateBackwardRating();
			final List<DigitalJoystickDirection> options = unit.calcMoveOptions();
			ret = unit.filterOptions(getGameView(), options, StepRatingPurpose.NORMAL);
		}
		return ret;
	}
}
