package de.ped.empire.logic;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;

/**
 * The order that can be given to a military {@link Unit} to go escort another unit.
 * Depending on the orders of the other unit, this unit moves directly to the other unit, or to the destination of the other unit.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class OrderEscort extends OrderGoTo {

	private static final OrderTemplate TEMPLATE = OrderTemplate.ESCORT;

	/**
	 * Checks before creation of the order object if the order would make sense,
	 * based on the creation parameters.
	 * 
	 * @param gameState
	 *            the game the order will be used in
	 * @param commandable
	 *            the commandable the order will be given to. <code>null</code>
	 *            is not allowed
	 * @param target
	 *            the commandable to escort. <code>null</code> is not allowed
	 * @return <code>true</code> if the order creation makes sense
	 */
	public static boolean makesSense(final GameState gameState, final Commandable commandable, final CommandableKey target) {
		boolean ret = false;
		if (commandable instanceof Unit) {
			final Unit unit = (Unit) commandable;
			if (MilitaryBranch.SEA_UNIT == unit.getUnitTemplate().militaryBranch) {
				final CommandableTemplate targetTemplate = CommandableTemplate.findTemplateById(target.getTemplateId());
				if (MilitaryBranch.SEA_UNIT == targetTemplate.militaryBranch) {
					ret = true;
				}
			}
		}
		return ret;
	}

	public OrderEscort(final GameState gameState, final Commandable commandable, final CommandableKey target) throws CircularReferenceException {
		super(gameState, commandable, TEMPLATE, target);
	}

	OrderEscort(final GameState gameState, final Commandable commandable, final Marshaller marshaller) {
		super(gameState, commandable, TEMPLATE, marshaller);
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		super.marshal(marshaller);
	}

	/** 
	 * @return <code>false</code>, because the escort command is never fulfilled
	 */
	@Override
	public boolean isFulfilled() {
		return false;
	}

	@Override
	public boolean makesSense(final Commandable commandable) {
		boolean ret1 = makesSense(commandable, true);
//		ret1 = orderMakesSense(commandable);
		boolean ret = ret1;
//		if (commandable.getKey().getTemplateId() != ImageKey.CMDL_ID_CITY) {
//			final Unit unit = (Unit) commandable;
//			final StepRatingPosition srp = unit.getStepRatings().get(getPosition());
//			if (null != srp) {
//				ret = true;
//			}
//		}
		if (!ret) {
			logger.trace("super.makesSense returned false");
		} else {
			boolean ret2 = makesSense(getGameState(), commandable, getTargetCommandableKey());
			if (!ret2) {
				logger.trace("this.makesSense returned false");
			}
			ret = ret && ret2;
		}
		return ret;
	}

	private static final int DISTRIBUTION_SIZE = 5;
	private static final int[] DISTRIBUTION = new int[] {
		1, 1, 1, 1, 1, 
		1, 4, 4, 4, 1, 
		1, 4, 1, 4, 1, 
		1, 4, 4, 4, 1, 
		1, 1, 1, 1, 1, 
	};

	@Override
	boolean refreshPosition(final boolean isWithLockedTargetPosition) {
		boolean ret = true;
		if (isWithLockedTargetPosition) {
			ret = super.refreshPosition(false);
		} else {
			final Unit unit = (Unit) getCommandable();
			PlayfieldPosition targetPos = null;
			final CommandableKey targetCommandableKey = getTargetCommandableKey();
			if (null != targetCommandableKey) {
				final GameView gameView = getGameView();
				final GameState gameState = gameView.getGameState();
				final Commandable target = gameState.findCommandable(targetCommandableKey);
				if (0 == target.getHitsTillDeath()) {
					ret = false;
				} else {
					try {
						targetPos = calcTargetPosition();
					} catch (final CircularReferenceException ex) {
						throw new IllegalStateException("CircularReference", ex);
					}
				}
			}
			if (ret) {
				final GameMap map = getGameView().getPlayerMap();
				final int indexOffset = DISTRIBUTION_SIZE / 2;
				final int[] currentDistribution = DISTRIBUTION.clone();
				final PlayfieldPosition[] currentPositions = new PlayfieldPosition[DISTRIBUTION.length];
				int index = 0;
				final PlayfieldDistance dist = new PlayfieldDistance(); // re-used for performance reasons
				for (int y = -indexOffset; y <= indexOffset; y++) {
					final StringBuffer buf = new StringBuffer();
					for (int x = -indexOffset; x <= indexOffset; x++) {
						if (0 < currentDistribution[index]) {
							dist.setLocation(x, y);
							final PlayfieldPosition nearpos = targetPos.add(dist);
							map.getProperties().wrapInside(nearpos);
							currentPositions[index] = nearpos;
							boolean isOkay = false;
							if (!PlayfieldPosition.equals(unit.getPosition(), nearpos) 
									&& !PlayfieldPosition.equals(getTargetCommandable().getPosition(), nearpos)) {
								if (unit instanceof Unit) {
									Unit u = (Unit) unit;
									calculateForwardRating(StepRatingPurpose.NORMAL);
									final StepRatingPosition rating = u.getStepRating(nearpos);
									if (null != rating) {
										isOkay = true;
									}
								}
							}
							if (!isOkay) {
								currentDistribution[index] = 0;
								buf.append("(-,-)");
							} else {
								buf.append(nearpos.toString());
							}
						}
						index++;
					}
					logger.trace("OrderEscort.refreshPosition(): " + buf.toString());
				}
				try {
					setPosition(currentPositions[MathUtil.randomDistributed(currentDistribution)]);
				} catch (final ArrayIndexOutOfBoundsException ex) {
					ret = false;
				}
			}
			logger.trace("OrderEscort.refreshPosition(): unit=" + unit + " target=" + targetCommandableKey + " --> ret=" + ret);
		}
		return ret;
	}

	@Override
	public MoveResult evaluate(final boolean isNow) {
		MoveResult ret = null;
		final Unit unit = (Unit) getCommandable();
		if (isFulfilled()) {
			unit.setOrder(null);
		} else if (!isNow || unit.canMove()) {
			ret = findNextMove();
		}
		return ret;
	}

	@Override
	public I18NStringWithFillIns getToolTip(final Commandable commandable) {
		final Commandable targetUnit = getTargetCommandable();
		final I18NStringWithFillIns part0 = new I18NStringWithFillIns(getMessageKeyPrefix() + Messages.TEXT_POSTFIX, null, null);
		final I18NStringWithFillIns ret;
		if (null != targetUnit) {
			ret = new I18NStringWithFillIns(null, "%0 %1", new I18NStringWithFillIns[] { 
					part0, targetUnit.getName(),
			});
		} else {
			ret = part0;
		}
		return ret;
	}
}
