package de.ped.empire.logic;

import java.util.ArrayList;
import java.util.List;

import de.ped.tools.Marshaller;

/**
 * The order that can be given to a military {@link Unit} to go explore the world.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class OrderExplore extends OrderHome {

	private static final OrderTemplate TEMPLATE = OrderTemplate.EXPLORE;

	public OrderExplore(final GameState gameState, final Commandable commandable) {
		super(gameState, commandable, TEMPLATE);
	}

	OrderExplore(final GameState gameState, final Commandable commandable, final Marshaller marshaller) {
		super(gameState, commandable, TEMPLATE, marshaller);
	}

	/**
	 * If low on fuel, behave like {@link OrderHome}. 
	 * Otherwise, sets the nearest unexplored field as target seed for the step evaluation process. 
	 * If there are several unexplored fields in the same distance, sets them all.
	 * 
	 * @return <code>true</code> if the order goal was reachable and further
	 *         rating makes sense
	 */
	@Override
	public boolean prepareRating() {
		final Unit unit = (Unit) getCommandable();
		final boolean ret;
		if (unit.isLowOnFuel()) {
			ret = super.prepareRating();
		} else {
			calculateForwardRating(StepRatingPurpose.NORMAL);
			final ArrayList<FieldAlike> unexplored = findSomeUnexplored();
			ret = unit.setTargetSeeds(unexplored);
		}
		return ret;
	}

	@Override
	public boolean isFulfilled() {
		calculateForwardRating(StepRatingPurpose.NORMAL);
		final ArrayList<FieldAlike> unexplored = findSomeUnexplored();
		boolean ret = (unexplored.isEmpty());
		return ret;
	}

	@Override
	public MoveResult evaluate(final boolean isNow) {
		final MoveResult ret = super.evaluate(isNow);
		logger.trace("evaluate(" + isNow + ") -> " + ret);
		return ret;
	}

	private ArrayList<FieldAlike> findSomeUnexplored() {
		final Unit unit = (Unit) getCommandable();
		final int maxDistance = Math.min(20, unit.getStepsLeftTillDeath() / 2);
		final ArrayList<FieldAlike> targets = new ArrayList<FieldAlike>();
		int bestNumOfUnexplored = 0;
		int d;
		for (d = 1; (d <= maxDistance) && (targets.isEmpty()); d++) {
			// Collect all fields in air distance to filter them later on
			final List<FieldAlike> targetFieldCandidates = 
					unit.getGameView().getMap().getAllFieldsInAirDistance(unit.getField(), d, true, unit.getSightingRange(MilitaryBranch.NO_UNIT) + 1);
			for (final FieldAlike targetField : targetFieldCandidates) {
				// Filter for fields reachable in <distance> steps by the unit
				final StepRatingPosition stepRating = unit.getStepRating(targetField.getPosition());
				if ((null != stepRating) && (d == stepRating.distanceFromStart) && (!stepRating.isOwnedByEnemy)) {
					// Filter for fields with unexplored fields in the sighting range of the unit
					final List<FieldAlike> unexploredFieldCandidates = 
							unit.getGameView().getMap().getAllFieldsInAirDistance(targetField, unit.getSightingRange(MilitaryBranch.NO_UNIT), true, 0);
					int numOfUnexplored = 0;
					for (final FieldAlike unexploredField : unexploredFieldCandidates) {
						if (unexploredField.isUnexplored()) {
							numOfUnexplored++;
						}
					}
					// Keep the fields that cover the most unexplored area
					if (bestNumOfUnexplored < numOfUnexplored) {
						bestNumOfUnexplored = numOfUnexplored;
						targets.clear();
						targets.add(targetField);
					} else if ((0 < numOfUnexplored) && (bestNumOfUnexplored == numOfUnexplored)) {
						targets.add(targetField);
					}
				}
			}
		}
		logger.trace("OrderExplore.findSomeUnexplored for pos " + getCommandable().getPosition().toString() 
				+ " found " + targets.size() + " targets in distance " + (d - 1) 
				+ " with " + bestNumOfUnexplored + " unexplored fields to cover." 
//				, new IllegalArgumentException("don't worry")
		);
		return targets;
	}
}
