package de.ped.empire.logic;

import java.util.List;

import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Level;

/**
 * The order that can be given to a military {@link Unit} to go to a location on the map.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class OrderGoTo extends Order {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	private static final OrderTemplate TEMPLATE = OrderTemplate.GOTO;

	/**
	 * Creates a relative move path description from a start position to a
	 * target position.
	 * 
	 * @param map
	 *            the map where the move path is defined in
	 * @param startPosition
	 *            the start position
	 * @param targetPosition
	 *            the target position
	 * @return the relative path description
	 */
	public static I18NStringWithFillIns createRelativeMovePathDescription(final GameMap map, final PlayfieldPosition startPosition, final PlayfieldPosition targetPosition) {
		final I18NStringWithFillIns ret;
		final GameProperties props = map.getProperties();
		final PlayfieldDistance distanceVector = props.calcAirDistanceVector(startPosition, targetPosition);
		props.denormDistanceVectorInplace(distanceVector, startPosition);
		DistanceVectorStringPart[] parts = calcDistanceVectorStringParts(distanceVector, props.getTopologyMode());
		final String directionKeyPrefix = "Action.GoTo.Status.Direction.";
		if (null == parts[1]) { // One direction is enough
			ret = new I18NStringWithFillIns("Action.GoTo.Status.Single.Text", null, 
					new Object[] { 
						new I18NStringWithFillIns(directionKeyPrefix + parts[0].getDir(), null, null),
						Integer.toString(parts[0].getDiff()),
					});
		} else { // Two directions needed
			ret = new I18NStringWithFillIns("Action.GoTo.Status.Both.Text", null, 
					new Object[] { 
						new I18NStringWithFillIns(directionKeyPrefix + parts[0].getDir(), null, null),
						Integer.toString(parts[0].getDiff()),
						new I18NStringWithFillIns(directionKeyPrefix + parts[1].getDir(), null, null),
						Integer.toString(parts[1].getDiff()),
					});
		}
		return ret;
	}

	/**
	 * When showing the distance vector in human-readable form, this structure
	 * keeps one part containing the value on a single axis.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class DistanceVectorStringPart {
		public final int diff;
		public final String[] dirs;

		/**
		 * Constructor.
		 * @param ndir the postfix of the message key for the direction if <em>diff</em> is negative
		 * @param pdir the postfix of the message key for the direction if <em>diff</em> is positive
		 * @param diff the distance on this axis
		 */
		public DistanceVectorStringPart(final String ndir, final String pdir, final int diff) {
			this.diff = diff;
			this.dirs = new String[] { ndir, pdir };
		}

		/** @return the postfix of the message key for the direction */
		public String getDir() {
			final String ret;
			if (diff < 0) {
				ret = dirs[0];
			} else if (diff > 0) {
				ret = dirs[1];
			} else {
				ret = null;
			}
			return ret;
		}

		/** @return the distance to show for the direction */
		public int getDiff() {
			return Math.abs(diff);
		}
	}

	public static DistanceVectorStringPart[] calcDistanceVectorStringParts(final PlayfieldDistance distanceVector,
			final TopologyMode topologyMode) {
		int xdiff = distanceVector.getX();
		int ydiff = distanceVector.getY();
		DistanceVectorStringPart[] parts = new DistanceVectorStringPart[] { null, null, }; 
		DistanceVectorStringPart xp;
		DistanceVectorStringPart yp;
		DistanceVectorStringPart zp;
		switch(topologyMode) {
		case HEXAGON: {
			int zdiff = 0;
			if ((xdiff < 0) && (ydiff < 0)) {
				zdiff = Math.max(xdiff, ydiff);
				xdiff -= zdiff;
				ydiff -= zdiff;
			} else if ((xdiff > 0) && (ydiff > 0)) {
				zdiff = Math.min(xdiff, ydiff);
				xdiff -= zdiff;
				ydiff -= zdiff;
			}
			xp = new DistanceVectorStringPart("LDL", "RUR", xdiff);
			yp = new DistanceVectorStringPart("U", "D", ydiff);
			zp = new DistanceVectorStringPart("LUL", "RDR", zdiff);
			if (0 == xdiff) {
				if (0 == ydiff) {
					parts[0] = zp;
				} else if (0 == zdiff) {
					parts[0] = yp;
				} else {
					parts[0] = yp;
					parts[1] = zp;
				}
			} else if (0 == ydiff) {
				if (0 == zdiff) {
					parts[0] = xp;
				} else {
					parts[0] = xp;
					parts[1] = zp;
				}
			} else if (0 == zdiff) {
				parts[0] = xp;
				parts[1] = yp;
			}
			break;
		}
		case HEXAGON_H: {
			int zdiff = 0;
			if ((xdiff < 0) && (ydiff < 0)) {
				zdiff = Math.max(xdiff, ydiff);
				xdiff -= zdiff;
				ydiff -= zdiff;
			} else if ((xdiff > 0) && (ydiff > 0)) {
				zdiff = Math.min(xdiff, ydiff);
				xdiff -= zdiff;
				ydiff -= zdiff;
			}
			xp = new DistanceVectorStringPart("DDL", "UUR", xdiff);
			yp = new DistanceVectorStringPart("L", "R", ydiff);
			zp = new DistanceVectorStringPart("UUL", "DDR", zdiff);
			if (0 == xdiff) {
				if (0 == ydiff) {
					parts[0] = zp;
				} else if (0 == zdiff) {
					parts[0] = yp;
				} else {
					parts[0] = yp;
					parts[1] = zp;
				}
			} else if (0 == ydiff) {
				if (0 == zdiff) {
					parts[0] = xp;
				} else {
					parts[0] = xp;
					parts[1] = zp;
				}
			} else if (0 == zdiff) {
				parts[0] = xp;
				parts[1] = yp;
			}
			break;
		}
		case SQUARE8:
		case SQUARE4:
		default:
			xp = new DistanceVectorStringPart("L", "R", xdiff);
			yp = new DistanceVectorStringPart("U", "D", ydiff);
			if ((0 == xdiff) || (0 == ydiff)) {
				if (0 == xdiff) {
					parts[0] = yp;
				} else {
					parts[0] = xp;
				}
			} else {
				parts[0] = yp;
				parts[1] = xp;
			}
		}
		return parts;
	}

	public OrderGoTo(final GameState gameState, final Commandable commandable, final PlayfieldPosition position) {
		this(gameState, commandable, TEMPLATE, position);
	}

	public OrderGoTo(final GameState gameState, final Commandable commandable, final CommandableKey target) throws CircularReferenceException {
		this(gameState, commandable, TEMPLATE, target);
	}

	OrderGoTo(final GameState gameState, final Commandable commandable, final Marshaller marshaller) {
		this(gameState, commandable, TEMPLATE, marshaller);
	}

	protected OrderGoTo(final GameState gameState, final Commandable commandable, final OrderTemplate template) {
		super(gameState, commandable, template);
		this.targetCommandableKey = null;
	}

	protected OrderGoTo(final GameState gameState, final Commandable commandable, final OrderTemplate template, final PlayfieldPosition targetPosition) {
		this(gameState, commandable, template);
		this.targetPosition.setLocation(targetPosition);
		logger.trace("OrderGoTo(" + targetPosition.toString() + ")");
	}

	protected OrderGoTo(final GameState gameState, final Commandable commandable, final OrderTemplate template, final CommandableKey targetCommmandableKey) throws CircularReferenceException {
		super(gameState, commandable, template);
		this.targetCommandableKey = targetCommmandableKey;
		calcTargetPosition(); // As a side effect throws CircularReferenceException
	}

	protected OrderGoTo(final GameState gameState, final Commandable commandable, final OrderTemplate template, final Marshaller marshaller) {
		super(gameState, commandable, template, marshaller);
		if (!marshaller.readNullIndicator()) {
			targetCommandableKey = de.ped.empire.logic.CommandableKey.Factory.create(marshaller);
		} else {
			targetCommandableKey = null;
		}
		targetPosition = PlayfieldPosition.demarshalWithNull(marshaller);
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		super.marshal(marshaller);
		if (!marshaller.writeNullIndicator(targetCommandableKey)) {
			marshaller.writeMarshallable(targetCommandableKey);
		}
		PlayfieldPosition.marshalWithNull(marshaller, targetPosition);
	}

	/** 
	 * @return the current position to move to, no matter if the target is moving or not
	 */
	PlayfieldPosition getPosition() {
		return targetPosition;
	}

	protected CommandableKey getTargetCommandableKey() {
		return targetCommandableKey;
	}

	protected Commandable getTargetCommandable() {
		return getGameView().getGameState().findCommandable(targetCommandableKey);
	}

	/**
	 * Refreshes the target position based on the target commandable key.
	 * @param isWithLockedTargetPosition ignored in default implementation
	 * @return <code>true</code> if successful. <code>false</code> indicates that the target does no longer exist.
	 */
	boolean refreshPosition(final boolean isWithLockedTargetPosition) {
		boolean ret = true;
		if (null != targetCommandableKey) {
			final GameView gameView = getGameView();
			final GameState gameState = gameView.getGameState();
			final Commandable target = gameState.findCommandable(targetCommandableKey);
			if (0 == target.getHitsTillDeath()) {
				ret = false;
			} else {
				PlayfieldPosition targetPos = null;
				try {
					targetPos = calcTargetPosition();
				} catch (final CircularReferenceException ex) {
					throw new IllegalStateException("CircularReference", ex);
				}
				if (null != targetPos) {
					setPosition(gameState.findCommandable(targetCommandableKey).getPosition());
				}
			}
		}
		return ret;
	}

	protected void setPosition(final PlayfieldPosition position) {
		targetPosition.setLocation(position);
		moveResultCache = null;
	}

	@Override
	public I18NStringWithFillIns getToolTip(final Commandable commandable) {
		final GameView gameView = getGameView();
		final I18NStringWithFillIns ret;
		logger.trace("!!!" + commandable.getKey() + "!!!" + targetCommandableKey);
		if (null != targetCommandableKey) {
			final Commandable targetCmdl = gameView.findCommandable(targetCommandableKey);
			if (isAttack(commandable)) {
				final I18NStringWithFillIns part0 = new I18NStringWithFillIns("Action.Attack.Text", null, null);
				ret = new I18NStringWithFillIns("Action.Attack.Commandable.Text", "%0 %1's %2", 
						new Object[] { 
							part0, 
							getGameView().getPlayer(targetCmdl.getPlayerId()).getName(),
							targetCmdl.getName()
						} );
			} else {
				ret = new I18NStringWithFillIns("Action.GoTo.Status.Unit.Text", null, 
						new I18NStringWithFillIns[] { targetCmdl.getName(), });
			}
		} else {
			final FieldAlike field = gameView.getMap().getAt(targetPosition);
			if (field.isCity()) {
				ret = new I18NStringWithFillIns("Action.GoTo.Status.City.Text", null, 
						new I18NStringWithFillIns[] { field.getField().getName(), });
			} else {
				ret = new I18NStringWithFillIns("Action.GoTo.Status.Position.Text", null, new I18NStringWithFillIns[] {
						createRelativeMovePathDescription(gameView.getMap(), commandable.getPosition(), targetPosition), } );
			}
		}
		return ret;
	}

	private boolean isAttack(final Commandable commandable) {
		logger.trace("!!isAttack: " + targetCommandableKey + "!!" + commandable.getPlayerId()); 
		return ((null != targetCommandableKey)
				&& (targetCommandableKey.getPlayerId() != commandable.getPlayerId()));
	}

	@Override
	public MoveResult evaluate(final boolean isNow) {
		return evaluate(isAttack(getCommandable()), !isNow);
	}

	protected MoveResult evaluate(final boolean isFightAnOption, final boolean isPlanningAhead) {
		MoveResult ret = null;
		final GameView gameView = getGameView();
		final GameMap map = gameView.getPlayerMap();
		final Unit unit = (Unit) getCommandable();
		if (refreshPosition(false)) {
			final GameProperties props = map.getProperties();
			final PlayfieldPosition upos = unit.getPosition();
			final PlayfieldDistance distanceVector = props.calcAirDistanceVector(upos, targetPosition);
			final int distance = props.calcAirDistanceScalar(upos, distanceVector);
			if (0 == distance) { // Command fulfilled
				if (!isPlanningAhead && isFulfilled()) {
					unit.setOrder(null);
				}
			} else if (isPlanningAhead || unit.canMove()) {
				if (1 == distance) { // This is the last step to fulfill the command
					props.denormDistanceVectorInplace(distanceVector, upos);
					final DigitalJoystickDirection dir = DigitalJoystickDirection.getDirection(distanceVector);
					final MoveResult result = evaluateMove(unit, dir, isFightAnOption);
					if (result.getResultMask() == MoveResult.MOVE_MASK_NOTHING) {
						// The move was impossible, e.g., I was loading and the ship was already full
						if (!isPlanningAhead) {
							result.setResultMask(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT);
						}
					}
					ret = result;
					logger.trace("OrderGoTo.evaluate: ret = result in (1 == distance): ret = " + ret);
				} else {
					logger.trace("OrderGoTo.evaluate: 1 < distance for " + unit + " at " + unit.getPosition().toString() + " --> " + targetPosition.toString());
					ret = findNextMove();
					logger.trace("OrderGoTo.evaluate: ret = findNextMove: ret = " + ret);
				}
			}
		} else {
			if (!isPlanningAhead) {
				final MoveResult result = new MoveResult(getGameView(), unit, null);
				result.setResultMask(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT);
				ret = result;
				logger.trace("OrderGoTo.evaluate: ret = result in !isPlanningAhead: ret = " + ret);
			}
		}
		logger.trace("OrderGoTo.evaluate: cmdl=" + unit + " isFightAnOption=" + isFightAnOption + " isPlanningAhead=" + isPlanningAhead + " --> ret = " + ret);
		return ret;
	}

	/**
	 * is a proxy to
	 * {@link GameView#evaluateMove(Unit, DigitalJoystickDirection, boolean, StepRatingPurpose)}
	 * with added caching for performance reasons.
	 * 
	 * @param unit
	 *            the unit to move
	 * @param dir
	 *            the direction to move the unit to
	 * @param isFightAnOption
	 *            <code>true</code> if the unit is allowed to start a fight
	 *            within this move. (Can be <code>false</code> when executing a
	 *            long-lasting {@link Order}.)
	 * @return the result of the evaluation if to move
	 */
	protected MoveResult evaluateMove(
			final Unit unit, 
			final DigitalJoystickDirection dir, 
			final boolean isFightAnOption) {
		final GameView gameView = getGameView();
		final MoveResult ret;
		if (isFightAnOption) { // No cache so far
			ret = gameView.evaluateMove(unit, dir, isFightAnOption, stepRatingPurpose);
		} else {
			if (null == moveResultCache) {
				moveResultCache = gameView.evaluateMove(unit, dir, isFightAnOption, stepRatingPurpose);
			}
			ret = moveResultCache;
		}
		return ret;
	}

	@Override
	public boolean prepareRating() {
		final Unit unit = (Unit) getCommandable();
		calculateForwardRating(stepRatingPurpose);
//		unit.getStepRatings().logDirections(7, 0);
		boolean ret = false;
		if (refreshPosition(false)) {
			ret = unit.setTargetSeed(targetPosition);
		}
		logger.trace("OrderGoTo.prepareRating: --> ret = " + ret);
		return ret;
	}

	/**
	 * Finds the next move to execute for a given destination by doing forward
	 * and backward planning.
	 * 
	 * @return a move result describing the move, or <code>null</code> if no
	 *         next move found
	 */
	protected MoveResult findNextMove() {
		MoveResult ret = null;
		final Unit unit = (Unit) getCommandable();
		if (prepareRating()) {
			final int logLevel = Level.TRACE;
			logger.log(logLevel, "findNextMove: Calculating backwards rating for " + unit.toString() + " with order " + toString());
			unit.calculateBackwardRating();
			final List<DigitalJoystickDirection> options = unit.calcMoveOptions();
			if (logger.isEnabledFor(logLevel)) {
				logger.log(logLevel, "Move options are:");
				for (final DigitalJoystickDirection dir : options) {
					logger.log(logLevel, "- " + dir.key);
				}
			}
			ret = unit.filterOptions(getGameView(), options, stepRatingPurpose);
		}
		return ret;
	}

	@Override
	public boolean makesSense(final Commandable commandable) {
		return makesSense(commandable, false);
	}

	protected boolean makesSense(final Commandable commandable, final boolean isWithLockedTargetPosition) {
		boolean ret = false;
		final GameView gameView = getGameView();
		final GameState gameState = gameView.getGameState();
		if (commandable != getCommandable()) {
			throw new IllegalStateException("Wrong commandable");
		}
		if (isFulfilled()) { // GoTo to same position
			logger.trace("OrderGoTo.makesSense: isFulfilled");
			// Nothing to do
		} else {
			if (super.makesSense(commandable)) {
				final Unit unit = (Unit) commandable;
				boolean isOkay = true;
				if (null != targetCommandableKey) {
					final Commandable targetUnit = gameState.findCommandable(targetCommandableKey);
					if ((null == targetUnit) || (0 == targetUnit.getHitsTillDeath())) {
						isOkay = false;
					} else {
						isOkay = refreshPosition(isWithLockedTargetPosition);
					}
				}
				if (isOkay && gameView.isPossiblePlanAheadMove(unit, targetPosition, true, StepRatingPurpose.AHEAD)) {
					final MoveResult result = evaluate(true, true);
					if ((null != result) && (result.getResultMask() != MoveResult.MOVE_MASK_NOTHING)) {
						ret = true;
					} else {
						if (null == result) {
							logger.trace("OrderGoTo.makesSense: (null != result) was false!!!");
						} else {
							logger.trace("OrderGoTo.makesSense: (result.getResultMask() != MoveResult.MOVE_MASK_NOTHING) was false!!!");
						}
					}
				} else {
					if (!isOkay) {
						logger.trace("OrderGoTo.makesSense: isOkay was false");
					} else {
						logger.trace("OrderGoTo.makesSense: gameView.isPossiblePlanAheadMove was false");
					}
				}
			} else {
				logger.trace("OrderGoTo.makesSense: OrderGoTo.super.makesSense was false");
			}
		}
		logger.trace("OrderGoTo.makesSense: cmdl=" + commandable + " --> ret = " + ret);
		return ret;
	}

	/** 
	 * @return if the target position is reached
	 */
	@Override
	public boolean isFulfilled() {
		boolean ret = false;
		if (PlayfieldPosition.equals(getCommandable().getPosition(), targetPosition)) { // GoTo to same position
			ret = true;
		}
		return ret;
	}

	/**
	 * This command might lead to another unit as target that may have another
	 * unit as target, and so on. So, this method follows this chain till the
	 * end.
	 * 
	 * @return the end of the chain of targeted commands
	 * @throws CircularReferenceException
	 *             if this chain really became a ring and the command should be
	 *             discarded for that
	 */
	protected PlayfieldPosition calcTargetPosition() throws CircularReferenceException {
		final GameView gameView = getGameView();
		final GameState gameState = gameView.getGameState();
		// Find the last commandable in the chain.
		CommandableKey targetKey = targetCommandableKey;
		Commandable target = null;
		while (null != targetKey) {
			target = gameState.findCommandable(targetKey);
			targetKey = null;
			if (null != target) {
				if (target.equals(getCommandable())) {
					throw new CircularReferenceException("Command with circular reference for unit " + target);
				}
				// Follow the chain if the unit is of the same player, and the unit has an order that again is targeted.
				if ((target.getPlayerId() == getCommandable().getPlayerId()) 
						&& (0 < target.getHitsTillDeath())) {
					final Order order = target.getOrder();
					if ((null != order) && (order instanceof OrderGoTo)) {
						final OrderGoTo ogt = (OrderGoTo) order;
						targetKey = ogt.getTargetCommandableKey();
					}
				}
			}
		}
		// Check which position of that last target to use as target position.
		PlayfieldPosition ret = null;
		if (null != target) {
			ret = target.getPosition();
			// Follow the chain if the unit is of the same player, and the unit has an order that again is targeted.
			if (target.getPlayerId() == getCommandable().getPlayerId()) {
				final Order order = target.getOrder();
				if ((null != order) && (order instanceof OrderGoTo)) {
					final OrderGoTo ogt = (OrderGoTo) order;
					if (ogt.refreshPosition(false)) {
						ret = ogt.getPosition();
					}
				}
			}
		}
		final PlayfieldPosition that = ret;
		logger.trace("OrderGoTo.calcTargetPosition: --> ret = " + (null == ret ? "null" : that.toString()));
		return ret;
	}

	@Override
	public String toString() {
		String targetString;
		if (null != targetCommandableKey) {
			targetString = targetCommandableKey.toString();
		} else if (null != targetPosition) {
			final PlayfieldPosition that = targetPosition;
			targetString = that.toString();
		} else {
			targetString = "(null)";
		}
		return super.toString() + targetString;
	}

	void setStepRatingPurpose(final StepRatingPurpose stepRatingPurpose) {
		this.stepRatingPurpose = stepRatingPurpose;
	}

	private StepRatingPurpose stepRatingPurpose = StepRatingPurpose.NORMAL;

	/** The target position to reach with this order. */
	private PlayfieldPosition targetPosition = new PlayfieldPosition();

	/** If not <code>null</code>, this order tries to follow a moving target. */
	private final CommandableKey targetCommandableKey;

	/** Keeps the move result, if already calculated, for performance reasons. */
	private MoveResult moveResultCache = null;
}
