package de.ped.empire.logic;

import java.awt.event.KeyEvent;

import de.ped.tools.CollectionUtil;
import de.ped.tools.Bitfield32;

/**
 * A collection of static information accessible for each order or command.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class OrderTemplate {
	public static final byte TYPE_CLEAR_ALERTS = 0;
	public static final byte TYPE_SKIP_MOVE    = 1;
	public static final byte TYPE_CLEAR_ORDERS = 2;
	public static final byte TYPE_DISBAND      = 3;
	public static final byte TYPE_RENAME       = 4;
	public static final byte TYPE_SENTRY       = 5;
	public static final byte TYPE_LOAD         = 6;
	public static final byte TYPE_UNLOAD       = 7;
	public static final byte TYPE_HOME         = 8;

	public static final byte TYPE_GOTO         = 9;
	public static final byte TYPE_ATTACK       = 10;
	public static final byte TYPE_ESCORT       = 11;
	public static final byte TYPE_PATROL       = 12;
	public static final byte TYPE_EXPLORE      = 13;
	public static final byte TYPE_CHANGE_PRODUCTION = 14;
	public static final byte TYPE_CHANGE_MOVE_PATHS = 15;
	public static final byte TYPE_BUILD_AIRBASE     = 16;

	public static final OrderTemplate CLEAR_ALERTS = new OrderTemplate(TYPE_CLEAR_ALERTS, false, false, false, false, "Action.ClearAlerts", KeyEvent.VK_PERIOD);
	public static final OrderTemplate SKIP_MOVE    = new OrderTemplate(TYPE_SKIP_MOVE,    false, false, false, false, "Action.SkipTurn",    KeyEvent.VK_SPACE);
	public static final OrderTemplate CLEAR_ORDERS = new OrderTemplate(TYPE_CLEAR_ORDERS, true,  false, false, false, "Action.ClearOrders", KeyEvent.VK_MINUS);
	public static final OrderTemplate DISBAND      = new OrderTemplate(TYPE_DISBAND,      true,  false, false, true,  "Action.Disband",     KeyEvent.VK_DELETE);
	public static final OrderTemplate RENAME       = new OrderTemplate(TYPE_RENAME,       true,  false, false, true,  "Action.Rename",      KeyEvent.VK_R);
	public static final OrderTemplate SENTRY       = new OrderTemplate(TYPE_SENTRY,       true,  false, true,  false, "Action.Sentry",      KeyEvent.VK_S);
	public static final OrderTemplate LOAD         = new OrderTemplate(TYPE_LOAD,         true,  false, true,  false, "Action.LoadUnit",    KeyEvent.VK_L);
	public static final OrderTemplate UNLOAD       = new OrderTemplate(TYPE_UNLOAD,       true,  false, false, false, "Action.UnloadUnit",  KeyEvent.VK_U);
	public static final OrderTemplate HOME         = new OrderTemplate(TYPE_HOME,         true,  false, false, false, "Action.Home",        KeyEvent.VK_H);
	public static final OrderTemplate GOTO         = new OrderTemplate(TYPE_GOTO,         true,  true,  false, false, "Action.GoTo",        KeyEvent.VK_G);
	public static final OrderTemplate ATTACK       = new OrderTemplate(TYPE_ATTACK,       true,  true,  false, false, "Action.Attack",      KeyEvent.VK_A);
	public static final OrderTemplate ESCORT       = new OrderTemplate(TYPE_ESCORT,       true,  true,  false, false, "Action.Escort",      KeyEvent.VK_E);
	public static final OrderTemplate PATROL       = new OrderTemplate(TYPE_PATROL,       true,  true,  false, false, "Action.Patrol",      KeyEvent.VK_P);
	public static final OrderTemplate EXPLORE      = new OrderTemplate(TYPE_EXPLORE,      true,  true,  false, false, "Action.Explore",     KeyEvent.VK_X);
	public static final OrderTemplate CHANGE_PRODUCTION = new OrderTemplate(TYPE_CHANGE_PRODUCTION, true, false, false, true, "Action.ChangeProduction", KeyEvent.VK_P);
	public static final OrderTemplate CHANGE_MOVE_PATHS = new OrderTemplate(TYPE_CHANGE_MOVE_PATHS, true, false, false, true, "Action.ChangeMovePaths", 0);
	public static final OrderTemplate BUILD_AIRBASE     = new OrderTemplate(TYPE_BUILD_AIRBASE,     true, false, false, true, "Action.BuildAirbase",  KeyEvent.VK_B);

	public static final OrderTemplate[] INSTANCES = new OrderTemplate[]{
		CLEAR_ORDERS,
		CLEAR_ALERTS,
		DISBAND,
		SKIP_MOVE,
		SENTRY,
		LOAD,
		UNLOAD,
		BUILD_AIRBASE,
		HOME,
		RENAME,
		GOTO,
		ATTACK,
		ESCORT,
		PATROL,
		EXPLORE,
		CHANGE_PRODUCTION,
		CHANGE_MOVE_PATHS,
	};

	public static OrderTemplate findTemplateById(final int unitId) {
		for (final OrderTemplate template : INSTANCES) {
			if (template.id == unitId) {
				return template;
			}
		}
		return null;
	}

	public final byte id;
	public final String messageKeyPrefix;
	public final boolean isOrder;
	public final boolean isTargeted;
	public final boolean isWaiting;
	public final boolean isStartingDialog;
	public final int defaultKeyEvent;

	private OrderTemplate(
			final byte orderType, 
			final boolean isOrder, 
			final boolean isTargeted, 
			final boolean isWaiting, 
			final boolean isStartingDialog, 
			final String messageKeyPrefix, 
			final int defaultKeyEvent) {
		this.id = orderType;
		this.messageKeyPrefix = messageKeyPrefix;
		this.isOrder = isOrder;
		this.isTargeted = isTargeted;
		this.isWaiting = isWaiting;
		this.isStartingDialog = isStartingDialog;
		this.defaultKeyEvent = defaultKeyEvent;
	}

	public boolean canBeOverriddenToBeLoaded() {
		return ((SENTRY == this) || (SKIP_MOVE == this));
	}

	public boolean canBeOverriddenToFollowPath() {
		return canBeOverriddenToBeLoaded();
	}

	/**
	 * Allows for early checks if the creation of an order of that kind makes
	 * any sense at all for the given unit in the given game.
	 * 
	 * @param gameView
	 *            the game played
	 * @param commandable
	 *            the commandable the command shall be given
	 * @return <code>true</code> if the type of command might make sense in that
	 *         context
	 */
	public boolean makesSense(final GameView gameView, final Commandable commandable) {
		boolean ret = false;
		if (gameView.isDefaultPerspective()) {
			final Unit unit = (commandable instanceof Unit ? (Unit) commandable : null);
			final FieldCity city = (commandable instanceof FieldCity ? (FieldCity) commandable : null);
			switch(id) {
			case TYPE_CLEAR_ORDERS:
				if (null != unit) {
					ret = (null != unit.getOrder());
				}
				break;
			case TYPE_CLEAR_ALERTS:
				if (null != commandable) {
					ret = commandable.hasSeenEnemiesLately();
				}
				break;
			case TYPE_DISBAND:
				if (null != unit) {
					ret = true;
				}
				break;
			case TYPE_SKIP_MOVE:
				if (null != unit) {
					if (!gameView.getGameState().isInMoveOver()
							&& (0 != unit.getAttack())) { // Makes no sense for airports or cities.
						ret = true; // Skipping twice is nothing bad
					}
				}
				break;
			case TYPE_SENTRY:
				if (null != unit) {
					// Sentry makes no sense for units in city, and if already in sentry mode.
					ret = true;
					if ((gameView.getPlayerMap().getAt(unit.getPosition()).isCity())
							|| (unit.getOrder() instanceof OrderSentry)) {
						ret = false;
					}
					// Sentry also makes no sense for units with limited range, unless where they can re-fuel.
					if ((unit.getStepsLeftTillDeath() < Integer.MAX_VALUE) && (null == unit.getParentKey())) {
						ret = false;
					}
					// Makes no sense for airports or cities.
					if (0 == unit.getAttack()) {
						ret = false;
					}
				}
				break;
			case TYPE_LOAD:
				if (null != unit) {
					final CommandableTemplate template = unit.getUnitTemplate();
					if ((0 != unit.getAttack()) // Makes no sense for airports or cities.
							&& (template.canLoad(gameView))) {
						final GameMap map = gameView.getPlayerMap();
						final FieldAlike field = map.getAt(unit.getPosition());
						ret = false;
						if (!(field.isCity()) // Not in a city
								&& !(unit.getOrder() instanceof OrderLoad) // Not already in load mode
								&& (0 < unit.calcRemainingCapacity())) { // Capacity left
							// Only makes sense with land nearby
							for(final FieldAlike neighbor : map.getNeighborsOf(field)) {
								if (neighbor.isLand()) {
									ret = true;
									break;
								}
							}
						}
					}
				}
				break;
			case TYPE_UNLOAD:
				if ((null != unit) 
						&& (0 != unit.getAttack()) // Makes no sense for airports or cities.
						&& !(unit.getOrder() instanceof OrderUnload)) {
					final CommandableTemplate template = unit.getUnitTemplate();
					if (template.canLoad(gameView)
							&& (!CollectionUtil.isNullOrEmpty(unit.findChildren()))) {
						ret = true;
					}
				}
				break;
			case TYPE_BUILD_AIRBASE:
				if ((null != unit) 
						&& (0 != unit.getAttack()) // Makes no sense for airports or cities.
						&& !(unit.getOrder() instanceof OrderUnload)) {
					final RulesComplexity rulesComplexity = unit.getGameState().getProperties().getRulesComplexity();
					if ((0 < Templates.instance().AIRBASE.getStepsTillDeath(rulesComplexity)) // Ruleset knows airbases
							&& (unit.getUnitTemplate().militaryBranch == MilitaryBranch.LAND_UNIT) 
							&& (null == unit.getParentKey())
							&& !(unit.getGameState().getMap().getAt(unit.getPosition()).isCity())) {
						ret = true;
					}
				}
				break;
			case TYPE_HOME:
				if ((null != unit)
					&& (0 != unit.getAttack())) { // Makes no sense for airports or cities.
					ret = !OrderHome.isOrderAlreadyFulfilled(gameView, commandable);
				}
				break;
			case TYPE_RENAME:
				if (null != commandable) {
					ret = true;
				}
				break;
			case TYPE_GOTO:
				if ((null != unit) 
						&& (0 != unit.getAttack())) { // Makes no sense for airports or cities.
					ret = true;
				}
				break;
			case TYPE_ATTACK:
				if ((null != unit) 
						&& (0 != unit.getAttack())) { // Makes no sense for airports or cities.
					ret = true;
				}
				break;
			case TYPE_ESCORT:
				if ((null != unit) && (MilitaryBranch.SEA_UNIT == unit.getUnitTemplate().militaryBranch)) {
					ret = true;
				}
				break;
			case TYPE_EXPLORE:
				if (gameView.getProperties().getRulesComplexity().isWithFogOfUncertainty
						&& (null != unit)
						&& (0 != unit.getAttack())) { // Makes no sense for airports or cities.
					ret = true;
				}
				break;
			case TYPE_PATROL:
				if ((null != unit) 
						&& (0 != unit.getAttack())) { // Makes no sense for airports or cities.
					ret = false; // TODO: Implement patrol order
				}
				break;
			case TYPE_CHANGE_PRODUCTION:
				if ((null != city) && (city.getPlayerId() == gameView.getGameState().getCurrentPlayerId())) {
					ret = true;
				}
				break;
			case TYPE_CHANGE_MOVE_PATHS:
				if ((null != city) && (city.getPlayerId() == gameView.getGameState().getCurrentPlayerId())) {
					ret = true;
				} else if ((null != unit) 
						&& Bitfield32.isBit(MapViewMode.PATH.bitmaskUnitsToShow, unit.getUnitTemplate().id)) {
					ret = true;
				}
				break;
			default:
				// Nothing to do
			}
		}
		return ret;
	}

	@Override
	public String toString() {
		return messageKeyPrefix;
	}
}