package de.ped.empire.logic;

import java.awt.Color;
import java.util.HashMap;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayerColor;

/**
 * All that can be said about a player and how to display it. It says nothing about strategy.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Player implements Marshallable, Cloneable {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	/** The maximum of players that can be part of a single game. */
	public static final int MAX_NUMBER_OF_PLAYERS = 6;

	public static final Player PLAYER_NONE = new Player();

	/** The player ID for <em>no player</em>. */
	public static final int ID_NONE = 0xf; //Must be identical to Bitfield32.normalizedBitmask(ImageKey.BA_PLAYER_ID);

	/** The id (0..5) of the player. Persistent. */
	private final int id;
	/** The unique id (0..5) of the coloring of this player. Persistent. */
	private int colorId;
	/** The typeId (human, computer or inactive) of player. Persistent. */
	private int typeId;
	/** The handicap for fights of player. Persistent. */
	private int fightHandicapId;
	/** The handicap for production of player. Persistent. */
	private int productionHandicapId;
	/** The name of the player. Persistent. */
	private String name;
	/** If the player is still part of the game. Persistent. */
	private boolean isPartOfTheGame;
	/** The statistics of the player. Persistent. */
	private HashMap<CommandableTemplate, PlayerUnitStatistics> unitsStatistics = new HashMap<CommandableTemplate, PlayerUnitStatistics>();

	/** 
	 * Creates the neutral player
	 */
	public Player() {
		this.id = ID_NONE;
		this.colorId = ID_NONE;
		this.typeId = PlayerType.NONE.id;
		this.fightHandicapId = PlayerHandicap.NORMAL.id;
		this.productionHandicapId = PlayerHandicap.NORMAL.id;
		this.name = "neutral";
		this.isPartOfTheGame = false;
	}

	/**
	 * Creates one of the 6 possible real players with default values
	 * 
	 * @param id
	 *            the player id
	 */
	public Player(final int id) {
		this.id = id;
		this.colorId = id;
		this.typeId = (id < 2 ? PlayerType.DEFAULT.id : PlayerType.NONE.id);
		this.fightHandicapId = PlayerHandicap.NORMAL.id;
		this.productionHandicapId = PlayerHandicap.NORMAL.id;
		this.name = "Player " + (id + 1);
		this.isPartOfTheGame = false;
	}

	public Player(final int id, final int colIdx, final String name) {
		this.id = id;
		this.colorId = colIdx;
		this.typeId = PlayerType.DEFAULT.id;
		this.fightHandicapId = PlayerHandicap.NORMAL.id;
		this.productionHandicapId = PlayerHandicap.NORMAL.id;
		this.name = name;
		this.isPartOfTheGame = false;
	}

	public Player(final Marshaller marshaller) {
		this.id = marshaller.readInt();
		this.colorId = marshaller.readInt();
		this.typeId = marshaller.readInt();
		this.fightHandicapId = marshaller.readInt();
		this.productionHandicapId = marshaller.readInt();
		this.name = marshaller.readString();
		this.isPartOfTheGame = marshaller.readBoolean();
		// Read the player unit typeId statistics
		final int units = marshaller.readInt();
		for (int i = 0; i < units; i++) {
			final int unitType = marshaller.readInt();
			final PlayerUnitStatistics statistics = new PlayerUnitStatistics();
			marshaller.readMarshallable(statistics);
			unitsStatistics.put(CommandableTemplate.findTemplateById(unitType), statistics);
		}
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(id);
		marshaller.writeInt(colorId);
		marshaller.writeInt(typeId);
		marshaller.writeInt(fightHandicapId);
		marshaller.writeInt(productionHandicapId);
		marshaller.writeString(name);
		marshaller.writeBoolean(isPartOfTheGame);
		// Write the player unit typeId statistics
		marshaller.writeInt(unitsStatistics.size());
		for (final CommandableTemplate template : unitsStatistics.keySet()) {
			marshaller.writeInt(template.id);
			marshaller.writeMarshallable(unitsStatistics.get(template));
		}
	}

	void initForGame(final GameState gameState, final boolean isNewGame) {
		if (isNewGame) {
			isPartOfTheGame = (typeId != PlayerType.NONE.id);
		}
		initUnitsProduced(gameState.getProperties().getRulesComplexity(), isNewGame);
	}

	public boolean isPartOfTheGame() {
		return isPartOfTheGame;
	}

	void removeFromGame() {
		isPartOfTheGame = false;
	}

	@Override
	public Player clone() {
		Player ret = null;
		try {
			ret = (Player) super.clone();
		} catch (final CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return ret;
	}

	private void initUnitsProduced(final RulesComplexity rulesComplexity, final boolean isNewGame) {
		for (final CommandableTemplate template : CommandableTemplate.allUnitsToShowStatsFor(rulesComplexity)) {
			if (!template.isUndefinedOrNothing()) {
				final PlayerUnitStatistics playerUnitStatistics = unitsStatistics.get(template);
				if (null == playerUnitStatistics) {
					unitsStatistics.put(template, new PlayerUnitStatistics());
				} else if (isNewGame) {
					playerUnitStatistics.reset();
				}
			}
		}
	}

	public int getId() {
		return id;
	}

	public int getPlayerColorId() {
		return colorId;
	}

	/**
	 * Sets the color id for this player. 
	 * Does not check (and cannot) if other players might have the same color.
	 * @param colorId the new color id of that player
	 */
	public void setPlayerColorId(final int colorId) {
		this.colorId = colorId;
	}

	PlayerColor getPlayerColor() {
		return PlayerColor.get(colorId);
	}

	public Color getForeground() {
		return getPlayerColor().getColor();
	}

	public Color getBackground() {
		return getPlayerColor().getBackground();
	}

	public String getName() {
		return name;
	}

	public void setName(final String name) {
		this.name = name;
	}

	public int getPlayerTypeId() {
		return typeId;
	}

	public void setPlayerTypeId(final int id) {
		typeId = id;
	}

	public int getHandicapFightId() {
		return fightHandicapId;
	}

	public void setHandicapFightId(final int id) {
		fightHandicapId = id;
	}

	public int getHandicapProductionId() {
		return productionHandicapId;
	}

	public void setHandicapProductionId(final int id) {
		productionHandicapId = id;
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		return new Player(marshaller);
	}

	public int getCntUnitsProduced(final CommandableTemplate template) {
		final PlayerUnitStatistics stat = unitsStatistics.get(template);
		int ret;
		if (null == stat) {
			ret = 0;
		} else {
			ret = stat.getUnitsProduced();
		}
		return ret;
	}

	/**
	 * Keeps track how many units have been produced of that unit template for this player.
	 * This is needed for statistics and unique naming of unnamed units like armies.
	 * @param template the unit template of the unit added
	 * @return the (per game and player) unique id of that unit
	 */
	int addUnit(final CommandableTemplate template) {
		final PlayerUnitStatistics stat = unitsStatistics.get(template);
		return stat.incUnitsProduced();
	}

	public int getCntUnitsKilled(final CommandableTemplate template) {
		final PlayerUnitStatistics stat = unitsStatistics.get(template);
		int ret;
		if (null == stat) {
			ret = 0;
		} else {
			ret = stat.getUnitsKilled();
		}
		return ret;
	}

	void incCntUnitsKilled(final CommandableTemplate template) {
		final PlayerUnitStatistics stat = unitsStatistics.get(template);
		stat.incUnitsKilled();
	}

	public int getCntUnitsLost(final CommandableTemplate template) {
		final PlayerUnitStatistics stat = unitsStatistics.get(template);
		int ret;
		if (null == stat) {
			ret = 0;
		} else {
			ret = stat.getUnitsLost();
		}
		return ret;
	}

	void incCntUnitsLost(final CommandableTemplate template) {
		final PlayerUnitStatistics stat = unitsStatistics.get(template);
		stat.incUnitsLost();
	}

	/** 
	 * @return <code>false</code> if the player type is {@link PlayerType#NONE}, and <em>true</em> otherwise 
	 */
	public boolean isPlayer() {
		return (PlayerType.NONE.id != typeId);
	}

	public boolean isComputerPlayer() {
		return (PlayerType.COMPUTER_SIMPLE.id == typeId);
	}

	public boolean isHumanPlayer() {
		return (isPlayer() && !isComputerPlayer());
	}

	public double getProductionHandicapValue() {
		return PlayerHandicap.values()[productionHandicapId].factor;
	}

	public double getFightHandicapValue() {
		return PlayerHandicap.values()[fightHandicapId].factor;
	}
}
