package de.ped.empire.logic;

import java.util.ArrayList;

import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Keeps all the ratings for all possible positions on the map for planning ahead a command.
 * The rating is calculated twofold:
 * <ol>
 *   <li>
 *     First, a forward analysis is done, to see which fields are reachable at all, and how many steps are necessary to get there.
 *     This is done without any knowledge of the possible targets.
 *   </li>
 *   <li>
 *     Now the strategy can set target seeds, meaning good target locations to aim for. It is in the responsibility of the strategy to only pick targets which are marked as reachable in the previous step.
 *   </li>
 *   <li>
 *     Beginning from those targets, a backwards analysis is done. All paths leading to those targets are marked as good. If a field is part of several paths, the rating increases.
 *   </li>
 * </ol>
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class StepRatingMap {

	/**
	 * Creates an initial step rating map
	 * 
	 * @param unit
	 *            the unit to prepare the step rating map for
	 */
	public StepRatingMap(final Unit unit) {
		this.unit = unit;
		this.gameView = unit.getGameView();
		this.startPosition = unit.getPosition();
		final GameProperties props = gameView.getProperties();
		final PlayfieldDimension size = props.getSize();
		final PlayfieldDimension memSize = props.calcMapMemorySize(size);
		map = new StepRatingPosition[memSize.getWidth()][memSize.getHeight()];
		setAt(startPosition, (StepRatingPosition) StepRatingPosition.START.clone());
		logger = Logger.getLogger(getClass());
	}

	/**
	 * Creates an initial step rating map.
	 * TODO: Handle more than 1 unit in one map, or refactor back to a single unit.
	 * 
	 * @param units
	 *            currently expects an array of size 1, containing the unit to prepare the step rating map for
	 */
	public StepRatingMap(final Unit[] units) {
		// TODO: Fill with life. Usage in transporter/unit loading rule for computer strategy
		this.unit = units[0];
		this.gameView = unit.getGameView();
		this.startPosition = unit.getPosition();
		final PlayfieldDimension size = gameView.getProperties().getSize();
		map = new StepRatingPosition[size.getWidth()][size.getHeight()];
		setAt(startPosition, (StepRatingPosition) StepRatingPosition.START.clone());
		logger = Logger.getLogger(getClass());
	}

	public void setStepRatingPurpose(final StepRatingPurpose stepRatingPurpose) {
		this.stepRatingPurpose = stepRatingPurpose;
	}

	/**
	 * @param dimension
	 *            0 for x, 1 for y
	 * @return the size of the map in the given dimension
	 */
	public int getSize(final int dimension) {
		int ret;
		switch(dimension) {
		case 0:
			ret = map.length;
			break;
		case 1:
			ret = map[0].length;
			break;
		default:
			throw new IllegalArgumentException("Unknown dimension " + dimension);
		}
		return ret;
	}

	public StepRatingPosition get(final int x, final int y) {
		return map[x][y];
	}

	public StepRatingPosition get(final PlayfieldPosition pos) {
		return map[pos.getX()][pos.getY()];
	}

	/**
	 * Does a backward calculation from possible target points back to the unit to move.
	 * All positions which are on a path of the shortest length are marked.
	 * <p>
	 * Precondition: The target point 
	 * (or the nearest target points which must be of equal distance to the unit)
	 *  must already be marked, to work as seed for this algorithm.
	 */
	public void calculateBackwardRating() {
		boolean hasChanged = true;
		final int max = Math.max(map.length, map[0].length);
		for (int step = 0; (step < max) && (hasChanged) && (!get(startPosition).isOnShortestPath()); step++) {
			hasChanged = calculateBackwardStepRatings(step);
		};
	}

	/*
	 *  TODO: For speedup on large maps, set this to e.g. 50, but allow for non step rating based routing (e.g. direct route) on longer distances.
	 *  To avoid endless routing loops (the unit always switches between two positions), 
	 *  this value should still be greater than the hugest land mass that may need routing around.
	 */
	public static final int MAX_RATING_STEPS = Integer.MAX_VALUE;

	/** 
	 * Calculates which positions on the map are reachable and with how many steps.
	 * Forward calculation does not need to know target points, only the starting position.
	 */
	public void calculateForwardRating() {
		final int max = Math.min(Math.max(map.length, map[0].length), MAX_RATING_STEPS);
		for (int step = 0; (step < max) && (!currentForwardRatings.isEmpty()); step++) {
			calculateForwardStepRatings(step);
			final int logLevel = Level.TRACE;
			if (logger.isEnabledFor(logLevel)) {
				logger.log(logLevel, "Step " + step + "; Current forward ratings size: " + currentForwardRatings.size());
				logger.log(logLevel, dump());
			}
		}
	}

	public void cleanBackwardRating() {
		for (int ix = 0; ix < getSize(PlayfieldPosition.X); ix++) {
			for (int iy = 0; iy < getSize(PlayfieldPosition.Y); iy++) {
				final StepRatingPosition posRating = get(ix, iy);
				if (null != posRating) {
					posRating.setTargetSeed(false);
				}
			}
		}
	}

	private void calculateForwardStepRatings(final int step) {
		ArrayList<PlayfieldPosition> prevForwardRatings = currentForwardRatings;
		currentForwardRatings = new ArrayList<PlayfieldPosition>();
		final GameMap map = gameView.getPlayerMap();
		final int stepsLeftTillDeath = unit.getStepsLeftTillDeath();
		for (PlayfieldPosition pos : prevForwardRatings) {
			final StepRatingPosition startRating = this.get(pos);
			if (null != startRating) {
				if (startRating.distanceFromStart == step) { // No need to re-calculate things already visited
					if ((startRating.distanceFromStart < stepsLeftTillDeath)
							&& (!startRating.isOwnedByEnemy) // Don't move further from an enemy field
							&& (!startRating.isHomeBase)) { // Don't move further from a home base field
						for (final DigitalJoystickDirection dir : map.getProperties().getNeighborDirections(pos)) {
							final PlayfieldPosition dpos = map.getProperties().wrap(pos, dir);
							if (ImageKey.FIELD_ID_BORDER != gameView.getGameState().getMap().getAt(dpos).getTemplateId()) {
								final StepRatingPosition destRating = get(dpos);
								if ((null == destRating) || (destRating.distanceFromStart > step)) {
									final boolean isPossibleMoveAhead = gameView.isPossiblePlanAheadMove(unit, dpos, true, stepRatingPurpose);
									if (isPossibleMoveAhead) {
										final boolean isDiagMove = isDiagMove(dir);
										final boolean isEnemy = gameView.isOwnedByEnemy(dpos);
										final boolean isHomeBase = gameView.isHomeBase(unit, dpos);
										final int stepsConsumed = gameView.getMap().calcStepsToConsume(unit.getKey().getTemplateId(), dpos);
										final StepRatingPosition d = new StepRatingPosition(startRating.distanceFromStart + stepsConsumed, dir, isDiagMove, isEnemy, isHomeBase);
										if ((null == destRating) || (0 < d.compareTo(destRating))) {
											if (d.distanceFromStart < step + 1) {
												throw new IllegalStateException("Why the hell do I create a rating of distance " + d.distanceFromStart + " in step " + step + ", winning over " + destRating.distanceFromStart + "?");
											}
											setAt(dpos, d);
											logger.trace(dpos.toString() + "->(" + d.distanceFromStart + "," + d.stepTaken.key + "," + d.forwardRating + ")");
										}
									}
								}
							}
						}
					}
				} else if (startRating.distanceFromStart > step) { 
					// There are positions to move from in a later step, e.g., for tanks near rough
					currentForwardRatings.add(pos);
				}
			}
		}
	}

	protected boolean isDiagMove(final DigitalJoystickDirection dir) {
		// TODO: Unify with backwards rating
		final boolean isDiagMove;
		switch(gameView.getPlayerMap().getProperties().getTopologyMode()) {
		case SQUARE8:
			if (dir.is4Direction()) {
				isDiagMove = false;
			} else {
				isDiagMove = true;
			}
			break;
		default:
			isDiagMove = false;
		}
		return isDiagMove;
	}

	private boolean calculateBackwardStepRatings(final int step) {
		boolean ret = false;
		final GameMap map = gameView.getPlayerMap();
		final PlayfieldPosition pos = new PlayfieldPosition(); //re-used for performance reasons
		for (int ix = 0; ix < getSize(PlayfieldPosition.X); ix++) {
			pos.setX(ix);
			for (int iy = 0; iy < getSize(PlayfieldPosition.Y); iy++) {
				pos.setY(iy);
				final StepRatingPosition startRating = this.get(pos);
				if ((null != startRating) && (startRating.isOnShortestPath())) {
					final int stepsConsumed = gameView.getMap().calcStepsToConsume(unit.getKey().getTemplateId(), pos);
					for (final DigitalJoystickDirection dir : map.getProperties().getNeighborDirections(pos)) {
						final PlayfieldPosition dpos = map.getProperties().wrap(pos, dir);
						final StepRatingPosition destRating = this.get(dpos);
						if ((null != destRating) 
								&& ((!destRating.isHomeBase) || (0 == destRating.distanceFromStart))
								&& (destRating.distanceFromStart + stepsConsumed == startRating.distanceFromStart)) {
							ret = true;
							destRating.addBackwardRating(startRating, isDiagMove(dir));
						}
					}
				}
			}
		}
		return ret;
	}

	/**
	 * Prints the map to a string. Both step taken and the forward rating is printed out.
	 * @param size the size in fields to show
	 * @param min the first field to show
	 * @return a loggable representation of (parts of) the rating map. Lines are separated by the newline character.
	 */
	public String getLogDirections(final int size, final int min) {
		final StringBuffer sb = new StringBuffer();
		for (int y = min; y < min + size; y++) {
			for (int x = min; x < min + size; x++) {
				final StepRatingPosition pos = get(x, y);
				if (null == pos) {
					sb.append('.');
				} else {
					sb.append(pos.getStepTakenIndicator());
				}
			}
			sb.append("  ");
			for (int x = min; x < min + size; x++) {
				final StepRatingPosition pos = get(x, y);
				if (null == pos) {
					sb.append(" null ");
				} else {
					sb.append(TextFormatter.format(pos.forwardRating, 6, 1));
				}
			}
			sb.append("\n");
		}
		return sb.toString();
	}

	/**
	 * Dumps the whole step rating map, including forward and backwards rating.
	 * @return a very long string containing the whole step rating map
	 */
	public String dump() {
		String ret = stepRatingPurpose.toString() + "\n";
		final int LW = 3;
		final int WIDTH = 24;
		final String DELIM = "|";
		ret += TextFormatter.format("", LW);
		ret += DELIM;
		for (int x = 0; x < map.length; x++) {
			ret += TextFormatter.format(x, WIDTH, TextFormatter.CENTER);
			ret += DELIM;
		}
		ret += "\n";
		for (int y = 0; y < map[0].length; y++) {
			ret += TextFormatter.format(y, LW).toString();
			ret += DELIM;
			for (int x = 0; x < map.length; x++) {
				StepRatingPosition pos = map[x][y];
				ret += (null == pos ? TextFormatter.format("null", WIDTH, TextFormatter.CENTER) : pos.toString()) + DELIM;
			}
			ret += "\n";
		}
		return ret;
	}

	private void setAt(final PlayfieldPosition pos, final StepRatingPosition value) {
		map[pos.getX()][pos.getY()] = value;
		currentForwardRatings.add(pos.clone());
	}

	private Logger logger;
	private StepRatingPosition[][] map;
	private final GameView gameView;
	private final Unit unit;
	private final PlayfieldPosition startPosition;
	private StepRatingPurpose stepRatingPurpose = StepRatingPurpose.NORMAL;

	/**
	 * For optimization purposes on large maps: Contains all positions which are
	 * worth starting from in the forward rating steps to come. So only starting
	 * from those positions, new steps need to be evaluated.
	 */
	private ArrayList<PlayfieldPosition> currentForwardRatings = new ArrayList<PlayfieldPosition>();
}
