package de.ped.empire.logic;

import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.MathUtil;
import de.ped.tools.TextFormatter;

/**
 * Keeps the rating for a single step of planning ahead a command.
 * The rating is calculated twofold:
 * <ol>
 *   <li>
 *     First, a forward analysis is done, to see which fields are reachable at all, and how many steps are necessary to get there.
 *     This is done without any knowledge of the possible targets.
 *   </li>
 *   <li>
 *     Now the strategy can set target seeds, meaning good target locations to aim for. It is in the responsibility of the strategy to only pick targets which are marked as reachable in the previous step.
 *   </li>
 *   <li>
 *     Beginning from those targets, a backwards analysis is done. All paths leading to those targets are marked as good. If a field is part of several paths, the rating increases.
 *   </li>
 * </ol>
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class StepRatingPosition implements Comparable<StepRatingPosition>, Cloneable {
	/** The rating for the start position. Be careful to use a clone when inserting into a map, as following operations are likely to change it. */
	public static final StepRatingPosition START = new StepRatingPosition(0, null, false, false, false);

	/** The backward rating for a horizontal or vertical move. Normally, a diagonal move makes more ground and is preferred. */
	static final double BACKWARD_RATING_4DIR_MOVE = 1.2;
	/** The backward rating for a diagonal move. Normally, a diagonal move makes more ground and is preferred. */
	static final double BACKWARD_RATING_DIAG_MOVE = 1.0;
	/** The forward rating for the start position. */
	static final double FORWARD_RATING_START = 9.9;
	/** The forward rating for a horizontal or vertical move. */
	static final double FORWARD_RATING_4DIR_MOVE = 1.2;
	/** The forward rating for a diagonal move. If an up/down/left/right move suffices, this is preferred. */
	static final double FORWARD_RATING_DIAG_MOVE = 1.0;

	/** The direction the unit would have taken to reach this position (forward calculated) */
	public final DigitalJoystickDirection stepTaken;

	/** If <code>true</code>, this position is owned by an enemy */
	public final boolean isOwnedByEnemy;

	/** If <code>true</code>, this position is a home base, e.g., own city, airbase, carrier. */
	public final boolean isHomeBase;

	/** The distance from the current position of the unit (forward calculated) */
	public final int distanceFromStart;

	/** The forward rating of this move (forward calculated) */
	public final double forwardRating;

	/** The distance from the nearest target (backward calculated) */
	private int distanceFromNearestTarget;

	/** The backward rating of this move (backward calculated)*/
	private double backwardRating;

	/**
	 * Initializes a rating for a position on the field, based on forward
	 * rating. So the existence of such a rating indicates that this field is
	 * reachable by the unit.
	 * 
	 * @param distance
	 *            the distance from the current position of the unit
	 * @param stepTaken
	 *            the direction the unit would have taken to reach this position
	 * @param isDiagMove
	 *            if <code>true</code>, this move is a diagonal move in a topology where this makes sense
	 * @param isOwnedByEnemy
	 *            if <code>true</code>, this position is owned by an enemy
	 * @param isHomeBase
	 *            if <code>true</code>, this position is a home base for the unit
	 */
	public StepRatingPosition(
			final int distance, 
			final DigitalJoystickDirection stepTaken, 
			final boolean isDiagMove,
			final boolean isOwnedByEnemy, 
			final boolean isHomeBase) {
		this.distanceFromStart = distance;
		this.stepTaken = stepTaken;
		this.isOwnedByEnemy = isOwnedByEnemy;
		this.isHomeBase = isHomeBase;
		final double rating;
		if (null == stepTaken) {
			rating = FORWARD_RATING_START;
		} else if (isDiagMove) {
			rating = FORWARD_RATING_DIAG_MOVE;
		} else {
			rating = FORWARD_RATING_4DIR_MOVE;
		}
		this.forwardRating = rating;
		setTargetSeed(false);
	}

	@Override
	public StepRatingPosition clone() {
		StepRatingPosition ret = null;
		try {
			ret = (StepRatingPosition) super.clone();
		} catch (final CloneNotSupportedException e) {
			e.printStackTrace();
		}
		return ret;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + distanceFromStart;
		{
			long temp;
			temp = Double.doubleToLongBits(forwardRating);
			result = prime * result + (int) (temp ^ (temp >>> 32));
		} 
		result = prime * result + distanceFromNearestTarget;
		{
			long temp;
			temp = Double.doubleToLongBits(backwardRating);
			result = prime * result + (int) (temp ^ (temp >>> 32));
		}
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		StepRatingPosition other = (StepRatingPosition) obj;
		if (distanceFromStart != other.distanceFromStart)
			return false;
		if (Double.doubleToLongBits(forwardRating) != Double.doubleToLongBits(other.forwardRating))
			return false;
		if (distanceFromNearestTarget != other.distanceFromNearestTarget)
			return false;
		if (Double.doubleToLongBits(backwardRating) != Double.doubleToLongBits(other.backwardRating))
			return false;
		return true;
	}

	/**
	 * Compares <em>this</em> to an<em>other</em> step rating.
	 * It works both for forward rating (because at the time of forward rating the backward rating is the same for all ratings) and backwards rating.
	 * It rates based on the following rules:
	 * <ol>
	 *   <li>
	 *     Better backward rating wins first; if backwards rating is equal, better forward rating wins.
	 *   </li>
	 *   <li>
	 *     For both forward and backward rating, first, shorter distance wins. On equal distance, better details rating wins.
	 *   </li>
	 *   <li>
	 *     <code>null</code> always is less than any existing object.
	 *   </li>
	 * </ol>
	 * 
	 * @param other the object to compare with
	 */
	@Override
	public int compareTo(final StepRatingPosition other) {
		if (distanceFromNearestTarget != other.distanceFromNearestTarget) {
			// For distance, small is better
			return other.distanceFromNearestTarget - distanceFromNearestTarget;
		} else if (backwardRating != other.backwardRating) {
			// For backwrdRating, big is better
			return MathUtil.sgn(backwardRating - other.backwardRating);
		} else if (distanceFromStart != other.distanceFromStart) {
			// For distance, small is better
			return other.distanceFromStart - distanceFromStart;
		} else {
			// For forwardRating, big is better
			return MathUtil.sgn(forwardRating - other.forwardRating);
		}
	}

	int getDistanceFromNearestTarget() {
		return distanceFromNearestTarget;
	}

	/**
	 * @return if <em>this</em> position is part of one of the shortest paths.
	 *         Only has a chance to return <code>true</code> if backward rating
	 *         was executed.
	 */
	public boolean isOnShortestPath() {
		return (0.0 < backwardRating);
	}

	public double getBackwardRating() {
		return backwardRating;
	}

	/**
	 * Sets or unsets target seeds, before backtracking takes place.
	 * 
	 * @param isTargetSeed
	 *            if <code>true</code>, sets <em>this</em> as target seed;
	 *            otherwise unsets <em>this</em> as target seed.
	 */
	public void setTargetSeed(final boolean isTargetSeed) {
		if (isTargetSeed) {
			backwardRating = 1.0;
			distanceFromNearestTarget = 0;
		} else {
			backwardRating = 0.0;
			distanceFromNearestTarget = Integer.MAX_VALUE;
		}
	}

	/**
	 * While in the same distance from the target, several possible
	 * nearly-optimal paths leading through this position may increase the
	 * rating even more than the optimal path. So this method tells that there
	 * is one (more) nearly-optimal path.
	 * 
	 * @param sourceRating
	 *            the step rating of the position one step nearer to the path
	 *            target
	 * @param isDiagMove
	 *            if <code>true</code>, the position of <em>sourceRating</em> is
	 *            a diagonal move in a topology where this makes sense
	 */
	public void addBackwardRating(final StepRatingPosition sourceRating, final boolean isDiagMove) {
		if (distanceFromNearestTarget >= sourceRating.distanceFromNearestTarget + 1) {
			distanceFromNearestTarget = sourceRating.distanceFromNearestTarget + 1;
			backwardRating += sourceRating.getBackwardRating() * (isDiagMove ? BACKWARD_RATING_DIAG_MOVE : BACKWARD_RATING_4DIR_MOVE);
		}
	}

	/**
	 * @return a character that helps humans to understand the direction of the
	 *         step taken, e.g., in maps logged for debugging purposes
	 */
	public char getStepTakenIndicator() {
		char ret;
		final DigitalJoystickDirection dir = stepTaken;
		if (DigitalJoystickDirection.getDirection(1, 1) == dir) {
			ret = '\\';
		} else if (DigitalJoystickDirection.getDirection(-1, -1) == dir) {
			ret = '\\';
		} else if (DigitalJoystickDirection.getDirection(1, -1) == dir) {
			ret = '/';
		} else if (DigitalJoystickDirection.getDirection(-1, 1) == dir) {
			ret = '/';
		} else if (DigitalJoystickDirection.getDirection(1, 0) == dir) {
			ret = '-';
		} else if (DigitalJoystickDirection.getDirection(-1, 0) == dir) {
			ret = '-';
		} else if (DigitalJoystickDirection.getDirection(0, -1) == dir) {
			ret = '|';
		} else if (DigitalJoystickDirection.getDirection(0, 1) == dir) {
			ret = '|';
		} else {
			ret = 'o';
		}
		return ret;
	}

	@Override
	public String toString() {
		return "(b(" 
				+ (Integer.MAX_VALUE == distanceFromNearestTarget ? "---" : TextFormatter.format(distanceFromNearestTarget, 3)) 
				+ "," + TextFormatter.format(backwardRating, 3, 1) 
				+ "),f=(" 
				+ TextFormatter.format(distanceFromStart, 3) + "," 
				+ TextFormatter.format(forwardRating, 3, 1) + "))";
	}
}