package de.ped.empire.logic;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Shape;
import java.awt.image.BufferedImage;
import java.awt.image.RasterFormatException;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.HashMap;

import javax.imageio.ImageIO;

import de.ped.tools.BitArea;
import de.ped.tools.Bitfield32;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.PlayerColor;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Anything that can be drawn in a tile on the map. Knows how to be drawn, but not where.
 * Could be e.g. a field, a unit, a fight overlay or a crosshair.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class TileTemplate implements Tileable {

	/**
	 * Some basic data for any type of tile,
	 * mainly to be able to find the right map icon tile for a given {@link ImageKey}.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class TileTemplateData {
		/** The x position of the basic tile for that tile object type. */
		public final int iconX;
		/** The y position of the basic tile for that tile object type. */
		public final int iconY;
		/** The the translation vector for the alternative look tile for that tile object type. */
		public final Point iconModifier;
		/** If <code>true</code>, the white pixels from the mapicons resource files should get converted to transparent. */
		public final boolean isWhiteTransparent;
		/** The character to represent this type of tile for logging a map of tiles. */
		public final char logRepresentation;
		/** Defines which elements of the mask make sense for this type of tile. */
		public final int normalizationMask;

		/**
		 * Creates the data set identifying where in the mapicons resource files
		 * to look for these kind of image templates.
		 * 
		 * @param iconX
		 *            the x position of the basic tile for that tile object type
		 * @param iconY
		 *            the y position of the basic tile for that tile object type
		 * @param iconModifier
		 *            the translation vector for the alternative look tile for that tile
		 *            object type
		 * @param logRepresentation
		 *            the representation of this object in the logs
		 * @param isWhiteTransparent
		 *            if <em>true</em>, the white pixels from the mapicons
		 *            resource files should get converted to transparent
		 * @param areas
		 *            the list of bit areas of the bitfield used for this
		 */
		public TileTemplateData(
				final int iconX, 
				final int iconY, 
				final Point iconModifier, 
				final char logRepresentation,
				final boolean isWhiteTransparent,
				final BitArea[] areas) {
			this.iconX = iconX;
			this.iconY = iconY;
			this.iconModifier = iconModifier;
			this.logRepresentation = logRepresentation;
			this.isWhiteTransparent = isWhiteTransparent;
			this.normalizationMask = Bitfield32.bitfield(areas);
		}

		@Override
		public String toString() {
			return "" + logRepresentation + "(" + iconX + "," + iconY + ")";
		}
	}

	public static final Point TRANS_VECTOR_NONEMPTY = new Point(0, 1);
	public static final Point TRANS_VECTOR_LOOK = new Point(0, 8);
	public static final Point TRANS_VECTOR_NONE = new Point(0, 0);

	private static HashMap<ImageKey, TileTemplateData> TTD_MAP = null;

	/**
	 * @param imageKey
	 *            the image key
	 * @return the {@link TileTemplateData} object fitting for the
	 *         <em>imageKey</em>
	 */
	public static synchronized TileTemplateData getTTD(final int imageKey) {
		if (null == TTD_MAP) {
			final BitArea[] bitAreaNothing = new BitArea[] {
					ImageKey.BA_TTYPE,
					ImageKey.BA_TEMPLATE,
			};
			final BitArea[] bitAreasNormalField = new BitArea[] {
					ImageKey.BA_TTYPE,
					ImageKey.BA_TEMPLATE,
					ImageKey.BA_OUTDATED,
					ImageKey.BA_UNEXPLORED_MOD,
					ImageKey.BA_LOOK,
					ImageKey.BA_NEIGHBOR_MOD,
			};
			final BitArea[] bitAreasNormalUnit = new BitArea[] {
					ImageKey.BA_TTYPE,
					ImageKey.BA_TEMPLATE,
					ImageKey.BA_OUTDATED,
					ImageKey.BA_PLAYER_COLOR,
					ImageKey.BA_SENTRIED,
			};
			final BitArea[] bitAreasLoadableUnit = new BitArea[] {
					ImageKey.BA_TTYPE,
					ImageKey.BA_TEMPLATE,
					ImageKey.BA_OUTDATED,
					ImageKey.BA_PLAYER_COLOR,
					ImageKey.BA_SENTRIED,
					ImageKey.BA_NONEMPTY,
			};
			final int ux = 6*8+2;
			final TileTemplateData dodCity = new TileTemplateData( ux-1,  0, TRANS_VECTOR_LOOK, '$', false,
					new BitArea[] {
						ImageKey.BA_TTYPE,
						ImageKey.BA_TEMPLATE,
						ImageKey.BA_OUTDATED,
						ImageKey.BA_UNEXPLORED_MOD,
						ImageKey.BA_LOOK,
						ImageKey.BA_NONEMPTY,
						ImageKey.BA_PLAYER_COLOR,
			});
			final HashMap<ImageKey, TileTemplateData> builder = new HashMap<ImageKey, TileTemplateData>(); 
			builder.put(ImageKey.CROSSHAIR,
					new TileTemplateData(6*8,  2, TRANS_VECTOR_NONE, ' ', false, bitAreaNothing));
			builder.put(ImageKey.FIGHT,
					new TileTemplateData(6*8,  5, TRANS_VECTOR_NONE, ' ', true,
							new BitArea[] {
								ImageKey.BA_TTYPE,
								ImageKey.BA_TEMPLATE,
								ImageKey.BA_NEIGHBOR_MOD, // Side-used for animation step
					}));
			builder.put(ImageKey.UNEXPLORED_NEIGHBOR,
					new TileTemplateData(8*0, 8*0, TileTemplate.TRANS_VECTOR_NONE, ' ', true,
							new BitArea[] {
								ImageKey.BA_TTYPE,
								ImageKey.BA_TEMPLATE,
								ImageKey.BA_NEIGHBOR_MOD,
					}));
			builder.put(ImageKey.OUTDATED,
					new TileTemplateData(8*6, 4, TileTemplate.TRANS_VECTOR_NONE, ' ', true, bitAreaNothing));
			builder.put(ImageKey.UNEXPLORED,
					new TileTemplateData(6*8,  3, TRANS_VECTOR_NONE, ' ', false, bitAreaNothing));
			builder.put(ImageKey.BORDER,
					new TileTemplateData(6*8,  1, TRANS_VECTOR_NONE, '#', false,
							new BitArea[] {
								ImageKey.BA_TTYPE,
								ImageKey.BA_TEMPLATE,
								ImageKey.BA_OUTDATED,
								ImageKey.BA_UNEXPLORED_MOD,
					}));
			builder.put(ImageKey.SEA,
					new TileTemplateData( 1*8,  0, TRANS_VECTOR_LOOK, '~', false, bitAreasNormalField));
			builder.put(ImageKey.LAND,
					new TileTemplateData(6*8,  0, TRANS_VECTOR_LOOK, '.', false,
							new BitArea[] {
								ImageKey.BA_TTYPE,
								ImageKey.BA_TEMPLATE,
								ImageKey.BA_OUTDATED,
								ImageKey.BA_UNEXPLORED_MOD,
								ImageKey.BA_LOOK,
					}));
			builder.put(ImageKey.ROUGH,
					new TileTemplateData(4*8,  0, TRANS_VECTOR_LOOK, ':', false, bitAreasNormalField));
			builder.put(ImageKey.RIVER,
					new TileTemplateData(2*8,  0, TRANS_VECTOR_LOOK, '+', false, bitAreasNormalField));
			builder.put(ImageKey.FOREST,
					new TileTemplateData(3*8,  0, TRANS_VECTOR_LOOK, '*', false, bitAreasNormalField));
			builder.put(ImageKey.MOUNTAIN,
					new TileTemplateData(5*8,  0, TRANS_VECTOR_LOOK, 'o', false, bitAreasNormalField));
			builder.put(ImageKey.CITY,
					dodCity);
			builder.put(new ImageKey(ImageKey.TType.Unit,  ImageKey.CMDL_ID_CITY),
					dodCity);
			builder.put(ImageKey.ARMY,        
					new TileTemplateData(ux+ 0, 0, TRANS_VECTOR_NONE, 'a', false, bitAreasNormalUnit));
			builder.put(ImageKey.TANK,        
					new TileTemplateData(ux+ 1, 0, TRANS_VECTOR_NONE, 't', false, bitAreasNormalUnit));
			builder.put(ImageKey.FIGHTER,      
					new TileTemplateData(ux+ 2, 0, TRANS_VECTOR_NONE, 'b', false, bitAreasNormalUnit));
			builder.put(ImageKey.BOMBER,     
					new TileTemplateData(ux+ 3, 0, TRANS_VECTOR_NONE, 'f', false, bitAreasNormalUnit));
			builder.put(ImageKey.TRANSPORTER, 
					new TileTemplateData(ux+ 4, 0, TRANS_VECTOR_NONE, 'T',false, bitAreasLoadableUnit));
			builder.put(ImageKey.DESTROYER,   
					new TileTemplateData(ux+ 5, 0, TRANS_VECTOR_NONE, 'D', false, bitAreasNormalUnit));
			builder.put(ImageKey.SUBMARINE,
					new TileTemplateData(ux+ 6, 0, TRANS_VECTOR_NONE, 'S', false, bitAreasNormalUnit));
			builder.put(ImageKey.CRUISER,     
					new TileTemplateData(ux+ 7, 0, TRANS_VECTOR_NONE, 'c', false, bitAreasNormalUnit));
			builder.put(ImageKey.BATTLESHIP,  
					new TileTemplateData(ux+ 8, 0, TRANS_VECTOR_NONE, 'B', false, bitAreasNormalUnit));
			builder.put(ImageKey.CARRIER,     
					new TileTemplateData(ux+ 9, 0, TRANS_VECTOR_NONE, 'C', false, bitAreasLoadableUnit));
			builder.put(ImageKey.AIRBASE,
					new TileTemplateData(ux+10, 0, TRANS_VECTOR_NONE, 'A', false,
							new BitArea[] {
								ImageKey.BA_TTYPE,
								ImageKey.BA_TEMPLATE,
								ImageKey.BA_OUTDATED,
								ImageKey.BA_UNEXPLORED_MOD,
								ImageKey.BA_PLAYER_COLOR,
								ImageKey.BA_NONEMPTY,
					}));
			builder.put(ImageKey.UNDEFINED,
					new TileTemplateData(ux+0, 1, TRANS_VECTOR_NONE, '?', false,
							new BitArea[] {
								ImageKey.BA_TTYPE,
								ImageKey.BA_TEMPLATE,
					}));
			TTD_MAP = builder;
			logger.debug("DOD_MAP built with " + TTD_MAP.size() + " elements.");
			for (ImageKey key : builder.keySet()) {
				logger.debug("DOD_MAP contains 0x" + Integer.toHexString(key.getAll()));
			}
		}
		final int typeMask = ImageKey.bitmask(ImageKey.BA_TTYPE) | ImageKey.bitmask(ImageKey.BA_TEMPLATE);
		final int typeKey = imageKey & typeMask;
		final TileTemplateData ret = TTD_MAP.get(new ImageKey(typeKey));
		if (null == ret) {
			throw new IllegalArgumentException("No DOD found for image key 0x" + Integer.toHexString(imageKey));
		}
		return ret;
	}

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	protected static final int TEMPLATE_PLAYER_FOREGROUND_RGB = 0xff0000;
	protected static final int TEMPLATE_PLAYER_BACKGROUND_RGB = 0x800000;

	/**
	 * Loads an image from the resources.
	 * 
	 * @param resourceName
	 *            the name of the resource
	 * @return the image
	 * @throws IOException
	 *             if loading went wrong
	 */
	public static BufferedImage loadImage(final String resourceName) throws IOException {
		final URL url = Resources.getURLToResource(resourceName);
		final BufferedInputStream is = new BufferedInputStream(url.openStream());
		final BufferedImage ret = ImageIO.read(is);
		return ret;
	}

	protected static Logger logger = Logger.getLogger("DrawableObject");

	/**
	 * Paints the object in its current state
	 * 
	 * @param g
	 *            where to paint to
	 * @param offset
	 *            where to start painting in <em>g</em>
	 * @param imageKey
	 *            the image key of the image to paint
	 * @param pos
	 *            the map position
	 * @param zoomLevel
	 *            the zoom level to paint the object for
	 * @param topologyMode
	 *            the topology mode to paint the object for
	 */
	public static void paintTile(final Graphics g, final Point offset, final int imageKey, final PlayfieldPosition pos, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		paintTile(g, offset, imageKey, pos.getX(), pos.getY(), false, zoomLevel, topologyMode);
	}

	/**
	 * Paints the object in its current state
	 * 
	 * @param g
	 *            where to paint to
	 * @param offset
	 *            where to start painting in <em>g</em>
	 * @param imageKey
	 *            the image key of the image to paint
	 * @param zoomLevel
	 *            the zoom level to paint the object for
	 * @param topologyMode
	 *            the topology mode to paint the object for
	 */
	public static void paintTile(final Graphics g, final Point offset, final int imageKey, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		paintTile(g, offset, imageKey, 0, 0, false, zoomLevel, topologyMode);
	}

	/**
	 * Paints the object in its current state
	 * 
	 * @param g
	 *            where to paint to
	 * @param offset
	 *            where to start painting in <em>g</em>
	 * @param imageKey
	 *            the image key of the image to paint
	 * @param x
	 *            the map x-position
	 * @param y
	 *            the map y-position
	 * @param zoomLevel
	 *            the zoom level to paint the object for
	 * @param topologyMode
	 *            the topology mode to paint the object for
	 * @param isWithClip
	 *            if <code>true</code>, the rectangle is clipped to the floor
	 *            tile.
	 */
	public static void paintTile(final Graphics g, final Point offset, final int imageKey, final float x, final float y, final boolean isWithClip, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		final BufferedImage image = createImage(imageKey, zoomLevel, topologyMode);
		if (null == image) {
			throw new IllegalArgumentException("No image found for image key " + ImageKey.toString(imageKey));
		} else {
			final Shape clip = g.getClip();
			final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(topologyMode);
			if (isWithClip) {
				g.setClip(
						offset.x + ((int) Math.floor(x)) * playfieldTiling.getSpaceWidth(), 
						offset.y + ((int) Math.floor(y)) * playfieldTiling.getSpaceHeight(), 
						playfieldTiling.getSurroundingWidth(), playfieldTiling.getSurroundingHeight());
			}
			g.drawImage(image, 
					offset.x + Math.round(x * playfieldTiling.getSpaceWidth()), 
					offset.y + Math.round(y * playfieldTiling.getSpaceHeight()), 
					playfieldTiling.getSurroundingWidth(), playfieldTiling.getSurroundingHeight(), null);
			if (isWithClip) {
				g.setClip(clip);
			}
		}
	}

	/**
	 * Loads the image for this kind of field and player.
	 * 
	 * @param zoom
	 *            the zoom level to create the image for
	 * @param topologyMode
	 *            the topology mode to create the image for
	 * @param imageKey
	 *            the image key to load the image for
	 * @return the image
	 */
	public static BufferedImage createImage(final int imageKey, final ZoomLevel zoom, final TopologyMode topologyMode) {
		BufferedImage ret = null;
		final TileTemplateData dod = getTTD(imageKey);
		final ImageKey ikey = new ImageKey(imageKey);
		final ImageSet imageSet = zoom.getImageSet(topologyMode);
		final PlayfieldTiling playfieldTiling = zoom.getPlayfieldTiling(topologyMode);
		if (null == dod) {
			ret = imageSet.getTransparentImage(); 
		} else {
			ret = imageSet.getImage(ikey); 
			if (null == ret) {
				// Find and load the right tile from the mapicons resource file.
				int x = dod.iconX;
				int y = dod.iconY;
				if (0 != ikey.getBits(ImageKey.BA_NONEMPTY)) {
					x += TRANS_VECTOR_NONEMPTY.x;
					y += TRANS_VECTOR_NONEMPTY.y;
				}
				final int iconModifier = ikey.getBits(ImageKey.BA_LOOK);
				y += iconModifier * dod.iconModifier.y;
				x += iconModifier * dod.iconModifier.x;
				final int iconNeighborCode = ikey.getBits(ImageKey.BA_NEIGHBOR_MOD);
				final int iconNeighborCodeX = (iconNeighborCode >> 0) & 0x07;
				final int iconNeighborCodeY = (iconNeighborCode >> 3) & 0x07;
				logger.trace("For tile " + dod + " neighbor code=" + iconNeighborCode + ", modifies to (" + iconNeighborCodeX + "," + iconNeighborCodeY + ")");
				x += iconNeighborCodeX;
				y += iconNeighborCodeY;
				BufferedImage imgn = getTemplateImage(zoom, x, y, topologyMode);
				if (null != imgn) {
					if (dod.isWhiteTransparent) {
						ret = replaceTransparentColor(imgn, Color.white.getRGB());
					} else {
						ret = imgn;
						// Work in sentry modification.
						if (0 != ikey.getBits(ImageKey.BA_SENTRIED)) {
							ret = replaceRGB(ret, TileTemplate.TEMPLATE_PLAYER_BACKGROUND_RGB, Color.gray.getRGB());
						}
						// Work in player color.
						final int playerColorId = ikey.getBits(ImageKey.BA_PLAYER_COLOR);
						ret = replacePlayerRGB(ret, PlayerColor.get(playerColorId));
						// Work in fog of war.
						final int iconUnexploredNeighborCode = ikey.getBits(ImageKey.BA_UNEXPLORED_MOD);
						final int outdatedCode = ikey.getBits(ImageKey.BA_OUTDATED);
						if (0 < iconUnexploredNeighborCode) {
							final BufferedImage biCopy = new BufferedImage(ret.getWidth(), ret.getHeight(), ret.getType());
							biCopy.setData(ret.getData());
							if (0 != iconUnexploredNeighborCode) {
								final TileTemplateData ttdUnexploredNeighbor = TileTemplate.getTTD(ImageKey.UNEXPLORED_NEIGHBOR.getAll());
								BufferedImage biUnexploredNeighbor = getTemplateImage(zoom, 
										ttdUnexploredNeighbor.iconX + ((iconUnexploredNeighborCode >> 0) & 0x07), 
										ttdUnexploredNeighbor.iconY + ((iconUnexploredNeighborCode >> 3) & 0x07), 
										topologyMode);
								if (null != biUnexploredNeighbor) {
									biUnexploredNeighbor = replaceTransparentColor(biUnexploredNeighbor, Color.white.getRGB());
									biCopy.getGraphics().drawImage(biUnexploredNeighbor, 0, 0, playfieldTiling.getSurroundingWidth(), playfieldTiling.getSurroundingHeight(), null);
								}
							}
							ret = biCopy;
						}
						if (0 != outdatedCode) {
							ret = applyFogOfWarToColor(ret);
						}
					}
					final BufferedImage img = ret;
					zoom.getImageSet(topologyMode).storeImage(ikey, img); 
				}
			}
		}
		return ret;
	}

	/**
	 * Snips the sub-image at tile (<em>x</em>, <em>y</em>) from the templates
	 * image.
	 * 
	 * @param zoom
	 *            the zoom level
	 * @param tileXpos
	 *            the x coordinate of the tile to snip in the template
	 * @param tileYpos
	 *            the y coordinate of the tile to snip in the template
	 * @param topologyMode
	 *            the topology mode to get the sub-image for
	 * @return the sub-image
	 */
	protected static BufferedImage getTemplateImage(final ZoomLevel zoom, 
			final int tileXpos, final int tileYpos, final TopologyMode topologyMode) {
		try {
			return zoom.getImageSet(topologyMode).getTemplateImageTile(tileXpos, tileYpos);
		} catch (final RasterFormatException e) {
			logger.error("Unexpected exception for drawable with image position (" + tileXpos + "," + tileYpos + ")", e);
			return null;
		}
	}

	public static BufferedImage replacePlayerRGB(final BufferedImage imgn, final PlayerColor pl) {
		BufferedImage ret = imgn;
		ret = replaceRGB(ret, TEMPLATE_PLAYER_FOREGROUND_RGB, pl.getColor().getRGB());
		ret = replaceRGB(ret, TEMPLATE_PLAYER_BACKGROUND_RGB, pl.getBackground().getRGB());
		logger.trace("replacePlayerRGB: playerColorKey=" + pl.getKey() + ", foreground=0x" + Integer.toHexString(pl.getColor().getRGB()));
		return ret;
	}

	/**
	 * Creates a new image with one color replaced by another, but keeping the transparency
	 * @param input the input image
	 * @param sourceRGB the color to be replaced (alpha is ignored in comparison, but kept in result)
	 * @param destinationRGB the color to replace to (alpha is ignored)
	 * @return a new image with one color replaced by another, but keeping the transparency
	 */
	public static BufferedImage replaceRGB(final BufferedImage input, final int sourceRGB, final int destinationRGB) {
		final int h = input.getHeight();
		final int w = input.getWidth();
		final BufferedImage bid = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
		for (int x = 0; x < w; x++) {
			for (int y = 0; y < h; y++) {
				final int rgb = input.getRGB(x, y);
				int newRGB = rgb;
				if (Bitfield32.getBits(newRGB, BA_RGB) == Bitfield32.getBits(sourceRGB, BA_RGB)) {
					newRGB = Bitfield32.setBits(newRGB, BA_RGB, destinationRGB);
				}
				bid.setRGB(x, y, newRGB);
			}
		}
		return bid;
	}

	/**
	 * Creates a new image with for of war applied to it
	 * @param input the input image
	 * @return a new image based on the input but with fog of war applied
	 */
	public static BufferedImage applyFogOfWarToColor(final BufferedImage input) {
		final int h = input.getHeight();
		final int w = input.getWidth();
		final BufferedImage bid = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
		for (int x = 0; x < w; x++) {
			for (int y = 0; y < h; y++) {
				final int rgb = input.getRGB(x, y);
				final int newRGB = applyFogOfWarToColor(rgb);
				bid.setRGB(x, y, newRGB);
			}
		}
		return bid;
	}

	public static int applyFogOfWarToColor(int argb) {
		final int r = (argb >> 16) & 0xff;
		final int g = (argb >>  8) & 0xff;
		final int b = (argb >>  0) & 0xff;
		final int factor = 85;
		final int nr = r * factor / 100;
		final int ng = g * factor / 100;
		final int nb = b * factor / 100;
		final int nrgb = (nr << 16) | (ng << 8) | (nb << 0) | (argb & 0xff000000);
		return nrgb;
	}

	/**
	 * Creates a new image with one color set to transparent
	 * @param input the input image
	 * @param transparentColorRGB
	 *            the RGB color that will be made transparent (alpha part is ignored)
	 * @return a new image with one color set to transparent
	 */
	public static BufferedImage replaceTransparentColor(final BufferedImage input, final int transparentColorRGB) {
		final int h = input.getHeight();
		final int w = input.getWidth();
		final BufferedImage bid = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
		for (int x = 0; x < w; x++) {
			for (int y = 0; y < h; y++) {
				final int rgb = input.getRGB(x, y);
				final int newRGB = changeARGBColorToTransparent(rgb, transparentColorRGB);
				bid.setRGB(x, y, newRGB);
			}
		}
		return bid;
	}

	/**
	 * Sees that a pixel in the image with the given transparentColor gets 100%
	 * transparent. The code is based on a <a href=
	 * "http://forum.java.sun.com/thread.jspa?threadID=5227456&tstart=150">thread
	 * in the java forum</a>.
	 * 
	 * @param argb
	 *            the ARGB color of the pixel to make transparent, if equal to
	 *            transparent color
	 * @param transparentColorRGB
	 *            the RGB color that will be made transparent (alpha part is ignored)
	 * @return the transformed ARGB color value
	 */
	private static final int changeARGBColorToTransparent(final int argb, final int transparentColorRGB) {
		final int ret;
		if (Bitfield32.getBits(argb, BA_RGB) == Bitfield32.getBits(transparentColorRGB, BA_RGB)) {
			// Mark the alpha bits as transparent
			ret = Bitfield32.setBits(argb, BA_ALPHA, 0x00);
		} else {
			ret = argb;
		}
		return ret;
	}

	public static final BitArea BA_RGB = new BitArea(0, 24);
	public static final BitArea BA_ALPHA = new BitArea(24, 8);

	private int imageKey;

	public TileTemplate() {
		this.imageKey = 0;
		logger.setToShowThread(true);
		logger.setLogLevel(Level.INFO);
	}

	@Override
	public void paint(final Graphics g, Point offset, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		paintTile(g, offset, getImageKey(), zoomLevel, topologyMode);
	}

	@Override
	public void paint(final Graphics g, Point offset, final float x, final float y, final int imageKey, final float dyingProgress, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		paintTile(g, offset, imageKey, x, y + dyingProgress, (Animation.NOT_DYING < dyingProgress), zoomLevel, topologyMode);
	}

	/**
	 * Loads the image for this kind of field and player.
	 * @param zoomLevel the zoom level to create the image for
	 * @param topologyMode the topology mode to create the image for
	 * @return the image
	 */
	public final BufferedImage createImage(final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		return createImage(getImageKey(), zoomLevel, topologyMode);
	};

	/**
	 * @return the image key for that tile object. Meaning of the key is defined
	 *         in the {@link ImageKey} class.
	 */
	@Override
	public int getImageKey() {
		return imageKey;
	}

	/**
	 * @param imageKey
	 *            the image key for that tile object. Meaning of the key is
	 *            defined in the {@link ImageKey} class.
	 */
	@Override
	public void setImageKey(final int imageKey) {
		this.imageKey = imageKey & TileTemplate.getTTD(imageKey).normalizationMask;
	}

	/**
	 * Initializes the image key to the correct ttype and template id.
	 * 
	 * @param ttype
	 *            the ttype to set
	 * @param templateId
	 *            the template id within the ttype
	 */
	protected void initImageKey(final ImageKey.TType ttype, final int templateId) {
		final Bitfield32 bits = new Bitfield32();
		bits.setBits(ImageKey.BA_TTYPE, ttype.ordinal());
		bits.setBits(ImageKey.BA_TEMPLATE, templateId);
		setImageKey(bits.getAll());
	}

	/** 
	 * @return the player id of that tile object. Default implementation always returns no player.
	 */
	@Override
	public int getPlayerId() {
		return Player.ID_NONE;
	}

	/** 
	 * @return a tooltip for the current tile object. Default implementation returns the name.
	 */
	@Override
	public I18NStringWithFillIns getToolTip(final GameView gameView) {
		return getName();
	}

	/** 
	 * @return a 1-character representation, e.g., to allow logging of whole maps
	 */
	@Override
	public char getLogRepresentation() {
		return getTTD(imageKey).logRepresentation;
	}
}
