package de.ped.empire.logic;

import java.awt.Graphics;
import java.awt.Point;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.TopologyMode;

/**
 * Anything that can be drawn in a tile on the map. Knows how to be drawn, but not where.
 * Could be e.g. a field, a unit, a fight overlay or a crosshair.
 * 
 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
 */
public interface Tileable {
	/**
	 * The image key contains all information necessary to identify which tile
	 * to draw.
	 * 
	 * @return the image key
	 * @see ImageKey
	 */
	public abstract int getImageKey();

	/**
	 * The image key contains all information necessary to identify which tile
	 * to draw.
	 * 
	 * @param imageKey
	 *            the image key
	 * @see ImageKey
	 */
	public abstract void setImageKey(int imageKey);

	/**
	 * @return the player ID of the tile, or {@link Player#ID_NONE} if it belongs to no player
	 */
	public abstract int getPlayerId();

	/** 
	 * @return the name of the object, or <code>null</code> if the object has no name
	 */
	public abstract I18NStringWithFillIns getName();

	/** 
	 * @param gameView the game view
	 * @return the tooltip
	 */
	public abstract I18NStringWithFillIns getToolTip(final GameView gameView);

	/** 
	 * @return a 1-character representation, e.g., to allow logging of whole maps
	 */
	public abstract char getLogRepresentation();

	/**
	 * Paints the object in its current state
	 * 
	 * @param g
	 *            where to paint to
	 * @param offset
	 *            where to start painting in <em>g</em>
	 * @param zoomLevel
	 *            the zoom level to paint the object for
	 * @param topologyMode
	 *            the topology mode to paint the object for
	 * @see TileTemplate#paintTile(Graphics, Point, int, PlayfieldPosition, ZoomLevel, TopologyMode)
	 */
	public abstract void paint(final Graphics g, final Point offset, final ZoomLevel zoomLevel, final TopologyMode topologyMode);

	/**
	 * Paints the object in its current state
	 * 
	 * @param g
	 *            where to paint to
	 * @param offset
	 *            where to start painting in <em>g</em>
	 * @param x
	 *            the modification to x coordinate of the map position (e.g., to
	 *            indicate a move)
	 * @param y
	 *            the modification to y coordinate of the map position (e.g., to
	 *            indicate a move)
	 * @param dyingProgress
	 *            a value from 0.0f to 1.0f describing the position in the dying
	 *            animation
	 * @param zoomLevel
	 *            the zoom level to paint the object for
	 * @param topologyMode
	 *            the topology mode to paint the object for
	 * @param imageKey
	 *            the image key to paint; may differ from the current one of
	 *            <em>this</em>
	 */
	public abstract void paint(final Graphics g, final Point offset, final float x, final float y, final int imageKey, final float dyingProgress, final ZoomLevel zoomLevel, final TopologyMode topologyMode);
}
