package de.ped.empire.logic;

import java.util.List;

import de.ped.tools.BorderMode;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Logger;

/**
 * A factory for maps and scenarios used in tutorial and test cases.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Tutorial {

	public static GameMap createMap(final GameProperties props, final String[] mapDescription) {
		final GameMap map = new GameMap(props);
		int id = 0;
		for (int y = 0; y < mapDescription.length; y++) {
			final String row = mapDescription[y];
			for (int x = 0; x < row.length(); x++) {
				final char c = row.charAt(x);
				Field field;
				switch(c) {
				case '#':
					field = new Field(x, y, ImageKey.FIELD_ID_BORDER);
					break;
				case '~':
					field = new Field(x, y, ImageKey.FIELD_ID_SEA);
					break;
				case '+':
					field = new Field(x, y, ImageKey.FIELD_ID_LAND);
					break;
				case ':':
					field = new Field(x, y, ImageKey.FIELD_ID_ROUGH);
					break;
				case 'o':
					field = new Field(x, y, ImageKey.FIELD_ID_MOUNTAIN);
					break;
				case '$':
					field = new FieldCity(id, x, y);
					id++;
					break;
				case ' ':
				default:
					field = null;
				}
				if (null != field) {
					map.setField(field);
				}
			}
		}
		map.finish();
		return map;
	}

	public static GameProperties createProperties(final String[] mapForMoveTests, final TutorialLevel tutorialLevel) {
		final GameProperties ret = new GameProperties();
		ret.setRulesComplexity(RulesComplexity.ADVANCED);
		ret.setSize(PlayfieldPosition.X, mapForMoveTests[mapForMoveTests.length - 1].length());
		if (mapForMoveTests[0].length() == 2) {
			ret.setTopologyMode(TopologyMode.HEXAGON);
			ret.setSize(PlayfieldPosition.Y, mapForMoveTests.length - (mapForMoveTests[mapForMoveTests.length - 1].length() / 2 - 1));
		} else {
			ret.setSize(PlayfieldPosition.Y, mapForMoveTests.length);
		}
		if (tutorialLevel == TutorialLevel.HEXAGON_COMPACT) {
			ret.setTopologyMode(TopologyMode.HEXAGON);
		}
		if ('#' == mapForMoveTests[0].charAt(1)) { // Top border
			if ('#' == mapForMoveTests[1].charAt(0)) { // Left border
				ret.setBorderMode(BorderMode.ALL);
			} else {
				ret.setBorderMode(BorderMode.HORIZONTALLY);
			}
		} else {
			ret.setBorderMode(BorderMode.NONE);
		}
		ret.setTimestampNow();
		ret.setName("Tutorial" + tutorialLevel.ordinal());
		return ret;
	}

	public static Unit createUnit(final GameState gameState, final FieldCity city, final CommandableTemplate template, final PlayfieldPosition position) {
		city.setProduction(template);
		final Unit unit = Unit.create(gameState, city);
		gameState.addUnit(city, unit);
		if (null != position) {
			unit.setXY(position);
			gameState.getGameView(unit.getPlayerId()).updateViewMap(unit, null);
		}
		return unit;
	}

	public static Unit createUnit(final GameView gameView, final FieldCity city, final CommandableTemplate template, final PlayfieldPosition position) {
		return createUnit(gameView.getGameState(), city, template, position);
	}

	public static final String[] MAP_DESC_SQUARE_SINGLE_PATH_MOUNT = new String[]{
		"##########",
		"#~~~~~~~~#",
		"#~$+o++$~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"##########",
	};
	public static final String[] MAP_DESC_SQUARE_SINGLE_PATH_ROUGH = new String[]{
		"##########",
		"#~~~~~~~~#",
		"#~$+:++$~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"##########",
	};
	public static final String[] MAP_DESC_SQUARE_SINGLE_PATH_LAND = new String[]{
		"##########",
		"#~~~~~~~~#",
		"#~$++++$~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"#~~~~~~~~#",
		"##########",
	};
	public static final String[] MAP_DESC_SQUARE_BORDER_ALL = new String[]{
		"###############",
		"#~~~~~~~~~~~~~#",
		"#~$+$~+++++$+~#",
		"#~+++~+++++++~#",
		"#~++:~o::++++~#",
		"#~~~~:+++$+++~#",
		"#~+$~++++++++~#",
		"#~++~++++++$+~#",
		"#~~~~~~~~~~~~~#",
		"###############",
	};
	public static final String[] MAP_DESC_SQUARE_BORDER_NONE = new String[]{
		"~~~~~~~~~~~~~~~~",
		"~~~~~~~~~~~~~~~~",
		"~~$+$~+++++$+~~~",
		"~~+++~+++++++~~~",
		"++++:~o::++++~~~",
		"~~~~~:+++$++++++",
		"~~+$~++++++++~~~",
		"~~++~++++++$+~~~",
		"~~~~~~~~~~~~~~~~",
		"~~~~~~~~~~~~~~~~",
	};
	
	public static final String[] MAP_DESC_HEXAGON_BORDER_NONE = new String[]{
			"+~",
			"~~~~",
			"++++$~",
			"~~+++~++",
			"++++:~o::+",
			"~~~~~:+++$++",
			"~~+$~++++++++~",
			"~~++~++++++$+~~+",
			"~~~~~~~~~~++~~++",
			"+~~~~~~~~~~++++$",
			"  ~~~~~~~~~~++~+",
			"    ~~~~~~~~~+++",
			"      ~~~~~~~~~~",
			"        ~~~~~~~~",
			"          ~~+$+~",
			"            ~~~~",
			"              ~+",
		};
	public static final String[] MAP_DESC_SQUARE_FOUR_CITIES = new String[]{
			"############",
			"#~~~~~~~~~~#",
			"#~+++~~+++~#",
			"#~+$+~~+$+~#",
			"#~+++~~+++~#",
			"#~~~~~~~~~~#",
			"#~~~~~~~~~~#",
			"#~+++~~+++~#",
			"#~+$+~~+$+~#",
			"#~+++~~+++~#",
			"#~~~~~~~~~~#",
			"############",
		};

	public GameMap createMap(final TutorialLevel level) {
		final String[] mapDescription = level.mapDescription;
		final GameProperties props = createProperties(mapDescription, level);
		final GameMap map = createMap(props, mapDescription);
		for (final FieldCity city : map.getAllCities()) {
			city.setPlayerId(Player.ID_NONE);
		}
		Logger logger = Logger.getLogger(getClass().getName());
		logger.trace(map.getAt(level.posCity1).toString());
		final FieldCity ownCity = (FieldCity) map.getAt(level.posCity1);
		ownCity.setPlayerId(0);
		final FieldCity enemyCity = (FieldCity) map.getAt(level.posCity2);
		enemyCity.setPlayerId(1);
		return map;
	}

	public void createTutorial(final GameState gameState, final TutorialLevel id) {
		final int playerId = 0;
		final FieldCity ownCity = gameState.getMap().findCities(playerId).get(0);
		final FieldCity enemyCity = gameState.getMap().findCities(1).get(0);
		switch (id) {
		case LAND_UNITS: {
				// Rough is at (7, 4) and (8, 4)
//				final int[] destination = new int[] { 8, 3 };
				createUnit(gameState, ownCity, Templates.instance().TANK, new PlayfieldPosition(8, 5));
				createUnit(gameState, ownCity, Templates.instance().ARMY, new PlayfieldPosition(8, 7));
			}
			break;
		case AIR_UNITS: {
//				final int[] destination = new int[] { 8, 3 };
				createUnit(gameState, ownCity, Templates.instance().FIGHTER, new PlayfieldPosition(2, 1));
			}
			break;
		case VISIB_FIGHTER_SUBS: {
				createUnit(gameState, enemyCity, Templates.instance().FIGHTER, enemyCity.getPosition());
				final PlayfieldPosition start       = new PlayfieldPosition(13, 1);
				final PlayfieldPosition destination = new PlayfieldPosition(13, 4);
				createUnit(gameState, ownCity, Templates.instance().SUBMARINE, destination);
	//			createUnit(gameState, ownCity, Templates.instance().DESTROYER, start);
				createUnit(gameState, ownCity, Templates.instance().TRANSPORTER, start);
			}
			break;
		case VISIB_IF_LOADED: {
				createUnit(gameState, enemyCity, Templates.instance().TRANSPORTER, enemyCity.getPosition());
				createUnit(gameState, enemyCity, Templates.instance().ARMY, enemyCity.getPosition());
				createUnit(gameState, enemyCity, Templates.instance().FIGHTER, enemyCity.getPosition());
				final PlayfieldPosition start = new PlayfieldPosition(12, 8);
				createUnit(gameState, ownCity, Templates.instance().CRUISER, start);
			}
			break;
		case SINGLE_PATH_LAND:
		case SINGLE_PATH_ROUGH:
		case SINGLE_PATH_MOUNT: {
				// Nothing to do
			}
			break;
		case HEXAGON_NOCOMPACT: {
				// Nothing to do
			}
			break;
		case HEXAGON_COMPACT: {
				createUnit(gameState, enemyCity, Templates.instance().TRANSPORTER, enemyCity.getPosition());
				createUnit(gameState, enemyCity, Templates.instance().ARMY, enemyCity.getPosition());
				createUnit(gameState, enemyCity, Templates.instance().FIGHTER, enemyCity.getPosition());
				final PlayfieldPosition start = new PlayfieldPosition(12, 8);
				createUnit(gameState, ownCity, Templates.instance().CRUISER, start);
			}
			break;
		case ENEMY_ISLANDS: {
				final List<FieldAlike> neighbors = gameState.getMap().getAllFieldsInAirDistance(enemyCity, 1, false, 2);
				for (FieldAlike n : neighbors) {
					createUnit(gameState, ownCity, Templates.instance().TANK, n.getPosition());
				}
			}
			break;
		default:
			throw new IllegalArgumentException("Unknown tutorial #" + id);
		}
	}
}
