package de.ped.empire.logic;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.Bitfield32;

/**
 * The abstract class for a unit.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Unit extends TileTemplate implements UnitAlike {
	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;

	/**
	 * Creates a unit from the marshalling stream.
	 * <p>
	 * Implementation considered bad style, as supertype knows all subtypes. But
	 * this spares a separate factory class.
	 * 
	 * @param gameState
	 *            the game to create the unit in
	 * @param marshaller
	 *            the marshaller to create the unit from
	 * @return the unit
	 */
	public static Unit create(final GameState gameState, final Marshaller marshaller) {
		final ImageKey imageKey = new ImageKey(marshaller.readInt());
		final int templateId = imageKey.getBits(ImageKey.BA_TEMPLATE);
		CommandableTemplate template = null;
		Unit ret;
		switch(templateId) {
		case ImageKey.CMDL_ID_ARMY:
			template = Templates.instance().ARMY;
			break;
		case ImageKey.CMDL_ID_TANK:
			template = Templates.instance().TANK;
			break;
		case ImageKey.CMDL_ID_BOMBER:
			template = Templates.instance().BOMBER;
			break;
		case ImageKey.CMDL_ID_FIGHTER:
			template = Templates.instance().FIGHTER;
			break;
		case ImageKey.CMDL_ID_TRANSPORTER:
			template = Templates.instance().TRANSPORTER;
			break;
		case ImageKey.CMDL_ID_DESTROYER:
			template = Templates.instance().DESTROYER;
			break;
		case ImageKey.CMDL_ID_SUBMARINE:
			template = Templates.instance().SUBMARINE;
			break;
		case ImageKey.CMDL_ID_CRUISER:
			template = Templates.instance().CRUISER;
			break;
		case ImageKey.CMDL_ID_BATTLESHIP:
			template = Templates.instance().BATTLESHIP;
			break;
		case ImageKey.CMDL_ID_CARRIER:
			template = Templates.instance().CARRIER;
			break;
		case ImageKey.CMDL_ID_AIRBASE:
			template = Templates.instance().AIRBASE;
			break;
		default:
			throw new IllegalArgumentException("Unknown unit type " + Integer.toHexString(templateId));
		}
		if (null != template) {
			ret = new Unit(gameState, template, marshaller, imageKey);
		} else {
			ret = null;
		}
		return ret;
	}

	/**
	 * Creates a unit in a city.
	 * <p>
	 * Implementation considered bad style, but mimics behavior of the other
	 * creator for marshaller.
	 * 
	 * @param gameState
	 *            the game to create the unit in
	 * @param playerId
	 *            the ID of the player the unit is created for
	 * @param template
	 *            the kind of unit to create
	 * @param city
	 *            the city the unit is created in. If <code>null</code>, the
	 *            unit is only created for drawing purposes, but not part of the
	 *            game
	 * @return the unit
	 */
	static Unit create(final GameState gameState, final int playerId, final CommandableTemplate template, final FieldAlike city) {
		Unit ret;
		if ((null == template) || (template.isUndefinedOrNothing())) {
			ret = null;
		} else if ((Templates.instance().ARMY == template)
				|| (Templates.instance().TANK == template)
				|| (Templates.instance().BOMBER == template)
				|| (Templates.instance().FIGHTER == template)
				|| (Templates.instance().TRANSPORTER == template)
				|| (Templates.instance().DESTROYER == template)
				|| (Templates.instance().SUBMARINE == template)
				|| (Templates.instance().CRUISER == template)
				|| (Templates.instance().BATTLESHIP == template)
				|| (Templates.instance().CARRIER == template)
				|| (Templates.instance().AIRBASE == template)) {
			ret = new Unit(gameState, template, playerId, city);
		} else {
			throw new IllegalArgumentException("Unknown unit type " + Integer.toHexString(template.id));
		}
		return ret;
	}

	/**
	 * Creates a unit in a city.
	 * 
	 * @param gameState
	 *            the game to create the unit in
	 * @param city
	 *            the city the unit is created in
	 * @return the unit
	 */
	public static Unit create(final GameState gameState, final FieldCity city) {
		return create(gameState, city.getPlayerId(), city.getUnitInProduction(), city);
	}

	/**
	 * Defined for testing and initialization purposes. During game, use easier
	 * factory method.
	 * 
	 * @param gameState
	 *            the game to create the unit in
	 * @param template
	 *            the unit to create
	 * @param city
	 *            the city the unit is created in
	 * @return the unit
	 */
	public static Unit create(final GameState gameState, final CommandableTemplate template, final FieldCity city) {
		return create(gameState, city.getPlayerId(), template, city);
	}

	/**
	 * Creates a tile that is only needed for drawing, but not part of the game.
	 * 
	 * @param gameState
	 *            the game to create the tile in
	 * @param playerId
	 *            the id of the player
	 * @param template
	 *            the unit template to create the tile for
	 * @return the unit to show
	 */
	public static Unit create(final GameState gameState, final int playerId, final CommandableTemplate template) {
		return create(gameState, playerId, template, null);
	}

	/** Backwards reference to the game this unit is part of. Non-persistent. */
	private final GameState gameState;
	/** The primary key of this unit. Persistent. */
	private final CommandableKey key;
	/** The name of this unit, if given. Persistent. */
	private String nickName;
	/** The position of this unit on the map. Persistent as position. */
	private Field field;
	/** The number of hits this unit can still receive. Persistent. */
	private int hitsTillDeath;
	/** The number of rounds this unit was unable to refuel. Persistent. */
	private int roundsSinceRefuel;
	/** The number of steps this unit has moved this turn. Persistent. */
	private int stepsDoneThisTurn;
	/** Keeps if this unit has seen enemies since its last turn. Persistent. */
	private boolean hasSeenEnemiesLately;
	/** Keeps if this unit has moved this turn. Persistent. */
	private boolean hasMovedThisTurn;
	/** Keeps if skip turn command was given to this unit this turn. Persistent. */
	private boolean isSkippingThisTurn;
	/** If the unit was already selected in this round of nexting units. Persistent. */
	private boolean wasNexted;
	/** The order this unit has, if any. Persistent. */
	private Order order;
	/** The unique key of the parent of this unit, if any. Persistent. */
	private CommandableKey parentKey;
	/** Non-persistent. */
	private I18NStringWithFillIns i18nName = null;
	/** The step ratings this unit has on the map. Non-persistent. */
	private StepRatingMap stepRatings = null;

	/**
	 * Creates a unit.
	 * 
	 * @param gameState
	 *            the game to create the unit in
	 * @param template
	 *            the template the unit is based on
	 * @param playerId
	 *            the ID of the player the unit is created for
	 * @param field
	 *            the field where the unit will be placed. If <code>null</code>,
	 *            this unit object is only created for drawing purposes, but not
	 *            part of the game.
	 */
	public Unit(final GameState gameState, final CommandableTemplate template, final int playerId, final FieldAlike field) {
		initImageKey(ImageKey.TType.Unit, template.id);
		this.gameState = gameState;
		final Player player = gameState.getGameView(playerId).getPlayer(playerId);
		int id;
		if (null != field) {
			id = player.addUnit(template);
		} else {
			id = -1;
		}
		this.key = new CommandableKey(template.id, player.getId(), id);
		this.hitsTillDeath = template.getHitsTillDeath(gameState.getProperties().getRulesComplexity());
		this.roundsSinceRefuel = 0;
		this.stepsDoneThisTurn = 0;
		this.hasSeenEnemiesLately = false;
		this.hasMovedThisTurn = false;
		this.isSkippingThisTurn = false;
		this.wasNexted = false;
		this.nickName = "";
		this.parentKey = null;
		// Now that the unit is initialized, place it on the map
		if (null != field) {
			setField(field.getField());
		} else {
			this.field = null;
		}
	}

	public Unit(final GameState gameState, final CommandableTemplate template, final Marshaller marshaller, final ImageKey imageKey) {
		super.setImageKey(imageKey.getAll());
		this.gameState = gameState;
		final int playerId = marshaller.readInt();
		final int id = marshaller.readInt();
		this.key = new CommandableKey(template.id, playerId, id);
		this.nickName = marshaller.readString();
		final PlayfieldPosition pos = PlayfieldPosition.demarshalWithNull(marshaller);
		if (null == gameState) {
			this.field = null;
		} else {
			setField(gameState.getMap().getAt(pos).getField());
		}
		this.hitsTillDeath = marshaller.readInt();
		this.roundsSinceRefuel = marshaller.readInt();
		this.stepsDoneThisTurn = marshaller.readInt();
		this.hasSeenEnemiesLately = marshaller.readBoolean();
		this.hasMovedThisTurn = marshaller.readBoolean();
		this.isSkippingThisTurn = marshaller.readBoolean();
		this.wasNexted = marshaller.readBoolean();
		this.order = (Order) marshaller.readNullOrMarshallable(new Order.Factory(gameState, this));
		this.parentKey = (CommandableKey) marshaller.readNullOrMarshallable(CommandableKey.factory);
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(getImageKey());
		marshaller.writeInt(key.getPlayerId());
		marshaller.writeInt(key.getId());
		marshaller.writeString(nickName);
		final PlayfieldPosition pos;
		if (null == field) {
			pos = null;
		} else {
			pos = field.getPosition();
		}
		PlayfieldPosition.marshalWithNull(marshaller, pos);
		marshaller.writeInt(hitsTillDeath);
		marshaller.writeInt(roundsSinceRefuel);
		marshaller.writeInt(stepsDoneThisTurn);
		marshaller.writeBoolean(hasSeenEnemiesLately);
		marshaller.writeBoolean(hasMovedThisTurn);
		marshaller.writeBoolean(isSkippingThisTurn);
		marshaller.writeBoolean(wasNexted);
		marshaller.writeNullOrMarshallable(order);
		marshaller.writeNullOrMarshallable(parentKey);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	public int getId() {
		return key.getId();
	}

	public Field getField() {
		return field;
	}

	@Override
	public PlayfieldPosition getPosition() {
		return field.getPosition();
	}

	@Override
	public int getPlayerId() {
		return key.getPlayerId();
	}

	GameView getGameView() {
		return gameState.getGameView(key.getPlayerId());
	}

	public Player getPlayer() {
		return gameState.getPlayer(key.getPlayerId());
	}

	@Override
	public int getHitsTillDeath() {
		return hitsTillDeath;
	}

	@Override
	public float getFightStrength(final Commandable opponent, final boolean isAttacker) {
		float ret;
		final GameMap map = gameState.getMap();
		final RulesComplexity rulesComplexity = map.getProperties().getRulesComplexity();
		if (isAttacker) {
			ret = getAttack();
			if (opponent instanceof Unit) {
				final Unit oppUnit = (Unit) opponent;
				// Empire Deluxe Manual page 135, rule e: Advanced game only - attack ratings are x 1/2 when attacking air units
				if (MilitaryBranch.AIR_UNIT == oppUnit.getUnitTemplate().militaryBranch) {
					ret *= getUnitTemplate().getRulesComplexityDependingValues(rulesComplexity).factorRuleE;
				}
				// Empire Deluxe Manual page 135, rule l: Advanced game only - attack ratings are x 1/2 when crippled
				if (rulesComplexity.isWithFightMalusWhenDamaged && isDamaged()) {
					ret *= 0.5f;
				}
			}
		} else {
			ret = getDefense();
			final Unit oppUnit = (Unit) opponent;
			// Empire Deluxe Manual page 39-40: Defense rating is up for land units in special game map field types.
			// Empire Deluxe Manual page 135, rule h: Defense rating is x2 
			// Empire Deluxe Manual page 135, rule i: Defense rating is x2 
			if (MilitaryBranch.LAND_UNIT == getUnitTemplate().militaryBranch) {
				final int defenseBonusAgainst = field.getTemplate().rcdvs[rulesComplexity.ordinal()].defenseBonusAgainst; 
				if (Bitfield32.isBit(defenseBonusAgainst, oppUnit.getUnitTemplate().id)) {
					ret *= 2.0f;
				}
			}
			// Empire Deluxe Manual page 135, rule c: Defense rating is x 1/2 against shore bombardment
			if ((MilitaryBranch.LAND_UNIT == getUnitTemplate().militaryBranch) 
					&& (MilitaryBranch.SEA_UNIT == oppUnit.getUnitTemplate().militaryBranch)) {
				ret *= 0.5f;
			}
		}
		// Empire Deluxe Manual page 135, rule d: Attack and defense rating is x 1/2 against fighters
		if (Templates.instance().BOMBER.equals(getUnitTemplate()) 
				&& (opponent.getKey().getTemplateId() == ImageKey.CMDL_ID_FIGHTER)) {
			ret /= 2.0f;
		}
		ret /= getPlayer().getFightHandicapValue(); // Take fight handicap into account
		return ret;
	}

	@Override
	public int getAttack() {
		return getUnitTemplate().attack;
	}

	@Override
	public int getDefense() {
		return getUnitTemplate().defense;
	}

	@Override
	public int getShotDamage() {
		return getUnitTemplate().shotDamage;
	}

	@Override
	public boolean hasMovedThisTurn() {
		return hasMovedThisTurn;
	}

	boolean isSkippingThisTurn() {
		return isSkippingThisTurn;
	}

	@Override
	public void setSkippingThisTurn(final boolean value) {
		isSkippingThisTurn = value;
	}

	/**
	 * Unit receives hits (positive number) or gets repaired (negative number).
	 * 
	 * @param numberOfHits
	 *            the number of hits to receive, if positive, or the amount of
	 *            repair, if negative
	 */
	@Override
	public void receiveHits(final int numberOfHits) {
		hitsTillDeath = MathUtil.bound(hitsTillDeath - numberOfHits, 0, getUnitTemplate().getHitsTillDeath(gameState.getProperties().getRulesComplexity()));
		if (0 < numberOfHits) {
			unload();
		}
	}

	@Override
	public I18NStringWithFillIns getToolTip(final GameView gameView) {
		I18NStringWithFillIns ret;
		if (getPlayer().getId() == gameView.getGameState().getCurrentPlayerId()) {
			final Collection<Unit> children = findChildren();
			final Order order = getOrder();
			I18NStringWithFillIns orderString;
			if (null == order) {
				orderString = new I18NStringWithFillIns("Game.Unit.Command.Nothing", null, null);
			} else {
				orderString = order.getToolTip();
			}
			final int stepsLeftTillDeath = getStepsLeftTillDeath();
			final Object[] fillIns = new Object[] {
					getName(), 																	// #0 = name
					Integer.toString(getUnitTemplate().getHitsTillDeath(gameState.getProperties().getRulesComplexity()) - getHitsTillDeath()),// #1 = hits
					Integer.toString(getUnitTemplate().getHitsTillDeath(gameState.getProperties().getRulesComplexity())),						// #2 = max hits
					Integer.toString(getStepsLeftThisTurn()),									// #3 = moves left this turn
					Integer.toString(stepsLeftTillDeath),										// #4 = moves left till death
					getLoadDescription(children),												// #5 = number of units loaded
					orderString,																// #6 = command string
			};
			I18NStringWithFillIns part0 = new I18NStringWithFillIns("Game.Unit.Description.Text", null, fillIns);
			I18NStringWithFillIns part1 = I18NStringWithFillIns.EMPTY;
			if (Integer.MAX_VALUE != stepsLeftTillDeath) {
				part1 = new I18NStringWithFillIns("Game.Unit.MovesTillDeath.Text", null, fillIns);
			}
			I18NStringWithFillIns part2 = I18NStringWithFillIns.EMPTY;
			if (!children.isEmpty()) {
				part2 = new I18NStringWithFillIns("Game.Unit.Load.Text", null, fillIns);
			}
			I18NStringWithFillIns part3 = new I18NStringWithFillIns("Game.Unit.Command.Text", null, fillIns);
			I18NStringWithFillIns part4 = I18NStringWithFillIns.EMPTY;
			if (hasSeenEnemiesLately()) {
				part4 = new I18NStringWithFillIns("Game.Unit.SightedEnemy", null, fillIns);
			}
			I18NStringWithFillIns fuelEmptyPart = I18NStringWithFillIns.EMPTY;
			if (isLowOnFuel()) {
				fuelEmptyPart = new I18NStringWithFillIns("Game.Unit.FuelLow", null, null);
			}
			ret = new I18NStringWithFillIns("Game.Unit.Own.ToolTip", "%0%1%2%3%4%5", new Object[]{ fuelEmptyPart, part0, part1, part2, part3, part4 });
		} else {
			ret = new I18NStringWithFillIns("Game.Unit.Other.ToolTip", "%0%1", new Object[]{ 
					getName(), 
					new I18NStringWithFillIns("Game.City.Player.Text", null, new String[] { getPlayer().getName() })});
		}
		return ret;
	}

	boolean isLowOnFuel() {
		boolean ret = false;
		final int stepsLeftTillDeath = getStepsLeftTillDeath();
		if (stepsLeftTillDeath != Integer.MAX_VALUE) { // Only for units with limited range
			final Collection<FieldAlike> nearest = findNearestOf(findHomeBases()); 
			final Iterator<FieldAlike> it = nearest.iterator();
			if (it.hasNext()) {
				final int distance = gameState.getProperties().calcAirDistance(getPosition(), it.next().getPosition());
				final int stepsLeftExploring = stepsLeftTillDeath - distance;
				if ((0 <= stepsLeftExploring) && ((stepsLeftExploring <= 2))) {
					ret = true;
				}
			}
		}
		return ret;
	}

	private I18NStringWithFillIns getLoadDescription(final Collection<Unit> children) {
//		final String size = Integer.toString(children.size());
		final I18NStringWithFillIns ret;
		String format = "";
		final Collection<I18NStringWithFillIns> listItems = new ArrayList<I18NStringWithFillIns>(); 
		for (final CommandableTemplate template : CommandableTemplate.allRealUnitTypes(gameState.getProperties().getRulesComplexity())) {
			int cnt = 0;
			for (final Unit child : children) {
				if (child.getKey().getTemplateId() == template.id) {
					cnt++;
				}
			}
			if (0 < cnt) {
				if (0 < listItems.size()) {
					format += " + ";
				}
				format += Integer.toString(cnt) + " %" + Integer.toString(listItems.size());
				listItems.add(new I18NStringWithFillIns(template.getShortKey(), null, null));
			}
		}
		if (0 == listItems.size()) {
			ret = new I18NStringWithFillIns("Game.Unit.Load.Nothing", "-", null);
		} else {
			ret = new I18NStringWithFillIns(null, format, listItems.toArray());
		}
		return ret;
	}

	@Override
	public void startTurn() {
		// Check if from last turn the unit can heal
		if (isReadyForHealing()) {
			receiveHits(-1);
		}
		// Now prepare all for next turn
		this.stepsDoneThisTurn = 0;
		this.hasMovedThisTurn = false;
		this.isSkippingThisTurn = false;
		resetNext();
		cleanRating();
		if ((null != order) && (order.isFulfilled())) {
			order = null; // Destination reached.
		}
	}

	private void checkForFlightPathToApply() {
		if (!isDamaged() && ((null == order) || (order.template.canBeOverriddenToFollowPath()))) {
			final MovePath movePath = gameState.getMovePaths().findMovePath(key.getPlayerId(), field.getPosition(), getUnitTemplate().militaryBranch);
			if (null != movePath) {
				order = new OrderGoTo(gameState, this, movePath.getEndPos());
			}
		}
	}

	boolean isReadyForHealing() {
		boolean ret = false;
		if (field.isCity() 
				&& (field.getPlayerId() == getPlayerId()) 
				&& !hasMovedThisTurn
				&& ((MilitaryBranch.SEA_UNIT == getUnitTemplate().militaryBranch) // for backwards compatibility I want to see status message for docking even for undamaged ships
						|| isDamaged())) {
			ret = true;
		}
		return ret;
	}

	@Override
	public void finishTurn() {
		hasSeenEnemiesLately = false;
		if ((field.isCity())
				|| ((null != getParentKey()) 
						&& ((getParentKey().getTemplateId() == Templates.instance().CARRIER.id) 
								|| (getParentKey().getTemplateId() == Templates.instance().AIRBASE.id)))) {
			roundsSinceRefuel = 0;
		} else {
			roundsSinceRefuel++;
		}
		cleanRating();
		resetNext();
	}

	@Override
	public I18NStringWithFillIns getName() {
		if (null == i18nName) {
			final Object[] fillIns = new Object[] {
					new I18NStringWithFillIns(getUnitTemplate().getShortKey(), null, null),
					new I18NStringWithFillIns(getUnitTemplate().getNameKey(), null, null),
					Integer.toString(getId()),
					nickName,
			};
			final String nameKey;
			if (CollectionUtil.isNullOrEmpty(nickName)) {
				nameKey = "Game.Unit.Name.Default.Text";
			} else {
				nameKey = "Game.Unit.Name.Named.Text";
			}
			i18nName = new I18NStringWithFillIns(nameKey, "%2. %1", fillIns);
			logger.trace(getKey().toString() + ".getName(): unit has nickname " + nickName);
		}
		return i18nName;
	}

	@Override
	public String getNickName() {
		return nickName;
	}

	@Override
	public void setNickName(final String name) {
		this.nickName = name;
		this.i18nName = null;
		logger.debug(getKey().toString() + ".setNickName(" + nickName + ")");
	}

	@Override
	public boolean needsNickName() {
		return (CollectionUtil.isNullOrEmpty(nickName));
	}

	@Override
	public String getNickNameSuggestion() {
		final String suggestion = getUnitTemplate().getNameSuggestion();
		return suggestion;
	}

	GameState getGameState() {
		return gameState;
	}

	@Override
	public CommandableTemplate getUnitTemplate() {
		return CommandableTemplate.findTemplateById(key.getTemplateId());
	}

	@Override
	public char getLogRepresentation() {
		return getUnitTemplate().getLogRepresentation();
	}

	@Override
	public int getImageKey() {
		int mkey = super.getImageKey();
		mkey = Bitfield32.setBits(mkey, ImageKey.BA_NONEMPTY, (findChildren().isEmpty() ? 0 : 1));
		mkey = Bitfield32.setBits(mkey, ImageKey.BA_SENTRIED, ((null != order) && (order instanceof OrderSentry) ? 1 : 0));
		mkey = Bitfield32.setBits(mkey, ImageKey.BA_PLAYER_COLOR, getPlayer().getPlayerColorId());
		super.setImageKey(mkey);
		return mkey;
	}

	public boolean canMove() {
		return (0 < getStepsLeftThisTurn());
	}

	public void move(final DigitalJoystickDirection dir) {
		stepRatings = null;
		final GameMap map = gameState.getMap();
		final RulesComplexity rulesComplexity = map.getProperties().getRulesComplexity();
		final Field targetField = map.getAt(map.getProperties().wrap(field.getPosition(), dir)).getField();
		logger.trace(getKey().toString() + ".move(" + dir.key + ") from " + field.getPosition().toString() + " to " + targetField.getPosition().toString() + " - which sets hasSeenEnemiesLately to false.");
		setField(targetField);
		hasSeenEnemiesLately = false;
		hasMovedThisTurn = true;
		if ((null != order) && (order.isFulfilled())) {
			order = null;
		}
		// Handle transporter automatic cargo unload or sentry
		if (getUnitTemplate().canHold(rulesComplexity, Templates.instance().ARMY)) {
			if (!(field.isCity())) {
				final boolean isLandNearby = map.isLandNearby(field);
				if (!isLandNearby) {
					for (final Unit cargo : findChildren()) {
						cargo.setOrder(new OrderSentry(getGameState(), cargo));
					}
				}
				final boolean isCityToConquerNearby = map.isCityToConquerNearby(field, getPlayerId());
				if (isCityToConquerNearby) {
					unload();
				}
			}
		}
		if ((null == order) && (0 < getStepsLeftThisTurn())) {
			calculateRating();
		}
	}

	private void setField(final FieldAlike target) {
		if (null != field) {
			boolean ret = field.getUnits().remove(this);
			logger.debug("setField(" + field.getPosition().toString() + ") remove result=" + ret);
		}
		if (null != target) {
			field = (Field) target;
			Collection<UnitAlike> units = field.getUnits();
//			if (units.contains(this)) {
//				// We remove any other instance of this, as it might be only a ViewUnit.
//				units.remove(this);
//			}
			units.add(this);
		} else {
//			Nothing to do, otherwise there is no history where a unit has been
//			field = null;
		}
	}

	/**
	 * Changes the position without all the side effects of a move, e.g., when in
	 * a transporter
	 * 
	 * @param pos
	 *            an array of size zwo containing the new position
	 */
	public void setXY(final PlayfieldPosition pos) {
		stepRatings = null;
		setField(gameState.getMap().getAt(pos).getField());
		gameState.getMap().addUnit(this);
		hasSeenEnemiesLately = false;
		logger.trace(getKey().toString() + ".setXY(" + pos.toString() + ") - which sets hasSeenEnemiesLately to false.");
	}

	/** 
	 * Must be called after the movement has been done. 
	 */
	public void consumeStep() {
		stepsDoneThisTurn += gameState.getMap().calcStepsToConsume(key.getTemplateId(), field.getPosition());
	}

	public int getStepsLeftTillDeath() {
		int ret = Integer.MAX_VALUE; // Default value for non-fueling units
		final RulesComplexity rulesComplexity = gameState.getProperties().getRulesComplexity();
		final int stepsTillDeath = getUnitTemplate().getStepsTillDeath(rulesComplexity);
		if (Integer.MAX_VALUE != stepsTillDeath) {
			ret = stepsTillDeath - roundsSinceRefuel * getUnitTemplate().getMaximumStepsPerTurn(rulesComplexity) - stepsDoneThisTurn;
		}
		return ret;
	}

	public int getStepsLeftThisTurn() {
		final int stepsPerTurn = getUnitTemplate().getStepsPerTurn(
				gameState.getProperties().getRulesComplexity(),
				hitsTillDeath);
		return MathUtil.bound(stepsPerTurn - stepsDoneThisTurn, 0, getStepsLeftTillDeath());
	}

	public void consumeAllStepsThisTurn() {
		stepsDoneThisTurn += getStepsLeftThisTurn();
	}

	@Override
	public boolean isAlive() {
		return (0 < getHitsTillDeath());
	}

	@Override
	public boolean isDamaged() {
		return (getHitsTillDeath() < getUnitTemplate().getHitsTillDeath(gameState.getProperties().getRulesComplexity()));
	}

	@Override
	public boolean isNextable() {
		return (isAlive() 
				&& getUnitTemplate().isFocusable
				&& ((0 < getStepsLeftThisTurn()) 
						&& !isSkippingThisTurn()
						&& ((null == order) 
								|| (!order.template.isWaiting) 
								|| (null != gameState.findUnitToLoadInto(this)))
					|| hasSeenEnemiesLately));
	}

	@Override
	public int getSightingRange(final MilitaryBranch observed) {
		return getUnitTemplate().getSightingRange(gameState.getProperties().getRulesComplexity(), observed);
	}

	@Override
	public boolean hasSeenEnemiesLately() {
		return hasSeenEnemiesLately;
	}

	@Override
	public void setSeenEnemiesLately(final boolean value) {
		logger.setToShowStack(true);
		logger.trace(getKey().toString() + ".setSeenEnemiesLately(" + value + ")");
		hasSeenEnemiesLately = value;
	}

	@Override
	public boolean wasNexted() {
		return (wasNexted);
	}

	@Override
	public void resetNext() {
		wasNexted = false;
	}

	@Override
	public void nextFocusGained() {
		wasNexted = true;
		cleanRating();
		calculateRating();
		checkForFlightPathToApply();
	}

	private void calculateRating() {
		calculateForwardRating(StepRatingPurpose.NORMAL);
		if (null != order) {
			logger.trace(getKey().toString() + ".calculateRating: Calculating backwards rating for " + toString() + " with order " + order.toString());
			boolean ret = order.prepareRating();
			if (ret) {
				calculateBackwardRating();
			}
		}
	}

	@Override
	public void nextFocusLost() {
		hasSeenEnemiesLately = false;
		cleanRating();
		logger.trace(getKey().toString() + ".nextFocusLost() - which sets hasSeenEnemiesLately to false.");
	}

	@Override
	public Order getOrder() {
		return order;
	}

	@Override
	public void setOrder(final Order order) {
		this.order = order;
		if (null != order) {
			calculateRating();
		}
		logger.debug(getKey().toString() + ".setOrder(" + order + ")");
	}

	@Override
	public MoveResult evaluateOrder(final GameView gameView, final boolean isNow) {
		MoveResult ret = null;
		if (null != order) {
			ret = order.evaluate(isNow);
		}
		return ret;
	}

	void cleanRating() {
		stepRatings = null;
	}

	public void calculateForwardRating(final StepRatingPurpose stepRatingPurpose) {
		final long startTime = Calendar.getInstance().getTimeInMillis();
		stepRatings = new StepRatingMap(this);
		stepRatings.setStepRatingPurpose(stepRatingPurpose);
		stepRatings.calculateForwardRating();
		final long endTime = Calendar.getInstance().getTimeInMillis();
		logger.debug(getKey().toString() + ".calculateForwardRating(" + stepRatingPurpose + ") took " + (endTime - startTime) + " ms");
	}

	/**
	 * Does a backward calculation from possible target points back to the unit to move.
	 * All positions which are on a path of the shortest length are marked.
	 * <p>
	 * Precondition: The target point 
	 * (or the nearest target points which must be of equal distance to the unit)
	 *  must already be marked, to work as seed for this algorithm.
	 */
	void calculateBackwardRating() {
		final int logLevel = Level.TRACE; // Level for dumping the whole map
		if (logger.isEnabledFor(logLevel)) {
			logger.log(logLevel, "calculateBackwardRating(): (before) " + stepRatings.dump());
		}
		final long startTime = Calendar.getInstance().getTimeInMillis();
		stepRatings.calculateBackwardRating();
		final long endTime = Calendar.getInstance().getTimeInMillis();
		logger.debug(getKey().toString() + ".calculateBackwardRating() took " + (endTime - startTime) + " ms");
		if (logger.isEnabledFor(logLevel)) {
			logger.log(logLevel, "calculateBackwardRating(): (after) " + stepRatings.dump());
		}
	}

	boolean setTargetSeed(final PlayfieldPosition pos) {
		boolean ret = false;
		StepRatingPosition rating = getStepRating(pos);
		if (null != rating) {
			rating.setTargetSeed(true);
			ret = true;
		}
		return ret;
	}

	void clearAsTarget(final PlayfieldPosition pos) {
		StepRatingPosition rating = getStepRating(pos);
		if (null != rating) {
			rating.setTargetSeed(false);
		}
	}

	boolean setTargetSeeds(final Collection<FieldAlike> fields) {
		boolean ret = false;
		if (!CollectionUtil.isNullOrEmpty(fields)) {
			for (final FieldAlike field : fields) {
				ret = ret || setTargetSeed(field.getPosition());
			}
		}
		return ret;
	}

	/**
	 * Returns all the fields with the nearest distance from the list of fields.
	 * The distance is calculated using the step rating.
	 * 
	 * @param fields
	 *            the list of fields to take into account
	 * @return the subset of all the fields with the nearest equal distance,
	 *         could be empty if none of the fields is reachable
	 */
	List<FieldAlike> findNearestOf(final Collection<FieldAlike> fields) {
		final List<FieldAlike> ret = new LinkedList<FieldAlike>();
		if (!CollectionUtil.isNullOrEmpty(fields)) {
			if (null == getStepRatings()) {
				calculateForwardRating(StepRatingPurpose.NORMAL);
			}
			// calculate the nearest distance
			int nearestDistance = Integer.MAX_VALUE; 
			for (final FieldAlike field : fields) {
				final StepRatingPosition rating = getStepRating(field.getPosition());
				if ((null != rating) && (rating.distanceFromStart < nearestDistance)) {
					nearestDistance = rating.distanceFromStart;
				}
			}
			// Return all fields with nearest distance
			for (final FieldAlike field : fields) {
				final StepRatingPosition rating = getStepRating(field.getPosition());
				if ((null != rating) && (rating.distanceFromStart == nearestDistance)) {
					ret.add(field);
				}
			}
		}
		return ret;
	}

	/**
	 * Calculates the directions to move to.
	 * 
	 * @return a list of directions, ordered by decreasing preference
	 */
	List<DigitalJoystickDirection> calcMoveOptions() {
		Assert.assertNotNull(stepRatings);
		final List<DigitalJoystickDirection> topDirections = new ArrayList<DigitalJoystickDirection>(8);
		final GameMap map = gameState.getMap();
		Assert.assertNotNull(map);
		// First, add all directions on shortest path
		final PlayfieldPosition pos = field.getPosition();
		for (final DigitalJoystickDirection dir : map.getProperties().getNeighborDirections(pos)) {
			final StepRatingPosition result = stepRatings.get(map.getProperties().wrap(pos, dir));
			if ((null != result) && result.isOnShortestPath()) {
				boolean wasInserted = false;
				for (int i = 0; i < topDirections.size(); i++) {
					if (i < topDirections.size()) {
						DigitalJoystickDirection otherDir = topDirections.get(i);
						final StepRatingPosition otherRating = stepRatings.get(map.getProperties().wrap(pos, otherDir));
						if (/* (null == otherRating) || */ (0 < result.compareTo(otherRating))) {
							wasInserted = true;
							topDirections.add(i, dir);
							break;
						}
					}
				}
				if (!wasInserted) {
					topDirections.add(dir);
				}
			}
		}
		// Then, add all 45 and 90 degree deviations from those ideal paths
		final List<DigitalJoystickDirection> ret = new ArrayList<DigitalJoystickDirection>(8);
		ret.addAll(topDirections);
		final int[] deviations = new int[]{ -1, 1, -2, 2 };
		for (final int deviation : deviations) {
			for (final DigitalJoystickDirection dir : topDirections) {
				final DigitalJoystickDirection dir45 = dir.turnLeft8(deviation);
				if (!ret.contains(dir45)) {
					ret.add(dir45);
				}
			}
		}
		return topDirections;
	};

	/**
	 * Takes a list of move options and filters them if they lead anywhere.
	 * Currently only used in route planning orders.
	 * 
	 * @param gameView
	 *            the game to live in
	 * @param options
	 *            the list of move options
	 * @param stepRatingPurpose
	 *            if not NORMAL, other units in the way are not taken into
	 *            account
	 * @return the first non-empty move result. Empty would be <code>null</code>
	 *         or {@link MoveResult#MOVE_MASK_NOTHING}.
	 */
	MoveResult filterOptions(final GameView gameView, final List<DigitalJoystickDirection> options, final StepRatingPurpose stepRatingPurpose) {
		MoveResult ret = null;
		for (final DigitalJoystickDirection dir : options) {
			final MoveResult result = gameView.evaluateMove(this, dir, false, stepRatingPurpose);
			if (result.getResultMask() != MoveResult.MOVE_MASK_NOTHING) {
				ret = result;
				break;
			}
		}
		return ret;
	}

	public StepRatingMap getStepRatings() {
		return stepRatings;
	}

	StepRatingPosition getStepRating(final PlayfieldPosition pos) {
		if (null == stepRatings) {
			logger.warn(getKey().toString() + ".getStepRating(" + pos.toString() + ") without previous rating calc???", 
					new IllegalStateException(""));
			calculateRating();
		}
		return stepRatings.get(pos);
	}

	@Override
	public CommandableKey getKey() {
		return key;
	}

	@Override
	public Unit getUnit() {
		return this;
	}

	@Override
	public CommandableKey getParentKey() {
		return parentKey;
	}

	public void setParent(final CommandableKey key) {
		this.parentKey = key;
	}

	Collection<Unit> findChildren() {
		return gameState.findChildren(this);
	}

	/**
	 * @return the free capacity of this unit in load units
	 */
	int calcRemainingCapacity() {
		final int capacity = getUnitTemplate().getCapacity(gameState.getProperties().getRulesComplexity(), hitsTillDeath);
		final int holdsUsed = calcCargoHoldsUsed();
		return capacity - holdsUsed;
	}

	private int calcCargoHoldsUsed() {
		int ret = 0;
		final Collection<Unit> children = findChildren();
		final CommandableTemplate template = getUnitTemplate();
		if (null != children) {
			final RulesComplexity rulesComplexity = gameState.getProperties().getRulesComplexity();
			for (final Unit child : children) {
				final int holds = template.getCargoHoldsFor(rulesComplexity, child.getUnitTemplate());
				if (holds == Integer.MAX_VALUE) {
					throw new IllegalStateException("I have loaded a unit I shouldn't!");
				}
				ret += holds;
			}
		}
		return ret;
	}

	@Override
	public boolean canLoad(final UnitAlike unitToLoad, final boolean isNow) {
		boolean ret = false;
		if (getPlayerId() == unitToLoad.getPlayerId()) {
			int holdsRemaining;
			final RulesComplexity rulesComplexity = gameState.getProperties().getRulesComplexity();
			if (isNow) {
				holdsRemaining = calcRemainingCapacity();
			} else {
				holdsRemaining = getUnitTemplate().getMaximumCapacity(rulesComplexity);
			}
			final int holdsNeeded = getUnitTemplate().getCargoHoldsFor(rulesComplexity, unitToLoad.getUnitTemplate());
			ret = (holdsNeeded <= holdsRemaining);
		}
		return ret;
	}

	@Override
	public String toString() {
		return (getUnitTemplate().toString() + "(" + Integer.toString(getId()) + ")");
	}

	Collection<FieldAlike> findHomeBases() {
		final Collection<FieldCity> cities = getGameState().getMap().findCities(getPlayerId());
		final Collection<FieldAlike> fields = new LinkedList<FieldAlike>();
		fields.addAll(cities);
		if (getUnitTemplate().militaryBranch == MilitaryBranch.AIR_UNIT) {
			// Add all carriers (for fighters only) and airbases
			final Collection<Commandable> cmds = getGameState().findCommandables(getPlayerId());
			for (Commandable cmd : cmds) {
				if (cmd instanceof Unit) {
					final Unit target = (Unit) cmd;
					if (target.canLoad(this, false)) {
						fields.add(getGameState().getMap().getAt(target.getPosition()));
					}
				}
			}
		}
		return fields;
	}

	@Override
	public AttackMode getAttackModeAgainst(final Commandable opponent) {
		final RulesComplexity rulesComplexity = getGameState().getProperties().getRulesComplexity();
		final CommandableTemplate opponentTemplate;
		if (opponent instanceof FieldCity) {
			opponentTemplate = Templates.instance().CMDL_CITY;
		} else if (opponent instanceof Unit) {
			final Unit unit = (Unit) opponent;
			if ((getUnitTemplate().militaryBranch == MilitaryBranch.SEA_UNIT)
					&& (getGameState().getMap().getAt(unit.getPosition()).isCity())) {
				opponentTemplate = null;
			} else {
				opponentTemplate = unit.getUnitTemplate();
			}
		} else {
			throw new IllegalArgumentException("Unknown opponent type " + opponent.getClass().getName());
		}
		final AttackMode ret;
		if (null == opponentTemplate) {
			ret = AttackMode.IMPOSSIBLE;
		} else {
			ret = getUnitTemplate().getAttackModeAgainst(rulesComplexity, opponentTemplate);
		}
		return ret;
	}

	/**
	 * Skips the move; this might include healing.
	 * 
	 * @return a move result to show healing, or <code>null</code> otherwise
	 */
	MoveResult skipMove() {
		MoveResult ret = null;
		setSkippingThisTurn(true);
		// Is it docking for repair?
		if (isReadyForHealing()) {
			ret = new MoveResult(getGameView(), this, null);
			ret.setAffectedCommandable((FieldCity) (getGameView().getPlayerMap().getAt(getPosition()).getField()));
			ret.setResultMask(MoveResult.MOVE_MASK_DOCKING);
		}
		return ret;
	}

	/** 
	 * Clears all sentry orders for all containing units.
	 */
	void unload() {
		final Collection<Unit> children = findChildren();
		if (null != children) {
			for (final Unit child : children) {
				final Order order = child.getOrder();
				if ((null != order) && (order instanceof OrderSentry)) {
					child.setOrder(null);
				}
			}
		}
	}

	/**
	 * Returns if this unit at <em>unitPos</em> is visible to the
	 * <em>observer</em> at <em>observerPos</em>.
	 * 
	 * @param unitPos
	 *            the position of this unit to assume. May be different than
	 *            current position for animations
	 * @param observer
	 *            the observer commandable to check for if this unit is visible
	 * @param moveResult
	 *            the move the unit is part of (or <code>null</code> if the unit
	 *            is not involved in a move) to check if observed
	 * @return <code>true</code> if visible
	 */
	public boolean isVisibleFor(final PlayfieldPosition unitPos, final Commandable observer, final MoveResult moveResult) {
		final CommandableTemplate unitTemplate = getUnitTemplate();
		final int bitmask = unitTemplate.visibilityBitmask;
		boolean ret;
		if (PlayfieldPosition.equals(unitPos, observer.getPosition())) {
			ret = true;
		} else if (Bitfield32.isBit(bitmask, observer.getKey().getTemplateId()) 
				&& (gameState.getProperties().calcAirDistance(unitPos, observer.getPosition()) 
						<= observer.getSightingRange(unitTemplate.militaryBranch))) {
			ret = ((null != moveResult) && (0 != (moveResult.getResultMask() | MoveResult.MOVE_MASK_FIGHTING)))
						|| ((null == getParentKey()) 
								&& (ImageKey.FIELD_ID_CITY != gameState.getMap().getAt(unitPos).getTemplateId()));
		} else {
			ret = false;
		}
		return ret;
	}

	@Override
	public int hashCode() {
		return ((key == null) ? 0 : key.hashCode());
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (!(obj instanceof UnitAlike)) {
			return false;
		}
		UnitAlike other = (UnitAlike) obj;
		if (key == null) {
			if (other.getKey() != null) {
				return false;
			}
		} else if (!key.equals(other.getKey())) {
			return false;
		}
		return true;
	}

	@Override
	public int compareTo(final UnitAlike other) {
		return key.compareTo(other.getKey());
	}

	/**
	 * Disbands the unit, so it won't be on the map any more.
	 */
	public void disband() {
		setParent(null);
		setSeenEnemiesLately(false);
		setSkippingThisTurn(true);
//		Don't set field to null, as then there is no memory where the unit has been, for logging etc
//		field = null;
	}

	@Override
	public int getTurnOfSighting() {
		return UnitAlike.CURRENT;
	}
}
