package de.ped.empire.logic;

import java.awt.Graphics;
import java.awt.Point;
import java.util.Collection;
import java.util.TreeSet;

import de.ped.tools.Assert;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.TopologyMode;

/**
 * A field how a player sees it.
 * 
 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ViewField implements FieldAlike {

	public static FieldAlike create(final Marshaller marshaller, final GameView view, final int x, final int y) {
		return new ViewField(marshaller, (Field) view.getGameState().getMap().getAt(x, y));
	}

	ViewField(final Field field, final boolean isUnexplored) {
		this.stateField = field;
		setUnexplored(isUnexplored);
	}

	ViewField(final Marshaller marshaller, final Field field) {
		this.stateField = field;
		// TODO: Persist only isUnexplored and unitInProduction, should fit in one byte.
		this.imageKey = marshaller.readInt();
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(imageKey);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		throw Marshaller.createConstantMarshallableDemarshallingException();
	}

	@Override
	public boolean canLoad(final UnitAlike unitToLoad, final boolean isNow) {
		return stateField.canLoad(unitToLoad, isNow);
	}

	@Override
	public PlayfieldPosition getPosition() {
		return stateField.getPosition();
	}

	@Override
	public int getImageKey() {
		return imageKey;
	}

	@Override
	public int getPlayerId() {
		final int ret;
		if (isUnexplored()) {
			ret = Player.ID_NONE;
		} else {
			ret = stateField.getPlayerId();
		}
		return ret;
	}

	@Override
	public I18NStringWithFillIns getName() {
		return getTemplate().name;
	}

	@Override
	public I18NStringWithFillIns getToolTip(final GameView gameView) {
		final I18NStringWithFillIns ret;
		if (Templates.instance().FIELD_CITY.id == getTemplateId()) {
			FieldCity city = (FieldCity) stateField;
			if (0 == ImageKey.getBits(imageKey, ImageKey.BA_OUTDATED)) {
				ret = city.getToolTip(gameView);
			} else {
				Player player = ImageKey.getPlayer(gameView.getGameState(), imageKey);
				ret = FieldCity.getToolTip(gameView, city, player.getId(), city.getProductivity(), Templates.instance().UNDEFINED, 0, false);
				TileTemplate.logger.debug("Seen outdated city.");
			}
		} else {
			//ret = new I18NStringWithFillIns(null, "%0 Soso", new I18NStringWithFillIns[] { getName() } );
			ret = getName();
		}
		return ret;
	}

	@Override
	public char getLogRepresentation() {
		return TileTemplate.getTTD(imageKey).logRepresentation;
	}

	@Override
	public void paint(final Graphics g, final Point offset, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		stateField.paint(g, offset, zoomLevel, topologyMode);
	}

	@Override
	public void paint(final Graphics g, final Point offset, final float x, final float y, final int imgKey, final float dyingProgress, final ZoomLevel zoomLevel, final TopologyMode topologyMode) {
		stateField.paint(g, offset, x, y, imgKey, dyingProgress, zoomLevel, topologyMode);
	}

	@Override
	public int compareTo(final FieldAlike other) {
		return stateField.compareTo(other);
	}

	@Override
	public boolean isLand() {
		return getTemplate().isLand;
	}

	@Override
	public boolean isCity() {
		return (getTemplate().id == ImageKey.FIELD_ID_CITY);
	}

	@Override
	public boolean isUnexplored() {
		return (ImageKey.FIELD_ID_UNEXPLORED == ImageKey.getBits(imageKey, ImageKey.BA_TEMPLATE))
				&& (ImageKey.TType.Field.ordinal() == ImageKey.getBits(imageKey, ImageKey.BA_TTYPE));
	}
	
	public void setUnexplored(final boolean isUnexplored) {
		int val;
		if (isUnexplored) {
			val = 0x00;
			val = ImageKey.setBits(val, ImageKey.BA_TTYPE, ImageKey.TType.Field.ordinal());
			val = ImageKey.setBits(val, ImageKey.BA_TEMPLATE, ImageKey.FIELD_ID_UNEXPLORED);
		} else {
			val = stateField.getImageKey();
		}
		setImageKey(val);
	}
	
	@Override
	public int getCapacity(final Unit unit) {
		final int ret;
		if (isUnexplored()) {
			ret = getTemplate().getCapacity(unit.getGameState(), unit.getUnitTemplate());
		} else {
			ret = stateField.getCapacity(unit);
		}
		return ret;
	}

	@Override
	public byte getTemplateId() {
		return getTemplate().id;
	}

	@Override
	public FieldTemplate getTemplate() {
		final FieldTemplate ret;
		if (isUnexplored()) {
			ret = Templates.instance().UNEXPLORED;
		} else {
			ret = stateField.getTemplate();
		}
		return ret;
	}

	@Override
	public void setImageKey(final int imageKey) {
		this.imageKey = imageKey;
	}
	
	@Override
	public Field getField() {
		Assert.assertFalse(isUnexplored());
		return stateField;
	}


	@Override
	public Collection<UnitAlike> getUnits() {
		return units;
	}

	/** Non-persistent. */
	private TreeSet<UnitAlike> units = new TreeSet<UnitAlike>();
	/** Non-persistent. */
	private final Field stateField;
	/** Non-persistent. */
	private int imageKey;
}
