package de.ped.empire.logic;

import java.util.ArrayList;

import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;

/**
 * Knows all about a zoom level.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ZoomLevel {

	/** All zoom levels (resolutions) selectable to show the map. */
	public static final ZoomLevel[] INSTANCES = new ZoomLevel[] {
		new ZoomLevel(32, 4.0f, "mapicons32.png", "mapicons32Hex.png", true, true),
		new ZoomLevel(24, 4.0f, "mapicons24.png", "mapicons24Hex.png", true, true),
		new ZoomLevel(16, 3.0f, "mapicons16.png", "mapicons16Hex.png", false, true),
		new ZoomLevel( 8, 2.0f, "mapicons08.png", "mapicons08Hex.png", false, false),
		new ZoomLevel( 4, 1.0f, "mapicons04.png", "mapicons04Hex.png", false, false),
		new ZoomLevel( 2, 0.5f, "mapicons02.png", "mapicons02Hex.png", false, false),
	};
	static {
		for (int i = 0; i < INSTANCES.length; i++) {
			INSTANCES[i].init(i);
		}
	}

	/** The zoom level (resolution) in which to draw the icons in dialogs. */
	public static final ZoomLevel DIALOG_ZOOM_LEVEL = INSTANCES[2];

	/** The zoom level (resolution) in which to draw the icons in context menus. */
	public static final ZoomLevel MENU_ZOOM_LEVEL = INSTANCES[2];

	/** The zoom level (resolution) a new player starts with. */
	public static final int DEFAULT_ZOOM_LEVEL_INDEX = 1;

	/** The size in pixels of a map field when drawn in this zoom level. As all icons are square, the value is the same for width and height. */
	public final int resolution;
	/** The width in pixels of a stroke to draw. */
	public final float strokeWidth;
	/** If permitted to draw distance infos on map in this zoom level. */
	public final boolean isDistanceInfoPermitted;
	/** If hexagonal maps should be drawn in real hexagonal shapes. */
	public final boolean isHexagonalShaped;
	/** The image sets of that zoom level for the different topologies */
	private ImageSet[] imageSets;
	/** The name of the resource file holding the icons of this zoom level for square maps. */
	public final String squareResourceName;
	/** The name of the resource file holding the icons of this zoom level for hexagon maps. */
	public final String hexagonResourceName;

	private ZoomLevel(final int resolution, final float strokeWidth, final String resourceName, final String hexagonResourceName,
			final boolean isDistanceInfoPermitted, final boolean isHexagonalShaped) {
		this.resolution = resolution;
		this.strokeWidth = strokeWidth;
		this.squareResourceName = resourceName;
		this.hexagonResourceName = hexagonResourceName;
		this.isDistanceInfoPermitted = isDistanceInfoPermitted;
		this.isHexagonalShaped = isHexagonalShaped;
	}

	private void init(final int zoomIdx) {
		this.imageSets = new ImageSet[] {
				new ImageSet(new ImageSet.Key(zoomIdx, TopologyMode.SQUARE8), squareResourceName),
				new ImageSet(new ImageSet.Key(zoomIdx, TopologyMode.HEXAGON), hexagonResourceName), };
	}

	public ImageSet getImageSet(final TopologyMode topologyMode) {
		final int idx = ImageSet.Key.getTopologyIndex(topologyMode);
		return imageSets[idx];
	}

	public PlayfieldTiling getPlayfieldTiling(final TopologyMode topologyMode) {
		if (null == tilings.get(topologyMode.ordinal())) {
			final TopologyMode shapeMode;
			if (((TopologyMode.HEXAGON == topologyMode) || (TopologyMode.HEXAGON_H == topologyMode)) 
					&& (!isHexagonalShaped)) {
				shapeMode = TopologyMode.SQUARE8;
			} else {
				shapeMode = topologyMode;
			}
			tilings.set(topologyMode.ordinal(), new PlayfieldTiling(topologyMode, shapeMode, resolution, resolution, 1, 1, 2));
		}
		return tilings.get(topologyMode.ordinal());
	}

	private ArrayList<PlayfieldTiling> tilings = new ArrayList<PlayfieldTiling>(TopologyMode.values().length);
	{
		for (int i = 0; i < TopologyMode.values().length; i++) {
			tilings.add(null);
		}
	}

	@Override
	public String toString() {
		return "ZoomLevel" + String.format("%02d", resolution);
	}
}
