package de.ped.empire.test;

import static org.junit.Assert.*;

import java.util.Arrays;
import java.util.Collection;

import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import de.ped.empire.logic.GameFacade;
import de.ped.empire.logic.GameMap;
import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.GameState;
import de.ped.tools.BorderMode;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * A test for
 * {@link GameFacade#createGame(de.ped.empire.logic.GameObserver, int, GameProperties, TutorialLevel)}
 * on non-tutorial games that implicitly checks the creation of a {@link GameState} including its
 * {@link GameMap}.
 * 
 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
 */
@RunWith(value=Parameterized.class)
public class GameFacadeCreateGameTest {

	/**
	 * Parameters for a single test run.
	 * 
	 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class GameMapTestPropertyParams {
		final BorderMode borderMode;
		final int x;
		final int y;
		final int city;
		final int land;
		final int cont;

		public GameMapTestPropertyParams(BorderMode borderMode, int x, int y, int city, int land, int cont) {
			this.borderMode = borderMode;
			this.x = x;
			this.y = y;
			this.city = city;
			this.land = land;
			this.cont = cont;
		}

		@Override
		public String toString() {
			return "GameMapTestProperties(" + borderMode.name() + "," + x + "," + y + "," + city + "," + land + ","
					+ cont + ")";
		}
	}
	
	@Parameters
	public static Collection<GameMapTestPropertyParams> data() {
		return Arrays.asList(new GameMapTestPropertyParams[] {
			// Minimum size tests with borders
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.ALL, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max), 
			// Minimum size tests with no borders
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max), 
			// Maximum size tests
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.max, GameProperties.SLIDERDEF_SIZE.max, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.NONE, 
					GameProperties.SLIDERDEF_SIZE.max, GameProperties.SLIDERDEF_SIZE.max, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max), 
			// Other sizes
			new GameMapTestPropertyParams(BorderMode.HORIZONTALLY, 
					GameProperties.SLIDERDEF_SIZE.max, GameProperties.SLIDERDEF_SIZE.min, 
					GameProperties.SLIDERDEF_CONTENT.min, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max), 
			new GameMapTestPropertyParams(BorderMode.HORIZONTALLY, 
					GameProperties.SLIDERDEF_SIZE.min, GameProperties.SLIDERDEF_SIZE.max, 
					GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.max, GameProperties.SLIDERDEF_CONTENT.min), 
			}
		);
	}

	private GameMapTestPropertyParams prop;

	public GameFacadeCreateGameTest(GameMapTestPropertyParams params) {
		this.prop = params;
	}

	@Test
	public void testMultipleMaps() {
		checkMultiple(createProps(prop.borderMode, prop.x, prop.y, prop.city, prop.land, prop.cont));
	}

	private GameProperties createProps(BorderMode borderMode, int x, int y, int city, int land, int cont) {
		final GameProperties props = new GameProperties();
		props.setBorderMode(borderMode);
		props.setSize(0, x);
		props.setSize(1, y);
		props.setCityDensity(city);
		props.setLandRatio(land);
		props.setContinentSize(cont);
		return props;
	}

	private void checkMultiple(final GameProperties props) {
		final GameObserverStub obs = new GameObserverStub();
		final GameFacade gameFacade = new GameFacade();
		for (int i = 0; i < RUNS; i++) {
			logger.info("Props=" + props);
			int tries = gameFacade.createGame(obs, 0, props, GameFacade.TUTORIAL_LEVEL_NONE);
			logger.info("Tries=" + tries);
			assertNotEquals(GameFacade.FAILED, tries);
		}
	}

	private static final int RUNS = 1;

	private Logger logger = Logger.getLogger(this.getClass());
	{
		logger.setLogLevel(Level.DEBUG);
	}
}
