package de.ped.empire.test;

import static org.junit.Assert.*;

import java.util.List;

import org.junit.Test;

import de.ped.empire.logic.FieldAlike;
import de.ped.empire.logic.GameMap;
import de.ped.empire.logic.GameProperties;
import de.ped.tools.BorderMode;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;

/**
 * Tests for the {@link GameMap} class.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameMapTest {

	static GameMap checkInitGame(final GameProperties props) {
		final GameMap map = new GameMap(props);
		map.addAtRandom(props);
		map.finish();
		// Check for null entries in the map
		final int width = map.getSize().getWidth();
		for (int x = 0; x < width; x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final FieldAlike field = map.getAt(x, y);
				assertNotNull("x=" + x + ",y=" + y, field);
			}
		}
		// marshal/unmarshal
		marshalDemarshal(props, map);
		assertTrue(6 <= map.getAllCities().size());
		return map;
	}

	/**
	 * Marshals and demarshals a map, and does basic checks on it.
	 * @param props the game properties the map was based on
	 * @param map the map itself
	 * @return a copy of the map as result of the serialization and deserialization
	 */
	static GameMap marshalDemarshal(final GameProperties props, final GameMap map) {
		final Marshaller m = new Marshaller();
		final GameMap.Factory factory = new GameMap.Factory(props, null);
		m.writeMarshallable(map);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		final GameMap received = (GameMap)d.readMarshallable(factory);
		compareMaps(map, received);
		return received;
	}

	static void compareMaps(final GameMap map, final GameMap received) {
		assertEquals(map.getSize(), received.getSize());
		final PlayfieldDimension size = map.getSize();
		final GameProperties props = map.getProperties();
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				final FieldAlike expected = map.getAt(x, y);
				final FieldAlike result = received.getAt(x, y);
				assertEquals("x=" + x + ",y=" + y, expected.getTemplateId(), result.getTemplateId());
				assertEquals("x=" + x + ",y=" + y, expected.getClass(), result.getClass());
			}
		}
	}

	@Test
	public void testDefaultCreation() {
		final GameProperties props = new GameProperties();
		checkInitGame(props);
	}

	@Test
	public void testNoBorder() {
		final GameProperties props = new GameProperties();
		props.setBorderMode(BorderMode.HORIZONTALLY);
		checkInitGame(props);
	}

	@Test
	public void testGetAllFieldsInAirDistance() {
		final GameProperties props = new GameProperties();
		checkGetAllFieldsInAirDistance(props, TopologyMode.SQUARE8, new int[] { 1, 8, 16, }, 0);
		checkGetAllFieldsInAirDistance(props, TopologyMode.SQUARE4, new int[] { 1, 4,  8, }, 1);
		checkGetAllFieldsInAirDistance(props, TopologyMode.HEXAGON, new int[] { 1, 6, 12, }, 2);
	}

	public void checkGetAllFieldsInAirDistance(
			final GameProperties props, 
			final TopologyMode topologyMode, 
			final int[] sizes, 
			final int checkMode) {
		List<FieldAlike> result;
		props.setTopologyMode(topologyMode);
		final GameMap map = new GameMap(props);
		int sizesSum = 0;
		for (int s = 0; s < sizes.length; s++) {
			sizesSum += sizes[s];
			final String st = "s=" + s;
			result = map.getAllFieldsInAirDistance(map.getAt(CENTER), s, false, 0);
			assertEquals(st, sizes[s], result.size());
			for (final FieldAlike f : result) {
				final PlayfieldPosition pos = f.getField().getPosition();
				final String str = st + ",f=" + pos.toString();
				switch(checkMode) {
				case 0:
					assertTrue(str, Math.abs(pos.get(0) - CENTER.get(0)) <= s);
					assertTrue(str, Math.abs(pos.get(1) - CENTER.get(1)) <= s);
					assertTrue(str, Math.abs(pos.get(0) - CENTER.get(0)) == s || Math.abs(pos.get(1) - CENTER.get(1)) == s);
					break;
				case 1:
					assertTrue(str, Math.abs(pos.get(0) - CENTER.get(0)) + Math.abs(pos.get(1) - CENTER.get(1)) == s);
					break;
				case 2:
					assertTrue(str, Math.abs(pos.get(0) - CENTER.get(0)) <= s);
					assertTrue(str, Math.abs(pos.get(1) - CENTER.get(1)) <= s);
					break;
				}
			}
			checkForDuplicates(st, result);
			checkForZeroSum(st, result);
			result = map.getAllFieldsInAirDistance(map.getAt(CENTER), s, true, 0);
			assertEquals(st, sizesSum, result.size());
			for (final FieldAlike f : result) {
				final PlayfieldPosition pos = f.getField().getPosition();
				final String str = "s=" + s + ",f=" + pos.toString();
				switch(checkMode) {
				case 0:
				case 2:
					assertTrue(str, Math.abs(pos.get(0) - CENTER.get(0)) <= s);
					assertTrue(str, Math.abs(pos.get(1) - CENTER.get(1)) <= s);
					break;
				case 1:
					assertTrue(str, Math.abs(pos.get(0) - CENTER.get(0)) + Math.abs(pos.get(1) - CENTER.get(1)) <= s);
					break;
				}
			}
			checkForDuplicates(st, result);
			checkForZeroSum(st, result);
		}
	}

	public void checkForDuplicates(final String str, final List<FieldAlike> result) {
		for (int i = 0; i < result.size(); i++) {
			for (int j = 0; j < i; j++) {
				assertFalse(str + ":" + i + "=>" + result.get(i).getPosition().toString() + "," + j + "=>" + result.get(j).getPosition().toString(),
						PlayfieldPosition.equals(result.get(i).getPosition(), result.get(j).getPosition()));
			}
		}
	}

	public void checkForZeroSum(final String str, final List<FieldAlike> result) {
		int x = 0;
		@SuppressWarnings("unused")
		int y = 0;
		String s = str;
		for (final FieldAlike f : result) {
			final PlayfieldPosition pos = f.getField().getPosition();
			x += pos.get(0) - CENTER.get(0); 
			y += pos.get(1) - CENTER.get(1); 
			s += pos.toString();
		}
		assertEquals(s, 0, x);
		if (PlayfieldTiling.IS_HEX_COMPACT) {
			// Nothing to do
		} else {
			assertEquals(s, 0, y);
		}
	}

	private final PlayfieldPosition CENTER = new PlayfieldPosition(2, 4);
}
