package de.ped.empire.test;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import org.junit.After;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;

import de.ped.empire.logic.CircularReferenceException;
import de.ped.empire.logic.Commandable;
import de.ped.empire.logic.GameInitializationException;
import de.ped.empire.logic.GameState;
import de.ped.empire.logic.GameView;
import de.ped.empire.logic.GameMap;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.MapViewMode;
import de.ped.empire.logic.MilitaryBranch;
import de.ped.empire.logic.MovePath;
import de.ped.empire.logic.MoveResult;
import de.ped.empire.logic.Order;
import de.ped.empire.logic.OrderBuildAirbase;
import de.ped.empire.logic.OrderDefineMovePaths;
import de.ped.empire.logic.OrderDisband;
import de.ped.empire.logic.OrderEscort;
import de.ped.empire.logic.OrderGoTo;
import de.ped.empire.logic.OrderLoad;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.ProcessorSpeedType;
import de.ped.empire.logic.Templates;
import de.ped.empire.logic.Tutorial;
import de.ped.empire.logic.TutorialLevel;
import de.ped.empire.logic.Unit;
import de.ped.empire.logic.CommandableTemplate;
import de.ped.empire.logic.UnitAlike;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Tests gameplay scenarios where units are moved.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameMoveTest {

	/**
	 * Clears the animation sequence.
	 * @param gameView the game view to clear the animation sequence for
	 * @param observer the observer for the game view
	 * @return the result mask of the first animation
	 */
	static int clearAnimation(
			final GameView gameView, 
			final GameObserverStub observer) {
		int ret = 0;
		final ProcessorSpeedType pst = ProcessorSpeedType.SLOW;
		gameView.step(pst);
		if ((null != observer) && (null != observer.getNotif() && (null != observer.getNotif().getAnimationMoveResult()))) {
			ret = observer.getNotif().getAnimationMoveResult().getResultMask();
		}
		while (null != gameView.getCurrentAnimation()) {
			gameView.step(pst);
		}
		return ret;
	}

	/**
	 * Clears the animation sequence.
	 * @param gameView The game view to clear the animation sequence for.
	 * @param observer The observer for the game view.
	 * @param expectedNumberOfAnimations The expected number of animations. If the found number is different, an assertion fails.
	 * @return the result mask of the first animation
	 */
	static int clearAnimation(
			final GameView gameView, 
			final GameObserverStub observer, 
			final int expectedNumberOfAnimations) {
		assertEquals(expectedNumberOfAnimations, gameView.getNumberOfCurrentAnimations());
		return clearAnimation(gameView, observer);
	}

	@Before
	public void setUp() throws GameInitializationException {
		final Tutorial tutorial = new Tutorial();
		final GameMap map = tutorial.createMap(TutorialLevel.LAND_UNITS);
		observer = new GameObserverStub();
		ownCity = map.findCities(0).get(0);
		enemyCity = map.findCities(1).get(0);
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		gameView = gameState.getCurrentGameView();
		gameView.clearGetReady();
	}

	@After
	public void tearDown() {
		testIfUnitsClonedForViews(gameView, gameView.getUnits());
		if (null != gameView.getMap()) {
		    final Marshaller marshaller = new Marshaller();
		    marshaller.writeMarshallable(gameView.getGameState());
		    final byte[] data = marshaller.pack();
			final Marshaller d = new Marshaller(data);
			final GameState received = new GameState(observer, d);
			GameMapTest.compareMaps(gameView.getGameState().getMap(), received.getMap());
			for (int i = 0; i < Player.MAX_NUMBER_OF_PLAYERS; i++) {
				GameMapTest.compareMaps(gameView.getGameState().getGameView(i).getMap(), received.getGameView(i).getMap());
			}
			testIfUnitsClonedForViews(gameView, received.getGameView(0).getUnits());
		}
	}

	static void testIfUnitsClonedForViews(final GameView gameView, final Collection<UnitAlike> units) {
		for (final UnitAlike unit : units) {
			for (final Unit original : gameView.getGameState().getAllUnits()) {
				if (unit.getKey().equals(original.getKey())) {
					if (gameView.isUsingFogOfWar()) {
						assertFalse(original == unit);
					} else {
						assertEquals(original, unit);
					}
				}
				break;
			}
		}
	}

	@Test
	public void testAddMovePath_ToUnknownTerritory_Fails() throws CircularReferenceException {
		final int playerId = 0;
		final PlayfieldPosition startPos = ownCity.getPosition();
		final PlayfieldPosition endPos = new PlayfieldPosition(9, 5);
		final MovePath path = new MovePath(playerId, startPos, 0x0, endPos);
		assertFalse(gameView.makesSense(path, MilitaryBranch.LAND_UNIT));
	}

	@Test
	public void testAddMovePath_ToOtherCity_Works() throws CircularReferenceException {
		final int playerId = 0;
		final PlayfieldPosition startPos = ownCity.getPosition();
		final PlayfieldPosition endPos = new PlayfieldPosition(9, 5);
		final GameState gameState = gameView.getGameState();
		final FieldCity endField = (FieldCity) gameState.getMap().getAt(endPos);
		endField.setPlayerId(playerId);
		gameView.updateViewMap();
		final MovePath path = new MovePath(playerId, startPos, 0x0, endPos);
		assertTrue(gameView.makesSense(path, MilitaryBranch.LAND_UNIT));
		try {
			gameView.getGameState().getMovePaths().add(path);
		} catch (final CircularReferenceException ex) {
			throw ex;
		}
		Order order = new OrderDefineMovePaths(gameView.getGameState(), ownCity);
		assertTrue(order.makesSense());
	}

	@Test
	public void testAddMovePath_ToOtherCityWithStartingCitySurrounded_Works() throws CircularReferenceException {
		final int playerId = 0;
		final PlayfieldPosition startPos = ownCity.getPosition();
		final PlayfieldPosition endPos = new PlayfieldPosition(9, 5);
		final GameState gameState = gameView.getGameState();
		final FieldCity endField = (FieldCity) gameState.getMap().getAt(endPos);
		endField.setPlayerId(playerId);
		Tutorial.createUnit(gameState, ownCity, Templates.instance().TANK, new PlayfieldPosition(2, 3));
		Tutorial.createUnit(gameState, ownCity, Templates.instance().TANK, new PlayfieldPosition(3, 2));
		Tutorial.createUnit(gameState, ownCity, Templates.instance().TANK, new PlayfieldPosition(3, 3));
		gameView.updateViewMap();
		final MovePath path = new MovePath(playerId, startPos, 0x0, endPos);
		assertTrue(gameView.makesSense(path, MilitaryBranch.LAND_UNIT));
		try {
			gameView.getGameState().getMovePaths().add(path);
		} catch (final CircularReferenceException ex) {
			throw ex;
		}
		Order order = new OrderDefineMovePaths(gameView.getGameState(), ownCity);
		assertTrue(order.makesSense());
	}

	@Test
	public void testArmyRemove_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(1, 2);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderDisband order = new OrderDisband(gameView.getGameState(), unit1);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		assertEquals(MoveResult.MOVE_MASK_LOSING_FIGHT, clearAnimation(1));
	}

	@Test
	public void testArmyMovesIntoSea_NotOffered() {
		final PlayfieldPosition destination = new PlayfieldPosition(1, 2);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertFalse(order.makesSense());
	}

	@Test
	public void testArmyMovesOntoOtherIsland_NotOffered() {
		final PlayfieldPosition destination = new PlayfieldPosition(2, 7);
		assertEquals(Templates.instance().LAND.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		// With fog of war, user can set a goto command into the unknown. Without, known sea is in the way.
		assertEquals(gameView.isUsingFogOfWar(), order.makesSense());
	}

	@Test
	public void testArmyMovesOntoEmptyLandDistance1_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(3, 2);
		assertEquals(Templates.instance().LAND.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(1));
		assertEquals(destination, unit1.getPosition());
	}

	@Test
	public void testArmyGoToNeutralCityDistance2_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(4, 2);
		assertEquals(Templates.instance().FIELD_CITY.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(1));
		assertEquals(new PlayfieldPosition(3, 2), unit1.getPosition());
	}

	@Test
	public void testArmyGoToNeutralCityDistance1_Works() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(4, 2);
		final FieldCity enemy = (FieldCity) gameView.getGameState().getMap().getAt(destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, new PlayfieldPosition(3, 2));
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		{
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT, clearAnimation(1));
		} {
			assertEquals(unit1, gameView.getVisibleCommandableInFocus());
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertTrue(0 != (MoveResult.MOVE_MASK_FIGHTING & clearAnimation(1)));
		}
	}

	private void checkFighterScenarios(final int[][] destination, final int[] moveResultMasks) {
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().FIGHTER, new PlayfieldPosition(2, 3));
		createUnit(gameView, ownCity, Templates.instance().CARRIER, new PlayfieldPosition(5, 1));
		createUnit(gameView, ownCity, Templates.instance().AIRBASE, new PlayfieldPosition(6, 2));
		// Already virtually let the figher lose 12 step points
		unit1.finishTurn();
		unit1.finishTurn();
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final Logger logger = Logger.getLogger(getClass());
		logger.setLogLevel(Level.INFO);
		for (int step = 0; step < destination.length; step++) {
			logger.debug("testFighter step=" + step);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, new PlayfieldPosition(destination[step]));
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(moveResultMasks[step], clearAnimation(1));
		}
		assertEquals(false, unit1.isNextable());
	}

	@Test
	public void testFighterGoToWithLowFuel_Dies() {
		final int[][] destination = new int[][] { 
				{ 3, 3 }, { 4, 4 }, { 5, 5 }, { 6, 6 }, { 5, 5 }, { 4, 4 }, };
		final int[] masks = new int[] { 
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_FUEL_EMPTY,
		};
		checkFighterScenarios(destination, masks);
	}

	@Test
	public void testFighterGoToHomeCityWithLowFuel_Works() {
		final int[][] destination = new int[][] { 
				{ 3, 3 }, { 4, 4 }, { 5, 5 }, { 4, 4 }, { 3, 3 }, { 2, 2 }, };
		final int[] masks = new int[] { 
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_DOCKING,
		};
		checkFighterScenarios(destination, masks);
	}

	@Test
	public void testFighterGoToCarrierWithLowFuel_Works() {
		final int[][] destination = new int[][] { 
				{ 3, 3 }, { 4, 4 }, { 5, 4 }, { 5, 3 }, { 5, 2 }, { 5, 1 }, };
		final int[] masks = new int[] { 
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_LOADING,
		};
		checkFighterScenarios(destination, masks);
	}

	@Test
	public void testFighterGoToCarrierWithoutLowFuel_Works() {
		final int[][] destination = new int[][] { 
				{ 3, 3 }, { 4, 4 }, { 5, 3 }, { 5, 2 }, { 5, 1 }, };
		final int[] masks = new int[] { 
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_LOADING,
		};
		checkFighterScenarios(destination, masks);
	}

	@Test
	public void testFighterGoToAirbaseWithLowFuel_Works() {
		final int[][] destination = new int[][] { 
				{ 3, 3 }, { 4, 4 }, { 5, 4 }, { 6, 4 }, { 6, 3 }, { 6, 2 }, };
		final int[] masks = new int[] { 
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING,
				MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_LOADING,
		};
		checkFighterScenarios(destination, masks);
	}

	/** The test is broken, but the real use case works... */
	@Ignore
	public void testFighterGoToMovesAroundCity_Works() {
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().FIGHTER, new PlayfieldPosition(1, 1));
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		PlayfieldPosition destination = new PlayfieldPosition(3, 3);
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(3));
		assertEquals(destination, unit1.getPosition());
	}

	@Test
	public void testArmyMovesOntoEnemyLandDistance1_Works() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(3, 2);
		assertEquals(Templates.instance().LAND.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit enemy = createUnit(gameView, enemyCity, Templates.instance().ARMY, destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		{
			gameView.nextCommandable(unit1);
			assertEquals(unit1, gameView.getVisibleCommandableInFocus());
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT, clearAnimation(1));
		} {
			assertEquals(unit1, gameView.getVisibleCommandableInFocus());
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertTrue(0 != (MoveResult.MOVE_MASK_FIGHTING & clearAnimation(1)));
		}
	}

	@Test
	public void testArmyMovesOntoOwnLandDistance1_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(3, 2);
		assertEquals(Templates.instance().LAND.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		@SuppressWarnings("unused")
		final Unit other = createUnit(gameView, ownCity, Templates.instance().ARMY, destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		{ // Move directly onto the unit, should fail.
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertFalse(order.makesSense());
		} { // Move behind the unit onto a neutral city, should route around.
			gameView.nextCommandable(unit1);
			final PlayfieldPosition destination2 = new PlayfieldPosition(4, 2);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination2);
//			gameView.getMap().logMap();
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(1));
			assertEquals(new PlayfieldPosition(3, 3), unit1.getPosition());
		}
	}

	@Test
	public void testTransporterMovesOntoLand_Fails() {
		final PlayfieldPosition destination = new PlayfieldPosition(4, 3);
		assertEquals(Templates.instance().LAND.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, null);
		assertEquals(2, unit1.getStepsLeftThisTurn());
		{
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertFalse(order.makesSense());
		}
	}

	@Test
	public void testTransporterMovesOntoSeaDistance2_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(4, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, null);
		assertEquals(2, unit1.getStepsLeftThisTurn());
		{
			gameView.nextCommandable(unit1);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			{
//				assertEquals(MoveResult.MOVE_MASK_MOVED, clearAnimation(1));
//				assertEquals(1, unit1.getStepsLeftThisTurn());
//				assertArrayEquals(new int[]{ 3, 1}, unit1.getPosition());
//			} {
//				gameView.nextMove();
				assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(2));
				assertEquals(new PlayfieldPosition(4, 1), unit1.getPosition());
				assertEquals(0, unit1.getStepsLeftThisTurn());
			}
		}
	}

	@Test
	public void testTransporterMovesOntoSeaDistance3_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(5, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, null);
//		gameView.getMap().logMap();
		assertEquals(2, unit1.getStepsLeftThisTurn());
		{
			gameView.nextCommandable(unit1);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			{
//				assertEquals(MoveResult.MOVE_MASK_MOVED, clearAnimation(1));
//				assertArrayEquals(new int[]{ 3, 1}, unit1.getPosition());
//				assertEquals(1, unit1.getStepsLeftThisTurn());
//			} {
//				gameView.nextMove();
				assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(2));
				assertEquals(new PlayfieldPosition(4, 1), unit1.getPosition());
				assertEquals(0, unit1.getStepsLeftThisTurn());
			}
		}
	}

	@Test
	public void testTransporterMovesOntoNeutralCity_Fails() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(4, 2);
		final FieldCity enemy = (FieldCity) gameView.getGameState().getMap().getAt(destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, new PlayfieldPosition(3, 2));
		assertEquals(2, unit1.getStepsLeftThisTurn());
		{ // Just moving is not enough
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertFalse(order.makesSense());
		} { // Now try an attack, should also fail
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertFalse(order.makesSense());
		}
	}

	@Test
	public void testTransporterMovesOntoEnemySeaDistance1_Works() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(3, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit enemy = createUnit(gameView, enemyCity, Templates.instance().TRANSPORTER, destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, null);
		{
			gameView.nextCommandable(unit1);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT, clearAnimation(1));
		} {
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertTrue(0 != (MoveResult.MOVE_MASK_FIGHTING & clearAnimation(1)));
		}
	}

	@Test
	public void testTransporterMovesOntoEnemySeaDistance2_Works() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(4, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit enemy = createUnit(gameView, enemyCity, Templates.instance().TRANSPORTER, destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, null);
		final PlayfieldPosition posAfterStep1 = new PlayfieldPosition();
		{ // Step 1
			gameView.nextCommandable(unit1);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(2));
			posAfterStep1.setLocation(unit1.getPosition());
		} { // Step 2 as GoTo
			gameView.nextCommandable(unit1);
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT, clearAnimation(1));
			assertEquals(posAfterStep1, unit1.getPosition());
		} { // Step 2 as Attack
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertTrue(0 != (MoveResult.MOVE_MASK_FIGHTING & clearAnimation(1)));
		}
	}

	/**
	 * Checks if a sea unit attacking an enemy city populated by a land unit,
	 * sea unit and air unit correctly fails to find a target to attack.
	 * 
	 * @throws CircularReferenceException if the new move path would lead to a circular reference for at least one military branch
	 */
	@Test
	public void testCarrierAttacksEnemyCity_Fails() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(11, 7);
		final FieldCity enemyCity = (FieldCity) gameView.getGameState().getMap().getAt(destination);
		createUnit(gameView, enemyCity, Templates.instance().TRANSPORTER, enemyCity.getPosition());
		createUnit(gameView, enemyCity, Templates.instance().ARMY, enemyCity.getPosition());
		createUnit(gameView, enemyCity, Templates.instance().FIGHTER, enemyCity.getPosition());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().CARRIER, new PlayfieldPosition(12, 8));
		assertEquals(2, unit1.getStepsLeftThisTurn());
		{ // Just moving is not enough
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertFalse(order.makesSense());
		} { // Now try an attack, should also fail
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemyCity.getKey());
			assertFalse(order.makesSense());
		}
	}

	private int clearAnimation() {
		return clearAnimation(gameView, observer);
	}

	private int clearAnimation(int expectedNumberOfAnimations) {
		return clearAnimation(gameView, observer, expectedNumberOfAnimations);
	}

	@Test
	public void testArmyAttacksTransporter_Works() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(1, 2);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit enemy = createUnit(gameView, enemyCity, Templates.instance().TRANSPORTER, destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, null);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		{ // Step 1 as GoTo
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_ALERT_IMPEDIMENT, clearAnimation(1));
		} { // Step 2 as Attack
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			final int mask = clearAnimation(1);
			assertTrue(0 != (MoveResult.MOVE_MASK_FIGHTING & mask));
			assertTrue(0 == (MoveResult.MOVE_MASK_MOVING & mask));
		}
	}

	@Test
	public void testTankMovesIntoTransporter_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(1, 2);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final ArrayList<Unit> cargo = new ArrayList<Unit>();
		for (int i = 0; i < 6; i++) {
			cargo.add(createUnit(gameView, ownCity, Templates.instance().ARMY, null));
		}
		final Unit transporter = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, destination);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		for (Unit unit1 : cargo) {
			nextCommandableTo(unit1);
			{ // Load as GoTo
				OrderGoTo order = null;
				try {
					order = new OrderGoTo(gameView.getGameState(), unit1, transporter.getKey());
				} catch (final CircularReferenceException e) {
					fail("Circular reference");
				}
				assertTrue(order.makesSense());
				gameView.setOrder(order);
				assertEquals(MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_LOADING, clearAnimation(1));
				assertEquals(transporter.getPosition(), unit1.getPosition());
				assertEquals(transporter.getKey(), unit1.getParentKey());
			}
		} 
		// Move transporter one step. Army changes position accordingly.
		gameView.nextCommandable(transporter);
		final PlayfieldPosition transporterDestination = new PlayfieldPosition(1, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(transporterDestination).getTemplateId());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), transporter, transporterDestination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(1));
		for (Unit unit1 : cargo) {
			assertEquals(transporterDestination, unit1.getPosition());
		}
	}

	private void nextCommandableTo(final Commandable cmdl) {
		int cnt = 0;
		while ((gameView.getVisibleCommandableInFocus() != cmdl) && (cnt < 10)) {
			gameView.nextCommandable();
			cnt++;
		}
		assertTrue(cnt < 10);
	}

	@Test
	public void testTransporterLoad_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(1, 2);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final ArrayList<Unit> cargo = new ArrayList<Unit>();
		for (int i = 0; i < 6; i++) {
			cargo.add(createUnit(gameView, ownCity, Templates.instance().ARMY, null));
		}
		final Unit transporter = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, destination);
		// Set load order to transporter
		gameView.setMapViewMode(MapViewMode.NORMAL);
		nextCommandableTo(transporter);
		transporter.setOrder(new OrderLoad(gameView.getGameState(), transporter));
		assertNotNull(transporter.getOrder());
		// give the surrounding units a chance to load
		for (int cnt = 0; cnt < 6; cnt++) {
			gameView.nextCommandable();
			assertEquals(MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_LOADING, clearAnimation(1));
		}
		// Transporter should wake up again when fully loaded
		gameView.nextCommandable();
		clearAnimation(0);
		assertNull(transporter.getOrder());
		assertEquals(transporter, gameView.getVisibleCommandableInFocus());
		for (Unit unit1 : cargo) {
			assertEquals(transporter.getPosition(), unit1.getPosition());
			assertEquals(transporter.getKey(), unit1.getParentKey());
		} 
		// Move transporter one step. Army changes position accordingly.
		final PlayfieldPosition transporterDestination = new PlayfieldPosition(1, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(transporterDestination).getTemplateId());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), transporter, transporterDestination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(1));
		for (Unit unit1 : cargo) {
			assertEquals(transporterDestination, unit1.getPosition());
		}
	}

	@Test
	public void testArmyWithinTransporterAttacksTransporter_Forbidden() throws CircularReferenceException {
		final PlayfieldPosition startpos = new PlayfieldPosition(1, 2);
		final PlayfieldPosition destination = new PlayfieldPosition(2, 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(startpos).getTemplateId());
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, startpos);
		final Unit transporter = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, startpos);
		unit1.setParent(transporter.getKey());
		final Unit enemy = createUnit(gameView, enemyCity, Templates.instance().TRANSPORTER, destination);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		{ // Step 1 as GoTo
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertFalse(order.makesSense());
		} { // Step 2 as Attack
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertFalse(order.makesSense());
		}
	}

	@Test
	public void testTransporterMovesOntoSeaDistance1OverBorder_Works() {
		final PlayfieldDimension size = gameView.getMap().getSize();
		final PlayfieldPosition destination = new PlayfieldPosition( size.getWidth() - 1, size.getHeight() - 1);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, new PlayfieldPosition(0, 0));
		{
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
			assertTrue(order.makesSense());
			gameView.nextCommandable(unit1);
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(1));
		}
	}

	@Test
	public void testDestroyerEscortsTransporter_Works() {
		final PlayfieldPosition start       = new PlayfieldPosition(13, 1);
		final PlayfieldPosition destination = new PlayfieldPosition(13, 4);
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(start).getTemplateId());
		assertEquals(Templates.instance().SEA.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit destUnit = createUnit(gameView, ownCity, Templates.instance().TRANSPORTER, destination);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().DESTROYER, start);
		{
			gameView.nextCommandable(unit1);
			OrderEscort order = null;
			try {
				order = new OrderEscort(gameView.getGameState(), unit1, destUnit.getKey());
			} catch (final CircularReferenceException ex) {
				fail("Circular reference");
			}
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_MOVING, clearAnimation(2));
//			assertEquals(1, result.direction.getY());
		}
	}

	@Test
	public void testTankMovesIntoMountain_NotOffered() {
		final PlayfieldPosition destination = new PlayfieldPosition(6, 4);
		assertEquals(Templates.instance().MOUNTAIN.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TANK, new PlayfieldPosition(6, 5));
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertFalse(order.makesSense());
	}

	@Test
	public void testTankAttacksIntoMountain_NotMoved() throws CircularReferenceException {
		final PlayfieldPosition destination = new PlayfieldPosition(6, 4);
		assertEquals(Templates.instance().MOUNTAIN.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());
		final Unit enemy = createUnit(gameView, enemyCity, Templates.instance().ARMY, destination);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		boolean hasWon = false;
		int mask = 0x00;
		do {
			final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TANK, new PlayfieldPosition(6, 5));
			gameView.nextCommandable();
			assertEquals(unit1, gameView.getVisibleCommandableInFocus());
			final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, enemy.getKey());
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			mask = clearAnimation(); // Could be 1 or 2, depending on length of fight
			assertTrue(0 != (MoveResult.MOVE_MASK_FIGHTING & mask));
			if (0 != (MoveResult.MOVE_MASK_WINNING_FIGHT & mask)) {
				hasWon = true;
			}
		} while (!hasWon);
		assertTrue(0 == (MoveResult.MOVE_MASK_MOVING & mask));
	}

	@Test
	public void testTankMovesIntoRough_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(7, 4);
		// Check for precondition: Rough is at (7, 4).
		assertEquals(Templates.instance().ROUGH.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());

		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TANK, new PlayfieldPosition(7, 5));
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		final int mask = clearAnimation(1);
		assertTrue(0 != (MoveResult.MOVE_MASK_MOVING & mask));
		assertEquals(0, unit1.getStepsLeftThisTurn());
	}

	/**
	 * Tank has no choice than to move through a series of roughs.
	 */
	@Test
	public void testTankMovesThroughRough_Works() {
		final PlayfieldPosition destination = new PlayfieldPosition(6, 6);
		final PlayfieldPosition p44 = new PlayfieldPosition(4, 4);
		final PlayfieldPosition p55 = new PlayfieldPosition(5, 5);
		assertEquals(Templates.instance().ROUGH.id, gameView.getGameState().getMap().getAt(p44).getTemplateId());
		assertEquals(Templates.instance().ROUGH.id, gameView.getGameState().getMap().getAt(p55).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TANK, new PlayfieldPosition(3, 3));
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		assertEquals(2, unit1.getStepRatings().get(p44).distanceFromStart);
		assertEquals(3, unit1.getStepRatings().get(p55).distanceFromStart); // It's unknown so far, so we don't know it's rough.
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		final int mask = clearAnimation(1);
		assertTrue(0 != (MoveResult.MOVE_MASK_MOVING & mask));
		assertEquals(0, unit1.getStepsLeftThisTurn());
		assertEquals(new PlayfieldPosition(4, 4), unit1.getPosition());
	}

	@Test
	public void testTankMovesIntoRough_NotMoved() {
		final PlayfieldPosition destination = new PlayfieldPosition(7, 4);
		// Check for precondition: Rough is at (7, 4).
		assertEquals(Templates.instance().ROUGH.id, gameView.getGameState().getMap().getAt(destination).getTemplateId());

		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TANK, new PlayfieldPosition(7, 6));
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		final int mask = clearAnimation(2);
		assertTrue(0 != (MoveResult.MOVE_MASK_MOVING & mask));

		assertEquals(new PlayfieldPosition(7, 5), unit1.getPosition());
		assertEquals(1, unit1.getStepsLeftThisTurn());
	}

	@Test
	public void testTankMovesAroundRough_Moved() {
		// Rough is at (7, 4) and (8, 4)
		final PlayfieldPosition destination = new PlayfieldPosition(8, 3);
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().TANK, new PlayfieldPosition(8, 5));
		assertNull(gameView.getCurrentAnimation());
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		{
//			final int mask = clearAnimation(1); // Start location is directly beneath neutral city
//			assertTrue(0 != (MoveResult.MOVE_MASK_ALERT_SIGHTING & mask));
		}
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		final OrderGoTo order = new OrderGoTo(gameView.getGameState(), unit1, destination);
		assertTrue(order.makesSense());
		gameView.setOrder(order);
		final int mask = clearAnimation(2);
		assertTrue(0 != (MoveResult.MOVE_MASK_MOVING & mask));
		assertEquals(0, unit1.getStepsLeftThisTurn());
		assertEquals(destination, unit1.getPosition());
	}

	@Test
	public void testArmyBecomesAirbase_Works() {
		final PlayfieldPosition start = new PlayfieldPosition(3, 2);
		assertEquals(Templates.instance().LAND.id, gameView.getGameState().getMap().getAt(start).getTemplateId());
		final Unit unit1 = createUnit(gameView, ownCity, Templates.instance().ARMY, start);
		gameView.setMapViewMode(MapViewMode.NORMAL);
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		{ // Move directly onto the unit, should fail.
			final OrderBuildAirbase order = new OrderBuildAirbase(gameView.getGameState(), unit1);
			assertTrue(order.makesSense());
			gameView.setOrder(order);
			assertEquals(MoveResult.MOVE_MASK_GARRISON, clearAnimation(1));
			final Collection<UnitAlike> units = gameView.findUnitsAtPosition(start);
			assertEquals(1, units.size());
			final Iterator<UnitAlike> it = units.iterator();
			final UnitAlike airbase = it.next();
			assertEquals(Templates.instance().AIRBASE, airbase.getUnitTemplate());
		}
	}

	static Unit createUnit(final GameView gameView, final FieldCity city, final CommandableTemplate template, final PlayfieldPosition position) {
		return Tutorial.createUnit(gameView, city, template, position);
	}

	private GameObserverStub observer;
	private GameView gameView;
	private FieldCity ownCity;
	private FieldCity enemyCity;
}
