package de.ped.empire.test;

import static org.junit.Assert.*;

import java.awt.Point;
import java.util.List;

import org.junit.Test;

import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.ZoomLevel;
import de.ped.empire.logic.OrderGoTo;
import de.ped.tools.BorderMode;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Logger;
import de.ped.tools.test.GenericTest;

/**
 * Unit tests for the {@link GameProperties} class.
 * 
 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GamePropertiesTest {

	/**
	 * @see GameProperties#equals(Object)
	 * @see GameProperties#clone()
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(false, new GameProperties(), new GameProperties());
	}

	@Test
	public void testMap2Canvas() {
		final GameProperties props = new GameProperties();
		props.setTopologyMode(TopologyMode.HEXAGON);
		final PlayfieldPosition pos = new PlayfieldPosition(5, 8);
		final Point point = new Point();
		final Point offset = new Point(100, 75);
		final ZoomLevel zoomLevel = ZoomLevel.DIALOG_ZOOM_LEVEL;
		props.map2canvas(point, pos, offset, zoomLevel, PlayfieldTiling.CENTER);
		final PlayfieldPosition res = props.canvas2map(point.x, point.y, offset, zoomLevel);
		assertEquals(pos, res);
	}

	@Test
	public void testWrapInside() {
		final GameProperties props = new GameProperties();
		props.setTopologyMode(TopologyMode.HEXAGON);
		props.setSize(PlayfieldPosition.X, 10);
		props.setSize(PlayfieldPosition.Y, 10);

//		2020-01-07 12:42:03,243 [INFO ] {GameMap} ##        
//		2020-01-07 12:42:03,243 [INFO ] {GameMap} #~##      
//		2020-01-07 12:42:03,243 [INFO ] {GameMap} #.~~##    
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} #~$~~~##  
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} #~o~~~~~##
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} #~~~$~~$~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} #~~~~~~~~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} #~~~.~$~~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} #~~$~~~~~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap} ##~~.~*~~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap}   ##~~$~~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap}     ##.~~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap}       ##~#
//		2020-01-07 12:42:03,244 [INFO ] {GameMap}         ##

		final int[][][] positions;
		if (PlayfieldTiling.IS_HEX_COMPACT) {
			positions = new int[][][] {
				{ {  0, 0 }, { 0, 0 } }, // corner
				{ {  0, 9 }, { 0, 9 } }, // corner
				{ {  9, 0 }, { 9, 0 } }, // corner
				{ {  9, 9 }, { 9, 9 } }, // corner
				{ {  9, 4 }, { 9, 4 } }, // somewhere in the middle
				{ {  9,13 }, { 9, 3 } }, // wrap top to down
				{ {  5, 8 }, { 5, 8 } }, // somewhere in the middle
				{ {  0,-1 }, { 0, 9 } }, // wrap top to down
				{ {  0,-2 }, { 0, 8 } }, // wrap top to down
				{ {  9, 3 }, { 9, 3 } }, // somewhere in the middle
				{ {  9,-1 }, { 9, 9 } }, // wrap top to down
				{ {  9,-2 }, { 9, 8 } }, // wrap top to down
				{ {  0,10 }, { 0, 0 } }, // wrap top to down
				{ {  0,11 }, { 0, 1 } }, // wrap top to down
				{ {  9,10 }, { 9, 0 } }, // wrap top to down
				{ {  9,15 }, { 9, 5 } }, // wrap top to down
				{ { -1, 0 }, { 9, 0 } }, // wrap left to right
				{ { -2, 0 }, { 8, 0 } }, // wrap left to right
				{ { -1, 8 }, { 9, 8 } }, // wrap left to right
				{ { -2, 8 }, { 8, 8 } }, // wrap left to right
				{ { 10, 0 }, { 0, 0 } }, // wrap left to right
				{ { 11, 0 }, { 1, 0 } }, // wrap left to right
				{ { 10, 9 }, { 0, 9 } }, // wrap left to right
				{ { 11, 9 }, { 1, 9 } }, // wrap left to right
				{ { -1, 0 }, { 9, 0 } }, // wrap left to right
				{ { -2, 0 }, { 8, 0 } }, // wrap left to right
				{ { -1,-1 }, { 9, 9 } }, // wrap diagonally over lower left
				{ { -2,-2 }, { 8, 8 } }, // wrap diagonally over lower left
				{ { -1,10 }, { 9, 0 } }, // wrap diagonally over lower left
				{ { -2,11 }, { 8, 1 } }, // wrap diagonally over lower left
				{ { 10,10 }, { 0, 0 } }, // wrap diagonally over lower right
				{ { 11,11 }, { 1, 1 } }, // wrap diagonally over lower right
				{ { 10,-1 }, { 0, 9 } }, // wrap diagonally over upper right
				{ { 11,-2 }, { 1, 8 } }, // wrap diagonally over upper right
			};
		} else {
			positions = new int[][][] {
				{ {  0, 0 }, { 0, 0 } }, // corner
				{ {  0, 9 }, { 0, 9 } }, // corner
				{ {  9, 4 }, { 9, 4 } }, // corner
				{ {  9,13 }, { 9,13 } }, // corner
				{ {  5, 8 }, { 5, 8 } }, // somewhere in the middle
				{ {  0,-1 }, { 0, 9 } }, // wrap top to down
				{ {  0,-2 }, { 0, 8 } }, // wrap top to down
				{ {  9, 3 }, { 9,13 } }, // wrap top to down
				{ {  9, 2 }, { 9,12 } }, // wrap top to down
				{ {  0,10 }, { 0, 0 } }, // wrap top to down
				{ {  0,11 }, { 0, 1 } }, // wrap top to down
				{ {  9,14 }, { 9, 4 } }, // wrap top to down
				{ {  9,15 }, { 9, 5 } }, // wrap top to down
				{ { -1, 0 }, { 9, 5 } }, // wrap left to right
				{ { -2, 0 }, { 8, 5 } }, // wrap left to right
				{ { -1, 8 }, { 9,13 } }, // wrap left to right
				{ { -2, 8 }, { 8,13 } }, // wrap left to right
				{ { 10, 5 }, { 0, 0 } }, // wrap left to right
				{ { 11, 5 }, { 1, 0 } }, // wrap left to right
				{ { 10,14 }, { 0, 9 } }, // wrap left to right
				{ { 11,14 }, { 1, 9 } }, // wrap left to right
				{ { -1,-1 }, { 9, 4 } }, // wrap left to right
				{ { -2,-1 }, { 8, 4 } }, // wrap left to right
				{ { -1,-2 }, { 9,13 } }, // wrap diagonally over lower left
				{ { -2,-3 }, { 8,12 } }, // wrap diagonally over lower left
				{ { -1, 9 }, { 9, 4 } }, // wrap diagonally over lower left
				{ { -2,10 }, { 8, 5 } }, // wrap diagonally over lower left
				{ { 10,15 }, { 0, 0 } }, // wrap diagonally over lower right
				{ { 11,16 }, { 1, 1 } }, // wrap diagonally over lower right
				{ { 10, 4 }, { 0, 9 } }, // wrap diagonally over upper right
				{ { 11, 3 }, { 1, 8 } }, // wrap diagonally over upper right
			};
		}
		for (final int[][] testdata : positions) {
			final PlayfieldPosition pos = new PlayfieldPosition(testdata[0]);
			final PlayfieldPosition exp = new PlayfieldPosition(testdata[1]);
			final PlayfieldPosition res = new PlayfieldPosition(pos);
			props.wrapInside(res);
			assertEquals("Position to wrap was " + pos.toString() + ", result was " + res.toString(), exp, res);
		}
	}

	@Test
	public void testGetAllVectorsToAirDistance() {
		final GameProperties props = new GameProperties();
		props.setTopologyMode(TopologyMode.HEXAGON);
		props.setSize(PlayfieldPosition.X, 10);
		props.setSize(PlayfieldPosition.Y, 10);
		props.setBorderMode(BorderMode.NONE);
		final PlayfieldPosition middle = new PlayfieldPosition(3, 3);
		List<PlayfieldDistance> ret = props.getAllVectorsToAirDistance(middle, 0, false);
		assertEquals(1, ret.size());
		assertEquals(new PlayfieldDistance(0, 0), ret.get(0));
		if (PlayfieldTiling.IS_HEX_COMPACT) {
			ret = props.getAllVectorsToAirDistance(middle, 1, false);
			assertEquals(6, ret.size());
			Logger.getRootLogger().debug("Distance 1 on middle " + middle.toString() + ":");
			for (PlayfieldDistance pos : ret) {
				final PlayfieldDistance that = pos;
				Logger.getRootLogger().debug(that.toString());
			}
			assertEquals(new PlayfieldDistance( 1, -1), ret.get(0));
			assertEquals(new PlayfieldDistance(-0, -1), ret.get(1));
			assertEquals(new PlayfieldDistance(-1, -1), ret.get(2));
			assertEquals(new PlayfieldDistance( 1,  0), ret.get(3));
			assertEquals(new PlayfieldDistance(-1,  0), ret.get(4));
			assertEquals(new PlayfieldDistance( 0,  1), ret.get(5));

			ret = props.getAllVectorsToAirDistance(middle, 2, false);
			assertEquals(12, ret.size());
			Logger.getRootLogger().debug("Distance 2 on middle " + middle.toString() + ":");
			for (PlayfieldDistance pos : ret) {
				final PlayfieldDistance that = pos;
				Logger.getRootLogger().debug(that.toString());
			}
			assertEquals(new PlayfieldDistance( 1, -2), ret.get( 0));
			assertEquals(new PlayfieldDistance( 0, -2), ret.get( 1));
			assertEquals(new PlayfieldDistance(-1, -2), ret.get( 2));
			assertEquals(new PlayfieldDistance( 2, -1), ret.get( 3));
			assertEquals(new PlayfieldDistance(-2, -1), ret.get( 4));
			assertEquals(new PlayfieldDistance( 2,  0), ret.get( 5));
			assertEquals(new PlayfieldDistance(-2,  0), ret.get( 6));
			assertEquals(new PlayfieldDistance( 2,  1), ret.get( 7));
			assertEquals(new PlayfieldDistance( 1,  1), ret.get( 8));
			assertEquals(new PlayfieldDistance(-1,  1), ret.get( 9));
			assertEquals(new PlayfieldDistance(-2,  1), ret.get(10));
			assertEquals(new PlayfieldDistance( 0,  2), ret.get(11));
		}
	}	

	@Test
	public void testCalcAirDistanceVector() {
		final GameProperties props = new GameProperties();
		props.setTopologyMode(TopologyMode.HEXAGON);
		props.setSize(PlayfieldPosition.X, 10);
		props.setSize(PlayfieldPosition.Y, 10);
		props.setBorderMode(BorderMode.NONE);
		final int[][] testsets;
		if (PlayfieldTiling.IS_HEX_COMPACT) {
			testsets = new int[][] {
				{ 1, 2, 1, 2, 0, 0 }, // null vector
				{ 0, 0, 0, 1, 0, 1 },
				{ 0, 0, 1, 1, 1, 1 },
				{ 0, 0, 1, 0, 1, 0 },
				{ 0, 0, 9, 0,-1, 0 }, // with wrap 
			};
		} else {
			testsets = new int[][] {
//				{ 0, 0, 0, 1, 0, 1 },
//				{ 0, 0, 1, 1, 1, 1 },
//				{ 0, 0, 1, 0, 1, 0 },
//				{ 1, 2, 1, 2, 0, 0 }, // null vector
				{ 0, 0, 9, 5,-1, 0 }, // with wrap 
			};
		}
		for (final int[] testset : testsets) {
			final PlayfieldPosition start = new PlayfieldPosition(testset[0], testset[1]);
			final PlayfieldPosition target = new PlayfieldPosition(testset[2], testset[3]);
			final PlayfieldPosition exp = new PlayfieldPosition(testset[4], testset[5]);
			final PlayfieldDistance result = props.calcAirDistanceVector(start, target);
			assertEquals(
					"Air distance from " + start.toString() + " to " + target.toString()
							+ " was expected " + exp.toString() + " but was " + result.toString(),
					exp, result);
		}
	}

	@Test
	public void testCalcAirDistanceScalar() {
		final GameProperties props = new GameProperties();
		props.setTopologyMode(TopologyMode.HEXAGON);
		props.setSize(PlayfieldPosition.X, 10);
		props.setSize(PlayfieldPosition.Y, 10);
		props.setBorderMode(BorderMode.NONE);
		final int[][] testsets;
		if (PlayfieldTiling.IS_HEX_COMPACT) {
			testsets = new int[][] {
					{ 3, 3,  0,  0, 0 },
					{ 3, 3,  0, -1, 1 },
					{ 3, 3,  0,  1, 1 },
					{ 3, 3, -1, -1, 1 },
					{ 3, 3, -1,  0, 1 },
					{ 3, 3, -1,  1, 2 },
					{ 3, 3,  1, -1, 1 },
					{ 3, 3,  1,  0, 1 },
					{ 3, 3,  1,  1, 2 },
					{ 2, 2,  0,  0, 0 },
					{ 2, 2,  0, -1, 1 },
					{ 2, 2,  0,  1, 1 },
					{ 2, 2, -1, -1, 2 },
					{ 2, 2, -1,  0, 1 },
					{ 2, 2, -1,  1, 1 },
					{ 2, 2,  1, -1, 2 },
					{ 2, 2,  1,  0, 1 },
					{ 2, 2,  1,  1, 1 },
					//        (1, 0)        (3, 0)        (5, 0)        (7, 0)
					// (0, 0)        (2, 0)        (4, 0)        (6, 0)
					//        (1, 1)        (3, 1)        (5, 1)        (7, 1)
					// (0, 1)        (2, 1)        (4, 1)        (6, 1)
					//        (1, 2)        (3, 2)        (5, 2)        (7, 2)
					// (0, 2)        (2, 2)        (4, 2)        (6, 2)
					//        (1, 3)        (3, 3)        (5, 3)        (7, 3)
					// (0, 3)        (2, 3)        (4, 3)        (6, 3)
					//        (1, 4)        (3, 4)        (5, 4)        (7, 4)
					// (0, 4)        (2, 4)        (4, 4)        (6, 4)
					//        (1, 5)        (3, 5)        (5, 5)        (7, 5)
					// (0, 5)        (2, 5)        (4, 5)        (6, 5)
			};
		} else {
			testsets = new int[][] {
				{ 3, 3,  0,  0, 0 },
				{ 3, 3,  0, -1, 1 },
				{ 3, 3,  0,  1, 1 },
				{ 2, 2,  0,  0, 0 },
				{ 2, 2,  0, -1, 1 },
				{ 2, 2,  0,  1, 1 },
			};
		}
		for (int[] testset : testsets) {
			final PlayfieldPosition startPos = new PlayfieldPosition(testset[0], testset[1]);
			final PlayfieldDistance distanceVector = new PlayfieldDistance(testset[2], testset[3]);
			final int exp = testset[4];
			final int res = props.calcAirDistanceScalar(startPos, distanceVector);
			assertEquals("startPos=" + startPos.toString() + ", distanceVector=" + distanceVector.toString(), exp, res);
		}
	}

	@Test
	public void testCalcDistanceVectorStringParts() {
		checkDistanceVectorStringPart(new int[] {  0,  0 }, 0, null, 0, null);
		checkDistanceVectorStringPart(new int[] {  0,  1 }, 1, "D", 0, null);
		checkDistanceVectorStringPart(new int[] {  0, -1 }, 1, "U", 0, null);
		checkDistanceVectorStringPart(new int[] {  1,  0 }, 1, "RUR", 0, null);
		checkDistanceVectorStringPart(new int[] { -1,  0 }, 1, "LDL", 0, null);
		checkDistanceVectorStringPart(new int[] {  1,  1 }, 1, "RDR", 0, null);
		checkDistanceVectorStringPart(new int[] { -1, -1 }, 1, "LUL", 0, null);
		checkDistanceVectorStringPart(new int[] { -1,  1 }, 1, "LDL", 1, "D");
	}

	private void checkDistanceVectorStringPart(final int[] distanceVector, int diff0, String dir0, int diff1, String dir1) {
		OrderGoTo.DistanceVectorStringPart[] result;
		result = OrderGoTo.calcDistanceVectorStringParts(new PlayfieldDistance(distanceVector), TopologyMode.HEXAGON);
		assertEquals(diff0, result[0].getDiff());
		assertEquals(dir0, result[0].getDir());
		if (null == dir1) {
			assertNull(result[1]);
		} else {
			assertEquals(diff1, result[1].getDiff());
			assertEquals(dir1, result[1].getDir());
		}
	}
}
