package de.ped.empire.test;

import static org.junit.Assert.*;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.junit.Ignore;
import org.junit.Test;

import de.ped.empire.logic.Commandable;
import de.ped.empire.logic.CommandableKey;
import de.ped.empire.logic.GameInitializationException;
import de.ped.empire.logic.GameState;
import de.ped.empire.logic.GameView;
import de.ped.empire.logic.GameMap;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.GameFacade;
import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.MoveResult;
import de.ped.empire.logic.OrderDisband;
import de.ped.empire.logic.OrderLoad;
import de.ped.empire.logic.OrderSentry;
import de.ped.empire.logic.OrderSkipMove;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.ProcessorSpeedType;
import de.ped.empire.logic.Templates;
import de.ped.empire.logic.Unit;
import de.ped.empire.logic.UnitAlike;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Tests for the {@link GameView} class.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameTest {

	protected static FieldCity findCity(final GameView gameView) {
		return gameView.getMap().findCities(0).get(0);
	}

	/** 
	 * Tests creation of a game, creation of a unit, and marshalling/demarshalling of that game. 
	 * @throws GameInitializationException if game initialization failed
	 */
	@Test
	public void testCreationAndMarshalling() throws GameInitializationException {
		final GameObserverStub observer = new GameObserverStub();
		final GameProperties props = new GameProperties();
		final GameMap map = new GameMap(props);
		map.addAtRandom(props);
		map.finish();
		final GameState gameState = new GameState(observer, map);
		gameState.initForQuickStart();
		gameState.startGame();
		final GameView gameView = gameState.getCurrentGameView();
		final FieldCity city = findCity(gameView);
		gameView.setCityProduction(city, Templates.instance().ARMY);
		final Unit unit = Unit.create(gameState, city);
		gameState.addUnit(city, unit);
		for (int playerId = 0; playerId < gameView.getPlayers().length; playerId++) {
			final Player player = gameState.getGameView(playerId).getPlayer(playerId);
			if (player.isPlayer()) {
				final int id = player.getCntUnitsProduced(Templates.instance().ARMY);
				assertNotEquals("playerID=" + playerId, 0, id);
			}
		}
		// marshal/unmarshal
		final GameState received = checkMarshallingGame(observer, gameState, gameView);
		final GameView receivedView = received.getCurrentGameView();
		for (int playerId = 0; playerId < receivedView.getPlayers().length; playerId++) {
			final Player player = received.getGameView(playerId).getPlayer(playerId);
			if (player.isPlayer()) {
				final int cntArmiesProduced = player.getCntUnitsProduced(Templates.instance().ARMY);
				assertNotEquals("playerID=" + playerId, 0, cntArmiesProduced);
			}
		}
	}

	public static GameState checkMarshallingGame(final GameObserverStub observer, final GameState gameState,
			final GameView gameView) {
		final Marshaller m = new Marshaller();
		m.writeMarshallable(gameState);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		final GameState received = new GameState(observer, d);
		final GameView receivedView = received.getCurrentGameView();
		assertEquals(gameView.getPlayers().length, receivedView.getPlayers().length);
		assertEquals(gameView.getMap().getSize(), receivedView.getMap().getSize());
		return received;
	}

	@Test
	public void testFindCommandablesAtPosition() throws GameInitializationException {
		final ProcessorSpeedType pst = ProcessorSpeedType.SLOW;
		final GameObserverStub observer = new GameObserverStub();
		final GameProperties props = new GameProperties();
		final GameMap map = new GameMap(props);
		map.addAtRandom(props);
		map.finish();
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		final GameView gameView = gameState.getCurrentGameView();
		gameView.clearGetReady();
		gameView.nextCommandable();
		while (null != gameView.getCurrentAnimation()) {
			gameView.step(pst);
		}
		final Commandable current = gameView.getVisibleCommandableInFocus();
		final PlayfieldPosition currpos = current.getPosition();
		final List<Commandable> cmdls = gameView.findCommandablesAtPosition(currpos, true);
		assertEquals(pst.name(), 1, cmdls.size());
		assertEquals(pst.name(), current, cmdls.get(0));
	}

	/** 
	 * Tests switching players until the first unit was produced. 
	 * @throws GameInitializationException if game initialization failed
	 */
	@Test
	public void testProduction() throws GameInitializationException {
		final GameObserverStub observer = new GameObserverStub();
		final GameProperties props = new GameProperties();
		final GameMap map = new GameMap(props);
		map.addAtRandom(props);
		map.finish();
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		final GameView gameView0 = gameState.getCurrentGameView();
		assertEquals(0, gameState.getCurrentPlayerId());
		assertEquals(1, gameState.getTurn());
		// First there is a city
		gameView0.clearGetReady();
		gameView0.nextCommandable();
		Commandable cmdle = gameView0.getVisibleCommandableInFocus();
		assertNotNull(cmdle);
		assertEquals(FieldCity.class, cmdle.getClass());
		final FieldCity city = (FieldCity) cmdle;
		// That city produces an army
		gameView0.setCityProduction(city, Templates.instance().ARMY);
		final int turns = city.getTurnsToProduce(Templates.instance().ARMY);
		for (int i = 0; i < turns - 1; i++) {
			final String prefix = "turn=" + (2 + i);
			logger.debug("testProduction: " + prefix);
			assertEquals(gameState.getGameView(0), gameState.getCurrentGameView());
			assertEquals(gameView0, gameState.getCurrentGameView());
			gameView0.finishTurn();
			assertEquals(gameState.getGameView(1), gameState.getCurrentGameView());
			gameState.getGameView(1).finishTurn();
			assertEquals(prefix, 0, gameState.getCurrentPlayerId());
			assertEquals(prefix, 2 + i, gameState.getTurn());
			assertEquals(prefix, Templates.instance().ARMY.turnsToProduce - 1 - i, city.getTurnsToProduce());
			// While in production, that city is not nextable.
			gameView0.nextCommandable();
			cmdle = gameView0.getVisibleCommandableInFocus();
			assertNull(prefix, cmdle);
		}
		gameView0.finishTurn();
		final GameView gameView1 = gameState.getCurrentGameView();
		assertFalse("gameViews should differ, but both are " + gameView1, gameView1.equals(gameView0));
		gameView1.clearGetReady();
		gameView1.finishTurn();
		assertEquals(0, gameState.getCurrentPlayerId());
		assertEquals(1 + turns, gameState.getTurn());
		assertEquals(Templates.instance().ARMY.turnsToProduce, city.getTurnsToProduce());
		// Then we have an army
		gameView0.clearGetReady();
		gameView0.nextCommandable();
		cmdle = gameView0.getVisibleCommandableInFocus();
		assertNotNull(cmdle);
		assertEquals(ImageKey.CMDL_ID_ARMY, cmdle.getKey().getTemplateId());
		Unit army = (Unit) cmdle;
		assertEquals(1, army.getStepsLeftThisTurn());
		final boolean ok = army.canMove();
		assertEquals(true, ok);
		army.consumeStep();
		assertEquals(0, army.getStepsLeftThisTurn());
	}

	/** 
	 * Tests switching units and selecting units. 
	 * @throws GameInitializationException if game initialization failed
	 */
	@Test
	public void testUnitsNextAndSelection() throws GameInitializationException {
		final GameObserverStub observer = new GameObserverStub();
		final GameProperties props = new GameProperties();
		final GameMap map = new GameMap(props);
		map.addAtRandom(props);
		map.finish();
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		final GameView gameView = gameState.getCurrentGameView();
		final FieldCity city = findCity(gameView);
		gameView.setCityProduction(city, Templates.instance().NOTHING);
		final Unit unit1 = Unit.create(gameState, Templates.instance().ARMY, city);
		gameState.addUnit(city, unit1);
		final Unit unit2 = Unit.create(gameState, Templates.instance().ARMY, city);
		gameState.addUnit(city, unit2);
		assertEquals(1, unit1.getStepsLeftThisTurn());
		gameView.clearGetReady();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		gameView.nextCommandable();
		assertEquals(unit2, gameView.getVisibleCommandableInFocus());
		gameView.nextCommandable();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		gameView.nextCommandable();
		assertEquals(unit2, gameView.getVisibleCommandableInFocus());
	}

	/** 
	 * Tests removing units. 
	 * @throws GameInitializationException if game initialization failed
	 */
	@Test
	public void testUnitsRemove() throws GameInitializationException {
		final GameObserverStub observer = new GameObserverStub();
		final GameProperties props = new GameProperties();
		final GameMap map = new GameMap(props);
		map.addAtRandom(props);
		map.finish();
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		final GameView gameView = gameState.getCurrentGameView();
		final FieldCity city = findCity(gameView);
		gameView.setCityProduction(city, Templates.instance().NOTHING);
		final Unit unit1 = Unit.create(gameState, Templates.instance().ARMY, city);
		gameState.addUnit(city, unit1);
		final Unit unit2 = Unit.create(gameState, Templates.instance().ARMY, city);
		gameState.addUnit(city, unit2);
		assertEquals(1, unit1.getStepsLeftThisTurn());
		gameView.clearGetReady();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		gameView.setOrder(new OrderDisband(gameState, unit1));
		assertEquals(MoveResult.MOVE_MASK_LOSING_FIGHT, GameMoveTest.clearAnimation(gameView, observer, 1));
		gameView.nextCommandable();
		assertEquals(unit2, gameView.getCommandableInFocus());
		gameView.nextCommandable();
		assertEquals(unit2, gameView.getCommandableInFocus());
	}

	@Ignore
	public void testLoadingStoredFileBugNotBlinking() throws IOException {
		final GameObserverStub observer = new GameObserverStub();
		final File file = new File(
				"C:/Users/peter.diefenbach/Documents/BugNotBlinking." + GameFacade.SAVED_GAME_FILE_EXTENSION);
		assertTrue(0 < file.length());
		final Marshaller marshaller = Marshaller.loadFromFile(file);
//		final Marshaller mcopy = (Marshaller) marshaller.clone();
//		final byte[] arr = mcopy.pack();
//		HexDumpEncoder encoder = new HexDumpEncoder();
//		String ret = encoder.encode(arr);
//		logger.debug(ret);
//		assertArrayEquals(new byte[]{0, 0, 0, 0}, Arrays.copyOfRange(arr, 0, 4));
		final GameState gameState = new GameState(observer, marshaller);
		assertEquals(6, gameState.getProperties().getPlayers().length);
		assertEquals(0, gameState.getCurrentPlayerId());
		assertNotNull(gameState.getGameView(0));
		final GameView gameView = gameState.getCurrentGameView();
//		final LinkedList<Unit> units = gameView.getUnits();
		gameView.clearGetReady();
		Commandable cmdl = gameView.getCommandableInFocus();
		Commandable vcmd = gameView.getVisibleCommandableInFocus();
		assertEquals(cmdl, vcmd);
		assertEquals(30, cmdl.getKey().getId());
		assertEquals(0, cmdl.getKey().getPlayerId());
		assertEquals(Templates.instance().ARMY.id, cmdl.getKey().getTemplateId());
		logger.debug(gameView.getCommandableInFocus().toString());
		Unit unit = (Unit) cmdl;
		final CommandableKey parentKey = unit.getParentKey();
//		final Commandable parent = gameView.findCommandable(parentKey);
		assertNotNull(parentKey);
		assertEquals(0, parentKey.getPlayerId());
		assertEquals(Templates.instance().TRANSPORTER.id, parentKey.getTemplateId());
		final Collection<UnitAlike> units = gameView.getUnits();
		assertTrue(units.contains(unit));
	}

	@Ignore
	public void testLoadingStoredFileBugNexting() throws IOException {
		testLoadingStoredFileBugNexting(false);
	}

	@Ignore
	public void testLoadingStoredFileBugSkipping() throws IOException {
		testLoadingStoredFileBugNexting(true);
	}

	public void testLoadingStoredFileBugNexting(final boolean isSkipping) throws IOException {
		final GameObserverStub observer = new GameObserverStub();
		final File file = new File(
				"C:/Users/peter.diefenbach/Documents/BugNextableOrder3." + GameFacade.SAVED_GAME_FILE_EXTENSION);
		assertTrue(0 < file.length());
		final Marshaller marshaller = Marshaller.loadFromFile(file);
		final GameState gameState = new GameState(observer, marshaller);
		assertEquals(6, gameState.getProperties().getPlayers().length);
		assertEquals(0, gameState.getCurrentPlayerId());
		assertNotNull(gameState.getGameView(0));
		final GameView gameView = gameState.getCurrentGameView();
		final OrderSkipMove generalOrderSkipMove = new OrderSkipMove(gameState, null);
		// Traverse through all the units by nexting
		final Collection<UnitAlike> allUnits = gameView.getUnits();
		final LinkedList<Unit> nextedUnits = new LinkedList<Unit>();
		Commandable prev = null;
		do {
			Commandable cmdl = gameView.getCommandableInFocus();
			if (null == cmdl) {
				assertTrue(gameView.isFinishTurnAllowed());
				break;
			}
			assertEquals(0, cmdl.getKey().getPlayerId());
			logger.debug(cmdl.toString());
			if (cmdl instanceof Unit) {
				Unit unit = (Unit) cmdl;
				if (nextedUnits.contains(unit)) { // skip 2nd round
					break;
				}
				nextedUnits.add(unit);
			}
			if (null != prev) {
				final int dist = gameView.getProperties().calcAirDistance(prev.getPosition(), cmdl.getPosition());
				for (final UnitAlike unit : allUnits) {
					if ((0 == unit.getKey().getPlayerId()) 
							&& !(unit.getOrder() instanceof OrderLoad)
							&& !(unit.getOrder() instanceof OrderSentry)
							&& unit.isNextable()
							&& !nextedUnits.contains(unit)) {
						final int udist = gameView.getProperties().calcAirDistance(prev.getPosition(), unit.getPosition());
						assertTrue("From unit " + prev + prev.getPosition().toString() +
								" was " + unit + unit.getPosition().toString() +
								" nearer than nexted " + cmdl + cmdl.getPosition().toString(), 
								dist <= udist);
					}
				}
			}
			if (isSkipping) {
				gameView.setOrder(generalOrderSkipMove);
			} else {
				gameView.nextCommandable();
			}
			prev = cmdl;
		} while(true);
		// Check that all units were nexted
		for (final UnitAlike unit : allUnits) {
			if ((0 == unit.getKey().getPlayerId()) 
					&& unit.isNextable()
					&& !(unit.getOrder() instanceof OrderLoad)
					&& !(unit.getOrder() instanceof OrderSentry)) {
				assertTrue("Not contained unit was " + unit + " with order " + unit.getOrder(), nextedUnits.contains(unit));
			}
		}
	}

	private Logger logger = Logger.getLogger(getClass());

	public GameTest() {
		logger.setLogLevel(Level.INFO);
	}
}
