package de.ped.empire.test;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.TreeMap;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import de.ped.empire.logic.CommandableKey;
import de.ped.empire.logic.CommandableTemplate;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.GameInitializationException;
import de.ped.empire.logic.GameMap;
import de.ped.empire.logic.GameState;
import de.ped.empire.logic.GameView;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.MapViewMode;
import de.ped.empire.logic.MilitaryBranch;
import de.ped.empire.logic.MoveResult;
import de.ped.empire.logic.Order;
import de.ped.empire.logic.OrderGoTo;
import de.ped.empire.logic.OrderSkipMove;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.Templates;
import de.ped.empire.logic.Tutorial;
import de.ped.empire.logic.TutorialLevel;
import de.ped.empire.logic.Unit;
import de.ped.empire.logic.UnitAlike;
import de.ped.empire.logic.ViewUnit;
import de.ped.tools.Marshaller;
import de.ped.tools.PlayfieldPosition;

/**
 * A test that checks if a move is visible.
 * 
 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
 */
@RunWith(value=Parameterized.class)
public class MoveVisibilityTest {

	/**
	 * A unit observed.
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ObservedUnit {
		public final CommandableTemplate unitType;
		public final int[] pos;
		public ObservedUnit(
				final CommandableTemplate unitType,
				final int[] pos) {
			this.unitType = unitType;
			this.pos = pos;
		}
	}

	/**
	 * A move from one field to an adjacent field.
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class MoveStep {
		public final PlayfieldPosition destPos;
		public final byte mapAt;
		public final int moveResult;
		public MoveStep(
				final int x,
				final int y,
				final byte mapAt,
				final int moveResult) {
			this.destPos = (x < 0 ? null : new PlayfieldPosition(x, y));
			this.mapAt = mapAt;
			this.moveResult = moveResult;
		}
	}

	/**
	 * A moving unit.
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class MovingUnit {
		public final CommandableTemplate unitType;
		public final MoveStep[] steps;
		public MovingUnit(
				final CommandableTemplate unitType,
				final MoveStep[] steps) {
			this.unitType = unitType;
			this.steps = steps;
		}
	}

	/**
	 * The parameters of a single test run.
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class TestParams {
		@SuppressWarnings("unused")
		public final String testDescription;
		public final MovingUnit enemyUnit;
		public final MovingUnit movingUnit;
		public final int enemyNumberOfAnimations;
		public final int expectedEnemyResultMask;
		public final ObservedUnit[] expectedObservedPlayer1Units;
		public final ObservedUnit[] expectedObservedPlayer2Units;
		public TestParams(
				final String testDescription,
				final MovingUnit enemyUnit,
				final MovingUnit movingUnit,
				final ObservedUnit[] expectedObservedPlayer2Units,
				final int enemyNumberOfAnimations, 
				final int expectedEnemyResultMask,
				final ObservedUnit[] expectedObservedPlayer1Units) {
			this.testDescription = testDescription;
			this.enemyUnit = enemyUnit;
			this.movingUnit = movingUnit;
			this.enemyNumberOfAnimations = enemyNumberOfAnimations;
			this.expectedEnemyResultMask = expectedEnemyResultMask;
			this.expectedObservedPlayer1Units = expectedObservedPlayer1Units;
			this.expectedObservedPlayer2Units = expectedObservedPlayer2Units;
		}
	}

	@Parameters
	public static Collection<TestParams> data() {
		return Arrays.asList( new TestParams[] {
				new TestParams("No enemy unit nearby to view anything",
						null, // No observer unit, as we use enemy city as observer
						new MovingUnit(Templates.instance().ARMY, new MoveStep[] {
								new MoveStep(-1, -1, ImageKey.FIELD_ID_CITY, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 3,  2, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						0, MoveResult.MOVE_MASK_NOTHING, 
						new ObservedUnit[] { } 
				),
				new TestParams("Both start and end positon of move are in observable range",
						null, // No observer unit, as we use enemy city as observer
						new MovingUnit(Templates.instance().ARMY, new MoveStep[] {
								new MoveStep( 9,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep(10,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						1, MoveResult.MOVE_MASK_MOVING, 
						new ObservedUnit[] { new ObservedUnit(Templates.instance().ARMY, new int[] { 10, 7 } ) } 
				),
				new TestParams("Start positon of move is in observable range",
						null, // No observer unit, as we use enemy city as observer
						new MovingUnit(Templates.instance().ARMY, new MoveStep[] {
								new MoveStep( 9,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 8,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						1, MoveResult.MOVE_MASK_MOVING, 
						new ObservedUnit[] { } 
				),
				new TestParams("End positon of move is in observable range",
						null, // No observer unit, as we use enemy city as observer
						new MovingUnit(Templates.instance().ARMY, new MoveStep[] {
								new MoveStep( 8,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 9,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						1, MoveResult.MOVE_MASK_MOVING, 
						new ObservedUnit[] { 
								new ObservedUnit(Templates.instance().ARMY, new int[] {  9, 7 } ),
						} 
				),
				new TestParams("A move just out of observable range",
						null, // No observer unit, as we use enemy city as observer
						new MovingUnit(Templates.instance().ARMY, new MoveStep[] {
								new MoveStep( 8,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 8,  6, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						0, MoveResult.MOVE_MASK_NOTHING, 
						new ObservedUnit[] { } 
				),
				new TestParams("A submarine passing by the enemy city",
						null, // No observer unit, as we use enemy city as observer
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] {
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 9,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						2, MoveResult.MOVE_MASK_MOVING, 
						new ObservedUnit[] { } 
				),
				new TestParams("A submarine passing by an enemy sub is fully visible",
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] {
								new MoveStep( 7,  9, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
						} ),
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] {
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 7,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { 
								new ObservedUnit(Templates.instance().SUBMARINE, new int[] { 7, 9 } ),
						},
						2, MoveResult.MOVE_MASK_MOVING, 
						new ObservedUnit[] { 
								new ObservedUnit(Templates.instance().SUBMARINE, new int[] {  8, 8 } ), 
						} 
				),
				new TestParams("A submarine passing by an enemy army is invisible",
						new MovingUnit(Templates.instance().ARMY, new MoveStep[] {
								new MoveStep( 7,  7, ImageKey.FIELD_ID_LAND, MoveResult.MOVE_MASK_NOTHING),
						} ),
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] {
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 7,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						0, MoveResult.MOVE_MASK_NOTHING, 
						new ObservedUnit[] { } 
				),
				new TestParams("A submarine passing by an enemy fighter is invisible",
						new MovingUnit(Templates.instance().FIGHTER, new MoveStep[] {
								new MoveStep( 7,  9, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
						} ),
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] {
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 7,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { },
						0, MoveResult.MOVE_MASK_NOTHING, 
						new ObservedUnit[] { } 
				),
				new TestParams("A fighter staying and docking in the city stays invisible",
						new MovingUnit(Templates.instance().FIGHTER, new MoveStep[] {
								new MoveStep( 1,  1, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
						} ),
						new MovingUnit(Templates.instance().FIGHTER, new MoveStep[] {
								new MoveStep(-1, -1, ImageKey.FIELD_ID_CITY, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep(-1, -1, ImageKey.FIELD_ID_CITY, MoveResult.MOVE_MASK_NOTHING),
						}),
						new ObservedUnit[] { 
								new ObservedUnit(Templates.instance().FIGHTER, new int[] {  1, 1 } ), 
						},
						0, MoveResult.MOVE_MASK_NOTHING, 
						new ObservedUnit[] { 
						} 
				),
				new TestParams("A ship moving into city is visible, but ends hidden in city",
						new MovingUnit(Templates.instance().FIGHTER, new MoveStep[] {
								new MoveStep( 1,  1, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
						} ),
						new MovingUnit(Templates.instance().TRANSPORTER, new MoveStep[] {
								new MoveStep( 1,  2, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 2,  2, ImageKey.FIELD_ID_CITY, MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_DOCKING),
						}),
						new ObservedUnit[] { 
								new ObservedUnit(Templates.instance().FIGHTER, new int[] {  1, 1 } ), 
						},
						1, MoveResult.MOVE_MASK_MOVING | MoveResult.MOVE_MASK_DOCKING, 
						new ObservedUnit[] { 
						} 
				),
				new TestParams("A ship staying and docking in the city stays invisible",
						new MovingUnit(Templates.instance().FIGHTER, new MoveStep[] {
								new MoveStep( 1,  1, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
						} ),
						new MovingUnit(Templates.instance().TRANSPORTER, new MoveStep[] {
								new MoveStep(-1, -1, ImageKey.FIELD_ID_CITY, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep(-1, -1, ImageKey.FIELD_ID_CITY, MoveResult.MOVE_MASK_DOCKING),
						}),
						new ObservedUnit[] { 
								new ObservedUnit(Templates.instance().FIGHTER, new int[] {  1, 1 } ), 
						},
						0, MoveResult.MOVE_MASK_NOTHING, 
						new ObservedUnit[] {
						} 
				),
				new TestParams("A submarine just showing up and moving away from enemy sub, then this enemy sub goes even further away. Enemy sub image should stay in place.",
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] { // Player 2 sub moving away
								new MoveStep( 6,  9, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 5,  9, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
								new MoveStep( 4,  9, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
						} ),
						new MovingUnit(Templates.instance().SUBMARINE, new MoveStep[] { // Player 1 sub showing up
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_NOTHING),
								new MoveStep( 7,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
								new MoveStep( 8,  8, ImageKey.FIELD_ID_SEA, MoveResult.MOVE_MASK_MOVING),
						}),
						new ObservedUnit[] { // Player 2 sub seen through player 1 eyes
								new ObservedUnit(Templates.instance().SUBMARINE, new int[] { 6, 9 } ),
						},
						2, MoveResult.MOVE_MASK_MOVING, 
						new ObservedUnit[] { // No player 1 units seen by player 2
						} 
				),
		});
	}

	public MoveVisibilityTest(final TestParams testParams) {
		this.testParams = testParams;
	}

	@Before
	public void setUp() throws GameInitializationException {
		final Tutorial tutorial = new Tutorial();
		final GameMap map = tutorial.createMap(TutorialLevel.LAND_UNITS);
		observer = new GameObserverStub();
		playerCities[0] = map.findCities(0).get(0);
		playerCities[1] = map.findCities(1).get(0);
		gameState = new GameState(observer, map);
		gameState.startGame();
		player1GameView = gameState.getCurrentGameView();
		player1GameView.clearGetReady();
	}

	@After
	public void tearDown() {
		GameMoveTest.testIfUnitsClonedForViews(player1GameView, player1GameView.getUnits());
		if (null != player1GameView.getMap()) {
		    final Marshaller marshaller = new Marshaller();
		    marshaller.writeMarshallable(player1GameView.getGameState());
		    final byte[] data = marshaller.pack();
			final Marshaller d = new Marshaller(data);
			final GameState received = new GameState(observer, d);
			GameMapTest.compareMaps(player1GameView.getGameState().getMap(), received.getMap());
			for (int i = 0; i < Player.MAX_NUMBER_OF_PLAYERS; i++) {
				GameMapTest.compareMaps(player1GameView.getGameState().getGameView(i).getMap(), received.getGameView(i).getMap());
			}
			GameMoveTest.testIfUnitsClonedForViews(player1GameView, received.getGameView(0).getUnits());
		}
	}

	public void testMove(final TestParams testParams) {
		assertEquals(0, player1GameView.getPlayer().getId());
		final MovingUnit[] movingUnits = new MovingUnit[] { testParams.movingUnit, testParams.enemyUnit };
		final Unit[] units = new Unit[2];
		// Place units
		for (int i = 1; i >= 0; i--) {
			units[i] = placeUnit(i, movingUnits[i]);
		}
		// Move unit
		player1GameView.setMapViewMode(MapViewMode.NORMAL);
		player1GameView.nextCommandable();
		moveUnit(0, player1GameView, movingUnits[0], units[0]);
		{ // Finish turn and change perspective to enemy
			final boolean isNextPlayer = player1GameView.finishTurn();
			assertTrue(isNextPlayer);
		}
		final GameView player2GameView = gameState.getCurrentGameView();
		assertNotEquals(player1GameView, player2GameView);
		assertEquals(1, player2GameView.getPlayer().getId());
		player2GameView.clearGetReady();
		final int resultMask = GameMoveTest.clearAnimation(player2GameView, observer, testParams.enemyNumberOfAnimations); 
		assertEquals(testParams.expectedEnemyResultMask, resultMask);
		// Check all the view units in the enemy view
		checkViewUnits(player2GameView, testParams.expectedObservedPlayer1Units);
		checkViewUnits(player1GameView, testParams.expectedObservedPlayer2Units);

		player2GameView.setMapViewMode(MapViewMode.NORMAL);
		player2GameView.nextCommandable();
		moveUnit(1, player2GameView, movingUnits[1], units[1]);
		
		{ // Check all the view units in the player 1 view
			final boolean isNextPlayer = player2GameView.finishTurn();
			assertTrue(isNextPlayer);
		}
		assertEquals(player1GameView, gameState.getCurrentGameView());
		player1GameView.clearGetReady();
		// Save and load, and check if the view units are still at the same place. 
		final GameState received = GameTest.checkMarshallingGame(observer, gameState, player1GameView);
		checkViewUnits(received.getGameView(1), testParams.expectedObservedPlayer1Units);
		checkViewUnits(received.getGameView(0), testParams.expectedObservedPlayer2Units);
	}

	public void moveUnit(final int playerId, final GameView view, final MovingUnit mu, final Unit unit) {
		assertEquals(playerId, view.getPlayer().getId());
		assertEquals(playerId, gameState.getCurrentGameView().getPlayer().getId());
		if (null != mu) {
			PlayfieldPosition moveStartPos = mu.steps[0].destPos;
			assertEquals(unit, view.getVisibleCommandableInFocus());
			for (int i = 1; i < mu.steps.length; i++) {
				final MoveStep step = mu.steps[i];
				final PlayfieldPosition destination = step.destPos;
				final PlayfieldPosition moveEndPos;
				if (null == destination) { // It's skipping turn
					if (null == moveStartPos) {
						moveEndPos = playerCities[playerId].getPosition();
					} else {
						moveEndPos = moveStartPos;
					}
				} else {
					moveEndPos = destination;
				}
				final Order order;
				final int numberOfAnimations;
				if (null == destination) { // It's skipping turn
					order = new OrderSkipMove(gameState, unit);
					if (unit.getUnitTemplate().militaryBranch == MilitaryBranch.SEA_UNIT) {
						numberOfAnimations = 1;
					} else {
						numberOfAnimations = 0;
					}
				} else {
					if (null != moveStartPos) {
						assertEquals(step.mapAt, gameState.getMap().getAt(moveEndPos).getTemplateId());
					}
					order = new OrderGoTo(gameState, unit, destination);
					numberOfAnimations = 1;
				}
				assertTrue(order.makesSense());
				view.setOrder(order);
				assertEquals(step.moveResult, clearAnimation(view, numberOfAnimations));
				assertEquals(moveEndPos, unit.getPosition());
				moveStartPos = moveEndPos;
			}
		}
	}

	public Unit placeUnit(final int playerId, final MovingUnit mu) {
		Unit ret = null;
		if (null != mu) {
			PlayfieldPosition moveStartPos = mu.steps[0].destPos;
			ret = GameMoveTest.createUnit(player1GameView, playerCities[playerId], mu.unitType, moveStartPos);
			if (null != moveStartPos) {
				assertEquals(mu.steps[0].mapAt, gameState.getMap().getAt(moveStartPos).getTemplateId());
			}
		}
		return ret;
	}

	public void checkViewUnits(final GameView gameView, final ObservedUnit[] expectedObservedUnits) {
		final TreeMap<CommandableKey, UnitAlike> uams = (TreeMap<CommandableKey, UnitAlike>)(gameView.getMap().getUnitsMap());
		ArrayList<ViewUnit> vus = new ArrayList<ViewUnit>();
		for (final UnitAlike ua : uams.values()) {
			if (ua instanceof ViewUnit) {
				if (ua.getPlayerId() != gameView.getPlayer().getId()) {
					System.out.println(ua.getLogRepresentation() + ua.getPosition().toString());
					vus.add((ViewUnit) ua);
				}
			}
		}
		if (null != expectedObservedUnits) {
			assertEquals(expectedObservedUnits.length, vus.size());
			for (int i = 0; i < expectedObservedUnits.length; i++) {
				final ObservedUnit ou = expectedObservedUnits[i];
				final ViewUnit vu = vus.get(i);
				assertEquals(ou.unitType, vu.getUnitTemplate());
				assertEquals(new PlayfieldPosition(ou.pos), vu.getPosition());
			}
		}
	}

	@Test
	public void testMoves() {
		testMove(testParams);
	}

	private int clearAnimation(final GameView view, final int expectedNumberOfAnimations) {
		return GameMoveTest.clearAnimation(view, observer, expectedNumberOfAnimations);
	}
	
	private final TestParams testParams;
	private GameObserverStub observer;
	private GameState gameState;
	private GameView player1GameView;
	private FieldCity[] playerCities = new FieldCity[2]; 
}
