package de.ped.empire.test;

import static org.junit.Assert.*;

import org.junit.Test;

import de.ped.empire.logic.GameInitializationException;
import de.ped.empire.logic.GameState;
import de.ped.empire.logic.GameView;
import de.ped.empire.logic.GameMap;
import de.ped.empire.logic.Field;
import de.ped.empire.logic.FieldCity;
import de.ped.empire.logic.GameObserver;
import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.ImageKey;
import de.ped.empire.logic.StepRatingMap;
import de.ped.empire.logic.StepRatingPosition;
import de.ped.empire.logic.Templates;
import de.ped.empire.logic.Unit;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.test.GenericTest;

/**
 * Tests for the {@link Unit} class.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class UnitTest {

	@Test
	public void testGetStepRatings3() throws GameInitializationException {
		checkGetStepRatings(3);
	}

	@Test
	public void testGetStepRatings5() throws GameInitializationException {
		checkGetStepRatings(5);
	}

	@Test
	public void testGetStepRatings7() throws GameInitializationException {
		checkGetStepRatings(7);
	}

	/**
	 * @see Unit#equals(Object)
	 * @see Unit#clone()
	 * @throws GameInitializationException if game initialization failed
	 */
	@Test
	public void testObjectBasics() throws GameInitializationException {
		final int size = 3;
		final GameObserver observer = new GameObserverStub();
		final int min = 2;
		final int mid = min + size / 2;
		final GameMap map = createMap(size, min, mid);
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		final GameView gameView = gameState.getCurrentGameView();
		assertEquals(1, 1);
		final FieldCity city = GameTest.findCity(gameView);
		gameView.setCityProduction(city, Templates.instance().NOTHING);
		final Unit unit1 = Unit.create(gameState, Templates.instance().ARMY, city);
		final Unit unit2 = Unit.create(gameState, Templates.instance().ARMY, city);
		gameState.addUnit(city, unit1);
		GenericTest.checkObjectBasics(false, unit1, unit2);
		GenericTest.checkObjectBasics(false, map.getAt(3, 2), map.getAt(2, 3));
		final GameMap vmap = gameView.getMap();
		GenericTest.checkObjectBasics(false, vmap.getAt(3, 2), vmap.getAt(2, 3));
	}

	/** 
	 * @param size should be odd to work out, and not greater than 7 for a field of size 10
	 * @throws GameInitializationException if game initialization failed
	 */
	public void checkGetStepRatings(final int size) throws GameInitializationException {
		final GameObserver observer = new GameObserverStub();
		final int min = 2;
		final int mid = min + size / 2;
		final GameMap map = createMap(size, min, mid);
		final GameState gameState = new GameState(observer, map);
		gameState.startGame();
		final GameView gameView = gameState.getCurrentGameView();
		assertEquals(1, 1);
		final FieldCity city = GameTest.findCity(gameView);
		gameView.setCityProduction(city, Templates.instance().NOTHING);
		final Unit unit1 = Unit.create(gameState, Templates.instance().ARMY, city);
		gameState.addUnit(city, unit1);
		assertEquals(1, unit1.getStepsLeftThisTurn());
		gameView.clearGetReady();
		assertEquals(unit1, gameView.getVisibleCommandableInFocus());
		StepRatingMap ratings = unit1.getStepRatings();
		assertEquals(null, ratings.get(mid, mid).stepTaken);
		if (size > 1) {
			assertEquals(DigitalJoystickDirection.UP,    ratings.get(mid, mid - 1).stepTaken);
			assertEquals(DigitalJoystickDirection.DOWN,  ratings.get(mid, mid + 1).stepTaken);
			assertEquals(DigitalJoystickDirection.LEFT,  ratings.get(mid - 1, mid).stepTaken);
			assertEquals(DigitalJoystickDirection.RIGHT, ratings.get(mid + 1, mid).stepTaken);
		}
		logDirections(size, min, mid, ratings);
		for (int y = min; y < min + size; y++) {
			for (int x = min; x < min + size; x++) {
				if ((mid == x) && (mid == y)) {
					// Nothing to do
				} else {
					final boolean expected = !((x == y) || (x + y == 2 * min + size - 1));
					final StepRatingPosition posRating = ratings.get(x, y);
					assertEquals("x=" + x + ",y=" + y + "->" + posRating.stepTaken.key, expected, posRating.stepTaken.is4Direction());
				}
			}
		}
	}

	private void logDirections(final int size, final int min, final int mid, StepRatingMap ratings) {
		final Logger logger = Logger.getLogger(getClass());
		logger.setLogLevel(Level.INFO);
		final String rating = ratings.getLogDirections(size, min);
		final String[] lines = rating.split("\n");
		for (String line : lines) {
			logger.debug(line);
		}
	}

	protected GameMap createMap(final int size, final int min, final int mid) {
		// assumes default topology is SQUARE8.
		final GameMap map = new GameMap(new GameProperties());
		final PlayfieldDimension mapSize = map.getSize();
		int id = 0;
		for (int x = min; x < mapSize.getWidth() - 1; x++) {
			for (int y = min; y < mapSize.getHeight() - 1; y++) {
				Field field;
				if ((mid == x) && (mid == y)) {
					FieldCity city = new FieldCity(id, x, y);
					id++;
					city.setPlayerId(0);
					field = city;
				} else if ((x < min + size) && (y < min + size)) {
					field = new Field(x, y, ImageKey.FIELD_ID_LAND);
				} else {
					field = new Field(x, y, ImageKey.FIELD_ID_SEA);
				}
				map.setField(field);
			}
		}
		map.finish();
		return map;
	}
}
