package de.ped.gedicht.logic;

import de.ped.tools.CollectionUtil;
import de.ped.tools.HtmlHelper;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextDocumentType;
import de.ped.tools.TextFormatter;

/**
 * Abstract class to create poems.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Poem {
	/**
	 * The gender of a noun.
	 * 
	 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Genus {
		MALE,
		FEMALE,
		NEUTRUM,
	};

	/**
	 * The type (algorithm) of poem, laid out as factory.
	 * 
	 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Algorithm {
		WITHOUT_RHYME("Action.New1", "Gedicht ohne Reim"),
		WITH_RHYME   ("Action.New2", "Gedicht mit Reim");

		public final I18NStringWithFillIns heading;
		public final String actionKey;

		private Algorithm(final String key, final String alt) {
			this.actionKey = key;
			this.heading = new I18NStringWithFillIns(key + ".Type", alt, null);
		}
	};

	/**
	 * @param str
	 *            a string to convert
	 * @return the input string, but the first character of the string is now
	 *         upper case.
	 */
	public static String upperFirst(final String str) {
		if (CollectionUtil.isNullOrEmpty(str)) {
			return str;
		} else if (str.length() == 1) {
			return str.toUpperCase();
		} else {
			return str.substring(0, 1).toUpperCase() + str.substring(1, str.length());
		}
	}

	/**
	 * @param arr
	 *            the array to pick from
	 * @return one of <em>arr</em> is picked with equal probability, and
	 *         returned with a blank at the end.
	 */
	public static String createOneOf(final String[] arr) {
		return createOneOf(arr, 1);
	}

	/**
	 * @param arr
	 *            the array to pick from
	 * @param prob
	 *            the inverse probability if there should be picked something.
	 *            So, e.g., a 6 means, that only in 1/6 of all calls any element
	 *            from the array will be selected at all.
	 * @return a <em>prob</em>-sided dice is thrown, and if that fell on a 1,
	 *         one of <em>arr</em> is picked with equal probability, and
	 *         returned with a blank at the end. Otherwise, <code>null</code> is
	 *         returned.
	 */
	public static String createOneOf(final String[] arr, final int prob) {
		String ret = "";
		if (0 == MathUtil.random(prob)) {
			ret = arr[MathUtil.random(arr.length)] + " ";
		}
		return ret;
	}

	protected static void debug(final Algorithm algo) {
		System.out.println(PoemFacade.instance().create(algo).createPoem(TextDocumentType.TEXT, null));
	}
	
	protected Poem(final Algorithm algo) {
		this.algo = algo;
	}

	public String createPoem(final TextDocumentType type, String heading) {
		final TextDocumentBuilder doc = new TextDocumentBuilder((Messages) null);
		doc.setTextDocumentType(type);
		doc.setEmbedded(true);
		doc.tf().setIndentString("\t");
		if (null != heading) {
			doc.addSectionHeading(2, algo.name().toLowerCase(), heading);
		}
		beginPoem(doc);
		createPoem(doc);
		endPoem(doc);
		return doc.result(null);
	}

	protected void beginPoem(final TextDocumentBuilder doc) {
		// Nothing to do
	}

	protected void endPoem(final TextDocumentBuilder doc) {
		// Nothing to do
	}

	protected void beginVerse(final TextDocumentBuilder doc) {
		doc.beginParagraph(TextFormatter.htmlAttribute(TextFormatter.CLASS, "verse"));
	}

	protected void endVerse(final TextDocumentBuilder doc) {
		doc.endParagraph();
	}

	protected void addLine(final TextDocumentBuilder doc, final int lineNo, final String line) {
		beginLine(doc, lineNo);
		String str = line;
		if (doc.getTextDocumentType().isPlain()) {
			str = HtmlHelper.deSgmlify(str);
		}
		doc.add(str);
		endLine(doc, lineNo);
	}

	private void beginLine(final TextDocumentBuilder doc, final int lineNo) {
		if (doc.getTextDocumentType().isPlain()) {
			if (0 < lineNo) {
				doc.add(doc.tf().newLine());
			}
		} else {
			final TextFormatter tf = doc.tf();
			doc.add(tf.openHtmlElement("span", TextFormatter.htmlAttribute(TextFormatter.CLASS, "line")));
		}
	}

	private void endLine(final TextDocumentBuilder doc, final int lineNo) {
		if (doc.getTextDocumentType().isPlain()) {
			// Nothing to do
		} else {
			final TextFormatter tf = doc.tf();
			doc.add(tf.closeHtmlElement("span", TextFormatter.INLINE) + tf.newLine());
		}
	}

	public abstract void createPoem(final TextDocumentBuilder doc);
	
	public final Algorithm algo;
}
