package de.ped.gedicht.logic;

import java.util.HashSet;

import de.ped.tools.MathUtil;
import de.ped.tools.TextDocumentBuilder;

/**
 * Creates a poem based on a vocabulary of less than 100 words.
 * A verse follows the rhyme pattern <em>abab</em>;
 * for each of the four lines a different sentence strucure is used.
 * 
 * @author Copyright (c) 1987-2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Poem2 extends Poem {

	/**
	 * Sends a poem to the console.
	 * 
	 * @param args ignored
	 */
	public static void main(final String[] args) {
		Poem.debug(Poem.Algorithm.WITH_RHYME);
	}

	private static final String[] ARTICLES = new String[] {
		"die",
		"diese",
		"eine",
		"keine",
		"so eine",
		"deine",
		"jene",
		"jede",
	};

	private static final String[][] SUBJECTS = new String[][] {
		{
			"Vergebung",
			"Deutung",
			"Vernichtung",
			"Achtung",
			"Geltung",
			"Verleugnung",
			"Wandlung",
			"Entstehung",
			"Läuterung",
			"Haltung",
		}, {
			"Kühnheit",
			"Dunkelheit",
			"Wahrheit",
			"Vergangenheit",
			"Zeit",
			"Einigkeit",
			"Menschheit",
			"Gottheit",
			"Ewigkeit",
			"Schönheit",
		}, {
			"Antipathie",
			"Sympathie",
			"Monotonie",
			"Phantasie",
			"Energie",
			"Hierarchie",
			"Harmonie",
			"Anomalie",
			"Autonomie",
			"Parodie",
		}
	};

	private static final String[] ADDITION1 = new String[] {
		"besonders",
		"sehr",
		"überaus",
		"kaum",
		"allzu",
	};

	private static final String[][] ADVERBS = new String[][] {
		{
			"dominant",
			"tolerant",
			"ignorant",
			"gesandt",
			"bekannt",
			"verbannt",
			"markant",
			"verkannt",
			"verwandt",
			"elegant",
		}, {
			"unglaublich",
			"deutlich",
			"göttlich",
			"unermesslich",
			"vergänglich",
			"unsterblich",
			"phantastisch",
			"unheimlich",
			"vergeblich",
			"wahrscheinlich",
		}
	};

	private enum VerbType {
		ACTING (new String[] {
				"endet",
				"stirbt",
				"klagt",
				"entsteht",
				"ruht",
				"wartet",
				"strahlt",
				"leuchtet",
				"verrinnt",
				"warnt",
				"vergeht",
		}),
		BEING(new String[] {
				"ist",
				"war",
				"sei",
				"wäre",
				"bleibt",
		});

		private VerbType(String[] verbs) {
			this.verbs = verbs;
		}

		private final String[] verbs;
	};

	private static final String[] QUESTIONS = new String[] {
		"wann",
		"weshalb",
		"wodurch",
		"wo",
	};

	private static final String[] ADDITION2 = new String[] {
		"seit Anbeginn",
		"auf Ewigkeit",
		"oft",
		"bisweilen",
		"trotzdem",
		"nie",
		"schließlich",
		"selten",
		"nun",
		"vielleicht",
		"jedoch",
		"daher",
	};

	private static final int RHYME_INDEX_ADVERBS = 0;
	private static final int RHYME_INDEX_SUBJECTS = 1;

	private enum RhymeMode {
		NO_RHYME,
		DEFINE_RHYME,
		FOLLOW_RHYME;
	}

	Poem2(final Algorithm algo) {
		super(algo);
	}

	@Override
	public void createPoem(final TextDocumentBuilder doc) {
		for (int verse = 0; verse < 3; verse++) {
			beginVerse(doc);
			createVerse(doc);
			endVerse(doc);
		}
	}

	private void createVerse(final TextDocumentBuilder doc) {
		usedWords.clear();
		for (int lineNo = 0; lineNo < 4; lineNo++) {
			addLine(doc, lineNo, createLine(lineNo));
		}
	}

	private String createLine(final int line) {
		final StringBuffer buf = new StringBuffer();
		isQuestion = false;
		switch (line) {
			case 0: {
				buf.append(createSubstantive(RhymeMode.NO_RHYME));
				buf.append(createVerb(VerbType.ACTING));
				buf.append(createAdverb(RhymeMode.DEFINE_RHYME));
			}
			break;
			case 1: {
				buf.append(createSubstantive(RhymeMode.NO_RHYME));
				buf.append(createVerb(VerbType.BEING));
				buf.append(createSubstantive(RhymeMode.DEFINE_RHYME));
			}
			break;
			case 2: {
				buf.append(createAddition());
				buf.append(createVerb(VerbType.ACTING));
				buf.append(createSubstantive(RhymeMode.NO_RHYME));
				buf.append(createAdverb(RhymeMode.FOLLOW_RHYME));
			}
			break;
			case 3: {
				buf.append(createAddition());
				buf.append(createVerb(VerbType.BEING));
				buf.append(createSubstantive(RhymeMode.NO_RHYME));
				buf.append(createSubstantive(RhymeMode.FOLLOW_RHYME));
			}
			break;
		}
		String lineStr = Poem.upperFirst(buf.toString().trim());
		if (isQuestion) {
			lineStr += "?";
		} else {
			lineStr += createOneOf(new String[]{".", ".", "&nbsp;&hellip;", "!"});
		}
		return lineStr;
	}

	private String createSubstantive(final RhymeMode mode) {
		final StringBuffer buf = new StringBuffer();
		buf.append(createOneOf(ARTICLES));
		final int endingIdx;
		switch (mode) {
		case DEFINE_RHYME:
			endingIdx = MathUtil.random(SUBJECTS.length);
			rhymeIndex[RHYME_INDEX_SUBJECTS] = endingIdx;
			break;
		case FOLLOW_RHYME:
			endingIdx = rhymeIndex[RHYME_INDEX_SUBJECTS];
			break;
		case NO_RHYME:
		default:
			endingIdx = MathUtil.random(SUBJECTS.length);
		}
		String subst;
		do {
			subst = createOneOf(SUBJECTS[endingIdx]);
		} while (usedWords.contains(subst));
		usedWords.add(subst);
		buf.append(subst);
		return buf.toString();
	}

	private String createAdverb(final RhymeMode mode) {
		final StringBuffer buf = new StringBuffer();
		buf.append(createOneOf(ADDITION1, 2));
		final int endingIdx;
		switch (mode) {
		case DEFINE_RHYME:
			endingIdx = MathUtil.random(ADVERBS.length);
			rhymeIndex[RHYME_INDEX_ADVERBS] = endingIdx;
			break;
		case FOLLOW_RHYME:
			endingIdx = rhymeIndex[RHYME_INDEX_ADVERBS];
			break;
		case NO_RHYME:
		default:
			endingIdx = MathUtil.random(ADVERBS.length);
		}
		String subst;
		do {
			subst = createOneOf(ADVERBS[endingIdx]);
		} while (usedWords.contains(subst));
		usedWords.add(subst);
		buf.append(subst);
		return buf.toString();
	}

	private String createVerb(final VerbType mode) {
		return createOneOf(mode.verbs);
	}

	private String createAddition() {
		final String ret;
		if (0 == MathUtil.random(3)) {
			ret = createOneOf(QUESTIONS);
			isQuestion = true;
		} else {
			ret = createOneOf(ADDITION2);
		}
		return ret;
	}

	private HashSet<String> usedWords = new HashSet<String>();

	/**
	 * An array containing the current rhyme index for adverbs (at index
	 * {@link #RHYME_INDEX_ADVERBS}) and subjects (@link at index
	 * {@link #RHYME_INDEX_SUBJECTS})
	 */
	private int rhymeIndex[] = new int[2];

	private boolean isQuestion = false;
}
