package de.ped.ituneslibconverter;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.io.StringWriter;

import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

/**
 * Converts the iTunes library (in XML).
 * Current only conversion implemented is the move of all the files
 * from localfile/C/public/MP3 to localfile/X/public/MP3.  
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class LibConverter {

	/**
	 * Executes an XSLT transformation and returns the transformed stream as
	 * String.
	 * 
	 * @param xmlSource
	 *            the source to transform
	 * @param xslSource
	 *            the source containing the transformation rules
	 * @return the transformed stream
	 * @throws Exception
	 *             if transformation was not possible
	 */
	public String xsltTransform(final Source xmlSource, final Source xslSource) throws Exception {
		final StringWriter s = new StringWriter();
		try {
			// das Factory-Pattern untersttzt verschiedene XSLT-Prozessoren
			final TransformerFactory transFact = TransformerFactory.newInstance();
			final Transformer trans = transFact.newTransformer(xslSource);
			trans.transform(xmlSource, new StreamResult(s));
		} catch (final Exception e) {
			s.append("Tried parsing:");
			s.append("<br/>\n");
			s.append(e.getMessage());
			s.append("<br/>\n");
			for (final StackTraceElement el : e.getStackTrace()) {
				s.append(el.toString());
				s.append("<br/>\n");
			}
		}
		final String str = s.toString();
		return str.replaceFirst("<\\?xml [^\\?]*\\?>", "");
	}

	/**
	 * Executes an XSLT transformation and returns the transformed stream as
	 * String.
	 * 
	 * @param xmlFileName
	 *            the name of the file to transform
	 * @param xslFileName
	 *            the name of the file containing the transformation rules
	 * @return the transformed stream
	 * @throws Exception
	 *             if transformation was not possible
	 */
	public String xsltTransformFilenames(final String xmlFileName, final String xslFileName) throws Exception {
		final StringWriter s = new StringWriter();
		final File xmlFile = new File(xmlFileName);
		if (!xmlFile.exists()) { return "File " + xmlFileName + " not found!"; }
		final File xslFile = new File(xslFileName);
		if (!xslFile.exists()) { return "File " + xslFileName + " not found!"; }

//        s.append(xmlUrl.toExternalForm());
//        s.append("\n");
//        s.append(xslUrl.toExternalForm());
//        s.append("\n");
//        try {
//            InputStream xmlIS = xmlUrl.openStream();
//            xmlIS.read();
//            InputStream xslIS = xslUrl.openStream();
//            xslIS.read();
//        } catch (final Exception e) {
//            s.append("Tried to open input streams: \n");
//            s.append(e.getMessage());
//            s.append("\n");
//        }
		try {
			// JAXP liest Daten ber die Source-Schnittstelle
			Source xmlSource = new StreamSource(xmlFile);
			Source xslSource = new StreamSource(xslFile);
			s.append(xsltTransform(xmlSource, xslSource));
		} catch (final Exception e) {
			s.append("Tried parsing:");
			s.append("<br/>\n");
			s.append(e.getMessage());
			s.append("<br/>\n");
			for (final StackTraceElement el : e.getStackTrace()) {
				s.append(el.toString());
				s.append("<br/>\n");
			}
		}
		return s.toString();
	}

	/**
	 * Executes an XSLT transformation and returns the transformed stream as
	 * String.
	 * 
	 * @param xmlFileName
	 *            the name of the file to transform
	 * @param xslFileContent
	 *            the content of the file containing the transformation rules
	 * @return the transformed stream
	 * @throws Exception
	 *             if transformation was not possible
	 */
	public String xsltTransform(final String xmlFileName, final String xslFileContent) throws Exception {
		final StringWriter s = new StringWriter();
		final File xmlFile = new File(xmlFileName);
		if (!xmlFile.exists()) {
			return "File " + xmlFileName + " not found!";
		}
		try {
			// JAXP liest Daten ber die Source-Schnittstelle
			final Source xmlSource = new StreamSource(xmlFile);
			final Source xslSource = new StreamSource(new StringReader(xslFileContent));
			s.append(xsltTransform(xmlSource, xslSource));
		} catch (final Exception e) {
			s.append("Tried parsing:");
			s.append("<br/>\n");
			s.append(e.getMessage());
			s.append("<br/>\n");
			for (final StackTraceElement el : e.getStackTrace()) {
				s.append(el.toString());
				s.append("<br/>\n");
			}
		}
		return s.toString();
	}

	public void actionXSLT(final String xmlFileName, final String xslFileName, final String destFileName) {
		final FileWriter xmlDestFile;
		try {
			xmlDestFile = new FileWriter(destFileName);
			final String content = xsltTransformFilenames(xmlFileName, xslFileName);
			xmlDestFile.write(content);
			xmlDestFile.close();
		} catch (final IOException e) {
			e.printStackTrace();
		} catch (final Exception e) {
			e.printStackTrace();
		}
	}

	public void actionSimple(final String xmlFileName, final String xslFileName, final String destFileName) {
		try {
			final BufferedReader xmlSourceFile = new BufferedReader(new FileReader(xmlFileName));
			final BufferedWriter xmlDestFile = new BufferedWriter(new FileWriter(destFileName));
			String srcLine;
			String destLine;
			while ((srcLine = xmlSourceFile.readLine()) != null) {
				destLine = srcLine.replace("file://localhost/C:/Users/Public/MP3", "file://localhost/X:/Public/MP3");
				xmlDestFile.write(destLine + '\n');
			}
			xmlSourceFile.close();
			xmlDestFile.close();
		} catch (final IOException e) {
			e.printStackTrace();
		} catch (final Exception e) {
			e.printStackTrace();
		}
	}

	public void action(final String xmlFileName, final String xslFileName, final String destFileName) {
		actionSimple(xmlFileName, xslFileName, destFileName);
	}

	/**
	 * @param args currently ignored
	 */
	public static void main(final String[] args) {
		final LibConverter lc = new LibConverter();
		lc.action("X:/Public/MediathekLucy2010-07-11.xml", "X:/Public/MediathekConverter.xsl", "X:/Public/MediathekConverted.xml");
	}
}
