package de.ped.ituneslibconverter;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;

import de.ped.tools.TextFormatter;

/**
 * Represents a playlist.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Playlist {
	public static final String CHAR_ENCODING = "UTF-16";

	String[] legend = null;
	ArrayList<String[]> entries = new ArrayList<String[]>();

	public String[] getLegend() {
		return legend;
	}

	public int getLegendIndex(final String legendKey) {
		for (int i = 0; i < legend.length; i++) {
			if (legend[i].equals(legendKey)) {
				return i;
			}
		}
		return -1;
	}

	public int size() {
		return entries.size();
	}

	public String[] getEntry(final int index) {
		return entries.get(index);
	}

	public void setEntry(final int index, final String[] value) {
		entries.set(index, value);
	}

	public String getEntry(final int index, final int column) {
		final String[] entry = entries.get(index);
		return entry[column];
	}

	public void setEntry(final int index, final int column, final String value) {
		final String[] entry = entries.get(index);
		entry[column] = value;
	}

	public void load(final Reader reader) throws IOException {
		final BufferedReader file = new BufferedReader(reader);
		String line = null;
		int lineNumber = 0;
		do {
			line = file.readLine();
			if (null == line) {
				break;
			}
//			System.out.println(lineNumber + ": " + line);
			String[] parsedLine = parseLine(line);
			if (0 == lineNumber) {
				legend = parsedLine;
			} else {
				entries.add(parsedLine);
			}
			lineNumber++;
		} while (null != line);
		file.close();
	}

	public void save(final Writer writer) throws IOException {
		final BufferedWriter file = new BufferedWriter(writer);
		file.write(composeLine(legend));
		file.write("\n");
		for (int i = 0; i < entries.size(); i++) {
			file.write(composeLine(entries.get(i)));
			file.write("\n");
		}
		file.close();
	}

	public void loadFile(final String playlistFileName) throws IOException {
		// We need control over the charset, as the default charset is wrong for the files to read.
		final FileInputStream fis = new FileInputStream(playlistFileName);
		final InputStreamReader isr = new InputStreamReader(fis, CHAR_ENCODING);
//		System.out.println(isr.getEncoding());
		load(isr);
	}

	public void saveFile(final String playlistFileName) throws IOException {
		// We need control over the charset, as the default charset is wrong for the files to write.
		final FileOutputStream fos = new FileOutputStream(playlistFileName);
		final OutputStreamWriter osw = new OutputStreamWriter(fos, CHAR_ENCODING);
		save(osw);
	}

	public String[] parseLine(final String line) {
		if (null != line) {
			return line.split("\t");
		}
		return null;
	}

	public String composeLine(final String[] entry) {
		if (null != entry) {
			final StringBuffer ret = new StringBuffer();
			for (int i = 0; i < entry.length; i++) {
				if (i > 0) {
					ret.append("\t");
				}
				ret.append(entry[i]);
			}
			return ret.toString();
		}
		return null;
	}

	public void replaceForAllLines(final int columnIndex, final String toReplace, final String replaceBy) {
		for (int i = 0; i < size(); i++) {
			String name = getEntry(i, columnIndex);
			name = name.replace(toReplace, replaceBy);
			setEntry(i, columnIndex, name);
		}
	}

	/**
	 * Does a binary file copy.
	 * @param srcFilename the complete source file name and path.
	 * @param destFilename the complete destination file name and path.
	 * @throws IOException
	 *             if something happened during the file operation
	 */
	public static void copy(final String srcFilename, final String destFilename) throws IOException {
		BufferedInputStream is = new BufferedInputStream(new FileInputStream(srcFilename));
		BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(destFilename));
		int oneChar;
		while ((oneChar = is.read()) != -1) {
			os.write(oneChar);
		}
		is.close();
		os.close();
	}

	public static String createToName(final String fromName, final String toPath, int newIndex) {
		final int fileNameIndex = fromName.lastIndexOf("\\") + 1;
		final StringBuffer toName = TextFormatter.format(newIndex, 3, TextFormatter.RIGHT, '0');
		toName.append("-");
		toName.append(fromName.substring(fileNameIndex));
		final String ret = toName.toString();
		return toPath + ret;
	}

	public void collectAll(final String path) throws IOException {
		if (null != path) {
			final int columnIndex = getLegendIndex("Ort");
			for (int i = 0; i < size(); i++) {
				final String fromName = getEntry(i, columnIndex);
				final String toName = createToName(fromName, path, i + 1);
				System.out.println(fromName + " -> " + toName);
				copy(fromName, toName);
			}
		}
	}
	
}
