package de.ped.kitten.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.util.HashMap;

import de.ped.kitten.logic.Atari8bitScreenUtil;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Defines how to draw objects like {@link TrueTypeMonospaceFont} on a graphics
 * object.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Atari8bitScreen {

	public static final int X = 0;
	public static final int Y = 1;

	/**
	 * Defines the border around the screen.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Border {
		public final int top;
		public final int bottom;
		public final int left;
		public final int right;

		public Border(final int top, final int bottom, final int left, final int right) {
			this.top = top;
			this.bottom = bottom;
			this.left = left;
			this.right = right;
		}
		
		public Border(final int topAndBottom, final int leftAndRight) {
			this(topAndBottom, topAndBottom, leftAndRight, leftAndRight);
		}
	}

	public static Border DEFAULT_BORDER = new Border(3 * 8, 1 * 8);

	public static Border NULL_BORDER = new Border(0, 0);

	private final Logger logger = Logger.getLogger(getClass());
	{
		logger.setLogLevel(Level.INFO);
	}
   
	private Border border;

	public Border getBorder() {
		return border;
	}

	public void setBorder(final Border border) {
		this.border = border;
	}

	private final boolean isCaching;

	public Dimension getDimension(final int stretch, final boolean isWithBorder) {
		final Dimension ret = new Dimension(40 * 8 * stretch, 24 * 8 * stretch);
		if (isWithBorder) {
			ret.setSize(ret.width + (border.left + border.right), ret.height + (border.top + border.bottom));
		}
		return ret;
	}

	private int[] stretch = new int[] { 1, 1 };

	private ColorMap colorMap = new ColorMap("default");
	private TrueTypeMonospaceFont font;

	private final HashMap<CharRepresentation, CharImage> symbolImages = new HashMap<CharRepresentation, CharImage>();

	public Atari8bitScreen(final Border border, final boolean isCaching) {
		this.border = border;
		this.isCaching = isCaching;
		colorMap = ColorMap.create(Color.white, "white");
		font = new TrueTypeMonospaceFont();
	}

	public TrueTypeMonospaceFont getFont() {
		return font;
	}

	public void setFont(final TrueTypeMonospaceFont font) {
		this.font = font;
	}

	public void setColorMap(final ColorMap colors) {
		colorMap = colors;
	}

	public void setStretch(final int factor) {
		setStretch(X, factor);
		setStretch(Y, factor);
	}

	public void setStretch(final int index, final int factor) {
		stretch[index] = factor;
	}

	public void drawStringAt(final Graphics g, final String str, final double x, final double y) {
		if (null != str) {
			for (int i = 0; i < str.length(); i++) {
				drawCharAt(g, str.charAt(i), x + i, y);
			}
			
		}
	}

	public void draw4ColCharAt(
			final Graphics g, 
			final byte c, 
			final ColorMap[] colors, 
			final double x, 
			final double y) {
		final int colorIndex = Atari8bitScreenUtil.getColorIndexFor4ColChar(c);
		final char ch = Atari8bitScreenUtil.getCharIndexFor4ColChar(c);
		drawCharAt(g, ch, x, y, colors[colorIndex]);
	}

	public void drawCharAt(final Graphics g, final char c, final double x, final double y) {
		drawCharAt(g, c, x, y, colorMap);
	}

	public void drawCharAt(final Graphics g, final char c, final double x, final double y, final ColorMap colorMap) {
		if (isCaching) {
			final CharRepresentation rep = new CharRepresentation(c, font, colorMap, stretch[X], stretch[Y]);
			drawAndCacheCharAt(g, rep, x, y);
		} else {
			final int xpos = (int)Math.round(x * 8 * stretch[X]);
			final int ypos = (int)Math.round(y * 8 * stretch[Y]);
//			g.setColor(Color.YELLOW);
//			g.drawRect(xpos + border.LEFT, ypos + border.TOP, 1, 1);
			final String[] sdef = font.getCharDef(c, stretch[X], stretch[Y]);
			drawChar(g, sdef, xpos, ypos, colorMap);
		}
	}

	/**
	 * The most fundamental method for drawing a character of this font. All
	 * more convenient methods in the end call this method.
	 * 
	 * @param g
	 *            the graphics context to draw to
	 * @param def
	 *            the definition of the character. Assumes that the definition
	 *            was already stretched, e.g., by
	 *            {@link TrueTypeMonospaceSingleSizeChar#stretch}
	 * @param x
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the character
	 * @param y
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the character
	 * @param colorMap
	 *            the colors to use when drawing that character
	 */
	private void drawChar(final Graphics g, final String[] def, final int x, final int y, final ColorMap colorMap) {
		if (null != def) {
			final Color background = colorMap.get(null);
			if (null != background) {
				g.setColor(background);
				g.fillRect(x + border.left, y + border.top, def[0].length(), def.length);
			}
			for (int yi = 0; yi < def.length; yi++) {
				final int ypos = y + yi;
				final String line = def[yi];
				for (int xi = 0; xi < line.length(); xi++) {
					final int xpos = x + xi;
					final Color col = colorMap.get(Character.valueOf(line.charAt(xi)));
					drawPixel(g, xpos, ypos, 1, 1, col);
				}
			}
		}
	}

	public void drawPixel(
			final Graphics g, 
			final int xpos, 
			final int ypos, 
			final int stretchX, 
			final int stretchY, 
			final Color color) {
		if (null != color) {
			g.setColor(color);
			drawPixel(g, xpos, ypos, stretchX, stretchY);
		}
	}

	/**
	 * The most fundamental method to draw a pixel of the font. Overload for
	 * different optical effects.
	 * 
	 * @param g
	 *            the graphics context to draw to
	 * @param xpos
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the pixel
	 * @param ypos
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the pixel
	 * @param stretchX
	 *            the stretch factor to resize the drawing of the character
	 * @param stretchY
	 *            the stretch factor to resize the drawing of the character
	 */
	protected void drawPixel(
			final Graphics g, 
			final int xpos, 
			final int ypos, 
			final int stretchX, 
			final int stretchY) {
		g.fillRect(xpos + border.left, ypos + border.top, stretchX, stretchY);
	}

	/**
	 * Draws the given image at the given character position. This is the most
	 * efficient drawing method.
	 * 
	 * @param g
	 *            the graphics context to draw to
	 * @param image
	 *            the image to draw
	 * @param x
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the image
	 * @param y
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the image
	 */ 
	public void drawImageAt(final Graphics g, final Image image, final double x, final double y) {
		final int xpos = (int)Math.round(x * 8 * stretch[X]);
		final int ypos = (int)Math.round(y * 8 * stretch[Y]);
    	g.drawImage(image, xpos + border.left, ypos + border.top, null);
	}

	/**
	 * Draws a symbol as image, and caches that image. This is the fastest and
	 * therefore preferred drawing method for characters.
	 * 
	 * @param g
	 *            the graphics context to draw to
	 * @param rep
	 *            the char representation
	 * @param x
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the character
	 * @param y
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the character
	 */
	public synchronized void drawAndCacheCharAt(
			final Graphics g, 
			final CharRepresentation rep, 
			final double x, 
			final double y) {
		if (null != rep) {
			CharImage img = symbolImages.get(rep);
			if (null == img) {
				img = new CharImage(rep);
		    	symbolImages.put(rep, img);
		    	logger.debug("Created image for " + rep.ch + " in color " + rep.colorMap.name + " and size (" + rep.stretch[X] + "," + rep.stretch[Y] + ")");
			} 
			drawImageAt(g, img.image, x, y);
		}
	}

	public synchronized void cleanSymbolCache() {
		symbolImages.clear();
	}
}
