package de.ped.kitten.gui;

/**
 * Defines all parameters necessary to uniquely represent a symbol to draw in an {@link Atari8bitScreen}.
 * This data structure allows caching of symbol images needed several times and therefore performance optimization.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class CharRepresentation {
	public final ColorMap colorMap;
	public final char ch;
	public final int[] stretch;
	public final TrueTypeMonospaceFont font;

	public CharRepresentation(
			final char ch, 
			final TrueTypeMonospaceFont font, 
			final ColorMap colorMap, 
			final int stretchX, 
			final int stretchY) {
		this.ch = ch;
		this.font = font;
		this.colorMap = colorMap;
		this.stretch = new int[]{ stretchX, stretchY };
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ch;
		result = prime * result
				+ ((colorMap == null) ? 0 : colorMap.hashCode());
		return result;
	}

	/**
	 * Two symbol representations are equal if they have
	 * <ul>
	 *   <li>the same character,</li>
	 *   <li>the same color map (see {@link ColorMap#equals(Object)}),</li>
	 *   <li>the same font (by object identity)</li>
	 *   <li>the same stretch values (by integer comparison)</li>
	 * </ul>
	 * 
	 * @param obj the object to compare with
	 * @return if the two objects are equal
	 */
	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final CharRepresentation other = (CharRepresentation) obj;
		if (ch != other.ch)
			return false;
		if (colorMap == null) {
			if (other.colorMap != null)
				return false;
		} else if (!colorMap.equals(other.colorMap))
			return false;
		if (font == null) {
			if (other.font != null)
				return false;
		} else if (!font.equals(other.font))
			return false;
		if (stretch == null) {
			if (other.stretch != null)
				return false;
		} else if (stretch.length != other.stretch.length) {
			return false;
		} else {
			for (int i = 0; i < stretch.length; i++) {
				if (stretch[i] != other.stretch[i]) {
					return false;
				}
			}
		}
		return true;
	}
}