package de.ped.kitten.gui;

import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.imageio.ImageIO;

import de.ped.kitten.logic.Characters;

/**
 * The data for a game field description in the HTML help.
 * 
 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GameFieldDescription {

	/**
	 * For testing purposes. Calls {@link #createImageFiles()}.
	 * @param args ignored
	 * @throws IOException if files could not be created
	 */
	public static void main(final String[] args) throws IOException {
		createImageFiles();
	}

	/**
	 * Creates all the immage files used for HTML help explaining all the building blocks of a level.
	 * @throws IOException if files could not be created
	 */
	public static void createImageFiles() throws IOException {
		for (GameFieldDescription desc : instances()) {
			desc.createImageFile();
		}
	}

	/**
	 * Implements the singleton pattern for lazy initialization.
	 * @return all the game field descriptions to show in the HTML help, in this order. 
	 */
	public static final GameFieldDescription[] instances() {
		if (null == instances) {
			final int globalStretch = 2;
			final int xs = 2 * globalStretch;
			final int ys = 1 * globalStretch;
			final TrueTypeMonospaceFont modernFont = KittenFontsInitializer.initFonts(new TrueTypeMonospaceFont());
			instances = new GameFieldDescription[] {
					new GameFieldDescription("Game.Field.Tomcat",
							new CharRepresentation(Characters.CAT_IDLE, modernFont, KittenMainCanvas.tomcatColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Cat",
							new CharRepresentation(Characters.CAT_IDLE, modernFont, KittenMainCanvas.catColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Wall",
							new CharRepresentation(Characters.WALL, modernFont, KittenMainCanvas.redColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Block",
							new CharRepresentation(Characters.BLOCK, modernFont, KittenMainCanvas.blockColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Flower",
							new CharRepresentation(Characters.FLOWER, modernFont, KittenMainCanvas.greenColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Hurdle",
							new CharRepresentation(Characters.HURDLE, modernFont, KittenMainCanvas.orangeColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Spring",
							new CharRepresentation(Characters.SPRING, modernFont, KittenMainCanvas.redColorMap, xs, ys)),
					new GameFieldDescription("Game.Field.Microflicro",
							new CharRepresentation(Characters.MICROFLICRO2, modernFont, KittenMainCanvas.greenColorMap, xs, ys)),
			};
		}
		return instances;
	}

	/** All the game field descriptions to show in the HTML help, in this order. */
	private static GameFieldDescription[] instances;

	/** The message key prefix for i18n. */
	public final String messageKeyPrefix;

	/** The file name of the image resource. */
	public final String imageResourceName;

	/** The symbol in the level description. */
	private final CharRepresentation rep;

	/**
	 * Constructor.
	 * @param messageKeyPrefix the message key prefix for i18n
	 * @param rep the character representation of the image
	 */
	private GameFieldDescription(
			final String messageKeyPrefix, 
			final CharRepresentation rep) {
		this.rep = rep;
		this.messageKeyPrefix = messageKeyPrefix;
		this.imageResourceName = messageKeyPrefix + ".png";
	}

	private void createImageFile() throws IOException {
		final CharImage img = new CharImage(rep);
		final BufferedImage bi = img.image;
		final int h = bi.getHeight();
		final int w = bi.getWidth();
		final int border = 1;
		final BufferedImage bid = new BufferedImage(w + 2 * border, h + 2 * border, BufferedImage.TYPE_INT_RGB);
		// Black border
		for (int b = 0; b < border; b++) {
			for (int x = 0; x < w + 2 * border; x++) {
				bid.setRGB(x, b, 0x00000000);
				bid.setRGB(x, h + 2 * border - 1 - b, 0x00000000);
			}
			for (int y = 0; y < h + 2 * border; y++) {
				bid.setRGB(b, y, 0x00000000);
				bid.setRGB(w + 2 * border - 1 - b, y, 0x00000000);
			}
		}
		// Copy content of input image (with offset of border)
		for (int x = 0; x < w; x++) {
			for (int y = 0; y < h; y++) {
				final int rgb = bi.getRGB(x, y);
				final int newRGB = 0x00FFFFFF & rgb;
				bid.setRGB(x + border, y + border, newRGB);
			}
		}
		final File file = new File("src/" + Kitten.RESOURCES_PATH + "/" + imageResourceName);
		final BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(file));
		ImageIO.write(bid, "PNG", os);
		os.close();
	}
}