package de.ped.kitten.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Polygon;
import java.io.IOException;
import java.util.HashMap;

import javax.swing.JPanel;

import de.ped.kitten.logic.Characters;
import de.ped.kitten.logic.Duration;
import de.ped.kitten.logic.Game;
import de.ped.kitten.logic.HTSLogo;
import de.ped.kitten.logic.Level;
import de.ped.kitten.logic.LevelInfo;
import de.ped.kitten.logic.TomcatState;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.gui.ApplicationOrApplet;
import de.ped.tools.log.Logger;

/**
 * The drawing area for the level and all.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class KittenMainCanvas extends JPanel {
	private static final long serialVersionUID = 1L;

    /** How many rows are there before the real level screen starts, to leave space for points, lives and other information */
    private static final int ROWS_OFFSET = 2;

    /**
     * Data structure to keep where to draw what cat in the title screen.
     * 
     * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
     */
	private static class TitleCat {
		public final CharRepresentation rep;
		public final int x;
		public final int y;

		public TitleCat(final CharRepresentation rep, final int x, final int y) {
			this.rep = rep;
			this.x = x;
			this.y = y;
		}
	}

	private final Logger logger = Logger.getLogger(getClass());

	static final ColorMap orangeColorMap = ColorMap.create(Color.orange, "orange");
	static final ColorMap redColorMap = ColorMap.create(Color.red, "red");
	static final ColorMap greenColorMap = ColorMap.create(Color.green, "green");
	static final ColorMap whiteColorMap = ColorMap.create(Color.white, "white");
	static final ColorMap titleColorMap = ColorMap.create(Color.red.brighter(), "bright red");
	static final ColorMap gameOverColorMap = ColorMap.create(Color.black, Color.black, Color.green, "game over");
	static final ColorMap successHeartsColorMap = ColorMap.create(Color.black, Color.black, Color.red, "success");
	static final ColorMap tomcatColorMap = whiteColorMap;
	static final ColorMap catColorMap = orangeColorMap;
	static final ColorMap blockColorMap = orangeColorMap;

	private final ColorMap[] fourColCharColorMaps = new ColorMap[] {
			redColorMap, whiteColorMap, greenColorMap, orangeColorMap,
	};

	private final HashMap<Level.Symbol, CharRepresentation> paintChars = new HashMap<Level.Symbol, CharRepresentation>();

	private CharRepresentation springLoadedRep;
	private CharRepresentation tomcatMeowRep;
	private CharRepresentation tomcatLooksLeftRep;
	private CharRepresentation tomcatLooksRightRep;
	private CharRepresentation tomcatDownLeftRep;
	private CharRepresentation tomcatDownRightRep;
	private CharRepresentation tomcatUpLeftRep;
	private CharRepresentation tomcatUpRightRep;
	private CharRepresentation microflicro1;
	private CharRepresentation microflicro2;
	private CharRepresentation heart;

	private final KittenMainWindow parent;

	private final TrueTypeMonospaceFont classicFont = new TrueTypeMonospaceFont();
	private final TrueTypeMonospaceFont modernFont;
	private final Atari8bitScreen screen = new Atari8bitScreen(Atari8bitScreen.DEFAULT_BORDER, true);
	private Dimension minimumSize = new Dimension();

	/** Defines how many Graphics pixels an Atari pixel has. */
	private int globalStretch = 2;

	public int getMagnification() {
		return globalStretch;
	}

	public void setMagnification(final int magnification) {
		globalStretch = magnification;
		recreateSymbolImages();
	}

	public void onUsingAlternateGraphicsChange() {
		recreateSymbolImages();
	}

	private TrueTypeMonospaceFont getCurrentFont() {
		return (parent.isUsingAlternateGraphics() ? modernFont : classicFont);
	}

	private synchronized void recreateSymbolImages() {
		paintChars.put(Level.Symbol.BLANK, createSymbolRep(' ', redColorMap, 2, 1));
		paintChars.put(Level.Symbol.WALL, createSymbolRep(Characters.WALL, redColorMap, 2, 1));
		paintChars.put(Level.Symbol.BLOCK, createSymbolRep(Characters.BLOCK, blockColorMap, 2, 1));
		paintChars.put(Level.Symbol.HURDLE, createSymbolRep(Characters.HURDLE, orangeColorMap, 2, 1));
		paintChars.put(Level.Symbol.SPRING, createSymbolRep(Characters.SPRING, redColorMap, 2, 1));
		paintChars.put(Level.Symbol.FLOWER, createSymbolRep(Characters.FLOWER, greenColorMap, 2, 1));
		paintChars.put(Level.Symbol.TOMCAT, createSymbolRep(Characters.CAT_IDLE, tomcatColorMap, 2, 1));
		paintChars.put(Level.Symbol.CAT, createSymbolRep(Characters.CAT_IDLE, catColorMap, 2, 1));

		springLoadedRep = createSymbolRep(Characters.SPRING_LOADED, redColorMap, 2, 1);
		tomcatMeowRep = createSymbolRep(Characters.CAT_MEAOW, tomcatColorMap, 2, 1);
		tomcatLooksLeftRep = createSymbolRep(Characters.CAT_LOOKS_LEFT, tomcatColorMap, 2, 1);
		tomcatLooksRightRep = createSymbolRep(Characters.CAT_LOOKS_RIGHT, tomcatColorMap, 2, 1);
		tomcatDownLeftRep = createSymbolRep(Characters.CAT_DOWN_LEFT, tomcatColorMap, 2, 1);
		tomcatDownRightRep = createSymbolRep(Characters.CAT_DOWN_RIGHT, tomcatColorMap, 2, 1);
		tomcatUpLeftRep = createSymbolRep(Characters.CAT_UP_LEFT, tomcatColorMap, 2, 1);
		tomcatUpRightRep = createSymbolRep(Characters.CAT_UP_RIGHT, tomcatColorMap, 2, 1);
		microflicro1 = createSymbolRep(Characters.MICROFLICRO1, greenColorMap, 2, 1);
		microflicro2 = createSymbolRep(Characters.MICROFLICRO2, greenColorMap, 2, 1);
		heart = createSymbolRep(Characters.HEART, successHeartsColorMap, 2, 1);
		screen.cleanSymbolCache();
	}

//	private Image loadImage(String name) {
//		Image ret = null;
//    	URL url = Resources.getURLToResource(name);
//		try {
//		    BufferedInputStream is = new BufferedInputStream(url.openStream());
//		    BufferedImage hurdleImg = ImageIO.read(is);
//		    ret = TransformColorToTransparency(hurdleImg, Color.black, Color.black);
//		} catch (final IOException e) {
//			logger.error("Unable to read image " + name, e);
//		}
//		return ret;
//	}

	public KittenMainCanvas(final KittenMainWindow parent) {
		this.parent = parent;
		setBackground(Color.black);
		minimumSize.setSize(screen.getDimension(1, false));
		modernFont = KittenFontsInitializer.initFonts(classicFont);
		screen.setColorMap(ColorMap.create(Color.green, Color.blue, Color.white, "test"));
		screen.setFont(classicFont);
		recreateSymbolImages();
	}

	@Override
	public Dimension getPreferredSize() {
		return getMinimumSize();
	}

	@Override
	public Dimension getMinimumSize() {
		return minimumSize;
	}

	protected void drawCenteredString(final Graphics g, final String string, final int screenWidth, final double y) {
		final double x = (screenWidth - string.length()) / 2.0;
		screen.drawStringAt(g, string, x, y);
	}

	private void drawTextLine4Col(final Graphics g, final byte[] title, final int y) {
		for (int i = 0; i < title.length; i++) {
			screen.draw4ColCharAt(g, title[i], fourColCharColorMaps, i, y);
		}
	}

	private void paintTitleScreen(final Graphics g, final Game game) {
		try {
			g.setColor(Color.blue.darker().darker());
			g.fillRect(0 + screen.getBorder().left, 15 * 8 * globalStretch + screen.getBorder().top,
					40 * 8 * globalStretch, 5 * 8 * globalStretch);

			screen.setStretch(4 * globalStretch);
			screen.setColorMap(titleColorMap);
			screen.drawStringAt(g, "Kitten", 2, 0.75);

			screen.setStretch(0, 2 * globalStretch);
			screen.setStretch(1, 1 * globalStretch);
			final TitleCat[] TITLE_CATS = new TitleCat[] { new TitleCat(paintChars.get(Level.Symbol.TOMCAT), 9, 12),
					new TitleCat(tomcatLooksLeftRep, 12, 10), new TitleCat(tomcatLooksRightRep, 6, 10),
					new TitleCat(tomcatUpLeftRep, 15, 13), new TitleCat(tomcatDownLeftRep, 16, 9),
					new TitleCat(tomcatUpRightRep, 3, 12), new TitleCat(tomcatDownRightRep, 2, 8),
					// new TitleCat(heart, 4, 8),
			};
			final long step = game.getStepsElapsedInGameState() / Duration.ANIMATION.steps;
			for (int i = 0; i < TITLE_CATS.length; i++) {
				final TitleCat titleCatData = TITLE_CATS[i];
				CharRepresentation rep = titleCatData.rep;
				if ((0 == i) && (null != game) && game.isInMeow()) {
					rep = tomcatMeowRep;
				}
				final CharRepresentation cr = animate(rep, step);
				drawSymbol(g, cr, titleCatData.x, titleCatData.y);
			}
			screen.setStretch(1 * globalStretch);
			screen.setColorMap(whiteColorMap);
			final String copyright = parent.getMessages().getText("Game.Title.Copyright", "by Happy Trash Software");
			final String dedication = parent.getMessages().getText("Game.Title.Dedication",
					"dedicated to our two cats");
			drawCenteredString(g, copyright, 40, 16);
			drawCenteredString(g, dedication, 40, 18);
			// drawCenteredString(g, " SPIEL 1 EDITOR ", 40, 24);
			drawHtsLogo(g, globalStretch);
			final String christmas = parent.getApplicationEnvironment()
					.findParameter(ApplicationOrApplet.PARAM_KEY_CHRISTMAS).getValue();
			if (null != christmas) {
				drawChristmasTree(g);
			}
		} catch (final NullPointerException ex) {
			logger.warn("Not fully initialized?", ex);
		}
	}

	protected void drawChristmasLight(final Graphics g, final int x, final int y) {
		g.setColor(Color.yellow);
		g.fillRect(x, y - 8, 2, 2);
		g.setColor(Color.red);
		g.fillRect(x, y - 6, 2, 6);
	}

	protected void drawChristmasTree(final Graphics g) {
		final Polygon p = new Polygon();
		p.addPoint(50 - 0, 0);
		p.addPoint(50 - 20, 20);
		p.addPoint(50 - 10, 21);
		p.addPoint(50 - 30, 40);
		p.addPoint(50 - 10, 41);
		p.addPoint(50 - 40, 60);
		p.addPoint(50 - 20, 63);
		p.addPoint(50 + 20, 63);
		p.addPoint(50 + 40, 60);
		p.addPoint(50 + 10, 41);
		p.addPoint(50 + 30, 40);
		p.addPoint(50 + 10, 21);
		p.addPoint(50 + 20, 20);
		p.addPoint(50 + 0, 0);
		g.setColor(Color.green.darker());
		g.fillPolygon(p);
		g.setColor(Color.getColor(null, 0x808000));
		g.fillRect(50 - 5, 61, 10, 20);
		drawChristmasLight(g, 50 - 15, 18);
		drawChristmasLight(g, 50 + 23, 35);
		drawChristmasLight(g, 50 - 3, 55);
	}

//    private Image TransformColorToTransparency(BufferedImage image, Color c1, Color c2)
//    {
//      // Primitive test, just an example
//      final int r1 = c1.getRed();
//      final int g1 = c1.getGreen();
//      final int b1 = c1.getBlue();
//      final int r2 = c2.getRed();
//      final int g2 = c2.getGreen();
//      final int b2 = c2.getBlue();
//      ImageFilter filter = new RGBImageFilter()
//      {
//        public final int filterRGB(int x, int y, int rgb)
//        {
//          int r = (rgb & 0xFF0000) >> 16;
//          int g = (rgb & 0xFF00) >> 8;
//          int b = rgb & 0xFF;
//          if (r >= r1 && r <= r2 &&
//              g >= g1 && g <= g2 &&
//              b >= b1 && b <= b2)
//          {
//            // Set fully transparent but keep color
//            return rgb & 0xFFFFFF;
//          }
//          return rgb;
//        }
//      };
//
//      ImageProducer ip = new FilteredImageSource(image.getSource(), filter);
//      return Toolkit.getDefaultToolkit().createImage(ip);
//    }

	private HTSLogo htslogo;

	private Color[] RANDOM_COLORS = new Color[] { Color.blue, Color.cyan, Color.darkGray, Color.gray, Color.lightGray,
			Color.green, Color.magenta, Color.orange, Color.pink, Color.red, Color.white, Color.yellow };

	private Color getRandomColor() {
		return RANDOM_COLORS[MathUtil.random(RANDOM_COLORS.length)];
	}

	private void drawHtsLogo(final Graphics g, final int globalStretch) {
		if (null == htslogo) {
			try {
				htslogo = new HTSLogo(parent.getApplicationEnvironment().resources());
			} catch (final IOException e) {
				logger.warn("", e);
			}
		}
		final int ulx = screen.getDimension(1, false).width - HTSLogo.XSIZE;
		final int uly = 0;
		for (int y = 0; y < HTSLogo.YSIZE; y++) {
			for (int x = 0; x < HTSLogo.XSIZE; x++) {
				if (0 == MathUtil.random(300)) {
					g.setColor(getRandomColor());
				}
				if (htslogo.get(x, y)) {
					screen.drawPixel(g, (ulx + x + 12) * globalStretch, (uly + y) * globalStretch, globalStretch, globalStretch);
				}
			}
		}
	}

	public void paintFont(final Graphics g) {
		for (int i = 0; i < 256; i++) {
			screen.drawCharAt(g, (char) (i), (i % 16), (i / 16));
		}
	}

	private CharRepresentation createSymbolRep(
			final char c, 
			final ColorMap colorMap, 
			final int stretchX, 
			final int stretchY) {
		return new CharRepresentation(c, getCurrentFont(), colorMap, 2 * globalStretch, 1 * globalStretch);
	}

	private CharRepresentation getTomcatChar(final Game game) {
		CharRepresentation rep = null;
		char c = game.getTomcatState().character;
		if (game.isInMeow()) {
			rep = tomcatMeowRep;
		} else {
			rep = createSymbolRep(c, tomcatColorMap, 2, 1);
		}
		return rep;
	}

	private void drawMicroflicros(final Graphics g, final Game game, final Level level) {
		final boolean[] microflicros = level.getMicroflicros();
		final double microflicrosXPos = game.getMicroflicrosXPosition();
		final int microflicrosLoopState = game.getMicroflicrosLoopState();
		final CharRepresentation microflicro = (microflicrosLoopState == 0 ? microflicro1 : microflicro2);
		for(int i = 0; i < microflicros.length; i++) {
			if (microflicros[i]) {
				drawSymbol(g, microflicro, microflicrosXPos, ROWS_OFFSET + i);
			}
		}
	}

	private void paintLevelScreen(final Graphics g, final Game game) {
		final Level level = game.getLevel();
		screen.setStretch(0, 2*globalStretch);
		screen.setStretch(1, 1*globalStretch);
		drawMicroflicros(g, game, level);

		// Draw headings and footer
		final String pointsStr = parent.getMessages().getText("Game.Screen.Points.Text", "POINTS");
		final String flowersStr = parent.getMessages().getText("Game.Screen.Flowers.Text", "FLOWERS");
		final String timeStr = parent.getMessages().getText("Game.Screen.Time.Text", "TIME");
		screen.setColorMap(redColorMap);
		screen.drawStringAt(g, pointsStr, 0, 0);
		screen.setColorMap(greenColorMap);
		screen.drawStringAt(g, flowersStr, 0, 1);
		final CharRepresentation lifeIndicator = paintChars.get(Level.Symbol.TOMCAT);
		for (int i = 1; i < game.getLives(); i++) {
			drawSymbol(g, lifeIndicator, 20 - i, 0);
		}
		screen.setColorMap(orangeColorMap);
		screen.drawStringAt(g, timeStr, 20 - 4 - timeStr.length(), 1);
		screen.setColorMap(redColorMap);
		screen.drawStringAt(g, "" + game.getRemainingLevelTime(), 20 - 3, 1);
		screen.drawStringAt(g, "" + game.getPoints(), pointsStr.length() + 1, 0);
		screen.drawStringAt(g, "" + level.getRemainingFlowers(), flowersStr.length() + 1, 1);
		screen.setColorMap(redColorMap);
		final String screenStr = parent.getMessages().getText("Game.Screen.Screen.Text", "SCREEN");
		drawCenteredString(g, screenStr + " " + (game.getLevelIndex() + 1), 20, ROWS_OFFSET + 20);
		screen.setColorMap(whiteColorMap);
		drawTextLine4Col(g, level.getTitle(), ROWS_OFFSET + 21);

		// Draw level
		final PlayfieldPosition pos = level.getPosition();
		final TomcatState tomcatState = game.getTomcatState();
		for(int y = 0; y < LevelInfo.SIZE.getHeight(); y++) {
			for (int x = 0; x < LevelInfo.SIZE.getWidth(); x++) {
				final Level.Symbol symbol = level.getAt(x, y);
				CharRepresentation rep = paintChars.get(symbol);
				if ((symbol == Level.Symbol.SPRING) && (y == pos.getY()) && (x == pos.getX()) &&
						(tomcatState == TomcatState.SPRING_LEFT0 || tomcatState == TomcatState.SPRING_RIGHT0)) {
					rep = springLoadedRep;
				}
		    	final long step = game.getStepsElapsedInLevel() / Duration.ANIMATION.steps;
				final CharRepresentation cr = animate(rep, step);
				drawSymbol(g, cr, x, ROWS_OFFSET + y);
			}
		}

		// Draw tomcat
		final CharRepresentation rep = getTomcatChar(game);
		final long step = game.getStepsElapsedInTomcatState() / Duration.ANIMATION.steps;
		final CharRepresentation cr = animate(rep, step);
		//screen.setColorMap(whiteColorMap);
		double smoothX = 0;
		double smoothY = 0;
		if (parent.isUsingAlternateGraphics()) {
			final PlayfieldPosition lpos = level.getLastPosition();
			final long smoothSteps = game.getTomcatState().getSmoothSteps();
			final double smooth = MathUtil.bound(smoothSteps - 1 - game.getStepsElapsedInTomcatState(), 0, smoothSteps - 1);
			smoothX = (lpos.getX() - pos.getX()) * smooth / (double) smoothSteps;
			smoothY = (lpos.getY() - pos.getY()) * smooth / (double) smoothSteps;
		}
		drawSymbol(g, cr, pos.getX() + smoothX, ROWS_OFFSET + pos.getY() + smoothY);
		if (tomcatState == TomcatState.SUCCEEDED) {
			screen.setColorMap(successHeartsColorMap);
			drawSymbol(g, heart, pos.getX()    , ROWS_OFFSET + pos.getY() - 1);
			drawSymbol(g, heart, pos.getX() + 1, ROWS_OFFSET + pos.getY() - 1);
		}
		// Draw GameOver
		if (game.getState() == Game.State.GAME_OVER) {
			screen.setColorMap(gameOverColorMap);
			drawCenteredString(g, "           ", 20, 10);
			drawCenteredString(g, " GAME OVER ", 20, 11);
			drawCenteredString(g, "           ", 20, 12);
		}
	}

	private CharRepresentation animate(final CharRepresentation rep, final long step) {
		CharRepresentation cr = rep;
		if (parent.isUsingAlternateGraphics()) {
			char ret = rep.ch;
			char[] phases = getAnimationSequence(rep.ch);
			if (null != phases) {
				ret = phases[(int)MathUtil.modulo((long)step, phases.length)];
			}
			cr = new CharRepresentation(ret, rep.font, rep.colorMap, rep.stretch[0], rep.stretch[1]);
		}
		return cr;
	}

	private char[] getAnimationSequence(final char symbol) {
		char[] phases = null;
		switch(symbol) {
		case Characters.CAT_IDLE:
			phases = Characters.CAT_IDLE_SEQUENCE;
			break;
		case Characters.CAT_LOOKS_LEFT:
			phases = Characters.CAT_LOOKS_LEFT_SEQUENCE;
			break;
		case Characters.CAT_LOOKS_RIGHT:
			phases = Characters.CAT_LOOKS_RIGHT_SEQUENCE;
			break;
		default:
			// Nothing to do
		}
		return phases;
	}

	/** 
	 * Draws a symbol as image. If that image does not exist yet, it is created. 
	 * 
	 * @param g
	 *            the graphics context to draw to
	 * @param rep
	 *            the char representation
	 * @param x
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the character
	 * @param y
	 *            the x position in <em>g</em> where to draw the upper left
	 *            corner of the character
	 */
	private synchronized void drawSymbol(
			final Graphics g, 
			final CharRepresentation rep, 
			final double x, 
			final double y) {
//		if (null != rep) {
//			SymbolsImage img = symbolImages.get(rep);
//			if (null == img) {
//				img = new SymbolsImage(rep);
//		    	symbolImages.put(rep, img);
//		    	logger.debug("Created image for " + rep.ch);
//			} 
//			screen.drawImage(g, img.image, x, y);
//		}
		screen.drawAndCacheCharAt(g, rep, x, y);
	}

	@Override
	public void paintComponent(final Graphics g) {
		super.paintComponent(g);
		if (parent.isUsingAlternateGraphics()) {
			screen.setFont(modernFont);
		} else {
			screen.setFont(classicFont);
		}
		// paintFont(g);
		final Game game = parent.getModel();
		final Dimension s = getSize();
		final Dimension m1 = screen.getDimension(1, false);
		final int stretch = Math.max(1, Math.min(s.width / m1.width, s.height / m1.height));
		logger.trace("Stretch=" + stretch);
		if (globalStretch != stretch) {
			setMagnification(stretch);
		}
		final Dimension mn = screen.getDimension(globalStretch, false);
		final Atari8bitScreen.Border border = new Atari8bitScreen.Border(Math.max(0, (s.height - mn.height) / 2),
				Math.max(0, (s.width - mn.width) / 2));
		screen.setBorder(border);
		switch (game.getState()) {
		case IN_TITLE:
			paintTitleScreen(g, game);
			break;
		case IN_GAME:
		case GAME_OVER:
			paintLevelScreen(g, game);
			break;
		}
	}

	/** To minimize graphics time, repaints only those parts of the screen that change within a game step. */
	public void repaintStep() {
		final Game game = parent.getModel();
		// Repaint Microflicros
		final double mx = game.getMicroflicrosXPosition();
//		if (mx < 2.0) { // Redraw also the right side to clear the microflicros, when they came just in left
//			repaint(globalStretch * 16 * 18 + screen.getBorder().left,
//					globalStretch * 8 * 3 + screen.getBorder().top,
//					globalStretch * 32, 
//					globalStretch * 8 * 18);
//		}
		repaint((int)(globalStretch * 16 * (mx - 1)) + screen.getBorder().left,
				globalStretch * 8 * 3 + screen.getBorder().top,
				globalStretch * 48, 
				globalStretch * 8 * 18);
//		// Repaint Tomcat and surrounding
//		PlayfieldPosition pos = game.getLevel().getPosition();
//		final int tx = (pos.getX() - 1) * 16 + screen.getBorder().left;
//		final int ty = (pos.getY() - 1) * 8 + screen.getBorder().top;
//		repaint(tx, ty, 48, 24);
		// Repaint top of screen with points etc
		repaint(screen.getBorder().left, screen.getBorder().top, 
				globalStretch * 8 * 40, 
				globalStretch * 8 * 2);
//		repaint();
	}
}
