package de.ped.kitten.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.List;
import java.util.Locale;

import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JScrollPane;

import de.ped.kitten.logic.Duration;
import de.ped.kitten.logic.Game;
import de.ped.kitten.logic.GameProperties;
import de.ped.kitten.logic.Game.Event;
import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.FinishableThread;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.CheckBoxMenuItemAction;
import de.ped.tools.gui.GameThread;
import de.ped.tools.gui.PedAction;
import de.ped.tools.gui.UserGameActionInput;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.sound.SoundClip;

/**
 * The main window of the application.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class KittenMainWindow extends ApplicationMainWindow {
	private static final long serialVersionUID = 1L;

	public KittenMainWindow(final ApplicationEnvironment appEnv) throws IOException {
		super(appEnv);
		logger.setLogLevel(Level.INFO);
		sounds = new KittenSounds(getApplicationEnvironment().resources());

		this.model = new Game(getApplicationEnvironment().resources());

		worldCanvas = new KittenMainCanvas(this);
		final JScrollPane worldScrollPane = new JScrollPane(worldCanvas);
		getContentPane().add(worldScrollPane, BorderLayout.CENTER);

		if (getApplicationEnvironment().isSilent()) {
			eventSoundOnOffAction.setOn(false);
			backgroundSoundOnOffAction.setOn(false);
		}
		setUsingAlternateGraphics(alternateGraphicsOnOffAction.isOn());

		setJMenuBar(createMenuBar());
		pack(); // Resize the window to the preferred size of its components.
		final String christmas = Kitten.getInstance().findParameter(Kitten.PARAM_KEY_CHRISTMAS).getValue();
		if (null != christmas) {
			switchLocale(getMessages().getAvailableLocales()[0]);
		} else {
			switchLocale(getMessages().getCurrentLocale());
		}
		addKeyListener(this);
		pauseOrResumeRun(true);
	}

	private PedAction dieAction = new PedAction("Action.Die", KittenMainWindow.this, ResourceFinder.Folder.APP, false);
	private CheckBoxMenuItemAction runAction = new CheckBoxMenuItemAction(false, "Action.Pause", KittenMainWindow.this, ResourceFinder.Folder.COMMON);
	private CheckBoxMenuItemAction eventSoundOnOffAction = new CheckBoxMenuItemAction(true, "Action.SoundOnOff", this, ResourceFinder.Folder.COMMON);
	private CheckBoxMenuItemAction backgroundSoundOnOffAction = new CheckBoxMenuItemAction(true, "Action.BackgroundSoundOnOff", this, ResourceFinder.Folder.APP);
	private CheckBoxMenuItemAction alternateGraphicsOnOffAction = new CheckBoxMenuItemAction(true, "Action.AlternateGraphicsOnOff", this, ResourceFinder.Folder.APP);

	@Override
	public void pedAbstractActionPerformed(final PedAction action, final ActionEvent event) {
		super.pedAbstractActionPerformed(action, event);
		if (action == dieAction) {
			die();
		} else if (action == runAction) {
			togglePause();
		} else if (action == eventSoundOnOffAction) {
			// Nothing to do
		} else if (action == backgroundSoundOnOffAction) {
			// Nothing to do
		} else if (action == alternateGraphicsOnOffAction) {
			setUsingAlternateGraphics(((CheckBoxMenuItemAction) action).isOn());
		}
	}

	private JMenuBar createMenuBar() {
		final JMenuBar menuBar = createJMenuBar();
		final JMenu menuFile = createJMenu("Menu.File");
		createActionItem(menuFile, startNewGameAction);
		createActionItem(menuFile, exitGameAction);
		createActionItem(menuFile, dieAction);
		menuFile.addSeparator();
		menuFile.add(createLocalesMenu());
		menuFile.addSeparator();
		createActionItem(menuFile, quitAction);
		menuBar.add(menuFile);
		final JMenu menuRun = createJMenu("Menu.Run");
		createActionItem(menuRun, alternateGraphicsOnOffAction);
		// createActionItem(menuRun, runAction);
		createActionItem(menuRun, eventSoundOnOffAction);
		createActionItem(menuRun, backgroundSoundOnOffAction);
		menuBar.add(menuRun);
		final JMenu menuHelp = createJMenu("Menu.Help");
		createActionItem(menuHelp, helpAction);
		createActionItem(menuHelp, aboutAction);
		menuBar.add(menuHelp);
		try {
			menuBar.setHelpMenu(menuHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}
		return menuBar;
	}

	Game getModel() {
		return model;
	}

	private NewGameSelectionDialog getOrCreateNewGameSelectionDialog() {
		if (null == newGameSelectionDialog) {
			newGameSelectionDialog = new NewGameSelectionDialog(this);
		}
		return newGameSelectionDialog;
	}

	public UserGameActionInput findUserGameActionInput(final int playerIdx, final int playerAction) {
		return UserGameActionInput.find(userGameActionInputs, playerIdx, playerAction);
	}

	@Override
	public boolean isRunning() {
		return isRunningRun();
	}

	@Override
	public FinishableThread createRunThread() {
		final FinishableThread ret = new GameThread(this);
		ret.setLogger(Logger.getLogger(ret.getClass()));
		return ret;
	}

	@Override
	public void pauseOrResume(final boolean isToResume) {
//    	pauseOrResumeRun(isToResume);
	}

	public void togglePause() {
		pauseOrResume(!isRunning());
		worldCanvas.repaint();
	}

	@Override
	public void startNewGame() {
		final boolean wasRunning = isRunning();
		pauseOrResume(false);
		try {
			Thread.sleep(0);
		} catch (final InterruptedException e) {
			// Nothing to do - that's what we intend to do
		}
		final NewGameSelectionDialog dialog = getOrCreateNewGameSelectionDialog();
		dialog.enterDialog(model.getProperties());
		logger.debug("New game selection dialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (dialog.wasAcceptedDialogChanges()) {
			dialog.readUIFieldsIntoProperties();
			try {
				setModel(dialog.getProperties());
			} catch (final IOException e) {
				logger.error("Unable to read game data!", e);
			}
			repaint();
			pauseOrResume(true);
		} else {
			pauseOrResume(wasRunning);
		}
	}

	private void setModel(final GameProperties properties) throws IOException {
		model.initGame(properties, true);
	}

	@Override
	public void exitGame() {
		try {
			model.initGame(model.getProperties(), false);
		} catch (final IOException e) {
			logger.error("Unable to read game data!", e);
		}
	}

	public void die() {
		model.die();
	}

	/** To minimize graphics time, repaints only those parts of the screen that change within a game step. */
	private void repaintStep() {
		worldCanvas.repaintStep();
	}

	/**
	 * Called by the {@link GameThread} regularly.
	 */
	@Override
	public void step() {
		final boolean wasRepaintedAll = handleGameEvents(model.step());
		if (!wasRepaintedAll) {
			repaintStep();
		}
	}

	@Override
	public synchronized long getSleepDelay() {
		return Duration.GAME_STEP_DELAY;
	}

	@Override
	public void switchLocale(final Locale locale) {
		super.switchLocale(locale);
		userGameActionInputs = new UserGameActionInput[] {
				new UserGameActionInput(getMessages(), "Game.Player0L.KeyStroke",  KeyEvent.VK_LEFT,    0, DigitalJoystickDirection.MASK_LEFT),
				new UserGameActionInput(getMessages(), "Game.Player0R.KeyStroke",  KeyEvent.VK_RIGHT,   0, DigitalJoystickDirection.MASK_RIGHT),
				new UserGameActionInput(getMessages(), "Game.Player0U.KeyStroke",  KeyEvent.VK_UP,      0, DigitalJoystickDirection.MASK_UP),
				new UserGameActionInput(getMessages(), "Game.Player0D.KeyStroke",  KeyEvent.VK_DOWN,    0, DigitalJoystickDirection.MASK_DOWN),
				new UserGameActionInput(getMessages(), "Game.Player0UL.KeyStroke", KeyEvent.VK_NUMPAD7, 0, DigitalJoystickDirection.MASK_UP_LEFT),
				new UserGameActionInput(getMessages(), "Game.Player0UR.KeyStroke", KeyEvent.VK_NUMPAD9, 0, DigitalJoystickDirection.MASK_UP_RIGHT),
				new UserGameActionInput(getMessages(), "Game.Player0DL.KeyStroke", KeyEvent.VK_NUMPAD1, 0, DigitalJoystickDirection.MASK_DOWN_LEFT),
				new UserGameActionInput(getMessages(), "Game.Player0DR.KeyStroke", KeyEvent.VK_NUMPAD3, 0, DigitalJoystickDirection.MASK_DOWN_RIGHT),
		};
		newGameSelectionDialog = null;
		repaint();
	}

	@Override
	public void keyPressed(final KeyEvent keyEvent) {
		final Game game = getModel();
		boolean isToRepaint = false;
		for (final UserGameActionInput tsv : userGameActionInputs) {
			if (keyEvent.getKeyCode() == tsv.keyCode) {
				handleGameEvents(game.suggestUserGameAction(tsv.playerIdx, tsv.playerAction));
				isToRepaint = true;
			}
		}
		if (isToRepaint) {
			repaintStep();
		}
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		final Messages messages = getMessages();
		doc.setTitle(messages.getString("HelpWindow.Title"));
		doc.setHeading("Help");
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");
		buildMenuDescriptions(doc, getJMenuBar());
		buildPropertyDescriptions(doc, getOrCreateNewGameSelectionDialog());
		buildInputDescriptions(doc, userGameActionInputs, false);
		doc.addContentPart(2, "Help.ScreenContent");
		buildPropertyDescriptions(doc, 3, "Help.ScreenContent.AboveLevel",
				new String[] { "Game.Screen.Points", "Game.Screen.Lifes", "Game.Screen.Flowers", "Game.Screen.Time" },
				false);
		buildPropertyDescriptions(doc, 3, "Help.ScreenContent.BelowLevel", new String[] { "Game.Screen.Screen" },
				false);
		doc.addContentPart(3, "Help.ScreenContent.WithinLevel");
		for (GameFieldDescription field : GameFieldDescription.instances()) {
			doc.addContentPart(4, field.messageKeyPrefix, true, messages.getText(field.messageKeyPrefix),
					field.imageResourceName);
		}
		doc.endDocument();
	}

	@Override
	public void buildHtmlHelpResources() {
		try {
			GameFieldDescription.createImageFiles();
		} catch (IOException e) {
			logger.error(e.getMessage(), e);
		}
	}

	@Override
	protected String getKeyboardActionDescriptionText(final UserGameActionInput ugai) {
		return getMessages().getString(DigitalJoystickDirection.getDirection(ugai.playerAction).key);
	}

	private int backgroundSoundIndex = 0;

	private boolean handleGameEvents(final List<Game.Event> events) {
		boolean wasRepaintedAll = false;
		if (null != events) {
			for (final Game.Event event : events) {
				wasRepaintedAll = wasRepaintedAll || handleGameEvent(event);
			}
		}
		return wasRepaintedAll;
	}

	private boolean handleGameEvent(final Game.Event event) {
		boolean wasRepaintedAll = false;
		if (event == Event.BACKGROUND) {
			if (backgroundSoundOnOffAction.isOn()) {
				final SoundClip sound = sounds.findSound(sounds.beatToSoundKey(backgroundSoundIndex));
				currentClips.play(sound, "background" + backgroundSoundIndex);
			}
			backgroundSoundIndex++;
		} else {
			if (event == Game.Event.START_LEVEL) {
				backgroundSoundIndex = 0;
				worldCanvas.repaint();
				wasRepaintedAll = true;
			} else if (event == Game.Event.START_GAME) {
				exitGameAction.setEnabled(true);
				dieAction.setEnabled(true);
				worldCanvas.repaint();
				wasRepaintedAll = true;
			} else if (event == Game.Event.START_TITLE) {
				exitGameAction.setEnabled(false);
				dieAction.setEnabled(false);
				worldCanvas.repaint();
				wasRepaintedAll = true;
			}
			playSound(event);
		}
		return wasRepaintedAll;
	}

	private void playSound(final Game.Event event) {
		if (eventSoundOnOffAction.isOn()) {
			int y = 0;
			if (null != model.getLevel()) {
				y = model.getLevel().getPosition().getY();
			}
			final SoundClip sound = sounds.findSound(new KittenSounds.SoundKey(event, y));
			logger.debug("findSound " + event + " " + y);
			currentClips.play(sound, event.toString());
		}
	}

	public boolean isUsingAlternateGraphics() {
		return alternateGraphicsOnOffAction.isOn();
	}

	private void setUsingAlternateGraphics(final boolean isUsingAlternateGraphics) {
		alternateGraphicsOnOffAction.setOn(isUsingAlternateGraphics);
		worldCanvas.onUsingAlternateGraphicsChange();
	}

	private Logger logger = Logger.getLogger(getClass());

	private final KittenSounds sounds;

	private final SoundClip.RingBuffer currentClips = new SoundClip.RingBuffer(5);

	private final Game model;

	private KittenMainCanvas worldCanvas;

	private UserGameActionInput[] userGameActionInputs;

	private NewGameSelectionDialog newGameSelectionDialog;
}
