package de.ped.kitten.gui;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

import javax.sound.sampled.AudioInputStream;

import de.ped.kitten.logic.Duration;
import de.ped.kitten.logic.Game.Event;
import de.ped.kitten.logic.LevelInfo;
import de.ped.tools.MathUtil;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationSounds;
import de.ped.tools.sound.ConstantWave;
import de.ped.tools.sound.RepeatingWave;
import de.ped.tools.sound.AddedWave;
import de.ped.tools.sound.AmplitudeModulationWave;
import de.ped.tools.sound.ConcatWave;
import de.ped.tools.sound.EnvelopeWave;
import de.ped.tools.sound.MonoWave;
import de.ped.tools.sound.MultipliedWave;
import de.ped.tools.sound.Note;
import de.ped.tools.sound.SawDownWave;
import de.ped.tools.sound.SawUpWave;
import de.ped.tools.sound.SinusWave;
import de.ped.tools.sound.SoundClip;
import de.ped.tools.sound.DynamicSoundMonoWaveCreator;
import de.ped.tools.sound.DynamicSoundMonoWaveCreator.Sound;
import de.ped.tools.sound.SquareWave;
import de.ped.tools.sound.TriangleWave;
import de.ped.tools.sound.Tuning;
import de.ped.tools.sound.WhiteNoiseWave;

/**
 * Bundles all the sounds for Kitten.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class KittenSounds extends ApplicationSounds {
	/**
	 * Defines what sound to return.
	 * The following values have the following meanings:
	 * <ul>
	 *   <li>if <em>event</em> is <code>null</code>, <em>y</em> contains the note ID of the background melody.
	 *   <li>if <em>event</em> is {@link Event#JUMP} or {@link Event#FALL}, <em>y</em> contains the height of our tomcat.
	 *   <li>if <em>event</em> is some other game event, <em>y</em> is ignored. 
	 *   	The {@link Event#BACKGROUND} event creates the whole background melody.
	 *   	This is not needed for the Java game, but for the JavaScript implementation.
	 * </ul>
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class SoundKey implements Comparable<SoundKey> {
		/** 
		 * Constructor for game events
		 * @param event the game event
		 * @param y the height of the tomcat
		 */
		public SoundKey(final Event event, final int y) {
			setNormed(event, y);
		}

		public Event getEvent() {
			return event;
		}

		public int getY() {
			return y;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((event == null) ? 0 : event.hashCode());
			result = prime * result + y;
			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			return (0 == this.compareTo((SoundKey) obj));
		}

		@Override
		public int compareTo(final SoundKey other) {
			int ret = 0;
			if (null == this.event) {
				if (null == other.event) {
					// Two nulls are equal, so nothing to do
				} else {
					ret = -1;
				}
			} else {
				if (null == other.event) {
					ret = 1;
				} else {
					ret = this.event.compareTo(other.event);
				}
			}
			if (0 != ret) {
				return ret;
			}
			ret = this.y - other.y;
			if (0 != ret) {
				return ret;
			}
			return 0;
		}
		
		private void setNormed(final Event event, final int y) {
			if (null == event) {
				this.event = event;
				this.y = y;
			} else {
				switch(event) {
				case JUMP:
				case FALL:
					this.event = Event.JUMP;
					this.y = y;
					break;
				case START_TITLE:
				case DIED:
					this.event = Event.MEOW;
					this.y = 0;
					break;
				default:
					this.event = event;
					this.y = 0;
				}
			}
		}
		
		private Event event;
		private int y;
	}

	/**
	 * Constructor for background sound
	 * @param beat the heartbeat number of the background sound (counted up)
	 * @return the sound key for the sound to play, or <code>null</code> if there is a pause to play
	 */
	public SoundKey beatToSoundKey(final int beat) {
		final SoundKey ret;
		final int noteIndex = MathUtil.modulo(beat, backgroundMelody.length);
		final String tone = backgroundMelody[noteIndex];
		if (null != tone) {
			final Note note = Note.get(tone);
			ret = new SoundKey(null, note.getID());
		} else {
			ret = null;
		}
		return ret;
	}

	public KittenSounds(final ResourceFinder resources) {
		super(resources);
	}

	/**
	 * Returns the sound clip for the given key.
	 * @param key the key (e.g., event) to find the sound for. Accepts {@link SoundKey} objects.
	 * @return the sound clip for the given key, or <code>null</code> if there is no such sound.
	 */
	@Override
	public SoundClip findSound(final Object key) {
		init();
		final SoundClip ret;
		if (key instanceof SoundKey) {
			ret = sounds.get((SoundKey) key);
		} else {
			ret = null;
		}
		return ret;
	}

	@Override
	public void init() {
		if (null == sounds) {
			final HashMap<SoundKey, SoundClip> buf = new HashMap<SoundKey, SoundClip>();
			for (Event event : Event.values()) {
				SoundClip clip = null;
				final SoundKey key = new SoundKey(event, 0);
				switch (event) {
				case MEOW:
				case START_TITLE:
				case LEVEL_DONE:
				case DIED:
				case GAME_OVER:
				case SPRING:
				case COLLISION:
				case FALL:
				case FLOWER:
					clip = new SoundClip(createSound(key));
				default:
					// Nothing to do
				}
				if (null != clip) {
					buf.put(key, clip);
				}
			}
			for (int y = 0; y < LevelInfo.SIZE.getHeight(); y++) {
				final SoundKey key = new SoundKey(Event.JUMP, y);
				final SoundClip clip = new SoundClip(createSound(key));
				clip.setConcurrencyMode(SoundClip.ConcurrencyMode.ALL);
				buf.put(key, clip);
			}
			for (final String tone : backgroundMelody) {
				if (null != tone) {
					final SoundKey key = new SoundKey(null, Note.get(tone).getID());
					if (null == buf.get(key)) {
						buf.put(key, new SoundClip(createSound(key)));
					}
				}
			}
			sounds = buf;
		}
	}

	@Override
	public AudioInputStream createSound(final Object key) {
		AudioInputStream ret = null;
		DynamicSoundMonoWaveCreator.Sound sound = null;
		final DynamicSoundMonoWaveCreator smwc = new DynamicSoundMonoWaveCreator();
		final SoundKey soundKey = (SoundKey) key;
		final Event event = soundKey.getEvent();
		if (null == event) { // background melody
			sound = createBackgroundSound(smwc, Note.get(soundKey.getY()));
		} else { // event sound
			switch (event) {
			case MEOW:
				ret = read("cat13.wav");
				break;
			case LEVEL_DONE:
				sound = createSuccessSound(smwc);
				break;
			case GAME_OVER:
				sound = createGameOverSound(smwc);
				break;
			case SPRING:
				sound = createSpringSound(smwc);
				break;
			case COLLISION:
				sound = createCollisionSound(smwc);
				break;
			case FLOWER:
				sound = createCollectFlowerSound(smwc);
				break;
			case JUMP:
				sound = createJumpSound(smwc, soundKey.getY());
				break;
			case BACKGROUND:
				sound = createBackgroundMelody(smwc);
				break;
			default:
				// Nothing to do
			}
		}
		if ((null == ret) && (null != sound)) {
			ret = smwc.toAudioInputStream(sound);
		}
		return ret;
	}

	private static DynamicSoundMonoWaveCreator.Sound createJumpSound(final DynamicSoundMonoWaveCreator smwc, final int y) {
		final double duration = Duration.JUMP_PART.getDurationInSeconds() * 1.2;
		final double loudness = 0.35 + 0.05 * y;
		final MonoWave freq = new TriangleWave(smwc,
				Tuning.defaultTuning().getFrequency(Note.get(Note.get("C2").getID() - y)));
		final MonoWave hull = new EnvelopeWave(new EnvelopeWave.Parameters(smwc,
				new double[][] { { 0.05, 0.0, }, { duration - 0.1, loudness, }, { 0.05, loudness, } }));
		final MonoWave allWave = new MultipliedWave(smwc, freq, hull);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, duration);
	}

	private static Sound createSpringSound(final DynamicSoundMonoWaveCreator smwc) {
		final double duration = 0.4;
		final double loudness = 0.8;
		final MonoWave freqfmSawDn = new SawUpWave(new RepeatingWave.Params(smwc, 1 / duration, null, 2.0));
		final MonoWave freqfmfm = new SawDownWave(new RepeatingWave.Params(smwc, 1 / duration, null, 5.0));
		final MonoWave freqfmWoop = new SinusWave(new RepeatingWave.Params(smwc, 0.0, 0.75, 7.0, freqfmfm, new ConstantWave(5.0)));
		final MonoWave freqfm = new AddedWave(smwc, freqfmSawDn, 1.0, freqfmWoop, 1.0);
		final MonoWave freq = new SquareWave(smwc, Tuning.defaultTuning().getFrequency(Note.get("A1")), freqfm);
		final MonoWave hull = new EnvelopeWave(new EnvelopeWave.Parameters(smwc, new double[][]{ { 0.05, 0.0, }, { duration - 0.1, loudness, }, { 0.05, 0.0, } }));
		final MonoWave allWave = new MultipliedWave(smwc, freq, hull);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createCollectFlowerSound(final DynamicSoundMonoWaveCreator smwc) {
		final Note[] notes = new Note[]{ Note.get("F2"), Note.get("A2"), Note.get("C3"), };
		final double noteDistance = 0.075;
		final double noteDuration = 0.25;
		final double duration = (notes.length - 1) * noteDistance + noteDuration;
		final AddedWave.Part[] parts = new AddedWave.Part[notes.length];
		for (int i = 0; i < parts.length; i++) {
			final MonoWave base = new SinusWave(smwc, Tuning.defaultTuning().getFrequency(notes[i]));
			final MonoWave hull = new EnvelopeWave(smwc, noteDistance * i, noteDuration, 2.5, 0.0);
			final MonoWave allWave = new MultipliedWave(smwc, base, hull);
			parts[i] = new AddedWave.Part(allWave, noteDuration);
		}
		final MonoWave allWave = new AddedWave(smwc, Arrays.asList(parts));
		return new DynamicSoundMonoWaveCreator.Sound(allWave, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createCollisionSound(final DynamicSoundMonoWaveCreator smwc) {
		final double duration = 0.1;
		final MonoWave noiseFM = new SquareWave(smwc, Tuning.defaultTuning().getFrequency(Note.get("C-2")));
		final MonoWave noise = new WhiteNoiseWave(smwc);
		final MonoWave added = new AddedWave(smwc, noiseFM, 1.0, noise, 0.2);
		final MonoWave hull = new EnvelopeWave(smwc, duration, 1.0, 0.0);
		final MonoWave allWave = new MultipliedWave(smwc, added, hull);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createSuccessSound(final DynamicSoundMonoWaveCreator smwc) {
		final double noteDuration = 0.4;
		final Note[][] tones = new Note[][]{ 
				{ Note.get("G1"), Note.get("E1"), Note.get("C1"), },
				{ Note.get("G1"), Note.get("E1"), Note.get("C1"), },
				{ Note.get("F1"), Note.get("D1"), Note.get("F0"), },
				{ Note.get("F1"), Note.get("D1"), Note.get("F0"), },
				{ Note.get("E1"), Note.get("C1"), Note.get("G0"), },
				{ Note.get("D1"), Note.get("B0"), Note.get("G0"), },
				{ Note.get("C1"), Note.get("G0"), Note.get("C0"), },
		};
		final ConcatWave.Part[] waves = new ConcatWave.Part[tones.length];
		
		final MonoWave hullWave = new SawDownWave(smwc, 1.0 / noteDuration);
		final Tuning tuning = Tuning.defaultTuning();
		for (int i = 0; i < tones.length; i++) {
			final MonoWave tone0 = new TriangleWave(smwc, tuning.getFrequency(tones[i][0]));
			final MonoWave tone1 = new SquareWave(smwc, tuning.getFrequency(tones[i][1]));
			final MonoWave tone2 = new SawDownWave(smwc, tuning.getFrequency(tones[i][2]));
			final AddedWave.Part[] parts = new AddedWave.Part[]{
					new AddedWave.Part(tone0, 0.5),
					new AddedWave.Part(tone1, 0.25),
					new AddedWave.Part(tone2, 0.25),
			};
			final MonoWave added    = new AddedWave(smwc, Arrays.asList(parts));
			final MonoWave hSinusWave   = new AmplitudeModulationWave(smwc, added, 0.8, hullWave);
			waves[i] = new ConcatWave.Part(hSinusWave, noteDuration);
		}
		final MonoWave allWave      = new ConcatWave(smwc, waves);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, noteDuration * tones.length);
	}

	private static DynamicSoundMonoWaveCreator.Sound createGameOverSound(final DynamicSoundMonoWaveCreator smwc) {
		final double noteDuration = 0.4;
		final Note[] bass = new Note[]{ Note.get("A-1"), Note.get("F-1"), Note.get("D-1"), Note.get("A-2"), };
		final Tuning tuning = Tuning.defaultTuning();
		final double[] melody = new double[]{ 
				tuning.getFrequency(Note.get("C1")), 
				tuning.getFrequency(Note.get("A0")), 
				tuning.getFrequency(Note.get("F1")),
				tuning.getFrequency(Note.get("E1")),
				tuning.getFrequency(Note.get("C1")), 
				tuning.getFrequency(Note.get("A0")), 
				tuning.getFrequency(Note.get("F1")),
				tuning.getFrequency(Note.get("E1")),
				tuning.getFrequency(Note.get("C1")), 
				tuning.getFrequency(Note.get("A0")), 
				tuning.getFrequency(Note.get("F1")),
				tuning.getFrequency(Note.get("E1")),
				tuning.getFrequency(Note.get("A0")), 
				tuning.getFrequency(Note.get("A0")) * 0.99, 
				0.0,
				0.0,
		};
		final double bassHullDur1 = noteDuration * 0.3;
		final double bassHullDur2 = noteDuration * 0.05;
		final double bassHullDur3 = noteDuration * 0.6;
		final double bassHullDur4 = noteDuration * 0.05;
		final double[][] bassNoteHullParts = new double[][] { // A complicated way to get a saw in a linear approximation.
				{ bassHullDur1, 8.0/16.0, },
				{ bassHullDur2, 8.0/16.0, },
				{ bassHullDur3, 5.0/16.0, },
				{ bassHullDur4, 5.0/16.0, },
				{ bassHullDur1, 7.0/16.0, },
				{ bassHullDur2, 7.0/16.0, },
				{ bassHullDur3, 4.0/16.0, },
				{ bassHullDur4, 4.0/16.0, },
				{ bassHullDur1, 6.0/16.0, },
				{ bassHullDur2, 6.0/16.0, },
				{ bassHullDur3, 3.0/16.0, },
				{ bassHullDur4, 3.0/16.0, },
				{ bassHullDur1, 5.0/16.0, },
				{ bassHullDur2, 5.0/16.0, },
				{ bassHullDur3, 2.0/16.0, },
				{ bassHullDur4, 2.0/16.0, },
		};
		final ConcatWave.Part[] bassline = new ConcatWave.Part[bass.length];
		for (int i = 0; i < bassline.length; i++) {
			final EnvelopeWave.Parameters params = new EnvelopeWave.Parameters(smwc, noteDuration * 4 * i, bassNoteHullParts); 
			final MonoWave bassHullWave = new EnvelopeWave(params); 
			final MonoWave bassNoteWave = new SawDownWave(smwc, tuning.getFrequency(bass[i]));
			final MonoWave bassHNoteWave = new MultipliedWave(smwc, bassNoteWave, bassHullWave) ;
			bassline[i] = new ConcatWave.Part(bassHNoteWave, noteDuration * 4);
		}
		final MonoWave bassWave = new ConcatWave(smwc, bassline);
		final ConcatWave.Part[] melodyline = new ConcatWave.Part[melody.length];
		for (int i = 0; i < melodyline.length; i++) {
			final double echoFreq = (i == 0 ? 0.0 : melody[i - 1]);
			final MonoWave tone0    = new TriangleWave(smwc, melody[i]);
			final MonoWave tone1    = new TriangleWave(smwc, echoFreq); 
			final AddedWave.Part[] parts = new AddedWave.Part[]{
					new AddedWave.Part(tone0, 0.6),
					new AddedWave.Part(tone1, 0.3),
			};
			final MonoWave added    = new AddedWave(smwc, Arrays.asList(parts));
			melodyline[i] = new ConcatWave.Part(added, noteDuration);
		}
		final MonoWave melodyWave = new ConcatWave(smwc, melodyline);
		final MonoWave allWave      = new AddedWave(smwc, bassWave, 0.9, melodyWave, 1.0);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, noteDuration * melody.length);
	}

	private final static double BACKGROUND_NOTE_DURATION = 0.1;

	private DynamicSoundMonoWaveCreator.Sound createBackgroundMelody(final DynamicSoundMonoWaveCreator smwc) {
		final double melodyDuration = BACKGROUND_NOTE_DURATION * backgroundMelody.length;
		final ArrayList<AddedWave.Part> list = new ArrayList<AddedWave.Part>();
		for (int i = 0; i < backgroundMelody.length; i++) {
			final SoundKey key = beatToSoundKey(i);
			if (null != key) {
				final double offset = BACKGROUND_NOTE_DURATION * i;
				final Note note = Note.get(key.getY());
		    	final MonoWave noteWave = createBackgroundSoundWave(smwc, offset, note);
		        final AddedWave.Part part = new AddedWave.Part(noteWave);
		        list.add(part);
			}
		}
		final MonoWave allWave = new AddedWave(smwc, list);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, melodyDuration); 
	}

	private static DynamicSoundMonoWaveCreator.Sound createBackgroundSound(final DynamicSoundMonoWaveCreator smwc, final Note note) {
		final MonoWave noteWave = createBackgroundSoundWave(smwc, 0.0, note);
		return new DynamicSoundMonoWaveCreator.Sound(noteWave, BACKGROUND_NOTE_DURATION);
	}

	private static MonoWave createBackgroundSoundWave(final DynamicSoundMonoWaveCreator smwc, final double offset, final Note note) {
		final double freq = (note == null ? 0.0 : Tuning.defaultTuning().getFrequency(note));
		final RepeatingWave.Params twp = new RepeatingWave.Params(smwc, offset, 0, freq, null, null);
		final MonoWave tone = new TriangleWave(twp); 
		final EnvelopeWave.Parameters evp = new EnvelopeWave.Parameters(smwc, offset, BACKGROUND_NOTE_DURATION * 0.1, BACKGROUND_NOTE_DURATION * 0.1, BACKGROUND_NOTE_DURATION * 0.7, BACKGROUND_NOTE_DURATION * 0.1, 0.0, 0.3, 0.3);
		final MonoWave hull = new EnvelopeWave(evp);
		final MonoWave noteWave = new MultipliedWave(smwc, tone, hull);
		return noteWave;
	}

	private final String[] backgroundMelody = new String[]{ 
			"D1", "C1", "F0", null, "F1", null, "F1", null, 
			"D1", "C1", "F0", null, "F1", null, "F1", null, 
			"D1", "C1", "F0", null, "F1", null, "D0", null, 
			"F1", null, "C0", null, "E1", null, "E1", null, 

			"D1", "C1", "C0", null, "E1", null, "E1", null, 
			"D1", "C1", "C0", null, "E1", null, "E1", null, 
			"D1", "C1", "C0", null, "E1", null, "D0", null, 
			"E1", null, "F0", null, "F1", null, "F1", null, 

			"D1", "C1", "A1", null, "F1", null, "F1", null, 
			"D1", "C1", "A1", null, "F1", null, "F1", null, 
			"D1", "C1", "A1", null, "F1", null, "C2", null, 
			"F1", null, "D2", null, "E1", null, "E1", null, 

			"D1", "C1", "D2", null, "E1", null, "E1", null, 
			"D1", "C1", "D2", null, "E1", null, "E1", null, 
			"D1", "C1", "D2", null, "E1", null, "C2", null, 
			"E1", null, "A1", null, "F1", null, "F1", null, 
	};

	public final int BACKGROUND_BEATS = backgroundMelody.length;

	private HashMap<SoundKey, SoundClip> sounds;
}
