package de.ped.kitten.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.IOException;
import java.util.Vector;

import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.ListCellRenderer;

import de.ped.kitten.logic.GameProperties;
import de.ped.kitten.logic.LevelInfo;
import de.ped.kitten.logic.KittenLevelSet;
import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.MathUtil;
import de.ped.tools.Messages;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.JIconComboBox;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * The dialog to start a new game.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class NewGameSelectionDialog extends PedJDialog implements ItemListener {
	private static final long serialVersionUID = 1L;

	/**
	 * Keeps all information displayed for an entry in the levelset combobox.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class LevelSetComboBoxItemModel {
		public static final String LEVELSET_KEY_PREFIX = "GameProperties.LevelSet.";

		private final int index;
		private final boolean isAvailable;
		private final Messages messages;

		LevelSetComboBoxItemModel(final int index, final boolean isAvailable, final Messages messages) {
			this.index = index;
			this.isAvailable = isAvailable;
			this.messages = messages;
		}

		@Override
		public String toString() {
			return messages.getText(LEVELSET_KEY_PREFIX + index);
		}

		public String getToolTipText() {
			return messages.getShortDescription(LEVELSET_KEY_PREFIX + index);
		}
	}

	/**
	 * Allows to disable entries in the levelset combobox if the resources for the levelset are not available.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class LevelSetComboBoxItemRenderer extends JLabel implements ListCellRenderer<Object> {
		private static final long serialVersionUID = 1L;

		@Override
		public Component getListCellRendererComponent(
				final JList<?> list,
				final Object value,  
				final int index,
				final boolean isSelected,
				final boolean cellHasFocus) {
			setText(value.toString());
			LevelSetComboBoxItemModel model = (LevelSetComboBoxItemModel) value;
			setToolTipText(model.getToolTipText());
			setFocusable(model.isAvailable);
			setEnabled(model.isAvailable);
			if (isSelected) {
				setBackground(list.getSelectionBackground());
				setForeground(list.getSelectionForeground());
			} else {
				setBackground(list.getBackground());
				setForeground(list.getForeground());
			}
			return this;
		}
	}

	/**
	 * Draws the level as thumbnail into an icon.
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
//	private static class LevelIcon implements Icon {
//	    private int width = LevelInfo.XSIZE;
//	    private int height = LevelInfo.YSIZE;
//
//	    @Override
//	    public void paintIcon(Component c, Graphics g, int x, int y) {
//	        g.setColor(Color.WHITE);
//	        g.fillRect(x +1 ,y + 1,width -2 ,height -2);
//
//	        g.setColor(Color.BLACK);
//	        g.drawRect(x +1 ,y + 1,width -2 ,height -2);
//
//	        g.setColor(Color.RED);
//
//	        g.drawLine(x +10, y + 10, x + width -10, y + height -10);
//	        g.drawLine(x +10, y + height -10, x + width -10, y + 10);
//
//	        g.dispose();
//	    }
//
//	    @Override
//	    public int getIconWidth() {
//	        return width;
//	    }
//
//	    @Override
//	    public int getIconHeight() {
//	        return height;
//	    }
//	}

	/**
	 * Keeps all information displayed for an entry in the levelset combobox.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	static class LevelComboBoxItemModel {
		private final int index;
		public final String title;
//		public final Icon icon;
		
		LevelComboBoxItemModel(final int index, final LevelInfo levelInfo) {
			this.index = index;
//			this.icon = new LevelIcon();
			this.title = levelInfo.getTitleAsString();
		}
		
		@Override
		public String toString() {
			return Integer.toString(index + 1) + " " + title;
		}
	}

	/**
	 * Draws a level combobox item.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	static class LevelComboBoxItemRenderer extends JLabel implements ListCellRenderer<Object> {
		private static final long serialVersionUID = 1L;
		
//		private static final Dimension preferredSize = new Dimension(150, LevelInfo.YSIZE + 50);

		public LevelComboBoxItemRenderer() {
			setOpaque(true);
			setVerticalAlignment(CENTER);
		}
		
//		@Override
//		public Dimension getPreferredSize() {
//			return preferredSize;
//		}

		/*
		 * This method finds the image and text corresponding to the selected
		 * value and returns the label, set up to display the text and image.
		 */
		@Override
		public Component getListCellRendererComponent(
				final JList<?> list, 
				final Object value,
				final int index, 
				final boolean isSelected, 
				final boolean cellHasFocus) {
			LevelComboBoxItemModel item = (LevelComboBoxItemModel)value;
			if (isSelected) {
				setBackground(list.getSelectionBackground());
				setForeground(list.getSelectionForeground());
			} else {
				setBackground(list.getBackground());
				setForeground(list.getForeground());
			}
			setText(item.toString());
//			setIcon(item.icon);
			return this;
		}
	}

	public NewGameSelectionDialog(final ApplicationMainWindow parent) {
		super(parent, parent);
		final Messages messages = getMessages();
		setTitle(messages.getString("GamePropertiesWindow.Title"));
		logger.setLogLevel(Level.INFO);
		choiceLevelSet = new JIconComboBox<LevelSetComboBoxItemModel>(createLevelSetComboBoxItems(parent.getMessages()));
		choiceLevel = new JIconComboBox<LevelComboBoxItemModel>();
		setPropertyElements(new PropertyElement[] {
				new PropertyElement("GameProperties.LevelSet", choiceLevelSet),
				new PropertyElement("GameProperties.Level", choiceLevel),
		});
		setResizable(false);
		setTitlePanel(messages.getString("GamePropertiesWindow.Heading"));
		JPanel mainPanel = createMainPanel();
		GridBagLayout gridbag = new GridBagLayout();
		GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);
		addAllPropertyElementsTo(mainPanel, c);

		add(mainPanel, BorderLayout.CENTER);
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		addHelpButton("Help.Properties");
		this.setDefaultButton(getButtonOk());
		finishLayout();

		choiceLevelSet.setRenderer(new LevelSetComboBoxItemRenderer());
		choiceLevelSet.addItemListener(this);
		choiceLevel.setRenderer(new LevelComboBoxItemRenderer());
	}

	protected int levelToIndex(final int level) {
		return MathUtil.bound(level, 0, 1);
	}

	protected int indexToLevel(final int index) {
		return index;
	}

	@Override
	public void updateUIFieldsBasedOnProperties() {
		readPredefinedLevelSetParameter();
		final int levelSetIndex = gameProperties.getLevelSetIndex();
		choiceLevelSet.setSelectedIndex(levelSetIndex);
		logger.debug("Called updateChoiceLevelOnLevelSet " + levelSetIndex);
		choiceLevelSet.setSelectedIndex(levelSetIndex);
		Vector<LevelComboBoxItemModel> list = createLevelComboBoxItems(levelSetIndex);
		choiceLevel.removeAllItems();
		for (final LevelComboBoxItemModel item : list) {
			choiceLevel.addItem(item);
		}
		final int level = MathUtil.bound(gameProperties.getLevel(), 0, list.size() - 1);
		choiceLevel.setSelectedIndex(level);
		if (gameProperties.isSingleLevelGame()) {
			choiceLevel.setEnabled(false);
			choiceLevelSet.setEnabled(false);
		}
	}

	@Override
	public void readUIFieldsIntoProperties() {
		gameProperties.setLevelSetIndex(choiceLevelSet.getSelectedIndex());
		gameProperties.setLevel(indexToLevel(choiceLevel.getSelectedIndex()));
	}

	private void readPredefinedLevelSetParameter() {
		final String predefLevel = Kitten.getInstance().findParameter(ApplicationEnvironment.PARAM_KEY_GAME).getValue();
		if (null != predefLevel) {
			try {
				final String[] parts = predefLevel.split("\\.");
				int predefLevelSetIndex = 1;
				int predefLevelIndex;
				if (0 < parts.length) {
					if (1 == parts.length) {
						predefLevelIndex = Integer.parseInt(parts[0]);
					} else {
						predefLevelSetIndex = Integer.parseInt(parts[0]);
						predefLevelIndex = Integer.parseInt(parts[1]);
					}
					gameProperties.setLevelSetIndex(predefLevelSetIndex);
					gameProperties.setLevel(predefLevelIndex);
					gameProperties.setSingleLevelGame(true);
				} else {
					logger.debug("Did not understand level parameter " + predefLevel);
				}
			} catch (final Throwable ex) {
				logger.info("Could not read parameter " + Kitten.PARAM_KEY_GAME, ex);
			}
		}
	}

	public GameProperties getProperties() {
		return gameProperties;
	}

	public void enterDialog(final GameProperties props) {
		gameProperties = (GameProperties) props.clone();
		enterDialog();
	}

	@Override
	public void itemStateChanged(final ItemEvent event) {
		if (event.getSource() == choiceLevelSet) {
			if (event.getStateChange() == ItemEvent.SELECTED) {
				logger.debug("Event ItemStateChanged " + event);
				gameProperties.setLevelSetIndex(choiceLevelSet.getSelectedIndex());
				gameProperties.setLevel(choiceLevel.getSelectedIndex());
				updateUIFieldsBasedOnProperties();
				repaint();
			}
		}
	}

	private Vector<LevelSetComboBoxItemModel> createLevelSetComboBoxItems(final Messages messages) {
		Vector<LevelSetComboBoxItemModel> list = new Vector<LevelSetComboBoxItemModel>(); 
		for(int index = 0; null != messages.getText(LevelSetComboBoxItemModel.LEVELSET_KEY_PREFIX + index, null); index++) {
			logger.debug("Creating LevelSetComboBoxItemModel " + index + " " + messages.getText(LevelSetComboBoxItemModel.LEVELSET_KEY_PREFIX + index, null));
			final ApplicationMainWindow parent = (ApplicationMainWindow) getParent();
			LevelSetComboBoxItemModel item = new LevelSetComboBoxItemModel(index, KittenLevelSet.isAvailable(parent.getApplicationEnvironment().resources(), index), messages);
			list.add(item);
		}
		return list;
	}

	private Vector<LevelComboBoxItemModel> createLevelComboBoxItems(final int levelSetIndex) {
		final Vector<LevelComboBoxItemModel> list = new Vector<LevelComboBoxItemModel>(); 
		Vector<LevelInfo> levelset;
		try {
			final ApplicationMainWindow parent = (ApplicationMainWindow) getParent();
			levelset = gameProperties.getLevelSet(parent.getApplicationEnvironment().resources());
		} catch (final IOException e) {
			levelset = new Vector<LevelInfo>();
		}
		for (int i = 0; i < levelset.size(); i++) {
			list.add(new LevelComboBoxItemModel(i, levelset.get(i)));
		}
		return list;
	}

	private Logger logger = Logger.getLogger(getClass());

	private JIconComboBox<LevelSetComboBoxItemModel> choiceLevelSet;
	private JIconComboBox<LevelComboBoxItemModel> choiceLevel;

	private GameProperties gameProperties;
}
