package de.ped.kitten.gui;

import java.io.IOException;
import java.io.InputStream;

/**
 * Defines Atari 8bit monospace font.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TrueTypeMonospaceFont implements Cloneable {

	private static final Character[] PIXEL_REPRESENTATIONS = { Character.valueOf('.'), Character.valueOf('*')}; 

	private static final String[][] ATARI8BIT_CHARDEFS = new String[][] {
		{ // 0x00 = Heart
			"........",
			"..**.**.",
			".*******",
			".*******",
			"..*****.",
			"...***..",
			"....*...",
			"........",
		}, { // 0x01 = 
			"...**...",
			"...**...",
			"...**...",
			"...*****",
			"...*****",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x02 = 
			"......**",
			"......**",
			"......**",
			"......**",
			"......**",
			"......**",
			"......**",
			"......**",
		}, { // 0x03 = 
			"...**...",
			"...**...",
			"...**...",
			"*****...",
			"*****...",
			"........",
			"........",
			"........",
		}, { // 0x04 = 
			"...**...",
			"...**...",
			"...**...",
			"*****...",
			"*****...",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x05 = 
			"........",
			"........",
			"........",
			"*****...",
			"*****...",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x06 = 
			"......**",
			".....***",
			"....***.",
			"...***..",
			"..***...",
			".***....",
			"***.....",
			"**......",
		}, { // 0x07 = 
			"**......",
			"***.....",
			".***....",
			"..***...",
			"...***..",
			"....***.",
			".....***",
			"......**",
		}, { // 0x08 = 
			".......*",
			"......**",
			".....***",
			"....****",
			"...*****",
			"..******",
			".*******",
			"********",
		}, { // 0x09 = 
			"........",
			"........",
			"........",
			"........",
			"....****",
			"....****",
			"....****",
			"....****",
		}, { // 0x0a = 
			"*.......",
			"**......",
			"***.....",
			"****....",
			"*****...",
			"******..",
			"*******.",
			"********",
		}, { // 0x0b = 
			"....****",
			"....****",
			"....****",
			"....****",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x0c = 
			"****....",
			"****....",
			"****....",
			"****....",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x0d = 
			"********",
			"********",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x0e = 
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"********",
			"********",
		}, { // 0x0f = 
			"........",
			"........",
			"........",
			"........",
			"****....",
			"****....",
			"****....",
			"****....",
		}, { // 0x10 = 
			"........",
			"...***..",
			"...***..",
			".***.***",
			".***.***",
			"....*...",
			"...***..",
			"........",
		}, { // 0x11 = 
			"........",
			"........",
			"........",
			"...*****",
			"...*****",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x12 = 
			"........",
			"........",
			"........",
			"********",
			"********",
			"........",
			"........",
			"........",
		}, { // 0x13 = 
			"...**...",
			"...**...",
			"...**...",
			"********",
			"********",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x14 = 
			"........",
			"........",
			"..****..",
			".******.",
			".******.",
			".******.",
			"..****..",
			"........",
		}, { // 0x15 = 
			"........",
			"........",
			"........",
			"........",
			"********",
			"********",
			"********",
			"********",
		}, { // 0x16 = 
			"**......",
			"**......",
			"**......",
			"**......",
			"**......",
			"**......",
			"**......",
			"**......",
		}, { // 0x17 = 
			"........",
			"........",
			"........",
			"********",
			"********",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x18 = 
			"...**...",
			"...**...",
			"...**...",
			"********",
			"********",
			"........",
			"........",
			"........",
		}, { // 0x19 = 
			"****....",
			"****....",
			"****....",
			"****....",
			"****....",
			"****....",
			"****....",
			"****....",
		}, { // 0x1a = 
			"...**...",
			"...**...",
			"...**...",
			"...*****",
			"...*****",
			"........",
			"........",
			"........",
		}, { // 0x1b = ESC
			"........",
			".*****..",
			".*......",
			".****...",
			".*......",
			".*****..",
			"...*....",
			"...***..",
		}, { // 0x1c = UP
			"........",
			"...**...",
			"..****..",
			".******.",
			"...**...",
			"...**...",
			"...**...",
			"........",
		}, { // 0x1d = DOWN
			"........",
			"...**...",
			"...**...",
			"...**...",
			".******.",
			"..****..",
			"...**...",
			"........",
		}, { // 0x1e = LEFT
			"........",
			"...*....",
			"..**....",
			".******.",
			".******.",
			"..**....",
			"...*....",
			"........",
		}, { // 0x1f = RIGHT
			"........",
			"....*...",
			"....**..",
			".******.",
			".******.",
			"....**..",
			"....*...",
			"........",
		}, { // 0x20 = BLANK
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x21 = !
			"........",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"........",
			"...**...",
			"........",
		}, { // 0x22 = "
			"........",
			".**..**.",
			".**..**.",
			"........",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x23 = #
			"........",
			"..**.**.",
			".*******",
			"..**.**.",
			"..**.**.",
			".*******",
			"..**.**.",
			"........",
		}, { // 0x24 = $
			"...**...",
			"..*****.",
			".**.....",
			"..****..",
			".....**.",
			".*****..",
			"...**...",
			"........",
		}, { // 0x25 = %
			"........",
			".**..**.",
			".**.**..",
			"...**...",
			"..**....",
			".**..**.",
			".*...**.",
			"........",
		}, { // 0x26 = &
			"........",
			"..***...",
			".**.**..",
			"..***...",
			".**.**..",
			".**.***.",
			"..**.**.",
			"........",
		}, { // 0x27 = '
			"........",
			"...**...",
			"...**...",
			"........",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x28 = (
			"........",
			"....**..",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"....**..",
			"........",
		}, { // 0x29 = )
			"........",
			"..**....",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"..**....",
			"........",
		}, { // 0x2a = *
			"........",
			".**..**.",
			"..****..",
			"********",
			"..****..",
			".**..**.",
			"........",
			"........",
		}, { // 0x2b = +
			"........",
			"...**...",
			"...**...",
			".******.",
			"...**...",
			"...**...",
			"........",
			"........",
		}, { // 0x2c = ,
			"........",
			"........",
			"........",
			"........",
			"........",
			"...**...",
			"...**...",
			"..**....",
		}, { // 0x2d = -
			"........",
			"........",
			"........",
			".******.",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x2e = .
			"........",
			"........",
			"........",
			"........",
			"........",
			"...**...",
			"...**...",
			"........",
		}, { // 0x2f = /
			"........",
			".....**.",
			"....**..",
			"...**...",
			"..**....",
			".**.....",
			".*......",
			"........",
		}, { // 0x30 = 0
			"........",
			"..****..",
			".**..**.",
			".**.***.",
			".***.**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x31 = 1
			"........",
			"...**...",
			"..***...",
			"...**...",
			"...**...",
			"...**...",
			".******.",
			"........",
		}, { // 0x32 = 2
			"........",
			"..****..",
			".**..**.",
			"....**..",
			"...**...",
			"..**....",
			".******.",
			"........",
		}, { // 0x33 = 3
			"........",
			".******.",
			"....**..",
			"...**...",
			"....**..",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x34 = 4
			"........",
			"....**..",
			"...***..",
			"..****..",
			".**.**..",
			".******.",
			"....**..",
			"........",
		}, { // 0x35 = 5
			"........",
			".******.",
			".**.....",
			".*****..",
			".....**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x36 = 6
			"........",
			"..****..",
			".**.....",
			".*****..",
			".**..**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x37 = 7
			"........",
			".******.",
			".....**.",
			"....**..",
			"...**...",
			"..**....",
			"..**....",
			"........",
		}, { // 0x38 = 8
			"........",
			"..****..",
			".**..**.",
			"..****..",
			".**..**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x39 = 9
			"........",
			"..****..",
			".**..**.",
			"..*****.",
			".....**.",
			"....**..",
			"..***...",
			"........",
		}, { // 0x3a = :
			"........",
			"........",
			"...**...",
			"...**...",
			"........",
			"...**...",
			"...**...",
			"........",
		}, { // 0x3b = ;
			"........",
			"........",
			"...**...",
			"...**...",
			"........",
			"...**...",
			"...**...",
			"..**....",
		}, { // 0x3c = <
			".**.....",
			"..**....",
			"...**...",
			"....**..",
			"...**...",
			"..**....",
			".**.....",
			"........",
		}, { // 0x3d = =
			"........",
			"........",
			".******.",
			"........",
			"........",
			".******.",
			"........",
			"........",
		}, { // 0x3e = >
			".....**.",
			"....**..",
			"...**...",
			"..**....",
			"...**...",
			"....**..",
			".....**.",
			"........",
		}, { // 0x3f = ?
			"........",
			"..****..",
			".**..**.",
			"....**..",
			"...**...",
			"........",
			"...**...",
			"........",
		}, { // 0x40 = @
			"........",
			"..****..",
			".**..**.",
			".**.***.",
			".**.***.",
			".**.....",
			"..****..",
			"........",
		}, { // 0x41 = A
			"........",
			"...**...",
			"..****..",
			".**..**.",
			".**..**.",
			".******.",
			".**..**.",
			"........",
		}, { // 0x42 = B
			"........",
			".*****..",
			".**..**.",
			".*****..",
			".**..**.",
			".**..**.",
			".*****..",
			"........",
		}, { // 0x43 = C
			"........",
			"..****..",
			".**..**.",
			".**.....",
			".**.....",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x44 = D
			"........",
			".****...",
			".**.**..",
			".**..**.",
			".**..**.",
			".**.**..",
			".****...",
			"........",
		}, { // 0x45 = E
			"........",
			".******.",
			".**.....",
			".*****..",
			".**.....",
			".**.....",
			".******.",
			"........",
		}, { // 0x46 = F
			"........",
			".******.",
			".**.....",
			".*****..",
			".**.....",
			".**.....",
			".**.....",
			"........",
		}, { // 0x47 = G
			"........",
			"..****..",
			".**..**.",
			".**.....",
			".**.***.",
			".**..**.",
			"..*****.",
			"........",
		}, { // 0x48 = H
			"........",
			".**..**.",
			".**..**.",
			".******.",
			".**..**.",
			".**..**.",
			".**..**.",
			"........",
		}, { // 0x49 = I
			"........",
			".******.",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			".******.",
			"........",
		}, { // 0x4a = J
			"........",
			".....**.",
			".....**.",
			".....**.",
			".....**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x4b = K
			"........",
			".**..**.",
			".**.**..",
			".****...",
			".****...",
			".**.**..",
			".**..**.",
			"........",
		}, { // 0x4c = L
			"........",
			".**.....",
			".**.....",
			".**.....",
			".**.....",
			".**.....",
			".******.",
			"........",
		}, { // 0x4d = M
			"........",
			".**...**",
			".***.***",
			".*******",
			".**.*.**",
			".**...**",
			".**...**",
			"........",
		}, { // 0x4e = N
			"........",
			".**..**.",
			".***.**.",
			".******.",
			".******.",
			".**.***.",
			".**..**.",
			"........",
		}, { // 0x4f = O
			"........",
			"..****..",
			".**..**.",
			".**..**.",
			".**..**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x50 = P
			"........",
			".*****..",
			".**..**.",
			".**..**.",
			".*****..",
			".**.....",
			".**.....",
			"........",
		}, { // 0x51 = Q
			"........",
			"..****..",
			".**..**.",
			".**..**.",
			".**..**.",
			".**.**..",
			"..**.**.",
			"........",
		}, { // 0x52 = R
			"........",
			".*****..",
			".**..**.",
			".**..**.",
			".*****..",
			".**.**..",
			".**..**.",
			"........",
		}, { // 0x53 = S
			"........",
			"..****..",
			".**.....",
			"..****..",
			".....**.",
			".....**.",
			"..****..",
			"........",
		}, { // 0x54 = T
			"........",
			".******.",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"........",
		}, { // 0x55 = U
			"........",
			".**..**.",
			".**..**.",
			".**..**.",
			".**..**.",
			".**..**.",
			".******.",
			"........",
		}, { // 0x56 = V
			"........",
			".**..**.",
			".**..**.",
			".**..**.",
			".**..**.",
			"..****..",
			"...**...",
			"........",
		}, { // 0x57 = W
			"........",
			".**...**",
			".**...**",
			".**.*.**",
			".*******",
			".***.***",
			".**...**",
			"........",
		}, { // 0x58 = X
			"........",
			".**..**.",
			".**..**.",
			"..****..",
			"..****..",
			".**..**.",
			".**..**.",
			"........",
		}, { // 0x59 = Y
			"........",
			".**..**.",
			".**..**.",
			"..****..",
			"...**...",
			"...**...",
			"...**...",
			"........",
		}, { // 0x5a = Z
			"........",
			".******.",
			"....**..",
			"...**...",
			"..**....",
			".**.....",
			".******.",
			"........",
		}, { // 0x5b = [
			"........",
			"...***..",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"...***..",
			"........",
		}, { // 0x5c = \
			"........",
			".**.....",
			"..**....",
			"...**...",
			"....**..",
			".....**.",
			"......*.",
			"........",
		}, { // 0x5d = ]
			"........",
			"..***...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"..***...",
			"........",
		}, { // 0x5e = ^
			"........",
			"...**...",
			"..****..",
			".**..**.",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x5f = _
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			".******.",
			"........",
		}, { // 0x60 = Karo
			"........",
			"...**...",
			"..****..",
			".******.",
			".******.",
			"..****..",
			"...**...",
			"........",
		}, { // 0x61 = a
			"........",
			"........",
			"..****..",
			".....**.",
			"..*****.",
			".**..**.",
			"..*****.",
			"........",
		}, { // 0x62 = b
			"........",
			".**.....",
			".**.....",
			".*****..",
			".**..**.",
			".**..**.",
			".*****..",
			"........",
		}, { // 0x63 = c
			"........",
			"........",
			"..****..",
			".**.....",
			".**.....",
			".**.....",
			"..****..",
			"........",
		}, { // 0x64 = d
			"........",
			".....**.",
			".....**.",
			"..*****.",
			".**..**.",
			".**..**.",
			"..*****.",
			"........",
		}, { // 0x65 = e
			"........",
			"........",
			"..****..",
			".**..**.",
			".******.",
			".**.....",
			"..****..",
			"........",
		}, { // 0x66 = f
			"........",
			"....***.",
			"...**...",
			"..*****.",
			"...**...",
			"...**...",
			"...**...",
			"........",
		}, { // 0x67 = g
			"........",
			"........",
			"..*****.",
			".**..**.",
			".**..**.",
			"..*****.",
			".....**.",
			".*****..",
		}, { // 0x68 = h
			"........",
			".**.....",
			".**.....",
			".*****..",
			".**..**.",
			".**..**.",
			".**..**.",
			"........",
		}, { // 0x69 = i
			"........",
			"...**...",
			"........",
			"..***...",
			"...**...",
			"...**...",
			"..****..",
			"........",
		}, { // 0x6a = j
			"........",
			".....**.",
			"........",
			".....**.",
			".....**.",
			".....**.",
			".....**.",
			"..****..",
		}, { // 0x6b = k
			"........",
			".**.....",
			".**.....",
			".**.**..",
			".****...",
			".**.**..",
			".**..**.",
			"........",
		}, { // 0x6c = l
			"........",
			"..***...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"..****..",
			"........",
		}, { // 0x6d = m
			"........",
			"........",
			".**..**.",
			".*******",
			".*******",
			".**.*.**",
			".**...**",
			"........",
		}, { // 0x6e = n
			"........",
			"........",
			".*****..",
			".**..**.",
			".**..**.",
			".**..**.",
			".**..**.",
			"........",
		}, { // 0x6f = o
			"........",
			"........",
			"..****..",
			".**..**.",
			".**..**.",
			".**..**.",
			"..****..",
			"........",
		}, { // 0x70 = p
			"........",
			"........",
			".*****..",
			".**..**.",
			".**..**.",
			".*****..",
			".**.....",
			".**.....",
		}, { // 0x71 = q
			"........",
			"........",
			"..*****.",
			".**..**.",
			".**..**.",
			"..*****.",
			".....**.",
			".....**.",
		}, { // 0x72 = r
			"........",
			"........",
			".*****..",
			".**..**.",
			".**.....",
			".**.....",
			".**.....",
			"........",
		}, { // 0x73 = s
			"........",
			"........",
			"..*****.",
			".**.....",
			"..****..",
			".....**.",
			".*****..",
			"........",
		}, { // 0x74 = t
			"........",
			"...**...",
			".******.",
			"...**...",
			"...**...",
			"...**...",
			"....***.",
			"........",
		}, { // 0x75 = u
			"........",
			"........",
			".**..**.",
			".**..**.",
			".**..**.",
			".**..**.",
			"..*****.",
			"........",
		}, { // 0x76 = v
			"........",
			"........",
			".**..**.",
			".**..**.",
			".**..**.",
			"..****..",
			"...**...",
			"........",
		}, { // 0x77 = w
			"........",
			"........",
			".**...**",
			".**.*.**",
			".*******",
			"..*****.",
			"..**.**.",
			"........",
		}, { // 0x78 = x
			"........",
			"........",
			".**..**.",
			"..****..",
			"...**...",
			"..****..",
			".**..**.",
			"........",
		}, { // 0x79 = y
			"........",
			"........",
			".**..**.",
			".**..**.",
			".**..**.",
			"..*****.",
			"....**..",
			".****...",
		}, { // 0x7a = z
			"........",
			"........",
			".******.",
			"....**..",
			"...**...",
			"..**....",
			".******.",
			"........",
		}, { // 0x7b = SPADE
			"........",
			"...**...",
			"..****..",
			".******.",
			".******.",
			"...**...",
			"..****..",
			"........",
		}, { // 0x7c = 
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
			"...**...",
		}, { // 0x7d = CLS
			"........",
			".*****..",
			".***....",
			".****...",
			".*..**..",
			".*..**..",
			"....**..",
			"........",
		}, { // 0x7e = 
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
		}, { // 0x7f = 
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
			"........",
		}
	};

	private TrueTypeMonospaceChar[] chardefs;

	public TrueTypeMonospaceFont() {
		setChardefs(ATARI8BIT_CHARDEFS);
	}

	/**
	 * Creates a deep copy
	 * 
	 * @param other
	 *            the object to clone
	 */
	public TrueTypeMonospaceFont(final TrueTypeMonospaceFont other) {
		chardefs = new TrueTypeMonospaceChar[other.chardefs.length];
		for (int i = 0; i < chardefs.length; i++) {
			chardefs[i] = other.chardefs[i].clone();
		}
	}

	protected void setChardefs(final String[][] chardefs) {
		this.chardefs = new TrueTypeMonospaceChar[chardefs.length];
		for (int i = 0; i < chardefs.length; i++) {
			setChardef(i, chardefs[i]);
		}
	}

	public void setChardefs(final InputStream is, final int bytesPerChar) {
		final byte[] buffer = new byte[bytesPerChar];
		final int[]  ibuffer = new int[bytesPerChar];
		int charsCount = 0;
		try {
			while (bytesPerChar == is.read(buffer)) {
				for (int i = 0; i < buffer.length; i++) {
					ibuffer[i] = (int)(buffer[i] & 0xff);
				}
				int index;
				if (charsCount < 0x40) {
					index = charsCount + 0x20;
				} else if (charsCount < 0x60) {
					index = charsCount - 0x40;
				} else {
					index = charsCount;
				}
				setChardef(index, ibuffer);
				charsCount++;
			}
		} catch (final IOException e) {
			// Nothing to do
		}
	}

	public void setChardef(final int c, final String[] chardef) {
		if (null == chardefs[c]) {
			chardefs[c] = new TrueTypeMonospaceChar();
		}
		chardefs[c].add(new TrueTypeMonospaceSingleSizeChar(c, chardef));
	}

	public void setChardef(final int c, final int[] defs) {
		final String[] defstr = new String[defs.length];
		for (int y = 0; y < defs.length; y++) {
			int value = defs[y];
			String def = "";
			int mask = 0x1;
			for (int x = 0; x < defs.length; x++) {
				final Character ch = ((value & mask) == 0 ? PIXEL_REPRESENTATIONS[0] : PIXEL_REPRESENTATIONS[1]);
				def = ch + def;
				mask = mask << 1;
			}
			defstr[y] = def;
		}
		setChardef(c, defstr);
	}

	protected String[] getCharDef(final int c) {
		return getCharDef(c, 1, 1);
	}

	protected String[] getCharDef(final int c, final int xstretch, final int ystretch) {
		String[] ret = null;
		final char basicChar = (char) (c & 0x7f);
		final int index = basicChar;
		if ((0 <= index) && (index < chardefs.length)) {
			ret = chardefs[index].get(8 * xstretch, 8 * ystretch);
			if (0x80 <= c) {
				ret = TrueTypeMonospaceSingleSizeChar.invert(ret);
			}
		}
		return ret;
	}

	@Override
	public TrueTypeMonospaceFont clone()  {
		return new TrueTypeMonospaceFont(this);
	}
}
