package de.ped.kitten.gui;

import java.awt.Dimension;

import de.ped.tools.log.Logger;

/**
 * Defines one char of one size of a true type monospace font.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TrueTypeMonospaceSingleSizeChar implements Cloneable {

	/**
	 * Stretches the given character definition by the given factors.
	 * 
	 * @param def
	 *            the initial character definition
	 * @param xstretch
	 *            the factor to stretch the width
	 * @param ystretch
	 *            the factor to stretch the height
	 * @return the stretched character definition
	 */
	public static String[] stretch(final String[] def, final int xstretch, final int ystretch) {
		final String[] ret = new String[def.length * ystretch];
		final int width = def[0].length() * xstretch;
		for (int y = 0; y < ret.length; y++) {
			final int sy = y / ystretch;
			final StringBuffer line = new StringBuffer();
			for (int x = 0; x < width; x++) {
				final int sx = x / xstretch;
				try {
					line.append(def[sy].charAt(sx));
				} catch (StringIndexOutOfBoundsException ex) {
					Logger.getLogger("TrueTypeMonospaceSingleSizeChar").info(ex.toString(), ex);
					System.exit(1);
				}
			}
			ret[y] = line.toString();
		}
		return ret;
	}

	/**
	 * Inverts the given character definition - switches background and
	 * foreground symbol representations.
	 * 
	 * @param def
	 *            the character definition to invert
	 * @return the inverted character definition
	 */
	public static String[] invert(final String[] def) {
		String[] ret = null;
		if (null != def) {
			ret = new String[def.length];
			for (int yi = 0; yi < def.length; yi++) {
				final String lineIn = def[yi];
				final StringBuffer lineOut = new StringBuffer();
				for (int xi = 0; xi < lineIn.length(); xi++) {
					char c = '*';
					if (lineIn.charAt(xi) == '*') {
						c = '.';
					}
					lineOut.append(c);
				}
				ret[yi] = lineOut.toString();
			}
		}
		return ret;
	}

	/**
	 * Mirrors the given character definition horizontally and/or vertically.
	 * 
	 * @param def
	 *            the character definition to invert
	 * @param isHorizontal
	 *            if <code>true</code>, left/right mirroring is applied
	 * @param isVertical
	 *            if <code>true</code>, up/down mirroring is applied
	 * @return the inverted character definition
	 */
	public static String[] mirror(final String[] def, final boolean isHorizontal, final boolean isVertical) {
		String[] ret = null;
		if (null != def) {
			ret = new String[def.length];
			for (int yi = 0; yi < def.length; yi++) {
				final int syi = (isVertical ? def.length - 1 - yi : yi);
				final String lineIn = def[syi];
				final StringBuffer lineOut = new StringBuffer();
				for (int xi = 0; xi < lineIn.length(); xi++) {
					final int sxi = (isHorizontal? lineIn.length() - 1 - xi : xi);
					lineOut.append(lineIn.charAt(sxi));
				}
				ret[yi] = lineOut.toString();
			}
		}
		return ret;
	}

	private final int index;
	private final String[] def;
	private final Dimension size;

	public TrueTypeMonospaceSingleSizeChar(final int index, final String[] def) {
		this.index = index;
		this.def = def;
		this.size = new Dimension(def[0].length(), def.length);
	}

	public int getIndex() {
		return index;
	}

	public Dimension getSize() {
		return size;
	}

	public char get(final int x, final int y) {
		return def[y].charAt(x);
	}

	public String[] getDef() {
		return def;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + index;
		result = prime * result + ((size == null) ? 0 : size.hashCode());
		return result;
	}

	/**
	 * @param obj
	 *            the object to compare with
	 * @return <code>true</code> if both objects are of the same class, and if
	 *         they have the same index and the same size
	 */
	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final TrueTypeMonospaceSingleSizeChar other = (TrueTypeMonospaceSingleSizeChar) obj;
		if (index != other.index)
			return false;
		if (size == null) {
			if (other.size != null)
				return false;
		} else if (!size.equals(other.size))
			return false;
		return true;
	}

	/** 
	 * @return a deep copy
	 */
	@Override
	public TrueTypeMonospaceSingleSizeChar clone() {
		final TrueTypeMonospaceSingleSizeChar target = new TrueTypeMonospaceSingleSizeChar(index, def.clone());
		return target;
	}
}
