package de.ped.kitten.logic;

import java.io.IOException;
import java.util.LinkedList;
import java.util.Vector;

import de.ped.kitten.logic.Level.Symbol;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.ResourceFinder;
import de.ped.tools.log.Logger;

/**
 * The business logic of the game.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Game {

	/** 
	 * The possible global states the game can be in. 
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum State {
		IN_TITLE, IN_GAME, GAME_OVER
	};

	/** 
	 * The events that the game can tell the outside world, e.g., to play sounds. 
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Event {
		/** Plays the next tone of the background music */
		BACKGROUND,
		/** Cat was bored */
		MEOW,
		/** Title screen was started */
		START_TITLE, 
		/** Game was started */
		START_GAME, 
		/** Plays the game over music */
		GAME_OVER,
		/** Resets the background music */
		START_LEVEL, 
		/** Plays the success music */
		LEVEL_DONE, 
		/** Plays the death sound */
		DIED,
		/** Tomcat did a step of a jump */
		JUMP,
		/** Tomcat just hit a spring */
		SPRING,
		/** Tomcat fell a step */
		FALL,
		/** Tomcat collided with a wall or block */
		COLLISION,
		/** Tomcat just collected a flower */
		FLOWER,
	};

	/** The time it takes to count done one time step in a level. */
	public static final Duration LEVEL_TIME_DURATION = new Duration(5.0);

	/** The time in game steps it takes until an idle cat meows. */
	private static final Duration IDLE_MEOW_TIME = LEVEL_TIME_DURATION;

	/** The time in game steps how long a meow takes. */
	private static final Duration IDLE_MEOW_DURATION = new Duration(0.8);

	private static final Duration GAME_OVER_DURATION = new Duration(0.4 * 16L);
	
	//	private static final double MICROFLICRO_FLYOVER_IN_SEC = 6.0;
//	private static final long MICROFLICRO_FLYOVER_STEPS = Math.round(MICROFLICRO_FLYOVER_IN_SEC * 1000L / GAME_STEP_DELAY);

	private Logger logger = Logger.getLogger(getClass());

	private Vector<LevelInfo> levels;
	private int levelIndex = 0;

	private GameProperties props = new GameProperties();

	private State state;
	private int lives = 3;
	private int points = 0;

	private Level level;
	private long stepsElapsedInGameState = 0;
	private long stepsElapsedInLevel = 0;
	private TomcatState tomcatState;
	private TomcatState previousTomcatState = null;
	private long stepsElapsedInTomcatState = 0;
	private int fallingSteps = 0;
	private double microflicrosXPosition = 0.0;
	private final ResourceFinder resources;

	private LinkedList<Event> eventsStore = new LinkedList<Event>();

	public Game(final ResourceFinder resources) throws IOException {
		this.resources = resources;
		logger.setLogLevel(de.ped.tools.log.Level.INFO);
		initGame(props, false);
	}

	public GameProperties getProperties() {
		return props;
	}

	public synchronized void addEventToStore(final Event event) {
		eventsStore.add(event);
	}

	public synchronized void cleanEventStoreTo(final LinkedList<Event> sink) {
		if (null != sink) {
			sink.addAll(eventsStore);
		}
		eventsStore.clear();
	}

	public synchronized void initGame(final GameProperties props, final boolean isInGame) throws IOException {
		this.props = props;
		if (!isInGame) {
			setState(State.IN_TITLE);
		} else {
			setState(State.IN_GAME);
			levels = props.getLevelSet(resources);
			lives = 3;
			points = 0;
			initLevel(props.getLevel());
//			logger.debug("Microflicro speed is " + level.getLevelInfo().microflicroSpeed);
		}
	}

	public synchronized void initLevel(final int nextLevelIndex) {
		levelIndex = MathUtil.modulo(nextLevelIndex, levels.size());
		props.setLevel(levelIndex);
		level = new Level(levels.get(levelIndex));
		state = State.IN_GAME;
		stepsElapsedInLevel = 0L;
		tomcatState = TomcatState.FALLING_LEFT;
		previousTomcatState = null;
		fallingSteps = 0;
		microflicrosXPosition = -2.0;
		addEventToStore(Event.START_LEVEL);
	}

	public Level getLevel() {
		return level;
	}

	public int getLevelIndex() {
		return levelIndex;
	}

	private void setState(final Game.State nextState) {
		state = nextState;
		switch(state) {
		case IN_TITLE:
			levels = null;
			addEventToStore(Event.START_TITLE);
			break;
		case IN_GAME:
			addEventToStore(Event.START_GAME);
			break;
		case GAME_OVER:
			addEventToStore(Event.GAME_OVER);
			break;
		}
		stepsElapsedInGameState = 0L;
	}

	public State getState() {
		return state;
	}

	public TomcatState getTomcatState() {
		return tomcatState;
	}

	public long getStepsElapsedInTomcatState() {
		return stepsElapsedInTomcatState;
	}

	public void setTomcatState(
			final TomcatState state, 
			final boolean isWithFlowerDelay, 
			final LinkedList<Event> events) {
		if (((tomcatState == TomcatState.DYING) || (tomcatState == TomcatState.SUCCEEDED)) && (state != tomcatState.naturalNextState)) {
			logger.debug("Prohibited state change " + tomcatState.name() + " -> " + state.name());
			// Don't do the change.
		} else {
			logger.debug("setState " + tomcatState.name() + " -> " + state.name());
			previousTomcatState = tomcatState;
			tomcatState = state;
			if (isWithFlowerDelay) {
				stepsElapsedInTomcatState = -Duration.FLOWER.steps;
			} else {
				stepsElapsedInTomcatState = 0;
			}
			if (state.overallState == TomcatOverallState.FALLING) {
				fallingSteps++;
//				logger.debug("Falling steps = " + fallingSteps);
			} else {
				fallingSteps = 0;
			}
			// Add events
			if ((tomcatState == TomcatState.DYING) && (tomcatState != previousTomcatState)) {
				events.add(Event.DIED);
			} else if ((tomcatState == TomcatState.SUCCEEDED) && (tomcatState != previousTomcatState)) {
				events.add(Event.LEVEL_DONE);
			} else if (tomcatState.overallState == TomcatOverallState.FALLING) {
				events.add(Event.FALL);
			} else if ((tomcatState.overallState == TomcatOverallState.JUMPING) && (tomcatState != previousTomcatState)) {
				events.add(Event.JUMP);
			} else if ((tomcatState.overallState == TomcatOverallState.SPRING) && (tomcatState != previousTomcatState)) {
				events.add(Event.JUMP);
			}
		}
	}

	public static int calcRemainingLevelTime(final long stepsInLevel, final int levelInfoTime) {
		return Math.max(0, levelInfoTime - (int)(stepsInLevel/LEVEL_TIME_DURATION.steps));
	}

	public int getRemainingLevelTime() {
		return calcRemainingLevelTime(stepsElapsedInLevel, level.getLevelInfo().time);
	}

	public int getPoints() {
		return points;
	}

	public int getLives() {
		return lives;
	}

	public long getStepsElapsedInLevel() {
		return stepsElapsedInLevel;
	}

	public double getMicroflicrosXPosition() {
		return microflicrosXPosition;
	}

	public int getMicroflicrosLoopState() {
		return (int)((stepsElapsedInLevel / Duration.MICROFLICRO_CHEW.steps) % 2L);
	}

	public long getStepsElapsedInGameState() {
		return stepsElapsedInGameState;
	}

	public LinkedList<Event> step() {
		if (null != level) {
			logger.trace("Step " + stepsElapsedInLevel + " --> " + getMicroflicrosXPosition() + ", " + getRemainingLevelTime());
		}
		LinkedList<Event> events = new LinkedList<Event>();
		cleanEventStoreTo(events);
		if (state == State.IN_TITLE) {
			inTitleStep(events);
		} else {
			inGameStep(events);
		}
		stepsElapsedInGameState++;
		return events;
	}

	private void inTitleStep(final LinkedList<Event> events) {
		// Nothing to do
	}

	private void inGameStep(final LinkedList<Event> events) {
		if (tomcatState == TomcatState.FINISHED) {
			if (previousTomcatState == TomcatState.SUCCEEDED) {
				int nextLevel = levelIndex + 1;
				if (props.isSingleLevelGame()) {
					nextLevel = levelIndex;
				}
				initLevel(nextLevel);
			} else {
				lives--;
				if (0 < getLives()) {
					initLevel(levelIndex);
				} else {
					if (State.IN_GAME == state) {
						setState(State.GAME_OVER);
						events.add(Event.GAME_OVER);
					} else if (State.GAME_OVER == state) {
						if (GAME_OVER_DURATION.steps < stepsElapsedInGameState) {
							setState(State.IN_TITLE);
						}
					}
				}
			}
		} else {
			inLevelStep(events);
		}
	}

	private boolean wasSuicideRequested = false;

	/** 
	 * Just a bundle of data transferred between the methods handling a step. 
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private class TomcatMoveResult {
		/** 
		 * The next state tomcat will go into, depending on where he landed with the last move.
		 * If <code>null</code>, the move itself forces no new state.
		 */
		TomcatState nextState;
		/** If <code>true</code>, tomcat hit a flower, so he needs some time to collect it. */
		boolean isWithFlowersDelay;

		TomcatMoveResult() {
			set(null, false);
		}

		void set(final TomcatState nextState, final boolean isWithFlowersDelay) {
			this.nextState = nextState;
			this.isWithFlowersDelay = isWithFlowersDelay;
		}
	}

	private void inLevelStep(final LinkedList<Event> events) {
		microflicrosXPosition = MathUtil.modulo(microflicrosXPosition + Duration.GAME_STEP_DELAY / 75.0 / level.getLevelInfo().microflicroSpeed, 19.0);
		TomcatState nextState = null;
		boolean isWithFlowerDelay = false;
		if ((!tomcatState.timeTillNaturalNextState.equals(Duration.FOREVER)) && (stepsElapsedInTomcatState >= tomcatState.timeTillNaturalNextState.steps)) {
			// Default next state is natural next state
			if (null == tomcatState.naturalNextState) {
				nextState = tomcatState; // As we can't define the current state as next state, we need to use null for this.
			} else {
				nextState = tomcatState.naturalNextState;
			}
			// If tomcat moves, derive any actions from that
			if ((tomcatState.directionX != 0) || (tomcatState.directionY != 0)) {
				TomcatMoveResult result = tomcatMove(events);
				if (null != result.nextState) {
					nextState = result.nextState;
				} else if (result.isWithFlowersDelay) {
					isWithFlowerDelay = true;
				}
			}
		} 
		// Check for microflicro collision
		final double mfxp = getMicroflicrosXPosition();
		final PlayfieldPosition tomcatPos = level.getPosition();
		if ((Math.abs(mfxp - tomcatPos.getX()) < 1.0) && (level.getMicroflicros()[tomcatPos.getY()])) {
			nextState = TomcatState.DYING;
		}
		// Timeout
		if ((tomcatState != TomcatState.DYING) && (tomcatState != TomcatState.SUCCEEDED) && (getRemainingLevelTime() <= 0)) {
			nextState = TomcatState.DYING;
		}
		// Suicide
		if (wasSuicideRequested) {
			nextState = TomcatState.DYING;
			wasSuicideRequested = false;
		}
		// Finish
		if (null != nextState) {
			setTomcatState(nextState, isWithFlowerDelay, events);
		}
		// Idle meow
		final TomcatState s = getTomcatState();
		if ((s == TomcatState.IDLE_UNDIRECTED) && (IDLE_MEOW_TIME.steps == stepsElapsedInTomcatState)) {
			events.add(Event.MEOW);
		}
		// See for background music
		if ((s.overallState != TomcatOverallState.DYING) &&
				(s.overallState != TomcatOverallState.SUCCEEDED) &&
				(s.overallState != TomcatOverallState.FINISHED)) {
			if (0 == stepsElapsedInLevel % Duration.BACKGROUND_NOTE.steps) {
				events.add(Event.BACKGROUND);
			}
		}
		stepsElapsedInLevel++;
		stepsElapsedInTomcatState++;
	}

	/**
	 * As part of a step, moves the tomcat to another field. 
	 * Depending on the field, several actions occur.
	 * @param events this method adds new events to this collection
	 * @return all information necessary to handle the other aspects of the step
	 */
	private TomcatMoveResult tomcatMove(final LinkedList<Event> events) {
		final TomcatMoveResult result = new TomcatMoveResult(); 
		if (tomcatState.overallState == TomcatOverallState.SPRING) {
			// When on spring, only watch to not jump out of screen, but ignore all else on screen.
			if (level.getPosition().getY() <= 1) {
				result.nextState = tomcatState.getFallingState();
			} else {
				level.move(tomcatState.directionX, tomcatState.directionY, false);
			}
		} else {
			// Look where tomcat is, and act accordingly.
			final Symbol symbolAtNewPos = level.symbolAtPositionAfterMove(tomcatState.directionX, tomcatState.directionY);
			switch(symbolAtNewPos) {
			case BLANK:
				level.move(tomcatState.directionX, tomcatState.directionY, false);
				break;
			case WALL:
			case BLOCK:
				level.confirmPosition();
				if (tomcatState.overallState == TomcatOverallState.FALLING) {
					logger.debug("Fallen steps = " + fallingSteps);
					if (6 < fallingSteps) {
						result.nextState = TomcatState.DYING;
					} else {
						result.nextState = tomcatState.getIdleState();
					}
				} else {
					if (tomcatState.overallState == TomcatOverallState.JUMPING) {
						events.add(Event.COLLISION);
					}
					result.nextState = tomcatState.getFallingState();
				}
				logger.debug("Hit wall. State " + tomcatState.name() + " -> " + result.nextState.name());
				break;
			case HURDLE:
				level.move(tomcatState.directionX, tomcatState.directionY, false);
				result.nextState = TomcatState.DYING;
				break;
			case SPRING:
				level.move(tomcatState.directionX, tomcatState.directionY, false);
				events.add(Event.SPRING);
				result.nextState = tomcatState.getSpringState();
				break;
			case FLOWER:
				level.move(tomcatState.directionX, tomcatState.directionY, false);
				if (level.eatsFlower()) {
					points += 50;
					result.isWithFlowersDelay = true;
					events.add(Event.FLOWER);
				}
				break;
			case CAT:
				level.move(tomcatState.directionX, tomcatState.directionY, true);
				if (0 < level.getRemainingFlowers()) {
					result.nextState = TomcatState.DYING;
				} else {
					points += 200;
					result.nextState = TomcatState.SUCCEEDED;
				}
				break;
			case TOMCAT:
				throw new IllegalStateException("Found another tomcat in level?!?");
			default:
				throw new IllegalStateException("Unknown symbol " + symbolAtNewPos.name());
			}
		}
		return result;
	}

	/**
	 * Tells tomcat to die. May be called by other threads.
	 */
	public void die() {
		wasSuicideRequested = true;
	}

	/**
	 * Tells tomcat where to move to. May be called by other threads.
	 * 
	 * @param playerIdx
	 *            the player id
	 * @param playerAction
	 *            the player action, given as {@link DigitalJoystickDirection}
	 *            mask
	 * @return the game events that happened because of the move
	 */
	public LinkedList<Event> suggestUserGameAction(final int playerIdx, final int playerAction) {
		final LinkedList<Event> events = new LinkedList<Event>();
		TomcatState lastState = tomcatState;
		TomcatState nextState = null;
		if (DigitalJoystickDirection.is(playerAction, DigitalJoystickDirection.MASK_LEFT)) {
			if (lastState.overallState == TomcatOverallState.IDLE) {
				nextState = TomcatState.IDLE_LEFT;
			}
		} else if (DigitalJoystickDirection.is(playerAction, DigitalJoystickDirection.MASK_RIGHT)) {
			if (lastState.overallState == TomcatOverallState.IDLE) {
				nextState = TomcatState.IDLE_RIGHT;
			}
		}
		TomcatState currentState = lastState;
		if (null != nextState) {
			currentState = nextState;
		}
		if (DigitalJoystickDirection.is(playerAction, DigitalJoystickDirection.MASK_UP)) {
			if (currentState == TomcatState.IDLE_LEFT) {
				nextState = TomcatState.JUMPING_LEFT_HIGH0;
			} else if (currentState == TomcatState.IDLE_RIGHT) {
				nextState = TomcatState.JUMPING_RIGHT_HIGH0;
			}
		} else if (DigitalJoystickDirection.is(playerAction, DigitalJoystickDirection.MASK_DOWN)) {
			if (currentState == TomcatState.IDLE_LEFT) {
				nextState = TomcatState.JUMPING_LEFT_LOW0;
			} else if (currentState == TomcatState.IDLE_RIGHT) {
				nextState = TomcatState.JUMPING_RIGHT_LOW0;
			}
		}
		if (null != nextState) {
			level.confirmPosition();
			setTomcatState(nextState, false, events);
		}
		return events;
	}

	/**
	 * @return if the tomcat is right in a meow. Could be in title, or in game
	 *         because it is bored. (The dying meow is not handled here.)
	 */
	public boolean isInMeow() {
		boolean ret = false;
		switch(state) {
		case IN_TITLE:
			if ((5 <= stepsElapsedInGameState) && (stepsElapsedInGameState <= IDLE_MEOW_DURATION.steps)) {
				ret = true;
			}
			break;
		case IN_GAME:
			if ((getTomcatState() == TomcatState.IDLE_UNDIRECTED) && (IDLE_MEOW_TIME.steps <= stepsElapsedInTomcatState) && (stepsElapsedInTomcatState < IDLE_MEOW_TIME.steps + IDLE_MEOW_DURATION.steps)) {
				ret = true;
			}
			break;
		default:
			// Nothing to do
			break;
		}
		return ret;
	}
}
