package de.ped.kitten.logic;


import java.util.Arrays;

import de.ped.tools.PlayfieldPosition;

/**
 * The description and logic of a game level.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Level {

	/**
	 * One of the building blocks a level can be built of.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static enum Symbol {
		/** A blank field. */
		BLANK, 
		/** A wall. */
		WALL, 
		/** A block; behaves the same as a wall, but looks more open. */
		BLOCK, 
		/** A hurdle. */
		HURDLE, 
		/** A spring. */
		SPRING, 
		/** A flower. */
		FLOWER, 
		/** The tom cat main character - should exist only once. */
		TOMCAT, 
		/** The cat to reach - should exist only once. */
		CAT,
	};

	private LevelInfo levelInfo;

	private Symbol[][] screen;
	private boolean[] microflicros;

	private PlayfieldPosition position;
	private PlayfieldPosition lastPosition;
	private int remainingFlowers;

	protected Symbol[][] initScreen(final LevelInfo levelInfo) {
		Symbol[][] source = levelInfo.screen;
		Symbol[][] ret = new Symbol[source.length][source[0].length];
		for (int x = 0; x < source.length; x++) {
			ret[x] = Arrays.copyOf(source[x], source[x].length);
		}
		return ret;
	}

	private boolean[] initMicroflicros(final LevelInfo levelInfo) {
		return Arrays.copyOf(levelInfo.microflicros, levelInfo.microflicros.length);
	}

	public Level(final LevelInfo levelInfo) {
		this.levelInfo = levelInfo;
		screen = initScreen(levelInfo);
		microflicros = initMicroflicros(levelInfo);
		position = new PlayfieldPosition(levelInfo.start);
		lastPosition = new PlayfieldPosition(position);
		remainingFlowers = levelInfo.flowers;
	}

	public LevelInfo getLevelInfo() {
		return levelInfo;
	}

	public synchronized Symbol getAt(final int x, final int y) {
		return screen[x][y];
	}

	protected void setAt(final int x, final int y, final Symbol value) {
		screen[x][y] = value;
	}

	public boolean[] getMicroflicros() {
		return microflicros;
	}

	public synchronized PlayfieldPosition getPosition() {
		return position;
	}

	public synchronized PlayfieldPosition getLastPosition() {
		return lastPosition;
	}

	public int getRemainingFlowers() {
		return remainingFlowers;
	}

	public void confirmPosition() {
		lastPosition.setLocation(position.getX(), position.getY());
	}

	/**
	 * Moves tomcat to the given position.
	 * @param x the x coordinate of the direction to move; may be -1, 0 or 1
	 * @param y the y coordinate of the direction to move; may be -1, 0 or 1
	 * @param isAtCat if the given position is the cat our tomcat tries to reach
	 */
	public synchronized void move(final int x, final int y, final boolean isAtCat) {
		confirmPosition();
		int nx = position.getX() + x;
		final int ny = position.getY() + y;
		if (isAtCat) {
			nx--;
		}
		position.setLocation(nx, ny);
	}

	public synchronized Symbol symbolAtPositionAfterMove(final int x, final int y) {
		return screen[position.getX() + x][position.getY() + y];
	}

	public synchronized boolean eatsFlower() {
		boolean ret = false;
		if (screen[position.getX()][position.getY()] == Symbol.FLOWER) {
			setAt(position.getX(), position.getY(), Symbol.BLANK);
			if (0 < remainingFlowers) {
				remainingFlowers--;
			}
			ret = true;
		}
		return ret;
	}

	public byte[] getTitle() {
		return levelInfo.title;
	}
}