package de.ped.kitten.logic;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashMap;

import de.ped.kitten.logic.Level.Symbol;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;

/**
 * All the basic data to build up a level.
 * 
 * Contains only final data elements, so it can be safely referenced to. 
 */
public class LevelInfo {

	/** The size of the levels, including borders. */
	public static final PlayfieldDimension SIZE = new PlayfieldDimension(20, 20);

	private static final HashMap<Byte, Symbol> LEVEL_INFO_BYTE_CONVERTER = new HashMap<Byte, Symbol>();
	private static final HashMap<Character, Symbol> LEVEL_INFO_CHAR_CONVERTER = new HashMap<Character, Symbol>();

	static {
		LEVEL_INFO_CHAR_CONVERTER.put('.', Symbol.BLANK);
		LEVEL_INFO_CHAR_CONVERTER.put('1', Symbol.BLOCK);
		LEVEL_INFO_CHAR_CONVERTER.put('2', Symbol.WALL);
		LEVEL_INFO_CHAR_CONVERTER.put('3', Symbol.SPRING);
		LEVEL_INFO_CHAR_CONVERTER.put('4', Symbol.HURDLE);
		LEVEL_INFO_CHAR_CONVERTER.put('5', Symbol.FLOWER);
		LEVEL_INFO_CHAR_CONVERTER.put('6', Symbol.TOMCAT);
		LEVEL_INFO_CHAR_CONVERTER.put('7', Symbol.CAT);
	}

	static {
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte)   0), Symbol.BLANK);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte) -53), Symbol.BLOCK);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte)  10), Symbol.WALL);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte)  13), Symbol.SPRING);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte) -52), Symbol.HURDLE);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte)-114), Symbol.FLOWER);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte)  68), Symbol.TOMCAT);
		LEVEL_INFO_BYTE_CONVERTER.put(Byte.valueOf((byte) -60), Symbol.CAT);
	}

	private static Symbol[][] initScreen() {
		final Symbol[][] ret = new Symbol[LevelInfo.SIZE.getWidth()][LevelInfo.SIZE.getHeight()];
		for (int y = 0; y < ret[0].length; y++) {
			ret[0][y] = Symbol.WALL;
			ret[ret.length - 1][y] = Symbol.WALL;
		}
		for (int x = 0; x < ret.length; x++) {
			ret[x][0] = Symbol.WALL;
			ret[x][ret[0].length - 1] = Symbol.WALL;
		}
		for (int y = 1; y < ret[0].length - 1; y++) {
			for (int x = 1; x < ret.length - 1; x++) {
				ret[x][y] = Symbol.BLANK;
			}
		}
		return ret;
	}

	private static Symbol[][] convertScreen(final String[] screen) {
		final Symbol[][] ret = initScreen();
		for (int y = 1; y < ret[0].length - 1; y++) {
			for (int x = 1; x < ret.length - 1; x++) {
				final Character c = screen[y - 1].charAt(x - 1);
				final Symbol s = LEVEL_INFO_CHAR_CONVERTER.get(c);
				if (null == s) {
					throw new IllegalArgumentException("Found symbol " + c + " at position (" + x + "," + y + ")");
				}
				ret[x][y] = s;
			}
		}
		return ret;
	}

	private static Symbol[][] convertScreen(final byte[][] screen) {
		final Symbol[][] ret = initScreen();
		for (int y = 1; y < ret[0].length - 1; y++) {
			for (int x = 1; x < ret.length - 1; x++) {
				final Byte b = Byte.valueOf(screen[y][x]);
				Symbol s = LEVEL_INFO_BYTE_CONVERTER.get(b);
				if (null == s) {
					throw new IllegalArgumentException("Found symbol " + Integer.toString(b) + " at position (" + x + "," + y + ")");
				}
				ret[x][y] = s;
			}
		}
		return ret;
	}

	private static boolean[] initMicroflicros() {
		boolean[] ret = new boolean[SIZE.getHeight()];
		return ret;
	}

	private static boolean[] convertMicroflicros(final String microflicros) {
		boolean[] ret = initMicroflicros();
		for (int i = 0; i < microflicros.length(); i++) {
			if ('.' != microflicros.charAt(i)) {
				ret[i + 1] = true;
			}
		}
		return ret;
	}

	private static boolean[] convertMicroflicros(final byte[] microflicroSprite) {
		final boolean[] ret = initMicroflicros();
		for (int i = 0; i < 18; i++) {
			final byte b = microflicroSprite[8 * i + 62];
//			logger.debug("" + Byte.toString(b));
			if((byte)0 != b) {
				ret[i + 1] = true;
			}
		}
		return ret;
	}

	private static void convertTitle(final byte[] title, final String string) {
		Arrays.fill(title, (byte)0);
		final int start = (SIZE.getWidth() - string.length()) / 2;
		for (int i = 0; i < string.length(); i++) {
			title[start + i] = Atari8bitScreenUtil.get4ColCharForChar(string.charAt(i), 1);
		}
	}

	/** The speed of the microflicros. Allowed values are 1 (for high speed) and 2 (for low speed). */
	public final int microflicroSpeed;
	/** The number of flowers tomcat needs to collect on this level to be accepted by his mate */
	public final int flowers;
	/** The time (in multitudes of 5 seconds) tomcat has to finish this level. */
	public final int time;
	/** The start position of tomcat. */
	public final PlayfieldPosition start;
	/** The destination position for tomcat. That's where the cat is. */
//	public final PlayfieldPosition destination;
	/** The name of the screen, coded as it would be on Atari - that means bytes representing 4-color chars. */
	public final byte[] title;
	/** An array describing if there is a microflicro in that row. */
	public final boolean[] microflicros;
	/** An array of size {@link #SIZE} describing the content of that level. */
	public final Symbol[][] screen;

	/**
	 * Initializes the level by defining all parameters in human-readable form.
	 * @param microflicroSpeed allowed values are 1 and 2
	 * @param flowers the number of flowers that tomcat needs to collect on this level
	 * @param time the time (in multitudes of 5 seconds) tomcat has to finish this level
	 * @param microflicros an array describing which rows are threatened by microflicros
	 * @param titleString the string to show in the level title line
	 * @param screen the screen content
	 */
	public LevelInfo(
			final int microflicroSpeed, 
			final int flowers, 
			final int time, 
			final String microflicros, 
			final String titleString, 
			final String[] screen) {
		this.microflicroSpeed = microflicroSpeed;
		this.flowers = flowers;
		this.time = time;
		this.microflicros = convertMicroflicros(microflicros);
		this.screen = convertScreen(screen);
		this.start = find(Symbol.TOMCAT);
//		this.destination = find(Symbol.CAT);
		this.screen[start.getX()][start.getY()] = Symbol.BLANK;
		this.title = new byte[20];
		convertTitle(this.title, titleString);
	}

	private void read(final InputStream is, final byte[] arr) throws IOException {
		final int numOfReadBytes = is.read(arr);
		if (numOfReadBytes < arr.length) {
			throw new IOException("Unable to read a full set of " + arr.length + " bytes, just found " + numOfReadBytes + " bytes!");
		}
	}

	/**
	 * Initializes the level by reading the data as it was stored by the original Kitten program on Atari 8 bit systems.
	 * @param is input stream delivering the data
	 * @throws IOException if something went wrong in reading the data
	 */
	public LevelInfo(final InputStream is) throws IOException {
		final byte[][] inputLinesData = new byte[23][SIZE.getWidth()];
		for (int y = 0; y < inputLinesData.length; y++) {
			read(is, inputLinesData[y]);
		}
		this.screen = convertScreen(inputLinesData);

		String titleLine = "";
		for (int x = 0; x < SIZE.getWidth(); x++) {
			final byte b = inputLinesData[SIZE.getHeight() + 1][x];
//			titleLine = titleLine + Character.valueOf((char)b);
			titleLine = titleLine + Integer.toHexString(b) + " ";
		}
		this.title = inputLinesData[SIZE.getHeight() + 1];

		final byte[] positionData = new byte[2];
		read(is, positionData); //Tomcat position 
		final PlayfieldPosition startPositionByFile = new PlayfieldPosition(positionData[0], positionData[1] - 2);
		read(is, positionData); //Cat position 
//		final PlayfieldPosition destPositionByFile = new PlayfieldPosition(positionData[0], positionData[1] - 2);

		final byte[] levelData = new byte[4];
		read(is, levelData);
		this.microflicroSpeed = levelData[0];
		this.flowers = levelData[1];
		this.time = levelData[2];

		final byte[] microflicroSprite = new byte[256];
		read(is, microflicroSprite); 
		this.microflicros = convertMicroflicros(microflicroSprite);
		read(is, microflicroSprite); 

		final PlayfieldPosition startPositionBySymbol = find(Symbol.TOMCAT);
//		final PlayfieldPosition destPositionBySymbol = find(Symbol.CAT);

		this.start = (startPositionBySymbol == null ? startPositionByFile : startPositionBySymbol);
//		this.destination = (destPositionBySymbol == null ? destPositionByFile : destPositionBySymbol);
		try {
			this.screen[start.getX()][start.getY()] = Symbol.BLANK;
		} catch (final ArrayIndexOutOfBoundsException ex) {
			// Nothing to do
		}
	}

	private PlayfieldPosition find(final Symbol s) {
		for (int x = 0; x < screen.length; x++) {
			for (int y = 0; y < screen[x].length; y++) {
				if (s == screen[x][y]) {
					return new PlayfieldPosition(x, y);
				}
			}
		}
		return null;
	}

	public String getTitleAsString() {
		final StringBuffer ret = new StringBuffer();
		for (int i = 0; i < title.length; i++) {
			final char c = Atari8bitScreenUtil.getCharIndexFor4ColChar(title[i]);
			ret.append(c);
		}
		return ret.toString();
	}
}