package de.ped.motivation.logic;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Locale;
import java.util.concurrent.TimeUnit;

import de.ped.tools.HtmlHelper;
import de.ped.tools.Messages;
import de.ped.tools.TextFormatter;
import de.ped.tools.web.RootBean;

/**
 * The business logic behind the motivation JSP.
 * 
 * @author Copyright (c) 2022 Peter Diefenbach
 */
public class Motivation {
	public static final String DEFAULT_CONTENT = "<p>Was machst Du hier?!?</p>";
	public static final String PARAM_KEY_FOR   = "for";
	public static final String PARAM_KEY_ACTION = "action";
	public static final String PARAM_KEY_MOTIVE = "motive";
	public static final String PARAM_KEY_DEMO  = "demo";
	public static final String PARAM_KEY_STAT  = "stat";

	public static final String PARAM_VAL_ACTION_SUBMIT = "submit";
	public static final String PARAM_VAL_ACTION_SELECTMOTIVE = "selectmotive";
	public static final String[] PARAM_VAL_ACTIONS = { PARAM_VAL_ACTION_SUBMIT, PARAM_VAL_ACTION_SELECTMOTIVE };
	public static final String[] PARAM_VAL_MOTIVES = { "progressbar", "flower" };
	public static final String PARAM_VAL_TRUE = "true";
	public static final int DURATION_NEVER = -999999;

	public static String href(final String valueFor, final String valueAction, final String valueMotive, final String valueDemo, final boolean isStat) {
		String ret = "index.jsp";
		if (null != valueFor) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_FOR, valueFor);
		}
		if (null != valueAction) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_ACTION, valueAction);
		}
		if (null != valueMotive) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_MOTIVE, valueMotive);
		}
		if (null != valueDemo) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_DEMO, valueDemo);
		}
		if (isStat) {
			ret = HtmlHelper.addParameter(ret, PARAM_KEY_STAT, PARAM_VAL_TRUE);
		}
		ret = HtmlHelper.sgmlify(ret);
		return ret;
	}

	/**
	 * Parses a parameter and returns a normalized form of the value.
	 * 
	 * @param key
	 *            the parameter key to parse
	 * @param valueRead
	 *            the parameter value to parse
	 * @return the normalized parameter value
	 */
	public static String parseParam(final String key, final String valueRead) {
		String ret = null;
		if (PARAM_KEY_FOR.equalsIgnoreCase(key)) {
			ret = valueRead;
		} else if (PARAM_KEY_ACTION.equalsIgnoreCase(key)) {
			for (String action : PARAM_VAL_ACTIONS) {
				if (action.equalsIgnoreCase(valueRead)) {
					ret = action;
					break;
				}
			}
		} else if (PARAM_KEY_MOTIVE.equalsIgnoreCase(key)) {
			for (String motive : PARAM_VAL_MOTIVES) {
				if (motive.equalsIgnoreCase(valueRead)) {
					ret = motive;
					break;
				}
			}
		} else if (PARAM_KEY_DEMO.equalsIgnoreCase(key)) {
			ret = valueRead;
		} else if (PARAM_KEY_STAT.equalsIgnoreCase(key)) {
			if (PARAM_VAL_TRUE.equalsIgnoreCase(valueRead)) {
				ret = PARAM_VAL_TRUE;
			}
		} else {
			throw new IllegalArgumentException("Unknown parameter key: " + key);
		}
		return ret;
	}

	/*
CREATE TABLE Motivations (
	STRING(10) id,
	STRING name,
	STRING motive,
	DATE created NOT NULL,
	PRIMARY KEY(id)
)
CREATE TABLE History (
	STRING(10) id FOREIGN KEY Motivations.id;
	DATE start NOT NULL,
	INTEGER duration NOT NULL DEFAULT 0,
	PRIMARY KEY(id, start)
)
INSERT INTO(Motivations) VALUES("1234567890", "Peter", "28.12.2022")
INSERT INTO(History) VALUES("1234567890", "28.12.2022", 0)
SELECT * FROM History WHERE id="1234567890";
UPDATE History.duration = 1 WHERE id="1234567890" AND start="28.12.2022";
	 */

	final static SimpleDateFormat SDF = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN);

	/**
	 * Java representation of data table Motivation.UserData
	 * @author Copyright (c) 2022 Peter Diefenbach
	 */
	public static class UserData {
		public String id;
		public String name;
		public Date created;
		public String motive;
	}
	public static UserData createUserData(final String id, final String name, final String created, final String motive) throws ParseException {
		final UserData ret = new UserData();
		ret.id = id;
		ret.name = name;
		ret.created = SDF.parse(created);
		ret.motive = motive;
		return ret;
	}

	/**
	 * Java representation of data table Motivation.HistoryData
	 * @author Copyright (c) 2022 Peter Diefenbach
	 */
	public static class HistoryData {
		public String id;
		public Date date;
		public int days;
	}
	public static HistoryData createHistoryData(final String id, final String date, final int days) throws ParseException {
		final HistoryData ret = new HistoryData();
		ret.id = id;
		ret.date = SDF.parse(date);
		ret.days = days;
		return ret;
	}

	/**
	 * A structure to deliver to the JSP when creating the page
	 * @author Copyright (c) 2022 Peter Diefenbach
	 */
	public static class DataForPage {
		public UserData data = new UserData();
		public int duration; /** how old in days is the account */
		public int days = 0; /** how long is the current streak */
		public boolean isToday = false; /** was today already trained */
		public int latest;	/** since how many days not trained */
		public int longestRun; /** how long was the longest streak */
		public double avg; /** ratio of days trained since account exists */
	}

	private final ArrayList<UserData> userData = new ArrayList<UserData>();
	private final ArrayList<HistoryData> historyData = new ArrayList<HistoryData>();
	private Date today = new Date();
	{
		try {
			userData.add(createUserData("XXXXXX", "BärTiger", "01.02.2023", PARAM_VAL_MOTIVES[0]));
			userData.add(createUserData("YYYYYY", "McBlack",  "01.02.2023", PARAM_VAL_MOTIVES[0]));
			userData.add(createUserData("ZZZZZZ", "Faule Socke", "02.02.2023", PARAM_VAL_MOTIVES[0]));
			historyData.add(createHistoryData("XXXXXX", "01.02.2023", 3));
			historyData.add(createHistoryData("YYYYYY", "01.02.2023", 1));
			historyData.add(createHistoryData("YYYYYY", "03.02.2023", 1));
			today = SDF.parse("04.02.2023"); // Hack for testing
		} catch (ParseException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public long diffInDays(final Date firstDate, final Date secondDate) {
	    long diffInMillies = Math.abs(secondDate.getTime() - firstDate.getTime());
	    long diff = TimeUnit.DAYS.convert(diffInMillies, TimeUnit.MILLISECONDS);
	    return diff;
	}

	public DataForPage getDataFor(final String vFor, final String vAction, final String vMotive) {
		DataForPage ret = null;
		if (null != vFor) {
			// Find user data
			UserData udata = null;
			for (UserData ud : userData) {
				if (ud.id.equalsIgnoreCase(vFor)) {
					udata = ud;
					break;
				}
			}
			if (null != udata) {
				if (null != vMotive) {
					udata.motive = vMotive;
				}
				if (PARAM_VAL_ACTION_SUBMIT.equalsIgnoreCase(vAction)) {
					// Find fitting history data
					HistoryData hdata = null;
					boolean isForToday = false;
					for (HistoryData hd : historyData) {
						if (hd.id.equalsIgnoreCase(udata.id)) {
							long diff = diffInDays(hd.date, today);
							if ((0 <= diff) && (diff < hd.days)) {
								hdata = hd;
								isForToday = true;
								break;
							}
							if ((0 <= diff) && (diff == hd.days)) {
								hdata = hd;
								break;
							}
						}
					}
					if (isForToday) {
						// Nothing to do
					} else {
						if (null != hdata) {
							hdata.days++;
						} else {
							try {
								hdata = createHistoryData(udata.id, SDF.format(today), 1);
								historyData.add(hdata);
							} catch (ParseException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
						isForToday = true;
					}
				}
				ret = new DataForPage();
				ret.data = udata;
				boolean isForToday = false;
				int daystrained = 0;
				HistoryData latestData = null;
				for (HistoryData hd : historyData) {
					if (hd.id.equalsIgnoreCase(udata.id)) {
						// Find current streak
						long diff = diffInDays(hd.date, today);
						if ((0 <= diff) && (diff < hd.days)) {
							ret.days = (int) diffInDays(hd.date, today);
							ret.isToday = true;
						}
						if ((0 <= diff) && (diff == hd.days)) {
							ret.days = (int) diffInDays(hd.date, today);
							ret.isToday = false;
						}
						// Find longest streak
						if (ret.longestRun < hd.days) {
							ret.longestRun = hd.days;
						}
						// Sum up days trained
						daystrained += hd.days;
						// Find latest entry
						if ((latestData == null) || (latestData.date.before(hd.date))) {
							latestData = hd;
						}
					}
				}
				ret.duration = (int) diffInDays(udata.created, today) + 1;
				ret.avg = (double) daystrained / (ret.duration + 1);
				ret.latest = ((null == latestData) ? DURATION_NEVER : (int) diffInDays(latestData.date, today) - latestData.days + 1);
			}
		}
		return ret;
	}

	public String getStat() {
		StringBuffer s = new StringBuffer();
		for (UserData udata : userData) {
			DataForPage e = getDataFor(udata.id, null, null);
			s.append("<tr>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: left; ") + ">");
			s.append(e.data.name);
			s.append("</td>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: right; ") + ">");
			s.append(SDF.format(e.data.created));
			s.append("</td>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: right; ") + ">");
			s.append(e.duration);
			s.append("</td>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: right; ") + ">");
			s.append(e.days);
			s.append("</td>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: right; ") + ">");
			s.append((e.latest == DURATION_NEVER) ? "noch nie" : e.latest);
			s.append("</td>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: right; ") + ">");
			s.append(e.longestRun);
			s.append("</td>");
			s.append("<td" + TextFormatter.htmlAttribute(TextFormatter.STYLE, "text-align: right; ") + ">");
			s.append(Math.round(e.avg * 100));
			s.append("%");
			s.append("</td>");
			s.append("<td>");
			s.append("\n");
		}
		return s.toString();
	}

	public String getMotivesHtml(final String valueFor) {
		StringBuffer s = new StringBuffer();
		UserData udata = null;
		for (UserData ud : userData) {
			if (ud.id == valueFor) {
				udata = ud;
				break;
			}
		}
		if (null != udata) {
			s.append("&ndash;");
			for (String st : PARAM_VAL_MOTIVES) {
				s.append("[");
				if (!st.equalsIgnoreCase(udata.motive)) {
					s.append("<a" + RootBean.htmlAttribute(TextFormatter.HREF, Motivation.href(valueFor, null, st, null, false)) + ">");
				}
				s.append(st);
				if (!st.equalsIgnoreCase(udata.motive)) {
					s.append("</a>");
				}
				s.append("]");
				s.append("&ndash;");
			}
		}
		return s.toString();
	}
	
	private Messages messages = null;

	private Messages messages() {
		if (null == messages) {
			messages = new Messages("de.ped.motivation.logic.messages",
					"Messages",
					"1.0", 
					"Copyright 2024 Peter Diefenbach (peter@pdiefenbach.de)",
					"Motivation",
					"Motivation");
		}
		return messages;
	}

	/**
	 * Decides if this door is already open. Rules to follow:
	 * <ul>
	 *   <li>If the date to show is only a year without full date info, then the result is always <em>no</em>.</li>
	 *   <li>If <em>demoDateStr</em> is <code>null</code>, then the door is open if <em>dateToShowStr</em> is in the past or today.</li>
	 *   <li>If <em>demoDateStr</em> is not <code>null</code>, then the door is open if <em>dateToShowStr</em> is in the past or equal to <em>demoDateStr</em>.</li>
	 * </ul>
	 * 
	 * @param dateToShowStr
	 *            the date representation of the door to look for
	 * @param demoDateStr
	 *            the date representation of the demo mode (to test future
	 *            doors), or <code>null</code> for today
	 * @return if this door is already open
	 */
	public boolean isOpen(final String dateToShowStr, final String demoDateStr) {
		boolean ret = true;
		return ret;
	}

	private final String url = "jdbc:mysql://localhost:3306/"; 
	private final String dbName = "usr_web468_1";
	private final String driver = "com.mysql.jdbc.Driver"; 
	private final String userName = "web468"; 
	private final String password = "W57xNXx5"; 

	private String loadStory(final String vItem, final String vFor, final int vPart) {
		final StringBuffer sb = new StringBuffer();
		try { 
			Class.forName(driver).newInstance();
			final Connection conn = DriverManager.getConnection(url + dbName, userName, password);
			final int day = 1;
			final Statement st = conn.createStatement(); 
//			final ResultSet res = st.executeQuery("SELECT storyId, partId, subpartId, content FROM stories ORDER BY storyId, partId, subpartId ASC"); 
			final ResultSet res = st.executeQuery("SELECT content FROM stories WHERE storyId=0 AND partID=" + day + " AND subpartId=" + vPart); 
			while (res.next()) { 
				final String content = res.getString("content"); 
				sb.append(content);
			} 
			conn.close(); 
		} catch (final Exception e) { 
			sb.append(e.toString()); 
		} 
		return sb.toString();
	}

	public String saveStory(final String vItem, final String vFor, final String vPart, final String content) {
		String ret = null;
		if ((null != content) && (!content.isEmpty())) {
			try { 
				Class.forName(driver).newInstance();
				final Connection conn = DriverManager.getConnection(url + dbName, userName, password);

				final int day = 1;

				final PreparedStatement prepped = conn.prepareStatement("INSERT into stories(storyId, partId, subpartId, content) VALUES (?, ?, ?, ?) ON DUPLICATE KEY UPDATE content=?;");
				prepped.setInt(1, 0);
				prepped.setInt(2, day);
				prepped.setInt(3, Integer.parseInt(vPart));
				prepped.setString(4, content);
				prepped.setString(5, content);
				prepped.executeUpdate();

//				final Statement st = conn.createStatement(); 
//				final int val = st.executeUpdate("INSERT into stories(storyId, partId, subpartId, content) VALUES (" 
//						+ 0 + "," + day + "," + vPart + "," + "'" + content + "'" + ") ON DUPLICATE KEY UPDATE content=" + "'" + content + "'" + ";");
//				if ((val != 1) && (val != 2)){
//					ret = "Updated " + val + " rows?!?";
//				}
				conn.close(); 
			} catch (final Exception e) { 
				ret = e.toString(); 
			} 
		}
		return ret;
	}

	@SuppressWarnings("unused")
	private String getFilename(final String vItem, final String vFor, final String vPart) {
		return vFor + vItem + "_" + vPart + ".txt";
	}
}
