package de.ped.mp3sync;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Date;

import de.ped.tools.TextFormatter;
import de.ped.tools.log.Logger;

/**
 * Compares two files for synchronization, 
 * without changing anything within the file structure.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FileComparator implements Comparable<FileComparator> {
	/** Time in milliseconds to tolerate that destination could be newer than source. */
	public static final long TIMESTAMP_TOLERANCE = 60L * 1000L;

	public static final String DATETIME = "yyyy-MM-dd HH:mm:ss,SSS";
	public static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat(DATETIME);

	/** The parent folder for destination. Necessary if destination is <code>null</code>. */
	public final File destinationParent;
	/** The destination file or <code>null</code>, if there is no corresponding file. */
	public final File destination;
	/** The source file or <code>null</code>, if there is no corresponding file. */
	public final File source;
	/** The logger where to log output to. */
	public final Logger logger;

	/**
	 * The constructor initializes constants.
	 * 
	 * @param destinationParent
	 *            the parent folder for destination. Necessary if destination is
	 *            <code>null</code>.
	 * @param destination
	 *            the destination file or <code>null</code>, if there is no
	 *            corresponding file
	 * @param source
	 *            the source file or <code>null</code>, if there is no
	 *            corresponding file
	 * @param logger
	 *            the logger where to log output to
	 */
	public FileComparator(
			final File destinationParent, 
			final File destination, 
			final File source, 
			final Logger logger) {
		this.destinationParent = destinationParent;
		this.destination = destination;
		this.source = source;
		this.logger = logger;
	}

	/**
	 * Outputs a file for logging and debugging.
	 * 
	 * @param destination
	 *            the file to log
	 * @return a human-readable representation of that file suitable for
	 *         directory listings
	 */
	public static String listFile(final File destination) {
		final StringBuffer buf = new StringBuffer();
		buf.append(destination.isDirectory() ? "D" : "F");
		buf.append(" ");
		buf.append(DATE_FORMAT.format(new Date(destination.lastModified())));
		buf.append(" ");
		buf.append(TextFormatter.format(destination.length(), 8));
		buf.append(" ");
		buf.append(destination.getAbsolutePath());
		return buf.toString();
	}

	/** 
	 * @return <em>this</em> for logging and debugging
	 */
	@Override
	public String toString() {
		final StringBuffer buf = new StringBuffer();
		buf.append(isToCopy() ? "c" : "-"); 
		buf.append(isToDelete() ? "x" : "-");
		if (null == source) {
			if (null == destination) {
				// Nothing to do
			} else {
				buf.append("T");
				buf.append(listFile(destination));
			}
		} else {
			buf.append("S");
			buf.append(listFile(source));
		}
		return buf.toString();
	}

	/** 
	 * @return if the source should be copied to destination
	 */
	public boolean isToCopy() {
		boolean ret = false;
		if (null != source) {
			if (null == destination) {
				ret = true;
			} else if (!source.isDirectory()) { // Source is file
				if (!destination.isDirectory()) { // Destination is file
					if ((source.length() != destination.length()) ||
						(source.lastModified() > destination.lastModified() + TIMESTAMP_TOLERANCE)) {
						ret = true;
					}
				}
			}
		}
		return ret;
// Currently the code return always false if both are directories. If this should change, it needed to look somewhat like:
//		return ((null != source) && 
//				((null == destination) || 
//				 ((source.isDirectory() == destination.isDirectory()) && 
//				  ((source.length() != destination.length()) ||
//				   (source.lastModified() > destination.lastModified() + 3000L)))));
	}

	/** 
	 * @return if the destination should be deleted
	 */
	public boolean isToDelete() {
		return ((null == source) || 
				((null != destination) && 
				 ((source.isDirectory() != destination.isDirectory()))));
	}

	/** 
	 * @return if this really describes a file
	 */
	public boolean isToCollect() {
		return (true);
	}

	public boolean isDirectory() {
		return ((null != source) && (source.isDirectory()) ||
				(null != destination) && (destination.isDirectory()));
	}

	public String getName() {
		String ret = null;
		if (null != source) {
			ret = source.getName();
		} else if (null != destination) {
			ret = destination.getName();
		}
		return ret;
	}

	public File fullDestinationPath() {
		File ret = destination;
		if (null == ret) {
			ret = new File(destinationParent + File.separator + source.getName());
		}
		return ret;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((destination == null) ? 0 : destination.hashCode());
		result = prime
				* result
				+ ((destinationParent == null) ? 0 : destinationParent
						.hashCode());
		result = prime * result + ((source == null) ? 0 : source.hashCode());
		return result;
	}

	/**
	 * @param obj
	 *            the object to compare with
	 * @return if both objects are of the same type, and if destinationParent,
	 *         source and destination are equal
	 */
	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final FileComparator other = (FileComparator) obj;
		if (destination == null) {
			if (other.destination != null)
				return false;
		} else if (!destination.equals(other.destination))
			return false;
		if (destinationParent == null) {
			if (other.destinationParent != null)
				return false;
		} else if (!destinationParent.equals(other.destinationParent))
			return false;
		if (source == null) {
			if (other.source != null)
				return false;
		} else if (!source.equals(other.source))
			return false;
		return true;
	}

	/**
	 * Compares by the following criteria:
	 * <ol>
	 * <li>The destination parent folder, so nearby files are sorted nearby.</li>
	 * <li>The distinction file vs directory</li>
	 * <li>The lexical order of the file / directory name</li>
	 * </ol>
	 * 
	 * @param o
	 *            the object to compare with
	 * @return 0 if equal, a negative number if this is "smaller", and a
	 *         positive number if this is "larger"
	 */
	@Override
	public int compareTo(final FileComparator o) {
		int ret = destinationParent.compareTo(o.destinationParent);
		if (0 != ret) {
			return ret;
		}
		ret = (isDirectory() ? -1 : 0) + (o.isDirectory() ? 1 : 0);
		if (0 != ret) {
			return ret;
		}
		ret = getName().compareTo(o.getName());
		return ret;
	}
}