package de.ped.mp3sync;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;

import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.log.PrintWriterAppender;

/**
 * Synchronizes two folders.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MP3Sync {

	/**
	 * The goal to traverse the file tree for.
	 * 
	 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum TraversalMode {
		/** Only collect information, don't change the files */
		COLLECT(false, false),
		/** Only copy missing or outdated files, but don't delete files */
		COPY(true, false),
		/** Only delete files no longer present in the source repository */
		DELETE(false, true),
		/** Copy and delete in the same run */
		BOTH(true, true);
		public final boolean isDelete;
		public final boolean isCopy;
		private TraversalMode(final boolean isCopy, final boolean isDelete) {
			this.isCopy = isCopy;
			this.isDelete = isDelete;
		}
	};

	/** Milliseconds interval to poll until source and destination file are available. */
	private static final long POLL_TIME_IN_MSECS = 1000L * 60L;

	private static final String[][] EU_PATHS = new String[][] {
		{ "C:/Projekte/Peter/eclipse", "E:/eclipse" },
		{ "D:/Public",                 "E:/Public" },
		{ "D:/Eigene Dokumente",       "E:/Eigene Dokumente" },
		{ "D:/MP3",                    "E:/MP3" },
		{ "C:/Users/claud/Desktop",         "E:/Claudia/Desktop" },
		{ "C:/Users/claud/Documents",       "E:/Claudia/Documents" },
		{ "C:/Users/claud/Outlook-Dateien", "E:/Claudia/Outlook-Dateien" },
	};

	private static final String[][] CAR_PATHS = new String[][] {
		{ "D:/MP3",                    "E:/MP3" },
	};

	/**
	 * @param args call with argument <code>?</code> to see parameter options
	 */
	public static void main(final String[] args) {
		final String globalDestinationRootPath;
		switch(args.length) {
		case 0: // Standard MP3 sync with 1&1 drive
			globalDestinationRootPath = DESTINATION_ROOT_DEFAULT;
			break;
		case 1: // Standard MP3 sync with phone SD card OR help
			final String arg = args[0];
			if (arg.equals("?") || arg.equals("-?") || arg.equals("/?") || arg.equals("help") || arg.equals("-help") || arg.equals("--help") || arg.equals("/help")) {
				showHelp();
				globalDestinationRootPath = null;
			} else if (arg.equals("-toEU")) {
				toOrFromEU(0);
				globalDestinationRootPath = null;
			} else if (arg.equals("-fromEU")) {
				toOrFromEU(1);
				globalDestinationRootPath = null;
			} else if (arg.equals("-toCar")) {
				toCar();
				globalDestinationRootPath = null;
			} else {
				globalDestinationRootPath = arg;
				fileActionFilters = FILE_FILTERS_MP3_SMARTPHONE;
			}
			break;
		default: // Free full 1-way sync
			globalSourceRoot = new File(args[0]);
			globalDestinationRootPath = args[1];
			fileActionFilters = FILE_FILTERS_ALL_BACKUP;
		}
		if (null != globalDestinationRootPath) {
			globalDestinationRoot = new File(globalDestinationRootPath);
			final MP3Sync sync = new MP3Sync();
			sync.run();
		}
	}

	private static void toOrFromEU(final int sourceIndex) {
		fileActionFilters = FILE_FILTERS_ALL_BACKUP;
		traversalModes = TRAVERSAL_MODES_QUICK;
		final MP3Sync sync = new MP3Sync();
		for (String[] paths : EU_PATHS) {
			globalSourceRoot = new File(paths[sourceIndex]);
			globalDestinationRoot = new File(paths[1 - sourceIndex]);
			sync.run();
		}
	}

	private static void toCar() {
		fileActionFilters = FILE_FILTERS_MP3_SMARTPHONE;
		traversalModes = TRAVERSAL_MODES_QUICK;
		final MP3Sync sync = new MP3Sync();
		for (String[] paths : CAR_PATHS) {
			globalSourceRoot = new File(paths[0]);
			globalDestinationRoot = new File(paths[1]);
			sync.run();
		}
	}

	private static void showHelp() {
		System.out.println("Updates a destination directory based on a source directory.");
		System.out.println("Usage:");
		System.out.println("  mp3sync - copies music and data content to " + DESTINATION_ROOT_DEFAULT);
		System.out.println("  mp3sync -help - this online help");
		System.out.println("  mp3sync -toEU - copies all important stuff to SD card");
		System.out.println("  mp3sync -fromEU - copies all important stuff from SD card");
		System.out.println("  mp3sync -toCar - copies music to SD card, intended for car");
		System.out.println("  mp3sync <dir1> - copies music to <dir1>, intended for smartphones");
		System.out.println("  mp3sync <dir1> <dir2> - copies files from <dir1> to <dir2>");
	}

	/** 
	 * File filter what NOT to handle for copying music to smartphone SD card.
	 */
	public static final FileActionFilter[] FILE_FILTERS_MP3_SMARTPHONE = new FileActionFilter[] {
		// Directories
		new FileActionFilter("Automatisch zu iTunes hinzufgen", FileActionFilter.COPY),
		new FileActionFilter("Books", FileActionFilter.COPY),
		new FileActionFilter("Neu", FileActionFilter.COPY),
		new FileActionFilter("MCC Party Songs", FileActionFilter.COPY),
		new FileActionFilter("Playlists", FileActionFilter.COPY),
		new FileActionFilter("TV Shows", FileActionFilter.COPY),
		new FileActionFilter("rzte \\(Backup\\)", FileActionFilter.COPY),
		new FileActionFilter("Bilder", FileActionFilter.COPY),
		new FileActionFilter("Veoh", FileActionFilter.COPY),
		// Files
		new FileActionFilter("AlbumArtSmall\\.jpg", FileActionFilter.COPY),
		new FileActionFilter("AlbumArt_\\{.*\\.jpg", FileActionFilter.COPY),
		new FileActionFilter("[Dd]esktop\\.ini", FileActionFilter.COPY),
		new FileActionFilter("[Tt]humbs\\.db", FileActionFilter.COPY),
		new FileActionFilter("MP3\\.xls", FileActionFilter.COPY),
		new FileActionFilter("[Ff]older\\.jpg", FileActionFilter.COPY),
		new FileActionFilter("[Ff]older\\.JPG", FileActionFilter.COPY),
		new FileActionFilter(".*\\.lnk", FileActionFilter.COPY),
		new FileActionFilter(".*\\.[Bb][Mm][Pp]", FileActionFilter.COPY),
		new FileActionFilter(".*\\.[Gg][Ii][Ff]", FileActionFilter.COPY),
		new FileActionFilter(".*\\.[Jj][Pp][Gg]", FileActionFilter.COPY),
		new FileActionFilter(".*\\.[Jj][Pp][Ee][Gg]", FileActionFilter.COPY),
		new FileActionFilter(".*\\.[Pp][Nn][Gg]", FileActionFilter.COPY),
		new FileActionFilter(".*\\.flac", FileActionFilter.COPY),
		new FileActionFilter(".*\\.txt", FileActionFilter.COPY),
		new FileActionFilter(".*\\.xls", FileActionFilter.COPY),
		new FileActionFilter(".*\\.xlsx", FileActionFilter.COPY),
		new FileActionFilter(".*\\.htm", FileActionFilter.COPY),
		new FileActionFilter(".*\\.html", FileActionFilter.COPY),
		new FileActionFilter(".*\\.doc", FileActionFilter.COPY),
		new FileActionFilter(".*\\.docx", FileActionFilter.COPY),
		new FileActionFilter(".*\\.m3u", FileActionFilter.COPY),
 		new FileActionFilter(".*\\.epub", FileActionFilter.COPY),
		new FileActionFilter(".*\\.ini", FileActionFilter.COPY),
		new FileActionFilter(".*\\.nfo", FileActionFilter.COPY),
		new FileActionFilter("R-L\\.MP3", FileActionFilter.COPY),
	};

	/** 
	 * File filter what NOT to handle for copying music data to the cloud.
	 */
	public static final FileActionFilter[] FILE_FILTERS_MP3_BACKUP = new FileActionFilter[] {
		// Directories
		new FileActionFilter("Automatisch zu iTunes hinzufgen", FileActionFilter.COPY),
		new FileActionFilter("Neu", FileActionFilter.COPY),
		new FileActionFilter("Playlists", FileActionFilter.COPY),
		new FileActionFilter("MCC Party Songs", FileActionFilter.DELETE),
		new FileActionFilter("rzte \\(Backup\\)", FileActionFilter.BOTH),
		new FileActionFilter("Queen - Greatest Video Hits 2 - CD[12] - XviD.avi", FileActionFilter.BOTH),
		// Files
		new FileActionFilter("AlbumArtSmall\\.jpg", FileActionFilter.BOTH),
		new FileActionFilter("AlbumArt_\\{.*\\.jpg", FileActionFilter.BOTH),
		new FileActionFilter("[Dd]esktop\\.ini", FileActionFilter.BOTH),
		new FileActionFilter("[Tt]humbs\\.db", FileActionFilter.BOTH),
		new FileActionFilter("MP3\\.xls", FileActionFilter.BOTH),
		new FileActionFilter("[Ff]older\\.jpg", FileActionFilter.BOTH),
		new FileActionFilter("[Ff]older\\.JPG", FileActionFilter.BOTH),
		new FileActionFilter(".*\\.lnk", FileActionFilter.BOTH),
	};

	/** 
	 * File filter what NOT to handle for creating a backup.
	 */
	public static final FileActionFilter[] FILE_FILTERS_ALL_BACKUP = new FileActionFilter[] {
		// Directories
		new FileActionFilter(".metadata", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter(".recommenders", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("RemoteSystemsTempFiles", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter(".settings", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("Servers", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("javadoc", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("bin", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("build", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("generated", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter("rzte \\(Backup\\)", FileActionFilter.COPY),
		// Files
		new FileActionFilter("AlbumArtSmall\\.jpg", FileActionFilter.BOTH),
		new FileActionFilter("AlbumArt_\\{.*\\.jpg", FileActionFilter.BOTH),
		new FileActionFilter("[Dd]esktop\\.ini", FileActionFilter.BOTH),
		new FileActionFilter("[Tt]humbs\\.db", FileActionFilter.BOTH),
		new FileActionFilter("[Ff]older\\.jpg", FileActionFilter.BOTH),
		new FileActionFilter("[Ff]older\\.JPG", FileActionFilter.BOTH),
		new FileActionFilter("~$.*\\..*", FileActionFilter.BOTH), // MS Office tilde files as mutexes
		new FileActionFilter(".classpath", FileActionFilter.BOTH), // for workspace5 or workspace8
		new FileActionFilter(".project", FileActionFilter.BOTH), // for workspace5 or workspace8
	};

	final Logger logger;

	public Logger getLogger() {
		return logger;
	}

	public MP3Sync() {
		logger = Logger.getLogger(getClass());
		logger.removeAppender(Logger.DEFAULT_STRING_WRITER_APPENDER);
		try {
			final String logfilename = "C:\\Programme2\\mp3sync.log";
			final File file = new File(logfilename);
			logger.addAppender(new PrintWriterAppender(file));
		} catch (final IOException e) {
			logger.warn("Unable to create logfile, I'll do without.");
		}
		logger.setLogLevel(Level.DEBUG);
	}

	private LinkedList<FileComparator> fileList = new LinkedList<FileComparator>();

	/** 
	 * Waits until source and destination folder are available, and then synchronizes the folders. 
	 */
	public void run() {
		boolean done = false;
		do {
			done = execute();
			if (!done) {
				try {
					Thread.sleep(POLL_TIME_IN_MSECS);
				} catch (final InterruptedException e) {
					// Nothing to do
				}
			}
		} while (!done);
	}

	private static final File SOURCE_ROOT = new File("D:/MP3/"); // new File("C:\\Users\\Public\\MP3");
	private static final String DESTINATION_ROOT_DEFAULT = "O:/Public/MP3/"; // "https://sd2dav.1und1.de/Public/MP3/"; //"\\\\sd2dav.1und1.de@SSL\\DavWWWRoot\\Public\\MP3\\";
	private static File globalDestinationRoot;
	private static File globalSourceRoot = SOURCE_ROOT;
	private static FileActionFilter[] fileActionFilters = FILE_FILTERS_MP3_BACKUP;
	private static final TraversalMode[] TRAVERSAL_MODES_VERBOSE = new TraversalMode[] {
			TraversalMode.COPY,
			TraversalMode.COLLECT,
			TraversalMode.DELETE,
			TraversalMode.COLLECT,
		};
	private static final TraversalMode[] TRAVERSAL_MODES_QUICK = new TraversalMode[] {
			TraversalMode.BOTH,
		};
	private static TraversalMode[] traversalModes = TRAVERSAL_MODES_VERBOSE;

	/**
	 * Executes the synchronization.
	 * 
	 * @return <code>true</code> if synchronization was successful
	 */
	public boolean execute() {
		boolean ret = false;
		if (isSafe(globalDestinationRoot, globalSourceRoot)) {
			logger.info("Source root is " + globalSourceRoot.getPath() + ", destination root is " + globalDestinationRoot.getPath() + ".");
			for (final TraversalMode mode : traversalModes) {
				execute(globalDestinationRoot, globalSourceRoot, mode);
			}
			ret = true;
		}
		return ret;
	}

	private File sourceRoot;
	private File destinationRoot;
	

	public LinkedList<FileComparator> execute(final File destination, final File source, final TraversalMode mode) {
		logger.info("Started traversal " + mode.name() + ".");
		destination.mkdir();
		fileList.clear();
		sourceRoot = source;
		destinationRoot = destination;
		traverse(destination, source, mode);
		for (FileComparator f : fileList) {
			logger.trace(f.toString());
		}
		logger.info("Finished traversal " + mode.name() + ".");
		return fileList;
	}

	/**
	 * Traverses recursively the file tree with the action indicated by mode.
	 * 
	 * @param destination
	 *            The destination path to match and traverse
	 * @param source
	 *            The source path to match and traverse
	 * @param mode
	 *            The mode describing the action to execute.
	 */
	public void traverse(final File destination, final File source, final TraversalMode mode) {
		if (isSafe(destinationRoot, sourceRoot)) {
			final DirectoryComparator dc = new DirectoryComparator(destination, source, fileActionFilters, logger);
			final LinkedList<FileComparator> files = dc.result();
			fileList.addAll(filterForAction(files));
			execute(files, mode);
			for (final FileComparator fc : files) {
				if (fc.isDirectory()) {
					traverse(fc.fullDestinationPath(), fc.source, mode);
				}
			}
		}
	}

	/**
	 * Checks if both source and destination exist.
	 * 
	 * @param source
	 *            the source file
	 * @param destination
	 *            the destination file
	 * @return if both source and destination exist
	 */
	private boolean isSafe(final File destination, final File source) {
		boolean ret = false;
		if (source.exists()) {
			if (destination.exists()) {
				ret = true;
			} else {
				logger.error("Unable to access destination root " + globalDestinationRoot.getPath() + ".");
			}
		} else {
			logger.error("Unable to access source root " + globalSourceRoot.getPath() + ".");
		}
		return ret;
	}

	/**
	 * Returns only those {@link FileComparator} entries with work, which means
	 * either {@link FileComparator#isToCopy()} or
	 * {@link FileComparator#isToDelete()} resulted to <code>true</code>.
	 * 
	 * @param input
	 *            all file comparators to consider
	 * @return only those file comparators with either
	 *         {@link FileComparator#isToCopy()} or
	 *         {@link FileComparator#isToDelete()} resulted to <code>true</code>
	 */
	public LinkedList<FileComparator> filterForAction(final LinkedList<FileComparator> input) {
		final LinkedList<FileComparator> result = new LinkedList<FileComparator>();
		if (null != input) {
			for (final FileComparator fc : input) {
				if (fc.isToCopy() || fc.isToDelete()) {
					result.add(fc);
				}
			}
		}
		return result;
	}

	public void execute(final LinkedList<FileComparator> files, final TraversalMode mode) {
		for (final FileComparator fc : files) {
			switch(mode) {
			case COLLECT:
				if (fc.isToCopy() || fc.isToDelete()) {
					logger.debug(fc.toString());
				}
				break;
			case COPY:
			case DELETE:
			case BOTH:
				if (mode.isCopy && fc.isToCopy()) {
					logger.debug(fc.toString());
					final File destination = fc.fullDestinationPath();
					if (fc.isDirectory()) {
						destination.mkdir();
						// We don't copy recursively to give the filters of the comparator a chance.
					} else {
						copy(destination, fc.source);
					}
					destination.setLastModified(fc.source.lastModified());
				}
				if (mode.isDelete && fc.isToDelete()) {
					logger.debug(fc.toString());
					final File destination = fc.fullDestinationPath();
					delete(destination);
				}
				break;
			}
		}
	}

	/**
	 * Does a binary file copy.
	 * 
	 * @param src
	 *            the source file
	 * @param dest
	 *            the destination file
	 */
	public void copy(final File dest, final File src) {
		try {
			final BufferedInputStream bis = new BufferedInputStream(new FileInputStream(src));
			final BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(dest));
			// Working with a huge buffer should speed things up
			byte[] buffer = new byte[1024 * 1024];
			int read = 0;
			while ((read = bis.read(buffer)) != -1) {
				bos.write(buffer, 0, read);
			}
			bis.close();
			bos.close();
		} catch (final IOException e) {
			logger.info("Unable to copy", e);
		}
	}

	/**
	 * Deletes the file or directory (recursively).
	 * 
	 * @param file
	 *            the file to delete
	 */
    public void delete(final File file) {
    	if (null != file) {
			if (file.isDirectory()) {
				final File[] content = file.listFiles();
				if (null != content) {
					for (File f : content) {
						delete(f);
					}
				}
			}
			file.delete();
    	}
    }
}
