package de.ped.mp3sync.test;

import static org.junit.Assert.*;

import java.io.File;
import java.util.LinkedList;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import de.ped.mp3sync.FileComparator;
import de.ped.mp3sync.MP3Sync;
import de.ped.mp3sync.MP3Sync.TraversalMode;
import de.ped.tools.log.Level;

/**
 * Tests for the {@link MP3Sync} class.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MP3SyncTest {

	@Before
	public void setUp() throws Exception {
		final MP3Sync sync = new MP3Sync();
		sync.getLogger().setLogLevel(Level.ERROR);
		sync.delete(DESTINATION);
	}

	@After
	public void tearDown() throws Exception {
		final MP3Sync sync = new MP3Sync();
		sync.delete(DESTINATION);
	}

	private static final File SOURCE = new File("src/de/ped/mp3sync/test/data");
	private static final File DESTINATION = new File("src/de/ped/mp3sync/test/data2");

	@Test
	public void testRunCollect() {
		final MP3Sync sync = new MP3Sync();
		final LinkedList<FileComparator> fileList = sync.execute(DESTINATION, SOURCE, TraversalMode.COLLECT);
		assertEquals(5, fileList.size());
		for (final FileComparator fc : fileList) {
			assertEquals("c-", fc.toString().substring(0, 2));
		}
	}

	@Test
	public void testRunCopyAndDelete() {
		final MP3Sync sync = new MP3Sync();
		final File sourceFile = new File(SOURCE.getAbsolutePath() + File.separator + "folder.txt");
		{ 	// First do a copy.
			final LinkedList<FileComparator> fileList = sync.execute(DESTINATION, SOURCE, TraversalMode.COPY);
			assertEquals(5, fileList.size());
			for (final FileComparator fc : fileList) {
				assertEquals("c-", fc.toString().substring(0, 2));
			}
		}
		{ 	// Create one more file in the destination tree to be found as removable
			sync.copy(new File(DESTINATION.getAbsolutePath() + File.separator + "deletable.txt"), 
					  sourceFile);
		}
		{ 	// Delete the file created for deletion
			final LinkedList<FileComparator> fileList = sync.execute(DESTINATION, SOURCE, TraversalMode.DELETE);
			assertEquals(1, fileList.size());
		}
		{ 	// Is everything all right?
			final LinkedList<FileComparator> fileList = sync.execute(DESTINATION, SOURCE, TraversalMode.COLLECT);
			assertEquals(0, fileList.size());
			for (final FileComparator fc : fileList) {
				assertEquals("--", fc.toString().substring(0, 2));
			}
		}
	}

	@Test
	public void testRunCopyWithModificationDate() {
		final MP3Sync sync = new MP3Sync();
		{ 	// First do a copy.
			final LinkedList<FileComparator> fileList = sync.execute(DESTINATION, SOURCE, TraversalMode.COPY);
			assertEquals(5, fileList.size());
			for (final FileComparator fc : fileList) {
				assertEquals("c-", fc.toString().substring(0, 2));
			}
		}
		final long upper = FileComparator.TIMESTAMP_TOLERANCE + 1000L;
		checkForFileModification(sync, -upper, true);  // Destination is much older, so it should be copyable.
		checkForFileModification(sync, -2000L, false); // Destination is older, but not old enough.
		checkForFileModification(sync,     0L, false); // Destination is same age, so it should NOT be copyable.
		checkForFileModification(sync,  2000L, false); // Destination is newer, so it should NOT be copyable.
		checkForFileModification(sync,  upper, false); // Destination is much newer, so it should NOT be copyable.
	}

	public void checkForFileModification(final MP3Sync sync, final long modifier, final boolean shouldCopy) {
		// Modify the modification date. 
		final File sourceFile = new File(SOURCE.getAbsolutePath() + File.separator + "folder.txt");
		final boolean ok = new File(DESTINATION.getAbsolutePath() + File.separator + "folder.txt").setLastModified(sourceFile.lastModified() + modifier);
		assertTrue(ok);
		// Is everything all right?
		final LinkedList<FileComparator> fileList = sync.execute(DESTINATION, SOURCE, TraversalMode.COLLECT);
//		assertEquals(5, fileList.size());
		int counter = 0;
		for (final FileComparator fc : fileList) {
			if (fc.isToCopy()) {
				counter++;
			}
		}
		assertEquals((shouldCopy ? 1 : 0), counter);
	}
}
