package de.ped.pacman.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Polygon;
import java.awt.image.BufferedImage;
import java.util.ConcurrentModificationException;
import java.util.HashMap;

import javax.swing.JPanel;

import de.ped.pacman.logic.Being;
import de.ped.pacman.logic.Direction;
import de.ped.pacman.logic.Game;
import de.ped.pacman.logic.Labyrinth;
import de.ped.pacman.logic.PointsIndicator;
import de.ped.pacman.logic.Sprite;
import de.ped.pacman.logic.WallMaskFilter;
import de.ped.pacman.logic.Game.Mode;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.log.Logger;

/**
 * The drawing area for the labyrinth and all.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PacmanMainCanvas extends JPanel {
	private static final long serialVersionUID = 1L;

	private static final Logger logger = Logger.getRootLogger();

	/**
	 * Darkens the given color
	 * 
	 * @param color
	 *            the original color
	 * @param howmuch
	 *            a negative numer brightens, a positive number darkens
	 * @return the darkened or brightened color
	 */
	public static Color darken(final Color color, final int howmuch) {
		Color ret = color;
		if (howmuch < 0) {
			for (int i = 0; i > howmuch; i--) {
				ret = ret.brighter();
			}
		} else {
			for (int i = 0; i < howmuch; i++) {
				ret = ret.darker();
			}
		}
		return ret;
	}

	/**
	 * Darkens the given color, based on game mode and step.
	 * 
	 * @param color
	 *            the original color
	 * @param mode
	 *            the mode
	 * @param step
	 *            the step
	 * @return the darkened or brightened color
	 */
	protected static Color darken(final Color color, final Mode mode, final int step) {
		Color ret = color;
		switch (mode) {
		case LEVELINTRO:
			ret = darken(ret, mode.duration - step);
			break;
		case LEVELDONE:
		case OUTRO:
			ret = darken(ret, step);
			break;
		default:
			// Nothing to do
			break;
		}
		return ret;
	}

	/**
	 * Darkens the given color, based on game mode and step, as defined in the game state
	 * 
	 * @param color
	 *            the original color
	 * @param gamestate
	 *            the game state
	 * @return the darkened or brightened color
	 */
	public static Color darken(final Color color, final GlobalGameStateForDrawing gamestate) {
		return darken(color, gamestate.getMode(), gamestate.getStep());
	}

	/**
	 * Keeps global information about those parts of the game state that have influence on how to draw things. 
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class GlobalGameStateForDrawing {
		public static boolean isToDrawFrightened(final int remainingTicksPillIsActive) {
			boolean ret = false;
			if (Game.PILL_NOT_ACTIVE == remainingTicksPillIsActive) {
				ret = false;
			} else {
				if (remainingTicksPillIsActive < FRIGHTENED_BLINKS * FRIGHTENED_BLINK_STEPS) {
					final int ticksmod = FRIGHTENED_BLINK_STEPS - 1 - (remainingTicksPillIsActive % FRIGHTENED_BLINK_STEPS);
					final int ticksdiv = FRIGHTENED_BLINKS - (remainingTicksPillIsActive / FRIGHTENED_BLINK_STEPS);
					if (ticksmod < FRIGHTENED_PHASE_STEPS * ticksdiv) {
						ret = false;
					} else {
						ret = true;
					}
				} else {
					ret = true;
				}
			}
			return ret;
		}

		protected int level;
		protected Mode mode;
		protected int step;
		protected boolean isToDrawFrightened;

		public int getLevel() {
			return level;
		}

		public Mode getMode() {
			return mode;
		}

		public int getStep() {
			return step;
		}

		public boolean isToDrawFrightened() {
			return isToDrawFrightened;
		}

		public void set(final Game game) {
			level = game.getLevel();
			mode = game.getMode();
			step = game.getStep();
			isToDrawFrightened = isToDrawFrightened(game.remainingTicksPillIsActive());
		}
	}

	private final PacmanMainWindow parent;

	private final WallMaskFilter wallMaskFilter = new WallMaskFilter();

	private Font font;
	private GlobalGameStateForDrawing gamestate = new GlobalGameStateForDrawing();

	public PacmanMainCanvas(final PacmanMainWindow parent) {
		this.parent = parent;
		setBackground(Color.black);
	}

	@Override
	public Dimension getPreferredSize() {
		return getPreferredSize(new Dimension(16, 12));
	}

	@Override
	public Dimension getMinimumSize() {
		return getPreferredSize(new Dimension(8, 6));
	}

	/** Must be multiple of 2 to work properly. */
	private static final int PHASES = 4;

	private static final int PHASES2 = PHASES / 2;
	private static final int PHASES4 = PHASES2 / 2;

	/** The size of one labyrinth field in pixels */
	private Dimension stretch = new Dimension(16, 12);

	/** The screen offset in pixels of the top left pixel of the top left labyrinth field */
	private Dimension offset = new Dimension(0, 0);

	protected int maskWalls(final Logger logger, final int x, final int y) {
		int ret = 0;
		final Labyrinth labyrinth = parent.getCurrentModel().getBoard();
		for (final int[] maskData : WallMaskFilter.MASK_WALL_POSITIONS) {
			final int nx = x + maskData[0];
			final int ny = y + maskData[1];
			if (labyrinth.bounds(nx, ny)) {
				final char c = labyrinth.get(nx, ny);
				if ((Labyrinth.WALL == c) || (Labyrinth.DOOR == c)) {
					ret |= maskData[2];
				}
			}
		}
		return ret;
	}

	protected void drawPacmanArc(
			final Graphics g, 
			final int x, 
			final int y, 
			final int direction, 
			final int arcAngle) {
		g.setColor(Color.YELLOW);
		final Direction dir = Direction.DIRECTIONS[direction];
		final int arcStart = arcAngle + dir.angle;
		final int arcSize = 360 - 2 * arcAngle;
		g.fillArc(x, y, stretch.width, stretch.height, arcStart, arcSize);
		g.setColor(Color.BLACK);
		int eyesX = 0;
		int eyesY = 0;
		switch (direction) {
		case Direction.UP:
			eyesX = stretch.width / 4;
			eyesY = stretch.height / 2;
			break;
		case Direction.DOWN:
			eyesX = stretch.width / 4;
			eyesY = stretch.height / 2 - stretch.height / 8;
			break;
		case Direction.LEFT:
			eyesX = stretch.width / 2;
			eyesY = stretch.height / 4;
			break;
		case Direction.RIGHT:
			eyesX = stretch.width / 2 - stretch.width / 8;
			eyesY = stretch.height / 4;
			break;
		default:
		}
		g.fillRect(x + eyesX, y + eyesY, stretch.width / 8, stretch.height / 8);
	}

	public void drawPacman(final Graphics g, final int x, final int y, final int direction, final int mouthOpen) {
		final int arcAngle = 45 - (int) ((35.0 * mouthOpen) / PHASES2);
		drawPacmanArc(g, x, y, direction, arcAngle);
	}

	public void drawPacman(
			final Graphics g, 
			final int x, 
			final int y, 
			final int direction, 
			final int phase, 
			final GlobalGameStateForDrawing gamestate) {
		g.setColor(Color.YELLOW);
		final Mode mode = gamestate.getMode();
		switch (mode) {
		case LIFELOST:
			final Direction[] dirs = Direction.DIRECTIONS;
			drawPacman(g, x, y, gamestate.getStep() % dirs.length, 0);
			break;
		case GAMEOVER:
			drawPacmanArc(g, x, y, Direction.UP, (gamestate.getStep() + 1) * (120 / mode.duration));
			break;
		case WAITAFTERGAME:
		case OUTRO:
		case GAMEFINISHED:
			// No pacman left to draw
			break;
		default:
			final int safePhase = MathUtil.modulo(phase, PHASES);
			final int mouthOpen = (safePhase <= PHASES2 ? safePhase : PHASES - safePhase);
			drawPacman(g, x, y, direction, mouthOpen);
			break;
		}
	}

	public void drawOrange(
			final Graphics g, 
			int x, 
			int y, 
			final int direction, 
			final int phase) {
		final int safePhase = MathUtil.modulo(phase, PHASES);
		final int mouthOpen = (safePhase <= PHASES2 ? safePhase : PHASES - safePhase);
		final int phaseX = mouthOpen - PHASES4;
		final int safePhase2 = MathUtil.modulo(phase, PHASES2);
		final int phaseY = (safePhase2 <= PHASES4 ? safePhase2 : PHASES2 - safePhase2);
		x += phaseX * stretch.width / 16;
		y -= phaseY * stretch.height / 16;
		g.setColor(Color.ORANGE);
		g.fillOval(x, y, stretch.width, stretch.height);
		g.setColor(Color.orange.darker());
		final int[][] dots = new int[][]{ {8, 8}, {12, 12}, {9, 13}, {13, 9}, {12, 4}, {4, 12}, {3, 7}, {7, 3}, };
		for (int[] dot : dots) {
			g.fillOval(x + stretch.width * dot[0] / 16, y + stretch.height * dot[1] / 16, stretch.width / 8, stretch.height / 8);
		}
	}

	public void drawBanana(
			final Graphics g, 
			int x, 
			int y, 
			final int direction, 
			final int phase) {
		final int safePhase = MathUtil.modulo(phase, PHASES);
		final int mouthOpen = (safePhase <= PHASES2 ? safePhase : PHASES - safePhase);
		final int phaseX = mouthOpen - PHASES4;
		final int safePhase2 = MathUtil.modulo(phase, PHASES2);
		final int phaseY = (safePhase2 <= PHASES4 ? safePhase2 : PHASES2 - safePhase2);
		x += phaseX * stretch.width / 16;
		y -= phaseY * stretch.height / 16;
		g.setColor(Color.YELLOW);
		final int[][] dots = new int[][]{ {1, 1}, {7, 9}, {15, 12}, {15, 15}, {10, 15}, {4, 12}, {1, 6}, };
		Polygon p = new Polygon();
		for (int[] dot : dots) {
			p.addPoint(x + dot[0] * stretch.width / 16, y + dot[1] * stretch.height / 16);
		}
		g.fillPolygon(p);
	}

	public void drawCherry(
			final Graphics g, 
			int x, 
			int y, 
			final int direction, 
			final int phase) {
		final int safePhase = MathUtil.modulo(phase, PHASES);
		final int mouthOpen = (safePhase <= PHASES2 ? safePhase : PHASES - safePhase);
		final int phaseX = mouthOpen - PHASES4;
		final int safePhase2 = MathUtil.modulo(phase, PHASES2);
		final int phaseY = (safePhase2 <= PHASES4 ? safePhase2 : PHASES2 - safePhase2);
		x += phaseX * stretch.width / 16;
		y -= phaseY * stretch.height / 16;
		g.setColor(Color.GREEN);
		final int[][] dots = new int[][]{ 
				{10, 1}, {9, 3}, {13, 2}, {15, 5}, {10, 5}, {9, 4}, // right leaf 
				{8, 6}, {10, 11}, {9, 11}, {7, 7}, {5, 11}, {4, 11}, {7, 6}, 
				{9, 1}, };
		Polygon p = new Polygon();
		for (int[] dot : dots) {
			p.addPoint(x + dot[0] * stretch.width / 16, y + dot[1] * stretch.height / 16);
		}
		g.fillPolygon(p);
		g.setColor(Color.RED);
		g.fillOval(x + stretch.width * 9 / 16, y + stretch.height * 10 / 16, 
				stretch.width * 5 / 16, stretch.height * 6 / 16);
		g.fillOval(x + stretch.width * 1 / 16, y + stretch.height * 10 / 16, 
				stretch.width * 5 / 16, stretch.height * 6 / 16);
	}

	public void drawFruit(
			final Graphics g, 
			final int x, 
			final int y, 
			final int direction, 
			final int phase, 
			final GlobalGameStateForDrawing gamestate) {
		final int level = gamestate.getLevel();
		switch(level % 3) {
		case 0:
			drawCherry(g, x, y, direction, phase);
			break;
		case 1:
			drawOrange(g, x, y, direction, phase);
			break;
		case 2:
			drawBanana(g, x, y, direction, phase);
			break;
		default:
			throw new IllegalStateException("Negative level? " + level);
		}
	}

	public static final Color[] MONSTER_COLOR = new Color[]{ Color.RED, Color.MAGENTA, Color.CYAN, Color.ORANGE};

	public void drawOpponent(
			final Graphics g, 
			int x, 
			int y, 
			final int direction, 
			final int phase, 
			final int opponent, 
			final boolean isGhost, 
			final GlobalGameStateForDrawing gamestate) {
		final boolean isFrightened = gamestate.isToDrawFrightened() && !isGhost;
		// Body
		if (!isGhost) {
			g.setColor(isFrightened ? Color.BLUE : MONSTER_COLOR[opponent]);
			final int safePhase = MathUtil.modulo(phase, PHASES);
			final int mouthOpen = (safePhase <= PHASES2 ? safePhase : PHASES - safePhase);
			g.fillArc(x, y, stretch.width, stretch.height, 0, 180);
			g.fillRect(x, y + stretch.height / 2, stretch.width, stretch.height * 3 / 8);
			// Body
			for (int i = 0; i < 4; i++) {
				g.fillRect(
						x + stretch.width * i / 4 + stretch.width / 8 * mouthOpen / PHASES2, 
						y + stretch.height * 6 / 8, 
						stretch.width / 8, stretch.height / 4);
			}
		}
		// White of eyes
		g.setColor(Color.WHITE);
		int eyesWhitePosX;
		int eyesWhitePosY;
		int eyesWhiteSizeX;
		int eyesWhiteSizeY;
		if (isFrightened) {
			eyesWhitePosX = stretch.width / 4;
			eyesWhitePosY = stretch.height / 4;
			eyesWhiteSizeX = stretch.width * 1 / 8;
			eyesWhiteSizeY = stretch.height * 1 / 8;
		} else {
			eyesWhitePosX = stretch.width / 8;
			eyesWhitePosY = stretch.height / 8;
			eyesWhiteSizeX = stretch.width * 3 / 8;
			eyesWhiteSizeY = stretch.height * 3 / 8;
			x += Direction.DIRECTIONS[direction].signs[0];
			y += Direction.DIRECTIONS[direction].signs[1];
		}
		g.fillOval(x + eyesWhitePosX, 
				   y + eyesWhitePosY, eyesWhiteSizeX, eyesWhiteSizeY);
		g.fillOval(x + stretch.width - eyesWhiteSizeX - eyesWhitePosX, 
				   y + eyesWhitePosY, eyesWhiteSizeX, eyesWhiteSizeY);
		if (isFrightened) {
			// Frown mouth
			int mouth1X = x + stretch.width / 8;
			int mouth1Y = y + stretch.height * 5 / 8;
			for (int i = 1; i < 7; i++) {
				int mouth2X = mouth1X + stretch.width / 8;
				int mouth2Y = mouth1Y;
				if (0 == i % 2) {
					mouth2Y++;
				} else {
					mouth2Y--;
				}
				g.drawLine(mouth1X, mouth1Y, mouth2X, mouth2Y);
				mouth1X = mouth2X;
				mouth1Y = mouth2Y;
			}
		} else {
			// pupils
			final int eyesBlackPosX = stretch.width / 8;
			final int eyesBlackPosY = stretch.height / 8;
			final int eyesBlackSizeX = stretch.width * 3 / 16;
			final int eyesBlackSizeY = stretch.height * 3 / 16;
			g.setColor(Color.BLACK);
			g.fillRect(x + eyesWhitePosX + eyesBlackPosX 
							+ stretch.width / 16 * Direction.DIRECTIONS[direction].signs[0], 
					   y + eyesWhitePosY + eyesBlackPosY 
					   		+ stretch.height / 16 * Direction.DIRECTIONS[direction].signs[1], 
					   eyesBlackSizeX, eyesBlackSizeY);
			g.fillRect(x + stretch.width - eyesWhiteSizeX - eyesWhitePosX + eyesBlackPosX 
							+ stretch.width / 16 * Direction.DIRECTIONS[direction].signs[0], 
					   y + eyesWhitePosY + eyesBlackPosY 
					   		+ stretch.height / 16 * Direction.DIRECTIONS[direction].signs[1], 
					   eyesBlackSizeX, eyesBlackSizeY);
		}
	}

	public void drawDirOptions(
			final Graphics g, 
			int x, 
			int y, 
			final int[] dirOptions, 
			final GlobalGameStateForDrawing gamestate) {
		g.setColor(Color.GRAY);
		x += stretch.width / 2;
		y += stretch.height / 2;
		for (int dir = 0; dir < dirOptions.length; dir++) {
			Direction d = Direction.DIRECTIONS[dir];
			final int x2 = x + dirOptions[dir] * d.signs[0] * 5;
			final int y2 = y + dirOptions[dir] * d.signs[1] * 5;
			g.drawLine(x + 1, y + 0, x2 + 1, y2 + 0);
			g.drawLine(x - 1, y + 0, x2 - 1, y2 + 0);
			g.drawLine(x + 0, y + 1, x2 + 0, y2 + 1);
			g.drawLine(x + 0, y - 1, x2 + 0, y2 - 1);
		}
	}

	public void drawPointsIndicator(
			final Graphics g, 
			final int x, 
			final int y, 
			final String text, 
			final GlobalGameStateForDrawing gamestate) {
		g.setColor(Color.YELLOW);
		g.setFont(font);
		g.drawString(text, x, y + stretch.height);
	}

	public void drawDot(
			final Graphics g, 
			final int x, 
			final int y, 
			final GlobalGameStateForDrawing gamestate) {
		g.setColor(darken(Color.YELLOW, gamestate));
		g.drawRect(x + stretch.width * 7 / 16, y + stretch.height * 7 / 16, stretch.width / 8, stretch.height / 8);
	}

	public void drawPill(
			final Graphics g, 
			final int x, 
			final int y, 
			final GlobalGameStateForDrawing gamestate) {
		g.setColor(darken(Color.YELLOW, gamestate));
		g.fillOval(x + stretch.width / 8, y + stretch.height / 8, stretch.width * 3 / 4, stretch.height * 3 / 4);
	}

	/**
	 * Internal class to draw a wall correctly.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class WallImagesKey {
		public final int wallMask;
		public final boolean isDoor;

		public WallImagesKey(final boolean isDoor, final int wallMask) {
			this.isDoor = isDoor;
			this.wallMask = wallMask;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + (isDoor ? 1231 : 1237);
			result = prime * result + wallMask;
			return result;
		}

		@Override
		public boolean equals(final Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			final WallImagesKey other = (WallImagesKey) obj;
			if (isDoor != other.isDoor)
				return false;
			if (wallMask != other.wallMask)
				return false;
			return true;
		}

		@Override
		public String toString() {
			return ("(" + Integer.toBinaryString(wallMask) + (isDoor ? "D" : "W") + ")");
		}
	}

	private HashMap<WallImagesKey, Image> wallImages = new HashMap<WallImagesKey, Image>();

	public void drawWall(
			final Graphics g, 
			final int x, 
			final int y, 
			int surroundingWallsMask, 
			final boolean isDoor, 
			final GlobalGameStateForDrawing gamestate) {
		WallImagesKey key = new WallImagesKey(isDoor, surroundingWallsMask);
		Image img = wallImages.get(key);
		if (null == img) {
			img = new BufferedImage(stretch.width, stretch.height, BufferedImage.TYPE_INT_ARGB);
			final Graphics ig = img.getGraphics();
			Color color;
			if (isDoor) {
				color = darken((Color.PINK), gamestate);
			} else {
				color = darken((Color.BLUE), gamestate);
			}
			ig.setColor(color);
			surroundingWallsMask = wallMaskFilter.filterMaskWalls(surroundingWallsMask);
			if (0 == surroundingWallsMask) { // Just a column, not a wall
				ig.drawRect(stretch.width * 3 / 8, stretch.height * 3 / 8, stretch.width / 4, stretch.height / 4);
			} else {
				for (int i = 0; i < WallMaskFilter.WALL_PARTS_TO_PAINT; i++) {
					if (0 != (surroundingWallsMask & WallMaskFilter.MASK_WALL_POSITIONS[i][2])) {
						ig.fillRect(stretch.width  * 3 / 8 + stretch.width  / 4 * WallMaskFilter.MASK_WALL_POSITIONS[i][0], 
									stretch.height * 3 / 8 + stretch.height / 4 * WallMaskFilter.MASK_WALL_POSITIONS[i][1], 
									stretch.width / 4, stretch.height / 4);
					}
				}
			}
			if (gamestate.getMode() == Mode.RUNNING) {
				wallImages.put(key, img);
			}
		}
//		g.drawRect(x, y, STRETCH.width, STRETCH.height);
		g.drawImage(img, x, y, null);
	}

	public void draw(
			final Graphics g, 
			final int x, 
			final int y, 
			final char c, 
			final GlobalGameStateForDrawing gamestate) {
		final int xp = offset.width + x * stretch.width;
		final int yp = offset.height + y * stretch.height;
		switch (c) {
		case Labyrinth.DOT:
			drawDot(g, xp, yp, gamestate);
			break;
		case Labyrinth.PILL:
			drawPill(g, xp, yp, gamestate);
			break;
		case Labyrinth.WALL:
			drawWall(g, xp, yp, maskWalls(logger, x, y), false, gamestate);
			break;
		case Labyrinth.DOOR:
			drawWall(g, xp, yp, maskWalls(logger, x, y), true, gamestate);
			break;
		case Labyrinth.EMPTY:
		case Labyrinth.DEN:
			break;
		default:
			logger.debug("Unknown labyrinth character: " + c);
		}
	}

	public void draw(final Graphics g, final Sprite sprite, final GlobalGameStateForDrawing gamestate) {
		final int xp = offset.width  + (int)Math.round(sprite.getX() * stretch.width);
		final int yp = offset.height + (int)Math.round(sprite.getY() * stretch.height);
		final int phase = parent.getCurrentModel().getStep() % PHASES;
		if (sprite instanceof Being) {
			final Being being = (Being) sprite;
			if (being.isPacPan()) {
				drawPacman(g, xp, yp, being.getDirection(), phase, gamestate);
			} else if (being.isFruit()) {
				drawFruit(g, xp, yp, being.getDirection(), phase, gamestate);
			} else if (being.isOpponent()) {
				if (null == being.getId()) {
					throw (new IllegalStateException("Being without ID"));
				}
				int id = Integer.parseInt(being.getId()) - 1;
				drawOpponent(g, xp, yp, being.getDirection(), phase, id, being.isGhost(), gamestate);
//        		Opponent opp = (Opponent)being;
//        		drawDirOptions(g, xp, yp, opp.getDirectionRatings(), gamestate);
			}
		} else if (sprite instanceof PointsIndicator) {
			PointsIndicator pi = (PointsIndicator) sprite;
			drawPointsIndicator(g, xp, yp, pi.getText(), gamestate);
		}
	}

	public void drawText(final Graphics g, final int x, final int y, final String text) {
		int xp = offset.width + x * stretch.width + stretch.width / 4;
		final int yp = offset.height + y * stretch.height;
		g.setColor(Color.YELLOW);
		g.setFont(font);
		for (int i = 0; i < text.length(); i++) {
			g.drawString(text.substring(i, i + 1), xp, yp);
			xp += stretch.width;
		}
	}

	public static final int FRIGHTENED_BLINKS = 4;
	public static final int FRIGHTENED_BLINK_STEPS = 5;
	public static final int FRIGHTENED_PHASE_STEPS = 1;

	protected Dimension getPreferredSize(final Dimension preferredStretch) {
		PlayfieldDimension lcs = getLabyrinthCharSize();
		return new Dimension(lcs.getWidth() * preferredStretch.width, lcs.getHeight() * preferredStretch.height);
	}

	/**
	 * @return the size of the labyrinth plus some room for score, lives etc.
	 */
	protected PlayfieldDimension getLabyrinthCharSize() {
		PlayfieldDimension ret = new PlayfieldDimension(28, 31);
		final Game model = parent.getCurrentModel();
		if (null != model) {
			final Labyrinth labyrinth = model.getBoard();
			final PlayfieldDimension d = labyrinth.getSize();
			ret.setLocation(d.getWidth(), d.getHeight() + 2);
		}
		return ret;
	}

	@Override
	public void paintComponent(final Graphics g) {
		super.paintComponent(g);
		final Game model = parent.getCurrentModel();
		gamestate.set(model);
		if (null != model) {
			final Labyrinth labyrinth = model.getBoard();
			final PlayfieldDimension ls = labyrinth.getSize();
			final PlayfieldDimension lcs = getLabyrinthCharSize();
			final Dimension s = getSize();
			stretch.width = s.width / lcs.getWidth();
			stretch.height = s.height / lcs.getHeight();
			offset.width  = Math.max(0, (s.width  - lcs.getWidth()  * stretch.width ) / 2);
			offset.height = Math.max(0, (s.height - lcs.getHeight() * stretch.height) / 2);
			font = new Font(Font.MONOSPACED, Font.BOLD, Math.min(stretch.width, stretch.height * 4 / 3));
			wallImages.clear(); // possible Optimization: Do this only when the screen size has changed
			drawLabyrinth(g, labyrinth);
			try {
				for (final Being being : labyrinth.getBeings()) {
					draw(g, being, gamestate);
				}
				for (final Sprite sprite : model.getSprites()) {
					draw(g, sprite, gamestate);
				}
				final String pts = Integer.toString(model.getPoints());
				drawText(g, 5 - pts.length(), ls.getHeight() + 1, pts);
//				final String steps = Integer.toString(model.getSprites().size());
//				drawText(g, d.width - 3, d.height + 1, steps);
				for (int i = 1; i < model.getLives(); i++) {
					drawPacman(g, 
							offset.width + (ls.getWidth() / 2 - 1 + i) * stretch.width, 
							offset.height + ls.getHeight() * stretch.height, Direction.LEFT, 0);
				}
	//    		final String remainingDots = Integer.toString(model.getBoard().getRemainingDots());
	//    		drawText(g, d.width / 2, d.height + 2, remainingDots);
			} catch (final ConcurrentModificationException e) {
				// Nothing to do
			}
		}
	}

	private void drawLabyrinth(final Graphics g, final Labyrinth labyrinth) {
		final PlayfieldDimension d = labyrinth.getSize();
		for (int y = 0; y < d.getHeight(); y++) {
			for (int x = 0; x < d.getWidth(); x++) {
				draw(g, x, y, labyrinth.get(x, y), gamestate);
			}
		}
	}

	public void step() {
		repaint();
	}
}
