package de.ped.pacman.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.util.List;
import java.util.Locale;

import javax.sound.sampled.Clip;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JScrollPane;

import de.ped.pacman.logic.Direction;
import de.ped.pacman.logic.Game;
import de.ped.pacman.logic.GameProperties;
import de.ped.pacman.logic.PacmanLevelSet;
import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.Assert;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.FinishableThread;
import de.ped.tools.MathUtil;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.CheckBoxMenuItemAction;
import de.ped.tools.gui.GameThread;
import de.ped.tools.gui.PedAction;
import de.ped.tools.gui.UserGameActionInput;
import de.ped.tools.log.Logger;
import de.ped.tools.sound.SoundClip;

/**
 * The main window of the application.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PacmanMainWindow extends ApplicationMainWindow {
	private static final long serialVersionUID = 1L;

	public PacmanMainWindow(final ApplicationEnvironment appEnv) {
		super(appEnv);
		final String levelSet = Pacman.getInstance().findParameter(ApplicationEnvironment.PARAM_KEY_GAME).getValue();
		sounds = new PacmanSounds(getApplicationEnvironment().resources());

		this.model = new Game(levelSet);
		this.titleModel = new Game(PacmanLevelSet.TITLE.name);
		worldCanvas = new PacmanMainCanvas(this);
		final JScrollPane worldScrollPane = new JScrollPane(worldCanvas);
		getContentPane().add(worldScrollPane, BorderLayout.CENTER);
		if (getApplicationEnvironment().isSilent()) {
			eventSoundOnOffAction.setOn(false);
		}
		setJMenuBar(createMenuBar());
		pack(); // Resize the window to the preferred size of its components.
		switchLocale(getMessages().getCurrentLocale());
		addKeyListener(this);
		pauseOrResume(true);
	}

	private CheckBoxMenuItemAction runAction             = new CheckBoxMenuItemAction(false, "Action.Pause", this, ResourceFinder.Folder.COMMON);
	private CheckBoxMenuItemAction eventSoundOnOffAction = new CheckBoxMenuItemAction(true, "Action.SoundOnOff", this, ResourceFinder.Folder.COMMON);

	@Override
	public void pedAbstractActionPerformed(final PedAction action, final ActionEvent event) {
		super.pedAbstractActionPerformed(action, event);
		if (action == runAction) {
			togglePause();
		} else if (action == eventSoundOnOffAction) {
			// Nothing to do
		}
	}

	private JMenuBar createMenuBar() {
		final JMenuBar menuBar = createJMenuBar();
		final JMenu menuFile = createJMenu("Menu.File");
		createActionItem(menuFile, startNewGameAction);
		menuFile.addSeparator();
		menuFile.add(createLocalesMenu());
		menuFile.addSeparator();
		createActionItem(menuFile, quitAction);
		menuBar.add(menuFile);
		final JMenu menuRun = createJMenu("Menu.Run");
		createActionItem(menuRun, runAction);
		createActionItem(menuRun, eventSoundOnOffAction);
		menuBar.add(menuRun);
		final JMenu menuHelp = createJMenu("Menu.Help");
		createActionItem(menuHelp, helpAction);
		createActionItem(menuHelp, aboutAction);
		menuBar.add(menuHelp);
		try {
			menuBar.setHelpMenu(menuHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}
		return menuBar;
	}

	private NewGameSelectionDialog getOrCreateNewGameSelectionDialog() {
		if (null == newGameSelectionDialog) {
			newGameSelectionDialog = new NewGameSelectionDialog(this);
		}
		return newGameSelectionDialog;
	}

	public UserGameActionInput findUserGameActionInput(final int playerIdx, final int playerAction) {
		return UserGameActionInput.find(userGameActionInputs, playerIdx, playerAction);
	}

	@Override
	protected void startNewGame() {
		final boolean wasRunning = isRunning();
		pauseOrResume(false);
		try {
			Thread.sleep(0);
		} catch (final InterruptedException e) {
			// Nothing to do - that's what we intend to do
		}
		final NewGameSelectionDialog dialog = getOrCreateNewGameSelectionDialog();
		dialog.enterDialog(model.getProperties());
		logger.debug("New game selection dialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (dialog.wasAcceptedDialogChanges()) {
			dialog.readUIFieldsIntoProperties();
			setModel(dialog.getProperties());
			repaint();
			pauseOrResume(true);
		} else {
			pauseOrResume(wasRunning);
		}
	}

	public void setModel(final GameProperties props) {
		model.initGame(props);
		isTitle = false;
	}

	@Override
	public void switchLocale(final Locale locale) {
		super.switchLocale(locale);
		userGameActionInputs = new UserGameActionInput[] {
				new UserGameActionInput(getMessages(), "Game.Player0L.KeyStroke", KeyEvent.VK_LEFT, 0, Direction.LEFT),
				new UserGameActionInput(getMessages(), "Game.Player0R.KeyStroke", KeyEvent.VK_RIGHT, 0,
						Direction.RIGHT),
				new UserGameActionInput(getMessages(), "Game.Player0U.KeyStroke", KeyEvent.VK_UP, 0, Direction.UP),
				new UserGameActionInput(getMessages(), "Game.Player0D.KeyStroke", KeyEvent.VK_DOWN, 0,
						Direction.DOWN), };
		newGameSelectionDialog = null;
		repaint();
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		doc.setTitle(getMessages().getString("HelpWindow.Title"));
		doc.setHeading("Help");
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");
		buildMenuDescriptions(doc, getJMenuBar());
		buildInputDescriptions(doc, userGameActionInputs, false);
		buildPropertyDescriptions(doc, getOrCreateNewGameSelectionDialog());
		doc.endDocument();
	}

	@Override
	protected String getKeyboardActionDescriptionText(final UserGameActionInput ugai) {
		String strActionKey = "";
		switch (ugai.playerAction) {
		case Direction.LEFT:
			strActionKey = "Game.PlayerAction.L.Text";
			break;
		case Direction.RIGHT:
			strActionKey = "Game.PlayerAction.R.Text";
			break;
		case Direction.UP:
			strActionKey = "Game.PlayerAction.U.Text";
			break;
		case Direction.DOWN:
			strActionKey = "Game.PlayerAction.D.Text";
			break;
		default:
			Assert.assertTrue(true);
		}
		return getMessages().getString(strActionKey);
	}

	public synchronized Game getModel() {
		return model;
	}

	@Override
	public boolean isRunning() {
		return isRunningRun();
	}

	@Override
	public FinishableThread createRunThread() {
		return new GameThread(this);
	}

	@Override
	public void pauseOrResume(final boolean isToResume) {
		pauseOrResumeRun(isToResume);
	}

	private void togglePause() {
		pauseOrResume(!isRunning());
		worldCanvas.step();
	}

	Game getCurrentModel() {
		return (isTitle ? titleModel : model);
	}

	@Override
	public void step() {
		final Game currentModel = getCurrentModel();
		final List<Game.Event> events = currentModel.step();
		handleGameEvents(currentModel, events);
		if (Game.Mode.GAMEFINISHED == currentModel.getMode()) {
			final GameProperties props = new GameProperties();
			props.setLevelSet(PacmanLevelSet.TITLE);
			props.setLevel(MathUtil.random(3));
			titleModel.initGame(props);
			isTitle = true;
		}
		worldCanvas.step();
	}

	@Override
	public synchronized long getSleepDelay() {
		int ret = Game.GAME_STEP_DELAY;
		final Game m = getCurrentModel();
		if (null != m) {
			ret = m.getSleepDelay();
		}
		return ret;
	}

	@Override
	public void keyPressed(final KeyEvent keyEvent) {
		logger.trace("keyPressed(" + keyEvent.getKeyCode() + ")");
		if (!isTitle) {
			final Game game = getCurrentModel();
			boolean isToRepaint = false;
			for (final UserGameActionInput tsv : userGameActionInputs) {
				if (keyEvent.getKeyCode() == tsv.keyCode) {
					game.suggestUserGameAction(tsv.playerIdx, tsv.playerAction);
					isToRepaint = true;
				}
			}
			if (isToRepaint) {
				worldCanvas.repaint();
			}
		}
	}

	private boolean handleGameEvents(final Game game, final List<Game.Event> events) {
		boolean wasRepaintedAll = false;
		if (null != events) {
			for (final Game.Event event : events) {
				wasRepaintedAll = wasRepaintedAll || handleGameEvent(game, event);
			}
		}
		return wasRepaintedAll;
	}

	private boolean handleGameEvent(final Game game, final Game.Event event) {
		if ((null != game) && (PacmanLevelSet.TITLE != game.getProperties().getLevelSet())) {
			playSound(event);
		}
		return false;
	}

	private void playSound(final Game.Event event) {
		if (eventSoundOnOffAction.isOn()) {
			logger.trace("playSound(" + event + ")");
			final SoundClip sound = sounds.findSound(event);
			switch (event) {
			case PILL_ENDED:
				SoundClip.stop(pillSound);
				break;
			case FRUIT_ENDED:
				SoundClip.stop(fruitSound);
				break;
			case EATING_PILL:
				if (null != sound) {
					sound.setConcurrencyMode(SoundClip.ConcurrencyMode.ALL);
					try {
						pillSound = sound.getOrCreateClip();
						pillSound.loop(Clip.LOOP_CONTINUOUSLY);
						pillSound.start();
					} catch (final NullPointerException ex) {
						logger.warn("Unexpected exception playing sound for event " + event.toString(), ex);
					}
				}
				break;
			case FRUIT_ENTERED:
				if (null != sound) {
					sound.setConcurrencyMode(SoundClip.ConcurrencyMode.ALL);
					try {
						fruitSound = sound.getOrCreateClip();
						fruitSound.loop(Clip.LOOP_CONTINUOUSLY);
						fruitSound.start();
					} catch (final NullPointerException ex) {
						logger.warn("Unexpected exception playing sound for event " + event.toString(), ex);
					}
				}
				break;
			default:
				if (null != sound) {
					sound.setConcurrencyMode(SoundClip.ConcurrencyMode.ALL);
					currentClips.play(sound, event.toString());
				}
			}
		}
	}

	private Logger logger = Logger.getRootLogger();

	private final PacmanSounds sounds;
	private Clip pillSound;
	private Clip fruitSound;
	private final SoundClip.RingBuffer currentClips = new SoundClip.RingBuffer(5);

	protected final Game model;
	protected final Game titleModel;
	private boolean isTitle = true;

	private PacmanMainCanvas worldCanvas;

	private UserGameActionInput[] userGameActionInputs;

	private NewGameSelectionDialog newGameSelectionDialog;
}
