package de.ped.pacman.gui;

import java.util.HashMap;

import javax.sound.sampled.AudioInputStream;

import de.ped.pacman.logic.Game;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationSounds;
import de.ped.tools.sound.AddedWave;
import de.ped.tools.sound.AmplitudeModulationWave;
import de.ped.tools.sound.ConcatWave;
import de.ped.tools.sound.ConstantWave;
import de.ped.tools.sound.MonoWave;
import de.ped.tools.sound.MultipliedWave;
import de.ped.tools.sound.Note;
import de.ped.tools.sound.SinusWave;
import de.ped.tools.sound.SoundClip;
import de.ped.tools.sound.SquareWave;
import de.ped.tools.sound.DynamicSoundMonoWaveCreator;
import de.ped.tools.sound.DynamicSoundMonoWaveCreator.Sound;
import de.ped.tools.sound.EnvelopeWave;
import de.ped.tools.sound.TriangleWave;
import de.ped.tools.sound.Tuning;

/**
 * Bundles all the sounds for Pacman.
 * 
 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PacmanSounds extends ApplicationSounds {
	public PacmanSounds(final ResourceFinder resources) {
		super(resources);
	}

	/**
	 * Returns the sound clip for the given key.
	 * @param key the key (e.g., event) to find the sound for. Accepts {@link de.ped.pacman.logic.Game.Event} objects.
	 * @return the sound clip for the given key, or <code>null</code> if there is no such sound.
	 */
	@Override
	public SoundClip findSound(final Object key) {
		init();
		final SoundClip ret = sounds.get((Game.Event) key);
		return ret;
	}

	@Override
	public void init() {
		if (null == sounds) {
			final HashMap<Game.Event, SoundClip> ret = new HashMap<Game.Event, SoundClip>();
			for (Game.Event event : Game.Event.values()) {
				final AudioInputStream stream = createSound(event);
				SoundClip clip = null;
				if (null != stream) {
					clip = new SoundClip(stream);
				}
				if (null != clip) {
					ret.put(event, clip);
				}
			}
			ret.put(Game.Event.EATING_FRUIT, ret.get(Game.Event.EATING_GHOST));
			sounds = ret;
		}
	}

	@Override
	public AudioInputStream createSound(final Object key) {
		AudioInputStream ret = null;
		DynamicSoundMonoWaveCreator.Sound sound = null;
		final DynamicSoundMonoWaveCreator smwc = new DynamicSoundMonoWaveCreator();
		final Game.Event event = (Game.Event) key;
		switch (event) {
		case GAME_STARTED:
			sound = createGameStartedSound(smwc);
			break;
		case LEVEL_STARTED:
			// Nothing to do
			break;
		case EATING_DOT:
			sound = createEatingDotSound(smwc);
			break;
		case EATING_PILL:
			sound = createEatingPillSound(smwc);
			break;
		case FRUIT_ENTERED:
			sound = createFruitEnteredSound(smwc);
			break;
		case EATING_GHOST:
			sound = createEatingGhostSound(smwc);
			break;
		case EATING_FRUIT:
			// Nothing to do
			break;
		case FRUIT_ENDED:
			// Nothing to do
			break;
		case LEVEL_DONE:
			// Nothing to do
			break;
		case LIFE_LOST:
			sound = createLifeLostSound(smwc);
			break;
		case GAME_OVER:
			// Nothing to do
			break;
		default:
			// Nothing to do
		}
		if (null != sound) {
			ret = smwc.toAudioInputStream(sound);
		}
		return ret;
	}

	private static DynamicSoundMonoWaveCreator.Sound createLifeLostSound(DynamicSoundMonoWaveCreator smwc) {
		final Tuning tuning = Tuning.defaultTuning();
		final double duration = 2.0;
		final EnvelopeWave fmDown = new EnvelopeWave(smwc, duration, 4.0, 0.0);
		final SinusWave fmSin = new SinusWave(smwc, 4.0 / duration);
		final AddedWave fm = new AddedWave(smwc, fmSin, 1.0, fmDown, 1.0);
		final EnvelopeWave am = new EnvelopeWave(smwc, duration, 0.0, -1.0);
		final TriangleWave main = new TriangleWave(smwc, tuning.getFrequency(Note.get("A1")), fm);
		final AmplitudeModulationWave amw = new AmplitudeModulationWave(smwc, main, 1.0, am);
		return new DynamicSoundMonoWaveCreator.Sound(amw, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createEatingDotSound(DynamicSoundMonoWaveCreator smwc) {
		final Tuning tuning = Tuning.defaultTuning();
		final double duration = 0.12;
		final EnvelopeWave fm = new EnvelopeWave(smwc, duration, -5.0, 5.0);
		final EnvelopeWave am = new EnvelopeWave(smwc, duration, 0.0, -1.0);
		final SinusWave main = new SinusWave(smwc, tuning.getFrequency(Note.get("A1")), fm);
		final AmplitudeModulationWave amw = new AmplitudeModulationWave(smwc, main, 1.0, am);
		return new DynamicSoundMonoWaveCreator.Sound(amw, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createEatingPillSound(DynamicSoundMonoWaveCreator smwc) {
		final Tuning tuning = Tuning.defaultTuning();
		final double duration = 1.0;
		final MonoWave ams = new SinusWave(smwc, duration);
		final MonoWave amf = new ConstantWave(0.6);
		final MonoWave am  = new MultipliedWave(smwc, ams, amf);
		final MonoWave main = new SinusWave(smwc, tuning.getFrequency(Note.get("A0")));
		final AmplitudeModulationWave amw = new AmplitudeModulationWave(smwc, main, 0.6, am);
		return new DynamicSoundMonoWaveCreator.Sound(amw, duration);
	}

	private Sound createFruitEnteredSound(DynamicSoundMonoWaveCreator smwc) {
		final Tuning tuning = Tuning.defaultTuning();
		final double duration = 1.0;
		final EnvelopeWave.Parameters ep1 = new EnvelopeWave.Parameters(smwc, 0.00, 0.03, 0.02, 0.10, 0.05, 0.0, 0.1, 0.05);
		final EnvelopeWave.Parameters ep2 = new EnvelopeWave.Parameters(smwc, 0.25, 0.03, 0.02, 0.10, 0.05, 0.0, 0.1, 0.05);
		final EnvelopeWave.Parameters ep3 = new EnvelopeWave.Parameters(smwc, 0.50, 0.05, 0.05, 0.20, 0.10, 0.0, 0.1, 0.05);
		final MonoWave main1 = new SquareWave(smwc, tuning.getFrequency(Note.get("G1")));
		final MonoWave main2 = new SquareWave(smwc, tuning.getFrequency(Note.get("C1")));
		final MonoWave main3 = new SquareWave(smwc, tuning.getFrequency(Note.get("E1")));
		final MonoWave amw1  = new MultipliedWave(smwc, main1, new EnvelopeWave(ep1));
		final MonoWave amw2  = new MultipliedWave(smwc, main2, new EnvelopeWave(ep2));
		final MonoWave amw3  = new MultipliedWave(smwc, main3, new EnvelopeWave(ep3));
		final MonoWave amw = new ConcatWave(smwc, new ConcatWave.Part[] {
				new ConcatWave.Part(amw1, 0.25),
				new ConcatWave.Part(amw2, 0.25),
				new ConcatWave.Part(amw3, 0.50),
		});
		return new DynamicSoundMonoWaveCreator.Sound(amw, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createEatingGhostSound(DynamicSoundMonoWaveCreator smwc) {
		final Tuning tuning = Tuning.defaultTuning();
		final double duration = 0.3;
		final EnvelopeWave fm = new EnvelopeWave(smwc, duration, -15.0, 15.0);
		final EnvelopeWave am = new EnvelopeWave(smwc, duration, 0.0, -1.0);
		final TriangleWave main = new TriangleWave(smwc, tuning.getFrequency(Note.get("A1")), fm);
		final AmplitudeModulationWave amw = new AmplitudeModulationWave(smwc, main, 1.0, am);
		return new DynamicSoundMonoWaveCreator.Sound(amw, duration);
	}

	private static DynamicSoundMonoWaveCreator.Sound createGameStartedSound(final DynamicSoundMonoWaveCreator smwc) {
		final Tuning tuning = Tuning.defaultTuning();
		final String[] melody = { 
				// First block
				"C1", "",
				"C2", "",
				"G1", "",
				"E1", "C2",
				"G1", "",
				"E1", "",
				"", "",
				"", "",
				// Repeated half-tone up
				"C#1", "",
				"C#2", "",
				"G#1", "",
				"E#1", "C#2",
				"G#1", "",
				"E#1", "",
				"", "",
				"", "",
				// First block repeated
				"C1", "",
				"C2", "",
				"G1", "",
				"E1", "C2",
				"G1", "",
				"E1", "",
				"", "",
				"", "",
				// Coda
				"E1", "F1",
				"F#1", "",
				"F1", "G1",
				"G#1", "",
				"A1", "A#1",
				"B1", "",
				"C2", "",
				"", "",
		};
		final double noteDuration = 6.0 / melody.length;
		final ConcatWave.Part[] melodyline = new ConcatWave.Part[melody.length];
		for (int i = 0; i < melodyline.length; i++) {
			final String mel = melody[i];
			final MonoWave added;
			if (!mel.isEmpty()) {
				added = new TriangleWave(smwc, tuning.getFrequency(Note.get(melody[i])));
			} else {
				added = new ConstantWave(0.0);
			}
			melodyline[i] = new ConcatWave.Part(added, noteDuration);
		}
		final MonoWave melodyWave = new ConcatWave(smwc, melodyline);
		final MonoWave allWave = new AmplitudeModulationWave(smwc, melodyWave, 0.4);
		return new DynamicSoundMonoWaveCreator.Sound(allWave, noteDuration * melody.length);
	}

	private HashMap<Game.Event, SoundClip> sounds;
}
