package de.ped.pacman.logic;

import java.util.LinkedList;

import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Logger;

/**
 * Knows the current position for pacman, a monster or a fruit - anything able to walk around and being uniquely identifiable.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Being extends Sprite implements Comparable<Being>, Cloneable {
	protected Logger logger = Logger.getRootLogger();

	public static final String PACMAN = "P";
	public static final String FRUIT = "F";
	public static final String BLINKY = "1";
	public static final String PINKY = "2";
	public static final String INKY = "3";
	public static final String CLYDE = "4";
	public static final String BEINGS_ORDER = PACMAN + FRUIT + BLINKY + PINKY + INKY + CLYDE;

	/** For testing purposes made routing rating visible */
	protected int[] directionRatings = new int[Direction.DIRECTIONS.length];

	/** The step counter to calculate the correct speed of this being (e.g., when to omit a step) */ 
	private int stepCounter;
	
	/** The number of steps this being is caged in the den before it can pass doors. */
	public final int stepsInDen;

	/** The number of steps this being is hidden before it enters the stage. */
	public final int stepsBeforeInStage;

	/** 
	 * For testing purposes made routing rating visible.
	 * @return for all {@link Direction#DIRECTIONS} the routing rating
	 */
	public int[] getDirectionRatings() {
		return directionRatings;
	}

	public boolean isPacPan() {
		return (id.equals(PACMAN));
	}

	public boolean isFruit() {
		return (id.equals(FRUIT));
	}

	public boolean isOpponent() {
		return (id.equals(INKY) || id.equals(BLINKY)  || id.equals(PINKY)  || id.equals(CLYDE));
	}

	/**
	 * Returns if the being is currently a ghost. E.g. applies for opponents
	 * eaten and on their way to the den.
	 * <p>
	 * Default implementation always returns <code>false</code>, overload if
	 * needed.
	 * 
	 * @return if the being is currently a ghost
	 */
	public boolean isGhost() {
		return false;
	}

	/**
	 * Returns if a step is possible, given on the content of the last labyrinth
	 * position and the possible next position. Overload if necessary.
	 * 
	 * @param game
	 *            the environment to look into
	 * @param newPosContent
	 *            the content of the possible destination field
	 * @param oldPosContent
	 *            the content of the field this being is coming from
	 * @return <code>true</code> if the step in this direction is allowed,
	 *         considering the contents
	 */
	public boolean canEnterOnContent(final Game game, final char newPosContent, final char oldPosContent) {
		return !((newPosContent == Labyrinth.WALL) || (newPosContent == Labyrinth.DOOR));
	}

	public Being(final String id, final float x, final float y, final int direction, final int stepsBeforeInStage, final int stepsInDen) {
		this.id = id;
		set (x, y);
		this.direction = direction;
		this.stepsBeforeInStage = stepsBeforeInStage;
		this.stepsInDen = stepsInDen;
	}

	protected String id;

	protected int direction = Direction.LEFT;
	protected int nextDirection = Direction.UNDEFINED;

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((id == null) ? 0 : id.hashCode());
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Being other = (Being) obj;
		if (id == null) {
			if (other.id != null)
				return false;
		} else if (!id.equals(other.id))
			return false;
		return true;
	}

	@Override
	public int compareTo(final Being other) {
		return BEINGS_ORDER.indexOf(getId()) - BEINGS_ORDER.indexOf(other.getId());
	}

	public String getId() {
		return id;
	}

	/**
	 * Returns the current direction.
	 * @return either {@link Direction#UP}, {@link Direction#DOWN}, {@link Direction#LEFT} or {@link Direction#RIGHT}
	 */
	public int getDirection() {
		return direction;
	}

	/**
	 * Sets the current direction.
	 * @param direction either {@link Direction#UP}, {@link Direction#DOWN}, {@link Direction#LEFT} or {@link Direction#RIGHT}
	 */
	public void setDirection(final int direction) {
		if (Direction.UNDEFINED == direction) {
			this.direction = direction;
			logger.debug("Being " + id + " set direction to undefined.");
		} else {
			this.direction = Direction.norm(direction);
			logger.debug("Being " + id + " set direction to " + Direction.DIRECTIONS[direction].name + ".");
		}
	}

	/**
	 * Returns the next direction to move whenever the next opportunity to change the direction appears.
	 * @return either {@link Direction#UP}, {@link Direction#DOWN}, {@link Direction#LEFT} or {@link Direction#RIGHT}
	 */
	protected int getNextDirection() {
		return nextDirection;
	}

	/**
	 * Sets the next direction to move whenever the next opportunity to change the direction appears.
	 * @param direction either {@link Direction#UP}, {@link Direction#DOWN}, {@link Direction#LEFT} or {@link Direction#RIGHT}
	 */
	public void setNextDirection(final int direction) {
		this.nextDirection = Direction.norm(direction);
//		logger.debug("Being " + id + " set next direction to " + DIRECTIONS[direction].name + ".");
	}

	/**
	 * Overload with the strategy of the being to change direction. Default
	 * implementation rates all directions.
	 * 
	 * @param game
	 *            the game the being is in
	 * @return the direction to take. {@link Direction#UNDEFINED} means there is none.
	 * @see #rateAllDirections(Game)
	 */
	public int planNextDirection(final Game game) {
		int ret = getNextDirection();
		// Only calculate new direction if on grid, to save cycle power
		if (isOnGrid()) {
			final int[] dirOptions = rateAllDirections(game);
			this.directionRatings = dirOptions;
			try {
				ret = MathUtil.randomDistributed(dirOptions);
			} catch (final ArrayIndexOutOfBoundsException ex) {
				logger.error(id + " has no direction to move to from position (" + getX() + "," + getY() + ")");
				ret = Direction.UNDEFINED;
			}
		}
		return ret;
	}

	/**
	 * Checks if a move in the given direction is possible, without moving.
	 * 
	 * @param game
	 *            the game the being is in
	 * @param direction
	 *            either {@link Direction#UP}, {@link Direction#DOWN},
	 *            {@link Direction#LEFT} or {@link Direction#RIGHT}.
	 * @return Information about the move, or <code>null</code> if move is
	 *         impossible.
	 */
	public MoveResultInformation calcMoveDestinationIfPossible(final Game game, final int direction) {
		float xDest = round(vector[0] + RESOLUTION * (float)Direction.DIRECTIONS[direction].signs[0]);
		float yDest = round(vector[1] + RESOLUTION * (float)Direction.DIRECTIONS[direction].signs[1]);
		if (null != game) {
			final PlayfieldDimension size = game.getBoard().getSize();
			xDest = MathUtil.modulo(xDest, (float)(size.getWidth() - 1));
			yDest = MathUtil.modulo(yDest, (float)(size.getHeight() - 1));
		}
		if (!isLegalPosition(xDest, yDest)) {
			return null;
		}
		if (null != game) {
			int xfrom = floor(xDest);
			int yfrom = floor(yDest);
			int xto   = xfrom;
			int yto   = yfrom;
			switch(direction) {
			case Direction.UP:
				yfrom = ceil(vector[1]);
				yto   = floor(yDest);
				break;
			case Direction.DOWN:
				yfrom = floor(vector[1]);
				yto   = ceil(yDest);
				break;
			case Direction.LEFT:
				xfrom = ceil(vector[0]);
				xto   = floor(xDest);
				break;
			case Direction.RIGHT:
				xfrom = floor(vector[0]);
				xto   = ceil(xDest);
				break;
			}
			final char valueAtDestination = game.getBoard().get(xto, yto);
			final char valueAtCurrentPos = game.getBoard().get(xfrom, yfrom);
			if (canEnterOnContent(game, valueAtDestination, valueAtCurrentPos)) {
				return new MoveResultInformation(xDest, yDest, valueAtCurrentPos, valueAtDestination);
			} else {
				return null;
			}
		}
		return new MoveResultInformation(xDest, yDest);
	}

	/**
	 * Tries to move in the given direction. If successful, the position is
	 * changed.
	 * 
	 * @param game
	 *            the game the being is in
	 * @param direction
	 *            either {@link Direction#UP}, {@link Direction#DOWN},
	 *            {@link Direction#LEFT} or {@link Direction#RIGHT}.
	 * @return <code>true</code> iff the move was possible
	 */
	protected boolean moveIfPossible(final Game game, final int direction) {
		boolean ret = false;
		final MoveResultInformation mri = calcMoveDestinationIfPossible(game, direction);
		if (null != mri) {
			vector[0] = mri.xDestPos;
			vector[1] = mri.yDestPos;
			ret = true;
		}
		return ret;
	}

	/**
	 * Returns how many steps the being is allowed to move in the current step.
	 * Is called exactly once per time slice.
	 * <p>
	 * Default implementation always returns <em>1</em>; overload to slow down
	 * or speed up being.
	 * 
	 * @param game
	 *            the game the being is in
	 * @return <em>1</em>, meaning that the being is doing one move per game
	 *         step.
	 */
	public int numberOfMovesForThisStep(final Game game) {
		int ret = 1;
		final int[] stepWaits = getStepWaits(game);
		stepCounter = MathUtil.modulo(stepCounter, stepWaits.length);
//		logger.debug("Wait step for " + getId() + ": " + stepCounter + " (" + STEP_WAITS[stepCounter] + ")");
		ret = stepWaits[stepCounter];
		stepCounter++;
		return ret;
	}

	public void move(final Game game, final LinkedList<Game.Event> events) {
		final int moves = numberOfMovesForThisStep(game);
		for (int i = 0; i < moves; i++) {
			final int dir = planNextDirection(game);
			if (dir != Direction.UNDEFINED) {
				setNextDirection(dir);
			}
			boolean couldMove = false;
			if ((nextDirection != Direction.UNDEFINED) && (direction != nextDirection)) {
				couldMove = moveIfPossible(game, nextDirection);
				if (couldMove) {
					direction = nextDirection;
					nextDirection = Direction.UNDEFINED;
				}
			}
			if (!couldMove) {
				couldMove = moveIfPossible(game, direction);
			}
			if (couldMove) {
				if (null != game) {
					evalAfterMove(game, events);
				}
			}
		}
	}

	/**
	 * Overload with actions that need to happen after the move.
	 * 
	 * @param game the game the being is in
	 * @param events might be filled by this method with events that lead to actions on the UI, like playing sounds.
	 */
	protected void evalAfterMove(final Game game, LinkedList<Game.Event> events) {
		// Default implementation does nothing
	}

	public PlayfieldPosition field() {
		PlayfieldPosition ret = null;
		if (isOnGrid()) {
			ret = new PlayfieldPosition((int)Math.round(vector[0]), (int)Math.round(vector[1]));
		}
		return ret;
	}

	public boolean touches(final Being being) {
		final double distance = Math.abs(getX() - being.getX()) + Math.abs(getY() - being.getY()); 
		return (distance < 0.8);
	}

	public void pillEatenNotification() {
		// Nothing to do
	}

	@Override
	public Being clone() {
		final Being ret = (Being)super.clone();
		return ret;
	}

	/**
	 * Rates a step of this being. Can safely assume that the step is possible.
	 * Default implementation just returns a constant positive value, so all
	 * directions have the same probability.
	 * 
	 * @param dir
	 *            the direction to try
	 * @param game
	 *            the game the being is in
	 * @param mri
	 *            the move result information
	 * @return a non-negative rating. 0 means impossible, and a higher number
	 *         means higher rating
	 */
	protected int rateStep(final int dir, final Game game, final MoveResultInformation mri) {
		return 1;
	}

	public int[] rateAllDirections(final Game game) {
		final int[] dirOptions = new int[Direction.DIRECTIONS.length]; // All fields are initialized to 0
		for (int dir = 0; dir < Direction.DIRECTIONS.length; dir++) {
			final MoveResultInformation mri = calcMoveDestinationIfPossible(game, dir);
			if (null != mri) {
				dirOptions[dir] = rateStep(dir, game, mri);
			}
		}
		forbidTurnaround(dirOptions);
		return dirOptions;
	}

	/**
	 * Changes dirOptions so that the opposite direction to the current
	 * direction is forbidden. This helps to avoid brownian motion of the being.
	 * 
	 * @param dirOptions
	 *            this array is altered
	 */
	protected void forbidTurnaround(final int[] dirOptions) {
		if (MathUtil.count(dirOptions) > 1) {
			final int dir = getDirection();
			if (Direction.UNDEFINED != dir) {
				final int oppdir = Direction.getOppositeDirection(getDirection());
				dirOptions[oppdir] = Math.min(dirOptions[oppdir], 0);
			}
		}
	}
	
	public static final int[] DEFAULT_STEP_WAITS = new int[]{ 1, };

	/**
	 * Defines when to omit a move or take to moves in a step. The returned
	 * array is followed step by step in round robin fashion. A 1 means, for a
	 * step the being moves one step. A 0 means to omit a step. A 2 means two
	 * moves in one step, and so on.
	 * <p>
	 * Default implementation is always 1 move per step.
	 * 
	 * @param game
	 *            the outcome may be influenced by the game the being is in
	 * @return an array of <code>int</code> defining how many moves the being
	 *         should move in subsequent steps
	 */
	protected int[] getStepWaits(final Game game) {
		return DEFAULT_STEP_WAITS;
	}
}
