package de.ped.pacman.logic;

import java.util.LinkedList;
import java.util.List;

import de.ped.tools.MathUtil;
import de.ped.tools.log.Logger;

/**
 * The business logic of the game.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Game {
	private transient Logger logger = Logger.getLogger(getClass());

	/** How many ticks the effect of eating a pill will last */
	public static final int PILL_DURATION = 100;
	public static final int PILL_NOT_ACTIVE = -1;

	/**
	 * The modes the game can be in.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Mode {
		/** The intro of a level is running; the initial state of this state machine */
		LEVELINTRO(10),
		/** This is the first level of the game, so an intro melody is played */
		GAMEINTRO(6000 / Game.GAME_STEP_DELAY),
		/** Pacman is in action in a level */
		RUNNING(-1),
		/** The level done sequence is running */
		LEVELDONE(10),
		/** The life lost sequence is running */
		LIFELOST(10),
		/** The game over sequence is running */
		GAMEOVER(10),
		/** The game is over but still visible on the screen */
		WAITAFTERGAME(100),
		/** The game is over and the outro sequence is running */
		OUTRO(10),
		/** The game has finished; the final state of this state machine */
		GAMEFINISHED(1);
		
		public final int duration;
		
		private Mode(final int duration) {
			this.duration = duration;
		}
	}

	/**
	 * The events that might trigger something special in the UI, like playing a sound
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum Event {
		LEVEL_STARTED,
		GAME_STARTED,
		FRUIT_ENTERED,
		EATING_DOT,
		EATING_PILL,
		PILL_ENDED,
		FRUIT_ENDED,
		EATING_FRUIT,
		EATING_GHOST,
		LEVEL_DONE,
		LIFE_LOST,
		GAME_OVER,
	}

	private GameProperties props = new GameProperties();
	private int points = 0;
	private LevelInfo levelInfo;
	private int lives = 3;
	private Labyrinth board = new Labyrinth();
	private Mode mode = Mode.LEVELINTRO;
	private int stepCnt = 0;
	private int stepOfPillEaten = -PILL_DURATION;
	private int opponentsEaten = 0;
	private int suggestedUserGameDirection = Direction.UNDEFINED;
	protected LinkedList<Sprite> sprites = new LinkedList<Sprite>();

	public static final int GAME_STEP_DELAY = 120;

	/**
	 * Some infos defining a game level.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class LevelInfo {
		int level;
		String lab;
		Being[] beings;
		int waitTicks;
	}

	public LevelInfo levelFactory(final int level) {
		LevelInfo ret = null;
		if (0 <= level) {
			ret = new LevelInfo();
			ret.level = level;
			final PacmanLevelSet levelSet = props.getLevelSet();
			final int numOfLabs;
			switch(levelSet) {
			case STANDARD:
				numOfLabs = 2;
				if (0 == (level % numOfLabs)) {
					ret.lab = Labyrinth.LABYRINTH_PACMAN;
				} else {
					ret.lab = Labyrinth.LABYRINTH_PACMAN2;
				}
				ret.beings = Labyrinth.BEINGS;
				break;
			case ADVENT:
				numOfLabs = 1;
				ret.lab = Labyrinth.LABYRINTH_PACMAN3;
				ret.beings = Labyrinth.BEINGS;
				break;
			case TITLE:
				numOfLabs = 3;
				ret.lab = Labyrinth.LABYRINTH_TITLE;
				ret.beings = Labyrinth.BEINGS_TITLE;
				break;
			default:
				numOfLabs = 3;
				assert(false);
			}
			ret.waitTicks = Game.GAME_STEP_DELAY * (100 - 25 * MathUtil.bound(level / numOfLabs, 0, 3)) / 100;
		}
		return ret;
	}

	public Game(final String levelSet) {
		props.setLevelSet(levelSet);
		initGame(props);
	}

	public synchronized Mode getMode() {
		return mode;
	}

	protected void setMode(final Mode mode) {
		this.mode = mode;
		stepCnt = 0;
	}

	public synchronized int getLevel() {
		return levelInfo.level;
	}

	public int getSleepDelay() {
		int ret = Game.GAME_STEP_DELAY;
		switch(mode) {
		case RUNNING:
			ret = levelInfo.waitTicks;
			break;
		default:
			// Nothing to do
		}
		return ret;
	}

	public GameProperties getProperties() {
		return props;
	}

	public synchronized void initGame(final GameProperties props) {
		this.props = props;
		points = 0;
		lives = props.getLevelSet().numberOfLives;
		logger.debug("Level=" + props.getLevel());
		initLevel(props.getLevel());
	}

	public synchronized void initLevel(final int level) {
		levelInfo = levelFactory(level);
		board.init(levelInfo.lab, levelInfo.beings);
		mode = Mode.LEVELINTRO;
		stepCnt = 0;
		suggestedUserGameDirection = Direction.UNDEFINED;
		sprites.clear();
		stepOfPillEaten = -PILL_DURATION;
	}

	public synchronized int getStep() {
		return stepCnt;
	}

	/** 
	 * For debugging purposes only. 
	 * @param step the new value for step
	 */
	public synchronized void setStep(final int step) {
		this.stepCnt = step;
	}

	public synchronized int getPoints() {
		return points;
	}

	public synchronized void addPoints(final int howMany, final Being reason) {
		points += howMany;
		if (null != reason) {
			sprites.add(new PointsIndicator(reason, getStep(), Integer.toString(howMany)));
		}
	}

	public synchronized Labyrinth getBoard() {
		return board;
	}

	public synchronized LinkedList<Sprite> getSprites() {
		return sprites;
	}

	/**
	 * Executes one step in the game object.
	 * @return list of events that lead to actions on the UI, like playing sounds.
	 */
	public synchronized LinkedList<Event> step() {
		LinkedList<Event> events = new LinkedList<Event>();
		switch (mode) {
		case LEVELINTRO:
			if (stepCnt == 1) {
				suggestedUserGameDirection = Direction.UNDEFINED;
				sprites.clear();
				board.initBeings(levelInfo.beings);
				stepOfPillEaten = -PILL_DURATION;
				if (!(PacmanLevelSet.TITLE == props.getLevelSet())) {
					events.add(Event.LEVEL_STARTED);
				}
			}
			if (stepCnt >= mode.duration) {
				setMode(Mode.GAMEINTRO);
			}
			break;
		case GAMEINTRO:
			if (!(PacmanLevelSet.TITLE == props.getLevelSet()) && (0 == points) && (lives == props.getLevelSet().numberOfLives)) {
				if (stepCnt == 1) {
					events.add(Event.GAME_STARTED);
				} else {
					// Nothing to do
				}
			} else {
				stepCnt = mode.duration;
			}
			if (stepCnt >= mode.duration) {
				setMode(Mode.RUNNING);
			}
			break;
		case RUNNING:
			gameStep(events);
			if (0 == board.getRemainingDots()) {
				setMode(Mode.LEVELDONE);
			}
			break;
		case LEVELDONE:
			if ((stepCnt == 1) && (!(PacmanLevelSet.TITLE == props.getLevelSet()))) {
				events.add(Event.PILL_ENDED);
				events.add(Event.FRUIT_ENDED);
				events.add(Event.LEVEL_DONE);
			}
			if (stepCnt >= mode.duration) {
				initLevel(levelInfo.level + 1);
				setMode(Mode.LEVELINTRO);
			}
			break;
		case LIFELOST:
			if ((stepCnt == 1) && (!(PacmanLevelSet.TITLE == props.getLevelSet()))) {
				events.add(Event.PILL_ENDED);
				events.add(Event.FRUIT_ENDED);
				events.add(Event.LIFE_LOST);
			}
			if (stepCnt >= mode.duration) {
				lives--;
				if (lives <= 0) {
					setMode(Mode.GAMEOVER);
				} else {
					setMode(Mode.LEVELINTRO);
				}
			}
			break;
		case GAMEOVER:
			if ((stepCnt == 1) && (!(PacmanLevelSet.TITLE == props.getLevelSet()))) {
				events.add(Event.GAME_OVER);
			}
			if (stepCnt >= mode.duration) {
				stepOfPillEaten = -PILL_DURATION;
				setMode(Mode.WAITAFTERGAME);
			}
			break;
		case WAITAFTERGAME:
			if (stepCnt >= mode.duration) {
				setMode(Mode.OUTRO);
			}
			break;
		case OUTRO:
			if (stepCnt >= mode.duration) {
				setMode(Mode.GAMEFINISHED);
			}
			break;
		case GAMEFINISHED:
			if (stepCnt >= mode.duration) {
				stepCnt = mode.duration;
			}
			break;
		default:
			throw new IllegalStateException("Unknown game mode " + mode);
		}
		stepCnt++;
		return events;
	}

	/**
	 * Executes a step within a running game.
	 * @param events might be filled by this method with events that lead to actions on the UI, like playing sounds.
	 */
	protected void gameStep(final LinkedList<Event> events) {
		for (final Sprite sprite : sprites) {
			sprite.step(this);
		}
		final List<Being> beings = board.getBeings();
		for (final Being being : beings) {
			try {
				being.move(this, events);
			} catch (final IllegalArgumentException ex) {
				// Nothing to do
			}
		}
		for (Being being : board.getBeingTemplates()) {
			if (being.stepsBeforeInStage == stepCnt) {
				final boolean wasAdded = board.addBeing(being.getId());
				if (wasAdded && being.isFruit()) {
					events.add(Event.FRUIT_ENTERED);
				}
			}
		}
		for (final Being being1 : beings) {
			for (final Being being2 : beings) {
				if ((0 > being1.compareTo(being2))  && (being1.touches(being2))) {
					onCollision(being1, being2, events);
				}
			}
		}
		removeDeadBeings();
		removeDeadSprites();
		if (1 == remainingTicksPillIsActive()) {
			events.add(Event.PILL_ENDED);
		}
		logger.trace("gameStep() : step=" + stepCnt);
	}

	protected void removeDeadBeings() {
		boolean isDone = true;
		do {
			isDone = true;
			for (final Being being1 : board.getBeings()) {
				if (!being1.isAlive()) {
					removeBeing(being1);
					isDone = false;
					break;
				}
			}
		} while (!isDone);
	}

	protected void removeDeadSprites() {
		boolean isDone = true;
		do {
			isDone = true;
			for (final Sprite sprite : sprites) {
				if (!sprite.isAlive()) {
					sprites.remove(sprite);
					isDone = false;
					break;
				}
			}
		} while (!isDone);
	}

	public void onCollision(final Being being1, final Being being2, final LinkedList<Game.Event> events) {
		if (being1.isPacPan()) {
			if (being2.isFruit()) {
				addPoints(50, being2);
				being2.die();
				events.add(Event.FRUIT_ENDED);
				events.add(Event.EATING_FRUIT);
			} else if (being2.isOpponent() && !being2.isGhost()) {
				if (PILL_NOT_ACTIVE != remainingTicksPillIsActive()) {
					opponentsEaten++;
					addPoints(opponentsEaten * 100, being2);
					((Opponent)being2).setGhost(true);
					events.add(Event.EATING_GHOST);
				} else {
					setMode(Mode.LIFELOST);
				}
			}
		}
	}

	public boolean removeBeing(final Being being) {
		return board.removeBeing(being);
	}

	/**
	 * The system got any input, e.g., from the keyboard, how a player should act for the next move.
	 * @param playerIdx the index of the player to suggest a turn for
	 * @param suggestion either {@link Direction#UP}, {@link Direction#DOWN}, {@link Direction#LEFT} or {@link Direction#RIGHT}
	 */
	public synchronized void suggestUserGameAction(final int playerIdx, final int suggestion) {
		suggestedUserGameDirection = suggestion;
		final Being pacman = board.getBeing(Being.PACMAN);
		if (null != pacman) {
			pacman.setNextDirection(suggestion);
		}
	}

	int getSuggestedUserGameAction() {
		return suggestedUserGameDirection;
	}

	public void eatPill() {
		opponentsEaten = 0;
		stepOfPillEaten = stepCnt;
		for (final Being being : board.getBeings()) {
			being.pillEatenNotification();
		}
	}

	/**
	 * For how many steps the pill effect will continue. 
	 * @return a number between 0 (pill effect over) and {@link #PILL_DURATION} (pill just eaten). Or, if no pill is currently active, {@link #PILL_NOT_ACTIVE}.
	 */
	public synchronized int remainingTicksPillIsActive() {
		int ret = PILL_NOT_ACTIVE;
		final int diff = stepCnt - stepOfPillEaten;
		if ((0 < diff) && (diff < PILL_DURATION)) {
			ret = PILL_DURATION - diff;
		}
		return ret;
	}

	public int getLives() {
		return lives;
	}
}
