package de.ped.pacman.logic;

import java.util.LinkedList;

import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldPosition;

/**
 * One of the opponents trying to catch PacMan.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Opponent extends Being {

	public static final int[] NORMAL_MOVES     = new int[]{1, 1, 0, 1, 1};
	public static final int[] FRIGHTENED_MOVES = new int[]{1, 0, 0};
	public static final int[] GHOST_MOVES      = new int[]{4};

	public static final int STEPS_IN_DEN = 20;

	protected boolean isGhost = false;

	public Opponent(final float x, final float y, final int direction, final String id, final int stepsBeforeInStage) {
		super(id, x, y, direction, stepsBeforeInStage, stepsBeforeInStage + Integer.parseInt(id) * STEPS_IN_DEN);
	}

	/**
	 * We need special rules for the den and the doors, depending on if this
	 * opponent is currently a ghost allowed to get in or a living opponent
	 * allowed to get out.
	 * 
	 * @param game
	 *            the game this being is in
	 * @param newPosContent
	 *            the content of the labyrinth position to enter
	 * @param oldPosContent
	 *            the content of the labyrinth position to leave
	 * @return if that being can enter the new position
	 */
	@Override
	public boolean canEnterOnContent(final Game game, final char newPosContent, final char oldPosContent) {
		boolean ret = true;
		switch(newPosContent) {
		case Labyrinth.WALL:
			ret = false;
			break;
		case Labyrinth.EMPTY:
		case Labyrinth.DOT:
		case Labyrinth.PILL:
			if (isGhost) {
				ret = (Labyrinth.EMPTY == oldPosContent) || (Labyrinth.DOT == oldPosContent) || (Labyrinth.PILL == oldPosContent);
			} else {
				ret = true;
			}
			break;
		case Labyrinth.DOOR:
			if (isGhost) {
				ret = (Labyrinth.EMPTY == oldPosContent) || (Labyrinth.DOT == oldPosContent) || (Labyrinth.PILL == oldPosContent);
			} else {
				ret = (Labyrinth.DEN == oldPosContent);
				if (null != game) {
					ret = ret && (game.getStep() >= stepsInDen);
				}
			}
			break;
		case Labyrinth.DEN:
			if (isGhost) {
				ret = (Labyrinth.DOOR == oldPosContent) || (Labyrinth.EMPTY == oldPosContent) || (Labyrinth.DOT == oldPosContent) || (Labyrinth.PILL == oldPosContent);
			} else {
				ret = (Labyrinth.DEN == oldPosContent);
			}
			break;
		default:
			// Nothing to do
		}
		return ret;
	}

	@Override
	public void pillEatenNotification() {
		setDirection(Direction.getOppositeDirection(getDirection()));
	}

	public static final int IMPOSSIBLE = 0;
	public static final int NOT_ADVISABLE = 1;
	public static final int POSSIBLE = 2;
	public static final int DESIRABLE = 6;
	public static final int RECOMMENDED = 10;

	/**
	 * Rates a step of this opponent. Can safely assume that the step is
	 * possible.
	 * 
	 * @param dir
	 *            either {@link Direction#UP}, {@link Direction#DOWN},
	 *            {@link Direction#LEFT} or {@link Direction#RIGHT}.
	 * @param game
	 *            the game the being is in
	 * @param mri
	 *            the move result
	 * @return Rating for the move; a higher number means a better move, a
	 *         rating of 0 or lower will not be considered.
	 */
	@Override
	protected int rateStep(final int dir, final Game game, final MoveResultInformation mri) {
		if (!isOnGrid()) {
			throw new IllegalStateException("Only allowed when on grid");
		}
		final Being pacman = game.getBoard().getBeing(PACMAN);
		int rating = NOT_ADVISABLE;
		if (isGhost) {
			if (Labyrinth.DOOR == mri.valAtCurentPos) {
				rating = (Labyrinth.DEN == mri.valAtDestinationPos ? DESIRABLE : IMPOSSIBLE);
			} else if (Labyrinth.DEN == mri.valAtCurentPos) {
				rating = IMPOSSIBLE;
			} else {
				if (Labyrinth.DOOR == mri.valAtDestinationPos) {
					rating = RECOMMENDED;
				} else {
					PlayfieldPosition door = game.getBoard().getNearestDoor(getX(), getY());
					rating = rateStep(dir, door.getX(), door.getY(), true);
				}
			}
//			if (Sprite.areEqual(13.0F, getX()) && Sprite.areEqual(11.0F, getY()) && (dir == Direction.DOWN)) {
//				logger.debug("Before door ghost rating towards den was " + rating);
//			}
		} else {
			if (Labyrinth.DOOR == mri.valAtCurentPos) {
				rating = (Labyrinth.DEN == mri.valAtDestinationPos ? IMPOSSIBLE : DESIRABLE);
			} else if (Labyrinth.DEN == mri.valAtCurentPos) {
				if (Labyrinth.DOOR == mri.valAtDestinationPos) {
					if (isFrightened(game)) {
						rating = IMPOSSIBLE;
					} else {
						rating = RECOMMENDED;
					}
				} else {
					PlayfieldPosition door = game.getBoard().getNearestDoor(getX(), getY());
					rating = rateStep(dir, door.getX(), door.getY(), true);
				}
			} else {
				if (null != pacman) {
					final boolean isTowardsRef = (!isFrightened(game));
					rating = rateStep(dir, pacman.getX(), pacman.getY(), isTowardsRef);
				}
			}
		}
		return rating;
	}

	public static final int[][] RATINGS = new int[][]{ {DESIRABLE, POSSIBLE, NOT_ADVISABLE}, {NOT_ADVISABLE, POSSIBLE, DESIRABLE} };

	/**
	 * Rates a move of this opponent. Can safely assume that the move is
	 * possible.
	 * 
	 * @param dir
	 *            either {@link Direction#UP}, {@link Direction#DOWN},
	 *            {@link Direction#LEFT} or {@link Direction#RIGHT}
	 * @param refX
	 *            the x-position of the target (e.g., pacman or nearest door to
	 *            den)
	 * @param refY
	 *            the y-position of the target (e.g., pacman or nearest door to
	 *            den)
	 * @param isTowardsRef
	 *            if <code>true</code>, this opponent tries to move towards the
	 *            reference position; if <code>false</code>, tries to flee from
	 *            the reference position
	 * @return rating for the move; a higher number means a better move, a
	 *         rating of 0 or lower will not be considered
	 */
	protected int rateStep(final int dir, final double refX, final double refY, final boolean isTowardsRef) {
		int rating = NOT_ADVISABLE;
		final double diffY = refY - getY();
		final double diffX = refX - getX();
		final int pillIdx = (isTowardsRef ? 0 : 1);
		if ((dir == Direction.UP) || (dir == Direction.DOWN)) {
			final int absDiffY = MathUtil.sgn(diffY) - Direction.DIRECTIONS[dir].signs[1];
			rating = RATINGS[pillIdx][Math.abs(absDiffY)];
		} else if ((dir == Direction.LEFT) || (dir == Direction.RIGHT)) {
			final int absDiffX = MathUtil.sgn(diffX) - Direction.DIRECTIONS[dir].signs[0];
			rating = RATINGS[pillIdx][Math.abs(absDiffX)];
		} else {
			throw new IllegalArgumentException("Unknown direction " + dir);
		}
		return rating;
	}

	/**
	 * Opponent is slower than PacMan, especially when PacMan has eaten pill, so
	 * some steps ir has to wait. As a ghost, however, opponent is much faster
	 * than PacMan.
	 * 
	 * @param game
	 *            the game the being is in
	 * @return an array of <code>int</code> defining how many moves the being
	 *         should move in subsequent steps
	 */
	@Override
	protected int[] getStepWaits(final Game game) {
		int[] ret;
		if (isGhost) {
			ret = GHOST_MOVES;
		} else if (isFrightened(game)){
			ret = FRIGHTENED_MOVES;
		} else {
			ret = NORMAL_MOVES;
		}
		return ret;
	}

	/**
	 * A ghost returned to den gets to life again.
	 * 
	 * @param game
	 *            the game the being is in
	 */
	@Override
	protected void evalAfterMove(final Game game, final LinkedList<Game.Event> events) {
		if (isGhost() && isOnGrid() && (Labyrinth.DEN == game.getBoard().get(Math.round(getX()), Math.round(getY())))) {
			setGhost(false);
		}
	}

	public boolean isFrightened(final Game game) {
		return (Game.PILL_NOT_ACTIVE != game.remainingTicksPillIsActive());
	}

	@Override
	public boolean isGhost() {
		return isGhost;
	}

	public void setGhost(final boolean isGhost) {
		this.isGhost = isGhost;
	}
}
