package de.ped.pacman.logic;

import de.ped.tools.MathUtil;

/**
 * Anything that can be drawn somewhere, not necessarily aligned with the grid.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Sprite implements Cloneable {

	public static final float RESOLUTION = 1.0F / 4.0F;

	public static final float EPSILON = RESOLUTION / 2.0F;

	public static boolean areEqual(final float val1, final float val2) {
		return (Math.abs(val1 - val2) < EPSILON);
	}

	public static boolean isOnGrid(final float value) {
		return areEqual(value, Math.round(value));
	}

	/**
	 * @param x
	 *            the x coordinate of the position
	 * @param y
	 *            the y coordinate of the position
	 * @return if both <em>x</em> and <em>y</em> are on grid
	 * @see #isOnGrid(float)
	 */
	public static boolean isOnGrid(final float x, final float y) {
		return (isOnGrid(x) && (isOnGrid(y)));
	}

	/**
	 * @param x
	 *            the x coordinate of the position
	 * @param y
	 *            the y coordinate of the position
	 * @return if either <em>x</em> or <em>y</em> is on grid
	 * @see #isOnGrid(float)
	 */
	public static boolean isLegalPosition(final float x, final float y) {
		return (isOnGrid(x) || (isOnGrid(y)));
	}

	public static int floor(final float value) {
		return (int)(Math.floor(value));
	}

	public static int ceil(final float value) {
		return (int)(Math.ceil(value));
	}

	public static float round(final float value) {
		return MathUtil.round(value, RESOLUTION);
	}

	protected float[] vector = new float[2];

	protected boolean isAlive = true;

	public Sprite() {
		// Nothing to do
	}

	public Sprite(final float x, final float y) {
		set(x, y);
	}

	public float getX() {
		return vector[0];
	}

	public float getY() {
		return vector[1];
	}

	public float get(final int dimension) {
		return vector[dimension];
	}

	public void set(float x, float y) {
		x = round(x);
		y = round(y);
		if (isLegalPosition(x, y)) {
			vector[0] = x;
			vector[1] = y;
		} else {
			throw new IllegalArgumentException("At least one value (x or y) should be integer");
		}
	}

	/**
	 * @return if its position is on grid
	 * @see #isOnGrid(float, float)
	 */
	public boolean isOnGrid() {
		return (isOnGrid(getX()) && (isOnGrid(getY())));
	}

	/** 
	 * Overload if necessary. 
	 * @param game
	 *            the game the sprite is in
	 */
	public void step(final Game game) {
		// Nothing to do
	}

	public boolean isAlive() {
		return isAlive;
	}

	public void die() {
		isAlive = false;
	}

	@Override
	public Sprite clone() {
		try {
			final Sprite ret = (Sprite)super.clone();
			ret.vector = vector.clone();
			return ret;
		} catch (final CloneNotSupportedException e) {
			return null;
		}
	}
}
