package de.ped.pacman.logic;

/**
 * Identifies which wall parts to draw, based on the information which of the neighboring fields are also walls.
 * The current implementation is a bit tricky as it tries to only draw the outline of a wall area to mimic the look of the original Pacman screen.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class WallMaskFilter {

	public static final int MASK_UP    = 0x01;
	public static final int MASK_LEFT  = 0x02;
	public static final int MASK_DOWN  = 0x04;
	public static final int MASK_RIGHT = 0x08;
	public static final int MASK_UP_LEFT    = 0x10;
	public static final int MASK_DOWN_LEFT  = 0x20;
	public static final int MASK_DOWN_RIGHT = 0x40;
	public static final int MASK_UP_RIGHT   = 0x80;

	public static final int MASK_4NEIGHBORS = MASK_UP | MASK_LEFT | MASK_DOWN | MASK_RIGHT;

	public static final int WALL_PARTS_TO_PAINT = 4;

	public static final int[][] MASK_WALL_POSITIONS = new int[][] {
		{  0, -1, MASK_UP },
		{ -1,  0, MASK_LEFT },
		{  0,  1, MASK_DOWN },
		{  1,  0, MASK_RIGHT },
		{ -1, -1, MASK_UP_LEFT },
		{ -1,  1, MASK_DOWN_LEFT },
		{  1,  1, MASK_DOWN_RIGHT },
		{  1, -1, MASK_UP_RIGHT },
	};

	/**
	 * Filter for which parts of a wall field to paint, so that in an area of walls not a full grid but only the outline is drawn.
	 * In the inner array, the wall part in element 0 will be removed if
	 * <ul>
	 *   <li>all wall parts in element 1 are present, and</li>
	 *   <li>true</li>
	 * </ul>
	 */
	public static final int[][] MASK_WALLS_FILTER = new int[][] {
		{ MASK_UP,    MASK_UP_LEFT   | MASK_UP_RIGHT   },
		{ MASK_LEFT,  MASK_UP_LEFT   | MASK_DOWN_LEFT  },
		{ MASK_DOWN,  MASK_DOWN_LEFT | MASK_DOWN_RIGHT },
		{ MASK_RIGHT, MASK_UP_RIGHT  | MASK_DOWN_RIGHT },
	};

	public static final int[][] MASK_WALLS_META_FILTER = new int[][] {
		{ MASK_UP,    MASK_RIGHT | MASK_DOWN_RIGHT },
		{ MASK_UP,    MASK_LEFT  | MASK_DOWN_LEFT  },
		{ MASK_LEFT,  MASK_UP    | MASK_UP_RIGHT   },
		{ MASK_LEFT,  MASK_DOWN  | MASK_DOWN_RIGHT },
		{ MASK_DOWN,  MASK_RIGHT | MASK_UP_RIGHT   },
		{ MASK_DOWN,  MASK_LEFT  | MASK_UP_LEFT    },
		{ MASK_RIGHT, MASK_UP    | MASK_UP_LEFT    },
		{ MASK_RIGHT, MASK_DOWN  | MASK_DOWN_LEFT  },
	};

	private int[] converterLookup;

	public WallMaskFilter() {
		converterLookup = new int[256];
		for (int i = 0; i < converterLookup.length; i++) {
			converterLookup[i] = calcFilterMaskWalls(i);
		}
	}

	private int calcFilterMaskWalls(final int surroundingWallsMask) {
		int ret = surroundingWallsMask & MASK_4NEIGHBORS;
		for (final int[] filter : MASK_WALLS_FILTER) {
			if (filter[1] == (surroundingWallsMask & filter[1])) {
				boolean isMetafiltered = false;
				for (final int[] metafilter : MASK_WALLS_META_FILTER) {
					if ((filter[0] == metafilter[0]) && 
							(0 == (surroundingWallsMask & metafilter[1]))) {
						isMetafiltered = true;
						break;
					}
				}
				if (!isMetafiltered) {
					ret &= ~filter[0];
				}
			}
		}
		return ret;
	}

	public int filterMaskWalls(int surroundingWallsMask) {
		return converterLookup[surroundingWallsMask];
	}
}
