package de.ped.pacman.test;

import static org.junit.Assert.*;

import java.util.LinkedList;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.ped.pacman.logic.Being;
import de.ped.pacman.logic.Direction;
import de.ped.pacman.logic.Fruit;
import de.ped.pacman.logic.Game;
import de.ped.pacman.logic.Labyrinth;
import de.ped.pacman.logic.MoveResultInformation;
import de.ped.pacman.logic.Opponent;
import de.ped.pacman.logic.Pacman;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.test.GenericTest;

/**
 * Tests {@link Being} and its subclasses.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class BeingTest {

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		Logger.getRootLogger().setLogLevel(Level.WARN);
	}

	@AfterClass
	public static void tearDownAfterClass() throws Exception {
		// Nothing to do
	}

	@Before
	public void setUp() throws Exception {
		// Nothing to do
	}

	@After
	public void tearDown() throws Exception {
		// Nothing to do
	}

	protected static final float res = Being.RESOLUTION;

	public static final float EPSILON = res / 4;

	@Test
	public void testInit_LegalPosition_Worked() {
		final Being pacman = new Pacman(1.0F, 2.75F, Direction.UP, 0);
		assertEquals(1.0, pacman.getX(), EPSILON);
		assertEquals(2.75, pacman.getY(), EPSILON);
	}

	@Test(expected=IllegalArgumentException.class)
	public void testInit_IllegalPosition() {
		new Being(Being.PACMAN, 1.5F, 2.75F, Direction.UP, 0, 0);
	}

	@Test
	public void testSet_LegalPositions_Worked() {
		final Being pacman = new Pacman(1.0F, 2.75F, Direction.UP, 0);
		assertEquals(1.0, pacman.getX(), EPSILON);
		assertEquals(2.75, pacman.getY(), EPSILON);
		pacman.set(1.5F, 2.0F);
		assertEquals(1.5, pacman.getX(), EPSILON);
		assertEquals(2.0, pacman.getY(), EPSILON);
	}

	@Test
	public void testSet_LegalPositionsNeedsRounding_Worked() {
		final Being pacman = new Pacman(1.0F, 2.75F + EPSILON, Direction.UP, 0);
		assertEquals(1.0, pacman.getX(), EPSILON);
		assertEquals(2.75, pacman.getY(), EPSILON);
	}

	@Test
	public void testMove_Worked() {
		final LinkedList<Game.Event> events = new LinkedList<Game.Event>();
		final Being pacman = new Pacman(1.5F, 2.0F, Direction.LEFT, 0);
		assertEquals(1.5, pacman.getX(), EPSILON);
		assertEquals(2.0, pacman.getY(), EPSILON);
		pacman.move(null, events);
		assertEquals(1.5 - Being.RESOLUTION, pacman.getX(), EPSILON);
		assertEquals(2.0, pacman.getY(), EPSILON);
	}

	@Test
	public void testMoveGame_Worked() {
		final LinkedList<Game.Event> events = new LinkedList<Game.Event>();
		final Game game = new Game(null);
		final Being pacman = game.getBoard().getBeing(Being.PACMAN);
		pacman.set(13.0F, 23.0F);
		pacman.setDirection(Direction.LEFT);
		final double x = pacman.getX();
		final double y = pacman.getY();
		pacman.move(game, events);
		assertEquals(x - Being.RESOLUTION, pacman.getX(), EPSILON);
		assertEquals(y, pacman.getY(), EPSILON);
	}

	@Test
	public void testMove_Failed() {
		final LinkedList<Game.Event> events = new LinkedList<Game.Event>();
		final Being pacman = new Pacman(1.5F, 2.0F, Direction.LEFT, 0);
		assertEquals(1.5, pacman.getX(), EPSILON);
		assertEquals(2.0, pacman.getY(), EPSILON);
		pacman.setDirection(Direction.UP);
		pacman.move(null, events);
		assertEquals(1.5, pacman.getX(), EPSILON);
		assertEquals(2.0, pacman.getY(), EPSILON);
	}

	@Test
	public void testMoveGame_Wall_Unchanged() {
		final LinkedList<Game.Event> events = new LinkedList<Game.Event>();
		final Game game = new Game(null);
		final Being pacman = game.getBoard().getBeing(Being.PACMAN);
		pacman.set(13.0F, 23.0F);
		pacman.setDirection(Direction.DOWN);
		final double x = pacman.getX();
		final double y = pacman.getY();
		pacman.move(game, events);
		assertEquals(x, pacman.getX(), EPSILON);
		assertEquals(y, pacman.getY(), EPSILON);
	}

	@Test
	public void testField_Worked() {
		assertNull(new Pacman(1.5F, 2.0F, Direction.LEFT, 0).field());
		assertEquals(new PlayfieldPosition(1, 2), new Pacman(1.0F, 2.0F, Direction.LEFT, 0).field());
	}

	@Test
	public void testTouches_Worked() {
		assertEquals(false, new Pacman(1.5F, 2.0F, Direction.LEFT, 0).touches(new Fruit(3.5F, 2.0F, Direction.LEFT, 0)));
		assertEquals(true,  new Pacman(1.5F, 2.0F, Direction.LEFT, 0).touches(new Fruit(1.5F, 2.0F, Direction.LEFT, 0)));
	}

	@Test
	public void testCanEnterOnContent() {
		final Being being = new Opponent(1.5F, 2.0F, Direction.LEFT, Being.PINKY, 0);
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.EMPTY, Labyrinth.EMPTY));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.EMPTY, Labyrinth.DOT));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.EMPTY, Labyrinth.PILL));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.EMPTY, Labyrinth.DOOR));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.EMPTY, Labyrinth.DEN));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DOT, Labyrinth.EMPTY));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DOT, Labyrinth.DOT));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DOT, Labyrinth.PILL));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DOT, Labyrinth.DOOR));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DOT, Labyrinth.DEN));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.WALL, Labyrinth.EMPTY));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.WALL, Labyrinth.DOT));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.WALL, Labyrinth.PILL));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.WALL, Labyrinth.DOOR));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.WALL, Labyrinth.DEN));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DOOR, Labyrinth.EMPTY));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DOOR, Labyrinth.DOT));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DOOR, Labyrinth.PILL));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DOOR, Labyrinth.DOOR));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DOOR, Labyrinth.DEN));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DEN, Labyrinth.EMPTY));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DEN, Labyrinth.DOT));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DEN, Labyrinth.PILL));
		assertEquals(false, being.canEnterOnContent(null, Labyrinth.DEN, Labyrinth.DOOR));
		assertEquals(true, being.canEnterOnContent(null, Labyrinth.DEN, Labyrinth.DEN));
	}

	public static final String LABYRINTH_WITH_DEN = 
		"#######\n" +
		"# .#::#\n" +
		"#==#==#\n" +
		"#::# .#\n" +
		"#######\n";

	public static final Being[] LABYRINTH_WITH_DEN_BEINGS = new Being[] {
		new Pacman(1, 1, Direction.LEFT, 0),
		new Opponent(1, 3, Direction.UP, Being.BLINKY, 0),
		new Opponent(2, 3, Direction.DOWN, Being.PINKY, 0),
	};

	protected void checkCalcMoveDestinationIfPossible(
			final Game game, 
			final Being being, 
			final MoveResultInformation[] exp) {
		final String text = "Step=" + game.getStep() + ", " + being.getId() + "(" + being.getX() + "," + being.getY() + "," + (being.isGhost() ? "Ghost" : "normal") + ") ";
		for (int directionIdx = 0; directionIdx < Direction.DIRECTIONS.length; directionIdx++) {
			final MoveResultInformation mri = being.calcMoveDestinationIfPossible(game, directionIdx);
			assertEquals(text + Direction.DIRECTIONS[directionIdx].name, 
					exp[directionIdx], mri);
		}
	}

	protected void checkCalcMoveDestinationIfPossible(
			final Game game, 
			final Being being, 
			final float x, 
			final float y, 
			final MoveResultInformation[][] expectedValues) {
		being.set(x, y);
		game.setStep(0);
		if (being instanceof Opponent) {
			Opponent opp = (Opponent)being;
			opp.setGhost(false);
		}
		checkCalcMoveDestinationIfPossible(game, being, expectedValues[0]);

		game.setStep(Opponent.STEPS_IN_DEN * 5);
		checkCalcMoveDestinationIfPossible(game, being, expectedValues[1]);

		if (being instanceof Opponent) {
			Opponent opp = (Opponent)being;
			opp.setGhost(true);
			checkCalcMoveDestinationIfPossible(game, being, expectedValues[2]);
			opp.setGhost(false);
		}
	}

	@Test
	public void testCalcMoveDestinationIfPossible() {
		final Game game = new Game(null);
		game.getBoard().init(LABYRINTH_WITH_DEN, LABYRINTH_WITH_DEN_BEINGS);
		game.setStep(Opponent.STEPS_IN_DEN * 5);
		final Being being = game.getBoard().getBeing(Being.BLINKY);
		// Left side of the labyrinth, where the way out of the den is up
		checkCalcMoveDestinationIfPossible(game, being, 1.00F, 3.00F, // In the den
			new MoveResultInformation[][] {	{ //Step 0
					null, 
					null, 
					null, 
					new MoveResultInformation(1.25F, 3.00F, Labyrinth.DEN, Labyrinth.DEN),
				}, { //Step n
					new MoveResultInformation(1.00F, 2.75F, Labyrinth.DEN, Labyrinth.DOOR),
					null,
					null,
					new MoveResultInformation(1.25F, 3.00F, Labyrinth.DEN, Labyrinth.DEN),
				}, { //Ghost
					null, 
					null, 
					null, 
					null, 
				} } );
		checkCalcMoveDestinationIfPossible(game, being, 1.00F, 2.00F, // On the door
				new MoveResultInformation[][] {	{ //Step 0
					new MoveResultInformation(1.00F, 1.75F, Labyrinth.DOOR, Labyrinth.EMPTY), 
					null, 
					null, 
					null,
				}, { //Step n
					new MoveResultInformation(1.00F, 1.75F, Labyrinth.DOOR, Labyrinth.EMPTY), 
					null,
					null,
					null,
				}, { //Ghost
					null, 
					new MoveResultInformation(1.00F, 2.25F, Labyrinth.DOOR, Labyrinth.DEN), 
					null, 
					null, 
				} } );
		checkCalcMoveDestinationIfPossible(game, being, 1.00F, 1.00F, // On empty
				new MoveResultInformation[][] {	{ //Step 0
					null, 
					null, 
					null, 
					new MoveResultInformation(1.25F, 1.00F, Labyrinth.EMPTY, Labyrinth.DOT), 
				}, { //Step n
					null, 
					null, 
					null, 
					new MoveResultInformation(1.25F, 1.00F, Labyrinth.EMPTY, Labyrinth.DOT), 
				}, { //Ghost
					null, 
					new MoveResultInformation(1.00F, 1.25F, Labyrinth.EMPTY, Labyrinth.DOOR), 
					null, 
					new MoveResultInformation(1.25F, 1.00F, Labyrinth.EMPTY, Labyrinth.DOT), 
				} } );
		checkCalcMoveDestinationIfPossible(game, being, 1.00F, 2.25F, // From the den to the door
				new MoveResultInformation[][] {	{ //Step 0
					null, 
					null, 
					null, 
					null,
				}, { //Step n
					new MoveResultInformation(1.00F, 2.00F, Labyrinth.DEN, Labyrinth.DOOR),
					null,
					null,
					null,
				}, { //Ghost
					null, 
					new MoveResultInformation(1.00F, 2.50F, Labyrinth.DOOR, Labyrinth.DEN),
					null, 
					null, 
				} } );
		checkCalcMoveDestinationIfPossible(game, being, 1.00F, 1.25F, // From the door to empty 
				new MoveResultInformation[][] {	{ //Step 0
					new MoveResultInformation(1.00F, 1.00F, Labyrinth.DOOR, Labyrinth.EMPTY), 
					null, 
					null, 
					null,
				}, { //Step n
					new MoveResultInformation(1.00F, 1.00F, Labyrinth.DOOR, Labyrinth.EMPTY), 
					null,
					null,
					null,
				}, { //Ghost
					null, 
					new MoveResultInformation(1.00F, 1.50F, Labyrinth.EMPTY, Labyrinth.DOOR), 
					null, 
					null, 
				} } );

		// Right side of the labyrinth, where the way out of the den is down
		checkCalcMoveDestinationIfPossible(game, being, 5.00F, 1.00F, // In the den 
				new MoveResultInformation[][] {	{ //Step 0
					null, 
					null, 
					new MoveResultInformation(4.75F, 1.00F, Labyrinth.DEN, Labyrinth.DEN), 
					null,
				}, { //Step n
					null, 
					new MoveResultInformation(5.00F, 1.25F, Labyrinth.DEN, Labyrinth.DOOR), 
					new MoveResultInformation(4.75F, 1.00F, Labyrinth.DEN, Labyrinth.DEN), 
					null,
				}, { //Ghost
					null, 
					null, 
					null, 
					null, 
				} } );
		checkCalcMoveDestinationIfPossible(game, being, 5.00F, 1.75F, // From the den to the door
				new MoveResultInformation[][] {	{ //Step 0
					null, 
					null, 
					null, 
					null,
				}, { //Step n
					null, 
					new MoveResultInformation(5.00F, 2.00F, Labyrinth.DEN, Labyrinth.DOOR), 
					null, 
					null,
				}, { //Ghost
					new MoveResultInformation(5.00F, 1.50F, Labyrinth.DOOR, Labyrinth.DEN), 
					null, 
					null, 
					null, 
				} } );
	}

	public void checkRateAllDirections(final Game game, final Opponent being, final int[] expected) {
		final String text = "Frightened=" + (game.remainingTicksPillIsActive() != Game.PILL_NOT_ACTIVE) + ", " + being.getId() + "(" + being.getX() + "," + being.getY() + "," + (being.isGhost() ? "Ghost" : "normal") + ") ";
		being.setDirection(Direction.UNDEFINED);
		assertArrayEquals(text, expected, being.rateAllDirections(game));
	}

	public void checkRateAllDirections(final int x, final int y, final int[][] expected) {
		final Game game = new Game(null);
		game.getBoard().init(LABYRINTH_WITH_DEN, LABYRINTH_WITH_DEN_BEINGS);
		final Opponent being = (Opponent)game.getBoard().getBeing(Being.BLINKY);
		being.set(x, y);
		game.setStep(Opponent.STEPS_IN_DEN * 5);

		// Normal opponent, no ghost, not frightened
		being.setGhost(false);
		checkRateAllDirections(game, being, expected[0]);

		// frightened
		game.eatPill();
		checkRateAllDirections(game, being, expected[1]);

		// Ghost
		being.setGhost(true);
		checkRateAllDirections(game, being, expected[2]);
	}

	@Test
	public void testRateAllDirections() {
		// Left side of the labyrinth, where the way out of the den is up
		checkRateAllDirections(1, 3, new int[][]{ {10, 0, 0, 2}, {10, 0, 0, 2}, {0, 0, 0, 0} });
		checkRateAllDirections(1, 2, new int[][]{ {6, 0, 0, 0}, {6, 0, 0, 0}, {0, 6, 0, 0} });
		checkRateAllDirections(1, 1, new int[][]{ {0, 0, 0, 2}, {0, 0, 0, 2}, {0, 10, 0, 2} });
		// Right side of the labyrinth, where the way out of the den is down
		checkRateAllDirections(5, 1, new int[][]{ {0, 10, 2, 0}, {0, 10, 2, 0}, {0, 0, 0, 0} });
	}

	@Test
	public void testDefectGhostsDontFindDen() {
		final Game game = new Game(null);
		game.getBoard().init(Labyrinth.LABYRINTH_PACMAN, Labyrinth.BEINGS);
		final Opponent being = (Opponent)game.getBoard().getBeing(Being.BLINKY);
		game.setStep(Opponent.STEPS_IN_DEN * 5);
		being.set(13, 11);
		being.setGhost(true);
		being.setDirection(Direction.UNDEFINED);
		final int[] rating = being.rateAllDirections(game);
		assertArrayEquals(new int[]{0, 10, 2, 2}, rating);
	}

	/**
	 * @see Being#equals(Object)
	 * @see Being#clone()
	 */
	@Test
	public void testObjectBasics() {
		final Being pacman = new Pacman(1.0F, 2.75F, Direction.UP, 0);
		final Being opponent = new Opponent(1, 3, Direction.UP, Being.BLINKY, 0);
		GenericTest.checkObjectBasics(false, pacman, opponent);
	}
}
