package de.ped.tools;

/**
 * Allows integration of assertions into your program.
 * An assertion is a condition (boolean expression) that must be met (true) at this position.
 * If the assertion fails, a programmer made a mistake somewhere.
 * <br>
 * Traditionally there are so-called preconditions (assertions at the beginning of a method),
 * postconditions (assertions when leaving a method) and other assertions inside a method.
 * 
 * @author Qusay H. Mahmoud for the April 1998 issue of Visual J++ Developer's Journal.
 */
public abstract class Assert {
	/** Iff this variable is <code>true</code>, assertions will be checked. */
	public static boolean NDEBUG = true;

	/**
	 * Checks a given assertion.
	 * 
	 * @param text
	 *            the message text of the exception if the assertion is not met
	 * @param expression
	 *            the assertion that must be met
	 * @return always <code>true</code>
	 * @throws AssertionException
	 *             if the assertion is not met
	 */
	public static boolean assertTrue(final String text, final boolean expression) throws AssertionException {
		assert expression;
		boolean ret = true;
		if (NDEBUG && !expression) {
			ret = false;
			if (null == text) {
				throw new AssertionException();
			} else {
				throw new AssertionException(text);
			}
		}
		return ret;
	}

	/**
	 * Checks a given assertion.
	 * 
	 * @param expression
	 *            the assertion that must be met
	 * @return always <code>true</code>
	 * @throws AssertionException
	 *             if the assertion is not met
	 */
	public static boolean assertTrue(final boolean expression) throws AssertionException {
		return assertTrue(null, expression);
	}

	public static boolean assertFalse(final String text, final boolean expression) {
		return assertTrue(text, !expression);
	}

	public static boolean assertFalse(final boolean expression) {
		return assertTrue(null, !expression);
	}

	public static boolean assertNull(final String text, final Object object) {
		return assertTrue(text, null == object);
	}

	public static boolean assertNull(final Object object) {
		return assertNull("Expected " + object + " to be null", object);
	}

	public static boolean assertNotNull(final String text, final Object object) {
		return assertTrue(text, null != object);
	}

	public static boolean assertNotNull(final Object object) {
		return assertNotNull("Unexpected null value", object);
	}

	public static boolean assertEquals(final String text, final boolean expected, final boolean observed) {
		return assertTrue(text, expected == observed);
	}

	public static boolean assertEquals(final boolean expected, final boolean observed) {
		return assertEquals("Expected " + expected + " but was " + observed, expected, observed);
	}

	public static boolean assertEquals(final String text, final long expected, final long observed) {
		return assertTrue(text, expected == observed);
	}

	public static boolean assertEquals(final long expected, final long observed) {
		return assertEquals("Expected " + expected + " but was " + observed, expected, observed);
	}

	public static boolean assertEquals(final String text, final Object expected, final Object observed) {
		if (null == expected) {
			return assertNull(text, observed);
		} else {
			return assertTrue(text, expected.equals(observed));
		}
	}

	public static boolean assertEquals(final Object expected, final Object observed) {
		return assertEquals("Expected " + expected + " but was " + observed, expected, observed);
	}

	public static boolean assertNotEquals(final String text, final boolean expected, final boolean observed) {
		return assertFalse(text, expected == observed);
	}

	public static boolean assertNotEquals(final boolean expected, final boolean observed) {
		return assertNotEquals("Expected no " + expected + ", but was " + observed, expected, observed);
	}

	public static boolean assertNotEquals(final String text, final long expected, final long observed) {
		return assertFalse(text, expected == observed);
	}

	public static boolean assertNotEquals(final long expected, final long observed) {
		return assertNotEquals("Expected no " + expected + ", but was " + observed, expected, observed);
	}

	public static boolean assertNotEquals(final String text, final Object expected, final Object observed) {
		if (null == expected) {
			return assertNotNull(text, observed);
		} else {
			return assertFalse(text, expected.equals(observed));
		}
	}

	public static boolean assertNotEquals(final Object expected, final Object observed) {
		return assertNotEquals("Expected no " + expected + ", but was " + observed, expected, observed);
	}
}
