package de.ped.tools;

/**
 * A bit field containing 32 bits.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Bitfield32 implements Cloneable, Comparable<Bitfield32> {
	public static int SIZE = Integer.SIZE;

	/**
	 * @param bitno
	 *            the index of the bit (0 = the least significant bit)
	 * @return the bitmask with only the bit with the given index set
	 */
	public static int bitmask(final int bitno) {
		return (0x1 << bitno);
	}

	/**
	 * @param firstBit
	 *            the index of the least significant bit set
	 * @param numberOfBits
	 *            the number of bits to set, starting at the least significant
	 *            bit
	 * @return the bitmask with only the bits starting at the given index set.
	 */
	public static int bitmask(final int firstBit, final int numberOfBits) {
		int ret = 0x0;
		if (firstBit < SIZE) {
			for (int i = 0; i < numberOfBits; i++) {
				ret = (ret << 1) | 0x1;
			}
			return (ret << firstBit);
		}
		return ret;
	}

	/**
	 * @param area
	 *            the bit area
	 * @return the bitmask with only the bits defined by the bit area set
	 */
	public static int bitmask(final BitArea area) {
		return area.bits32mask;
	}

	/**
	 * @param area
	 *            The bit area.
	 * @return a bitmask starting at the least significant bit and filling as
	 *         many least significant bits as the bit area has number of bits.
	 *         So it defines a mask for the result of {@link #getBits(BitArea)}
	 *         or the input to {@link #setBits(BitArea, int)}.
	 */
	public static int normalizedBitmask(final BitArea area) {
		return (area.bits32mask >>> area.firstBit);
	}

	/**
	 * @param areas
	 *            all areas to see
	 * @return a bitmask where all the bits of the bitmasks of the
	 *         <em>areas</em> are set
	 */
	public static int bitfield(final BitArea[] areas) {
		int mask = 0x00;
		if (null != areas) {
			for (BitArea area : areas) {
				mask |= area.bits32mask;
			}
		}
		return mask;
	}

	/**
	 * Sets or clears the given bit in the given mask.
	 * 
	 * @param field
	 *            the field to change the bit for
	 * @param bitno
	 *            the index of the bit to change (0 = the least significant bit)
	 * @param value
	 *            the value to change that bit to
	 * @return the resulting field
	 */
	public static int setBit(int field, final int bitno, final boolean value) {
		final int bitmask = bitmask(bitno);
		if (value) {
			field |= bitmask;
		} else {
			field &= (~bitmask);
		}
		return field;
	}

	/**
	 * Returns the given bit in the given mask.
	 * 
	 * @param field
	 *            the field to check the bit from
	 * @param bitno
	 *            the index of the bit to return (0 = the least significant bit)
	 * @return <code>true</code> if that bit was 1, <code>false</code> if that bit was 0
	 */
	public static boolean isBit(final int field, final int bitno) {
		return (0 != (field & bitmask(bitno)));
	}

	public static int setBits(final int field, final int firstBit, final int numberOfBits, final int value) {
		final int bitmask = bitmask(firstBit, numberOfBits);
		return ((field & (~bitmask)) | (value << firstBit) & bitmask);
	}

	public static int setBits(final int field, final BitArea bitArea, final int value) {
		return ((field & (~bitArea.bits32mask)) | (value << bitArea.firstBit) & bitArea.bits32mask);
	}

	public static int getBits(final int field, final int firstBit, final int numberOfBits) {
		return (field & bitmask(firstBit, numberOfBits)) >>> firstBit;
	}

	public static int getBits(final int field, final BitArea bitArea) {
		return (field & bitArea.bits32mask) >>> bitArea.firstBit;
	}

	public Bitfield32() {
		this.field = 0x0;
	}

	public Bitfield32(final int field) {
		this.field = field;
	}

	public void setBit(final int bitno, final boolean value) {
		field = setBit(field, bitno, value);
	}

	public boolean isBit(final int bitno) {
		return isBit(field, bitno);
	}

	public void setBits(final int firstBit, final int numberOfBits, final int value) {
		field = setBits(field, firstBit, numberOfBits, value);
	}

	public void setBits(final BitArea bitArea, final int value) {
		field = setBits(field, bitArea, value);
	}

	public int getBits(final int firstBit, final int numberOfBits) {
		return getBits(field, firstBit, numberOfBits);
	}

	public int getBits(final BitArea bitArea) {
		return getBits(field, bitArea);
	}

	public int getAll() {
		return field;
	}

	public void setAll(int field) {
		this.field = field;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + field;
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final Bitfield32 other = (Bitfield32) obj;
		if (field != other.field)
			return false;
		return true;
	}

	@Override
	public int compareTo(Bitfield32 other) {
		return field - other.field;
	}

	@Override
	public Bitfield32 clone() {
		return new Bitfield32(field);
	}

	/** The full 32-bit bitmask */
	private int field;
}
