package de.ped.tools;

/**
 * A bit field containing 64 bits.
 * 
 * @author Copyright (c) 2015 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Bitfield64 implements Cloneable, Comparable<Bitfield64> {
	public static int SIZE = Long.SIZE;

	/**
	 * @param bitno
	 *            the index of the bit
	 * @return the bitmask with only the bit with the given index set
	 */
	public static long bitmask(final int bitno) {
		return (1L << bitno);
	}

	/**
	 * @param firstBit
	 *            the index of the least significant bit set
	 * @param numberOfBits
	 *            the number of bits to set, starting at the least significant
	 *            bit
	 * @return the bitmask with only the bits starting at the given index set
	 */
	public static long bitmask(final int firstBit, final int numberOfBits) {
		long ret = 0x0;
		if (firstBit < SIZE) {
			for (int i = 0; i < numberOfBits; i++) {
				ret = (ret << 1) | 0x1;
			}
			return (ret << firstBit);
		}
		return ret;
	}

	/**
	 * @param area
	 *            the bit area
	 * @return the bitmask with only the bits defined by the bit area set
	 */
	public static long bitmask(final BitArea area) {
		return area.bits64mask;
	}

	/**
	 * @param area
	 *            the bit area
	 * @return a bitmask starting at the least significant bit and filling as
	 *         many least significant bits as the bit area has number of bits.
	 *         So it defines a mask for the result of {@link #getBits(BitArea)}
	 *         or the input to {@link #setBits(BitArea, long)}.
	 */
	public static long normalizedBitmask(final BitArea area) {
		return (area.bits64mask >>> area.firstBit);
	}

	/**
	 * @param areas
	 *            all areas to see
	 * @return a bitmask where all the bits of the bitmasks of the
	 *         <em>areas</em> are set
	 */
	public static long bitfield(final BitArea[] areas) {
		long mask = 0x00;
		if (null != areas) {
			for (BitArea area : areas) {
				mask |= area.bits64mask;
			}
		}
		return mask;
	}

	/**
	 * Sets or clears the given bit in the given mask.
	 * 
	 * @param field
	 *            the field to change the bit for
	 * @param bitno
	 *            the index of the bit to change
	 * @param value
	 *            the value to change that bit to
	 * @return the resulting field
	 */
	public static long setBit(long field, final int bitno, final boolean value) {
		final long bitmask = bitmask(bitno);
		if (value) {
			field |= bitmask;
		} else {
			field &= (~bitmask);
		}
		return field;
	}

	public static boolean isBit(final long field, final int bitno) {
		return (0 != (field & bitmask(bitno)));
	}

	public static long setBits(final long field, final int firstBit, final int numberOfBits, final long value) {
		final long bitmask = bitmask(firstBit, numberOfBits);
		return ((field & (~bitmask)) | (value << firstBit) & bitmask);
	}

	public static long setBits(final long field, final BitArea bitArea, final long value) {
		return ((field & (~bitArea.bits64mask)) | (value << bitArea.firstBit) & bitArea.bits64mask);
	}

	public static long getBits(final long field, final int firstBit, final int numberOfBits) {
		return (field & bitmask(firstBit, numberOfBits)) >>> firstBit;
	}

	public static long getBits(final long field, final BitArea bitArea) {
		return (field & bitArea.bits64mask) >>> bitArea.firstBit;
	}

	public Bitfield64() {
		this.field = 0x0;
	}

	public Bitfield64(final long field) {
		this.field = field;
	}

	public void setBit(final int bitno, final boolean value) {
		field = setBit(field, bitno, value);
	}

	public boolean isBit(final int bitno) {
		return isBit(field, bitno);
	}

	public void setBits(final int firstBit, final int numberOfBits, final long value) {
		field = setBits(field, firstBit, numberOfBits, value);
	}

	public void setBits(final BitArea bitArea, final long value) {
		field = setBits(field, bitArea, value);
	}

	public long getBits(final int firstBit, final int numberOfBits) {
		return getBits(field, firstBit, numberOfBits);
	}

	public long getBits(final BitArea bitArea) {
		return getBits(field, bitArea);
	}

	public long getAll() {
		return field;
	}

	public void setAll(long field) {
		this.field = field;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (int) (field ^ (field >>> 32));
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Bitfield64 other = (Bitfield64) obj;
		if (field != other.field)
			return false;
		return true;
	}

	@Override
	public int compareTo(Bitfield64 other) {
		final long upper = (field >>> 32) - (other.field >>> 32);
		if (0 != upper) {
			return (int) upper;
		}
		final long lowerMask = 0xffffffff;
		return (int)((field & lowerMask) - (other.field & lowerMask));
	}

	@Override
	public Bitfield64 clone() {
		return new Bitfield64(field);
	}

	/** The full 64-bit bitmask */
	private long field;
}
