package de.ped.tools;

import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Useful routines for string, array and collection handling.
 *
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class CollectionUtil {
	/** An offset value indicating that something does not exist. Can be safely assumed to be -1. */
	public static final int NOT_FOUND = -1;
	
	/**
	 * @param string
	 *            the string to check
	 * @return if the string is <code>null</code> or an empty string.
	 */
	public static boolean isNullOrEmpty(final String string) {
		return ((null == string) || string.isEmpty());
	}

	/**
	 * @param collection
	 *            the collection to check
	 * @return if <em>collection</em> is <code>null</code> or contains no element
	 */
	public static boolean isNullOrEmpty(final Collection<?> collection) {
		return ((null == collection) || collection.isEmpty());
	}

	/**
	 * @param <T>
	 *            any type
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element
	 *         equal to <em>value</em>. If <em>value</em> is <code>null</code>,
	 *         then any <code>null</code> element fits.
	 */
	public static <T> boolean contains(final T[] array, final T value) {
		return (NOT_FOUND != indexOf(array, value));
	}

	/**
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element equal to
	 *         <em>value</em>
	 */
	public static boolean contains(final boolean[] array, final boolean value) {
		return (NOT_FOUND != indexOf(array, value));
	}

	/**
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element equal to
	 *         <em>value</em>
	 */
	public static boolean contains(final byte[] array, final byte value) {
		return (NOT_FOUND != indexOf(array, value));
	}

	/**
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element equal to
	 *         <em>value</em>
	 */
	public static boolean contains(final int[] array, final int value) {
		return (NOT_FOUND != indexOf(array, value));
	}

	/**
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element equal to
	 *         <em>value</em>
	 */
	public static boolean contains(final long[] array, final long value) {
		return (NOT_FOUND != indexOf(array, value));
	}

	/**
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element equal to
	 *         <em>value</em>
	 */
	public static boolean contains(final char[] array, final char value) {
		return (NOT_FOUND != indexOf(array, value));
	}

	/**
	 * @param array
	 *            the array to check
	 * @param value
	 *            the element to search for
	 * @return <code>true</code> if the array contains at least one element equal to
	 *         <em>value</em>
	 */
	public static boolean containsIgnoreCase(final String[] array, final String value) {
		return (NOT_FOUND != indexOfIgnoreCase(array, value));
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em>. If <em>value</em> is <code>null</code>, then any
	 * <code>null</code> element fits.
	 * 
	 * @param <T>
	 *            any type
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static <T> int indexOf(final T[] array, final T value) {
		return indexOf(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em>. If <em>value</em>
	 * is <code>null</code>, then any <code>null</code> element fits.
	 * 
	 * @param <T>
	 *            any type
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search 
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static <T> int indexOf(final T[] array, final T value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				final T e = array[i];
				if ((e == value) || (value != null) && value.equals(e)) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final boolean[] array, boolean value) {
		return indexOf(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final boolean[] array, final boolean value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				if (array[i] == value) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final byte[] array, byte value) {
		return indexOf(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final byte[] array, final byte value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				if (array[i] == value) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final int[] array, int value) {
		return indexOf(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final int[] array, final int value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				if (array[i] == value) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final long[] array, long value) {
		return indexOf(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final long[] array, final long value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				if (array[i] == value) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final char[] array, char value) {
		return indexOf(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOf(final char[] array, final char value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				if (array[i] == value) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of an element equal to
	 * <em>value</em> (compared by {@link String#equalsIgnoreCase(String)}). If
	 * <em>value</em> is <code>null</code>, then any <code>null</code> element fits.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOfIgnoreCase(final String[] array, final String value) {
		return indexOfIgnoreCase(array, value, 0);
	}

	/**
	 * Searches <em>array</em>, starting at index <em>fromIndex</em>, for the
	 * first occurrence of an element equal to <em>value</em> (compared by
	 * {@link String#equalsIgnoreCase(String)}). If <em>value</em> is
	 * <code>null</code>, then any <code>null</code> element fits.
	 * 
	 * @param array
	 *            the array to search
	 * @param value
	 *            the element to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none.
	 */
	public static int indexOfIgnoreCase(final String[] array, final String value, final int fromIndex) {
		if (null != array) {
			for (int i = fromIndex; i < array.length; i++) {
				final String e = array[i];
				if ((e == value) || (value != null) && value.equalsIgnoreCase(e)) {
					return i;
				}
			}
		}
		return NOT_FOUND;
	}

	/**
	 * Searches <em>array</em> for the first occurrence of a sequence of
	 * elements equal to <em>part</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param part
	 *            the part to look for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOfPart(final byte[] array, final byte[] part) {
		return indexOfPart(array, part, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of a sequence of elements equal to <em>part</em>.
	 * 
	 * @param array
	 *            the array to search
	 * @param part
	 *            the part to look for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOfPart(final byte[] array, final byte[] part, final int fromIndex) {
		int ret = NOT_FOUND;
		int posInPart = 0;
		if ((null != array) && (null != part)) {
			for (int posInArray = fromIndex; posInArray < array.length; posInArray++) {
				if (array[posInArray] == part[posInPart]) {
					posInPart++;
					if (posInPart >= part.length) {
						ret = posInArray - posInPart + 1;
						break;
					}
				} else {
					posInArray -= posInPart;
					posInPart = 0;
				}
			}
		}
		return ret;
	}

	/**
	 * Searches <em>array</em> for the
	 * first occurrence of a sequence of elements equal to one of the
	 * <em>partsToSearchFor</em>. Current implementation assumes that parts to
	 * search for are all of same length, and are all a repetition of a single
	 * byte.
	 * 
	 * @param array
	 *            the array to search
	 * @param partsToSearchFor
	 *            the parts to search for
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOfFirstPart(final byte[] array, final byte[][] partsToSearchFor) {
		return indexOfFirstPart(array, partsToSearchFor, 0);
	}

	/**
	 * Searches <em>array</em>, starting at position <em>fromIndex</em>, for the
	 * first occurrence of a sequence of elements equal to one of the
	 * <em>partsToSearchFor</em>. Current implementation assumes that parts to
	 * search for are all of same length, and are all a repetition of a single
	 * byte.
	 * 
	 * @param array
	 *            the array to search
	 * @param partsToSearchFor
	 *            the parts to search for
	 * @param fromIndex
	 *            the first index of the array where to start the search
	 * @return the index of the first occurrence, or {@link #NOT_FOUND} if there
	 *         was none
	 */
	public static int indexOfFirstPart(final byte[] array, final byte[][] partsToSearchFor, final int fromIndex) {
		int ret = NOT_FOUND;
		if ((null != array) && (null != partsToSearchFor)) {
			int posInPart = 0;
			int matching = NOT_FOUND;
			for (int posInArray = fromIndex; posInArray < array.length; posInArray++) {
				final byte current = array[posInArray];
				int matchingCurrent = NOT_FOUND;
				for (int pat = 0; pat < partsToSearchFor.length; pat++) {
					if ((null != partsToSearchFor[pat]) && (0 < partsToSearchFor[pat].length) &&  (current == partsToSearchFor[pat][0])) {
						matchingCurrent = pat;
						break;
					}
				}
				if (NOT_FOUND == matchingCurrent) { // Current char fits no pattern
					posInPart = 0;
					matching = NOT_FOUND;
				} else if (matching == matchingCurrent) { // Current char fits current pattern
					posInPart++;
					if (posInPart >= partsToSearchFor[matching].length) {
						ret = posInArray - posInPart + 1;
						break;
					}
				} else { // Current char may be the start of another pattern
					posInPart = 1;
					matching = matchingCurrent;
				}
			}
		}
		return ret;
	}

	/**
	 * Works like {@link String#replaceFirst(String, String)}, but for the last
	 * occurrence.
	 * 
	 * Thanks to <a href=
	 * "http://stackoverflow.com/questions/2282728/java-replacelast">stackoverflow</a>.
	 * 
	 * @param input
	 *            the string to run the changes on
	 * @param regex
	 *            the regular expression to to search for
	 * @param replacement
	 *            the string that shall replace the last occurrence of
	 *            <em>regex</em>
	 * @return the modified <em>input</em>
	 */
	public static String replaceLast(final String input, final String regex, final String replacement) {
		final Pattern pattern = Pattern.compile(regex);
		final Matcher matcher = pattern.matcher(input);
		if (!matcher.find()) {
			return input;
		}
		int lastMatchStart = 0;
		do {
			lastMatchStart = matcher.start();
		} while (matcher.find());
		matcher.find(lastMatchStart);
		final StringBuffer sb = new StringBuffer(input.length());
		matcher.appendReplacement(sb, replacement);
		matcher.appendTail(sb);
		return sb.toString();
	}

	/**
	 * Reverts the array in-place - the previously last element becomes the
	 * first one, and so on.
	 * 
	 * @param <T>
	 *            any type
	 * @param array
	 *            the array to revert in-place
	 */
	public static <T> void revert(T[] array) {
		for (int i = 0; i < array.length / 2; i++) {
			final int j = array.length - 1 - i;
			final T buf = array[i];
			array[i] = array[j];
			array[j] = buf;
		}
	}
}
