package de.ped.tools;

import java.util.ArrayList;

/**
 * All the 8 directions a digital joystick could take.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DigitalJoystickDirection {
	public static final int MASK_UP = 0x01;
	public static final int MASK_DOWN = 0x02;
	public static final int MASK_LEFT = 0x04;
	public static final int MASK_RIGHT = 0x08;
	public static final int MASK_ALL_DIRECTIONS = MASK_UP | MASK_DOWN | MASK_LEFT | MASK_RIGHT;

	public static final int MASK_UP_LEFT    = MASK_UP   | MASK_LEFT;
	public static final int MASK_UP_RIGHT   = MASK_UP   | MASK_RIGHT;
	public static final int MASK_DOWN_LEFT  = MASK_DOWN | MASK_LEFT;
	public static final int MASK_DOWN_RIGHT = MASK_DOWN | MASK_RIGHT;

	public static final DigitalJoystickDirection UP         = new DigitalJoystickDirection( 0, -1, "Game.PlayerAction.U.Text",    90, MASK_UP,         MASK_UP_LEFT,    MASK_UP_RIGHT);
	public static final DigitalJoystickDirection DOWN       = new DigitalJoystickDirection( 0,  1, "Game.PlayerAction.D.Text",   -90, MASK_DOWN,       MASK_DOWN_RIGHT, MASK_DOWN_LEFT);
	public static final DigitalJoystickDirection LEFT       = new DigitalJoystickDirection(-1,  0, "Game.PlayerAction.L.Text",   180, MASK_LEFT,       MASK_DOWN_LEFT,  MASK_UP_LEFT);
	public static final DigitalJoystickDirection RIGHT      = new DigitalJoystickDirection( 1,  0, "Game.PlayerAction.R.Text",     0, MASK_RIGHT,      MASK_UP_RIGHT,   MASK_DOWN_RIGHT);
	public static final DigitalJoystickDirection UP_LEFT    = new DigitalJoystickDirection(-1, -1, "Game.PlayerAction.UL.Text",  135, MASK_UP_LEFT,    MASK_LEFT,       MASK_UP);
	public static final DigitalJoystickDirection UP_RIGHT   = new DigitalJoystickDirection( 1, -1, "Game.PlayerAction.UR.Text",   45, MASK_UP_RIGHT,   MASK_UP,         MASK_RIGHT);
	public static final DigitalJoystickDirection DOWN_LEFT  = new DigitalJoystickDirection(-1,  1, "Game.PlayerAction.DL.Text", -135, MASK_DOWN_LEFT,  MASK_DOWN,       MASK_LEFT);
	public static final DigitalJoystickDirection DOWN_RIGHT = new DigitalJoystickDirection( 1,  1, "Game.PlayerAction.DR.Text",  -45, MASK_DOWN_RIGHT, MASK_RIGHT,      MASK_DOWN);

	/**
	 * Checks if a direction contains as part another direction, e.g.
	 * {@link #MASK_UP_LEFT} contains {@link #MASK_UP} and {@link #MASK_LEFT}.
	 * 
	 * @param current
	 *            the bitmask representation of the (maybe composed) direction
	 *            to check
	 * @param part
	 *            the bitmask representation of the direction part to check for
	 * @return if the current mask contains the part
	 */
	public static boolean is(int current, int part) {
		return ((current & part) != 0);
	}

	protected static DigitalJoystickDirection[] directions = new DigitalJoystickDirection[0x10]; 
	static {
		directions[MASK_UP]         = UP;
		directions[MASK_DOWN]       = DOWN;
		directions[MASK_LEFT]       = LEFT;
		directions[MASK_RIGHT]      = RIGHT;
		directions[MASK_UP_LEFT]    = UP_LEFT;
		directions[MASK_UP_RIGHT]   = UP_RIGHT;
		directions[MASK_DOWN_LEFT]  = DOWN_LEFT;
		directions[MASK_DOWN_RIGHT] = DOWN_RIGHT;
	}

	private static ArrayList<DigitalJoystickDirection> directions4;
	private static ArrayList<DigitalJoystickDirection> directions8;

	/** @return the 4 standard directions in counter-clockwise order. */
	public static ArrayList<DigitalJoystickDirection> getDirections4() {
		if (null == directions4) {
			final ArrayList<DigitalJoystickDirection> arr = new ArrayList<DigitalJoystickDirection>(8);
			arr.add(UP);
			arr.add(LEFT);
			arr.add(DOWN);
			arr.add(RIGHT);
			directions4 = arr;
		}
		return directions4;
	}

	/** @return the 8 directions (including diagonals) in counter-clockwise order. */
	public static ArrayList<DigitalJoystickDirection> getDirections8() {
		if (null == directions8) {
			final ArrayList<DigitalJoystickDirection> arr = new ArrayList<DigitalJoystickDirection>(8);
			arr.add(UP);
			arr.add(UP_LEFT);
			arr.add(LEFT);
			arr.add(DOWN_LEFT);
			arr.add(DOWN);
			arr.add(DOWN_RIGHT);
			arr.add(RIGHT);
			arr.add(UP_RIGHT);
			directions8 = arr;
		}
		return directions8;
	}

	public static DigitalJoystickDirection getDirection(final int mask) {
		return directions[mask];
	}

	public static DigitalJoystickDirection getDirection(final int xdiff, final int ydiff) {
		DigitalJoystickDirection ret = null; 
		for (final DigitalJoystickDirection dir : directions) {
			if ((null != dir) && (dir.getX() == xdiff) && (dir.getY() == ydiff)) {
				ret = dir;
				break;
			}
		}
		return ret;
	}

	public static DigitalJoystickDirection getDirection(final int[] vector) {
		return getDirection(vector[0], vector[1]);
	}

	public static DigitalJoystickDirection getDirection(final PlayfieldDistance vector) {
		return getDirection(vector.getX(), vector.getY());
	}

	/** The bitmask that represents the direction. */
	public final int mask;
	/** The message key that represents the direction textually. */
	public final String key;
	/** The (x,y) representation of the normalized vector of that direction in a coordinate system with (0, 0) in the upper left corner. */
	private final PlayfieldDistance vector;
	/** The mask of the direction if you turn 45° counter-clockwise from this direction. */
	public final int maskNextCounterClockwise;
	/** The mask of the direction if you turn 45° clockwise from this direction. */
	public final int maskNextClockwise;
	/** The angle of the direction, counted counter-clockwise from right direction (like in a coordinate system). */
	public final int angleCounterClockwise;

	protected DigitalJoystickDirection(final int x, final int y, final String key, final int angle, final int mask, final int maskNextCounterClockwise, final int maskNextClockwise) {
		this.vector = new PlayfieldDistance(x, y);
		this.mask = mask;
		this.key = key;
		this.angleCounterClockwise = angle;
		this.maskNextCounterClockwise = maskNextCounterClockwise;
		this.maskNextClockwise = maskNextClockwise;
	}

	public int getId() {
		for (int i = 0; i < directions.length; i++) {
			final DigitalJoystickDirection dir = directions[i];
			if (dir == this) {
				return i;
			}
		}
		throw new IllegalStateException("This instance should never exist!");
	}

	public int getX() {
		return vector.getX();
	}

	public int getY() {
		return vector.getY();
	}

	public PlayfieldDistance getDistanceVector() {
		return vector;
	}

	/**
	 * @param steps
	 *            the number of steps to turn left by 45 degrees
	 * @return the direction you get when you turn counter-clockwise by 45
	 *         degrees for the number of steps
	 */
	public DigitalJoystickDirection turnLeft8(final int steps) {
		DigitalJoystickDirection ret = this;
		final int normedSteps = MathUtil.modulo(steps, 8);
		if (steps < 4) {
			for (int i = 0; i < normedSteps; i++) {
				ret = getDirection(ret.maskNextCounterClockwise);
			}
		} else {
			for (int i = 8; i > normedSteps; i--) {
				ret = getDirection(ret.maskNextClockwise);
			}
		}
		return ret;
	}

	/**
	 * @param steps
	 *            the number of steps to turn right by 45 degrees
	 * @return the direction you get when you turn clockwise by 45 degrees for
	 *         the number of steps
	 */
	public DigitalJoystickDirection turnRight8(final int steps) {
		return turnLeft8(-steps);
	}

	/**
	 * @return the opposite direction
	 */
	public DigitalJoystickDirection getOppositeDirection() {
		return turnLeft8(4);
	}

	public boolean is4Direction() {
		return (getDirections4().contains(this));
	}

	@Override
	public String toString() {
		return key;
	}
}
