package de.ped.tools;

import java.util.HashSet;

import de.ped.tools.log.Logger;


/**
 * A thread that repeatedly runs through a step and can be finished by another thread.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FinishableThread extends Thread {

	public static final long NO_SLEEP = -1L;

	private static int nextInstanceID = 0;

	private static int calculateInstanceID() {
		int instanceID = nextInstanceID;
		nextInstanceID++;
		return instanceID;
	}

	private static HashSet<FinishableThread> threads = new HashSet<FinishableThread>();

	public static void monitorAllThreads(final Logger logger) {
		for (final FinishableThread t : threads) {
			logger.debug(t.getName() + ": " + t.getState().name());
			final StackTraceElement[] trace = t.getStackTrace();
			for (final StackTraceElement te : trace) {
				logger.debug(te.toString());
			}
		}
	}

	transient protected Logger logger;

	private int instanceID;

	public int getInstanceID() {
		return instanceID;
	}

	/**
	 * The states a finishable threads can be in.
	 * 
	 * @author Copyright (c) 2022 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static enum State {
		/** Initial state, not running. */
		INIT,
		/** Started, but not running yet. */
		STARTING,
		/** Running. */
		RUNNING,
		/** Running, asked to terminate. */
		TERMINATING,
		/** Final state. */
		TERMINATED,
	}
	
	/** The state this finishable thread is in. */
	private State state = State.INIT;

	/** Time between two run steps in millis. A negative value means no sleep at all. */
	private long tickLength = 100L;

	public FinishableThread(final String name) {
		super(name + nextInstanceID);
		instanceID = calculateInstanceID();
	}

	public FinishableThread(final String name, final long tickLength) {
		super(name + nextInstanceID);
		instanceID = calculateInstanceID();
		setTickLength(tickLength);
	}

	public void setLogger(final Logger logger) {
		this.logger = logger;
	}

	public boolean isStarted() {
		return State.RUNNING.ordinal() <= state.ordinal();
	}

	public boolean isToTerminate() {
		return State.TERMINATING == state;
	}

	public long getTickLength() {
		return tickLength;
	}

	public void setTickLength(final long tickLength) {
		this.tickLength = tickLength;
	}

	public long getSleepDelay() {
		return tickLength;
	}

	/** Overload to do something useful. */
	public void onRunStart() {
		// Default implementation does nothing
	}

	/** Overload to do something useful. */
	public void onRunEnd() {
		// Default implementation does nothing
	}

	/** Overload to do something useful. */
	public void threadStep() {
		// Default implementation does nothing
	}

	@Override
	public synchronized void start() {
		state = State.STARTING;
		super.start();
	}

	@Override
	public void run() {
		try {
			threads.add(this);
			onRunStart();
			if (State.STARTING == state) {
				state = State.RUNNING;
			}
			while (State.TERMINATING != state) {
				threadStep();
				final long delay = getSleepDelay();
				if (0 <= delay) {
					Thread.sleep(delay);
				}
			}
		} catch (final InterruptedException e) {
			// Expected behavior to force termination
		} catch (final Throwable e) {
			if (null != logger) {
				logger.error("Unexpected exception", e);
			}
		}
		state = State.TERMINATED;
		threads.remove(this);
		onRunEnd();
	}

	public void onTerminate() {
		// Default implementation does nothing
	}

	/** 
	 * Call this from another thread's context to ask this thread to finish gracefully.
	 * Waits till this is finished, and then returns. 
	 */
	public void terminate() {
		state = State.TERMINATING;
		onTerminate();
		try {
			notify();
		} catch (final IllegalMonitorStateException ex) {
			state = State.TERMINATED;
		}
		while (state != State.TERMINATED) {
			try {
				Thread.sleep(0L);
			} catch (final InterruptedException e) {
				// Nothing to do
			}
		}
	}
}
