package de.ped.tools;

import java.util.Iterator;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;

/**
 * Code helpful to create HTML/XHTML code for use within web pages.
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class HtmlHelper {
	/**
	 * Most basic SGML representations of characters, to be safe for HTML pages, which are also predefined in XML.
	 * These are the less-than (&lt;) and greater-than (&gt;) signs, 
	 * the apersand (&amp;), the double quotation mark (&quot;) and the apostrophe (').
	 * Don't change the order of the conversions, as we don't want recursive representations.
	 */
	public static final String[][] SGML_CONVERSIONS = new String[][] {
		{ "&",  "&amp;"  },
		{ "<",  "&lt;"   },
		{ ">",  "&gt;"   },
		{ "\"", "&quot;" },
	};

	/**
	 * Lots more SGML representations of characters, to be safe for HTML pages.
	 * This is no longer a 1:1 mapping, so it is not reversible. So it should only be used 
	 * to convert from HTML to plain text.
	 */
	public static final String[][] EXTENDED_SGML_CONVERSIONS = new String[][] {
		{ "!",   "&exclamation;" },
		{ "%",   "&percent;" },
		{ "+",   "&add;"    },
		{ "=",   "&equal;"  },
		{ " ",   "&nbsp;"   },
		{ "§",   "&sect;"   },
		{ "(c)", "&copy;"   },
		{ "\"",  "&laquo;"  }, // <<
		{ "(R)", "&reg;"    },
		{ "°",   "&deg;"    },
		{ "\"",  "&raquo;"  }, // >>
		{ "Ä",   "&Auml;"   },
		{ "ä",   "&auml;"   },
		{ "Ö",   "&Ouml;"   },
		{ "ö",   "&ouml;"   },
		{ "Ü",   "&Uuml;"   },
		{ "ü",   "&uuml;"   },
		{ "ß",   "&szlig;"  },
		{ "-",   "&ndash;"  },
		{ "-",   "&mdash;"  },
		{ "-",   "&minus;"  }, // 2018: Not accepted by all browsers
		{ "\"",  "&bdquo;"  }, // ,,
		{ "\"",  "&ldquo;"  }, // ´´
		{ "\"",  "&rdquo;"  }, // ``
		{ "'",   "&lsquo;"  }, // ´
		{ "'",   "&rsquo;"  }, // `
		{ "'",   "&sbquo;"  }, // ,
		{ "'",   "&rsaquo;" }, // >
		{ "'",   "&lsaquo;" }, // <
		{ "...", "&hellip;" },
		{ "<-",  "&larr;" },
		{ "->",  "&rarr;" },
		{ "<->", "&harr;" },
		{ "<=",  "&lArr;" },
		{ "=>",  "&rArr;" },
		{ "<=>", "&hArr;" },
	};

	/**
	 * Helper class to allow conversion between the SGML name of a character and the unicode code point of that character.  
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static final class SgmlUnicodeRepresentation {
		public SgmlUnicodeRepresentation(final int unicode, final String name) {
			this.name = name;
			this.unicode = unicode;
		}
		
		public final String name;
		public final int unicode;
	}

	/**
	 * Lots more SGML representations of characters, to be safe for HTML pages.
	 * This is no longer a 1:1 mapping, so it is not reversable. So it should only be used 
	 * to convert from HTML to plain text.
	 * See https://en.wikipedia.org/wiki/List_of_XML_and_HTML_character_entity_references
	 */
	public static final SgmlUnicodeRepresentation[] SGML_UNICODE_REPRESENTATIONS = new SgmlUnicodeRepresentation[] {
		new SgmlUnicodeRepresentation(33, "&exclamation;"),
		new SgmlUnicodeRepresentation(34, "&quot;"),
		new SgmlUnicodeRepresentation(37, "&percent;"),
		new SgmlUnicodeRepresentation(38, "&amp;"),
		new SgmlUnicodeRepresentation(39, "&apos;"),
		new SgmlUnicodeRepresentation(43, "&add;"),
		new SgmlUnicodeRepresentation(60, "&lt;"),
		new SgmlUnicodeRepresentation(61, "&equal;"),
		new SgmlUnicodeRepresentation(62, "&gt;"),
		new SgmlUnicodeRepresentation(160, "&nbsp;"),
		new SgmlUnicodeRepresentation(161, "&iexcl;"),
		new SgmlUnicodeRepresentation(162, "&cent;"),
		new SgmlUnicodeRepresentation(163, "&pound;"),
		new SgmlUnicodeRepresentation(164, "&curren;"),
		new SgmlUnicodeRepresentation(165, "&yen;"),
		new SgmlUnicodeRepresentation(166, "&brvbar;"),
		new SgmlUnicodeRepresentation(167, "&sect;"),
		new SgmlUnicodeRepresentation(168, "&uml;"),
		new SgmlUnicodeRepresentation(169, "&copy;"),
		new SgmlUnicodeRepresentation(170, "&ordf;"),
		new SgmlUnicodeRepresentation(171, "&laquo;"),
		new SgmlUnicodeRepresentation(172, "&not;"),
		new SgmlUnicodeRepresentation(173, "&shy;"),
		new SgmlUnicodeRepresentation(174, "&reg;"),
		new SgmlUnicodeRepresentation(175, "&macr;"),
		new SgmlUnicodeRepresentation(176, "&deg;"),
		new SgmlUnicodeRepresentation(177, "&plusmn;"),
		new SgmlUnicodeRepresentation(178, "&sup2;"),
		new SgmlUnicodeRepresentation(179, "&sup3;"),
		new SgmlUnicodeRepresentation(180, "&acute;"),
		new SgmlUnicodeRepresentation(181, "&micro;"),
		new SgmlUnicodeRepresentation(182, "&para;"),
		new SgmlUnicodeRepresentation(183, "&middot;"),
		new SgmlUnicodeRepresentation(184, "&cedil;"),
		new SgmlUnicodeRepresentation(185, "&sup1;"),
		new SgmlUnicodeRepresentation(186, "&ordm;"),
		new SgmlUnicodeRepresentation(187, "&raquo;"),
		new SgmlUnicodeRepresentation(188, "&frac14;"),
		new SgmlUnicodeRepresentation(189, "&frac12;"),
		new SgmlUnicodeRepresentation(190, "&frac34;"),
		new SgmlUnicodeRepresentation(191, "&iquest;"),
		new SgmlUnicodeRepresentation(192, "&Agrave;"),
		new SgmlUnicodeRepresentation(193, "&Aacute;"),
		new SgmlUnicodeRepresentation(194, "&Acirc;"),
		new SgmlUnicodeRepresentation(195, "&Atilde;"),
		new SgmlUnicodeRepresentation(196, "&Auml;"),
		new SgmlUnicodeRepresentation(197, "&Aring;"),
		new SgmlUnicodeRepresentation(198, "&AElig;"),
		new SgmlUnicodeRepresentation(199, "&Ccedil;"),
		new SgmlUnicodeRepresentation(200, "&Egrave;"),
		new SgmlUnicodeRepresentation(201, "&Eacute;"),
		new SgmlUnicodeRepresentation(202, "&Ecirc;"),
		new SgmlUnicodeRepresentation(203, "&Euml;"),
		new SgmlUnicodeRepresentation(204, "&Igrave;"),
		new SgmlUnicodeRepresentation(205, "&Iacute;"),
		new SgmlUnicodeRepresentation(206, "&Icirc;"),
		new SgmlUnicodeRepresentation(207, "&Iuml;"),
		new SgmlUnicodeRepresentation(208, "&ETH;"),
		new SgmlUnicodeRepresentation(209, "&Ntilde;"),
		new SgmlUnicodeRepresentation(210, "&Ograve;"),
		new SgmlUnicodeRepresentation(211, "&Oacute;"),
		new SgmlUnicodeRepresentation(212, "&Ocirc;"),
		new SgmlUnicodeRepresentation(213, "&Otilde;"),
		new SgmlUnicodeRepresentation(214, "&Ouml;"),
		new SgmlUnicodeRepresentation(215, "&times;"),
		new SgmlUnicodeRepresentation(216, "&Oslash;"),
		new SgmlUnicodeRepresentation(217, "&Ugrave;"),
		new SgmlUnicodeRepresentation(218, "&Uacute;"),
		new SgmlUnicodeRepresentation(219, "&Ucirc;"),
		new SgmlUnicodeRepresentation(220, "&Uuml;"),
		new SgmlUnicodeRepresentation(221, "&Yacute;"),
		new SgmlUnicodeRepresentation(222, "&THORN;"),
		new SgmlUnicodeRepresentation(223, "&szlig;"),
		new SgmlUnicodeRepresentation(224, "&agrave;"),
		new SgmlUnicodeRepresentation(225, "&aacute;"),
		new SgmlUnicodeRepresentation(226, "&acirc;"),
		new SgmlUnicodeRepresentation(227, "&atilde;"),
		new SgmlUnicodeRepresentation(228, "&auml;"),
		new SgmlUnicodeRepresentation(229, "&aring;"),
		new SgmlUnicodeRepresentation(230, "&aelig;"),
		new SgmlUnicodeRepresentation(231, "&ccedil;"),
		new SgmlUnicodeRepresentation(232, "&egrave;"),
		new SgmlUnicodeRepresentation(233, "&eacute;"),
		new SgmlUnicodeRepresentation(234, "&ecirc;"),
		new SgmlUnicodeRepresentation(235, "&euml;"),
		new SgmlUnicodeRepresentation(236, "&igrave;"),
		new SgmlUnicodeRepresentation(237, "&iacute;"),
		new SgmlUnicodeRepresentation(238, "&icirc;"),
		new SgmlUnicodeRepresentation(239, "&iuml;"),
		new SgmlUnicodeRepresentation(240, "&eth;"),
		new SgmlUnicodeRepresentation(241, "&ntilde;"),
		new SgmlUnicodeRepresentation(242, "&ograve;"),
		new SgmlUnicodeRepresentation(243, "&oacute;"),
		new SgmlUnicodeRepresentation(244, "&ocirc;"),
		new SgmlUnicodeRepresentation(245, "&otilde;"),
		new SgmlUnicodeRepresentation(246, "&ouml;"),
		new SgmlUnicodeRepresentation(247, "&divide;"),
		new SgmlUnicodeRepresentation(248, "&oslash;"),
		new SgmlUnicodeRepresentation(249, "&ugrave;"),
		new SgmlUnicodeRepresentation(250, "&uacute;"),
		new SgmlUnicodeRepresentation(251, "&ucirc;"),
		new SgmlUnicodeRepresentation(252, "&uuml;"),
		new SgmlUnicodeRepresentation(253, "&yacute;"),
		new SgmlUnicodeRepresentation(254, "&thorn;"),
		new SgmlUnicodeRepresentation(255, "&yuml;"),
		new SgmlUnicodeRepresentation(338, "&OElig;"),
		new SgmlUnicodeRepresentation(339, "&oelig;"),
		new SgmlUnicodeRepresentation(352, "&Scaron;"),
		new SgmlUnicodeRepresentation(353, "&scaron;"),
		new SgmlUnicodeRepresentation(376, "&Yuml;"),
		new SgmlUnicodeRepresentation(402, "&fnof;"),
		new SgmlUnicodeRepresentation(710, "&circ;"),
		new SgmlUnicodeRepresentation(732, "&tilde;"),
		new SgmlUnicodeRepresentation(913, "&Alpha;"),
		new SgmlUnicodeRepresentation(914, "&Beta;"),
		new SgmlUnicodeRepresentation(915, "&Gamma;"),
		new SgmlUnicodeRepresentation(916, "&Delta;"),
		new SgmlUnicodeRepresentation(917, "&Epsilon;"),
		new SgmlUnicodeRepresentation(918, "&Zeta;"),
		new SgmlUnicodeRepresentation(919, "&Eta;"),
		new SgmlUnicodeRepresentation(920, "&Theta;"),
		new SgmlUnicodeRepresentation(921, "&Iota;"),
		new SgmlUnicodeRepresentation(922, "&Kappa;"),
		new SgmlUnicodeRepresentation(923, "&Lambda;"),
		new SgmlUnicodeRepresentation(924, "&Mu;"),
		new SgmlUnicodeRepresentation(925, "&Nu;"),
		new SgmlUnicodeRepresentation(926, "&Xi;"),
		new SgmlUnicodeRepresentation(927, "&Omicron;"),
		new SgmlUnicodeRepresentation(928, "&Pi;"),
		new SgmlUnicodeRepresentation(929, "&Rho;"),
		new SgmlUnicodeRepresentation(931, "&Sigma;"),
		new SgmlUnicodeRepresentation(932, "&Tau;"),
		new SgmlUnicodeRepresentation(933, "&Upsilon;"),
		new SgmlUnicodeRepresentation(934, "&Phi;"),
		new SgmlUnicodeRepresentation(935, "&Chi;"),
		new SgmlUnicodeRepresentation(936, "&Psi;"),
		new SgmlUnicodeRepresentation(937, "&Omega;"),
		new SgmlUnicodeRepresentation(945, "&alpha;"),
		new SgmlUnicodeRepresentation(946, "&beta;"),
		new SgmlUnicodeRepresentation(947, "&gamma;"),
		new SgmlUnicodeRepresentation(948, "&delta;"),
		new SgmlUnicodeRepresentation(949, "&epsilon;"),
		new SgmlUnicodeRepresentation(950, "&zeta;"),
		new SgmlUnicodeRepresentation(951, "&eta;"),
		new SgmlUnicodeRepresentation(952, "&theta;"),
		new SgmlUnicodeRepresentation(953, "&iota;"),
		new SgmlUnicodeRepresentation(954, "&kappa;"),
		new SgmlUnicodeRepresentation(955, "&lambda;"),
		new SgmlUnicodeRepresentation(956, "&mu;"),
		new SgmlUnicodeRepresentation(957, "&nu;"),
		new SgmlUnicodeRepresentation(958, "&xi;"),
		new SgmlUnicodeRepresentation(959, "&omicron;"),
		new SgmlUnicodeRepresentation(960, "&pi;"),
		new SgmlUnicodeRepresentation(961, "&rho;"),
		new SgmlUnicodeRepresentation(962, "&sigmaf;"),
		new SgmlUnicodeRepresentation(963, "&sigma;"),
		new SgmlUnicodeRepresentation(964, "&tau;"),
		new SgmlUnicodeRepresentation(965, "&upsilon;"),
		new SgmlUnicodeRepresentation(966, "&phi;"),
		new SgmlUnicodeRepresentation(967, "&chi;"),
		new SgmlUnicodeRepresentation(968, "&psi;"),
		new SgmlUnicodeRepresentation(969, "&omega;"),
		new SgmlUnicodeRepresentation(977, "&thetasym;"),
		new SgmlUnicodeRepresentation(978, "&upsih;"),
		new SgmlUnicodeRepresentation(982, "&piv;"),
		new SgmlUnicodeRepresentation(8194, "&ensp;"),
		new SgmlUnicodeRepresentation(8195, "&emsp;"),
		new SgmlUnicodeRepresentation(8201, "&thinsp;"),
		new SgmlUnicodeRepresentation(8204, "&zwnj;"),
		new SgmlUnicodeRepresentation(8205, "&zwj;"),
		new SgmlUnicodeRepresentation(8206, "&lrm;"),
		new SgmlUnicodeRepresentation(8207, "&rlm;"),
		new SgmlUnicodeRepresentation(8211, "&ndash;"),
		new SgmlUnicodeRepresentation(8212, "&mdash;"),
		new SgmlUnicodeRepresentation(8213, "&horbar;"),
		new SgmlUnicodeRepresentation(8216, "&lsquo;"),
		new SgmlUnicodeRepresentation(8217, "&rsquo;"),
		new SgmlUnicodeRepresentation(8218, "&sbquo;"),
		new SgmlUnicodeRepresentation(8220, "&ldquo;"),
		new SgmlUnicodeRepresentation(8221, "&rdquo;"),
		new SgmlUnicodeRepresentation(8222, "&bdquo;"),
		new SgmlUnicodeRepresentation(8224, "&dagger;"),
		new SgmlUnicodeRepresentation(8225, "&Dagger;"),
		new SgmlUnicodeRepresentation(8226, "&bull;"),
		new SgmlUnicodeRepresentation(8230, "&hellip;"),
		new SgmlUnicodeRepresentation(8240, "&permil;"),
		new SgmlUnicodeRepresentation(8242, "&prime;"),
		new SgmlUnicodeRepresentation(8243, "&Prime;"),
		new SgmlUnicodeRepresentation(8249, "&lsaquo;"),
		new SgmlUnicodeRepresentation(8250, "&rsaquo;"),
		new SgmlUnicodeRepresentation(8254, "&oline;"),
		new SgmlUnicodeRepresentation(8260, "&frasl;"),
		new SgmlUnicodeRepresentation(8364, "&euro;"),
		new SgmlUnicodeRepresentation(8465, "&image;"),
		new SgmlUnicodeRepresentation(8472, "&weierp;"),
		new SgmlUnicodeRepresentation(8476, "&real;"),
		new SgmlUnicodeRepresentation(8482, "&trade;"),
		new SgmlUnicodeRepresentation(8501, "&alefsym;"),
		new SgmlUnicodeRepresentation(8592, "&larr;"),
		new SgmlUnicodeRepresentation(8593, "&uarr;"),
		new SgmlUnicodeRepresentation(8594, "&rarr;"),
		new SgmlUnicodeRepresentation(8595, "&darr;"),
		new SgmlUnicodeRepresentation(8596, "&harr;"),
		new SgmlUnicodeRepresentation(8629, "&crarr;"),
		new SgmlUnicodeRepresentation(8656, "&lArr;"),
		new SgmlUnicodeRepresentation(8657, "&uArr;"),
		new SgmlUnicodeRepresentation(8658, "&rArr;"),
		new SgmlUnicodeRepresentation(8659, "&dArr;"),
		new SgmlUnicodeRepresentation(8660, "&hArr;"),
		new SgmlUnicodeRepresentation(8704, "&forall;"),
		new SgmlUnicodeRepresentation(8706, "&part;"),
		new SgmlUnicodeRepresentation(8707, "&exist;"),
		new SgmlUnicodeRepresentation(8709, "&empty;"),
		new SgmlUnicodeRepresentation(8711, "&nabla;"),
		new SgmlUnicodeRepresentation(8712, "&isin;"),
		new SgmlUnicodeRepresentation(8713, "&notin;"),
		new SgmlUnicodeRepresentation(8715, "&ni;"),
		new SgmlUnicodeRepresentation(8719, "&prod;"),
		new SgmlUnicodeRepresentation(8721, "&sum;"),
		new SgmlUnicodeRepresentation(8722, "&minus;"),
		new SgmlUnicodeRepresentation(8727, "&lowast;"),
		new SgmlUnicodeRepresentation(8730, "&radic;"),
		new SgmlUnicodeRepresentation(8733, "&prop;"),
		new SgmlUnicodeRepresentation(8734, "&infin;"),
		new SgmlUnicodeRepresentation(8736, "&ang;"),
		new SgmlUnicodeRepresentation(8743, "&and;"),
		new SgmlUnicodeRepresentation(8744, "&or;"),
		new SgmlUnicodeRepresentation(8745, "&cap;"),
		new SgmlUnicodeRepresentation(8746, "&cup;"),
		new SgmlUnicodeRepresentation(8747, "&int;"),
		new SgmlUnicodeRepresentation(8756, "&there4;"),
		new SgmlUnicodeRepresentation(8764, "&sim;"),
		new SgmlUnicodeRepresentation(8773, "&cong;"),
		new SgmlUnicodeRepresentation(8776, "&asymp;"),
		new SgmlUnicodeRepresentation(8800, "&ne;"),
		new SgmlUnicodeRepresentation(8801, "&equiv;"),
		new SgmlUnicodeRepresentation(8804, "&le;"),
		new SgmlUnicodeRepresentation(8805, "&ge;"),
		new SgmlUnicodeRepresentation(8834, "&sub;"),
		new SgmlUnicodeRepresentation(8835, "&sup;"),
		new SgmlUnicodeRepresentation(8836, "&nsub;"),
		new SgmlUnicodeRepresentation(8838, "&sube;"),
		new SgmlUnicodeRepresentation(8839, "&supe;"),
		new SgmlUnicodeRepresentation(8853, "&oplus;"),
		new SgmlUnicodeRepresentation(8855, "&otimes;"),
		new SgmlUnicodeRepresentation(8869, "&perp;"),
		new SgmlUnicodeRepresentation(8901, "&sdot;"),
		new SgmlUnicodeRepresentation(8968, "&lceil;"),
		new SgmlUnicodeRepresentation(8969, "&rceil;"),
		new SgmlUnicodeRepresentation(8970, "&lfloor;"),
		new SgmlUnicodeRepresentation(8971, "&rfloor;"),
		new SgmlUnicodeRepresentation(9001, "&lang;"),
		new SgmlUnicodeRepresentation(9002, "&rang;"),
		new SgmlUnicodeRepresentation(9674, "&loz;"),
		new SgmlUnicodeRepresentation(9824, "&spades;"),
		new SgmlUnicodeRepresentation(9827, "&clubs;"),
		new SgmlUnicodeRepresentation(9829, "&hearts;"),
		new SgmlUnicodeRepresentation(9830, "&diams;"),
	};

	/**
	 * Escapes typical SGML characters to be safe for HTML pages.
	 * 
	 * @param text the string to convert
	 * @return the converted string
	 * @see #SGML_CONVERSIONS
	 */
	public static String sgmlify(final String text) {
		String str = text;
		for (int i = 0; i < SGML_CONVERSIONS.length; i++) {
			final String[] conversion = SGML_CONVERSIONS[i];
			str = str.replaceAll(conversion[0], conversion[1]);
		}
		return str;
	}

	/**
	 * Replaces typical SGML representations of characters by the characters themselves.
	 * 
	 * @param text the string to convert
	 * @return the converted string
	 * @see #SGML_CONVERSIONS
	 */
	public static String deSgmlify(final String text) {
		String str = text;
		for (int i = SGML_UNICODE_REPRESENTATIONS.length - 1; i >= 0; i--) {
			final SgmlUnicodeRepresentation conversion = SGML_UNICODE_REPRESENTATIONS[i];
			str = str.replaceAll("&#" + Integer.toString(conversion.unicode, 10) + ";", conversion.name);
			str = str.replaceAll("&#x" + Integer.toString(conversion.unicode, 16) + ";", conversion.name);
			str = str.replaceAll("&#x0" + Integer.toString(conversion.unicode, 16) + ";", conversion.name);
			str = str.replaceAll("&#x00" + Integer.toString(conversion.unicode, 16) + ";", conversion.name);
			str = str.replaceAll("&#x000" + Integer.toString(conversion.unicode, 16) + ";", conversion.name);
		}
		for (int i = SGML_CONVERSIONS.length - 1; i >= 0; i--) {
			final String[] conversion = SGML_CONVERSIONS[i];
			str = str.replaceAll(conversion[1], conversion[0]);
		}
		for (int i = EXTENDED_SGML_CONVERSIONS.length - 1; i >= 0; i--) {
			final String[] conversion = EXTENDED_SGML_CONVERSIONS[i];
			str = str.replaceAll(conversion[1], conversion[0]);
		}
		return str;
	}

	/**
	 * Takes an URL string and escapes characters not allowed in an URL.
	 * @param url the URL string
	 * @return the converted URL string
	 */
	public static String urlify(final String url) {
		return url.replace(" ", "%20");
	}

	/**
	 * Takes an URL string and escapes characters not allowed in an URL.
	 * @param url the URL string
	 * @return the converted URL string
	 */
	public static String deUrlify(final String url) {
		return url.replace("%20", " ");
	}

	/**
	 * Creates depending on <em>condition</em>, either a link or a text without
	 * a link.
	 * 
	 * @param text
	 *            the text to be shown
	 * @param href
	 *            where the link shows to when the text is clicked
	 * @param isLink
	 *            only if <code>true</code>, the text will be a link, otherwise it's plain
	 *            text
	 * @return the HTML snippet for a link or a plain text
	 */
	public static String linkOrShow(final String text, final String href, final boolean isLink) {
		final StringBuffer s = new StringBuffer();
		if (isLink) {
			s.append("<a" + TextFormatter.htmlAttribute(TextFormatter.HREF, sgmlify(href)));
			s.append(">");
			s.append(text);
			s.append("</a>");
		} else {
			s.append(text);
		}
		return s.toString();
	}

	/**
	 * @param condition
	 *            only if <code>true</code>, a selected attribute is returned
	 * @return either a selected attribute or an empty string
	 */
	public static String selectedIf(final boolean condition) {
		return (condition ? TextFormatter.htmlAttribute(TextFormatter.SELECTED, TextFormatter.SELECTED) : "");
	}

	/**
	 * @param condition
	 *            only if <code>true</code>, a checked attribute is returned
	 * @return either a checked attribute or an empty string
	 */
	public static String checkedIf(final boolean condition) {
		return (condition ? TextFormatter.htmlAttribute(TextFormatter.CHECKED, TextFormatter.CHECKED) : "");
	}

	/**
	 * Lists all the keys and values in the map. Useful to list all parameters
	 * of a page.
	 * 
	 * @param map
	 *            it is assumed that keys are of type String, and values of type
	 *            String[]
	 * @return HTML code that presents all keys and values in human-readable
	 *         form, in this case a HTML table
	 */
	public static String listParameters(final Map<String, String[]> map) {
		final StringBuffer s = new StringBuffer();
		final int ind = 0;
		s.append(line(ind + 0, "<div" + TextFormatter.htmlAttribute(TextFormatter.CLASS, "scrollable") + ">"));
		s.append(line(ind + 1, "<table>"));
		s.append(line(ind + 2, "<caption>Parameters</caption>"));
		s.append(line(ind + 2, "<thead>"));
		s.append(line(ind + 3, "<tr>"));
		s.append(line(ind + 4, "<th>Parameter</th><th>Values</th>"));
		s.append(line(ind + 3, "</tr>"));
		s.append(line(ind + 2, "</thead>"));
		s.append(line(ind + 2, "<tbody>"));
		final Iterator<String> it = map.keySet().iterator();
		while (it.hasNext()) {
			final String key = (String) it.next();
			final String[] value = (String[]) map.get(key);
			String valueStr = "";
			for (int i = 0; i < value.length; i++) {
				valueStr += value[i];
				if (i < value.length - 1) {
					valueStr += "|";
				}
			}
			s.append(line(ind + 3, "<tr>"));
			s.append(line(ind + 4, "<td>" + TextFormatter.tableCellContent(TextDocumentType.XHTML10_STRICT, key) + "</td>"));
			s.append(line(ind + 4, "<td>" + TextFormatter.tableCellContent(TextDocumentType.XHTML10_STRICT, valueStr) + "</td>"));
			s.append(line(ind + 3, "</tr>"));
		}
		s.append(line(ind + 2, "</tbody>"));
		s.append(line(ind + 1, "</table>"));
		s.append(line(ind + 0, "</div>"));
		return s.toString();
	}

	/**  
	 * If <code>true</code>, {@link #toHtmlCheckList(String, String[], int[])} 
	 * will use a real multi selection.
	 * Otherwise it will use several check boxes.
	 */
	public static final boolean IS_USING_MULTI_SELECTION = false;

	/**
	 * Creates a HTML snippet containing a multi-selection (represented as real
	 * multi-selection or lots of checkboxes), filled with the given values. To
	 * retrieve the parameters of that multi-selection, use
	 * {@link #readCheckListParams(HttpServletRequest, String, int)}.
	 * 
	 * @param id
	 *            the id (or prefix) of the HTML elements
	 * @param texts
	 *            those items will be filled into the multi-selection
	 * @param checkedIDs
	 *            those indexes are pre-selected
	 * @return HTML snippet
	 * @see #IS_USING_MULTI_SELECTION
	 */
	public static String toHtmlCheckList(final String id, final String[] texts, final int[] checkedIDs) {
		final StringBuffer s = new StringBuffer();
		if (null != texts) {
			final int max = texts.length;
			final int ind = 0;
			if (IS_USING_MULTI_SELECTION) {
				s.append(line(ind + 0, "<select"
						+ TextFormatter.htmlAttribute(TextFormatter.ID, id)
						+ TextFormatter.htmlAttribute(TextFormatter.NAME, id)
						+ TextFormatter.htmlAttribute(TextFormatter.SIZE, Integer.toString(max))
						+ TextFormatter.htmlAttribute(TextFormatter.MULTIPLE, TextFormatter.MULTIPLE)
						+ ">"));
				for (int idx = 0; idx < max; idx++) {
					s.append(line(ind + 1, "<option"
							+ TextFormatter.htmlAttribute(TextFormatter.VALUE, Integer.toString(idx))
							+ selectedIf(CollectionUtil.contains(checkedIDs, idx))
							+ ">"
							+ texts[idx]
							+ "</option>"));
				}
				s.append(line(ind + 0, "</select>"));
			} else {
				s.append(line(ind + 0, "<div"
						+ TextFormatter.htmlAttribute(TextFormatter.CLASS, id)
						+ ">"));
				s.append(line(ind + 1, "<div"
						+ TextFormatter.htmlAttribute(TextFormatter.CLASS, "multicolumn")
						+ ">"));
				for (int idx = 0; idx < max; idx++) {
					final String name = id + idx;
					s.append(line(ind + 2, "<input"
							+ TextFormatter.htmlAttribute(TextFormatter.TYPE, TextFormatter.CHECKBOX)
							+ TextFormatter.htmlAttribute(TextFormatter.ID, name)
							+ TextFormatter.htmlAttribute(TextFormatter.NAME, name)
							+ TextFormatter.htmlAttribute(TextFormatter.VALUE, Integer.toString(idx))
							+ checkedIf(CollectionUtil.contains(checkedIDs, idx))
							+ "/>&nbsp;<label"
							+ TextFormatter.htmlAttribute(TextFormatter.FOR, name)
							+ ">" 
							+ texts[idx] 
							+ "</label>"));
					s.append(line(ind + 2, "<br/>"));
				}
				s.append(line(ind + 1, "</div>"));
				s.append(line(ind + 0, "</div>"));
			}
		}
		return s.toString();
	}

	/**
	 * Parses param for an int value. If this fails, a human-readable error
	 * message is appended to errorMsgs.
	 * 
	 * @param errorMsgs
	 *            takes human-readable error messages if a problem occurs
	 * @param param
	 *            this string is parsed for an int value
	 * @return the parsed int value, or 0
	 */
	public static int parseInt(final StringBuffer errorMsgs, final String param) {
		int ret = 0;
		try {
			ret = Integer.parseInt(param);
		} catch (final NumberFormatException ex) {
			if (null != errorMsgs) {
				errorMsgs.append("Expected a number ");
				errorMsgs.append(ex.getLocalizedMessage());
				errorMsgs.append("\n");
			}
		} catch (final NullPointerException ex) {
			if (null != errorMsgs) {
				errorMsgs.append(ex.getLocalizedMessage());
				errorMsgs.append("\n");
			}
		}
		return ret;
	}

	/**
	 * Reads the parameters of a multi-selection created by
	 * {@link #toHtmlCheckList(String, String[], int[])}.
	 * 
	 * @param request
	 *            the servlet request containing the parameters
	 * @param id
	 *            the parameter name (or prefix)
	 * @param maxNumber
	 *            the number of (selected and unselected) elements in the
	 *            multi-selection
	 * @return a list of indexes of the selected elements
	 */
	public static int[] readCheckListParams(final HttpServletRequest request, final String id, final int maxNumber) {
		int[] selectedIndexes = null;
		String[] parameterValues = request.getParameterValues(id);
		if (null != parameterValues) {
			// Read parameter created by multi-selection.
			selectedIndexes = new int[parameterValues.length];
			for (int i = 0; i < parameterValues.length; i++) {
				selectedIndexes[i] = Integer.parseInt(parameterValues[i]);
			}
		} else {
			// Read parameter created by multiple checkboxes.
			int numberOfIndexes = 0;
			for (int i = 0; i < maxNumber; i++) {
				final String param = request.getParameter(id + i);
				if (null != param) {
					numberOfIndexes++;
				}
			}
			selectedIndexes = new int[numberOfIndexes];
			int idx = 0;
			for (int i = 0; i < maxNumber; i++) {
				final String paramValue = request.getParameter(id + i);
				if (null != paramValue) {
					selectedIndexes[idx] = i;
					idx++;
				}
			}
		}
		return selectedIndexes;
	}

	/**
	 * Adds a parameter to the given URL.
	 * 
	 * @param url
	 *            the URL where to add the parameter to
	 * @param key
	 *            the key of the parameter to add
	 * @param value
	 *            the value of the parameter to add
	 * @return the new URL with the key/value pair appended as parameter
	 */
	public static String addParameter(final String url, final String key, final String value) {
		final StringBuffer ret = new StringBuffer(url);
		if (url.indexOf("?") < 0) {
			ret.append("?");
		} else {
			ret.append("&");
		}
		ret.append(key);
		ret.append("=");
		ret.append(value);
		return ret.toString();
	}

	/**
	 * Creates an error message telling that a value was not expected for a
	 * parameter.
	 * 
	 * @param key
	 *            the parameter key
	 * @param unknownValue
	 *            the unexpected value
	 * @param knownValues
	 *            the expected values
	 * @return the error message
	 */
	public static String buildErrorMsgUnexpectedParameterValue(
			final String key, 
			final String unknownValue, 
			final String[] knownValues) {
		final StringBuffer errorMsgs = new StringBuffer();
		errorMsgs.append("Unexpected value (");
		errorMsgs.append(unknownValue);
		errorMsgs.append(") for parameter ");
		errorMsgs.append(key);
		errorMsgs.append(" - I only accept one of ");
		for (int i = 0; i < knownValues.length; i++) {
			errorMsgs.append(knownValues[i]);
			if (i < knownValues.length - 1) {
				errorMsgs.append(", ");
			}
		}
		errorMsgs.append(".");
		errorMsgs.append("\n");
		return errorMsgs.toString();
	}

	/**
	 * Finds for a given key the index in an array of expected values.
	 * 
	 * @param request
	 *            the request containing the parameter
	 * @param key
	 *            the parameter key
	 * @param expectedValues
	 *            the list of expected values for the key
	 * @param errorMsgs
	 *            will be unchanged if found - and an error message will be
	 *            appended if value not found
	 * @return the index of the value, or -1 if not found
	 */
	public static int readRadioParameter(
			final HttpServletRequest request, 
			final String key, 
			final String[] expectedValues, 
			final StringBuffer errorMsgs) {
		final String param = request.getParameter(key);
		int idx = -1;
		if (null != param) {
			for (int i = 0; (i < expectedValues.length) && (idx < 0); i++) {
				if (param.equalsIgnoreCase(expectedValues[i])) {
					idx = i;
				}
			}
		}
		if (idx < 0) {
			errorMsgs.append(buildErrorMsgUnexpectedParameterValue(key, param, expectedValues));
		}
		return idx;
	}

	public static String createLabel(final String param, final Messages messages, final String key) {
		final StringBuffer sb = new StringBuffer();
		sb.append("<label " 
				+ TextFormatter.htmlAttribute(TextFormatter.FOR, param)
				+ TextFormatter.htmlAttribute(TextFormatter.TITLE, messages.getShortDescription(key)) 
				+ ">");
		sb.append(messages.getText(key));
		sb.append("</label>" + "\n");
		return sb.toString();
	}
	
	private static String line(final int indent, final String text) {
		return TextFormatter.indent(TextFormatter.DEFAULT_INDENT, indent, text) + TextFormatter.NL;
	}
}
