package de.ped.tools;

import java.util.Arrays;

/**
 * Defines I18N'ed complex strings containing non-I18N'ed fill-ins.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class I18NStringWithFillIns implements Marshallable, Cloneable {
	public static final I18NStringWithFillIns EMPTY = new I18NStringWithFillIns(null, "", null);
	protected static final char ID_I18NSTRING = 'N';
	protected static final char ID_STRING     = 'S';
	protected static final char ID_BOOLEAN    = 'B';
	protected static final char ID_CHAR       = 'C';
	protected static final char ID_INTEGER    = 'I';
	protected static final char ID_LONG       = 'L';
	protected static final char ID_FLOAT      = 'F';
	protected static final char ID_DOUBLE     = 'D';
	protected static final char ID_NULL       = '0';

	/** Creates an empty string. */
	public I18NStringWithFillIns() {
		// Nothing to do
	}

	/**
	 * Creates a string object from the parameters.
	 * 
	 * @param key
	 *            the message key where to search for an i18n'ed value string
	 * @param alternative
	 *            the alternative text to use if there was no value found
	 * @param fillIns
	 *            the (non i18n'ed) string snippets that will be included
	 */
	public I18NStringWithFillIns(final String key, final String alternative, final Object[] fillIns) {
		this.key = key;
		this.alternative = alternative;
		this.fillIns = fillIns;
	}

	/**
	 * Reads a string object from the marshaller.
	 * 
	 * @param marshaller
	 *            the marshaller to read the object from
	 */
	public I18NStringWithFillIns(Marshaller marshaller) {
		marshaller.readMarshallable(this);
	}

	/**
	 * Does a deep serialization of this object. The fill-ins that are not basic
	 * types or themselves {@link I18NStringWithFillIns} are stored as simple
	 * strings using {@link Object#toString()}.
	 */
	@Override
	public void marshal(Marshaller marshaller) {
		marshaller.writeString(key);
		marshaller.writeString(alternative);
		// write fillIns
		if (null == fillIns) {
			marshaller.writeInt(Marshaller.ARRAY_NULL_INDICATOR);
		} else {
			marshaller.writeInt(fillIns.length);
			for (int i = 0; i < fillIns.length; i++) {
				Object fillIn = fillIns[i];
				if (null == fillIn) {
					marshaller.writeChar(ID_NULL);
				} else if (fillIn instanceof I18NStringWithFillIns) {
					marshaller.writeChar(ID_I18NSTRING);
					marshaller.writeMarshallable((I18NStringWithFillIns) fillIn);
				} else if (fillIn instanceof Boolean) {
					marshaller.writeChar(ID_BOOLEAN);
					marshaller.writeBoolean(((Boolean) fillIn).booleanValue());
				} else if (fillIn instanceof Character) {
					marshaller.writeChar(ID_CHAR);
					marshaller.writeChar(((Character) fillIn).charValue());
				} else if (fillIn instanceof Integer) {
					marshaller.writeChar(ID_INTEGER);
					marshaller.writeInt(((Integer) fillIn).intValue());
				} else if (fillIn instanceof Long) {
					marshaller.writeChar(ID_LONG);
					marshaller.writeLong(((Long) fillIn).longValue());
				} else if (fillIn instanceof Float) {
					marshaller.writeChar(ID_FLOAT);
					marshaller.writeFloat(((Float) fillIn).floatValue());
				} else if (fillIn instanceof Double) {
					marshaller.writeChar(ID_DOUBLE);
					marshaller.writeDouble(((Double) fillIn).doubleValue());
				} else {
					marshaller.writeChar(ID_STRING);
					marshaller.writeString(fillIn.toString()); 
				}
			}
		}
	}

	@Override
	public Marshallable demarshal(Marshaller marshaller) {
		key = marshaller.readString();
		alternative = marshaller.readString();
		// read fillIns
		final int fillInsLength = marshaller.readInt();
		if (Marshaller.ARRAY_NULL_INDICATOR == fillInsLength) {
			fillIns = null;
		} else {
			fillIns = new Object[fillInsLength];
			for (int i = 0; i < fillIns.length; i++) {
				final char c = marshaller.readChar();
				switch(c) {
				case ID_NULL:
					fillIns[i] = null;
					break;
				case ID_I18NSTRING:
					fillIns[i] = new I18NStringWithFillIns(marshaller); 
					break;
				case ID_BOOLEAN:
					fillIns[i] = marshaller.readBoolean();
					break;
				case ID_CHAR:
					fillIns[i] = marshaller.readChar();
					break;
				case ID_INTEGER:
					fillIns[i] = marshaller.readInt();
					break;
				case ID_LONG:
					fillIns[i] = marshaller.readLong();
					break;
				case ID_FLOAT:
					fillIns[i] = marshaller.readFloat();
					break;
				case ID_DOUBLE:
					fillIns[i] = marshaller.readDouble();
					break;
				case ID_STRING:
				default:
					fillIns[i] = marshaller.readString(); 
				}
			}
		}
		return this;
	}

	/**
	 * @return a shallow copy of this object
	 */
	@Override
	public I18NStringWithFillIns clone() {
		I18NStringWithFillIns ret = null;
		try {
			ret = (I18NStringWithFillIns) super.clone();
		} catch (final CloneNotSupportedException e) {
			// Nothing to do
		}
		return ret;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((alternative == null) ? 0 : alternative.hashCode());
		result = prime * result + Arrays.hashCode(fillIns);
		result = prime * result + ((key == null) ? 0 : key.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		I18NStringWithFillIns other = (I18NStringWithFillIns) obj;
		if (key == null) {
			if (other.key != null)
				return false;
		} else if (!key.equals(other.key))
			return false;
		if (alternative == null) {
			if (other.alternative != null)
				return false;
		} else if (!alternative.equals(other.alternative))
			return false;
		if (!Arrays.equals(fillIns, other.fillIns))
			return false;
		return true;
	}

	@Override
	public String toString() {
		final StringBuffer buf = new StringBuffer();
		buf.append(key);
		buf.append('/');
		buf.append(alternative);
		buf.append('(');
		if (null == fillIns) {
			buf.append(fillIns);
		} else {
			for(int i = 0; i < fillIns.length; i++) {
				if (0 < i) {
					buf.append(',');
				}
				buf.append(fillIns[i]);
			}
		}
		buf.append(')');
		return buf.toString();
	}

	/**
	 * @return <code>true</code> if both key and alternative are <code>null</code> or
	 *         empty strings. No effort is taken to check if the value behind
	 *         the key is <code>null</code> or empty.
	 */
	public boolean isEmpty() {
		return CollectionUtil.isNullOrEmpty(key) && CollectionUtil.isNullOrEmpty(alternative);
	}

	public String getKey() {
		return key;
	}

	public void setKey(String key) {
		this.key = key;
	}

	public String getAlternative() {
		return alternative;
	}

	public void setAlternative(String alternative) {
		this.alternative = alternative;
	}

	public Object[] getFillIns() {
		return fillIns;
	}

	public void setFillIns(Object[] fillIns) {
		this.fillIns = fillIns;
	}

	private String key;
	private String alternative;
	private Object[] fillIns;
}