package de.ped.tools;

/**
 * Supertype of any object that can be serialized/deserialized via {@link Marshaller}.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public interface Marshallable {

	/**
	 * Defines how to marshal itself into the marshaller.
	 * Normally you don't call this method yourself,
	 * but use {@link Marshaller#writeMarshallable(Marshallable)}.
	 * <p>
	 * For objects with an inheritance hierarchy 
	 * and different data to demarshal depending on the subtype,
	 * the easiest way is to code <em>this</em> to
	 * first marshal the information which subtype <em>this</em> is, 
	 * and then marshal the data of <em>this</em>.
	 * For objects without that complication, you can omit the subtype information,
	 * as the reader knows what data to expect.
	 * <p>
	 * Beware: Is called in a <code>synchronized</code> context, 
	 * so don't use time or resource consuming operations.
	 * 
	 * @param marshaller the marshaller to write into
	 */
	void marshal(final Marshaller marshaller);

	/**
	 * Defines how to demarshal a 
	 * Marshallalable object (e.g., itself) from <em>m</em>.
	 * Normally you don't call this method yourself,
	 * but use {@link Marshaller#readMarshallable(Marshallable)}.
	 * <p>
	 * For objects with an inheritance hierarchy 
	 * and different data to demarshal depending on the subtype,
	 * the easiest way is to code <em>this</em> to
	 * first demarshal some type information, create an instance of that type,
	 * demarshal the data into it and return that instance.
	 * For objects without that complication, it may be easier to
	 * just demarshal the data into <em>this</em> and return <em>this</em>.
	 * <p>
	 * Beware: Is called in a <code>synchronized</code> context, 
	 * so don't use time or resource consuming operations.
	 * 
	 * @param marshaller the marshaller to read from
	 * @return the object that was read - normally <em>this</em>
	 */
	Marshallable demarshal(final Marshaller marshaller);
}
