package de.ped.tools;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;

/**
 * Marshals and demarshals all the basic types and {@link Marshallable} objects; realized as thread safe queue.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Marshaller implements Cloneable {

	/**
	 * Thrown when the reader tries to read beyond the end of the stream.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class UnexpectedEOFException extends ArrayIndexOutOfBoundsException {
		private static final long serialVersionUID = 1L;
		
		public UnexpectedEOFException() {
			super("Reading past end of buffer");
		}
	}

	/**
	 * Thrown to tell that the {@link Marshallable#demarshal} method should not
	 * be used.
	 * 
	 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ConstantMarshallableDemarshallingException extends IllegalStateException {
		private static final long serialVersionUID = 1L;

		public ConstantMarshallableDemarshallingException() {
			super("This is a const or abstract object - please use the appropriate constructor or create method to unmarshal this object.");
		}
	}

	/** The default chunk size to grow the buffer. */
	public static final int CHUNK_SIZE = 1024;

	/** For arrays of variable length, <code>null</code> arrays are coded with this value as length. */
	public static final int ARRAY_NULL_INDICATOR = -1;

	/** As boolean values are marshalled as bytes, that's the representation for <code>true</code>. */
	public static final byte TRUE_INDICATOR = (byte)0xff;

	/** As boolean values are marshalled as bytes, that's the representation for <code>false</code>. */
	public static final byte FALSE_INDICATOR = (byte)0x00;

	/** Objects like {@link Marshallable} can be <code>null</code>, so here's the representation for a non-<code>null</code> object. */
	public static final byte OBJECT_NOT_NULL_INDICATOR = FALSE_INDICATOR;

	/** Objects like {@link Marshallable} can be <code>null</code>, so here's the representation for a <code>null</code> object. */
	public static final byte OBJECT_NULL_INDICATOR = TRUE_INDICATOR;

	/**
	 * @return a {@link ConstantMarshallableDemarshallingException} ready to be thrown.
	 */
	public static ConstantMarshallableDemarshallingException createConstantMarshallableDemarshallingException() {
		return new ConstantMarshallableDemarshallingException();
	}

	public static Marshaller loadFromFile(final File file) throws IOException {
		Marshaller ret = null;
		final FileInputStream fw = new FileInputStream(file);
		final BufferedInputStream bw = new BufferedInputStream(fw);
		final byte[] data = new byte[(int) file.length()];
		bw.read(data);
		final Marshaller marshaller = new Marshaller(data);
		bw.close();
		fw.close();
		ret = marshaller;
		return ret;
	}

	public static byte[] writeToFile(final Marshallable object, final File file) throws IOException {
		final FileOutputStream fw = new FileOutputStream(file);
		final BufferedOutputStream bw = new BufferedOutputStream(fw);
		final Marshaller marshaller = new Marshaller();
		marshaller.writeMarshallable(object);
		final byte[] data = marshaller.pack();
		bw.write(data);
		bw.close();
		fw.close();
		return data;
	}

	private byte[] buffer;

	private int writePos;
	private int readPos;

	/** Creates an empty queue. */
	public Marshaller() {
		init();
	}

	/**
	 * Creates an already filled queue.
	 * 
	 * @param arr
	 *            the queue content
	 */
	public Marshaller(final byte[] arr) {
		init(arr);
	}

	/** Reinitializes to an empty queue. */
	public synchronized void init() {
		buffer = new byte[CHUNK_SIZE];
		writePos = 0;
		readPos = 0;
	}

	/** 
	 * Reinitializes to an already filled queue. 
	 * 
	 * @param arr
	 *            the queue content
	 */
	public synchronized void init(final byte[] arr) {
		buffer = arr;
		writePos = (null == arr ? 0 : arr.length);
		readPos = 0;
	}

	public synchronized int size() {
		return writePos - readPos;
	}

	protected void grow() {
		grow(CHUNK_SIZE);
	}

	/**
	 * First, tries to keep the buffer size, but shift the buffer by eliminating
	 * all the already read bytes. If that fails, shifts the buffer, but also
	 * adds <em>amount</em> bytes.
	 * 
	 * @param amount the number of bytes to add to this buffer, if necessary
	 */
	protected void grow(final int amount) {
		final int size = writePos - readPos;
		if (size + amount < buffer.length) {
			for (int i = 0; i < size; i++) {
				buffer[i] = buffer[i + readPos];
			}
			writePos -= readPos;
			readPos = 0;
		} else {
			final byte[] newBuffer = new byte[buffer.length + amount];
			for (int i = 0; i < size; i++) {
				newBuffer[i] = buffer[i + readPos];
			}
			buffer = newBuffer;
			writePos -= readPos;
			readPos = 0;
		}
	}

	protected void growIfNeeded(final int amount) {
		final int diff = writePos + amount - buffer.length;
		if (0 < diff) {
			grow(Math.max(diff, CHUNK_SIZE));
		}
	}

	/** 
	 * Returns the packed content of the buffer, and clears <em>this</em>' buffer.
	 * 
	 * @return a copy of the marshalled data
	 */
	public synchronized byte[] pack() {
		byte[] ret = new byte[writePos - readPos];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = buffer[i + readPos];
		}
		writePos = 0;
		readPos = 0;
		return ret;
	}

	/** Shrinks the buffer to only keep all the written, but non-read bytes. */
	public synchronized void trimToSize() {
		byte[] buf = pack();
		init(buf);
	}

	public synchronized void writeByte(final byte value) {
		growIfNeeded(1);
		buffer[writePos++] = value;
	}

	/**
	 * Reads a single <code>byte</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized byte readByte() {
		final int restorePos = readPos;
		try {
			if (writePos <= readPos) {
				throw new UnexpectedEOFException();
			}
			return buffer[readPos++];
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public static byte toByte(final boolean value) {
		return (value ? TRUE_INDICATOR : FALSE_INDICATOR);
	}

	public static boolean toBoolean(final byte value) {
		return (FALSE_INDICATOR != value);
	}

	public synchronized void writeBoolean(final boolean value) {
		writeByte(toByte(value));
	}

	/**
	 * Reads a <code>boolean</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized boolean readBoolean() {
		return (toBoolean(readByte()));
	}

	public synchronized void writeShort(final short value) {
		growIfNeeded(2);
		buffer[writePos++] = (byte)((value >>  8) & 0xff);
		buffer[writePos++] = (byte)((value >>  0) & 0xff);
	}

	/**
	 * Reads a <code>short</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized int readShort() {
		final int restorePos = readPos;
		try {
			return
				((readByte() & 0xff) <<  8) | 
				((readByte() & 0xff) <<  0); 
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeInt(final int value) {
		growIfNeeded(4);
		buffer[writePos++] = (byte)((value >> 24) & 0xff);
		buffer[writePos++] = (byte)((value >> 16) & 0xff);
		buffer[writePos++] = (byte)((value >>  8) & 0xff);
		buffer[writePos++] = (byte)((value >>  0) & 0xff);
	}

	/**
	 * Reads an <code>int</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized int readInt() {
		final int restorePos = readPos;
		try {
			return
				((readByte() & 0xff) << 24) | 
				((readByte() & 0xff) << 16) | 
				((readByte() & 0xff) <<  8) | 
				((readByte() & 0xff) <<  0); 
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeLong(final long value) {
		growIfNeeded(8);
		buffer[writePos++] = (byte)((value >> 56) & 0xff);
		buffer[writePos++] = (byte)((value >> 48) & 0xff);
		buffer[writePos++] = (byte)((value >> 40) & 0xff);
		buffer[writePos++] = (byte)((value >> 32) & 0xff);
		buffer[writePos++] = (byte)((value >> 24) & 0xff);
		buffer[writePos++] = (byte)((value >> 16) & 0xff);
		buffer[writePos++] = (byte)((value >>  8) & 0xff);
		buffer[writePos++] = (byte)((value >>  0) & 0xff);
	}

	/**
	 * Reads a <code>long</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized long readLong() {
		final int restorePos = readPos;
		try {
			return
				(((long)(readByte() & 0xff)) << 56) | 
				(((long)(readByte() & 0xff)) << 48) | 
				(((long)(readByte() & 0xff)) << 40) | 
				(((long)(readByte() & 0xff)) << 32) | 
				(((long)(readByte() & 0xff)) << 24) | 
				(((long)(readByte() & 0xff)) << 16) | 
				(((long)(readByte() & 0xff)) <<  8) | 
				(((long)(readByte() & 0xff)) <<  0); 
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeFloat(final float value) {
		writeInt(Float.floatToRawIntBits(value));
	}

	/**
	 * Reads a <code>float</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized float readFloat() {
		return Float.intBitsToFloat(readInt());
	}

	public synchronized void writeDouble(final double value) {
		writeLong(Double.doubleToRawLongBits(value));
	}

	/**
	 * Reads a <code>double</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized double readDouble() {
		return Double.longBitsToDouble(readLong());
	}

	public synchronized void writeChar(final char value) {
		writeInt(value);
	}

	/**
	 * Reads a <code>char</code>.
	 * 
	 * @return the value read from the queue
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized char readChar() {
		return (char)readInt();
	}

	/**
	 * Writes the first <em>numOfEntries</em> elements of <em>arr</em> to the
	 * buffer.
	 * 
	 * @param arr
	 *            from this array entries are written to the queue
	 * @param numOfEntries
	 *            so many entries are written to the queue
	 */
	public synchronized void writeFixedLengthBytesArray(final byte[] arr, final int numOfEntries) {
		growIfNeeded(numOfEntries);
		for (int i = 0; i < numOfEntries; i++) {
			buffer[writePos++] = arr[i];
		}
	}

	public synchronized void writeFixedLengthBytesArray(byte[] arr) {
		writeFixedLengthBytesArray(arr, arr.length);
	}

	/**
	 * Reads a byte array of fixed length (so the length information was not
	 * stored in the stream).
	 * 
	 * @param length
	 *            The number of bytes to read
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized byte[] readFixedLengthBytesArray(final int length) {
		final int restorePos = readPos;
		try {
			final byte[] arr = new byte[length];
			for (int i = 0; i < length; i++) {
				arr[i] = readByte();
			}
			return arr;
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeFixedLengthBooleanArray(final boolean[] arr) {
		final byte[] barr = new byte[arr.length];
		for (int i = 0; i < arr.length; i++) {
			barr[i] = toByte(arr[i]);
		}
		writeFixedLengthBytesArray(barr, barr.length);
	}

	/**
	 * Reads a boolean array of fixed length (so the length information was not
	 * stored in the stream).
	 * 
	 * @param length
	 *            the number of booleans to read
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized boolean[] readFixedLengthBooleanArray(final int length) {
		final int restorePos = readPos;
		try {
			final boolean[] arr = new boolean[length];
			for (int i = 0; i < length; i++) {
				arr[i] = toBoolean(readByte());
			}
			return arr;
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeFixedLengthIntArray(final int[] arr) {
		growIfNeeded(4 * arr.length);
		for (int value : arr) {
			writeInt(value);
		}
	}

	/**
	 * Reads an int array of fixed length (so the length information was not
	 * stored in the stream).
	 * 
	 * @param length
	 *            the number of ints to read
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized int[] readFixedLengthIntArray(final int length) {
		final int restorePos = readPos;
		try {
			final int[] arr = new int[length];
			for (int i = 0; i < length; i++) {
				arr[i] = readInt();
			}
			return arr;
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeVariableLengthByteArray(final byte[] arr) {
		if (null == arr) {
			writeInt(ARRAY_NULL_INDICATOR);
		} else {
			growIfNeeded(arr.length + 4);
			writeInt(arr.length);
			writeFixedLengthBytesArray(arr);
		}
	}

	/**
	 * Reads a byte array of variable length (so the length information was
	 * stored in the stream).
	 * 
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized byte[] readVariableLengthBytesArray() {
		final int restorePos = readPos;
		try {
			final int length = readInt();
			if (ARRAY_NULL_INDICATOR == length) {
				return null;
			}
			return readFixedLengthBytesArray(length);
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public void writeVariableLengthBooleanArray(final boolean[] arr) {
		if (null == arr) {
			writeInt(ARRAY_NULL_INDICATOR);
		} else {
			growIfNeeded(arr.length + 4);
			writeInt(arr.length);
			writeFixedLengthBooleanArray(arr);
		}
	}

	public boolean[] readVariableLengthBooleanArray() {
		final int restorePos = readPos;
		try {
			final int length = readInt();
			if (ARRAY_NULL_INDICATOR == length) {
				return null;
			}
			return readFixedLengthBooleanArray(length);
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public void writeVariableLengthIntArray(final int[] arr) {
		if (null == arr) {
			writeInt(ARRAY_NULL_INDICATOR);
		} else {
			growIfNeeded(arr.length * 4 + 4);
			writeInt(arr.length);
			writeFixedLengthIntArray(arr);
		}
	}

	public int[] readVariableLengthIntArray() {
		final int restorePos = readPos;
		try {
			final int length = readInt();
			if (ARRAY_NULL_INDICATOR == length) {
				return null;
			}
			return readFixedLengthIntArray(length);
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeString(final String value) {
		final byte[] arr = (null == value ? null : value.getBytes());
		writeVariableLengthByteArray(arr);
	}

	/**
	 * Reads a String.
	 * 
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized String readString() {
		final int restorePos = readPos;
		try {
			final byte[] arr = readVariableLengthBytesArray();
			if (null == arr) {
				return null;
			}
			return new String(arr);
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeStringArray(final String[] arr) {
		if (null == arr) {
			writeInt(ARRAY_NULL_INDICATOR);
		} else {
			writeInt(arr.length);
			for (int i = 0; i < arr.length; i++) {
				writeString(arr[i]);
			}
		}
	}

	/**
	 * Reads a String array.
	 * 
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized String[] readStringArray() {
		final int restorePos = readPos;
		try {
			final int arrLength = readInt();
			if (Marshaller.ARRAY_NULL_INDICATOR == arrLength) {
				return null;
			}
			final String[] arr = new String[arrLength];
			for (int i = 0; i < arrLength; i++) {
				arr[i] = readString();
			}
			return arr;
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	/**
	 * Writes <em>marshallable</em> into <em>this</em>. Calls
	 * {@link Marshallable#marshal(Marshaller)}.
	 * 
	 * @param marshallable
	 *            the object to write
	 * 
	 */
	public synchronized void writeMarshallable(final Marshallable marshallable) {
		marshallable.marshal(this);
	}

	/**
	 * Reads a {@link Marshallable} from <em>this</em>, using
	 * {@link Marshallable#demarshal(Marshaller)} in <em>marshallable</em>.
	 * 
	 * @param marshallable
	 *            the object that reads (most times the object to read)
	 * @return the object read
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized Marshallable readMarshallable(final Marshallable marshallable) {
		final int restorePos = readPos;
		try {
			return marshallable.demarshal(this);
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	/**
	 * Checks if the given marshallable is <code>null</code>, and writes a null indicator or
	 * not-null indicator into the stream.
	 * 
	 * @param marshallable
	 *            the object to stream
	 * @return <code>true</code> if the object was <code>null</code>
	 */
	public synchronized boolean writeNullIndicator(final Object marshallable) {
		boolean ret;
		if (marshallable == null) {
			writeByte(OBJECT_NULL_INDICATOR);
			ret = true;
		} else {
			writeByte(OBJECT_NOT_NULL_INDICATOR);
			ret = false;
		}
		return ret;
	}

	/**
	 * Reads if the next marshallable to read is <code>null</code> and should be skipped.
	 * 
	 * @return <code>true</code> if the object was <code>null</code>
	 */
	public synchronized boolean readNullIndicator() {
		final byte isNull = readByte();
		return (toBoolean(isNull));
	}

	/**
	 * Writes <em>marshallable</em> into <em>this</em>. Calls
	 * {@link Marshallable#marshal(Marshaller)}.
	 * 
	 * @param marshallable
	 *            the object to stream
	 */
	public synchronized void writeNullOrMarshallable(final Marshallable marshallable) {
		final boolean isNull = writeNullIndicator(marshallable);
		if (!isNull) {
			marshallable.marshal(this);
		}
	}

	/**
	 * Reads a {@link Marshallable} from <em>this</em>, using
	 * {@link Marshallable#demarshal(Marshaller)} in <em>marshallable</em>.
	 * 
	 * @param marshallable
	 *            the object that reads (most times the object to read)
	 * @return the object read
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 */
	public synchronized Marshallable readNullOrMarshallable(final Marshallable marshallable) {
		final int restorePos = readPos;
		try {
			final boolean isNull = readNullIndicator();
			if (isNull) {
				return null;
			} else {
				return marshallable.demarshal(this);
			}
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		}
	}

	public synchronized void writeMarshallableArray(final Marshallable[] arr) {
		if (null == arr) {
			writeInt(ARRAY_NULL_INDICATOR);
		} else {
			writeInt(arr.length);
			for (int i = 0; i < arr.length; i++) {
				writeNullOrMarshallable(arr[i]);
			}
		}
	}

	/**
	 * Reads a Marshallable array.
	 * 
	 * @param marshallable a template of the type to create.
	 * @return the extracted content
	 * @throws UnexpectedEOFException
	 *             There was nothing (or not enough) to read. The read pointer
	 *             was restored to where it was before this call.
	 * @deprecated The problem how to create objects of the proper type is yet unsolved. So you need to roll your own.
	 */
	public synchronized Marshallable[] readMarshallableArray(final Marshallable marshallable) {
		final int restorePos = readPos;
		try {
			final int arrLength = readInt();
			if (Marshaller.ARRAY_NULL_INDICATOR == arrLength) {
				return null;
			}
			final Marshallable[] arr = new Marshallable[arrLength];
			for (int i = 0; i < arrLength; i++) {
				// How to create an object of subtype? clone() does not seem to work, even with reflections.
				if (marshallable instanceof Cloneable) {
					Object clone = Object.class.getDeclaredMethod("clone").invoke(marshallable);
					arr[i] = readNullOrMarshallable((Marshallable) clone);
				} else {
					throw new IllegalArgumentException("Marshallable must also be cloneable!");
				}
			}
			return arr;
		} catch (final UnexpectedEOFException ex) {
			readPos = restorePos;
			throw ex;
		} catch (IllegalArgumentException ex) {
			readPos = restorePos;
			throw ex;
		} catch (SecurityException ex) {
			readPos = restorePos;
			throw ex;
		} catch (IllegalAccessException ex) {
			readPos = restorePos;
			throw new IllegalArgumentException("Marshallable must also be cloneable!");
		} catch (InvocationTargetException ex) {
			readPos = restorePos;
			throw new IllegalArgumentException("Marshallable must also be cloneable!");
		} catch (NoSuchMethodException ex) {
			readPos = restorePos;
			throw new IllegalArgumentException("Marshallable must also be cloneable!");
		}
	}

	@Override
	public Marshaller clone() {
		Marshaller other;
		try {
			other = (Marshaller) super.clone();
			other.buffer = Arrays.copyOf(buffer, buffer.length);
		} catch (final CloneNotSupportedException e) {
			throw new IllegalStateException("Unable to clone " + getClass().getSimpleName(), e);
		}
		return other;
	}

	public int getReadPos() {
		return readPos;
	}

	public int getWritePos() {
		return writePos;
	}
}
