package de.ped.tools;

import java.util.Arrays;

/**
 * A simple mathematical matrix class for dense matrices of scalar type double.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MathMatrix implements Cloneable {
	/** Index of the x coordinate */
	public static final int X = 0;

	/** Index of the y coordinate */
	public static final int Y = 1;

	/**
	 * @param size1
	 *            the x-size of the matrx
	 * @param size2
	 *            the y-size of the matrix
	 * @return a zero matrix (a matrix full of zero values) of given size
	 */
	public static MathMatrix zero(final int size1, final int size2) {
		final MathMatrix ret = new MathMatrix(size1, size2);
		return ret;
	}

	/**
	 * @param size
	 *            the size of the matrx (x-size and y-size are the same)
	 * @return an identity matrix (a matrix with 1 values in the diagonale, and
	 *         all other values zero) of given size
	 */
	public static MathMatrix id(final int size) {
		final MathMatrix ret = new MathMatrix(size, size);
		for (int i = 0; i < ret.coord.length; i++) {
			ret.coord[i][i] = 1.0;
		}
		return ret;
	}

	/**
	 * @param size
	 *            the number of dimensions the matrix shall be used for. Should
	 *            be higher or equal to <em>idx1</em> and <em>idx2</em>.
	 * @param idx1
	 *            the index of the first dimension the rotation should be in
	 * @param idx2
	 *            the index of the second dimension the rotation should be in
	 * @param angle
	 *            the rotation angle, in mathematical notation (2*pi is full
	 *            rotation)
	 * @return a rotation matrix that rotates in the plane spanned by idx1 and
	 *         idx2
	 */
	public static MathMatrix rot(final int size, final int idx1, final int idx2, final double angle) {
		if (idx1 == idx2) {
			throw new ArrayIndexOutOfBoundsException();
		}
		final MathMatrix ret = id(size);
		final double cos = Math.cos(angle);
		final double sin = Math.sin(angle);
		ret.coord[idx1][idx1] = cos;
		ret.coord[idx2][idx1] = -sin;
		ret.coord[idx1][idx2] = sin;
		ret.coord[idx2][idx2] = cos;
		return ret;
	}

	/**
	 * Creates a projection matrix.
	 * 
	 * @param size
	 *            the dimension of the vector
	 * @param dimension
	 *            the dimension to zero out
	 * @return the matrix to multiply a coordinate vector with to execute this
	 *         projection
	 */
	public static MathMatrix project(final int size, final int dimension) {
		final MathMatrix ret = MathMatrix.id(size);
		ret.set(dimension, dimension, 0.0);
		return ret;
	}

	/**
	 * Creates a translation matrix for <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous
	 * coordinates</a>.
	 * 
	 * @param translation
	 *            the Cartesian vector to translate with
	 * @return the matrix to multiply a homogeneous coordinate vector with to
	 *         execute this translation
	 */
	public static MathMatrix homTrans(final MathVector translation) {
		final int dim = translation.getLength();
		final MathMatrix ret = MathMatrix.id(dim + 1);
		for (int i = 0; i < dim; i++) {
			ret.coord[dim][i] = translation.get(i);
		}
		return ret;
	}

	/**
	 * Creates a scaling matrix for <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous
	 * coordinates</a>.
	 * 
	 * @param size
	 *            the dimension of the Cartesian vector
	 * @param value
	 *            the value to scale all dimensions with
	 * @return the matrix to multiply a homogeneous coordinate vector with to
	 *         execute this scaling
	 */
	public static MathMatrix homScale(final int size, final double value) {
		final MathMatrix ret = MathMatrix.id(size + 1);
		for (int i = 0; i < ret.getLength(0) - 1; i++) {
			ret.set(i, i, value);
		}
		return ret;
	}

	/**
	 * Creates a scaling matrix for <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous
	 * coordinates</a>.
	 * 
	 * @param scale
	 *            the Cartesian vector to scale with
	 * @return the matrix to multiply a homogeneous coordinate vector with to
	 *         execute this scaling
	 */
	public static MathMatrix homScale(final MathVector scale) {
		final int dim = scale.getLength();
		final MathMatrix ret = MathMatrix.id(dim + 1);
		for (int i = 0; i < dim; i++) {
			ret.coord[i][i] = scale.get(i);
		}
		return ret;
	}

	/**
	 * Creates a <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous</a>.
	 * perspective matrix.
	 * 
	 * @param size
	 *            the dimension of the Cartesian vector
	 * @param dimension
	 *            the dimension to zero out
	 * @param distance
	 *            the distance of the eye point in that dimension, must be non-zero.
	 * @return the matrix to multiply a homogeneous coordinate vector with to
	 *         execute the central projection from eye point onto the plane spanned by all other dimensions
	 */
	public static MathMatrix homPerspective(final int size, final int dimension, final double distance) {
		final MathMatrix ret = MathMatrix.id(size + 1);
		ret.set(dimension, dimension, 0.0);
		ret.set(dimension, size + 1, -1.0 / distance);
		return ret;
	}

	/**
	 * Creates a <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous</a>
	 * rotation matrix.
	 * 
	 * @param size
	 *            The dimension of the Cartesian vector. Should be higher or
	 *            equal to <em>idx1</em> and <em>idx2</em>.
	 * @param idx1
	 *            the index of the first dimension the rotation should be in
	 * @param idx2
	 *            the index of the second dimension the rotation should be in
	 * @param angle
	 *            the rotation angle, in mathematical notation (2*pi is full
	 *            rotation)
	 * @return the matrix to multiply a homogeneous coordinate vector with to
	 *         execute this rotation
	 */
	public static MathMatrix homRot(final int size, final int idx1, final int idx2, final double angle) {
		if ((idx1 >= size) || (idx2 >= size)) {
			throw new ArrayIndexOutOfBoundsException();
		}
		return rot(size + 1, idx1, idx2, angle);
	}

	/**
	 * Creates a <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous</a>
	 * projection matrix.
	 * 
	 * @param size
	 *            the dimension of the Cartesian vector
	 * @param dimension
	 *            the dimension to zero out
	 * @return the matrix to multiply a homogeneous coordinate vector with to
	 *         execute this projection
	 */
	public static MathMatrix homProject(final int size, final int dimension) {
		if (size <= dimension) {
			throw new ArrayIndexOutOfBoundsException();
		}
		return project(size + 1, dimension);
	}

	/**
	 * Constructor resulting in a zero matrix (a matrix full of zero values) of
	 * given size.
	 * 
	 * @param size1
	 *            the x-size of the matrx
	 * @param size2
	 *            the y-size of the matrix
	 */
	public MathMatrix(final int size1, final int size2) {
		coord = new double[size1][size2];
		for (int i = 0; i < coord.length; i++) {
			for (int j = 0; j < coord[0].length; j++) {
				coord[i][j] = 0.0;
			}
		}
	}

	/**
	 * Defines the full matrix. Be careful - this method does no deep copy.
	 * 
	 * @param values
	 *            with this content the matrix will be created
	 */
	public MathMatrix(final double[][] values) {
		setAll(values);
	}

	/**
	 * Copy constructor. Creates a deep copy.
	 * 
	 * @param other
	 *            this matrix will be cloned
	 */
	public MathMatrix(final MathMatrix other) {
		copyFrom(other);
	}

	@Override
	public MathMatrix clone() {
		return new MathMatrix(this);
	}

	/**
	 * Overwrites self with a deep copy of <em>from</em>.
	 * 
	 * @param from
	 *            the matrix that will be closed to self
	 */
	public void copyFrom(final MathMatrix from) {
		if (this != from) {
			if ((coord == null) || (coord[0] == null) || 
				(coord.length != from.coord.length) || 
				(coord[0].length != from.coord[0].length)) {
				coord = new double[from.coord.length][from.coord[0].length];
			}
			for (int i = 0; i < coord.length; i++) {
				for (int j = 0; j < coord[0].length; j++) {
					coord[i][j] = from.coord[i][j];
				}
			}
		}
	}

	/**
	 * Gets an element from the matrix.
	 * 
	 * @param index1
	 *            the x-coordinate of the element to get
	 * @param index2
	 *            the y-coordinate of the element to get
	 * @return the element at the given position in the matrix
	 */
	public double get(final int index1, final int index2) {
		return (coord[index1][index2]);
	}

	/**
	 * Sets an element of the matrix.
	 * 
	 * @param index1
	 *            the x-coordinate of the element to set
	 * @param index2
	 *            the y-coordinate of the element to set
	 * @param value
	 *            the new value for the element
	 */
	public void set(final int index1, final int index2, final double value) {
		coord[index1][index2] = value;
	}

	/**
	 * @param dim
	 *            The dimension to get the the length of. Either {@link #X} or
	 *            {@link #Y}.
	 * @return the size of the given direction of this matrix
	 */
	public int getLength(final int dim) {
		int ret = 0;
		switch(dim) {
		case X:
			ret = coord.length;
			break;
		case Y:
			ret = coord[0].length;
			break;
		default:
			throw new ArrayIndexOutOfBoundsException();
		}
		return ret;
	}

	/**
	 * Gets the full matrix. Be careful - this method gives access to the inner
	 * parts.
	 * 
	 * @return the array representation of the matrix
	 */
	public double[][] getAll() {
		return coord;
	}

	/** 
	 * Sets the full matrix. Be careful - this method does no deep copy. 
	 * 
	 * @param values
	 *            with this content the matrix will be set
	 */
	public void setAll(final double[][] values) {
		coord = values;
	}

	/**
	 * @param row
	 *            the row to return
	 * @return a new math vector containing the values of that row
	 */
	public MathVector getRow(final int row) {
		final MathVector ret = MathVector.zero(coord.length);
		for (int col = 0; col < coord.length; col++) {
			ret.set(col, coord[col][row]);
		}
		return ret;
	}

	/**
	 * @param col
	 *            the column to return
	 * @return a new math vector containing the values of that column
	 */
	public MathVector getColumn(final int col) {
		final MathVector ret = MathVector.zero(coord[0].length);
		for (int row = 0; row < coord[0].length; row++) {
			ret.set(row, coord[col][row]);
		}
		return ret;
	}

	/**
	 * Adds other matrix to this. It's an in-place operation, the result is kept in <em>this</em>.
	 * 
	 * @param other
	 *            the matrix to add
	 */
	public void addSelf(final MathMatrix other) {
		if ((coord.length != other.coord.length) || (coord[0].length != other.coord[0].length))  {
			throw new ArrayIndexOutOfBoundsException();
		}
		for (int i = 0; i < coord.length; i++) {
			for (int j = 0; j < coord[0].length; j++) {
				coord[i][j] += other.coord[i][j];
			}
		}
	}

    /** 
     * Does a matrix addition. <em>this</em> is not changed.
	 * 
	 * @param other
	 *            the matrix to add
	 * @return the resulting matrix
     */
	public MathMatrix add(final MathMatrix other) {
		final MathMatrix ret = new MathMatrix(this);
		ret.addSelf(other);
		return ret;
	}

	/** 
	 * Unary minus. It's an in-place operation, the result is kept in <em>this</em>.
	 */
	public void minusSelf() {
		for (int i = 0; i < coord.length; i++) {
			for (int j = 0; j < coord[0].length; j++) {
				coord[i][j] = -coord[i][j];
			}
		}
	}

	/**
	 * Unary minus. <em>this</em> is not changed.
	 * 
	 * @return the resulting matrix
	 */
	public MathMatrix minus() {
		final MathMatrix ret = new MathMatrix(this);
		ret.minusSelf();
		return ret;
	}

	/** 
	 * Binary minus. It's an in-place operation, the result is kept in <em>this</em>.
	 * 
	 * @param other
	 *            the matrix to substract
	 */
	public void subtractSelf(final MathMatrix other) {
		if ((coord.length != other.coord.length) || (coord[0].length != other.coord[0].length))  {
			throw new ArrayIndexOutOfBoundsException();
		}
		for (int i = 0; i < coord.length; i++) {
			for (int j = 0; j < coord[0].length; j++) {
				coord[i][j] -= other.coord[i][j];
			}
		}
	}

    /**
	 * Binary minus. <em>this</em> is not changed.
	 * 
	 * @param other
	 *            the matrix to substract
	 * @return the resulting matrix
	 */
	public MathMatrix subtract(final MathMatrix other) {
		final MathMatrix ret = new MathMatrix(this);
		ret.subtractSelf(other);
		return ret;
	}

	/** 
	 * Multiplies with scalar factor. It's an in-place operation, the result is kept in <em>this</em>.
	 * 
	 * @param factor
	 *            the scalar factor
	 */
	public void multSelf(final double factor) {
		for (int i = 0; i < coord.length; i++) {
			for (int j = 0; j < coord[0].length; j++) {
				coord[i][j] *= factor;
			}
		}
	}

	/**
	 * Multiplies with scalar factor. <em>this</em> is not changed.
	 * 
	 * @param factor
	 *            the scalar factor
	 * @return the resulting matrix
	 */
	public MathMatrix mult(final double factor) {
		final MathMatrix ret = new MathMatrix(this);
		ret.multSelf(factor);
		return ret;
	}

	/**
	 * Multiplies with other matrix. <em>this</em> is not changed.
	 * 
	 * @param other
	 *            the matrix to multiply with
	 * @return the resulting matrix
	 */
	public MathMatrix mult(final MathMatrix other) {
		if (coord.length != other.coord[0].length) {
			throw new ArrayIndexOutOfBoundsException();
		}
		final MathMatrix ret = new MathMatrix(coord[0].length, other.coord.length);
		for (int i = 0; i < ret.coord.length; i++) {
			for (int j = 0; j < ret.coord[0].length; j++) {
				for (int k = 0; k < coord.length; k++) {
					ret.coord[i][j] += coord[k][j] * other.coord[i][k];
				}
			}
		}
		return ret;
	}

	/** 
	 * Multiplies with vector. <em>this</em> is not changed.
	 * 
	 * @param other
	 *            the vector to multiply with
	 * @return the resulting matrix
	 */
	public MathVector mult(final MathVector other) {
		if (coord.length != other.getLength()) {
			throw new ArrayIndexOutOfBoundsException();
		}
		final MathVector ret = MathVector.zero(coord[0].length);
		for (int i = 0; i < ret.coord.length; i++) {
			for (int k = 0; k < coord.length; k++) {
				ret.coord[i] += coord[k][i] * other.coord[k];
			}
		}
		return ret;
	}

	/** 
	 * @return a transposition of <em>this</em>
	 */
	public MathMatrix transpose() {
		final MathMatrix ret = new MathMatrix(coord[0].length, coord.length);
		for (int i = 0; i < ret.coord.length; i++) {
			for (int j = 0; j < ret.coord[0].length; j++) {
				ret.coord[i][j] = coord[j][i];
			}
		}
		return ret;
	}

	/**
	 * Rounds each coordinate of this, result is stored in this.
	 * @param precision the precision to round to
	 * @see MathUtil#round(double, double)
	 */
	public void roundSelf(final double precision) {
		for (int i = 0; i < coord.length; i++) {
			for (int j = 0; j < coord[0].length; j++) {
				coord[i][j] = MathUtil.round(coord[i][j], precision);
			}
		}
	}

	/**
	 * Rounds each coordinate of this, result is returned.
	 * @param precision the precision to round to
	 * @return the resulting matrix
	 * @see MathUtil#round(double, double)
	 */
	public MathMatrix round(final double precision) {
		final MathMatrix ret = new MathMatrix(this);
		ret.roundSelf(precision);
		return ret;
	}

	private static final double EPS = 0.000000001;

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append("(");
		for (int i = 0; i < coord[0].length; i++) {
			ret.append("(");
			for (int j = 0; j < coord.length; j++) {
				double value = coord[j][i];
				if (Math.abs(value) < EPS) {
					value = 0.0;
				}
				ret.append(value);
				if (j < coord.length - 1) {
					ret.append(", ");
				}
			}
			ret.append(")");
			if (i < coord[0].length - 1) {
				ret.append(", ");
			}
		}
		ret.append(")");
		return ret.toString();
	}

	/**
	 * @param o
	 *            the object to compare with
	 * @return if the other object is also a matrix, and if they have the same
	 *         dimensions and all elements are equal
	 */
	@Override
	public boolean equals(final Object o) {
		if (o instanceof MathMatrix) {
			final MathMatrix other = (MathMatrix)o;
			if (coord.length == other.coord.length) {
				for (int i = 0; i < coord.length; i++) {
					if (!Arrays.equals(coord[i], other.coord[i])) {
						return false;
					}
				}
				return true;
			}
		}
		return false;
	}

	@Override
	public int hashCode() {
		int hashCode = 0;
		for (int i = 0; i < coord.length; i++) {
			hashCode ^= Arrays.hashCode(coord[i]);
		}
		return hashCode;
	}

	/** Keeps the coordinates. */
	protected double coord[][];
}
