package de.ped.tools;

import java.util.List;
import java.util.Random;

/**
 * Simple mathematical routines.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class MathUtil {

	protected static final String DIVISION_BY_ZERO = "Division by zero";
	protected static final String ALL_PROBABILITIES_IN_THE_DISTRIBUTION_WERE_0_OR_LESS = "All probabilities in the distribution were 0 or less";

	private static Random randomGenerator = new Random();

	/**
	 * Generates a (nearly) equally distributed integer random number in the
	 * range from 0 to <em>m</em> - 1 (if <em>m</em> is positive) or from
	 * <em>m</em> + 1 to 0 (if <em>m</em> is negative).
	 * 
	 * @param m
	 *            the upper bound for the random number (if positive)
	 * @param random
	 *            the random generator to use, or <code>null</code> to use the
	 *            internal default random generator.
	 * @return the random number
	 */
	public static int random(final int m, final Random random) {
		final Random rg = (null == random ? randomGenerator : random);
		if (0 == m) {
			// random(0) should behave nicely
			return 0;
		}
		if (m < 0) {
			return -random(-m);
		}
		// Get a random number, shift the sign away and do a modulo.
		return ((rg.nextInt() >>> 1) % m);
	}

	/**
	 * Generates a (nearly) equally distributed integer random number in the
	 * range from 0 to <em>m</em> - 1 (if <em>m</em> is positive) or from
	 * <em>m</em> + 1 to 0 (if <em>m</em> is negative).
	 * 
	 * @param m
	 *            the upper bound for the random number (if positive)
	 * @return the random number
	 */
	public static int random(final int m) {
		return random(m, null);
	}

	/**
	 * Generates a (nearly) equally distributed long random number in the range
	 * from 0 to <em>m</em> - 1 (if <em>m</em> is positive) or from <em>m</em> +
	 * 1 to 0 (if <em>m</em> is negative).
	 * 
	 * @param m
	 *            the upper bound for the random number (if positive)
	 * @return the random number
	 */
	public static long random(final long m) {
		if (0L == m) {
			// random(0) should behave nicely
			return 0L;
		}
		if (m < 0L) {
			return -random(-m);
		}
		// Get a random number, shift the sign away and do a modulo.
		return ((randomGenerator.nextLong() >>> 1) % m);
	}

	/**
	 * Generates a (nearly) equally distributed double random number in the
	 * range from 0 to <em>m</em> (including 0, excluding <em>m</em>) if
	 * <em>m</em> is positive, or from <em>m</em> to 0 (including 0, excluding
	 * <em>m</em>) if <em>m</em> is negative.
	 * 
	 * @param m
	 *            the upper bound for the random number (if positive)
	 * @return the random number
	 */
	public static float random(final float m) {
		if (0.0F == m) {
			// random(0) should behave nicely
			return 0.0F;
		}
		if (m < 0.0F) {
			return -random(-m);
		}
		// Get a random number in the range [0,1) and multiply with m to get the range [0,m)
		return (randomGenerator.nextFloat() * m);
	}

	/**
	 * Generates a (nearly) equally distributed double random number in the
	 * range from 0 to <em>m</em> (including 0, excluding <em>m</em>) if
	 * <em>m</em> is positive, or from <em>m</em> to 0 (including 0, excluding
	 * <em>m</em>) if <em>m</em> is negative.
	 * 
	 * @param m
	 *            the upper bound for the random number (if positive)
	 * @return the random number
	 */
	public static double random(final double m) {
		if (0.0 == m) {
			// random(0) should behave nicely
			return 0.0;
		}
		if (m < 0.0) {
			return -random(-m);
		}
		// Get a random number in the range [0,1) and multiply with m to get the range [0,m)
		return (randomGenerator.nextDouble() * m);
	}

	/**
	 * Counts the number of non-zero entries in the array.
	 * 
	 * @param distribution the array to analyze
	 * @return the number of non-zero entries in the array
	 */
	public static int count(final int[] distribution) {
		int ret = 0;
		for (int i = 0; i < distribution.length; i++) {
			if (0 != distribution[i]) {
				ret++;
			}
		}
		return ret;
	}

	/**
	 * Calculates the sum of all entries with positive values in the array.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return the sum 
	 */
	public static int sum(final int[] distribution) {
		int ret = 0;
		for (int i = 0; i < distribution.length; i++) {
			if (0 < distribution[i]) {
				ret += distribution[i];
			}
		}
		return ret;
	}

	/**
	 * Calculates the sum of all entries with positive values in the array.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return the sum 
	 */
	public static long sum(final long[] distribution) {
		long ret = 0L;
		for (int i = 0; i < distribution.length; i++) {
			if (0 < distribution[i]) {
				ret += distribution[i];
			}
		}
		return ret;
	}

	/**
	 * Calculates the sum of all entries with positive values in the array.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return the sum 
	 */
	public static float sum(final float[] distribution) {
		float ret = 0.0F;
		for (int i = 0; i < distribution.length; i++) {
			if (0.0F < distribution[i]) {
				ret += distribution[i];
			}
		}
		return ret;
	}

	/**
	 * Calculates the sum of all entries with positive values in the array.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return the sum 
	 */
	public static double sum(final double[] distribution) {
		double ret = 0.0;
		for (int i = 0; i < distribution.length; i++) {
			if (0.0 < distribution[i]) {
				ret += distribution[i];
			}
		}
		return ret;
	}

	/**
	 * chooses an integer random number in the range from 0 to
	 * distribution.length-1. If distribution[0] is two times the value of
	 * distribution[1], then it's twice as likely that the result is 0 than 1.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return index in the range from 0 to distribution.length - 1
	 * @exception java.lang.ArrayIndexOutOfBoundsException
	 *                if all probabilities in <em>distribution</em>
	 *                are 0 or less.
	 * @exception java.lang.NullPointerException
	 *                if <em>distribution</em> is <code>null</code>
	 */
	public static int randomDistributed(final int[] distribution) throws ArrayIndexOutOfBoundsException {
		int probabilitySum = sum(distribution);
		int probabilityValue = random(probabilitySum);
		probabilitySum = 0;
		for (int index = 0; index < distribution.length; index++) {
			if (0 < distribution[index]) {
				probabilitySum += distribution[index];
				if (probabilityValue < probabilitySum) {
					return index;
				}
			}
		}
		throw new ArrayIndexOutOfBoundsException(ALL_PROBABILITIES_IN_THE_DISTRIBUTION_WERE_0_OR_LESS);
	}

	/**
	 * chooses an integer random number in the range from 0 to
	 * distribution.length-1. If distribution[0] is two times the value of
	 * distribution[1], then it's twice as likely that the result is 0 than 1.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return index in the range from 0 to distribution.length - 1
	 * @exception java.lang.ArrayIndexOutOfBoundsException
	 *                if all probabilities in <em>distribution</em>
	 *                are 0 or less.
	 * @exception java.lang.NullPointerException
	 *                if <em>distribution</em> is <code>null</code>
	 */
	public static int randomDistributed(final long[] distribution) throws ArrayIndexOutOfBoundsException {
		long probabilitySum = sum(distribution);
		long probabilityValue = random(probabilitySum);
		probabilitySum = 0L;
		for (int index = 0; index < distribution.length; index++) {
			if (0 < distribution[index]) {
				probabilitySum += distribution[index];
				if (probabilityValue < probabilitySum) {
					return index;
				}
			}
		}
		throw new ArrayIndexOutOfBoundsException(ALL_PROBABILITIES_IN_THE_DISTRIBUTION_WERE_0_OR_LESS);
	}

	/**
	 * chooses an integer random number in the range from 0 to
	 * distribution.length-1. If distribution[0] is two times the value of
	 * distribution[1], then it's twice as likely that the result is 0 than 1.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return index in the range from 0 to distribution.length - 1
	 * @exception java.lang.ArrayIndexOutOfBoundsException
	 *                if all probabilities in <em>distribution</em>
	 *                are 0 or less.
	 * @exception java.lang.NullPointerException
	 *                if <em>distribution</em> is <code>null</code>
	 */
	public static int randomDistributed(final float[] distribution) throws ArrayIndexOutOfBoundsException {
		float probabilitySum = sum(distribution);
		float probabilityValue = random(probabilitySum);
		probabilitySum = 0.0F;
		for (int index = 0; index < distribution.length; index++) {
			if (0.0F < distribution[index]) {
				probabilitySum += distribution[index];
				if (probabilityValue < probabilitySum) {
					return index;
				}
			}
		}
		throw new ArrayIndexOutOfBoundsException(ALL_PROBABILITIES_IN_THE_DISTRIBUTION_WERE_0_OR_LESS);
	}

	/**
	 * chooses an integer random number in the range from 0 to
	 * distribution.length-1. If distribution[0] is two times the value of
	 * distribution[1], then it's twice as likely that the result is 0 than 1.
	 * 
	 * @param distribution
	 *            the array to analyze
	 * @return index in the range from 0 to distribution.length - 1
	 * @exception java.lang.ArrayIndexOutOfBoundsException
	 *                if all probabilities in <em>distribution</em>
	 *                are 0 or less.
	 * @exception java.lang.NullPointerException
	 *                if <em>distribution</em> is <code>null</code>
	 */
	public static int randomDistributed(final double[] distribution) throws ArrayIndexOutOfBoundsException {
		double probabilitySum = sum(distribution);
		double probabilityValue = random(probabilitySum);
		probabilitySum = 0.0;
		for (int index = 0; index < distribution.length; index++) {
			if (0.0 < distribution[index]) {
				probabilitySum += distribution[index];
				if (probabilityValue < probabilitySum) {
					return index;
				}
			}
		}
		throw new ArrayIndexOutOfBoundsException(ALL_PROBABILITIES_IN_THE_DISTRIBUTION_WERE_0_OR_LESS);
	}
	
	/**
	 * Shuffles the array randomly in-place.
	 * 
	 * @param array
	 *            the array to shuffle
	 * @param random
	 *            the random generator to use, or <code>null</code> to use the
	 *            internal default random generator
	 */
	public static void shuffle(final int[] array, final Random random) {
		for (int i = array.length - 1; 0 < i; i--) {
			final int j = random(i + 1, random);
			if (i != j) {
				int buf = array[j];
				array[j] = array[i];
				array[i] = buf;
			}
		}
	}

	/**
	 * Shuffles the array randomly in-place.
	 * 
	 * @param <T>
	 *            any type
	 * @param array
	 *            the array to shuffle
	 * @param random
	 *            the random generator to use, or <code>null</code> to use the
	 *            internal default random generator
	 */
	public static <T> void shuffle(final T[] array, final Random random) {
		for (int i = array.length - 1; 0 < i; i--) {
			final int j = random(i + 1, random);
			if (i != j) {
				T buf = array[j];
				array[j] = array[i];
				array[i] = buf;
			}
		}
	}

	/**
	 * Shuffles the list randomly in-place. Implementation can still be
	 * optimized for linked lists.
	 * 
	 * @param <T>
	 *            any type
	 * @param array
	 *            the list to shuffle
	 * @param random
	 *            the random generator to use, or <code>null</code> to use the
	 *            internal default random generator
	 */
	public static <T> void shuffle(final List<T> array, final Random random) {
		for (int i = array.size() - 1; 0 < i; i--) {
			final int j = random(i + 1, random);
			if (i != j) {
				T buf = array.get(j);
				array.set(j, array.get(i));
				array.set(i, buf);
			}
		}
	}

	/**
	 * A friendlier modulo function, that behaves more like the counter model, or kitchen clock model:
	 * If it's 1 o'clock and I turn back the clock about 1 hour, it's not zero o'clock, but 12 o'clock.
	 * If I turn it back one more hour, it's not -1 o'clock, but 11 o'clock. And so on.
	 * <br>
	 * To give some more hints for the concepts: The following equations are always true.
	 * <pre>
	 *     modulo(val, x, y) = -modulo(-val, -x, -y)
	 *     modulo(val, x, y) = modulo(val - y, x, 0) + y
	 * </pre>
	 * Examples:
	 * <br>
	 * If offset is 0, the behavior is nearly like normal the modulo operator (%)
	 * Otherwise it tells the lower bound of the destination range. Even more examples:
	 * <br>
	 * <table border="1">
	 * <caption>Examples</caption>
	 * <thead>
	 * <tr><th> call                     </th><th>             result </th><th> comment </th></tr>
	 * </thead>
	 * <tbody>
	 * <tr><td><tt>modulo( 12, 5, 0)</tt></td><td align="right">  2 </td><td> this behaves like the normal modulo operation (12 % 5)</td></tr>
	 * <tr><td><tt>modulo(-12, 5, 0)</tt></td><td align="right">  3 </td><td> To negative numbers, we add full circles (mod) till it's positive </td></tr>
	 * <tr><td><tt>modulo( 12,-5, 0)</tt></td><td align="right"> -3 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * <tr><td><tt>modulo(-12,-5, 0)</tt></td><td align="right"> -2 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * <tr><td><tt>modulo(  1,12, 1)</tt></td><td align="right">  1 </td><td> limits any number to the range 1..12, like the kitchen clock. In the range all stays the same. </td></tr>
	 * <tr><td><tt>modulo( 12,12, 1)</tt></td><td align="right"> 12 </td><td> limits any number to the range 1..12, like the kitchen clock. In the range all stays the same. </td></tr>
	 * <tr><td><tt>modulo( 13,12, 1)</tt></td><td align="right">  1 </td><td> limits any number to the range 1..12, like the kitchen clock. </td></tr>
	 * <tr><td><tt>modulo( 14,12, 1)</tt></td><td align="right">  2 </td><td> limits any number to the range 1..12, like the kitchen clock. </td></tr>
	 * <tr><td><tt>modulo(  0,12, 1)</tt></td><td align="right"> 12 </td><td> limits any number to the range 1..12, like the kitchen clock. </td></tr>
	 * <tr><td><tt>modulo(  0, 3,-1)</tt></td><td align="right">  0 </td><td> limits any number to the range -1..1. In the range all stays the same. </td></tr>
	 * <tr><td><tt>modulo(  2, 3,-1)</tt></td><td align="right"> -1 </td><td> limits any number to the range -1..1. </td></tr>
	 * </tbody>
	 * </table>
	 *
	 * @param val
	 *            the value to be modulated
	 * @param mod
	 *            the upper bound (if offset is 0), or the length of the
	 *            destination range (if offset is not 0)
	 * @param offset
	 *            the lower bound of the destination range
	 * @return the result
	 */
	public static int modulo(int val, final int mod, final int offset) {
		int ret;
		if (0 == mod) { // Filter out invalid values and behave nicely
			ret = 0;
		} else if ((offset <= val) && (val < mod + offset)) { // We're already in range
			ret = val;
		} else {
			// First do a normation and get rid of the offset
			val = val - offset;
			// Now have a look at all the different cases
			if (mod < 0) {
				ret = -modulo(-val, -mod, 0) + offset;
			} else if (val < 0) {
				int a = (-val) % mod;
				if (0 == a) {
					ret = offset;
				} else {
					ret = (mod - a) + offset;
				}
			} else {
				ret = (val % mod) + offset;
			}
		}
		return ret;

	}

	/**
	 * A shortcut for modulo(<em>val</em>, <em>mod</em>, 0).
	 *
	 * @param val
	 *            the value to be modulated
	 * @param mod
	 *            the upper bound 
	 * @return the result
	 * @see #modulo(int, int, int)
	 */
	public static int modulo(final int val, final int mod) {
		return modulo(val, mod, 0);
	}

	public static long modulo(long val, final long mod, final long offset) {
		long ret;
		if (0 == mod) { // Filter out invalid values and behave nicely
			ret = 0;
		} else if ((offset <= val) && (val < mod + offset)) { // We're already in range
			ret = val;
		} else {
			// First do a normation and get rid of the offset
			val = val - offset;
			// Now have a look at all the different cases
			if (mod < 0) {
				ret = -modulo(-val, -mod, 0) + offset;
			} else if (val < 0) {
				long a = (-val) % mod;
				if (0 == a) {
					ret = offset;
				} else {
					ret = (mod - a) + offset;
				}
			} else {
				ret = (val % mod) + offset;
			}
		}
		return ret;

	}

	/**
	 * A shortcut for modulo(<em>val</em>, <em>mod</em>, 0).
	 *
	 * @param val
	 *            the value to be modulated
	 * @param mod
	 *            the upper bound
	 * @return the result
	 * @see #modulo(long, long, long)
	 */
	public static long modulo(final long val, final long mod) {
		return modulo(val, mod, 0L);
	}

	/**
	 * An adequate division method fitting for modulo.
	 * The intention is that for every non-zero x and every y the following equation is true:
	 * <pre>
	 *     val = x * divisio(val, x, y) + modulo(val, x, y)
	 * </pre>
	 * To understand that concept, start out with y=0 and x&gt;0.
	 * <br>
	 * To give some more hints for the concepts: The following equations are always true.
	 * <pre>
	 *     divisio(val, x, y) = divisio(-val, -x, -y)
	 *     divisio(val, x, y) = divisio(val - y, x, 0)
	 * </pre>
	 * Examples:
	 * <br>
	 * If offset is 0, the behavior is nearly like normal the division operator (/)
	 * Otherwise it tells the lower bound of the destination range. Even more examples:
	 * <br>
	 * <table border="1">
	 * <caption>Examples</caption>
	 * <thead>
	 * <tr><th> call                      </th><th>             result </th><th> comment </th></tr>
	 * </thead>
	 * <tbody>
	 * <tr><td><tt>divisio(  0, 5, 0)</tt></td><td align="right">  0 </td><td> this behaves like the normal division operation (0 / 5)</td></tr>
	 * <tr><td><tt>divisio( 12, 5, 0)</tt></td><td align="right">  2 </td><td> this behaves like the normal division operation (12 / 5)</td></tr>
	 * <tr><td><tt>divisio( -1, 5, 0)</tt></td><td align="right"> -1 </td><td> -1 is the last number of the smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio( -5, 5, 0)</tt></td><td align="right"> -1 </td><td> -5 is the first number of the smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio( -6, 5, 0)</tt></td><td align="right"> -2 </td><td> -6 is the last number of the second smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio(  1,-5, 0)</tt></td><td align="right"> -1 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * <tr><td><tt>divisio(  0,-5, 0)</tt></td><td align="right">  0 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * </tbody>
	 * </table>
	 * 
	 * @param val
	 *            the value to be divided
	 * @param div
	 *            the upper bound (if offset is 0), or the length of the
	 *            destination range (if offset is not 0)
	 * @param offset
	 *            the lower bound of the destination range
	 * @return the result
	 * @throws ArithmeticException
	 *             thrown if <em>div</em> is 0.
	 * @see #modulo(int, int, int)
	 */
	public static int divisio(int val, final int div, final int offset) {
		// First do a normation and get rid of the offset
		val = val - offset;
		// Now have a look at all the different cases
		if (div < 0) {
			return divisio(-val, -div, 0);
		}
		if (0 <= val) {
			// For positive val, it's a normal division
			return (val / div);
		} else {
			// For negative val, we have to take special care, as there is no "-0"
			return ((val + (1 - div)) / div);
		}
	}

	/**
	 * A shortcut for divisio(<em>val</em>, <em>div</em>, 0).
	 *
	 * @param val
	 *            the value to be divided
	 * @param div
	 *            the upper bound
	 * @return the result
	 * @see #divisio(int, int, int)
	 */
	public static int divisio(final int val, final int div) {
		return divisio(val, div, 0);
	}

	/**
	 * An adequate division method fitting for modulo.
	 * The intention is that for every non-zero x and every y the following equation is true:
	 * <pre>
	 *     val = x * divisio(val, x, y) + modulo(val, x, y)
	 * </pre>
	 * To understand that concept, start out with y=0 and x&gt;0.
	 * <br>
	 * To give some more hints for the concepts: The following equations are always true.
	 * <pre>
	 *     divisio(val, x, y) = divisio(-val, -x, -y)
	 *     divisio(val, x, y) = divisio(val - y, x, 0)
	 * </pre>
	 * Examples:
	 * <br>
	 * If offset is 0, the behavior is nearly like normal the division operator (/)
	 * Otherwise it tells the lower bound of the destination range. Even more examples:
	 * <br>
	 * <table border="1">
	 * <caption>Examples</caption>
	 * <thead>
	 * <tr><th> call                      </th><th>             result </th><th> comment </th></tr>
	 * </thead>
	 * <tbody>
	 * <tr><td><tt>divisio(  0, 5, 0)</tt></td><td align="right">  0 </td><td> this behaves like the normal division operation (0 / 5)</td></tr>
	 * <tr><td><tt>divisio( 12, 5, 0)</tt></td><td align="right">  2 </td><td> this behaves like the normal division operation (12 / 5)</td></tr>
	 * <tr><td><tt>divisio( -1, 5, 0)</tt></td><td align="right"> -1 </td><td> -1 is the last number of the smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio( -5, 5, 0)</tt></td><td align="right"> -1 </td><td> -5 is the first number of the smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio( -6, 5, 0)</tt></td><td align="right"> -2 </td><td> -6 is the last number of the second smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio(  1,-5, 0)</tt></td><td align="right"> -1 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * <tr><td><tt>divisio(  0,-5, 0)</tt></td><td align="right">  0 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * </tbody>
	 * </table>
	 *
	 * @param val
	 *            the value to be divided
	 * @param div
	 *            the upper bound (if offset is 0), or the length of the
	 *            destination range (if offset is not 0)
	 * @param offset
	 *            the lower bound of the destination range
	 * @return the result
	 * @throws ArithmeticException
	 *             thrown if <em>div</em> is 0.
	 * @see #modulo(double, double, double)
	 */
	public static int divisio(float val, final float div, final float offset) {
		// First do a normation and get rid of the offset
		val = val - offset;
		float dres = val / div;
		if (Float.isNaN(dres)) {
			throw new ArithmeticException(DIVISION_BY_ZERO);
		}
		return (int) Math.floor(dres);
	}

	/**
	 * A shortcut for divisio(<em>val</em>, <em>div</em>, 0).
	 *
	 * @param val
	 *            the value to be divided
	 * @param div
	 *            the upper bound
	 * @return the result
	 * @see #divisio(double, double, double)
	 */
	public static int divisio(final float val, final float div) {
		return divisio(val, div, 0.0F);
	}

	/**
	 * The counterpart of my extended modulo, adapted for doubles.
	 * <br>
	 * Example: 11.0 mod 2.5 = (11.0 - 4 * 2.5) mod 2.5 = (11.0 - 10.0) mod 2.5 = 1.0
	 *
	 * @param val
	 *            the value to be modulated
	 * @param mod
	 *            the upper bound (if offset is 0), or the length of the
	 *            destination range (if offset is not 0)
	 * @param offset
	 *            the lower bound of the destination range
	 * @return the result
	 */
	public static float modulo(final float val, final float mod, final float offset) {
		float ret;
		if (0.0F == mod) { // Filter out invalid values and behave nicely
			ret = 0.0F;
		} else if ((offset <= val) && (val < mod + offset)) { // We're already in range
			ret = val;
		} else {
			ret = val - mod * divisio(val, mod, offset);
		}
		return ret;
	}

	public static float modulo(final float val, final float mod) {
		return modulo(val, mod, 0.0F);
	}

	/**
	 * An adequate division method fitting for modulo.
	 * The intention is that for every non-zero x and every y the following equation is true:
	 * <pre>
	 *     val = x * divisio(val, x, y) + modulo(val, x, y)
	 * </pre>
	 * To understand that concept, start out with y=0 and x&gt;0.
	 * <br>
	 * To give some more hints for the concepts: The following equations are always true.
	 * <pre>
	 *     divisio(val, x, y) = divisio(-val, -x, -y)
	 *     divisio(val, x, y) = divisio(val - y, x, 0)
	 * </pre>
	 * Examples:
	 * <br>
	 * If offset is 0, the behavior is nearly like normal the division operator (/)
	 * Otherwise it tells the lower bound of the destination range. Even more examples:
	 * <br>
	 * <table border="1">
	 * <caption>Examples</caption>
	 * <thead>
	 * <tr><th> call                      </th><th>             result </th><th> comment </th></tr>
	 * </thead>
	 * <tbody>
	 * <tr><td><tt>divisio(  0, 5, 0)</tt></td><td align="right">  0 </td><td> this behaves like the normal division operation (0 / 5)</td></tr>
	 * <tr><td><tt>divisio( 12, 5, 0)</tt></td><td align="right">  2 </td><td> this behaves like the normal division operation (12 / 5)</td></tr>
	 * <tr><td><tt>divisio( -1, 5, 0)</tt></td><td align="right"> -1 </td><td> -1 is the last number of the smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio( -5, 5, 0)</tt></td><td align="right"> -1 </td><td> -5 is the first number of the smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio( -6, 5, 0)</tt></td><td align="right"> -2 </td><td> -6 is the last number of the second smallest "negative" circle</td></tr>
	 * <tr><td><tt>divisio(  1,-5, 0)</tt></td><td align="right"> -1 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * <tr><td><tt>divisio(  0,-5, 0)</tt></td><td align="right">  0 </td><td> With negative bounds, the results are reverted (whoever that needs...) </td></tr>
	 * </tbody>
	 * </table>
	 *
	 * @param val
	 *            the value to be divided
	 * @param div
	 *            the upper bound (if offset is 0), or the length of the
	 *            destination range (if offset is not 0)
	 * @param offset
	 *            the lower bound of the destination range
	 * @return the result
	 * @throws ArithmeticException
	 *             thrown if <em>div</em> is 0.
	 * @see #modulo(double, double, double)
	 */
	public static int divisio(double val, final double div, final double offset) {
		// First do a normation and get rid of the offset
		val = val - offset;
		double dres = val / div;
		if (Double.isNaN(dres)) {
			throw new ArithmeticException(DIVISION_BY_ZERO);
		}
		return (int) Math.floor(dres);
	}

	/**
	 * A shortcut for divisio(<em>val</em>, <em>div</em>, 0).
	 *
	 * @param val
	 *            the value to be divided
	 * @param div
	 *            the upper bound
	 * @return the result
	 * @throws ArithmeticException
	 *             thrown if <em>div</em> is 0.
	 * @see #divisio(double, double, double)
	 */
	public static int divisio(final double val, final double div) {
		return divisio(val, div, 0.0);
	}

	/**
	 * The counterpart of my extended modulo, adapted for doubles.
	 * <br>
	 * Example: 11.0 mod 2.5 = (11.0 - 4 * 2.5) mod 2.5 = (11.0 - 10.0) mod 2.5 = 1.0
	 *
	 * @param val
	 *            the value to be modulated
	 * @param mod
	 *            the upper bound (if offset is 0), or the length of the
	 *            destination range (if offset is not 0)
	 * @param offset
	 *            the lower bound of the destination range
	 * @return the result
	 */
	public static double modulo(final double val, final double mod, final double offset) {
		double ret;
		if (0.0 == mod) { // Filter out invalid values and behave nicely
			ret = 0.0;
		} else if ((offset <= val) && (val < mod + offset)) { // We're already in range
			ret = val;
		} else {
			ret = val - mod * divisio(val, mod, offset);
		}
		return ret;
	}

	public static double modulo(final double val, final double mod) {
		return modulo(val, mod, 0.0);
	}

	/**
	 * Rounds <em>value</em> to the nearest grid position, where precision
	 * defines the distance between those grid positions.
	 * 
	 * @param value
	 *            the value to round
	 * @param precision
	 *            the distance between grid positions
	 * @return the nearest grid position to value
	 */
	public static int round(final int value, final int precision) {
		return precision * MathUtil.divisio(value + precision / 2, precision);
	}

	/**
	 * Rounds <em>value</em> to the nearest grid position, where precision
	 * defines the distance between those grid positions. If you set
	 * <em>precision</em> to 1.0F, this behaves like
	 * {@link java.lang.Math#round(float)}.
	 * 
	 * @param value
	 *            the value to round
	 * @param precision
	 *            the distance between grid positions
	 * @return the nearest grid position to value
	 */
	public static float round(final float value, final float precision) {
		return precision * MathUtil.divisio(value + precision / 2.0F, precision);
	}

	/**
	 * Rounds <em>value</em> to the nearest grid position, where precision
	 * defines the distance between those grid positions. If you set
	 * <em>precision</em> to 1.0, this behaves like
	 * {@link java.lang.Math#round(double)}.
	 * 
	 * @param value
	 *            the value to round
	 * @param precision
	 *            the distance between grid positions
	 * @return the nearest grid position to value
	 */
	public static double round(final double value, final double precision) {
		return precision * MathUtil.divisio(value + precision / 2.0, precision);
	}

	/**
	 * A bounds function, cutting values exceeding the given bounds into the
	 * nearest bound.
	 * 
	 * @param val
	 *            the value to be bound
	 * @param min
	 *            the lower bound. If value is lesser than this bound, this
	 *            bound is returned.
	 * @param max
	 *            the upper bound. If value is greater than this bound, this
	 *            bound is returned.
	 * @return <em>val</em> if it is in the range, otherwise <em>min</em> or
	 *         <em>max</em>
	 */
	public static int bound(final int val, final int min, final int max) {
		if (val < min)	return min;
		if (val > max)	return max;
		return val;
	}

	/**
	 * A bounds function, cutting values exceeding the given bounds into the
	 * nearest bound.
	 * 
	 * @param val
	 *            the value to be bound
	 * @param min
	 *            the lower bound. If value is lesser than this bound, this
	 *            bound is returned.
	 * @param max
	 *            the upper bound. If value is greater than this bound, this
	 *            bound is returned.
	 * @return <em>val</em> if it is in the range, otherwise <em>min</em> or
	 *         <em>max</em>
	 */
	public static long bound(final long val, final long min, final long max) {
		if (val < min)	return min;
		if (val > max)	return max;
		return val;
	}

	/**
	 * A bounds function, cutting values exceeding the given bounds into the
	 * nearest bound.
	 * 
	 * @param val
	 *            the value to be bound
	 * @param min
	 *            the lower bound. If value is lesser than this bound, this
	 *            bound is returned.
	 * @param max
	 *            the upper bound. If value is greater than this bound, this
	 *            bound is returned.
	 * @return <em>val</em> if it is in the range, otherwise <em>min</em> or
	 *         <em>max</em>
	 */
	public static float bound(final float val, final float min, final float max) {
		if (val < min)	return min;
		if (val > max)	return max;
		return val;
	}

	/**
	 * A bounds function, cutting values exceeding the given bounds into the
	 * nearest bound.
	 * 
	 * @param val
	 *            the value to be bound
	 * @param min
	 *            the lower bound. If value is lesser than this bound, this
	 *            bound is returned.
	 * @param max
	 *            the upper bound. If value is greater than this bound, this
	 *            bound is returned.
	 * @return <em>val</em> if it is in the range, otherwise <em>min</em> or
	 *         <em>max</em>
	 */
	public static double bound(final double val, final double min, final double max) {
		if (val < min)	return min;
		if (val > max)	return max;
		return val;
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @param epsilon
	 *            the rounding error allowed to be considered as zero
	 * @return -1, if <em>val</em> is smaller than -epsilon, +1 if <em>val</em>
	 *         is larger than epsilon, and 0 if <em>val</em> is within epsilon
	 *         range to 0
	 */
	public static int sgn(final int val, final int epsilon) {
		if (val < epsilon) return -1;
		if (val > epsilon) return  1;
		return 0;
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @return -1, if <em>val</em> is negative, +1 if <em>val</em> is positive,
	 *         and 0 if <em>val</em> is 0
	 */
	public static int sgn(final int val) {
		return sgn(val, 0);
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @param epsilon
	 *            the rounding error allowed to be considered as zero
	 * @return -1, if <em>val</em> is smaller than -epsilon, +1 if <em>val</em>
	 *         is larger than epsilon, and 0 if <em>val</em> is within epsilon
	 *         range to 0
	 */
	public static int sgn(final long val, final long epsilon) {
		if (val < epsilon) return -1;
		if (val > epsilon) return  1;
		return 0;
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @return -1, if <em>val</em> is negative, +1 if <em>val</em> is positive,
	 *         and 0 if <em>val</em> is 0
	 */
	public static int sgn(final long val) {
		return sgn(val, 0L);
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @param epsilon
	 *            the rounding error allowed to be considered as zero
	 * @return -1, if <em>val</em> is smaller than -epsilon, +1 if <em>val</em>
	 *         is larger than epsilon, and 0 if <em>val</em> is within epsilon
	 *         range to 0
	 */
	public static int sgn(final float val, final float epsilon) {
		if (val < epsilon) return -1;
		if (val > epsilon) return  1;
		return 0;
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @return -1, if <em>val</em> is negative, +1 if <em>val</em> is positive,
	 *         and 0 if <em>val</em> is 0
	 */
	public static int sgn(final float val) {
		return sgn(val, 0.0F);
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @param epsilon
	 *            the rounding error allowed to be considered as zero
	 * @return -1, if <em>val</em> is smaller than -epsilon, +1 if <em>val</em>
	 *         is larger than epsilon, and 0 if <em>val</em> is within epsilon
	 *         range to 0
	 */
	public static int sgn(final double val, final double epsilon) {
		if (val < epsilon) return -1;
		if (val > epsilon) return  1;
		return 0;
	}

	/**
	 * A signum function.
	 * 
	 * @param val
	 *            the value the sign is to be checked of
	 * @return -1, if <em>val</em> is negative, +1 if <em>val</em> is positive,
	 *         and 0 if <em>val</em> is 0
	 */
	public static int sgn(final double val) {
		return sgn(val, 0.0);
	}
}
