package de.ped.tools;

import java.util.Arrays;

/**
 * A simple mathematical vector class for dense vectors of scalar type double.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MathVector implements Cloneable, Comparable<MathVector> {
	/** Index of the x coordinate. */
	public static final int X = 0;

	/** Index of the y coordinate. */
	public static final int Y = 1;

	/** Index of the z coordinate. */
	public static final int Z = 2;

	/**
	 * @param size
	 *            the size of the vector
	 * @return a zero vector (a vector full of zero values) of given size
	 */
	public static MathVector zero(final int size) {
		final MathVector ret = new MathVector(size);
		return ret;
	}

	/**
	 * @param size
	 *            the size of the vector
	 * @return an id vector (a vector full of 1.0 values) of given size
	 */
	public static MathVector id(final int size) {
		final MathVector ret = new MathVector(size);
		for (int i = 0; i < ret.coord.length; i++) {
			ret.coord[i] = 1.0;
		}
		return ret;
	}

	/** 
	 * Convenient factory for vectors.
	 * @param a the list of doubles to create a vector from
	 * @return the vector object containing those 
	 */
	public static MathVector create(final double ...a) {
		return new MathVector(a);
	}

	protected static final boolean IS_PRECALCING_HASH = false;

	/**
	 * Constructor resulting in a Nullvector of given dimension.
	 * 
	 * @param dimension
	 *            the dimension of the vector
	 */
	public MathVector(final int dimension) {
		coord = new double[dimension];
		for (int i = 0; i < coord.length; i++) {
			coord[i] = 0.0;
		}
	}

	/**
	 * Defines the full vector. Be careful - this method does no deep copy.
	 * 
	 * @param values
	 *            the content of the vector
	 */
	public MathVector(final double[] values) {
		coord = values;
	}

	/**
	 * Creates a deep copy.
	 * 
	 * @param other
	 *            the vector to clone from
	 */
	public MathVector(final MathVector other) {
		coord = new double[other.coord.length];
		copyFrom(other);
	}

	@Override
	public MathVector clone() {
		MathVector ret = null;
		try {
			ret = (MathVector)super.clone();
			ret.coord = new double[coord.length];
			ret.copyFrom(this);
		} catch (final CloneNotSupportedException e) {
			// Nothing to do
		}
		return ret;
	}

	/**
	 * Creates a deep copy.
	 * 
	 * @param from
	 *            the vector to clone from
	 */
	public void copyFrom(final MathVector from) {
		if (this != from) {
			if (coord.length != from.coord.length) {
				coord = new double[from.coord.length];
			}
			for (int i = 0; i < coord.length; i++) {
				coord[i] = from.coord[i];
			}
			isHashCodeValid = false;
		}
	}

	/**
	 * Gets an element from the vector.
	 * 
	 * @param index
	 *            the index of the element to get
	 * @return the <em>index</em>th element
	 */
	public double get(final int index) {
		return (coord[index]);
	}

	/**
	 * Sets an element of the vector.
	 * 
	 * @param index
	 *            the index of the element to set
	 * @param value
	 *            the value to set
	 */
	public void set(final int index, final double value) {
		coord[index] = value;
		isHashCodeValid = false;
	}

	public int getLength() {
		return coord.length;
	}
	
	/** 
	 * @return the full content of the vector. Be careful - this method gives access to the inner parts. 
	 */
	public double[] getAll() {
		return coord;
	}

	/**
	 * Sets the full vector. Be careful - this method does no deep copy.
	 * 
	 * @param values
	 *            the full content of the vector
	 */
	public void setAll(final double[] values) {
		coord = values;
		isHashCodeValid = false;
	}

	public void setAllFrom(final double ...values) {
		setAll(values);
	}

	/**
	 * Does a vector addition from other vector to this, result is stored in this.
	 * 
	 * @param other
	 *            the vector to add
	 */
	public void addSelf(final MathVector other) {
		if (coord.length != other.coord.length) {
			throw new ArrayIndexOutOfBoundsException();
		}
		for (int i = 0; i < coord.length; i++) {
			coord[i] = coord[i] + other.coord[i];
		}
		isHashCodeValid = false;
	}

	/**
	 * Adds other vector to this; this is untouched.
	 * 
	 * @param other
	 *            the vector to add
	 * @return the result of the addition
	 */
	public MathVector add(final MathVector other) {
		final MathVector ret = new MathVector(this);
		ret.addSelf(other);
		return ret;
	}

	/** 
	 * Unary minus. Result is stored in this.
	 */
	public void minusSelf() {
		for (int i = 0; i < coord.length; i++) {
			coord[i] = -coord[i];
		}
		isHashCodeValid = false;
	}

	public MathVector minus() {
		final MathVector ret = new MathVector(this);
		ret.minusSelf();
		return ret;
	}

	/**
	 * Binary minus. Result is stored in this.
	 * 
	 * @param other
	 *            the vector to substract
	 */
	public void subtractSelf(final MathVector other) {
		if (coord.length != other.coord.length) {
			throw new ArrayIndexOutOfBoundsException();
		}
		for (int i = 0; i < coord.length; i++) {
			coord[i] = coord[i] - other.coord[i];
		}
		isHashCodeValid = false;
	}

	public MathVector subtract(final MathVector other) {
		final MathVector ret = new MathVector(this);
		ret.subtractSelf(other);
		return ret;
	}

	/**
	 * Multiplies with scalar factor. Result is stored in this.
	 * 
	 * @param factor
	 *            the scalar factor
	 */
	public void multSelf(final double factor) {
		for (int i = 0; i < coord.length; i++) {
			coord[i] = coord[i] * factor;
		}
		isHashCodeValid = false;
	}

	public MathVector mult(final double factor) {
		final MathVector ret = new MathVector(this);
		ret.multSelf(factor);
		return ret;
	}

	/**
	 * Multiplies with other vector. Result is stored in this.
	 * 
	 * @param other
	 *            the vector to multiply with
	 */
	public void multSelf(final MathVector other) {
		if (coord.length != other.coord.length) {
			throw new ArrayIndexOutOfBoundsException();
		}
		for (int i = 0; i < coord.length; i++) {
			coord[i] = coord[i] * other.coord[i];
		}
		isHashCodeValid = false;
	}

	public MathVector mult(final MathVector other) {
		final MathVector ret = new MathVector(this);
		ret.multSelf(other);
		return ret;
	}

	/**
	 * Multiplies with matrix. this is not changed.
	 * 
	 * @param other
	 *            the matrix to multiply with
	 * @return the resulting matrix
	 */
	public MathVector mult(final MathMatrix other) {
		if (coord.length != other.getLength(MathMatrix.X)) {
			throw new ArrayIndexOutOfBoundsException();
		}
		final MathVector ret = MathVector.zero(other.getLength(MathMatrix.Y));
		for (int i = 0; i < ret.coord.length; i++) {
			for (int k = 0; k < coord.length; k++) {
				ret.coord[i] += coord[k] * other.coord[i][k];
			}
		}
		return ret;
	}

	/**
	 * @return the length of the vector in the vectorspace, to the power of two.
	 */
	public double getRadiusSquare() {
		double result = 0.0;
		for (int i = 0; i < coord.length; i++) {
			result += (coord[i] * coord[i]);
		}
		return result;
	}

	/**
	 * @return the length of the vector in the vectorspace
	 */
	public double getRadius() {
		return Math.sqrt(getRadiusSquare());
	}

	/** 
	 * Norms the vector to the radius of 1. Result is stored in this.
	 */
	public void normSelf() {
		multSelf(1 / getRadius());
	}

	public MathVector norm() {
		final MathVector ret = new MathVector(this);
		ret.normSelf();
		return ret;
	}

	/**
	 * @param idx1 the index of the first dimension spanning the plane
	 * @param idx2 the index of the second dimension spanning the plane
	 * @return the angle of the vector in the plane spanned by idx1 and idx2
	 */
	public double getAngle(final int idx1, final int idx2) {
		return Math.atan2(coord[idx1], coord[idx2]);
	}

	/**
	 * Rotates the vector about 90 degrees in the plane spanned by idx1 and idx2 for howOften times in-place.
	 * If the coordinates are positive to the right and down (like e.g. screen coordinates), 
	 * then this is a turn right.
	 * If the coordinates are positive to the right and up (like e.g. a mathematical coordinate system), 
	 * then this is a turn left.
	 * 
	 * @param idx1 the index of the first dimension spanning the plane
	 * @param idx2 the index of the second dimension spanning the plane
	 * @param howOften so many turns. May be negative, then it turns the other way round.
	 */
	public void rotateSelfQuarter(final int idx1, final int idx2, final int howOften) {
		final int normed = MathUtil.modulo(howOften, 4);
		for (int i = 0; i < normed; i++) {
			final double buf = coord[idx1];
			coord[idx1] = -coord[idx2];
			coord[idx2] = buf;
		}
	}

	private static final double EPS = 0.000000001;

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append("(");
		for (int i = 0; i < coord.length; i++) {
			double value = coord[i];
			if (Math.abs(value) < EPS) {
				value = 0.0;
			}
			ret.append(Math.round(value * 10000) / 10000.0);
			if (i < coord.length - 1) {
				ret.append(", ");
			}
		}
		ret.append(")");
		return ret.toString();
	}

	/**
	 * @param epsilon
	 *            the rounding error to accept
	 * @return if the vector radius is smaller than the given epsilon
	 */
	public boolean isNullVector(final double epsilon) {
		return (getRadiusSquare() < (epsilon * epsilon));
	}

	/**
	 * Rounds each coordinate of this, result is stored in this.
	 * @param precision the precision to round to
	 * @see MathUtil#round(double, double)
	 */
	public void roundSelf(final double precision) {
		for (int i = 0; i < coord.length; i++) {
			coord[i] = MathUtil.round(coord[i], precision);
		}
	}

	public MathVector round(final double precision) {
		final MathVector ret = new MathVector(this);
		ret.roundSelf(precision);
		return ret;
	}

	/**
	 * Creates <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous
	 * coordinates</a> from Cartesian coordinates.
	 * 
	 * @return the homogeneous vector for this Cartesian vector
	 * @see #dehomogenize()
	 */
	public MathVector homogenize() {
		final MathVector ret = MathVector.zero(this.getLength() + 1);
		for (int i = 0; i < ret.coord.length - 1; i++) {
			ret.coord[i] = this.coord[i];
		}
		ret.coord[ret.coord.length - 1] = 1.0;
		return ret;
	}

	/**
	 * Creates Cartesian coordinates from <a href=
	 * "http://de.wikipedia.org/wiki/Homogene_Koordinaten#Homogene_Matrizen">homogeneous
	 * coordinates</a>.
	 * 
	 * @return the Cartesian vector for this homogeneous vector
	 * @see #homogenize()
	 */
	public MathVector dehomogenize() {
		final MathVector ret = MathVector.zero(this.getLength() - 1);
		final double distance = this.coord[this.coord.length - 1];
		for (int i = 0; i < ret.coord.length; i++) {
			ret.coord[i] = this.coord[i] / distance;
		}
		return ret;
	}

	/**
	 * @param o
	 *            the object to compare with
	 * @return if both objects are math vectors, and if they have the same
	 *         dimension and all coordinates are equal
	 */
	@Override
	public boolean equals(final Object o) {
		if (o instanceof MathVector) {
			final MathVector other = (MathVector)o;
//			if (coord.length == other.coord.length) {
//				for (int i = 0; i < coord.length; i++) {
//					if (coord[i] != other.coord[i]) {
//						return false;
//					}
//				}
//				return true;
//			}
			return Arrays.equals(coord, other.coord);
		}
		return false;
	}

	@SuppressWarnings("unused")
	@Override
	public int hashCode() {
		if ((IS_PRECALCING_HASH) && (isHashCodeValid)) {
			// Nothing to do
		} else {
			hashCode = Arrays.hashCode(coord);
		}
		return hashCode;
	}

	/**
	 * Compares two vectors by radius.
	 * 
	 * @param other
	 *            the vector to compare with
	 * @return a positive number if <em>this</em> has a larger radius than the
	 *         other; zero for equal radius size; a negative number if
	 *         <em>this</em> has a smaller radius than the other
	 * @see #getRadius()
	 */
	@Override
	public int compareTo(final MathVector other) {
		return (int)Math.signum(getRadiusSquare() - other.getRadiusSquare());
	}

	/** Keeps the coordinates. */
	protected double coord[];

	protected int hashCode = 0;
	protected boolean isHashCodeValid = false;
}
