package de.ped.tools;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.net.URLConnection;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.PropertyResourceBundle;
import java.util.ResourceBundle;
import java.util.ResourceBundle.Control;

/**
 * Defines I18N'ed key/value string properties through a resource bundle. 
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Messages {

	/**
	 * Helps to define if the internationalized value should be returned, or if
	 * the value should be the same for all languages (e.g., for logging or
	 * debugging purposes).
	 * 
	 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum I18NedOrNot {
		/** Internationalized */
		I18N,
		/** Non-internationalized, e.g., for logging or debugging purposes */
		NonI18N,
	}

	/**
	 * Custom <code>ResourceBundle.Control</code> implementation wherein you
	 * explicitly read the properties files as UTF-8 using
	 * <code>InputStreamReader</code>, so that you can just save them as UTF-8
	 * without the need to hassle with <code>native2ascii</code>.
	 * 
	 * Thanks to <a href=
	 * "https://stackoverflow.com/questions/4659929/how-to-use-utf-8-in-resource-properties-with-resourcebundle#4660195">stackoverflow</a>.
	 * 
	 * @author BalusC
	 * @author Copyright (c) 2023 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class CharsetControl extends Control {
		/** 
		 * Creates a control with default character set. 
		 * The default depends on the Java version and the OS.
		 * Up to Java 8, the default depends on the OS Java is developed on.
		 * Till Java 9, the default is UTF-8.
		 */
		public CharsetControl() {
			// Nothing to do
		}

		/**
		 * Creates a control with a defined character set with which to read the resource bundle files.
		 * @param charset the name of the character set; <code>null</code> for default character set.
		 */
		public CharsetControl(final String charset) {
			this.charset = charset;
		}

		/**
		 * @return the character set with which to read the resource bundle files; <code>null</code> for default character set.
		 */
		public String getCharset() {
			return charset;
		}

		/**
		 * Sets the character set with which to read the resource bundle files.
		 * @param charset the name of the character set; <code>null</code> for default character set
		 */
		public void setCharset(final String charset) {
			this.charset = charset;
		}

		@Override
		public ResourceBundle newBundle(final String baseName, final Locale locale, final String format,
				final ClassLoader loader, final boolean reload)
				throws IllegalAccessException, InstantiationException, IOException	    {
	        // The below is a copy of the default implementation.
	        final String bundleName = toBundleName(baseName, locale);
	        final String resourceName = toResourceName(bundleName, "properties");
	        ResourceBundle bundle = null;
	        InputStream stream = null;
	        if (reload) {
	        	final URL url = loader.getResource(resourceName);
	            if (url != null) {
	            	final URLConnection connection = url.openConnection();
	                if (connection != null) {
	                    connection.setUseCaches(false);
	                    stream = connection.getInputStream();
	                }
	            }
	        } else {
	            stream = loader.getResourceAsStream(resourceName);
	        }
	        if (stream != null) {
	            try {
	            	if (null == charset) {
		                bundle = new PropertyResourceBundle(new InputStreamReader(stream));
	            	} else {
		                // Only this line is changed to make it to read properties files as UTF-8.
		                bundle = new PropertyResourceBundle(new InputStreamReader(stream, charset));
	            	}
	            } finally {
	                stream.close();
	            }
	        }
	        return bundle;
	    }

		/**
		 * The character set with which to read the resource bundle files;
		 * <code>null</code> for default character set
		 */
		private String charset;
	}

	/** How names are postfixed in the properties file. */
	public static final String TEXT_POSTFIX = ".Text";

	/** How keystrokes are postfixed in the properties file. */
	public static final String KEYSTROKE_POSTFIX = ".KeyStroke";

	/** How descriptions (work for short and long description) are postfixed in the properties file. */
	public static final String DESCRIPTION_POSTFIX = ".Description";

	/** How short descriptions (often used as tool tips) are postfixed in the properties file. */
	public static final String SHORT_DESCRIPTION_POSTFIX = ".ShortDescription";

	/** 
	 * How long descriptions (often used in help texts and user manuals) are postfixed in the properties file.
	 * Any value for those should be xhtml formatted, e.g., at least embedded in an xhtml paragraph (<code>&lt;p&gt;</code>).
	 */
	public static final String LONG_DESCRIPTION_POSTFIX = ".LongDescription";

	/** How mnemonics are postfixed in the properties file. */
	public static final String MNEMONIC_POSTFIX = ".Mnemonic";

	public static final String APPLICATION_NAME_KEY = "Application" + TEXT_POSTFIX;
	public static final String APPLICATION_VERSION_KEY = "Application.Version";
	public static final String APPLICATION_COPYRIGHT_KEY = "Application.Copyright";
	public static final String APPLICATION_SHORTDESC_KEY = "Application" + SHORT_DESCRIPTION_POSTFIX;
	public static final String APPLICATION_DESCRIPTION_KEY = "Application" + DESCRIPTION_POSTFIX;


	/** Defines the locales that are defined for most applications. */
	public static final Locale[] DEFAULT_LOCALES = new Locale[] {
			new Locale("de", "DE"),
			new Locale("en", "US"),
	};

	/** Defines the locales that are defined for most applications. */
	public static final Locale[] DE_LOCALES = new Locale[] {
			DEFAULT_LOCALES[0],
	};

	/** Defines the index of the locales in {@link #DEFAULT_LOCALES} that will be the starting locale. */
	public static final int DEFAULT_LOCALE_INDEX = 1;

	public static final String DEFAULT_RESOURCE_BUNDLE_NAME = "de.ped.tools.messages";
	/** Only used for logging purposes, so should be language-independent. */
	public static final String SEPARATOR = "=====================================================================";

	/**
	 * Finds the best fitting locale within the given locales.
	 * 
	 * @param locales
	 *            The locales to search through
	 * @param localeDescription
	 *            the locale to find. See {@link #createLocale(String)} for
	 *            details.
	 * @param defaultLocaleIndex
	 *            the index to pick if none of the locales fit to the
	 *            localeDescription
	 * @return the index of the entry within locales identified as best fitting.
	 *         If <em>localeDescription</em> is <code>null</code>,
	 *         <em>defaultLocaleIndex</em> is returned.
	 */
	public static final int findLocaleIndex(final Locale[] locales, final String localeDescription, final int defaultLocaleIndex) {
		int best = defaultLocaleIndex;
		final Locale loc = createLocale(localeDescription);
		if (null != loc) {
			for (int i = 0; i < locales.length; i++) {
				if (localeEqualityLevel(locales[i], loc) > localeEqualityLevel(locales[i], locales[best])) {
					best = i;
				}
			}
		}
		return best;
	}

	public static int localeEqualityLevel(final Locale locale1, final Locale locale2) {
		int ret = 0;
		if (null == locale1) {
			if (null == locale2) {
				ret = 1;
			} 
		} else if (null != locale2) {
			if (locale1.getLanguage().equalsIgnoreCase(locale2.getLanguage())) {
				ret++;
				if (locale1.getCountry().equalsIgnoreCase(locale2.getCountry())) {
					ret++;
					if (locale1.getVariant().equalsIgnoreCase(locale2.getVariant())) {
						ret++;
					}
				}
			}
		}
		return ret;
	}

	/**
	 * Creates a locale fitting to the locale description.
	 * 
	 * @param localeDescription
	 *            the locale to find, given as String of format
	 *            <em>&lt;language&gt;</em> or
	 *            <em>&lt;language&gt;</em><code>_</code><em>&lt;country&gt;</em>
	 *            or
	 *            <em>&lt;language&gt;</em><code>_</code><em>&lt;country&gt;</em><code>_</code><em>&lt;variant&gt;</em>
	 * @return the appropriate locale, or <code>null</code> if the locale
	 *         description was <code>null</code>
	 */
	public static Locale createLocale(final String localeDescription) {
		Locale ret = null;
		if (null != localeDescription) {
			final String[] localeParts = localeDescription.split("_");
			switch(localeParts.length) {
			case 1:
				ret = new Locale(localeParts[0]);
				break;
			case 2:
				ret = new Locale(localeParts[0], localeParts[1]);
				break;
			case 3:
			default:
				ret = new Locale(localeParts[0], localeParts[1], localeParts[3]);
			}
		}
		return ret;
	}

	/** The location of the properties file containing the key/value pairs */
	private String resourceBundleName;
	/** The properties bundle containing the key/value pairs */
	private ResourceBundle resourceBundle;
	/** The default properties bundle containing the key/value pairs */
	private ResourceBundle defaultResourceBundle;
	/** The current locale selected */
	private Locale currentLocale;
	/** All locales available */
	private Locale[] availableLocales;

	/** Knows the non-I18N'ed name of the application. */
	private String applicationName = getUnknownString(APPLICATION_NAME_KEY);
	/** Knows the non-I18N'ed version string of the application. */
	private String applicationVersion = getUnknownString(APPLICATION_VERSION_KEY);
	/** Knows the non-I18N'ed copyright message for application. */
	private String applicationCopyrightMessage = getUnknownString(APPLICATION_COPYRIGHT_KEY);
	/** Knows the non-I18N'ed short description message for application. */
	private String applicationShortDescription = getUnknownString(APPLICATION_SHORTDESC_KEY);
	/** Knows the non-I18N'ed description message for application. */
	private String applicationDescription = getUnknownString(APPLICATION_DESCRIPTION_KEY);

	public Messages() {
		// Nothing to do
	}

	public Messages(final String resourceBundleName) {
		setResourceBundleName(resourceBundleName);
	}

	public Messages(
			final String resourceBundleName, 
			final String appName, 
			final String appVersion, 
			final String appCopyrightMessage,
			final String appShortDescription,
			final String appDescription) {
		setResourceBundleName(resourceBundleName);
		setApplicationStrings(appName, appVersion, appCopyrightMessage, appShortDescription, appDescription);
	}

	public Messages(
			final String resourceBundleName, 
			final String charset,
			final String appName, 
			final String appVersion, 
			final String appCopyrightMessage,
			final String appShortDescription,
			final String appDescription) {
		setCharset(charset);
		setResourceBundleName(resourceBundleName);
		setApplicationStrings(appName, appVersion, appCopyrightMessage, appShortDescription, appDescription);
	}

	private Messages(
			final String resourceBundleName, 
			final String appName, 
			final String appVersion, 
			final String appCopyrightMessage, 
			final String appShortDescription,
			final String appDescription,
			final Locale[] availableLocales, 
			final Locale currentLocale) {
		this.resourceBundleName = resourceBundleName;
		this.currentLocale = currentLocale;
		this.availableLocales = availableLocales;
		loadBundles();
		setApplicationStrings(appName, appVersion, appCopyrightMessage, appShortDescription, appDescription);
	}

	public Messages(
			final String resourceBundleName, 
			final String appName, 
			final String appVersion, 
			final String appCopyrightMessage, 
			final String appShortDescription,
			final String appDescription,
			final Locale[] availableLocales, 
			final String preferredLocaleDescription, 
			final int defaultLocaleIndex) {
		this(resourceBundleName, appName, appVersion, appCopyrightMessage, null, null, 
				availableLocales, 
				availableLocales[findLocaleIndex(availableLocales, preferredLocaleDescription, defaultLocaleIndex)]);
	}

	public String getResourceBundleName() {
		return resourceBundleName;
	}

	/**
	 * Sets the resource bundle name (like
	 * <code>de.ped.gedicht.gui.messages</code>) and tries to load the resource
	 * bundle. Expects a file with the ending <code>.properties</code> at that
	 * location.
	 * 
	 * @param name
	 *            the resource bundle name
	 */
	public void setResourceBundleName(final String name) {
		resourceBundleName = name;
		loadBundles();
	}

	public Locale getCurrentLocale() {
		return currentLocale;
	}

	public void setCurrentLocale(final Locale locale) {
		currentLocale = locale;
		loadBundles();
	}

	public Locale[] getAvailableLocales() {
		return availableLocales;
	}

	public void setAvailableLocales(final Locale[] locales) {
		availableLocales = locales;
	}

	/**
	 * @return the character set with which to read the resource bundle files; <code>null</code> for default character set.
	 */
	public String getCharset() {
		return utf8control.getCharset();
	}

	/**
	 * Sets the character set with which to read the resource bundle files.
	 * @param charset the name of the character set; <code>null</code> for default character set.
	 */
	public void setCharset(final String charset) {
		utf8control.setCharset(charset);
	}

	protected void loadBundles() {
		if (null == currentLocale) {
			defaultResourceBundle = ResourceBundle.getBundle(DEFAULT_RESOURCE_BUNDLE_NAME, utf8control);
			resourceBundle = ResourceBundle.getBundle(resourceBundleName, utf8control);
		} else {
			defaultResourceBundle = ResourceBundle.getBundle(DEFAULT_RESOURCE_BUNDLE_NAME, currentLocale, utf8control);
			resourceBundle = ResourceBundle.getBundle(resourceBundleName, currentLocale, utf8control);
		}
	}

	private CharsetControl utf8control = new CharsetControl();

	/**
	 * Defines what to show when the string was not defined in the resource
	 * bundle. Overload for useful behavior.
	 * 
	 * @param key
	 *            the message key
	 * @return the message key, preceeded and succeeded by an exclamation mark
	 */
	public String getUnknownString(final String key) {
		return '!' + key + '!';
	}

	/**
	 * Finds the value to the key in the properties file.
	 * 
	 * @param key
	 *            the key
	 * @return the value to the key
	 */
	public String getString(final String key) {
		return getString(key, getUnknownString(key));
	}

	/**
	 * Finds the value to the key in the properties file.
	 * 
	 * @param key
	 *            the key to find
	 * @param alternative
	 *            the string to return if key not found
	 * @return the value to the key, or <em>alternative</em>
	 */
	public String getString(final String key, final String alternative) {
		String ret = alternative;
		ResourceBundle[] arr = { resourceBundle, defaultResourceBundle };
		boolean found = false;
		for (ResourceBundle bundle : arr) {
			try {
				ret = bundle.getString(key);
				found = true;
			} catch (final MissingResourceException e) {
				// Nothing to do
			} catch (final NullPointerException e) {
				found = true;
			} catch (final ClassCastException e) {
				// Nothing to do
			}
			if (found) {
				break;
			}
		}
		return ret;
	}

	/**
	 * Finds the value to the key in the properties file, and replaces
	 * placeholders by real texts.
	 * 
	 * @param key
	 *            the key to find
	 * @param alternative
	 *            the string to return if key not found
	 * @param fillIns
	 *            An array of non- or pre-i18n'ed texts to fill into the
	 *            template. Currently only a size up to 10 is supported.
	 * @return the value to the key, or <em>alternative</em>
	 * @see #parseString(String, Object[])
	 * @see #getString(String, String)
	 */
	public String getString(final String key, final String alternative, final Object[] fillIns) {
		return parseString(getString(key, alternative), fillIns);
	}

	public String getString(final I18NStringWithFillIns text) {
		return getString(text.getKey(), text.getAlternative(), text.getFillIns());
	}

	/**
	 * In case a value for the <em>key</em> exists, just behaves like {@link #getString(String, String, Object[])}.
	 * However, if not, it looks for alternative keys:
	 * 
	 * <ul>
	 *   <li>For keys ending with {@link #TEXT_POSTFIX}, 
	 *   		it also looks for a key without that postfix.</li>
	 *   <li>For keys ending with {@link #SHORT_DESCRIPTION_POSTFIX},
	 *   		it also looks for the {@link #DESCRIPTION_POSTFIX}.
	 *   		If still nothing found, it looks for the {@link #LONG_DESCRIPTION_POSTFIX}
	 *   		and if found there, tries to strip result from HTML to plain text.</li>
	 *   <li>For keys ending with {@link #LONG_DESCRIPTION_POSTFIX},
	 *   		it also looks for the {@link #DESCRIPTION_POSTFIX} or {@link #SHORT_DESCRIPTION_POSTFIX},
	 *   		and if found there, embeds the result in an xhtml paragraph (<code>&lt;p&gt;</code>).</li>
	 * </ul>
	 * 
	 * @param key
	 *            the key to find
	 * @param alternative
	 *            the string to return if key not found
	 * @param fillIns
	 *            An array of non- or pre-i18n'ed texts to fill into the
	 *            template. Currently only a size up to 10 is supported.
	 * @return the value to the key, or <em>alternative</em>
	 */
	public String getStringOr(final String key, final String alternative, final Object[] fillIns) {
		String ret = null;
		final int dotPos = key.lastIndexOf(".");
		final String mypostfix;
		final String nakedKey;
		if (0 <= dotPos) {
			nakedKey = key.substring(0, dotPos);
			mypostfix = key.substring(dotPos);
		} else {
			nakedKey = key;
			mypostfix = "";
		}
		String keyToUse = key;
		String val = getString(keyToUse, null, null);
		if (TEXT_POSTFIX.equals(mypostfix)) {
			if (null == val) {
				keyToUse = nakedKey;
			}
		} else if (SHORT_DESCRIPTION_POSTFIX.equals(mypostfix)) {
			if (null == val) {
				keyToUse = nakedKey + DESCRIPTION_POSTFIX;
				val = getString(keyToUse, null, null);
				if (null == val) {
					keyToUse = nakedKey + LONG_DESCRIPTION_POSTFIX;
					ret = getString(keyToUse, alternative, fillIns);
					if (null != ret) {
						ret = TextFormatter.toPlainText(TextDocumentType.XHTML10_STRICT, ret);
					}
				}
			}
		} else if (LONG_DESCRIPTION_POSTFIX.equals(mypostfix)) {
			if (null == val) {
				keyToUse = nakedKey + DESCRIPTION_POSTFIX;
				val = getString(keyToUse, null, null);
				if (null == val) {
					keyToUse = nakedKey + SHORT_DESCRIPTION_POSTFIX;
				}
				ret = getString(keyToUse, alternative, fillIns);
				if (null != ret) {
					ret = "<p>" + TextFormatter.sgmlify(TextDocumentType.XHTML10_STRICT, ret) + "</p>";
				}
			}
		}
		if (null == ret) {
			ret = getString(keyToUse, alternative, fillIns);
		}
		return ret;
	}

	public String getStringOr(final I18NStringWithFillIns text) {
		return getStringOr(text.getKey(), text.getAlternative(), text.getFillIns());
	}

	/**
	 * @param key
	 *            the key to find
	 * @return the text for the given key
	 */
	public String getText(final String key) {
		return getStringOr(key + TEXT_POSTFIX, null, null);
	}

	/**
	 * @param key
	 *            the key to find
	 * @param alternative
	 *            the string to return if key not found
	 * @return the text for the given key, or the alternative if not found
	 */
	public String getText(final String key, final String alternative) {
		return getStringOr(key + TEXT_POSTFIX, alternative, null);
	}

	public String getText(final String key, final String alternative, final Object[] fillIns) {
		return getStringOr(key + TEXT_POSTFIX, alternative, fillIns);
	}

	public String getText(final I18NStringWithFillIns text) {
		return getText(text.getKey(), text.getAlternative(), text.getFillIns());
	}

	/**
	 * @param key
	 *            the key to find
	 * @return the keystroke of the given key
	 */
	public String getKeyStroke(final String key) {
		return getStringOr(key + KEYSTROKE_POSTFIX, null, null);
	}

	/**
	 * @param key
	 *            the key to find
	 * @return the short description for the given key. It should be usable for
	 *         tool tips for example. So it should contain only plain text.
	 */
	public String getShortDescription(final String key) {
		return getStringOr(key + SHORT_DESCRIPTION_POSTFIX, null, null);
	}

	/**
	 * @param i18n
	 *            the internationalized string object
	 * @return the short description for the given key. It should be usable for
	 *         tool tips for example. So it should contain only plain text.
	 */
	public String getShortDescription(I18NStringWithFillIns i18n) {
		return getStringOr(i18n.getKey() + SHORT_DESCRIPTION_POSTFIX, i18n.getAlternative(), i18n.getFillIns());
	}

	/**
	 * @param key
	 *            the key to find
	 * @return the long description for the given key. The return value should
	 *         be valid HTML and XHTML directly inside the
	 *         <code>&lt;body&gt;</code> tag, for example embrace plain text
	 *         within the <code>&lt;p&gt;</code> tag. This implementation looks
	 *         if a long description exists in the resource bundle. If not, it
	 *         looks for a short description and embraces that within a
	 *         <code>&lt;p&gt;</code> tag.
	 */
	public String getLongDescription(final String key) {
		return getStringOr(key + LONG_DESCRIPTION_POSTFIX, null, null);
	}

	/**
	 * @param i18n
	 *            the internationalized string object
	 * @return the long description for the given key. The return value should
	 *         be valid HTML and XHTML directly inside the
	 *         <code>&lt;body&gt;</code> tag, for example embrace plain text
	 *         within the <code>&lt;p&gt;</code> tag. This implementation looks
	 *         if a long description exists in the resource bundle. If not, it
	 *         looks for a short description and embraces that within a
	 *         <code>&lt;p&gt;</code> tag.
	 */
	public String getLongDescription(I18NStringWithFillIns i18n) {
		return getStringOr(i18n.getKey() + LONG_DESCRIPTION_POSTFIX, i18n.getAlternative(), i18n.getFillIns());
	}

	/**
	 * @param key
	 *            the key to find
	 * @return the mnemonic for the given key
	 */
	public String getMnemonic(final String key) {
		return getStringOr(key + MNEMONIC_POSTFIX, null, null);
	}

	/**
	 * @param i18ned
	 *            if the internationalized value should be returned, or if the
	 *            value should be the same for all languages
	 * @return the name of the application
	 */
	public String getApplicationName(final I18NedOrNot i18ned) {
		final String ret;
		switch (i18ned) {
		case NonI18N:
			ret = applicationName;
			break;
		case I18N:
		default:
			ret = getString(APPLICATION_NAME_KEY, applicationName);
		}
		return ret;
	}

	/**
	 * Sets the name of the application.
	 * 
	 * @param applicationName
	 *            the non-I18N'ed name of the application
	 */
	public void setApplicationName(final String applicationName) {
		this.applicationName = applicationName;
	}

	/** 
	 * @return the version string of the application
	 */
	public String getApplicationVersion() {
		return getString(APPLICATION_VERSION_KEY, applicationVersion);
	}

	/**
	 * Sets the version string of the application.
	 * 
	 * @param applicationVersion
	 *            the non-I18N'ed version string of the application
	 */
	public void setApplicationVersion(final String applicationVersion) {
		this.applicationVersion = applicationVersion;
	}

	/** 
	 * @return the copyright message for the application. 
	 */
	public String getApplicationCopyrightMessage() {
		return getString(APPLICATION_COPYRIGHT_KEY, applicationCopyrightMessage);
	}

	/**
	 * Sets the non-I18N'ed copyright message of the application.
	 * 
	 * @param applicationCopyrightMessage
	 *            the non-I18N'ed copyright message of the application
	 */
	public void setApplicationCopyrightMessage(final String applicationCopyrightMessage) {
		this.applicationCopyrightMessage = applicationCopyrightMessage;
	}

	/** 
	 * @return the short description for the application. 
	 */
	public String getApplicationShortDescription() {
		return getString(APPLICATION_SHORTDESC_KEY, applicationShortDescription);
	}

	/**
	 * Sets the non-I18N'ed short description of the application.
	 * 
	 * @param applicationShortDescription
	 *            the non-I18N'ed short description of the application
	 */
	public void setApplicationShortDescription(final String applicationShortDescription) {
		this.applicationShortDescription = applicationShortDescription;
	}

	/** 
	 * @return the description for the application. 
	 */
	public String getApplicationDescription() {
		return getString(APPLICATION_DESCRIPTION_KEY, applicationDescription);
	}

	/**
	 * Sets the non-I18N'ed description of the application.
	 * 
	 * @param applicationDescription
	 *            the non-I18N'ed description of the application
	 */
	public void setApplicationDescription(final String applicationDescription) {
		this.applicationDescription = applicationDescription;
	}

	public void setApplicationStrings(
			final String appName, 
			final String appVersion, 
			final String appCopyrightMessage,
			final String appShortDescription,
			final String appDescription) {
		setApplicationName(appName);
		setApplicationVersion(appVersion);
		setApplicationCopyrightMessage(appCopyrightMessage);
		setApplicationShortDescription(appShortDescription);
		setApplicationDescription(appDescription);
	}

	/**
	 * @return the title for the application. Default is a concatenation of the
	 *         application name and the version.
	 */
	public String getApplicationTitle() {
		return getApplicationName(I18NedOrNot.I18N) + " " + getApplicationVersion();
	}

	/**
	 * @return the long title of the application. Default is a concatenation of
	 *         the application name, the version and (if available) the short
	 *         description.
	 */
	public String getApplicationTitleLong() {
		final String sd = getApplicationShortDescription();
		return getApplicationTitle() + (CollectionUtil.isNullOrEmpty(sd) ? "" : " - " + sd);
	}

	/**
	 * Parses the template and performs the following operations:
	 * <ol>
	 * <li>Where template contains the substring '%<em>x</em>' (with <em>x</em> being a number from 0 to the length of fill-ins - 1), fills in the fill-in with index <em>x</em>.
	 * <li>Where template contains the substring '%' followed by another message key, this is replaced by the value for that message key.
	 * <li>Where template contains the substring '%%', this is replaced by a single '%'.
	 * </ol>
	 * @param template
	 *            the string to parse
	 * @param fillins
	 *            An array of non- or pre-i18n'ed texts to fill into the template. 
	 * @return the transformed string
	 */
	public String parseString(final String template, final Object[] fillins) {
		String ret = template;
		if (null != template) {
			// Handle %i for fillins
			if (null != fillins) {
				for (int i = fillins.length - 1; 0 <= i; i--) {
					final Object fillin = fillins[i];
					if (null != fillin) {
						String replacement;
						if (fillin instanceof I18NStringWithFillIns) {
							replacement = getText((I18NStringWithFillIns) fillin);
						} else {
							replacement = fillin.toString();
						}
						if (null == replacement) {
							throw new IllegalArgumentException("fillin " + i + " resulted in null: " + fillin);
						}
						ret = ret.replace("%" + i, replacement);
					}
				}
			}
			int pos = ret.length();
			// Handle %key for key redirection
			while (true) {
				pos = ret.lastIndexOf("%", pos);
				if (-1 == pos) {
					break;
				}
				if ((0 < pos) && ('%' == ret.charAt(pos - 1))) {
					// Ignore %% for now, will be handled later
					pos -= 2;
				} else {
					int endpos = pos + 1;
					while (endpos < ret.length()) {
						final char c = ret.charAt(endpos);
						if (('.' == c) || ('_' == c)
								|| ((endpos == pos + 1) ? Character.isLetter(c) : Character.isLetterOrDigit(c))) {
							endpos++;
						} else {
							break;
						}
					}
					String key = ret.substring(pos + 1, endpos);
					if (!key.isEmpty()) {
						ret = ret.replace("%" + key, getString(key, getUnknownString(key), fillins));
					} else {
						pos--;
					}
				}
			}
			// Handle %%
			ret = ret.replace("%%", "%");
		}
		return ret;
	}
}
