package de.ped.tools;

import java.awt.Point;
import java.io.Serializable;

/**
 * Simple data type, represents the position of a playfield on the board.
 * In fact, it's a tuple of 2 integers.
 * Needed internally to uniquely identify playfields.
 * The current implementation assumes that both coordinates are within the range [0, 255], 
 * as only this range will be marshalled and demarshalled. 
 * 
 * @author Copyright (c) 2001 Peter Diefenbach
 */
public class PlayfieldPosition implements Cloneable, Comparable<PlayfieldPosition>, Serializable, Marshallable {
	private static final long serialVersionUID = 1L;
	/** The number of coordinates to define a position on the map */
	public static final int COORDS = 2;
	/** The index of the horizontal ordinate in the coordinate array */
	public static final int X = 0;
	/** The index of the vertical ordinate in the coordinate array */
	public static final int Y = 1;

	public static final int BITS = 8;
	public static final BitArea BA_X = new BitArea(    0, BITS);
	public static final BitArea BA_Y = new BitArea( BITS, BITS);
	public static final BitArea BA_NULL = new BitArea(2 * BITS, 1);

	/** the upper limits (that means, the highest legal value + 1) for both dimensions. */
	public static final PlayfieldDimension BOUNDS = new PlayfieldDimension(1 << BITS, 1 << BITS);

	public static void marshalWithNull(final Marshaller marshaller, final PlayfieldPosition that) {
		int pos = 0x00;
		if (null == that) {
			pos = Bitfield32.setBits(pos, BA_NULL, 1);
		} else {
			that.assertBounds();
			pos = Bitfield32.setBits(pos, BA_X, that.x);
			pos = Bitfield32.setBits(pos, BA_Y, that.y);
		}
		marshaller.writeInt(pos);
	}

	public static PlayfieldPosition demarshalWithNull(final Marshaller marshaller) {
		final int pos = marshaller.readInt();
		final PlayfieldPosition that;
		if (0 != Bitfield32.getBits(pos, BA_NULL)) {
			that = null;
		} else {
			final int x = Bitfield32.getBits(pos, BA_X);
			final int y = Bitfield32.getBits(pos, BA_Y);
			that = new PlayfieldPosition(x, y);
			that.assertBounds();
		}
		return that;
	}

	/** 
	 * Creates a playfield position (0, 0). 
	 */
	public PlayfieldPosition() {
		// Nothing to do
	}

	/**
	 * Creates a playfield position (x, y).
	 * @param x the x-coordinate
	 * @param y the y-coordinate
	 */
	public PlayfieldPosition(final int x, final int y) {
		this.x = x;
		this.y = y;
	}

	/**
	 * Copies a playfield position.
	 * @param other the position to clone
	 */
	public PlayfieldPosition(final PlayfieldPosition other) {
		this.x = other.x;
		this.y = other.y;
	}

	public PlayfieldPosition(final Point point) {
		this.x = point.x;
		this.y = point.y;
	}

	public PlayfieldPosition(final int[] arr) {
		setLocation(arr);
	}

	public boolean isWithin(final PlayfieldDimension bounds) {
		return ((0 <= x) && (x < bounds.getWidth()) && (0 <= y) && (y < bounds.getHeight()));
	}

	/**
	 * @param index One of {@link #X} or {@link #Y}
	 * @return the x or y value
	 */
	public int get(final int index) {
		final int ret;
		switch (index) {
		case 0:
			ret = x;
			break;
		case 1:
			ret = y;
			break;
		default:
			throw new IndexOutOfBoundsException("Only [0,1] accepted, " + index + " is out of range");
		}
		return ret;
	}

	/**
	 * @param index One of {@link #X} or {@link #Y}
	 * @param value the x or y value to set
	 */
	public void set(final int index, final int value) {
		switch (index) {
		case 0:
			x = value;
			break;
		case 1:
			y = value;
			break;
		default:
			throw new IndexOutOfBoundsException("Only [0,1] accepted, " + index + " is out of range");
		}
	}

	public int getX() {
		return x;
	}

	public void setX(final int x) {
		this.x = x;
	}

	public int getY() {
		return y;
	}

	public void setY(final int y) {
		this.y = y;
	}

	public void setLocation(final int x, final int y) {
		this.x = x;
		this.y = y;
	}

	public void setLocation(final PlayfieldPosition that) {
		this.x = that.x;
		this.y = that.y;
	}

	public void setLocation(final Point point) {
		this.x = point.x;
		this.y = point.y;
	}

	public void setLocation(final int[] arr) {
		this.x = arr[X];
		this.y = arr[Y];
	}

	public int[] toArray(final int[] arr) {
		if (null == arr) {
			return new int[] { x, y, };
		} else {
			arr[0] = x;
			arr[1] = y;
			return arr;
		}
	}

	public void assertBounds() {
		Assert.assertTrue(isWithin(BOUNDS));
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		assertBounds();
		int pos = 0x00;
		pos = Bitfield32.setBits(pos, BA_X, x);
		pos = Bitfield32.setBits(pos, BA_Y, y);
		marshaller.writeInt(pos);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		final int pos = marshaller.readInt();
		x = Bitfield32.getBits(pos, BA_X);
		y = Bitfield32.getBits(pos, BA_Y);
		assertBounds();
		return this;
	}

	@Override
	public String toString() {
		return "(" + x + "," + y + ")";
	}

	@Override
	public PlayfieldPosition clone() {
		return new PlayfieldPosition(this);
	}

	@Override
	public boolean equals(final Object o) {
		if (o instanceof PlayfieldPosition) {
			final PlayfieldPosition other = (PlayfieldPosition) o;
			return ((x == other.x) && (y == other.y));
		} else {
			return false;
		}
	}

	@Override
	public int compareTo(final PlayfieldPosition other) {
		if (y != other.y) {
			return y - other.y;
		}
		return x - other.x;
	}

	@Override
	public int hashCode() {
		return ((y << 16) | (x & 0xffff));
	}

	public static boolean equals(final PlayfieldPosition that, final PlayfieldPosition other) {
		return that.equals(other);
	}

	public void addInplace(final PlayfieldDistance dist) {
		x += dist.getX();
		y += dist.getY();
	}

	public void addInplace(final PlayfieldPosition pos, final PlayfieldDistance dist) {
		x = pos.x + dist.getX();
		y = pos.y + dist.getY();
	}

	public PlayfieldPosition add(final PlayfieldDistance dist) {
		final PlayfieldPosition ret = new PlayfieldPosition(this);
		ret.addInplace(dist);
		return ret;
	}

	public void substractInplace(final PlayfieldDistance dist) {
		x -= dist.getX();
		y -= dist.getY();
	}

	public void substractInplace(final PlayfieldPosition pos, final PlayfieldDistance dist) {
		x = pos.x - dist.getX();
		y = pos.y - dist.getY();
	}

	public PlayfieldPosition substract(final PlayfieldDistance dist) {
		final PlayfieldPosition ret = new PlayfieldPosition(this);
		ret.substractInplace(dist);
		return ret;
	}

	/**
	 * @param start the start position
	 * @param end the end position
	 * @return the vector from start to end position
	 */
	public static PlayfieldDistance distance(final PlayfieldPosition start, final PlayfieldPosition end) {
		return new PlayfieldDistance(end.getX() - start.getX(), end.getY() - start.getY());
	}

	private int x = 0;
	private int y = 0;
}
