package de.ped.tools;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.geom.Rectangle2D;
import java.io.Serializable;
import java.util.ArrayList;

/**
 * Bundling all the knowledge how the playfields on a board are tiled.
 * 
 * @author Copyright (c) 2020 Peter Diefenbach
 */
public class PlayfieldTiling implements Cloneable, Comparable<PlayfieldTiling>, Serializable, Marshallable {
	private static final long serialVersionUID = 1L;

	/** the top left corner of the surrounding rectangle of a playfield tile is meant */
	public static final boolean TOP_LEFT = false;
	/** the center of a playfield tile is meant */
	public static final boolean CENTER = true;

	/** 
	 * implementation switch: 
	 * <ul>
	 * <li>If <code>false</code>, hexagonal topology is implemented linearly, but more memory consumptive. 
	 * (Here, {@link TopologyMode#HEXAGON} is listed; {@link TopologyMode#HEXAGON_H} is handled the like.
	 * <pre>
	 *        (1, 0)        (3, 1)        (5, 2)        (7, 3)
	 * (0, 0)        (2, 1)        (4, 2)        (6, 3)
	 *        (1, 1)        (3, 2)        (5, 3)        (7, 4)
	 * (0, 1)        (2, 2)        (4, 3)        (6, 4)
	 *        (1, 2)        (3, 3)        (5, 4)        (7, 5)
	 * (0, 2)        (2, 3)        (4, 4)        (6, 5)
	 * </pre>
	 * <li>If <code>true</code>, it is implemented non-linearly, but less memory consumptive: 
	 * (Again, {@link TopologyMode#HEXAGON} is listed; {@link TopologyMode#HEXAGON_H} is handled the like.
	 * <pre>
	 *        (1, 0)        (3, 0)        (5, 0)        (7, 0)
	 * (0, 0)        (2, 0)        (4, 0)        (6, 0)
	 *        (1, 1)        (3, 1)        (5, 1)        (7, 1)
	 * (0, 1)        (2, 1)        (4, 1)        (6, 1)
	 *        (1, 2)        (3, 2)        (5, 2)        (7, 2)
	 * (0, 2)        (2, 2)        (4, 2)        (6, 2)
	 * </pre>
	 * </ul>
	 */
	public static final boolean IS_HEX_COMPACT = true;

	/** Position type for topologies where neighbor vectors are depending on the position, e.g., hexagonal in compact form */
	private static final int HEX_COMPACT_ODD = 2;
	/** Position type for topologies where neighbor vectors are depending on the position, e.g., hexagonal in compact form */
	private static final int HEX_COMPACT_EVEN = 1;
	/** Position type for topologies where neighbor vectors are invariant to the position */
	private static final int NONE = 0;

	/** 
	 * Calculates the air distance absolute value based on a distance vector.
	 * Air distance means that obstacles are not taken into account.
	 * The result is always non-negative and depends on the chosen topology mode.
	 * @param topologyMode the topologyMode
	 * @param startPos starting position of the vector
	 * @param distanceVector the distance vector
	 * @return the air distance between the positions
	 */
	public static int calcAirDistanceScalar(final TopologyMode topologyMode, final PlayfieldPosition startPos, final PlayfieldPosition distanceVector) {
		final int ret;
		switch(topologyMode) {
		case SQUARE8:
			ret = Math.max(Math.abs(distanceVector.getX()), Math.abs(distanceVector.getY()));
			break;
		case SQUARE4:
			ret = Math.abs(distanceVector.getX()) + Math.abs(distanceVector.getY());
			break;
		case HEXAGON: {
			final int xdist = distanceVector.getX();
			final int ydist = renormHexY(xdist, distanceVector.getY(), getPositionType(topologyMode, startPos), 1);
			if (MathUtil.sgn(xdist) * MathUtil.sgn(ydist) > 0) {
				ret = Math.max(Math.abs(xdist), Math.abs(ydist));
			} else {
				ret = Math.abs(xdist) + Math.abs(ydist);
			}
			break;
		}
		case HEXAGON_H: {
			final int ydist = distanceVector.getY();
			final int xdist = renormHexX(distanceVector.getX(), ydist, getPositionType(topologyMode, startPos), 1);
			if (MathUtil.sgn(xdist) * MathUtil.sgn(ydist) > 0) {
				ret = Math.max(Math.abs(xdist), Math.abs(ydist));
			} else {
				ret = Math.abs(xdist) + Math.abs(ydist);
			}
			break;
		}
		case TRIANGLE:
			int dx = distanceVector.getX();
			int dy = distanceVector.getY();
			// if ((xd == 0) && (yd == 0)) { //0: Nullpunkt, Spezialfall von 1
			// return 0;
			// }
			// if (dy == 0) { //1: Horizontal, Spezialfall von 3
			// return Math.abs(dx);
			// }
			// if (Math.abs(dx) == Math.abs(dy)) { //2: Diagonal, Spezialfall von 3
			// return (2 * Math.abs(dy));
			// }
			if (Math.abs(dx) >= Math.abs(dy)) { // 3: Diagonal und links und rechts davon
				ret = (Math.abs(dy) + Math.abs(dx));
			} else {
				// Erst mal die Diagonale Richtung Zielpunkt gehen, bis über oder unter dem anderen Punkt.
				int dist = 2 * Math.abs(dx);
				if (0 < dy) {
					dy = dy - Math.abs(dx);
				} else {
					dy = dy + Math.abs(dx);
				}
				dx = 0;
				// Nun in Zweiersprüngen ans Ziel
				dist += 4 * (Math.abs(dy)/2); // Hier ist das Abrunden bei der Integer-Division wichtig
				if (0 < dy) {
					dy = dy % 2;
				} else {
					dy = -((-dy) % 2);
				}
				// Nun sind wir entweder schon da, oder nur noch einen Schritt entfernt
				if (0 != dy) {
					dist += 1;
					if ((startPos.getY() + startPos.getX()) % 2 == 0) { //Dreieck \/
						dist += ((dy > 0) ? 2 : 0);
					} else { // Dreieck /\
						dist += ((dy < 0) ? 2 : 0);
					}
				}
				ret = dist;
			}
			break;
		default:
			throw new IllegalStateException("Not implemented topology " + topologyMode);
		}
		return ret;
	}

	public static int renormHexY(final int xdist, final int ydist, final int positionType, final int sgn) {
		int ret = ydist;
		switch(positionType) {
		case HEX_COMPACT_EVEN:
			ret += sgn * MathUtil.divisio(xdist, 2);
			break;
		case HEX_COMPACT_ODD:
			ret += sgn * MathUtil.divisio(xdist + 1, 2);
			break;
		default:
			// Nothing to do
		}
		return ret;
	}

	public static int renormHexX(final int xdist, final int ydist, final int positionType, final int sgn) {
		int ret = xdist;
		switch(positionType) {
		case HEX_COMPACT_EVEN:
			ret += sgn * MathUtil.divisio(ydist, 2);
			break;
		case HEX_COMPACT_ODD:
			ret += sgn * MathUtil.divisio(ydist + 1, 2);
			break;
		default:
			// Nothing to do
		}
		return ret;
	}

	public static final ArrayList<Polygon> getPolys() {
		return PlayfieldTiling.hack;
	}

	private static ArrayList<Polygon> hack = new ArrayList<Polygon>();

	public PlayfieldTiling() {
		this(TopologyMode.SQUARE8, TopologyMode.SQUARE8, 0, 0, 1, 1, 1);
	}

	public PlayfieldTiling(final TopologyMode positionTopologyMode, final TopologyMode shapeTopologyMode, final int sw,
			final int sh, final int ratioX, final int ratioY, final int stretchStep) {
		this.positionTopologyMode = positionTopologyMode;
		this.shapeTopologyMode = shapeTopologyMode;
		this.space = new Dimension();
		this.surroundingBox = new Dimension();
		setSpace(sw, sh);
		this.ratio= new Dimension(ratioX, ratioY);
		this.stretchStep = 1;
	}

	public PlayfieldTiling(final PlayfieldTiling other) {
		this.shapeTopologyMode = other.shapeTopologyMode;
		this.positionTopologyMode = other.positionTopologyMode;
		this.space = (Dimension) other.space.clone();
		this.surroundingBox = (Dimension) other.surroundingBox.clone();
		this.ratio = (Dimension) other.ratio.clone();
		this.stretchStep = other.stretchStep;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		// TODO: Now the unique set is topologyMode, space, ratio, strechStep
		marshaller.writeInt(space.width);
		marshaller.writeInt(space.height);
		marshaller.writeInt(surroundingBox.width);
		marshaller.writeInt(surroundingBox.height);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		space.width = marshaller.readInt();
		space.height = marshaller.readInt();
		surroundingBox.width = marshaller.readInt();
		surroundingBox.height = marshaller.readInt();
		return this;
	}

	@Override
	public String toString() {
		return "(space(" + space.width + "," + space.height + "),surr(" + surroundingBox.width + "," + surroundingBox.height + "))";
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((space == null) ? 0 : space.hashCode());
		result = prime * result + ((surroundingBox == null) ? 0 : surroundingBox.hashCode());
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		PlayfieldTiling other = (PlayfieldTiling) obj;
		if (space == null) {
			if (other.space != null) {
				return false;
			}
		} else if (!space.equals(other.space)) {
			return false;
		}
		if (surroundingBox == null) {
			if (other.surroundingBox != null) {
				return false;
			}
		} else if (!surroundingBox.equals(other.surroundingBox)) {
			return false;
		}
		return true;
	}

	@Override
	public int compareTo(final PlayfieldTiling other) {
		if (space.width != other.space.width) {
			return space.width - other.space.width;
		}
		if (space.height != other.space.height) {
			return space.height - other.space.height;
		}
		if (surroundingBox.width != other.surroundingBox.width) {
			return surroundingBox.width - other.surroundingBox.width;
		}
		if (surroundingBox.height != other.surroundingBox.height) {
			return surroundingBox.height - other.surroundingBox.height;
		}
		return 0;
	}

	@Override
	public PlayfieldTiling clone() {
		return new PlayfieldTiling(this);
	}

	public PlayfieldPosition canvas2map(final int cx, final int cy,
			final Point mapOffset) {
		int mx = 0;
		int my = 0;
		final int w = space.width;
		final int h = space.height;
		// Which neighbor fields to check if a screen position belongs to one of those? null if guess is already accurate.
		PlayfieldDistance dists[] = null;
		// First get a rough guess which playfield the position is in
		final TopologyMode tm = getPositionTopology();
		switch(tm) {
		case SQUARE8:
		case SQUARE4:
			mx = MathUtil.divisio(cx - mapOffset.x, w);
			my = MathUtil.divisio(cy - mapOffset.y, h);
			break;
		case HEXAGON:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				mx = MathUtil.divisio(cx - mapOffset.x, w);
				my = MathUtil.divisio(cy - mapOffset.y + ((0 == mx % 2) ? 0 :  h / 2), h);
			} else {
				mx = MathUtil.divisio(cx - mapOffset.x, w);
				my = MathUtil.divisio(cy - mapOffset.y + (mx * h / 2), h);
			}
			dists = new PlayfieldDistance[] {
					new PlayfieldDistance( 0,  0), // guessed field
					new PlayfieldDistance(-1, -1),
					new PlayfieldDistance(-1,  0),
					new PlayfieldDistance(-1,  1),
					new PlayfieldDistance( 1, -1),
					new PlayfieldDistance( 1,  0),
					new PlayfieldDistance( 1,  1),
			};
			break;
		case HEXAGON_H:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				mx = MathUtil.divisio(cx - mapOffset.x + ((0 == my % 2) ? 0 :  w / 2), w);
				my = MathUtil.divisio(cy - mapOffset.y, h);
			} else {
				mx = MathUtil.divisio(cx - mapOffset.x + (my * w / 2), w);
				my = MathUtil.divisio(cy - mapOffset.y, h);
			}
			dists = new PlayfieldDistance[] {
					new PlayfieldDistance( 0,  0), // guessed field
					new PlayfieldDistance(-1, -1),
					new PlayfieldDistance(-1,  0),
					new PlayfieldDistance(-1,  1),
					new PlayfieldDistance( 0, -1), //added
					new PlayfieldDistance( 0,  1), //added
					new PlayfieldDistance( 1, -1),
					new PlayfieldDistance( 1,  0),
					new PlayfieldDistance( 1,  1),
			};
			break;
		case TRIANGLE:
			// First get a rough guess which playfield the position is in
			mx = MathUtil.divisio(cx - mapOffset.x, w);
			my = MathUtil.divisio(cy - mapOffset.y, h);
			dists = new PlayfieldDistance[] {
					new PlayfieldDistance( 0,  0), // guessed field
					new PlayfieldDistance(-1,  0), // right end of the field to the left
					new PlayfieldDistance( 1,  0), // left end of the field to the right
			};
			break;
		default:
			throw new IllegalStateException("Unknown topology mode " + tm);
		}
		PlayfieldTiling.hack.clear();
		if (null == dists) {
			return new PlayfieldPosition(mx, my);
		} else {
			PlayfieldPosition ret = null;
			for (PlayfieldDistance dist : dists) {
				final PlayfieldPosition pos = new PlayfieldPosition(mx + dist.getX(), my + dist.getY());
				final Point s = new Point();
				map2canvas(s, pos, mapOffset, TOP_LEFT);
				final Polygon poly = polygon(pos, s, 0);
//				hack.add(poly);
				if (poly.contains(new Point(cx, cy))) {
					ret = pos;
					break;
				}
			}
			// Nothing of those? Then the user must have hit between the fields.
			return ret;
		}
	}

	/**
	 * Calculates the upper left point of the surrounding box of a playfield.
	 * @param point output parameter containign the resulting upper left point of the surrounding box of a playfield
	 * @param position the playfield position
	 * @param mapOffset the offset of the upper left point of the whole map on the canvas
	 * @param isCenter if {@link PlayfieldTiling#TOP_LEFT}, the canvas position of the upper left corner of the map tile is calculated; 
	 * 		if {@link PlayfieldTiling#CENTER}, the center of the map tile is calculated
	 */
	public void map2canvas(final Point point, final PlayfieldPosition position,
			final Point mapOffset, boolean isCenter) {
		point.setLocation(mapOffset);
		final int mx = position.getX();
		final int my = position.getY();
		final int spaceWidth = space.width;
		final int spaceHeight = space.height;
		final TopologyMode ptm = getPositionTopology();
		switch(ptm) {
		case SQUARE8:
		case SQUARE4:
			point.translate(Math.round(mx * spaceWidth), Math.round(my * spaceHeight));
			break;
		case HEXAGON:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				point.translate(
						mx * spaceWidth, 
						my * spaceHeight - ((0 == mx % 2) ? 0 : spaceHeight / 2));
			} else {
				point.translate(
						mx * spaceWidth, 
						my * spaceHeight - (mx * spaceHeight / 2));
			}
			break;
		case HEXAGON_H:
			if (PlayfieldTiling.IS_HEX_COMPACT) {
				point.translate(
						mx * spaceWidth - ((0 == my % 2) ? 0 : spaceWidth / 2),
						my * spaceHeight);
			} else {
				point.translate(
						mx * spaceWidth - (my * spaceWidth / 2),
						my * spaceHeight);
			}
			break;
		case TRIANGLE:
			point.translate(
					mx * spaceWidth, 
					my * spaceHeight);
			break;
		default:
			throw new IllegalStateException("Unknown topology mode " + ptm);
		}
		if (isCenter) {
			point.translate(surroundingBox.width / 2, surroundingBox.height / 2);
			switch(getShapeTopology()) {
			case TRIANGLE:
				final int surroundingHeight = surroundingBox.height;
				// The last complex term in the formula below describes that a downside triangle \/ has another optical Schwerpunkt than an upside triangle /\. 
				point.translate(0, ((position.getX() + my) % 2 == 0) ? -(surroundingHeight/6) : (surroundingHeight/6));
				break;
			default:
				// Nothing to do
			}
		}
	}

	/**
	 * @param position the playfield position
	 * @param s the point at the upper left corner of the surrounding rectangle of this polygon
	 * @param outboundPercent a 0 draws exactly the defined polygon; negative values define the polgyon inbound to the field borders (-100 is zero size); positive values define it outbound.
	 * @return a polygon outlining shape of that playfield
	 */
	public Polygon polygon(final PlayfieldPosition position, final Point s, final int outboundPercent) {
		final int[] xarr;
		final int[] yarr;
		final int xstretch = surroundingBox.width;
		final int ystretch = surroundingBox.height;
		final int xinbound = (int) Math.round(xstretch * -outboundPercent / 200.0F);
		final int yinbound = (int) Math.round(ystretch * -outboundPercent / 200.0F);
		final TopologyMode stm = getShapeTopology();
		switch(stm) {
		case HEXAGON:
			xarr = new int[8];
			yarr = new int[8];
			xarr[0] = s.x                          + xinbound; //left part 1
			xarr[1] = s.x             + xstretch/4 + xinbound / 2; //upper left
			xarr[2] = s.x + xstretch  - xstretch/4 - xinbound / 2; //upper right
			xarr[3] = s.x + xstretch               - xinbound; //right part 1
			xarr[4] = xarr[3];                      //right part 2
			xarr[5] = xarr[2];                      //lower right
			xarr[6] = xarr[1];                      //lower left
			xarr[7] = xarr[0];                      //left part 2
			yarr[0] = s.y + ystretch/2 - 1;
			yarr[1] = s.y                + yinbound;
			yarr[2] = yarr[1];
			yarr[3] = yarr[0];
			yarr[4] = yarr[0] + 1;
			yarr[5] = s.y + ystretch - 1 - yinbound;
			yarr[6] = yarr[5];
			yarr[7] = yarr[4];
			break;
		case HEXAGON_H:
			xarr = new int[8];
			yarr = new int[8];
			yarr[0] = s.y                          + yinbound; //left part 1
			yarr[1] = s.y             + ystretch/4 + yinbound / 2; //upper left
			yarr[2] = s.y + ystretch  - ystretch/4 - yinbound / 2; //upper right
			yarr[3] = s.y + ystretch               - yinbound; //right part 1
			yarr[4] = yarr[3];                      //right part 2
			yarr[5] = yarr[2];                      //lower right
			yarr[6] = yarr[1];                      //lower left
			yarr[7] = yarr[0];                      //left part 2
			xarr[0] = s.x + xstretch/2 - 1;
			xarr[1] = s.x                + xinbound;
			xarr[2] = xarr[1];
			xarr[3] = xarr[0];
			xarr[4] = xarr[0] + 1;
			xarr[5] = s.x + xstretch - 1 - xinbound;
			xarr[6] = xarr[5];
			xarr[7] = xarr[4];
			break;
		case SQUARE4:
		case SQUARE8:
			xarr = new int[4];
			yarr = new int[4];
			xarr[0] = s.x                + xinbound; //left
			xarr[1] = s.x + xstretch - 1 - xinbound; //right
			xarr[2] = xarr[1];
			xarr[3] = xarr[0];
			yarr[0] = s.y                + yinbound; // upper
			yarr[1] = yarr[0]; 
			yarr[2] = s.y + ystretch - 1 - yinbound; // lower
			yarr[3] = yarr[2];
			break;
		case TRIANGLE:
			xarr = new int[3];
			yarr = new int[3];
			xarr[0] = s.x            + (2 + xinbound);
			xarr[1] = s.x + xstretch / 2;
			xarr[2] = s.x + xstretch - (2 + xinbound);
			final int y12 = yinbound / 2;
			final int y16 = yinbound / 6;
			if ((position.getX() + position.getY()) % 2 == 0) { // upside triangle \/
				yarr[0] = s.y            + (1            + y12 + y16);
				yarr[1] = s.y + ystretch - (2 + yinbound + y12 - y16);
			} else { // downside triangle /\
				yarr[0] = s.y + ystretch - (1            + y12 + y16);
				yarr[1] = s.y            + (2 + yinbound + y12 - y16);
			}
			yarr[2] = yarr[0];
			break;
		default:
			throw new IllegalStateException("Not implemented topology " + stm);
		}
		return new Polygon(xarr, yarr, xarr.length);
	}

	public static int getPositionType(final TopologyMode topologyMode, final PlayfieldPosition position) {
		final int ret;
		switch(topologyMode) {
		case HEXAGON:
			if (IS_HEX_COMPACT) {
				if (0 != position.getX() % 2) {
					ret = HEX_COMPACT_ODD;
				} else {
					ret = HEX_COMPACT_EVEN;
				}
			} else {
				ret = NONE;
			}
			break;
		case HEXAGON_H:
			if (IS_HEX_COMPACT) {
				if (0 != position.getY() % 2) {
					ret = HEX_COMPACT_ODD;
				} else {
					ret = HEX_COMPACT_EVEN;
				}
			} else {
				ret = NONE;
			}
			break;
		default:
			ret = NONE;
		}
		return ret;
	}

	/**
	 * @param topologyMode the topologyMode
	 * @param size the size of the playfield
	 * @return the surrounding rectangle to the map, relative to the upper left corner of tile (0, 0), and in tile size.
	 */
	public static Rectangle2D.Float getMapBoundary(final TopologyMode topologyMode, final PlayfieldDimension size) {
		final Rectangle2D.Float ret = new Rectangle2D.Float(0f, 0f, size.getWidth(), size.getHeight());
		switch(topologyMode) {
		case HEXAGON:
			ret.y -= -0.5f;
			ret.height +=  0.5f;
			break;
		case HEXAGON_H:
			ret.x -= -0.5f;
			ret.width +=  0.5f;
			break;
		case SQUARE8:
		case SQUARE4:
		default:
			// Nothing to do
		}
		return ret;
	}

	public Dimension getMapSizeInPixels(final TopologyMode topologyMode, final Rectangle2D.Float mapBoundary) {
		final Dimension ret = new Dimension();
		ret.width  = Math.max(0, Math.round((mapBoundary.width  - 1) * getSpaceWidth()  + getSurroundingWidth()));
		ret.height = Math.max(0, Math.round((mapBoundary.height - 1) * getSpaceHeight() + getSurroundingHeight()));
		return ret;
	}

	/** 
	 * @param idx the index of the dimension, either {@link PlayfieldPosition#X} or {@link PlayfieldPosition#Y}
	 * @return dimension of packed space a field holds
	 */
	public int getSpaceDimension(final int idx) {
		switch(idx) {
		case PlayfieldPosition.X:
			return space.width;
		case PlayfieldPosition.Y:
			return space.height;
		default:
			throw new IllegalArgumentException("Index out of bounds [0, 1]: " + idx);
		}
	}

	/** 
	 * @param idx the index of the dimension, either {@link PlayfieldPosition#X} or {@link PlayfieldPosition#Y}
	 * @return dimension of surrounding rectangle 
	 */
	public int getSurroundingDimension(final int idx) {
		switch(idx) {
		case PlayfieldPosition.X:
			return surroundingBox.width;
		case PlayfieldPosition.Y:
			return surroundingBox.height;
		default:
			throw new IllegalArgumentException("Index out of bounds [0, 1]: " + idx);
		}
	}

	public int getSpaceWidth() {
		return space.width;
	}

	public int getSpaceHeight() {
		return space.height;
	}

	public void setSpace(final int spaceWidth, final int spaceHeight) {
		space.width  = spaceWidth;
		space.height = spaceHeight;
		switch(shapeTopologyMode) {
		case HEXAGON:
			surroundingBox.width = space.width + space.width / 4;
			surroundingBox.height = space.height;
			break;
		case HEXAGON_H:
			surroundingBox.width = space.width;
			surroundingBox.height = space.height + space.height / 4;
			break;
		case TRIANGLE:
			surroundingBox.width = space.width * 2;
			surroundingBox.height = space.height;
			break;
		default:
			surroundingBox.width = space.width;
			surroundingBox.height = space.height;
		}
	}

	public int getSurroundingWidth() {
		return surroundingBox.width;
	}

	public int getSurroundingHeight() {
		return surroundingBox.height;
	}

	public int getRatioWidth() {
		return ratio.width;
	}

	public int getRatioHeight() {
		return ratio.height;
	}

	/** @return how each tile is shaped */
	public TopologyMode getShapeTopology() {
		return shapeTopologyMode;
	}

	/** @return how the tiles are arranged on the board, including neighboring relationships */
	public TopologyMode getPositionTopology() {
		return positionTopologyMode;
	}

	private final TopologyMode shapeTopologyMode;
	private final TopologyMode positionTopologyMode;
	private final Dimension space;
	private final Dimension surroundingBox;
	private final Dimension ratio;
	public final int stretchStep;
}
