package de.ped.tools;

import java.net.MalformedURLException;
import java.net.URL;

import de.ped.tools.log.Logger;

/**
 * Finds the resources of that program. This may be different, 
 * depending on if the project is started as application, applet or servlet, 
 * and if started from inside a jar file or from the development environment.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ResourceFinder {
	/**
	 * Defines where to search for a resource.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static enum Folder {
		/** application specific */
		APP,
		/** shared between the applications */
		COMMON
	}

	/**
	 * All data about a folder where resources could reside.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ResourcesFolder {
		/** The package path to look for resources underneath. */
		private String rootPackage;
		private URL rootURL;
		private ResourcesFolder(final String folder) {
			setRootPackage(folder);
		}
		private void setRootPackage(final String folder) {
			this.rootPackage = folder;
			resetURL();
		}
		private void resetURL() {
			rootURL = null;
		}
	}

	private static Logger logger = Logger.getRootLogger();

	/**
	 * Creates a resource loader object looking for objects at the root of the package path.
	 */
	public ResourceFinder() {
		// Nothing to do
	}

	/**
	 * Creates a resource loader object. Implicitly calls
	 * {@link #setResourcesRootPackage(String)}.
	 * 
	 * @param resourcesRootPackage
	 *            the root of the resources path. 
	 *            Might be <code>null</code> if no application specific resources are needed; 
	 *            the common resources will be available nevertheless.
	 */
	public ResourceFinder(final String resourcesRootPackage) {
		setResourcesRootPackage(resourcesRootPackage);
	}

	/**
	 * Creates a resource loader object. Implicitly calls
	 * {@link #setPathToJavaRootPackage(String)}.
	 * 
	 * @param resourcesRootPackage
	 *            the root of the resources path. 
	 *            Might be <code>null</code> if no application specific resources are needed; 
	 *            the common resources will be available nevertheless.
	 * @param pathToJavaRootPackage
	 *            the root of the package path
	 */
	public ResourceFinder(final String resourcesRootPackage, final String pathToJavaRootPackage) {
		setResourcesRootPackage(resourcesRootPackage);
		setPathToJavaRootPackage(pathToJavaRootPackage);
	}

	private ResourcesFolder[] resourcesFolders = new ResourcesFolder[Folder.values().length];
	{
		resourcesFolders[Folder.APP.ordinal()] = new ResourcesFolder("");
		resourcesFolders[Folder.COMMON.ordinal()] = new ResourcesFolder("de/ped/tools/resources");
	};

	/**
	 * Sets where in the file system the root of the package path lies. This
	 * differs from the mode the system is running - application, applet or web
	 * service. The default value fits for application during development. Must
	 * be called before the first call of {@link #getURLToResource(String)}.
	 * 
	 * @param pathToJavaRootPackage
	 *            the root of the package path
	 */
	public synchronized void setPathToJavaRootPackage(final String pathToJavaRootPackage) {
		this.pathToJavaRootPackage = pathToJavaRootPackage;
		for (ResourcesFolder folder : resourcesFolders) {
			folder.resetURL();
		}
		logger.debug("Path to java root package set as " + pathToJavaRootPackage + ".");
	}

	public synchronized String getPathToJavaRootPackage() {
		return pathToJavaRootPackage;
	}

	/**
	 * Where to find all the resources like maps? This differs from the mode the
	 * system is running - application, applet or web service. The default value
	 * fits for application during development. Must be called before the first
	 * call of {@link #getURLToResource(String)}.
	 * 
	 * @param resourcesRootPackage
	 *            the root of the resources path
	 */
	public synchronized void setResourcesRootPackage(final String resourcesRootPackage) {
		resourcesFolders[Folder.APP.ordinal()].setRootPackage(resourcesRootPackage);
		logger.debug("Resources root package set as " + resourcesRootPackage + ".");
	}

	public synchronized String getResourcesRootPackage() {
		return resourcesFolders[Folder.APP.ordinal()].rootPackage;
	}

	/**
	 * Finds and returns the URL leading to the application specific resource root directory. First,
	 * the system tries to find a resource with that name, and if that fails, it
	 * builds a file URL to the path set with
	 * {@link #setPathToJavaRootPackage(String)}.
	 * 
	 * @return the URL leading to the resource directory
	 */
	public URL getURLToResources() {
		return getURLToResources(resourcesFolders[Folder.APP.ordinal()]);
	}

	private synchronized URL getURLToResources(ResourcesFolder folder) {
		if (null == folder.rootURL) {
			folder.rootURL = getClass().getResource("/" + folder.rootPackage);
			logger.debug("URL for resource root " + folder.rootPackage + " found was " + folder.rootURL);
		}
		if (null == folder.rootURL) {
			try {
				folder.rootURL = new URL("file:" + pathToJavaRootPackage + "/" + folder.rootPackage);
			} catch (final MalformedURLException e) {
				e.printStackTrace();
			}
			logger.debug("URL for resource root " + folder.rootPackage + " assumed as " + folder.rootURL);
		}
		return folder.rootURL;
	}

	/**
	 * Finds a resource file, given by the path and filename starting from the
	 * resource root. (Basically, it concatenates the URL to the resource root
	 * and the given path and filename.)
	 * 
	 * @param resourceName
	 *            path and filename relative to the resource root
	 * @return full URL
	 */
	public URL getURLToResource(final String resourceName) {
		return getURLToResource(resourceName, ResourceFinder.Folder.APP);
	}

	public URL getURLToResource(final String resourceName, final Folder folder) {
		final URL resUrl = getURLToResources(resourcesFolders[folder.ordinal()]);
		URL ret = null;
		try {
			ret = new URL(resUrl.toString() + "/" + resourceName);
		} catch (final MalformedURLException e) {
			logger.warn("", e);
		}
		return ret;
	}

	public String getRelativePathToResource(final String resourceName, final Folder folder) {
		final String resUrl = resourcesFolders[folder.ordinal()].rootPackage;
		return resUrl + "/" + resourceName;
	}

	/** Where to find all the resources like maps? */
	private String pathToJavaRootPackage = "src";

}
