package de.ped.tools;

import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.List;

import de.ped.tools.TextFormatter.TableCellInfo;
import de.ped.tools.TextFormatter.TableInfo;
import de.ped.tools.log.Logger;

/**
 * A builder for text documents.
 * Besides the pure text content, it can also contain a text title.
 * 
 * @see TextDocumentType for document types that can be built.
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TextDocumentBuilder {
	/**
	 * A data keeper for an invenory entry.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class InventoryEntry {
		public final int depth;
		public final String href;
		public final String text;
		public InventoryEntry(final int depth, final String href, final String text) {
			this.depth = depth;
			this.href = href;
			this.text = text;
		}
	}

	private static final String INVENTORY_MARKER = "<inventory/>";
	private static final int INVENTORY_MIN_DEPTH = 2;

	/**
	 * Internally used to mark the last paragraph of a section. Used to convert
	 * xhtml to html.
	 */
	private static final String P_CLASS_HACK = "<p class=\"hack\">";

	public TextDocumentBuilder(final Messages messages) {
		this.messages = messages;
		tf();
	}

	public TextDocumentType getTextDocumentType() {
		return textDocumentType;
	}

	public void setTextDocumentType(final TextDocumentType type) {
		this.textDocumentType = type;
		tf = null;
		tf();
	}

	/**
	 * @return if the document created will be embedded within another HTML
	 *         body. In this case this builder will create no embracing
	 *         &lt;HTML&gt;, &lt;HEAD&gt; or &lt;BODY&gt; elements.
	 */
	public boolean isEmbedded() {
		return isEmbedded;
	}

	/**
	 * @param isEmbedded
	 *            if the document created will be embedded within another HTML
	 *            body. In this case this builder will create no embracing
	 *            &lt;HTML&gt;, &lt;HEAD&gt; or &lt;BODY&gt; elements.
	 */
	public void setEmbedded(final boolean isEmbedded) {
		this.isEmbedded = isEmbedded;
	}

	/**
	 * Creates (if necessary) the text formatter for the defined text document type.
	 * @return the text formatter to build the document with
	 */
	public TextFormatter tf() {
		if (null == tf) {
			tf = new TextFormatter(textDocumentType);
		}
		return tf;
	}

	public Messages messages() {
		return messages;
	}

	public Charset getCharset() {
		return tf.getDocType().getCharset();
	}

	/**
	 * @return up to which level the section headings will be listed in the inventory
	 */
	public int getInventoryMaxDepth() {
		return inventoryMaxDepth;
	}

	/**
	 * @param maxDepth the maximum depth that should be shown in the inventory
	 */
	public void setInventoryMaxDepth(final int maxDepth) {
		this.inventoryMaxDepth = maxDepth;
	}

	/**
	 * @return the heading text
	 */
	public String getHeadingText() {
		return headingText;
	}

	/**
	 * @return the heading id used for the inventory to link to the heading
	 */
	public String getHeadingId() {
		return headingId;
	}

	/**
	 * Sets the headline for this page.
	 * 
	 * @param key
	 *            The I18N message file key. The key will be used as anchor, and
	 *            the message value of the key will used as headline text.
	 */
	public void setHeading(final String key) {
		setHeading(key, messages.getText(key));
	}

	/**
	 * Sets the headline for this page.
	 * 
	 * @param anchor
	 *            An anchor to be able to refer to the beginning of the page.
	 *            May be <code>null</code> if not needed.
	 * @param text
	 *            the text to display as headline
	 */
	public void setHeading(final String anchor, final String text) {
		this.headingId = anchor;
		this.headingText = text;
	}

	/**
	 * @return the title
	 */
	public String getTitle() {
		return title;
	}

	/**
	 * Sets the title. As a side effect, if the heading text is <code>null</code>,
	 * also the heading text will be set.
	 * 
	 * @param title
	 *            the title to set
	 */
	public void setTitle(final String title) {
		this.title = title;
		if (null == headingText) {
			headingId = null;
			headingText = title;
		}
	}

	/**
	 * Adds any text to this document.
	 * 
	 * @param text
	 *            the text to add
	 */
	public void add(final String text) {
		final String t;
		if (tf.getDocType() == TextDocumentType.HTML401_TRANSITIONAL) {
			t = text.replace("&apos;", "'"); // Java library bug fix
		} else {
			t = text;
		}
		mainBody.append(t);
	}

	public void addContentPart(final int depth, final String key) {
		addContentPart(depth, key, true, messages.getText(key), null);
	}
	
	public void addContentPart(final int depth, final String key, final boolean isNamedAnchor, final String text, final String imageFilename) {
		final String id = (isNamedAnchor ? key : null);
		addSectionHeading(depth, id, text, imageFilename);
		if (null != key) {
			final String desc = convertToTextDocumentType(messages.getLongDescription(key));
			if (null != desc) {
				add(tf.indent(desc));
				add(tf.ppEndLine());
				add(TextFormatter.NL);
			}
		}
	}

	/**
	 * Converts a long description from the messages that may contain paragraphs
	 * in xhtml style to html 4.0, if necessary.
	 * 
	 * @param desc
	 *            the long description as read from the messages file
	 * @return the converted (if necessary, otherwise identical) description
	 */
	public String convertToTextDocumentType(final String desc) {
		String ret = desc;
		if (null != desc) {
			switch(textDocumentType) {
			case TEXT:
			case FIXEDFONT:
				ret = ret.replace("<img.*</img>", "");
				ret = ret.replace("<br/>", "\n");
				ret = ret.replace("<p>", "");
				ret = CollectionUtil.replaceLast(ret, "</p>", P_CLASS_HACK);
				ret = ret.replace("</p>", "\n");
				ret = ret.replace("<p><ul>", "");
				ret = ret.replace("<p><ol>", "");
				ret = ret.replace("<p><dl>", "");
				if (ret.contains(P_CLASS_HACK)) {
					isParagraphTagHackApplicable = true;
				}
				break;
			case HTML401_TRANSITIONAL:
				ret = ret.replace("</img>", "");
				ret = ret.replace("<br/>", "<br>");
				ret = ret.replace("<p>", "");
				ret = CollectionUtil.replaceLast(ret, "</p>", P_CLASS_HACK);
				ret = ret.replace("</p>", "<p>");
				ret = ret.replace("<p><ul>", "<ul>");
				ret = ret.replace("<p><ol>", "<ol>");
				ret = ret.replace("<p><dl>", "<dl>");
				ret = ret.replace(P_CLASS_HACK + "<ul>", "<ul>");
				ret = ret.replace(P_CLASS_HACK + "<ol>", "<ol>");
				ret = ret.replace(P_CLASS_HACK + "<dl>", "<dl>");
				if (ret.contains(P_CLASS_HACK)) {
					isParagraphTagHackApplicable = true;
				}
				break;
			case XHTML10_TRANSITIONAL:
			case XHTML10_STRICT:
			case XHTML11:
				ret = ret.replace("<br>", "<br/>");
				ret = ret.replace("</br>", "");
				ret = ret.replace("></img>", "/>");
				ret = ret.replace(" " + TextFormatter.LANG + "=", " " + TextFormatter.XML_LANG + "=");
				break;
			default:
				// Nothing to do
			}
		}
		return ret;
	}

	/**
	 * Helps to overcome some limitations of the Java built-in HTML renderer.
	 * If this text document is of HTML4.01 transitional type (that is currently mainly used for the built-in renderer),
	 * removes the last HTML paragraph tag that was added by {@link #convertToTextDocumentType(String)}, if there is any. 
	 * That way, the vertical space between the previous text and the following text is not overly long.
	 */
	private void removePreviousHtmlTransitionalParagraphTag() {
		if (null != mainBody) {
			switch(textDocumentType) {
			case TEXT:
			case FIXEDFONT:
			case HTML401_TRANSITIONAL:
				if (isParagraphTagHackApplicable) {
					final String body = mainBody.toString();
					final int idx = body.lastIndexOf(P_CLASS_HACK);
					if (0 <= idx) {
						boolean isToRemove = true;
						final String end = body.substring(idx + P_CLASS_HACK.length());
						int tagIdx = -1;
						do {
							tagIdx = end.indexOf("<", tagIdx + 1);
							if (0 <= tagIdx) {
								int type = Character.getType(end.charAt(tagIdx + 1));
								if ((type == Character.UPPERCASE_LETTER) || (type == Character.LOWERCASE_LETTER)) {
									if (!end.substring(tagIdx).startsWith(INVENTORY_MARKER)) {
										isToRemove = false;
									}
								}
							}
						} while (0 <= tagIdx);
						if (isToRemove) {
							final int idxEnd = idx + P_CLASS_HACK.length();
							mainBody.delete(idx, idxEnd);
						}
					}
				}
				break;
			default:
				// Nothing to do
			}
		}
		isParagraphTagHackApplicable = false;
	}

	public void addSectionHeading(final int depth, final String key) {
		addSectionHeading(depth, key, messages.getText(key));
	}

	public void addSectionHeading(final int depth, String anchor, final String text) {
		addSectionHeading(depth, anchor, text, null);
	}

	/**
	 * Creates a section heading.
	 * @param depth the heading depth (1-6), or {@link #DEPTH_SAME} or {@link #DEPTH_NEXT}. 
	 * @param anchor the name of the html anchor, or <code>null</code> if generic name should be used
	 * @param text the text of the heading
	 * @param imageFilename the image file name, or <code>null</code> if there is no image attached to this heading
	 */
	public void addSectionHeading(final int depth, String anchor, final String text, final String imageFilename) {
		if (null == anchor) {
			anchor = "chapter" + Integer.toString(inventory.size());
		}
		final String textWithImg = htmlTextWithImg(text, imageFilename);
		removePreviousHtmlTransitionalParagraphTag();
		final int d;
		if ((depth == DEPTH_SAME) || (depth == DEPTH_NEXT)) {
			final int cd = getInventoryCurrentDepth();
			if (depth == DEPTH_SAME) {
				d = cd;
			} else {
				d = cd + 1;
			}
		} else {
			d = depth;
		}
		mainBody.append(TextFormatter.NL);
		mainBody.append(tf.ppBeginLine());
		mainBody.append(tf.addSectionHeading(d, tf.anchor(anchor, textWithImg)));
		mainBody.append(tf.ppEndLine());
		if ((1 < d) && (d <= inventoryMaxDepth)) {
			inventory.add(new InventoryEntry(d, anchor, text));
		}
	}

	public String htmlTextWithImg(final String text, final String imageFilename) {
		final String textWithImg;
		if (null != imageFilename) {
			textWithImg = TextFormatter.img(tf.getDocType(), imageFilename, "", null, null)
					+ (CollectionUtil.isNullOrEmpty(text) ? "" : "&nbsp;&nbsp;" + text);
		} else {
			textWithImg = text;
		}
		return textWithImg;
	}

	public static final int DEPTH_SAME = -1;
	public static final int DEPTH_NEXT = -2;

	public int getInventoryCurrentDepth() {
		final int size = inventory.size();
		final int depth;
		if (0 < size) {
			depth = inventory.get(size - 1).depth;
		} else {
			depth = 1;
		}
		return depth;
	}

	public void beginDocument() {
		if (!isEmbedded) {
			tf.setLocale(messages.getCurrentLocale());
			mainBody.append(tf.beginDocument());
			mainBody.append(tf.addHead(getTitle(), null));
			mainBody.append(tf.beginBody());
		}
		if (null != headingText) {
			mainBody.append(TextFormatter.NL);
			mainBody.append(tf.ppBeginLine());
			final String key = headingId;
			final String text = headingText;
			addContentPart(1, key, true, text, null);
			mainBody.append(tf.ppEndLine());
		}
		if (0 < inventoryMaxDepth) {
			mainBody.append(TextFormatter.NL);
			mainBody.append(tf.ppBeginLine());
			mainBody.append(tf.addSectionHeading(2, messages.getText("Help.TableOfContent")));
			mainBody.append(tf.ppEndLine());
			mainBody.append(INVENTORY_MARKER);
		}
	}

	public void endDocument() {
		String inv = buildInventory(tf.indent(""));
		if (!isEmbedded) {
			mainBody.append(tf.endBody());
			mainBody.append(tf.endDocument());
		}
		final String result = mainBody.toString().replaceAll(INVENTORY_MARKER, inv);
		mainBody = new StringBuffer(result);
	}

	private String buildInventory(final String indent) {
		final StringBuffer ret = new StringBuffer();
		int depth = INVENTORY_MIN_DEPTH - 1;
		boolean isInventory = false;
		for (int i = 0; i < inventory.size(); i++) {
			final InventoryEntry entry = inventory.get(i);
			adjustInventoryDepth(ret, depth, entry.depth);
			ret.append(tf.ppBeginLine());
			ret.append(tf.href("#" + entry.href, entry.text));
			ret.append(tf.ppEndLine());
			depth = entry.depth;
			isInventory = true;
		}
		if (isInventory) {
			adjustInventoryDepth(ret, depth, INVENTORY_MIN_DEPTH - 1);
		}
		return ret.toString();
	}

	private void adjustInventoryDepth(final StringBuffer buf, final int oldDepth, final int depth) {
		final int diff = depth - oldDepth;
		if (0 < diff) { // <ul><li> for every entry depth increase
			for (int d = oldDepth; d < depth; d++) {
				final String attributes;
				if (oldDepth < INVENTORY_MIN_DEPTH) {
					attributes = TextFormatter.htmlAttribute(TextFormatter.CLASS, "toc");
				} else {
					attributes = null;
				}
				buf.append(tf.openHtmlElement("ul", attributes));
				buf.append(tf.ppEndLine());
				buf.append(tf.openHtmlElement("li", null));
				buf.append(tf.ppEndLine());
			}
		} else if (0 == diff) { // </li><li> for every entry of same depth
			buf.append(tf.closeHtmlElement("li"));
			buf.append(tf.ppEndLine());
			buf.append(tf.openHtmlElement("li", null));
			buf.append(tf.ppEndLine());
		} else { // </li></ul> for every entry depth decrease, ending with <li> if not closing depth
			for (int d = oldDepth; d > depth; d--) {
				buf.append(tf.closeHtmlElement("li"));
				buf.append(tf.ppEndLine());
				buf.append(tf.closeHtmlElement("ul"));
				buf.append(tf.ppEndLine());
			}
			if (INVENTORY_MIN_DEPTH <= depth) {
				buf.append(tf.closeHtmlElement("li"));
				buf.append(tf.ppEndLine());
				buf.append(tf.openHtmlElement("li", null));
				buf.append(tf.ppEndLine());
			}
		}
	}

	/** 
	 * @return the document created by this builder
	 */
	public String result() {
		return result(null);
	}

	public String result(final ResourceFinder finder) {
		return result(finder, false);
	}

	/**
	 * @param finder if not <code>null</code>, replaces in the result all HTML image source paths by their resource URLs
	 * @param isImgFiles if <code>true</code>, images are not inserted as URLs but a local files.
	 * @return the document created by this builder
	 */
	public String result(final ResourceFinder finder, final boolean isImgFiles) {
		Logger logger = Logger.getLogger(getClass());
		removePreviousHtmlTransitionalParagraphTag();
		String content = mainBody.toString();
		if (null != finder) {
			int index = 0;
			while (0 <= index) {
				final String initialPart = "src=\"";
				index = content.indexOf(initialPart, index);
				if (0 <= index) {
					int startIndex = index + initialPart.length();
					int endIndex = content.indexOf("\"", startIndex);
					final String filename = content.substring(startIndex, endIndex);
					final ResourceFinder.Folder folder;
					final String path;
					final String prefix = COMMON_IMAGE_PREFIX;
					if (filename.startsWith(prefix)) {
						folder = ResourceFinder.Folder.COMMON;
						path = filename.substring(prefix.length());
					} else {
						folder = ResourceFinder.Folder.APP;
						path = filename;
					}
					final String urlString;
					if (isImgFiles) {
						urlString = finder.getRelativePathToResource(path, folder);
					} else {
						final URL url = finder.getURLToResource(path, folder);
						urlString = url.toString();
					}
					logger.trace("filename=" + filename + " --> ImageURL=" + urlString);
					content = content.substring(0, startIndex) + urlString + content.substring(endIndex);
					index = endIndex + urlString.length() - path.length();
				}
			}
		}
		return content;
	}

	public static String getCodedImageFilename(final String filename, final ResourceFinder.Folder folder) {
		final String prefix;
		if (null == folder) {
			prefix = null;
		} else if (ResourceFinder.Folder.COMMON == folder) {
			prefix = COMMON_IMAGE_PREFIX;
		} else {
			prefix = "";
		}
		final String ret;
		if ((null != prefix) && (null != filename)) {
			ret = prefix + filename;
		} else {
			ret = null;
		}
		return ret;
	}

	public void beginTable(final String id, final TableInfo tableInfo) {
		removePreviousHtmlTransitionalParagraphTag();
		this.tableInfo = tableInfo;
		add(tf.beginTable(id, tableInfo));
		if ((null != tableInfo) && (null != tableInfo.columns) && tableInfo.isWithHeading) {
			add(tf.formatTableHeading(tableInfo.columns));
		}
		add(tf.beginTableBody());
	}

	public void endTable() {
		add(tf.endTableBody());
		add(tf.endTable());
		tableInfo = null;
	}

	public void addTableRow(final ArrayList<TableCellInfo> cellsData) {
		add(tf.beginTableRow());
		// Because of colspan (and - not implemented yet - rowspan), table column number might differ from array index.
		int tableColumn = 0;
		for (int column = 0; column < cellsData.size(); column++) {
			add(tf.ppBeginLine());
			final TableCellInfo tci = cellsData.get(column);
			add(tf.formatTableCell(tableInfo.columns[tableColumn], tci));
			add(tf.ppEndLine());
			tableColumn += tci.getColspan();
		}
		add(tf.endTableRow());
	}

	public void beginParagraph(final String htmlAttributes) {
		add(tf.beginParagraph(htmlAttributes));
	}

	public void endParagraph() {
		add(tf.endParagraph());
	}

	public void beginDL(final String attributes) {
		removePreviousHtmlTransitionalParagraphTag();
		if (textDocumentType.isPlain()) {
			// Nothing to do
		} else {
			add(tf.openHtmlElement(TextFormatter.DL, attributes));
			add(tf.ppEndLine());
		}
	}

	public void endDL() {
		if (textDocumentType.isPlain()) {
			// Nothing to do
		} else {
			add(tf.closeHtmlElement(TextFormatter.DL));
			add(tf.ppEndLine());
		}
	}

	public void addDT(final String text) {
		if (textDocumentType.isPlain()) {
			add(text);
			add(TextFormatter.NL);
		} else {
			add(tf.openHtmlElement(TextFormatter.DT, null));
			add(tf.ppEndLine());
			add(tf.ppBeginLine());
			add(text);
			add(tf.ppEndLine());
			add(tf.closeHtmlElement(TextFormatter.DT));
			add(tf.ppEndLine());
		}
	}

	public void beginDD() {
		if (textDocumentType.isPlain()) {
			// Nothing to do
		} else {
			add(tf.openHtmlElement(TextFormatter.DD, null));
			add(tf.ppEndLine());
		}
	}

	public void endDD() {
		if (textDocumentType.isPlain()) {
			add(TextFormatter.NL);
		} else {
			add(tf.closeHtmlElement(TextFormatter.DD));
			add(tf.ppEndLine());
		}
	}

	/**
	 * Adds a full {@link TextTree} structure description to this document. The presentation of the tree structure is implemented by HTML heading elements and/or by a HTML definition list structure.
	 * @param textTree the root element of the tree
	 * @param headingDepth the heading depth to show the root element with. If the heading depth exceeds the maximum depth to create headings for, then a definition list structure is created.
	 * @param maxHeadingDepth the maximum depth to create headings for
	 * @param isToCreateAnchors if HTML anchor elements should be created (can be overruled by the {@link TextTreeNodeType} of the tree elements)
	 */
	public void add(final TextTree textTree, final int headingDepth, final int maxHeadingDepth, final boolean isToCreateAnchors) {
		if (null != textTree) {
			String key = null;
			String text = null;
			String description = null;
			final I18NStringWithFillIns node = textTree.getNode();
			// Open this node
			if (null != node) {
				key = node.getKey();
				text = node.getAlternative();
				if (null == text) {
					text = messages.getText(key);
				}
				description = textTree.getLongDescription();
				if (null == description) {
					description = messages.getLongDescription(key);
				}
				if (CollectionUtil.isNullOrEmpty(description)) {
					// Description is a long text, with Paragraph surrounding. So for proper formatting, we need at least an empty paragraph.
					description = "<p></p>";
				}
				description = convertToTextDocumentType(description);
				if (null != text) {
					final boolean isNamedAnchor;
					switch(textTree.getType()) {
					case ITEM:
					case PROPERTY_ROOT:
						isNamedAnchor = true;
						break;
					default:
						isNamedAnchor = isToCreateAnchors;
					}
					final String imageFilename = textTree.getImageFilename();
					if (maxHeadingDepth < headingDepth) {
						final String id = (isNamedAnchor ? key : null);
						beginHtmlDlEntry(id, text, description, imageFilename);
					} else {
						addContentPart(headingDepth, key, isNamedAnchor, text, imageFilename);
					}
				}
			}
			// Traverse downwards
			final List<TextTree> children = textTree.getChildren();
			if (!CollectionUtil.isNullOrEmpty(children)) {
				// Open list of children
				if (maxHeadingDepth <= headingDepth) {
					beginDL(textTree.getHtmlAttribs());
				}
				// Traverse all children
				for (final TextTree child : children) {
					add(child, headingDepth + 1, maxHeadingDepth, isToCreateAnchors);
				}
				// Close list of children
				if (maxHeadingDepth <= headingDepth) {
					endDL();
				}
			}
			// Close this node
			if (null != text) {
				if (maxHeadingDepth < headingDepth) {
					endHtmlDlEntry(description);
				} else {
					// Nothing to do
				}
			}
		}
	}

	private void beginHtmlDlEntry(final String id, final String title, final String description, final String imageFilename) {
		final TextFormatter tf = tf();
		final String sgmlTitle = tf.sgmlify(title);
		final String textWithImg = htmlTextWithImg(sgmlTitle, imageFilename);
		String titleText;
		if (null != id) {
			titleText = tf.anchor(id, textWithImg);
		} else {
			titleText = textWithImg;
		}
		addDT(titleText);
		beginDD();
		if (null != description) {
			add(tf.ppBeginLine());
			add(description);
			add(tf.ppEndLine());
		}
	}

	private void endHtmlDlEntry(final String description) {
		endDD();
	}

	public int getMinIndent() {
		return tf.getMinIndent();
	}

	public void setMinIndent(final int indent) {
		tf.setMinIndent(indent);
	}

	/** The text formatter to build the document with. */
	protected TextFormatter tf = null;

	/** The object to internationalize the document with. */
	protected final Messages messages;

	private TableInfo tableInfo = null;

	private boolean isEmbedded = false;
	private static String COMMON_IMAGE_PREFIX = "§";

	/** Is the text to display with inventory? */
	private int inventoryMaxDepth = 3;
	/** The document title to display in the top of the window */
	private String title;
	/** The id for the document title to use as anchor */
	private String headingId;
	/** The text of the document title to display as heading */
	private String headingText;
	/** The mainBody of the document, after the inventory */
	private StringBuffer mainBody = new StringBuffer();

	private boolean isParagraphTagHackApplicable = false;

	private ArrayList<InventoryEntry> inventory = new ArrayList<InventoryEntry>();

	private TextDocumentType textDocumentType = TextDocumentType.HTML401_TRANSITIONAL;

}
