package de.ped.tools;

import java.nio.charset.Charset;

/**
 * Different document types, including HTML types, xhtml types, and plain text types,
 * in which documents can be formatted into.
 *
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 * @see <a href="http://www.w3schools.com/tags/tag_doctype.asp">w3schools on HTML and xhtml doctypes</a>
 */
public enum TextDocumentType {
	/** For normal text with variable font width. */
	TEXT                (""),
	/** For normal text with fixed font width. */
	FIXEDFONT           (""),
	/** For HTML 4.01 transitional formatting. */
	HTML401_TRANSITIONAL("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">"),
	/** For HTML 4.01 strict formatting. */
	HTML401_STRICT      ("<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">"),
	/** For HTML 5 formatting. */
	HTML5               ("<!DOCTYPE html>"),
	/** For xhtml 1.0 transitional formatting. */
	XHTML10_TRANSITIONAL("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">"),
	/** For xhtml 1.0 strict formatting. */
	XHTML10_STRICT      ("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"),
	/** For xhtml 1.1 formatting. */
	XHTML11             ("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">");

	public static final String CHARSET_UTF_8 = "UTF-8";
	public static final String CHARSET_ISO_8859_1 = "ISO-8859-1";
	public static final String CHARSET_WINDOWS_1252 = "windows-1252";

	/**
	 * @param doctypeTag an introducing description in the document naming the document type
	 */
	private TextDocumentType(final String doctypeTag) {
		this.doctypeTag = doctypeTag;
	}

	public boolean isPlain() {
		return (this == TEXT) || (this == FIXEDFONT);
	}

	public boolean isHtml() {
		return (this == HTML401_TRANSITIONAL) 
				|| (this == HTML401_STRICT) 
				|| (this == HTML5);
	}

	public boolean isXhtml() {
		return (this == XHTML10_TRANSITIONAL) 
				|| (this == XHTML10_STRICT) 
				|| (this == XHTML11);
	}

	public boolean isOldHtml() {
		return (this == HTML401_TRANSITIONAL);
	}

	public String[] getTypicalFileExtensions() {
		String[] ret;
		switch(this) {
		case TEXT:
		case FIXEDFONT:
			ret = new String[] { "txt" };
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
			ret = new String[] { "html", "htm" };
			break;
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
			ret = new String[] { "html", "htm", "xml" };
			break;
		default:
			ret = null;
			break;
		}
		return ret;
	}

	public String getTypicalFileFilterDescription() {
		String ret;
		switch(this) {
		case TEXT:
		case FIXEDFONT:
			ret = "Text files (*.txt)";
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
			ret = "HTML files (*.html, *.htm)";
			break;
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
			ret = "XHTML files (*.html, *.htm, *.xml)";
			break;
		default:
			ret = null;
			break;
		}
		return ret;
	}

	public Charset getCharset() {
//		switch (this) {
//		case FIXEDFONT:
//		case TEXT:
//		case HTML401_TRANSITIONAL:
//			return Charset.forName(CHARSET_ISO_8859_1);
//		default:
//			return Charset.forName(CHARSET_ISO_8859_1);
			return Charset.forName(CHARSET_UTF_8);
//		}
	}

	/** An introducing description in the document naming the document type. */
	public final String doctypeTag;
}