package de.ped.tools;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Locale;
import java.util.Stack;

import de.ped.tools.log.Logger;

/**
 * Useful routines for output formatting.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TextFormatter {
	private static Logger logger = Logger.getRootLogger();

	/** allowed alignment value */
	public static final int LEFT = 0;

	/** allowed alignment value */
	public static final int CENTER = 1;

	/** allowed alignment value */
	public static final int RIGHT = 2;

	/** allowed alignment value */
	public static final int UNDEFINED = -1;

	/** takes alignment value as index, and returns how this alignment is called as HTML style. */
	public static final String[] ALIGNMENT_NAME = new String[]{"left", "center", "right"}; 
	/** takes alignment value as index, and returns how this alignment is called as HTML class. */
	public static final String[] ALIGNMENT_CLASS_NAME = new String[]{"textleft", "textcentered", "textright"}; 

	public static final String NL = "\n";

	public static final String DEFAULT_INDENT = "\t";

	/** HTML element */
	private static final String BODY = "body";
	/** HTML element */
	private static final String DIV = "div";
	/** HTML element */
	private static final String PRE = "pre";
	/** HTML element */
	private static final String TABLE = "table";
	/** HTML element */
	private static final String TBODY = "tbody";
	/** HTML element */
	private static final String THEAD = "thead";
	/** HTML element */
	private static final String TR = "tr";
	/** HTML element */
	public static final String DL = "dl";
	/** HTML element */
	public static final String DD = "dd";
	/** HTML element */
	public static final String DT = "dt";
	/** HTML element */
	public static final String IMG = "img";

	/** HTML attribute. */
	public static final String CLASS = "class";
	/** HTML attribute. */
	public static final String TITLE = "title";
	/** HTML attribute. */
	public static final String FOR   = "for";
	/** HTML attribute. */
	public static final String ID    = "id";
	/** HTML attribute. */
	public static final String NAME  = "name";
	/** HTML attribute. */
	public static final String SIZE  = "size";
	/** HTML attribute. */
	public static final String MULTIPLE  = "multiple";
	/** HTML attribute. */
	public static final String HREF  = "href";
	/** HTML attribute. */
	public static final String BORDER = "border";
	/** HTML attribute. */
	public static final String CELLSPACING = "cellspacing";
	/** HTML attribute. */
	public static final String CELLPADDING = "cellpadding";
	/** HTML attribute. */
	public static final String COLSPAN = "colspan";
	/** HTML attribute. */
	public static final String ROWSPAN = "rowspan";
	/** Default value for HTML attributes {@link #COLSPAN} and {@link #ROWSPAN}. */
	public static final int SPAN_DEFAULT = 1;
	/** HTML attribute. */
	public static final String SRC = "src";
	/** HTML attribute. */
	public static final String ALT = "alt";
	/** HTML attribute. */
	public static final String ALIGN = "align";
	/** HTML attribute. */
	public static final String STYLE = "style";
	/** HTML attribute. */
	public static final String XMLNS = "xmlns";
	/** HTML attribute. */
	public static final String LANG = "lang";
	/** HTML attribute. */
	public static final String XML_LANG = "xml:lang";
	/** HTML attribute. */
	public static final String HTTP_EQUIV = "http-equiv";
	/** HTML attribute. */
	public static final String CHARSET = "charset";
	/** HTML attribute. */
	public static final String CONTENT = "content";
	/** HTML attribute. */
	public static final String REL = "rel";
	/** HTML attribute. */
	public static final String TYPE = "type";
	/** Value for HTML attribute {@link #TYPE}. */
	public static final String CHECKBOX = "checkbox";
	/** HTML attribute. */
	public static final String VALUE = "value";
	/** HTML attribute. */
	public static final String SELECTED = "selected";
	/** HTML attribute. */
	public static final String CHECKED = "checked";
	/** HTML attribute. */
	public static final String WIDTH = "width";
	/** HTML attribute. */
	public static final String HEIGHT = "height";
	/** HTML attribute. */
	public static final String ROLE = "role";
	/** Value for HTML attribute {@link #ROLE}, indicating copright info, footer, etc. */
	public static final String CONTENTINFO = "contentinfo";
	/** HTML attribute. */
	public static final String SCOPE = "scope";
	/** Value for HTML attribute {@link #SCOPE}. */
	public static final String COL = "col";
	/** Value for HTML attribute {@link #SCOPE}. */
	public static final String ROW = "row";
	/** Value for HTML attribute {@link #SCOPE}. */
	public static final String COLGROUP = "colgroup";
	/** Value for HTML attribute {@link #SCOPE}. */
	public static final String ROWGROUP = "rowgroup";

	private static String decideOnHtml(final TextDocumentType doctype, final String plain, final String html) {
		String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = plain;
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = html;
			break;
		}
		return ret;
	}

	/**
	 * @param doctype
	 *            the document type to use
	 * @return the representation of a line break in the given format style
	 */
	protected static String newLine(final TextDocumentType doctype) {
		String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = NL;
			break;
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
			ret = "<br/>" + NL;
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		default:
			ret = "<br>" + NL;
		}
		return ret;
	}

	/**
	 * @param doctype
	 *            the document type to use
	 * @return the representation of a line break in the given format style, if
	 *         the format style is a plain text style. (HTML renderers are able
	 *         to do the line breaks themselves.)
	 */
	public static String newLineIfPlainText(final TextDocumentType doctype) {
		return decideOnHtml(doctype, newLine(doctype), " ");
	}

	/**
	 * Escapes typical SGML characters to be safe for HTML pages.
	 * <ul>
	 *   <li>The &lt; and &gt; characters, </li>
	 *   <li>the &amp;,</li>
	 *   <li>The &quot;.</li>
	 * </ul>
	 * @param doctype according to this format style the conversion will be done
	 * @param text the string to be converted
	 * @return the converted string
	 * @see HtmlHelper#sgmlify(String)
	 */
	public static String sgmlify(final TextDocumentType doctype, final String text) {
		String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = text;
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = HtmlHelper.sgmlify(text);
			break;
		}
		return ret;
	}

	/**
	 * Takes a HTML snippet and strips it down to plain text.
	 * @param doctype according to this format style the conversion will be done
	 * @param text the string to convert
	 * @return the converted string
	 */
	public static String toPlainText(TextDocumentType doctype, String text) {
		String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = text;
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			final String[] HTML_ELEMENTS_TO_IGNORE = new String[] {
					"html", BODY, "article", "aside", 
					"a", 
					"b", "strong", "i", "em", "mark", "small", "del", "ins", "sub", "sup", 
					"q", "cite", "bdi", "big", "small", "abbr", "dfn", 
					"h1", "h2", "h3", "h4", "h5", "h6",
					"hr", 
					"pre", "blockquote", "address", "bdo", 
					"img", 
					"div", "span", "center", 
			};
			final String[] HTML_BLOCKS_TO_IGNORE = new String[] {
					"head",  
					"map", 
					"audio",
					"form",
					"details",
			};
			ret = text;
			for (final String element : HTML_BLOCKS_TO_IGNORE) {
				ret = ret.replaceAll("<" + element + "\\b.*?>.*?</" + element + "\\b.*?>", ""); // full content of <element>...</element>
			}
			for (final String element : HTML_ELEMENTS_TO_IGNORE) {
				ret = ret.replaceAll("</?" + element + "\\b.*?>", ""); // <element>, </element>, <element/>
			}
			if (doctype.isXhtml()) {
				ret = ret.replaceAll("<p\\b.*?/>", "\n"); // <p/>
				ret = ret.replaceAll("<p\\b.*?>", "");    // <p>
				ret = ret.replaceAll("</p\\b.*?>", "\n"); // </p>
			} else {
				ret = ret.replaceAll("<p\\b.*?>", "\n");  // <p>, <p/>
				ret = ret.replaceAll("</p\\b.*?>", "");   // </p>
			}
			ret = ret.replaceAll("<br\\b.*?>", "\n"); // <br>, <br/>
			ret = ret.replaceAll("</br\\b.*?>", "");  // </br>
			ret = ret.replaceAll("<br\\b.*?>", "\n"); // <br>, <br/>
			ret = ret.replaceAll("</br\\b.*?>", "");  // </br>
			ret = HtmlHelper.deSgmlify(ret);
			break;
		}
		return ret;
	}

	/**
	 * If in html mode, creates an anchor tag around the text.
	 * 
	 * @param doctype
	 *            depending of the format style, the anchor is set
	 * @param text
	 *            the text to display
	 * @param id
	 *            the name of the anchor to define
	 * @return the extended text
	 */
	public static String anchor(final TextDocumentType doctype, final String id, final String text) {
		String str = text;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			// Nothing to do
			break;
		case HTML5:
		case XHTML11:
			str = "<a" + htmlAttribute(ID, id) + ">" + (null == text ? "" : text) + "</a>";
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		default:
			str = "<a" + htmlAttribute(ID, id) + htmlAttribute(NAME, id) + ">" + (null == text ? "" : text) + "</a>";
			break;
		}
		return str;
	}

	public static String htmlAttribute(final String key, final String value) {
		String ret = "";
		if (null != value) {
			return " " + key + "=\"" + value + "\"";
		}
		return ret;
	}

	/**
	 * If in html mode, creates a href (a link) around the text.
	 * 
	 * @param doctype
	 *            depending of the format style, the href is set
	 * @param text
	 *            the text to display
	 * @param href
	 *            the name of the anchor to link to
	 * @return the extended text
	 */
	protected static String href(final TextDocumentType doctype, final String href, final String text) {
		String str = text;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			// Nothing to do
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			str = "<a" + htmlAttribute(HREF, sgmlify(doctype, href)) + ">" + text + "</a>";
		}
		return str;
	}

	/**
	 * Creates a HTML snippet containing an <code>img</code> element
	 * @param doctype the text document type.
	 * @param src the image source. If <code>null</code>, only an empty string will be returned for no image at all.
	 * @param alt the alternative text for visually impaired
	 * @param size the image width and height
	 * @param htmlAttribs other HTML attributes in HTML syntax
	 * @return a HTML snippet containing an <code>img</code> element
	 */
	public static String img(final TextDocumentType doctype, final String src, final String alt, final Dimension size, final String htmlAttribs) {
		return (null == src ? "" : 
			"<" + TextFormatter.IMG 
				+ htmlAttribute(TextFormatter.SRC, src) 
				+ htmlAttribute(TextFormatter.ALT, (null == alt ? "" : alt))
				+ (null == size ? "" : 
						htmlAttribute(TextFormatter.WIDTH, Integer.toString(size.width))
						+ htmlAttribute(TextFormatter.HEIGHT, Integer.toString(size.height)))
				+ (null == htmlAttribs ? "" : htmlAttribs)
				+ TextFormatter.closingTag(doctype, TextFormatter.IMG));
	}

	public static String closingTag(final TextDocumentType doctype, final String htmlElement) {
		final String closingTag;
		if (doctype.isOldHtml() || (doctype == TextDocumentType.HTML5)) {
			closingTag = ">";
		} else if (doctype.isXhtml()) {
			closingTag = "/>";
		} else {
			closingTag = "></" + htmlElement + ">";
		}
		return closingTag;
	}

	public static final String CONTENT_TYPE_PLAINTEXT = "text/plain";

//    public static final String CONTENT_TYPE_HTML = "text/html; charset=iso-8859-1";
	public static final String CONTENT_TYPE_HTML = "text/html; charset=UTF-8";

	private static final String[][] HEADING_PRE_POSTFIX = new String[][] {
		{"### ", " ###"},
		{"*** ", " ***"},
		{"=== ", " ==="},
		{"--- ", " ---"},
		{"- - ", " - -"},
		{"", ""},
	};

	/**
	 * Checks if the string is good for a table cell content (that means, is not
	 * empty), and repairs if necessary.
	 * 
	 * @param doctype
	 *            the formatting style
	 * @param value
	 *            the cell value to display
	 * @return <em>value</em>, or if <em>doctype</em> is HTML and <em>value</em>
	 *         is empty, a non-breakable space character in HTML
	 */
	protected static String tableCellContent(final TextDocumentType doctype, final String value) {
		String ret = value;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			// Nothing to do
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			if (CollectionUtil.isNullOrEmpty(value)) {
				ret = "&nbsp;";
			}
			break;
		}
		return ret;
	}

	/**
	 * Adds characters to a given string to achieve the right width.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the string (and to the inverse, where to
	 *            put the fill chars)
	 *            (one of {@link #LEFT}, {@link #CENTER}, {@link #RIGHT}; default is LEFT)
	 * @param fillChar
	 *            this char will be repeatedly inserted as filling
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final String val, final int width, final int alignment, final char fillChar) {
		final StringBuffer ret = new StringBuffer(width);
		switch (alignment) {
		case CENTER: {
			final int full = width - val.length();
			final int half = full / 2;
			for (int i = 0; i < half; i++) {
				ret.append(fillChar);
			}
			ret.append(val);
			for (int i = half; i < full; i++) {
				ret.append(fillChar);
			}
			break;
		}
		case RIGHT: {
			final int full = width - val.length();
			for (int i = 0; i < full; i++) {
				ret.append(fillChar);
			}
			ret.append(val);
			break;
		}
		case LEFT:
		default:
			ret.append(val);
			while (ret.length() < width) {
				ret.append(fillChar);
			}
			break;
		}
		return ret;
	}

	/**
	 * Adds blanks to a given string to achieve the right width.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the string (and to the inverse, where to
	 *            put the blanks)
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final String val, final int width, final int alignment) {
		return format(val, width, alignment, ' ');
	}

	/**
	 * Adds blanks to a given string to achieve the right width. The blanks are
	 * always added at the end.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final String val, final int width) {
		return format(val, width, LEFT);
	}

	/**
	 * Adds blanks to a given number to achieve the right width.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the string (and to the inverse, where to
	 *            put the fill chars)
	 * @param fillChar
	 *            this char will be repeatedly inserted as filling
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final int val, final int width, final int alignment, final char fillChar) {
		return format(Integer.toString(val), width, alignment, fillChar);
	}

	/**
	 * Adds blanks to a given number to achieve the right width.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the string (and to the inverse, where to
	 *            put the fill chars)
	 * @param fillChar
	 *            this char will be repeatedly inserted as filling
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final long val, final int width, final int alignment, final char fillChar) {
		return format(Long.toString(val), width, alignment, fillChar);
	}

	/**
	 * Adds blanks to the output of a number to achieve the right width.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the integer (and to the inverse, where to
	 *            put the blanks)
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final int val, final int width, final int alignment) {
		return format(Integer.toString(val), width, alignment);
	}

	/**
	 * Adds blanks to the output of a number to achieve the right width.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the integer (and to the inverse, where to
	 *            put the blanks)
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final long val, final int width, final int alignment) {
		return format(Long.toString(val), width, alignment);
	}

	/**
	 * Adds blanks to the output of a number to achieve the right width. The
	 * blanks are always added at the end.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final int val, final int width) {
		return format(Integer.toString(val), width, RIGHT);
	}

	/**
	 * Adds blanks to the output of a number to achieve the right width. The
	 * blanks are always added at the end.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final long val, final int width) {
		return format(Long.toString(val), width, RIGHT);
	}

	/**
	 * Adds blanks to the output of a double to achieve the right width.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @param afterDotPositions
	 *            so many positions after the dot are shown
	 * @param alignment
	 *            tells where to put the number (and to the inverse, where to
	 *            put the blanks) (one of {@link #LEFT}, {@link #CENTER} or
	 *            {@link #RIGHT})
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(double val, final int width, final int afterDotPositions, final int alignment) {
		for (int i = 0; i < afterDotPositions; i++) {
			val *= 10.0;
		}
		val = Math.round(val);
		for (int i = 0; i < afterDotPositions; i++) {
			val /= 10.0;
		}
		return format(Double.toString(val), width, alignment);
	}

	/**
	 * Adds blanks to the output of a double to achieve the right width. The
	 * blanks are always added to the left.
	 * 
	 * @param val
	 *            the number
	 * @param width
	 *            the length in characters that's to achieve
	 * @param afterDotPositions
	 *            so many positions after the dot are shown
	 * @return the formatted string buffer
	 */
	public static StringBuffer format(final double val, final int width, final int afterDotPositions) {
		return format(val, width, afterDotPositions, RIGHT);
	}

    public static final String HEXDIGITS = "0123456789ABCDEF";

	public static void hexdump(final StringBuffer buf, final byte b) {
		// buf.append(b);
		// buf.append("=");
		buf.append(HEXDIGITS.charAt((b >>> 4) & 0x0f));
		buf.append(HEXDIGITS.charAt((b >>> 0) & 0x0f));
	}
	
	public static String hexdump(final byte b) {
		final StringBuffer buf = new StringBuffer();
		hexdump(buf, b);
		return buf.toString();
	}

	public static void hexdump(final StringBuffer buf, final byte[] arr, final int start, final int len,
			final String filling, final int blocksize, final String blockPrefix, final String blockPostfix) {
		for (int block = 0; block < len; block += blocksize) {
			buf.append(blockPrefix);
			for (int i = 0; (i < blocksize) && (start + block + i < len); i++) {
				hexdump(buf, arr[start + block + i]);
				buf.append(filling);
			}
			buf.append(blockPostfix);
		}
	}

	public static String hexdump(final byte[] arr, final int start, final int len,
			final String filling, final int blocksize, final String blockPrefix, final String blockPostfix) {
		final StringBuffer buf = new StringBuffer();
		hexdump(buf, arr, start, len, filling, blocksize, blockPrefix, blockPostfix);
		return buf.toString();
	}

	public static void hexdump(final StringBuffer buf, final byte[] arr, final int start, final int len, final String filling) {
		hexdump(buf, arr, start, len, filling, Integer.MAX_VALUE, "", "");
	}

	public static String hexdump(final byte[] arr, final int start, final int len, final String filling) {
		return hexdump(arr, start, len, filling, Integer.MAX_VALUE, "", "");
	}

	public static void hexdump(final StringBuffer buf, final byte[] arr, final int start, final int len) {
		hexdump(buf, arr, start, len, " ");
	}

	public static String hexdump(final byte[] arr, final int start, final int len) {
		return hexdump(arr, start, len, " ");
	}

	public static void hexdumpCode(final StringBuffer buf, final byte[] arr, final int start, final int len) {
		hexdump(buf, arr, start, len, "", 0x20, "\"", "\" +\n");
	}

	public static String hexdumpCode(final byte[] arr, final int start, final int len) {
		return hexdump(arr, start, len, "", 0x20, "\"", "\" +\n");
	}

	public static void hexdump(final StringBuffer buf, final byte[] arr) {
		hexdump(buf, arr, 0, arr.length);
	}

	public static void hexdumpCode(final StringBuffer buf, final byte[] arr) {
		hexdumpCode(buf, arr, 0, arr.length);
	}

	public static String hexdump(final byte[] arr) {
		final StringBuffer buf = new StringBuffer();
		hexdump(buf, arr);
		return buf.toString();
	}

	public static String hexdumpCode(final byte[] arr) {
		final StringBuffer buf = new StringBuffer();
		hexdumpCode(buf, arr);
		return buf.toString();
	}

	/**
	 * Repeats the given string n times.
	 * 
	 * @param str
	 *            string to repeat
	 * @param times
	 *            number of times to repeat
	 * @return The concatenation of the given string, n times.
	 */
	public static String repeat(final String str, final int times) {
		String ret = null;
		if (null != str) {
			final StringBuffer buf = new StringBuffer();
			for (int i = 0; i < times; i++) {
				buf.append(str);
			}
			ret = buf.toString();
		}
		return ret;
	}

	/**
	 * Indents the given text by prefixing the given indent string n times.
	 * 
	 * @param strIndent
	 *            the string with which to prefix
	 * @param numIndent
	 *            number of times to repeat the prefix
	 * @param text
	 *            text to be prefixed
	 * @return The concatenation of the prefix, n times, and the text.
	 */
	public static String indent(final String strIndent, final int numIndent, final String text) {
		return repeat(strIndent, numIndent) + text;
	}

	/**
	 * Parses a string and handles special white characters:
	 * <ul>
	 *   <li>
	 * Newlines ('\n') are separators between strings in the returned string
	 * array
	 *   </li>
	 *   <li>
	 * Tabs ('\t') are filled up with spaces to the next tab stop (given a tab
	 * width of 8 characters)
	 *   </li>
	 * </ul>
	 * 
	 * @param s
	 *            the string to wrap
	 * @return the resulting array of strings
	 */
	public static String[] wrap(final String s) {
		String[] ret = null;
		if (null != s) {
			// Find all the newlines ('\n') and split the string up
			// As an array can't grow, we use a vector as buffer
			final ArrayList<String> v = new ArrayList<String>();
			int sourceIndex = 0;
			int sourceNewLineIndex;
			while (sourceIndex < s.length()) {
				sourceNewLineIndex = s.indexOf('\n', sourceIndex);
				if (sourceNewLineIndex < 0) {
					sourceNewLineIndex = s.length();
				}
				try {
					final String line = new String(s.substring(sourceIndex, sourceNewLineIndex));
					v.add(line);
				} catch (final StringIndexOutOfBoundsException e) {
					logger.error("Exception while newlines: ", e);
				}
				sourceIndex = sourceNewLineIndex + 1;
			}
			// Find all the tabs ('\t') and insert spaces instead
			for (int line = 0; line < v.size(); line++) {
				final StringBuffer buf = new StringBuffer();
				final String str = v.get(line);
				sourceIndex = 0;
				boolean ready = false;
				while (sourceIndex < str.length()) {
					sourceNewLineIndex = str.indexOf('\t', sourceIndex);
					if (sourceNewLineIndex < 0) {
						sourceNewLineIndex = str.length();
						ready = true;
					}
					try {
						buf.append(str.substring(sourceIndex, sourceNewLineIndex));
					} catch (final StringIndexOutOfBoundsException e) {
						logger.error("Exception while tabs: ", e);
					}
					if (!ready) {
						do {
							buf.append(' ');
						} while ((buf.length() % 8) != 0);
					}
					sourceIndex = sourceNewLineIndex + 1;
				}
				v.set(line, buf.toString());
			}
			// Now we know the number of lines, so copy it to a string array
			ret = v.toArray(new String[] {});
		}
		return ret;
	}

	public static String unwrap(final String[] lines) {
		String ret = null;
		if (null != lines) {
			final StringBuffer sb = new StringBuffer();
			for (final String line : lines) {
				sb.append(line);
				sb.append('\n');
			}
			ret = sb.toString();
		}
		return ret;
	}

	public static String asButton(final String text) {
		return " <b>\"<span" + htmlAttribute(CLASS, "button") + ">" + text + "</span>\"</b> ";
	}

	private static final int SECTION_HEADING_NONE = -1;

	public TextFormatter(final TextDocumentType doctype) {
		setDocType(doctype);
	}

	/** 
	 * @return the locale for this document
	 */
	public Locale getLocale() {
		return locale;
	}

	public void setLocale(final Locale locale) {
		this.locale = locale;
	}

	/** 
	 * @return the doc type for this document
	 */
	public TextDocumentType getDocType() {
		return doctype;
	}

	public void setDocType(final TextDocumentType doctype) {
		this.doctype = doctype;
	}

	/**
	 * @return the doc type at the current position of the document. Might be
	 *         different from {@link #getDocType()}, e.g., inside a pre-formatted
	 *         block.
	 */
	public TextDocumentType getLocalDocType() {
		return (elementsOpen.contains(PRE) ? TextDocumentType.FIXEDFONT : doctype);
	}

	public String getIndentString() {
		return indentString;
	}

	public void setIndentString(String indentString) {
		this.indentString = indentString;
	}

	/**
	 * Does nice indentation for non-plain text formats. The corresponding closing operation is {@link #ppEndLine()}.
	 * 
	 * @return the indented string
	 */
	public String ppBeginLine() {
		final String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = "";
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = repeat(indentString, minIndent + elementsOpen.size());
			break;
		}
		return ret;
	}

	/**
	 * Does nice indentation for non-plain text formats. Is the closing
	 * operation for {@link #ppBeginLine()}.
	 * 
	 * @return a newline for non-plain text formats
	 */
	public String ppEndLine() {
		return decideOnHtml(doctype, "", NL);
	}

	public String indent(String string) {
		return ppBeginLine() + string;
	}

	public String openHtmlElement(String element, String attributes) {
		final StringBuffer ret = new StringBuffer();
		ret.append(ppBeginLine());
		ret.append("<");
		ret.append(element);
		if (!CollectionUtil.isNullOrEmpty(attributes)) {
			if (!attributes.startsWith(" ")) {
				ret.append(" ");
			}
			ret.append(attributes);
		}
		ret.append(">");
		elementsOpen.push(element);
		return ret.toString();
	}

	public static final boolean INLINE = true;
	public static final boolean OWNLINE = false;

	public String closeHtmlElement(final String element) {
		return closeHtmlElement(element, OWNLINE);
	}

	/**
	 * Closes a HTML element.
	 * @param element the HTML element to close
	 * @param isInline either {@link #INLINE} or {@link #OWNLINE}.
	 * @return HTML snippet
	 */
	public String closeHtmlElement(final String element, final boolean isInline) {
		final String elo = elementsOpen.pop();
		Assert.assertEquals(elo, element);
		String ret = "</" + element + ">";
		if (!isInline) {
			ret = indent(ret);
		}
		return ret;
	}

	/**
	 * Own simple implementation of the JDK 1.7 method <code>Locale.toLanguageTag()</code>, to be available under JDK 1.6.
	 * @param locale The locale to output
	 * @return a BCP47 language tag representing the locale
	 */
	private String toLanguageTag(final Locale locale) {
		String ret = "";
		final String delimeter = "-";
		final String language = locale.getISO3Language();
		final String country = locale.getISO3Country();
		final String variant = locale.getVariant();
		ret += (!CollectionUtil.isNullOrEmpty(language) ? language : "");
		ret += (!CollectionUtil.isNullOrEmpty(country) ? delimeter + country : "");
		ret += (!CollectionUtil.isNullOrEmpty(variant) ? delimeter + variant : "");
		return ret;
	}

	public String beginDocument() {
		final StringBuffer text = new StringBuffer();
		elementsOpen.clear();
		String htmlAttribs = "";
		String xhtmlAttribs = htmlAttribute(XMLNS, "http://www.w3.org/1999/xhtml");
		if (null != locale) {
			final String lang = toLanguageTag(locale);
			htmlAttribs = htmlAttribs + htmlAttribute(LANG, lang);
			xhtmlAttribs = xhtmlAttribs + htmlAttribute(XML_LANG, lang);
		}
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			// Nothing to do;
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
			text.append(doctype.doctypeTag + ppEndLine());
			text.append(openHtmlElement("html", htmlAttribs) + ppEndLine());
			break;
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
			text.append(doctype.doctypeTag + ppEndLine());
			text.append(openHtmlElement("html", xhtmlAttribs) + ppEndLine());
			break;
		default:
			throw new IllegalStateException("Unknown DocType " + doctype);
		}
		return text.toString();
	}

	public String endDocument() {
		String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = "";
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
			ret = closeHtmlElement("html") + ppEndLine();
			Assert.assertEquals(0, elementsOpen.size());
			break;
		default:
			ret = "";
			break;
		}
		return ret;
	}

	/**
	 * Adds a HTML head element, if doctype is appropriate.
	 * 
	 * @param title
	 *            the (non-I18N'ed) title to show in the window frame
	 * @param script
	 *            any script to define for the HTML page
	 * @return a HTML snippet
	 */
	public String addHead(final String title, final String script) {
		final StringBuffer s = new StringBuffer();
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			// Nothing to do
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
			s.append(openHtmlElement("head", null) + ppEndLine());
			final String metaHttpEquiv = "<meta" + htmlAttribute(HTTP_EQUIV, "Content-Type")
					+ htmlAttribute(CONTENT, CONTENT_TYPE_HTML);
			final String cssFilename = null; // If needed, set to "dsatool.css"
			final String linkStylesheet;
			if (CollectionUtil.isNullOrEmpty(cssFilename)) {
				linkStylesheet = null;
			} else {
				linkStylesheet = "<link" + htmlAttribute(REL, "stylesheet") + htmlAttribute(TYPE, "text/css")
					+ htmlAttribute(HREF, cssFilename);
			}
			// Requirement from
			// https://developers.google.com/speed/docs/insights/ConfigureViewport
			final String metaViewport = "<meta" + htmlAttribute(NAME, "viewport")
					+ htmlAttribute(CONTENT, "width=device-width, initial-scale=1");
			if (doctype.isOldHtml()) {
				// Meta http-equiv is not supported by old JDK html-renderer.
				// :-(
				if (!CollectionUtil.isNullOrEmpty(linkStylesheet)) {
					s.append(indent(linkStylesheet + ">") + ppEndLine());
				}
				s.append(indent(metaViewport + ">") + ppEndLine());
			} else if (doctype.isXhtml()) {
				if (!CollectionUtil.isNullOrEmpty(linkStylesheet)) {
					s.append(indent(linkStylesheet + " />") + ppEndLine());
				}
				s.append(indent(metaViewport + " />") + ppEndLine());
				s.append(indent(metaHttpEquiv + " />") + ppEndLine());
			} else {
				if (!CollectionUtil.isNullOrEmpty(linkStylesheet)) {
					s.append(indent(linkStylesheet + "></link>") + ppEndLine());
				}
				s.append(indent(metaViewport + ">") + ppEndLine());
				s.append(indent(metaHttpEquiv + ">") + ppEndLine());
			}

			if (null != title) {
				s.append(indent("<title>" + title + "</title>") + ppEndLine());
			}
			// if (null != styleAdditions)
			// {
			// s.append("<style type=\"text/css\">" + ppEndLine());
			// s.append(styleAdditions);
			// s.append("</style>" + ppEndLine());
			// }
			if (null != script) {
				s.append("<script type=\"text/javascript\">" + ppEndLine());
				s.append(script);
				s.append("</script>" + ppEndLine());
			}
			s.append(closeHtmlElement("head") + ppEndLine());
			break;
		}
		return s.toString();
	}

	public String beginBody() {
		return beginHtmlElement(BODY);
	}

	public String endBody() {
		return endHtmlElement(BODY);
	}

	public String newLine() {
		return newLine(getLocalDocType());
	}

	public String newLineIfPlainText() {
		return newLineIfPlainText(getLocalDocType());
	}

	public String sgmlify(final String text) {
		return sgmlify(getLocalDocType(), text);
	}

	public String anchor(final String id, final String text) {
		return anchor(getLocalDocType(), id, text);
	}

	public String href(final String href, final String text) {
		return href(getLocalDocType(), href, text);
	}

	private String beginSectionHeading(final int depth) {
		Assert.assertEquals(SECTION_HEADING_NONE, sectionHeadingOpen);
		sectionHeadingOpen = depth;
		Assert.assertTrue((0 < depth) && (depth <= 6));
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			return HEADING_PRE_POSTFIX[depth - 1][0];
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			return ("<h" + depth + ">");
		}
	}

	private String endSectionHeading(final int depth) {
		Assert.assertEquals(depth, sectionHeadingOpen);
		sectionHeadingOpen = SECTION_HEADING_NONE;
		Assert.assertTrue((0 < depth) && (depth <= 6));
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			return HEADING_PRE_POSTFIX[depth - 1][1] + NL + NL;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			return "</h" + depth + ">";
		}
	}

	public String addSectionHeading(final int depth, final String content) {
		final StringBuffer ret = new StringBuffer();
		ret.append(beginSectionHeading(depth));
		ret.append(content);
		ret.append(endSectionHeading(depth));
		ret.append(ppEndLine());
		return ret.toString();
	}

	public String beginParagraph(final String htmlAttribs) {
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			return NL;
		case HTML401_TRANSITIONAL:
			return "";
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			return openHtmlElement("p", htmlAttribs) + ppEndLine();
		}
	}

	public String beginParagraph() {
		return beginParagraph(null);
	}

	public String endParagraph() {
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			return NL;
		case HTML401_TRANSITIONAL:
			return ppBeginLine() + "<p>" + ppEndLine();
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			return closeHtmlElement("p") + ppEndLine();
		}
	}

	public String beginPreformatted() {
		final String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = NL; // as beginParagraph
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = openHtmlElement(DIV, htmlAttribute(CLASS, "scrollable"))
					+ ppEndLine() 
					+ openHtmlElement(PRE, null)
					+ ppEndLine();
		}
		return ret;
	}

	public String endPreformatted() {
		final String ret;
		switch (doctype) {
		case TEXT:
		case FIXEDFONT:
			ret = NL; // as endParagraph
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = closeHtmlElement(PRE) 
					+ ppEndLine() 
					+ closeHtmlElement(DIV) 
					+ ppEndLine();
		}
		return ret;
	}

	public String beginBold() {
		return decideOnHtml(doctype, "", "<b>");
	}

	public String endBold() {
		return decideOnHtml(doctype, "", "</b>");
	}

	/**
	 * Defines all the key information about a column to show for a table model.
	 * 
	 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ColumnInfo implements Cloneable {
		/**
		 * Default constructor - useless in itself, at least set mode and getter
		 * afterwards.
		 */
		public ColumnInfo() {
			// Nothing to do
		}

		/**
		 * Constructor defining the minimum information for this object to be
		 * useful.
		 * 
		 * @param length
		 *            the length of the column in characters
		 * @param alignment
		 *            how to align the text in that column (one of
		 *            {@link #LEFT}, {@link #CENTER}, {@link #RIGHT}, or
		 *            {@link #UNDEFINED})
		 */
		public ColumnInfo(final int length, final int alignment) {
			this.columnDefaultLength = length;
			this.columnAlignment = alignment;
		}

		/**
		 * Constructor defining the minimum information for this object to be
		 * useful.
		 * 
		 * @param length
		 *            the length of the column in characters
		 * @param alignment
		 *            how to align the text in that column (one of
		 *            {@link #LEFT}, {@link #CENTER}, {@link #RIGHT}, or
		 *            {@link #UNDEFINED})
		 * @param heading
		 *            the column heading text
		 */
		public ColumnInfo(final int length, final int alignment, final String heading) {
			this.columnDefaultLength = length;
			this.columnAlignment = alignment;
			this.headingName = heading;
		}

		public ColumnInfo(final int length, final int alignment, final Messages messages, final String key) {
			this(length, alignment, messages, key, false);
		}

		/**
		 * Constructor defining the minimum information for this object to be
		 * useful.
		 * 
		 * @param length
		 *            the length of the column in characters
		 * @param alignment
		 *            how to align the text in that column (one of
		 *            {@link #LEFT}, {@link #CENTER}, {@link #RIGHT}, or
		 *            {@link #UNDEFINED})
		 * @param messages
		 *            the messages object to internationalize
		 * @param key
		 *            the i18n text key
		 * @param isWithHref
		 *            if for the key a href should be created
		 */
		public ColumnInfo(final int length, final int alignment, final Messages messages, final String key,
				final boolean isWithHref) {
			this.columnDefaultLength = length;
			this.columnAlignment = alignment;
			setHeadingKey(messages, key, isWithHref);
		}

		@Override
		public Object clone() {
			try {
				return super.clone();
			} catch (final CloneNotSupportedException ex) {
				logger.error("Cannot clone!", ex);
				return null;
			}
		}

		public String getHeadingName() {
			return headingName;
		}

		public String getHeadingToolTip() {
			return headingToolTip;
		}

		public String getHeadingHref() {
			return headingHref;
		}

		protected void setHeadingName(final String headingName) {
			this.headingName = headingName;
		}

		protected void setHeadingToolTip(final String headingToolTip) {
			this.headingToolTip = headingToolTip;
		}

		protected void setHeadingHref(final String headingHref) {
			this.headingHref = headingHref;
		}

		protected void setHeadingKey(final Messages messages, final String key, final boolean isWithHref) {
			this.headingName = messages.getText(key);
			this.headingToolTip = messages.getShortDescription(key);
			if (isWithHref) {
				this.headingHref = key;
			} else {
				this.headingHref = null;
			}
		}

		/** The default alignment, or {@link #UNDEFINED}. */
		public int columnAlignment = UNDEFINED;

		/**
		 * The default length to format the value, or {@link #UNDEFINED} if no
		 * extra formatting should happen.
		 */
		public int columnDefaultLength = UNDEFINED;

		/** The heading to show on top of the column. */
		protected String headingName = null;

		/** The internationalized heading to show on top of the column. */
		protected String headingToolTip = null;

		/** The href to link to on top of the column. */
		protected String headingHref = null;
	}

	/**
	 * Defines all the key information about a table to format.
	 * 
	 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class TableInfo {
		public static final String TABLE_COLUMN_SEPARATOR = "|";
		public static final char TABLE_FILLER = '_';
		public static final boolean IS_WITH_BORDER = true;

		public final TextDocumentBuilder doc;
		public final ColumnInfo[] columns;
		public boolean isWithBorder = IS_WITH_BORDER;
		public boolean isWithHeading = true;
		public String tableAttributes = null;
		public String tableColumnSeparator = TABLE_COLUMN_SEPARATOR;
		public char tableFiller = TABLE_FILLER;

		public TableInfo(final TextDocumentBuilder doc, final boolean isWithBorder, final boolean isWithHeading, final String tableAttributes, final ColumnInfo[] columns) {
			this.doc = doc;
			this.isWithBorder = isWithBorder;
			this.isWithHeading = isWithHeading;
			this.tableAttributes = tableAttributes;
			this.columns = columns;
		}

		public String beginRow() {
			return doc.tf().beginHtmlElement(TR);
		}

		public String endRow() {
			return doc.tf().endHtmlElement(TR);
		}

		/**
		 * Formats a value.
		 * 
		 * @param ci
		 *            the column info
		 * @param value
		 *            the value to show within the cell
		 * @param alignment
		 *            one of {@link #LEFT}, {@link #CENTER}, {@link #RIGHT}, or
		 *            {@link #UNDEFINED} for using default column alignment
		 * @return the formatted value
		 */
		@SuppressWarnings("unused")
		private String format(final ColumnInfo ci, final String value, int alignment) {
			final String ret;
			if (ci.columnDefaultLength < 0) {
				// No formatting for length
				ret = value;
			} else {
				if (alignment < 0) {
					alignment = ci.columnAlignment;
				}
				ret = TextFormatter.format(value, ci.columnDefaultLength, alignment, tableFiller).toString();
			}
			return ret;
		}

		public String format(ColumnInfo ci, TableCellInfo tci) {
			final StringBuffer s = new StringBuffer();
			final TextFormatter tf = doc.tf();
			final Messages messages = doc.messages();
			final TextDocumentType localDocType = tf.getLocalDocType();
			final int rowspan = tci.getRowspan();
			final int colspan = tci.getColspan();
			final String elem = tci.getHtmlElementName();
			String scope = tci.getHeadingScope();
			int alignment = tci.getAlignment();
			final Object cellData = tci.getContent();
			String toolTip = null;
			final String value;
			String headingName = null; // Was an optional parameter to format(ColumnInfo)
			if (null == cellData) {
				value = null;
			} else if (cellData instanceof ColumnInfo) {
				final ColumnInfo tcici = (ColumnInfo) cellData;
				if (null == headingName) {
					final String headingHref = tcici.getHeadingHref();
					if (CollectionUtil.isNullOrEmpty(headingHref)) {
						headingName = tcici.getHeadingName();
					} else {
						headingName = href(localDocType, "#" + headingHref, tcici.getHeadingName());
					}
					toolTip = tcici.getHeadingToolTip();
				}
				if (null != headingName) {
					tcici.columnDefaultLength = Math.max(tcici.columnDefaultLength, headingName.length());
				}
				alignment = CENTER;
				value = headingName;
				scope = COL;
			} else if (cellData instanceof I18NStringWithFillIns) {
				final I18NStringWithFillIns msgKey = (I18NStringWithFillIns) cellData;
				value = tf.href("#" + msgKey.getKey(), messages.getText(msgKey));
				toolTip = messages.getShortDescription(msgKey);
			} else {
				value = cellData.toString();
			}
			if (UNDEFINED == alignment) {
				alignment = ci.columnAlignment;
			}
			switch (localDocType) {
			case TEXT:
			case FIXEDFONT:
				if (SPAN_DEFAULT != rowspan) {
					throw new IllegalArgumentException("rowspan " + rowspan + " not supported for doctype " + localDocType);
				}
				final String content;
				if (ci.columnDefaultLength < 0) {
					// No formatting for length
					content = value;
				} else {
					if (alignment < 0) {
						alignment = ci.columnAlignment;
					}
					content = TextFormatter.format(value, ci.columnDefaultLength, alignment, tableFiller).toString();
				}
				s.append(content);
				s.append(tableColumnSeparator);
				break;
			case HTML401_TRANSITIONAL:
			case HTML401_STRICT:
			case HTML5:
			case XHTML10_TRANSITIONAL:
			case XHTML10_STRICT:
			case XHTML11:
			default:
				s.append("<" + elem);
				String classes = "";
				if (0 <= alignment) {
					if (localDocType.isOldHtml()) {
						s.append(htmlAttribute(ALIGN, TextFormatter.ALIGNMENT_NAME[alignment]));
					} else {
//						s.append(htmlAttribute(STYLE, "text-align:" + TextFormatter.ALIGNMENT_NAME[alignment] + ";"));
						classes += TextFormatter.ALIGNMENT_CLASS_NAME[alignment] + " ";
					}
				}
				classes = classes.trim();
				if (!CollectionUtil.isNullOrEmpty(classes)) {
					s.append(htmlAttribute(CLASS, classes));
				}
				if (SPAN_DEFAULT != colspan) {
					s.append(htmlAttribute(COLSPAN, String.valueOf(colspan)));
				}
				if (SPAN_DEFAULT != rowspan) {
					s.append(htmlAttribute(ROWSPAN, String.valueOf(rowspan)));
				}
				if (!CollectionUtil.isNullOrEmpty(toolTip)) {
					s.append(htmlAttribute(TITLE, toolTip));
				}
				if (!CollectionUtil.isNullOrEmpty(scope)) {
					s.append(htmlAttribute(SCOPE, scope));
				}
				s.append(">");
				s.append(TextFormatter.tableCellContent(localDocType, value));
				s.append("</" + elem + ">");
				break;
			}
			return s.toString();
		}
	}

	/**
	 * Defines all the key information about a table to format.
	 * TODO: Currently work in progress and yet unused.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class TableCellInfo {
		private Object content;
		private int alignment = UNDEFINED;
		private int colspan = 1;
		private int rowspan = 1;
		private String headingScope = null;
		public TableCellInfo(final Object content, final String headingScope, final int alignment, final int colspan, final int rowspan) {
			this.content = content;
			this.alignment = alignment;
			this.headingScope = headingScope;
			this.colspan = colspan;
			this.rowspan = rowspan;
		}
		public TableCellInfo(final Object content) {
			this.content = content;
		}
		public TableCellInfo(final Object content, final int alignment) {
			this.content = content;
			this.alignment = alignment;
		}
		public Object getContent() {
			return content;
		}
		public int getAlignment() {
			return alignment;
		}
		public int getColspan() {
			return colspan;
		}
		public int getRowspan() {
			return rowspan;
		}
		public String getHeadingScope() {
			return headingScope;
		}
		public boolean isHeading() {
			return (null != headingScope);
		}
		public String getHtmlElementName() {
			return (isHeading() ? "th" : "td");
		}
	}

	public TableInfo tableInfo = null;

	public String beginTable(final String id, final TableInfo tableInfo) {
		this.tableInfo = tableInfo;
		String ret;
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			ret = "";
			break;
		case HTML5:
			ret = openHtmlElement(DIV, htmlAttribute(CLASS, "scrollable"))
					+ ppEndLine()
					+ openHtmlElement(TABLE, 
							(null == id ? "" : htmlAttribute(ID, id)) 
							+ (null == tableInfo.tableAttributes ? "" : tableInfo.tableAttributes))
					+ ppEndLine();
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = openHtmlElement(DIV, htmlAttribute(CLASS, "scrollable"))
					+ ppEndLine()
					+ openHtmlElement(TABLE, 
							(null == id ? "" : htmlAttribute(ID, id))
							+ htmlAttribute(BORDER, (tableInfo.isWithBorder ? "1" : "0")) 
							+ htmlAttribute(CELLSPACING, "0") 
							+ htmlAttribute(CELLPADDING, "3")
							+ (null == tableInfo.tableAttributes ? "" : tableInfo.tableAttributes))
					+ ppEndLine();
			break;
		}
		return ret;
	}

	public String endTable() {
		tableInfo = null;
		String ret;
		switch(getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			ret = "";
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = closeHtmlElement(TABLE) + ppEndLine() + closeHtmlElement(DIV) + ppEndLine();
			break;
		}
		return ret;
	}

	public String beginTableHead() {
		return beginHtmlElement(THEAD);
	}

	public String endTableHead() {
		return endHtmlElement(THEAD);
	}

	public String beginTableBody() {
		return beginHtmlElement(TBODY);
	}

	public String endTableBody() {
		return endHtmlElement(TBODY);
	}

	public String beginTableRow() {
		return tableInfo.beginRow();
	}

	public String endTableRow() {
		return tableInfo.endRow();
	}

	public String beginHtmlElement(final String element) {
		String ret;
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			ret = "";
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = openHtmlElement(element, null) + ppEndLine();
			break;
		}
		return ret;
	}

	public String endHtmlElement(final String element) {
		String ret;
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			if ((THEAD.equals(element)) || (TBODY.equals(element)) || (BODY.equals(element))) {
				ret = "";
			} else {
				ret = NL;
			}
			break;
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			ret = closeHtmlElement(element) + ppEndLine();
			break;
		}
		return ret;
	}

	public String formatTableHeading(final ColumnInfo[] ci) {
		final StringBuffer ret = new StringBuffer();
		ret.append(beginTableHead());
		ret.append(beginTableRow());
		for (int c = 0; c < ci.length; c++) {
			ret.append(formatTableHeadingCell(ci[c], null));
		}
		ret.append(endTableRow());
		ret.append(endTableHead());
		return ret.toString();
	}

	/**
	 * Formats a table heading output.
	 * 
	 * @param ci
	 *            the column info
	 * @param headingName
	 *            the heading. If <code>null</code>, the heading name from the column info is used.
	 * @return the formatted table heading
	 */
	public String formatTableHeadingCell(final ColumnInfo ci, final String headingName) {
		final StringBuffer ret = new StringBuffer();
		ret.append(ppBeginLine());
		final ColumnInfo ci2;
		if (null != headingName) {
			ci2 = (ColumnInfo) ci.clone();
			ci2.setHeadingName(headingName);
			ci2.setHeadingToolTip(null);
		} else {
			ci2 = ci;
		}
		ret.append(tableInfo.format(ci, new TableCellInfo(ci2, COL, CENTER, SPAN_DEFAULT, SPAN_DEFAULT)));
		ret.append(ppEndLine());
		return ret.toString();
	}

	public String formatTableCell(final ColumnInfo ci, final String text, final int alignment) {
		return tableInfo.format(ci, new TableCellInfo(text, alignment));
	}

	public String formatTableCell(final ColumnInfo ci, final String text) {
		return tableInfo.format(ci, new TableCellInfo(text));
	}

	public String formatTableCell(final ColumnInfo ci, final TableCellInfo tci) {
		return tableInfo.format(ci, tci);
	}

	/**
	 * Builds a text spanning the full row, depending on the doctype, containing
	 * a title part and a definition part.
	 * 
	 * @param titleText
	 *            The title part of the text.
	 * @param definitionText
	 *            The definition part of the text.
	 * @return the text.
	 */
	public String formatTableFullTextRow(final String titleText, final String definitionText) {
		final StringBuffer ret = new StringBuffer();
		String value = titleText;
		if (null != definitionText) {
			value += ": " + definitionText;
		}
		ret.append(beginTableRow());
		// FIXME: Should be done inside tableInfo.format, based on the column info array
		int length = -1;
		for (int c = 0; c < tableInfo.columns.length; c++) {
			length += tableInfo.columns[c].columnDefaultLength + 1;
		}
		final ColumnInfo ci = new ColumnInfo(length, TextFormatter.LEFT);
		ret.append(tableInfo.format(ci, new TableCellInfo(value, null, TextFormatter.LEFT, tableInfo.columns.length, TextFormatter.SPAN_DEFAULT)));
		ret.append(endTableRow());
		return ret.toString();
	}

	/**
	 * If in fixed width mode, adds characters to a given string to achieve the
	 * right width; if in any other mode, returns <em>val</em> unchanged.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the string (and to the inverse, where to
	 *            put the blanks)
	 * @param fillChar
	 *            this char will be repeatedly inserted as filling
	 * @return the formatted text
	 */
	public String formatIfPlainText(final String val, final int width, final int alignment, final char fillChar) {
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			return format(val, width, alignment, fillChar).toString();
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			return val;
		}
	}

	/**
	 * If in fixed width mode, blanks to a given string to achieve the right
	 * width; if in any other mode, returns <em>val</em> unchanged.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the string (and to the inverse, where to
	 *            put the blanks)
	 * @return the formatted text
	 */
	public String formatIfPlainText(final String val, final int width, final int alignment) {
		return formatIfPlainText(val, width, alignment, ' ');
	}

	/**
	 * If in fixed width mode, adds blanks to a given string to achieve the
	 * right width. The blanks are always added at the end. If in any other
	 * mode, returns <em>val</em> unchanged.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @return the formatted text
	 */
	public String formatIfPlainText(final String val, final int width) {
		return formatIfPlainText(val, width, LEFT);
	}

	/**
	 * If in fixed width mode, adds blanks to the output of an integer to
	 * achieve the right width. If in any other mode, returns <em>val</em> as
	 * unformatted string.
	 * 
	 * @param val
	 *            the value to format
	 * @param width
	 *            the length in characters that's to achieve
	 * @param alignment
	 *            tells where to put the integer (and to the inverse, where to
	 *            put the blanks)
	 * @return the formatted text
	 */
	public String formatIfPlainText(final int val, final int width, final int alignment) {
		return formatIfPlainText(Integer.toString(val), width, alignment);
	}

	/**
	 * If in fixed width mode, adds blanks to the output of an integer to
	 * achieve the right width. If in any other mode, returns <em>val</em> as
	 * unformatted string.
	 * 
	 * @param val
	 *            the value to format
	 * @param width
	 *            the length in characters that's to achieve
	 * @return the formatted text
	 */
	public String formatIfPlainText(final int val, final int width) {
		return formatIfPlainText(val, width, RIGHT);
	}

	/**
	 * If in fixed width mode, adds blanks to the output of a double to achieve
	 * the right width. If in any other mode, returns <em>val</em> as
	 * unformatted string.
	 * 
	 * @param val
	 *            the value to format
	 * @param width
	 *            the length in characters that's to achieve
	 * @param afterDotPositions
	 *            so many positions after the dot are shown
	 * @param alignment
	 *            tells where to put the number (and to the inverse, where to
	 *            put the blanks)
	 * @return the formatted text
	 */
	public String formatIfPlainText(double val, final int width, final int afterDotPositions, final int alignment) {
		for (int i = 0; i < afterDotPositions; i++) {
			val *= 10.0;
		}
		val = Math.round(val);
		for (int i = 0; i < afterDotPositions; i++) {
			val /= 10.0;
		}
		switch (getLocalDocType()) {
		case TEXT:
		case FIXEDFONT:
			return format(Double.toString(val), width, alignment).toString();
		case HTML401_TRANSITIONAL:
		case HTML401_STRICT:
		case HTML5:
		case XHTML10_TRANSITIONAL:
		case XHTML10_STRICT:
		case XHTML11:
		default:
			return Double.toString(val);
		}
	}

	/**
	 * If in fixed width mode, adds blanks to the output of a double to achieve
	 * the right width. If in any other mode, returns <em>val</em> as
	 * unformatted string. The blanks are always added to the left.
	 * 
	 * @param val
	 *            the string
	 * @param width
	 *            the length in characters that's to achieve
	 * @param afterDotPositions
	 *            so many positions after the dot are shown
	 * @return the formatted text
	 */
	public String formatIfPlainText(final double val, final int width, final int afterDotPositions) {
		return formatIfPlainText(val, width, afterDotPositions, RIGHT);
	}

	public int getMinIndent() {
		return minIndent;
	}

	public void setMinIndent(final int indent) {
		this.minIndent = indent;
	}

	private TextDocumentType doctype;

	private Locale locale = null;

	private int sectionHeadingOpen = SECTION_HEADING_NONE;

	private String indentString = DEFAULT_INDENT;

	private Stack<String> elementsOpen = new Stack<String>();
	private int minIndent;
}
