package de.ped.tools;

/**
 * Defines what kind of neighboring relationships a field on the board can have.
 * 
 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
 */
public enum TopologyMode {
	/** Original Empire topology - Squares, all 8 neighbors. */
	SQUARE8(8, TopologyMode.ALL_MODES),
	/** Squares, only 4 direct neighbors. */
	SQUARE4(4, TopologyMode.ALL_MODES),

	/** 
	 * Triangles, with coordinates as defined like: 
	 * <pre>
	 * \ 0,0/\ 2,0/\ 4,0/
	 *  \  /  \  /  \  / 
	 *   \/ 1,0\/ 3,0\/  
	 *-------------------
	 *   /\ 1,1/\ 3,1/\  
	 *  /  \  /  \  /  \ 
	 * / 0,1\/ 2,1\/ 4,1\
	 * </pre>
	 */
	TRIANGLE(3, TopologyMode.HORIZONTAL_MODE),

	/** 
	 * Hexagons, with coordinates as defined like: 
	 * <pre>
	 * (0,0)     (2,0)
	 *      (1,0)     (3,0)
	 * (0,1)     (2,1)
	 *      (1,1)     (3,1)
	 * (0,2)     (2,2)
	 *      (1,2)     (3,2)
	 * </pre>
	 * See {@link PlayfieldTiling#IS_HEX_COMPACT} for exact neighboring relationships.
	 */
	HEXAGON(6, TopologyMode.VERTICAL_MODE),

	/** 
	 * Hexagons, with coordinates as defined like: 
	 * <pre>
	 * (0,0) (1,0) (2,0) (3,0)
	 *    (0,1) (1,1) (2,1) (3,1)
	 * (0,2) (1,2) (2,2) (3,2)
	 *    (0,3) (1,3) (2,3) (3,3)
	 * </pre>
	 * See {@link PlayfieldTiling#IS_HEX_COMPACT} for exact neighboring relationships.
	 */
	HEXAGON_H(6, TopologyMode.HORIZONTAL_MODE);

	public static final int HORIZONTAL_MODE = 1;
	public static final int VERTICAL_MODE = 2;
	public static final int ALL_MODES = HORIZONTAL_MODE | VERTICAL_MODE;
	public static final String KEY = "GameProperties.MapTopologyMode";

	private TopologyMode(final int neighbors, final int modes) {
		this.neighbors = neighbors;
		this.modes = modes;
	}

	public final int neighbors;
	public final int modes;
}