package de.ped.tools.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.InputVerifier;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JTable;
import javax.swing.KeyStroke;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.Assert;
import de.ped.tools.CollectionUtil;
import de.ped.tools.FinishableThread;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.ResourceFinder.Folder;
import de.ped.tools.LocalizedString;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextDocumentType;
import de.ped.tools.TextFormatter;
import de.ped.tools.TextTree;
import de.ped.tools.TextTreeNodeType;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * The main window of the application.
 * Laid out as abstract class. Contains loading of a resource for I18N, several standard actions like exit, help and info, 
 * support for distinguishing applet and application behavior on exit, and support for creating a HTML help and plain text info dialog.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class ApplicationMainWindow extends JFrame implements KeyListener {
	private static final long serialVersionUID = 1L;

	protected final static Logger logger = Logger.getRootLogger();
	{
		logger.setLogLevel(Level.INFO);
	}

	/** 
	 * The action to select a locale. 
	 * Locales are always displayed in their own locale. The rationale behind this decision is
	 * the scenario that the user may select a language he is unfamiliar with and then is unable
	 * to identify the representation of a known locale in that unknown language.
	 * 
	 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class SwitchLocaleAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public SwitchLocaleAction(final Locale locale) {
			super("Menu.Locales." + locale.toString(), ApplicationMainWindow.this, ResourceFinder.Folder.COMMON);
			this.locale = locale;
			putValue(Action.NAME, locale.getDisplayName(locale));
			setLogger(logger);
			logger.trace("SwitchLocaleAction(" + getMessageKeyPrefix() + ")");
		}

		public Locale getLocale() {
			return locale;
		}

		public JRadioButtonMenuItem getRadioButtonMenuItem() {
			return menuItem;
		}

		public void setRadioButtonMenuItem(final JRadioButtonMenuItem menuItem) {
			this.menuItem = menuItem;
		}

		public void setSelected(final Locale localeToSet) {
			if (null != menuItem) {
				menuItem.setSelected(locale.equals(localeToSet));
			}
			logger.trace("SwitchLocaleAction(" + locale + ").setSelected(" + localeToSet + ")");
		}

		public final Locale locale;
		private JRadioButtonMenuItem menuItem;
	}

	/**
	 * An action to switch the Swing look and feel.
	 * 
	 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public class LAFAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public LAFAction(final String className, final String i18nKey) {
			super(i18nKey, ApplicationMainWindow.this, null);
			this.className = className;
		}

		public LAFAction(final UIManager.LookAndFeelInfo lafInfo) {
			super(null, ApplicationMainWindow.this, null);
			this.lafInfo = lafInfo;
		}

		public String getClassName() {
			if (null != lafInfo) {
				return lafInfo.getClassName();
			}
			return className;
		}

		public void setClassName(final String className) {
			this.className = className;
		}

		public JRadioButtonMenuItem getRadioButtonMenuItem() {
			return menuItem;
		}

		public void setRadioButtonMenuItem(final JRadioButtonMenuItem menuItem) {
			this.menuItem = menuItem;
			this.menuItem.setText(lafInfo.getName());
		}

		UIManager.LookAndFeelInfo lafInfo;
		String className;
		JRadioButtonMenuItem menuItem;
	}

	/**
	 * The action to show the info about the program.
	 * 
	 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	class AboutAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public AboutAction() {
			super("Action.Info", ApplicationMainWindow.this, ResourceFinder.Folder.COMMON, true);
		}

		@Override
		protected Object[] getI18NParams() {
			return new String[] { getMessages().getApplicationName(Messages.I18NedOrNot.I18N), };
		}
	}

	/**
	 * The action to start a new game.
	 * 
	 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class StartNewGameAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public StartNewGameAction(final ApplicationMainWindow parent) {
			super("Action.New", parent, ResourceFinder.Folder.COMMON, true);
		}
	}

	/**
	 * Exits a running game.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ExitGameAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public ExitGameAction(final ApplicationMainWindow parent) {
			super("Action.ExitGame", parent, ResourceFinder.Folder.COMMON);
		}
	}

	/**
	 * Quits the program.
	 * 
	 * @author Copyright (c) 2021 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class QuitAction extends PedAction {
		private static final long serialVersionUID = 1L;

		public QuitAction(final ApplicationMainWindow parent) {
			super("Action.Quit", parent, ResourceFinder.Folder.COMMON);
			updateEnabled();
		}

		@Override
		public boolean isToBeEnabled() {
			return getParent().isExitable();
		}
	}

	protected Action quitAction;
	protected Action helpAction;
	protected Action aboutAction;
	protected SwitchLocaleAction[] switchLocaleActions;
	protected List<Action> allActions = new ArrayList<Action>();
	protected UIManager.LookAndFeelInfo[] lookAndFeelInfos;
	protected Action lafCrossPlatformAction;
	protected LAFAction lafSystemAction;
	protected LAFAction[] lafActions;
	protected StartNewGameAction startNewGameAction;
	protected ExitGameAction exitGameAction;
	protected JRadioButtonMenuItem miLAFCrossPlatform;
	protected JRadioButtonMenuItem miLAFSystem;

	private ApplicationEnvironment appEnv;

	public ApplicationMainWindow(final ApplicationEnvironment appEnv) {
		this.appEnv = appEnv;
		final Messages messages = getMessages();
		final String title = (messages == null ? null : messages.getApplicationTitle());
		super.setTitle(title);
		final boolean isExitable = isExitable();
		try {
			setDefaultCloseOperation(isExitable ? JFrame.EXIT_ON_CLOSE : JFrame.DO_NOTHING_ON_CLOSE);
		} catch (final SecurityException ex) {
			logger.debug("Could not set the default close operation. Missing some rights?");
		}
		quitAction = new QuitAction(this);
		helpAction = new PedAction("Action.Help", ApplicationMainWindow.this, ResourceFinder.Folder.COMMON, true);
		aboutAction = new AboutAction();
		lafCrossPlatformAction = new LAFAction(UIManager.getCrossPlatformLookAndFeelClassName(), "Action.LAFCrossPlatform");
		lafSystemAction = new LAFAction(UIManager.getSystemLookAndFeelClassName(), "Action.LAFSystem");
		if (null != messages) {
			Locale[] availableLocales = messages.getAvailableLocales();
			if (null != availableLocales) {
				switchLocaleActions = new SwitchLocaleAction[availableLocales.length];
				for (int i = 0; i < availableLocales.length; i++) {
					Locale aLocale = availableLocales[i];
					switchLocaleActions[i] = new SwitchLocaleAction(aLocale);
				}
			}
		}
		startNewGameAction = new StartNewGameAction(this);
		exitGameAction = new ExitGameAction(this);
	}

	public ApplicationEnvironment getApplicationEnvironment() {
		return appEnv;
	}

	/**
	 * @return <code>true</code> if the project can be exited - for example, an applet is not by default
	 */
	public boolean isExitable() {
		return appEnv.isExitable();
	}

	public Messages getMessages() {
		return getApplicationEnvironment().messages();
	}

	/**
	 * Will be called whenever a {@link PedAction} was selected to be performed.
	 * @param action the action picked to perform
	 * @param event the UI event that lead to firing this
	 */
	public void pedAbstractActionPerformed(final PedAction action, final ActionEvent event) {
		logger.trace("pedAbstractActionPerformed(" + action + "," + event + ")");
		if (action == quitAction) {
			logger.info("Thank you for playing.");
			ApplicationOrApplet.exit();
		} else if (action instanceof SwitchLocaleAction) {
			switchLocale(((SwitchLocaleAction) action).locale);
		} else if (action instanceof LAFAction) {
			setLookAndFeel(((LAFAction) action).getClassName());
		} else if (action == helpAction) {
			showHelpBox();
		} else if (action == aboutAction) {
			showAboutBox();
		} else if (action == startNewGameAction) {
			startNewGame();
		} else if (action == exitGameAction) {
			exitGame();
		}
	}

	public JRadioButtonMenuItem createRadioButtonMenuItem(final ButtonGroup buttonGroup, final Action action) {
		JRadioButtonMenuItem menuItem = new JRadioButtonMenuItem(action);
		buttonGroup.add(menuItem);
		return menuItem;
	}

	public JIconComboBox<Object> createComboBox(final Object[] items) {
		return createComboBox(items, true, null);
	}

	public JIconComboBox<Object> createComboBox(final Object[] items, final boolean isHelpWithElements) {
		return createComboBox(items, isHelpWithElements, null);
	}

	/**
	 * Creates a combo box filled with {@link LocalizedString} objects representing the given items. 
	 * Additionally checks for png image resources named like the message keys and adds them as icons, if existing.
	 * @param items an array of objects which are either of type {@link LocalizedString}, or at least return the i18n message key within {@link Object#toString()}.
	 * @param isHelpWithElements if help for this combo box should list all the elements within this combo
	 * @param folder the resource folder where to search for the icon resources of the items, or <code>null</code> to indicate that there are no icons to display
	 * @return a combo box representing the given objects.
	 */
	public JIconComboBox<Object> createComboBox(final Object[] items, final boolean isHelpWithElements, ResourceFinder.Folder folder) {
		final Messages messages = getMessages();
		final JIconComboBox<Object> ret = new JIconComboBox<Object>();
		for (int i = 0; i < items.length; i++) {
			final Object item = items[i];
			final LocalizedString itemToAdd;
			if (item instanceof LocalizedString) {
				itemToAdd = (LocalizedString) item;
			} else {
				itemToAdd = new LocalizedString(messages, item.toString());
			}
			if (null != folder) {
				final String imageFilename = itemToAdd.getImageFilename();
				logger.trace("Looking for icon " + imageFilename + " for item " + itemToAdd.key);
				final ResourceFinder res = getApplicationEnvironment().resources();
				if (null != res) {
					final java.net.URL url = res.getURLToResource(imageFilename + ".png", folder);
					ret.addIcon(itemToAdd, new ImageIcon(url));
				}
			}
			ret.addItem(itemToAdd);
		}
		ret.setHelpWithElements(isHelpWithElements);
		ret.setFolder(folder);
		return ret;
	}

	/**
	 * Creates a combo box filled with{@link LocalizedString} objects representing values from an enum type. 
	 * Additionally checks for png image resources named like the message keys and adds them as icons, if existing. 
	 * @param values the values from an enum type
	 * @param prefix the prefix of the message key; the names of the enum elements are appended to the prefix to find the i18n'ed text
	 * @param imageFilenamePrefix the prefix of the image file name; the names of the enum elements are appended to the prefix to find the i18n'ed text
	 * @param folder the resource folder where to search for the icon resources of the items, or <code>null</code> to indicate that there are no icons to display
	 * @return a combo box representing the given enum values.
	 */
	public JIconComboBox<?> createComboBox(final Enum<?>[] values, final String prefix, final String imageFilenamePrefix, ResourceFinder.Folder folder) {
		final Messages messages = getMessages();
		final LocalizedString[] strs = new LocalizedString[values.length];
		for (int i = 0; i < values.length; i++) {
			final String imageFilename = (null == imageFilenamePrefix ? null
					: imageFilenamePrefix + "." + values[i].name());
			strs[i] = new LocalizedString(messages, prefix + "." + values[i].name(), imageFilename);
		}
		return createComboBox(strs, true, folder);
	}

	public JButton createButton(final String key) {
		return createButton(new I18NStringWithFillIns(key, getMessages().getUnknownString(key), null));
	}

	public JButton createButton(final I18NStringWithFillIns istr) {
		String name = istr.getAlternative();
		final JButton button = new JButton(name);
		putValues(button, istr);
		return button;
	}

	public JRadioButton createRadioButton(final I18NStringWithFillIns istr) {
		String name = istr.getAlternative();
		final JRadioButton button = new JRadioButton(name);
		putValues(button, istr);
		return button;
	}

	public JLabel createLabel(final String i18nKeyPrefix) {
		final JLabel ret = createLabel(getMessages().getText(i18nKeyPrefix), getMessages().getShortDescription(i18nKeyPrefix));
		return ret;
	}

	public static JLabel createLabel(final String text, final String tooltip, final GridBagLayout gridbag, final GridBagConstraints constraint) {
		final JLabel label = createLabel(text, tooltip);
		constraint.gridwidth = 1;
		gridbag.setConstraints(label, constraint);
		return label;
	}

	public static JLabel createLabel(final String text, final String tooltip) {
		final JLabel label = new JLabel(text);
		if (null != tooltip) {
			label.setToolTipText(tooltip);
		}
		return label;
	}

	public static JPanel createPanel(final Component component) {
		final JPanel p = new JPanel();
		p.setLayout(new FlowLayout(FlowLayout.LEFT));
		p.add(component);
		return p;
	}

	public static JPanel createPanel(final Component component, final GridBagLayout gridbag, final GridBagConstraints constraint) {
		final JPanel p = createPanel(component);
		constraint.gridwidth = GridBagConstraints.REMAINDER; // end row
		gridbag.setConstraints(p, constraint);
		return p;
	}

	public JMenuBar createJMenuBar() {
		JMenuBar mb = new JMenuBar();
		mb.setName("Menu");
		return mb;
	}

	public PedJMenu createJMenu(final String key) {
		return createJMenu(key, null);
	}

	/**
	 * Creates a JMenu object and sets the text, mnemonic, tool tip and icon for it, based on the message key.
	 * @param key the message key to internationalize the JMenu object
	 * @param resourceFolder the resource folder to search for an icon image fitting to the key, or <code>null</code> for no icon
	 * @return the JMenu object
	 */
	public PedJMenu createJMenu(final String key, final ResourceFinder.Folder resourceFolder) {
		PedJMenu menu = new PedJMenu();
		putValues(menu, key);
		if (null != resourceFolder) {
			menu.setResourceFolder(resourceFolder);
			final java.net.URL url = getApplicationEnvironment().resources().getURLToResource(menu.getResourceFilename(), menu.getResourceFolder());
			menu.setIcon(new ImageIcon(url));
		}
		return menu;
	}

	/**
	 * Sets the text, name, tool tip and mnemonic for the button.
	 * 
	 * @param button
	 *            the button to change
	 * @param i18n
	 *            the internationalized string object to get the values from
	 */
	public void putValues(final AbstractButton button, final I18NStringWithFillIns i18n) {
		final Messages messages = getMessages();
		putValues((JComponent) button, i18n);
		int mnemonic = 0;
		if (null != messages) {
			final String mnemonicDesc = messages.getMnemonic(i18n.getKey());
			if (!CollectionUtil.isNullOrEmpty(mnemonicDesc)) {
				final KeyStroke stroke = KeyStroke.getKeyStroke(mnemonicDesc);
				if (null != stroke) {
					mnemonic = stroke.getKeyCode();
				}
			}
		}
		button.setText(button.getName());
		button.setMnemonic(mnemonic);
	}

	public void putValues(final JComponent button, final I18NStringWithFillIns i18n) {
		final Messages messages = getMessages();
		final String name;
		final String desc;
		if (null == messages) {
			name = i18n.getAlternative();
			desc = null;
		} else {
			name = messages.getText(i18n.getKey(), i18n.getAlternative());
			desc = messages.getShortDescription(i18n.getKey());
		}
		button.setName(name);
		button.setToolTipText(desc);
	}

	/**
	 * Sets the name, accelerator key and description (tool tip) for the action.
	 * 
	 * @param action
	 *            the action to change
	 * @param key
	 *            the message key to get the new texts from
	 */
	public void putValues(final Action action, final String key) {
		if (action instanceof PedAction) {
			PedAction a = (PedAction) action;
			a.putValue(PedAction.MESSAGE_KEY, key);
		}
		final Messages messages = getMessages();
		String text = getI18NString(action, key, Messages.TEXT_POSTFIX);
		action.putValue(PedAction.PURE_NAME, text);
		if (action instanceof PedAction) {
			PedAction a = (PedAction) action;
			if (a.isStartingDialog()) {
				text += messages.getString("Action.DialogIndicator", "...");
			}
		}
		action.putValue(Action.NAME, text);
		final String shortDesc = getI18NString(action, key, Messages.SHORT_DESCRIPTION_POSTFIX);
		final String longDesc = getI18NString(action, key, Messages.LONG_DESCRIPTION_POSTFIX);
		final String keystroke = getI18NString(action, key, Messages.KEYSTROKE_POSTFIX);
		if (null != keystroke) {
			action.putValue(Action.ACCELERATOR_KEY, KeyStroke.getKeyStroke(keystroke));
		}
		if (null != shortDesc) {
			action.putValue(Action.SHORT_DESCRIPTION, shortDesc);
		}
		if (null != longDesc) {
			action.putValue(Action.LONG_DESCRIPTION, longDesc);
		}
		//        action.putValue(Action.DISPLAYED_MNEMONIC_INDEX_KEY, 0); 
	}

	private String getI18NString(final Action action, final String key, final String postfix) {
		String text;
		final Messages messages = getMessages();
		if (action instanceof SwitchLocaleAction) {
			if (Messages.TEXT_POSTFIX.equals(postfix)) {
				text = (String) action.getValue(Action.NAME);
			} else {
				text = null;
			}
		} else if (action instanceof PedAction) {
			PedAction a = (PedAction) action;
			Object[] params = a.getI18NParams();
			if (null == params) {
				text = messages.getStringOr(a.getMessageKeyPrefix() + postfix, "", params);
			} else {
				text = messages.getStringOr(a.getMessageKeyPrefix() + postfix, "%0", params);
			}
		} else {
			text = messages.getStringOr(key + postfix, null, null);
		}
		return text;
	}

	/**
	 * Sets the name, accelerator key and description (tool tip) for the action.
	 * 
	 * @param action
	 *            the action to change
	 */
	public void putValues(final PedAction action) {
		putValues(action, action.getMessageKeyPrefix());
	}

	/**
	 * Sets the text, mnemonic and tool tip for the menu, based on the message
	 * key.
	 * 
	 * @param menu
	 *            the JMenu object
	 * @param key
	 *            the message key to get the texts from
	 */
	public void putValues(final JMenu menu, final String key) {
		final Messages messages = getMessages();
		if (null != messages) {
			final String text = messages.getText(key);
			final String mnemonic = messages.getMnemonic(key);
			final String desc = messages.getShortDescription(key);
			menu.setName(key);
			menu.setText(text);
			if (!CollectionUtil.isNullOrEmpty(mnemonic)) {
				menu.setMnemonic(mnemonic.charAt(0));
			}
			if (null != desc) {
				menu.setToolTipText(desc);
			}
		}
	}

	public static void adjustRowHeightsToRenderers(final JTable table) {
		try {
			for (int row = 0; row < table.getRowCount(); row++) {
				int rowHeight = table.getRowHeight();
				for (int column = 0; column < table.getColumnCount(); column++) {
					final Component comp = table.prepareRenderer(table.getCellRenderer(row, column), row, column);
					rowHeight = Math.max(rowHeight, comp.getPreferredSize().height);
				}
				table.setRowHeight(row, rowHeight);
			}
		} catch (final ClassCastException e) {
			// Nothing to do
		}
	}

	/**
	 * Adds a row to the panel in {@link GridBagLayout}, containing a label and
	 * (if not <code>null</code>) a component. If the component is a {@link JRadioButton} or
	 * {@link JCheckBox}, there is no explicit label, but the text capabilities
	 * of those components are used instead.
	 * 
	 * @param parent
	 *            here the elements get added to.
	 * @param pe
	 *            the property element to add
	 * @param constraint
	 *            the gridbag constraints describing the layout constraints how
	 *            to add
	 */
	public void addLabelAndComponentTo(final JPanel parent, final GridBagConstraints constraint, final PropertyElement pe) {
		final GridBagLayout gridbag = (GridBagLayout)parent.getLayout();
		final Messages messages = getMessages();
		final String key = pe.getKey();
		final Component component = pe.getComponent();
		final String name = (CollectionUtil.isNullOrEmpty(key) ? "" : messages.getText(key));
		final String tooltip = messages.getShortDescription(new I18NStringWithFillIns(key, "%0", getToolTipFillIns(component)));
		if ((component != null) && (component instanceof JRadioButton)) {
			final JRadioButton box = (JRadioButton) component;
			box.setText(name);
			box.setToolTipText(tooltip);
	        parent.add(createPanel(component, gridbag, constraint));
// Commented out - so checkboxes are presented like other properties, with label on the left and edit element on the right
//        } else if ((component != null) && (component instanceof JCheckBox)) {
//        	final JCheckBox box = (JCheckBox) component;
//        	box.setText(name);
//        	box.setToolTipText(tooltip);
//	        parent.add(createPanel(component, gridbag, constraint));
		} else {
			if (null == component) { // This must be a heading label
				final JLabel label = createLabel(name, tooltip);
				addImageIconTo(label, pe);
				final JPanel p = new JPanel() {
					private static final long serialVersionUID = 1L; 
					@Override
					public Dimension getPreferredSize() {
						return new Dimension(200, super.getPreferredSize().height);
					}
				};
				label.setForeground(Color.blue);
				p.add(label);
				p.setBackground(Color.white);

				constraint.gridwidth = GridBagConstraints.REMAINDER; //end row
				constraint.fill = GridBagConstraints.HORIZONTAL; // Make the label fill the whole row
				gridbag.setConstraints(p, constraint);
				parent.add(p);
			} else {
				final JLabel label = createLabel(name, tooltip, gridbag, constraint);
				addImageIconTo(label, pe);
				parent.add(label);
				label.setLabelFor(component);
				parent.add(createPanel(component, gridbag, constraint));
			}
		}
	}

	public void addImageIconTo(final JLabel label, final PropertyElement pe) {
		if (null != pe.getResourceFolder()) {
			final ResourceFinder res = getApplicationEnvironment().resources();
			if (null != res) {
				final java.net.URL url = res.getURLToResource(pe.getImageKey() + ".png", pe.getResourceFolder());
				label.setIcon(new ImageIcon(url));
				logger.trace("Looking for icon " + pe.getKey() + " in " + url);
			}
		}
	}

	protected I18NStringWithFillIns[] getToolTipFillIns(final Component component) {
		I18NStringWithFillIns i18ns = I18NStringWithFillIns.EMPTY;
		if (component instanceof JFormattedTextField) {
			final InputVerifier inputVerifier = ((JFormattedTextField) component).getInputVerifier();
			if ((null != inputVerifier) && (inputVerifier instanceof FormattedTextFieldInputVerifier)) {
				i18ns = ((FormattedTextFieldInputVerifier) inputVerifier).getToolTip();
			}
		}
		final I18NStringWithFillIns[] params = new I18NStringWithFillIns[] { i18ns };
		return params;
	}

	public void addPropertyElementTo(final JPanel parent, final PropertyElement pe, final GridBagConstraints constraint) {
		final PropertyElement[] childs = pe.getChilds();
		if (null != childs) {
			addLabelAndComponentTo(parent, constraint, pe);
			for (PropertyElement child : childs) {
				addPropertyElementTo(parent, child, constraint);
			}
		} else {
			addLabelAndComponentTo(parent, constraint, pe);
		}
	}

	protected JMenuItem createActionItem(final JMenu menu, final Action action) {
		final JMenuItem menuItem;
		allActions.add(action);
		if (action instanceof CheckBoxMenuItemAction) {
			menuItem = ((CheckBoxMenuItemAction) action).createCheckBoxMenuItem();
		} else {
			menuItem = new JMenuItem(action);
		}
		menu.add(menuItem);
		return menuItem;
	}

	protected JRadioButtonMenuItem createRadioButtonActionItem(final JMenu menu, final Action action, final ButtonGroup buttonGroup) {
		allActions.add(action);
		final JRadioButtonMenuItem menuItem = createRadioButtonMenuItem(buttonGroup, action);
		if (null != menu) {
			menu.add(menuItem);
		}
		return menuItem;
	}

	public JMenu createLocalesMenu() {
		final PedJMenu ret = createJMenu("Menu.Locales", ResourceFinder.Folder.COMMON);
		final ButtonGroup localeButtonGroup = new ButtonGroup();
		if (null != switchLocaleActions) {
			for (final SwitchLocaleAction action : switchLocaleActions) {
				action.setRadioButtonMenuItem(createRadioButtonActionItem(ret, action, localeButtonGroup));
			}
		}
		return ret;
	}

	public JMenu createLAFMenu() {
		final PedJMenu ret = createJMenu("Menu.LookAndFeel", ResourceFinder.Folder.COMMON);
		addLAFMenuItems(ret);
		return ret;
	}

	private void addLAFMenuItems(final JMenu menu) {
		final ButtonGroup lafButtonGroup = new ButtonGroup();
//		JFrame.setDefaultLookAndFeelDecorated(true);
//		UIManager.installLookAndFeel("Napkin",     "net.sourceforge.napkinlaf.NapkinLookAndFeel");

		try {
			UIManager.setLookAndFeel(UIManager.getCrossPlatformLookAndFeelClassName());
		} catch (final Exception e) {
			Assert.assertTrue(false);
		}
		lookAndFeelInfos = UIManager.getInstalledLookAndFeels();
		lafActions = new LAFAction[lookAndFeelInfos.length];
		for (int i = 0; i < lookAndFeelInfos.length; i++) {
			lafActions[i] = new LAFAction(lookAndFeelInfos[i]);
			lafActions[i].setRadioButtonMenuItem(createRadioButtonMenuItem(lafButtonGroup, lafActions[i]));
		}
		miLAFCrossPlatform = createRadioButtonActionItem(menu, lafCrossPlatformAction, lafButtonGroup);
		miLAFSystem        = createRadioButtonActionItem(menu, lafSystemAction, lafButtonGroup);
		menu.addSeparator();
		for (final LAFAction lafAction : lafActions) {
			menu.add(lafAction.getRadioButtonMenuItem());
		}
		actualizeLAFRadioButtons();
	}

	/**
	 * Initializes the correct selection of the radio buttons.
	 * Called only once during initialization.
	 */
	protected void actualizeLAFRadioButtons() {
		final String lafClassName = UIManager.getLookAndFeel().getClass().getName();
		if (null != lafClassName) {
			if (lafClassName.equals(UIManager.getSystemLookAndFeelClassName())) {
				miLAFSystem.setSelected(true);
			} else if (lafClassName.equals(UIManager.getCrossPlatformLookAndFeelClassName())) {
				miLAFCrossPlatform.setSelected(true);
			} else {
				for (int i = 0; i < lookAndFeelInfos.length; i++) {
					if (lafClassName.equals(lookAndFeelInfos[i].getName())) {
						lafActions[i].getRadioButtonMenuItem().setSelected(true);
					}
				}
			}
		}
	}

	/**
	 * Lists all actions within <em>menu</em>. Can be used as part of the HTML
	 * help.
	 * 
	 * All direct children menus are shown as separate HTML chapters, with a
	 * definition list (<em>dl</em> HTML element) for the entries. Menus within
	 * menus are shown as nested definition lists.
	 * 
	 * @param doc
	 *            the document where to write into
	 * @param menuBar
	 *            the menu bar to get the content from
	 */
	public void buildMenuDescriptions(final TextDocumentBuilder doc, final JMenuBar menuBar) {
		final TextTree tree = new TextTree(TextTreeNodeType.MENU_BAR, "Help.Menu", null, null, null, null, null);
		for (int i = 0; i < menuBar.getMenuCount(); i++) {
			final JMenu element = menuBar.getMenu(i);
			if (element instanceof JMenu) {
				final JMenu menu = (JMenu) element;
				final TextTree child = new TextTree(TextTreeNodeType.MENU, menu.getName(), menu.getText(), null, null, null, null);
				tree.add(child);
				buildMenuDescriptions(child, menu, 1);
			}
		}
		logger.debug(tree.dumpRecursive(0));
		doc.add(tree, 2, 3, true);
	}

	/**
	 * Translates a single menu into a {@link TextTree} structure that can be used for e.g. help text creation.
	 * 
	 * @param tree
	 *            the root element to create the menu structure in.
	 * @param menu
	 *            the menu to get the content from
	 * @param depth
	 *            the depth where to stop going into detail
	 */
	private void buildMenuDescriptions(final TextTree tree, final JMenu menu, final int depth) {
		tree.setHtmlAttribs(TextFormatter.htmlAttribute(TextFormatter.CLASS, "actions"));
		for (final Component element : menu.getMenuComponents()) {
			if (element instanceof JMenu) {
				final JMenu m = (JMenu) element;
				String imageFilename = null;
				if (element instanceof PedJMenu) {
					final PedJMenu pm = (PedJMenu) m;
					imageFilename = TextDocumentBuilder.getCodedImageFilename(pm.getResourceFilename(),
							pm.getResourceFolder());
				}
				final TextTree branch = new TextTree(TextTreeNodeType.MENU, m.getName(), m.getText(), null, null, imageFilename, null);
				buildMenuDescriptions(branch, (JMenu) element, depth + 1);
				tree.add(branch);
			}
			if (element instanceof JMenuItem) {
				final JMenuItem mi = (JMenuItem) element;
				tree.add(buildActionDescription(mi.getAction()));
			}
		}
	}

	/**
	 * Translates a single action into a {@link TextTree} element that can be used for e.g. help text creation.
	 * 
	 * @param action
	 *            the action to get the content from
	 * @return the text tree element
	 */
	private static TextTree buildActionDescription(final Action action) {
		final TextTree ret;
		if (null != action) {
			ret = new TextTree();
			ret.setType(TextTreeNodeType.ACTION);
			String id = null;
			if (action instanceof PedAction) {
				PedAction maa = (PedAction) action;
				id = maa.getMessageKeyPrefix();
				ret.setKey(id);
				final Object img = action.getValue(Action.SMALL_ICON);
				// Only put images with an URL into the help file - some images are created within the game that should not show up.
				final Object url = action.getValue(PedAction.SMALL_ICON_URL);
				if ((null != img) && (null != url)) {
					final String filename = TextDocumentBuilder.getCodedImageFilename(id + ".png",
							(maa.isCommonResource() ? Folder.COMMON : Folder.APP));
					ret.setImageFilename(filename);
				}
			} 
			String text = null;
			final Object value = action.getValue(Action.NAME);
			if (null != value) {
				text = value.toString();
				final String description = (String) action.getValue(Action.LONG_DESCRIPTION);
				ret.setText(text);
				ret.setLongDescription(description);
			}
		} else {
			ret = null;
		}
		return ret;
	}

	/**
	 * Translates a list of property elements into a {@link TextTree} structure that can be used for e.g. help text creation.
	 * 
	 * @param tree
	 *            the root element to create the menu structure in.
	 * @param propertyElements
	 *            the properties to describe. Each element could be of type
	 *            String (which is interpreted as simple message key) or a
	 *            {@link PropertyElement} (which contains the message key and
	 *            may even contain child elements).
	 */
	private void buildPropertyDescriptions(final TextTree tree, final Object[] propertyElements) {
		if (null != propertyElements) {
			final Messages messages = getMessages();
			final Iterator<?> propertyIterator = Arrays.asList(propertyElements).iterator();
			tree.setHtmlAttribs(TextFormatter.htmlAttribute(TextFormatter.CLASS, "properties"));
			while (propertyIterator.hasNext()) {
				final TextTree node = new TextTree();
				node.setType(TextTreeNodeType.PROPERTY);
				final Object p = propertyIterator.next();
				String key = null;
				String dd = null;
				Object[] childs = null;
				if (p instanceof String) {
					key = (String) p;
					dd = messages.getLongDescription(key);
				} else if (p instanceof PropertyElement) {
					final PropertyElement pe = (PropertyElement) p;
					final Component component = pe.getComponent();
					node.setImageFilename(TextDocumentBuilder.getCodedImageFilename(pe.getImageKey() + ".png", pe.getResourceFolder()));
					key = pe.getKey();
					dd = messages.getLongDescription(new I18NStringWithFillIns(key, "%0", getToolTipFillIns(component)));
					childs = pe.getChilds();
					if (component instanceof JComboBox) {
						final JComboBox<?> box = (JComboBox<?>) component;
						JIconComboBox<?> iconComboBox = null;
						boolean isHelpWithElements = true;
						Folder folder = null;
						if (box instanceof JIconComboBox) {
							iconComboBox = (JIconComboBox<?>) box;
							isHelpWithElements = iconComboBox.isHelpWithElements();
							folder = iconComboBox.getFolder();
						}
						if (isHelpWithElements) {
							final ArrayList<LocalizedString> items = new ArrayList<LocalizedString>();
							for (int i = 0; i < box.getItemCount(); i++) {
								Object item = box.getItemAt(i);
								if (item instanceof TableCellComboBoxRenderer.TableCellComboBoxItem) {
									item = ((TableCellComboBoxRenderer.TableCellComboBoxItem) item).str;
								}
								if (item instanceof LocalizedString) {
									items.add((LocalizedString) item);
									LocalizedString ls = (LocalizedString) item;
									String imageFilename = null;
									if ((null != iconComboBox) && (null != folder)) {
										imageFilename = TextDocumentBuilder.getCodedImageFilename(iconComboBox.getImageFilename(item), folder);
									}
									node.add(new TextTree(TextTreeNodeType.ITEM, ls.key, ls.getText(), ls.getLongDescription(), null, imageFilename, null));
								}
							}
						}
					}
				}
				node.setKey(key);
				node.setText(messages.getText(key));
				node.setLongDescription(dd);
				tree.add(node);
				if (null != childs) {
					buildPropertyDescriptions(node, childs);
				}
			}
		}
	}

	/**
	 * Builds the descriptions of the properties that can be set in the given dialog. So far tailored for the properties of a new game / run dialog.
	 * 
	 * @param doc
	 *            the document to build the descriptions into
	 * @param dialog
	 *            the dialog to take the properties from to describe
	 */
	public void buildPropertyDescriptions(final TextDocumentBuilder doc, final PedJDialog dialog) {
		buildPropertyDescriptions(doc, 2, "Help.Properties", dialog.getAllPropertyElements(), true);
	}

	/**
	 * Builds the descriptions of the given properties into the given document.
	 * 
	 * @param doc
	 *            the document to build the descriptions into
	 * @param depth
	 *            the heading depth
	 * @param key
	 *            the i18n key that defines heading text and first paragraph
	 * @param propertyElements
	 *            the properties to describe. Each element could be of type
	 *            String (which is interpreted as simple message key) or a
	 *            {@link PropertyElement} (which contains the message key and
	 *            may even contain child elements).
	 * @param isToCreateAnchors
	 *            if HTML anchor elements should be created for each property.
	 *            Could lead to problems if this way several anchors on the same
	 *            page could get the same id.
	 */
	public void buildPropertyDescriptions(final TextDocumentBuilder doc, final int depth, final String key,
			final Object[] propertyElements, final boolean isToCreateAnchors) {
		final Messages messages = doc.messages();
		final TextTree tree = new TextTree(TextTreeNodeType.PROPERTY_ROOT, key, messages.getText(key),
				messages.getLongDescription(key), null, null, null);
		buildPropertyDescriptions(tree, propertyElements);
		logger.debug(tree.dumpRecursive(0));
		doc.add(tree, depth, depth, isToCreateAnchors);
	}

	protected static final TextFormatter.ColumnInfo[] CI_TEMPLATE = new TextFormatter.ColumnInfo[] {};
	protected static final TextFormatter.TableCellInfo[] TCI_TEMPLATE = new TextFormatter.TableCellInfo[] {};
	protected static final String[] STR_TEMPLATE = new String[] {};

	/**
	 * @param doc the document to add the description to
	 * @param userGameActionInputs the list of action keys defined
	 * @param isDifferentPerPlayer if <code>true</code>, an additional column is created to also list the player the key is assigned to
	 */
	protected void buildInputDescriptions(final TextDocumentBuilder doc, final UserGameActionInput[] userGameActionInputs, final boolean isDifferentPerPlayer) {
		final Messages messages = getMessages();
		final String keyHeading = "Help.Input";
		if (!CollectionUtil.isNullOrEmpty(messages.getLongDescription(keyHeading))) {
			final String keyInputMouse = "Help.Input.Mouse";
			final boolean isWithMouse = !CollectionUtil.isNullOrEmpty(messages.getLongDescription(keyInputMouse));
			if ((null != userGameActionInputs) || isWithMouse) {
				doc.addContentPart(2, keyHeading);
				buildKeyboardActionsDescription(doc, userGameActionInputs, isDifferentPerPlayer);
				if (isWithMouse) {
					doc.addContentPart(3, keyInputMouse);
				}
			}
		}
	}

	/**
	 * @param doc the document to add the description to
	 * @param userGameActionInputs the list of action keys defined
	 * @param isDifferentPerPlayer if <code>true</code>, an additional column is created to also list the player the key is assigned to
	 */
	protected void buildKeyboardActionsDescription(final TextDocumentBuilder doc, final UserGameActionInput[] userGameActionInputs, final boolean isDifferentPerPlayer) {
		if (null != userGameActionInputs) {
			final Messages messages = getMessages();
			doc.addContentPart(3, "Help.KeyboardActions");
			final ArrayList<TextFormatter.ColumnInfo> cis = new ArrayList<TextFormatter.ColumnInfo>();
			if (isDifferentPerPlayer) {
				cis.add(new TextFormatter.ColumnInfo(10, TextFormatter.RIGHT, messages, "Game.Player"));
			}
			cis.add(new TextFormatter.ColumnInfo(40, TextFormatter.LEFT, messages, "Game.PlayerAction"));
			cis.add(new TextFormatter.ColumnInfo(20, TextFormatter.LEFT, messages, "Game.KeyStroke"));
			final TextFormatter.TableInfo ti = new TextFormatter.TableInfo(doc, true, true,
					TextFormatter.htmlAttribute(TextFormatter.CLASS, "keyboardActions inv"), 
					cis.toArray(CI_TEMPLATE));
			doc.beginTable("keys", ti);
			for (final UserGameActionInput ugai : userGameActionInputs) {
				final ArrayList<TextFormatter.TableCellInfo> cs = new ArrayList<TextFormatter.TableCellInfo>(3);
				if (isDifferentPerPlayer) {
					cs.add(new TextFormatter.TableCellInfo(Integer.toString(ugai.playerIdx + 1)));
				}
				cs.add(new TextFormatter.TableCellInfo(getKeyboardActionDescriptionText(ugai)));
				cs.add(new TextFormatter.TableCellInfo(ugai.keyCodeString));
				doc.addTableRow(cs);
			}
			doc.endTable();
		}
	}

	/**
	 * Called by
	 * {@link #buildKeyboardActionsDescription(TextDocumentBuilder, UserGameActionInput[], boolean)}
	 * per key. Override when needed.
	 * 
	 * Default implementation returns <code>null</code>. Thought about making this
	 * method abstract, but I didn't want to force subclasses not using the
	 * calling method to implement a method never needed.
	 * 
	 * @param ugai
	 *            the key that could be pressed
	 * @return a human-readable description what this key does
	 */
	protected String getKeyboardActionDescriptionText(final UserGameActionInput ugai) {
		return null;
	}

	protected void startNewGame() {
		// Default implementation does nothing
	}

	protected void exitGame() {
		// Default implementation does nothing
	}

	/**
	 * @return <code>true</code> if the application is running, or <code>false</code> if
	 *         paused. Default implementation returns always <code>true</code>.
	 */
	public boolean isRunning() {
		return true;
	}

	/**
	 * Can pause or resume the application. Default implementation does nothing.
	 * 
	 * @param isToResume
	 *            <code>true</code> if to resume the application, <code>false</code> to
	 *            pause
	 */
	protected void pauseOrResume(final boolean isToResume) {
		handleTitleStateTransition(isToResume ? TitleStateTransition.RESUME : TitleStateTransition.PAUSE);
	}

	private boolean isRunning = false;
	private FinishableThread finishableThread = null;

	/**
	 * Overload this to implement real pause or resume behavior.
	 * 
	 * @return the run thread of the application that runs independently to the GUI thread
	 */
	public FinishableThread createRunThread() {
		return null;
	}

	protected boolean isRunningRun() {
		return isRunning;
	}

	protected synchronized void pauseOrResumeRun(final boolean isToResume) {
		if (isToResume) {
			isRunning = true;
			if (null == finishableThread) {
				finishableThread = createRunThread();
			}
			finishableThread.start();
			handleTitleStateTransition(TitleStateTransition.RESUME);
		} else {
			isRunning = false;
			if (null != finishableThread) {
				finishableThread.interrupt();
				try {
					finishableThread.join();
				} catch (final InterruptedException e) {
					logger.error(e.getMessage(), e);
				}
				finishableThread = null;
			}
			handleTitleStateTransition(TitleStateTransition.PAUSE);
		}
	}
	
	/**
	 * Can be called to start a pause, e.g., for showing a dialog, no matter if
	 * the game was already paused or not.
	 * 
	 * @return if the game was running or not
	 * @see #pauseEnd(boolean, boolean)
	 */
	protected boolean pauseBegin() {
		final boolean wasRunning = isRunning();
		pauseOrResume(false);
		try {
			Thread.sleep(0);
		} catch (final InterruptedException e) {
			// Nothing to do - that's what we intend to do
		}
		return wasRunning;
	}
	
	/**
	 * Can be called to resume from a pause, e.g., for showing a dialog.
	 * 
	 * @param wasRunning
	 *            typically the value returned from {@link #pauseBegin()}
	 * @param shouldRun
	 *            <em>true</em> if this is a new start of a game, e.g., for a new
	 *            game dialog
	 */
	protected void pauseEnd(final boolean wasRunning, final boolean shouldRun) {
		pauseOrResume(wasRunning | shouldRun);
	}

	public enum TitleStateTransition {
		/** (Re-)Starts the title screen */
		START,
		/** Leaves the title screen, e.g., to start the game */
		STOP,
		/** Pauses the title screen, e.g., to show info message box */
		PAUSE,
		/** Resumes the title screen, e.g., after showing the info message box */
		RESUME,
	};
	
	/**
	 * Can be called to indicate that the title screen has performed one of the
	 * transitions in {@link TitleStateTransition}, and therefore e.g. playing
	 * the title music has to be adapted to the new state. The default
	 * implementation does nothing.
	 * 
	 * @param transition
	 *            the transition to perform
	 */
	protected void handleTitleStateTransition(final TitleStateTransition transition) {
		// Default implementation does nothing.
	}

	/** 
	 * Zooms in or out. Developed to work with {@link JScrollAndZoomPane}. 
	 * Default implementation does nothing.
	 * 
	 * @param amount the amount by which to change the zoom level - typically -1 or +1.
	 */
	public void changeZoomLevel(final int amount) {
		// Default implementation does nothing.
	}

	/** Displays the user help from the menu; implicitly pauses the application. */
	public void showHelpBox() {
		// Stop the game
		final boolean wasRunning = pauseBegin();
		final HtmlPaneDialog d = createHelpBox();
		d.enterDialog();
		// Resume the game
		pauseEnd(wasRunning, false);
	}

	public HtmlPaneDialog createHelpBox() {
		HtmlPaneDialog d = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			final TextDocumentBuilder doc = new TextDocumentBuilder(getMessages());
			doc.setEmbedded(false);
			buildHelpDocument(doc);
			String content = doc.result(getApplicationEnvironment().resources());
			d = new HtmlPaneDialog(this, doc.getTitle(), content);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		return d;
	}

	/**
	 * Builds the help document.
	 * 
	 * @param doc
	 *            the document to write into
	 */
	public abstract void buildHelpDocument(final TextDocumentBuilder doc);

	public void buildHtmlHelp(final TextDocumentType type, final boolean isEmbedded, final String folder,
			final String fileNameInfix) throws IOException {
		final Locale[] availableLocales = getMessages().getAvailableLocales();
		if (null != availableLocales) {
			for (Locale locale : availableLocales) {
				buildHtmlHelp(type, isEmbedded, folder, fileNameInfix, locale);
			}
		} else {
			buildHtmlHelp(type, isEmbedded, folder, fileNameInfix, null);
		}
	}

	/**
	 * Builds resources like images for the help document. 
	 * Default implementation does nothing.
	 */
	public void buildHtmlHelpResources() {
		// Default implementation does nothing
	}

	public void buildHtmlHelp(final TextDocumentType type, final boolean isEmbedded, final String folder,
			final String fileNameInfix, final Locale locale) throws IOException {
		final TextDocumentBuilder doc = new TextDocumentBuilder(getMessages());
		doc.setTextDocumentType(type);
		doc.setEmbedded(isEmbedded);
		if (null != locale) {
			switchLocale(locale);
		}
		buildHelpDocument(doc);
		final String filename = "Html_" 
				+ (null == fileNameInfix ? getMessages().getApplicationName(Messages.I18NedOrNot.NonI18N) : fileNameInfix) + "_" 
				+ (null == locale ? "" : locale.toString() + "_")
				+ type.name()
				+ "." + type.getTypicalFileExtensions()[0];
		final String outfileName = (folder == null ? "" : folder + File.separator ) + filename;
		writeToFile(outfileName, doc.getCharset(), doc.result(getApplicationEnvironment().resources(), true));
	}

	public static void writeToFile(final String fileName, final Charset charset, final String content) throws IOException {
		Writer fs = null;
		try {
			fs = new OutputStreamWriter(new FileOutputStream(fileName), charset);
			fs.write(content);
		} finally {
			if (null != fs) {
				fs.close();
			}
		}
	}

	/** Displays the about box as plain text; implicitly pauses the application. */
	public void showAboutBox() {
		// Stop the game
		final boolean wasRunning = pauseBegin();
		// Show the box
		final Messages messages = getMessages();
		PedJDialog d = null;
		try {
			setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
			d = new JInfoDialog(
					this, 
					messages.getString("AboutWindow.Title"), 
					getTitle() + "\n" + messages.getApplicationCopyrightMessage(),
					JOptionPane.INFORMATION_MESSAGE);
			d.setResizable(false);
		} finally {
			setCursor(Cursor.getDefaultCursor());
		}
		d.enterDialog();
		// Resume the game
		pauseEnd(wasRunning, false);
	}

	/**
	 * Update the localized string for the menu, and search recursively for
	 * submenus to do the same. However, this does not update the actions or
	 * menu entries within those menus.
	 * 
	 * @param locale
	 *            The locale to switch to
	 * @param menu
	 *            The menu to update
	 */
	protected void switchLocale(final Locale locale, final JMenu menu) {
		putValues(menu, menu.getName());
		logger.trace("switchLocale(" + locale + "," + menu + ")");
		final int menuSize = menu.getMenuComponentCount();
		for (int i = 0; i < menuSize; i++) {
			final Component component = menu.getMenuComponent(i);
			if (component instanceof JMenu) {
				switchLocale(locale, (JMenu)component);
			}
		}
	}

	/**
	 * Switches the locale of the UI. Overload to support runtime switching of
	 * the locale for application specific parts of the UI. This stub
	 * implementation changes all JMenu entries of the application main window,
	 * and all actions defined in this class.
	 * 
	 * @param locale
	 *            the locale to switch to
	 */
	public void switchLocale(final Locale locale) {
		final Messages messages = getMessages();
		logger.trace("switchLocale(" + locale + ")");
		messages.setCurrentLocale(locale);
		if (null != switchLocaleActions) {
			for (final SwitchLocaleAction switchLocaleAction : switchLocaleActions) {
				switchLocaleAction.setSelected(locale);
			}
		}
		for (final Action action : allActions) {
			if (action instanceof PedAction) {
				putValues((PedAction) action);
			}
		}
		final JMenuBar menuBar = getJMenuBar();
		if (null != menuBar) {
			final int menuBarSize = menuBar.getMenuCount();
			for (int i = 0; i < menuBarSize; i++) {
				final JMenu menu =  menuBar.getMenu(i);
				switchLocale(locale, menu);
				final Component[] components = menu.getMenuComponents();
				for (final Component c : components) {
					if (c instanceof JMenu) {
						final JMenu menu2 =  (JMenu) c;
						switchLocale(locale, menu2);
					}
				}
			}
		}
	}

	public void setLookAndFeel(final String className) {
		try {
			UIManager.setLookAndFeel(className);
			SwingUtilities.updateComponentTreeUI(this);
			pack();
		} catch (final ClassNotFoundException e) {
			logger.warn("Could not set look and feel " + className, e);
		} catch (final InstantiationException e) {
			logger.warn("Could not set look and feel " + className, e);
		} catch (final IllegalAccessException e) {
			logger.warn("Could not set look and feel " + className, e);
		} catch (final UnsupportedLookAndFeelException e) {
			logger.warn("Could not set look and feel " + className, e);
		}
	}

	/** 
	 * Called by the GameThread.
	 * Default implementation does nothing, override by useful behavior. 
	 */
	public void step() {
		// Default implementation does nothing.
	}

	/**
	 * Called by the GameThread to determine how long to sleep.
	 * 
	 * @return the number of milliseconds to sleep to the next call to step.
	 *         Override by useful behavior; default implementation returns 0.
	 */
	public long getSleepDelay() {
		return 0;
	}

	/** 
	 * @return <code>true</code> if this application is waiting for a modal dialog
	 */
	public boolean isWaitingForModal() {
		return (null != dialogWaitingFor);
	}
	
	/** 
	 * @return the modal dialog the application is waiting for
	 */
	public PedJDialog getDialogWaitingFor() {
		return dialogWaitingFor;
	}
	
	/**
	 * Called whenever a {@link PedJDialog} gets opened or closed. Override with
	 * your own behavior, but don't forget to call this super implementation
	 * from there.
	 * 
	 * @param dialog
	 *            the dialog that was opened or closed.
	 * @param wasOpened
	 *            <code>true</code> if the dialog was entered, <code>false</code> if the
	 *            dialog was closed
	 */
	public void setWaitingForModal(final PedJDialog dialog, final boolean wasOpened) {
		if (wasOpened) {
			dialogWaitingFor = dialog;
		} else {
			dialogWaitingFor = null;
		}
	}

	@Override
	public void keyTyped(KeyEvent e) {
		// Default implementation does nothing
	}

	@Override
	public void keyPressed(KeyEvent e) {
		// Default implementation does nothing
	}

	@Override
	public void keyReleased(KeyEvent e) {
		// Default implementation does nothing
	}

	private PedJDialog dialogWaitingFor = null;
}
