package de.ped.tools.gui;

import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Frame;
import java.awt.Label;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.ApplicationParameter;
import de.ped.tools.CollectionUtil;
import de.ped.tools.Messages;
import de.ped.tools.Messages.I18NedOrNot;
import de.ped.tools.ResourceFinder;
import de.ped.tools.TextDocumentType;
import de.ped.tools.log.Logger;

/**
 * An Applet opening a new frame. This new frame should be the "real thing" for the user.
 * This class contains some code to make it easy to be run as application as well.
 *
 * @author Copyright (c) 2001 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class ApplicationOrApplet extends Applet implements ApplicationEnvironment {
	private static final long serialVersionUID = 1L;

	/**
	 * The instance of this class. Is known after init() is called.
	 */
	private static ApplicationOrApplet instance  = null;

	/**
	 * @return the instance
	 */
	public static ApplicationOrApplet getInstance() {
		return instance;
	}

	/**
	 * Is <code>true</code> iff started as application.
	 */
	private static boolean isApplication = false;

	/**
	 * @return <code>true</code> if the program is run as application via the main method
	 */
	public static boolean isApplication() {
		return isApplication;
	}

	public ApplicationOrApplet() {
		params = new ArrayList<ApplicationParameter>();
		params.add(paramLabel);
		params.add(paramBG);
		params.add(paramFG);
		params.add(paramLanguage);
		params.add(paramSilent);
		params.add(paramHtmlHelp);
		params.add(paramResource);
	}

	private Messages messages;

	@Override
	public Messages messages() {
		if (null == messages) {
			messages = createMessages();
		}
		return messages;
	}

	/**
	 * Factory to create the messages object that will handle all the localized stuff.
	 * 
	 * @return the messages object that will handle all the localized stuff
	 */
	protected abstract Messages createMessages();

	private ResourceFinder resources;

	@Override
	public ResourceFinder resources() {
		if (null == resources) {
			resources = new ResourceFinder(getPathToResources());
		}
		return resources;
	}

	/**
	 * Relative file path to the resources. Default implementation always
	 * returns <code>null</code>, meaning there are no resources. Subclasses can
	 * overload this method. This object guarantees that this method will be
	 * called only once.
	 * 
	 * @return the relative file path to the resources
	 */
	protected String getPathToResources() {
		return null;
	}

	@Override
	public boolean isExitable() {
		return isApplication;
	}

	/**
	 * Contains the frame that is to open.
	 */
	private Frame frame = null;

	public Frame getFrame() {
		return frame;
	}

	/**
	 * Factory to create the main frame to display. Should be overloaded.
	 * 
	 * @return the main frame to display
	 */
	protected Frame createFrame() {
		return null;
	}

	/**
	 * @return <code>true</code> if the frame should be deleted each time the user leaves the page
	 * (if started as applet).
	 * Default behavior is to keep it, so the user may go back to the page
	 * and continue.
	 */
	protected boolean isFrameToDeleteOnStop() {
		return false;
	}

	/**
	 * Should be called in main method before the instance of this class is created.
	 */
	public static void applicationPreCreate() {
		isApplication = true;
	}

	/**
	 * Should be called in main method after the instance of this class is
	 * created.
	 * 
	 * @param frameApplet
	 *            the application or applet
	 * @param commandLineParams
	 *            the command line params
	 */
	protected static void applicationPostCreate(final ApplicationOrApplet frameApplet, final String[] commandLineParams) {
		if (null != frameApplet) {
			frameApplet.parseCommandLineParameters(commandLineParams);
			frameApplet.init();
			frameApplet.start();
		}
	}

	/**
	 * Call this method when the application should come to an end. If called in
	 * an applet, nothing happens.
	 */
	public static void exit() {
		if (isApplication()) {
			System.exit(0);
		} else {
			// Nothing to do, an applet stops when the browser stops.
		}
	}

	private ApplicationParameter paramLabel = new ApplicationParameter("label", "Default label", "Label string to be displayed");
	private ApplicationParameter paramBG = new ApplicationParameter("background", "C0C0C0", "Label background color, format \"rrggbb\" in hex notation");
	private ApplicationParameter paramFG = new ApplicationParameter("foreground", "000000", "Label foreground color, format \"rrggbb\" in hex notation");
	private ApplicationParameter paramLanguage = new ApplicationParameter(PARAM_KEY_LANGUAGE, "", "Currently most of my apps support \"de_DE\" or \"en_US\"");
	private ApplicationParameter paramSilent = new ApplicationParameter("silent", "no", "If not \"no\", will start without opening a visible window - used for testing purposes");
	private ApplicationParameter paramHtmlHelp = new ApplicationParameter("htmlhelp", "", "If not empty, will write help web page to this path (use \".\" for current directory)");
	private ApplicationParameter paramResource = new ApplicationParameter("resources", "no", "If not \"no\", will create some resource files");

	private ApplicationParameter paramDebug = new ApplicationParameter(PARAM_KEY_DEBUG, null, "If set, will enable game specific testing views");
	private ApplicationParameter paramChristmas = new ApplicationParameter(PARAM_KEY_CHRISTMAS, null, "if set, defines that this is a christmas edition");

	public List<ApplicationParameter> getParameters() {
		return params;
	}

	protected void addParameter(final ApplicationParameter param) {
		params.add(param);
	}

	protected ApplicationParameter getParamChristmas() {
		return paramChristmas;
	}

	protected ApplicationParameter getParamDebug() {
		return paramDebug;
	}

	@Override
	public ApplicationParameter findParameter(String key) {
		final List<ApplicationParameter> params = getParameters();
		ApplicationParameter ret = null;
		for (final ApplicationParameter param : params) {
			if (param.key.equals(key)) {
				ret = param;
				break;
			}
		}
		return ret;
	}

	/**
	 * Evaluates the parameters read and does something with it.
	 * Default implementation reads some default parameters.
	 */
	protected void evaluateParameters() {
		final Color bg = Color.decode("0x" + paramBG.getValue());
		final Color fg = Color.decode("0x" + paramFG.getValue());
		label.setText(paramLabel.getValue());
		try {
			label.setBackground(bg);
			label.setForeground(fg);
			setBackground(bg);
			setForeground(fg);
		} catch (final IllegalAccessError e) {
			Logger.getLogger(getClass()).info(e + "Ah, using an old browser, eh? Ok, I'm ignoring the color settings of the small applet frame...");
		}
	}

	/**
	 * Applet-specific: External interface used by design tools to show properties of an applet.
	 * 
	 * @return the parameter infos
	 */
	@Override
	public String[][] getParameterInfo() {
		final String[][] info = new String[params.size()][];
		for (int i = 0; i < info.length; i++) {
			info[i] = params.get(i).toParameterInfo();
		}
		return info;
	}

	private Label label = new Label();

	/**
	 * Parses the command line arguments (if called as application) for parameter definitions.
	 * 
	 * @param commandLineArguments the command line arguments to parse
	 */
	protected void parseCommandLineParameters(String[] commandLineArguments) {
		for (String arg : commandLineArguments) {
			final String[] parts = arg.split("=");
			if (2 == parts.length) {
				final ApplicationParameter param = findParameter(parts[0]);
				if (null != param) {
					param.setValue(parts[1]);
				}
			}
		}
	}

	/**
	 * Applet-specific: Delivers information about the applet to the user.
	 * 
	 * @return the applet info
	 */
	@Override
	public String getAppletInfo() {
		String ret;
		Messages messages = messages();
		if (null != messages) {
			final String desc = messages.getApplicationDescription();
			ret = messages.getApplicationName(I18NedOrNot.I18N) + " " + messages.getApplicationVersion() + "\n" 
					+ (CollectionUtil.isNullOrEmpty(desc) ? "" : desc + "\n")
					+ messages.getApplicationCopyrightMessage();
		} else {
			ret = super.getAppletInfo();
		}
		return ret;
	}

	/**
	 * Applet-specific: Is called when the user loads the page rooting the
	 * applet.
	 */
	@Override
	public void init() {
		instance = this;
		setBackground(Color.lightGray);
		setForeground(Color.black);
		label.setText("label");
		setLayout(new BorderLayout());
		add(label, BorderLayout.NORTH);
		
		final List<ApplicationParameter> params = getParameters();
		if (!isApplication) {
			// For applets, this is the place to read the page parameters.
			try {
				for (final ApplicationParameter param : params) {
					if (null != param.key) {
						final String value = getParameter(param.key);
						param.setValue(value);
		//				Logger.getRootLogger().debug("Parameter " + param.key + " read as " + value);
					}
				}
			} catch (final NullPointerException ex) {
				Logger.getRootLogger().info(
						"Seems like I have been started as neither applet nor application. I give up reading applet parameters. ");
			}
		}
		evaluateParameters();
		System.out.println( Messages.SEPARATOR + "\n" +
							getAppletInfo() +
							"\n[Started at: " + new Date() + "]\n" +
							Messages.SEPARATOR
							);
		frame = createFrame();
	}

	@Override
	public boolean isSilent() {
		return (!paramSilent.defaultValue.equalsIgnoreCase(paramSilent.getValue()) || isHtmlHelp() || isCreatingResources());
	}

	private boolean isHtmlHelp() {
		return !paramHtmlHelp.defaultValue.equalsIgnoreCase(paramHtmlHelp.getValue());
	}

	private boolean isCreatingResources() {
		return !paramResource.defaultValue.equalsIgnoreCase(paramResource.getValue());
	}

	/**
	 * Applet-specific: Is called when the user (re)enters the page rooting the
	 * applet.
	 */
	@Override
	public void start() {
		super.start();
		if (null == frame) {
			frame = createFrame();
		}
		if (null != frame) {
			if (frame instanceof ApplicationMainWindow) {
				final ApplicationMainWindow amw = (ApplicationMainWindow) frame;
				boolean isToExit = false;
				if (isHtmlHelp()) {
					final String path = paramHtmlHelp.getValue();
					Logger.getRootLogger().debug("Building HTML help files to " + path);
					try {
						amw.buildHtmlHelp(TextDocumentType.HTML5, true, path, null);
					} catch (final IOException e) {
						Logger.getRootLogger().error("Unable to build HTML help file", e);
					}
					isToExit = true;
				}
				if (isCreatingResources()) {
					Logger.getRootLogger().info("Building HTML help resources");
					amw.buildHtmlHelpResources();
					isToExit = true;
				}
				if (isToExit) {
					exit();
				}
			}
			if (!isSilent()) {
				frame.setVisible(true);
			}
		}
	}

	/**
	 * Applet-specific: Is called when the user leaves the page rooting the applet.
	 */
	@Override
	public void stop() {
		if (null != frame) {
			frame.setVisible(false);
			if (isFrameToDeleteOnStop()) {
				frame.dispose();
				frame = null;
			}
		}
		super.stop();
	}

	/**
	 * Applet-specific: Is called when the user leaves the browser.
	 */
	@Override
	public void destroy() {
		if (null != frame) {
			frame.dispose();
			frame = null;
		}
		instance = null;
		super.destroy();
	}
	
	private ArrayList<ApplicationParameter> params;
}
