package de.ped.tools.gui;

import java.io.File;
import java.io.IOException;

import javax.sound.sampled.AudioFileFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;

import de.ped.tools.ResourceFinder;
import de.ped.tools.log.Logger;
import de.ped.tools.sound.SoundClip;

/**
 * Bundles all the sounds for an application.
 * 
 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ApplicationSounds {
	/**
	 * Subclasses could implmement time-consuming pre-loading or creation of sounds, and using objects could call this explicitly. 
	 * Default implementation does nothing.
	 */
	public void init() {
		// Default implementation does nothing
	}

	/**
	 * Returns the sound clip for the given key. 
	 * Typically this method is called within the {@link #init()} method, or from a sound converter, so nothing is pre-fetched.
	 * @param key the key (e.g., event) to find the sound for
	 * @return the sound clip for the given key, or <code>null</code> if there is no such sound.
	 */
	public AudioInputStream createSound(final Object key) {
		return null;
	}

	/**
	 * Returns the sound clip for the given key. Typically the sound is pre-fetched and ready to play.
	 * @param key the key (e.g., event) to find the sound for
	 * @return the sound clip for the given key, or <code>null</code> if there is no such sound.
	 */
	public SoundClip findSound(final Object key) {
		return null;
	}

	/**
	 * Writes the given sound into a wave sound file.
	 * @param soundKey The ID of the sound
	 * @param name The file name
	 * @throws IOException writing the file was impossible
	 */
	public void writeSoundFile(final Object soundKey, final String name) throws IOException {
		final AudioInputStream audioInputStream = createSound(soundKey);
		final AudioFileFormat.Type fileType = AudioFileFormat.Type.WAVE;
		final String fileName = name + fileType.getExtension();
		final File fileOut = new File(fileName);
//		if (AudioSystem.isFileTypeSupported(fileType, audioInputStream)) {
			AudioSystem.write(audioInputStream, fileType, fileOut);
//		}
		Logger.getRootLogger().debug("Written " + fileName);
	}

	protected ApplicationSounds(final ResourceFinder resources) {
		this.resources = resources;
	}

	protected ResourceFinder getResources() {
		return resources;
	}

	protected AudioInputStream read(final String resourceName) {
		AudioInputStream ret = null;
		try {
			ret = AudioSystem.getAudioInputStream(getResources().getURLToResource(resourceName));
		} catch (UnsupportedAudioFileException ex) {
			logger.error(ex.getMessage(), ex);
		} catch (IOException ex) {
			logger.error(ex.getMessage(), ex);
		}
		return ret;
	}

	protected final Logger logger = new Logger(getClass().getName());

	private final ResourceFinder resources;
}
