package de.ped.tools.gui;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JCheckBox;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.TreePath;

/**
 * A tree manager to support JTree with a checkbox for each node.
 * 
 * @author Santhosh Kumar T - santhosh@in.fiorano.com, from a <a href=
 *         "http://www.jroller.com/page/santhosh/Weblog/jtree_with_checkboxes?catname=">thread
 *         at jroller.com</a>.
 * @author Peter Diefenbach (peter@pdiefenbach.de) did modifications to use
 *         directly the TreeModel, not the TreeSelectionModel.
 */
public class CheckTreeManager extends MouseAdapter implements TreeSelectionListener {
	private CheckTreeSelectionModel selectionModel;
	private JTree tree = new JTree();
	int hotspot = new JCheckBox().getPreferredSize().width;

	/**
	 * Calling this changes the tree to display checkboxes for each node. Each
	 * node is deselected as default.
	 * 
	 * @param tree
	 *            The tree to manage input events for.
	 * @param isUsingSelectionModel
	 *            if <code>true</code>, Kumar's implementation is used and the
	 *            manager uses the selectionModel. Otherwise Peter's
	 *            implemenation uses directly the treeModel, assuming that the
	 *            nodes in the model implement {@link CheckTreeNode}.
	 */
	public CheckTreeManager(final JTree tree, final boolean isUsingSelectionModel) {
		this.tree = tree;
		if (isUsingSelectionModel) {
			selectionModel = new CheckTreeSelectionModel(tree.getModel());
			tree.setCellRenderer(new CheckTreeCellRenderer(tree.getCellRenderer(), selectionModel));
			selectionModel.addTreeSelectionListener(this);
//            TreePath p = tree.getPathForRow(0);
//            selectionModel.addSelectionPath(p); 
		} else {
			tree.setCellRenderer(new CheckTreeCellRenderer(tree.getCellRenderer(), tree.getModel()));
		}
		tree.addMouseListener(this);
	}

	@Override
	public void mouseClicked(final MouseEvent me) {
		final TreePath path = tree.getPathForLocation(me.getX(), me.getY());
		if (null == path) {
			return;
		}
		if (me.getX() > tree.getPathBounds(path).x + hotspot) {
			return;
		}
		if (null != selectionModel) {
			// Kumar's implementation
			final boolean selected = selectionModel.isPathSelected(path, true);
			selectionModel.removeTreeSelectionListener(this);
			try {
				if (selected) {
					selectionModel.removeSelectionPath(path);
				} else {
					selectionModel.addSelectionPath(path);
				}
			} finally {
				selectionModel.addTreeSelectionListener(this);
				tree.treeDidChange();
			}
		} else {
			// Peter's implementation
			final Object o = path.getLastPathComponent();
			if (o instanceof CheckTreeNode) {
				final CheckTreeNode anode = (CheckTreeNode) o;
				final TristateCheckBox.State oldState = anode.getSelectionState();
				try {
					anode.setSelected(TristateCheckBox.SELECTED != oldState);
				} finally {
					tree.treeDidChange();
				}
			}
		}
	}

	public CheckTreeSelectionModel getSelectionModel() {
		return selectionModel;
	}
 
	@Override
	public void valueChanged(final TreeSelectionEvent e) {
		tree.treeDidChange();
	}
}
