package de.ped.tools.gui;

import java.util.ArrayList;
import java.util.Stack;

import javax.swing.tree.DefaultTreeSelectionModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

/**
 * A tree selection model to support JTree with a checkbox for each node.
 * 
 * @author Santhosh Kumar T - santhosh@in.fiorano.com, from a <a href=
 *         "http://www.jroller.com/page/santhosh/Weblog/jtree_with_checkboxes?catname=">thread
 *         at jroller.com</a>
 */
public class CheckTreeSelectionModel extends DefaultTreeSelectionModel {
	private static final long serialVersionUID = 1L;

	private TreeModel model;

	public CheckTreeSelectionModel(final TreeModel model) {
		this.model = model;
		setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
		model.getRoot();
		this.rowMapper = null;
	}

	/**
	 * @param path
	 *            the path to check
	 * @return if there is any unselected node in the subtree of given path
	 */
	public boolean isPartiallySelected(final TreePath path) {
		if (isPathSelected(path, true)) {
			return false;
		}
		final TreePath[] selectionPaths = getSelectionPaths();
		if (null == selectionPaths) {
			return false;
		}
		for (int j = 0; j < selectionPaths.length; j++) {
			if (isDescendant(selectionPaths[j], path)) {
				return true;
			}
		}
		return false;
	}
 
	/**
	 * @param path
	 *            the path to check
	 * @param dig
	 *            if <code>true</code>, then a path is assumed to be selected,
	 *            if one of its ancestors is selected
	 * @return whether the given path is selected
	 */
	public boolean isPathSelected(TreePath path, final boolean dig) {
		if (!dig) {
			return super.isPathSelected(path);
		}
		while (null != path && !super.isPathSelected(path)) {
			path = path.getParentPath();
		}
		return (null != path);
	}
 
	/**
	 * @param path1
	 *            the path to check
	 * @param path2
	 *            the path to check
	 * @return if path1 descendant of path2
	 */  
	private boolean isDescendant(final TreePath path1, final TreePath path2) {
		final Object obj1[] = path1.getPath();
		final Object obj2[] = path2.getPath();
		for (int i = 0; i < obj2.length; i++) {
			if (obj1[i] != obj2[i]) {
				return false;
			}
		}
		return true;
	}
 
	@Override
	public void setSelectionPaths(final TreePath[] pPaths) {
		throw new UnsupportedOperationException("not implemented yet!!!");
	}

	@Override
	public void addSelectionPaths(final TreePath[] paths) {
		// unselect all descendants of paths[]
		for (int i = 0; i < paths.length; i++) {
			TreePath path = paths[i];
			TreePath[] selectionPaths = getSelectionPaths();
			if (null == selectionPaths) {
				break;
			}
			final ArrayList<TreePath> toBeRemoved = new ArrayList<TreePath>();
			for (int j = 0; j < selectionPaths.length; j++) {
				if (isDescendant(selectionPaths[j], path)) {
					toBeRemoved.add(selectionPaths[j]);
				}
			}
			super.removeSelectionPaths((TreePath[]) toBeRemoved.toArray(new TreePath[0]));
		}

		// if all siblings are selected then unselect them and select parent
		// recursively
		// otherwize just select that path.
		for (int i = 0; i < paths.length; i++) {
			TreePath path = paths[i];
			TreePath temp = null;
			while (areSiblingsSelected(path)) {
				temp = path;
				if (null == path.getParentPath()) {
					break;
				}
				path = path.getParentPath();
			}
			if (null != temp) {
				if (null != temp.getParentPath()) {
					addSelectionPath(temp.getParentPath());
				} else {
					if (!isSelectionEmpty()) {
						removeSelectionPaths(getSelectionPaths());
					}
					super.addSelectionPaths(new TreePath[] { temp });
				}
			} else {
				super.addSelectionPaths(new TreePath[] { path });
			}
		}
	}

	/**
	 * @param path
	 *            the path to check
	 * @return whether all siblings of the given path are selected
	 */ 
	private boolean areSiblingsSelected(final TreePath path) {
		final TreePath parent = path.getParentPath();
		if (null == parent) {
			return true;
		}
		final Object node = path.getLastPathComponent();
		final Object parentNode = parent.getLastPathComponent();

		final int childCount = model.getChildCount(parentNode);
		for (int i = 0; i < childCount; i++) {
			final Object childNode = model.getChild(parentNode, i);
			if (childNode == node) {
				continue;
			}
			if (!isPathSelected(parent.pathByAddingChild(childNode))) {
				return false;
			}
		}
		return true;
	}

	@Override
	public void removeSelectionPaths(final TreePath[] paths) {
		for (int i = 0; i < paths.length; i++) {
			final TreePath path = paths[i];
			if (1 == path.getPathCount()) {
				super.removeSelectionPaths(new TreePath[] { path });
			} else {
				toggleRemoveSelection(path);
			}
		}
	}

	/**
	 * If any ancestor node of given path is selected, then unselects it and
	 * selection all its descendants except given path and descendants.
	 * Otherwise just unselects the given path.
	 * 
	 * @param path
	 *            the path to check
	 */
	private void toggleRemoveSelection(final TreePath path) {
		final Stack<TreePath> stack = new Stack<TreePath>();
		TreePath parent = path.getParentPath();
		while ((null != parent) && !isPathSelected(parent)) {
			stack.push(parent);
			parent = parent.getParentPath();
		}
		if (null != parent) {
			stack.push(parent);
		} else {
			super.removeSelectionPaths(new TreePath[] { path });
			return;
		}
		while (!stack.isEmpty()) {
			final TreePath temp = (TreePath) stack.pop();
			final TreePath peekPath = stack.isEmpty() ? path : (TreePath) stack.peek();
			final Object node = temp.getLastPathComponent();
			final Object peekNode = peekPath.getLastPathComponent();
			final int childCount = model.getChildCount(node);
			for (int i = 0; i < childCount; i++) {
				final Object childNode = model.getChild(node, i);
				if (childNode != peekNode) {
					super.addSelectionPaths(new TreePath[] { temp
							.pathByAddingChild(childNode) });
				}
			}
		}
		super.removeSelectionPaths(new TreePath[] { parent });
	}
}
