package de.ped.tools.gui;

import java.text.NumberFormat;
import java.text.ParseException;

import javax.swing.JFormattedTextField;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.MathUtil;

/**
 * Verifies an input given in a formatted text field when leaving the text field, and corrects it.
 * Checks the formatted text to be parsable as double, and allows for checks against
 * min and max values.
 * If a dialog is given, this dialog is presented when the check fails and the value is corrected. 
 *
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FormattedTextFieldInputDoubleVerifier extends FormattedTextFieldInputVerifier {
	public static final double MIN_DEFAULT = Double.MIN_VALUE;
	public static final double MAX_DEFAULT = Double.MAX_VALUE;

	private double min = MIN_DEFAULT;
	private double max = MAX_DEFAULT;

	public double getMin() {
		return min;
	}

	public void setMin(final double min) {
		this.min = min;
	}

	public double getMax() {
		return max;
	}

	public void setMax(final double max) {
		this.max = max;
	}

	public FormattedTextFieldInputDoubleVerifier() {
		// Nothing to do
	}

	public FormattedTextFieldInputDoubleVerifier(final PedJDialog parent, final double min, final double max) {
		setParent(parent);
		this.min = min;
		this.max = max;
	}

	@Override
	public boolean verifyText(final JFormattedTextField ftf) throws ParseException {
		final String text = ftf.getText();
		final double value = NumberFormat.getInstance().parse(text).doubleValue();
		final double newValue = MathUtil.bound(value, min, max);
		if (value != newValue) {
			ftf.setText(Double.toString(newValue));
			showInputChangedDialog();
		}
		return true;
	}

	@Override
	protected I18NStringWithFillIns getMessage(final boolean isWarningDialog) {
		final String key;
		int idx = 0;
		if (MIN_DEFAULT != min) {
			idx += 1;
		}
		if (MAX_DEFAULT != max) {
			idx += 2;
		}
		switch(idx) {
		case 1:
			if (0.0 == min) {
				key = "InputVerifier.NonNegative";
			} else {
				key = "InputVerifier.Min";
			}
			break;
		case 2:
			key = "InputVerifier.Max";
			break;
		case 3:
			key = "InputVerifier.MinMax";
			break;
		case 4:
			key = "InputVerifier.Step";
			break;
		default:
			key = "InputVerifier.None";
		}
		final I18NStringWithFillIns finalizer;
		if (isWarningDialog) {
			finalizer = new I18NStringWithFillIns("InputVerifier.Finalizer", "", null);
		} else {
			finalizer = I18NStringWithFillIns.EMPTY;
		}
		return new I18NStringWithFillIns(key, null, new Object[] { 
					finalizer, 
					Double.valueOf(min), 
					Double.valueOf(max), } );
	}
}
