package de.ped.tools.gui;

import java.text.NumberFormat;
import java.text.ParseException;

import javax.swing.JFormattedTextField;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.MathUtil;

/**
 * Verifies an input given in a formatted text field when leaving the text field, and corrects it.
 * Checks the formatted text to be parsable as int, and allows for checks against
 * <ul><li>min and max values,</li><li>the value being divisible without remainder by a given step number.</li></ul>
 * If a dialog is given, this dialog is presented when the check fails and the value is corrected. 
 *
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class FormattedTextFieldInputIntegerVerifier extends FormattedTextFieldInputVerifier {
	public static final int MIN_DEFAULT = Integer.MIN_VALUE;
	public static final int MAX_DEFAULT = Integer.MAX_VALUE;
	public static final int STEP_DEFAULT = 1;

	private int min = MIN_DEFAULT;
	private int max = MAX_DEFAULT;
	private int step = STEP_DEFAULT;

	public int getMin() {
		return min;
	}

	public void setMin(final int min) {
		this.min = min;
	}

	public int getMax() {
		return max;
	}

	public void setMax(final int max) {
		this.max = max;
	}

	public int getStep() {
		return step;
	}

	public void setStep(final int step) {
		this.step = step;
	}

	public FormattedTextFieldInputIntegerVerifier() {
		// Nothing to do
	}

	public FormattedTextFieldInputIntegerVerifier(final PedJDialog parent, final int min, final int max) {
		this(parent, min, max, STEP_DEFAULT);
	}

	public FormattedTextFieldInputIntegerVerifier(final PedJDialog parent, final int min, final int max, final int step) {
		setParent(parent);
		this.step = step;
		this.min = roundDown(min);
		this.max = roundUp(max);
	}

	public int roundUp(final int val) {
		int ret = val;
		if (1 < step) {
			int rem = MathUtil.modulo(val, step);
			if (0 != rem) {
				ret += (step - rem);
			}
		}
		return ret;
	}

	public int roundDown(final int val) {
		int ret = val;
		if (1 < step) {
			ret -= MathUtil.modulo(val, step);
		}
		return ret;
	}

	@Override
	public boolean verifyText(final JFormattedTextField ftf) throws ParseException {
		final String text = ftf.getText();
		final int value = NumberFormat.getInstance().parse(text).intValue();
		final int newValue = roundUp(MathUtil.bound(value, min, max));
		if (value != newValue) {
			ftf.setText(Integer.toString(newValue));
			showInputChangedDialog();
		}
		return true;
	}

	@Override
	protected I18NStringWithFillIns getMessage(final boolean isWarningDialog) {
		final String key;
		int idx = 0;
		if (MIN_DEFAULT != min) {
			idx += 1;
		}
		if (MAX_DEFAULT != max) {
			idx += 2;
		}
		if (STEP_DEFAULT != step) {
			idx += 4;
		}
		switch(idx) {
		case 1:
			if (0 == min) {
				key = "InputVerifier.NonNegative";
			} else {
				key = "InputVerifier.Min";
			}
			break;
		case 2:
			key = "InputVerifier.Max";
			break;
		case 3:
			key = "InputVerifier.MinMax";
			break;
		case 4:
			key = "InputVerifier.Step";
			break;
		case 5:
			key = "InputVerifier.MinStep";
			break;
		case 6:
			key = "InputVerifier.MaxStep";
			break;
		case 7:
			key = "InputVerifier.MinMaxStep";
			break;
		default:
			key = "InputVerifier.None";
		}
		final I18NStringWithFillIns finalizer;
		if (isWarningDialog) {
			finalizer = new I18NStringWithFillIns("InputVerifier.Finalizer", "", null);
		} else {
			finalizer = I18NStringWithFillIns.EMPTY;
		}
		return new I18NStringWithFillIns(key, null, new Object[] { 
				finalizer, 
				Integer.valueOf(min), 
				Integer.valueOf(max), 
				Integer.valueOf(step), } );
	}
}
