package de.ped.tools.gui;

import java.text.ParseException;

import javax.swing.InputVerifier;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFormattedTextField;
import javax.swing.JOptionPane;
import javax.swing.JFormattedTextField.AbstractFormatter;

import de.ped.tools.I18NStringWithFillIns;

/**
 * Verifies an input given in a formatted text field when leaving the text field.
 * Might alter the value to an allowed range, or revert to the last accepted value.
 *
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class FormattedTextFieldInputVerifier extends InputVerifier {

	/** 
	 * The parent dialog containing the field to be checked. Needed to position an input changed dialog correctly. 
	 */
	private PedJDialog parent;

	public PedJDialog getParent() {
		return parent;
	}

	public void setParent(final PedJDialog parent) {
		this.parent = parent;
	}

	@Override
	public boolean verify(final JComponent input) {
		if (input instanceof JFormattedTextField) {
			final JFormattedTextField ftf = (JFormattedTextField) input;
			final AbstractFormatter formatter = ftf.getFormatter();
			if (null != formatter) {
				try {
					final boolean ret = verifyText(ftf);
					if (ret) {
						formatter.stringToValue(ftf.getText());
					}
					return ret;
				} catch (final ParseException pe) {
					return false;
				}
			}
		}
		return true;
	}

	@Override
	public boolean shouldYieldFocus(final JComponent input) {
		return verify(input);
	}

	/**
	 * Override this method to perform the checks. Default implementation
	 * accepts all.
	 * @param ftf the text field
	 * @return if the text was valid
	 * @throws ParseException if the text could not be parsed
	 */
	public boolean verifyText(final JFormattedTextField ftf) throws ParseException {
		return true;
	}

	/**
	 * Should be called from within {@link #verifyText(JFormattedTextField)} whenever the values needed to be corrected.
	 */
	protected void showInputChangedDialog() {
		JDialog d = null;
		if (null != getParent()) {
			d = getDefaultInputChangedDialog();
		}
		if (null != d) {
			d.setVisible(true);
		}
	}

	protected JDialog getDefaultInputChangedDialog() {
		PedJDialog ret = null;
		final PedJDialog parent = getParent();
		if (null != parent) {
			ret = new JInfoDialog(parent, 
					parent.getMessages().getString("InputVerifier.Title"), 
					parent.getMessages().getString(getMessage(true)),
					JOptionPane.WARNING_MESSAGE);
		}
		return ret;
	}

	public I18NStringWithFillIns getToolTip() {
		return getMessage(false);
	}

	/**
	 * @param isWarningDialog
	 *            if it's the text for a warning dialog
	 * @return either the text for the warning dialog, or the text for a tool
	 *         tip or help file.
	 */
	protected abstract I18NStringWithFillIns getMessage(final boolean isWarningDialog);
}
