package de.ped.tools.gui;

import java.awt.Container;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;

import javax.swing.BorderFactory;
import javax.swing.JTextArea;
import javax.swing.UIManager;

import de.ped.tools.log.Logger;

/**
 * A helper class for defining some algorithms around layout. Currently it only
 * contains methods for positioning a child window in the center of the parent
 * window.
 * 
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GUIAlgorithms {
	private static Logger logger = Logger.getRootLogger();

	/**
	 * Calculates the point of the upper left corner of a child rectangle with a
	 * given dimension so that this child rectangle is in the center of a parent
	 * rectangle.
	 * 
	 * @param childBounds
	 *            the size of the child
	 * @param parentBounds
	 *            the position and size of the parent
	 * @return the position the child should have
	 */
	public static Point center(Dimension childBounds, Rectangle parentBounds) {
		if (null == childBounds) {
			childBounds = new Dimension(0, 0);
		}
		if (null == parentBounds) {
			parentBounds = new Rectangle(0, 0, 800, 600);
		}
		int x = Math.max(0, parentBounds.x + (parentBounds.width - childBounds.width) / 2);
		int y = Math.max(0, parentBounds.y + (parentBounds.height - childBounds.height) / 2);
		return new Point(x, y);
	}

	/**
	 * Sets the location of the child to the center of the parent. 
	 * So call it after calculating the size!
	 * 
	 * @param child
	 *            the child that will be located
	 * @param parent
	 *            the parent
	 */
	public static void center(final Container child, final Container parent) {
		try {
			Dimension childBounds = child.getSize();
			Rectangle parentBounds;
			if (null != parent) {
				parentBounds = parent.getBounds();
			} else {
				parentBounds = new Rectangle(0, 0, 800, 600);
			}
			int x = Math.max(0, parentBounds.x + (parentBounds.width - childBounds.width) / 2);
			int y = Math.max(0, parentBounds.y + (parentBounds.height - childBounds.height) / 2);
			child.setLocation(x, y);
		} catch (final java.lang.NoSuchMethodError e) {
			logger.info(e + "Ah, using an old browser, eh? Ok, I won't center the dialog on the screen...");
		}
	}

	/** 
	 * Centers the given window in its owner. 
	 * So call it after calculating the size!
	 * 
	 * @param child
	 *            the child that will be located
	 */
	public static void center(final Window child) {
		center(child, child.getOwner());
	}

	/**
	 * Takes a text area fields and modifies it in a way that it looks and
	 * behaves like a label, but with the capacities of the area like
	 * multi-lining.
	 * 
	 * @param area
	 *            the text area to modify
	 */
	public static void modifyToLabelLAF(final JTextArea area) {
		area.setEditable(false);
		area.setFont(UIManager.getFont("Label.font"));
		area.setBackground(UIManager.getColor("Label.background"));
		area.setForeground(UIManager.getColor("Label.foreground"));
		area.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		area.setFocusable(false);
	}
}
