package de.ped.tools.gui;

import java.awt.Component;
import java.awt.MediaTracker;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.swing.DefaultListCellRenderer;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.ListCellRenderer;

import de.ped.tools.LocalizedString;
import de.ped.tools.ResourceFinder.Folder;
import de.ped.tools.log.Logger;

/**
 * A combo box with added functionality for an icon per entry and additional information for the help creation.
 * 
 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class JIconComboBox<T> extends JComboBox<T> {
	private static final long serialVersionUID = 1L;

	/**
	 * A list renderer that can render any object as containing an icon and text.
	 * 
	 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de), based on a 
	 * 	<a href="http://helpdesk.objects.com.au/java/how-do-i-display-a-jcombobox-with-icons-and-text">solution</a> found on the web
	 */
	public class IconListRenderer extends DefaultListCellRenderer {
		private static final long serialVersionUID = 1L;
		@Override
		public Component getListCellRendererComponent(JList<?> list, Object value, int index, boolean isSelected,
				boolean cellHasFocus) {
			// Get the renderer component from parent class
			final JLabel label = (JLabel) super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
			// Set icon to display for value
			label.setIcon(icons.get(value));
			return label;
		}
	}

	public JIconComboBox() {
		super();
	}

	public JIconComboBox(final Vector<T> items) {
		super(items);
	}

	/**
	 * @return if help for this combo box should list all the elements within this combo
	 */
	public boolean isHelpWithElements() {
		return isHelpWithElements;
	}

	/**
	 * @param isHelpWithElements
	 *            if help for this combo box should list all the elements within this combo
	 */
	public void setHelpWithElements(final boolean isHelpWithElements) {
		this.isHelpWithElements = isHelpWithElements;
	}

	/**
	 * @param folder
	 *            the resources folder where to find the icon images
	 */
	public void setFolder(final Folder folder) {
		this.folder = folder;
	}

	/**
	 * @return the resources folder where to find the icon images
	 */
	public Folder getFolder() {
		return folder;
	}

	/**
	 * Adds an icon to an item
	 * @param item the item to add the icon for
	 * @param icon the icon to add
	 */
	public void addIcon(final T item, final Icon icon) {
		if (null != icon) {
			icons.put(item, icon);
			final ListCellRenderer<? super T> renderer = getRenderer();
			if (renderer instanceof JIconComboBox.IconListRenderer) {
				// Nothing to do
			} else {
				setRenderer(new IconListRenderer());
			}
		}
	}
	
	/**
	 * Returns the image filename of the icon for a given item, if any. 
	 * @param item the item to search the icon for
	 * @return the image filename of the icon or <code>null</code> if no icon was defined or icon was not loadable
	 */
	public String getImageFilename(final Object item) {
		String ret = null;
		if (item instanceof LocalizedString) {
			final Icon icon = getIcon(item);
			if ((null != icon) && (icon instanceof ImageIcon)) {
				final ImageIcon ii = (ImageIcon) icon;
				Logger.getRootLogger().trace("LoadStatus=" + ii.getImageLoadStatus() + " " + MediaTracker.ERRORED + " " + MediaTracker.COMPLETE);
				if (MediaTracker.ERRORED != ii.getImageLoadStatus()) {
					final String url = ii.getDescription();
					String[] split = url.split("/");
					ret = split[split.length - 1];
					Logger.getRootLogger().trace("getImageFilename(" + item + ") has URL=" + url + " and returns filename=" + ret);
				}
			}
		}
		return ret;
	}
	
	/**
	 * Returns the icon to a given item, if any. There is no guarantee that the icon is loadable.
	 * @param item the item to search the icon for
	 * @return the icon or <code>null</code> if no icon was defined
	 */
	public Icon getIcon(final Object item) {
		return icons.get(item);
	}
	
	private Map<Object, Icon> icons = new HashMap<Object, Icon>();
	private boolean isHelpWithElements = true;
	private Folder folder = null;
}