package de.ped.tools.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.util.Dictionary;
import java.util.Hashtable;

import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.CollectionUtil;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.SliderDef;
import de.ped.tools.log.Logger;

/**
 * Slider and text field which both show the same value
 * 
 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class JSliderWithTextFieldPanel extends JPanel {
	private static final long serialVersionUID = 1L;

	/**
	 * Slider controller that updates a text field or label when a slider
	 * changed its value
	 * 
	 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class SliderToTextChangeListener implements ChangeListener {
		
		public SliderToTextChangeListener(final JLabel target) {
			this.targetLabel = target;
			this.targetText = null;
		}

		public SliderToTextChangeListener(final JFormattedTextField target) {
			this.targetText= target;
			this.targetLabel = null;
		}

		@Override
		public void stateChanged(ChangeEvent e) {
			final JSlider source = (JSlider) e.getSource(); 
			final int value = source.getValue();
			if (null != targetLabel) {
				targetLabel.setText(Integer.toString(value));
			}
			if (null != targetText) {
				targetText.setValue(Integer.valueOf(value));
			}
		}
		
		private final JLabel targetLabel;
		private final JFormattedTextField targetText;
		
	}

	public JSliderWithTextFieldPanel(final PedJDialog dialog, final int min, final int max, final int step, final String keysPrefix, ResourceFinder.Folder folder) {
		super();
		defineSlider(dialog.getApplicationEnvironment(), slider, min, max, step, keysPrefix, folder);
		text.setInputVerifier(new FormattedTextFieldInputIntegerVerifier(dialog, min, max, step));
		text.setEditable(false);
		text.setFocusable(false);
		definePanel();
	}

	public JSliderWithTextFieldPanel(final PedJDialog dialog, final SliderDef def, final String keysPrefix, ResourceFinder.Folder folder) {
		this(dialog, def.min, def.max, def.step, keysPrefix, folder);
	}

	public int getValue() {
		return slider.getValue();
	}

	public void setValue(final int value) {
		slider.setValue(value);
		text.setValue(Integer.valueOf(value));
		text.setColumns(4);
		text.setHorizontalAlignment(JTextField.RIGHT);
	}

	@Override
	public void setEnabled(final boolean enabled) {
		slider.setEnabled(enabled);
		slider.setFocusable(enabled);
	}

	private void defineSlider(final ApplicationEnvironment env, final JSlider slider, final int min,
			final int max, final int step, final String keysPrefix, ResourceFinder.Folder folder) {
		slider.setMinimum(min);
		slider.setMaximum(max);
		if (1 < step) {
			slider.setMinorTickSpacing(step);
			slider.setSnapToTicks(true);
		} else {
			slider.setMajorTickSpacing(Math.round((max - min) / 10.0F));
		}
		slider.setPaintTicks(true);
		final Hashtable<Integer, JLabel> labelDict = new Hashtable<Integer, JLabel>();
		labelDict.put(new Integer(min), createMinMaxLabel(env, keysPrefix + ".Min", folder));
		labelDict.put(new Integer(max), createMinMaxLabel(env, keysPrefix + ".Max", folder));
		slider.setLabelTable(labelDict);
		slider.setPaintLabels(false);
	}

	private JLabel createMinMaxLabel(final ApplicationEnvironment env, final String key, final ResourceFinder.Folder folder) {
		final Messages messages = env.messages();
		final ResourceFinder finder = env.resources();
		final JLabel label = new JLabel(messages.getText(key));
		final String toolTip = messages.getShortDescription(key);
		if (!CollectionUtil.isNullOrEmpty(toolTip)) {
			label.setToolTipText(toolTip);
		}
		if ((null != finder) && (null != folder)) {
			final java.net.URL url = finder.getURLToResource(key + ".png", folder);
			label.setIcon(new ImageIcon(url));
			Logger.getLogger(getClass()).trace("createMinMaxLabel(" + key + "," + folder.name() + ") -> URL=" + url);
		}
		return label;
	}

	private void definePanel() {
		setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		final Dimension minMaxSize = new Dimension(95, 15);
		@SuppressWarnings("unchecked")
		final Dictionary<Integer, JLabel> dict = slider.getLabelTable();
		if (null != dict) {
			final JLabel minLabel = dict.get(Integer.valueOf(slider.getMinimum()));
			final JLabel maxLabel = dict.get(Integer.valueOf(slider.getMaximum()));
			final JPanel minPanel = new JPanel();
			final JPanel maxPanel = new JPanel();
			slider.addChangeListener(new SliderToTextChangeListener(text));
			minLabel.setPreferredSize(minMaxSize);
			maxLabel.setPreferredSize(minMaxSize);
			// Set icon near to slider
			minLabel.setHorizontalTextPosition(SwingConstants.LEFT);
			maxLabel.setHorizontalTextPosition(SwingConstants.RIGHT);
			// Align text to slider
			minLabel.setHorizontalAlignment(SwingConstants.RIGHT);
			maxLabel.setHorizontalAlignment(SwingConstants.LEFT);
			
			final Color foreground = Color.GRAY;
			minLabel.setForeground(foreground);
			maxLabel.setForeground(foreground);
			minPanel.add(minLabel);
			maxPanel.add(maxLabel);
			add(minPanel);
			add(slider);
			add(maxPanel);
			add(text);
		} else {
			add(slider);
			add(text);
		}
	}

	private final JSlider slider = new JSlider();
	private final JFormattedTextField text = new JFormattedTextField();
}