package de.ped.tools.gui;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;

import de.ped.tools.TextFormatter;

/**
 * A label that may contain several lines of text
 *
 * @author David Flanagan, (c) 1996 O'Reilly &amp; Associates
 */
public class MultiLineLabel extends Canvas {
	private static final long serialVersionUID = 1L;

	protected String[] lines;
	protected int numLines;
	protected int marginWidth;
	protected int marginHeight;
	protected int lineHeight;
	protected int lineAscent;
	protected int[] lineWidths;
	protected int maxWidth;
	protected int alignment = TextFormatter.LEFT;

	/**
	 * Tokenizes the string that may contains newlines and breaks it up.
	 * 
	 * @param label
	 *            the text to be broken in several strings
	 */
	protected void newLabel(final String label) {
		this.lines = TextFormatter.wrap(label);
		this.numLines = lines.length;
		this.lineWidths = new int[numLines];
	}

	/**
	 * Calculates the maximum width of all text lines.
	 */
	protected void measure() {
		final FontMetrics fm = getFontMetrics(getFont());
		if (null == fm) {
			return;
		}
		lineHeight = fm.getHeight();
		lineAscent = fm.getAscent();
		maxWidth = 0;
		for (int i = 0; i < numLines; i++) {
			lineWidths[i] = fm.stringWidth(lines[i]);
			if (lineWidths[i] > maxWidth) {
				maxWidth = lineWidths[i];
			}
		}
	}

	/**
	 * Creates a multi line label.
	 * 
	 * @param aLabel
	 *            the text to be displayed
	 * @param aMargin_width
	 *            the space at the left and right
	 * @param aMargin_height
	 *            the space at the upper and lower border
	 * @param anAlignment
	 *            one of {@link TextFormatter#LEFT}, {@link TextFormatter#CENTER} or {@link TextFormatter#RIGHT}
	 * @see TextFormatter
	 */
	public MultiLineLabel(final String aLabel, final int aMargin_width, final int aMargin_height, final int anAlignment) {
		newLabel(aLabel);
		this.marginWidth  = aMargin_width;
		this.marginHeight = aMargin_height;
		this.alignment    = anAlignment;
	}

	/**
	 * Creates a multi line label with left alignment.
	 * 
	 * @param aLabel
	 *            the text to be displayed
	 * @param aMargin_width
	 *            the space at the left and right
	 * @param aMargin_height
	 *            the space at the upper and lower border
	 */
	public MultiLineLabel(final String aLabel, final int aMargin_width, final int aMargin_height) {
		newLabel(aLabel);
		this.marginWidth  = aMargin_width;
		this.marginHeight = aMargin_height;
		this.alignment    = TextFormatter.LEFT;
	}

	/**
	 * Creates a multi line label with default size.
	 * 
	 * @param aLabel
	 *            the text to be displayed
	 * @param anAlignment
	 *            one of {@link TextFormatter#LEFT}, {@link TextFormatter#CENTER} or {@link TextFormatter#RIGHT}
	 */
	public MultiLineLabel(final String aLabel, final int anAlignment) {
		newLabel(aLabel);
		this.marginWidth  = 10;
		this.marginHeight = 10;
		this.alignment    = anAlignment;
	}

	/**
	 * Creates a multi line label with default size and left alignment.
	 * 
	 * @param aLabel
	 *            the text to be displayed
	 */
	public MultiLineLabel(final String aLabel) {
		newLabel(aLabel);
		this.marginWidth  = 10;
		this.marginHeight = 10;
		this.alignment    = TextFormatter.LEFT;
	}

	/**
	 * Sets the text to be displayed
	 * 
	 * @param aLabel
	 *            the text to be displayed
	 */
	public void setLabel(final String aLabel) {
		newLabel(aLabel);
		measure();
		repaint();
	}

	/**
	 * Sets the font in which to display the label.
	 * 
	 * @param font
	 *            the font to set
	 */
	@Override
	public void setFont(final Font font) {
		super.setFont(font);
		measure();
		repaint();
	}

	/**
	 * Sets the foreground color in which to display the label.
	 * 
	 * @param color
	 *            the color to set
	 */
	@Override
	public void setForeground(final Color color) {
		super.setForeground(color);
		repaint();
	}

	/**
	 * @param anAlignment
	 *            one of {@link TextFormatter#LEFT}, {@link TextFormatter#CENTER} or {@link TextFormatter#RIGHT}
	 */
	public void setAlignment(final int anAlignment) {
		this.alignment = anAlignment;
		repaint();
	}

	public void setMarginWidth(final int mw) {
		this.marginWidth = mw;
		repaint();
	}

	public void setMarginHeight(final int mh) {
		this.marginHeight = mh;
		repaint();
	}

	/**
	 * @return one of {@link TextFormatter#LEFT}, {@link TextFormatter#CENTER} or {@link TextFormatter#RIGHT}
	 */
	public int getAlignment() {
		return alignment;
	}

	public int getMarginWidth() {
		return marginWidth;
	}

	public int getMarginHeight() {
		return marginHeight;
	}

	@Override
	public void addNotify() {
		super.addNotify();
		measure();
	}

	@Override
	public Dimension getPreferredSize() {
		return new Dimension(
			maxWidth              + 2 * marginWidth,
			numLines * lineHeight + 2 * marginHeight);
	}

	@Override
	public Dimension getMinimumSize() {
		return new Dimension(
			maxWidth,
			numLines * lineHeight);
	}

	@Override
	public void paint(final Graphics g) {
		final Dimension d = getSize();
		int y = lineAscent + (d.height - numLines * lineHeight) / 2;
		int x;
		for (int i = 0; i < numLines; i++, y += lineHeight) {
			switch(alignment) {
			case TextFormatter.LEFT:
				x = marginWidth;
				break;
			case TextFormatter.RIGHT:
				x = d.width - marginWidth - lineWidths[i];
				break;
			case TextFormatter.CENTER:
			default:
				x = (d.width - lineWidths[i]) / 2;
				break;
			}
			g.drawString(lines[i], x, y);
		}
	}
}
