package de.ped.tools.gui;

import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;

import de.ped.tools.ResourceFinder;
import de.ped.tools.log.Logger;

/**
 * Actions, as defined in Swing, with support for I18N with UI texts.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PedAction extends AbstractAction {
	private static final long serialVersionUID = 1L;

	/** The key to map the message key string for I18N to. */
	public static final String MESSAGE_KEY = "MESSAGE_KEY";

	/** The key to map the translated pure name (without dialolg indicator) to. */
	public static final String PURE_NAME = "PURE_NAME";

	/** Info if the resources like images are common between applications. */
	public static final String IS_COMMON = "IS_COMMON";

	/** The URL of the small icon, if any. */
	public static final String SMALL_ICON_URL = "SMALL_ICON_URL";

	public PedAction() {
		super();
	}

	public PedAction(final String text) {
		super(text);
	}

	public PedAction(final String text, final Icon icon) {
		super(text, icon);
	}

	public PedAction(String messageKey, final ApplicationMainWindow appMainWindow, final ResourceFinder.Folder folder) {
		this(messageKey, appMainWindow, folder, false);
	}

	public PedAction(String messageKey, final ApplicationMainWindow appMainWindow, final ResourceFinder.Folder folder, final boolean isOpeningDialog) {
		setMessageKeyPrefix(messageKey, appMainWindow, folder);
		setStartingDialog(isOpeningDialog);
		setLogger(Logger.getLogger(appMainWindow.getClass()));
	}

	/**
	 * @return whether this action starts a dialog, as this might be indicated
	 *         with the action representation in a menu
	 */
	public boolean isStartingDialog() {
		return isStartingDialog;
	}

	/**
	 * Sets whether this action starts a dialog, as this might be indicated with
	 * the action representation in a menu.
	 * 
	 * @param isStartingDialog
	 *            <code>true</code> if this action starts a dialog
	 */
	public void setStartingDialog(final boolean isStartingDialog) {
		this.isStartingDialog = isStartingDialog;
	}

	private boolean isStartingDialog = false;

	public void putValues(final ApplicationMainWindow appMainWindow) {
		final String key = (String) getValue(MESSAGE_KEY);
		if (null != key) {
			appMainWindow.putValues(this, key);
		}
	}

	public void setMessageKeyPrefix(final String key) {
		setMessageKeyPrefix(key, null, ResourceFinder.Folder.APP);
	}

	public void setMessageKeyPrefix(final String key, final ApplicationMainWindow parent, final ResourceFinder.Folder folder) {
		putValue(MESSAGE_KEY, key);
		if (null != parent) {
			listener = parent;
			if (null != folder) {
				final ResourceFinder res = parent.getApplicationEnvironment().resources();
				if (null != res) {
					final java.net.URL url = res.getURLToResource(key + ".png", folder);
					putValue(Action.SMALL_ICON, new ImageIcon(url));
					putValue(SMALL_ICON_URL, url);
					putValue(IS_COMMON, ResourceFinder.Folder.COMMON == folder);
				}
			}
		}
	}

	public String getMessageKeyPrefix() {
		return (String) getValue(MESSAGE_KEY);
	}

	public boolean isCommonResource() {
		boolean ret = false;
		final Object val = getValue(IS_COMMON);
		if (null != val) {
			ret = (Boolean) val;
		}
		return ret;
	}

	/**
	 * @return any potential I18N params which will be applied to any I18N'ed
	 *         keys. Default implementation just returns <code>null</code>.
	 */
	protected Object[] getI18NParams() {
		return null;
	}

	/**
	 * @return <code>true</code>, if this action should show up in the menu - no
	 *         matter if enabled or disabled. Default implementation returns
	 *         always <code>true</code>.
	 */
	public boolean isToBeDefined() {
		return true;
	}

	/**
	 * @return If <code>true</code>, this action should show up enabled in the menu;
	 *         if <code>false</code>, this action should show up disabled
	 *         (grayed-out) in the menu. Default implementation returns always
	 *         <code>true</code>.
	 */
	public boolean isToBeEnabled() {
		return true;
	}

	/**
	 * Sets the action enabled or disabled, based on {@link #isEnabled()}.
	 */
	public void updateEnabled() {
		setEnabled(isToBeEnabled());
	}

	public void setLogger(final Logger logger) {
		this.paalogger = logger;
	}

	@Override
	public void actionPerformed(final ActionEvent event) {
		try {
			securedActionPerformed(event);
		} catch (final Throwable ex) {
			if (null != paalogger) {
				paalogger.errorWithStackTrace("This should never happen!", ex);
			} else {
				System.out.println("This should never happen: " + ex.toString());
				ex.printStackTrace();
			}
		}
		if (null != listener) {
			listener.pedAbstractActionPerformed(this, event);
		}
	}

	@Override
	public String toString() {
		String ret = getMessageKeyPrefix();
		if (null != ret) {
			ret = "PedAction(" + ret + ")";
		} else {
			ret = super.toString();
		}
		return ret;
	}

	public ApplicationMainWindow getParent() {
		return listener;
	}

	protected void securedActionPerformed(final ActionEvent event) {
		// Default implementation does nothing
	}

	protected Logger paalogger = null;

	protected ApplicationMainWindow listener;
}
