package de.ped.tools.gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JDialog;
import javax.swing.JFormattedTextField;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.CollectionUtil;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.ResourceFinder;
import de.ped.tools.log.Logger;

/**
 * An abstract class for a modal dialog containing the following:
 * <ul>
 * <li>Border layout</li>
 * <li>Support for an icon on the left side of the dialog</li>
 * <li>Support for an explanatory string at the top of the dialog</li>
 * <li>A button bar at the bottom of the dialog, containing at least an OK button,
 *  optionally a Cancel button, and support for some more buttons before these two.</li>
 * <li>Support for the display of properties (data elements displayed as label/UI component pair)</li>
 * <li>Support for data synchronization with an underlying data model (e.g., for dependencies in the display of those properties)</li>
 * </ul>
 * 
 * @author Copyright (c) 2004 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class PedJDialog extends JDialog implements WindowListener, ActionListener {
	/**
	 * Shows the online help and scrolls directly to the given anchor in the online help document.
	 * 
	 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class OnlineHelpActionListener implements ActionListener {
		/**
		 * @param parent the application main window that can build the online help document.
		 * @param anchor the anchor within the online help document to scroll to, or <code>null</code> to start at the top of the document.
		 */
		public OnlineHelpActionListener(final ApplicationMainWindow parent, final String anchor) {
			this.parent = parent;
			this.anchor = anchor;
		}

		@Override
		public void actionPerformed(final ActionEvent ev) {
			final HtmlPaneDialog d = parent.createHelpBox();
			if (!CollectionUtil.isNullOrEmpty(anchor)) {
				final String a;
				if ("#".equals(anchor.substring(0,  1))) {
					a = anchor;
				} else {
					a = "#" + anchor;
				}
				d.scrollToInternalAnchor(a);
			}
			d.enterDialog();
		}
		
		public final ApplicationMainWindow parent;
		public final String anchor;
	}

	private static final long serialVersionUID = 1L;
	private static final String BUTTON_OK_NAME     = "   OK   ";
	private static final String BUTTON_CANCEL_NAME = "Abbrechen";

	protected static final int MAIN_PANEL_BORDER_TOP = 3;
	protected static final int MAIN_PANEL_BORDER_LEFT = 3;
	protected static final int MAIN_PANEL_BORDER_RIGHT = 3;
	protected static final int MAIN_PANEL_BORDER_BOTTOM = 3;

	private static final int TITLE_PANEL_BORDER_TOP = 3;
	private static final int TITLE_PANEL_BORDER_LEFT = 3;
	private static final int TITLE_PANEL_BORDER_RIGHT = 3;
	private static final int TITLE_PANEL_BORDER_BOTTOM = 3;

	private static final int BUTTON_PANEL_BORDER_TOP = 3;
	private static final int BUTTON_PANEL_BORDER_LEFT = 3;
	private static final int BUTTON_PANEL_BORDER_RIGHT = 3;
	private static final int BUTTON_PANEL_BORDER_BOTTOM = 3;

	public static void setValue(final JFormattedTextField tf, final Object value, final int width) {
		tf.setValue(value);
		tf.setColumns(width);
		tf.setHorizontalAlignment(JTextField.RIGHT);
	}

	/**
	 * Initializes the dialog, when the UI parent is a frame.
	 * 
	 * @param mainWindow
	 *            the application main window, needed for internationalization
	 *            support
	 * @param parent
	 *            the direct UI parent
	 */
	protected PedJDialog(final ApplicationMainWindow mainWindow, final Frame parent) {
		super(parent, true);
		this.mainWindow = mainWindow;
		init();
	}

	/**
	 * Initializes the dialog, when the UI parent is a dialog.
	 * 
	 * @param mainWindow
	 *            the application main window, needed for internationalization
	 *            support
	 * @param parent
	 *            the direct UI parent
	 */
	protected PedJDialog(final ApplicationMainWindow mainWindow, final Dialog parent) {
		super(parent, true);
		this.mainWindow = mainWindow;
		init();
	}

	private void init() {
		getContentPane().setLayout(new BorderLayout());
		addWindowListener(this);
		buttonPanel.setLayout(new FlowLayout(FlowLayout.CENTER));
		buttonPanel.setBorder(BorderFactory.createEmptyBorder(BUTTON_PANEL_BORDER_TOP, BUTTON_PANEL_BORDER_LEFT, BUTTON_PANEL_BORDER_BOTTOM, BUTTON_PANEL_BORDER_RIGHT));
		getContentPane().add(buttonPanel, BorderLayout.SOUTH);
		buttonOk = mainWindow.createButton(new I18NStringWithFillIns("Dialog.OK", BUTTON_OK_NAME, null));
		buttonCancel = mainWindow.createButton(new I18NStringWithFillIns("Dialog.Cancel", BUTTON_CANCEL_NAME, null));
	}

	public void setDefaultButton(final JButton button) {
		getRootPane().setDefaultButton(button);
	}

	public JButton getDefaultButton() {
		return getRootPane().getDefaultButton();
	}

	/**
	 * Helping method. Call it from constructors only, and this should be the
	 * last call within the constructor!
	 */
	protected void finishLayout() {
		addButton(buttonOk);
		final boolean isWithCancelButton = (JOptionPane.OK_CANCEL_OPTION == optionType);
		if (isWithCancelButton) {
			addButton(buttonCancel);
		}
		pack();
		setMinimumSizeTo(buttonPanel);
		GUIAlgorithms.center(this);
	}

	protected void setMinimumSizeTo(final JPanel buttonPanel) {
		final Dimension buttonSize = buttonPanel.getPreferredSize();
		setMinimumSize(new Dimension(buttonSize.width + 16, buttonSize.height));
	}

	public ApplicationMainWindow getMainWindow() {
		return mainWindow;
	}

	public ApplicationEnvironment getApplicationEnvironment() {
		return mainWindow.getApplicationEnvironment();
	}

	public Messages getMessages() {
		return getApplicationEnvironment().messages();
	}

	public JPanel createMainPanel() {
		final JPanel mainPanel = new JPanel();
		mainPanel.setBorder(BorderFactory.createEmptyBorder(MAIN_PANEL_BORDER_TOP, MAIN_PANEL_BORDER_LEFT, MAIN_PANEL_BORDER_BOTTOM, MAIN_PANEL_BORDER_RIGHT));
		return mainPanel;
	}

	protected void addHelpButton(final String anchor) {
		buttonPanel.add(createHelpButton(anchor));
		buttonPanel.add(Box.createHorizontalStrut(30));
	}

	protected void addButton(final JButton button) {
		buttonPanel.add(button);
		button.addActionListener(this);
	}

	protected JButton getButtonOk() {
		return buttonOk;
	}

	protected JButton getButtonCancel() {
		return buttonCancel;
	}
	
	protected JPanel getButtonPanel() {
		return buttonPanel;
	}

	public JLabel getTitleLabel() {
		return titleLabel;
	}

	protected void addTitlePanel(final JPanel panel) {
		panel.setBorder(BorderFactory.createEmptyBorder(TITLE_PANEL_BORDER_TOP, TITLE_PANEL_BORDER_LEFT, TITLE_PANEL_BORDER_BOTTOM, TITLE_PANEL_BORDER_RIGHT));
		getContentPane().add(panel, BorderLayout.NORTH);
	}

	public void setTitlePanel(final String title) {
		titleLabel = new JLabel(title);
		final JPanel titlePanel = new JPanel();
		titlePanel.setLayout(new FlowLayout(FlowLayout.CENTER));
		titlePanel.add(titleLabel);
		addTitlePanel(titlePanel);
	}

	JLabel titleLabel;
	protected int optionType = JOptionPane.DEFAULT_OPTION;

	@Override
	public void windowOpened(final WindowEvent e) {
		wasOkPressed = false;
	}

	/**
	 * Invoked when the user attempts to close the window from the window's system menu.
	 * Default implementation leaves the dialog without accepting the changes done in the dialog.
	 */
	@Override
	public void windowClosing(final WindowEvent e) {
		onCancel();
	}

	@Override
	public void windowClosed(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowIconified(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowDeiconified(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowActivated(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void windowDeactivated(final WindowEvent e) {
		// Nothing to do
	}

	@Override
	public void actionPerformed(final ActionEvent e) {
		if (e.getSource() == buttonOk) {
			onOk();
		} else if (e.getSource() == buttonCancel) {
			onCancel();
		}
	}

	/**
	 * Can be called after the dialog was left.
	 * 
	 * @return <code>true</code> if the user left the dialog in acceptance of all
	 *         changes applied, e.g., by pressing the OK button
	 */
	public boolean wasAcceptedDialogChanges() {
		return wasOkPressed;
	}

	/** (Re-)Initializes and opens the dialog. */
	public void enterDialog() {
		wasOkPressed = false;
		updateUIFieldsBasedOnProperties();
		showDialogThreadSafe();
	}

	/**
	 * Wraps {@link JDialog#setVisible(boolean)} to be started within another
	 * thread, if not called from the event dispatcher thread.
	 */
	public void showDialogThreadSafe() {
		if(SwingUtilities.isEventDispatchThread()){
			setVisible(true);
		} else {
			if (null != mainWindow) {
				mainWindow.setWaitingForModal(this, true);
			}
			final Logger logger = Logger.getRootLogger();
			logger.info(getClass().getName() + " was called in a non-thread-safe context!");
			SwingUtilities.invokeLater(new Runnable() {
				@Override
				public void run() {
					setVisible(true);
				}
			});
		}
	}

	@Override
	public void setVisible(final boolean b) {
		if (b && (null != mainWindow)) {
			mainWindow.setWaitingForModal(this, true);
		}
		super.setVisible(b);
		if (!b && (null != mainWindow)) {
			mainWindow.setWaitingForModal(this, false);
		}
	}

	/** Leaves and closes the dialog. */
	protected void leaveDialog() {
		setVisible(false);
	}

	/** Called when OK button is pressed. Default implementation leaves dialog. */
	protected void onOk() {
		wasOkPressed = true;
		leaveDialog();
	}

	/** Called when Cancel button is pressed. Default implementation leaves dialog. */
	public void onCancel() {
		leaveDialog();
	}

	/** 
	 * Writes the UI fields according to the properties. 
	 * Inverse operation to {@link #readUIFieldsIntoProperties()}.
	 * Default implementation does nothing, override in subclasses. 
	 */
	public void updateUIFieldsBasedOnProperties() {
		// Default implementation does nothing
	}

	/** 
	 * Writes the model according to the GUI fields. 
	 * Inverse operation to {@link #updateUIFieldsBasedOnProperties()}.
	 * Default implementation does nothing, override in subclasses. 
	 */
	public void readUIFieldsIntoProperties() {
		// Default implementation does nothing
	}

	/**
	 * Sets all property elements.
	 * 
	 * @param propertyElements
	 *            the property elements
	 */
	public void setPropertyElements(final PropertyElement[] propertyElements) {
		this.propertyElements = propertyElements;
	}

	public PropertyElement[] getAllPropertyElements() {
		return propertyElements;
	}

	/**
	 * Adds all property elements to the given panel.
	 * 
	 * @param propertyElements
	 *            the property elements to add
	 * @param panel
	 *            the panel to change
	 * @param constraint
	 *            the gridbag constraints describing the layout constraints how
	 *            to add
	 */
	public void addPropertyElementsTo(final PropertyElement[] propertyElements, final JPanel panel, final GridBagConstraints constraint) {
		if (null != propertyElements) {
			for (final PropertyElement pe : propertyElements) {
				getMainWindow().addPropertyElementTo(panel, pe, constraint);
			}
		}
	}

	/**
	 * Adds all property elements to the given panel.
	 * 
	 * @param panel
	 *            the panel to change
	 * @param constraint
	 *            the gridbag constraints describing the layout constraints how
	 *            to add
	 */
	public void addAllPropertyElementsTo(final JPanel panel, final GridBagConstraints constraint) {
		addPropertyElementsTo(propertyElements, panel, constraint);
	}

	/**
	 * Sets a message icon in the west of the dialog.
	 * 
	 * @param icon
	 *            the icon
	 */
	private void setMessageIcon(final Icon icon) {
		if (null != icon) {
			final JPanel iconPanel = new JPanel();
			final JLabel label = new JLabel();
			label.setIcon(icon);
			label.setBorder(BorderFactory.createEmptyBorder(10,  10,  0,  0));
			iconPanel.add(label);
			getContentPane().add(iconPanel, BorderLayout.WEST);
		}
	}

	/**
	 * Sets a message icon in the west of the dialog.
	 * 
	 * @param messageIconIdx
	 *            one of {@link JOptionPane#ERROR_MESSAGE},
	 *            {@link JOptionPane#WARNING_MESSAGE},
	 *            {@link JOptionPane#INFORMATION_MESSAGE},
	 *            {@link JOptionPane#QUESTION_MESSAGE}, or
	 *            {@link JOptionPane#PLAIN_MESSAGE}
	 */
	protected void setMessageIcon(final int messageIconIdx) {
		final String resourceName;
		switch(messageIconIdx) {
		case JOptionPane.ERROR_MESSAGE:
			resourceName = "OptionPane.errorIcon";
			break;
		case JOptionPane.WARNING_MESSAGE:
			resourceName = "OptionPane.warningIcon";
			break;
		case JOptionPane.INFORMATION_MESSAGE:
			resourceName = "OptionPane.informationIcon";
			break;
		case JOptionPane.QUESTION_MESSAGE:
			resourceName = "OptionPane.questionIcon";
			break;
		default:
			resourceName = null;
		}
		final Icon icon;
		if (null == resourceName) {
			icon = null;
		} else {
			icon = UIManager.getIcon(resourceName);
		}
		setMessageIcon(icon);
	}

	/**
	 * Sets the option type.
	 * 
	 * @param optionType
	 *            one of {@link JOptionPane#DEFAULT_OPTION},
	 *            {@link JOptionPane#OK_CANCEL_OPTION}
	 */
	protected void setOptionType(final int optionType) {
		this.optionType = optionType;
		if (JOptionPane.DEFAULT_OPTION == optionType) {
			setDefaultButton(getButtonOk());
		} else {
			setDefaultButton(null);
		}
	}

	protected void setOptionPaneParams(final String message, final int messageIconIdx) {
		setOptionPaneParams(message, messageIconIdx, null, null);
	}

	protected void setOptionPaneParams(final String message, final int messageIconIdx, final String imageKey, final ResourceFinder.Folder imageFolder) {
		if ((null != imageKey) && (null != imageFolder)) {
			final ResourceFinder res = getApplicationEnvironment().resources();
			if (null != res) {
				setMessageIcon(new ImageIcon(res.getURLToResource(imageKey + ".png", imageFolder)));
			}
		} else {
			setMessageIcon(messageIconIdx);
		}
		final JTextArea textArea = new JTextArea(message);
		GUIAlgorithms.modifyToLabelLAF(textArea);
		getContentPane().add(textArea, BorderLayout.CENTER);
		setDefaultButton(getButtonOk());
	}

	public JButton createHelpButton(final String anchor) {
		final JButton helpButton;
		final String key = "Action.Help";
		helpButton = new JButton();
		getMainWindow().putValues(helpButton, new I18NStringWithFillIns(key, "?", null));
		helpButton.addActionListener(new OnlineHelpActionListener(getMainWindow(), anchor));
		return helpButton;
	}

	/** The application main window. Contains e.g. internationalization support. */
	private ApplicationMainWindow mainWindow;

	private JButton buttonOk;
	private JButton buttonCancel;
	private JPanel  buttonPanel = new JPanel();

	private boolean wasOkPressed = false;

	private PropertyElement[] propertyElements;
}
