package de.ped.tools.gui;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.AbstractAction;
import javax.swing.ActionMap;
import javax.swing.ButtonGroup;
import javax.swing.ButtonModel;
import javax.swing.Icon;
import javax.swing.JCheckBox;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.ActionMapUIResource;

/**
 * A checkbox able to show three states - selected, unselected and a third one 
 * that can be used for meaning like <em>don't care</em> or <em>partially selected</em>.
 * 
 * Maintenance tip - There were some tricks to getting this code
 * working:
 * <ul>
 *   <li>You have to overwrite addMouseListener() to do nothing.</li>
 *   <li>You have to add a mouse event on mousePressed by calling
 *      super.addMouseListener().</li>
 *   <li>You have to replace the UIActionMap for the keyboard event
 *      "pressed" with your own one.</li>
 *   <li>You have to remove the UIActionMap for the keyboard event
 *      "released".</li>
 *   <li>You have to grab focus when the next state is entered,
 *      otherwise clicking on the component won't get the focus.</li>
 *   <li>You have to make a TristateDecorator as a button model that
 *      wraps the original button model and does state management.</li>
 * </ul>
 * 
 * @author Dr. Heinz M. Kabutz, 
 *  from an 
 *  <a href="http://www.javaspecialists.co.za/archive/Issue082.html">example at Java specialists</a>.
 */
public class TristateCheckBox extends JCheckBox {
	private static final long serialVersionUID = 1L;

	/** This is a type-safe enumerated type */
	public static class State {
		private State() {
			// Nothing to do
		}
	}

	public static final State NOT_SELECTED = new State();
	public static final State SELECTED = new State();
	public static final State DONT_CARE = new State();

	private final TristateDecorator model;

	@SuppressWarnings("serial")
	public TristateCheckBox(final String text, final Icon icon, final State initial) {
		super(text, icon);
		// Add a listener for when the mouse is pressed
		super.addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(final MouseEvent e) {
				grabFocus();
				model.nextState();
			}
		});
		// Reset the keyboard action map
		final ActionMap map = new ActionMapUIResource();
		map.put("pressed", new AbstractAction() {
			@Override
			public void actionPerformed(final ActionEvent e) {
				grabFocus();
				model.nextState();
			}
		});
		map.put("released", null);
		SwingUtilities.replaceUIActionMap(this, map);
		// set the model to the adapted model
		model = new TristateDecorator(getModel());
		setModel(model);
		setState(initial);
	}

	public TristateCheckBox(final String text, final State initial) {
		this(text, null, initial);
	}

	public TristateCheckBox(final String text) {
		this(text, DONT_CARE);
	}

	public TristateCheckBox() {
		this(null);
	}

	/**
	 * No one may add mouse listeners, not even Swing!
	 * 
	 * @param listener
	 *            the listener
	 */
	@Override
	public synchronized void addMouseListener(final MouseListener listener) {
		// Nothing to do
	}

	/**
	 * Set the new state.
	 * 
	 * @param state
	 *            either {@link #SELECTED}, {@link #NOT_SELECTED} or
	 *            {@link #DONT_CARE}. If <code>null</code>, it is treated as
	 *            DONT_CARE.
	 */
	public void setState(final State state) {
		model.setState(state);
	}

	/**
	 * @return the current state, which is determined by the selection status of
	 *         the model
	 */
	public State getState() {
		return model.getState();
	}

	@Override
	public void setSelected(final boolean isToSelect) {
		if (isToSelect) {
			setState(SELECTED);
		} else {
			setState(NOT_SELECTED);
		}
	}

	/**
	 * A button model with three states.
	 * <p>
	 * Exactly which Design Pattern is this?  Is it an Adapter,
	 * a Proxy or a Decorator?  In this case, my vote lies with the
	 * Decorator, because we are extending functionality and
	 * "decorating" the original model with a more powerful model.
	 * 
	 * @author Dr. Heinz M. Kabutz, 
	 *  from an 
	 *  <a href="http://www.javaspecialists.co.za/archive/Issue082.html">example at Java specialists</a>.
	 */
	private class TristateDecorator implements ButtonModel {
		private final ButtonModel other;

		private TristateDecorator(final ButtonModel other) {
			this.other = other;
		}

		private void setState(final State state) {
			if (NOT_SELECTED == state) {
				other.setArmed(false);
				setPressed(false);
				setSelected(false);
			} else if (SELECTED == state) {
				other.setArmed(false);
				setPressed(false);
				setSelected(true);
			} else { // either "null" or DONT_CARE
				other.setArmed(true);
				setPressed(true);
				setSelected(true);
			}
		}

		/**
		 * The current state is embedded in the selection / armed state of the
		 * model.
		 * 
		 * @return {@link TristateCheckBox#SELECTED} state when the checkbox is
		 *         selected but not armed, {@link TristateCheckBox#DONT_CARE}
		 *         state when the checkbox is selected and armed (grey) and
		 *         {@link TristateCheckBox#NOT_SELECTED} when the checkbox is
		 *         deselected.
		 */
		private State getState() {
			if (isSelected() && !isArmed()) {
				// normal black tick
				return SELECTED;
			} else if (isSelected() && isArmed()) {
				// don't care grey tick
				return DONT_CARE;
			} else {
				// normal deselected
				return NOT_SELECTED;
			}
		}

		/**
		 * Rotates between {@link TristateCheckBox#NOT_SELECTED},
		 * {@link TristateCheckBox#SELECTED} and
		 * {@link TristateCheckBox#DONT_CARE}.
		 */
		private void nextState() {
			final State current = getState();
			if (NOT_SELECTED == current) {
				setState(SELECTED);
			} else if (SELECTED == current) {
				setState(DONT_CARE);
			} else if (DONT_CARE == current) {
				setState(NOT_SELECTED);
			}
		}

		/**
		 * Filter: No one may change the armed status except us.
		 * 
		 * @param isToArm
		 *            ignored
		 */
		@Override
		public void setArmed(final boolean isToArm) {
			// Nothing to do
		}

		/** 
		 * We disable focusing on the component when it is not enabled. 
		 * 
		 * @param isToEnable
		 *            ignored
		 */
		@Override
		public void setEnabled(final boolean isToEnable) {
			setFocusable(isToEnable);
			other.setEnabled(isToEnable);
		}

		/*
		 * All these methods simply delegate to the "other" model
		 * that is being decorated. 
		 */
		@Override
		public boolean isArmed() {
			return other.isArmed();
		}

		@Override
		public boolean isSelected() {
			return other.isSelected();
		}

		@Override
		public boolean isEnabled() {
			return other.isEnabled();
		}

		@Override
		public boolean isPressed() {
			return other.isPressed();
		}

		@Override
		public boolean isRollover() {
			return other.isRollover();
		}

		@Override
		public void setSelected(final boolean b) {
			other.setSelected(b);
		}

		@Override
		public void setPressed(final boolean b) {
			other.setPressed(b);
		}

		@Override
		public void setRollover(final boolean b) {
			other.setRollover(b);
		}

		@Override
		public void setMnemonic(final int key) {
			other.setMnemonic(key);
		}

		@Override
		public int getMnemonic() {
			return other.getMnemonic();
		}

		@Override
		public void setActionCommand(final String s) {
			other.setActionCommand(s);
		}

		@Override
		public String getActionCommand() {
			return other.getActionCommand();
		}

		@Override
		public void setGroup(final ButtonGroup group) {
			other.setGroup(group);
		}

		@Override
		public void addActionListener(final ActionListener l) {
			other.addActionListener(l);
		}

		@Override
		public void removeActionListener(final ActionListener l) {
			other.removeActionListener(l);
		}

		@Override
		public void addItemListener(final ItemListener l) {
			other.addItemListener(l);
		}

		@Override
		public void removeItemListener(final ItemListener l) {
			other.removeItemListener(l);
		}

		@Override
		public void addChangeListener(final ChangeListener l) {
			other.addChangeListener(l);
		}

		@Override
		public void removeChangeListener(final ChangeListener l) {
			other.removeChangeListener(l);
		}

		@Override
		public Object[] getSelectedObjects() {
			return other.getSelectedObjects();
		}
	}
}
