package de.ped.tools.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Insets;
import java.util.ArrayList;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.border.EmptyBorder;

import de.ped.tools.CollectionUtil;
import de.ped.tools.TextDocumentBuilder;

/**
 * The class defining a wizard, that is a series of dialogs to perform a complex task.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de), 
 * 		based on an <a href="http://www.oracle.com/technetwork/articles/javase/wizard-136789.html">article</a> 
 * 		by Robert Eckstein, February 10, 2005 
 */
public class Wizard extends PedJDialog {
	public enum NavigationAction {
		BACK,
		NEXT,
		FINISH,
		CANCEL,
	}

	private static final long serialVersionUID = 1L;

	private WizardModel wizardModel;
	private WizardController wizardController;
	private JPanel cardPanel;
	private CardLayout cardLayout;
	private JButton backButton;
	private JButton nextButton;

	public Wizard(final ApplicationMainWindow mainWindow, final Frame parent, final String title, final String helpKey) {
		super(mainWindow, parent);
		setTitle(title);
		init(helpKey);
	}

	public Wizard(final ApplicationMainWindow mainWindow, final Dialog parent, final String title, final String helpKey) {
		super(mainWindow, parent);
		setTitle(title);
		init(helpKey);
	}

	private void init(final String helpKey) {
		wizardModel = new WizardModel();
		wizardController = new WizardController(this);

		JPanel buttonPanel = new JPanel();
		Box buttonBox = new Box(BoxLayout.X_AXIS);

		cardPanel = new JPanel();
		cardPanel.setBorder(new EmptyBorder(new Insets(5, 10, 5, 10)));

		cardLayout = new CardLayout();
		cardPanel.setLayout(cardLayout);
		backButton = getMainWindow().createButton("Dialog.Back");
		nextButton = getMainWindow().createButton("Dialog.Next");

		backButton.addActionListener(wizardController);
		nextButton.addActionListener(wizardController);
		getButtonCancel().addActionListener(wizardController);

		buttonPanel.setLayout(new BorderLayout());
		buttonPanel.add(new JSeparator(), BorderLayout.NORTH);

		buttonBox.setBorder(new EmptyBorder(new Insets(5, 10, 5, 10)));
		if (!CollectionUtil.isNullOrEmpty(helpKey)) {
			buttonBox.add(createHelpButton(helpKey));
			buttonBox.add(Box.createHorizontalStrut(30));
		}
		buttonBox.add(backButton);
		buttonBox.add(Box.createHorizontalStrut(10));
		buttonBox.add(nextButton);
		buttonBox.add(Box.createHorizontalStrut(30));
		buttonBox.add(getButtonCancel());
		buttonPanel.add(buttonBox, java.awt.BorderLayout.EAST);
		setMinimumSizeTo(buttonPanel);
		getContentPane().add(buttonPanel, java.awt.BorderLayout.SOUTH);
		getContentPane().add(cardPanel, java.awt.BorderLayout.CENTER);
	}

	public void registerWizardPanel(WizardPanelDescriptor panel) {
		cardPanel.add(panel.getPanelComponent(), panel.getPanelId());
		wizardModel.registerPanel(panel);
	}

	public WizardPanelDescriptor getPanelDescriptor(Object id) {
		return wizardModel.getPanelDescriptor(id);
	}

	void setBackButtonEnabled(final boolean b) {
		backButton.setEnabled(b);
	}

	void setNextButtonEnabled(final boolean b) {
		nextButton.setEnabled(b);
	}

	public ArrayList<Object> getPanelDescriptorIds() {
		return wizardModel.getPanelDescriptorIds();
	}

	/** Forwards the call to all the panels contained in this wizard. */
	@Override
	public void readUIFieldsIntoProperties() {
		for (Object id : getPanelDescriptorIds()) {
			WizardPanelDescriptor wpd = wizardModel.getPanelDescriptor(id);
			wpd.readUIFieldsIntoProperties();
		}
	}

	/** Forwards the call to all the panels contained in this wizard. */
	@Override
	public void updateUIFieldsBasedOnProperties() {
		for (Object id : getPanelDescriptorIds()) {
			WizardPanelDescriptor wpd = wizardModel.getPanelDescriptor(id);
			wpd.updateUIFieldsBasedOnProperties();
		}
	}

	public void buildHelpDocumentSections(final TextDocumentBuilder doc, final int depth, boolean isToCreateAnchors) {
		final Object[] temp = new Object[0];
		final Object[] panelIds = getPanelDescriptorIds().toArray(temp);
		for (Object panelId : panelIds) {
			buildHelpDocumentSection(doc, depth, panelId, isToCreateAnchors);
		}
	}

	public void buildHelpDocumentSection(final TextDocumentBuilder doc, final int depth, final Object panelId, boolean isToCreateAnchors) {
		final WizardPanelDescriptor wpd = getPanelDescriptor(panelId);
		getMainWindow().buildPropertyDescriptions(doc, depth, wpd.getHelpHeadingKey(), wpd.getAllPropertyElements(), isToCreateAnchors);
	}

	public Object getCurrentPanelId() {
		return wizardModel.getCurrentPanelDescriptor();
	}

	public void setCurrentPanelId(final Object id) {
		final WizardPanelDescriptor oldPanelDescriptor = wizardModel.getCurrentPanelDescriptor();
		if (oldPanelDescriptor != null) {
			oldPanelDescriptor.aboutToHidePanel();
		}
		wizardModel.setCurrentPanelId(id);
		// Show the panel
		wizardModel.getCurrentPanelDescriptor().aboutToDisplayPanel();
		cardLayout.show(cardPanel, id.toString());
		wizardModel.getCurrentPanelDescriptor().displayingPanel();
		// Update the wizard buttons according to the panel state
		update();
	}

	public JButton getButton(final NavigationAction navAction) {
		final JButton ret;
		switch(navAction) {
		case BACK:
			ret = backButton;
			break;
		case NEXT:
			ret = nextButton;
			break;
		case CANCEL:
			ret = getButtonCancel();
			break;
		case FINISH:
			ret = null;
			break;
		default:
			ret = null;
		}
		return ret;
	}

	protected void onNavigationAction(final NavigationAction navAction) {
		switch(navAction) {
		case BACK:
			onBack();
			break;
		case NEXT:
			onNext();
			break;
		case CANCEL:
			onCancel();
			break;
		case FINISH:
			onFinish();
			break;
		}
	}

	private void onBack() {
		setCurrentPanelId(wizardModel.getCurrentPanelDescriptor().getBackPanelId());
	}

	/**
	 * Called when the NEXT button was pressed.
	 */
	protected void onNext() {
		final Object next = wizardModel.getCurrentPanelDescriptor().getNextPanelId();
		if (WizardPanelDescriptor.FINISH_ID.equals(next)) {
			onFinish();
		} else {
			setCurrentPanelId(next);
		}
	}

	private void onFinish() {
		onOk();
	}

	/** 
	 * (Re-)Initializes and opens the dialog. 
	 */
	@Override
	public void enterDialog() {
		pack();
		super.enterDialog();
	}

	public void update() {
		// Update back button and next/finish button according to current panel
		final Object back = wizardModel.getCurrentPanelDescriptor().getBackPanelId();
		if (null == back) {
			backButton.setEnabled(false);
		} else {
			backButton.setEnabled(true);
		}
		final Object next = wizardModel.getCurrentPanelDescriptor().getNextPanelId();
		final String key;
		if (WizardPanelDescriptor.FINISH_ID.equals(next)) {
			key = "Dialog.Finish";
			nextButton.setEnabled(true);
		} else if (null == next) {
			key = "Dialog.Next";
			nextButton.setEnabled(false);
		} else {
			key = "Dialog.Next";
			nextButton.setEnabled(true);
		}
		nextButton.setText(getMessages().getText(key));
	}
}
