package de.ped.tools.gui;

import java.awt.Component;

/**
 * The class describing a single {@link Wizard} panel.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de), 
 * 		based on an <a href="http://www.oracle.com/technetwork/articles/javase/wizard-136789.html">article</a> 
 * 		by Robert Eckstein, February 10, 2005 
 */
public class WizardPanelDescriptor {
	/**
	 * Use this identifier as next panel identifier to indicate that the next action will finish the wizard.
	 */
	public static final Object FINISH_ID = "Finish";

	private Component panel;
	private Object panelIdentifier;
	private Object nextPanelIdentifier;
	private Object backPanelIdentifier;
	private final Wizard wizard;
	private PropertyElement[] propertyElements;

	public WizardPanelDescriptor(final Wizard wizard) {
		this.wizard = wizard;
	}

	public WizardPanelDescriptor(final Wizard wizard, final Component panel, final Object id, final Object back, final Object next) {
		this.wizard = wizard;
		set(panel, id, back, next);
	}

	/** 
	 * @return the component to show when this page of the wizard will be displayed
	 */
	public final Component getPanelComponent() {
		return panel;
	}

	/** 
	 * @param panel the component to show when this page of the wizard will be displayed
	 */
	public final void setPanelComponent(final Component panel) {
		this.panel = panel;
	}

	/** 
	 * @return the id for this page of the wizard 
	 */
	public final Object getPanelId() {
		return panelIdentifier;
	}

	/** 
	 * @param id the id for this page of the wizard
	 */
	public final void setPanelId(final Object id) {
		panelIdentifier = id;
	}

	/** 
	 * @return the id for the next page of the wizard (or <code>null</code>, if there is none). 
	 */
	public Object getNextPanelId() {
		return nextPanelIdentifier;
	}

	/** 
	 * @param id the id for the next page of the wizard (or <code>null</code>, if there is none)
	 */
	public void setNextPanelId(final Object id) {
		nextPanelIdentifier = id;
	}

	/** 
	 * @return the id for the previous page of the wizard (or <code>null</code>, if there is none)
	 */
	public Object getBackPanelId() {
		return backPanelIdentifier;
	}

	/** 
	 * @param id the id for the previous page of the wizard (or <code>null</code>, if there is none)
	 */
	public void setBackPanelId(final Object id) {
		backPanelIdentifier = id;
	}

	public void set(final Component panel, final Object id, final Object back, final Object next) {
		setPanelComponent(panel);
		setPanelId(id);
		setBackPanelId(back);
		setNextPanelId(next);
	}

	/** 
	 * Will be executed when the panel gets hidden. 
	 * Default implementation does nothing. 
	 */
	public void aboutToHidePanel() {
		// Default implementation does nothing.
	}

	/** 
	 * Will be executed before the panel gets displayed. 
	 * Default implementation does nothing. 
	 */
	public void aboutToDisplayPanel() {
		// Default implementation does nothing.
	}

	/** 
	 * Will be executed when the panel gets displayed. 
	 * Default implementation only calls {@link #update()}. 
	 */
	public void displayingPanel() {
		update();
	}

	/** 
	 * Should be called when something has changed and the wizard should update its buttons. 
	 * Calls {@link #updateNavDescriptors()} and then {@link Wizard#update()}. 
	 */
	public void update() {
		updateNavDescriptors();
		wizard.update();
	}

	/** 
	 * Will always be called when something has changed and the wizard should update its buttons. 
	 * Default implementation does nothing. 
	 */
	protected void updateNavDescriptors() {
		// Default implementation does nothing.
	}
	
	protected Wizard getWizard() {
		return wizard;
	}

	/**
	 * Sets all property elements.
	 * 
	 * @param propertyElements
	 *            the property elements
	 */
	public void setPropertyElements(final PropertyElement[] propertyElements) {
		this.propertyElements = propertyElements;
	}

	public PropertyElement[] getAllPropertyElements() {
		return propertyElements;
	}

	public String getHelpHeadingKey() {
		return panelIdentifier.toString();
	}

	/** 
	 * Default implementation does nothing.
	 */
	public void readUIFieldsIntoProperties() {
		
	}

	/** 
	 * Default implementation does nothing.
	 */
	public void updateUIFieldsBasedOnProperties() {
		
	}
}
