package de.ped.tools.log;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedHashSet;
import java.util.StringTokenizer;



/**
 * Allows logging. 
 * The interface is a subset of the Log4j interface, to minimize the effort when switching to that.
 * 
 * @author Copyright (c) 2006 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Logger {

	public static final String TIMESTAMP_FORMAT_STRING = "yyyy-MM-dd HH:mm:ss,SSS";
	public static final SimpleDateFormat TIMESTAMP_FORMAT = new SimpleDateFormat(TIMESTAMP_FORMAT_STRING);

	public static final String[] MESSAGE_LEVEL_STRINGS = {
		"OFF  ", "FATAL", "ERROR", "WARN ", "INFO ", "DEBUG", "TRACE", "ALL  ",
	};

	/** Up to which message level a message should lead to a halt? */
	protected static int fatalLevel = Level.FATAL;

	public static int normLogLevel(final int level) {
		return Math.max(Level.OFF, Math.min(Level.ALL, level));
	}

	public static void setFatalLevel(final int level) {
		fatalLevel = normLogLevel(level);
	}

	/** Counts the number of messages logged for each message level. */
	private static int[] messageLevelCounters = new int[Level.ALL]; 

	public static final StringWriterAppender DEFAULT_STRING_WRITER_APPENDER = new StringWriterAppender("memlogbuffer");

	public static final PrintStreamAppender SYSTEM_OUT_APPENDER = new PrintStreamAppender(System.out, "sysout");

	public static final PrintStreamAppender SYSTEM_ERR_APPENDER = new PrintStreamAppender(System.err, "syserr");

	private static volatile Logger rootLogger = null;

	private static LinkedHashSet<Logger> loggers = new LinkedHashSet<Logger>();

	/**
	 * @return the root logger instance. If it does not exist, it is created.
	 */
	public static synchronized final Logger getRootLogger() {
		if (null == rootLogger) {
			rootLogger = new Logger(null);
		}
		return rootLogger;
	}

	/**
	 * Returns the logger instance with the given name. If it does not exist, it is created.
	 * @param name the name of the logger
	 * @return the logger instance
	 */
	public static synchronized final Logger getLogger(final String name) {
		if (null == name) {
			return getRootLogger();
		}
		for (final Logger logger : loggers) {
			if (name.equals(logger.m_name)) {
				return logger;
			}
		}
		final Logger logger = new Logger(name);
		loggers.add(logger);
		return logger;
	}

	/**
	 * Short form for <code>getLogger(clazz.getName())</code>.
	 * 
	 * @param clazz
	 *            the class to create the logger for
	 * @return the logger instance
	 * @see #getLogger(String)
	 */
	public static final Logger getLogger(final Class<?> clazz) {
		return getLogger(stripPackagesFromClassName(clazz.getName()));
	}

	private String m_name;
	private boolean isToShowDate = true;
	private boolean isToShowThread = false;
	private boolean isToShowStack = true;

	private LinkedHashSet<Appender> appenders = new LinkedHashSet<Appender>();

	/** Up to which message level the messages should get logged? */
	protected int loglevel = Level.DEBUG;

	public Logger(String clazz) {
		this.m_name = clazz;
		appenders.add(SYSTEM_OUT_APPENDER);
	}

	public int getLogLevel() {
		return loglevel;
	}

	public void setLogLevel(final int level) {
		loglevel = Math.max(Level.OFF, Math.min(Level.ALL, level));
	}

	public synchronized void addAppender(final Appender appender) {
		appenders.add(appender);
	}

	public synchronized void removeAppender(final Appender appender) {
		appenders.remove(appender);
	}

	public boolean isToShowThread() {
		return isToShowThread;
	}

	public void setToShowThread(final boolean toShowThread) {
		this.isToShowThread = toShowThread;
	}

	public boolean isToShowDate() {
		return isToShowDate;
	}

	public void setToShowDate(final boolean toShowDate) {
		this.isToShowDate = toShowDate;
	}

	public boolean isToShowStack() {
		return isToShowStack;
	}

	public void setToShowStack(final boolean toShowStack) {
		this.isToShowStack = toShowStack;
	}

	public String getClassName() {
		return m_name;
	}

	public static boolean isFor(final int msglevel, final int levelToCompare) {
		return ((Level.OFF < msglevel) && (msglevel <= levelToCompare));
	}

	public boolean isEnabledFor(final int msglevel) {
		return isFor(msglevel, loglevel);
	}

	public boolean isFatalFor(final int msglevel) {
		return isFor(msglevel, fatalLevel);
	}

	private void showErrorDialog(final String title, final String message) {
		// Default implementation does nothing.
	}

	/**
	 * Logs the given message.
	 * 
	 * @param msglevel
	 *            the message level to log this message with
	 * @param message
	 *            the message to log
	 * @param ex
	 *            the exception to log with this message, or <code>null</code>
	 *            if none
	 * @param isAlwaysShowingException
	 *            if <code>true</code>, the exception is always shown, no matter
	 *            of the debug level; otherwise it is only shown in debug level
	 *            or higher
	 */
	public synchronized void log(
			final int msglevel, 
			final String message, 
			final Throwable ex, 
			final boolean isAlwaysShowingException) {
		if (isEnabledFor(msglevel)) {
			final StringBuffer s = new StringBuffer();
			if (isToShowDate) {
				final Date t = new Date();
				final String df = TIMESTAMP_FORMAT.format(t);
				s.append(df);
				s.append(' ');
			}
			s.append('[');
			s.append(MESSAGE_LEVEL_STRINGS[msglevel]);
			s.append("] ");
			if (null != m_name) {
				s.append("{");
				s.append(m_name);
				s.append("} ");
			}
			if (isToShowThread) {
				s.append("<");
				s.append(Thread.currentThread().getName());
				s.append("> ");
			}
			s.append(message);
			if (null != ex) {
				s.append("\n            ");
				s.append(ex.toString());
			}
//			s.append("\n");
			final String msg = s.toString();
			for (final Appender appender : appenders) {
				appender.println(msg);
				if ((null != ex) && ((Level.DEBUG <= loglevel) || (isAlwaysShowingException))) {
					appender.printStackTrace(ex);
				}
				appender.flush();
			}
			messageLevelCounters[msglevel - 1]++;
			if (isFatalFor(msglevel)) {
				throw new IllegalStateException("A fatal log message appeared: " + s.toString());
			}
		}
	}

	/**
	 * Logs the given message.
	 * 
	 * @param msglevel
	 *            the message level to log this message with
	 * @param message
	 *            the message to log
	 */
	public void log(final int msglevel, final String message) {
		log(msglevel, message, null, false);
	}

	/**
	 * Logs the given message on fatal log level.
	 * 
	 * @param message
	 *            the message to log
	 * @see Level#FATAL
	 */
	public void fatal(final String message) {
		showErrorDialog(message, message);
		log(Level.FATAL, message, null, false);
	}
	
	public void fatal(final String message, final Throwable ex) {
		showErrorDialog(message, message);
		log(Level.FATAL, message, ex, isToShowStack);
	}

	/**
	 * Logs the given message on error log level.
	 * 
	 * @param message
	 *            the message to log
	 * @see Level#ERROR
	 */
	public void error(final String message) {
		showErrorDialog(message, message);
		log(Level.ERROR, message, null, false);
	}

	public void error(final String message, final Throwable ex) {
		showErrorDialog(ex.toString(), message);
		log(Level.ERROR, message, ex, isToShowStack);
	}

	public void errorWithStackTrace(final String message, final Throwable ex) {
		showErrorDialog(ex.toString(), message);
		log(Level.ERROR, message, ex, true);
	}

	public void warn(final String message) {
		log(Level.WARN, message, null, false);
	}

	public void warn(final String message, final Throwable ex) {
		log(Level.WARN, message, ex, isToShowStack);
	}

	/**
	 * Logs the given message on info log level.
	 * 
	 * @param message
	 *            the message to log
	 * @see Level#INFO
	 */
	public void info(final String message) {
		log(Level.INFO, message, null, false);
	}

	public void info(final String message, final Throwable ex) {
		log(Level.INFO, message, ex, isToShowStack);
	}

	/**
	 * Logs the given message on debug log level.
	 * 
	 * @param message
	 *            the message to log
	 * @see Level#DEBUG
	 */
	public void debug(final String message) {
		log(Level.DEBUG, message, null, false);
	}

	public void debug(final String message, final Throwable ex) {
		log(Level.DEBUG, message, ex, isToShowStack);
	}

	/**
	 * Logs the given message on trace log level.
	 * 
	 * @param message
	 *            the message to log
	 * @see Level#TRACE
	 */
	public void trace(final String message) {
		log(Level.TRACE, message, null, false);
	}

	public void trace(final String message, final Throwable ex) {
		log(Level.TRACE, message, ex, isToShowStack);
	}

	/**
	 * Logs the given message on all log level.
	 * 
	 * @param message
	 *            the message to log
	 * @see Level#ALL
	 */
	public void log(final String message) {
		log(Level.ALL, message, null, false);
	}

	public void log(final String message, final Throwable ex) {
		log(Level.ALL, message, ex, isToShowStack);
	}

	/**
	 * @param className
	 *            the Java class name
	 * @return the last part of the class name, stripping away all the packages
	 */
	private static String stripPackagesFromClassName(final String className) {
		int index = className.lastIndexOf('.');
		index = (((index >= 0) && (index < className.length() + 1)) ? (index + 1) : 0);
		return className.substring(index);
	}

	/**
	 * Dumps an XML String by tokenizing it using the delimiter "&lt;&gt;".
	 * 
	 * @param xml
	 *            the string to be dumped
	 * @param indent
	 *            the prefix to add to each line written
	 */
	public void dumpXMLString(String xml, final String indent) {
		xml = xml.replace('\n', (char) 0);
		xml = xml.replace('\r', (char) 0);
		final StringTokenizer st = new StringTokenizer(xml, "<>");
		while(st.hasMoreTokens()) {
			String token = st.nextToken();
			debug(indent + "<" + token + ">");
		}
	}

	/**
	 * Retrieves a value from a String Array that contains key/value pair Strings.
	 * The appropriate key must be specified, of course.
	 * 
	 * @param  pairs  The array of Strings that must be searched for the value that
	 *                belongs to the appropriate key.
	 * @param  key    The key String.
	 * 
	 * @return Returns the value that corresponds to the specified key, or <code>null</code> if
	 *         the value couldn't be found.
	 * 
	 */
//	public static String getValueStringFromArray(String[] pairs, String key) {
//		if ((pairs != null) && (pairs.length > 0)) {
//			for (int i = 0; i < pairs.length; i++) {
//				final int startIndex = key.length() + 1;
//				if (pairs[i].startsWith(key + "=")) {
//					return pairs[i].substring(startIndex);
//				}
//			}
//		}
//		return null;
//	}

	/**
	 * @param logLevel
	 *            the log level
	 * @return the number of logged entries of the given log level
	 */
	public static int getErrorLevelCounter(final int logLevel) {
		return ((logLevel >= Level.FATAL) && (logLevel <= Level.ALL)) ? messageLevelCounters[logLevel - 1] : 0;
	}

	/**
	 * @param fromLogLevel
	 *            the log level to start counting with
	 * @param toLogLevel
	 *            the last log level to count with
	 * 
	 * @return the number of logged entries of the given range of log levels
	 */
	public static int getErrorLevelSum(final int fromLogLevel, final int toLogLevel) {
		int sum = 0;
		if (fromLogLevel <= toLogLevel) {
			for (int logLevel = fromLogLevel; logLevel <= toLogLevel; logLevel++) {
				sum += getErrorLevelCounter(logLevel);
			}
		} else {
			sum = getErrorLevelSum(toLogLevel, fromLogLevel);
		}
		return sum;
	}

	/**
	 * @param leastLogLevel
	 *            the least log level to count
	 * @return the number of logged entries being at least of the given log
	 *         level.
	 */
	public static int getErrorLevelSum(final int leastLogLevel) {
		return getErrorLevelSum(Level.FATAL, leastLogLevel);
	}

	private static void resetLogBuffer() {
		DEFAULT_STRING_WRITER_APPENDER.clearContent();
	}

	/**
	 * Resets the counters for number of logged entries.
	 */
	public static synchronized void resetErrorLevelCounters() {
		for (int loglevel = 0; loglevel < messageLevelCounters.length; loglevel++) {
			messageLevelCounters[loglevel] = 0;
		}
		resetLogBuffer();
	}

	public static synchronized String getWholeLog() {
		return DEFAULT_STRING_WRITER_APPENDER.getContent();
	}

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append(getClass().getName());
		ret.append("(");
		ret.append(m_name);
		ret.append(",");
		ret.append(MESSAGE_LEVEL_STRINGS[loglevel]);
		ret.append(",(");
		boolean isFirst = true;
		for (final Appender appender : appenders) {
			if (isFirst) {
				isFirst = false;
			} else {
				ret.append(",");
			}
			ret.append(appender.toString());
		}
		ret.append("))");
		return ret.toString();
	}
}
