package de.ped.tools.sound;

/**
 * Helps to build an efficient implementation of {@link MonoWave}.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class AbstractWave implements MonoWave {

	/**
	 * Parent for all other Params classes.
	 * 
	 * @author Copyright (c) 2010 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Params {
		public Params(final DynamicSoundMonoWaveCreator smwc, final double offset, final double duration) {
			this.smwc = smwc;
			this.offset = offset;
			this.duration = duration;
		}

		public Params(final DynamicSoundMonoWaveCreator smwc, final double offset) {
			this(smwc, offset, Double.MAX_VALUE);
		}

		public Params(final DynamicSoundMonoWaveCreator smwc) {
			this(smwc, 0.0);
		}

		/** the sound wave creator object to attach this wave to */
		public final DynamicSoundMonoWaveCreator smwc;
		/** the begin time (in seconds) of this sound */
		public final double offset;
		/** the duration (in seconds) of this sound */
		public final double duration;
	}

	public double getOffset() {
		return params.offset;
	}

	protected int getTimeslice() {
		final int smwcTimeslice = params.smwc.getTimeslice();
		int ret;
		if ((smwcTimeslice < offsetTimeslice) || (smwcTimeslice >= endingTimeslice)) {
			ret = -1;
		} else {
			ret = smwcTimeslice - offsetTimeslice;
		}
		return ret;
	}

	/** 
	 * @return the SoundStereoWaveCreator object this wave belongs to
	 */
	public DynamicSoundMonoWaveCreator getSoundSteroWaveCreator() {
		return params.smwc;
	}

	/**
	 * Creates a mono wave and registers it for the given wave creator.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param offset
	 *            duration in seconds of leading zero signal
	 */
	protected AbstractWave(final DynamicSoundMonoWaveCreator smwc, final double offset) {
		this(new Params(smwc, offset));
	}

	/**
	 * Creates a mono wave and registers it for the given wave creator.
	 * 
	 * @param params
	 *            the wave parameters
	 */
	protected AbstractWave(final Params params) {
		if (null == params) {
			throw new IllegalArgumentException("params must not be null");
		}
		this.params = params;
		this.offsetTimeslice = params.smwc.durationToSize(params.offset);
		this.endingTimeslice = params.smwc.durationToSize(params.offset + params.duration);
		params.smwc.register(this);
	}

	/**
	 * Called during {@link #next()}, the current wave value is calculated.
	 * 
	 * @return the current wave value
	 */
	protected abstract double calcCurrentValue();

	/**
	 * @return the next current value calculated via {@link #calcCurrentValue()}
	 */
	@Override
	public double next() {
		final int smwcTimeslice = params.smwc.getTimeslice();
		if ((smwcTimeslice < offsetTimeslice) || (smwcTimeslice >= endingTimeslice)) {
			currentValue = 0;
		} else {
			currentValue = calcCurrentValue();
		}
		return get();
	}

	/**
	 * @return the calculated value for this time slice
	 */
	@Override
	public double get() {
		return currentValue;
	}

	protected Params params;
	private final int offsetTimeslice;
	private final int endingTimeslice;
	private double currentValue;
}
