package de.ped.tools.sound;

import java.util.ArrayList;
import java.util.List;

/**
 * A mono wave that results from the addition of each value of other mono
 * waves. So all those sounds happen at the same time.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AddedWave extends AbstractWave  {

	/**
	 * A wave to add to the added wave.
	 * 
	 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Part {
		public static ArrayList<Part> createParts(final List<MonoWave> waves) {
			final ArrayList<Part> ret = new ArrayList<Part>();
			if (null != waves) {
				for (final MonoWave wave : waves) {
					ret.add(new Part(wave));
				}
			}
			return ret;
		}

		public Part(final MonoWave wave) {
			this.wave = wave;
			this.loudness = 1.0;
		}

		public Part(final MonoWave wave, final double multiplicand) {
			this.wave = wave;
			this.loudness = multiplicand;
		}

		@Override
		public String toString() {
			final StringBuffer ret = new StringBuffer();
			ret.append("(");
			ret.append(loudness);
			ret.append(",");
			ret.append(wave);
			ret.append(")");
			return ret.toString();
		}

		public final MonoWave wave;
		public final double loudness;
	}

	/**
	 * Creates a mono wave by adding the given summands.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param waves
	 *            the parts to add
	 */
	public AddedWave(final DynamicSoundMonoWaveCreator smwc, final List<Part> waves) {
		super(smwc, 0.0);
		this.summands = new ArrayList<Part>(waves.size());
		this.summands.addAll(waves);
	}

	/**
	 * Creates a mono wave by adding the given summands.
	 * 
	 * @param params
	 *            the abstract wave parameters
	 * @param waves
	 *            the parts to add
	 */
	public AddedWave(final AbstractWave.Params params, final List<Part> waves) {
		super(params);
		this.summands = new ArrayList<Part>(waves.size());
		this.summands.addAll(waves);
	}

	/**
	 * Creates an added mono wave objects for the given two summands.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param m1
	 *            the first mono wave
	 * @param loudness1
	 *            the loudness of the first mono wave
	 * @param m2
	 *            the second mono wave
	 * @param loudness2
	 *            the loudness of the second mono wave
	 */
	public AddedWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final MonoWave m1, 
			final double loudness1, 
			final MonoWave m2, 
			final double loudness2) {
		super(smwc, 0.0);
		summands = new ArrayList<Part>(2);
		summands.add(new Part(m1, loudness1));
		summands.add(new Part(m2, loudness2));
	}

	/**
	 * Creates an added mono wave objects for the given two summands.
	 * 
	 * @param params
	 *            the abstract wave parameters
	 * @param m1
	 *            the first mono wave
	 * @param loudness1
	 *            the loudness of the first mono wave
	 * @param m2
	 *            the second mono wave
	 * @param loudness2
	 *            the loudness of the second mono wave
	 */
	public AddedWave(
			final AbstractWave.Params params, 
			final MonoWave m1, 
			final double loudness1, 
			final MonoWave m2, 
			final double loudness2) {
		super(params);
		summands = new ArrayList<Part>(2);
		summands.add(new Part(m1, loudness1));
		summands.add(new Part(m2, loudness2));
	}

	@Override
	protected double calcCurrentValue() {
		double ret = 0.0;
		if (null != summands) {
			for (final Part part : summands) {
				ret += part.wave.get() * part.loudness;
			}
		}
		return ret;
	}

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append("add(");
		for (final Part part : summands) {
			ret.append(part);
		}
		ret.append(")");
		return ret.toString();
	}

	List<Part> summands;
}
