package de.ped.tools.sound;

/**
 * Modulate the amplitude of a main signal by a constant factor that can be further modulated by another mono wave signal.
 * The amplitude is modulated with the formula 
 * <blockquote>constant factor * (1 + amplitude modulator)</blockquote>
 * The main use for this is to create a hull curve to influence the loudness of a sound.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class AmplitudeModulationWave extends AbstractWave {
	/**
	 * Creates a mono wave object that multiplies the main signal with a
	 * constant factor.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param mainSignal
	 *            the main signal that will be amplitude modulated
	 * @param normalFactor
	 *            the constant factor that will be applied to the main signal
	 */
	public AmplitudeModulationWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final MonoWave mainSignal, 
			final double normalFactor) {
		this(smwc, mainSignal, normalFactor, new ConstantWave(0.0));
	}

	/**
	 * Creates a mono wave object that multiplies the main signal with a
	 * constant factor that is further modulated by the amplitude modulator
	 * wave.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param mainSignal
	 *            the main signal that will be amplitude modulated
	 * @param normalFactor
	 *            the constant factor that will be applied to the main signal
	 * @param amplitudeModulator
	 *            the modulator that is applied to the constant factor
	 */
	public AmplitudeModulationWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final MonoWave mainSignal, 
			final double normalFactor, 
			final MonoWave amplitudeModulator) {
		super(smwc, 0.0);
		this.mainSignal = mainSignal;
		this.normalFactor = normalFactor;
		this.amplitudeModulator = amplitudeModulator;
	}

	@Override
	protected double calcCurrentValue() {
		return mainSignal.get() * normalFactor * (1.0 + amplitudeModulator.get());
	}

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append("amplitudeMod(");
		ret.append(normalFactor);
		ret.append(",");
		ret.append(mainSignal);
		ret.append(")");
		return ret.toString();
	}

	private MonoWave mainSignal;
	private double normalFactor;
	private MonoWave amplitudeModulator;
}