package de.ped.tools.sound;

import java.util.LinkedList;

import javax.sound.sampled.AudioInputStream;

/**
 * A simple creator for wave sounds using additive synthesis. All wave sounds
 * created with the same creator are created with the same sample rate.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class DynamicSoundMonoWaveCreator extends SoundMonoWaveCreator {

	/**
	 * A sound wave to create.
	 * 
	 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Sound {
		public Sound(final MonoWave wave, final double duration) {
			this.wave = wave;
			this.duration = duration;
		}

		@Override
		public String toString() {
			return "sound(dur=" + duration + "," + wave + ")";
		}

		public final MonoWave wave;
		public final double duration;
	}

	int getTimeslice() {
		return timeslice;
	}

	/**
	 * Creates a wave creator with default sample rate @link{SoundMonoWaveCreator#DEFAULT_SAMPLE_RATE}.
	 */
	public DynamicSoundMonoWaveCreator() {
		super();
	}

	/**
	 * Creates a wave creator with given sample rate.
	 * 
	 * @param sampleRate
	 *            the sample rate
	 */
	public DynamicSoundMonoWaveCreator(final int sampleRate) {
		super(sampleRate);
	}

	/** All mono waves controlled by this creator. */
	private LinkedList<MonoWave> monoWaves = new LinkedList<MonoWave>();

	/**
	 * Should be called in each constructor of a mono wave.
	 * 
	 * @param monoWave
	 *            the mono wave to register to this creator
	 */
	public void register(final MonoWave monoWave) {
		monoWaves.add(monoWave);
	}

	/**
	 * Advances to the next time slot (as defined by the sample rate).
	 * Implicitly the next() method is called for all waves controlled by this
	 * creator.
	 * 
	 * @return the current value for the wave registered last at this creator
	 * @see MonoWave#next()
	 */
	public double next() {
		timeslice++;
		for (MonoWave wave : monoWaves) {
			wave.next();
		}
		return get();
	}

	/**
	 * @return the current value for the wave registered last at this creator
	 * @see MonoWave#get()
	 */
	public double get() {
		final MonoWave wave = monoWaves.getLast();
		return wave.get();
	}

	/**
	 * Executes the defined mono wave for the given duration.
	 * 
	 * @param sound
	 *            the sound to calculate
	 * @return the mono wave as array of double values
	 */
	public double[] execute(final Sound sound) {
		final double[] ret = new double[durationToSize(sound.duration)];
		for (int i = 0; i < ret.length; i++) {
			next();
			ret[i] = sound.wave.get();
		}
		return ret;
	}

	public AudioInputStream toAudioInputStream(final Sound sound) {
		final byte[] sample = DynamicSoundMonoWaveCreator.toStereoSample(execute(sound));
		return DynamicSoundMonoWaveCreator.toAudioInputStream(getSampleRate(), sample);
	}

	private int timeslice = -1;
}
