package de.ped.tools.sound;

import de.ped.tools.log.Logger;

/**
 * Creates a complex envelope shape. Basically, that's a concatenation of linear
 * interpolations in time between two amplitude levels.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class EnvelopeWave extends AbstractWave {
	@SuppressWarnings("unused")
	private static Logger logger = Logger.getRootLogger();

	/**
	 * ALl the parameters defining the envelope.
	 * 
	 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Parameters extends AbstractWave.Params {

		public static double[][] createParts(
				final double duration, 
				final double startLevel,
				final double endLevel) {
			return new double[][] {
				{ duration, startLevel },
				{ 0.0, endLevel, }, };
		}

		public static double[][] createParts(
				final double durationAttack, double durationDecay, 
				final double durationSustain, double durationRelease, 
				final double levelInitial, double levelPeak, double levelSustain) {
			return new double[][] {
				{ durationAttack,  levelInitial },
				{ durationDecay,   levelPeak },
				{ durationSustain, levelSustain },
				{ durationRelease, levelSustain }, };
		}

		public Parameters(final DynamicSoundMonoWaveCreator smwc, final double offset, final double[][] partParameters) {
			super(smwc, offset);
			this.parts = partParameters;
		}

		public Parameters(final DynamicSoundMonoWaveCreator smwc, final double[][] partParameters) {
			this(smwc, 0.0, partParameters);
		}

		public Parameters(final AbstractWave.Params params, 
				final double duration, 
				final double startLevel,
				final double endLevel) {
			this(params.smwc, params.offset, createParts(duration, startLevel, endLevel));
		}		

		public Parameters(final DynamicSoundMonoWaveCreator smwc, final double offset, 
				final double duration, 
				final double startLevel,
				final double endLevel) {
			this(smwc, offset, createParts(duration, startLevel, endLevel));
		}

		public Parameters(final DynamicSoundMonoWaveCreator smwc, final double offset, 
				final double durationAttack, double durationDecay, 
				final double durationSustain, double durationRelease, 
				final double levelInitial, double levelPeak, double levelSustain) {
			this(smwc, offset, 
					createParts(durationAttack, durationDecay, durationSustain, durationRelease,
							levelInitial, levelPeak, levelSustain));
		}		

		public Parameters(final AbstractWave.Params params, 
				final double durationAttack, double durationDecay, 
				final double durationSustain, double durationRelease, 
				final double levelInitial, double levelPeak, double levelSustain) {
			this(params.smwc, params.offset, 
					createParts(durationAttack, durationDecay, durationSustain, durationRelease,
							levelInitial, levelPeak, levelSustain));
		}

		public final double[][] parts;
	}

	public EnvelopeWave(final Parameters params) {
		super(params);
	}

	public EnvelopeWave(final DynamicSoundMonoWaveCreator smwc, final double offset, final double[][] partParameters) {
		this(new Parameters(smwc, offset, partParameters));
	}

	/**
	 * Creates a complex envelope shape. Basically, that's a concatenation of
	 * linear interpolations in time between two amplitude levels.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param partParameters
	 *            a two-dimensional array. The first dimension is of arbitrary
	 *            length and determines the number of parts. The second
	 *            dimension is always of size 2, and index 0 contains the
	 *            duration and index 1 the start level of that part. The end
	 *            level of a part is implicitly defined as always the start
	 *            level of the next part, or 0 for the last part.
	 */
	public EnvelopeWave(final DynamicSoundMonoWaveCreator smwc, final double[][] partParameters) {
		this(smwc, 0.0, partParameters);
	}

	public EnvelopeWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final double offset, 
			final double duration, 
			final double startLevel, 
			final double endLevel) {
		this(new Parameters(smwc, offset, duration, startLevel, endLevel));
	}

	/**
	 * Creates a very simple envelope shape, starting at a start value and
	 * sliding linearily to an end value. Basically, that's a linear
	 * interpolation in time between two values.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param duration
	 *            the duration in seconds
	 * @param startLevel
	 *            the start level of the interpolation
	 * @param endLevel
	 *            the end level of the interpolation
	 */
	public EnvelopeWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final double duration, 
			final double startLevel, 
			final double endLevel) {
		this(new Parameters(smwc, 0.0, duration, startLevel, endLevel));
	}

	/**
	 * Creates the famous ADRS (Attack-Decay-Sustain-Release) envelope shape.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param durationAttack
	 *            the duration of the attack phase in seconds
	 * @param durationDecay
	 *            the duration of the decay phase in seconds
	 * @param durationSustain
	 *            the duration of the sustain phase in seconds
	 * @param durationRelease
	 *            the duration of the release phase in seconds
	 * @param levelInitial
	 *            the level to start with
	 * @param levelPeak
	 *            the level reached after the attack phase
	 * @param levelSustain
	 *            the level to keep during the sustain phase
	 */
	public EnvelopeWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final double durationAttack, double durationDecay, 
			final double durationSustain, double durationRelease, 
			final double levelInitial, double levelPeak, double levelSustain) {
		this(new Parameters(smwc, 0.0, 
				durationAttack, durationDecay, durationSustain, durationRelease,
				levelInitial, levelPeak, levelSustain));
	}

	protected Parameters getParams() {
		return (Parameters) super.params;
	}
	
	public double calcEnvelopePart(
			final double startValue, 
			final double endValue, 
			final int currentTimeslice, 
			final int partLength) {
		return getSoundSteroWaveCreator().interpolate(startValue, endValue, (double)(currentTimeslice) / (double)(partLength));
	}

	@Override
	protected double calcCurrentValue() {
		double ret = 0.0;
		final int timeslice = getTimeslice();
		final Parameters parameters = getParams();
		int partStartTimeslice = 0;
		int partLength = 0;
		int i = 0;
		for (i = 0; i < parameters.parts.length; i++) {
			final double duration = parameters.parts[i][0];
			partLength = getSoundSteroWaveCreator().durationToSize(duration);
			if (partStartTimeslice + partLength >= timeslice) {
				break;
			} else {
				partStartTimeslice += partLength;
			}
		}
		if (i < parameters.parts.length) {
			final double endLevel = ((i + 1) < parameters.parts.length ? parameters.parts[i + 1][1] : 0.0);
			ret = calcEnvelopePart(parameters.parts[i][1], endLevel, timeslice - partStartTimeslice, partLength);
//			logger.debug("Envelope part " + i + ", timeslice " + timeslice + " gets " + ret);
		}
		return ret;
	}

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		final Parameters parameters = getParams();
		ret.append("env(");
		for(final double[] part : parameters.parts) {
			ret.append("(dur=");
			ret.append(part[0]);
			ret.append(",ampStart=");
			ret.append(part[1]);
			ret.append(")");
		}
		ret.append(")");
		return ret.toString();
	}
}
