package de.ped.tools.sound;

import java.util.HashMap;

/**
 * Calculates the frequencies of common notes.
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Note {
	/** The note C, unspecific to any octave */
	public static final int C = 0;
	/** The note D, unspecific to any octave */
	public static final int D = 2;
	/** The note E, unspecific to any octave */
	public static final int E = 4;
	/** The note F, unspecific to any octave */
	public static final int F = 5;
	/** The note G, unspecific to any octave */
	public static final int G = 7;
	/** The note A, unspecific to any octave */
	public static final int A = 9;
	/** The note B, unspecific to any octave */
	public static final int B = 11;

	/** The sharp modifier to a note, "#" in musical notation */
	public static final int SHARP = 1;
	/** The flat modifier to a note, "b" in musical notation */
	public static final int FLAT  = -1;

	/** The number of semitones per octave for western music. */
	static final int SEMITONES_PER_OCTAVE = 12;

	private static final HashMap<Integer, Note> instances = new HashMap<Integer, Note>();

	public static Note get(final int id) {
		Note ret = instances.get(Integer.valueOf(id));
		if (null == ret) {
			ret = new Note(id);
			instances.put(Integer.valueOf(id), ret);
		}
		return ret;
	}

	/**
	 * Returns the note for a semi-tone and octave.
	 * <p>
	 * Example: <code>Note.get(Note.A, 1)</code> returns the chamber tone.
	 * 
	 * @param semitone
	 *            the tone as part of the octave. Could be {@link #C},
	 *            {@link #D}, {@link #E}, {@link #F}, {@link #G}, {@link #A},
	 *            {@link #B}, and you can add {@link #SHARP} to create sharp notes or
	 *            {@link #FLAT} to create flat notes.
	 * @param octave
	 *            the octave the tone is in. 0 for the octave starting with C0,
	 *            positive values for higher octaves, negative values for lower
	 *            octaves.
	 * @return the note object
	 */
	public static Note get(final int semitone, final int octave) {
		return get(noteID(semitone, octave));
	}

	/** 
	 * Creates (if it does not exist) and returns a {@link Note} object from the given note description, which supports the following pattern in ENBF:
	 * <pre>
	 * 	 [A-G][#b]*-?\d+
	 * </pre>
	 * <table>
	 * 	 <caption>Examples for note descriptions</caption>
	 *   <tbody>
	 *     <tr><th>A1</th><td>The chamber tone</td></tr>
	 *     <tr><th>A#1</th><td>One semi-tone higher than chamber tone</td></tr>
	 *     <tr><th>Ab1</th><td>One semi-tone below chamber tone</td></tr>
	 *     <tr><th>A#b1</th><td>The chamber tone again</td></tr>
	 *     <tr><th>A##1</th><td>Two semi-tones higher than chamber tone, same as B1</td></tr>
	 *     <tr><th>B1</th><td>One tone higher than chamber tone</td></tr>
	 *     <tr><th>A2</th><td>One octave higher than chamber tone</td></tr>
	 *     <tr><th>A0</th><td>One octave below chamber tone</td></tr>
	 *     <tr><th>A-1</th><td>Two octaves below chamber tone</td></tr>
	 *   </tbody>
	 * </table>
	 * @param noteDescription the string representation of the note
	 * @return the note for a note description
	 * @throws java.lang.IllegalArgumentException if the noteDescription did not fit to the fitting pattern
	 */
	public static Note get(final String noteDescription) {
		return get(noteID(noteDescription));
	}

	/**
	 * @param semitone
	 *            the tone as part of the octave. Could be {@link #C},
	 *            {@link #D}, {@link #E}, {@link #F}, {@link #G}, {@link #A},
	 *            {@link #B}, and you can add {@link #SHARP} to create sharp notes or
	 *            {@link #FLAT} to create flat notes.
	 * @param octave
	 *            the octave the tone is in. 0 for the octave starting with C0,
	 *            positive values for higher octaves, negative values for lower
	 *            octaves.
	 * @return the ID of a note. C0 gets the ID 0, C#0 the ID 1, higher tones
	 *         get positive values, lower tones gets negative values.
	 */
	public static int noteID(final int semitone, final int octave) {
		return SEMITONES_PER_OCTAVE * octave + semitone;
	}

	private static int noteID(final String noteDescription) {
		int semitone;
		final char note = noteDescription.charAt(0);
		switch(note) {
		case 'C':
			semitone = C;
			break;
		case 'D':
			semitone = D;
			break;
		case 'E':
			semitone = E;
			break;
		case 'F':
			semitone = F;
			break;
		case 'G':
			semitone = G;
			break;
		case 'A':
			semitone = A;
			break;
		case 'B':
			semitone = B;
			break;
		default:
			throw new IllegalArgumentException("Unable to read note " + note);
		}
		int pos = 1;
		while (true) {
			char mod = noteDescription.charAt(pos);
			if ('b' == mod) {
				semitone += FLAT;
				pos++;
			} else if ('#' == mod) {
				semitone += SHARP;
				pos++;
			} else {
				break;
			}
		}
		int octave = Integer.parseInt(noteDescription.substring(pos));
		return noteID(semitone, octave);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + id;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		Note other = (Note) obj;
		if (id != other.id) {
			return false;
		}
		return true;
	}

	public int getID() {
		return id;
	}

	private Note(final int id) {
		this.id = id;
	}

	private Note(final int semitone, final int octave) {
		this(noteID(semitone, octave));
	}

	private final int id;
}
