package de.ped.tools.sound;

import de.ped.tools.MathUtil;

/**
 * Defines a repeating wave (like sinus, square, triangle or saw) with a given
 * frequency (that is the reverse to the length of when the signal starts to
 * repeat itself) and the option for a frequency modulation. When the frequency
 * modulator has a value of 0, nothing is modulated. A modulation value of 1
 * sets the pitch up one seminote, and a modulation value of -1 sets the pitch
 * down one seminote.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class RepeatingWave extends AbstractWave {

	/**
	 * All the parameters defining the wave (excluding the wave form).
	 * 
	 * @author Copyright (c) 2009 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Params extends AbstractWave.Params {
		/**
		 * Defines a repeating wave with the full parameter set.
		 * 
		 * @param smwc
		 *            the sound wave creator object to attach this wave to
		 * @param offset
		 *            duration in seconds of leading zero signal
		 * @param phaseShift
		 *            the starting phase of the repeating signal, in the range
		 *            of (including) 0 to (excluding) 1
		 * @param normalFrequency
		 *            the unmodulated frequency of the repeating signal
		 * @param fm
		 *            the signal to modulate the frequency with
		 * @param am
		 *            the signal to modulate the amplitude with
		 */
		public Params(
				final DynamicSoundMonoWaveCreator smwc,
				final double offset, 
				final double phaseShift, 
				final double normalFrequency, 
				final MonoWave fm, 
				final MonoWave am) {
			super(smwc, offset);
			this.phaseShift = phaseShift;
			this.normalFrequency = normalFrequency;
			this.frequencyModulator = (null == fm ? new ConstantWave(0.0) : fm);
			this.amplitudeModulator = (null == am ? new ConstantWave(1.0) : am);
		}

		public Params(final DynamicSoundMonoWaveCreator smwc, final double normalFrequency, final MonoWave fm, final MonoWave am) {
			this(smwc, 0.0, 0.0, normalFrequency, fm, am);
		}

		public Params(final DynamicSoundMonoWaveCreator smwc, final double normalFrequency, final MonoWave fm, final double amplitude) {
			this(smwc, 0.0, 0.0, normalFrequency, fm, new ConstantWave(amplitude));
		}

		public Params(final DynamicSoundMonoWaveCreator smwc, final double normalFrequency, final MonoWave fm) {
			this(smwc, 0.0, 0.0, normalFrequency, fm, null);
		}

		public Params(final DynamicSoundMonoWaveCreator smwc, final double normalFrequency) {
			this(smwc, 0.0, 0.0, normalFrequency, null, null);
		}
		
		public Params(final AbstractWave.Params params, final double normalFrequency, final MonoWave fm) {
			this(params.smwc, params.offset, 0.0, normalFrequency, fm, null);
		}
		
		public Params(final AbstractWave.Params params, final double normalFrequency) {
			this(params.smwc, params.offset, 0.0, normalFrequency, null, null);
		}

		public final double phaseShift;
		public final double normalFrequency;
		public final MonoWave frequencyModulator;
		public final MonoWave amplitudeModulator;
	}

	protected Params getParams() {
		return (Params) super.params;
	}

	/**
	 * Returns the phase the wave currently has. Only the fractional part of the
	 * number transports meaning, so the signal at (phase + 1) looks the same as
	 * the signal at (phase) for all values of phase - what is the definition of
	 * <em>repeating</em>.
	 * 
	 * @return the phase the wave currently has
	 */
	public double getPhase() {
		return phase;
	}

	/** 
	 * @return the normal, unmodulated frequency of the signal created
	 */
	public double getNormalFrequency() {
		return getParams().normalFrequency;
	}

	/** 
	 * @return the frequency modulator for the signal created
	 */
	public MonoWave getFrequencyModulator() {
		return getParams().frequencyModulator;
	}

	/**
	 * Creates a repeating mono wave without frequency modulation.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param normalFrequency
	 *            the normal, unmodulated frequency of the signal created
	 */
	protected RepeatingWave(final DynamicSoundMonoWaveCreator smwc, final double normalFrequency) {
		this(new Params(smwc, normalFrequency));
	}

	/**
	 * Creates a repeating mono wave with frequency modulation.
	 * 
	 * @param smwc
	 *            the SoundStereoWaveCreator object this wave belongs to
	 * @param normalFrequency
	 *            the normal, unmodulated frequency of the signal created
	 * @param frequencyModulator
	 *            the wave the frequency is modulated with
	 */
	protected RepeatingWave(
			final DynamicSoundMonoWaveCreator smwc, 
			final double normalFrequency, 
			final MonoWave frequencyModulator) {
		this(new Params(smwc, normalFrequency, frequencyModulator));
	}

	protected RepeatingWave(final Params params) {
		super(params);
		this.phase = getParams().phaseShift;
	}

	/** 
	 * @return the modulated frequency 
	 */
	public double calcCurrentFrequency() {
		final Params params = getParams();
		final double fm = params.frequencyModulator.get();
		final double ret;
		final double normalFreq = params.normalFrequency;
		if (Math.abs(fm) < 0.0001) {
			ret = normalFreq;
		} else {
			return normalFreq * Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO, fm);
		}
		return ret;
	}

	/** 
	 * @param frequency the frequency to calculate with
	 * @return how much the phase increases each step for the given frequency 
	 */
	public double calcStepSize(final double frequency) {
		return (frequency / (double) getSoundSteroWaveCreator().getSampleRate());
	}

	/** 
	 * @return how much the phase increases each step for the current modulated frequency
	 */
	public double calcStepSize() {
		return calcStepSize(calcCurrentFrequency());
	}

	@Override
	public double next() {
		phase += calcStepSize();
		phase = MathUtil.modulo(phase, 1.0, 0.0);
		return super.next();
	}

	/** This value represents the phase during sound creation and is therefore modified over time. */
	private double phase = 0.0;
}