package de.ped.tools.sound;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.security.AccessControlException;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;

import de.ped.tools.log.Logger;

/**
 * A clip of wave sound to be played asynchronously.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class SoundClip {
	private static Logger logger = Logger.getRootLogger();

	/**
	 * Defines what to play if some clips shall run concurrently.
	 * 
	 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static enum ConcurrencyMode {
		ALL,		// Runs all clips in parallel
		FIRST_WINS,	// Play requests are ignored as long as the previos clip is not finished.
		LAST_WINS,	// When a new play request comes in, stops the current one and starts the new one.
	}

	/**
	 * A ring buffer for current clips, because we have to stop clips after playing to avoid threads flooding.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class RingBuffer {
		public RingBuffer(final int size) {
			currentClips = new Clip[size];
		}

		public void play(final SoundClip sound, final String name) {
			if (null != sound) {
				try {
					if (null != currentClips[currentClipIdx]) {
						SoundClip.stop(currentClips[currentClipIdx]);
						currentClips[currentClipIdx] = null;
					}
					currentClips[currentClipIdx] = sound.play();
					currentClipIdx = (currentClipIdx + 1) % currentClips.length;
				} catch (final NullPointerException ex) {
					logger.warn("Unexpected exception playing sound " + name, ex);
				} catch (final OutOfMemoryError ex) {
					logger.warn("Unexpected exception playing sound " + name, ex);
				}
			}
		}

		private final Clip[] currentClips;
		private int currentClipIdx = 0;
	}

	public static SoundClip soundClip(final DynamicSoundMonoWaveCreator smwc, final DynamicSoundMonoWaveCreator.Sound sound) {
		return new SoundClip(smwc.toAudioInputStream(sound));
	}

	/**
	 * Plays the clip asynchronously.
	 * 
	 * @param clip
	 *            the clip to play
	 */
	public static void play(final Clip clip) {
		if (null != clip) {
			clip.stop();
			clip.setFramePosition(0);
			clip.start();
			logger.trace("SoundClip.play(" + clip + ")");
		}
	}

	/** 
	 * Stops playing the clip. Could continue later on. 
	 * 
	 * @param clip
	 *            the clip to stop playing
	 */
	public static void stop(final Clip clip) {
		if (null != clip) {
			clip.stop();
			logger.trace("SoundClip.stop(" + clip + ")");
		}
	}

	/** 
	 * Closes the clip and allows to release the resources. 
	 * This operation might be time-consuming for longer clips. 
	 * 
	 * @param clip
	 *            the clip to stop playing and close
	 */
	public static void close(final Clip clip) {
		if (null != clip) {
			clip.stop();
			clip.flush();
			clip.close();
			logger.trace("SoundClip.close(" + clip + ")");
		}
	}

	public static void pauseOrResume(final Clip clip, final boolean isToResume) {
		if (null != clip) {
			if (isToResume) {
				clip.start();
			} else {
				clip.stop();
			}
		}
	}

	/**
	 * Creates and loads a clip from file.
	 * 
	 * @param filename
	 *            the name (and path) of the file to load from
	 */
	public SoundClip(final String filename) {
		init(filename);
	}

	/**
	 * Creates and loads a clip from URL.
	 * 
	 * @param url
	 *            the URL to load from
	 */
	public SoundClip(final URL url) {
		init(url);
	}

	/**
	 * Creates and loads a clip from an audio input stream.
	 * 
	 * @param audioInputStream
	 *            the audio input stream to load from
	 */
	public SoundClip(final AudioInputStream audioInputStream) {
		init(audioInputStream);
	}

	/**
	 * Creates a clip from a wave created in-memory.
	 * 
	 * @param sampleRate
	 *            the sample rate
	 * @param sample
	 *            the wave data
	 */
	public SoundClip(final int sampleRate, final byte[] sample) {
		init(sampleRate, sample);
	}

	/** 
	 * Loads a clip from file. 
	 * 
	 * @param filename
	 *            the name (and path) of the file to load from
	 */
	public void init(final String filename) {
		this.filename = filename;
		try {
			init(AudioSystem.getAudioInputStream(new File(filename)));
		} catch (final UnsupportedAudioFileException e) {
			logger.debug(e.toString(), e);
		} catch (final IOException e) {
			logger.debug(e.toString(), e);
		} catch (final AccessControlException e) {
			logger.debug(e.toString(), e);
		}
	}

	/** 
	 * Loads a clip from URL. 
	 * 
	 * @param url
	 *            the URL to load from
	 */
	public void init(final URL url) {
		this.filename = url.getFile();
		try {
			init(AudioSystem.getAudioInputStream(url));
		} catch (final UnsupportedAudioFileException e) {
			logger.debug(e.toString(), e);
		} catch (final IOException e) {
			logger.debug(e.toString(), e);
		} catch (final AccessControlException e) {
			logger.debug(e.toString(), e);
		}
	}

	/** 
	 * Creates a clip from a wave created in-memory. 
	 * 
	 * @param sampleRate
	 *            the sample rate
	 * @param sample
	 *            the wave data
	 */
	public void init(final int sampleRate, final byte[] sample) {
		this.filename = null;
		final AudioInputStream ais = DynamicSoundMonoWaveCreator.toAudioInputStream(sampleRate, sample);
		init(ais);
	}

	/** 
	 * Loads a clip from an audio input stream. 
	 * 
	 * @param audioInputStream
	 *            the audio input stream to load from
	 */
	public void init(final AudioInputStream audioInputStream) {
		try {
			audioFormat = audioInputStream.getFormat();
			size = (int) (audioFormat.getFrameSize() * audioInputStream.getFrameLength());
			audio = new byte[size];
			int offset = 0;
			int read;
			do {
				read = audioInputStream.read(audio, offset, size - offset);
				if (0 <= read) {
					offset += read;
				}
			} while ((0 <= read) && (offset < size));
			logger.trace("Trying to read " + size + " bytes of sound, read " + offset + ", diff is " + (size - offset) + ", last chunk was " + read);
			dataLineInfo = new DataLine.Info(Clip.class, audioFormat, size);
		} catch (final IOException e) {
			logger.debug(e.toString(), e);
		} catch (final AccessControlException e) {
			logger.debug(e.toString(), e);
		}
	}

	/**
	 * @return The filename of the sound file, 
	 *     or <code>null</code> if this sound clip was created in memory.
	 */
	public String getFilename() {
	    return filename;
	}

	public ConcurrencyMode getConcurrencyMode() {
		return concurrencyMode;
	}

	public void setConcurrencyMode(final ConcurrencyMode concurrencyMode) {
		this.concurrencyMode = concurrencyMode;
	}

	public boolean isAllowedPlayingInParallel() {
		return (ConcurrencyMode.ALL == concurrencyMode);
	}

	/**
	 * Creates the clip object for playing.
	 * <p>
	 * This is an expensive operation, as it internally creates a new thread
	 * object. If called often enough without cleaning up the clips by calling
	 * {@link #close(Clip)}, the system will barf.
	 * 
	 * @return the clip object
	 */
	public Clip createClip() {
		Clip clip = null;
		try {
			clip = (Clip) AudioSystem.getLine(dataLineInfo);
			clip.open(audioFormat, audio, 0, size);
		} catch (final LineUnavailableException e) {
			logger.debug(e.toString(), e);
		} catch (final NullPointerException e) {
			// Nothing to do - most likely a file not found.
		}
		return clip;
	}

	/**
	 * Creates the clip object for playing, if the current state and concurrency
	 * mode allows.
	 * <p>
	 * The creation of a clip object is an expensive operation, as it internally
	 * creates a new thread object. If called often enough without cleaning up
	 * the clips by calling {@link #close(Clip)}, the system will barf.
	 * 
	 * @return the clip object
	 */
	public Clip getOrCreateClip() {
		Clip clip = null;
		switch(concurrencyMode) {
		case ALL:
		case LAST_WINS:
			clip = createClip();
			break;
		case FIRST_WINS:
			if (null == singleClip) {
				singleClip = createClip();
			}
			clip = singleClip;
			break;
		}
		return clip;
	}
	
	/**
	 * Creates the clip object for playing and plays the clip asynchronously.
	 * 
	 * @return the clip object
	 * @throws NullPointerException if no clip could be created
	 * @throws OutOfMemoryError if not enough memory available to create sound clip or thread
	 */
	public Clip play() throws NullPointerException, OutOfMemoryError {
		Clip clip = getOrCreateClip();
		play(clip);
		return clip;
	}

	private String filename;
	private AudioFormat audioFormat;
	private int size;
	private byte[] audio;
	private DataLine.Info dataLineInfo;
	private ConcurrencyMode concurrencyMode = ConcurrencyMode.FIRST_WINS;
	private Clip singleClip = null;
}
