package de.ped.tools.sound;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.SourceDataLine;

import de.ped.tools.MathUtil;

/**
 * A simple creator for wave sounds using additive synthesis.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class SoundMonoWaveCreator {
	/** The maximum amplitude to multiply with to convert to integers. */
	public static final int MAX_AMPLITUDE = 10000; //32000;
 
	/**
	 * Defines a sinus wave of amplitude 1, given frequency for given sample
	 * rate.
	 * 
	 * @param i
	 *            the current step
	 * @param freq
	 *            the frequency
	 * @param sampleRate
	 *            the sample rate
	 * @return the wave value at the given step
	 */
	public static double sinusWave(final int i, final double freq, final int sampleRate) {
		double waveSize = ((double) sampleRate) / freq;
		return (Math.sin((double) i / waveSize * 2.0 * Math.PI));
	}

	/** 
	 * Defines a square wave of amplitude 1, given frequency for given sample rate. 
	 * 
	 * @param i
	 *            the current step
	 * @param freq
	 *            the frequency
	 * @param sampleRate
	 *            the sample rate
	 * @return the wave value at the given step
	 */
	public static double squareWave(final int i, final double freq, final int sampleRate) {
		double ret = 1.0f;
		double waveSize = ((double) sampleRate) / freq;
		if (MathUtil.modulo(i, waveSize, 0) < (waveSize / 2.0)) {
			ret = -ret;
		}
		return ret;
	}

	/** 
	 * Defines a saw wave (with saws leading up) of amplitude 1, given frequency for given sample rate. 
	 * 
	 * @param i
	 *            the current step
	 * @param freq
	 *            the frequency
	 * @param sampleRate
	 *            the sample rate
	 * @return the wave value at the given step
	 */
	public static double sawUpWave(final int i, final double freq, final int sampleRate) {
		double waveSize = ((double) sampleRate) / freq;
		return (float) (-1.0 + 2.0 * MathUtil.modulo((double) i, waveSize, 0.0) / waveSize);
	}

	/** 
	 * Defines a saw wave (with saws leading down) of amplitude 1, given frequency for given sample rate. 
	 * 
	 * @param i
	 *            the current step
	 * @param freq
	 *            the frequency
	 * @param sampleRate
	 *            the sample rate
	 * @return the wave value at the given step
	 */
	public static double sawDownWave(final int i, final double freq, final int sampleRate) {
		return -sawUpWave(i, freq, sampleRate);
	}

	/** 
	 * Defines a triangle wave of amplitude 1, given frequency for given sample rate. 
	 * 
	 * @param i
	 *            the current step
	 * @param freq
	 *            the frequency
	 * @param sampleRate
	 *            the sample rate
	 * @return the wave value at the given step
	 */
	public static double triangleWave(final int i, final double freq, final int sampleRate) {
		double ret = 1.0;
		double waveSize = ((double) sampleRate) / freq;
		if (MathUtil.modulo(i, waveSize, 0) < (waveSize / 2.0)) {
			ret = sawUpWave(i, freq * 2.0, sampleRate);
		} else {
			ret = sawDownWave(i, freq * 2.0, sampleRate);
		}
		return ret;
	}

	/**
	 * Plays the sample. The thread is waiting while the sample plays.
	 * 
	 * @param sampleRate
	 *            the sample rate to play with
	 * @param sample
	 *            the sample to play
	 */
	public static void playActive(final int sampleRate, final byte[] sample) {
		try {
			final AudioFormat audioformat = getAudioFormat(sampleRate);
			final DataLine.Info datalineinfo = new DataLine.Info(SourceDataLine.class, audioformat);
			if (!AudioSystem.isLineSupported(datalineinfo)) {
				System.out.println("Line matching " + datalineinfo + " is not supported.");
			} else {
				final SourceDataLine sourcedataline = (SourceDataLine) AudioSystem.getLine(datalineinfo);
				sourcedataline.open(audioformat);
				sourcedataline.start();
				sourcedataline.write(sample, 0, sample.length);
				sourcedataline.drain();
				sourcedataline.stop();
				sourcedataline.close();
			}
		} catch (final LineUnavailableException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Plays a sample packed as {@link AudioInputStream}. The thread is waiting while the sample plays.
	 * 
	 * @param stream
	 *            the audio input stream to play
	 */
	public static void playActive(final AudioInputStream stream) {
		try {
			final AudioFormat audioformat = stream.getFormat();
			final int length = stream.available();
			final byte[] arr = new byte[length];
			stream.read(arr);
			playActive(Math.round(audioformat.getSampleRate()), arr);
		} catch (final IOException e) {
			e.printStackTrace();
		}
	}

	/** The default sample rate. */
	public static final int DEFAULT_SAMPLE_RATE = 22050;

	/** The sample rate to calculate with. */
	private int sampleRate = DEFAULT_SAMPLE_RATE;

	public int getSampleRate() {
		return sampleRate;
	}

	/** 
	 * Creates a wave creator with default sample rate @link{#DEFAULT_SAMPLE_RATE}. 
	 */
	public SoundMonoWaveCreator() {
		// Nothing to do
	}

	/**
	 * Creates a wave creator with given sample rate.
	 * 
	 * @param sampleRate
	 *            the sample rate
	 */
	public SoundMonoWaveCreator(final int sampleRate) {
		this.sampleRate = sampleRate;
	}

	public int durationToSize(final double length) {
		return (int) (length * (double) (sampleRate));
	}

	public double sizeToDuration(final int size) {
		return ((double) (size) * (double) (sampleRate));
	}

	/**
	 * Creates a sinus wave with given duration in seconds and given frequency.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param frequency
	 *            the frequency (defining the tone height) of the wave to
	 *            create
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] sinusWave(final double duration, final double frequency) {
		double[] ret = new double[durationToSize(duration)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = sinusWave(i, frequency, sampleRate);
		}
		return ret;
	}

	/**
	 * Creates a square wave with given duration in seconds and given frequency.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param frequency
	 *            the frequency (defining the tone height) of the wave to
	 *            create
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] squareWave(final double duration, final double frequency) {
		final double[] ret = new double[durationToSize(duration)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = squareWave(i, frequency, sampleRate);
		}
		return ret;
	}

	/**
	 * Creates a saw wave with given duration in seconds and given frequency.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param frequency
	 *            the frequency (defining the tone height) of the wave to
	 *            create
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] sawUpWave(final double duration, final double frequency) {
		final double[] ret = new double[durationToSize(duration)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = sawUpWave(i, frequency, sampleRate);
		}
		return ret;
	}

	/**
	 * Creates a saw wave with given duration in seconds and given frequency.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param frequency
	 *            the frequency (defining the tone height) of the wave to
	 *            create
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] sawDownWave(final double duration, final double frequency) {
		final double[] ret = new double[durationToSize(duration)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = sawDownWave(i, frequency, sampleRate);
		}
		return ret;
	}

	/**
	 * Creates a triangle wave with given duration in seconds and given
	 * frequency.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param frequency
	 *            the frequency (defining the tone height) of the wave to
	 *            create
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] triangleWave(final double duration, final double frequency) {
		final double[] ret = new double[durationToSize(duration)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = triangleWave(i, frequency, sampleRate);
		}
		return ret;
	}

	/**
	 * Creates a totally random noise sample of amplitude 1.0.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] whiteNoise(final double duration) {
		final int size = (int) (duration * (double) (sampleRate));
		final double[] ret = new double[size];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = -1.0 + 2.0 * Math.random();
		}
		return ret;
	}

	/**
	 * Creates a random noise sample of amplitude 1.0.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param maxIncrease
	 *            the higher maxIncrease is, the higher the frequencies added
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] pinkNoiseSaw(final double duration, final double maxIncrease) {
		final int size = (int) (duration * (double) (sampleRate));
		final double[] ret = new double[size];
		ret[0] = -1.0 + 2.0 * Math.random();
		for (int i = 1; i < ret.length; i++) {
			ret[i] = MathUtil.modulo(ret[i - 1] + 2.0 * maxIncrease * Math.random(), 2.0, -1.0);
		}
		return ret;
	}

	/**
	 * Creates a random noise sample of amplitude 1.0.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param maxIncrease
	 *            the maximum difference between two adjacent points. A value of
	 *            1.0 is nearly white noise, smaller values cut of higher
	 *            frequencies and make the sound deeper
	 * @return the wave as array of doubles containing values between -1.0 and
	 *         +1.0
	 */
	public double[] pinkNoise(final double duration, final double maxIncrease) {
		final int size = (int) (duration * (double) (sampleRate));
		final double[] ret = new double[size];
		ret[0] = -1.0 + 2.0 * Math.random();
		for (int i = 1; i < ret.length; i++) {
			ret[i] = MathUtil.modulo(ret[i - 1] - maxIncrease + 2.0 * maxIncrease * Math.random(), 2.0, -1.0);
		}
		return ret;
	}

	public double interpolate(final double level1, final double level2, final double partOf2nd) {
		return (level1 * (1.0 - partOf2nd) + level2 * partOf2nd);
	}

	/**
	 * Creates a single part of an envelope shape. Basically, that's a linear
	 * interpolation in time between two amplitude levels.
	 * 
	 * @param duration
	 *            the duration in seconds
	 * @param startLevel
	 *            the start level of the interpolation
	 * @param endLevel
	 *            the end level of the interpolation
	 * @return the envelope shape
	 */
	public double[] envelopePartWave(final double duration, final double startLevel, final double endLevel) {
		final double[] ret = new double[durationToSize(duration)];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = interpolate(startLevel, endLevel, (double) (i) / (double) (ret.length));
		}
		return ret;
	}

	/**
	 * Creates a complex envelope shape. Basically, that's a concatenation of
	 * linear interpolations in time between two amplitude levels.
	 * 
	 * @param partParameters
	 *            a two-dimensional array. The first dimension is of arbitrary
	 *            length and determines the number of parts. The second
	 *            dimension is always of size 2, and index 0 contains the
	 *            duration and index 2 the start level of that part. The end
	 *            level of a part is implicitly defined as always the start
	 *            level of the next part, or 0 for the last part.
	 * @return the envelope shape
	 */
	public double[] envelopeWave(final double[][] partParameters) {
		double[] ret = new double[0];
		for (int i = 0; i < partParameters.length; i++) {
			final double endLevel = ((i + 1) < partParameters.length ? partParameters[i + 1][1] : 0.0);
			final double[] part = envelopePartWave(partParameters[i][0], partParameters[i][1], endLevel);
			ret = concat(ret, part, 0);
		}
		return ret;
	}

	/**
	 * Creates the famous ADRS (Attack-Decay-Sustain-Release) envelope shape.
	 * 
	 * @param durationAttack
	 *            the duration of the attack phase in seconds
	 * @param durationDecay
	 *            the duration of the decay phase in seconds
	 * @param durationSustain
	 *            the duration of the sustain phase in seconds
	 * @param durationRelease
	 *            the duration of the release phase in seconds
	 * @param levelInitial
	 *            the level to start with
	 * @param levelPeak
	 *            the level reached after the attack phase
	 * @param levelSustain
	 *            the level to keep during the sustain phase
	 * @return the envelope shape
	 */
	public double[] ADSRWave(
			final double durationAttack, 
			final double durationDecay, 
			final double durationSustain, 
			final double durationRelease, 
			final double levelInitial, 
			final double levelPeak, 
			final double levelSustain) {
		final double[][] partParameters = new double[][] {
				{ durationAttack,  levelInitial },
				{ durationDecay,   levelPeak },
				{ durationSustain, levelSustain },
				{ durationRelease, levelSustain },
		};
		return envelopeWave(partParameters);
	}

	public double[] smoothSpot(final double[] arr, final int pos, final int width) {
		// TODO: Implement if ever needed.
		return arr;
	}

	/**
	 * Concatenates the given sounds to one.
	 * 
	 * @param arrs
	 *            the wave files to concat
	 * @param smoothWidth
	 *            currently unused
	 * @return a wave array of a duration that is equal to the sum of the
	 *         duration of all wave arrays given
	 */
	public double[] concat(final double[][] arrs, final int smoothWidth) {
		int size = 0;
		for (int a = 0; a < arrs.length; a++) {
			size += arrs[a].length;
		}
		final double[] ret = new double[size];
		int p = 0;
		for (int a = 0; a < arrs.length; a++) {
			for (int i = 0; i < arrs[a].length; i++) {
				ret[p++] = arrs[a][i];
			}
		}
		if (smoothWidth > 0) {
			int pos = 0;
			for (int a = 0; a < arrs.length; a++) {
				pos += arrs[a].length;
				smoothSpot(ret, pos, smoothWidth);
			}
		}
		return ret;
	}

	public double[] concat(final double[][] arrs) {
		return concat(arrs, 0);
	}

	/**
	 * Concatenates the given sounds to one.
	 * 
	 * @param arr1
	 *            the 1st wave array to concat
	 * @param arr2
	 *            the 2nd wave array to concat
	 * @param smoothWidth
	 *            currently unused
	 * @return a wave array of a duration that is equal to the sum of the
	 *         duration of both wave arrays given
	 */
	public double[] concat(final double[] arr1, final double[] arr2, final int smoothWidth) {
		return concat(new double[][] { arr1, arr2 }, smoothWidth);
	}

	public double[] concat(final double[] arr1, final double[] arr2) {
		return concat(new double[][] { arr1, arr2 }, 0);
	}

	/**
	 * Cuts a part out of a wave array.
	 * 
	 * @param arr
	 *            the original wave array
	 * @param start
	 *            the start point of the snippet (in seconds)
	 * @param end
	 *            the end point of the snippet (in seconds). If end is smaller
	 *            than start, an empty array is returned. If end is exceeding
	 *            the duration of arr, the result is clipped to the duration of
	 *            arr.
	 * @return the snippet as wave array
	 */
	public double[] snippet(final double[] arr, final double start, final double end) {
		final int startOffset = MathUtil.bound(durationToSize(start), 0, arr.length);
		final int endOffset = MathUtil.bound(durationToSize(end), startOffset, arr.length);
		final double[] ret;
		if (startOffset < endOffset) {
			ret = new double[endOffset - startOffset];
			for (int i = 0; i < ret.length; i++) {
				ret[i] = arr[i - startOffset];
			}
		} else {
			ret = new double[0];
		}
		return ret;
	}

	private double getAtOrZero(final double[] arr, final int idx) {
		return (((0 <= idx) && (idx < arr.length)) ? arr[idx] : 0.0);
	}

	/**
	 * Adds the given sounds to one, so that they happen at the same time.
	 * 
	 * @param arr1
	 *            the 1st array to add
	 * @param arr2
	 *            the 2nd array to add
	 * @return the resulting wave array. It's duration is the duration of the
	 *         longest input wave array
	 */
	public double[] add(final double[] arr1, final double[] arr2) {
		final int size = Math.max(arr1.length, arr2.length);
		final double[] ret = new double[size];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = getAtOrZero(arr1, i) + getAtOrZero(arr2, i);
		}
		return ret;
	}

	/**
	 * Multiplies the given sounds to one. Useful e.g. to apply a hull curve to
	 * a sound.
	 * 
	 * @param arr1
	 *            the 1st array to multiply
	 * @param arr2
	 *            the 2nd array to multiply
	 * @return the resulting wave array. It's duration is the duration of the
	 *         shortest input wave array
	 */
	public double[] multiply(final double[] arr1, final double[] arr2) {
		final int size = Math.min(arr1.length, arr2.length);
		final double[] ret = new double[size];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = arr1[i] * arr2[i];
		}
		return ret;
	}

	/**
	 * Multiplies the given sound with a constant factor - in other words,
	 * multiplies the volume of the sound by the factor.
	 * 
	 * @param arr
	 *            the array to multiply
	 * @param factor
	 *            the factor to multiply all values with
	 * @return the resulting wave array. It's duration is the duration of the
	 *         input wave array
	 */
	public double[] multiply(final double[] arr, final double factor) {
		final double[] ret = new double[arr.length];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = arr[i] * factor;
		}
		return ret;
	}

	/**
	 * Limits all wave values to be within the range [-maxValue, +maxValue].
	 * 
	 * @param arr
	 *            the array to limit
	 * @param maxValue
	 *            the array to limit to
	 * @return the resulting wave array. It's duration is the duration of the
	 *         input wave array
	 */
	public double[] cutAmplitude(final double[] arr, final double maxValue) {
		final double[] ret = new double[arr.length];
		for (int i = 0; i < ret.length; i++) {
			ret[i] = MathUtil.bound(arr[i], -maxValue, maxValue);
		}
		return ret;
	}

	/**
	 * Converts a mono wave to a stereo sample with equal left and right
	 * channel.
	 * 
	 * @param arr
	 *            the mono wave
	 * @return the stereo sample
	 */
	public static byte[] toStereoSample(final double[] arr) {
		final byte[] ret = new byte[4 * arr.length];
		int adr = 0;
		for (int i = 0; i < arr.length; i++, adr += 4) {
			int sample = MathUtil.bound((int) (arr[i] * (double) (MAX_AMPLITUDE)), -MAX_AMPLITUDE, MAX_AMPLITUDE);
			ret[adr + 0] = (byte) (sample);
			ret[adr + 1] = (byte) (sample >>> 8);
			ret[adr + 2] = (byte) (sample);
			ret[adr + 3] = (byte) (sample >>> 8);
		}
		return ret;
	}

	/**
	 * Returns the audio format object fitting for this sound creator and sample rate.
	 * @param sampleRate the sample rate
	 * @return the audio format object fitting for this sound creator and sample rate.
	 */
	public static AudioFormat getAudioFormat(final int sampleRate) {
		return new AudioFormat(
				AudioFormat.Encoding.PCM_SIGNED,
				sampleRate,
				16,
				2,
				4,
				sampleRate,
				false);
	}

	public static AudioInputStream toAudioInputStream(final int sampleRate, final byte[] arr) {
		final InputStream is = new ByteArrayInputStream(arr);
		final AudioFormat audioFormat = getAudioFormat(sampleRate);
		return new AudioInputStream(is, audioFormat, arr.length / 4);
	}

	public AudioInputStream toAudioInputStream(final double[] arr) {
		final byte[] barr = toStereoSample(arr);
		final InputStream is = new ByteArrayInputStream(barr);
		final AudioFormat audioFormat = getAudioFormat(sampleRate);
		return new AudioInputStream(is, audioFormat, arr.length / 4);
	}
}
