package de.ped.tools.sound;

/**
 * The tuning to apply when calculating the frequency of a note.
 * 
 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Tuning {
	/**
	 * A default tuning as singleton for ease of use.
	 * 
	 * @return the default tuning
	 */
	public static Tuning defaultTuning() {
		if (null == instance) {
			instance = new Tuning(TuningSystem.EQUAL_12_EDO, Note.get(0), Note.get("A1"), Tuning.CHAMBER_TONE_A1_FREQUENCY);
		}
		return instance;
	}

	/** A default tuning as singleton for ease of use. */
	private static Tuning instance;

	/** The difference factor between two frequencies being one semitone apart for the equal tuning system in western music. */
	public static final double SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO = Math.pow(2.0, 1.0 / ((double)Note.SEMITONES_PER_OCTAVE));

	public static final double CHAMBER_TONE_A1_FREQUENCY = 440.0;

	/** 
	 * Constructor.
	 * @param tuningSystem the tuning system to use
	 * @param baseTone the base tone for the tuning
	 * @param referenceTone the reference tone for the tuning
	 * @param referenceToneFrequency the frequency of the reference tone
	 */
	public Tuning(TuningSystem tuningSystem, Note baseTone, Note referenceTone, double referenceToneFrequency) {
		this.tuningSystem = tuningSystem;
		this.baseTone = baseTone;
		final int halfTones = referenceTone.getID() - baseTone.getID();
		this.baseToneFrequency = referenceToneFrequency * tuningSystem.getReverseFrequencyFactor(halfTones);
	}

	/**
	 * Returns the frequency of the given note.
	 * @param note the note
	 * @return the frequency of the given note
	 */
	public double getFrequency(final Note note) {
		return baseToneFrequency * tuningSystem.getFrequencyFactor(note.getID() - baseTone.getID());
	}

	/** The tuning system to use. */
	public final TuningSystem tuningSystem;

	/** The base tone. */
	public final Note baseTone;

	/** The frequency of the base tone. */
	public final double baseToneFrequency;
}