package de.ped.tools.sound;

import de.ped.tools.MathUtil;

/**
 * The tuning system to apply when calculating the frequency of a note.
 * See <a href="https://en.wikipedia.org/wiki/Musical_tuning#Tuning_systems">Wikipedia</a> for details.
 * 
 * @author Copyright (c) 2017 Peter Diefenbach (peter@pdiefenbach.de)
 */
public enum TuningSystem {
	PYTHAGOREAN (new double[] { 
			256.0 / 243.0, 
			  9.0 /   8.0, 
			 32.0 /  27.0, 
			 81.0 /  64.0, 
			  4.0 /   3.0, 
			729.0 / 512.0, 
			  3.0 /   2.0, 
			128.0 /  81.0, 
			 27.0 /  16.0, 
			 16.0 /   9.0, 
			243.0 / 128.0, 
	}),
	EQUAL_12_EDO(new double[] { 
			Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  2),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  3),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  4),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  5),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  6),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  7),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  8),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO,  9),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO, 10),
			Math.pow(Tuning.SEMITONE_FACTOR_EQUAL_TEMPERAMENT_12_EDO, 11),
	});
	
	private double[] ratios;
	
	/**
	 * Defines a tuning system.
	 * 
	 * @param ratios
	 *            an array defining the frequency ratios of the tones of an
	 *            octave compared to the base note of the octave. Element 0
	 *            defines the factor for the first half note above base
	 *            note, the last lement defines the factor the last half
	 *            note below one octave above base note. (The octave above
	 *            always has factor 2.0, so it's not part of the array.) So,
	 *            for western music with 12 half-tones in an octave, this
	 *            array is of size 11.
	 */
	private TuningSystem(final double[] ratios) {
		this.ratios = ratios;
	}
	
	public int getNumberOfHalfTones() {
		return ratios.length + 1;
	}
	
	/**
	 * Calculates the frequency of a tone.
	 * 
	 * @param halfTones
	 *            The number of half tones higher than the base note. So a
	 *            negative number represents half tones below the base note.
	 * @return the frequency of the tone
	 */
	public double getFrequencyFactor(final int halfTones) {
		final int halfTonesPerOctave = getNumberOfHalfTones();
		final int mod = MathUtil.modulo(halfTones, halfTonesPerOctave);
		final int div = MathUtil.divisio(halfTones, halfTonesPerOctave);
		final double modf = (0 == mod ? 1.0 : ratios[mod - 1]);
		return Math.pow(2.0, div) * modf;
	}
	
	/**
	 * Calculates the frequency of the base tone, if another tone is the
	 * frequency reference. So it's the reverse operation of
	 * {@link #getFrequencyFactor(int)}.
	 * 
	 * @param halfTones
	 *            The number of half tones the reference is higher than the base note. So a
	 *            negative number represents half tones down.
	 * @return the frequency of the base tone
	 */
	public double getReverseFrequencyFactor(final int halfTones) {
		return 1.0 / getFrequencyFactor(halfTones);
	}
}