package de.ped.tools.test;

import static org.junit.Assert.*;

import org.junit.Test;

import de.ped.tools.BitArea;
import de.ped.tools.Bitfield32;
import de.ped.tools.Bitfield64;

/**
 * Unit tests for the {@link BitArea}, {@link Bitfield32} and {@link Bitfield64} class.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class BitfieldTest {

	/**
	 * @see Bitfield32#SIZE
	 * @see Bitfield64#SIZE
	 */
	@Test
	public void testSize() {
		assertEquals(32, Bitfield32.SIZE);
		assertEquals(64, Bitfield64.SIZE);
	}

	/**
	 * @see Bitfield32#setBit(int, boolean)
	 * @see Bitfield32#isBit(int)
	 */
	@Test
	public void test32SetBit() {
		Bitfield32 bitfield = new Bitfield32();
		for (int i = 0; i < Bitfield32.SIZE; i++) {
			assertEquals(false, bitfield.isBit(i));
			bitfield.setBit(i, true);
			assertEquals(true, bitfield.isBit(i));
			bitfield.setBit(i, false);
			assertEquals(false, bitfield.isBit(i));
		}
	}

	/**
	 * @see Bitfield64#setBit(int, boolean)
	 * @see Bitfield64#isBit(int)
	 */
	@Test
	public void test64SetBit() {
		Bitfield64 bitfield = new Bitfield64();
		for (int i = 0; i < Bitfield64.SIZE; i++) {
			assertEquals(false, bitfield.isBit(i));
			bitfield.setBit(i, true);
			assertEquals(true, bitfield.isBit(i));
			bitfield.setBit(i, false);
			assertEquals(false, bitfield.isBit(i));
		}
	}

	private static final int CHUNKSIZE = 3;

	/**
	 * @see Bitfield32#setBits(int, int, int)
	 * @see Bitfield32#getBits(int, int)
	 * @see Bitfield32#clone()
	 */
	@Test
	public void test32SetBits() {
		Bitfield32 bf = new Bitfield32();
		for (int i = 0; i <= Bitfield32.SIZE - CHUNKSIZE; i += CHUNKSIZE) {
			final String msg = "i=" + i; 
			assertEquals(msg, 0, bf.getBits(i, CHUNKSIZE));
			bf.setBits(i, 3, 0x5);
			assertEquals(msg, 0x5, bf.getBits(i, CHUNKSIZE));
			Bitfield32 bitfield = (Bitfield32) bf.clone();
			assertEquals(msg, 0x5, bitfield.getBits(i, CHUNKSIZE));
		}
	}

	/**
	 * @see Bitfield64#setBits(int, int, long)
	 * @see Bitfield64#getBits(int, int)
	 */
	@Test
	public void test64SetBits() {
		Bitfield64 bf = new Bitfield64();
		for (int i = 0; i <= Bitfield64.SIZE - CHUNKSIZE; i += CHUNKSIZE) {
			final String msg = "i=" + i; 
			assertEquals(msg, 0, bf.getBits(i, CHUNKSIZE));
			bf.setBits(i, 3, 0x5);
			assertEquals(msg, 0x5, bf.getBits(i, CHUNKSIZE));
			Bitfield64 bitfield = (Bitfield64) bf.clone();
			assertEquals(msg, 0x5, bitfield.getBits(i, CHUNKSIZE));
		}
	}

	/**
	 * @see BitArea#equals(Object)
	 * @see BitArea#bits32mask
	 * @see BitArea#bits64mask
	 */
	@Test
	public void testBitArea() {
		GenericTest.checkObjectBasics(false, BA_00_4, BA_28_4);
		assertEquals(0xf, BA_00_4.bits32mask);
		assertEquals(0xfL, BA_00_4.bits64mask);
		assertEquals(0xf0000000, BA_28_4.bits32mask);
		assertEquals(0xf0000000L, BA_28_4.bits64mask);
		assertEquals(0xc0000000, BA_30_4.bits32mask);
		assertEquals(0x3c0000000L, BA_30_4.bits64mask);
		assertEquals(0x0, BA_60_4.bits32mask);
		assertEquals(0xf000000000000000L, BA_60_4.bits64mask);
		assertEquals(0x0, BA_62_4.bits32mask);
		assertEquals(0xc000000000000000L, BA_62_4.bits64mask);
	}

	/**
	 * @see Bitfield32#setBits(BitArea, int)
	 * @see Bitfield32#getBits(BitArea)
	 */
	@Test
	public void test32SetBitsArea() {
		Bitfield32 bitfield = new Bitfield32();
		bitfield.setBits(BA_00_4, 0xa);
		assertEquals(0xa, bitfield.getBits(BA_00_4));
		bitfield.setBits(BA_28_4, 0xb);
		assertEquals(0xb, bitfield.getBits(BA_28_4));
		bitfield.setBits(BA_30_4, 0xa);
		assertEquals(0x2, bitfield.getBits(BA_30_4));
		bitfield.setBits(BA_60_4, 0xc);
		assertEquals(0x0, bitfield.getBits(BA_60_4));
		bitfield.setBits(BA_62_4, 0xa);
		assertEquals(0x0, bitfield.getBits(BA_62_4));
	}

	/**
	 * @see Bitfield64#setBits(BitArea, long)
	 * @see Bitfield64#getBits(BitArea)
	 */
	@Test
	public void test64SetBitsArea() {
		Bitfield64 bitfield = new Bitfield64();
		bitfield.setBits(BA_00_4, 0xa);
		assertEquals(0xa, bitfield.getBits(BA_00_4));
		bitfield.setBits(BA_28_4, 0xb);
		assertEquals(0xb, bitfield.getBits(BA_28_4));
		bitfield.setBits(BA_30_4, 0xa);
		assertEquals(0xa, bitfield.getBits(BA_30_4));
		bitfield.setBits(BA_60_4, 0xc);
		assertEquals(0xc, bitfield.getBits(BA_60_4));
		bitfield.setBits(BA_62_4, 0xa);
		assertEquals(0x2, bitfield.getBits(BA_62_4));
	}

	public static final BitArea BA_00_4 = new BitArea( 0, 4);
	public static final BitArea BA_28_4 = new BitArea(Bitfield32.SIZE - 4, 4);
	public static final BitArea BA_30_4 = new BitArea(Bitfield32.SIZE - 2, 4);
	public static final BitArea BA_60_4 = new BitArea(Bitfield64.SIZE - 4, 4);
	public static final BitArea BA_62_4 = new BitArea(Bitfield64.SIZE - 2, 4);

	/**
	 * @see Bitfield32#bitfield(BitArea[])
	 */
	@Test
	public void test32Bitfield() {
		assertEquals(0x00, Bitfield32.bitfield(new BitArea[] { } ));
		assertEquals(0x0f, Bitfield32.bitfield(new BitArea[] { BA_00_4 } ));
		assertEquals(0xf000000f, Bitfield32.bitfield(new BitArea[] { BA_00_4, BA_28_4, BA_30_4, BA_60_4, BA_62_4 } ));
	}

	/**
	 * @see Bitfield32#bitfield(BitArea[])
	 */
	@Test
	public void test64Bitfield() {
		assertEquals(0x00L, Bitfield64.bitfield(new BitArea[] { } ));
		assertEquals(0x0fL, Bitfield64.bitfield(new BitArea[] { BA_00_4 } ));
		assertEquals(0xf0000003f000000fL, Bitfield64.bitfield(new BitArea[] { BA_00_4, BA_28_4, BA_30_4, BA_60_4, BA_62_4 } ));
	}

	/**
	 * @see Bitfield32#bitmask(int, int)
	 * @see Bitfield32#bitmask(BitArea)
	 */
	@Test
	public void test32Bitmask() {
		assertEquals(0x00, Bitfield32.bitmask(4, 0));
		assertEquals(0x01, Bitfield32.bitmask(0, 1));
		assertEquals(0x0f, Bitfield32.bitmask(0, 4));
		assertEquals(0x0f, Bitfield32.bitmask(BA_00_4));
		assertEquals(0xf0000000, Bitfield32.bitmask(BA_28_4));
	}

	/**
	 * @see Bitfield64#bitmask(int, int)
	 * @see Bitfield64#bitmask(BitArea)
	 */
	@Test
	public void test64Bitmask() {
		assertEquals(0x00L, Bitfield64.bitmask(4, 0));
		assertEquals(0x01L, Bitfield64.bitmask(0, 1));
		assertEquals(0x0fL, Bitfield64.bitmask(0, 4));
		assertEquals(0x0fL, Bitfield64.bitmask(BA_00_4));
		assertEquals(0xf0000000L, Bitfield64.bitmask(BA_28_4));
		assertEquals(0xf000000000000000L, Bitfield64.bitmask(BA_60_4));
	}

	/**
	 * @see Bitfield32#normalizedBitmask(BitArea)
	 */
	@Test
	public void test32NormalizedBitmask() {
		assertEquals(0x0f, Bitfield32.normalizedBitmask(BA_00_4));
		assertEquals(0x0f, Bitfield32.normalizedBitmask(BA_28_4));
	}

	/**
	 * @see Bitfield64#normalizedBitmask(BitArea)
	 */
	@Test
	public void test64NormalizedBitmask() {
		assertEquals(0x0f, Bitfield64.normalizedBitmask(BA_00_4));
		assertEquals(0x0f, Bitfield64.normalizedBitmask(BA_28_4));
		assertEquals(0x0f, Bitfield64.normalizedBitmask(BA_60_4));
	}

	/**
	 * @see Bitfield32#setAll(int)
	 * @see Bitfield32#getAll()
	 */
	@Test
	public void test32setAll() {
		Bitfield32 bf = new Bitfield32();
		bf.setAll(0xaffeaffe);
		assertEquals(0xaffeaffe, bf.getAll());
	}

	/**
	 * @see Bitfield64#setAll(long)
	 * @see Bitfield64#getAll()
	 */
	@Test
	public void test64setAll() {
		Bitfield64 bf = new Bitfield64();
		bf.setAll(0xaffeaffeaffeaffeL);
		assertEquals(0xaffeaffeaffeaffeL, bf.getAll());
	}

	/**
	 * @see Bitfield32#equals(Object)
	 * @see Bitfield32#clone()
	 */
	@Test
	public void test32ObjectBasics() {
		GenericTest.checkObjectBasics(true, new Bitfield32(), new Bitfield32(0xff));
	}

	/**
	 * @see Bitfield64#equals(Object)
	 * @see Bitfield64#clone()
	 */
	@Test
	public void test64ObjectBasics() {
		GenericTest.checkObjectBasics(true, new Bitfield64(), new Bitfield64(0xff));
	}
}
