package de.ped.tools.test;

import static org.junit.Assert.*;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.junit.Test;

import de.ped.tools.MathUtil;

/**
 * Generic tests that all classes should run
 * 
 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class GenericTest {

	/**
	 * Tests if the tests in {@link #checkObjectBasics} do what they should.
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(false, Integer.valueOf(1));
		GenericTest.checkObjectBasics(false, Integer.valueOf(1), Integer.valueOf(2));
		GenericTest.checkObjectBasics(false, Integer.valueOf(1), Integer.valueOf(2), Integer.valueOf(3));
	}

	/**
	 * Checks all the generic concepts of Java objects or typical Object
	 * interfaces like {@link Object#equals(Object)}, {@link Object#hashCode()}, 
	 * {@link Comparable}, {@link Cloneable}, {@link Object#clone()}.
	 * <p>
	 * For each object handed in, the following tests are performed:
	 * <ul>
	 * <li>object is not <code>null</code> (precondition)
	 * <li>object is not equal to <code>null</code>
	 * <li>object is not equal to <code>Integer.getValue(0)</code>
	 * <li>object is equal to self
	 * <li>object is equal to clone and vice versa (iff isCloneEqual is <code>true</code>)
	 * <li>object compared to <code>null</code> is throwing a {@link NullPointerException}
	 * <li>object compared to self is 0
	 * <li>object compared to self is 0
	 * <li>object compared to clone and vice versa is 0 (independent of isCloneEqual)
	 * <li>hash code of object and clone is same
	 * </ul>
	 * 
	 * If more than one object is handed in, the folling tests are performed 
	 * for each non-identical combination of two of those objects:
	 * <ul>
	 * <li>the two objects are not equal
	 * <li>the object listed first is lesser than the other one.
	 * </ul>
	 * 
	 * @param isCloneEqual if clones are treated as equal.
	 * @param args the objects to test; if they are comparable, they must be in ascending order
	 */
	@SuppressWarnings({ "unchecked", "rawtypes" })
	public static void checkObjectBasics(boolean isCloneEqual, final Object ...args) {
		// Tests that can be performed on single objects in the list
		for (int i = 0; i < args.length; i++) {
			final Object obj = args[i];
			final String message = "" + i + ":";
			assertNotNull(message, obj);
			assertFalse(message, obj.equals(null));
			assertFalse(message, obj.equals(Integer.valueOf(0)));
			assertTrue(message, obj.equals(obj));
			Object clone = null;
			boolean wasHashCodeCalled = false;
			if (obj instanceof Cloneable) {
				// Check if clone() method is public, and if yes, call it. Check the result if equal to original.
				try {
					final Method m = obj.getClass().getDeclaredMethod("clone");
					if (Modifier.isPublic(m.getModifiers())) {
						clone = m.invoke(obj);
						assertEquals(message, isCloneEqual, obj.equals(clone));
						assertEquals(message, isCloneEqual, clone.equals(obj));
						if (isCloneEqual) {
							assertEquals(message, obj.hashCode(), clone.hashCode());
							wasHashCodeCalled = true;
						}
					}
				} catch (SecurityException e) {
					// Nothing to do
				} catch (NoSuchMethodException e) {
					// Nothing to do
				} catch (IllegalArgumentException e) {
					// Nothing to do
				} catch (IllegalAccessException e) {
					// Nothing to do
				} catch (InvocationTargetException e) {
					// Nothing to do
				}
			}
			if (obj instanceof Comparable) {
				final Comparable ca = (Comparable) obj;
				try {
					assertNotEquals(message, 0, ca.compareTo(null));
					fail();
				} catch(final NullPointerException ex) {
					// Expected behavior, nothing to do
				}
				assertEquals(message, 0, ca.compareTo(ca));
				if (null != clone) {
					assertEquals(message, 0, ca.compareTo(clone));
					assertEquals(message, 0, ((Comparable) clone).compareTo(ca));
				}
			}
			if (!wasHashCodeCalled) {
				obj.hashCode(); // We just call it, but have nothing to test for
			}
		}
		// Tests that deal with the relationships between objects in the list
		for (int i = 0; i < args.length; i++) {
			for (int j = 0; j < args.length; j++) {
				if (i != j) {
					final String message = "" + i + "," + j + ":";
					assertFalse(message, args[i].equals(args[j]));
					if ((args[i] instanceof Comparable) && (args[j] instanceof Comparable)) {
						final Comparable ca = (Comparable) args[i];
						final Comparable cb = (Comparable) args[j];
						final int cab = ca.compareTo(cb);
						final int cba = cb.compareTo(ca);
						assertNotEquals(message, 0, cab);
						assertNotEquals(message, 0, cba);
						assertNotEquals(message, MathUtil.sgn(cab), MathUtil.sgn(cba));
						assertEquals(message, i < j, cab < cba);
					}
				}
			}
		}
	}
}
