package de.ped.tools.test;

import static org.junit.Assert.*;

import org.junit.Test;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.Marshaller.ConstantMarshallableDemarshallingException;
import de.ped.tools.TextFormatter;

/**
 * Unit tests for the behavior of the {@link Marshaller} class.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MarshallerTest {

	/**
	 * A {@link Marshallable} class that should not be demarshalled.
	 * 
	 * @author Copyright (c) 2019 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ConstantMarshallable implements Marshallable {
		@Override
		public void marshal(Marshaller marshaller) {
			// Nothing to do
		}
		@Override
		public Marshallable demarshal(Marshaller marshaller) {
			throw Marshaller.createConstantMarshallableDemarshallingException();
		}
		
	}

	/**
	 * Checks the assertion that there is no more data to read from the given marshaller.
	 * @param d the marshaller to read from
	 */
	public static void assertEOF(final Marshaller d) {
		try {
			d.readByte();
			fail();
		} catch (final Exception ex) {
			// Nothing to do
		}
	}

	/**
	 * @see Marshaller#readByte()
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testDemarshalByte_EmptyBuffer() {
		final Marshaller d = new Marshaller();
		d.readByte();
	}

	/**
	 * @see Marshaller#readByte()
	 * @see Marshaller#writeByte(byte)
	 * @see Marshaller#pack()
	 */
	@Test
	public void testMarshalByte_FullRange() {
		final Marshaller m = new Marshaller();
		for (int i = -128; i < 127; i++) {
			m.writeByte((byte)i);
		}
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		for (int i = -128; i < 127; i++) {
			final byte b = d.readByte();
			assertEquals((byte)i, b);
		}
		assertEOF(d);
	}

	/**
	 * @see Marshaller#readInt()
	 * @see Marshaller#writeInt(int)
	 * @see Marshaller#pack()
	 */
	@Test
	public void testMarshalInt_Sample() {
		final Marshaller m = new Marshaller();
		for (int i = -128; i < 127; i++) {
			m.writeInt(i);
		}
		m.writeInt(0x80808080);
		m.writeInt(0xffffffff);
		m.writeInt(0xaa55aa55);
		m.writeInt(0x55aa55aa);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		for (int i = -128; i < 127; i++) {
			final int b = d.readInt();
			assertEquals(i, b);
		}
		assertEquals(0x80808080, d.readInt());
		assertEquals(0xffffffff, d.readInt());
		assertEquals(0xaa55aa55, d.readInt());
		assertEquals(0x55aa55aa, d.readInt());
		assertEOF(d);
	}

	private static final double EPSILON = 0.0000001;
	
	/**
	 * A little class to be marshalled and demarshalled.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private class TestObject implements Marshallable {

		@Override
		public void marshal(Marshaller marshaller) {
			// Nothing to do
		}

		@Override
		public Marshallable demarshal(Marshaller marshaller) {
			return this;
		}
	}

	/**
	 * @see Marshaller#Marshaller()
	 * @see Marshaller#Marshaller(byte[])
	 * @see Marshaller#writeBoolean(boolean)
	 * @see Marshaller#writeByte(byte)
	 * @see Marshaller#writeChar(char)
	 * @see Marshaller#writeShort(short)
	 * @see Marshaller#writeInt(int)
	 * @see Marshaller#writeLong(long)
	 * @see Marshaller#writeFloat(float)
	 * @see Marshaller#writeDouble(double)
	 * @see Marshaller#writeVariableLengthBooleanArray(boolean[])
	 * @see Marshaller#writeNullOrMarshallable(Marshallable)
	 * @see Marshaller#readBoolean()
	 * @see Marshaller#readByte()
	 * @see Marshaller#readChar()
	 * @see Marshaller#readShort()
	 * @see Marshaller#readInt()
	 * @see Marshaller#readLong()
	 * @see Marshaller#readFloat()
	 * @see Marshaller#readDouble()
	 * @see Marshaller#readVariableLengthBooleanArray()
	 * @see Marshaller#readNullOrMarshallable(Marshallable)
	 * @see Marshaller#clone()
	 * @see Marshaller#pack()
	 */
	@Test
	public void testMarshalBasicTypesMix() {
		final Marshaller m = new Marshaller();
		// Write some mixed data
		m.writeBoolean(true);
		m.writeByte((byte)0xaa);
		m.writeChar('a');
		short s = 1025;
		m.writeShort(s);
		m.writeInt(512);
		m.writeLong(555555L);
		m.writeFloat(0.5F);
		m.writeDouble(Math.PI);
		m.writeVariableLengthBooleanArray((boolean[]) null);
		m.writeVariableLengthBooleanArray(new boolean[] { true, false });
		m.writeVariableLengthIntArray((int[]) null);
		m.writeVariableLengthIntArray(new int[] { 42, -4711 });
		m.writeNullOrMarshallable(null);
		m.writeNullOrMarshallable(new TestObject());
		// Test clone
		final Marshaller m2 = (Marshaller) m.clone();
		final byte[] pack = m2.pack();
		final Marshaller d = new Marshaller(pack);
		// Read the mixed data
		assertEquals(true, d.readBoolean());
		assertEquals((byte)0xaa, d.readByte());
		assertEquals('a', d.readChar());
		assertEquals(1025, d.readShort());
		assertEquals(512, d.readInt());
		assertEquals(555555L, d.readLong());
		assertEquals(0.5F, d.readFloat(), EPSILON);
		assertEquals(Math.PI, d.readDouble(), EPSILON);
		assertNull(d.readVariableLengthBooleanArray());
		assertArrayEquals(new boolean[] { true, false }, d.readVariableLengthBooleanArray());
		assertNull(d.readVariableLengthIntArray());
		assertArrayEquals(new int[] { 42, -4711 }, d.readVariableLengthIntArray());
		assertNull(d.readNullOrMarshallable(new TestObject()));
		TestObject to = new TestObject();
		assertEquals(to, d.readNullOrMarshallable(to));
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeFixedLengthBytesArray(byte[])
	 */
	@Test(expected=NullPointerException.class)
	public void testMarshalBytesArrayFixedLength_Null() {
		final Marshaller m = new Marshaller();
		m.writeFixedLengthBytesArray((byte[])null);
	}

	/**
	 * @see Marshaller#writeFixedLengthBytesArray(byte[])
	 * @see Marshaller#readFixedLengthBytesArray(int)
	 */
	@Test
	public void testMarshalBytesArrayFixedLength_Example() {
		final Marshaller m = new Marshaller();
		final byte[] data = new byte[20];
		for (int i = 0; i < data.length; i++) {
			data[i] = (byte)(7 * i - 100);
		}
		m.writeFixedLengthBytesArray(data);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		final byte[] result = d.readFixedLengthBytesArray(data.length);
		assertArrayEquals(data, result);
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeVariableLengthByteArray(byte[])
	 * @see Marshaller#readVariableLengthBytesArray()
	 */
	@Test
	public void testMarshalBytesArrayVariableLength_Null() {
		final Marshaller m = new Marshaller();
		m.writeVariableLengthByteArray((byte[])null);
		final Marshaller d = new Marshaller(m.pack());
		assertNull(d.readVariableLengthBytesArray());
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeVariableLengthByteArray(byte[])
	 * @see Marshaller#readVariableLengthBytesArray()
	 */
	@Test
	public void testMarshalBytesArrayVariableLength_Example() {
		final Marshaller m = new Marshaller();
		final byte[] data = new byte[20];
		for (int i = 0; i < data.length; i++) {
			data[i] = (byte)(7 * i - 100);
		}
		m.writeVariableLengthByteArray(data);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		final byte[] result = d.readVariableLengthBytesArray();
		assertArrayEquals(data, result);
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeString(String)
	 * @see Marshaller#readString()
	 */
	@Test
	public void testMarshalString_Null() {
		final Marshaller m = new Marshaller();
		m.writeString((String)null);
		final Marshaller d = new Marshaller(m.pack());
		assertNull(d.readString());
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeString(String)
	 * @see Marshaller#readString()
	 */
	@Test
	public void testMarshalString_Example() {
		final String str = "Hallöchen Test";
		final Marshaller m = new Marshaller();
		m.writeString(str);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		final String result = d.readString();
		assertEquals(str, result);
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeStringArray(String[])
	 * @see Marshaller#readStringArray()
	 */
	@Test
	public void testMarshalStringArray_Null() {
		final Marshaller m = new Marshaller();
		m.writeStringArray((String[])null);
		final Marshaller d = new Marshaller(m.pack());
		assertNull(d.readStringArray());
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeStringArray(String[])
	 * @see Marshaller#readStringArray()
	 */
	@Test
	public void testMarshalStringArray_SmallExample() {
		final String[] str = new String[]{"Hallöchen", "Test"};
		final Marshaller m = new Marshaller();
		m.writeStringArray(str);
		final byte[] pack = m.pack();
		final Marshaller d = new Marshaller(pack);
		final String[] result = d.readStringArray();
		assertArrayEquals(str, result);
		assertEOF(d);
	}

	/**
	 * @see Marshaller#writeStringArray(String[])
	 * @see Marshaller#readStringArray()
	 */
	@Test
	public void testMarshalStringArray_LargeExample() {
		final String ex = 
			"Client TestClientXYZ connected to Server Troff CPU\n" +
			"Client TestClientXYZ joined group TestGroupXYZ\n" + 
			"Client TestClientABC joined group TestGroupXYZ\n" + 
			"Client TestClientABC has left group TestGroupXYZ\n" +
			"Client TestClientXYZ has left group TestGroupXYZ\n";
		final String example = TextFormatter.repeat(ex, 10); // We want to exceed the chunk size of the Marshaller
		final String[] str = example.split("\n");
		final Marshaller m = new Marshaller();
		m.writeStringArray(str);
		final byte[] pack = m.pack();
		final int expected = 4 // int: length of array 
				+ 5*10*(4-1)   // per newline one int, but minus newline char
				+ example.length();  // all the other chars
		assertEquals(expected, pack.length);
		final Marshaller d = new Marshaller(pack);
		final String[] result = d.readStringArray();
		assertArrayEquals(str, result);
		assertEOF(d);
	}

	/**
	 * @see Marshaller#createConstantMarshallableDemarshallingException()
	 */
	@Test(expected=ConstantMarshallableDemarshallingException.class)
	public void testCreateConstantMarshallableDemarshallingException() {
		final Marshaller m = new Marshaller();
		(new ConstantMarshallable()).demarshal(m);
	}

	/**
	 * @see Marshaller#equals(Object)
	 * @see Marshaller#clone()
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(false, new Marshaller(), new Marshaller());
	}
}
