package de.ped.tools.test;

import static org.junit.Assert.*;

import java.util.Arrays;
import java.util.Random;

import org.junit.Test;

import de.ped.tools.MathUtil;

/**
 * Unit tests for the {@link MathUtil} class.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MathUtilTest {

	private static final float F_EPS = 1E-6F;
	private static final float F_DELTA = 1E-6F;

	private static final double D_EPS = 1E-15;
	private static final double D_DELTA = 1E-10;
	
	private static final int HOWOFTEN = 10;

	/**
	 * @see MathUtil#random(int)
	 */
	@Test
	public void testRandomInt() {
		assertEquals(0, MathUtil.random(0));
		for (int border : new int[] { 1, 6, 20 }) {
			for (int i = 0; i < HOWOFTEN; i++) {
				final int pos = MathUtil.random(border);
				assertTrue(0 <= pos);
				assertTrue(pos < border);
				final int neg = MathUtil.random(-border);
				assertTrue(0 >= neg);
				assertTrue(neg > -border);
			}
		}
	}

	/**
	 * @see MathUtil#random(long)
	 */
	@Test
	public void testRandomLong() {
		assertEquals(0L, MathUtil.random(0L));
		for (long border : new long[] { 1L, 6L, 20L }) {
			for (int i = 0; i < HOWOFTEN; i++) {
				final long pos = MathUtil.random(border);
				assertTrue(0 <= pos);
				assertTrue(pos < border);
				final long neg = MathUtil.random(-border);
				assertTrue(0 >= neg);
				assertTrue(neg > -border);
			}
		}
	}

	/**
	 * @see MathUtil#random(float)
	 */
	@Test
	public void testRandomFloat() {
		assertEquals(0.0, MathUtil.random(0.0), D_EPS);
		for (float border : new float[] { 1.0F, 6.0F, 20.0F, 0.1F }) {
			for (int i = 0; i < HOWOFTEN; i++) {
				final float pos = MathUtil.random(border);
				assertTrue(0 <= pos);
				assertTrue(pos < border);
				final float neg = MathUtil.random(-border);
				assertTrue(0 >= neg);
				assertTrue(neg > -border);
			}
		}
	}

	/**
	 * @see MathUtil#random(double)
	 */
	@Test
	public void testRandomDouble() {
		assertEquals(0.0, MathUtil.random(0.0), D_EPS);
		for (double border : new double[] { 1.0, 6.0, 20.0, 0.1, Math.PI }) {
			for (int i = 0; i < HOWOFTEN; i++) {
				final double pos = MathUtil.random(border);
				assertTrue(0 <= pos);
				assertTrue(pos < border);
				final double neg = MathUtil.random(-border);
				assertTrue(0 >= neg);
				assertTrue(neg > -border);
			}
		}
	}

	/**
	 * @see MathUtil#randomDistributed(int[])
	 */
	@Test(expected=NullPointerException.class)
	public void testRandomDistributedInt_Null() {
		final int[] arr = null;
		MathUtil.randomDistributed(arr);
	}

	/**
	 * @see MathUtil#randomDistributed(long[])
	 */
	@Test(expected=NullPointerException.class)
	public void testRandomDistributedLong_Null() {
		final long[] arr = null;
		MathUtil.randomDistributed(arr);
	}

	/**
	 * @see MathUtil#randomDistributed(float[])
	 */
	@Test(expected=NullPointerException.class)
	public void testRandomDistributedFloat_Null() {
		final float[] arr = null;
		MathUtil.randomDistributed(arr);
	}

	/**
	 * @see MathUtil#randomDistributed(double[])
	 */
	@Test(expected=NullPointerException.class)
	public void testRandomDistributedDouble_Null() {
		final double[] arr = null;
		MathUtil.randomDistributed(arr);
	}

	/**
	 * @see MathUtil#randomDistributed(int[])
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testRandomDistributedInt_Empty() {
		MathUtil.randomDistributed(new int[]{0});
	}

	/**
	 * @see MathUtil#randomDistributed(long[])
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testRandomDistributedLong_Empty() {
		MathUtil.randomDistributed(new long[]{0L});
	}

	/**
	 * @see MathUtil#randomDistributed(float[])
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testRandomDistributedFloat_Empty() {
		MathUtil.randomDistributed(new float[]{0.0F});
	}

	/**
	 * @see MathUtil#randomDistributed(double[])
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testRandomDistributedDouble_Empty() {
		MathUtil.randomDistributed(new double[]{0.0});
	}

	/**
	 * @see MathUtil#randomDistributed(int[])
	 */
	@Test
	public void testRandomDistributedInt_Ok() {
		assertEquals(0, MathUtil.randomDistributed(new int[]{1}));
		assertEquals(0, MathUtil.randomDistributed(new int[]{1000}));
		assertEquals(0, MathUtil.randomDistributed(new int[]{1000, 0}));
		assertEquals(1, MathUtil.randomDistributed(new int[]{0, 1000}));
		assertEquals(1, MathUtil.randomDistributed(new int[]{-1000, 1}));
	}

	/**
	 * @see MathUtil#randomDistributed(int[])
	 */
	@Test
	public void testRandomDistributedLong_Ok() {
		assertEquals(0, MathUtil.randomDistributed(new long[]{1L}));
		assertEquals(0, MathUtil.randomDistributed(new long[]{1000L}));
		assertEquals(0, MathUtil.randomDistributed(new long[]{1000L, 0L}));
		assertEquals(1, MathUtil.randomDistributed(new long[]{0L, 1000L}));
		assertEquals(1, MathUtil.randomDistributed(new long[]{-1000L, 1L}));
	}

	/**
	 * @see MathUtil#randomDistributed(double[])
	 */
	@Test
	public void testRandomDistributedFloat_Ok() {
		assertEquals(0, MathUtil.randomDistributed(new float[]{1.0F}));
		assertEquals(0, MathUtil.randomDistributed(new float[]{0.1F}));
		assertEquals(0, MathUtil.randomDistributed(new float[]{0.1F, 0.0F}));
		assertEquals(1, MathUtil.randomDistributed(new float[]{0.0F, 0.1F}));
		assertEquals(1, MathUtil.randomDistributed(new float[]{-0.1F, 1.0F}));
	}

	/**
	 * @see MathUtil#randomDistributed(double[])
	 */
	@Test
	public void testRandomDistributedDouble_Ok() {
		assertEquals(0, MathUtil.randomDistributed(new double[]{1.0}));
		assertEquals(0, MathUtil.randomDistributed(new double[]{Math.PI}));
		assertEquals(0, MathUtil.randomDistributed(new double[]{Math.PI, 0.0}));
		assertEquals(1, MathUtil.randomDistributed(new double[]{0.0, Math.PI}));
		assertEquals(1, MathUtil.randomDistributed(new double[]{-Math.PI, 1.0}));
	}

	/**
	 * @see MathUtil#divisio(int, int)
	 */
	@Test(expected=ArithmeticException.class)
	public void testDivisioInt_fail() {
		MathUtil.divisio(0, 0);
	}

	/**
	 * @see MathUtil#divisio(float, float)
	 */
	@Test(expected=ArithmeticException.class)
	public void testDivisioFloat_fail() {
		MathUtil.divisio(0.0F, 0.0F);
	}

	/**
	 * @see MathUtil#divisio(double, double)
	 */
	@Test(expected=ArithmeticException.class)
	public void testDivisioDouble_fail() {
		MathUtil.divisio(0.0, 0.0);
	}

	/**
	 * @see MathUtil#divisio(int, int, int)
	 * @see MathUtil#divisio(double, double, double)
	 */
	@Test
	public void testModuloDivisio() {
		assertEquals(0, MathUtil.modulo(0, 0));
		assertEquals(0L, MathUtil.modulo(0L, 0L));
		assertEquals(0.0F, MathUtil.modulo(0.0F, 0.0F), F_EPS);
		assertEquals(0.0, MathUtil.modulo(0.0, 0.0), D_EPS);

		testModuloDivisioWithMinus( 0, 0,   0, 1, 0);
		testModuloDivisioWithMinus( 0, 0,   0, 5, 0); //this behaves like the normal division operation (0 / 5)
		testModuloDivisioWithMinus( 1, 0,   1, 5, 0);
		testModuloDivisioWithMinus( 4, 0,   4, 5, 0);
		testModuloDivisioWithMinus( 0, 1,   5, 5, 0);
		testModuloDivisioWithMinus( 1, 1,   6, 5, 0);
		testModuloDivisioWithMinus( 4, 1,   9, 5, 0);
		testModuloDivisioWithMinus( 0, 2,  10, 5, 0);
		testModuloDivisioWithMinus( 1, 2,  11, 5, 0);
		testModuloDivisioWithMinus( 2, 2,  12, 5, 0); //this behaves like the normal division operation (12 / 5)
		testModuloDivisioWithMinus( 4,-1,  -1, 5, 0); //-1 is the last number of the smallest "negative" circle
		testModuloDivisioWithMinus( 1,-1,  -4, 5, 0);
		testModuloDivisioWithMinus( 0,-1,  -5, 5, 0); //-5 is the first number of the smallest "negative" circle
		testModuloDivisioWithMinus( 4,-2,  -6, 5, 0); //-6 is the last number of the second smallest "negative" circle
		testModuloDivisioWithMinus( 1,-2,  -9, 5, 0);
		testModuloDivisioWithMinus( 0,-2, -10, 5, 0);
		testModuloDivisioWithMinus( 4,-3, -11, 5, 0);
		testModuloDivisioWithMinus( 3,-3, -12, 5, 0);
		testModuloDivisioWithMinus( 1, 0,   1, 1, 1);
		testModuloDivisioWithMinus( 1, 0,   1,12, 1);
		testModuloDivisioWithMinus(12, 0,  12,12, 1);
		testModuloDivisioWithMinus( 1, 1,  13,12, 1);
		testModuloDivisioWithMinus( 2, 1,  14,12, 1);
		testModuloDivisioWithMinus(12,-1,   0,12, 1);
		testModuloDivisioWithMinus( 0, 0,   0, 3,-1);
		testModuloDivisioWithMinus(-1, 1,   2, 3,-1);
		// Tests for float and double precision
		testModuloDivisioFractionsWithMinus(1.0, 4.0, 11.0, 2.5, 0.0); //11.0 mod 2.5 = (11.0 - 4 * 2.5) mod 2.5 = (11.0 - 10.0) mod 2.5 = 1.0
		testModuloDivisioFractionsWithMinus(0.0, 0.0, 0.0, Math.PI, 0.0); // Trivial case
		testModuloDivisioFractionsWithMinus(Math.PI - F_DELTA, 0.0, Math.PI - F_DELTA, Math.PI, 0.0); // Shortly before the first turnover
		testModuloDivisioFractionsWithMinus(0.0, 1.0, Math.PI, Math.PI, 0.0); // The first turnover
		testModuloDivisioFractionsWithMinus(Math.PI - F_DELTA, -1.0, -F_DELTA, Math.PI, 0.0); //Slightly negative val
		testModuloDivisioFractionsWithMinus(0.0, -1.0, -Math.PI, Math.PI, 0.0); //The lowest number of the first negative circle
		testModuloDivisioFractionsWithMinus(Math.PI - F_DELTA, -2.0, -Math.PI - F_DELTA, Math.PI, 0.0); //Nearly the highest number of the second negative circle
		// Tests for double precision
		testModuloDivisioDoubleWithMinus(Math.PI - D_DELTA, 0.0, Math.PI - D_DELTA, Math.PI, 0.0); // Shortly before the first turnover
		testModuloDivisioDoubleWithMinus(Math.PI - D_DELTA, -1.0, -D_DELTA, Math.PI, 0.0); //Slightly negative val
		testModuloDivisioDoubleWithMinus(Math.PI - D_DELTA, -2.0, -Math.PI - D_DELTA, Math.PI, 0.0); //Nearly the highest number of the second negative circle
	}

	private void testModuloDivisioWithMinus(final int modResult, final int divResult, final int val, final int mod, final int offset) {
		testModuloDivisioIntWithMinus(modResult, divResult,  val, mod, offset);
		testModuloDivisioFractionsWithMinus(modResult, divResult, val, mod, offset);
	}

	private void testModuloDivisioIntWithMinus(final int modResult, final int divResult, final int val, final int mod, final int offset) {
		testModuloDivisioInt( modResult,  divResult,  val,  mod, offset);
		testModuloDivisioInt(-modResult,  divResult, -val, -mod, -offset);
		testModuloDivisioLong( modResult,  divResult,  val,  mod, offset);
		testModuloDivisioLong(-modResult,  divResult, -val, -mod, -offset);
	}

	private void testModuloDivisioInt(final int modResult, final int divResult, final int val, final int mod, final int offset) {
		assertEquals(val, divResult * mod + modResult);
		final int m = MathUtil.modulo(val, mod, offset);
		final int d = MathUtil.divisio(val, mod, offset);
		assertEquals(modResult, m);
		assertEquals(divResult, d);
		assertEquals(m, MathUtil.modulo(val - offset, mod, 0) + offset);
		assertEquals(d, MathUtil.divisio(val - offset, mod, 0));
	}

	private void testModuloDivisioLong(final long modResult, final long divResult, final long val, final long mod, final long offset) {
		assertEquals(val, divResult * mod + modResult);
		final long m = MathUtil.modulo(val, mod, offset);
		final long d = MathUtil.divisio(val, mod, offset);
		assertEquals(modResult, m);
		assertEquals(divResult, d);
		assertEquals(m, MathUtil.modulo(val - offset, mod, 0) + offset);
		assertEquals(d, MathUtil.divisio(val - offset, mod, 0));
	}

	private void testModuloDivisioFractionsWithMinus(final double modResult, final double divResult, final double val, final double mod, final double offset) {
		testModuloDivisioDoubleWithMinus( modResult,  divResult,  val,  mod, offset);
		testModuloDivisioFloatWithMinus((float) modResult, (float) divResult, (float) val, (float) mod, (float) offset);
	}

	private void testModuloDivisioDoubleWithMinus(final double modResult, final double divResult, final double val, final double mod, final double offset) {
		testModuloDivisioDouble( modResult,  divResult,  val,  mod, offset);
		testModuloDivisioDouble(-modResult,  divResult, -val, -mod, -offset);
	}

	private void testModuloDivisioDouble(final double modResult, final double divResult, final double val, final double mod, final double offset) {
		assertEquals(val, divResult * mod + modResult, D_EPS);
		final double m = MathUtil.modulo(val, mod, offset);
		final double d = MathUtil.divisio(val, mod, offset);
		assertEquals(modResult, m, D_EPS);
		assertEquals(divResult, d, D_EPS);
		assertEquals(m, MathUtil.modulo(val - offset, mod, 0) + offset, D_EPS);
		assertEquals(d, MathUtil.divisio(val - offset, mod, 0), D_EPS);
	}

	private void testModuloDivisioFloatWithMinus(final float modResult, final float  divResult, final float  val, final float  mod, final float  offset) {
		testModuloDivisioFloat( modResult,  divResult,  val,  mod, offset);
		testModuloDivisioFloat(-modResult,  divResult, -val, -mod, -offset);
	}

	private void testModuloDivisioFloat(final float modResult, final float divResult, final float val, final float mod, final float offset) {
		assertEquals(val, divResult * mod + modResult, F_EPS);
		final float m = MathUtil.modulo(val, mod, offset);
		final float d = MathUtil.divisio(val, mod, offset);
		assertEquals(modResult, m, F_EPS);
		assertEquals(divResult, d, F_EPS);
		assertEquals(m, MathUtil.modulo(val - offset, mod, 0) + offset, F_EPS);
		assertEquals(d, MathUtil.divisio(val - offset, mod, 0), F_EPS);
	}

	/**
	 * @see MathUtil#round(int, int)
	 */
	@Test
	public void testRoundInt() {
		assertEquals( 12, MathUtil.round( 13,  4));
		assertEquals(-12, MathUtil.round(-13,  4));
		assertEquals( 12, MathUtil.round( 13, -4));
		assertEquals(-12, MathUtil.round(-13, -4));
	}

	/**
	 * @see MathUtil#round(float, float)
	 */
	@Test
	public void testRoundFloat() {
		assertEquals( 0.75F, MathUtil.round( 0.8F,  0.25F), F_EPS);
		assertEquals(-0.75F, MathUtil.round(-0.8F,  0.25F), F_EPS);
		assertEquals( 0.75F, MathUtil.round( 0.8F, -0.25F), F_EPS);
		assertEquals(-0.75F, MathUtil.round(-0.8F, -0.25F), F_EPS);
	}

	/**
	 * @see MathUtil#round(double, double)
	 */
	@Test
	public void testRoundDouble() {
		assertEquals( 0.75, MathUtil.round( 0.8,  0.25), D_EPS);
		assertEquals(-0.75, MathUtil.round(-0.8,  0.25), D_EPS);
		assertEquals( 0.75, MathUtil.round( 0.8, -0.25), D_EPS);
		assertEquals(-0.75, MathUtil.round(-0.8, -0.25), D_EPS);
	}

	private void testBound(final int result, final int val, final int min, final int max) {
		assertEquals(         result, MathUtil.bound(         val,          min,          max));
		assertEquals((long)   result, MathUtil.bound((long)   val, (long)   min, (long)   max));
		assertEquals((float)  result, MathUtil.bound((float)  val, (float)  min, (float)  max), F_EPS);
		assertEquals((double) result, MathUtil.bound((double) val, (double) min, (double) max), D_EPS);
	}

	/**
	 * @see MathUtil#bound(int, int, int)
	 * @see MathUtil#bound(long, long, long)
	 * @see MathUtil#bound(float, float, float)
	 * @see MathUtil#bound(double, double, double)
	 */
	@Test
	public void testBound() {
		testBound(0, 0, 0, 0);
		testBound(0, 1, 0, 0);
		testBound(1, 0, 1, 5);
		testBound(1, 1, 1, 5);
		testBound(3, 3, 1, 5);
		testBound(5, 5, 1, 5);
		testBound(5, 6, 1, 5);
	}	

	private void testSgn(final int result, final int val) {
		assertEquals(result, MathUtil.sgn(val));
		assertEquals(result, MathUtil.sgn((long) val));
		assertEquals(result, MathUtil.sgn((float) val));
		assertEquals(result, MathUtil.sgn((double) val));
	}

	/**
	 * @see MathUtil#sgn(int)
	 * @see MathUtil#sgn(long)
	 * @see MathUtil#sgn(double)
	 * @see MathUtil#sgn(float)
	 */
	@Test
	public void testSgn() {
		testSgn( 0,  0);
		testSgn( 1,  1);
		testSgn(-1, -1);
		testSgn( 1,  2);
		testSgn(-1, -2);
		assertEquals( 1, MathUtil.sgn( F_EPS));
		assertEquals(-1, MathUtil.sgn(-F_EPS));
		assertEquals( 1, MathUtil.sgn( D_EPS));
		assertEquals(-1, MathUtil.sgn(-D_EPS));
	}

	/**
	 * @see MathUtil#count(int[])
	 */
	@Test
	public void testCount() {
		assertEquals(0, MathUtil.count(new int[] { } ));
		assertEquals(0, MathUtil.count(new int[] { 0 } ));
		assertEquals(1, MathUtil.count(new int[] { 5 } ));
	}

	/**
	 * @see MathUtil#shuffle(int[], java.util.Random)
	 * @see MathUtil#shuffle(Object[], Random)
	 * @see MathUtil#shuffle(java.util.List, Random)
	 */
	@Test
	public void testShuffle() {
		final Random random = new Random();
		MathUtil.shuffle(new int[] { 1, 2, 3, 4, 5, }, random);
		final Integer[] arr = new Integer[] { 
				Integer.valueOf(1), Integer.valueOf(2), Integer.valueOf(3),
				Integer.valueOf(4), Integer.valueOf(5), };
		MathUtil.shuffle(arr, random);
		MathUtil.shuffle(Arrays.asList(arr), random);
	}
}
