package de.ped.tools.test;

import static org.junit.Assert.*;

import org.junit.Test;

import de.ped.tools.MathMatrix;
import de.ped.tools.MathVector;

/**
 * Unit tests for the {@link MathVector} and {@link MathMatrix} classes.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class MathVectorTest {

	public static final double EPS = 1E-15;

	private void testCoord2D(
			final double x, 
			final double y, 
			final double angle, 
			final double radius, 
			final boolean isNullv, 
			final MathVector coord) {
		assertEquals(x, coord.get(0), EPS);
		assertEquals(y, coord.get(1), EPS);
		assertEquals(isNullv, coord.isNullVector(EPS));
		assertEquals(angle, coord.getAngle(MathVector.X, MathVector.Y), EPS);
		assertEquals(radius, coord.getRadius(), EPS);
	}

	/**
	 * @see MathVector#create(double...)
	 * @see MathVector#MathVector(int)
	 * @see MathVector#MathVector(MathVector)
	 * @see MathVector#get(int)
	 * @see MathVector#isNullVector(double)
	 * @see MathVector#getAngle(int, int)
	 * @see MathVector#getRadius()
	 * @see MathVector#setAll(double[])
	 * @see MathVector#addSelf(MathVector)
	 * @see MathVector#normSelf()
	 * @see MathVector#norm()
	 * @see MathVector#rotateSelfQuarter(int, int, int)
	 * @see MathVector#clone()
	 * @see MathVector#equals(Object)
	 * @see MathVector#compareTo(MathVector)
	 */
	@Test
	public void testCoord2D() {
		final double sqrt2 = Math.sqrt(2);
		final MathVector a = new MathVector(2);
		testCoord2D(0, 0, 0, 0, true, a);
		a.setAllFrom(1, 0);
		testCoord2D(1, 0, Math.PI / 2.0, 1, false, a);
		final MathVector b     = MathVector.create(0, 1);
		testCoord2D(0, 1, 0.0, 1, false, b);
		final MathVector c     = new MathVector(a);
		testCoord2D(1, 0, Math.PI / 2.0, 1, false, a);
		c.addSelf(b);
		testCoord2D(1, 1, Math.PI / 4.0, sqrt2, false, c);
		MathVector d = c.norm();
		c.normSelf();
		assertTrue(c.equals(d));
		assertEquals(0, c.compareTo(d));
		testCoord2D(sqrt2 / 2, sqrt2 / 2, Math.PI / 4.0, 1.0, false, c);
		b.rotateSelfQuarter(MathVector.X, MathVector.Y, 1);
		testCoord2D(1, 0, Math.PI / 2.0, 1, false, a);
		c.rotateSelfQuarter(MathVector.X, MathVector.Y, 1);
		testCoord2D(-sqrt2 / 2, sqrt2 / 2, -Math.PI / 4.0, 1.0, false, c);
		assertEquals(c, c.clone());
	}

	private void compareMatrixWithRef(final MathMatrix matrix, final double[][] ref) {
		assertEquals(ref.length, matrix.getLength(MathMatrix.X));
		assertEquals(ref[0].length, matrix.getLength(MathMatrix.Y));
		for (int x = 0; x < matrix.getLength(MathMatrix.X); x++) {
			for (int y = 0; y < matrix.getLength(MathMatrix.Y); y++) {
				assertEquals(ref[x][y], matrix.get(x, y), EPS);
			}
		}
	}

	/**
	 * @see MathMatrix#zero(int, int)
	 * @see MathMatrix#id(int)
	 * @see MathMatrix#MathMatrix(int, int)
	 * @see MathMatrix#MathMatrix(MathMatrix)
	 * @see MathMatrix#getLength(int)
	 * @see MathMatrix#get(int, int)
	 * @see MathMatrix#getAll()
	 * @see MathMatrix#set(int, int, double)
	 * @see MathMatrix#transpose()
	 * @see MathMatrix#mult(MathMatrix)
	 * @see MathMatrix#multSelf(double)
	 */
	@Test
	public void testMatrixBasics() {
		// Test zero matrix creator
		final MathMatrix zero = MathMatrix.zero(3, 3);
		compareMatrixWithRef(zero, new double[][] { { 0.0, 0.0, 0.0 }, { 0.0, 0.0, 0.0 }, { 0.0, 0.0, 0.0 } });
		// Test unit matrix creator
		final MathMatrix unit = MathMatrix.id(3);
		compareMatrixWithRef(unit, new double[][] { { 1.0, 0.0, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } });
		// Test copy constructor
		final MathMatrix a = new MathMatrix(unit);
		compareMatrixWithRef(a, new double[][] { { 1.0, 0.0, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } });
		compareMatrixWithRef(a, unit.getAll());
		// Test if copy constructor really does deep copy
		a.set(0, 1, 0.5);
		compareMatrixWithRef(unit, new double[][] { { 1.0, 0.0, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } });
		compareMatrixWithRef(a, new double[][] { { 1.0, 0.5, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } });
		// Test transpose
		final MathMatrix b = a.transpose();
		assertEquals(0.5, b.get(1, 0), EPS);
		compareMatrixWithRef(a, new double[][] { { 1.0, 0.5, 0.0 }, { 0.0, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } });
		compareMatrixWithRef(b, new double[][] { { 1.0, 0.0, 0.0 }, { 0.5, 1.0, 0.0 }, { 0.0, 0.0, 1.0 } });
		// Some test with non-quadratic matrices.
		final MathMatrix c = new MathMatrix(3, 2);
		c.set(0, 0, 1.0);
		c.set(1, 0, 2.0);
		c.set(2, 0, 3.0);
		c.set(0, 1, 4.0);
		c.set(1, 1, 5.0);
		c.set(2, 1, 6.0);
		compareMatrixWithRef(c, new double[][] { { 1.0, 4.0 }, { 2.0, 5.0 }, { 3.0, 6.0 } });
		final MathMatrix d = new MathMatrix(2, 3);
		d.set(0, 0, 6.0);
		d.set(1, 0, -1.0);
		d.set(0, 1, 3.0);
		d.set(1, 1, 2.0);
		d.set(0, 2, 0.0);
		d.set(1, 2, -3.0);
		compareMatrixWithRef(d, new double[][] { { 6.0, 3.0, 0.0 }, { -1.0, 2.0, -3.0 } });
		final MathMatrix e = c.mult(d);
		compareMatrixWithRef(e, new double[][] { { 12.0, 39.0 }, { -6.0, -12.0 } });
		assertEquals(2, e.getLength(MathMatrix.X));
		assertEquals(2, e.getLength(MathMatrix.Y));
		assertEquals(12.0, e.get(0, 0), EPS);
		assertEquals(-6.0, e.get(1, 0), EPS);
		assertEquals(39.0, e.get(0, 1), EPS);
		assertEquals(-12.0, e.get(1, 1), EPS);
		c.multSelf(2.0);
		assertEquals(12.0, c.get(2, 1), EPS);
		compareMatrixWithRef(c, new double[][] { { 2.0, 8.0 }, { 4.0, 10.0 }, { 6.0, 12.0 } });
	}

	public void assertVectorEquals(final MathVector a, final MathVector b) {
		assertEquals(a.getLength(), b.getLength());
		for (int i = 0; i < a.getLength(); i++) {
			assertEquals(a.get(i), b.get(i), EPS);
		}
	}

	/**
	 * @see MathVector#MathVector(double[])
	 * @see MathVector#isNullVector(double)
	 * @see MathVector#getLength()
	 * @see MathVector#homogenize()
	 * @see MathVector#dehomogenize()
	 * @see MathVector#clone()
	 * @see MathVector#set(int, double)
	 * @see MathVector#add(MathVector)
	 * @see MathVector#mult(double)
	 * @see MathVector#mult(MathVector)
	 * @see MathVector#mult(MathMatrix)
	 * @see MathMatrix#id(int)
	 * @see MathMatrix#zero(int, int)
	 * @see MathMatrix#project(int, int)
	 * @see MathMatrix#homProject(int, int)
	 * @see MathMatrix#rot(int, int, int, double)
	 * @see MathMatrix#homRot(int, int, int, double)
	 * @see MathMatrix#homTrans(MathVector)
	 * @see MathMatrix#homScale(int, double)
	 * @see MathMatrix#homScale(MathVector)
	 * @see MathMatrix#mult(MathVector)
	 */
	@Test
	public void testMatrixOperations() {
		final MathVector v = new MathVector(new double[]{3.0, 4.0, 5.0});
		final MathVector w = new MathVector(new double[]{5.0, 1.0, -2.0});

		// id:
		final MathMatrix id = MathMatrix.id(3);
		final MathVector v2 = id.mult(v);
		assertVectorEquals(v, v2);
		final MathVector v3 = v.mult(id);
		assertVectorEquals(v, v3);

		// zero (3x3)
		final MathMatrix zeroQ = MathMatrix.zero(3, 3);
		final MathVector z3 = zeroQ.mult(v);
		assertEquals(3, z3.getLength());
		assertTrue(z3.isNullVector(EPS));

		// zero (3x5)
		final MathMatrix zero35 = MathMatrix.zero(3, 5);
		final MathVector z5 = zero35.mult(v);
		assertEquals(5, z5.getLength());
		assertTrue(z5.isNullVector(EPS));

		// homTrans/dehomTrans:
		final MathVector hom = v.homogenize();
		final MathVector dehom = hom.dehomogenize();
		assertVectorEquals(v, dehom);

		// projection
		final MathMatrix p = MathMatrix.project(3, 2);
		final MathVector vp = p.mult(v);
		final MathVector vpComp = (MathVector)v.clone();
		vpComp.set(2, 0.0);
		assertVectorEquals(vpComp, vp);
		final MathMatrix ph = MathMatrix.homProject(3, 2);
		final MathVector vph = ph.mult(hom).dehomogenize();
		assertVectorEquals(vp, vph);

		for (int idx1 = 0; idx1 < 3; idx1++) { // rotation
			final int idx2 = (idx1 + 1) % 3;
			final MathMatrix rx90 = MathMatrix.rot(3, idx1, idx2, Math.PI / 2.0);
			final MathVector vrx90 = rx90.mult(v);
			final MathVector vrx90Comp = (MathVector)v.clone();
			vrx90Comp.set(idx1, -v.get(idx2));
			vrx90Comp.set(idx2,  v.get(idx1));
			assertVectorEquals(vrx90Comp, vrx90);
			final MathMatrix rx90h = MathMatrix.homRot(3, idx1, idx2, Math.PI / 2);
			final MathVector vrx90h = rx90h.mult(hom).dehomogenize();
			assertVectorEquals(vrx90, vrx90h);
		}
		// translation
		final MathVector vtrans = v.add(w);
		final MathMatrix transh = MathMatrix.homTrans(w);
		final MathVector vtransh = transh.mult(hom).dehomogenize();
		assertVectorEquals(vtrans, vtransh);

		// translation forth and back
		final MathVector vtrans2 = v.add(w).subtract(w);
		assertVectorEquals(v, vtrans2);
		final MathMatrix transh2 = MathMatrix.homTrans(w.minus());
		final MathMatrix transh12 = transh.mult(transh2);
		final MathVector vtransh12 = transh12.mult(hom).dehomogenize();
		assertVectorEquals(v, vtransh12);

		// scalar scaling
		final MathVector vss = v.mult(2.0);
		final MathMatrix ssh = MathMatrix.homScale(3, 2.0);
		final MathVector vssh = ssh.mult(hom).dehomogenize();
		assertVectorEquals(vss, vssh);

		// vector scaling
		final MathVector vvs = v.mult(w);
		final MathMatrix vsh = MathMatrix.homScale(w);
		final MathVector vvsh = vsh.mult(hom).dehomogenize();
		assertVectorEquals(vvs, vvsh);
	}

	/**
	 * @see MathVector#copyFrom(MathVector)
	 * @see MathVector#equals(Object)
	 */
	@Test
	public void testVectorCopyFrom() {
		final MathVector a = MathVector.create(1.0);
		final MathVector b = MathVector.create(2.0, 3.0);
		a.copyFrom(b);
		assertEquals(a, b);
		assertEquals(a.hashCode(), b.hashCode());
		assertNotEquals(a.getAll(), b.getAll()); // We did a deep copy
		assertNotEquals(a, new MathMatrix(new double[][] { { 2.0 }, { 3.0 } } )); // Matrix is not a vector
		assertNotEquals(a, new MathMatrix(new double[][] { { 2.0 , 3.0 } } )); // Matrix is not a vector
	}

	/**
	 * @see MathMatrix#MathMatrix(double[][])
	 * @see MathMatrix#copyFrom(MathMatrix)
	 * @see MathMatrix#clone()
	 * @see MathMatrix#equals(Object)
	 */
	@Test
	public void testMatrixCopyFrom() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		m1.copyFrom(m1); // Does nothing
		MathMatrix m2 = (MathMatrix) m1.clone();
		assertTrue(m2.equals(m1));
		m2 = new MathMatrix(new double[][] { { 1.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } } );
		m1.copyFrom(m2);
		assertTrue(m1.equals(m2));
		m2 = new MathMatrix(new double[][] { { 1.0, 2.0, 3.0 } } );
		m1.copyFrom(m2);
		assertTrue(m1.equals(m2));
	}

	/**
	 * @see MathVector#MathVector(double[])
	 * @see MathMatrix#MathMatrix(double[][])
	 * @see MathMatrix#getRow(int)
	 * @see MathMatrix#getColumn(int)
	 * @see MathMatrix#clone()
	 * @see MathMatrix#equals(Object)
	 */
	@Test
	public void testGetter() {
		final MathMatrix m1 = new MathMatrix(0, 0);
		m1.setAll(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		assertEquals(new MathVector(new double[] { 1.0, 3.0 } ), m1.getRow(0));
		assertEquals(new MathVector(new double[] { 1.0, 2.0 } ), m1.getColumn(0));
		final MathMatrix m2 = (MathMatrix) m1.clone();
		assertTrue(m2.equals(m1));
	}

	/**
	 * @see MathVector#MathVector(double[])
	 * @see MathVector#toString()
	 * @see MathVector#mult(MathVector)
	 */
	@Test
	public void testMathVectorMath() {
		final MathVector v2 = MathVector.create(2.0, 3.0, 4.0);
		assertEquals(v2, v2.mult(MathVector.id(3)));
	}

	/**
	 * @see MathMatrix#id(int)
	 * @see MathMatrix#MathMatrix(double[][])
	 * @see MathMatrix#minus()
	 * @see MathMatrix#add(MathMatrix)
	 * @see MathMatrix#subtract(MathMatrix)
	 * @see MathMatrix#mult(double)
	 * @see MathMatrix#mult(MathMatrix)
	 */
	@Test
	public void testMathMatrixMath() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		final MathMatrix m2 = new MathMatrix(new double[][] { { 2.0, 3.0 }, { 4.0, 5.0 } } );
		assertEquals(m1, m1.minus().minus());
		assertEquals(m1, m1.add(m2).subtract(m2));
		assertEquals(m1, m1.mult(5.0).mult(0.2));
		assertEquals(m1, m1.mult(MathMatrix.id(2)));
	}

	/**
	 * @see MathMatrix#rot(int, int, int, double)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testRot_IdenticalDimensions_Exception() {
		MathMatrix.rot(3,  1,  1,  Math.PI);
	}

	/**
	 * @see MathMatrix#homProject(int, int)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testHomProject_OutOfBounds() {
		MathMatrix.homProject(3, 3);
	}

	/**
	 * @see MathMatrix#homRot(int, int, int, double)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testHomRot_OutOfBounds1() {
		MathMatrix.homRot(3, 3, 1, Math.PI);
	}

	/**
	 * @see MathMatrix#homRot(int, int, int, double)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testHomRot_OutOfBounds2() {
		MathMatrix.homRot(3, 1, 3, Math.PI);
	}

	/**
	 * @see MathMatrix#homRot(int, int, int, double)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testHomRotXY_OutOfBounds() {
		MathMatrix.homRot(1, MathVector.X, MathVector.Y, Math.PI);
	}

	/**
	 * @see MathMatrix#homRot(int, int, int, double)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testHomRotYZ_OutOfBounds() {
		MathMatrix.homRot(2, MathVector.Y, MathVector.Z, Math.PI);
	}

	/**
	 * @see MathMatrix#homRot(int, int, int, double)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testHomRotZX_OutOfBounds() {
		MathMatrix.homRot(2, MathVector.Z, MathVector.X, Math.PI);
	}

	/**
	 * @see MathMatrix#getLength(int)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testMatrixGetLength_OutOfBounds() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		m1.getLength(MathMatrix.Y + 1);
	}

	/**
	 * @see MathMatrix#add(MathMatrix)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testVectorAdd_SizesDontFit() {
		final MathVector v1 = MathVector.create(1.0, 2.0);
		final MathVector v2 = MathVector.create(2.0, 3.0, 4.0);
		v1.add(v2);
	}

	/**
	 * @see MathMatrix#add(MathMatrix)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testVectorSubtract_SizesDontFit() {
		final MathVector v1 = MathVector.create(1.0, 2.0);
		final MathVector v2 = MathVector.create(2.0, 3.0, 4.0);
		v1.subtract(v2);
	}

	/**
	 * @see MathMatrix#add(MathMatrix)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testMatrixAdd_SizesDontFit() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		final MathMatrix m2 = new MathMatrix(new double[][] { { 2.0, 3.0, 4.0 }, { 4.0, 5.0, 6.0 } } );
		m1.add(m2);
	}

	/**
	 * @see MathMatrix#subtract(MathMatrix)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testMatrixSubtract_SizesDontFit() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		final MathMatrix m2 = new MathMatrix(new double[][] { { 2.0, 3.0, 4.0 }, { 4.0, 5.0, 6.0 } } );
		m1.subtract(m2);
	}

	/**
	 * @see MathVector#mult(MathVector)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testVectorMultVector_SizesDontFit() {
		final MathVector v1 = MathVector.create(1.0, 2.0);
		final MathVector v = MathVector.create(2.0, 3.0, 4.0);
		v.mult(v1);
	}

	/**
	 * @see MathMatrix#mult(MathVector)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testVectorMultMatrix_SizesDontFit() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		final MathVector v = new MathVector(new double[] { 2.0, 3.0, 4.0 } );
		v.mult(m1);
	}

	/**
	 * @see MathMatrix#mult(MathVector)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testMatrixMultVector_SizesDontFit() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		final MathVector v = new MathVector(new double[] { 2.0, 3.0, 4.0 } );
		m1.mult(v);
	}

	/**
	 * @see MathMatrix#mult(MathMatrix)
	 */
	@Test(expected=ArrayIndexOutOfBoundsException.class)
	public void testMatrixMultMatrix_SizesDontFit() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } } );
		final MathMatrix m2 = new MathMatrix(new double[][] { { 2.0, 3.0, 4.0 }, { 4.0, 5.0, 6.0 } } );
		m1.mult(m2);
	}

	/**
	 * @see MathMatrix#setAll(double[][])
	 * @see MathMatrix#equals(Object)
	 * @see MathMatrix#hashCode()
	 */
	@Test
	public void testSetterEquals() {
		final MathMatrix m1 = new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } });
		final MathMatrix m2 = new MathMatrix(2, 2);
		m2.setAll(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 } });
		assertTrue(m1.equals(m2));
		assertEquals(m1.hashCode(), m2.hashCode());
		assertFalse(m1.equals(new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 5.0 } })));
		assertFalse(m1.equals(new MathMatrix(new double[][] { { 1.0, 2.0 }, { 3.0, 4.0 }, { 5.0, 6.0 } })));
		assertFalse(m1.equals(new MathMatrix(new double[][] { { 1.0, 2.0, 3.0 }, { 3.0, 4.0, 5.0 } })));
		assertFalse(m1.equals(null));
	}

	/**
	 * @see MathMatrix#toString()
	 * @see MathVector#toString()
	 */
	@Test
	public void testToString() {
		assertEquals("((1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0))", MathMatrix.id(3).toString());
		assertEquals("(1.0, 1.0, 1.0)", MathVector.id(3).toString());
	}

	/**
	 * @see MathVector#equals(Object)
	 * @see MathVector#clone()
	 * @see MathMatrix#equals(Object)
	 * @see MathMatrix#clone()
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(true, new MathVector(2), MathVector.create(2, 3));
		GenericTest.checkObjectBasics(true, new MathMatrix(2, 2), MathMatrix.id(2));
	}
}
