package de.ped.tools.test;

import static org.junit.Assert.*;

import java.util.Locale;

import org.junit.Before;
import org.junit.Test;

import de.ped.tools.TextDocumentType;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Unit tests for the {@link TextFormatter} class.
 * 
 * @author Copyright (c) 2014 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TextFormatterTest {

	private Logger logger = Logger.getLogger(getClass());

	@Before
	public void setup() {
		logger.setLogLevel(Level.INFO);
	}

	/**
	 * @see TextFormatter#TextFormatter(TextDocumentType)
	 * @see TextFormatter#setLocale(Locale)
	 * @see TextFormatter#beginDocument()
	 * @see TextFormatter#addHead(String, String)
	 * @see TextFormatter#beginBody()
	 * @see TextFormatter#beginParagraph()
	 * @see TextFormatter#newLine()
	 * @see TextFormatter#endParagraph()
	 * @see TextFormatter#endBody()
	 * @see TextFormatter#endDocument()
	 */
	@Test
	public void testPattern() {
		final Locale[] locales = new Locale[] { null, new Locale("de", "DE") };
		for (final TextDocumentType docType : TextDocumentType.values()) {
			for (final Locale locale : locales) {
				final TextFormatter tf = new TextFormatter(docType);
				tf.setLocale(locale);
				final StringBuffer result = new StringBuffer();
				result.append(tf.beginDocument());
				result.append(tf.addHead("Test", null));
				result.append(tf.beginBody());
				result.append(tf.beginParagraph());
				result.append("Dies ist der erste Testsatz im ersten Absatz. ");
				result.append(tf.newLine());
				result.append("Dies ist der zweite Testsatz im ersten Absatz. ");
				result.append(tf.endParagraph());
				result.append(tf.endBody());
				result.append(tf.endDocument());
				logger.debug("***" + docType.name() + "*** :\n" + result.toString());
			}
		}
	}

	/**
	 * @see TextFormatter#sgmlify(TextDocumentType, String)
	 */
	@Test
	public void testSgmlify() {
		assertEquals("", TextFormatter.sgmlify(TextDocumentType.HTML401_STRICT, ""));
		assertEquals("&quot;&lt;&amp;&gt;&quot;", TextFormatter.sgmlify(TextDocumentType.HTML401_STRICT, "\"<&>\""));
		assertEquals("&quot;&lt;&amp;&gt;&quot;", TextFormatter.sgmlify(TextDocumentType.TEXT, "&quot;&lt;&amp;&gt;&quot;"));
	}

	/**
	 * @see TextFormatter#toPlainText(TextDocumentType, String)
	 */
	@Test
	public void testToPlainText() {
		assertEquals("", TextFormatter.toPlainText(TextDocumentType.HTML401_STRICT, ""));
		assertEquals("\"<&>\"", TextFormatter.toPlainText(TextDocumentType.HTML401_STRICT, "&quot;&lt;&amp;&gt;&quot;"));
		assertEquals("&quot;&lt;&amp;&gt;&quot;", TextFormatter.toPlainText(TextDocumentType.TEXT, "&quot;&lt;&amp;&gt;&quot;"));
		assertEquals("text", TextFormatter.toPlainText(TextDocumentType.HTML401_STRICT, "<a href=\"#target\">text</a>"));
		assertEquals("This was great and wonderful text", TextFormatter.toPlainText(TextDocumentType.HTML401_STRICT, "This was <a href=\"#target\">great</a> and <a href=\"#target\">wonderful</a> text"));
		assertEquals("This was wonderful text", TextFormatter.toPlainText(TextDocumentType.HTML401_STRICT, "This was <head>great and </head>wonderful text"));
	}

	/**
	 * @see TextFormatter#repeat(String, int)
	 */
	@Test
	public void testRepeat() {
		assertNull(TextFormatter.repeat(null, 5));
		assertEquals("", TextFormatter.repeat("", 5));
		assertEquals("", TextFormatter.repeat("Hallo", 0));
		assertEquals("Hallo", TextFormatter.repeat("Hallo", 1));
		assertEquals("HalloHallo", TextFormatter.repeat("Hallo", 2));
	}

	/**
	 * @see TextFormatter#wrap(String)
	 * @see TextFormatter#unwrap(String[])
	 */
	@Test
	public void testWrapUnwrap() {
		assertArrayEquals(null, TextFormatter.wrap(null));
		assertEquals(null, TextFormatter.unwrap(null));
		assertArrayEquals(new String[]{}, TextFormatter.wrap(""));
		assertEquals("", TextFormatter.unwrap(new String[]{}));
		assertArrayEquals(new String[]{" "}, TextFormatter.wrap(" "));
		assertArrayEquals(new String[]{" "}, TextFormatter.wrap(" \n"));
		assertEquals(" \n", TextFormatter.unwrap(new String[]{" "}));
	}
}
