package de.ped.tools.web;

import java.util.EnumSet;

/**
 * The data necessary to describe an entry in the navigation bar.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach
 */
public class NavBarEntry {

	public static final String USE_REF = "#";

	/**
	 * Constructor.
	 * @param depth the depth in the navigation tree
	 * @param title the title to display in the navigation tree for the reader of the page in German language
	 * @param title_en_US the title to display in the navigation tree for the reader of the page in English language
	 * @param description the description to add for the reader of the page in German language
	 * @param description_en_US the description to add for the reader of the page in English language
	 * @param ref the page reference this entry leads to, or <code>null</code> for a heading entry
	 * @param liClass the class attribute value of the HTML li (list item) element
	 * @param iconPath the URL to the icon belonging to that entry
	 * @param languagesSupported the languages for which this list is suggested
	 */
	public NavBarEntry(final int depth, final String title, final String title_en_US, final String description, final String description_en_US, final PageRef ref, final String liClass, final String iconPath, EnumSet<RootBean.Language> languagesSupported) {
		this.depth = depth;
		this.ref = ref;
		this.title = title;
		this.title_en_US = title_en_US;
		this.description = description;
		this.description_en_US = description_en_US;
		this.iconPath = iconPath;
		this.liClass = liClass;
		this.languagesSupported = languagesSupported;
	}

	public int getDepth() {
		return depth;
	}

	public PageRef getPageRef() {
		return ref;
	}

	public String getTitle(final RootBean.Language language) {
		String ret = null;
		if (null == title) {
			if (null != ref) {
				ret = ref.getTitle(language);
			}
		} else {
			if ((RootBean.Language.ENGLISH == language) && (null != title_en_US)) {
				ret = title_en_US;
			} else {
				ret = title;
			}
		}
		return ret;
	}

	public String getDescription(final RootBean.Language language) {
		final String ret;
		if (null == description_en_US) {
			if (null == description) {
				// If nothing is defined, return an empty string
				ret = "";
			} else {
				// If not both are defined, pick the one that is defined.
				ret = description;
			}
		} else if (null == description) {
			// If not both are defined, pick the one that is defined.
			ret = description_en_US;
		} else if (RootBean.Language.ENGLISH == language) {
			// Both are defined, so pick the right language for the page.
			ret = description_en_US;
		} else {
			ret = description;
		}
		return ret;
	}

	public boolean isFor(final RootBean.Language language) {
		final boolean ret;
		if (null == languagesSupported) {
			ret = true;
		} else if (languagesSupported.contains(language)) {
			ret = true;
		} else {
			ret = false;
		}
		return ret;
	}

	public String getIconPath() {
		final String ret;
		if (USE_REF.equals(iconPath)) {
			ret = ref.titleImg;
		} else {
			ret = iconPath;
		}
		return ret;
	}

	/** The depth in the navigation tree. */
	protected final int depth;
	/** The page the navigation bar entry is leading to. Might be <code>null</code>, e.g., for headings. */
	protected final PageRef ref;
	/** The class attribute of the HTML &lt;li&gt; element. */
	public final String liClass;
	/** The icon to show with the navigation. */
	protected final String iconPath;
	/** The title to display in the nav bar for the entry */
	public final String title;
	/** The title to display in the nav bar for the entry */
	protected final String title_en_US;
	/** The description to display for the entry */
	protected final String description;
	/** The description to display for the entry */
	protected final String description_en_US;
	/** The language for which this list is suggested, or <code>null</code> if the destination page supports both German and English. */
	public final EnumSet<RootBean.Language> languagesSupported;
}
