package de.ped.tools.web;

import java.awt.Dimension;
import java.io.File;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.EnumSet;
import java.util.Locale;

import javax.servlet.ServletContext;
import javax.servlet.http.HttpServletRequest;

import de.ped.tools.CollectionUtil;
import de.ped.tools.HtmlHelper;
import de.ped.tools.ResourceFinder;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.web.PageRef.HTMLFlavor;

/**
 * The bean containing the Java interface to Peter Diefenbach's homepage Java Server Pages.
 * 
 * @author Copyright (c) 2009 Peter Diefenbach
 */
public class RootBean {

	public static final String CANHIDE = "canhide";

	/**
	 * The human language a page is written in.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach
	 */
	public static enum Language {
		GERMAN ("de-DE", "de_DE", "",       "Deutsch", "/de/ped/tools/resources/Menu.Locales.de_DE.png"),
		ENGLISH("en-US", "en_US", "_en_US", "English", "/de/ped/tools/resources/Menu.Locales.en_US.png");

		/** the HTML language key */
		public final String htmlLangKey;
		/** the infix for the flag resource filename */
		public final String resFilenameInfix;
		/** the infix for the html file */
		public final String htmlFileInfix;
		/** the text to show */
		public final String text;
		/** the Java locale */
		public final Locale locale;
		/** The file path string representation to the flag symbol */
		public final String flagFilepath;

		private Language(final String htmlLangKey, final String resFilenameInfix, final String htmlFileInfix, final String text, final String flagFilepath) {
			this.htmlLangKey = htmlLangKey;
			this.resFilenameInfix = resFilenameInfix;
			this.htmlFileInfix = htmlFileInfix;
			this.text = text;
			final String[] parts = htmlLangKey.split("-");
			this.locale = new Locale(parts[0], parts[1]);
			this.flagFilepath = flagFilepath;
		}
	}

	public static final EnumSet<Language> LANGSET_GERMAN  = EnumSet.of(Language.GERMAN);
	public static final EnumSet<Language> LANGSET_ENGLISH = EnumSet.of(Language.ENGLISH);
	public static final EnumSet<Language> LANGSET_ALL     = EnumSet.of(Language.GERMAN, Language.ENGLISH);
	public static final EnumSet<Language> LANGSET_NONE    = EnumSet.noneOf(Language.class);

	/** Link class indicating indentation in the site menu. */
	private static final String LINK_CLASS_LEVEL1 = "level1";
	/** Link class indicating indentation in the site menu. */
	private static final String LINK_CLASS_LEVEL0 = "level0";
	/** Link class indicating indentation in the site menu. */
	private static final String LINK_CLASS_HEADING = "heading";
	/** Link class indicating the menu for language selection in the site menu. */
	private static final String LINK_CLASS_LANGSELECT = "langselect";
	/** Link class indicating last entry of an indentation in the site menu. */
	private static final String LINK_CLASS_LEVEL1_LAST = LINK_CLASS_LEVEL1 + " last";

	/* ------------------------------------------------------- */ 
	/* --- Static HTML pages                               --- */ 

	public static final String AUTHOR = "Peter Diefenbach";
	public static final String MAILADDRESS = "peter@pdiefenbach.de";

	public static final String ITEM_KEY = "item";
	public static final String ITEM_VALUE_TOC = "toc";
	public static final String ITEM_VALUE_CONTENT = "content";
	public static final String ITEM_VALUE_ALL = "all";

	/** 
	 * @param cssClass the value of the CSS class attribute for a HTML element
	 * @return a HTML snippet for the CSS class attribute
	 */
	protected static String htmlAttribClass(final String cssClass) {
		final String ret;
		if ((null != cssClass) && (!cssClass.isEmpty())) {
			ret = htmlAttribute(TextFormatter.CLASS, cssClass) + " ";
		} else {
			ret = "";
		}
		return ret;
	}

	public static String htmlAttribute(final String key, final String value) {
		return TextFormatter.htmlAttribute(key, value);
	}

	public RootBean() {
		Logger.getRootLogger().setLogLevel(Level.WARN);
	}

	public String url(final Subdomain destinationSubdomain, final String page) {
		String ret;
		if (Subdomain.EXTERNAL.equals(destinationSubdomain)) {
			ret = page;
		} else {
			final String subdomainPrefix = getSubdomainPrefixTo(destinationSubdomain);
			if (subdomainPrefix.isEmpty()) {
				ret = getLocalHomePath() + "/" + page;
			} else {
				ret = getSubdomainPrefixTo(destinationSubdomain) + page;
			}
		}
		return ret;
	}

	public String url(final PageRef entry) {
		return url(entry.getSubdomain(), entry.getLink(language));
	}

	protected String url(final NavBarEntry entry) {
		return url(entry.getPageRef());
	}

	/**
	 * @param id the HTML id attribute of the anchor element
	 * @param text the text that will be embraced by the anchor element
	 * @return an HTML anchor element.
	 */
	public String anchor(final String id, final String text) {
		return TextFormatter.anchor(thisPageRef.htmlFlavor.doctype, id, text);
	}

	/**
	 * @param id the HTML id attribute of the anchor element
	 * @return an HTML anchor element.
	 */
	public String anchor(final String id) {
		return anchor(id, null);
	}

	/**
	 * Returns an HTML href element.
	 * 
	 * @param targetSubdomain the subdomain the link is showing to
	 * @param url             the URL of the link within the subdomain
	 * @param text            the text to display on the HTML page
	 * @param cssClass        the CSS class name of this element, or
	 *                        <code>null</code> if none
	 * @return an HTML snippet
	 */
	public String link(final Subdomain targetSubdomain, final String url, final String text, final String cssClass) {
		final String cssClazz;
		if (Subdomain.EXTERNAL.equals(targetSubdomain)) {
			cssClazz = "extLink";
		} else {
			cssClazz = cssClass;
		}
		final String ret = ("<a" 
				+ htmlAttribClass(cssClazz) 
				+ htmlAttribute(TextFormatter.HREF, HtmlHelper.sgmlify(url(targetSubdomain, url)))
				+ ">"
				+ text + "</a>");
		return ret;
	}

	/**
	 * Returns an HTML href element.
	 * 
	 * @param ref      the page reference to link to (or at least the beginning of
	 *                 it), or <em>null</em> to create a non-link HTML text snippet
	 * @param postfix  anything that needs to be appended to the reference, e.g.
	 *                 page parameters (<code>null</code> adds nothing)
	 * @param text     the text to display on the HTML page. (If <code>null</code>,
	 *                 the text stored in <em>ref</em> is taken.)
	 * @param cssClass the CSS class name of this element, or <code>null</code> if
	 *                 none
	 * @param iconPath the path to the icon, or {@link NavBarEntry#USE_REF} to use
	 *                 the path in <em>ref</em>, or <code>null</code> for no icon
	 * @param dummy    currently unused
	 * @return an HTML snippet
	 */
	public String link(final PageRef ref, final String postfix, final String text, final String cssClass,
			final String iconPath, final boolean dummy) {
		final String pf = (null == postfix ? "" : postfix);
		final String txt;
		final String img;
		final String imgtxt;
		if (null == text) {
			if (null == ref) {
				txt = null;
			} else {
				txt = ref.getTitle(language);
			}
		} else {
			txt = text;
		}
		if (NavBarEntry.USE_REF.equals(iconPath)) {
			if (null == ref) {
				img = null;
			} else {
				img = ref.titleImg;
			}
		} else {
			img = iconPath;
		}
		final String content;
		if (null != img) {
			imgtxt = TextFormatter.img(getThisPageRef().htmlFlavor.doctype, img, "", ICON_DIMENSION, null) 
					+ "&nbsp;";
			content = "<span style=\"white-space:nowrap\">" + imgtxt + "<span style=\"white-space:normal\">" + txt + "</span></span>";
		} else {
			imgtxt = "";
			content = imgtxt + txt;
		}
		final String href;
		if (null == ref) {
			href = content;
		} else {
			href = link(ref.getSubdomain(), ref.getLink(language) + pf, content, cssClass);
		}
		return href;
	}

	/**
	 * Returns an HTML href element.
	 * 
	 * @param ref
	 *            the page reference to link to
	 * @param text
	 *            the text to display on the HTML page. (If <code>null</code>,
	 *            the text stored in <em>ref</em> is taken.)
	 * @param cssClass
	 *            the CSS class name of this element, or <code>null</code> if
	 *            none
	 * @return an HTML snippet
	 */
	public String link(final PageRef ref, final String text, final String cssClass) {
		return link(ref, null, text, cssClass, NavBarEntry.USE_REF, true);
	}

	/**
	 * Returns an HTML href element.
	 * 
	 * @param ref
	 *            the page reference to link to
	 * @param text
	 *            the text to display on the HTML page. (If <code>null</code>,
	 *            the text stored in <em>ref</em> is taken.)
	 * @return an HTML snippet
	 */
	public String link(final PageRef ref, final String text) {
		return link(ref, null, text, null, NavBarEntry.USE_REF, true);
	}

	/**
	 * Returns an HTML href element.
	 * 
	 * @param ref
	 *            the page reference to link to
	 * @return an HTML snippet
	 */
	public String link(final PageRef ref) {
		return link(ref, null, null, null, NavBarEntry.USE_REF, true);
	}

	/**
	 * Returns an HTML href element to a target not part of this domain.
	 * 
	 * @param url
	 *            the url to link to
	 * @param text
	 *            the text to display on the HTML page
	 * @param cssClass
	 *            the CSS class name of this element, or <code>null</code> if
	 *            none
	 * @return an HTML snippet
	 */
	public String extLink(final String url, final String text, final String cssClass) {
		return link(Subdomain.EXTERNAL, url, text, cssClass);
	}

	/**
	 * Returns an HTML href element to a target not part of this domain.
	 * 
	 * @param url
	 *            the url to link to
	 * @param text
	 *            the text to display on the HTML page
	 * @return an HTML snippet
	 */
	public String extLink(final String url, final String text) {
		return link(Subdomain.EXTERNAL, url, text, null);
	}

	/**
	 * Returns an HTML href element to a target not part of this domain.
	 * 
	 * @param url
	 *            the url to link to. Also used as the text to show on this page
	 *            for the link.
	 * @return an HTML snippet
	 */
	public String extLink(final String url) {
		return link(Subdomain.EXTERNAL, url, url, null);
	}

	/**
	 * @param title the window title
	 * @param keywords the keywords to define, or <code>null</code> for none
	 * @param description the description, or <code>null</code> for none
	 * @param cssfile the page specific cascading style sheet file name, or <code>null</code> for no specific file
	 * @return an HTML head element completely filled.
	 */
	public String head(final String title, final String keywords, final String description, final String cssfile) {
		final StringBuffer s = new StringBuffer();
		final String icon = getSubdomain().iconUrl;
		final String css = getSubdomain().cssPath;
		s.append("\n");
		addLine(s, 0, "<head>");
		if (null != title) {
			addLine(s, 1, "<title>" + title + "</title>");
		}
		final String charset = thisPageRef.charset;
		final String charsetElem;
		if (thisPageRef.htmlFlavor == HTMLFlavor.HTML5) {
			charsetElem = "<meta" + htmlAttribute(TextFormatter.CHARSET, charset) + "/>";
		} else {
			charsetElem = "<meta" + htmlAttribute(TextFormatter.HTTP_EQUIV, "Content-Type")
				+ htmlAttribute(TextFormatter.CONTENT, "text/html; charset=" + charset) 
				+ "/>";
		}
		addLine(s, 1, charsetElem);
		addLine(s, 1, "<link" 
				+ htmlAttribute(TextFormatter.REL, "stylesheet")
				+ htmlAttribute(TextFormatter.TYPE, "text/css")
				+ htmlAttribute(TextFormatter.HREF, css)
				+ "/>");
		if (null != cssfile) {
			addLine(s, 1, "<link" 
					+ htmlAttribute(TextFormatter.REL, "stylesheet")
					+ htmlAttribute(TextFormatter.TYPE, "text/css")
					+ htmlAttribute(TextFormatter.HREF, cssfile)
					+ "/>");
		}
		addLine(s, 1, "<link"
				+ htmlAttribute(TextFormatter.REL, "author")
				+ htmlAttribute(TextFormatter.HREF, "mailto:" + MAILADDRESS)
				+ "/>");
		addLine(s, 1, "<link"
				+ htmlAttribute(TextFormatter.REL, "shortcut icon")
				+ htmlAttribute(TextFormatter.HREF, icon)
				+ "/>");
		addLine(s, 1, "<script" 
				+ (thisPageRef.htmlFlavor == HTMLFlavor.HTML5 ? "" : htmlAttribute(TextFormatter.TYPE, "text/javascript"))
				+ htmlAttribute(TextFormatter.SRC, "/defines.js")
				+ "></script>");
		// Requirement from https://developers.google.com/speed/docs/insights/ConfigureViewport
		addLine(s, 1, "<meta"
				+ htmlAttribute(TextFormatter.NAME, "viewport")
				+ htmlAttribute(TextFormatter.CONTENT, "width=device-width, initial-scale=1")
				+ "/>");
		addLine(s, 1, "<meta"
				+ htmlAttribute(TextFormatter.NAME, "author")
				+ htmlAttribute(TextFormatter.CONTENT, AUTHOR)
				+ "/>");
		addLine(s, 1, "<meta"
				+ htmlAttribute(TextFormatter.NAME, "keywords")
				+ htmlAttribute(TextFormatter.CONTENT, "Diefenbach" + ((null == keywords) ? "" : (", " + keywords)))
				+ "/>");
		if (null != description) {
			addLine(s, 1, "<meta"
					+ htmlAttribute(TextFormatter.NAME, "description")
					+ htmlAttribute(TextFormatter.CONTENT, description)
					+ "/>");
		}
		if (getSubdomain().isSearchableForRobots) {
			addLine(s, 1, "<meta"
					+ htmlAttribute(TextFormatter.NAME, "rating")
					+ htmlAttribute(TextFormatter.CONTENT, "General")
					+ "/>");
			addLine(s, 1, "<meta"
					+ htmlAttribute(TextFormatter.NAME, "robots")
					+ htmlAttribute(TextFormatter.CONTENT, "follow")
					+ "/>");
			addLine(s, 1, "<meta"
					+ htmlAttribute(TextFormatter.NAME, "robots")
					+ htmlAttribute(TextFormatter.CONTENT, "index")
					+ "/>");
			addLine(s, 1, "<meta"
					+ htmlAttribute(TextFormatter.NAME, "revisit-after")
					+ htmlAttribute(TextFormatter.CONTENT, "31 days")
					+ "/>");
		}
		addLine(s, 0, "</head>");
		return s.toString();
	}

	public String head(final String title, final String keywords, final String description) {
		return head(title, keywords, description, null);
	}

	public String head(final String title, final String keywords) {
		return head(title, keywords, null);
	}

	public String head(final String title) {
		return head(title, null);
	}

	public String titleImg() {
		String ret = "";
		if ((null != thisPageRef) && (null != thisPageRef.titleImg)) {
			ret = TextFormatter.img(getThisPageRef().htmlFlavor.doctype, thisPageRef.titleImg, "", null, 
					htmlAttribute(TextFormatter.CLASS, "title"));
		}
		return ret;
	}

	/**
	 * @return a HTML snippet containing the attributes of the HTML body
	 */
	public String bodyAttribs() {
		return bodyAttribs(null);
	}

	/**
	 * @param classes HTML class content to add for this body element, 
	 * 		or <code>null</code> if no additional HTML classes to set
	 * @return a HTML snippet containing the attributes of the HTML body
	 */
	public String bodyAttribs(final String classes) {
		final String bodyClasses = (getSubdomain().isWithNav ? "html5nav" : "")
				+ (CollectionUtil.isNullOrEmpty(thisPageRef.bodyClasses) ? "" : " " + thisPageRef.bodyClasses)
				+ (CollectionUtil.isNullOrEmpty(classes) ? "" : " " + classes);
		return (CollectionUtil.isNullOrEmpty(bodyClasses) ? "" : htmlAttribute(TextFormatter.CLASS, bodyClasses));
	}

	/**
	 * @param pageHeading the heading
	 * @return a HTML snippet containing the beginning of a HTML body
	 */
	public String bodyStart(final String pageHeading) {
		final StringBuffer s = new StringBuffer();
		addLine(s, 0, "");
		addLine(s, 1, "<a" + htmlAttribute(TextFormatter.ID, "top") + "></a>");
		if (getSubdomain().isWithNav) {
			createHtml5Nav(s);
		}
		addLine(s, 1, "<main" 
				+ htmlAttribute(TextFormatter.ID, "main")
				+ ">");
		addLine(s, 1, "<div" + htmlAttribute(TextFormatter.CLASS, "contentL") + ">");
		addLine(s, 1, "<div" + htmlAttribute(TextFormatter.CLASS, "contentR") + ">");
		addLine(s, 1, "<div" + htmlAttribute(TextFormatter.CLASS, "contentC") + ">");
		if (null != pageHeading) {
			addLine(s, 1, "<h1>" + pageHeading + "</h1>");
		}
		return s.toString();
	}

	final Dimension ICON_DIMENSION = new Dimension(16, 16);

	private void createHtml5Nav(final StringBuffer s) {
		final String toMainText = (language ==  Language.ENGLISH ? "to main content" : "zum Hauptinhalt");
		addLine(s, 1, "<a" + htmlAttribute(TextFormatter.ID, "skip-link") + htmlAttribute(TextFormatter.HREF, "#main") + ">" + toMainText + "</a>");
		addLine(s, 1, "<nav" + htmlAttribute(TextFormatter.ID, "sitenav") + htmlAttribute(TextFormatter.CLASS, CANHIDE)
				+ htmlAttribute("aria-labelledby", "sitenav-label") + ">");
		addLine(s, 2, "<h2" + htmlAttribute(TextFormatter.ID, "sitenav-label") + htmlAttribute(TextFormatter.CLASS, "visually-hidden") + ">Site Navigation</h2>");
		int prevDepth = -1;
		int td = 2;
		for (int idx = 0; idx < navBarEntriesTotalNew.length; idx++) {
			final NavBarEntry entry = navBarEntriesTotalNew[idx];
			final NavBarEntry next = (idx < navBarEntriesTotalNew.length - 1 ?  navBarEntriesTotalNew[idx + 1] : null);
			final int depth = entry.getDepth();
			final PageRef ref = entry.getPageRef();
			final String title = entry.getTitle(language).replaceAll(" ", "&nbsp;");
			final String img = entry.getIconPath();
			if (LINK_CLASS_LANGSELECT.equals(entry.liClass)) {
				if (!thisPageRef.isMultiLingual()) {
					break;
				}
			}
			boolean isNewLevel = false;
			for (int i = prevDepth - 1; i >= depth; i--) {
				if (i <= 1) {
					addLine(s, --td, "</ul>");
					addLine(s, --td, "</details>");
					if (0 <= i) {
						addLine(s, --td, "<!-- </li> -->");
					}
				}
			}
			if ((depth <= 1) && (null != next) && (depth < next.getDepth())) {
				isNewLevel = true;
			} else if (depth == 3) {
				// No elbow, as defined via class="level1"
			}
			final Language langToShow;
			if ((null == entry.languagesSupported) 
				|| (entry.languagesSupported.isEmpty())
				|| (entry.languagesSupported.contains(this.language))) {
				langToShow = null;
			} else {
				langToShow = entry.languagesSupported.iterator().next();
			}
			final String text = (null == img ? "" : 
								TextFormatter.img(getThisPageRef().htmlFlavor.doctype, 
										img, "", ICON_DIMENSION, null)
								+ "&nbsp;")
							+ title
							+ (null == langToShow ? "" : 
								"&nbsp;(" 
								+ TextFormatter.img(getThisPageRef().htmlFlavor.doctype, 
										langToShow.flagFilepath, "", null, null)
								+ ")");
			String href;
			String current = null;
			if (null == ref) {
				if (isNewLevel) {
					href = null;
				} else {
					href = text;
				}
			} else if (!ref.getSubdomain().equals(getSubdomain())) {
				href = link(ref, null, text, null, null, false);
			} else if (ref.getLink(language).equalsIgnoreCase(getThisPageRef().getLink(language))) {
				href = "<a" + htmlAttribute(TextFormatter.HREF, "#") + ">" + text + "</a>";
				current = "page";
			} else {
				href = link(ref, null, text, null, null, false);
			}
			if ((null == current) && (1 == entry.getDepth())) {
				// Check if a child is current page
				NavBarEntry l1 = null;
				final Logger logger = Logger.getLogger(getClass());
				if (PageRef.HOME == getThisPageRef()) { logger.trace("entry=" + entry.title); }
				for (final NavBarEntry e : navBarEntriesTotalNew) {
					if (PageRef.HOME == getThisPageRef()) { logger.trace("e=" + e.title); }
					if (1 == e.getDepth()) {
						l1 = e;
						if (PageRef.HOME == getThisPageRef()) { logger.trace("l1=" + e.title); }
					}
					final PageRef r = e.getPageRef();
					if ((null != r) && (r == getThisPageRef())) {
						// We have found the current page
						if (PageRef.HOME == getThisPageRef()) { logger.trace("finishing with " + e.title + ", because " + r.bodyClasses + " == " + getThisPageRef().bodyClasses); }
						break;
					}
				}
				if (entry == l1) {
					current = "page";
				}
			}
			// Check for language entries
			for (Language lang : Language.values()) {
				if (lang.htmlLangKey.equals(entry.liClass)) {
					if (lang.htmlLangKey.equals(language.htmlLangKey)) {
						href = "<a" + htmlAttribute(TextFormatter.HREF, "#") + ">" + text + "</a>";
						current = "page";
					} else {
						final String paramStr = (CollectionUtil.isNullOrEmpty(itemToShow) ? "" : "?item=" + itemToShow);
						href = "<a" + htmlAttribute(TextFormatter.HREF, thisPageRef.getLink(lang) + paramStr) + ">" + text + "</a>";
					}
					s.append("\n");
					break;
				}
			}
			// Write the list item
			if (isNewLevel) {
				if (0 < depth) {
					addLine(s, td++, 
							"<li" + (null == entry.liClass ? "" : htmlAttribute(TextFormatter.CLASS, entry.liClass)) 
							+ (null == current ? "" : htmlAttribute("aria-current", current))
							+ ">");
				}
				addLine(s, td++, "<details>");
				addLine(s, td, "<summary" + (entry.depth == 0 ? htmlAttribute(TextFormatter.ID, "navrootsummary") : "") + ">"
						+ (null == img ? "" : 
								TextFormatter.img(getThisPageRef().htmlFlavor.doctype, img, "", ICON_DIMENSION, null)
								+ "&nbsp;")
						+ title 
						+ "</summary>");
				addLine(s, td++, "<ul>");
			} else if (null != href) {
				addLine(s, td, 
						"<li" 
						+ (null == entry.liClass ? "" : htmlAttribute(TextFormatter.CLASS, entry.liClass)) 
						+ (null == current ? "" : htmlAttribute("aria-current", current)) 
						+ ">" 
						+ href + "</li>");
			}
			prevDepth = depth;
		}
		for (int i = prevDepth - 1; i >= 0; i--) {
			if (i <= 1) {
				addLine(s, --td, "</ul>");
				addLine(s, --td, "</details>");
				if (0 <= i) {
					addLine(s, --td, "<!-- </li> -->");
				}
			}
		}
		addLine(s, 1, "</nav>");
	}

	public String printNavEntry(final NavBarEntry entry) {
		return printNavEntry(entry, null);
	}

	public String printNavEntry(final NavBarEntry entry, String text) {
		final String txt = null == text ? entry.getTitle(language) : text;
		final String href = link(entry.getPageRef(), null, txt, null, entry.getIconPath(), true);
		return href;
	}

	public String bodyStart() {
		return bodyStart(null);
	}

	public static final boolean IS_SHOWING_FILE_DATE = false;
	public static final String PROJECT_FILE_PATH_ROOT = "C:/Projekte/Peter/eclipse/workspace8/WebProject/";
	protected DateFormat getDateFormat() {
		if (null == df) {
			df = new SimpleDateFormat("yyyy-MM-dd");
		}
		return df;
	}

	/**
	 * @param filename
	 *            the name of the html file
	 * @return the date of the last change timestamp of the source file
	 */
	public String getDate(final String filename) {
		String ret = null;
		if (!IS_SHOWING_FILE_DATE) {
			// Nothing to do
		} else if (null == filename) {
			// Nothing to do
		} else {
			final String srcFilename = filename.replace(".html", ".jsp");
			final File srcFile = new File(
					PROJECT_FILE_PATH_ROOT + "src/html/" + getSubdomain().sourceJspPath + File.separator + srcFilename);
			final long srcTimestamp = srcFile.lastModified();
			if (0L == srcTimestamp) {
				// return "!!!File not found: " + srcFile;
				// Nothing to do
			} else {
				final Date d = new Date(srcTimestamp);
				final DateFormat df = getDateFormat();
				ret = df.format(d);
			}
		}
		return ret;
	}

	/**
	 * @param copyrightMessage
	 *            the copyright message to show at the end of the page
	 * @return a HTML snippet containing the ending of the HTML body tag
	 */
	public String bodyEnd(final String copyrightMessage) {
		final StringBuffer s = new StringBuffer();
		final String date = getDate(getThisPageRef().getLink(language));
		addLine(s, 0, "");
		// Clean that div orgy up at body end
		addLine(s, 1, "</div> <!-- contentC -->");
		addLine(s, 1, "</div> <!-- contentR -->");
		addLine(s, 1, "</div> <!-- contentL -->");
		addLine(s, 1, "<div" + htmlAttribute(TextFormatter.ID, "footer") 
				+ htmlAttribute(TextFormatter.CLASS, CANHIDE)
				+ htmlAttribute(TextFormatter.ROLE, TextFormatter.CONTENTINFO) + ">");
		addLine(s, 2, "<hr" + htmlAttribute(TextFormatter.ID, "footerHr") + "/>");
		final String toTopText = "⇱"; //(language ==  Language.ENGLISH ? "to top" : "zum Seitenanfang");
		addLine(s, 2, "<div" + htmlAttribute(TextFormatter.CLASS, "pagenav") 
				+ "><a" + htmlAttribute(TextFormatter.HREF, "#top") + ">" + toTopText + "</a></div>");
		addLine(s, 2, copyrightMessage);
		if (null != date) {
			final String lastUpdateText = (language ==  Language.ENGLISH ? " &mdash; latest changes: " : " &mdash; Zuletzt aktualisiert: ");
			addLine(s, 2, lastUpdateText + date);
		}
		addLine(s, 1, "</div> <!-- footer -->");
		addLine(s, 1, "</main>");
		return s.toString();
	}

	public String rangeTo() {
		return (getLanguage() == Language.GERMAN ? "&ndash;" : "/");
	}

	public String bodyEnd() {
		return bodyEnd(copyrightMessage(1995) + "<br/>");
	}

	public String copyrightMessage(final int startYear) {
		return "Copyright &copy; " + startYear + rangeTo() + "2024 "
				+ ((PageRef.IMPRINT == thisPageRef) ? AUTHOR : link(PageRef.IMPRINT, null, AUTHOR, null, null, true));
	}

	public NavBarEntry createNavBarEntry(final int depth, final String title, final String title_en_US, final PageRef ref, final String liClass, final String iconPath) {
		EnumSet<Language> langs = null;
		if (null != ref) {
			langs = EnumSet.noneOf(Language.class);
			if (null != ref.link) {
				langs.add(Language.GERMAN);
			}
			if (null != ref.link_en_US) {
				langs.add(Language.ENGLISH);
			}
		}
		return new NavBarEntry(depth, title, title_en_US, null, null, ref, liClass, iconPath, langs);
	}

	public final NavBarEntry PROFESSION	   = createNavBarEntry(1, "Beruflich", "Profession",  PageRef.PROFESSION, LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);

	public final NavBarEntry RECIPES	   = createNavBarEntry(2, null, null, PageRef.RECIPES, 	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry MUSIC		   = createNavBarEntry(2, null, null, PageRef.MUSIC, 	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry POEMS		   = createNavBarEntry(2, null, null, PageRef.POEMS, 	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry POEMSALEAT	   = createNavBarEntry(3, "aleatorische", "aleatoric", PageRef.POEMSALEAT, LINK_CLASS_LEVEL1_LAST, NavBarEntry.USE_REF);
	public final NavBarEntry POEMALEAT_WEB = createNavBarEntry(3, null, null, PageRef.POEMALEAT_DYNAMIC, LINK_CLASS_LEVEL1_LAST, NavBarEntry.USE_REF);
	public final NavBarEntry QUOTES		   = createNavBarEntry(2, null, null, PageRef.QUOTES,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry QUOTCOLL	   = createNavBarEntry(3, null, null, PageRef.QUOTCOLL,	LINK_CLASS_LEVEL1_LAST, NavBarEntry.USE_REF);
	public final NavBarEntry PICTURES      = createNavBarEntry(2, null, null, PageRef.PICTURES,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);

	public final NavBarEntry PROGRAMS_HOME = createNavBarEntry(2, "Programmieren", "Programming", PageRef.PROGRAMS_HOME, LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry ATARI8BIT     = createNavBarEntry(2, null, null, PageRef.ATARI8BIT, 	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry CONQUEST_APP  = createNavBarEntry(2, null, null, PageRef.CONQUEST_APP,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry DEINBAC_APP   = createNavBarEntry(2, null, null, PageRef.DEINBAC_APP,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry EMPIRE_APP    = createNavBarEntry(2, null, null, PageRef.EMPIRE_APP,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry KITTEN 	   = createNavBarEntry(2, "Kitten", "Kitten", null, LINK_CLASS_LEVEL0, "/resources/Sitemap.Kitten.png");
	public final NavBarEntry KITTEN_WEB    = createNavBarEntry(2, null, null, PageRef.KITTEN_WEB,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry KITTEN_APP    = createNavBarEntry(2, null, null, PageRef.KITTEN_APP,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry PACMAN 	   = createNavBarEntry(2, "Pacman", "Pacman", null, LINK_CLASS_LEVEL0, "/resources/Sitemap.Pacman.png");
	public final NavBarEntry PACMAN_WEB    = createNavBarEntry(2, null, null, PageRef.PACMAN_WEB,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry PACMAN_APP    = createNavBarEntry(2, null, null, PageRef.PACMAN_APP,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry TROFF_APP     = createNavBarEntry(2, null, null, PageRef.TROFF_APP,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);

	public final NavBarEntry DSA_HOME	   = createNavBarEntry(2, "Rollenspielen", "Role playing games", PageRef.DSA_HOME, LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_TOOL      = createNavBarEntry(2, "DSATool",   "The Dark Eye tool", null,         LINK_CLASS_LEVEL0, "/resources/Sitemap.DSATool.png");
	public final NavBarEntry DSA_TOOL_WEB  = createNavBarEntry(3, null, null, PageRef.DSA_TOOL_WEB,      LINK_CLASS_LEVEL1, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_TOOL_APP  = createNavBarEntry(3, null, null, PageRef.DSA_TOOL_APP, 		LINK_CLASS_LEVEL1_LAST, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_ADVENTURE = createNavBarEntry(2, "Abenteuer", "Adventures", null,       LINK_CLASS_LEVEL0, "/resources/Sitemap.Adventure.png");
	public final NavBarEntry DSA_URSUL     = createNavBarEntry(3, null, null, PageRef.DSA_URSUL,         LINK_CLASS_LEVEL1, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_DRACHENTAL= createNavBarEntry(3, null, null, PageRef.DSA_DRACHENTAL,    LINK_CLASS_LEVEL1, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_WIZVALLEY = createNavBarEntry(3, null, null, PageRef.DSA_WIZVALLEY,     LINK_CLASS_LEVEL1_LAST, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_HERO      = createNavBarEntry(2, "Heldenberichte", "Heroes' life stories", null, LINK_CLASS_LEVEL0, "/resources/Sitemap.Character.png");
	public final NavBarEntry DSA_RUMRAGH   = createNavBarEntry(3, null, null, PageRef.DSA_RUMRAGH,       LINK_CLASS_LEVEL1, NavBarEntry.USE_REF);
	public final NavBarEntry DSA_HESINOR   = createNavBarEntry(3, null, null, PageRef.DSA_HESINOR,       LINK_CLASS_LEVEL1_LAST, NavBarEntry.USE_REF);

	public final NavBarEntry BOOKMARK	   = createNavBarEntry(2, null, null, PageRef.BOOKMARK,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);
	public final NavBarEntry IMPRINT	   = createNavBarEntry(2, null, null, PageRef.IMPRINT,	LINK_CLASS_LEVEL0, NavBarEntry.USE_REF);

	/** The navigation tree for the private subdomain. */
	final NavBarEntry[] navBarEntriesTotalNew = new NavBarEntry[]{
			createNavBarEntry(0, "Site Navigation", null, null, null, null),
			createNavBarEntry(1, "Peter Diefenbach:", null,  null, LINK_CLASS_HEADING, "/resources/Sitemap.png"),
			PROFESSION,
			createNavBarEntry(1, "Privat",    "Private", null,          LINK_CLASS_LEVEL0, "/resources/Sitemap.Private.png"),
			createNavBarEntry(2, "Übersicht", "Overview", PageRef.HOME, LINK_CLASS_LEVEL0, "/resources/Sitemap.Overview.png"),
//			createNavBarEntry(2, "Programme", "Programs", null,         LINK_CLASS_LEVEL0, "/resources/Sitemap.Programming.png"),
			MUSIC,
			POEMS,
			POEMSALEAT,
			RECIPES,
//			createNavBarEntry(2, null, null,        PageRef.VIERSCHWINGENSAGA, null, null),
			PICTURES,
			QUOTES,
			QUOTCOLL,
//			createNavBarEntry(3, "Schmuddelkino", "adult cinema", PageRef.KINOATLANTIK, null, null),

			createNavBarEntry(1, "Programme",    "Programs", null,          LINK_CLASS_LEVEL0, "/resources/Sitemap.Programs.png"),
			createNavBarEntry(2, "Übersicht", "Overview", PageRef.PROGRAMS_HOME, LINK_CLASS_LEVEL0, "/resources/Sitemap.Overview.png"),
			KITTEN_WEB,
			KITTEN_APP,
			TROFF_APP,
			EMPIRE_APP,
			CONQUEST_APP,
			PACMAN_WEB,
			PACMAN_APP,
			DEINBAC_APP,
			ATARI8BIT,

			createNavBarEntry(1, "Rollenspiele", "RPGs", null,          LINK_CLASS_LEVEL0, "/resources/Sitemap.RPG.png"),
			createNavBarEntry(2, "Übersicht", "Overview", PageRef.DSA_HOME, LINK_CLASS_LEVEL0, "/resources/Sitemap.Overview.png"),
			DSA_TOOL,
			DSA_TOOL_WEB,
			DSA_TOOL_APP,
//			createNavBarEntry(3, "DSA Charaktere", PageRef.DSA_CHAR, null, null),
			DSA_ADVENTURE,
			DSA_URSUL,
			DSA_DRACHENTAL,
			DSA_WIZVALLEY,
			DSA_HERO,
			DSA_RUMRAGH,
			DSA_HESINOR,

			createNavBarEntry(1, "Allgemeines", "Common", null, LINK_CLASS_LEVEL0, "/resources/Sitemap.Info.png"),
			BOOKMARK,
			IMPRINT,
			createNavBarEntry(1, "Sprache", "Language", null, LINK_CLASS_LANGSELECT, "/resources/Sitemap.Locales.png"),
			createNavBarEntry(2, Language.GERMAN.text,  Language.GERMAN.text,  null, Language.GERMAN.htmlLangKey,  Language.GERMAN.flagFilepath),
			createNavBarEntry(2, Language.ENGLISH.text, Language.ENGLISH.text, null, Language.ENGLISH.htmlLangKey, Language.ENGLISH.flagFilepath),
	};

	public String javaHead(final String programType) {
		final StringBuffer s = new StringBuffer();
		addLine(s, 0, head("Peter Diefenbach: " + thisPageRef.getTitle(language), "Game, Java, Application", 
				thisPageRef.getTitle(language) + (null == programType ? "" : " " + programType) + " by Peter Diefenbach"));
		return s.toString();
	}

	public enum HowToExecMode {
		HOW_TO_START,
		FOR_PROGRAMMERS,
		OTHER_IMPLS,
		MANUAL_INTRO,
		ALL
	};

	/**
	 * 
	 * @param mode defines what to include in the HTML snippet
	 * @param jarRef the jar file to download
	 * @param webRef if not <code>null</code>, the web (JavaScript or JSP) implementation of this program
	 * @param atariRef if not <code>null</code>, the Atari8bit implementation of this program
	 * @param codeRemark1 if not <code>null</code>, this HTML snippet is inserted into the section for programmers
	 * @param codeRemark2 if not <code>null</code>, this HTML snippet is appended to the section for programmers
	 * @return a HTML snippet
	 */
	public String javaHowToExec(final HowToExecMode mode, final PageRef jarRef, final PageRef webRef, final PageRef atariRef, final String codeRemark1, final String codeRemark2) {
		final StringBuffer s = new StringBuffer();
		final String h2class = (jarRef == PageRef.DSATOOL_JAR ? htmlAttribute(TextFormatter.CLASS, "rpg") : "");
		addLine(s, 0, "");
		switch(language) {
		case ENGLISH:
			if ((mode == HowToExecMode.HOW_TO_START) || (mode == HowToExecMode.ALL)) {
				addLine(s, 2, "<h2" + h2class + ">How to start the program</h2>");
				addLine(s, 2, "<ol>");
				addLine(s, 3, "<li>");
				addLine(s, 4, "You need a Java Runtime Environment on your computer.");
				addLine(s, 4, "The easiest way is to get it from " + link(PageRef.JAVA_COM) + "."); 
				addLine(s, 4, "There you can also check if you already have it installed.");
				addLine(s, 3, "</li>");
				addLine(s, 3, "<li>");
				addLine(s, 4, "Download the program " + link(jarRef) + "."); 
				addLine(s, 3, "</li>");
				addLine(s, 3, "<li>");
				addLine(s, 4, "Start it by double-clicking the downloaded file.");
				addLine(s, 3, "</li>");
				addLine(s, 2, "</ol>");
				addLine(s, 2, "");
			}
			if ((mode == HowToExecMode.FOR_PROGRAMMERS) || (mode == HowToExecMode.ALL)) {
				addLine(s, 2, "<h2" + h2class + ">A word to the programmers among you</h2>");
				addLine(s, 2, "<p>");
				addLine(s, 3, "If you want, you can"); 
				addLine(s, 3, link(PageRef.DOWNLOAD_JAVA, "download the source code") + ".");
				addLine(s, 2, "</p>");
				if (null != codeRemark1) {
					addLine(s, 2, codeRemark1);
				}
				addLine(s, 2, "<p>");
				addLine(s, 3, "Feel free to use the files in your own projects, as long as you name me and all other contributors to the files you took in use from me.");
				addLine(s, 3, "<br/>If you apply changes to my files, please inform me of the result, maybe there are more interested in your improvements.");
				addLine(s, 2, "</p>");
				if (null != codeRemark2) {
					addLine(s, 2, codeRemark2);
				}
				addLine(s, 2, "");
			}
			if ((mode == HowToExecMode.OTHER_IMPLS) || (mode == HowToExecMode.ALL)) {
				if (null != webRef) {
					addLine(s, 2, "<h2" + htmlAttribute(TextFormatter.ID, "Help.OtherImplementations") + h2class + ">Other Implementations</h2>");
					addLine(s, 2, "<p>");
					final String atari = ((null == atariRef) ? "" : 
						", and the "
						 + link(atariRef, "original game for 8-bit Atari home computers"));
					addLine(s, 3, "There is also a " + link(webRef, "web implementation") + atari + ".");
					addLine(s, 2, "</p>");
				}
			}
			if ((mode == HowToExecMode.MANUAL_INTRO) || (mode == HowToExecMode.ALL)) {
				addLine(s, 2, "<hr/>");
				addLine(s, 2, "<p>");
				addLine(s, 3, "And here's the manual:");
				addLine(s, 2, "</p>");
			}
			break;
		case GERMAN:
		default:
			if ((mode == HowToExecMode.HOW_TO_START) || (mode == HowToExecMode.ALL)) {
				addLine(s, 2, "<h2>Wie starte ich das Programm?</h2>");
				addLine(s, 2, "<ol>");
				addLine(s, 3, "<li>");
				addLine(s, 4, "Installieren Sie ein Java Runtime Environment auf Ihrem Rechner.");
				addLine(s, 4, "Am einfachsten bekommen Sie das von  " + link(PageRef.JAVA_COM) + ".");
				addLine(s, 4, "Dort können Sie auch prüfen, ob Sie es vielleicht schon haben.");
				addLine(s, 3, "</li>");
				addLine(s, 3, "<li>");
				addLine(s, 4, "Laden Sie das Programm " + link(jarRef) + " herunter."); 
				addLine(s, 3, "</li>");
				addLine(s, 3, "<li>");
				addLine(s, 4, "Starten Sie es dann mit einem Doppelklick auf die heruntergeladene Datei.");
				addLine(s, 3, "</li>");
				addLine(s, 2, "</ol>");
				addLine(s, 2, "");
			}
			if ((mode == HowToExecMode.FOR_PROGRAMMERS) || (mode == HowToExecMode.ALL)) {
				addLine(s, 2, "<h2>Ein Wort an Programmierer</h2>");
				addLine(s, 2, "<p>");
				addLine(s, 3, "Wenn Sie wollen, können Sie das"); 
				addLine(s, 3, link(PageRef.DOWNLOAD_JAVA, "Projekt mit Quellcode herunterladen") + ".");
				addLine(s, 2, "</p>");
				if (null != codeRemark1) {
					addLine(s, 2, codeRemark1);
				}
				addLine(s, 2, "<p>");
				addLine(s, 3, "Sie dürfen die Dateien jederzeit in eigenen Projekten mitbenutzen,");
				addLine(s, 3, "solange ich und alle, die bisher zum von mir übernommenen Code beigetragen haben,");
				addLine(s, 3, "im Produkt genannt werden.");
				addLine(s, 3, "(Ich denke, das ist ein Gebot der Höflichkeit.)");
				addLine(s, 3, "<br/>Falls Sie Änderungen an den Quelldateien vornehmen, informieren Sie mich bitte,");
				addLine(s, 3, "vielleicht sind Ihre Verbesserungen ja von allgemeinem Interesse.");
				addLine(s, 2, "</p>");
				if (null != codeRemark2) {
					addLine(s, 2, codeRemark2);
				}
				addLine(s, 2, "");
			}
			if ((mode == HowToExecMode.OTHER_IMPLS) || (mode == HowToExecMode.ALL)) {
				if (null != webRef) {
					addLine(s, 2, "<h2" + htmlAttribute(TextFormatter.ID, "Help.OtherImplementations") + h2class + ">Andere Implementierungen</h2>");
					addLine(s, 2, "<p>");
					final String atari = ((null == atariRef) ? "" : 
							" und das "
							 + link(atariRef, "Originalspiel für 8-bit Atari Heimcomputer"));
					addLine(s, 3, "Es gibt auch eine " + link(webRef, "Web-Version") + atari + ".");
					addLine(s, 2, "</p>");
				}
			}
			if ((mode == HowToExecMode.MANUAL_INTRO) || (mode == HowToExecMode.ALL)) {
				addLine(s, 2, "<hr/>");
				addLine(s, 2, "<p>");
				addLine(s, 3, "Und hier die Anleitung:");
				addLine(s, 2, "</p>");
			}
			break;
		}
		return s.toString();
	}

	/**
	 * 
	 * @param jarRef the jar file to download
	 * @return a HTML snippet
	 */
	public String javaHowToExec(final PageRef jarRef) {
		return javaHowToExec(HowToExecMode.ALL, jarRef, null, null, null, null);
	}

	/**
	 * Reads all parameters for the dynamic page.
	 * As a side effect, a string describing in human words if something went wrong is stored via {@link #setErrorMessages(String)}.
	 * @param request the request containing the parameters
	 */
	public void readParameters(final HttpServletRequest request) {
		final StringBuffer errorMsgs = new StringBuffer();
		itemToShow = request.getParameter(ITEM_KEY);
		// Currently we can't do error handling, all we read are strings handled by XSLT.
		// So we always return an empty string.
		setErrorMessages(errorMsgs.toString());
	}

	protected void addLine(final StringBuffer s, final int indent, final String line) {
		s.append(TextFormatter.repeat("\t", indent));
		s.append(line);
		s.append(TextFormatter.NL);
	}

	protected String getLocalHomePath() { return localHomePath; }
	public void setLocalHomePath(final String localHomePath) { this.localHomePath = localHomePath; }

	public PageRef getThisPageRef() { return thisPageRef; }
	public void setThisPageRef(final PageRef thisPageRef) { this.thisPageRef = thisPageRef; }

	public String getSubdomainPrefixTo(final Subdomain destinationSubdomain) {
		return getSubdomain().getPrefixTo(destinationSubdomain);
	}

	public Subdomain getSubdomain() { return thisPageRef.getSubdomain(); }

	protected Language getLanguage() {
		return language;
	}

	private void setLanguage(final Language language) {
		this.language = language;
	}

	public boolean setLanguage(final String langKey) {
		boolean ret = false;
		for (Language lang : Language.values()) {
			if (lang.htmlLangKey.equals(langKey)) {
				setLanguage(lang);
				ret = true;
				break;
			}
		}
		return ret;
	}

	public String getItemToShow() {
		return itemToShow;
	}

	public void setItemToShow(final String itemToShow) {
		this.itemToShow = itemToShow;
	}

	public void setServletContext(final ServletContext servletContext) {
		getResourceFinder().setPathToJavaRootPackage(servletContext.getRealPath("."));
	}

	public ResourceFinder getResourceFinder() {
		if (null == rf) {
			rf = new ResourceFinder("html/resources");
		}
		return rf;
	}

	public String getErrorMessages() {
		return errorMessages;
	}

	protected void setErrorMessages(final String errorMessages) {
		this.errorMessages = errorMessages;
	}

	private Language language = Language.GERMAN;
	private String localHomePath = ".";
	private PageRef thisPageRef;
	private DateFormat df;
	private ResourceFinder rf;
	private String itemToShow;
	private String errorMessages;
}
