package de.ped.troff.client.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.BorderFactory;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.border.Border;

import de.ped.tools.Idable;
import de.ped.tools.Messages;
import de.ped.tools.PlayerColor;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.JIconComboBox;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PlayerColorEditor;
import de.ped.tools.gui.PropertyElement;
import de.ped.tools.gui.TableRowNumberRenderer;
import de.ped.tools.gui.TableStringRenderer;
import de.ped.tools.gui.TableTools;
import de.ped.tools.gui.PlayerColorRenderer;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * The dialog presenting the game properties for a run.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffGamePropertiesDialog extends PedJDialog implements ItemListener {
	private static final long serialVersionUID = 1L;

	public static final int MAX_NUMBER_OF_HUMAN_PLAYERS = 4;

	protected TroffGamePropertiesDialog(
			final ApplicationMainWindow mainWindow, 
			final TroffGameProperties aModel) {
		super(mainWindow, mainWindow);
		model = (TroffGameProperties) aModel.clone();
		final boolean isEditable = true; // To compare code easier with Empire version

		final Messages messages = getMessages();
		setTitle(messages.getText("GamePropertiesWindow.Title"));
		choiceBoard = mainWindow.createComboBox(model.getBoardInfos(), true);
		choiceSpeed = mainWindow.createComboBox(model.getSpeedInfos(), false);
		choiceAbsoluteView = mainWindow
				.createComboBox(new String[] { "GameProperties.View.Absolute", "GameProperties.View.Players" });
		final String[] numOfPlayersStrings = new String[TroffGameProperties.MAX_NUMBER_OF_PLAYERS];
		for (int i = 0; i < numOfPlayersStrings.length; i++) {
			numOfPlayersStrings[i] = "GameProperties.NumberOfPlayers." + Integer.toString(i + 1);
		}
		choiceNumberOfPlayers = mainWindow.createComboBox(numOfPlayersStrings, false);
		setPropertyElements(new PropertyElement[] {
				new PropertyElement("GameProperties.Boards", choiceBoard),
				new PropertyElement("GameProperties.View", choiceAbsoluteView),
				new PropertyElement("GameProperties.Speed", choiceSpeed),
				new PropertyElement("GameProperties.NumberOfPlayers", choiceNumberOfPlayers),
		});
		setTitlePanel(messages.getString("GamePropertiesWindow.Heading"));
		final JPanel mapPropertiesPanel = new JPanel();
		final Border lineBorder = BorderFactory.createLineBorder(Color.gray);
		mapPropertiesPanel.setBorder(BorderFactory.createTitledBorder(lineBorder, messages.getText("GameProperties.World")));
		mapPropertiesPanel.setLayout(new BorderLayout());
		{
			final JPanel mainPropsPanel = new JPanel();
			final GridBagLayout gridbag = new GridBagLayout();
			final GridBagConstraints c = new GridBagConstraints();
			c.anchor = GridBagConstraints.WEST; // All elements are left bound
												// to their space
			mainPropsPanel.setLayout(gridbag);
			addAllPropertyElementsTo(mainPropsPanel, c);
			mapPropertiesPanel.add(mainPropsPanel, BorderLayout.CENTER);
		}
		final JPanel playersPanel = new JPanel();
		{
			playersPanel.setName("Players");
			playersPanel.setLayout(new BorderLayout());
			playersPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.gray), messages.getText("GameProperties.Players")));
			tableModel = new TroffPlayersSelectionTableModel(this, model, messages, isEditable);
			table = new JTable(tableModel);
			table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			table.setDefaultRenderer(PlayerColor.class, new PlayerColorRenderer(messages));
			table.setDefaultEditor(PlayerColor.class, new PlayerColorEditor(messages));
			table.setDefaultRenderer(TroffPlayerType.class, new TroffPlayerTypeRenderer(messages));
			table.setDefaultEditor(TroffPlayerType.class, new TroffPlayerTypeEditor(messages));
			table.setDefaultRenderer(TroffPlayerView.class, new TroffPlayerViewRenderer(messages));
			table.setDefaultEditor(TroffPlayerView.class, new TroffPlayerViewEditor(messages));
			table.setDefaultRenderer(TroffPlayerMode.class, new TroffPlayerModeRenderer(messages));
			table.setDefaultEditor(TroffPlayerMode.class, new TroffPlayerModeEditor(messages));
			table.setDefaultRenderer(Integer.class, new TableRowNumberRenderer());
			table.setDefaultRenderer(String.class, new TableStringRenderer());
			TableTools.adjustColumns(table, 5, mainWindow.getApplicationEnvironment().resources());
			ApplicationMainWindow.adjustRowHeightsToRenderers(table);
			final JPanel tablePanel = new JPanel();
			tablePanel.setLayout(new BorderLayout());
			tablePanel.add(table, BorderLayout.CENTER);
			tablePanel.add(table.getTableHeader(), BorderLayout.NORTH);
			playersPanel.add(tablePanel, BorderLayout.CENTER);
		}
		mainPanel = createMainPanel();
		{
			final BorderLayout mainPanelLayout = new BorderLayout();
			mainPanel.setLayout(mainPanelLayout);
			mainPanel.add(mapPropertiesPanel, BorderLayout.CENTER);
			mainPanel.add(playersPanel, BorderLayout.SOUTH);
		}
		getContentPane().add(new JScrollPane(mainPanel), BorderLayout.CENTER);
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		addHelpButton("Help.Properties");
		finishLayout();
		updateUIFieldsBasedOnProperties();
		choiceNumberOfPlayers.addItemListener(this);
		choiceAbsoluteView.addItemListener(this);
	}

	/**
	 * Invoked when an item state change event occurs.
	 * @param e the event to handle
	 */
	@Override
	public void itemStateChanged(final ItemEvent e) {
		final Object source = e.getSource();
		if ((source == choiceNumberOfPlayers) || (source == choiceAbsoluteView)) {
			updateUIFields();
		}
	}

	@Override
	public void updateUIFieldsBasedOnProperties() {
		choiceBoard.setSelectedIndex(model.getBoardIndex());
		choiceAbsoluteView.setSelectedIndex(model.isAbsoluteView() ? 0 : 1);
		choiceSpeed.setSelectedIndex(model.getSpeedIndex());
		choiceNumberOfPlayers.setSelectedIndex(model.getNumberOfPlayers() - 1);
		for (int i = 0; i < TroffGameProperties.MAX_NUMBER_OF_PLAYERS; i++) {
			tableModel.setValueAt(model.getPlayerType(i), i, TroffPlayersSelectionTableModel.COL_IDX_TYPE);
			tableModel.setValueAt(model.getPlayerName(i), i, TroffPlayersSelectionTableModel.COL_IDX_NAME);
			tableModel.setValueAt(model.getPlayerView(i), i, TroffPlayersSelectionTableModel.COL_IDX_VIEW);
			tableModel.setValueAt(model.getPlayerMode(i), i, TroffPlayersSelectionTableModel.COL_IDX_MODE);
		}
		updateUIFields();
	}

	protected void updateUIFields() {
		// Let's react live in the table on changes to number of players box
		table.revalidate();
		table.repaint();
	}

	@Override
	public void readUIFieldsIntoProperties() {
		model.setBoardIndex(choiceBoard.getSelectedIndex());
		model.setAbsoluteView(0 == choiceAbsoluteView.getSelectedIndex());
		model.setSpeedIndex(choiceSpeed.getSelectedIndex());
		model.setNumberOfPlayers(choiceNumberOfPlayers.getSelectedIndex() + 1);
		for (int i = 0; i < TroffGameProperties.MAX_NUMBER_OF_PLAYERS; i++) {
			model.setPlayerType(i, ((Idable) tableModel.getValueAt(i, TroffPlayersSelectionTableModel.COL_IDX_TYPE)).getId());
			model.setPlayerView(i, ((Idable) tableModel.getValueAt(i, TroffPlayersSelectionTableModel.COL_IDX_VIEW)).getId());
			model.setPlayerMode(i, ((Idable) tableModel.getValueAt(i, TroffPlayersSelectionTableModel.COL_IDX_MODE)).getId());
			model.setPlayerName(i,  (String) tableModel.getValueAt(i, TroffPlayersSelectionTableModel.COL_IDX_NAME));
		}
	}

	public TroffGameProperties getModel() {
		return model;
	}

	public void enterDialog(final TroffGameProperties aModel) {
		model = (TroffGameProperties) aModel.clone();
		enterDialog();
	}

	PropertyElement[] getPlayerProperties() {
		final Messages messages = getMessages();
		return new PropertyElement[] {
			new PropertyElement("GameProperties.PlayerType", new TroffPlayerTypeRenderer(messages)),
			new PropertyElement("GameProperties.PlayerMode", new TroffPlayerModeRenderer(messages)),
			new PropertyElement("GameProperties.PlayerView", new TroffPlayerViewRenderer(messages)),
			new PropertyElement("GameProperties.Players.Color.Heading", new PlayerColorRenderer(messages)),
			new PropertyElement("GameProperties.PlayerName", new JTextField()),
		};
	};

	private final JPanel mainPanel;
	private JIconComboBox<?> choiceBoard;
	JIconComboBox<?> choiceNumberOfPlayers;
	JIconComboBox<?> choiceAbsoluteView;
	private JIconComboBox<?> choiceSpeed;
	private final JTable table;

	private TroffGameProperties model;
	private final TroffPlayersSelectionTableModel tableModel;
}
