package de.ped.troff.client.gui;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.util.Locale;

import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

import de.ped.tools.ApplicationEnvironment;
import de.ped.tools.Assert;
import de.ped.tools.TextDocumentBuilder;
import de.ped.tools.TextFormatter;
import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.PlayerColor;
import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.CheckBoxMenuItemAction;
import de.ped.tools.gui.JInfoDialog;
import de.ped.tools.gui.PedAction;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;
import de.ped.tools.gui.UserGameActionInput;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.sound.SoundClip;
import de.ped.troff.client.logic.ClientThread;
import de.ped.troff.client.logic.ThreadBundle;
import de.ped.troff.middleware.Environment;
import de.ped.troff.middleware.ServerInfo;
import de.ped.troff.middleware.ServerSelectionModel;
import de.ped.troff.server.TroffRoundsManagerImpl;
import de.ped.troff.server.logic.TroffPlayer;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * The main window, including menu.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffMainWindow extends ApplicationMainWindow {
	private static final long serialVersionUID = 1L;

	public TroffMainWindow(final ApplicationEnvironment appEnv) {
		super(appEnv);
		logger.setToShowThread(true);
		logger.setLogLevel(Level.INFO);
		if (getApplicationEnvironment().isSilent()) {
			soundOnOffAction.setOn(false);
		}
		setResizable(false);
		getContentPane().setLayout(new BorderLayout());
		initSounds();
		initColors();

		final JPanel textPanel = new JPanel();
		textPanel.setLayout(new BorderLayout());
		statusLabel = new JLabel();
		textPanel.add(statusLabel, BorderLayout.CENTER);

		mainCanvas = new TroffMainCanvas(this);
		getContentPane().add(mainCanvas, BorderLayout.CENTER);
		getContentPane().add(textPanel, BorderLayout.SOUTH);
		setJMenuBar(createMenuBar());
		pack(); // Resize the window to the preferred size of its components.
		switchLocale(getMessages().getCurrentLocale());

		addKeyListener(this);
		logger.debug("added " + this.getClass().getSimpleName() + " as key listener.");
		updateServerModelChecked(new ServerSelectionModel());

		// TODO: Remove those when server functionality is working well.
		selectServerGroupAction.setEnabled(true);
		selectGroupAction.setEnabled(false);
	}

	private PedAction selectServerGroupAction = new PedAction("Action.SelectServer", TroffMainWindow.this, null, true);
	private PedAction selectGroupAction = new PedAction("Action.SelectGroup", TroffMainWindow.this, null, true);
	private PedAction sendGroupMessageAction = new PedAction("Action.CreateMessage", TroffMainWindow.this, null, true);
	private PedAction pauseAction = new PedAction("Action.Pause", TroffMainWindow.this, ResourceFinder.Folder.COMMON, false);
	private PedAction resumeAction = new PedAction("Action.Resume", TroffMainWindow.this, ResourceFinder.Folder.COMMON, false);
	private CheckBoxMenuItemAction soundOnOffAction = new CheckBoxMenuItemAction(true, "Action.SoundOnOff", TroffMainWindow.this, ResourceFinder.Folder.COMMON);

	@Override
	public void pedAbstractActionPerformed(final PedAction action, final ActionEvent event) {
		super.pedAbstractActionPerformed(action, event);
		if (action == selectServerGroupAction) {
			selectServerGroup();
		} else if (action == selectGroupAction) {
			selectGroup();
		} else if (action == sendGroupMessageAction) {
			sendGroupMessage();
		} else if (action == pauseAction) {
			pauseOrResume(false);
		} else if (action == resumeAction) {
			pauseOrResume(true);
		} else if (action == soundOnOffAction) {
			// Nothing to do
		}
	}

	private JMenuBar createMenuBar() {
		final JMenuBar menuBar = createJMenuBar();
		final JMenu menuFile = createJMenu("Menu.File");
		createActionItem(menuFile, selectServerGroupAction);
		createActionItem(menuFile, selectGroupAction);
		createActionItem(menuFile, sendGroupMessageAction);
		menuFile.addSeparator();
		createActionItem(menuFile, startNewGameAction);
		menuFile.addSeparator();
		menuFile.add(createLocalesMenu());
		menuFile.addSeparator();
		createActionItem(menuFile, quitAction);
		menuBar.add(menuFile);
		final JMenu menuRun = createJMenu("Menu.Run");
		createActionItem(menuRun, pauseAction);
		createActionItem(menuRun, resumeAction);
		createActionItem(menuRun, soundOnOffAction);
		menuBar.add(menuRun);
		final JMenu menuHelp = createJMenu("Menu.Help");
		createActionItem(menuHelp, helpAction);
		createActionItem(menuHelp, aboutAction);
		menuBar.add(menuHelp);
		try {
			menuBar.setHelpMenu(menuHelp);
		} catch (final Throwable e) {
			// Nothing to do - this feature is a nice to have.
		}
		menuBar.add(menuHelp);
		return menuBar;
	}

	@Override
	public boolean isRunning() {
		return getClientThread().isRunning();
	}

	protected void sendGroupMessage() {
		final boolean wasRunning = isRunning();
		pauseOrResume(false);
		final CreateMessageDialog dialog = new CreateMessageDialog(this);
		dialog.enterDialog();
		if (dialog.wasAcceptedDialogChanges()) {
			dialog.readUIFieldsIntoProperties();
			final String message = dialog.getMessage();
			final ClientThread model = getClientThread();
			if (null != model) {
				model.requestSendMessage(message);
			}
		}
		pauseOrResume(wasRunning);
	}

	protected void selectServerGroup() {
		final boolean wasRunning = isRunning();
		pauseOrResume(false);
		boolean isServerOkay = selectServer();
		if (isServerOkay) {
			selectGroup();
		}
		pauseOrResume(wasRunning);
	}

	protected boolean selectServer() {
		ServerSelectionModel oldModel = threadBundle.getServerSelectionModel();
		final ServerSelectionDialog dialog = new ServerSelectionDialog(this, oldModel);
		boolean isOkayToLeave = false;
		boolean isOkay = false;
		while (!isOkayToLeave) {
			dialog.enterDialog();
			logger.trace("Server selection dialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
			if (dialog.wasAcceptedDialogChanges()) {
				dialog.readUIFieldsIntoProperties();
				isOkayToLeave = updateServerModelChecked(dialog.getModel());
				try {
					Thread.sleep(100L);
				} catch (final InterruptedException e) {
					// Nothing to do
				}
				if (isOkayToLeave) {
					oldModel = threadBundle.getServerSelectionModel();
				} else {
					updateServerModelChecked(oldModel);
				}
				isOkay = true;
			} else {
				// Cancel is always a reason to leave this dialog
				isOkayToLeave = true;
			}
		}
		return isOkay;
	}

	protected PedJDialog createConnectExceptionDialog() {
		final Messages messages = getMessages();
		final String title = messages.getString("ServerSelectionWindow.Title");
		final String message = messages.getString("ServerSelectionWindow.ConnectException");
		return new JInfoDialog(this, title, message, JOptionPane.ERROR_MESSAGE);
	}

	protected boolean updateServerModelChecked(final ServerSelectionModel serverSelectionModel) {
		boolean ret = true;
		try {
			updateServerModel(serverSelectionModel);
		} catch (final IOException e) {
			logger.info("Exception", e);
			final PedJDialog connectExceptionDialog = createConnectExceptionDialog();
			connectExceptionDialog.setVisible(true);
			ret = false;
		}
		updateGroupModel(null);
		getClientThread().requestStartGame(props);
		return ret;
	}

	public void updateServerModel(final ServerSelectionModel serverSelectionModel) throws IOException {
		threadBundle.setServerSelectionModel(serverSelectionModel);
		logger.debug(serverSelectionModel.toString());
		threadBundle.start();
		while (!threadBundle.isStarted()) {
			try {
				Thread.sleep(100L);
			} catch (final InterruptedException e) {
				// Nothing to do
			}
		}
		final ClientThread model = getClientThread();
		final ServerSelectionModel ssmodel = getServerSelectionModel();
		if (null == model) {
			logger.error("Client model is null!");
		} else if (null == ssmodel) {
			logger.error("Server selection model is null!");
		} else {
			model.requestLogin(ssmodel.getClientName(), (byte) ssmodel.getNumberOfPlayers());
		}
	}

	protected void selectGroup() {
		final ServerSelectionModel ssmodel = getServerSelectionModel();
		final ClientThread model = getClientThread();
		// model.requestLogout();
		model.requestLogin(ssmodel.getClientName(), (byte) ssmodel.getNumberOfPlayers());
		model.requestServerStatus();
		ServerInfo serverStatus = null;
		logger.debug("Waiting for ServerStatus");
		do {
			// TODO: Bring up dialog to wait for server, and allow user to cancel
			serverStatus = model.getServerStatus();
			if (null == serverStatus) {
				try {
					Thread.sleep(500L);
				} catch (final InterruptedException e) {
					// Nothing to do
				}
			}
		} while (null == serverStatus);
		logger.debug("Got ServerStatus");
		final Messages messages = getMessages();
		final String columnHeadingGroupName = messages.getText("GroupSelectionWindow.GroupsList.Group");
		final String columnHeadingGroupNumberOfPlayers = messages
				.getText("GroupSelectionWindow.GroupsList.NumberOfPlayers");
		final GroupSelectionModel gsmodel = new GroupSelectionModel(serverStatus, columnHeadingGroupName,
				columnHeadingGroupNumberOfPlayers);
		final GroupSelectionDialog dialog = new GroupSelectionDialog(this, gsmodel);
		dialog.setTitle(dialog.getTitle() + serverStatus.getName());
		dialog.enterDialog();
		logger.debug("Group selection dialog " + (dialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (dialog.wasAcceptedDialogChanges()) {
			dialog.readUIFieldsIntoProperties();
			updateGroupModel(gsmodel);
			try {
				Thread.sleep(100L);
			} catch (final InterruptedException e) {
				// Nothing to do
			}
		}
	}

	public void updateGroupModel(final GroupSelectionModel gsmodel) {
		final ClientThread model = getClientThread();
		if (null == model) {
			logger.error("Client model is null!");
		} else {
			String groupName = TroffRoundsManagerImpl.DEFAULT_GROUP_NAME;
			if (null != gsmodel) {
				groupName = gsmodel.getSelectedGroupName();
			}
			model.requestJoin(groupName);
		}
	}

	@Override
	protected void startNewGame() {
		boolean wasRunning = isRunning();
		pauseOrResume(false);
		final TroffGamePropertiesDialog dialog = getOrCreateNewGameSelectionDialog();
		final ServerInfo serverInfo = getClientThread().getServerStatus();
		String titlePostfix = "";
		if (null != serverInfo) {
			final I18NStringWithFillIns postfix = new I18NStringWithFillIns("GamePropertiesWindow.Postfix.Text", null,
					new String[] { serverInfo.getName(), serverInfo.getGroupOfClient(), });
			titlePostfix = getMessages().getString(postfix);
		}
		dialog.setTitle(getMessages().getString("GamePropertiesWindow.Title") + titlePostfix);
		dialog.enterDialog(props);
		logger.trace("New game selection dialog " + (newGameSelectionDialog.wasAcceptedDialogChanges() ? "OK" : "Cancel"));
		if (dialog.wasAcceptedDialogChanges()) {
			dialog.readUIFieldsIntoProperties();
			props = (TroffGameProperties) dialog.getModel().clone();
			getClientThread().requestStartGame(props);
		} else {
			pauseOrResume(wasRunning);
		}
	}

	public void startNewGame(final TroffGameProperties props) {
		mainCanvas = new TroffMainCanvas(this);
		getContentPane().add(mainCanvas, BorderLayout.CENTER);
		pack();
		repaint();
	}

	public ServerSelectionModel getServerSelectionModel() {
		return threadBundle.getServerSelectionModel();
	}

	public ClientThread getClientThread() {
		return threadBundle.getClientThread();
	}

	@Override
	public void buildHelpDocument(final TextDocumentBuilder doc) {
		doc.setTitle(getMessages().getString("HelpWindow.Title"));
		doc.setHeading("Help");
		doc.beginDocument();
		doc.addContentPart(2, "Help.Introduction");
		doc.addContentPart(2, "Help.Operation");
		buildInputDescriptions(doc, userGameActionInputs, true);
		buildMenuDescriptions(doc, getJMenuBar());
		final TroffGamePropertiesDialog diag = getOrCreateNewGameSelectionDialog();
		buildPropertyDescriptions(doc, diag);
		PropertyElement[] playerPEs = diag.getPlayerProperties();
		buildPropertyDescriptions(doc, 3, "GameProperties.Players", playerPEs, true);
		doc.endDocument();
	}

	@Override
	protected String getKeyboardActionDescriptionText(final UserGameActionInput ugai) {
		String strActionKey = "";
		switch (ugai.playerAction) {
		case TroffPlayer.TURN_LEFT:
			strActionKey = "GameProperties.PlayerAction.TurnLeft.Text";
			break;
		case TroffPlayer.TURN_RIGHT:
			strActionKey = "GameProperties.PlayerAction.TurnRight.Text";
			break;
		case TroffPlayer.ACTION:
			strActionKey = "GameProperties.PlayerAction.Action.Text";
			break;
		default:
			Assert.assertTrue(true);
		}
		return getMessages().getString(strActionKey);
	}

	@Override
	public void switchLocale(final Locale locale) {
		super.switchLocale(locale);
		final Messages messages = getMessages();
		userGameActionInputs = new UserGameActionInput[]{
			new UserGameActionInput(messages, "Game.Player0L.KeyStroke", KeyEvent.VK_A,         0, TroffPlayer.TURN_LEFT),
			new UserGameActionInput(messages, "Game.Player0R.KeyStroke", KeyEvent.VK_S,         0, TroffPlayer.TURN_RIGHT),
			new UserGameActionInput(messages, "Game.Player0A.KeyStroke", KeyEvent.VK_CAPS_LOCK, 0, TroffPlayer.ACTION),
			new UserGameActionInput(messages, "Game.Player1L.KeyStroke", KeyEvent.VK_LEFT,      1, TroffPlayer.TURN_LEFT),
			new UserGameActionInput(messages, "Game.Player1R.KeyStroke", KeyEvent.VK_RIGHT,     1, TroffPlayer.TURN_RIGHT),
			new UserGameActionInput(messages, "Game.Player1A.KeyStroke", KeyEvent.VK_DOWN,      1, TroffPlayer.ACTION),
			new UserGameActionInput(messages, "Game.Player2L.KeyStroke", KeyEvent.VK_K,         2, TroffPlayer.TURN_LEFT),
			new UserGameActionInput(messages, "Game.Player2R.KeyStroke", KeyEvent.VK_L,         2, TroffPlayer.TURN_RIGHT),
			new UserGameActionInput(messages, "Game.Player2A.KeyStroke", KeyEvent.VK_O,         2, TroffPlayer.ACTION),
			new UserGameActionInput(messages, "Game.Player3L.KeyStroke", KeyEvent.VK_NUMPAD4,   3, TroffPlayer.TURN_LEFT),
			new UserGameActionInput(messages, "Game.Player3R.KeyStroke", KeyEvent.VK_NUMPAD6,   3, TroffPlayer.TURN_RIGHT),
			new UserGameActionInput(messages, "Game.Player3A.KeyStroke", KeyEvent.VK_NUMPAD5,   3, TroffPlayer.ACTION),
		};
		newGameSelectionDialog = null;
		updateStatusText();
		repaint();
	}

	public void update() {
		if (null != mainCanvas) {
			mainCanvas.repaint();
		}
	}

	public void setStatusText(final I18NStringWithFillIns[] msgs) {
		log = msgs;
		updateStatusText();
	}

	protected void updateStatusText() {
		if ((null != mainCanvas) && (null != log)) {
			final String[] strs = new String[log.length];
			for (int i = 0; i < log.length; i++) {
				strs[i] = getMessages().getText(log[i]);
			}
			final String text = TextFormatter.unwrap(strs);
			mainCanvas.setStatusText(text == null ? "" : text);
		}
	}

	@Override
	protected void pauseOrResume(final boolean isToResume) {
		final ClientThread model = getClientThread();
		if (null != model) {
			model.requestPauseOrResumeRun(isToResume);
		}
		if (!isToResume) {
			try {
				Thread.sleep(0);
			} catch (final InterruptedException e) {
				// Nothing to do - that's what we intend to do
			}
		}
	}

	public boolean isSoundOn() {
		return soundOnOffAction.isOn();
	}

	public void playSound(final int eventIndex) {
		if (isSoundOn()) {
			final SoundClip sound = sounds.findSound(eventIndex);
			currentSounds.play(sound, "event " + eventIndex);
		}
	}

	/** Loads or creates the sound clips to play. */
	private void initSounds() {
		sounds = new TroffSounds(getApplicationEnvironment().resources());
		sounds.findSound(0);
	}

	private TroffSounds sounds;

	private void initColors() {
		PlayerColor.init(
				new PlayerColor[] {
						new PlayerColor(0, "GameProperties.Players.Color.0", "FF0000", "800000"), // red
						new PlayerColor(1, "GameProperties.Players.Color.1", "00FF00", "008000"), // green 
						new PlayerColor(2, "GameProperties.Players.Color.2", "00F0F0", "008080"), // cyan
						new PlayerColor(3, "GameProperties.Players.Color.3", "FF00FF", "600060"), // magenta
						new PlayerColor(4, "GameProperties.Players.Color.4", "8080FF", "000040"), // blue
						new PlayerColor(5, "GameProperties.Players.Color.5", "F0F000", "404000"), // yellow
						new PlayerColor(6, "GameProperties.Players.Color.6", "FF8000", "402000"), // orange
						new PlayerColor(7, "GameProperties.Players.Color.7", "808080", "303030"), // grey
				}
		);
	}

	@Override
	public void keyPressed(final KeyEvent keyEvent) {
		logger.debug("keyPressed: " + keyEvent.getKeyCode() + " source=" + keyEvent.getSource());
		boolean isToRepaint = false;
		final ClientThread model = getClientThread();
		for (final UserGameActionInput tsv : userGameActionInputs) {
			if (keyEvent.getKeyCode() == tsv.keyCode) {
				model.requestGameActionInput(tsv.playerIdx, tsv.playerAction);
				isToRepaint = true;
			}
		}
		if (isToRepaint) {
			mainCanvas.repaint();
		}
	}

	public final Logger logger = Logger.getRootLogger();

	public Environment getEnvironment() {
		return environment;
	}

	private final Environment environment = new Environment();

	public UserGameActionInput findUserGameActionInput(final int playerIdx, final int playerAction) {
		return UserGameActionInput.find(userGameActionInputs, playerIdx, playerAction);
	}

	private UserGameActionInput[] userGameActionInputs;

	private TroffGamePropertiesDialog getOrCreateNewGameSelectionDialog() {
		if (null == newGameSelectionDialog) {
			newGameSelectionDialog = new TroffGamePropertiesDialog(this, props);
		}
		return newGameSelectionDialog;
	}

	private TroffGamePropertiesDialog newGameSelectionDialog;

	private ThreadBundle threadBundle = new ThreadBundle(this);

	private TroffGameProperties props = new TroffGameProperties();

	/** The canvas painting the game board. */
	private TroffMainCanvas mainCanvas;

	private SoundClip.RingBuffer currentSounds = new SoundClip.RingBuffer(5);

	/** The label to show status messages. */
	private JLabel statusLabel;
	
	private I18NStringWithFillIns[] log = null;
}
