package de.ped.troff.client.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;

import javax.swing.JPanel;

import de.ped.tools.MathUtil;
import de.ped.tools.PlayerColor;
import de.ped.tools.gui.UserGameActionInput;
import de.ped.tools.log.Logger;
import de.ped.troff.client.logic.Abstract3DGeometry;
import de.ped.troff.client.logic.Circle;
import de.ped.troff.client.logic.ClientThread;
import de.ped.troff.client.logic.ColorUtil;
import de.ped.troff.client.logic.Real3DGeometry;
import de.ped.troff.client.logic.Side;
import de.ped.troff.server.logic.BoardInfo;
import de.ped.troff.server.logic.TroffGame;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffPlayer;
import de.ped.troff.server.logic.PlayerPosInfo;
import de.ped.troff.server.logic.TroffGameProperties;
import de.ped.troff.server.logic.TroffGameProperties.ModeInfo;

/**
 * The area where one player view is painted.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffPlayerViewCanvas extends JPanel {
	private static final long serialVersionUID = 1L;
	private static Logger logger = Logger.getRootLogger();

	public static final int SIZE_IN_PIXEL_X = 320;
	public static final int SIZE_IN_PIXEL_Y = 320;

	private final TroffMainWindow parent;
	private final Dimension preferredSize = new Dimension(SIZE_IN_PIXEL_X, SIZE_IN_PIXEL_Y);

	private final byte playerIdx;

	public TroffPlayerViewCanvas(final TroffMainWindow parent, final byte playerIdx) {
		this.parent = parent;
		setBackground(Color.black);
		this.playerIdx = playerIdx;
		final ClientThread model = parent.getClientThread();
		final TroffGameProperties props = model.getGameProperties();
		if (props.isAbsoluteView()) {
			final BoardInfo boardInfo = props.getBoardInfos()[props.getBoardIndex()];
			preferredSize.setSize(boardInfo.getSize().getWidth() * RADAR_XSTRETCH, boardInfo.getSize().getHeight() * RADAR_YSTRETCH);
		}
	}

	@Override
	public Dimension getPreferredSize() {
		return preferredSize;
	}

	@Override
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	public static final Color[] WALL_COLORS = new Color[] { Color.black, Color.gray, Color.white, };

	public Color fieldColor(final byte fieldValue, final BoardInfo.FieldInfo fieldInfo) {
		Color c;
		if (BoardInfo.isPlayer(fieldValue)) {
			final byte playerId = BoardInfo.fieldValue2playerIdx(fieldValue);
			c = getPlayerColor(playerId);
			if ((null != fieldInfo) && fieldInfo.isHead()) {
				c = c.brighter().brighter().brighter();
				logger.trace("Getting color for head.");
			}
		} else {
			c = WALL_COLORS[fieldValue];
		}
//		if ((null != fieldInfo) && (fieldInfo.getLastFieldValue() != fieldValue)) {
//			Color oldColor = fieldColor(fieldInfo.getLastFieldValue(), null);
//			int mixRGB = ColorUtil.mixRGB(c.getRGB(), oldColor.getRGB(), 33);
//			c = ColorUtil.rgbToColor(mixRGB);
//		}
		return c;
	}

	private Color getPlayerColor(final byte playerId) {
		final int colorId = parent.getClientThread().getGameProperties().getPlayers()[playerId].getPlayerColorId();
		return PlayerColor.get(colorId).getColor();
	}

	/**
	 * Returns the color to use, depending on the paremeters given.
	 * @param fieldValue the field value to paint
	 * @param fieldInfo additional info about the field, e.g., head and head direction. May be <code>null</code>
	 * @param distance the distance of the color - distant colors may be faded
	 * @param side the side to show. May be 
	 *   <ul>
	 *     <li>{@link Side#LEFT}, {@link Side#FRONT} or {@link Side#RIGHT} for real cubes in 3D,</li>
	 *     <li>{@link Side#BOTTOM} for empty fields, or</li>
	 *     <li><code>null</code> to indicate that a radar view is drawn.</li>
	 *   </ul>
	 * @return the color to use
	 */
	public Color fieldColor(
			final byte fieldValue, 
			final BoardInfo.FieldInfo fieldInfo, 
			final int distance, 
			final Side side) {
		Color ret;
		if (side == Side.BOTTOM) {
			ret = Color.green.darker().darker();
		} else {
			ret = fieldColor(fieldValue, fieldInfo);
			if (null == side) {
				ret = ret.brighter();
			} else {
				switch (side) {
				case RIGHT:
				case LEFT:
					ret = ret.darker();
					break;
				default:
					// Nothing to do
				}
			}
		}
		if (distance > 0) {
			ret = ColorUtil.fade(ret, distance);
		}
		return ret;
	}

	/**
	 * Returns the color to use, depending on the paremeters given.
	 * @param fieldValue the field value to paint
	 * @param fieldInfo additional info about the field, e.g., head and head direction. May be <code>null</code>
	 * @param xdistance the distance of the color - distant colors may be faded
	 * @param ydistance the distance of the color - distant colors may be faded
	 * @param distanceFactor the distance of the color - distant colors may be faded
	 * @param side the side to show. May be 
	 *   <ul>
	 *     <li>{@link Side#LEFT}, {@link Side#FRONT} or {@link Side#RIGHT} for real cubes in 3D,</li>
	 *     <li>{@link Side#BOTTOM} for empty fields, or</li>
	 *     <li><code>null</code> to indicate that a radar view is drawn.</li>
	 *   </ul>
	 * @return the color to use
	 */
	public Color fieldColor(
			final byte fieldValue, 
			final BoardInfo.FieldInfo fieldInfo, 
			final int xdistance, 
			final int ydistance, 
			final double distanceFactor, 
			final Side side) {
		return fieldColor(fieldValue, fieldInfo, (int)(Math.round(Circle.getDistance(xdistance, ydistance) * distanceFactor)), side);
	}

	public static final double RADAR_COLOR_DISTANCE_FADING = 0.5;

	public static final double PILOTVIEW_COLOR_DISTANCE_FADING = 1.0;

	public static final int PILOTVIEW_HORIZON_DISTANCE = 7;

	public static final Abstract3DGeometry GEOMETRY3D = new Real3DGeometry(
			SIZE_IN_PIXEL_X,
			SIZE_IN_PIXEL_Y,
			PILOTVIEW_HORIZON_DISTANCE);

	public static final int RADAR_XSTRETCH = 10;
	public static final int RADAR_YSTRETCH = 10;
	public static final int RADAR_XSIZE = 6;
	public static final int RADAR_YSIZE = 6;
	public static final int[][][] RADAR_EYES = new int[][][] {
		// North
		{ 	{                   RADAR_XSIZE / 4,                   RADAR_YSIZE / 4, 1, 1 },
			{ RADAR_XSIZE - 1 - RADAR_XSIZE / 4,                   RADAR_YSIZE / 4, 1, 1 },
		},
		// East
		{ 	{ RADAR_XSIZE - 1 - RADAR_XSIZE / 4,                   RADAR_YSIZE / 4, 1, 1 },
			{ RADAR_XSIZE - 1 - RADAR_XSIZE / 4, RADAR_YSIZE - 1 - RADAR_YSIZE / 4, 1, 1 },
		},
		// South
		{ 	{                   RADAR_XSIZE / 4, RADAR_YSIZE - 1 - RADAR_YSIZE / 4, 1, 1 },
			{ RADAR_XSIZE - 1 - RADAR_XSIZE / 4, RADAR_YSIZE - 1 - RADAR_YSIZE / 4, 1, 1 },
		},
		// West
		{ 	{                   RADAR_XSIZE / 4,                   RADAR_YSIZE / 4, 1, 1 },
			{                   RADAR_XSIZE / 4, RADAR_YSIZE - 1 - RADAR_YSIZE / 4, 1, 1 },
		},
	};

	protected static final Circle RADAR_CIRCLE = new Circle(PILOTVIEW_HORIZON_DISTANCE);

	public static final int RELATIVE_RADAR_FORWARD_MIN = -1;

	protected boolean isToDraw3DEyes(final int dist, final int xpos, final Side side, final int dir) {
		boolean ret = false;
		switch (side) {
		case FRONT:
			ret = (2 == dir);
			break;
		case LEFT:
		case RIGHT:
			ret = (((1 == dir) && (xpos < 0)) || ((3 == dir) && (xpos > 0)));
			break;
		default:
			// Nothing to do
			break;
		}
		return ret;
	}

	protected void paint3DBoxSide(
			final Graphics g, 
			final int dist, 
			final int xpos, 
			final Side side, 
			final BoardInfo.FieldInfo fieldInfo) {
		final BoardInfo boardInfo = parent.getClientThread().getGameState().getBoardInfo();
		final PlayerPosInfo thisPlayerPos = boardInfo.getPlayerPosInfo(playerIdx);
		final int[][] polySide = GEOMETRY3D.getBoxSide(dist, xpos, side, Abstract3DGeometry.BOX_SIDE);
		if (null != polySide) {
			g.fillPolygon(polySide[0], polySide[1], polySide[0].length);
			if (fieldInfo.isHead()) {
				// Draw eyes
				final int dir = PlayerPosInfo.normDirection(fieldInfo.getHeadDirection() - thisPlayerPos.getDirection());
				if (isToDraw3DEyes(dist, xpos, side, dir)) {
					g.setColor(ColorUtil.fade(Color.white, dist));
					for (int eye : Abstract3DGeometry.BOX_EYES) {
						int[][] polyEye = GEOMETRY3D.getBoxSide(dist, xpos, side, eye);
						if (null != polyEye) {
							g.fillPolygon(polyEye[0], polyEye[1], polyEye[0].length);
						}
					}
				}
			}
		}
	}

	/**
	 * Paints a 3D box in the given position. 
	 * It is important to draw from back to front, and from the outer to the inner;
	 * otherwise boxes in the background might be drawn on top of boxes in the foreground.
	 * @param g where to draw to
	 * @param dist how far ahead is the box
	 * @param xpos how much to the right is the box? Might be negative to express that the box is to the left
	 */
	protected void paint3DBox(final Graphics g, final int dist, final int xpos) {
		final BoardInfo boardInfo = parent.getClientThread().getGameState().getBoardInfo();
		if ((0 == dist) && (0 == xpos)) {
			// Nothing to do - my current box is drawn empty, so that I can look out
		} else {
			final BoardInfo.FieldInfo fieldInfo = new BoardInfo.FieldInfo();
			final byte field = boardInfo.getAtForwardPosition(playerIdx, dist, xpos, fieldInfo);
			if (BoardInfo.FREE == field) {
				final Color c = fieldColor(field, fieldInfo, dist, xpos, PILOTVIEW_COLOR_DISTANCE_FADING, Side.BOTTOM);
				g.setColor(c);
				final int[][] xs = GEOMETRY3D.getBoxSide(dist, xpos, Side.BOTTOM, Abstract3DGeometry.BOX_SIDE);
				if (null != xs) {
					g.fillPolygon(xs[0], xs[1], xs[0].length);
				}
			} else {
				// Draw front side
				final Color c = fieldColor(field, fieldInfo, dist, xpos, PILOTVIEW_COLOR_DISTANCE_FADING, Side.FRONT);
				g.setColor(c);
				paint3DBoxSide(g, dist, xpos, Side.FRONT, fieldInfo);
				Side cubeSide = null;
				if (xpos < 0) {
					cubeSide = Side.RIGHT;
				} else if (xpos > 0) {
					cubeSide = Side.LEFT;
				}
				if (null != cubeSide) {
					// Draw left or right side
					g.setColor(fieldColor(field, fieldInfo, dist, xpos, PILOTVIEW_COLOR_DISTANCE_FADING, cubeSide));
					paint3DBoxSide(g, dist, xpos, cubeSide, fieldInfo);
				}
			}
		}
	}

	/**
	 * Paints a full 3D landscape from the perspective of the player with the Player ID playerIdx.
	 * @param g where to draw to
	 */
	protected void paint3DLandscape(final Graphics g) {
		if (null != parent.getClientThread().getGameState()) {
			// Plot upper and lower trapez in player color - as the player sits
			// in the center of his cube.
			final int[][] playerCeiling = GEOMETRY3D.getPlayerCeiling();
			final int[][] playerFloor = GEOMETRY3D.getPlayerFloor();
			g.setColor(fieldColor(BoardInfo.playerIdx2fieldValue(playerIdx), null, 0, Side.LEFT));
			g.fillPolygon(playerCeiling[0], playerCeiling[1], playerCeiling[0].length);
			g.fillPolygon(playerFloor[0], playerFloor[1], playerFloor[0].length);
			// 3-D View
			for (int dist = GEOMETRY3D.getHorizonCircle().size() - 1; dist >= 0; dist--) {
				for (int xpos = GEOMETRY3D.getHorizonCircle().getX(dist); xpos > 0; xpos--) {
					// Left sided boxes
					paint3DBox(g, dist, -xpos);
					// Right sided boxes
					paint3DBox(g, dist, xpos);
				}
				// Boxes straight ahead
				paint3DBox(g, dist, 0);
			}
		}
	}

	protected void paintRadarBox(
			final Graphics g, 
			final int xoffset, 
			final int yoffset, 
			final int xi, 
			final int yi, 
			final BoardInfo.FieldInfo fieldInfo, 
			final int headDirectionToDrawOffset, 
			final double distanceFadingFactor) {
		final int x = xoffset + xi * RADAR_XSTRETCH;
		final int y = yoffset + yi * RADAR_YSTRETCH;
		if (null != fieldInfo) {
			if (BoardInfo.FREE != fieldInfo.getField()) {
				final Color c = fieldColor(fieldInfo.getField(), fieldInfo, xi, yi, distanceFadingFactor, null);
				g.setColor(c);
				g.drawRect(x, y, RADAR_XSIZE, RADAR_YSIZE);
				if (fieldInfo.isHead()) {
					final int headDirectionToDraw = MathUtil
							.modulo(fieldInfo.getHeadDirection() - headDirectionToDrawOffset, TroffGame.DIRECTIONS.length);
					final int[][] eyes = RADAR_EYES[headDirectionToDraw];
					for (int[] eye : eyes) {
						g.drawRect(x + eye[0], y + eye[1], eye[2], eye[3]);
					}
				}
			} else {
				final Color c = fieldColor(fieldInfo.getField(), fieldInfo, xi, yi, distanceFadingFactor, Side.BOTTOM);
				g.setColor(c);
				final int dotx = x + RADAR_XSIZE / 2;
				final int doty = y + RADAR_YSIZE / 2;
				g.drawLine(dotx, doty, dotx, doty);
			}
		}
	}

	/**
	 * Paints the radar view of the full field, from absolute view.
	 * @param g where to draw to
	 */
	protected void paintAbsoluteView(final Graphics g) {
		final int xoffset = (RADAR_XSTRETCH - RADAR_XSIZE) / 2;
		final int yoffset = (RADAR_YSTRETCH - RADAR_YSIZE) / 2;
		final TroffGameState gameState = parent.getClientThread().getGameState();
		if (null != gameState) {
			final BoardInfo boardInfo = gameState.getBoardInfo();
			final BoardInfo.FieldInfo fieldInfo = new BoardInfo.FieldInfo();
			for (int x = 0; x < boardInfo.getSize().getWidth(); x++) {
				for (int y = 0; y < boardInfo.getSize().getHeight(); y++) {
					boardInfo.calcFieldInfoAt(x, y, fieldInfo);
					paintRadarBox(g, xoffset, yoffset, x, y, fieldInfo, 0, 0.0);
				}
			}
		}
	}

	/**
	 * Paints the absolute radar view (north is always up) of the surrounding of this player's head.
	 * @param g where to draw to
	 */
	protected void paintAbsoluteRadar(final Graphics g) {
		final TroffGameState gameState = parent.getClientThread().getGameState();
		if (null != gameState) {
			final BoardInfo boardInfo = gameState.getBoardInfo();
			final BoardInfo.FieldInfo fieldInfo = new BoardInfo.FieldInfo();
			final PlayerPosInfo thisPlayerPos = boardInfo.getPlayerPosInfo(playerIdx);
			final int xoffset = (SIZE_IN_PIXEL_X - RADAR_XSIZE) / 2;
			final int yoffset = (SIZE_IN_PIXEL_Y - RADAR_YSIZE) / 2;
			for (int xi = -RADAR_CIRCLE.size() + 1; xi < RADAR_CIRCLE.size(); xi++) {
				final int xpos = MathUtil.modulo(thisPlayerPos.getXPos() + xi, boardInfo.getSize().getWidth());
				final int halfCircleX = RADAR_CIRCLE.getX(xi);
				for (int yi = -halfCircleX + 1; yi < halfCircleX; yi++) {
					final int ypos = MathUtil.modulo(thisPlayerPos.getYPos() + yi, boardInfo.getSize().getHeight());
					boardInfo.calcFieldInfoAt(xpos, ypos, fieldInfo);
					paintRadarBox(g, xoffset, yoffset, xi, yi, fieldInfo, 0, RADAR_COLOR_DISTANCE_FADING);
				}
			}
		}
	}

	/**
	 * Paints a full 3D landscape from the perspective of the player with the player ID playerIdx.
	 * Up is always the direction the player is heading to.
	 * @param g where to draw to
	 */
	protected void paintRelativeRadar(final Graphics g) {
		final TroffGameState gameState = parent.getClientThread().getGameState();
		if (null != gameState) {
			final BoardInfo boardInfo = gameState.getBoardInfo();
			final BoardInfo.FieldInfo fieldInfo = new BoardInfo.FieldInfo();
			final PlayerPosInfo thisPlayerPos = boardInfo.getPlayerPosInfo(playerIdx);
			final int xoffset = (SIZE_IN_PIXEL_X - RADAR_XSIZE) / 2;
			final int yoffset = SIZE_IN_PIXEL_Y + (RELATIVE_RADAR_FORWARD_MIN - 2) * RADAR_YSIZE;
			for (int dist = RELATIVE_RADAR_FORWARD_MIN; dist < RADAR_CIRCLE.size(); dist++) {
				final int halfCircleX = RADAR_CIRCLE.getX(dist);
				for (int xpos = -halfCircleX + 1; xpos < halfCircleX; xpos++) {
					boardInfo.getAtForwardPosition(playerIdx, dist, xpos, fieldInfo);
					paintRadarBox(g, xoffset, yoffset, xpos, -dist, fieldInfo, thisPlayerPos.getDirection(),
							RADAR_COLOR_DISTANCE_FADING);
				}
			}
		}
	}

	protected void paintIntro(final Graphics g) {
		final ClientThread model = parent.getClientThread();
		final int introDoorHalfHeight = (int) (preferredSize.height / 2 * model.getIntroTicks()
				/ ClientThread.INTRO_DRAW_STEPS);
		// Paint opening door
		g.setColor(getPlayerColor(playerIdx));
		g.fillRect(0, 0, preferredSize.width, introDoorHalfHeight);
		g.fillRect(0, preferredSize.height - introDoorHalfHeight, preferredSize.width, introDoorHalfHeight);
		g.setColor(Color.WHITE);
		// Paint player number into opening door
		final Font df = g.getFont();
		g.setFont(new Font(df.getFontName(), Font.BOLD, preferredSize.height / 4));
		drawIntroTextAtPosition(g, Integer.toString(playerIdx + 1), 1, 2, 1, 4, introDoorHalfHeight);
		g.setFont(df);
		// Paint keys into opening door, if human player
		final TroffGameProperties props = parent.getClientThread().getGameProperties();
		if (TroffGameProperties.PLAYERTYPE_HUMAN == props.getPlayerType(playerIdx)) {
			final UserGameActionInput ugaiLeft   = parent.findUserGameActionInput(playerIdx, TroffPlayer.TURN_LEFT);
			final UserGameActionInput ugaiRight  = parent.findUserGameActionInput(playerIdx, TroffPlayer.TURN_RIGHT);
			final UserGameActionInput ugaiAction = parent.findUserGameActionInput(playerIdx, TroffPlayer.ACTION);
			drawIntroTextAtPosition(g, ugaiLeft.keyCodeString, 1, 4, 5, 8, introDoorHalfHeight);
			drawIntroTextAtPosition(g, ugaiAction.keyCodeString, 2, 4, 5, 8, introDoorHalfHeight);
			drawIntroTextAtPosition(g, ugaiRight.keyCodeString, 3, 4, 5, 8, introDoorHalfHeight);
		} else {
			drawIntroTextAtPosition(g, "---", 2, 4, 5, 8, introDoorHalfHeight);
		}
	}

	protected static final String[] readySetGoStrings = new String[] {
			"Game.IntroGo",
			null,
			"Game.IntroSet",
			null,
			"Game.IntroReady",
	};

	public static final int HEADS_UP_DISPLAY_LAMPS_Y = 5;
	public static final int HEADS_UP_DISPLAY_LAMPS_HEIGHT = 5;
	public static final int HEADS_UP_DISPLAY_HUNTER_SLOT_WIDTH = 3;

	public static final int HEADS_UP_DISPLAY_HUNTER_WIDTH = TroffGameProperties.SHOOTING_FREQUENCY * HEADS_UP_DISPLAY_HUNTER_SLOT_WIDTH + 2;
	public static final int HEADS_UP_DISPLAY_WIDTH = HEADS_UP_DISPLAY_HUNTER_WIDTH + 2 + HEADS_UP_DISPLAY_LAMPS_HEIGHT + 2 + HEADS_UP_DISPLAY_LAMPS_HEIGHT;

	protected void paintHunterHeadsUpDisplay(
			final Graphics g, 
			final int howManyEmpty, 
			final int of, 
			final boolean isHunter) {
		final int arcSize = HEADS_UP_DISPLAY_LAMPS_HEIGHT;
		int x = (preferredSize.width - HEADS_UP_DISPLAY_WIDTH) / 2;
		g.setColor(Color.GRAY);
		g.drawRoundRect(x, HEADS_UP_DISPLAY_LAMPS_Y, of * HEADS_UP_DISPLAY_HUNTER_SLOT_WIDTH + 1,
				HEADS_UP_DISPLAY_LAMPS_HEIGHT + 1, arcSize, arcSize);
		for (int i = 0; i < of; i++) {
			Color color = Color.DARK_GRAY;
			if (isHunter) {
				color = Color.BLACK;
				if (i < (of - howManyEmpty)) {
					if (i < of / 2) {
						color = ColorUtil.rgbToColor(ColorUtil.mixRGB(Color.RED.getRGB(), Color.YELLOW.getRGB(), i * 100 / (of / 2)));
					} else {
						color = ColorUtil.rgbToColor(ColorUtil.mixRGB(Color.YELLOW.getRGB(), Color.GREEN.getRGB(), (i - of / 2) * 100 / (of / 2)));
					}
				}
			}
			g.setColor(color);
			g.fillRect(x + 1, HEADS_UP_DISPLAY_LAMPS_Y + 1, HEADS_UP_DISPLAY_HUNTER_SLOT_WIDTH,
					HEADS_UP_DISPLAY_LAMPS_HEIGHT);
			x += HEADS_UP_DISPLAY_HUNTER_SLOT_WIDTH;
		}
	}

	protected void paintIsTraceHeadsUpDisplay(final Graphics g, final boolean isPrey, final boolean isTraceOn) {
		final int arcSize = HEADS_UP_DISPLAY_LAMPS_HEIGHT;
		final int x = (preferredSize.width - HEADS_UP_DISPLAY_WIDTH) / 2 + HEADS_UP_DISPLAY_HUNTER_WIDTH + 2 + HEADS_UP_DISPLAY_LAMPS_HEIGHT;
		g.setColor(Color.GRAY);
		g.drawRoundRect(x, HEADS_UP_DISPLAY_LAMPS_Y, HEADS_UP_DISPLAY_LAMPS_HEIGHT + 1, HEADS_UP_DISPLAY_LAMPS_HEIGHT + 1, arcSize, arcSize);
		final Color color = (isTraceOn ? Color.YELLOW : Color.BLACK);
		g.setColor(color);
		g.fillRoundRect(x + 1, HEADS_UP_DISPLAY_LAMPS_Y + 1, HEADS_UP_DISPLAY_LAMPS_HEIGHT, HEADS_UP_DISPLAY_LAMPS_HEIGHT, arcSize, arcSize);
	}

	protected void paintHeadsUpDisplay(final Graphics g) {
		final ClientThread model = parent.getClientThread();
		final TroffGameProperties props = model.getGameProperties();
		final TroffGameState gameState = model.getGameState();
		final FontMetrics fm = g.getFontMetrics();
		final int y = 0 + fm.getMaxAscent();
		g.setColor(Color.LIGHT_GRAY);
		String text = props.getFullPlayerName(playerIdx);
		g.drawString(text, 5, y);
		final ModeInfo modeInfo = TroffGameProperties.MODE_INFOS[props.getPlayerMode(playerIdx)];
		text = parent.getMessages().getText(modeInfo.getKey());
		g.drawString(text, preferredSize.width - 5 - fm.stringWidth(text), y);

		final int modeMask = modeInfo.getModeMask();
		final boolean isHunter = (0 != (modeMask & TroffGameProperties.ModeInfo.PLAYERMODEMASK_HUNTER));
		paintHunterHeadsUpDisplay(g, gameState.nextShotIn(playerIdx), TroffGameProperties.SHOOTING_FREQUENCY, isHunter);
		final boolean isPrey = (0 != (modeMask & TroffGameProperties.ModeInfo.PLAYERMODEMASK_PREY));
		final boolean isTrace = (0 != (modeMask & TroffGameProperties.ModeInfo.PLAYERMODEMASK_TRACE))
				|| gameState.isTracing(playerIdx);
		paintIsTraceHeadsUpDisplay(g, isPrey, isTrace);

		if (TroffGameProperties.PLAYERTYPE_HUMAN == props.getPlayerType(playerIdx)) {
			final UserGameActionInput ugaiLeft = parent.findUserGameActionInput(playerIdx, TroffPlayer.TURN_LEFT);
			final UserGameActionInput ugaiRight = parent.findUserGameActionInput(playerIdx, TroffPlayer.TURN_RIGHT);
			final UserGameActionInput ugaiAction = parent.findUserGameActionInput(playerIdx, TroffPlayer.ACTION);
			final int keyCodesY = 30;
			g.setColor(Color.LIGHT_GRAY);
			g.drawString(ugaiLeft.keyCodeString, 5, keyCodesY);
			g.drawString(ugaiRight.keyCodeString, preferredSize.width - 5 - fm.stringWidth(ugaiRight.keyCodeString),
					keyCodesY);
			g.drawString(ugaiAction.keyCodeString, (preferredSize.width - fm.stringWidth(ugaiAction.keyCodeString)) / 2,
					keyCodesY);
		}
	}

	@Override
	public void paintComponent(final Graphics g) {
		super.paintComponent(g);
		final ClientThread model = parent.getClientThread();
		if (null != model) {
			final TroffGameProperties props = model.getGameProperties();
			final TroffGameState gameState = model.getGameState();
			if (null == gameState) {
				return;
			}
			String text = null;
			if (props.isAbsoluteView()) {
				paintAbsoluteView(g);
				// paintAbsoluteRadarAbsoluteReal3D(g);
				if (!gameState.isAlive()) {
					text = parent.getMessages().getString("Game.GameOver", "Game Over");
				}
			} else if (playerIdx < props.getNumberOfPlayers()) {
				final int viewIndex = props.getPlayerView(playerIdx);
				final int viewMask = TroffGameProperties.VIEW_INFOS[viewIndex].getViewMask();
				if (0 != (viewMask & TroffGameProperties.ViewInfo.VIEW_3D)) {
					paint3DLandscape(g);
				}
				if (0 != (viewMask & TroffGameProperties.ViewInfo.VIEW_RADAR_ABSOLUTE)) {
					paintAbsoluteRadar(g);
				}
				if (0 != (viewMask & TroffGameProperties.ViewInfo.VIEW_RADAR_RELATIVE)) {
					paintRelativeRadar(g);
				}
				final PlayerPosInfo thisPlayerPos = gameState.getBoardInfo().getPlayerPosInfo(playerIdx);
				if (!thisPlayerPos.isAlive()) {
					text = parent.getMessages().getString("Game.GameOver", "Game Over");
				}
				paintHeadsUpDisplay(g);
			}
			if (model.isIntroRunning()) {
				paintIntro(g);
				final int idx = (int) MathUtil.bound(
						readySetGoStrings.length * model.getIntroTicks() / ClientThread.INTRO_DRAW_STEPS, 0,
						readySetGoStrings.length - 1);
				text = readySetGoStrings[idx];
				if (null != text) {
					text = parent.getMessages().getString(text);
				}
			}
			if ((null == text) && !model.isRunning()) {
				text = parent.getMessages().getString("Game.Paused", "Pause");
			}
			if (null != text) {
				drawCenteredTextBox(g, text);
			}
		}
	}

	protected void drawCenteredTextBox(final Graphics g, final String text) {
		final FontMetrics fm = g.getFontMetrics();
		final int strWidth = fm.stringWidth(text);
		final int strHeight = fm.getMaxAscent() + fm.getMaxDescent();
		final int xUpperLeft = (preferredSize.width - strWidth) / 2;
		final int yUpperLeft = (preferredSize.height - strHeight) / 2;
		g.setColor(Color.darkGray);
		final int xborder = 10;
		final int yborder = 5;
		g.fillRect(xUpperLeft - xborder, yUpperLeft - yborder, strWidth + 2 * xborder, strHeight + 2 * yborder);
		g.setColor(Color.white);
		g.draw3DRect(xUpperLeft - xborder, yUpperLeft - yborder, strWidth + 2 * xborder, strHeight + 2 * yborder, true);
		g.drawString(text, xUpperLeft, yUpperLeft + fm.getMaxAscent());
	}

	protected void drawIntroTextAtPosition(
			final Graphics g, 
			final String text, 
			final int xNumerator, 
			final int xDenominator, 
			final int yNumerator, 
			final int yDenominator, 
			final int introDoorHalfCurrentHeight) {
		final FontMetrics fm = g.getFontMetrics();
		final int strWidth = fm.stringWidth(text);
		final int strHeight = fm.getMaxAscent() + fm.getMaxDescent();
		final int totalHeightHalf = preferredSize.height / 2;
		final int xTextCenter = (preferredSize.width * xNumerator) / xDenominator;
		final int yTextCenter = (preferredSize.width * yNumerator) / yDenominator;
		final int yabs = totalHeightHalf - introDoorHalfCurrentHeight;
		final int xUpperLeft = xTextCenter - strWidth / 2;
		final int yUpperLeft = yTextCenter - strHeight / 2;
		final int engravement = 2;
		if (yTextCenter < totalHeightHalf) {
			// Text in door moving up
//			g.clipRect(0, 0, preferredSize.width, introDoorHalfHeight);
			final int y = yUpperLeft + fm.getMaxAscent() - yabs;
			g.setColor(Color.black);
			g.drawString(text, xUpperLeft + engravement, y + engravement);
			g.setColor(Color.white);
			g.drawString(text, xUpperLeft, y);
		}
		if (yTextCenter >= totalHeightHalf) {
			// Text in door moving down
//			g.clipRect(0, preferredSize.height - introDoorHalfHeight, preferredSize.width, introDoorHalfHeight);
			final int y = yUpperLeft + fm.getMaxAscent() + yabs;
			g.setColor(Color.black);
			g.drawString(text, xUpperLeft + engravement, y + engravement);
			g.setColor(Color.white);
			g.drawString(text, xUpperLeft, y);
		}
//		g.clipRect(0, 0, preferredSize.width, preferredSize.height);
	}
}
