package de.ped.troff.client.gui;

import javax.sound.sampled.AudioInputStream;

import de.ped.tools.ResourceFinder;
import de.ped.tools.gui.ApplicationSounds;
import de.ped.tools.sound.SoundClip;
import de.ped.tools.sound.SoundMonoWaveCreator;
import de.ped.troff.server.logic.Event;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * Bundles all the sounds for Troff.
 * 
 * @author Copyright (c) 2018 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffSounds extends ApplicationSounds {
	public TroffSounds(final ResourceFinder resources) {
		super(resources);
	}

	/**
	 * Returns the sound clip for the given key.
	 * @param key the key (e.g., event) to find the sound for. Accepts {@link Integer} objects, with values from the {@link de.ped.troff.server.logic.Event} class.
	 * @return the sound clip for the given key, or <code>null</code> if there is no such sound.
	 */
	@Override
	public SoundClip findSound(final Object key) {
		if (null == sounds) {
			init();
		}
		final int eventIndex = ((Integer) key).intValue();
		final SoundClip ret = sounds[eventIndex];
		return ret;
	}

	@Override
	public void init() {
		final int[] keys = new int[] {
				Event.STEP, 
				Event.SHOOT_FAILED, 
				Event.SHOOT_SUCCEEDED, 
				Event.HIT_WALL, 
				Event.HIT_SOLID_WALL,
				Event.DEATH, 
				Event.TRACE_ON, 
				Event.TRACE_OFF, 
				Event.INTRO,
		};
		int max = 0;
		for (int key : keys) {
			max = Math.max(max,  key);
		}
		final SoundClip[] buf = new SoundClip[max + 1];
		for (int key : keys) {
			final AudioInputStream stream = createSound(key);
			if (null != stream) {
				buf[key] = new SoundClip(stream);
			}
		}
		buf[Event.HIT_SOLID_WALL] = buf[Event.SHOOT_FAILED];
		sounds = buf;
	}

	@Override
	public AudioInputStream createSound(final Object key) {
		final int eventIndex = ((Integer) key).intValue();
		final SoundMonoWaveCreator smwc = new SoundMonoWaveCreator();
		final double stepSoundDuration = 0.001 * TroffGameProperties.getShortestDelay() * 0.4;
		AudioInputStream stream = null;
		double[] sound = null;
		switch (eventIndex) {
		case Event.STEP:
			sound = smwc.multiply(smwc.triangleWave(stepSoundDuration, 220.0),
					smwc.envelopePartWave(stepSoundDuration, 0.7, 0.0));
			break;
		case Event.SHOOT_FAILED:
			stream = read("HOLZSCLG.WAV");
			break;
		case Event.SHOOT_SUCCEEDED:
			stream = read("ZISCAUF.WAV");
			break;
		case Event.HIT_WALL:
			sound = new double[0];
			for (double i = 0.7; i >= 0.0; i -= 0.01) {
				sound = smwc.concat(sound, smwc.pinkNoise(0.005, i));
			}
			break;
		case Event.DEATH:
			stream = read("BOING1.WAV");
			break;
		case Event.TRACE_ON:
			sound = smwc.multiply(smwc.sinusWave(stepSoundDuration, 220.0 * 2.0 / 3.0),
					smwc.envelopePartWave(stepSoundDuration, 0.0, 1.0));
			break;
		case Event.TRACE_OFF:
			sound = smwc.multiply(smwc.sinusWave(stepSoundDuration, 220.0 * 2.0 / 3.0),
					smwc.envelopePartWave(stepSoundDuration, 1.0, 0.0));
			break;
		case Event.INTRO:
			stream = read("WISCWOIS.WAV");
			break;
		default:
		}
		if ((null == stream) && (null != sound)) {
			stream = smwc.toAudioInputStream(sound);
		}
		return stream;
	}

	private SoundClip[] sounds;
}
