package de.ped.troff.client.logic;

import java.util.HashMap;

/**
 * Knows what polygons to draw for the 3D view, given the relative position of the cube to the current player position.
 * Adds to the real 3D math some useful stuff like hashing values once calculated for speed, or sight limits.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class Abstract3DGeometry {

	public static final int X = 0;
	public static final int Y = 1;

	public static final int BOX_SIDE = 0;
	public static final int BOX_EYE1 = 1;
	public static final int BOX_EYE2 = 2;
	public static final int[] BOX_EYES = new int[]{BOX_EYE1, BOX_EYE2};
	public static final String[] CUBE_NAMES = {"side", "eye1", "eye2"};

	public static final double EYE_SIZE = 1.0 / 4.0;
	public static final double EYE_CENTER_OFF_CUBE_CENTER_IN_HEIGHT = 1.0 / 6.0;
	public static final double EYE_CENTER_OFF_CUBE_CENTER_IN_WIDTH = 1.0 / 4.0;

	/**
	 * Returns a human-readable description of the polygon data structure for
	 * debugging purposes.
	 * 
	 * @param poly
	 *            a two-dimensional array. The first dimension defines the
	 *            stongpoints of the polygon, the second dimension is always
	 *            size 2 for x and y position
	 * @return a human-readable description of the polygon data structure 
 	 */
	public static String toString(final int[][] poly) {
		if (null == poly) {
			return null;
		}
		final StringBuffer buf = new StringBuffer("Polygon(");
		for (int i = 0; i < poly[0].length; i++) {
			buf.append("(");
			buf.append(poly[X][i]);
			buf.append(",");
			buf.append(poly[Y][i]);
			buf.append(")");
		}
		buf.append(")");
		return buf.toString();
	}

	public final int sizeInPixelX;
	public final int sizeInPixelY;
	protected final Circle horizonCircle;

	public Abstract3DGeometry(final int sizeInPixelX, final int sizeInPixelY, final int horizonDistance) {
		this.sizeInPixelX = sizeInPixelX;
		this.sizeInPixelY = sizeInPixelY;
		horizonCircle = new Circle(horizonDistance);
	}

	public Circle getHorizonCircle() {
		return horizonCircle;
	}

	protected abstract int[][] calcBoxSide(final int dist, final int xpos, final Side side, final int what);

	private HashMap<CubeSideKey, int[][]> map = new HashMap<CubeSideKey, int[][]>();

	public int[][] getBoxSide(final CubeSideKey key) {
		if ((key.side != Side.BACK) && 
			(key.side != Side.FRONT) && 
			(key.side != Side.LEFT) &&
			(key.side != Side.RIGHT) &&
			(key.side != Side.TOP) &&
			(key.side != Side.BOTTOM)) {
			throw new IllegalArgumentException("side was " + key.side);
		}
		if ((key.what != BOX_SIDE) && 
			(key.what != BOX_EYE1) && 
			(key.what != BOX_EYE2)) {
			throw new IllegalArgumentException("what was " + key.what);
		}
		if ((0 <= key.dist) && 
				(((Side.FRONT == key.side) && (key.dist < horizonCircle.size())) || 
				 ((Side.FRONT != key.side) && (key.dist < horizonCircle.size() - 1)))) {
			int[][] value = map.get(key);
			if (null == value) {
				value = calcBoxSide(key.dist, key.xpos, key.side, key.what);
				map.put(key, value);
			}
			return value;
		}
		return null;
	}

	public int[][] getBoxSide(final int dist, final int xpos, final Side side, final int what) {
		return getBoxSide(new CubeSideKey(dist, xpos, side, what));
	}

	public int[][] getPlayerCeiling() {
		return getBoxSide(0, 0, Side.TOP, BOX_SIDE);
	}

	public int[][] getPlayerFloor() {
		return getBoxSide(0, 0, Side.BOTTOM, BOX_SIDE);
	}
	
}
