package de.ped.troff.client.logic;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.Iterator;
import java.util.List;

import de.ped.tools.FinishableThread;
import de.ped.tools.log.Logger;
import de.ped.troff.client.gui.TroffMainWindow;
import de.ped.troff.middleware.Environment;
import de.ped.troff.middleware.Message;
import de.ped.troff.middleware.ServerInfo;
import de.ped.troff.middleware.ServerSelectionModel;
import de.ped.troff.middleware.TroffRoundsManager;
import de.ped.troff.server.logic.Event;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * All the client knows about the state of the game, 
 * and responsible for taking any actions if the state changes.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ClientThread extends FinishableThread {

	/** 
	 * The shortest time period this thread should wake up. 10 means 100fps, 1000 means 1fps :-)
	 * Should be the fastest game speed / 5 or smaller for good keyboard reactivity.
	 */
	public static final long HEARTBEAT_TICKS = 10;

	/** The number of intro steps to draw. For smooth drawing, should be a divisor of the canvas sizes. */
	public static final long INTRO_DRAW_STEPS = 80;

	private final TroffMainWindow window;
	private final Environment environment;

	private TroffGameProperties props = new TroffGameProperties();

    private volatile boolean isRunning = TroffRoundsManager.IS_RUNNING_DEFAULT;

	private long timeOfIntroStart = 0L;

	private TroffRoundsManager server;

	public ClientThread(
			final TroffMainWindow window, 
			final Environment environment, 
			final ServerSelectionModel model) throws UnknownHostException, IOException {
		super("ClientThread", HEARTBEAT_TICKS);
		setLogger(Logger.getLogger(this.getClass()));
		logger.setToShowThread(true);
		this.window = window;
		this.environment = (window == null ? environment : window.getEnvironment());
		server = new TroffRoundsManager(this, model.getHostName(), model.getHostPortNo(), logger);
	}

	public Environment getEnvironment() {
		return environment;
	}

	private void startNewGame(final TroffGameProperties props) {
		this.props = props;
		window.startNewGame(props);
		startIntro();
		isRunning = true;
		window.playSound(Event.INTRO);
	}

	public TroffGameProperties getGameProperties() {
		return props;
	}

	public TroffGameState getGameState() {
		return (null == server ? null : server.getLastReceivedGameState());
	}

	public ServerInfo getServerStatus() {
		return (null == server ? null : server.getLastReceivedServerStatus());
	}

	public long getIntroTicks() {
		return INTRO_DRAW_STEPS - ((long) (System.currentTimeMillis() - timeOfIntroStart)) * INTRO_DRAW_STEPS
				/ TroffRoundsManager.INTRO_DELAY;
	}

	private void startIntro() {
		timeOfIntroStart = System.currentTimeMillis();
	}

	public boolean isIntroRunning() {
		return (System.currentTimeMillis() - timeOfIntroStart < TroffRoundsManager.INTRO_DELAY);
	}

	private void introStep() {
		window.update();
	}

	public boolean isRunning() {
		return isRunning;
	}

	private void request(final Message req) {
		if (null != server) {
			server.sendMessage(req);
		}
	}

	/**
	 * May be called by other threads.
	 * 
	 * @param clientName
	 *            the name of the client
	 * @param numberOfPlayers
	 *            the number of players at that client
	 */
	public void requestLogin(final String clientName, final byte numberOfPlayers) {
		request(Message.createLoginReq(clientName, numberOfPlayers));
	}

	/**
	 * May be called by other threads.
	 */
	public void requestLogout() {
		request(Message.createLogoutReq());
	}

	/** 
	 * May be called by other threads. 
	 * @param groupName
	 *            the name of the group to join
	 */
	public void requestJoin(final String groupName) {
		request(Message.createJoinReq(groupName));
	}

	/**
	 * May be called by other threads.
	 */
	public void requestLeave() {
		request(Message.createLeaveReq());
	}

	/**
	 * May be called by other threads.
	 */
	public void requestServerStatus() {
		request(Message.createServerStatusReq());
	}

	/**
	 * May be called by other threads.
	 * 
	 * @param props
	 *            the game properties
	 */
	public void requestStartGame(final TroffGameProperties props) {
		request(Message.createStartGameReq(props));
		logger.debug("Still alive in requestStartGame.");
	}

	/**
	 * May be called by other threads.
	 * 
	 * @param isToResume
	 *            <code>true</code> to resume the game, <code>false</code> to
	 *            pause the game
	 */
	public void requestPauseOrResumeRun(final boolean isToResume) {
		request(Message.createPauseResumeReq(isToResume));
		logger.debug("Still alive in requestPauseOrResumeRun.");
	}

	/**
	 * May be called by other threads.
	 * 
	 * @param playerIdx
	 *            the player ID
	 * @param playerAction
	 *            the player action
	 */
	public void requestGameActionInput(final int playerIdx, final int playerAction) {
		request(Message.createGameActionReq(playerIdx, playerAction));
	}

	/**
	 * May be called by other threads.
	 * 
	 * @param message
	 *            the message to the other players
	 */
	public void requestSendMessage(final String message) {
		request(Message.createSendMessageReq(message));
	}

	private void gameStep() {
		if (null != server) {
			final TroffGameState state = server.getLastReceivedGameState();
			if (null != state) {
				final List<Event> events = state.getEvents();
				if (window.isSoundOn() && (null != events)) {
					final Iterator<Event> it = events.iterator();
					while (it.hasNext()) {
						final Event event = it.next();
						window.playSound(event.getType());
					}
				}
				if (!state.isAlive()) {
					isRunning = false;
				}
			}
		}
	}

	public void step() {
		if (isIntroRunning()) {
			introStep();
		} else {
			gameStep();
		}
		window.update();
	}

	public void update(final TroffRoundsManager who, Message msg) {
		logger.debug(this.getClass().getSimpleName() + " received notification " + msg);
		switch (msg.getMsgType()) {
		case Message.RESPONSE | Message.MESSAGES:
			window.setStatusText(msg.getGroupMessages());
			break;
		case Message.RESPONSE | Message.LOGIN:
			// Nothing to do
			break;
		case Message.RESPONSE | Message.LOGOUT:
			// Nothing to do
			break;
		case Message.RESPONSE | Message.JOIN:
			// Nothing to do
			break;
		case Message.RESPONSE | Message.LEAVE:
			// Nothing to do
			break;
		case Message.RESPONSE | Message.START_GAME:
			startNewGame(msg.getProperties());
			break;
		case Message.RESPONSE | Message.PAUSE_RESUME:
			isRunning = msg.isRunning();
			window.update();
			break;
		case Message.RESPONSE | Message.GAME_STATE:
			step();
			break;
		case Message.RESPONSE | Message.SEND_MESSAGE:
			// Nothing to do
			break;
		}
		logger.debug(this.getClass().getSimpleName() + " received notification " + msg + " finished.");
	}

	private void readMessages() {
		if (null != server) {
			try {
				server.readMessages();
			} catch (final IOException e) {
				logger.debug("", e);
			}
		}
	}

	@Override
	public void threadStep() {
		if (isIntroRunning()) {
			step();
		}
		readMessages();
	}
}
