package de.ped.troff.client.logic;

import java.awt.Color;
import java.util.ArrayList;

import de.ped.tools.MathVector;
import de.ped.troff.server.logic.BoardInfo;

/**
 * A data representation of a cube that helps to identify those rectangles of the cube 
 * that need to be drawn if viewed from a certain perspective.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Cube3D {

	/**
	 * A data representation of a rectangle, that is a side of a cube
	 * 
	 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class Rectangle3D {
		public final MathVector[] vertexes = new MathVector[4];
		public MathVector[] getVertexes() {
			return vertexes;
		}
	}

	public static final int CORNER_MASK_TOP    = 0x00;
	public static final int CORNER_MASK_BOTTOM = 0x01;
	public static final int CORNER_MASK_LEFT   = 0x00;
	public static final int CORNER_MASK_RIGHT  = 0x02;
	public static final int CORNER_MASK_FRONT  = 0x00;
	public static final int CORNER_MASK_BACK   = 0x04;

	public static final int CORNER_TOP_LEFT_FRONT     = CORNER_MASK_TOP    | CORNER_MASK_LEFT  | CORNER_MASK_FRONT;
	public static final int CORNER_TOP_LEFT_BACK      = CORNER_MASK_TOP    | CORNER_MASK_LEFT  | CORNER_MASK_BACK;
	public static final int CORNER_TOP_RIGHT_FRONT    = CORNER_MASK_TOP    | CORNER_MASK_RIGHT | CORNER_MASK_FRONT;
	public static final int CORNER_TOP_RIGHT_BACK     = CORNER_MASK_TOP    | CORNER_MASK_RIGHT | CORNER_MASK_BACK;
	public static final int CORNER_BOTTOM_LEFT_FRONT  = CORNER_MASK_BOTTOM | CORNER_MASK_LEFT  | CORNER_MASK_FRONT;
	public static final int CORNER_BOTTOM_LEFT_BACK   = CORNER_MASK_BOTTOM | CORNER_MASK_LEFT  | CORNER_MASK_BACK;
	public static final int CORNER_BOTTOM_RIGHT_FRONT = CORNER_MASK_BOTTOM | CORNER_MASK_RIGHT | CORNER_MASK_FRONT;
	public static final int CORNER_BOTTOM_RIGHT_BACK  = CORNER_MASK_BOTTOM | CORNER_MASK_RIGHT | CORNER_MASK_BACK;

	/** Human-readable descriptions for the corner IDs, for debugging purposes */
	public static final String[] CORNER_NAMES = {
		"TopLeftFront",    "TopLeftBack",    "TopRightFront",    "TopRightBack", 
		"BottomLeftFront", "BottomLeftBack", "BottomRightFront", "BottomRightBack"};

	public static final int[][] SIDE_VERTEX_INDEXES = new int[][] {
		{CORNER_TOP_LEFT_FRONT,    CORNER_TOP_RIGHT_FRONT,  CORNER_BOTTOM_RIGHT_FRONT, CORNER_BOTTOM_LEFT_FRONT},  //SIDE_FRONT
		{CORNER_TOP_LEFT_BACK,     CORNER_TOP_RIGHT_BACK,   CORNER_BOTTOM_RIGHT_BACK,  CORNER_BOTTOM_LEFT_BACK},   //SIDE_BACK
		{CORNER_TOP_LEFT_FRONT,    CORNER_TOP_LEFT_BACK,    CORNER_BOTTOM_LEFT_BACK,   CORNER_BOTTOM_LEFT_FRONT},  //SIDE_LEFT
		{CORNER_TOP_RIGHT_FRONT,   CORNER_TOP_RIGHT_BACK,   CORNER_BOTTOM_RIGHT_BACK,  CORNER_BOTTOM_RIGHT_FRONT}, //SIDE_RIGHT
		{CORNER_TOP_LEFT_FRONT,    CORNER_TOP_LEFT_BACK,    CORNER_TOP_RIGHT_BACK,     CORNER_TOP_RIGHT_FRONT},    //SIDE_TOP
		{CORNER_BOTTOM_LEFT_FRONT, CORNER_BOTTOM_LEFT_BACK, CORNER_BOTTOM_RIGHT_BACK,  CORNER_BOTTOM_RIGHT_FRONT}, //SIDE_BOTTOM
	};

	private MathVector[] vertexes = new MathVector[8];
	Color color = Color.white;
	BoardInfo.FieldInfo fieldInfo;

//	public Cube3D(double left, double right, double back, double front, double bottom, double top) {
//		vertexes[CORNER_BOTTOM_LEFT_BACK]   = new MathVector(new double[]{left, back, bottom});
//		vertexes[CORNER_BOTTOM_RIGHT_BACK]  = new MathVector(new double[]{right, back, bottom});
//		vertexes[CORNER_BOTTOM_LEFT_FRONT]  = new MathVector(new double[]{left, front, bottom});
//		vertexes[CORNER_BOTTOM_RIGHT_FRONT] = new MathVector(new double[]{right, front, bottom});
//		vertexes[CORNER_TOP_LEFT_BACK]      = new MathVector(new double[]{left, back, top});
//		vertexes[CORNER_TOP_RIGHT_BACK]     = new MathVector(new double[]{right, back, top});
//		vertexes[CORNER_TOP_LEFT_FRONT]     = new MathVector(new double[]{left, front, top});
//		vertexes[CORNER_TOP_RIGHT_FRONT]    = new MathVector(new double[]{right, front, top});
//	}

	public Cube3D(double mx, double my, double mz, double edgeLength) {
		final double r = edgeLength / 2.0;
		vertexes[CORNER_BOTTOM_LEFT_FRONT]  = new MathVector(new double[]{mx - r, my - r, mz - r});
		vertexes[CORNER_BOTTOM_RIGHT_FRONT] = new MathVector(new double[]{mx + r, my - r, mz - r});
		vertexes[CORNER_BOTTOM_LEFT_BACK]   = new MathVector(new double[]{mx - r, my + r, mz - r});
		vertexes[CORNER_BOTTOM_RIGHT_BACK]  = new MathVector(new double[]{mx + r, my + r, mz - r});
		vertexes[CORNER_TOP_LEFT_FRONT]     = new MathVector(new double[]{mx - r, my - r, mz + r});
		vertexes[CORNER_TOP_RIGHT_FRONT]    = new MathVector(new double[]{mx + r, my - r, mz + r});
		vertexes[CORNER_TOP_LEFT_BACK]      = new MathVector(new double[]{mx - r, my + r, mz + r});
		vertexes[CORNER_TOP_RIGHT_BACK]     = new MathVector(new double[]{mx + r, my + r, mz + r});
	}

	public MathVector[] getVertexes() {
		return vertexes;
	}

	public Color getColor() {
		return color;
	}

	public void setColor(final Color c) {
		this.color = c;
	}

	public BoardInfo.FieldInfo getFieldInfo() {
		return fieldInfo;
	}

	public void setFieldInfo(final BoardInfo.FieldInfo fieldInfo) {
		this.fieldInfo = fieldInfo;
	}

	public Rectangle3D createSide(final Side side) {
		final Rectangle3D ret = new Rectangle3D();
		for (int i = 0; i < ret.vertexes.length; i++) {
			final int vi = SIDE_VERTEX_INDEXES[side.ordinal()][i];
			ret.vertexes[i] = vertexes[vi];
		}
		return ret;
	}

	private int calcBoundaryValue(final double value, final double lower, final double upper) {
		int ret = 0;
		if (value < lower) {
			ret = -1;
		} else if (value > upper){
			ret = 1;
		}
		return ret;
	}

	public ArrayList<Cube3D.Rectangle3D> createPlanesVisibleFrom(final MathVector pos) {
		final ArrayList<Cube3D.Rectangle3D> ret = new ArrayList<Cube3D.Rectangle3D>();
		if (null == pos) {
			for (final Side side : Side.values()) {
				ret.add(createSide(side));
			}
		} else {
			final int xstat = calcBoundaryValue(pos.get(MathVector.X), vertexes[0].get(MathVector.X), vertexes[7].get(MathVector.X));
			final int ystat = calcBoundaryValue(pos.get(MathVector.Y), vertexes[0].get(MathVector.Y), vertexes[7].get(MathVector.Y));
			final int zstat = calcBoundaryValue(pos.get(MathVector.Z), vertexes[0].get(MathVector.Z), vertexes[7].get(MathVector.Z));
			if ((0 == xstat) && (0 == ystat) && (0 == zstat)) { // I'm inside
				for (final Side side : Side.values()) { // Draw all planes
					ret.add(createSide(side));
				}
			} else {
				if (-1 == xstat) {
					ret.add(createSide(Side.LEFT));
				}
				if (1 == xstat) {
					ret.add(createSide(Side.RIGHT));
				}
				if (-1 == ystat) {
					ret.add(createSide(Side.BACK));
				}
				if (1 == ystat) {
					ret.add(createSide(Side.FRONT));
				}
				if (-1 == zstat) {
					ret.add(createSide(Side.BOTTOM));
				}
				if (1 == zstat) {
					ret.add(createSide(Side.TOP));
				}
			}
		}
		return ret;
	}
}