package de.ped.troff.client.logic;

import de.ped.tools.MathMatrix;
import de.ped.tools.MathVector;

/**
 * Knows what polygons to draw for the 3D view, given the relative position of
 * the cube to the current player position.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Real3DGeometry extends Abstract3DGeometry {

	public Real3DGeometry(final int sizeInPixelX, final int sizeInPixelY, final int horizonDistance) {
		super(sizeInPixelX, sizeInPixelY, horizonDistance);
	}

	public static final double BORDER_THICKNESS = 0.03;

	public static final MathVector CAMERA_POSITION = new MathVector(new double[]{0.5, 0.5, 0.5});

	protected MathMatrix mTransRot;

	protected MathMatrix getHomogenousTranslationRotationMatrix() {
		if (null == mTransRot) {
			final MathMatrix mRotXY = MathMatrix.homRot(3, MathVector.X, MathVector.Y, 0.0);
			final MathMatrix mRotYZ = MathMatrix.homRot(3, MathVector.Y, MathVector.Z, Math.PI / 2.0);
			final MathMatrix mRot = mRotYZ.mult(mRotXY);
			final MathMatrix mTrans = MathMatrix.homTrans(CAMERA_POSITION.minus());
			mTransRot = mRot.mult(mTrans);
		}
		return mTransRot;
	}

	protected MathVector transform(final MathVector v) {
		final MathVector vHom = v.homogenize();
		final MathMatrix mhTransRot = getHomogenousTranslationRotationMatrix();
		final MathVector vHomT = mhTransRot.mult(vHom);
		final MathVector vt = vHomT.dehomogenize();
		return vt;
	}

	protected MathVector[] getVertexes(final int dist, final int xpos, final Side side, final int what) {
		if (Abstract3DGeometry.BOX_SIDE == what) {
			final Cube3D cube = new Cube3D(xpos + 0.5, dist + 0.5, 0.5, 1.0 - 2.0 * BORDER_THICKNESS);
			final Cube3D.Rectangle3D rect = cube.createSide(side);
			return rect.getVertexes();
		} else if ((Abstract3DGeometry.BOX_EYE1 == what) || (Abstract3DGeometry.BOX_EYE2 == what)) {
			final double sgn = (BOX_EYE1 == what ? 1.0 : -1.0);
			if (Side.FRONT == side) {
				final double eyeCenterX = xpos + 0.5 + sgn * Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_WIDTH;
				final double eyeCenterY = dist + Abstract3DGeometry.EYE_SIZE;
				final double eyeCenterZ = 0.5 + Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_HEIGHT;
				final Cube3D cube = new Cube3D(eyeCenterX, eyeCenterY, eyeCenterZ, Abstract3DGeometry.EYE_SIZE);
				final Cube3D.Rectangle3D rect = cube.createSide(side);
				return rect.getVertexes();
			} else if (Side.LEFT == side) {
				final double eyeCenterX = xpos + Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_HEIGHT;
				final double eyeCenterY = dist + 0.5 + sgn * Abstract3DGeometry.EYE_SIZE;
				final double eyeCenterZ = 0.5 + Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_WIDTH;
				final Cube3D cube = new Cube3D(eyeCenterX, eyeCenterY, eyeCenterZ, Abstract3DGeometry.EYE_SIZE);
				final Cube3D.Rectangle3D rect = cube.createSide(side);
				return rect.getVertexes();
			} else if (Side.RIGHT == side) {
				final double eyeCenterX = xpos + 1.0 - Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_HEIGHT;
				final double eyeCenterY = dist + 0.5 + sgn * Abstract3DGeometry.EYE_SIZE;
				final double eyeCenterZ = 0.5 + Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_WIDTH;
				final Cube3D cube = new Cube3D(eyeCenterX, eyeCenterY, eyeCenterZ, Abstract3DGeometry.EYE_SIZE);
				final Cube3D.Rectangle3D rect = cube.createSide(side);
				return rect.getVertexes();
			}
		}
		return null;
	}

	public static final double WIDE_ANGLE_FACTOR = 2.1;

	public static final double ZOOM = 1.3;

	@Override
	protected int[][] calcBoxSide(final int dist, final int xpos, final Side side, final int what) {
		final MathVector[] vector = getVertexes(dist, xpos, side, what);
		if (null == vector) {
			return null;
		}
		final int[][]ret = new int[2][vector.length];
		for (int i = 0; i < vector.length; i++) {
			final MathVector vt = transform(vector[i]);
			final double z = vt.get(MathVector.Z);
			final double zfactor = ZOOM / (1.0 + WIDE_ANGLE_FACTOR * z);
			final double x = vt.get(MathVector.X) * zfactor;
			final double y = vt.get(MathVector.Y) * zfactor;
			ret[X][i] = (int)Math.round(x * sizeInPixelX) + sizeInPixelX / 2;
			final int kruemmung = (int)Math.round(Math.sqrt(dist * dist + xpos * xpos));
			ret[Y][i] = (int)Math.round(y * sizeInPixelY) + sizeInPixelY / 2 + kruemmung;
		}
		return ret;
	}
}
