package de.ped.troff.client.logic;

/**
 * Knows what polygons to draw for the 3D view, given the relative position of the cube to the current player position.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Simple3DGeometry extends Abstract3DGeometry {
	protected static final int BORDER_THICKNESS = 1;

	/** The y position of the horizon. That's the highest possible bottom line. */
	protected final int horizonY;
	protected final int[] ypos;
	protected final int[] yh;
	protected final int[] xw;

	protected static final double LAIDBACK = 0.7;

	public Simple3DGeometry(final int sizeInPixelX, final int sizeInPixelY, final int horizonDistance) {
		super(sizeInPixelX, sizeInPixelY, horizonDistance);
		double eFactor  = Math.pow(2.0, 1.0 - Simple3DGeometry.LAIDBACK);
		int horizon = sizeInPixelY;
		ypos = new int[horizonDistance];
		yh = new int[horizonDistance];
		xw = new int[horizonDistance];
		for (int i = 0; i < horizonDistance; i++) {
			ypos[i] = (int)Math.round(sizeInPixelY / 2.0 * (1.0 - eFactor)) - (3 - i) * sizeInPixelY / 100;
			yh[i]   = (int)Math.round(sizeInPixelY * eFactor);
			xw[i]   = (int)Math.round(sizeInPixelX * eFactor * 1.2);
			horizon = Math.min(horizon, ypos[i] + yh[i]);
			eFactor /= 2.0;
		}
		horizonY = horizon;
	}

	/** 
	 * @param dist the distance of the box
	 * @param xpos the horizontal translation of the box
	 * @return the x value of the left side of the given box, without taking border thickness into account
	 */
	protected int calcLeftX(final int dist, final int xpos) {
		return sizeInPixelX /2 - xw[dist] / 2 + xpos * xw[dist];
	}

	public int getX(final int dist, final int xpos, final int corner) {
		int x;
		switch (corner) {
		case Cube3D.CORNER_TOP_LEFT_FRONT:
		case Cube3D.CORNER_BOTTOM_LEFT_FRONT: 
			{
				x = calcLeftX(dist, xpos) + BORDER_THICKNESS;
			}
			break;
		case Cube3D.CORNER_TOP_LEFT_BACK: 
		case Cube3D.CORNER_BOTTOM_LEFT_BACK:
			{
				x = calcLeftX(dist + 1, xpos) + BORDER_THICKNESS;
			}
			break;
		case Cube3D.CORNER_TOP_RIGHT_FRONT:
		case Cube3D.CORNER_BOTTOM_RIGHT_FRONT:
			{
				x = calcLeftX(dist, xpos + 1) - BORDER_THICKNESS;
			}
			break;
		case Cube3D.CORNER_TOP_RIGHT_BACK:
		case Cube3D.CORNER_BOTTOM_RIGHT_BACK:
			{
				x = calcLeftX(dist + 1, xpos + 1) - BORDER_THICKNESS;
			}
			break;
		default:
			throw new IllegalArgumentException();
		}
		return x;
	}

	public int getY(final int dist, final int xpos, final int corner) {
		int y;
		switch (corner) {
		case Cube3D.CORNER_TOP_LEFT_FRONT:
		case Cube3D.CORNER_TOP_RIGHT_FRONT:
			{
				y = ypos[dist] + BORDER_THICKNESS;
			}
			break;
		case Cube3D.CORNER_BOTTOM_LEFT_FRONT: 
		case Cube3D.CORNER_BOTTOM_RIGHT_FRONT:
			{
				y = ypos[dist] + yh[dist] - BORDER_THICKNESS;
			}
			break;
		case Cube3D.CORNER_TOP_LEFT_BACK: 
		case Cube3D.CORNER_TOP_RIGHT_BACK:
			{
				y = ypos[dist + 1] + BORDER_THICKNESS;
			}
			break;
		case Cube3D.CORNER_BOTTOM_LEFT_BACK:
		case Cube3D.CORNER_BOTTOM_RIGHT_BACK:
			{
				y = ypos[dist + 1] + yh[dist + 1] - BORDER_THICKNESS;
			}
			break;
		default:
			throw new IllegalArgumentException();
		}
		return y;
	}

	/** 
	 * @param dist the distance of the box
	 * @param xpos the horizontal translation of the box
	 * @return all data for a 2D rectangle: x, y, width, height. 
	 */
	public int[] calcBoxFront(final int dist, final int xpos) {
		final int x = calcLeftX(dist, xpos);
		final int rectX = x + BORDER_THICKNESS;
		final int rectY = ypos[dist] + BORDER_THICKNESS;
		final int rectW = xw[dist] - 2 * BORDER_THICKNESS;
		final int rectH = yh[dist] - 2 * BORDER_THICKNESS;
		return new int[] { rectX, rectY, rectW, rectH };
	}

	protected int[][] calcBoxSide(final int dist, final int xpos, final Side side) {
		final int[][] ret = new int[2][4];
		switch (side) {
		case FRONT:
			{
				ret[X][0] = getX(dist, xpos, Cube3D.CORNER_TOP_LEFT_FRONT);
				ret[Y][0] = getY(dist, xpos, Cube3D.CORNER_TOP_LEFT_FRONT);
				ret[X][1] = getX(dist, xpos, Cube3D.CORNER_TOP_RIGHT_FRONT);
				ret[Y][1] = getY(dist, xpos, Cube3D.CORNER_TOP_RIGHT_FRONT);
				ret[X][2] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_FRONT);
				ret[Y][2] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_FRONT);
				ret[X][3] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_FRONT);
				ret[Y][3] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_FRONT);
			}
			break;
		case BACK:
			break;
		case LEFT:
			{
				ret[X][0] = getX(dist, xpos, Cube3D.CORNER_TOP_LEFT_FRONT);
				ret[Y][0] = getY(dist, xpos, Cube3D.CORNER_TOP_LEFT_FRONT);
				ret[X][1] = getX(dist, xpos, Cube3D.CORNER_TOP_LEFT_BACK);
				ret[Y][1] = getY(dist, xpos, Cube3D.CORNER_TOP_LEFT_BACK);
				ret[X][2] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_BACK);
				ret[Y][2] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_BACK);
				ret[X][3] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_FRONT);
				ret[Y][3] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_FRONT);
			}
			break;
		case RIGHT:
			{
				ret[X][0] = getX(dist, xpos, Cube3D.CORNER_TOP_RIGHT_FRONT);
				ret[Y][0] = getY(dist, xpos, Cube3D.CORNER_TOP_RIGHT_FRONT);
				ret[X][1] = getX(dist, xpos, Cube3D.CORNER_TOP_RIGHT_BACK);
				ret[Y][1] = getY(dist, xpos, Cube3D.CORNER_TOP_RIGHT_BACK);
				ret[X][2] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_BACK);
				ret[Y][2] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_BACK);
				ret[X][3] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_FRONT);
				ret[Y][3] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_FRONT);
			}
			break;
		case TOP:
			{
				ret[X][0] = getX(dist, xpos, Cube3D.CORNER_TOP_LEFT_FRONT);
				ret[Y][0] = getY(dist, xpos, Cube3D.CORNER_TOP_LEFT_FRONT);
				ret[X][1] = getX(dist, xpos, Cube3D.CORNER_TOP_LEFT_BACK);
				ret[Y][1] = getY(dist, xpos, Cube3D.CORNER_TOP_LEFT_BACK);
				ret[X][2] = getX(dist, xpos, Cube3D.CORNER_TOP_RIGHT_BACK);
				ret[Y][2] = getY(dist, xpos, Cube3D.CORNER_TOP_RIGHT_BACK);
				ret[X][3] = getX(dist, xpos, Cube3D.CORNER_TOP_RIGHT_FRONT);
				ret[Y][3] = getY(dist, xpos, Cube3D.CORNER_TOP_RIGHT_FRONT);
			}
			break;
		case BOTTOM:
			{
				ret[X][0] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_FRONT);
				ret[Y][0] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_FRONT);
				ret[X][1] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_BACK);
				ret[Y][1] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_LEFT_BACK);
				ret[X][2] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_BACK);
				ret[Y][2] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_BACK);
				ret[X][3] = getX(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_FRONT);
				ret[Y][3] = getY(dist, xpos, Cube3D.CORNER_BOTTOM_RIGHT_FRONT);
			}
			break;
		default:
			throw new IllegalArgumentException("side was " + side);
		}
		return ret;
	}

	@Override
	protected int[][] calcBoxSide(final int dist, final int xpos, final Side side, final int what) {
		final int[][] xf = calcBoxSide(dist, xpos, side);
		if (Abstract3DGeometry.BOX_SIDE == what) {
			return xf;
		} else if ((Abstract3DGeometry.BOX_EYE1 == what) || (Abstract3DGeometry.BOX_EYE2 == what)) {
			final double sgn = (BOX_EYE1 == what ? 1.0 : -1.0);
			final int[][] eye = Simple3DGeometry.linearTransformTrapez(xf, 
					0.5 + sgn * Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_WIDTH - Abstract3DGeometry.EYE_SIZE / 2.0, 
					Abstract3DGeometry.EYE_CENTER_OFF_CUBE_CENTER_IN_WIDTH - Abstract3DGeometry.EYE_SIZE / 2.0, 
					Abstract3DGeometry.EYE_SIZE, Abstract3DGeometry.EYE_SIZE);
			return eye;
		}
		throw new IllegalArgumentException("what was " + what);
	}

	public static int linearTransform(final int value, final int min, final int max, final double offset, final double factor) {
		return min + (int)Math.round((value - min) * factor + (max - min) * offset);
	}

	public static void linearTransform(final int[] ret, final int[] values, final double offset, final double factor) {
		int min = Integer.MAX_VALUE;
		int max = Integer.MIN_VALUE;
		for (int value : values) {
			min = Math.min(min, value);
			max = Math.max(max, value);
		}
		for (int i = 0; i < values.length; i++) {
			ret[i] = linearTransform(values[i], min, max, offset, factor);
		}
	}

	/** 
	 * Does a linear transformation inside a trapez.
	 * @param original the original coordinates
	 * @param offsetX the x offset
	 * @param offsetY the y offset
	 * @param factorX the x factor
	 * @param factorY the y factor
	 * @return a polygon represented by two arrays containing x and y coordinates
	 */
	public static int[][] linearTransformTrapez(
			final int[][] original, 
			final double offsetX, 
			final double offsetY, 
			final double factorX, 
			final double factorY) {
		if ((null == original) || (null == original[0])) {
			return null;
		}
		if ((2 != original.length) || (4 != original[0].length) || (4 != original[1].length)) {
			return null;
		}
		final int[][] ret = new int[original.length][original[0].length];
		linearTransform(ret[X], original[X], offsetX, factorX);
		linearTransform(ret[Y], original[Y], offsetY, factorY);
		return ret;
	}
}
