package de.ped.troff.middleware;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.net.Socket;
import java.net.URL;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.Marshaller.UnexpectedEOFException;
import de.ped.tools.TextFormatter;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;

/**
 * Represents one side of the connection between client and server.
 * Only runs correctly if the two sides of one connection run in different threads.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Connector {

	public static final boolean CLIENT_SIDE = false;
	public static final boolean SERVER_SIDE = true;
	
	/**
	 * Creates this side of the connection.
	 * 
	 * @param environment
	 *            the environment of the Troff instance
	 * @param socket
	 *            the socket this connection uses, or <code>null</code> to
	 *            indicate the direct connection or web API connection between client and server of
	 *            this instance of Troff
	 * @param url the URL, if a web API is used
	 * @param isServerSide
	 *            {@link #SERVER_SIDE} if this is the connector on the server side, {@link #CLIENT_SIDE} otherwise
	 * @param logger
	 *            the logger where to log any error or debugging messages to. If
	 *            <code>null</code>, the root logger is used.
	 */
	public Connector(
			final Environment environment, 
			final Socket socket, 
			final URL url,
			final boolean isServerSide, 
			final Logger logger) {
		this.logger = (logger == null ? Logger.getRootLogger() : logger);
		this.logger.setLogLevel(Level.INFO);
		this.isServerSide = isServerSide;
		this.socket = socket;
		this.url = url;
		this.environment = environment;
		if (null == socket) {
			if (null == url) {
				try {
					environment.getOrCreateNullSocketInstance();
				} catch (final IOException e) {
					logger.error("Unable to create NullSocket", e);
				}
			} 
		}
	}

	public ClientInfo getClientInfo() {
		return clientInfo;
	}

	public String getClientName() {
		return (clientInfo == null ? null : clientInfo.getName());
	}

	public boolean isTerminating() {
		return isTerminating;
	}

	public String getSocketInfo() {
		String ret = NullSocket.NAME;
		if (null != socket) {
			ret = socket.getInetAddress().getCanonicalHostName();
		} else if (null != url) {
			ret = url.toString();
		}
		return ret;
	}

	protected void throwIllegalStateException(final boolean isInput, final Socket socket) {
		throw new IllegalStateException((isInput ? "input" : "output") + "Stream is null for Socket " + socket + " on " + (isServerSide ? "server" : "client") + " side");
	}

	public void connect() throws IOException {
		if (null != socket) {
			inputStream = new DataInputStream(socket.getInputStream());
			outputStream = new DataOutputStream(socket.getOutputStream());
		} else if (null == url){
			inputStream = new DataInputStream(environment.nullSocket.getInputStream(isServerSide));
			outputStream = new DataOutputStream(environment.nullSocket.getOutputStream(isServerSide));
		}
		if (null == url) {
			if (null == inputStream) {
				throwIllegalStateException(true, socket);
			}
			if (null == outputStream) {
				throwIllegalStateException(false, socket);
			}
		}
		clientInfo = null;
	}

	public void disconnect() {
		isTerminating = true;
		try {
			if (null != outputStream) {
				outputStream.close();
			}
		} catch (final IOException e) {
			// Nothing to do
		}
		try {
			if (null != inputStream) {
				inputStream.close();
			}
		} catch (final IOException e) {
			// Nothing to do
		}
		try {
			if (null != socket) {
				socket.close();
			}
		} catch (final IOException e) {
			// Nothing to do
		}
		clientInfo = null;
	}

	public void send(final Message message) throws IOException {
		final Marshaller marshaller = new Marshaller();
		marshaller.writeMarshallable(message);
		logger.debug("Sending " + message.toString());
		checkForClientInfoChange(message);
		final byte[] pack = marshaller.pack();
		if (null != outputStream) {
			outputStream.write(pack);
		} else {
			if (isServerSide) {
				// FIXME: Implement http protocol
			} else {
				/*
				 * A free Java sample program 
				 * to GET content from any URL using SOCKET connection
				 *
				 * @author William Alexander
				 * free for use as long as this comment is included 
				 * in the program as it is
				 * 
				 * More Free Java programs available for download 
				 * at http://www.java-samples.com
				 *
				 */
				String data = TextFormatter.hexdump(pack, 0, pack.length, "");
				URL sendURL = new URL(url.toString() + "?data=" + data);
				logger.info("URL=" + sendURL.toString());
				try {
					Socket socket = new Socket(sendURL.getHost(), sendURL.getPort());
					PrintWriter out = new PrintWriter(
							new BufferedWriter(new OutputStreamWriter(socket.getOutputStream())));
					final String outstr = "GET " + sendURL.getFile() + sendURL.getQuery() + " HTTP/1.0";
					logger.info("outstr=" + outstr);
					out.println(outstr);
					out.println();
					out.flush();

					BufferedReader inr = new BufferedReader(new InputStreamReader(socket.getInputStream()));
					String inputLine;
					int count = 0;

					while ((inputLine = inr.readLine()) != null) {
						count++;
						System.out.println(count);
						System.out.println(inputLine);
					}
					inr.close();
					System.out.println("PRINTING HERE!!!");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	protected void handleMessage(final Message message) {
		checkForClientInfoChange(message);
		logger.debug("Receiving " + message.toString());
	}

	protected void checkForClientInfoChange(final Message msg) {
		final byte msgType = msg.getMsgType();
		if (msgType == Message.getMsgType(Message.LOGIN, Message.REQUEST)) {
			clientInfo = new ClientInfo(msg.getName(), getSocketInfo(), msg.getNumberOfPlayers());
		} else if (msgType == Message.getMsgType(Message.LOGIN, Message.RESPONSE)) {
			if (!msg.wasSuccess()) {
				clientInfo = null;
			}
		}
	}

	/**
	 * Does one try to read a message.
	 * 
	 * @return the message received, or <code>null</code> if there was no
	 *         (complete) message to receive
	 * @throws IOException
	 *             if the message could not be read
	 */
	public Message receive() throws IOException {
		Message message = null;
		if (null == url) {
			try {
				if (0 < inputStream.available()) {
					final int bytesRead = inputStream.read(readBuffer);
					if (0 < bytesRead) {
						reader.writeFixedLengthBytesArray(readBuffer, bytesRead);
					}
				}
			} catch (final IOException ex) { // Typically a stream closed.
				isTerminating = true;
				throw ex;
			}
		} else { // http connection
			if (isServerSide) {
				// FIXME: Fill with life
			} else {
				// FIXME: Fill with life
			}
		}
		try {
			final Marshallable marshallable = reader.readMarshallable(messageFactory);
			message = (Message)marshallable;
			handleMessage(message);
		} catch (final UnexpectedEOFException ex) {
			// Nothing to do
		}
		return message;
	}

	/**
	 * Infinitely tries to read a message.
	 * 
	 * @param ticks
	 *            time to wait (in milliseconds) between two tries to read the
	 *            message
	 * @return the message received, or <code>null</code> if there was no
	 *         (complete) message to receive
	 * @throws IOException
	 *             if the message could not be read for technical reasons
	 */
	public Message receiveWaiting(final long ticks) throws IOException {
		final long millisStart = System.currentTimeMillis();
		while (true) {
			final Message msg = receive();
			if (null != msg) {
				return msg;
			}
			final long millisEnd = System.currentTimeMillis();
			if (millisStart + ticks < millisEnd) {
				return null;
			}
			try {
				Thread.sleep(10L);
			} catch (final InterruptedException x) {
				// Nothing to do
			}
		}
	}

	protected Environment environment;
	protected Socket socket;
	protected URL url;
	protected boolean isServerSide;
	protected DataInputStream inputStream;
	protected DataOutputStream outputStream;

	protected boolean isTerminating = false;
	protected Logger logger; 
	protected Marshaller reader = new Marshaller();
	protected byte[] readBuffer = new byte[0x100];
	protected Message messageFactory = Message.createMessageFactory();
	protected ClientInfo clientInfo;
}
