package de.ped.troff.middleware;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * The base class for all messages between client and server.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class Message implements Marshallable {
	// Types
	public static final byte TYPE_MASK = -0x80;
	public static final byte REQUEST = 0x00;
	public static final byte RESPONSE = -0x80;
	public static final String[] TYPE_NAMES = new String[] { "Req", "Res", };

	// Messages
	public static final byte MSG_MASK = 0x7f;
	public static final byte LOGIN = 0x00;
	public static final byte LOGOUT = 0x01;
	public static final byte JOIN = 0x02;
	public static final byte LEAVE = 0x03;
	public static final byte SERVER_STATUS = 0x04;
	public static final byte MESSAGES = 0x05;
	public static final byte START_GAME = 0x06;
	public static final byte PAUSE_RESUME = 0x07;
	public static final byte GAME_STATE = 0x08;
	public static final byte GAME_ACTION = 0x09;
	public static final byte SEND_MESSAGE = 0x0a;
	public static final String[] MSG_NAMES = new String[] {
		"Login", "Logout", "Join", "Leave", "ServerStatus", "Messages", "StartGame", "PauseResume", "GameState", "GameAction", "SendMessage", 
	};

	public static byte getMsgType(final byte msg, final byte type) {
		return (byte)(msg | type);
	}

//	static class Parameter {
//		final int id;
//		final String key;
//		@SuppressWarnings("rawtypes")
//		final Class type;
//		final byte[] containingMessages;
//		Parameter(final int id, final String key, @SuppressWarnings("rawtypes") final Class value, final byte[] containingMessages) {
//			this.id = id;
//			this.key = key;
//			this.type = value;
//			this.containingMessages = containingMessages;
//		}
//	}
//
//	protected static final Parameter[] params = new Parameter[] {
//			new Parameter(0, "name", String.class, new byte[] {
//					// On login, it's a clientName
//					Message.LOGIN | Message.REQUEST, 
//					Message.LOGIN | Message.RESPONSE, 
//					// On join, it's a groupName
//					Message.JOIN | Message.REQUEST, 
//					Message.JOIN | Message.RESPONSE, 
//					// On a message request, it's the message to broadcast
//					Message.MESSAGE | Message.REQUEST, 
//			}),
//			new Parameter(1, "playerIdx", Integer.class, new byte[] { 
//					// For login, this is the number of players
//					Message.LOGIN | Message.REQUEST, 
//					Message.LOGIN | Message.RESPONSE, 
//					// For game action, this is the player id
//					Message.GAME_ACTION | Message.REQUEST, 
//					Message.GAME_ACTION | Message.RESPONSE, 
//			}),
//			new Parameter(2, "action", Integer.class, new byte[] { 
//					Message.GAME_ACTION | Message.REQUEST, 
//					Message.GAME_ACTION | Message.RESPONSE, 
//			}),
//			new Parameter(3, "isOn", Boolean.class, new byte[] { 
//					// On login and join, it's a success flag
//					Message.LOGIN | Message.RESPONSE, 
//					Message.JOIN  | Message.RESPONSE, 
//					// On pause_resume request, it's to resume
//					Message.PAUSE_RESUME | Message.REQUEST, 
//					// On pause_resume response, it's to indicate it's running
//					Message.PAUSE_RESUME | Message.RESPONSE, 
//			}),
//			new Parameter(4, "serverInfo", ServerInfo.class, new byte[] { 
//					Message.LOGOUT        | Message.RESPONSE, 
//					Message.SERVER_STATUS | Message.RESPONSE, 
//			}),
//			new Parameter(5, "properties", Properties.class, new byte[] { 
//					Message.START_GAME   | Message.REQUEST, 
//					Message.START_GAME   | Message.RESPONSE, 
//			}),
//			new Parameter(6, "messages", String[].class, new byte[] { 
//					Message.MESSAGES     | Message.RESPONSE, 
//			}),
//			new Parameter(7, "gameState", GameState.class, new byte[] { 
//					Message.GAME_STATE   | Message.RESPONSE, 
//			}),
//	};
//
	/** 
	 * Creates a message factory, not a message itself. 
	 * @return a message factory object
	 */
	public static Message createMessageFactory() {
		// Message type is bogus, to not be confused with a real message if tried to use as that.
		return new Message(Message.getMsgType(TYPE_MASK, MSG_MASK), null, 0, 0, false, null, null, null, null);
	}

	public static Message createLeaveReq() {
		return Message.createNoParamMessage(Message.LEAVE, Message.REQUEST);
	}

	public static Message createLogoutReq() {
		return Message.createNoParamMessage(Message.LOGOUT, Message.REQUEST);
	}

	public static Message createServerStatusReq() {
		return Message.createNoParamMessage(Message.SERVER_STATUS, Message.REQUEST);
	}

	public static Message createLeaveRes() {
		return Message.createNoParamMessage(Message.LEAVE, Message.RESPONSE);
	}

	protected static Message createNoParamMessage(final byte msg, final byte type) {
		return new Message(Message.getMsgType(msg, type), null, 0, 0, false, null, null, null, null);
	}

	public static Message createJoinReq(final String name) {
		return Message.createStringParamMessage(Message.JOIN, Message.REQUEST, name);
	}

	public static Message createSendMessageReq(final String name) {
		return Message.createStringParamMessage(Message.SEND_MESSAGE, Message.REQUEST, name);
	}

	protected static Message createStringParamMessage(final byte msg, final byte type, final String name) {
		return new Message(Message.getMsgType(msg, type), name, 0, 0, false, null, null, null, null);
	}

	public static Message createLoginReq(final String name, final byte playerId) {
		return new Message(Message.getMsgType(Message.LOGIN, Message.REQUEST), name, playerId, 0, false, null, null, null, null);
	}

	public static Message createLoginRes(final String name, final byte playerIdx, final boolean wasSuccess) {
		return new Message(Message.getMsgType(Message.LOGIN, Message.RESPONSE), name, playerIdx, 0, wasSuccess, null, null, null, null);
	}

	public static Message createJoinRes(final String name, final boolean wasSuccess) {
		return new Message(Message.getMsgType(Message.JOIN, Message.RESPONSE), name, 0, 0, wasSuccess, null, null, null, null);
	}

	public static Message createGameActionReq(final int playerIdx, final int action) {
		return Message.createGameActionMessage(Message.GAME_ACTION, Message.REQUEST, playerIdx, action);
	}

	public static Message createGameActionRes(final int playerIdx, final int action) {
		return Message.createGameActionMessage(Message.GAME_ACTION, Message.RESPONSE, playerIdx, action);
	}

	protected static Message createGameActionMessage(final byte msg, final byte type, final int playerIdx, final int action) {
		return new Message(Message.getMsgType(msg, type), null, playerIdx, action, false, null, null, null, null);
	}

	public static Message createPauseResumeReq(final boolean isOn) {
		return Message.createPauseResumeMessage(Message.PAUSE_RESUME, Message.REQUEST, isOn);
	}

	public static Message createPauseResumeRes(final boolean isOn) {
		return Message.createPauseResumeMessage(Message.PAUSE_RESUME, Message.RESPONSE, isOn);
	}

	protected static Message createPauseResumeMessage(final byte msg, final byte type, final boolean isOn) {
		return new Message(Message.getMsgType(msg, type), null, 0, 0, isOn, null, null, null, null);
	}

	public static Message createLogoutRes(final ServerInfo result) {
		return Message.createServerInfoResponseMessage(Message.LOGOUT, Message.RESPONSE, result);
	}

	public static Message createServerStatusRes(final ServerInfo result) {
		return Message.createServerInfoResponseMessage(Message.SERVER_STATUS, Message.RESPONSE, result);
	}

	protected static Message createServerInfoResponseMessage(final byte msg, final byte type, final ServerInfo result) {
		return new Message(Message.getMsgType(msg, type), null, 0, 0, false, result, null, null, null);
	}

	public static Message createStartGameReq(final TroffGameProperties properties) {
		return Message.createStartGameMessage(Message.START_GAME, Message.REQUEST, properties);
	}

	public static Message createStartGameRes(final TroffGameProperties properties) {
		return Message.createStartGameMessage(Message.START_GAME, Message.RESPONSE, properties);
	}

	protected static Message createStartGameMessage(final byte msg, final byte type, final TroffGameProperties properties) {
		return new Message(Message.getMsgType(msg, type), null, 0, 0, false, null, properties, null, null);
	}

	public static Message createMessagesRes(final I18NStringWithFillIns[] logbook) {
		return new Message(Message.getMsgType(Message.MESSAGES, Message.RESPONSE), null, 0, 0, false, null, null, logbook, null);
	}

	public static Message createGameStateRes(final TroffGameState gameState) {
		return new Message(Message.getMsgType(Message.GAME_STATE, Message.RESPONSE), null, 0, 0, false, null, null, null, gameState);
	}

	protected Message(
			final byte msgtype, 
			final String name,
			final int playerIdx,
			final int action,
			final boolean isOn,
			final ServerInfo serverInfo,
			final TroffGameProperties properties,
			final I18NStringWithFillIns[] messages,
			final TroffGameState gameState) {
		this.msgtype = msgtype; 
		this.name = name;
		this.playerIdx = playerIdx;
		this.action = action;
		this.isOn = isOn;
		this.serverInfo = serverInfo;
		this.properties = properties;
		this.messages = messages;
		this.gameState = gameState;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeByte(msgtype);
		marshaller.writeString(name);
		marshaller.writeInt(playerIdx);
		marshaller.writeInt(action);
		marshaller.writeBoolean(isOn);
		marshaller.writeNullOrMarshallable(serverInfo);
		marshaller.writeNullOrMarshallable(properties);
		marshaller.writeMarshallableArray(messages);
		marshaller.writeNullOrMarshallable(gameState);
	}

	@Override
	public final Marshallable demarshal(final Marshaller marshaller) {
		return new Message(
				marshaller.readByte(),
				marshaller.readString(),
				marshaller.readInt(),
				marshaller.readInt(),
				marshaller.readBoolean(),
				(ServerInfo) marshaller.readNullOrMarshallable(new ServerInfo(null, null, null)),
				(TroffGameProperties) marshaller.readNullOrMarshallable(new TroffGameProperties()),
				readMessages(marshaller),
				(TroffGameState) marshaller.readNullOrMarshallable(new TroffGameState(0))
		);
	}

	private I18NStringWithFillIns[] readMessages(final Marshaller marshaller) {
//		return (I18NStringWithFillIns[]) marshaller.readMarshallableArray(I18NStringWithFillIns.EMPTY);
		final int arrLength = marshaller.readInt();
		if (Marshaller.ARRAY_NULL_INDICATOR == arrLength) {
			return null;
		}
		final I18NStringWithFillIns[] arr = new I18NStringWithFillIns[arrLength];
		for (int i = 0; i < arrLength; i++) {
			final boolean isNull = marshaller.readNullIndicator();
			final I18NStringWithFillIns is;
			if (isNull) {
				is = null;
			} else {
				is = new I18NStringWithFillIns(marshaller);
			}
			arr[i] = is;
		}
		return arr;
	}
	

	@Override
	public String toString() {
		final StringBuffer ret = new StringBuffer();
		ret.append(getClass().getSimpleName() + "(" + MSG_NAMES[getMsg()] + TYPE_NAMES[(getType() == REQUEST ? 0 : 1)]);
		ret.append("," + playerIdx);
		ret.append("," + action);
		ret.append("," + isOn);
		if (null != name) {
			ret.append("," + name);
		}
		if (null != serverInfo) {
			ret.append("," + "ServerInfo(" + serverInfo.getName() + ", " + serverInfo.getNumberOfPlayers() + ")");
		}
		if (null != properties) {
			ret.append("," + properties);
		}
		if (null != messages) {
			ret.append("," + messages);
		}
		if (null != gameState) {
			ret.append("," + "GameState(" + gameState.getStep() + ", " + gameState.isAlive() + ")");
		}
		ret.append(")");
		return ret.toString();
	}

	public byte getMsgType() {
		return msgtype;
	}

	public byte getType() {
		return (byte)(msgtype & TYPE_MASK);
	}

	public byte getMsg() {
		return (byte)(msgtype & MSG_MASK);
	}

	public String getText() {
		return name;
	}

	public String getName() {
		return name;
	}

	public byte getNumberOfPlayers() {
		return (byte) playerIdx;
	}

	public int getPlayerIdx() {
		return playerIdx;
	}

	public int getAction() {
		return action;
	}

	public boolean wasSuccess() {
		return isOn;
	}

	public boolean isToResume() {
		return isOn;
	}

	public boolean isRunning() {
		return isOn;
	}

	public ServerInfo getServerInfo() {
		return serverInfo;
	}

	public TroffGameProperties getProperties() {
		return properties;
	}

	public I18NStringWithFillIns[] getGroupMessages() {
		return messages;
	}

	public TroffGameState getGameState() {
		return gameState;
	}

	protected final byte msgtype;
	protected final String name;
	protected final int playerIdx;
	protected final int action;
	protected final boolean isOn;
	protected final ServerInfo serverInfo;
	protected final TroffGameProperties properties;
	protected final I18NStringWithFillIns[] messages;
	protected final TroffGameState gameState;
}
