package de.ped.troff.middleware;

/**
 * The model for server connection and selection.
 * Contains information if this instance is acting as server,
 * and to what server the client part of this instance is connecting.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ServerSelectionModel implements Cloneable {
	public static final int PORTNO_LOCALHOST_DEFAULT = 81;
	public static final String HOSTNAME_LOCALHOST = "localhost";
	public static final String HOSTNAME_EXTERNAL_DEFAULT = "localhost";
	public static final int PORTNO_EXTERNAL_DEFAULT = 81;
	public static final String CLIENTNAME_DEFAULT = "unnamed";
	public static final String SERVERNAME_DEFAULT = "Server";
	public static final String SERVERNAME_LOCAL = "local";

	/**
	 * Defines the host type. From that the system can derive what kind of
	 * protocol client and server use, and what the client can choose to define
	 * the server instance.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public enum HostType {
		LOCALLY,
		PORT_LOCALHOST,
		PORT_EXTERNAL_DEFAULT,
		PORT_EXTERNAL_OWN,
		HTTP_LOCALHOST;
		
		public static final String KEY = "ServerSelectionWindow.ClientPanel.HostType";
	}

	public ServerSelectionModel() {
		// Nothing to do, all members are already initialized.
	}

	public ServerSelectionModel(
			final boolean isServer, 
			final int serverPortNo, 
			final String serverName, 
			final HostType hostType, 
			final String hostName, 
			final int hostPortNo, 
			final String clientName, 
			final int numberOfPlayers) {
		this.isServer = isServer;
		this.serverPortNo = serverPortNo;
		this.serverName = serverName;
		setHostType(hostType);
		this.hostName = hostName;
		this.hostPortNo = hostPortNo;
		this.clientName = clientName;
		this.numberOfPlayers = numberOfPlayers;
	}

	public boolean isServer() {
		return isServer;
	}

	public void setServer(final boolean isServer) {
		this.isServer = isServer;
	}

	public int getServerPortNo() {
		return serverPortNo;
	}

	public void setServerPortNo(final int serverPortNo) {
		this.serverPortNo = serverPortNo;
	}

	public String getServerName() {
		return serverName;
	}

	public void setServerName(final String serverName) {
		this.serverName = serverName;
	}

	public HostType getHostType() {
		return hostType;
	}

	public void setHostType(final HostType hostType) {
		this.hostType = hostType;
	}

	public boolean isUsingDirectServer() {
		return (hostType == HostType.LOCALLY);
	}

	public String getOwnHostName() {
		return hostName;
	}

	public void setOwnHostName(final String hostName) {
		this.hostName = hostName;
	}

	public int getOwnHostPortNo() {
		return hostPortNo;
	}

	public void setOwnHostPortNo(final int hostPortNo) {
		this.hostPortNo = hostPortNo;
	}

	public String getHostName() {
		String ret = null;
		switch(hostType) {
		case LOCALLY:
			break;
		case PORT_LOCALHOST:
			ret = HOSTNAME_LOCALHOST;
			break;
		case PORT_EXTERNAL_DEFAULT:
			ret = HOSTNAME_EXTERNAL_DEFAULT;
			break;
		case PORT_EXTERNAL_OWN:
			ret = hostName;
			break;
		case HTTP_LOCALHOST:
			ret = "http://localhost:8080/troff/index.jsp";
			break;
		}
		return ret;
	}

	public int getHostPortNo() {
		int ret = 0;
		switch(hostType) {
		case LOCALLY:
			break;
		case PORT_LOCALHOST:
			ret = serverPortNo;
			break;
		case PORT_EXTERNAL_DEFAULT:
			ret = PORTNO_EXTERNAL_DEFAULT;
			break;
		case PORT_EXTERNAL_OWN:
			ret = hostPortNo;
			break;
		case HTTP_LOCALHOST:
			break;
		}
		return ret;
	}

	public String getClientName() {
		return clientName;
	}

	public void setClientName(final String clientName) {
		this.clientName = clientName;
	}

	public int getNumberOfPlayers() {
		return numberOfPlayers;
	}

	public void setNumberOfPlayers(final int numberOfPlayers) {
		this.numberOfPlayers = numberOfPlayers;
	}

	public boolean isToStartServer() {
//		return (isServer() || (CONNECT_LOCALLY == getHostType()));
		return true;
	}

	public String getServerNameForNaming() {
		String ret = getServerName();
		if (isServer()) {
			ret = getServerName();
		} else {
			ret = SERVERNAME_LOCAL;
		}
		return ret;
	}

	@Override
	public ServerSelectionModel clone() {
		try {
			return (ServerSelectionModel) super.clone();
		} catch (final CloneNotSupportedException e) {
			return null;
		}
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + "((" + isServer + "," + serverPortNo + "," + serverName + ")(" + hostType
				+ "," + hostName + "," + hostPortNo + "," + clientName + "," + numberOfPlayers + "))";
	}

	protected boolean isServer = false;
	protected int serverPortNo = PORTNO_LOCALHOST_DEFAULT;
	protected String serverName = SERVERNAME_DEFAULT;
	protected HostType hostType = HostType.LOCALLY;
	protected String hostName = HOSTNAME_LOCALHOST;
	protected int hostPortNo = PORTNO_LOCALHOST_DEFAULT;
	protected String clientName = CLIENTNAME_DEFAULT;
	protected int numberOfPlayers = 1;
}