package de.ped.troff.middleware;

import java.io.IOException;
import java.net.Socket;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.HashMap;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.log.Logger;
import de.ped.troff.client.logic.ClientThread;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * Offers an interface for any client connecting to a multi-client game.
 * In fact, acts as a proxy, encapsulating the client/server communication.
 * Also keeps all the last messages that went in or out.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffRoundsManager {

	/** The delay in milliseconds between two intro steps. */
	public static final long INTRO_DELAY = 2400L;

	public static final long DEFAULT_WAIT_TIME = 100L;

	public static final boolean IS_RUNNING_DEFAULT = true;

	public TroffRoundsManager(
			final ClientThread parent, 
			final String host, 
			final int portno, 
			final Logger logger) throws UnknownHostException, IOException {
		this.parent = parent;
		this.logger = (logger == null ? Logger.getRootLogger() : logger);
		Socket socket = null;
		URL url = null;
		if (null != host) {
			if (0 != portno) {
				socket = new Socket(host, portno);
			} else {
				url = new URL(host);
			}
		}
		connector = new Connector(parent.getEnvironment(), socket, url, Connector.CLIENT_SIDE, logger);
		connector.connect();
	}

	public Message getLastMessage(final byte msgType) {
		return map.get(Byte.valueOf(msgType));
	}

	protected void addLastMessage(final Message message) {
		if (null != message) {
			map.put(Byte.valueOf(message.getMsgType()), message);
		}
	}

	protected void clearLastMessage(final byte msgType) {
		map.remove(Byte.valueOf(msgType));
	}

	public void sendMessage(final Message req) {
		synchronized (this) {
			addLastMessage(req);
		}
		try {
			send(req);
		} catch (final IOException ex) {
			logger.error("", ex);
		}
	}

	public synchronized I18NStringWithFillIns[] getLastReceivedTextMessages() {
		final Message msg = getLastMessage(Message.getMsgType(Message.MESSAGES, Message.RESPONSE));
		return (msg == null ? null : msg.getGroupMessages());
	}

	public synchronized ServerInfo getLastReceivedServerStatus() {
		final Message msg = getLastMessage(Message.getMsgType(Message.SERVER_STATUS, Message.RESPONSE));
		return (msg == null ? null : msg.getServerInfo());
	}

	public synchronized TroffGameProperties getLastReceivedStartGame() {
		final Message msg = getLastMessage(Message.getMsgType(Message.START_GAME, Message.RESPONSE));
		return (msg == null ? null : msg.getProperties());
	}

	public synchronized TroffGameState getLastReceivedGameState() {
		final Message msg = getLastMessage(Message.getMsgType(Message.GAME_STATE, Message.RESPONSE));
		return (msg == null ? null : msg.getGameState());
	}

	/** 
	 * Mainly a shortcut to ease programming of test code. In productive environment considered harmful.
	 * Receives and discards all messages until it finds the expected response to a given request.
	 * 
	 * @param lastSentRequest the request we expect to receive a response for
	 * @return the response for the request
	 * @exception IOException if something went wrong while reading
	 */
	public Message readMessagesTillResponse(final Message lastSentRequest) throws IOException {
		Message response = null;
		if (null != lastSentRequest) {
			while (null == response) {
				final Message received = receive(DEFAULT_WAIT_TIME);
				if (null != received) {
					handleMessage(received);
					if (   (received.getMsg() == lastSentRequest.getMsg())
						&& (received.getType() == Message.RESPONSE)
						&& (lastSentRequest.getType() == Message.REQUEST)) {
						response = received;
					}
				}
			}
		}
		return response;
	}

	/** 
	 * Reads and handles messages from the queue, but does not wait for them. 
	 * @exception IOException if something went wrong while reading
	 */
	public void readMessages() throws IOException {
		Message received = null;
		do {
			received = receive(0L);
			handleMessage(received);
		} while (null != received);
	}

	/**
	 * Handles a message that came from the server, depending on if it is the
	 * answer to a request or a unrequested notification.
	 * 
	 * @param msg
	 *            the message to handle
	 */
	protected void handleMessage(final Message msg) {
		if (null != msg) {
			updateStateForReceivingMessage(msg);
			parent.update(this, msg);
		}
	}

	/**
	 * Handles a message that came from the server,
	 * depending on if it is the answer to a request or a unrequested notification.
	 * 
	 * @param msg
	 *            the message to handle
	 */
	protected synchronized void updateStateForReceivingMessage(final Message msg) {
		if (null != msg) {
			addLastMessage(msg);
			switch(msg.getMsgType()) {
			case Message.LOGOUT        | Message.RESPONSE: 
				clearLastMessage(Message.getMsgType(Message.LOGIN,       Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.JOIN,        Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.LEAVE,       Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.START_GAME,  Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.GAME_STATE,  Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.GAME_ACTION, Message.RESPONSE));
				break;
			case Message.LEAVE         | Message.RESPONSE: 
				clearLastMessage(Message.getMsgType(Message.JOIN,        Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.START_GAME,  Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.GAME_STATE,  Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.GAME_ACTION, Message.RESPONSE));
				break;
			case Message.LOGIN         | Message.RESPONSE: 
				clearLastMessage(Message.getMsgType(Message.LOGOUT,      Message.RESPONSE));
				break;
			case Message.JOIN          | Message.RESPONSE: 
				clearLastMessage(Message.getMsgType(Message.LEAVE,       Message.RESPONSE));
				clearLastMessage(Message.getMsgType(Message.LOGOUT,      Message.RESPONSE));
				break;
			case Message.GAME_ACTION   | Message.RESPONSE: 
			case Message.SERVER_STATUS | Message.RESPONSE: 
			case Message.MESSAGES      | Message.RESPONSE: 
			case Message.START_GAME    | Message.RESPONSE: 
			case Message.PAUSE_RESUME  | Message.RESPONSE: 
			case Message.GAME_STATE    | Message.RESPONSE: 
			case Message.SEND_MESSAGE  | Message.RESPONSE: 
				break;
			default:
				throw new IllegalArgumentException("I'm a client, not a doctor. Can't deal with message type " + msg.getMsgType());
			}
		}
	}

	public Connector getConnector() {
		return connector;
	}

	public void disconnect() {
		if (null != connector) {
			connector.disconnect();
		} else {
			logger.error("Connector is null!");
		}
	}

	protected void send(final Message req) throws IOException {
		if (null != connector) {
			connector.send(req);
		} else {
			logger.error("Connector is null!");
		}
	}

	protected Message receive(final long waitTime) throws IOException {
		return connector.receiveWaiting(waitTime);
	}

	protected final ClientThread parent;
	protected final Logger logger;

	protected HashMap<Byte, Message> map = new HashMap<Byte, Message>();

	protected final Connector connector;
}
