package de.ped.troff.server;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.TreeSet;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.log.Logger;
import de.ped.troff.middleware.ClientInfo;
import de.ped.troff.middleware.Connector;
import de.ped.troff.middleware.GroupInfo;
import de.ped.troff.middleware.ServerInfo;
import de.ped.troff.middleware.ServerSelectionModel;
import de.ped.troff.middleware.Message;
import de.ped.troff.middleware.TroffRoundsManager;
import de.ped.troff.server.logic.TroffGame;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * Implements the server facade for any client connecting to a multi-client game.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffRoundsManagerImpl implements Serializable {

	private static final long serialVersionUID = 1L;

	public static final String DEFAULT_GROUP_NAME = "default";

	static final TroffServerGroup[] GROUP_ARRAY_TEMPLATE = new TroffServerGroup[]{};
	static final TroffServerClient[] CLIENT_ARRAY_TEMPLATE = new TroffServerClient[]{};

	public TroffRoundsManagerImpl() {
		clientGroups.add(new TroffServerGroup(this, DEFAULT_GROUP_NAME));
	}

	public synchronized Message handleMessage(final Connector connector, final Message msg) {
		Message response = null;
		if (null != msg) {
			switch(msg.getMsgType()) {
			case Message.LOGIN         | Message.REQUEST: 
				response = Message.createLoginRes(msg.getName(), (byte) msg.getPlayerIdx(), loginConnector(connector));
				break;
			case Message.LOGOUT        | Message.REQUEST: 
				response = Message.createLogoutRes(logoutClientName(connector.getClientName()));
				break;
			case Message.JOIN          | Message.REQUEST:
				final String groupName = msg.getName(); 
				response = Message.createJoinRes(groupName, joinClientName(connector.getClientName(), groupName));
				break;
			case Message.LEAVE         | Message.REQUEST:
				leaveClientName(connector.getClientName()); 
				response = Message.createLeaveRes();
				break;
			case Message.SERVER_STATUS | Message.REQUEST:
				response = Message.createServerStatusRes(getState(getClientByConnector(connector)));
				break;
			case Message.MESSAGES      | Message.REQUEST: 
				response = Message.createMessagesRes(getTextMessages(connector.getClientName()));
				break;
			case Message.START_GAME    | Message.REQUEST:
				startGame(getClientByConnector(connector), msg.getProperties()); 
				break;
			case Message.PAUSE_RESUME  | Message.REQUEST:
				final String name = connector.getClientName();
				final TroffServerClient client = getClientByName(name);
				if (null != client) {
					pauseOrResumeGame(client, msg.isToResume());
				} else {
					throw new IllegalArgumentException("No client found named " + name + " for connector " + connector);
				} 
				break;
			case Message.GAME_STATE    | Message.REQUEST:
				final TroffGameState state = getClientGroupByClient(getClientByConnector(connector)).getGame().getGameState(); 
				response = Message.createGameStateRes(state);
				break;
			case Message.GAME_ACTION   | Message.REQUEST:
				gameActionInput(getClientByConnector(connector), msg.getPlayerIdx(), msg.getAction()); 
				break;
			case Message.SEND_MESSAGE       | Message.REQUEST:
				receiveGroupMessage(getClientByConnector(connector), msg.getText()); 
				break;
			default:
				throw new IllegalArgumentException("I'm a server, not a doctor. Can't deal with message type " + msg.getMsgType());
			}
			logger.debug(msg.toString() + " --> " + (null == response ? "n/a" : response.toString()));
		}
		return response;
	}

	/**
	 * Lets the given client join the given group. If necessary, the client
	 * leaves any other group.
	 * 
	 * @param client
	 *            the client
	 * @param clientGroup
	 *            the group
	 * @return <code>true</code> if the client is in that group afterwards
	 */
	public synchronized boolean join(final TroffServerClient client, final TroffServerGroup clientGroup) {
		loginClient(client);
		if (null == clientGroup) {
			return false;
		}
		final TroffServerGroup currentRound = getClientGroupByClient(client);
		if (null == currentRound) {
			return clientGroup.join(client);
		} else if (clientGroup == currentRound) {
			return true;
		} else {
			leave(client);
			return clientGroup.join(client);
		}
	}

	public synchronized void leave(final TroffServerClient client) {
		final TroffServerGroup clientGroup = getClientGroupByClient(client);
		if (null != clientGroup) {
			clientGroup.leave(client);
			final TroffServerClient[] clients = clientGroup.getClients();
			if (((null == clients) || (0 == clients.length)) 
					&& (!DEFAULT_GROUP_NAME.equals(clientGroup.getName()))) {
				clientGroups.remove(clientGroup);
			}
		}
	}

	private synchronized void startGame(final TroffServerClient client, final TroffGameProperties props) {
		final TroffServerGroup clientGroup = getOrJoinClientGroup(client);
		if (null != clientGroup) {
			clientGroup.startGame(props);
		}
	}

	private void gameActionInput(final TroffServerClient client, final int playerIdx, final int suggestion) {
		final TroffServerGroup clientGroup = getOrJoinClientGroup(client);
		if (null != clientGroup) {
			clientGroup.suggestUserGameAction(playerIdx, suggestion);
		}
	}

	private void receiveGroupMessage(final TroffServerClient client, final String message) {
		final TroffServerGroup clientGroup = getOrJoinClientGroup(client);
		if (null != clientGroup) {
			final I18NStringWithFillIns is = new I18NStringWithFillIns("Game.Message.Message", "%0 %1 %2: %3",
					new String[] { null, getName(), client.getName(), message});
			clientGroup.receiveMessage(is);
		}
	}

	private boolean pauseOrResumeGame(final TroffServerClient client, final boolean isToResume) {
		final TroffServerGroup clientGroup = getOrJoinClientGroup(client);
		boolean result = TroffRoundsManager.IS_RUNNING_DEFAULT;
		if (null != clientGroup) {
			result = clientGroup.pauseOrResumeGame(client, isToResume);
		}
		return result;
	}

	private boolean loginConnector(final Connector connector) {
		final TroffServerClient client = new TroffServerClient(connector);
		return loginClient(client);
	}

	private ServerInfo logoutClientName(final String clientName) {
		final TroffServerClient client = getClientByName(clientName);
		logoutClient(client);
		return getState(client);
	}

	public boolean loginClient(final TroffServerClient client) {
		boolean didSomething = false;
		if (null != client) {
			didSomething = clients.add(client);
			if (didSomething) {
				client.receiveMessage(new I18NStringWithFillIns("Game.Message.Login", "%0: +%1",
						new String[] { getName(), client.getName(), }));
				logger.debug("Client " + client.getName() + " connected to server " + getName());
			} else {
				logger.debug("Client " + client.getName() + " re-connected to Server " + getName());
			}
		} else {
			logger.error("Login with null client!");
		}
		return didSomething;
	}

	public void logoutClient(final TroffServerClient client) {
		leave(client);
		if (null != client) {
			clients.remove(client);
			client.receiveMessage(new I18NStringWithFillIns("Game.Message.Logout", "%0: +%1",
					new String[] { getName(), client.getName(), }));
			logger.debug("Client " + client.getName() + " disconnected from Server " + getName());
		}
	}

	private boolean joinClientName(final String clientName, final String groupName) {
		final TroffServerClient client = getClientByName(clientName);
		final boolean result = joinClient(client, groupName);
		return result;
	}

	/**
	 * Lets the given client join the given group. If necessary, a group with
	 * that name is created. If necessary, the client leaves any other group.
	 * 
	 * @param client
	 *            the client
	 * @param clientGroupName
	 *            the group name
	 * @return <code>true</code> if the client is in that group afterwards
	 */
	private synchronized boolean joinClient(final TroffServerClient client, final String clientGroupName) {
		TroffServerGroup clientGroup = getClientGroupByName(clientGroupName);
		if (null == clientGroup) {
			clientGroup = new TroffServerGroup(this, clientGroupName);
			clientGroups.add(clientGroup);
		}
		return join(client, clientGroup);
	}

	private void leaveClientName(final String clientName) {
		final TroffServerClient client = getClientByName(clientName);
		leave(client);
	}

	private I18NStringWithFillIns[] getTextMessages(final String clientName) {
		final TroffServerClient client = getClientByName(clientName);
		I18NStringWithFillIns[] msgs = null;
		if (null != client) {
			msgs = client.getTextMessages();
		} else {
			logger.debug("No client named " + clientName);
		}
		return msgs;
	}

	private TroffServerClient getClientByConnector(final Connector connector) {
		return getClientByName(connector.getClientName());
	}

	private TroffServerClient getClientByName(final String name) {
		TroffServerClient ret = null;
		for (final TroffServerClient client : clients) {
			if (null != client) {
				final String clientName = client.getName();
				if (null == clientName) {
					if (null == name) {
						ret = client;
						break;
					}
				} else {
					if (clientName.equals(name)) {
						ret = client;
						break;
					}
				}
			}
		}
		return ret;
	}

	private ServerInfo getState(final TroffServerClient callingClient) {
		final TroffServerGroup[] cgroups = getAllClientGroups();
		final ArrayList<GroupInfo> groups = new ArrayList<GroupInfo>();
		for (int i = 0; i < cgroups.length; i++) {
			final ArrayList<ClientInfo> clients = new ArrayList<ClientInfo>();
			final TroffServerClient[] cclients = cgroups[i].getClients();
			for (int c = 0; c < cclients.length; c++) {
				final ClientInfo ci = cclients[c].getClientInfo();
				if (null == ci) {
					throw new IllegalStateException("ClientInfo is null for client " + cclients[c] + " in group " + cgroups[i]);
				}
				clients.add(cclients[c].getClientInfo());
			}
			groups.add(new GroupInfo(cgroups[i].getName(), clients));
		}
		final TroffServerGroup group = getClientGroupByClient(callingClient);
		final String groupName = (group == null ? null : group.getName());
		final ServerInfo state = new ServerInfo(getName(), groups, groupName);
		return state;
	}

	private TroffServerGroup getClientGroupByName(final String name) {
		for (final TroffServerGroup round : clientGroups) {
			if (round.getName().equals(name)) {
				return round;
			}
		}
		return null;
	}

	@SuppressWarnings("unused")
	private String showState() {
		final StringBuffer s = new StringBuffer();
		s.append(getName());
		s.append("\n");
		for (final TroffServerGroup round : clientGroups) {
			s.append("  ");
			s.append(round.getName());
			s.append("\n");
			for (final TroffServerClient  client : round.getClients()) {
				s.append("    ");
				s.append(client.getName());
				s.append("\n");
			}
		}
		return s.toString();
	}

	@SuppressWarnings("unused")
	private TroffGameProperties getGameProperties(final TroffServerClient client) {
		TroffGameProperties ret = null;
		final TroffServerGroup clientGroup = getOrJoinClientGroup(client);
		if (null != clientGroup) {
			final TroffGame game = clientGroup.getGame();
			if (null != game) {
				ret = game.getProperties();
			}
		}
		return ret;
	}

	private TroffServerGroup getOrJoinClientGroup(final TroffServerClient client) {
		TroffServerGroup clientGroup = getClientGroupByClient(client);
		if (null == clientGroup) {
			clientGroup = getAllClientGroups()[0];
			join(client, clientGroup);
		}
		return clientGroup;
	}

	private TroffServerGroup getClientGroupByClient(final TroffServerClient client) {
		for (final TroffServerGroup round : clientGroups) {
			if (round.contains(client)) {
				return round;
			}
		}
		return null;
	}

	public TroffServerGroup[] getAllClientGroups() {
		return clientGroups.toArray(GROUP_ARRAY_TEMPLATE);
	}

	public TroffServerClient[] getAllClients() {
		return clients.toArray(CLIENT_ARRAY_TEMPLATE);
	}

	String getName() {
		return name;
	}

	public void setName(final String name) {
		this.name = name;
	}

	transient Logger logger = TroffServerThread.logger;

	protected LinkedList<TroffServerGroup> clientGroups = new LinkedList<TroffServerGroup>();

	protected TreeSet<TroffServerClient> clients = new TreeSet<TroffServerClient>();

	protected String name = ServerSelectionModel.SERVERNAME_DEFAULT;
}
