package de.ped.troff.server;

import java.io.IOException;
import java.io.Serializable;
import java.util.LinkedList;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.log.Logger;
import de.ped.troff.middleware.ClientInfo;
import de.ped.troff.middleware.Connector;
import de.ped.troff.middleware.GroupInfo;
import de.ped.troff.middleware.Message;
import de.ped.troff.middleware.TroffRoundsManager;
import de.ped.troff.server.logic.TroffGame;
import de.ped.troff.server.logic.TroffGameState;

/**
 * The server-side data representation of a client. 
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffServerClient implements Comparable<TroffServerClient>, Serializable {
	private static final long serialVersionUID = 1L;

	protected transient Logger logger = TroffServerThread.logger;

	protected transient final Connector connector;

	protected LinkedList<I18NStringWithFillIns> receivedMessages = new LinkedList<I18NStringWithFillIns>();

	protected boolean isToResume = TroffRoundsManager.IS_RUNNING_DEFAULT;

	public TroffServerClient(final Connector connector) {
		this.connector = connector;
		if (null == connector) {
			throw new IllegalArgumentException("Connector is null");
		}
	}

	public ClientInfo getClientInfo() {
		return connector.getClientInfo();
	}

	public String getName() {
		return connector.getClientName();
	}

	public byte getNumberOfPlayers() {
		final ClientInfo info = getClientInfo();
		if (null == info) {
			return GroupInfo.MAX_NUMBER_OF_PLAYERS;
		} else {
			return info.numberOfPlayers;
		}
	}

	public void setToResume(final boolean isToResume) {
		this.isToResume = isToResume;
	}

	public boolean isToResume() {
		return isToResume;
	}

	protected void sendNotification(final Message message) {
		if (null != connector) {
			try {
				connector.send(message);
			} catch (final IOException ex) {
				logger.debug("Could not send new " + message.toString() + " to client " + getName(), ex);
			}
		}
	}

	public void receiveMessage(final I18NStringWithFillIns message) {
		if (null != message) {
			receivedMessages.add(message);
		}
		sendNotification(Message.createMessagesRes(getTextMessages()));
	}

	public void receiveGameState(final TroffGameState state) {
		sendNotification(Message.createGameStateRes(state));
	}

	public void receiveStartGame(final TroffGame game) {
		sendNotification(Message.createStartGameRes(game.getProperties()));
	}

	public void receivePauseOrResume(final boolean isToResume) {
		sendNotification(Message.createPauseResumeRes(isToResume));
	}

	public void receiveGameActionInput(final int playerIdx, final int suggestion) {
		sendNotification(Message.createGameActionRes(playerIdx, suggestion));
	}

	private static final I18NStringWithFillIns[] I18NSTRING_ARRAY_TEMPLATE = new I18NStringWithFillIns[]{};
	
	public I18NStringWithFillIns[] getTextMessages() {
		return receivedMessages.toArray(I18NSTRING_ARRAY_TEMPLATE);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((getName() == null) ? 0 : getName().hashCode());
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final TroffServerClient other = (TroffServerClient) obj;
		if (getName() == null) {
			if (other.getName() != null)
				return false;
		} else if (!getName().equals(other.getName()))
			return false;
		return true;
	}

	@Override
	public int compareTo(final TroffServerClient other) {
		final String name = getName();
		final String otherName = other.getName();
		if (null == name) {
			if (null == otherName) {
				return 0;
			} else {
				return -1;
			}
		}
		return name.compareTo(otherName);
	}

}