package de.ped.troff.server;

import java.io.Serializable;
import java.util.LinkedList;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.log.Logger;
import de.ped.troff.middleware.GroupInfo;
import de.ped.troff.server.logic.TroffGame;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffGameProperties;

/**
 * The data representation of a round or group. 
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffServerGroup implements Comparable<TroffServerGroup>, Serializable {
	private static final long serialVersionUID = 1L;

	public static final String DEFAULT_NAME = "New round";

	/**
	 * A synchronization wrapper around a set of {@link TroffServerClient} objects.
	 * 
	 * @author Copyright (c) 2020 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	private static class ClientSet {
		private synchronized boolean add(final TroffServerClient client) {
			boolean didSomething = false;
			if (!clients.contains(client)) {
				didSomething = clients.add(client);
			}
			return didSomething;
		}

		private synchronized boolean remove(final TroffServerClient client) {
			boolean didSomething = false;
			while (clients.remove(client)) {
				didSomething = true;
			}
			return didSomething;
		}

		private synchronized int getNumberOfPlayers() {
			int numberOfPlayers = 0;
			for (final TroffServerClient c : clients) {
				numberOfPlayers += c.getNumberOfPlayers();
			}
			return numberOfPlayers;
		}

		private synchronized boolean contains(final TroffServerClient client) {
			return clients.contains(client);
		}

		private synchronized TroffServerClient[] getClients() {
			return clients.toArray(TroffRoundsManagerImpl.CLIENT_ARRAY_TEMPLATE);
		}

		private LinkedList<TroffServerClient> clients = new LinkedList<TroffServerClient>();
	}

	public TroffServerGroup() {
		this(null, DEFAULT_NAME);
	}

	public TroffServerGroup(final TroffRoundsManagerImpl server, final String name) {
		this.name = name;
		this.parent = server;
	}

	public boolean contains(final TroffServerClient client) {
		return clients.contains(client);
	}

	public boolean isAllowedToJoin(final TroffServerClient client) {
		boolean ret = false;
		if (null != client) {
			if (isNamed() && ((null == game) || (!game.getGameState().isAlive()))) {
				final int numberOfPlayers = clients.getNumberOfPlayers();
				if (numberOfPlayers + client.getNumberOfPlayers() <= GroupInfo.MAX_NUMBER_OF_PLAYERS) {
					ret = true;
				}
			}
		}
		return ret;
	}

	public boolean join(final TroffServerClient client) {
		boolean ret = false;
		if (isAllowedToJoin(client)) {
			ret = clients.add(client);
			if (ret) {
				receiveMessage(new I18NStringWithFillIns("Game.Message.Join", "%0: %1 +%2",
						new String[] { (null == parent ? null : parent.getName()), getName(), client.getName(), }));
				logger.debug("Client " + client.getName() + " joined group " + getName());
				pauseOrResumeGame();
			}
		}
		return ret;
	}

	public void leave(final TroffServerClient client) {
		boolean didSomething = clients.remove(client);
		if (didSomething) {
			final I18NStringWithFillIns is = new I18NStringWithFillIns("Game.Message.Leave", "%0: %1 -%2",
					new String[] { (null == parent ? null : parent.getName()), getName(), client.getName(), });
			receiveMessage(is);
			client.receiveMessage(is); // The client has to get it extra, as he's not in the group any more
			logger.debug("Client " + client.getName() + " has left group " + getName());
			pauseOrResumeGame();
		}
	}

	public void startGame(final TroffGameProperties props) {
		logger.trace("startGame()");
		if (!isNamed()) {
			throw new IllegalStateException("Round is not named");
		}
		game = new TroffGame(props, this);
		game.start();
		// Create a copy of all clients
		final TroffServerClient[] myClients = getClients();
		// now safely traverse copy, as traversal takes a while and calls outside world
		for (final TroffServerClient clientToInform : myClients) {
			clientToInform.receiveStartGame(game);
		}
		pauseOrResumeGame();
	}

	/**
	 * Handles a request to pause or resume the game.
	 * @param client the requesting client
	 * @param isToResume <code>true</code> if to resume the game, <code>false</code> if to pause the game
	 * @return <code>true</code> if the game state is running after handling the request, <code>false</code> if the game state is paused after handling the request
	 */
	public boolean pauseOrResumeGame(final TroffServerClient client, final boolean isToResume) {
		client.setToResume(isToResume);
		return pauseOrResumeGame();
	}

	/**
	 * Checks the pause or resume requests of the clients of this group and pauses or resumes the game accordingly.
	 * Should be called if a pause/resume request came in, or the members of the group change.
	 * @return <code>true</code> if the game state is running after checking, <code>false</code> if the game state is paused after checking.
	 */
	protected boolean pauseOrResumeGame() {
		// Create a copy of all clients
		final TroffServerClient[] myClients = getClients();
		// now safely traverse copy, as traversal takes a while and calls outside world
		boolean isToResume = true;
		for (final TroffServerClient clientToInform : myClients) {
			isToResume = isToResume && clientToInform.isToResume();
		}
		logger.trace("pauseOrResumeGame() to " + isToResume);
		if (null != game) {
			logger.trace("pauseOrResumeGame() found game, trying to pause or resume it to " + isToResume);
			game.pauseOrResume(isToResume);
			logger.trace("pauseOrResumeGame() paused or resumed game.");
		} else {
			logger.trace("pauseOrResumeGame() found no game.");
		}
		for (final TroffServerClient clientToInform : myClients) {
			logger.trace("pauseOrResumeGame() informs client.");
			clientToInform.receivePauseOrResume(isToResume);
		}
		return isToResume;
	}

	public void suggestUserGameAction(final int playerIdx, final int suggestion) {
		final TroffGame game = getGame();
		if (null != game) {
			game.suggestUserGameAction(playerIdx, suggestion);
		}
		// Create a copy of all clients
		final TroffServerClient[] myClients = getClients();
		// now safely traverse copy, as traversal takes a while and calls outside world
		for (final TroffServerClient clientToInform : myClients) {
			clientToInform.receiveGameActionInput(playerIdx, suggestion);
		}
	}

	public TroffGameProperties getGameProperties() {
		return (null == game ? null : game.getProperties());
	}

	public void receiveMessage(final I18NStringWithFillIns message) {
		// Create a copy of all clients
		final TroffServerClient[] myClients = getClients();
		// now safely traverse copy, as traversal takes a while and calls outside world
		for (final TroffServerClient clientToInform : myClients) {
			clientToInform.receiveMessage(message);
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final TroffServerGroup other = (TroffServerGroup) obj;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		return true;
	}

	@Override
	public int compareTo(final TroffServerGroup other) {
		return name.compareTo(other.name);
	}

	public void update(final TroffGame who, final TroffGameState what) {
		logger.debug("TroffServerGroup.update " + who + " - " + what);
		if (who == game) {
			// Create a copy of all clients
			final TroffServerClient[] myClients = getClients();
			// now safely traverse copy, as traversal takes a while and calls outside world
			for (final TroffServerClient client : myClients) {
				client.receiveGameState(game.getGameState());
			}
		}
		logger.debug("TroffServerGroup.update " + who + " - " + what + " finished.");
	}

	public TroffServerClient[] getClients() {
		return clients.getClients();
	}

	public String getName() {
		return name;
	}

	public void setName(final String name) {
		this.name = name;
	}

	public boolean isNamed() {
		return (!DEFAULT_NAME.equals(name));
	}

	public TroffGame getGame() {
		return game;
	}

	transient protected Logger logger = TroffServerThread.logger;

	protected String name;

	protected ClientSet clients = new ClientSet();

	protected transient TroffGame game;

	private TroffRoundsManagerImpl parent;
} 