package de.ped.troff.server;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashSet;
import java.util.Iterator;

import de.ped.tools.FinishableThread;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.troff.middleware.Environment;
import de.ped.troff.middleware.ServerSelectionModel;

/**
 * The Troff server, listening for and accepting client connections.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffServerThread extends FinishableThread {

	public static final Logger logger = new Logger("Server");

	static {
		logger.setToShowThread(true);
		logger.setLogLevel(Level.INFO);
	}

	public static void main(final String args[]) {
		instance();
	}

	private static TroffServerThread instance;

	public static TroffServerThread instance() {
		if (null == instance) {
			TroffServerThread server = null;
			try {
				final Environment environment = new Environment();
				final ServerSelectionModel model = new ServerSelectionModel(); 
				model.setServer(true);
				model.setServerPortNo(8082);
				model.setServerName(ServerSelectionModel.SERVERNAME_DEFAULT);
				server = new TroffServerThread(environment, model);
				server.start();
				instance = server;
			} catch (final IOException e) {
				e.printStackTrace();
			}
		}
		return instance;
	}

	public TroffServerThread(
			final Environment environment, 
			final ServerSelectionModel model) throws IOException {
		super("TroffServer", FinishableThread.NO_SLEEP);
		setLogger(logger);
		this.environment = environment;
		manager = new TroffRoundsManagerImpl();
		manager.setName(model.getServerNameForNaming());
		if (model.isServer()) {
			serverSocket = new ServerSocket(model.getServerPortNo());
		} else {
			serverSocket = null;
			setTickLength(100L);
		}
	}

	public Environment getEnvironment() {
		return environment;
	}

	protected synchronized void startConnection(final Socket socket) {
		final TroffServerConnectionThread connection = new TroffServerConnectionThread(this, manager, socket);
		add(connection);
		connection.start();
	}

	protected synchronized void add(final TroffServerConnectionThread connection) {
		threads.add(connection);
	}

	protected synchronized void remove(final TroffServerConnectionThread connectionThread) {
		connectionThread.terminate();
		threads.remove(connectionThread);
	}

	protected synchronized void removeAll() {
		while (!threads.isEmpty()) {
			final Iterator<TroffServerConnectionThread> it = threads.iterator();
			remove(it.next());
			try {
				wait(1L);
			} catch (final InterruptedException x) {
				// Nothing to do
			}
		}
	}

	@Override
	public void onRunStart() {
		logger.debug("Started.");
		startConnection(null);
	}

	@Override
	public void threadStep() {
		Socket socket = null;
		if (null != serverSocket) {
			try {
				socket = serverSocket.accept();
			} catch (final IOException ex) {
				if (!isToTerminate()) {
					logger.debug("Catched exception", ex);
				}
			}
			startConnection(socket);
		}
	}

	@Override
	public void onTerminate() {
		if (null != serverSocket) {
			try {
				serverSocket.close();
			} catch (final IOException ex) {
				// Nothing to do
			}
		}
	}

	@Override
	public void onRunEnd() {
		if (null != serverSocket) {
			try {
				serverSocket.close();
			} catch (final IOException ex) {
				logger.debug("Catched exception", ex);
			}
		}
		removeAll();
		logger.debug("Stopped.");
	}

	protected final TroffRoundsManagerImpl manager;
	protected final ServerSocket serverSocket;
	protected Environment environment;
	HashSet<TroffServerConnectionThread> threads = new HashSet<TroffServerConnectionThread>();
}
