package de.ped.troff.server.logic;

import java.io.Serializable;
import java.util.Arrays;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;

/**
 * All you need to know about the current board state to draw everything.
 * 
 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class BoardInfo implements Cloneable, Serializable, Marshallable {

	/**
	 * All you need to know about a field on the board to draw it.
	 * 
	 * @author Copyright (c) 2011 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class FieldInfo {
		int x;
		int y;
		byte field;
		boolean isHead;
		int headDirection;
		public int getX() {
			return x;
		}
		public int getY() {
			return y;
		}
		public byte getField() {
			return field;
		}
		public boolean isHead() {
			return isHead;
		}
		public int getHeadDirection() {
			return headDirection;
		}
	}

	private static final long serialVersionUID = 1L;

	public static final byte SOLID_WALL = 2;

	public static final byte REMOVABLE_WALL = 1;

	public static final byte FREE = 0;

	public static final byte PLAYER = 8;

	public static boolean isFree(final byte fieldValue) {
		return FREE == fieldValue;
	}

	public static boolean isPlayer(final byte fieldValue) {
		return fieldValue >= PLAYER;
	}

	public static byte fieldValue2playerIdx(final byte fieldValue) {
		return (byte) (fieldValue - PLAYER);
	}

	public static byte playerIdx2fieldValue(final byte playerIdx) {
		return (byte) (PLAYER + playerIdx);
	}

	public static boolean isPlayer(final byte fieldValue, final int playerNumber) {
		return playerNumber == fieldValue2playerIdx(fieldValue);
	}

	String key;
	PlayfieldDimension size;
	private byte[] board;
	PlayerPosInfo[] playerPosInfos;

	public BoardInfo() {
		size = new PlayfieldDimension(0, 0);
	}

	/** 
	 * Creates an empty board and sets all the players.
	 * @param key the message key for the board template, as shown e.g. in game template selection
	 * @param xsize the x dimension of the board
	 * @param ysize the y dimension of the board
	 * @param playerPosInfos the start positions of the players
	 * @see #init(String, int, int, PlayerPosInfo[])
	 */
	public BoardInfo(final String key, final int xsize, final int ysize, final PlayerPosInfo[] playerPosInfos) {
		init(key, xsize, ysize, playerPosInfos);
	}

	/**
	 * Creates a clone of another board, but can clip the number of players.
	 * @param other the board to clone from, typically a template
	 * @param numberOfPlayers the number of players who want to play that game
	 * @see #init(BoardInfo, int)
	 */
	public BoardInfo(final BoardInfo other, final int numberOfPlayers) {
		init(other, numberOfPlayers);
	}

	/** 
	 * Creates an empty board and sets all the players.
	 * @param key the message key for the board template, as shown e.g. in game template selection
	 * @param xsize the x dimension of the board
	 * @param ysize the y dimension of the board
	 * @param playerPosInfos the start positions of the players
	 */
	public void init(final String key, final int xsize, final int ysize, final PlayerPosInfo[] playerPosInfos) {
		this.key = key;
		this.size = new PlayfieldDimension(xsize, ysize);
		this.board = new byte[size.getWidth() * ((size.getHeight() + 1) / 2)];
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = 0; y < size.getHeight(); y++) {
				setAt(x, y, BoardInfo.FREE);
			}
		}
		this.playerPosInfos = new PlayerPosInfo[playerPosInfos.length];
		clonePlayerPosInfosFrom(playerPosInfos);
	}

	/**
	 * Creates a clone of another board, but can clip the number of players.
	 * @param other the board to clone from, typically a template
	 * @param numberOfPlayers the number of players who want to play that game
	 */
	public void init(final BoardInfo other, final int numberOfPlayers) {
		this.key = other.key;
		this.size = other.size.clone();
		this.board = new byte[size.getWidth() * ((size.getHeight() + 1) / 2)];
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = 0; y < size.getHeight(); y++) {
				setAt(x, y, other.getAt(x, y));
			}
		}
		final int nop = Math.min(numberOfPlayers, other.playerPosInfos.length);
		this.playerPosInfos = new PlayerPosInfo[nop];
		clonePlayerPosInfosFrom(other.playerPosInfos);
	}

	protected void clonePlayerPosInfosFrom(final PlayerPosInfo[] source) {
		if (source.length < playerPosInfos.length) {
			throw new IllegalArgumentException("Array size of PlayerPosInfo differ: mine was " + playerPosInfos.length + " but other was " + source.length);
		}
		for (int i = 0; i < playerPosInfos.length; i++) {
			if (null == playerPosInfos[i]) {
				playerPosInfos[i] = new PlayerPosInfo(source[i]);
			} else {
				playerPosInfos[i].copyFrom(source[i]);
			}
		}
	}

	/**
	 * @return the message key for the board template, as shown e.g. in game template selection
	 */
	public String getKey() {
		return key;
	}

	public PlayfieldDimension getSize() {
		return size;
	}

	public static byte getNibble(final byte fullByte, final boolean isHighNibble) {
		if (isHighNibble) {
			return (byte) ((fullByte >>> 4) & 0x0f);
		} else {
			return (byte) (fullByte & 0x0f);
		}
	}

	public static byte setNibble(final byte fullByte, final byte nibble, final boolean isHighNibble) {
		byte result;
		if (isHighNibble) {
			result = (byte)((fullByte & 0x0f) | ((nibble & 0x0f) << 4));
		} else {
			result = (byte)((fullByte & 0xf0) | (nibble & 0x0f));
		}
		return result;
	}

	protected int getByteIndex(final int x, final int y) {
		return x * ((size.getHeight() + 1) / 2) + (y >>> 1);
	}

	public byte getAt(final int x, final int y) {
		final int idx = getByteIndex(x, y);
		byte val = board[idx];
		return getNibble(val, 0 == (y & 0x1));
	}

	public void setAt(final int x, final int y, final byte value) {
		final int idx = getByteIndex(x, y);
		byte val = board[idx];
		board[idx] = setNibble(val, value, 0 == (y & 0x1));
	}

	public PlayerPosInfo[] getPlayerPosInfos() {
		return playerPosInfos;
	}

	public int getPlayerPosInfosSize() {
		return (playerPosInfos == null ? 0 : playerPosInfos.length);
	}

	public PlayerPosInfo getPlayerPosInfo(final int playerIdx) {
		return (playerPosInfos == null ? null : playerPosInfos[playerIdx]);
	}

	public void setPlayerPosInfo(final int playerIdx, final PlayerPosInfo ppinfo) {
		PlayerPosInfo info = playerPosInfos[playerIdx];
		if (null == info) {
			info = ppinfo;
		} else {
			info.copyFrom(ppinfo);
		}
	}

	public void setPlayerPosInfo(final int playerIdx, final int xpos, final int ypos, final int direction) {
		final PlayerPosInfo info = playerPosInfos[playerIdx];
		if (null == info) {
			playerPosInfos[playerIdx] = new PlayerPosInfo(xpos, ypos, direction);
		} else {
			info.setAll(xpos, ypos, direction);
		}
	}

	@Override
	public BoardInfo clone() {
		return new BoardInfo(this, Integer.MAX_VALUE);
	}

	/**
	 * Implicitly used in combo box to show selection of possible boards to play
	 * 
	 * @return the message key for the board template, as shown e.g. in game template selection
	 */
	@Override
	public String toString() {
		return key;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Arrays.hashCode(board);
		result = prime * result + ((key == null) ? 0 : key.hashCode());
		result = prime * result + Arrays.hashCode(playerPosInfos);
		result = prime * result + size.hashCode();
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final BoardInfo other = (BoardInfo) obj;
		if (!Arrays.equals(board, other.board))
			return false;
		if (key == null) {
			if (other.key != null)
				return false;
		} else if (!key.equals(other.key))
			return false;
		if (!Arrays.deepEquals(playerPosInfos, other.playerPosInfos))
			return false;
		if (!size.equals(other.size))
			return false;
		return true;
	}

	public void calcFieldInfoAt(final int x, final int y, final BoardInfo.FieldInfo fieldInfo) {
		if (null != fieldInfo) {
			final byte field = getAt(x, y);
			fieldInfo.x      = x;
			fieldInfo.y      = y;
			fieldInfo.field  = field;
			fieldInfo.isHead = false;
			fieldInfo.headDirection = -1;
			if (field >= 0) {
		        for (final PlayerPosInfo playerPosInfo : getPlayerPosInfos()) {
					if (playerPosInfo.isAlive() && (playerPosInfo.getXPos() == x) && (playerPosInfo.getYPos() == y)) {
						fieldInfo.isHead = true;
						fieldInfo.headDirection = playerPosInfo.getDirection();
	//						logger.debug("Found at " + forward + " forward and " + right + " right (" + x + ", " + y + ") the code " + ret);
					}
		        }
			}
		}
	}

	public BoardInfo.FieldInfo calcFieldInfoAt(final int x, final int y) {
		final FieldInfo fieldInfo = new FieldInfo();
		calcFieldInfoAt(x, y, fieldInfo);
		return fieldInfo;
	}

	public byte getAtForwardPosition(final int playerIdx, final int forward, final int right, final FieldInfo fieldInfo) {
		final PlayerPosInfo playerPosInfo = getPlayerPosInfo(playerIdx);
		int x = playerPosInfo.getXPos();
		int y = playerPosInfo.getYPos();
		final int pdir = playerPosInfo.getDirection();
		int dir = pdir;
		x += forward * TroffGame.DIRECTIONS[dir][0];
		y += forward * TroffGame.DIRECTIONS[dir][1];
		dir = PlayerPosInfo.calculateTurnRight(dir, 1);
		x += right * TroffGame.DIRECTIONS[dir][0];
		y += right * TroffGame.DIRECTIONS[dir][1];
		x = MathUtil.modulo(x, size.getWidth());
		y = MathUtil.modulo(y, size.getHeight());
		final byte ret = getAt(x, y);
		calcFieldInfoAt(x, y, fieldInfo);
		if (isFree(ret)) {
	//			logger.debug("Found at " + forward + " forward and " + right + " right (" + x + ", " + y + ") the code " + ret);
		}
		return ret;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeMarshallable(size);
		marshaller.writeVariableLengthByteArray(board);
		marshaller.writeInt(playerPosInfos.length);
		for (final PlayerPosInfo info : playerPosInfos) {
			marshaller.writeMarshallable(info);
		}
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		size.demarshal(marshaller);
		board = marshaller.readVariableLengthBytesArray();
		final int playerPosSize = marshaller.readInt();
		playerPosInfos = new PlayerPosInfo[playerPosSize];
		for (int i = 0; i < playerPosInfos.length; i++) {
			playerPosInfos[i] = new PlayerPosInfo(marshaller);
		}
		return this;
	}
}