package de.ped.troff.server.logic;

import java.io.Serializable;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;

/**
 * Bundles all information about the player's state on the playfield, 
 * including position, direction and the tiny bit of information if alive or not.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class PlayerPosInfo implements Cloneable, Serializable, Marshallable {
	private static final long serialVersionUID = 1L;

	public static int normDirection(final int direction) {
		return MathUtil.modulo(direction, TroffGame.DIRECTIONS.length);
	}

	/**
	 * Calculates the new direction, when you turn starting from the given direction.
	 * @param direction starting direction
	 * @param howmuch one of {@link TroffPlayer#TURN_LEFT}, {@link TroffPlayer#TURN_STRAIGHT}, {@link TroffPlayer#TURN_RIGHT}
	 * @return one of {@link TroffGame#NORTH}, {@link TroffGame#EAST}, {@link TroffGame#SOUTH}, {@link TroffGame#WEST} 
	 */
	public static int calculateTurnRight(final int direction, final int howmuch) {
		return normDirection(direction + howmuch);
	}

	public PlayerPosInfo(final int xpos, final int ypos, final int direction) {
		setAll(xpos, ypos, direction);
	}

	public PlayerPosInfo(final PlayerPosInfo other) {
		copyFrom(other);
	}
	
	public PlayerPosInfo(final Marshaller marshaller) {
		marshaller.readMarshallable(this);
	}

	@Override
	public PlayerPosInfo clone() {
		return new PlayerPosInfo(this);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + direction;
//		result = prime * result + (isAlive ? 1231 : 1237);
		result = prime * result + xpos;
		result = prime * result + ypos;
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		final PlayerPosInfo other = (PlayerPosInfo) obj;
		if (direction != other.direction)
			return false;
//		if (isAlive != other.isAlive)
//			return false;
		if (xpos != other.xpos)
			return false;
		if (ypos != other.ypos)
			return false;
		return true;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(xpos);
		marshaller.writeInt(ypos);
		marshaller.writeInt(direction);
		marshaller.writeBoolean(isAlive);
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		xpos = marshaller.readInt();
		ypos = marshaller.readInt();
		direction = marshaller.readInt();
		isAlive = marshaller.readBoolean();
		return this;
	}

	public void setAll(final int xpos, final int ypos, final int direction) {
		this.xpos = xpos;
		this.ypos = ypos;
		this.direction = normDirection(direction);
	}

	public void copyFrom(final PlayerPosInfo other) {
		this.xpos = other.xpos;
		this.ypos = other.ypos;
		this.direction = other.direction;
	}

	public int getXPos() {
		return xpos;
	}

	public void setXPos(final int xpos) {
		this.xpos = xpos;
	}

	public int getYPos() {
		return ypos;
	}

	public void setYPos(final int ypos) {
		this.ypos = ypos;
	}

	public void setXYPos(final int xpos, final int ypos) {
		this.xpos = xpos;
		this.ypos = ypos;
	}

	public int getDirection() {
		return direction;
	}

	public void setDirection(final int direction) {
		this.direction = normDirection(direction);
	}

	public boolean isAlive() {
		return isAlive;
	}

	/**
	 * Sets a player to dead or alive.
	 * @param isAlive <code>true</code> for alive, <code>false</code> for dead
	 */
	public void setAlive(final boolean isAlive) {
		this.isAlive = isAlive;
	}

	private int xpos;
	private int ypos;
	private int direction;
	private boolean isAlive = true;
}