package de.ped.troff.server.logic;

import java.util.ArrayList;

import de.ped.tools.log.Logger;
import de.ped.troff.middleware.TroffRoundsManager;
import de.ped.troff.server.TroffServerGroup;

/**
 * Represents all the rules of the game, knows the state of the game,
 * and contains all other logical elements of the game, like board or players.
 * Is central starting point for any call into the logical layer from higher layers.
 * 
 * @author Copyright (c) 2001 Peter Diefenbach
 */
public class TroffGame {
	private Logger logger = Logger.getLogger(this.toString());

	public static final int NORTH = 0;
	public static final int EAST = 1;
	public static final int SOUTH = 2;
	public static final int WEST = 3;
	public static final int[][] DIRECTIONS = new int[][] {
		{ 0,-1}, { 1, 0}, {0, 1}, {-1, 0}
	};

	public TroffGame() {
		// Nothing to do
		logger.info("Game created: " + toString());
	}

	public TroffGame(final TroffGameProperties props, final TroffServerGroup serverGroup) {
		this.serverGroup = serverGroup;
		setProperties(props);
		logger.info("Game created: " + toString());
	}

	public TroffGameProperties getProperties() {
		return properties;
	}

	/**
	 * Initializes the game. If the object was created by default constructor, this is the fist call you should do.
	 * @param props the game properties to create the game with
	 */
	protected void setProperties(final TroffGameProperties props) {
		properties = props;
		gameState = new TroffGameState(properties.getNumberOfPlayers());
		gameState.setBoardInfo(BoardFactory.createBoard(properties.getBoardIndex(), properties.getNumberOfPlayers()));
		players = new TroffPlayer[properties.getNumberOfPlayers()];
		final BoardInfo boardInfo = gameState.getBoardInfo();
		for (byte i = 0; i < players.length; i++) {
			final TroffPlayer player;
			switch(properties.getPlayerType(i)) {
			case TroffGameProperties.PLAYERTYPE_HUMAN:
				player = new HumanPlayer(this, i, boardInfo.playerPosInfos[i], properties.getPlayerMode(i));
				break;
			default:
				player = new ComputerPlayerRandom(this, i, boardInfo.playerPosInfos[i], properties.getPlayerMode(i));
			}
			players[i] = player;
		}
		for (final TroffPlayer player : players) {
			setAt(player.getX(), player.getY(), BoardInfo.playerIdx2fieldValue(player.getIndex()));
		}
	}

	public BoardInfo getBoardInfo() {
		return gameState.getBoardInfo();
	}

	public void setAt(final int x, final int y, final byte value) {
		gameState.getBoardInfo().setAt(x, y, value);
	}

	public TroffPlayer getPlayer(final int i) {
		return players[i];
	}

	public TroffPlayer[] getPlayers() {
		return players;
	}

	public boolean isLegalMove(final TroffPlayer player, final int nx, final int ny) {
		boolean ret = true;
		if (!BoardInfo.isFree(gameState.getBoardInfo().getAt(nx, ny))) {
			ret = false;
		}
		return ret;
	}

	/**
	 * Executes a single step, so moves all players and checks for events to present.
	 * @return list of events to show to the players, mostly by playing a sound
	 */
	private ArrayList<Event> step() {
		ArrayList<Event> ret = null;
		if (isAlive()) {
			ret = new ArrayList<Event>();
			gameState.incStep();
			ret.add(new Event(Event.STEP));
			for (int currentPlayerIdx = 0; currentPlayerIdx < players.length; currentPlayerIdx++) {
				final TroffPlayer player = players[currentPlayerIdx];
				final int mode = properties.getPlayerMode(currentPlayerIdx);
				final int mask = TroffGameProperties.MODE_INFOS[mode].getModeMask();
				if (0 != (mask & TroffGameProperties.ModeInfo.PLAYERMODEMASK_HUNTER)) {
					if (player.isShooting()) {
						shoot(currentPlayerIdx, ret);
					}
				}
				if ((0 != (mask & TroffGameProperties.ModeInfo.PLAYERMODEMASK_TRACE)) || 
					((0 != (mask & TroffGameProperties.ModeInfo.PLAYERMODEMASK_PREY)) && (player.isTracing()))) {
					// Nothing to do, tracing happens from alone
				} else {
					setAt(player.getX(), player.getY(), BoardInfo.FREE);
				}
				final boolean hasPlayerJustDied = player.move();
				if (hasPlayerJustDied) {
					ret.add(new Event(Event.DEATH, player.getX(), player.getY()));
				}
				setAt(player.getX(), player.getY(), BoardInfo.playerIdx2fieldValue(player.getIndex()));
				if (!player.getPlayerPosInfo().equals(gameState.getBoardInfo().playerPosInfos[currentPlayerIdx])) {
					throw new IllegalStateException("Player position redundancy fail");
				}
			}
		}
		return ret;
	}

	public void execStep() {
		if (isIntroRunning) {
			// Nothing to do
		} else {
			final ArrayList<Event> ret = step();
			gameState.setEvents(ret);
			for (int playerIdx = 0; playerIdx < players.length; playerIdx++) {
				gameState.setTracing(playerIdx, getPlayer(playerIdx).isTracing());
			}
			gameState.setAlive(isAlive());
		}
		serverGroup.update(this, gameState);
	}

	private void shoot(final int currentPlayerIdx, final ArrayList<Event> ret) {
		if (0 == gameState.nextShotIn(currentPlayerIdx)) {
			ret.add(new Event(Event.SHOOT_SUCCEEDED, players[currentPlayerIdx].getX(), players[currentPlayerIdx].getY()));
			gameState.setLastShot(currentPlayerIdx);
			boolean found = false;
			final BoardInfo.FieldInfo fieldInfo = new BoardInfo.FieldInfo();
			for (int dist = 1; !found && (dist <= TroffGameProperties.SHOOTING_DISTANCE); dist++) {
				final byte field = gameState.getBoardInfo().getAtForwardPosition(currentPlayerIdx, dist, 0, fieldInfo);
				if (!BoardInfo.isFree(field)) {
					found = true;
					final byte player = BoardInfo.fieldValue2playerIdx(field);
					if ((fieldInfo.isHead) && (player == currentPlayerIdx) && (players[player].isAlive())) {
						// Shot another player's head
						players[player].setAlive(false);
						// If a player's head should be like a solid wall....
//						ret.add(new Event(Event.HIT_SOLID_WALL, fieldInfo.getX(), fieldInfo.getY()));
						// If a player's head should be like a removable wall....
						setAt(fieldInfo.getX(), fieldInfo.getY(), BoardInfo.FREE);
						ret.add(new Event(Event.DEATH, fieldInfo.getX(), fieldInfo.getY()));
					} else if (BoardInfo.SOLID_WALL != field) {
						setAt(fieldInfo.getX(), fieldInfo.getY(), BoardInfo.FREE);
						ret.add(new Event(Event.HIT_WALL, fieldInfo.getX(), fieldInfo.getY()));					} else {
						ret.add(new Event(Event.HIT_SOLID_WALL, fieldInfo.getX(), fieldInfo.getY()));
					}
				}
			}
			if (!found) {
				ret.add(new Event(Event.NO_HIT, fieldInfo.getX(), fieldInfo.getY()));
			}
		} else {
			ret.add(new Event(Event.SHOOT_FAILED, players[currentPlayerIdx].getX(), players[currentPlayerIdx].getY()));
		}
	}

	/**
	 * @return if the game is still running
	 */
	private boolean isAlive() {
		boolean ret = true;
		if (null == players) {
			ret = false;
		} else if (1 == players.length) {
			ret = (players[0].isAlive());
		} else {
			int playersAlive = 0;
			for (final TroffPlayer player : players) {
				if (player.isAlive()) {
					playersAlive++;
				}
			}
			ret = (playersAlive > 1);
		}
		return ret;
	}

	/**
	 * The system got any input, e.g., from the keyboard, how a player should act
	 * for the next move.
	 * 
	 * @param playerIdx
	 *            the index of the player to suggest a turn for
	 * @param suggestion
	 *            one of {@link TroffPlayer#TURN_LEFT}, {@link TroffPlayer#TURN_STRAIGHT}, {@link TroffPlayer#TURN_RIGHT}
	 */
	public void suggestUserGameAction(final int playerIdx, final int suggestion) {
		if ((null != players) && (players.length > playerIdx)) {
			players[playerIdx].actionGameInput(suggestion);
		}
	}

	public long getSleepDelay() {
		if (isIntroRunning) {
			isIntroRunning = false;
			return TroffRoundsManager.INTRO_DELAY;
		} else {
			return TroffGameProperties.SPEED_INFOS[properties.getSpeedIndex()].getDelay();
		}
	}

	public TroffGameState getGameState() {
		return gameState;
	}

	public boolean isRunning() {
		return isRunning;
	}

	public boolean isIntroRunning() {
		return isIntroRunning;
	}

	public void start() {
		isIntroRunning = true;
		pauseOrResume(true);
	}

	public synchronized void pauseOrResume(final boolean isToResume) {
		if (isToResume) {
			logger.info("pauseOrResume(" + isToResume + ") - trying to resume");
			resumeRun();
			logger.info("pauseOrResume(" + isToResume + ") - finished resume");
		} else {
			logger.info("pauseOrResume(" + isToResume + ") - trying to pause");
			pauseRun();
			logger.info("pauseOrResume(" + isToResume + ") - finished pause");
		}
	}

	private void resumeRun() {
		if (!isRunning) {
			boolean wasCreated = false;
			if (null == gameThread) {
				gameThread = new GameThread(this);
				wasCreated = true;
			}
			try {
				if (!wasCreated) {
					logger.error("Starting a thread I have not created now??");
				}
				gameThread.start();
				isRunning = true;
			} catch (final IllegalThreadStateException ex) {
				// Game has ended, but I have not known this so far.
				gameThread = null;
			}
		}
	}

	private void pauseRun() {
		isRunning = false;
		if (null != gameThread) {
			logger.info("pauseRun() - trying to interrupt");
			gameThread.interrupt();
			try {
				logger.info("pauseRun() - trying to join");
				gameThread.join();
				logger.info("pauseRun() - joined");
			} catch (final InterruptedException e) {
				logger.error(e.getMessage(), e);
			}
			gameThread = null;
		} else {
			logger.info("pauseRun() - no game thread, so we are fine");
		}
	}

	private TroffGameProperties properties;
	private TroffServerGroup serverGroup;

	private TroffGameState gameState;

	private TroffPlayer[] players;

	private boolean isRunning;
	private boolean isIntroRunning = true;
	private GameThread gameThread;
}
