package de.ped.troff.server.logic;

import java.io.Serializable;

import de.ped.tools.Assert;
import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;

/** 
 * Contains all properties that remain constant for a run. 
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffGameProperties implements Cloneable, Serializable, Marshallable {
	private static final long serialVersionUID = 1L;

	public static final int MIN_NUMBER_OF_PLAYERS = 1;
	public static final int MAX_NUMBER_OF_PLAYERS = 8;
	public static final int PLAYERTYPE_HUMAN = 0;
	public static final int PLAYERTYPE_RANDOM = 1;

	public static final int SHOOTING_FREQUENCY = 10;
	public static final int SHOOTING_DISTANCE  =  5;

	/**
	 * Representations for the different speeds the game can run in.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class SpeedInfo {
		private int delay;
		private String key;
		public int getDelay() {
			return delay;
		}
		public String getKey() {
			return key;
		}
		public SpeedInfo(final int delay, final String key) {
			this.delay = delay;
			this.key = key;
		}
		@Override
		public String toString() {
			return key;
		}
	}

	/**
	 * Representations for the different views a player can have onto the board.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ViewInfo {
		public static int VIEW_3D             = 0x01;
		public static int VIEW_RADAR_ABSOLUTE = 0x02;
		public static int VIEW_RADAR_RELATIVE = 0x04;
		public static int VIEW_3D_MODEL       = 0x08;
		private int viewMask;
		private String key;
		public int getViewMask() {
			return viewMask;
		}
		public String getKey() {
			return key;
		}
		public ViewInfo(final int viewMask, final String key) {
			this.viewMask = viewMask;
			this.key = key;
		}
		@Override
		public String toString() {
			return key;
		}
	}

	/**
	 * Representations for the different modes/roles a player can have in the game.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ModeInfo {
		/** A prey action is to leave a trace. */
		public static final int PLAYERMODEMASK_PREY   = 0x01;
		/** A tracing player always leaves a trace, action button can be used otherwise. */
		public static final int PLAYERMODEMASK_TRACE  = 0x02;
		/** A hunter's action is to shoot a bullet the facing direction. A non-solid wall ahead is shot. */
		public static final int PLAYERMODEMASK_HUNTER = 0x04;

		public static final int PLAYERMODE_TRACE  = 0;
		public static final int PLAYERMODE_PREY   = 1;
		public static final int PLAYERMODE_HUNTER = 2;
		public static final int PLAYERMODE_TRACING_HUNTER = 3;

		private int modeMask;
		private String key;

		public int getModeMask() {
			return modeMask;
		}

		public String getKey() {
			return key;
		}

		public ModeInfo(final int modeMask, final String key) {
			this.modeMask = modeMask;
			this.key = key;
		}

		@Override
		public String toString() {
			return key;
		}
	}

	public static final SpeedInfo[] SPEED_INFOS = new SpeedInfo[] {
		new SpeedInfo(1000, "GameProperties.Speed.Slowest"),
		new SpeedInfo( 750, "GameProperties.Speed.VerySlow"),
		new SpeedInfo( 500, "GameProperties.Speed.Slow"),
		new SpeedInfo( 300, "GameProperties.Speed.Normal"),
		new SpeedInfo( 200, "GameProperties.Speed.Fast"),
		new SpeedInfo( 100, "GameProperties.Speed.VeryFast"),
		new SpeedInfo(  50, "GameProperties.Speed.Fastest"),
	};

	public static final ViewInfo[] VIEW_INFOS = new ViewInfo[] {
		new ViewInfo(ViewInfo.VIEW_3D, "GameProperties.PlayerView.3D"),
		new ViewInfo(ViewInfo.VIEW_RADAR_ABSOLUTE, "GameProperties.PlayerView.Abs"),
		new ViewInfo(ViewInfo.VIEW_RADAR_RELATIVE, "GameProperties.PlayerView.Rel"),
		new ViewInfo(ViewInfo.VIEW_3D | ViewInfo.VIEW_RADAR_ABSOLUTE, "GameProperties.PlayerView.3DAbs"),
		new ViewInfo(ViewInfo.VIEW_3D | ViewInfo.VIEW_RADAR_RELATIVE, "GameProperties.PlayerView.3DRel"),
	};

	public static final ModeInfo[] MODE_INFOS = new ModeInfo[] {
		new ModeInfo(ModeInfo.PLAYERMODEMASK_TRACE,  "GameProperties.PlayerMode.Trace"),
		new ModeInfo(ModeInfo.PLAYERMODEMASK_PREY,   "GameProperties.PlayerMode.Prey"),
		new ModeInfo(ModeInfo.PLAYERMODEMASK_HUNTER, "GameProperties.PlayerMode.Hunter"),
		new ModeInfo(ModeInfo.PLAYERMODEMASK_TRACE | ModeInfo.PLAYERMODEMASK_HUNTER, "GameProperties.PlayerMode.TracingHunter"),
	};

	public static final int getShortestDelay() {
		return SPEED_INFOS[SPEED_INFOS.length - 1].getDelay();
	}

	private int boardIndex = 4;
	private int numberOfPlayers = 8;
	private int speedIndex = 2;
	private boolean isAbsoluteView = false;
	private TroffPlayerProperties[] playerProperties = new TroffPlayerProperties[MAX_NUMBER_OF_PLAYERS];

	public BoardInfo[] getBoardInfos() {
		return BoardFactory.getBoardInfos();
	}
	public int getBoardIndex() {
		return boardIndex;
	}
	public void setBoardIndex(final int boardIndex) {
		this.boardIndex = boardIndex;
	}
	public int getNumberOfPlayers() {
		return numberOfPlayers;
	}
	public void setNumberOfPlayers(final int numberOfPlayers) {
		this.numberOfPlayers = MathUtil.bound(numberOfPlayers, MIN_NUMBER_OF_PLAYERS, MAX_NUMBER_OF_PLAYERS);
	}
	public TroffPlayerProperties[] getPlayers() {
		return playerProperties;
	}
	public int getPlayerType(final int i) {
		return playerProperties[i].getTypeId();
	}
	public void setPlayerType(final int i, final int value) {
		playerProperties[i].setTypeId(value);
	}
	public int getPlayerView(final int i) {
		return playerProperties[i].getViewId();
	}
	public void setPlayerView(final int i, final int value) {
		playerProperties[i].setViewId(MathUtil.bound(value, 0, VIEW_INFOS.length - 1));
	}
	public int getPlayerMode(final int i) {
		return playerProperties[i].getModeId();
	}
	public void setPlayerMode(final int i, final int value) {
		playerProperties[i].setModeId(MathUtil.bound(value, 0, MODE_INFOS.length - 1));
	}
	public String getPlayerName(final int i) {
		return playerProperties[i].getName();
	}
	public void setPlayerName(final int i, final String value) {
		playerProperties[i].setName(value);
	}
	public String getFullPlayerName(final int i) {
		return playerProperties[i].getFullPlayerName();
	}
	public SpeedInfo[] getSpeedInfos() {
		return SPEED_INFOS;
	}
	public int getSpeedIndex() {
		return speedIndex;
	}
	public void setSpeedIndex(final int speedIndex) {
		this.speedIndex = MathUtil.bound(speedIndex, 0, SPEED_INFOS.length - 1);
	}
	public boolean isAbsoluteView() {
		return isAbsoluteView;
	}
	public void setAbsoluteView(final boolean isAbsoluteView) {
		this.isAbsoluteView = isAbsoluteView;
	}

	public TroffGameProperties() {
		for (int i = 0; i < MAX_NUMBER_OF_PLAYERS; i++) {
			if (0 == i) {
				playerProperties[i] = new TroffPlayerProperties(PLAYERTYPE_HUMAN, ModeInfo.PLAYERMODE_HUNTER, ModeInfo.PLAYERMODE_PREY, i, "Player 0", null);
			} else {
				playerProperties[i] = new TroffPlayerProperties(PLAYERTYPE_RANDOM, i % VIEW_INFOS.length, ModeInfo.PLAYERMODE_TRACE, i, "Player " + i, null);
			}
		}
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(boardIndex);
		marshaller.writeInt(speedIndex);
		marshaller.writeBoolean(isAbsoluteView);
		marshaller.writeInt(numberOfPlayers);
		for (int i = 0; i < playerProperties.length; i++) {
			marshaller.writeMarshallable(playerProperties[i]);
		}
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		boardIndex = marshaller.readInt();
		speedIndex = marshaller.readInt();
		isAbsoluteView = marshaller.readBoolean();
		numberOfPlayers = marshaller.readInt();
		for (int i = 0; i < playerProperties.length; i++) {
			playerProperties[i] = (TroffPlayerProperties) marshaller.readMarshallable(playerProperties[i]);
		}
		return this;
	}

	@Override
	public TroffGameProperties clone() {
		TroffGameProperties ret = null;
		try {
			ret = (TroffGameProperties) super.clone();
			for (int i = 0; i < playerProperties.length; i++) {
				ret.playerProperties[i] = (TroffPlayerProperties) playerProperties[i].clone();
			}
		} catch (final CloneNotSupportedException e) {
			Assert.assertTrue(false);
		}
		return ret;
	}
}
