package de.ped.troff.server.logic;

import java.io.Serializable;
import java.util.ArrayList;

import de.ped.tools.Marshallable;
import de.ped.tools.Marshaller;
import de.ped.tools.MathUtil;

/**
 * Everything that needs to be told to the clients when a game does a step.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TroffGameState implements Cloneable, Serializable, Marshallable {
	private static final long serialVersionUID = 1L;

	public TroffGameState(final int numberOfPlayers) {
		this.step = 0;
		playerState = new TroffPlayerState[numberOfPlayers];
		for (int i = 0; i < playerState.length; i++) {
			playerState[i] = new TroffPlayerState();
		}
		isAlive = true;
	}

	public TroffGameState(final Marshaller marshaller) {
		marshaller.readMarshallable(this);
	}

	public int getStep() {
		return step;
	}

	public void incStep() {
		this.step++;
	}

	public BoardInfo getBoardInfo() {
		return boardInfo;
	}

	public void setBoardInfo(final BoardInfo boardInfo) {
		this.boardInfo = boardInfo;
	}

	public ArrayList<Event> getEvents() {
		return events;
	}

	public void setEvents(final ArrayList<Event> events) {
		this.events = events;
	}

	public void setLastShot(final int playerIdx) {
		playerState[playerIdx].setLastShot(getStep());
	}

	public int nextShotIn(final int playerIdx) {
		final int lastShot = playerState[playerIdx].getLastShot();
		return MathUtil.bound(lastShot + TroffGameProperties.SHOOTING_FREQUENCY - getStep(), 0, TroffGameProperties.SHOOTING_FREQUENCY);
	}

	public void setTracing(final int playerIdx, final boolean value) {
		playerState[playerIdx].setTracing(value);
	}

	public boolean isTracing(final int playerIdx) {
		return playerState[playerIdx].isTracing();
	}

	public void setAlive(final boolean value) {
		isAlive = value;
	}

	public boolean isAlive() {
		return isAlive;
	}

	@Override
	public void marshal(final Marshaller marshaller) {
		marshaller.writeInt(step);
		marshaller.writeBoolean(isAlive);
		marshaller.writeMarshallable(boardInfo);
		marshaller.writeInt(playerState.length);
		for (final TroffPlayerState state : playerState) {
			marshaller.writeMarshallable(state);
		}
		if (null == events) {
			marshaller.writeInt(Marshaller.ARRAY_NULL_INDICATOR);
		} else {
			marshaller.writeInt(events.size());
			for (Event event : events) {
				marshaller.writeMarshallable(event);
			}
		}
	}

	@Override
	public Marshallable demarshal(final Marshaller marshaller) {
		step = marshaller.readInt();
		isAlive = marshaller.readBoolean();
		boardInfo = (BoardInfo) marshaller.readMarshallable(new BoardInfo());
		final int playersSize = marshaller.readInt();
		playerState = new TroffPlayerState[playersSize];
		for (int i = 0; i < playersSize; i++) {
			playerState[i] = new TroffPlayerState(marshaller);
		}
		final int eventsSize = marshaller.readInt();
		if (Marshaller.ARRAY_NULL_INDICATOR == eventsSize) {
			events = null;
		} else {
			events = new ArrayList<Event>(eventsSize);
			for (int i = 0; i < eventsSize; i++) {
				Event event = new Event(marshaller);
				events.add(event);
			}
		}
		return this;
	}

	protected int step;
	protected boolean isAlive;
	protected BoardInfo boardInfo;
	protected TroffPlayerState[] playerState;
	protected ArrayList<Event> events;
}
