package de.ped.troff.server.logic;

import de.ped.tools.MathUtil;
import de.ped.tools.log.Logger;

/**
 * A player - may be human, computer or hybrid.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
public abstract class TroffPlayer {
	public static final int ACTION     =  0;
	/** The indication to change direction and turn left with the head. */
	public static final int TURN_LEFT     = -1; 
	/** The indication to just keep the direction. */
	public static final int TURN_STRAIGHT =  0;
	/** The indication to change direction and turn right with the head. */
	public static final int TURN_RIGHT    =  1;

	private static Logger logger = Logger.getRootLogger();

	/**
	 * Given the current x value, and the current direction, calculates the next x value.
	 * @param x The current x value of the position.
	 * @param direction The current direction.
	 * @return The next x value of the position.
	 */
	public int calculateMoveX(final int x, final int direction) {
		return MathUtil.modulo(x + TroffGame.DIRECTIONS[direction][0], game.getBoardInfo().getSize().getWidth());
	}

	/**
	 * Given the current y value, and the current direction, calculates the next y value.
	 * @param y The current y value of the position.
	 * @param direction The current direction.
	 * @return The next y value of the position.
	 */
	public int calculateMoveY(final int y, final int direction) {
		return MathUtil.modulo(y + TroffGame.DIRECTIONS[direction][1], game.getBoardInfo().getSize().getHeight());
	}

	private byte idx;
	private TroffGame game;
	private PlayerPosInfo posInfo;
	private int turnSuggestion = TURN_STRAIGHT;
	private boolean actionSuggestion = false;
	private int playerMode;

	public TroffPlayer(final TroffGame game, final byte i, final PlayerPosInfo playerPosInfo, final int playerMode) {
		this.game = game;
		this.idx = i;
		this.posInfo = playerPosInfo;
		this.playerMode = playerMode;
	}

	public TroffGame getGame() {
		return game;
	}

	public void setGame(final TroffGame game) {
		this.game = game;
	}

	/** 
	 * @return the unique index the player is identified with. Defined by the game. 
	 */
	public byte getIndex() {
		return idx;
	}

	/** 
	 * @param index the unique index the player is identified with. Defined by the game. 
	 */
	public void setIndex(final byte index) {
		this.idx = index;
	}

	public int getPlayerMode() {
		return playerMode;
	}
	
	PlayerPosInfo getPlayerPosInfo() {
		return posInfo;
	}

	/** 
	 * The direction the player is heading to.
	 * @return one of {@link TroffGame#NORTH}, {@link TroffGame#EAST}, {@link TroffGame#SOUTH}, {@link TroffGame#WEST}
	 */
	public int getDirection() {
		return posInfo.getDirection();
	}

	/** 
	 * The direction the player is heading to, as displayed on the screen.
	 * 
	 * @return one of {@link TroffGame#NORTH}, {@link TroffGame#EAST}, {@link TroffGame#SOUTH}, {@link TroffGame#WEST}
	 */
	public int getDirectionToShow() {
		return posInfo.getDirection();
	}

	/**
	 * The turn suggestion is provided by the user, e.g., via keyboard.
	 * 
	 * @return one of {@link TroffPlayer#TURN_LEFT}, {@link TroffPlayer#TURN_STRAIGHT}, {@link TroffPlayer#TURN_RIGHT}
	 */
	public int getTurnSuggestion() {
		return turnSuggestion;
	}

	/**
	 * The turn suggestion is provided by the user, e.g., via keyboard.
	 * 
	 * @param turnSuggestion
	 *            one of {@link TroffPlayer#TURN_LEFT}, {@link TroffPlayer#TURN_STRAIGHT},
	 *            {@link TroffPlayer#TURN_RIGHT}
	 */
	protected void setTurnSuggestion(final int turnSuggestion) {
		if (isAlive()) {
			this.turnSuggestion = turnSuggestion;
		}
	}

	public boolean getActionSuggestion() {
		return actionSuggestion;
	}

	public void  setActionSuggestion(final boolean actionSuggestion) {
		if (isAlive()) {
			this.actionSuggestion = actionSuggestion;
		}
	}

	public void actionGameInput(final int suggestion) {
		if (isAlive()) {
			if ((TURN_LEFT == suggestion) || (TURN_RIGHT == suggestion)) {
				setTurnSuggestion(suggestion);
			} else if (ACTION == suggestion) {
				setActionSuggestion(!actionSuggestion);
				logger.debug("Set action suggestion to " + getActionSuggestion());
			} else {
				assert(false);
			}
		}
	}

	/** 
	 * The direction the player is heading to, taking the suggestion into account.
	 * @return one of {@link TroffGame#NORTH}, {@link TroffGame#EAST}, {@link TroffGame#SOUTH}, {@link TroffGame#WEST} 
	 */
	public int getSuggestedDirection() {
		return PlayerPosInfo.calculateTurnRight(posInfo.getDirection(), turnSuggestion);
	}

	private int calculateTurnRight(final int howmuch) {
		int ret = posInfo.getDirection();
		if (isAlive()) {
			ret = PlayerPosInfo.calculateTurnRight(posInfo.getDirection(), howmuch);
		}
		return ret;
	}

	private int turnRight(final int howmuch) {
		int direction = calculateTurnRight(howmuch);
		posInfo.setDirection(direction);
		return direction;
	}

	public int getX() {
		return posInfo.getXPos();
	}

	public void setX(final int x) {
		posInfo.setXPos(x);
	}

	public int getY() {
		return posInfo.getYPos();
	}

	public void setY(final int y) {
		posInfo.setYPos(y);
	}

	public boolean isAlive() {
		return posInfo.isAlive();
	}

	/**
	 * Sets a player to dead or alive.
	 * @param isAlive <code>true</code>for alive, <code>false</code> for dead
	 */
	public void setAlive(final boolean isAlive) {
		posInfo.setAlive(isAlive);
	}

	public int calculateX(final int turnRight) {
		return calculateMoveX(posInfo.getXPos(), calculateTurnRight(turnRight));
	}

	public int calculateY(final int turnRight) {
		return calculateMoveY(posInfo.getYPos(), calculateTurnRight(turnRight));
	}

	public boolean isShooting() {
		boolean ret = false;
		if (isAlive() && (0 != (TroffGameProperties.MODE_INFOS[playerMode].getModeMask() & TroffGameProperties.ModeInfo.PLAYERMODEMASK_HUNTER))) {
			if (getActionSuggestion()) {
				ret = true;
				setActionSuggestion(false);
			}
		}
		return ret;
	}

	public boolean isTracing() {
		boolean ret = false;
		if (isAlive() && (0 != (TroffGameProperties.MODE_INFOS[playerMode].getModeMask() & TroffGameProperties.ModeInfo.PLAYERMODEMASK_PREY))) {
			ret = getActionSuggestion();
		}
		return ret;
	}

	/**
	 * Moves the player one step forward in the decided direction.
	 * @return <code>true</code> if the player just died
	 */
	public boolean move() {
		boolean ret = false;
		if (isAlive()) {
			int direction = turnRight(decideTurn());
			posInfo.setDirection(direction);
			turnSuggestion = 0;
			int nx = calculateMoveX(posInfo.getXPos(), direction);
			int ny = calculateMoveY(posInfo.getYPos(), direction);
			if (game.isLegalMove(this, nx, ny)) {
				posInfo.setXYPos(nx, ny);
			} else {
				setAlive(false);
				ret = true;
			}
		}
		return ret;
	}

	/**
	 * It's up to the concrete strategy to decide where to go for the next turn.
	 * 
	 * @return one of {@link #TURN_LEFT}, {@link #TURN_STRAIGHT}, {@link #TURN_RIGHT}
	 */
	public abstract int decideTurn();
}
