package de.ped.troff.test;

import static org.junit.Assert.*;

import java.util.Arrays;
import java.util.Collection;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;

import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.troff.client.gui.TroffPlayerViewCanvas;
import de.ped.troff.client.logic.Abstract3DGeometry;
import de.ped.troff.client.logic.CubeSideKey;
import de.ped.troff.client.logic.Real3DGeometry;
import de.ped.troff.client.logic.Side;
import de.ped.troff.client.logic.Simple3DGeometry;

/**
 * Unit tests for the {@link Abstract3DGeometry} class.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
@RunWith(value=Parameterized.class)
public class Abstract3DGeometryTest {

	@Parameters
	public static Collection<Abstract3DGeometry[]> data() {
		return Arrays.asList( new Abstract3DGeometry[][] {
			{ new Simple3DGeometry(
					TroffPlayerViewCanvas.SIZE_IN_PIXEL_X,
					TroffPlayerViewCanvas.SIZE_IN_PIXEL_Y,
					TroffPlayerViewCanvas.PILOTVIEW_HORIZON_DISTANCE), },
			{ new Real3DGeometry(
					TroffPlayerViewCanvas.SIZE_IN_PIXEL_X,
					TroffPlayerViewCanvas.SIZE_IN_PIXEL_Y,
					TroffPlayerViewCanvas.PILOTVIEW_HORIZON_DISTANCE), }
		});
	}

	@Before
	public void setup() {
		logger.setLogLevel(Level.INFO);
	}

	public Abstract3DGeometryTest(final Abstract3DGeometry testling) {
		this.testling = testling;
	}

	@Override
	public String toString() {
		return "Testling=" + testling.getClass().getSimpleName();
	}

	@Test(expected = IllegalArgumentException.class)
	public void testGetBoxSideIntIntInt_IllegalSide_Fails() {
		testling.getBoxSide(1, 1, null, Abstract3DGeometry.BOX_SIDE);
	}

	@Test(expected = IllegalArgumentException.class)
	public void testGetBoxSideIntIntInt_IllegalWhat_Fails() {
		testling.getBoxSide(1, 1, Side.FRONT, -1);
	}

	@Test
	public void testGetBoxSideIntIntInt_BoxInBack_Null() {
		final int[][] poly = testling.getBoxSide(-1, 0, Side.BACK, Abstract3DGeometry.BOX_SIDE);
		assertNull(toString(), poly);
	}

	@Test
	public void testGetBoxSideIntIntInt_BoxTooFar_Null() {
		final int[][] polyFront = testling.getBoxSide(testling.getHorizonCircle().size(), 1, Side.FRONT,
				Abstract3DGeometry.BOX_SIDE);
		assertNull(toString(), polyFront);
		final int[][] polyLeft = testling.getBoxSide(testling.getHorizonCircle().size() - 1, 1, Side.LEFT,
				Abstract3DGeometry.BOX_SIDE);
		assertNull(toString(), polyLeft);
	}

	public int[][] checkPoly(CubeSideKey key) {
		final String parameters = toString() + ";" + key.toString();
		final int[][] poly = testling.getBoxSide(key);
		// Check for right numbers
		assertNotNull(parameters, poly);
		assertEquals(parameters, 2, poly.length);
		assertEquals(parameters, 4, poly[Abstract3DGeometry.X].length);
		assertEquals(parameters, 4, poly[Abstract3DGeometry.Y].length);
		if (Abstract3DGeometry.BOX_SIDE == key.what) {
			// Check if all x values are on the right half, depending on xpos
			int leftHalfCountExpected = 2; 
			if (0 < key.xpos) {
				leftHalfCountExpected = 0;
			} else if (0 > key.xpos) {
				leftHalfCountExpected = 4;
			}
			int leftHalfCount = 0;
			for (int x : poly[Abstract3DGeometry.X]) {
				if (x < testling.sizeInPixelX / 2) {
					leftHalfCount++;
				}
			}
			assertEquals(parameters, leftHalfCountExpected, leftHalfCount);
			// Check if all y values are on the right half
	//		int topHalfCount = 0;
	//		for (int y : poly[Abstract3DGeometry.Y]) {
	//			if (y <= testling.sizeInPixelX / 2 + 8) {
	//				topHalfCount++;
	//			}
	//		}
	//		assertEquals(parameters, 2, topHalfCount);
		}
		return poly;
	}

	@Test
	public void testGetBoxSideIntIntInt_00Bottom() {
		final CubeSideKey key = new CubeSideKey(0, 0, Side.BOTTOM, Abstract3DGeometry.BOX_SIDE);
		final String parameters = toString() + ";" + key.toString();
		final int[][] poly = checkPoly(key);
		// Check if all y values are on the right half
		for (int y : poly[Abstract3DGeometry.Y]) {
			assertFalse(parameters + ";y=" + y, y <= testling.sizeInPixelX / 2);
		}
		logger.debug(parameters + ": " + Abstract3DGeometry.toString(poly));
	}

	@Test
	public void testGetBoxSideIntIntInt_01Bottom() {
		final CubeSideKey key = new CubeSideKey(0, 1, Side.BOTTOM, Abstract3DGeometry.BOX_SIDE);
		final String parameters = toString() + ";" + key.toString();
		final int[][] poly = checkPoly(key);
		// Check if all y values are on the right half
		for (int y : poly[Abstract3DGeometry.Y]) {
			assertFalse(parameters + ";y=" + y, y <= testling.sizeInPixelX / 2);
		}
		logger.debug(parameters + ": " + Abstract3DGeometry.toString(poly));
	}

	@Test
	public void testGetBoxSideIntIntInt_10Bottom() {
		final CubeSideKey key = new CubeSideKey(1, 0, Side.BOTTOM, Abstract3DGeometry.BOX_SIDE);
		final String parameters = toString() + ";" + key.toString();
		final int[][] poly = checkPoly(key);
		// Check if all y values are on the right half
		for (int y : poly[Abstract3DGeometry.Y]) {
			assertFalse(parameters + ";" + Abstract3DGeometry.toString(poly) + ";y=" + y, 
					y <= testling.sizeInPixelX / 2);
		}
		logger.debug(parameters + ": " + Abstract3DGeometry.toString(poly));
	}

	@Test
	public void testGetBoxSideIntIntInt_00Top() {
		final CubeSideKey key = new CubeSideKey(0, 0, Side.TOP, Abstract3DGeometry.BOX_SIDE);
		final String parameters = toString() + ";" + key.toString();
		final int[][] poly = checkPoly(key);
		// Check if all y values are on the right half
		for (int y : poly[Abstract3DGeometry.Y]) {
			assertTrue(parameters + ";y=" + y, y <= testling.sizeInPixelX / 2);
		}
		logger.debug(parameters + ": " + Abstract3DGeometry.toString(poly));
	}

	@Test
	public void testGetBoxSideIntIntInt_BoxNearlyTooFar_OnlyFrontSide() {
		checkPoly(new CubeSideKey(testling.getHorizonCircle().size() - 1, 1, Side.FRONT, Abstract3DGeometry.BOX_SIDE));
		checkPoly(new CubeSideKey(testling.getHorizonCircle().size() - 2, 1, Side.LEFT, Abstract3DGeometry.BOX_SIDE));
	}

	private Abstract3DGeometry testling;

	private Logger logger = Logger.getLogger(this.getClass());
}
