package de.ped.troff.test;

import static org.junit.Assert.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;

import org.junit.Before;
import org.junit.Test;

import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.tools.test.GenericTest;
import de.ped.troff.server.logic.BoardFactory;
import de.ped.troff.server.logic.BoardInfo;
import de.ped.troff.server.logic.TroffGame;
import de.ped.troff.server.logic.PlayerPosInfo;

/**
 * Unit tests for the {@link BoardInfo} class.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class BoardInfoTest {

	private Logger logger = Logger.getLogger(this.getClass());

	@Before
	public void setup() {
		logger.setLogLevel(Level.INFO);
	}

	@Test(expected = NullPointerException.class)
	public void test_NullBoard_AllNull() {
		final BoardInfo bi = new BoardInfo(null, 0, 0, null);
		assertNull(bi.getKey());
		assertNull(bi.getPlayerPosInfos());
		assertEquals(0, bi.getSize().getWidth());
		assertEquals(0, bi.getSize().getHeight());

		// int[][] cb = bi.cloneBoard();
		// assertEquals(0, cb.length);

		final BoardInfo cbi = (BoardInfo) bi.clone();
		assertNull(cbi.getKey());
		assertNull(cbi.getPlayerPosInfos());
		assertEquals(0, cbi.getSize().getWidth());
		assertEquals(0, cbi.getSize().getHeight());
	}

	@Test
	public void test_Board1x1_Empty() {
		final PlayerPosInfo ppi = new PlayerPosInfo(0, 0, TroffGame.NORTH);
		final BoardInfo bi = new BoardInfo("Hallo", 1, 1, new PlayerPosInfo[] { ppi });
		assertEquals("Hallo", bi.getKey());
		assertEquals(1, bi.getPlayerPosInfos().length);
		assertEquals(ppi, bi.getPlayerPosInfos()[0]);
		assertEquals(1, bi.getSize().getWidth());
		assertEquals(1, bi.getSize().getHeight());
		assertEquals(BoardInfo.FREE, bi.getAt(0, 0));
		// Clone board
		// int[][] cb = bi.cloneBoard();
		// assertEquals(1, cb.length);
		// assertEquals(1, cb[0].length);
		// assertEquals(BoardInfo.FREE, cb[0][0]);
		// // Is cloned board really a clone?
		// cb[0][0] = BoardInfo.SOLID_WALL;
		assertEquals(BoardInfo.FREE, bi.getAt(0, 0));
		// Clone whole object
		final BoardInfo cbi = (BoardInfo) bi.clone();
		assertEquals("Hallo", cbi.getKey());
		assertEquals(1, cbi.getPlayerPosInfos().length);
		assertEquals(ppi, cbi.getPlayerPosInfos()[0]);
		assertEquals(1, cbi.getSize().getWidth());
		assertEquals(1, cbi.getSize().getHeight());
		assertEquals(BoardInfo.FREE, cbi.getAt(0, 0));
		// Is cloned object really a clone?
		cbi.setAt(0, 0, BoardInfo.SOLID_WALL);
		assertEquals(BoardInfo.FREE, bi.getAt(0, 0));
	}

	@Test
	public void testSerialization() throws IOException, ClassNotFoundException {
		final BoardInfo b = BoardFactory.createBoard(1, 8);
		final ByteArrayOutputStream baos = new ByteArrayOutputStream();
		final ObjectOutputStream oos = new ObjectOutputStream(baos);
		oos.writeObject(b);
		final byte[] arr = baos.toByteArray();
		logger.debug("testSerialization(): Array length=" + arr.length);
		final ByteArrayInputStream bais = new ByteArrayInputStream(arr);
		final ObjectInputStream ois = new ObjectInputStream(bais);
		final Object o = ois.readObject();
		final BoardInfo result = (BoardInfo) o;
		assertEquals(b, result);
	}

	@SuppressWarnings("unused")
	@Test
	public void testNibbleConversion() {
		final byte b00 = (byte) 0x00;
		final byte b05 = (byte) 0x05;
		final byte b0a = (byte) 0x0a;
		final byte b0f = (byte) 0x0f;
		final byte b50 = (byte) 0x50;
		final byte b55 = (byte) 0x55;
		final byte b5a = (byte) 0x5a;
		final byte b5f = (byte) 0x5f;
		final byte ba0 = (byte) 0xa0;
		final byte ba5 = (byte) 0xa5;
		final byte baa = (byte) 0xaa;
		final byte baf = (byte) 0xaf;
		final byte bf0 = (byte) 0xf0;
		final byte bf5 = (byte) 0xf5;
		final byte bfa = (byte) 0xfa;
		final byte bff = (byte) 0xff;

		assertEquals(0x0, BoardInfo.getNibble(b00, true));
		assertEquals(0x0, BoardInfo.getNibble(b00, false));
		assertEquals(0xf, BoardInfo.getNibble(bff, true));
		assertEquals(0xf, BoardInfo.getNibble(bff, false));

		assertEquals(b00, BoardInfo.setNibble(b00, b00, false));
		assertEquals(bff, BoardInfo.setNibble(bff, b0f, false));
		assertEquals(bff, BoardInfo.setNibble(bff, b0f, true));
		assertEquals(bfa, BoardInfo.setNibble(bf0, b0a, false));
		assertEquals(bfa, BoardInfo.setNibble(b0a, b0f, true));
		assertEquals(bfa, BoardInfo.setNibble(b0a, b5f, true)); // See if high nibble of nibble parameter is ignored
	}

	/**
	 * @see BoardInfo#equals(Object)
	 * @see BoardInfo#clone()
	 */
	@Test
	public void testObjectBasics() {
		GenericTest.checkObjectBasics(true, 
				new BoardInfo(), 
				new BoardInfo("key", 1, 1, new PlayerPosInfo[] { new PlayerPosInfo(0, 0, TroffGame.NORTH), } ));
	}
}
