package de.ped.troff.test;

import static org.junit.Assert.*;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.HashMap;

import de.ped.tools.I18NStringWithFillIns;
import de.ped.tools.Messages;
import de.ped.tools.log.Level;
import de.ped.tools.log.Logger;
import de.ped.troff.client.gui.Troff;
import de.ped.troff.client.gui.TroffMainWindow;
import de.ped.troff.client.logic.ClientThread;
import de.ped.troff.middleware.Environment;
import de.ped.troff.middleware.Message;
import de.ped.troff.middleware.ServerInfo;
import de.ped.troff.middleware.ServerSelectionModel;
import de.ped.troff.middleware.TroffRoundsManager;
import de.ped.troff.server.TroffServerThread;
import de.ped.troff.server.logic.TroffGameState;
import de.ped.troff.server.logic.TroffGameProperties;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

/**
 * Tests if the communication protocol between client and server runs smoothly.
 * 
 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class ClientServerCommunicationTest {

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		// Nothing to do
	}

	@AfterClass
	public static void tearDownAfterClass() throws Exception {
		// Nothing to do
	}

	private static class TroffRoundsManagerProxyTestData {
		public int notified = 0;
	}

	public static final String SERVERNAME = ServerSelectionModel.SERVERNAME_LOCAL;

	/**
	 * Stub of the {@link ClientThread} for the {@link TroffRoundsManager} to
	 * have someone to notify on changes.
	 * 
	 * @author Copyright (c) 2012 Peter Diefenbach (peter@pdiefenbach.de)
	 */
	public static class ClientThreadStub extends ClientThread {
		private final ServerSelectionModel model;

		private HashMap<TroffRoundsManager, TroffRoundsManagerProxyTestData> proxies;

		public ClientThreadStub(
				final TroffMainWindow window, 
				final Environment environment, 
				final ServerSelectionModel model) throws UnknownHostException, IOException {
			super(window, environment, model);
			this.model = model;
			proxies = new HashMap<TroffRoundsManager, TroffRoundsManagerProxyTestData>();
		}

		public ServerSelectionModel getModel() {
			return model;
		}

		void addProxy(final TroffRoundsManager proxy) {
			proxies.put(proxy, new TroffRoundsManagerProxyTestData());
		}

		int getNotifiedCount(final TroffRoundsManager proxy) {
			return proxies.get(proxy).notified;
		}

		@Override
		public void update(final TroffRoundsManager who, final Message msg) {
			logger.debug("Update called.");
			if (proxies.containsKey(who)) {
				final TroffRoundsManager proxy = (TroffRoundsManager) who;
				proxies.get(who).notified++;
				logger.debug("Updated " + proxy.getConnector().getClientName() + " part " + msg.toString());
			}
		}

	}

	private Logger logger = Logger.getLogger(this.getClass());

	private ServerSelectionModel model = new ServerSelectionModel();
	private TroffServerThread serverThread;
	private ClientThreadStub clientThreadStub;

	@Before
	public void setUp() throws Exception {
		logger.setLogLevel(Level.INFO);
		final Environment environment = new Environment();
		model.setHostType(ServerSelectionModel.HostType.LOCALLY);
		serverThread = new TroffServerThread(environment, model);
		serverThread.start();
		clientThreadStub = new ClientThreadStub(null, environment, model);
	}

	@After
	public void tearDown() throws Exception {
		serverThread.terminate();
		clientThreadStub = null;
	}

	public static final String CLIENT_INET_ADDRESS_DEFAULT = "127.0.0.1";
	public static final String[] CLIENT_INET_ADDRESSES = new String[]{ CLIENT_INET_ADDRESS_DEFAULT, "CNKAR118.ict-se.de", "", "local" };

	public static final String CLIENT_NAME = "TestClientXYZ";
	public static final String CLIENT_NAME2 = "TestClientABC";
	public static final String GROUP_NAME = "TestGroupXYZ";

	public static final String CONNECT_RESPONSE_INITIAL = SERVERNAME + " (0)\n  default (0)\n";

	public static final String CONNECT_RESPONSE_AFTER_JOIN = SERVERNAME + " (1)\n  default (0)\n  " + GROUP_NAME
			+ " (1)\n    " + CLIENT_NAME + " (1): " + CLIENT_INET_ADDRESS_DEFAULT + "\n";
	public static final String CONNECT_RESPONSE_AFTER_JOIN2 = SERVERNAME + " (2)\n  default (0)\n  " + GROUP_NAME
			+ " (2)\n    " + CLIENT_NAME + " (1): " + CLIENT_INET_ADDRESS_DEFAULT + "\n    " + CLIENT_NAME2 + " (1): "
			+ CLIENT_INET_ADDRESS_DEFAULT + "\n";

	public void login(final TroffRoundsManager proxy, final String clientName, final boolean expected)
			throws IOException {
		final Message request = Message.createLoginReq(clientName, (byte) 1);
		proxy.sendMessage(request);
		final Message response = proxy.readMessagesTillResponse(request);
		assertEquals(expected, response.wasSuccess());
	}

	public void logout(final TroffRoundsManager proxy) throws IOException {
		final Message request = Message.createLogoutReq();
		proxy.sendMessage(request);
		/* Message response = */ proxy.readMessagesTillResponse(request);
	}

	public void join(final TroffRoundsManager proxy, final String groupName, final boolean expected)
			throws IOException {
		final Message request = Message.createJoinReq(groupName);
		proxy.sendMessage(request);
		final Message response = proxy.readMessagesTillResponse(request);
		assertEquals(expected, response.wasSuccess());
	}

	public void leave(final TroffRoundsManager proxy) throws IOException {
		final Message request = Message.createLeaveReq();
		proxy.sendMessage(request);
		/* Message response = */ proxy.readMessagesTillResponse(request);
	}

	public void startGame(final TroffRoundsManager proxy) throws IOException {
		final Message request = Message.createStartGameReq(new TroffGameProperties());
		proxy.sendMessage(request);
		/* Message response = */ proxy.readMessagesTillResponse(request);
	}

	protected void checkMessageLine(final String expectedTemplate, final String result) {
		boolean isFound = false;
		for (final String clientInetAddress : CLIENT_INET_ADDRESSES) {
			final String expected = expectedTemplate.replaceAll(CLIENT_INET_ADDRESS_DEFAULT, clientInetAddress);
			if (expected.equals(result)) {
				isFound = true;
				break;
			}
		}
		if (!isFound) {
			assertEquals(expectedTemplate, result);
		}
	}

	protected void checkGroupMessages(final String expectedGroupMessages, final I18NStringWithFillIns[] result) {
		if (null != result) {
			final String[] strs = convertToStrings(result);
			checkGroupMessages(expectedGroupMessages, strs);
		}
	}

	public static String[] convertToStrings(final I18NStringWithFillIns[] result) {
		final Messages messages = createMessages();
		final String[] strs = new String[result.length];
		for (int i = 0; i < result.length; i++) {
			strs[i] = messages.getText(result[i]);
		}
		return strs;
	}

	private synchronized static Messages createMessages() {
		if (null == messages) {
			final Messages msgs = new Messages(Troff.MESSAGES_PACKAGE);
			msgs.setCurrentLocale(Messages.DEFAULT_LOCALES[1]);
			messages = msgs;
		}
		return messages;
	}
	
	private static Messages messages = null;

	protected void checkGroupMessages(final String expectedGroupMessages, final String[] result) {
		if (null != expectedGroupMessages) {
			// for (String line : result) {
			// logger.debug(line);
			// }
			final String[] expectedLines = expectedGroupMessages.split("\n");
			assertEquals(expectedLines.length, result.length);
			for (int i = 0; i < expectedLines.length; i++) {
				checkMessageLine(expectedLines[i], result[i]);
			}
		}
	}

	protected void checkServerInfo(final TroffRoundsManager proxy, final String expectedState,
			final String expectedGroupMessages) throws IOException {
		final Message request = Message.createServerStatusReq();
		proxy.sendMessage(request);
		final Message response = proxy.readMessagesTillResponse(request);
		assertEquals(Message.getMsgType(Message.SERVER_STATUS, Message.RESPONSE), response.getMsgType());
		final ServerInfo state = response.getServerInfo();
		checkMessageLine(expectedState, state.toString());
		checkGroupMessages(expectedGroupMessages, proxy.getLastReceivedTextMessages());
	}

	@Ignore
	public void testLocalServerSingleClientNormalFlow() throws UnknownHostException, IOException, InterruptedException {
		final TroffRoundsManager proxy = new TroffRoundsManager(clientThreadStub, null, model.getHostPortNo(),
				new Logger("Client"));
		clientThreadStub.addProxy(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, null);

		login(proxy, CLIENT_NAME, true);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL,
				"Client TestClientXYZ connected to server " + SERVERNAME + ".\n");

		join(proxy, GROUP_NAME, true);
		checkServerInfo(proxy, CONNECT_RESPONSE_AFTER_JOIN, "Client TestClientXYZ connected to server " + SERVERNAME
				+ ".\n" + "Client TestClientXYZ joined group TestGroupXYZ.\n");

		assertNull(proxy.getLastReceivedGameState());

		startGame(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_AFTER_JOIN, "Client TestClientXYZ connected to server " + SERVERNAME
				+ ".\n" + "Client TestClientXYZ joined group TestGroupXYZ.\n");

		Thread.sleep(1000L);
		assertNotNull(proxy.getLastReceivedGameState());
		assertTrue(0 < clientThreadStub.getNotifiedCount(proxy));

		leave(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, 
				"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
				"Client TestClientXYZ joined group TestGroupXYZ.\n" +
				"Client TestClientXYZ has left group TestGroupXYZ.\n");

		logout(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, null);

		proxy.disconnect();
		logger.debug("Client: Done.");
	}

	@Test
	public void testSingleClientNormalFlow() throws UnknownHostException, IOException, InterruptedException {
		final TroffRoundsManager proxy = new TroffRoundsManager(clientThreadStub, model.getHostName(), model.getHostPortNo(), new Logger("Client"));
		clientThreadStub.addProxy(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, 
				null);

		login(proxy, CLIENT_NAME, true);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, 
				"Client TestClientXYZ connected to server " + SERVERNAME + ".\n");

		join(proxy, GROUP_NAME, true);
		checkServerInfo(proxy, CONNECT_RESPONSE_AFTER_JOIN, 
				"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
				"Client TestClientXYZ joined group TestGroupXYZ.\n");

		assertNull(proxy.getLastReceivedGameState());

		startGame(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_AFTER_JOIN, 
				"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
				"Client TestClientXYZ joined group TestGroupXYZ.\n");
		
//		GameState gameState = proxy.getGameState();

		Thread.sleep(20000L);
		final TroffGameState state = proxy.getLastReceivedGameState();
		assertNotNull(state);
		assertTrue(0 < clientThreadStub.getNotifiedCount(proxy));

		leave(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, 
				"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
				"Client TestClientXYZ joined group TestGroupXYZ.\n" +
				"Client TestClientXYZ has left group TestGroupXYZ.\n");

		logout(proxy);
		checkServerInfo(proxy, CONNECT_RESPONSE_INITIAL, null);

		proxy.disconnect();
		logger.debug("Client: Done.");
	}

// To make this test run, now we need two different threads and non-local connectors.
//    @Ignore
//	public void testMultipleClientsNormalFlow() throws IOException {
//    	TroffRoundsManagerProxy proxy1 = new TroffRoundsManagerProxy(model.getHostName(), model.getHostPortNo(), new Logger("Client1"));
//    	proxy1.addObserver(this);
//    	proxies.put(proxy1, new TroffRoundsManagerProxyTestData());
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_INITIAL, 
//    			null);
//
//    	TroffRoundsManagerProxy proxy2 = new TroffRoundsManagerProxy(model.getHostName(), model.getHostPortNo(), new Logger("Client2"));
//    	proxy2.addObserver(this);
//    	proxies.put(proxy2, new TroffRoundsManagerProxyTestData());
//    	checkServerInfo(proxy2, CONNECT_RESPONSE_INITIAL, 
//    			null);
//
//    	login(proxy1, CLIENT_NAME, true);
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_INITIAL, 
//    			"Client TestClientXYZ connected to server " + SERVERNAME + ".\n");
//    	login(proxy2, CLIENT_NAME, false);
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_INITIAL, 
//    			"Client TestClientXYZ connected to server " + SERVERNAME + ".\n");
//    	login(proxy2, CLIENT_NAME2, true);
//    	checkServerInfo(proxy2, CONNECT_RESPONSE_INITIAL, 
//    			"Client TestClientABC connected to server " + SERVERNAME + ".\n");
//
//    	join(proxy1, GROUP_NAME, true);
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_AFTER_JOIN, 
//    			"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
//    			"Client TestClientXYZ joined group TestGroupXYZ.\n");
//    	join(proxy2, GROUP_NAME, true);
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_AFTER_JOIN2, 
//    			"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
//    			"Client TestClientXYZ joined group TestGroupXYZ.\n" + 
//    			"Client TestClientABC joined group TestGroupXYZ.\n");
//
//    	logout(proxy2);
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_AFTER_JOIN, 
//    			"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
//    			"Client TestClientXYZ joined group TestGroupXYZ.\n" + 
//    			"Client TestClientABC joined group TestGroupXYZ.\n" + 
//    			"Client TestClientABC has left group TestGroupXYZ.\n");
//    	
//    	leave(proxy1);
//    	checkServerInfo(proxy1, CONNECT_RESPONSE_INITIAL, 
//    			"Client TestClientXYZ connected to server " + SERVERNAME + ".\n" +
//    			"Client TestClientXYZ joined group TestGroupXYZ.\n" + 
//    			"Client TestClientABC joined group TestGroupXYZ.\n" + 
//    			"Client TestClientABC has left group TestGroupXYZ.\n" +
//    			"Client TestClientXYZ has left group TestGroupXYZ.\n");
//
//    	proxy1.disconnect();
//    	proxy2.disconnect();
//		logger.debug("Client: Done.");
//	}
	
	@Ignore
	public void testDummy() {
		// Nothing to do
	}
}
