package de.ped.deinbac.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Iterator;
import java.util.List;

import javax.swing.JPanel;

import de.ped.deinbac.logic.Bacterion;
import de.ped.deinbac.logic.DeinbacGame;
import de.ped.deinbac.logic.DeinbacProperties;
import de.ped.deinbac.logic.GameModel;
import de.ped.deinbac.logic.SimulationType;
import de.ped.tools.Assert;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.log.Logger;

/**
 * Paints the simulated world.
 * 
 * @author Copyright (c) 2008 Peter Diefenbach (peter@pdiefenbach.de)
 */
class DeinbacMainCanvas extends JPanel implements MouseListener {
	private static final long serialVersionUID = 1L;

	public static final Dimension MINIMUM_SIZE_IN_PIXEL = new Dimension(100, 50);

	static final int[] MAGNIFICATIONS = new int[] { 8, 4, 2, 1, };
	static final int[] BOXSIZES       = new int[] { 7, 3, 2, 1, };

	public DeinbacMainCanvas(final DeinbacMainWindow parent) {
		this.parent = parent;
		setBackground(Color.black);
		setZoomLevel(1);
		addMouseListener(this);
	}

	@Override
	public void mouseClicked(final MouseEvent e) {
		logger.trace("mouseClicked: " + e.paramString());
		if (isDisplayed) {
			final int worldX = (e.getX() - offset[0]) / magnification;
			final int worldY = (e.getY() - offset[1]) / magnification;
			final PlayfieldDimension playfieldSize = parent.getModel().getSimulatable().getPlayfieldSize();
			// TODO: use PlayfieldPosition instead? Then we could use inBounds here.
			switch (parent.getModel().getProps().getSimulation()) {
			case DEINBAC:
				if ((0 <= worldX) && (worldX < playfieldSize.getWidth()) && (0 <= worldY) && (worldY < playfieldSize.getHeight())) {
					parent.showPositionInfo(worldX, worldY);
				}
				break;
			case GAMEOFLIFE:
				parent.getModel().life.togglePlayfieldAt(worldX,  worldY);
				repaint();
				break;
			}
		}
	}

	@Override
	public void mousePressed(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseReleased(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseEntered(final MouseEvent e) {
		// Nothing to do
	}

	@Override
	public void mouseExited(final MouseEvent e) {
		// Nothing to do
	}

	/**
	 * @return the preferred size of this canvas.
	 */
	@Override
	public Dimension getPreferredSize() {
		return getMinimumSize();
	}

	@Override
	public Dimension getMinimumSize() {
		final PlayfieldDimension playfieldSize = parent.getModel().getSimulatable().getPlayfieldSize();
		final int xsize = Math.max(playfieldSize.getWidth() * magnification, MINIMUM_SIZE_IN_PIXEL.width);
		final int ysize = Math.max(playfieldSize.getHeight() * magnification, MINIMUM_SIZE_IN_PIXEL.height);
		return new Dimension(xsize, ysize);
	}

	@Override
	public void paintComponent(final Graphics g) {
		super.paintComponent(g);
		if (isDisplayed) {
			mypaint(g);
		}
	}

	/**
	 * Paints the whole mana world.
	 * 
	 * @param g
	 *            the graphics object to paint into
	 */
	private void mypaint(final Graphics g) {
		final GameModel gameModel = parent.getModel();
		final PlayfieldDimension playfieldSize = gameModel.getSimulatable().getPlayfieldSize();
		final Dimension worldInPixel = new Dimension(
				playfieldSize.getWidth()  * magnification, 
				playfieldSize.getHeight() * magnification);
		final Dimension screenSize = getSize();
		offset[0] = Math.max(0,  (screenSize.width  - worldInPixel.width ) / 2);
		offset[1] = Math.max(0,  (screenSize.height - worldInPixel.height) / 2);
		final int boxsize = BOXSIZES[zoomLevel];
		g.setColor(getBackground());
		setBackground(Color.decode("0x000040"));
		g.setColor(Color.black);
		g.fillRect(offset[0], offset[1], worldInPixel.width, worldInPixel.height);
		g.setColor(Color.gray);
		for (int x = 0; x < playfieldSize.getWidth(); x++) {
			for (int y = 0; y < playfieldSize.getHeight(); y++) {
				final int value = gameModel.getSimulatable().getPlayfieldAt(x, y);
				if (0 != value) {
					switch(value) {
					case 1: 
						g.setColor(Color.gray);
						break;
					case -1:
						g.setColor(Color.darkGray.darker());
						break;
					default:
						// Nothing to do
					}
					g.fillRect(
							offset[0] + magnification * x, 
							offset[1] + magnification * y, boxsize, boxsize);
				}
			}
		}
		if (gameModel.getProps().getSimulation() == SimulationType.DEINBAC) {
			final DeinbacGame model = parent.getModel().deinbac;
			final DeinbacProperties props = model.getProps();
			final List<Bacterion> bacs = model.getBacterions();
			final Iterator<Bacterion> it = bacs.iterator();
			while (it.hasNext()) {
				final Bacterion b = it.next();
				Assert.assertNotNull(b);
				if (b.getEnergy() >= 0) {
					final long age = model.getAge(b);
					if (age > 10) {
						Color c = Color.magenta;
						final int energy = b.getEnergy();
						int colorCounter = 0;
						if (props.getEnergyForSplit() <= energy) { colorCounter++; }
						if (props.getEnergyForSex()   <= energy) { colorCounter++; }
						if (0 == colorCounter) {
							c = c.darker();
						} else if (2 == colorCounter) {
							c = c.brighter();
						}
						g.setColor(c);
					} else {
						g.setColor(Color.yellow);
					}
					// If bacterion is on border in a torus world, we might need to draw it several times, clipping away the double parts
					final int bx = b.getX();
					final int by = b.getY();
					final int xsize = props.getXsize();
					final int ysize = props.getYsize();
					Assert.assertTrue(0 <= bx);
					Assert.assertTrue(0 <= by);
					int sx = bx;
					int sy = by;
					boolean needsClip = false;
					if (xsize - Bacterion.XSIZE < bx) {
						sx -= xsize;
						needsClip = true;
					}
					if (ysize - Bacterion.YSIZE < by) {
						sy -= ysize;
						needsClip = true;
					}
					if (needsClip) {
						g.setClip(offset[0], offset[1], 
								magnification * playfieldSize.getWidth(), 
								magnification * playfieldSize.getHeight());
					}
					for (int x = sx; x <= bx; x+= xsize) {
						for (int y = sy; y <= by; y+= ysize) {
							g.drawRect(
									offset[0] + magnification * x,
									offset[1] + magnification * y, 
									magnification * Bacterion.XSIZE - 1, 
									magnification * Bacterion.YSIZE - 1);
						}
					}
					if (needsClip) {
						g.setClip(null);
					}
				}
			}
		}
	}

	public boolean isDisplayed() {
		return isDisplayed;
	}

	public void setDisplayed(final boolean isDisplayed) {
		this.isDisplayed = isDisplayed;
	}

	public int getMagnification() {
		return magnification;
	}

	public int getZoomLevel() {
		return zoomLevel;
	}

	public void setZoomLevel(final int level) {
		this.zoomLevel = level;
		magnification = MAGNIFICATIONS[zoomLevel];
	}

	public int[] offset = new int[] { 50, 30 };

	private DeinbacMainWindow parent;

	private boolean isDisplayed = true;

	private int magnification;
	private int zoomLevel;
	
	private Logger logger = Logger.getLogger(this.getClass());
}
