package de.ped.dsatool.gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Iterator;

import javax.swing.JComboBox;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import de.ped.tools.Messages;
import de.ped.tools.gui.ApplicationMainWindow;
import de.ped.tools.gui.PedJDialog;
import de.ped.tools.gui.PropertyElement;
import de.ped.tools.log.Logger;
import de.ped.dsatool.dsa.generated.RegionType;
import de.ped.dsatool.dsa.generated.TownType;
import de.ped.dsatool.logic.DSA;
import de.ped.dsatool.logic.Town;

/**
 * The dialog that asks the user for the kind of town he/she wants to create.
 *
 * @author Copyright (c) 1998 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class TownDialog extends PedJDialog implements ActionListener, ItemListener {
	private static final long serialVersionUID = 1L;
	private static Logger logger = DSATool.getLogger();

	/**
	 * Constructor.
	 * 
	 * @param parent
	 *            the parent frame
	 * @param model
	 *            the model to operate on
	 */
	public TownDialog(final ApplicationMainWindow parent, final TownType model) {
		super(parent, parent);
		final Messages messages = getMessages();
		setTitle(messages.getText("Action.CreateTown"));
		town = model;
		itemOwnTown = messages.getText("TownDialog.SelectTown.Predefined.Item.OwnTown");
		// Fill the choice box for predefined towns.
		choiceTownPredefined = new JComboBox<String>();
		choiceTownPredefined.addItem(itemOwnTown); 
		final Iterator<TownType> it = DSA.instance().getTown().iterator();
		while (it.hasNext()) {
			final TownType town = (TownType)it.next();
			choiceTownPredefined.addItem(town.getName());
		}
		setPropertyElements(new PropertyElement[] {
				new PropertyElement("TownDialog.SelectTown.Predefined.Selection", choiceTownPredefined),
				new PropertyElement("TownDialog.SelectTown.Own", textName),
				new PropertyElement("TownDialog.NumberOfInhabitants", textInhabitants),
				new PropertyElement("TownDialog.Region", choiceRegion),
		});

		setTitlePanel(messages.getText("TownDialog.Title")); 
		// Main panel
		final JPanel mainPanel = createMainPanel();
		final GridBagLayout gridbag = new GridBagLayout();
		final GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.WEST; // All elements are left bound to their space
		mainPanel.setLayout(gridbag);
		addAllPropertyElementsTo(mainPanel, c);

		getContentPane().add(new JScrollPane(mainPanel), BorderLayout.CENTER);
		// Button panel
		setOptionType(JOptionPane.OK_CANCEL_OPTION);
		addHelpButton("Action.CreateTown");
		finishLayout();

		updateUIFieldsBasedOnProperties();
		choiceTownPredefined.addItemListener(this);
		textInhabitants.addActionListener(this);
		textName.addActionListener(this);

		enterDialog();
	}

	/**
	 * Invoked when an item state change event occurs.
	 * @param e the event to handle
	 */
	@Override
	public void itemStateChanged(final ItemEvent e) {
		if ((e.getSource() == choiceTownPredefined) && (e.getStateChange() == ItemEvent.SELECTED)) {
			// We have to copy the values of the selected town to the other fields.
			textName.setText(itemOwnTown);
			readUIFieldsIntoProperties();
		}
	}

	/**
	 * Invoked when an action occurs.
	 * @param e the event to handle
	 */
	@Override
	public void actionPerformed(final ActionEvent e) {
		logger.debug(e.paramString());
		final Object source = e.getSource();
		if ((source == getButtonOk()) || (source == textInhabitants) || (source == textName)) {
			// Pressing return in a text field should be handled as pressing the OK button
			readUIFieldsIntoProperties();
			onOk();
		} else if (source == getButtonCancel()) {
			onCancel();
		}
	}

	/** 
	 * @return the model
	 */
	public TownType getTown() {
		return town;
	}

	@Override
	public void readUIFieldsIntoProperties() {
		final TownType predefinedTown = Town.findPredefinedTown((String)choiceTownPredefined.getSelectedItem());
		if (null == predefinedTown) {
			// The choice was "no predefined town", so create a town from scratch.
			town = Town.create(Integer.parseInt(textInhabitants.getText()), textName.getText(), choiceRegion.getSelectedRegion());
		} else {
			// It was a predefined town. So add what we have to add.
			town = Town.create(predefinedTown);
		}
		updateUIFieldsBasedOnProperties();
	}

	@Override
	public void updateUIFieldsBasedOnProperties() {
		final TownType predefinedTown = Town.findPredefinedTown(town.getName());
		final boolean isPredefinedTown = (null != predefinedTown);
		if (isPredefinedTown) {
			choiceTownPredefined.setSelectedItem(town.getName());
		} else {
			choiceTownPredefined.setSelectedItem(itemOwnTown);
		}
		textName.setText(town.getName());
		textInhabitants.setText(Integer.toString(town.getNumberOfInhabitants()));
		choiceRegion.setSelectedRegion((RegionType)town.getRegion());
	}

	/** Model for the dialog. */
	private TownType town;

	private final JTextField textName = new JTextField(20);
	private final JTextField textInhabitants = new JTextField(5);
	private final ChoiceRegion choiceRegion = new ChoiceRegion();
	private final JComboBox<String> choiceTownPredefined;
	private final String itemOwnTown;
}
