package de.ped.empire.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.MouseInfo;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.Collection;
import java.util.ConcurrentModificationException;

import javax.swing.JPanel;

import de.ped.empire.logic.Commandable;
import de.ped.empire.logic.GameProperties;
import de.ped.empire.logic.GameViewProperties;
import de.ped.empire.logic.MoveResult;
import de.ped.empire.logic.MoveResult.CommandableInfo;
import de.ped.empire.logic.PositionedTileable;
import de.ped.empire.logic.TileTemplate;
import de.ped.empire.logic.GameMapInfo;
import de.ped.empire.logic.MapViewMode;
import de.ped.empire.logic.MilitaryBranch;
import de.ped.empire.logic.MovePath;
import de.ped.empire.logic.Player;
import de.ped.empire.logic.StepRatingMap;
import de.ped.empire.logic.StepRatingPosition;
import de.ped.empire.logic.Tileable;
import de.ped.empire.logic.Unit;
import de.ped.empire.logic.ZoomLevel;
import de.ped.empire.logic.MoveResult.Role;
import de.ped.tools.CollectionUtil;
import de.ped.tools.DigitalJoystickDirection;
import de.ped.tools.MathUtil;
import de.ped.tools.PlayfieldDimension;
import de.ped.tools.PlayfieldDistance;
import de.ped.tools.PlayfieldPosition;
import de.ped.tools.PlayfieldTiling;
import de.ped.tools.TopologyMode;
import de.ped.tools.log.Logger;

/**
 * The drawing area for the level and all. 
 * 
 * @author Copyright (c) 2013 Peter Diefenbach (peter@pdiefenbach.de)
 */
public class EmpireMainCanvas extends JPanel implements MouseListener {

	private static final long serialVersionUID = 1L;

	private static final Dimension MINIMUM_SIZE = new Dimension(400, 300);

	private static PlayfieldDistance NULL_VECTOR = new PlayfieldDistance(0, 0);

	public EmpireMainCanvas(final EmpireMainWindow parent) {
		this.parent = parent;
		setBackground(Color.black);
		notifyMapSizeChanged();
		addMouseListener(this);
	}

	@Override
	public void mouseClicked(final MouseEvent ev) {
		try {
			logger.trace("mouseClicked: " + ev.paramString());
			if (ev.getButton() == MouseEvent.BUTTON1) {
				final PlayfieldPosition pos = canvas2map(ev.getX(), ev.getY());
				if (null != pos) {
					parent.actionAtPosition(pos);
				}
			}
		} catch (final Throwable ex) {
			logger.errorWithStackTrace("This should never happen!", ex);
		}
	}

	@Override
	public void mouseEntered(final MouseEvent ev) {
		// Nothing to do
	}

	@Override
	public void mouseExited(final MouseEvent ev) {
		// Nothing to do
	}

	@Override
	public void mousePressed(final MouseEvent ev) {
		// Nothing to do
	}

	@Override
	public void mouseReleased(final MouseEvent ev) {
		// Nothing to do
	}

	@Override
	public void paintComponent(final Graphics g) {
		try {
			super.paintComponent(g);
			if (!parent.isGameRunning()) {
				paintTitle(g);
			} else {
				final int playerId = parent.getCurrentPlayerId();
				if (Player.ID_NONE != playerId) {
					if (parent.isGetReady()) {
						paintPlayerGetReadyBackground(g);
					} else {
						paintGameMap((Graphics2D) g, parent.getCurrentPlayerId());
					}
				} else {
					paintPlayerGetReadyBackground(g);
					g.setColor(Color.DARK_GRAY);
					g.drawString("No game view!", 50, 50);
				}
			}
		} catch(final Throwable ex) {
			logger.errorWithStackTrace("This should never happen!", ex);
		}
	}

	private void paintPlayerGetReadyBackground(final Graphics g) {
		final Dimension screenSize = getSize();
		final Player player;
		final int playerId = parent.getCurrentPlayerId();
		if (Player.ID_NONE != playerId) {
			final GameProperties props = parent.getGameProperties();
			player = props.getPlayers()[playerId];
		} else {
			player = Player.PLAYER_NONE;
		}
		g.setColor(player.getBackground());
		g.fillRect(0, 0, screenSize.width, screenSize.height);
	}

	private void paintTitle(final Graphics g) {
		getTitleImage();
		final Dimension screenSize = getSize();
		g.setColor(Color.decode("0xb97a57"));
		g.fillRect(0, 0, screenSize.width, screenSize.height);
		if (null != titleImage) {
			if ((0 < screenSize.width) && (0 < screenSize.height)) {
				final float screenRatio = (float) screenSize.width / (float) screenSize.height;
				final float imageRatio = (float) titleImage.getWidth() / (float) titleImage.getHeight();
				final Rectangle rect = new Rectangle();
				logger.trace("Screen size=(" + screenSize.width + "," + screenSize.height + ")" + screenRatio + " vs " + imageRatio);
				if (screenRatio < imageRatio) { // screen is narrower than image - fill top and bottom
					rect.x = 0;
					rect.width = screenSize.width;
					rect.height = Math.round(rect.width / imageRatio);
					rect.y = (screenSize.height - rect.height) / 2;
				} else { // screen is wider than image - fill left and right
					rect.y = 0;
					rect.height = screenSize.height;
					rect.width = Math.round(rect.height * imageRatio);
					rect.x = (screenSize.width - rect.width) / 2;
				}
				g.drawImage(titleImage, rect.x, rect.y, rect.width, rect.height, null);
			}
		} else {
			g.setColor(Color.yellow);
			g.drawString("Empire", 50, 50);
		}
	}

	private void paintGameMap(final Graphics2D g, final int playerId) {
		final GameFacadeProxy gameFacade = parent.getGameFacadeProxy();
		final Dimension screenSize = getSize();
		// Paint background for anything outside the map
		g.setColor(Color.DARK_GRAY);
		g.fillRect(0, 0, screenSize.width, screenSize.height);
		// Paint the map
		final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[parent.getZoomLevel()];
		final GameMapInfo mapInfo = gameFacade.createGameMapInfo(playerId);
		final GameProperties props = gameFacade.getGameProperties();
		final TopologyMode topologyMode = props.getTopologyMode();
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(topologyMode);
		mapOffset = getOffset(props);
		final PlayfieldDimension size = mapInfo.getSize();
		final MapViewMode mapViewMode = gameFacade.getMapViewMode(playerId);
		final PlayfieldPosition fieldpos = new PlayfieldPosition(); // re-used for performance reasons.
		final Point tileOffset = new Point(); // re-used for performance reasons.
		for (int x = 0; x < size.getWidth(); x++) {
			for (int y = props.getYMin(x); y < props.getYMax(x); y++) {
				fieldpos.setLocation(x, y);
				map2canvas(tileOffset, fieldpos, PlayfieldTiling.TOP_LEFT);
				try {
					TileTemplate.paintTile(g, tileOffset, mapInfo.getMapAt(fieldpos), zoomLevel, topologyMode);
				} catch (final Exception ex) {
					logger.fatal("Catched exception at " + fieldpos.toString() + " with field " + mapInfo.getMapAt(fieldpos), ex);
				}
			}
		}
		if (parent.isCommandableInFocusBlinking()) {
			final PositionedTileable blinkingTile = mapInfo.getBlinking();
			if (null != blinkingTile) {
				map2canvas(tileOffset, blinkingTile.getPosition(), PlayfieldTiling.TOP_LEFT);
				blinkingTile.paint(g, tileOffset, zoomLevel, topologyMode);
				logger.trace("Painting blinking drawable " + blinkingTile + " at " + blinkingTile.getPosition().toString());
			}
		}
		final float strokeWidth = zoomLevel.strokeWidth;
		final BasicStroke defaultStroke = new BasicStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
		final BasicStroke pathStroke = new BasicStroke(strokeWidth, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
		if (0x00 != mapViewMode.bitmaskUnitsToShow) {
			try {
				if ((mapViewMode.isShowingAnimation())) {
					final Commandable cmdle = mapInfo.getVisibleCommandableInFocus(); 
					if (null != cmdle) {
						final PlayfieldPosition position = cmdle.getPosition();
						if (cmdle instanceof Unit) {
							final Unit unit = (Unit) cmdle;
							final StepRatingMap stepRatings = unit.getStepRatings();
							// If in target selection mode, draw line from commandable in focus
							if (null != parent.getTargetSelectionOrder()) {
								final Point mousePos = MouseInfo.getPointerInfo().getLocation();
								final Point canvasPos = getLocationOnScreen();
								final Point startPosScreen = map2canvas(position, PlayfieldTiling.CENTER);
								final PlayfieldPosition destPosMap = canvas2map(mousePos.x - canvasPos.x, mousePos.y - canvasPos.y);
								if ((null != destPosMap) && (0 != mapInfo.getMapAt(destPosMap))) {
									final Point destPosScreen = map2canvas(destPosMap, PlayfieldTiling.CENTER);
									Color lineColor = Color.lightGray;
									BasicStroke lineStroke = defaultStroke;
									if (null != stepRatings) {
										StepRatingPosition rating = stepRatings.get(destPosMap);
										if (null != rating) {
											final int distance = rating.distanceFromStart;
											g.setColor(Color.yellow);
											g.drawString(Integer.toString(distance), 
													destPosScreen.x - (playfieldTiling.getSurroundingWidth()  / 2), 
													destPosScreen.y - (playfieldTiling.getSurroundingHeight() / 2));
											lineColor = Color.white;
											lineStroke = pathStroke;
										}
									}
									g.setStroke(lineStroke);
									g.setColor(lineColor);
									g.drawLine(startPosScreen.x, startPosScreen.y, destPosScreen.x, destPosScreen.y);
								}
							} else {
								// Draw step ratings
								if (ZoomLevel.INSTANCES[parent.getZoomLevel()].isDistanceInfoPermitted) {
									if (null != stepRatings) {
										final Point ct = new Point(); // Canvas position of target of move
										final GameViewProperties gvp = gameFacade.getGameViewProperties(gameFacade.getCurrentPlayerId());
										if ((null != gvp) && gvp.isShowingDistanceInfo()) {
											g.setStroke(defaultStroke);
											final PlayfieldPosition pt = new PlayfieldPosition(); // target of move - re-used for performance reasons
											final PlayfieldPosition ps = new PlayfieldPosition(); // source of move - re-used for performance reasons
											final Point cs = new Point(); // Canvas position of start of move
											for (int x = 0; x < stepRatings.getSize(PlayfieldPosition.X); x++) {
												pt.setX(x);
												for (int y = 0; y < stepRatings.getSize(PlayfieldPosition.Y); y++) {
													pt.setY(y);
													StepRatingPosition rating = stepRatings.get(pt);
													if ((null != rating) && (null != rating.stepTaken)) {
														g.setColor(rating.isOnShortestPath() ? Color.white : Color.red);
														map2canvas(ct, pt, PlayfieldTiling.CENTER);
														props.calcSourcePositionOfMoveStep(ps, rating.stepTaken, pt);
														map2canvas(cs, ps, PlayfieldTiling.CENTER);
//														if (1 == rating.distanceFromStart) {
//															logger.trace("From " + ps + "/" + cs + " - " + rating.stepTaken.getDistanceVector() + " - " + pt + "/" + ct );
//														}
														g.drawLine(ct.x, ct.y, 
																ct.x + (cs.x - ct.x) / 3, 
																ct.y + (cs.y - ct.y) / 3);
														g.drawString(Integer.toString(rating.distanceFromStart), ct.x, ct.y);
														if (rating.isHomeBase) {
															g.setColor(Color.white);
															map2canvas(ct, pt, PlayfieldTiling.TOP_LEFT);
															g.drawPolygon(playfieldTiling.polygon(pt, ct, 0));
														}
//														final double cnt = rating.getBackwardRating();
//														if (0.0 < cnt) {
//															g.drawString(TextFormatter.format(cnt, 4, 0).toString(), xs, ys);
//														}
													}
												}
											}
										}
										// Draw expected optimal path
										final PlayfieldPosition pos = unit.getPosition().clone();
										final PlayfieldPosition tpos = new PlayfieldPosition();
										final Point t = new Point();
										int loops = 100;
										do {
											StepRatingPosition rating = stepRatings.get(pos);
											if (null == rating) {
												break;
											}
											DigitalJoystickDirection bestDir = null;
											StepRatingPosition bestRating = rating;
											for (final DigitalJoystickDirection dir : props.getNeighborDirections(pos)) {
												final PlayfieldPosition next = props.wrap(pos, dir);
//												if (Position.equals(pos, unit.getPosition())) {
//													logger.trace(Position.toString(pos) + " with dir=" + dir + " --> " + Position.toString(next));
//												}
												final StepRatingPosition nextRating = stepRatings.get(next);
												if ((null != nextRating) 
														&& (0 < nextRating.compareTo(bestRating))) {
													bestDir = dir;
													bestRating = nextRating;
												}
											}
											if (null == bestDir) {
												break;
											}
											logger.trace("Path continues from " + pos.toString() + " to " + bestDir.key + " --> " + bestRating);
											g.setStroke(pathStroke);
											g.setColor(Color.yellow);
											tpos.addInplace(pos, props.renormDistanceVector(bestDir.getDistanceVector(), pos));
											map2canvas(ct, pos, PlayfieldTiling.CENTER);
											map2canvas(t, tpos, PlayfieldTiling.CENTER);
											g.drawLine(ct.x, ct.y, t.x, t.y);
											pos.setLocation(tpos);
											props.wrapInside(pos);
										}
										while (--loops > 0);
										if (loops <= 0) {
											logger.debug("Too many loops!");
										}
									}
								}
							}
						} else {
							if (parent.isCommandableInFocusBlinking()) {
								map2canvas(tileOffset, position, PlayfieldTiling.TOP_LEFT);
								mapInfo.getCrosshair().paint(g, tileOffset, zoomLevel, topologyMode);
							}
						}
					}
					// Draw the current animation
					final Role[] roles = Role.values().clone();
					CollectionUtil.revert(roles);
					for (final Role role : roles) {
						final MoveResult moveResult = mapInfo.getAnimationMoveResult();
						if (null != moveResult) {
							final CommandableInfo info = moveResult.getCommandableInfo(role);
							if (null != info) {
								final PlayfieldPosition pos = info.position;
								final Commandable cmdl = info.getUnit();
								if ((null != pos) && (null != cmdl)) {
									paintTileOffGrid(g, zoomLevel, mapInfo, topologyMode, tileOffset, cmdl, role, pos,
											info.imageKey);
								}
							}
						}
					}
					final PositionedTileable[] fightOverlays = mapInfo.getFightOverlays();
					if (null != fightOverlays) {
						for (final Role role : Role.values()) {
							final Tileable overlay = fightOverlays[role.ordinal()];
							if (null != overlay) {
								final MoveResult moveResult = mapInfo.getAnimationMoveResult();
								if (null != moveResult) {
									final CommandableInfo info = moveResult.getCommandableInfo(role);
									if (null != info) {
										paintTileOffGrid(g, zoomLevel, mapInfo, topologyMode, tileOffset, overlay, role,
												info.position, overlay.getImageKey());
									}
								}
							}
						}
					}
				}
			} catch (final ConcurrentModificationException ex) {
				// Ignore; will be redrawn later
			}
		}
		if (mapViewMode.equals(MapViewMode.PATH)) {
			final float[] pattern = new float[] {1.0f, 3.0f * strokeWidth - 1.0f}; 
			float patternLength = 0.0f;
			for (float f : pattern) {
				patternLength += f;
			}
			final long step = parent.getStep();
			final BasicStroke[] strokes = new BasicStroke[3];
			for (int milTypeId = 0; milTypeId < 3; milTypeId++) {
				final float phase = (float) MathUtil.modulo((patternLength / 3.0f * milTypeId)
						- step / BlinkingState.DURATION_IN_STEPS[parent.PST.ordinal()], patternLength);
				strokes[milTypeId] = new BasicStroke(
						strokeWidth,
						BasicStroke.CAP_ROUND, 
						BasicStroke.JOIN_ROUND, 
						1f, 
						pattern, 
						phase);
			}
			final Collection<MovePath> paths = gameFacade.getMovePaths(playerId);
			final Point startScreenPos = new Point();
			final Point endScreen = new Point();
			for (final MovePath path : paths) {
				final PlayfieldPosition startPos = path.getStartPos();
				final PlayfieldPosition endPos = path.getEndPos();
				map2canvas(startScreenPos, startPos, PlayfieldTiling.CENTER);
				map2canvas(endScreen, endPos, PlayfieldTiling.CENTER);
				for (MilitaryBranch branch : MilitaryBranch.values()) {
					if ((null != branch.color) && (path.isForMilitaryBranch(branch))) {
						g.setStroke (strokes[branch.ordinal()]);
						g.setColor(branch.color);
						g.drawLine(startScreenPos.x,  startScreenPos.y,  endScreen.x,  endScreen.y);
					}
				}
			}
			if (null != parent.getTargetSelectionOrder()) {
				final Point mousePos = MouseInfo.getPointerInfo().getLocation();
				final Point canvasPos = getLocationOnScreen();
				final Commandable commandable = gameFacade.getVisibleCommandableInFocus(playerId);
				final PlayfieldPosition startMapPos = commandable.getPosition();
				final Point destScreenPos = new Point();
				map2canvas(startScreenPos, startMapPos, PlayfieldTiling.CENTER);
				final PlayfieldPosition destMapPos = canvas2map(mousePos.x - canvasPos.x, mousePos.y - canvasPos.y);
				if (null != destMapPos) {
					map2canvas(destScreenPos, destMapPos, PlayfieldTiling.CENTER);
					Color lineColor = Color.lightGray;
					if (destMapPos.isWithin(size)) {
						g.setColor(Color.yellow);
						final int distance = props.calcAirDistance(commandable.getPosition(), destMapPos);
						g.drawString(Integer.toString(distance), 
								destScreenPos.x - (playfieldTiling.getSurroundingWidth() / 2), 
								destScreenPos.y - (playfieldTiling.getSurroundingHeight() / 2));
						lineColor = Color.white;
					}
					g.setStroke (pathStroke);
					g.setColor(lineColor);
					g.drawLine(startScreenPos.x, startScreenPos.y, destScreenPos.x, destScreenPos.y);
				}
			}
		}
	}

	private void paintTileOffGrid(final Graphics2D g, final ZoomLevel zoomLevel, final GameMapInfo mapInfo,
			final TopologyMode topologyMode, final Point tileOffset, final Tileable tile, final Role role,
			final PlayfieldPosition pos, final int imageKey) {
		float dyingProgress = mapInfo.getAnimationDyingProgress(role);
		PlayfieldDistance moveVector = mapInfo.getAnimationMoveVector(role);
		float moveProgress = mapInfo.getAnimationMoveProgress(role);
		if (null == moveVector) {
			moveVector = NULL_VECTOR;
		}
		map2canvas(tileOffset, pos, PlayfieldTiling.TOP_LEFT);
		final PlayfieldPosition tpos = pos.add(moveVector);
		final Point targetOffset = new Point();
		map2canvas(targetOffset, tpos, PlayfieldTiling.TOP_LEFT);
		final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(topologyMode);
		final double mx = (targetOffset.getX() - tileOffset.getX()) / playfieldTiling.getSpaceWidth() * moveProgress;
		final double my = (targetOffset.getY() - tileOffset.getY()) / playfieldTiling.getSpaceHeight() * moveProgress;
		tile.paint(g, tileOffset, (float)mx, (float)my, imageKey, dyingProgress, zoomLevel, topologyMode);
	}

	/**
	 * Translates from map position to canvas position.
	 * @param position the map position to translate
	 * @param isCenter if {@link PlayfieldTiling#TOP_LEFT}, the canvas position of the upper left corner of the map tile is calculated; 
	 * 		if {@link PlayfieldTiling#CENTER}, the center of the map tile is calculated
	 * @return the resulting canvas position
	 * @see #map2canvas(Point, PlayfieldPosition, boolean)
	 */
	Point map2canvas(final PlayfieldPosition position, final boolean isCenter) {
		final Point ret = new Point();
		map2canvas(ret, position, isCenter);
		return ret;
	}

	/**
	 * Translates from map position to canvas position.
	 * @param point output parameter takes the resulting canvas position
	 * @param position the map position to translate
	 * @param isCenter if {@link PlayfieldTiling#TOP_LEFT}, the canvas position of the upper left corner of the map tile is calculated; 
	 * 		if {@link PlayfieldTiling#CENTER}, the center of the map tile is calculated
	 */
	void map2canvas(final Point point, final PlayfieldPosition position, final boolean isCenter) {
		parent.getGameFacadeProxy().getGameProperties().map2canvas(point, position, mapOffset, ZoomLevel.INSTANCES[parent.getZoomLevel()], isCenter);
	}

	/**
	 * Translates from canvas (e.g., mouse) position to on what map position this canvas position lies.
	 * @param cx the canvas x position
	 * @param cy the canvas y position
	 * @return The resulting map position. Might be out of the map boundaries. 
	 * 	<code>null</code> if the pixel position cannot be translated into a map position, e.g., because no game is running.
	 */
	private PlayfieldPosition canvas2map(final int cx, final int cy) {
		final PlayfieldPosition position;
		final GameFacadeProxy gameFacade = parent.getGameFacadeProxy();
		final GameProperties gameProperties = gameFacade.getGameProperties();
		if (null != gameProperties) {
			final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[parent.getZoomLevel()];
			position = gameProperties.canvas2map(cx, cy, mapOffset, zoomLevel);
		} else {
			position = null;
		}
		return position;
	}

	public void notifyMapSizeChanged() {
		final GameProperties props = parent.getGameFacadeProxy().getGameProperties();
		if (null == props) {
			getTitleImage();
			preferredSize.width = MINIMUM_SIZE.width;
			preferredSize.height = MINIMUM_SIZE.height;
		} else {
			final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[parent.getZoomLevel()];
			final Dimension d = props.getMapSizeInPixels(zoomLevel, false);
			preferredSize.width = d.width;
			preferredSize.height = d.height;
		}
		minimumSize.width = preferredSize.width;
		minimumSize.height = preferredSize.height;
		setPreferredSize(preferredSize);
		setMinimumSize(minimumSize);
		revalidate();
	}

	/**
	 * In lazy evaluation loads and stores the title image when needed.
	 * 
	 * @return the title image
	 */
	private BufferedImage getTitleImage() {
		if (null == titleImage) {
			final String resourceName = "Title.png";
			try {
				titleImage = TileTemplate.loadImage(resourceName);
			} catch (final IOException e) {
				logger.error("Unable to read image " + resourceName, e);
			}
		}
		return titleImage;
	}

	private Point getOffset(final GameProperties props) {
		final Rectangle2D.Float mapBoundary = props.getMapBoundary();
		final Dimension screenSize = getSize();
		final ZoomLevel zoomLevel = ZoomLevel.INSTANCES[parent.getZoomLevel()];
		final Point ret = new Point();
		if (null != mapBoundary) {
			final TopologyMode topologyMode = props.getTopologyMode();
			final PlayfieldTiling playfieldTiling = zoomLevel.getPlayfieldTiling(topologyMode);
			ret.x = calcOffsetDimension(mapBoundary.x, mapBoundary.width,  playfieldTiling.getSpaceWidth(), screenSize.width);
			ret.y = calcOffsetDimension(mapBoundary.y, mapBoundary.height, playfieldTiling.getSpaceHeight(), screenSize.height);
		}
		return ret;
	}

	private int calcOffsetDimension(final float mapOffset, final float mapSize, final int fieldSize, final int screenSize) {
		int ret = (screenSize - Math.round(mapSize * fieldSize)) / 2 + Math.round(mapOffset * fieldSize);
		if (ret < 0) {
			ret = 0;
		}
		return ret;
	}

	private final Logger logger = Logger.getLogger(getClass());

	private final EmpireMainWindow parent;

	private Dimension minimumSize = new Dimension(24 * 5, 24 * 5);

	private Dimension preferredSize = new Dimension(24 * 15, 24 * 10);

	private Point mapOffset = new Point();

	private BufferedImage titleImage = null;
}
